////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.ActionProcessor;
using Vela.BinarySerialization;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Hash;
using Vela.Crypto.Mac;
using Vela.Crypto.Symmetric;
using Vela.Kernel;
using Vela.Serialization;
using Vela.Storehouse;
using Vela.XmlSerialization;

namespace Vela.ApplicationBackend
{
	/// <summary>
	/// Algorithm pool.
	/// </summary>
	public class AlgorithmPool : IKernelAlgorithmPool, ISerializationAlgorithmPool, ICryptoAlgorithmPool, ICoreAlgorithmPool
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public AlgorithmPool()
		{
		}

		#endregion Constructors

		#region Private static fields

		/// <summary>
		/// The synchronization object for the <see cref="Instance"/> property. READ-ONLY.
		/// </summary>
		private static readonly object InstanceLock = new object();

		#endregion Private static fields

		#region Public static properties

		private static AlgorithmPool _Instance;
		/// <summary>
		/// The global algorithm pool. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// 
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public static AlgorithmPool Instance
		{
			get
			{
				lock( InstanceLock )
				{
					if( _Instance != null )
						return _Instance;

					_Instance = new AlgorithmPool();

					return _Instance;
				}
			}
		}

		#endregion Public static properties

		#region Public static methods

		/// <summary>
		/// Gets the name of the folder where custom emoticons are.
		/// </summary>
		/// <returns>Returns the name of the folder where custom emoticons are.</returns>
		public static string GetCustomEmoticonsFolderName()
		{
			return AssemblyUtils.GetApplicationDataFolderName( ApplicationBackendConsts.CustomEmoticonsFolderTitle );
		}

		/// <summary>
		/// Gets the default name of the root folder where files are saved.
		/// </summary>
		/// <returns>Returns the default name of the root folder where files are saved.</returns>
		public static string GetDefaultRootSaveFolderName()
		{
			return AssemblyUtils.GetApplicationDataFolderName( ApplicationBackendConsts.RootSaveFolderTitle );
		}

		/// <summary>
		/// Gets the default entropy algorithm.
		/// </summary>
		/// <returns>Returns the default entropy algorithm.</returns>
		public static EntropyAlg GetDefaultEntropyAlg()
		{
			return new Mash( new Utf8(), new Sha512(), new Rijndael256( SymmPaddings.None, SymmChainings.Cbc ) );
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Gets the text encoding.
		/// </summary>
		/// <returns>Returns the text encoding.</returns>
		public virtual TextEncoding GetTextEncoding()
		{
			return new Utf8();
		}

		/// <summary>
		/// Gets the communication frame algorithm.
		/// </summary>
		/// <returns>Returns the communication frame algorithm.</returns>
		public virtual ICommunicationFrameAlg GetCommunicationFrameAlg()
		{
			return new Cofa();
		}

		/// <summary>
		/// Gets the binary serializer.
		/// </summary>
		/// <returns>Returns the binary serializer.</returns>
		public virtual ISerializer GetBinarySerializer()
		{
			return new BinarySerializer( GetTextEncoding() );
		}

		/// <summary>
		/// Gets the XML serializer.
		/// </summary>
		/// <returns>Returns the XML serializer.</returns>
		public virtual ISerializer GetXmlSerializer()
		{
			return new XmlSerializer( GetTextEncoding() );
		}

		/// <summary>
		/// Gets the serializer.
		/// </summary>
		/// <returns>Returns the serializer.</returns>
		public virtual ISerializer GetSerializer()
		{
			return new BinarySerializer( GetTextEncoding() );
		}

		/// <summary>
		/// Gets the deserializer.
		/// </summary>
		/// <returns>Returns the deserializer.</returns>
		public virtual IDeserializer GetDeserializer()
		{
			return new BinaryDeserializer( GetTextEncoding() );
		}

		/// <summary>
		/// Gets the hash serializer.
		/// </summary>
		/// <param name="serializeValueTypeName">If this is true then the type names of the serialized values are stored in the
		/// serialization output. If this is true, the compatibility level of different versions of a serialized document is
		/// limited (see the "Serialization versioning" section from the documentation of the <see cref="IHashSerializer"/>
		/// interface). This must be true for a hash serializer which will be used to test the equality of documents.</param>
		/// <returns>Returns the hash serializer.</returns>
		public virtual IHashSerializer GetHashSerializer( bool serializeValueTypeName )
		{
			return new BinaryHashSerializer( GetTextEncoding(), serializeValueTypeName );
		}

		/// <summary>
		/// Gets the serializer for checking the equality of documents.
		/// </summary>
		/// <returns>Returns the serializer for checking the equality of documents.</returns>
		public virtual IHashSerializer GetHashSerializerForEquality()
		{
			return new BinaryHashSerializer( GetTextEncoding(), true );
		}

		/// <summary>
		/// Gets the serializer for export.
		/// </summary>
		/// <returns>Returns the serializer for export.</returns>
		public virtual ISerializer GetSerializerForExport()
		{
			return GetXmlSerializer();
		}

		/// <summary>
		/// Gets the entropy algorithm.
		/// </summary>
		/// <returns>Returns the entropy algorithm.</returns>
		public virtual EntropyAlg GetEntropyAlg()
		{
			return GetDefaultEntropyAlg();
		}

		/// <summary>
		/// Gets the hash algorithm.
		/// </summary>
		/// <returns>Returns the hash algorithm.</returns>
		public virtual HashAlg GetHashAlg()
		{
			return new Sha512();
		}

		/// <summary>
		/// Gets the MAC algorithm.
		/// </summary>
		/// <returns>Returns the MAC algorithm.</returns>
		public virtual MacAlg GetMacAlg()
		{
			return new Bck( new Sha512() );
		}

		/// <summary>
		/// Gets the symmetric encryption padding.
		/// </summary>
		/// <returns>Returns the symmetric encryption padding.</returns>
		public virtual SymmPaddings GetSymmPadding()
		{
			return SymmPaddings.Iso10126;
		}

		/// <summary>
		/// Gets the symmetric encryption chaining.
		/// </summary>
		/// <returns>Returns the symmetric encryption chaining.</returns>
		public virtual SymmChainings GetSymmChaining()
		{
			return SymmChainings.Cbc;
		}

		/// <summary>
		/// Gets the symmetric encryption algorithm.
		/// </summary>
		/// <returns>Returns the symmetric encryption algorithm.</returns>
		public virtual SymmAlg GetSymmAlg()
		{
			return new Rijndael256( GetSymmPadding(), GetSymmChaining() );
		}

		/// <summary>
		/// Gets the asymmetric encryption padding.
		/// </summary>
		/// <returns>Returns the asymmetric encryption padding.</returns>
		public IAsymmEncryptionPaddingAlg GetAsymmEncryptionPadding()
		{
			return new Vshe( GetTextEncoding(), GetHashAlg() );
		}

		/// <summary>
		/// Gets the asymmetric signing padding.
		/// </summary>
		/// <returns>Returns the asymmetric signing padding.</returns>
		public IAsymmSigningPaddingAlg GetAsymmSigningPadding()
		{
			return new Vshs( GetTextEncoding(), GetHashAlg() );
		}

		/// <summary>
		/// Gets the algorithm used for asymmetric encryption.
		/// </summary>
		/// <param name="algorithmSecurityLevel">The security level for the cryptographic algorithms to use.</param>
		/// <returns>Returns the algorithm used for asymmetric encryption.</returns>
		/// <remarks>
		/// <para>You must later call the <see cref="AsymmAlg.GenerateKeyPair"/> method for the returned algorithm.</para>
		/// </remarks>
		public virtual AsymmAlg GetAsymmAlg( AlgorithmSecurityLevel algorithmSecurityLevel )
		{
			switch( algorithmSecurityLevel )
			{
				case AlgorithmSecurityLevel.Normal:
					return new Rsa2048();
				case AlgorithmSecurityLevel.Strong:
					return new Rsa4096();
				case AlgorithmSecurityLevel.Strongest:
					return new Rsa8192();
				default:
					throw ValidationUtils.Unhandled( algorithmSecurityLevel, "algorithmSecurityLevel" );
			}
		}

		/// <summary>
		/// Generates the algorithm used for asymmetric encryption.
		/// </summary>
		/// <param name="algorithmSecurityLevel">The security level for the cryptographic algorithms to use.</param>
		/// <returns>Returns the generated algorithm used for asymmetric encryption.</returns>
		/// <param name="lop">Long operation progress used during the generation of a key pair. This may be null.</param>
		public virtual AsymmAlg GenerateAsymmAlg( AlgorithmSecurityLevel algorithmSecurityLevel, LongOperationProgress lop )
		{
			AsymmAlg asymmAlg = GetAsymmAlg( algorithmSecurityLevel );
			asymmAlg.GenerateKeyPair( lop );

			return asymmAlg;
		}

		/// <summary>
		/// Gets the compression algorithm.
		/// </summary>
		/// <returns>Returns the compression algorithm. This may be null.</returns>
		public virtual CompressionAlg GetCompressionAlg()
		{
			return new Gzip();
		}

		/// <summary>
		/// Gets the password blender.
		/// </summary>
		/// <returns>Returns the password blender.</returns>
		public virtual PasswordBlenderAlg GetPasswordBlender()
		{
			// The "Sha512" algorithm is specifically used because its 64-bit registry operation slows down dedicated rainbow
			// attacks.
			return new Hpb( GetTextEncoding(), new Bck( new Sha512() ), 14, 1000000, 4 );
		}

		/// <summary>
		/// Gets the symmetrically encrypted envelope.
		/// </summary>
		/// <returns>Returns the symmetrically encrypted envelope.</returns>
		public SymmEncryptedEnvelope GetSymmEncryptedEnvelope()
		{
			return new SymmEncryptedEnvelope( CryptoUtils.GetLocator(), GetMacAlg(), GetCompressionAlg(), GetSymmAlg()
				, GetPasswordBlender() );
		}

		/// <summary>
		/// Gets the envelope symmetrically encrypted with a UPN.
		/// </summary>
		/// <param name="recipientUpnFingerprint">The UPN fingerprint of the recipient of the encrypted envelope.</param>
		/// <returns>Returns the envelope symmetrically encrypted with a UPN.</returns>
		public UpnEncryptedEnvelope GetUpnEncryptedEnvelope( Locator recipientUpnFingerprint )
		{
			return new UpnEncryptedEnvelope( CryptoUtils.GetLocator(), GetMacAlg(), GetCompressionAlg(), GetSymmAlg()
				, GetPasswordBlender(), recipientUpnFingerprint );
		}

		/// <summary>
		/// Gets the asymmetrically encrypted envelope.
		/// </summary>
		/// <returns>Returns the asymmetrically encrypted envelope.</returns>
		public AsymmEncryptedEnvelope GetAsymmEncryptedEnvelope()
		{
			return new AsymmEncryptedEnvelope( CryptoUtils.GetLocator(), GetMacAlg(), GetCompressionAlg(), GetSymmAlg()
				, GetAsymmEncryptionPadding() );
		}

		/// <summary>
		/// Gets the algorithm used to generate UPNs.
		/// </summary>
		/// <param name="algorithmSecurityLevel">The security level for the returned algorithm.</param>
		/// <returns>Returns the algorithm used to generate UPNs.</returns>
		public virtual UpnAlg GetUpnAlg( AlgorithmSecurityLevel algorithmSecurityLevel )
		{
			switch( algorithmSecurityLevel )
			{
				case AlgorithmSecurityLevel.Normal:
					return new Pream16();
				case AlgorithmSecurityLevel.Strong:
					return new Pream32();
				case AlgorithmSecurityLevel.Strongest:
					return new Pream0();
				default:
					throw ValidationUtils.Unhandled( algorithmSecurityLevel, "algorithmSecurityLevel" );
			}
		}

		/// <summary>
		/// Gets the size (in bytes) to which the encrypted data of an exported document's (encrypted) envelope is aligned. If
		/// this is bigger than <see cref="EncryptedEnvelope.BlockSizeSize"/>, the real size of the encrypted data will be
		/// hidden.
		/// </summary>
		/// <returns>Returns the size (in bytes) to which the encrypted data of an exported document's (encrypted) envelope is
		/// aligned.</returns>
		public virtual int GetExportedDocumentEnvelopeEncryptedDocumentBlockSize()
		{
			return 1024;
		}

		/// <summary>
		/// Gets the upper bound of the disalignment size used by steganographic algorithms. The disalignment size may be between
		/// 0 and this value (excluding it). If this is 0 or a negative number, it's ignored.
		/// </summary>
		public virtual int GetExportedDocumentSteganographicDisalignmentSizeZenith()
		{
			return 1 << 16;
		}

		/// <summary>
		/// Gets the size (in bytes) to which the encrypted data of an action request (encrypted) envelope is aligned. If this is
		/// bigger than <see cref="EncryptedEnvelope.BlockSizeSize"/>, the real size of the encrypted data will be hidden.
		/// </summary>
		/// <returns>Returns the size (in bytes) to which the encrypted data of an action request (encrypted) envelope is
		/// aligned.</returns>
		public virtual int GetActionRequestEnvelopeEncryptedDocumentBlockSize()
		{
			return 1024;
		}

		/// <summary>
		/// Gets the size (in bytes) to which the (encrypted) data of a secured document is aligned. If this is bigger than 1,
		/// the real size of the data will be hidden.
		/// </summary>
		/// <returns>Returns the size (in bytes) to which the (encrypted) data of a secured document is aligned.</returns>
		public virtual int GetSecuredDocumentBlockSize()
		{
			return 1024;
		}

		/// <summary>
		/// Gets the storehouse header.
		/// </summary>
		/// <returns>Returns the storehouse header.</returns>
		public virtual StorehouseHeader GetStorehouseHeader()
		{
			return new StorehouseHeader( CryptoUtils.GetLocator(), GetTextEncoding(), GetMacAlg(), GetCompressionAlg()
				, GetSymmAlg(), GetPasswordBlender() );
		}

		/// <summary>
		/// Gets the action processor configuration.
		/// </summary>
		/// <returns>Returns the action client configuration.</returns>
		public ActionProcessorConfiguration GetActionProcessorConfiguration()
		{
			return new ActionProcessorConfiguration( true, true, GetDefaultRootSaveFolderName()
				, GetActionClientConfiguration(), GetActionServiceConfiguration() );
		}

		/// <summary>
		/// Gets the action client configuration.
		/// </summary>
		/// <returns>Returns the action client configuration.</returns>
		/// <remarks>
		/// <para>An action client retries to send an action request until it expires (which is normally
		/// <see cref="Vela.PeerDocuments.PeerDocumentsConsts.ActionRequestValidity"/> minutes after its creation).</para>
		/// </remarks>
		public virtual ActionClientConfiguration GetActionClientConfiguration()
		{
			ActionClientRetryScheduleAlg actionClientRetryScheduleAlg = new ActionClientRetryScheduleAlg();

			RecipientAccessStatesManagerConfiguration rasManager = new RecipientAccessStatesManagerConfiguration(
				actionClientRetryScheduleAlg );

			return new ActionClientConfiguration( rasManager );
		}

		/// <summary>
		/// Gets the action service configuration.
		/// </summary>
		/// <returns>Returns the action service configuration.</returns>
		public virtual ActionServiceConfiguration GetActionServiceConfiguration()
		{
			return new ActionServiceConfiguration();
		}

		/// <summary>
		/// Gets the action service listener configuration.
		/// </summary>
		/// <returns>Returns the action service listener configuration.</returns>
		public virtual ActionServiceListenerConfiguration GetActionServiceListenerConfiguration()
		{
			return new ActionServiceListenerConfiguration( GetCommunicationFrameAlg()
				, ActionProcessorUtils.GetRandomServiceProcessingPort() );
		}

		#endregion Public methods
	}
}
