////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Text;
using Vela.ActionProcessor;
using Vela.ApplicationDocuments;
using Vela.Controls;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Crypto.Symmetric;
using Vela.DatabaseStorehouse;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.StarkDocuments;
using Vela.Storehouse;

namespace Vela.ApplicationBackend
{
	/// <summary>
	/// Storehouse utilities for an application backend.
	/// </summary>
	public static class ApplicationBackendStorehouse
	{
		#region Public constants

		/// <summary>
		/// The default administration connection string.
		/// </summary>
		public const string AdminConnectionStringDef = "Server=(local); database=master; Integrated security=SSPI;"
			+ " Enlist=False;";

		/// <summary>
		/// The number of posts whose flags to update in a batch.
		/// </summary>
		public const int UpdatePostsFlagBatchSize = 50;

		/// <summary>
		/// The number of letters to delete in a batch.
		/// </summary>
		public const int DeleteLettersBatchSize = 20;

		/// <summary>
		/// The number of letters to get as text in a batch.
		/// </summary>
		public const int GetLettersAsTextFlagBatchSize = 100;

		/// <summary>
		/// The number of letters to import in a batch.
		/// </summary>
		public const int ImportLettersBatchSize = 20;

		/// <summary>
		/// The template used to format the letters to get as text. Placeholders: 0 - counter, 1 - last (processing state) time,
		/// 2 - sender's name, 3 - recipient's name, 4 - account alias, 5 - message.
		/// </summary>
		public const string GetLettersAsTextTemplate
			= "#{0}, {1}: {2} -> {3} ({4})"
			+ "\n{5}"
			+ "\n\n";

		/// <summary>
		/// Template for the file name of an exported letter. Placeholders: 0 - export time, 1 - instance locator.
		/// </summary>
		public const string ExportedLetterFileNameTemplate = "{0}-{1}";

		#endregion Public constants

		#region Private static methods

		/// <summary>
		/// Checks if there is an account for the remote peer with the specified UPN.
		/// </summary>
		/// <param name="transaction">Transaction to use.</param>
		/// <param name="remoteUpnFingerprint">UPN of the remote peer for which to check if an account exists.</param>
		/// <returns>Returns true if there is an account for the remote peer with the specified UPN.</returns>
		private static bool AccountExistsForRemotePeer( IStorehouseTransaction transaction, Locator remoteUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( remoteUpnFingerprint, "remoteUpnFingerprint" );

			SpeckleFilter remoteUpnFingerprintFilter = new SpeckleFilter( Account.RemoteUpnFingerprintFilterIdentifier
				, remoteUpnFingerprint );

			Speckle accountSpeckle = transaction.CollectFirstSpeckle( PortableTypes.FindAllDerived( typeof( Account ) )
				, new SpeckleFilters( remoteUpnFingerprintFilter ), StorehouseRealms.Accounts, SheafLocks.Shared, null, null );

			return accountSpeckle != null;
		}

		/// <summary>
		/// Deletes all the sheafs associated to the vault with the specified locator.
		/// </summary>
		/// <param name="transaction">Transaction to use.</param>
		/// <param name="vaultLocator">The locator of the vault whose referred sheafs to delete.</param>
		/// <param name="serviceManager">Action service manager which is used to stop processing action requests which have a
		/// multiframe action response.</param>
		/// <param name="clientManager">Action client manager which is used to stop processing action requests which have a
		/// multiframe action response.</param>
		/// <param name="lop">Long operation progress. This may be null.</param>
		private static void DeleteVaultSheafs( IStorehouseTransaction transaction, Locator vaultLocator
			, IActionServiceManager serviceManager, IActionClientManager clientManager, LongOperationProgress lop )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( vaultLocator, "vaultLocator" );
			ValidationUtils.ObjectHasContent( serviceManager, "serviceManager" );
			ValidationUtils.ObjectHasContent( clientManager, "clientManager" );

			if( lop != null )
				lop.Progress();

			// Load the vault speckle.
			Speckle vaultSpeckle = transaction.LoadSpeckle( vaultLocator, StorehouseRealms.ProcessingIdentityContainers
				, SheafLocks.Exclusive, null, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Vault ) ) );
			if( vaultSpeckle == null )
				return;

			Locator vaultUpnFingerprint = vaultSpeckle.GetLoadedSpeckleFilter<Locator>(
				Vault.LocalUpnFingerprintFilterIdentifier );

			// Delete the accounts.
			SpeckleFilter localVaultUpnFilter = new SpeckleFilter( Account.LocalUpnFingerprintFilterIdentifier
				, vaultUpnFingerprint );
			Locators accountKeys = transaction.CollectKeys( PortableTypes.FindAllDerived( typeof( Account ) )
				, new SpeckleFilters( localVaultUpnFilter ), StorehouseRealms.Accounts, null, null, null ) as Locators;
			foreach( Locator accountKey in accountKeys )
				DeleteAccountSheafs( transaction, accountKey, serviceManager, clientManager, lop );

			if( lop != null )
				lop.Progress();

			// Delete the active action requests; there can still be some action requests which are not yet associated to
			// accounts.
			SpeckleFilter activeRequestInitiatorLocatorFilter = new SpeckleFilter(
				ActionRequest.InitiatorUpnFingerprintFilterIdentifier, vaultUpnFingerprint );
			transaction.DeleteSheafs( PortableTypes.FindAllDerived( typeof( ActionRequest ) )
				, new SpeckleFilters( activeRequestInitiatorLocatorFilter ), StorehouseRealms.ActionRequestsToSend );

			if( lop != null )
				lop.Progress();

			// Delete the local action progresses; these could be associated to the above action requests.
			SpeckleFilter localProgressInitiatorLocatorFilter = new SpeckleFilter(
				ActionProgress.InitiatorUpnFingerprintFilterIdentifier, vaultUpnFingerprint );
			transaction.DeleteSheafs( PortableTypes.FindAllDerived( typeof( ActionProgress ) )
				, new SpeckleFilters( localProgressInitiatorLocatorFilter ), StorehouseRealms.ActionProgresses );

			if( lop != null )
				lop.Progress();

			// Delete the (remaining) posts; the letters were deleted with the accounts.
			SpeckleFilter postInitiatorLocatorFilter = new SpeckleFilter( Post.InitiatorUpnFingerprintFilterIdentifier
				, vaultUpnFingerprint );
			transaction.DeleteSheafs( PortableTypes.FindAllDerived( typeof( Post ) )
				, new SpeckleFilters( postInitiatorLocatorFilter ), StorehouseRealms.Posts );

			if( lop != null )
				lop.Progress();

			// Delete the photo.
			SpeckleFilter photoCreatorUpnFingerprintFilter = new SpeckleFilter( Photo.CreatorUpnFingerprintFilterIdentifier
				, vaultUpnFingerprint );
			transaction.DeleteFirstSheaf( PortableTypes.FindAllDerived( typeof( Photo ) )
				, new SpeckleFilters( photoCreatorUpnFingerprintFilter ), StorehouseRealms.Photos );

			// Delete the vault.
			transaction.DeleteSheaf( vaultLocator, StorehouseRealms.ProcessingIdentityContainers );

			// Delete the UPN fingerprint reservation.
			transaction.DeleteSheaf( vaultUpnFingerprint, StorehouseRealms.UpnFingerprintReservations );

			if( lop != null )
				lop.Progress();
		}

		/// <summary>
		/// Deletes all the sheafs associated to the account with the specified locator.
		/// </summary>
		/// <param name="transaction">Transaction to use.</param>
		/// <param name="accountLocator">The locator of the account whose referred sheafs to delete.</param>
		/// <param name="serviceManager">Action service manager which is used to stop processing action requests which have a
		/// multiframe action response.</param>
		/// <param name="clientManager">Action client manager which is used to stop processing action requests which have a
		/// multiframe action response.</param>
		/// <param name="lop">Long operation progress. This may be null.</param>
		private static void DeleteAccountSheafs( IStorehouseTransaction transaction, Locator accountLocator
			, IActionServiceManager serviceManager, IActionClientManager clientManager, LongOperationProgress lop )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( accountLocator, "accountLocator" );
			ValidationUtils.ObjectHasContent( serviceManager, "serviceManager" );
			ValidationUtils.ObjectHasContent( clientManager, "clientManager" );

			if( lop != null )
				lop.Progress();

			// Load the account speckle.
			Speckle accountSpeckle = transaction.LoadSpeckle( accountLocator, StorehouseRealms.Accounts, SheafLocks.Exclusive
				, null, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Account ) ) );
			if( accountSpeckle == null )
				return;

			Locator localUpnFingerprint = accountSpeckle.GetLoadedSpeckleFilter<Locator>(
				Account.LocalUpnFingerprintFilterIdentifier );
			Locator remoteUpnFingerprint = accountSpeckle.GetLoadedSpeckleFilter<Locator>(
				Account.RemoteUpnFingerprintFilterIdentifier );

			// If the remote UPN is a local one, don't delete the account (because the vault would also be deleted).
			SpeckleFilter remoteVaultUpnFilter = new SpeckleFilter( Vault.LocalUpnFingerprintFilterIdentifier
				, remoteUpnFingerprint );
			Speckle remoteVaultSpeckle = transaction.CollectFirstSpeckle( PortableTypes.FindAllDerived( typeof( Vault ) ),
				new SpeckleFilters( remoteVaultUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared
				, null, new SpeckleFilterIdentifiers( Vault.EntityNameFilterIdentifier ) );
			if( remoteVaultSpeckle != null )
			{
				throw new ValidationException( "Can't delete the account because there is a vault with the same UPN fingerprint,"
						+ " which means that the account's remote peer is actually the local peer.", null, null
					, new ExceptionDetail( "Account locator", accountLocator ) );
			}

			// Delete the received action requests.
			SpeckleFilter receivedRequestInitiatorLocatorFilter = new SpeckleFilter(
				ActionRequest.InitiatorUpnFingerprintFilterIdentifier, remoteUpnFingerprint );
			SpeckleFilter receivedRequestRecipientLocatorFilter = new SpeckleFilter(
				ActionRequest.RecipientUpnFingerprintFilterIdentifier, localUpnFingerprint );
			transaction.DeleteSheafs( PortableTypes.FindAllDerived( typeof( ActionRequest ) )
				, new SpeckleFilters( receivedRequestInitiatorLocatorFilter, receivedRequestRecipientLocatorFilter )
				, StorehouseRealms.ServiceReceivedActionRequests );

			if( lop != null )
				lop.Progress();

			// Delete the sent action responses.
			SpeckleFilter sentResponseInitiatorLocatorFilter = new SpeckleFilter(
				ActionResponse.InitiatorUpnFingerprintFilterIdentifier, localUpnFingerprint );
			SpeckleFilter sentResponseRecipientLocatorFilter = new SpeckleFilter(
				ActionResponse.RecipientUpnFingerprintFilterIdentifier, remoteUpnFingerprint );
			transaction.DeleteSheafs( PortableTypes.FindAllDerived( typeof( ActionRequest ) )
				, new SpeckleFilters( sentResponseInitiatorLocatorFilter, sentResponseRecipientLocatorFilter )
				, StorehouseRealms.ServiceSentActionResponses );

			if( lop != null )
				lop.Progress();

			#region Delete the active action requests

			SpeckleFilter activeRequestInitiatorLocatorFilter = new SpeckleFilter(
				ActionRequest.InitiatorUpnFingerprintFilterIdentifier, localUpnFingerprint );
			SpeckleFilter activeRequestRecipientLocatorFilter = new SpeckleFilter(
				ActionRequest.RecipientUpnFingerprintFilterIdentifier, remoteUpnFingerprint );

			Locators activeActionRequestKeys = transaction.CollectKeys( PortableTypes.FindAllDerived( typeof( ActionRequest ) )
				, new SpeckleFilters( activeRequestInitiatorLocatorFilter, activeRequestRecipientLocatorFilter )
				, StorehouseRealms.ActionRequestsToSend, null, null, null ) as Locators;

			// Stop the action processing in the action service and client.
			serviceManager.StopRequests( activeActionRequestKeys );
			clientManager.StopRequests( activeActionRequestKeys );

			// Delete the active action requests.
			transaction.DeleteSheafs( PortableTypes.FindAllDerived( typeof( ActionRequest ) )
				, new SpeckleFilters( activeRequestInitiatorLocatorFilter, activeRequestRecipientLocatorFilter )
				, StorehouseRealms.ActionRequestsToSend );

			#endregion Delete the active action requests

			if( lop != null )
				lop.Progress();

			// Delete the local action progresses.
			SpeckleFilter localProgressInitiatorLocatorFilter = new SpeckleFilter(
				ActionProgress.InitiatorUpnFingerprintFilterIdentifier, localUpnFingerprint );
			SpeckleFilter localProgressRecipientLocatorFilter = new SpeckleFilter(
				ActionProgress.RecipientUpnFingerprintFilterIdentifier, remoteUpnFingerprint );
			transaction.DeleteSheafs( PortableTypes.FindAllDerived( typeof( ActionProgress ) )
				, new SpeckleFilters( localProgressInitiatorLocatorFilter, localProgressRecipientLocatorFilter )
				, StorehouseRealms.ActionProgresses );

			if( lop != null )
				lop.Progress();

			// Delete the remote action progresses.
			SpeckleFilter remoteProgressInitiatorLocatorFilter = new SpeckleFilter(
				ActionProgress.InitiatorUpnFingerprintFilterIdentifier, remoteUpnFingerprint );
			SpeckleFilter remoteProgressRecipientLocatorFilter = new SpeckleFilter(
				ActionProgress.RecipientUpnFingerprintFilterIdentifier, localUpnFingerprint );
			transaction.DeleteSheafs( PortableTypes.FindAllDerived( typeof( ActionProgress ) )
				, new SpeckleFilters( remoteProgressInitiatorLocatorFilter, remoteProgressRecipientLocatorFilter )
				, StorehouseRealms.ActionProgresses );

			if( lop != null )
				lop.Progress();

			// Delete the local active sessions.
			SpeckleFilter localSessionInitiatorLocatorFilter = new SpeckleFilter(
				InteractionSession.InitiatorUpnFingerprintFilterIdentifier, localUpnFingerprint );
			SpeckleFilter localSessionRecipientLocatorFilter = new SpeckleFilter(
				InteractionSession.RecipientUpnFingerprintFilterIdentifier, remoteUpnFingerprint );
			transaction.DeleteSheafs( PortableTypes.FindAllDerived( typeof( InteractionSession ) )
				, new SpeckleFilters( localSessionInitiatorLocatorFilter, localSessionRecipientLocatorFilter )
				, StorehouseRealms.InteractionSessions );

			if( lop != null )
				lop.Progress();

			// Delete the remote active sessions.
			SpeckleFilter remoteSessionInitiatorLocatorFilter = new SpeckleFilter(
				InteractionSession.InitiatorUpnFingerprintFilterIdentifier, remoteUpnFingerprint );
			SpeckleFilter remoteSessionRecipientLocatorFilter = new SpeckleFilter(
				InteractionSession.RecipientUpnFingerprintFilterIdentifier, localUpnFingerprint );
			transaction.DeleteSheafs( PortableTypes.FindAllDerived( typeof( InteractionSession ) )
				, new SpeckleFilters( remoteSessionInitiatorLocatorFilter, remoteSessionRecipientLocatorFilter )
				, StorehouseRealms.InteractionSessions );

			if( lop != null )
				lop.Progress();

			// Delete the letters sent by the local peer to the remote peer.
			SpeckleFilter letterInitiatorLocatorFilter = new SpeckleFilter( Post.InitiatorUpnFingerprintFilterIdentifier
				, localUpnFingerprint );
			SpeckleFilter letterRecipientLocatorFilter = new SpeckleFilter( Letter.RecipientUpnFingerprintFilterIdentifier
				, remoteUpnFingerprint );
			transaction.DeleteSheafs( PortableTypes.FindAllDerived( typeof( Letter ) )
				, new SpeckleFilters( letterInitiatorLocatorFilter, letterRecipientLocatorFilter ), StorehouseRealms.Posts );

			if( lop != null )
				lop.Progress();

			// Delete the letters received by the local peer from the remote peer.
			letterInitiatorLocatorFilter = new SpeckleFilter( Post.InitiatorUpnFingerprintFilterIdentifier
				, remoteUpnFingerprint );
			letterRecipientLocatorFilter = new SpeckleFilter( Letter.RecipientUpnFingerprintFilterIdentifier
				, localUpnFingerprint );
			transaction.DeleteSheafs( PortableTypes.FindAllDerived( typeof( Letter ) )
				, new SpeckleFilters( letterInitiatorLocatorFilter, letterRecipientLocatorFilter ), StorehouseRealms.Posts );

			if( lop != null )
				lop.Progress();

			// Delete the account.
			transaction.DeleteSheaf( accountLocator, StorehouseRealms.Accounts );

			// If there is another account for the same remote peer, don't delete the information of the remote peer.
			if( AccountExistsForRemotePeer( transaction, remoteUpnFingerprint ) )
			{
				if( lop != null )
					lop.Progress();

				return;
			}

			// Delete the ephemerid.
			SpeckleFilter ephemeridRemoteUpnFingerprintFilter = new SpeckleFilter( Ephemerid.RemoteUpnFingerprintFilterIdentifier
				, remoteUpnFingerprint );
			transaction.DeleteFirstSheaf( PortableTypes.FindAllDerived( typeof( Ephemerid ) )
				, new SpeckleFilters( ephemeridRemoteUpnFingerprintFilter ), StorehouseRealms.Ephemerids );

			// Delete the tally.
			SpeckleFilter tallyRemoteUpnFingerprintFilter = new SpeckleFilter( Tally.RemoteUpnFingerprintFilterIdentifier
				, remoteUpnFingerprint );
			transaction.DeleteFirstSheaf( PortableTypes.FindAllDerived( typeof( Tally ) )
				, new SpeckleFilters( tallyRemoteUpnFingerprintFilter ), StorehouseRealms.ProcessingIdentityContainers );

			// Delete the photo.
			SpeckleFilter photoCreatorUpnFingerprintFilter = new SpeckleFilter( Photo.CreatorUpnFingerprintFilterIdentifier
				, remoteUpnFingerprint );
			transaction.DeleteFirstSheaf( PortableTypes.FindAllDerived( typeof( Photo ) )
				, new SpeckleFilters( photoCreatorUpnFingerprintFilter ), StorehouseRealms.Photos );

			// Delete the presenter.
			SpeckleFilter presenterCreatorUpnFingerprintFilter = new SpeckleFilter(
				Presenter.CreatorUpnFingerprintFilterIdentifier, remoteUpnFingerprint );
			transaction.DeleteFirstSheaf( PortableTypes.FindAllDerived( typeof( Presenter ) )
				, new SpeckleFilters( presenterCreatorUpnFingerprintFilter ), StorehouseRealms.ProcessingIdentityContainers );

			// Delete the UPN fingerprint reservation.
			transaction.DeleteSheaf( remoteUpnFingerprint, StorehouseRealms.UpnFingerprintReservations );

			if( lop != null )
				lop.Progress();
		}

		/// <summary>
		/// Deletes all the sheafs associated to the letter with the specified locator.
		/// </summary>
		/// <param name="transaction">Transaction to use.</param>
		/// <param name="letterLocator">The locator of the letter whose referred sheafs to delete.</param>
		/// <param name="serviceManager">Action service manager which is used to stop processing action requests which have a
		/// multiframe action response. This may be null.</param>
		/// <param name="clientManager">Action client manager which is used to stop processing action requests which have a
		/// multiframe action response. This may be null.</param>
		private static void DeleteLetterSheafs( IStorehouseTransaction transaction, Locator letterLocator
			, IActionServiceManager serviceManager, IActionClientManager clientManager )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( letterLocator, "letterLocator" );

			#region Delete the file download request

			// This must be done first because otherwise the specific files download progress would be deleted by the generic
			// delete performed in the next step.

			SpeckleFilter accessGranterLocatorFilter = new SpeckleFilter(
				DownloadFilesActionRequest.AccessGranterLocatorFilterIdentifier, letterLocator );
			Speckle downloadRequestSpeckle = transaction.CollectFirstSpeckle(
				PortableTypes.FindAllDerived( typeof( DownloadFilesActionRequest ) )
				, new SpeckleFilters( accessGranterLocatorFilter ), StorehouseRealms.ActionRequestsToSend
				, SheafLocks.Exclusive, null, null );

			if( downloadRequestSpeckle != null )
			{
				// Stop the files download in the action service and client.
				if( serviceManager != null )
					serviceManager.StopRequests( new Locators( downloadRequestSpeckle.StorehouseKey ) );
				if( clientManager != null )
					clientManager.StopRequests( new Locators( downloadRequestSpeckle.StorehouseKey ) );

				// Delete the "DownloadFilesActionRequest" action request.
				transaction.DeleteSheaf( downloadRequestSpeckle.StorehouseKey, StorehouseRealms.ActionRequestsToSend );
			}

			// Delete the service-side and client-side files download progresses.
			SpeckleFilter referredLocatorFilter = new SpeckleFilter( ActionProgress.ReferredSheafLocatorFilterIdentifier
				, letterLocator );
			transaction.DeleteFirstSheaf( PortableTypes.FindAllDerived( typeof( DownloadFilesActionServiceProgress ) )
				, new SpeckleFilters( referredLocatorFilter ), StorehouseRealms.ActionProgresses );
			transaction.DeleteFirstSheaf( PortableTypes.FindAllDerived( typeof( DownloadFilesActionClientProgress ) )
				, new SpeckleFilters( referredLocatorFilter ), StorehouseRealms.ActionProgresses );

			#endregion Delete the file download request

			#region Delete the letter itself

			// Get the instance locator of the "SendLetterActionRequest" action request in order to stop its processing.
			SpeckleFilter transportedLetterLocatorFilter = new SpeckleFilter(
				ActionRequest.TransportedSheafLocatorFilterIdentifier, letterLocator );
			Speckle letterActionSpeckle = transaction.CollectFirstSpeckle( PortableTypes.FindAllDerived(
				typeof( SendLetterActionRequest ) ), new SpeckleFilters( transportedLetterLocatorFilter )
				, StorehouseRealms.ActionRequestsToSend, SheafLocks.Shared, null, null );

			// Stop the letter processing in the action service and client.
			if( letterActionSpeckle != null )
			{
				if( serviceManager != null )
					serviceManager.StopRequests( new Locators( letterActionSpeckle.StorehouseKey ) );
				if( clientManager != null )
					clientManager.StopRequests( new Locators( letterActionSpeckle.StorehouseKey ) );
			}

			// Delete the "SendLetterActionRequest" action request.
			transaction.DeleteFirstSheaf( PortableTypes.FindAllDerived( typeof( SendLetterActionRequest ) )
				, new SpeckleFilters( transportedLetterLocatorFilter ), StorehouseRealms.ActionRequestsToSend );

			// Delete the progress.
			transaction.DeleteFirstSheaf( PortableTypes.FindAllDerived( typeof( ActionProgress ) )
				, new SpeckleFilters( referredLocatorFilter ), StorehouseRealms.ActionProgresses );

			// Delete the letter.
			transaction.DeleteSheaf( letterLocator, StorehouseRealms.Posts );

			#endregion Delete the letter itself
		}

		/// <summary>
		/// Sets the information of the specified letter speckle.
		/// </summary>
		/// <param name="letterSpeckle">Letter speckle whose information to set.</param>
		/// <param name="initiatorEntityNameFilterIdentifier">The filter identifier for the entity name of the initiator of the
		/// letter.</param>
		/// <param name="recipientEntityNameFilterIdentifier">The filter identifier for the entity name of the recipient of the
		/// letter.</param>
		/// <param name="initiatorSpeckle">The initiator speckle (vault or tally).</param>
		/// <param name="recipientSpeckle">The recipient speckle (vault or tally).</param>
		/// <param name="emoticonManager">Emoticon manager which is used to set the emoticon descriptors of the specified letter
		/// speckle, with those extracted from its message summary. If this is null then the emoticon descriptors are set with
		/// null.</param>
		private static void SetLetterSpeckleInformation( LetterSpeckle letterSpeckle, string initiatorEntityNameFilterIdentifier
			, string recipientEntityNameFilterIdentifier, Speckle initiatorSpeckle, Speckle recipientSpeckle
			, EmoticonManager emoticonManager )
		{
			ValidationUtils.TextHasContent( initiatorEntityNameFilterIdentifier, "initiatorEntityNameFilterIdentifier" );
			ValidationUtils.TextHasContent( recipientEntityNameFilterIdentifier, "recipientEntityNameFilterIdentifier" );
			ValidationUtils.ObjectHasContent( initiatorSpeckle, "initiatorSpeckle" );
			ValidationUtils.ObjectHasContent( recipientSpeckle, "recipientSpeckle" );

			letterSpeckle.InitiatorEntityName = initiatorSpeckle.GetLoadedSpeckleFilter<string>(
				initiatorEntityNameFilterIdentifier );
			letterSpeckle.RecipientEntityName = recipientSpeckle.GetLoadedSpeckleFilter<string>(
				recipientEntityNameFilterIdentifier );

			if( emoticonManager != null )
			{
				letterSpeckle.EmoticonDescriptors = emoticonManager.GetEmoticonDescriptors(
					letterSpeckle.MessageSummary, ApplicationBackendConsts.LetterEmoticonSymbolsCountMax );
			}
			else
			{
				letterSpeckle.EmoticonDescriptors = null;
			}
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Creates a storehouse.
		/// </summary>
		/// <param name="algorithmPool">Algorithm pool.</param>
		/// <param name="storehouseAccess">The storehouse access.</param>
		/// <param name="storehouseConfiguration">The storehouse configuration.</param>
		/// <param name="storehouseFileName">The (initial) storehouse file name. This may be null or empty.</param>
		/// <param name="currentStorehouseFileName">The current storehouse file name. This is used to not allow the user to
		/// overwrite the current storehouse. This may be null or empty.</param>
		/// <returns>Returns the storehouse provider of the created storehouse, or null if none was created.</returns>
		public static IStorehouseProvider CreateStorehouse( AlgorithmPool algorithmPool, IStorehouseAccess storehouseAccess
			, StorehouseConfiguration storehouseConfiguration, string storehouseFileName, string currentStorehouseFileName )
		{
			ValidationUtils.ObjectHasContent( algorithmPool, "algorithmPool" );
			ValidationUtils.ObjectHasContent( storehouseAccess, "storehouseAccess" );
			ValidationUtils.ObjectHasContent( storehouseConfiguration, "storehouseConfiguration" );

			#region Get the file name

			string windowTitle = "Create Storehouse";

			string selectedFileName;
			if( !ControlsUtils.ShowSaveFileDialog( windowTitle, FileExtensions.Storehouse, FileExtensions.StorehouseFilter
				, FileUtils.GetExistingFolderName( storehouseFileName ), null, true, out selectedFileName ) )
			{
				return null;
			}

			storehouseFileName = selectedFileName;

			// Don't allow the user to overwrite the current storehouse.
			if( FileUtils.EqualNames( storehouseFileName, currentStorehouseFileName ) )
			{
				throw new ValidationException( "The storehouse to create may not overwrite the current storehouse.", null, null
					, new ExceptionDetail( "File name", currentStorehouseFileName ) );
			}

			#endregion Get the file name

			#region Add real entropy

			// Add new (global) entropy (from the user). This must be done before anything random used by the storehouse is
			// generated.

			EntropyWindow entropyWindow = new EntropyWindow();
			if( !ControlsUtils.ShowDialog( entropyWindow ) )
				return null;

			// Merge and replace the global entropy algorithm with the one from the storehouse.
			EntropyAlg entropyAlg = algorithmPool.GetEntropyAlg();
			EntropyPool.MergeAndReplace( entropyAlg );

			#endregion Add real entropy

			#region Get the user password

			PasswordWindow passwordWindow = new PasswordWindow( windowTitle, storehouseFileName, null, true );
			if( !ControlsUtils.ShowDialog( passwordWindow ) )
				return null;

			string storehousePassword = passwordWindow.Password;

			#endregion Get the user password

			StorehouseHeader storehouseHeader = algorithmPool.GetStorehouseHeader();
			StorehouseEntropy storehouseEntropy = new StorehouseEntropy( entropyAlg );

			// Create the storehouse provider.
			ISerializer serializer = algorithmPool.GetSerializer();
			IStorehouseProvider storehouseProvider = StorehouseUtils.CreateStorehouseProvider(
				storehouseAccess.StorehouseProviderTypeName, storehouseAccess as DatabaseAccess, storehouseFileName
				, storehousePassword, storehouseHeader, serializer );

			// Get the administration connection string.
			string adminConnectionString = null;
			if( storehouseProvider.UsesServerEngine )
			{
				EditSingleLineNoteWindow window = new EditSingleLineNoteWindow( "Administration Connection String"
					, "This must allow the storehouse server engine to create a storehouse.", AdminConnectionStringDef
					, false, TextValidations.LongNote, true, KernelConsts.LongNoteSizeMax );
				if( !ControlsUtils.ShowDialog( window ) )
					return null;

				adminConnectionString = window.Note;
			}

			// Create the storehouse.
			StorehouseUtils.CreateStorehouse( storehouseProvider, adminConnectionString, storehouseEntropy
				, storehouseConfiguration );

			return storehouseProvider;
		}

		/// <summary>
		/// Opens an existing storehouse.
		/// </summary>
		/// <param name="storehouseAccess">The storehouse access.</param>
		/// <param name="storehouseFileName">The (initial) storehouse file name. This may be null or empty.</param>
		/// <param name="browseStorehouseFile">If this is true then the dialog to open a storehouse file is shown.</param>
		/// <param name="storehousePassword">If this has content, the storehouse is automatically opened with this password; no
		/// password dialog is displayed.</param>
		/// <param name="serializer">The serializer of the sheafs.</param>
		/// <returns>Returns the storehouse provider of the opened storehouse, or null if none was opened.</returns>
		/// <remarks>
		/// <para>The storehouse's entropy state (<see cref="Vela.Crypto.EntropyPool"/>) is restored in the application's
		/// entropy pool.</para>
		/// </remarks>
		public static IStorehouseProvider OpenStorehouse( IStorehouseAccess storehouseAccess, string storehouseFileName
			, bool browseStorehouseFile, string storehousePassword, ISerializer serializer )
		{
			ValidationUtils.ObjectHasContent( storehouseAccess, "storehouseAccess" );

			string windowTitle = "Open Storehouse";

			if( browseStorehouseFile || TextUtils.NoContent( storehouseFileName )
				|| !DatabaseProvider.HasStorehouseFile( storehouseFileName ) )
			{
				OpenFileDialog ofd = new OpenFileDialog();
				ofd.Title = windowTitle;
				ofd.AddExtension = true;
				ofd.DefaultExt = FileExtensions.Storehouse;
				ofd.Filter = FileExtensions.StorehouseFilter;
				ofd.InitialDirectory = FileUtils.GetExistingFolderName( storehouseFileName );
				ofd.CheckFileExists = true;
				ofd.CheckPathExists = true;
				if( !ControlsUtils.ShowDialog( ofd ) )
					return null;

				storehouseFileName = ofd.FileName;
			}

			bool askPassword = TextUtils.NoContent( storehousePassword );
			while( true )
			{
				if( askPassword )
				{
					PasswordWindow passwordWindow = new PasswordWindow( windowTitle, storehouseFileName, null, false );
					if( !ControlsUtils.ShowDialog( passwordWindow ) )
						return null;

					storehouseFileName = passwordWindow.StorehouseFileName;
					storehousePassword = passwordWindow.Password;
				}
				else
				{
					askPassword = true;
				}

				IStorehouseProvider storehouseProvider = null;

				// Capture the exceptions during opening.
				try
				{
					// Create the storehouse provider.
					storehouseProvider = StorehouseUtils.OpenStorehouse( storehouseAccess.StorehouseProviderTypeName
						, storehouseAccess as DatabaseAccess, storehouseFileName, storehousePassword, serializer );
				}
				catch( Exception ex )
				{
					if( ExceptionUtils.IsCritical( ex ) )
						throw ExceptionUtils.PreserveStackTrace( ex );

					// Log even critical exceptions.
					LogManager.Add( ex );

					ControlsUtils.ShowException( ex );
				}

				if( storehouseProvider != null )
				{
					// Load the entropy algorithm from the storehouse and replace the global entropy algorithm with it.
					StorehouseEntropy entropy = StorehouseUtils.LoadEntropy( storehouseProvider );
					EntropyPool.MergeAndReplace( entropy.EntropyAlg );

					return storehouseProvider;
				}
			}
		}

		/// <summary>
		/// Checks that the communication frame algorithms of all the vaults from the storehouse, with the same processing port
		/// as the one from the specified algorithm, are the same.
		/// </summary>
		/// <param name="transaction">The transaction with which to load the documents.</param>
		/// <param name="processingPort">The processing port for which to compare the algorithms.</param>
		/// <param name="communicationFrameAlg">The communication frame algorithm which to compare with the ones used by the
		/// vaults from the storehouse.</param>
		/// <returns>Returns true if the communication frame algorithms are the same.</returns>
		public static bool SameUpnAlgorithmsPerPort( IStorehouseTransaction transaction, UInt16 processingPort
			, ICommunicationFrameAlg communicationFrameAlg )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( communicationFrameAlg, "communicationFrameAlg" );

			// Load the vaults.
			IDeserializer deserializer;
			Sheafs vaults = transaction.CollectSheafs( PortableTypes.FindAllDerived( typeof( Vault ) ), null
				, StorehouseRealms.ProcessingIdentityContainers, null, SheafLocks.Shared, null, out deserializer ) as Sheafs;
			if( ListUtils.NoContent( vaults ) )
				return true;

			foreach( Vault vault in vaults )
			{
				Upn vaultUpn = vault.ProcessingIdentity.Upn;
				ICommunicationFrameAlg vaultCommunicationFrameAlg = vaultUpn.UpnAlg.CommunicationFrameAlg;

				if( vaultUpn.Port == processingPort )
				{
					string vaultAlgName = RegistrableAlgorithmUtils.GetAlgorithmName( vaultUpn.UpnAlg.CommunicationFrameAlg );
					string frameAlgName = RegistrableAlgorithmUtils.GetAlgorithmName( communicationFrameAlg );

					// Immutable algorithms with identical names are identical.
					if( !TextUtils.Equals( vaultAlgName, frameAlgName ) )
						return false;
				}
			}

			return true;
		}

		/// <summary>
		/// Collects the list of root presenters.
		/// </summary>
		/// <param name="assembly">The assembly from where to get the root presenters. If this is null, the application's
		/// assembly is used.</param>
		/// <param name="rootPresentersResourcePrefix">The prefix used to get the resource files which contain the root peer
		/// presenters.</param>
		/// <param name="onlyRemoteUpnFingerprints">The remote UPN fingerprints of the documents to collect. This may be
		/// empty.</param>
		/// <param name="sheafs">List where to add the collected root presenters. The dictionary key is the instance locator of
		/// the current sheaf. If an item with the same key already exists, it's updated. This may be empty, but not null.
		/// </param>
		/// <remarks>
		/// <para>Expired documents are not collected.</para>
		/// </remarks>
		public static void CollectRootPresenters( Assembly assembly, string rootPresentersResourcePrefix
			, Locators onlyRemoteUpnFingerprints, Sheafs sheafs )
		{
			ValidationUtils.TextHasContent( rootPresentersResourcePrefix, "rootPresentersResourcePrefix" );
			ValidationUtils.ObjectHasContent( onlyRemoteUpnFingerprints, "onlyUpnFingerprints" );
			ValidationUtils.ObjectHasContent( sheafs, "sheafs" );

			string[] names = ResourceUtils.GetEmbeddedNames( assembly, rootPresentersResourcePrefix );

			foreach( string name in names )
			{
				byte[] buffer = ResourceUtils.GetEmbeddedBinary( assembly, name );

				IDeserializer deserializer;
				Presenter presenter = SerializationUtils.LoadFromBuffer( buffer, typeof( Presenter ), out deserializer )
					as Presenter;

				if( (presenter != null) && onlyRemoteUpnFingerprints.ContainsKey( presenter.CreatorUpnFingerprint )
					&& !KernelUtils.HasExpired( presenter, 0, false ) )
				{
					sheafs.Add( presenter, true );
				}
			}
		}

		/// <summary>
		/// Collects from the storehouse the documents with the specified initiator UPN fingerprints. Only the sheafs derived
		/// from <see cref="ILocallyNamedPublicProcessingIdentitySheaf"/> or
		/// <see cref="ILocallyNamedPrivateProcessingIdentitySheaf"/> (depending on "includePublicIdentities") are collected.
		/// </summary>
		/// <param name="transaction">Transaction to use.</param>
		/// <param name="identityUpnFingerprints">The UPN fingerprints of the identities from the documents to collect. This may
		/// be empty.</param>
		/// <param name="includePublicIdentities">If this is true then the sheafs derived from
		/// <see cref="ILocallyNamedPublicProcessingIdentitySheaf"/> are collected, else only derived from
		/// <see cref="ILocallyNamedPrivateProcessingIdentitySheaf"/> are collected.</param>
		/// <param name="sheafs">List where to add the collected sheafs. This may be empty, but not null.</param>
		/// <remarks>
		/// <para>For each specified UPN fingerprint, only one (= the first) sheaf is collected.</para>
		/// 
		/// <para>For each found document, if a tally is also found, the document is collected only if the tally's approval is at
		/// least <see cref="PeerDocumentsConsts.KnownPeerTallyApprovalMin"/>).</para>
		/// 
		/// <para>Expired documents are not collected.</para>
		/// </remarks>
		public static void CollectSheafsWithIdentities( IStorehouseTransaction transaction, Locators identityUpnFingerprints
			, bool includePublicIdentities, Sheafs sheafs )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( identityUpnFingerprints, "upnFingerprints" );
			ValidationUtils.ObjectHasContent( sheafs, "sheafs" );

			foreach( Locator upnFingerprint in identityUpnFingerprints )
			{
				IDeserializer deserializer;

				SpeckleFilter identityUpnFilter = new SpeckleFilter( Sheaf.ProcessingIdentityUpnFingerprintFilterIdentifier
					, upnFingerprint );

				Type documentBaseType = includePublicIdentities ? typeof( ILocallyNamedPublicProcessingIdentitySheaf )
					: typeof( ILocallyNamedPrivateProcessingIdentitySheaf );

				// Collect the document.
				ISheaf sheaf = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( documentBaseType )
					, new SpeckleFilters( identityUpnFilter ), StorehouseRealms.ProcessingIdentityContainers
					, SheafLocks.Shared, out deserializer );

				if( (sheaf == null) || KernelUtils.HasExpired( sheaf, 0, false ) )
					continue;

				// If there is a tally for the current UPN fingerprint, load it in order to check the approval.
				if( sheaf as Tally == null )
				{
					// Find the tally.
					Tally tally = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Tally ) )
						, new SpeckleFilters( identityUpnFilter ), StorehouseRealms.ProcessingIdentityContainers
						, SheafLocks.Shared, out deserializer ) as Tally;

					if( tally != null )
					{
						if( (tally.Approval < PeerDocumentsConsts.KnownPeerTallyApprovalMin)
							|| KernelUtils.HasExpired( tally, 0, false ) )
						{
							continue;
						}

						// The tally has priority because it's modified locally, in particular its entity name.
						sheaf = tally;
					}
				}

				sheafs.Add( sheaf );
			}
		}

		/// <summary>
		/// Collects from the storehouse the documents with the specified initiator UPN fingerprints. Only the sheafs derived
		/// from <see cref="ILocallyNamedPublicProcessingIdentitySheaf"/> or
		/// <see cref="ILocallyNamedPrivateProcessingIdentitySheaf"/> (depending on "includePublicIdentities") are collected.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="identityUpnFingerprints">The UPN fingerprints of the identities from the documents to collect. This may
		/// be empty.</param>
		/// <param name="includePublicIdentities">If this is true then the sheafs derived from
		/// <see cref="ILocallyNamedPublicProcessingIdentitySheaf"/> are collected, else only derived from
		/// <see cref="ILocallyNamedPrivateProcessingIdentitySheaf"/> are collected.</param>
		/// <param name="sheafs">List where to add the collected sheafs. This may be empty, but not null.</param>
		/// <remarks>
		/// <para>For each specified UPN fingerprint, only one (= the first) sheaf is collected.</para>
		/// 
		/// <para>For each found document, if a tally is also found, the document is collected only if the tally's approval is at
		/// least <see cref="PeerDocumentsConsts.KnownPeerTallyApprovalMin"/>).</para>
		/// 
		/// <para>Expired documents are not collected.</para>
		/// </remarks>
		public static void CollectSheafsWithIdentities( IStorehouseProvider storehouseProvider, Locators identityUpnFingerprints
			, bool includePublicIdentities, Sheafs sheafs )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				CollectSheafsWithIdentities( transaction, identityUpnFingerprints, includePublicIdentities, sheafs );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Gets a list of the entity names of the specified sheafs which have the specified sheaf locators.
		/// </summary>
		/// <param name="sheafs">Sheafs from which to get the entity names. This may be null or empty.</param>
		/// <param name="sheafLocators">The sheaf locators of the sheaf whose entity names to get. This may be null or empty.
		/// </param>
		/// <returns>Returns the list of entity names. This may be null or empty.</returns>
		public static string GetEntityNames( Sheafs sheafs, Locators sheafLocators )
		{
			if( ListUtils.NoContent( sheafs ) || ListUtils.NoContent( sheafLocators ) )
				return null;

			StringBuilder sb = new StringBuilder();

			foreach( Locator sheafLocator in sheafLocators )
			{
				ISheaf namedDocument = sheafs[ sheafLocator ];
				if( namedDocument != null )
				{
					INamedEntity namedEntity = (INamedEntity)namedDocument;

					if( sb.Length <= 0 )
						sb.AppendFormat( "\"{0}\"", namedEntity.EntityName );
					else
						sb.AppendFormat( ", \"{0}\"", namedEntity.EntityName );
				}
			}

			return sb.ToString();
		}

		/// <summary>
		/// Gets a list of the entity names and times of the specified sheafs which have the specified locators.
		/// </summary>
		/// <param name="sheafs">Sheafs from which to get the entity names. This may be null or empty.</param>
		/// <param name="sheafLocatorTimes">The locator-times (in UTC format) of the sheafs whose entity names and times to get.
		/// This may be null or empty.</param>
		/// <param name="lineTemplate">The template of each line which shows an entity name and time. Placeholders: 0 - entity
		/// name, 1 - time.</param>
		/// <param name="lineSeparator">The separator between consecutive lines which show an entity name and time.</param>
		/// <returns>Returns the list of entity names and times. This may be null or empty.</returns>
		public static string GetEntityNamesAndTimes( Sheafs sheafs, LocatorTimes sheafLocatorTimes, string lineTemplate
			, string lineSeparator )
		{
			if( ListUtils.NoContent( sheafs ) || ListUtils.NoContent( sheafLocatorTimes ) )
				return null;

			StringBuilder sb = new StringBuilder();

			int count = sheafLocatorTimes.Count;
			int i = 1;

			foreach( LocatorTime sheafLocatorTime in sheafLocatorTimes )
			{
				ISheaf namedDocument = sheafs[ sheafLocatorTime.Locator ];
				if( namedDocument != null )
				{
					INamedEntity namedEntity = (INamedEntity)namedDocument;
					string sheafLocatorTimeText = sheafLocatorTime.Time.ToLocalTime().ToString( KernelConsts.TimeUiFormat );

					sb.AppendFormat( lineTemplate, namedEntity.EntityName, sheafLocatorTimeText );

					if( i < count )
						sb.Append( lineSeparator );
				}

				i++;
			}

			return sb.ToString();
		}

		/// <summary>
		/// Gets a list of the identities who may have signed the specified attesting document.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction. This may
		/// be null if "includeSheafs" is false.</param>
		/// <param name="attestingDocument">Attesting document whose valid signatures are collected.</param>
		/// <param name="includeRootPresenters">If this is true then the root presenters are also included.</param>
		/// <param name="assembly">The assembly from where to get the root presenters. If this is null, the application's
		/// assembly is used.</param>
		/// <param name="rootPresentersResourcePrefix">The prefix used to get the resource files which contain the root peer
		/// presenters. This may be null or empty if "includeRootPresenters" is false.</param>
		/// <param name="includeSheafs">If this is true then the sheafs derived from
		/// <see cref="ILocallyNamedPublicProcessingIdentitySheaf"/> are also included.</param>
		/// <param name="signatureVerifierSheaf">Sheaf whose identities to include. This may be null if "includeRootPresenters"
		/// is true. This must be null if "includeSheafs" is true.</param>
		/// <param name="signingSheafs">Is set with the list of sheafs which may have signed the specified attesting document.
		/// </param>
		/// <param name="identityPublishers">Is set with the list of identity publishers which may have signed the specified
		/// attesting document. The dictionary key is the instance locator of the current sheaf. This may be empty, but not
		/// null.</param>
		/// <param name="identities">Is set with the list with identities which may have signed the specified attesting document.
		/// The dictionary key is the instance locator of the current sheaf. This may be empty, but not null.</param>
		/// <returns>Returns true if the specified storehouse contains identities which may have signed the specified attesting
		/// document. This is determined by intersecting "signingSheafs" with "identityPublishers" and "identities".</returns>
		/// <remarks>
		/// <para>Expired documents are not included.</para>
		/// </remarks>
		public static bool GetPotentialSigningIdentities( IStorehouseProvider storehouseProvider
			, AttestingDocument attestingDocument, bool includeRootPresenters, Assembly assembly
			, string rootPresentersResourcePrefix, bool includeSheafs
			, ILocallyNamedPublicProcessingIdentitySheaf signatureVerifierSheaf, out Sheafs signingSheafs
			, out Dictionary<Locator, IIdentityPublisher> identityPublishers
			, out Dictionary<Locator, PublicIdentity> identities )
		{
			#region Validate data

			if( includeSheafs )
				ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			ValidationUtils.ObjectHasContent( attestingDocument, "attestingDocument" );

			if( includeRootPresenters )
				ValidationUtils.TextHasContent( rootPresentersResourcePrefix, "rootPresentersResourcePrefix" );

			if( !includeRootPresenters && !includeSheafs )
				ValidationUtils.ObjectHasContent( signatureVerifierSheaf, "signatureVerifierSheaf" );

			if( includeSheafs && (signatureVerifierSheaf != null) )
			{
				throw new ValidationException( "If the sheafs are used to get the signing entity names, the individual sheaf"
					+ " must be missing." );
			}

			#endregion Validate data

			// Collect the signing sheaf locators.
			Locators signingUpnFingerprints = attestingDocument.CollectSigningUpnFingerprints();

			signingSheafs = new Sheafs();

			// Collect the signing sheafs.
			if( includeSheafs )
				CollectSheafsWithIdentities( storehouseProvider, signingUpnFingerprints, true, signingSheafs );

			if( (signatureVerifierSheaf != null) && !KernelUtils.HasExpired( signatureVerifierSheaf, 0, false )
				&& CoreDocumentsUtils.HasIdentityWithUpnFingerprint( signatureVerifierSheaf, signingUpnFingerprints ) )
			{
				signingSheafs.Add( signatureVerifierSheaf );
			}

			// Collect the root presenters. These will update the existing sheafs.
			if( includeRootPresenters )
				CollectRootPresenters( assembly, rootPresentersResourcePrefix, signingUpnFingerprints, signingSheafs );

			// Collect the signing identities from the signing sheafs.
			identityPublishers = new Dictionary<Locator, IIdentityPublisher>();
			identities = new Dictionary<Locator, PublicIdentity>();
			CoreDocumentsUtils.CollectIdentities( signingSheafs, includeRootPresenters, identityPublishers, identities );

			return ListUtils.HasContent( identityPublishers ) || ListUtils.HasContent( identities );
		}

		/// <summary>
		/// Gets a list of the names of the entities who have signed the specified attesting document.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction. This may
		/// be null if "includeSheafs" is false.</param>
		/// <param name="attestingDocument">Attesting document whose valid signatures are collected.</param>
		/// <param name="macValue">MAC value which must be equal with the one from the attesting document in order to consider
		/// a signature valid.</param>
		/// <param name="includeRootPresenters">If this is true then the root presenters are also used to verify the
		/// signatures.</param>
		/// <param name="assembly">The assembly from where to get the root presenters. If this is null, the application's
		/// assembly is used.</param>
		/// <param name="rootPresentersResourcePrefix">The prefix used to get the resource files which contain the root peer
		/// presenters. This may be null or empty if "includeRootPresenters" is false.</param>
		/// <param name="includeSheafs">If this is true then the sheafs derived from
		/// <see cref="ILocallyNamedPublicProcessingIdentitySheaf"/> are also used to verify the signatures.</param>
		/// <param name="signatureVerifierSheaf">Sheaf with whose identities to verify the signatures of the
		/// attesting document. This may be null if "includeRootPresenters" is true. This must be null if
		/// "includeSheafs" is true.</param>
		/// <param name="lineTemplate">The template of each line which shows a signing entity name and time. Placeholders:
		/// 0 - signing entity name, 1 - signing time (for a group identity, it's the newest signing time).</param>
		/// <param name="lineSeparator">The separator between consecutive lines which show a signing entity name and time.
		/// </param>
		/// <returns>Returns a list of the names of the entities who have signed the specified attesting document.</returns>
		/// <remarks>
		/// <para>Expired documents are not used to verify the signatures.</para>
		/// </remarks>
		public static string GetSigningEntityNames( IStorehouseProvider storehouseProvider, AttestingDocument attestingDocument
			, byte[] macValue, bool includeRootPresenters, Assembly assembly, string rootPresentersResourcePrefix
			, bool includeSheafs, ILocallyNamedPublicProcessingIdentitySheaf signatureVerifierSheaf, string lineTemplate
			, string lineSeparator )
		{
			ValidationUtils.ArrayHasContent( macValue, "macValue" );

			Sheafs signingSheafs;
			Dictionary<Locator, IIdentityPublisher> identityPublishers;
			Dictionary<Locator, PublicIdentity> identities;
			bool hasSigners = GetPotentialSigningIdentities( storehouseProvider, attestingDocument, includeRootPresenters
				, assembly, rootPresentersResourcePrefix, includeSheafs, signatureVerifierSheaf, out signingSheafs
				, out identityPublishers, out identities );

			if( !hasSigners )
			{
				throw new ValidationException( "No potential signers of the attesting document were found.", null, null
					, new ExceptionDetail( "Attesting document locator", attestingDocument.InstanceLocator ) );
			}

			// The locator-times of the peer documents which have actually signed the securing document.
			LocatorTimes actualSigningSheafLocatorTimes = new LocatorTimes();

			// Verify the signatures.
			attestingDocument.CollectValidSignatures( identityPublishers, macValue, actualSigningSheafLocatorTimes );
			attestingDocument.CollectValidSignatures( identities, macValue, actualSigningSheafLocatorTimes );

			return GetEntityNamesAndTimes( signingSheafs, actualSigningSheafLocatorTimes, lineTemplate, lineSeparator );
		}

		/// <summary>
		/// Loads the document from the specified buffer and decrypts it.
		/// </summary>
		/// <param name="transaction">Transaction to use. This may be null if "includeSheafs" is false.</param>
		/// <param name="documentBuffer">The buffer from which to load the document.</param>
		/// <param name="originalFileName">The name of the file from which the document buffer was loaded. This may be null or
		/// empty. This is used only for exception details.</param>
		/// <param name="includeSheafs">If this is true then the sheafs derived from
		/// <see cref="ILocallyNamedPrivateProcessingIdentitySheaf"/> are used for decryption.</param>
		/// <param name="decryptorSheaf">Sheaf with whose identities to decrypt the document's envelope. This must be null if
		/// "includeSheafs" is true.</param>
		/// <param name="documentType">The type of the document to load; the document to load may either have this type or
		/// a type derived from this.</param>
		/// <returns>Returns the loaded document.</returns>
		public static IDocument LoadAsymmEncryptedDocument( IStorehouseTransaction transaction, byte[] documentBuffer
			, string originalFileName, bool includeSheafs, ILocallyNamedPrivateProcessingIdentitySheaf decryptorSheaf
			, Type documentType )
		{
			#region Validate data

			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ListHasContent( documentBuffer, "documentBuffer" );

			if( !includeSheafs )
				ValidationUtils.ObjectHasContent( decryptorSheaf, "decryptorSheaf" );

			if( includeSheafs && (decryptorSheaf != null) )
			{
				throw new ValidationException( "If the sheafs are used to get the encrypting and signing documents, the"
					+ " individual sheafs must be missing." );
			}

			ValidationUtils.ObjectHasContent( documentType, "documentType" );

			#endregion Validate data

			IDeserializer envelopeDeserializer;
			AsymmEncryptedEnvelope asymmEncryptedEnvelope = SerializationUtils.LoadFromBuffer( documentBuffer
				, typeof( AsymmEncryptedEnvelope ), out envelopeDeserializer ) as AsymmEncryptedEnvelope;
			if( asymmEncryptedEnvelope == null )
			{
				throw new ValidationException( "Failed to load the encrypted document.", null, null
					, TextUtils.HasContent( originalFileName ) ? new ExceptionDetail( "File name", originalFileName ) : null );
			}

			Locators encryptingUpnFingerprints = asymmEncryptedEnvelope.CollectEncryptingUpnFingerprints();

			Sheafs decryptorDocuments = new Sheafs();

			// Collect the decryptor sheafs.
			if( includeSheafs )
				CollectSheafsWithIdentities( transaction, encryptingUpnFingerprints, false, decryptorDocuments );

			if( (decryptorSheaf != null) && !KernelUtils.HasExpired( decryptorSheaf, 0, false ) )
				decryptorDocuments.Add( decryptorSheaf );

			// Collect the decryptor identities from the decryptor sheafs.
			Dictionary<Locator, PrivateIdentity> identities = new Dictionary<Locator, PrivateIdentity>();
			CoreDocumentsUtils.CollectPrivateIdentities( decryptorDocuments, identities );

			// Find the first identity which will decrypt the document. In theory, a private identity which can't decrypt the
			// document should not be in the list, yet exceptions are captured in order to make the mechanism more flexible.
			foreach( KeyValuePair<Locator, PrivateIdentity> kvp in identities )
			{
				try
				{
					PrivateIdentity decryptorIdentity = kvp.Value;

					AsymmDecryptionPassword asymmDecryptionPassword = new AsymmDecryptionPassword( decryptorIdentity );

					IDeserializer documentDeserializer;
					EncryptedEnvelope.LoadFromEnvelope( asymmEncryptedEnvelope, asymmDecryptionPassword, documentType
						, out documentDeserializer );

					IDocument document = asymmEncryptedEnvelope.EncryptableDocument;
					if( (document != null) && documentType.IsAssignableFrom( document.GetType() ) )
						return document;
				}
				catch( Exception ex )
				{
					// The loop must continue (if possible).

					if( ExceptionUtils.IsAbort( ex ) )
						throw ExceptionUtils.PreserveStackTrace( ex );

					LogManager.Add( ex );
				}
			}

			throw new ValidationException( "Failed to decrypt the document. It's possible that you are not the intended"
					+ " recipient. If you have created the encrypted document, you can't decrypt it.", null, null
				, TextUtils.HasContent( originalFileName ) ? new ExceptionDetail( "File name", originalFileName ) : null );
		}

		/// <summary>
		/// Loads the document from the specified file and decrypts it.
		/// </summary>
		/// <param name="transaction">Transaction to use. This may be null if "includeSheafs" is false.</param>
		/// <param name="fileName">The name of the file from which to load the document.</param>
		/// <param name="includeSheafs">If this is true then the sheafs derived from
		/// <see cref="ILocallyNamedPrivateProcessingIdentitySheaf"/> are used for decryption.</param>
		/// <param name="decryptorSheaf">Sheaf with whose identities to decrypt the document's envelope. This must be null if
		/// "includeSheafs" is true.</param>
		/// <param name="documentType">The type of the document to load; the document to load may either have this type or
		/// a type derived from this.</param>
		/// <returns>Returns the loaded document.</returns>
		public static IDocument LoadAsymmEncryptedDocument( IStorehouseTransaction transaction, string fileName
			, bool includeSheafs, ILocallyNamedPrivateProcessingIdentitySheaf decryptorSheaf, Type documentType )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			byte[] documentBuffer = FileUtils.Read( fileName, true );

			return LoadAsymmEncryptedDocument( transaction, documentBuffer, fileName, includeSheafs, decryptorSheaf
				, documentType );
		}

		/// <summary>
		/// Loads the document from the specified file and decrypts it.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction. This may
		/// be null if "includeSheafs" is false.</param>
		/// <param name="fileName">The name of the file from which to load the document.</param>
		/// <param name="includeSheafs">If this is true then the sheafs derived from
		/// <see cref="ILocallyNamedPrivateProcessingIdentitySheaf"/> are used for decryption.</param>
		/// <param name="decryptorSheaf">Sheaf with whose identities to decrypt the document's envelope. This must be null if
		/// "includeSheafs" is true.</param>
		/// <param name="documentType">The type of the document to load.</param>
		/// <returns>Returns the loaded document.</returns>
		public static IDocument LoadAsymmEncryptedDocument( IStorehouseProvider storehouseProvider, string fileName
			, bool includeSheafs, ILocallyNamedPrivateProcessingIdentitySheaf decryptorSheaf, Type documentType )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			if( storehouseProvider != null )
			{
				// Create a transaction.
				IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

				try
				{
					IDocument document = LoadAsymmEncryptedDocument( transaction, fileName, includeSheafs, decryptorSheaf
						, documentType );

					// Commit the transaction.
					transaction.Commit();

					return document;
				}
				catch( Exception ex )
				{
					// Rolls back the transaction.
					transaction.Rollback();

					throw ExceptionUtils.PreserveStackTrace( ex );
				}
			}
			else
			{
				return LoadAsymmEncryptedDocument( (IStorehouseTransaction)null, fileName, includeSheafs, decryptorSheaf
					, documentType );
			}
		}

		/// <summary>
		/// Loads the account context for the specified account locator.
		/// </summary>
		/// <param name="transaction">The transaction with which to load the documents.</param>
		/// <param name="accountLocator">Account locator.</param>
		/// <param name="accountLock">The lock for the account to load.</param>
		/// <param name="tallyLock">The lock for the tally to load.</param>
		/// <param name="throwIfMissing">If this is true and the account with the specified locator doesn't exist, an exception
		/// is thrown.</param>
		/// <returns>Returns the loaded account context. This is null if the account doesn't exist and "throwIfMissing" is
		/// false.</returns>
		/// <remarks>
		/// <para>The sheafs are not locked.</para>
		/// </remarks>
		public static AccountContext LoadAccountContext( IStorehouseTransaction transaction, Locator accountLocator
			, SheafLocks accountLock, SheafLocks tallyLock, bool throwIfMissing )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( accountLocator, "accountLocator" );

			IDeserializer deserializer;

			#region Load the account

			Account account = transaction.LoadSheaf( accountLocator, StorehouseRealms.Accounts, typeof( Account )
				, accountLock, out deserializer ) as Account;

			if( account == null )
			{
				if( throwIfMissing )
				{
					throw new ValidationException( "The account is missing.", null, null
						, new ExceptionDetail( "Account locator", accountLocator ) );
				}
				else
					return null;
			}

			#endregion Load the account

			#region Load the vault

			SpeckleFilter vaultUpnFilter = new SpeckleFilter( Vault.LocalUpnFingerprintFilterIdentifier
				, account.LocalUpnFingerprint );
			Vault vault = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Vault ) )
				, new SpeckleFilters( vaultUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared
				, out deserializer ) as Vault;

			if( vault == null )
			{
				throw new ValidationException( "The vault is missing.", null, null
					, new ExceptionDetail( "Local UPN fingerprint", account.LocalUpnFingerprint ) );
			}

			#endregion Load the vault

			#region Load the local photo

			SpeckleFilter localPhotoUpnFilter = new SpeckleFilter( Photo.CreatorUpnFingerprintFilterIdentifier
				, account.LocalUpnFingerprint );
			Photo localPhoto = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Photo ) )
				, new SpeckleFilters( localPhotoUpnFilter ), StorehouseRealms.Photos, SheafLocks.Shared
				, out deserializer ) as Photo;

			#endregion Load the local photo

			#region Load the presenter

			SpeckleFilter presenterUpnFilter = new SpeckleFilter( Presenter.CreatorUpnFingerprintFilterIdentifier
				, account.RemoteUpnFingerprint );
			Presenter presenter = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Presenter ) )
				, new SpeckleFilters( presenterUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared
				, out deserializer ) as Presenter;

			if( presenter == null )
			{
				throw new ValidationException( "The presenter is missing.", null, null
					, new ExceptionDetail( "Remote UPN fingerprint", account.RemoteUpnFingerprint ) );
			}

			#endregion Load the presenter

			#region Load the tally

			SpeckleFilter tallyUpnFilter = new SpeckleFilter( Tally.RemoteUpnFingerprintFilterIdentifier
				, account.RemoteUpnFingerprint );
			Tally tally = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Tally ) )
				, new SpeckleFilters( tallyUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, tallyLock
				, out deserializer ) as Tally;

			if( tally == null )
			{
				throw new ValidationException( "The tally is missing.", null, null
					, new ExceptionDetail( "Remote UPN fingerprint", account.RemoteUpnFingerprint ) );
			}

			#endregion Load the tally

			#region Load the remote photo

			SpeckleFilter remotePhotoUpnFilter = new SpeckleFilter( Photo.CreatorUpnFingerprintFilterIdentifier
				, account.RemoteUpnFingerprint );
			Photo remotePhoto = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Photo ) )
				, new SpeckleFilters( remotePhotoUpnFilter ), StorehouseRealms.Photos, SheafLocks.Shared
				, out deserializer ) as Photo;

			#endregion Load the remote photo

			return new AccountContext( account, vault, localPhoto, presenter, tally, remotePhoto );
		}

		/// <summary>
		/// Checks if the storehouse with the specified provider contains at least 1 vault.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <returns>Returns true if the storehouse contains at least 1 vault.</returns>
		public static bool HasVault( IStorehouseProvider storehouseProvider )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				Speckle speckle = transaction.CollectFirstSpeckle( PortableTypes.FindAllDerived( typeof( Vault ) )
					, null, StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared, null, null );

				// Commit the transaction.
				transaction.Commit();

				return speckle != null;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Gets the entity name of the vault with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="vaultLocator">The locator of the vault whose entity name to get.</param>
		/// <returns>Returns the entity name of the vault.</returns>
		public static string GetVaultEntityName( IStorehouseProvider storehouseProvider, Locator vaultLocator )
		{
			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				Speckle vaultSpeckle = transaction.LoadSpeckle( vaultLocator, StorehouseRealms.ProcessingIdentityContainers
					, SheafLocks.Shared, null, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Vault ) ) );
				ValidationUtils.ObjectHasContent( vaultSpeckle, "vaultSpeckle" );

				// Commit the transaction.
				transaction.Commit();

				return vaultSpeckle.GetLoadedSpeckleFilter<string>( Vault.EntityNameFilterIdentifier );
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				if( storehouseProvider.IsTimeoutException( ex ) )
				{
					throw new Vela.Kernel.TimeoutException( "The vault or one of its dependencies is being processed by a"
						+ " background operation. Try again later.", ex );
				}

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the vault speckles from the specified storehouse.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <returns>Returns the loaded vault speckles.</returns>
		public static VaultSpeckles CollectVaultSpeckles( IStorehouseProvider storehouseProvider )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the vault speckles.
				VaultSpeckles speckles = new VaultSpeckles();
				transaction.CollectSpeckles( PortableTypes.FindAllDerived( typeof( Vault ) ), null
					, StorehouseRealms.ProcessingIdentityContainers, null, speckles, typeof( VaultSpeckle )
					, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Vault ) ) );

				// Commit the transaction.
				transaction.Commit();

				return speckles;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the vault speckle with the specified local UPN fingerprint, with all its speckle filters.
		/// </summary>
		/// <param name="transaction">The transaction with which to load the speckle.</param>
		/// <param name="localUpnFingerprint">The local UPN fingerprint of the speckle to load.</param>
		/// <returns>Returns the loaded speckle, or null if none is found.</returns>
		public static Speckle LoadVaultSpeckleByUpnFingerprint( IStorehouseTransaction transaction, Locator localUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( localUpnFingerprint, "localUpnFingerprint" );

			SpeckleFilter vaultUpnFilter = new SpeckleFilter( Vault.LocalUpnFingerprintFilterIdentifier
				, localUpnFingerprint );

			return transaction.CollectFirstSpeckle( PortableTypes.FindAllDerived( typeof( Vault ) )
				, new SpeckleFilters( vaultUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared, null
				, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Vault ) ) );
		}

		/// <summary>
		/// Loads the vaults from the specified storehouse.
		/// </summary>
		/// <param name="transaction">The transaction with which to load the sheafs.</param>
		public static Sheafs CollectVaults( IStorehouseTransaction transaction )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );

			IDeserializer deserializer;

			// Load the vault sheafs.
			IKernelObservableList vaults = transaction.CollectSheafs( PortableTypes.FindAllDerived( typeof( Vault ) )
				, null, StorehouseRealms.ProcessingIdentityContainers, null, SheafLocks.Shared, null, out deserializer );

			return vaults as Sheafs;
		}

		/// <summary>
		/// Loads the vaults from the specified storehouse.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		public static Sheafs CollectVaults( IStorehouseProvider storehouseProvider )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the vault sheafs.
				Sheafs vaults = CollectVaults( transaction );

				// Commit the transaction.
				transaction.Commit();

				return vaults;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the vault with the specified vault locator, from the storehouse.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="vaultLocator">The locator of the vault to load.</param>
		/// <returns>Returns the loaded vault, or null if none is found.</returns>
		public static Vault LoadVault( IStorehouseProvider storehouseProvider, Locator vaultLocator )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( vaultLocator, "vaultLocator" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				IDeserializer deserializer;
				Vault vault = transaction.LoadSheaf( vaultLocator, StorehouseRealms.ProcessingIdentityContainers
					, typeof( Vault ), SheafLocks.Shared, out deserializer ) as Vault;

				// Commit the transaction.
				transaction.Commit();

				return vault;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the vault with the specified local UPN fingerprint.
		/// </summary>
		/// <param name="transaction">The transaction with which to load the vault.</param>
		/// <param name="localUpnFingerprint">The local UPN fingerprint of the vault to load.</param>
		/// <returns>Returns the loaded vault, or null if none is found.</returns>
		public static Vault LoadVaultByUpnFingerprint( IStorehouseTransaction transaction, Locator localUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( localUpnFingerprint, "localUpnFingerprint" );

			IDeserializer deserializer;

			SpeckleFilter vaultUpnFilter = new SpeckleFilter( Vault.LocalUpnFingerprintFilterIdentifier
				, localUpnFingerprint );

			return transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Vault ) )
				, new SpeckleFilters( vaultUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared
				, out deserializer ) as Vault;
		}

		/// <summary>
		/// Loads the vault and the tally of the account with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="accountLocator">The locator of the account for which to load the data.</param>
		/// <param name="vault">Is set with the loaded vault.</param>
		/// <param name="tally">Is set with the loaded tally.</param>
		/// <remarks>
		/// <para>The loaded sheafs are not locked.</para>
		/// </remarks>
		public static void LoadVaultTally( IStorehouseProvider storehouseProvider, Locator accountLocator, out Vault vault
			, out Tally tally )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( accountLocator, "accountLocator" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				IDeserializer deserializer;
				Account account = transaction.LoadSheaf( accountLocator, StorehouseRealms.Accounts, typeof( Account )
					, SheafLocks.Shared, out deserializer ) as Account;
				if( account == null )
				{
					throw new ValidationException( "The account is missing.", null, null
						, new ExceptionDetail( "Account locator", accountLocator ) );
				}

				SpeckleFilter tallyUpnFilter = new SpeckleFilter( Tally.RemoteUpnFingerprintFilterIdentifier
					, account.RemoteUpnFingerprint );
				Tally tallyTemp = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Tally ) )
					, new SpeckleFilters( tallyUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared
					, out deserializer ) as Tally;
				if( tallyTemp == null )
				{
					throw new ValidationException( "The tally is missing.", null, null
						, new ExceptionDetail( "Remote UPN fingerprint", account.RemoteUpnFingerprint ) );
				}

				SpeckleFilter vaultUpnFilter = new SpeckleFilter( Vault.LocalUpnFingerprintFilterIdentifier
					, account.LocalUpnFingerprint );
				Vault vaultTemp = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Vault ) )
					, new SpeckleFilters( vaultUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared
					, out deserializer ) as Vault;
				if( vaultTemp == null )
				{
					throw new ValidationException( "The vault is missing.", null, null
						, new ExceptionDetail( "Local UPN fingerprint", account.LocalUpnFingerprint ) );
				}

				// Commit the transaction.
				transaction.Commit();

				tally = tallyTemp;
				vault = vaultTemp;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Deletes the vault with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="vaultLocator">The locator of the vault to delete.</param>
		/// <param name="serviceManager">Action service manager which is used to stop processing action requests which have a
		/// multiframe action response.</param>
		/// <param name="clientManager">Action client manager which is used to stop processing action requests which have a
		/// multiframe action response.</param>
		/// <param name="lop">Long operation progress. This may be null.</param>
		/// <returns>Returns true if the vault with the specified locator was deleted.</returns>
		public static bool DeleteVault( IStorehouseProvider storehouseProvider, Locator vaultLocator
			, IActionServiceManager serviceManager, IActionClientManager clientManager, LongOperationProgress lop )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( vaultLocator, "vaultLocator" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				if( lop != null )
					lop.Progress( 0 );

				DeleteVaultSheafs( transaction, vaultLocator, serviceManager, clientManager, lop );

				if( lop != null )
					lop.Progress( 1 );

				// Commit the transaction.
				transaction.Commit();

				return true;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				if( storehouseProvider.IsTimeoutException( ex ) )
				{
					throw new Vela.Kernel.TimeoutException( "The vault or one of its dependencies is being processed by a"
						+ " background operation. Try again later.", ex );
				}

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the photo with the specified creator UPN fingerprint, from the storehouse.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="creatorUpnFingerprint">The creator UPN fingerprint for which to load the photo.</param>
		/// <returns>Returns the loaded photo, or null if none is found.</returns>
		public static Photo LoadPhoto( IStorehouseProvider storehouseProvider, Locator creatorUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( creatorUpnFingerprint, "creatorUpnFingerprint" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				IDeserializer deserializer;
				SpeckleFilter photoUpnFilter = new SpeckleFilter( Photo.CreatorUpnFingerprintFilterIdentifier
					, creatorUpnFingerprint );
				Photo photo = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Photo ) )
					, new SpeckleFilters( photoUpnFilter ), StorehouseRealms.Photos, SheafLocks.Shared
					, out deserializer ) as Photo;

				// Commit the transaction.
				transaction.Commit();

				return photo;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Exports the presenter of the vault with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="vaultLocator">The locator of the vault whose presenter to export.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="hashSerializer">Hash serializer for signing.</param>
		/// <param name="serializerForExport">Serializer for saving the exported file.</param>
		public static void ExportPresenter( IStorehouseProvider storehouseProvider, Locator vaultLocator, MacAlg macAlg
			, IAsymmSigningPaddingAlg asymmSigningPaddingAlg, IHashSerializer hashSerializer, ISerializer serializerForExport )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( vaultLocator, "vaultLocator" );
			ValidationUtils.ObjectHasContent( macAlg, "macAlg" );
			ValidationUtils.ObjectHasContent( asymmSigningPaddingAlg, "asymmSigningPaddingAlg" );
			ValidationUtils.ObjectHasContent( hashSerializer, "hashSerializer" );
			ValidationUtils.ObjectHasContent( serializerForExport, "serializerForExport" );

			Vault vault;
			Photo photo;

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the vault.
				IDeserializer deserializer;
				vault = transaction.LoadSheaf( vaultLocator, StorehouseRealms.ProcessingIdentityContainers, typeof( Vault )
					, SheafLocks.Shared, out deserializer ) as Vault;
				if( vault == null )
				{
					throw new ValidationException( "The vault is missing.", null, null
						, new ExceptionDetail( "Vault locator", vaultLocator ) );
				}

				// Load the photo.
				SpeckleFilter photoUpnFilter = new SpeckleFilter( Photo.CreatorUpnFingerprintFilterIdentifier
					, vault.LocalUpnFingerprint );
				photo = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Photo ) )
					, new SpeckleFilters( photoUpnFilter ), StorehouseRealms.Photos, SheafLocks.Shared
					, out deserializer ) as Photo;

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				if( storehouseProvider.IsTimeoutException( ex ) )
				{
					throw new Vela.Kernel.TimeoutException( "The vault or one of its dependencies is being processed by a"
						+ " background operation. Try again later.", ex );
				}

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			// Create the presenter.
			Presenter presenter = vault.GetPresenter( CryptoUtils.GetLocator(), macAlg, asymmSigningPaddingAlg, photo, true
				, hashSerializer );

			string selectedFileName;
			if( !ControlsUtils.ShowSaveFileDialog( "Export Presenter", FileExtensions.Presenter, FileExtensions.PresenterFilter
				, null, vault.EntityName, true, out selectedFileName ) )
			{
				return;
			}

			// Save the presenter.
			serializerForExport.SaveToFile( presenter, null, selectedFileName );

			TooltipWindow.ShowTooltip();
		}

		/// <summary>
		/// Loads the tally speckle with the specified remote UPN fingerprint, with all its speckle filters.
		/// </summary>
		/// <param name="transaction">The transaction with which to load the speckle.</param>
		/// <param name="remoteUpnFingerprint">The remote UPN fingerprint of the speckle to load.</param>
		/// <returns>Returns the loaded speckle, or null if none is found.</returns>
		public static Speckle LoadTallySpeckleByUpnFingerprint( IStorehouseTransaction transaction
			, Locator remoteUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( remoteUpnFingerprint, "remoteUpnFingerprint" );

			SpeckleFilter tallyFilter = new SpeckleFilter( Tally.RemoteUpnFingerprintFilterIdentifier
				, remoteUpnFingerprint );

			return transaction.CollectFirstSpeckle( PortableTypes.FindAllDerived( typeof( Tally ) )
				, new SpeckleFilters( tallyFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared, null
				, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Tally ) ) );
		}

		/// <summary>
		/// Loads the tally speckle with the specified remote UPN fingerprint, with all its speckle filters.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="remoteUpnFingerprint">The remote UPN fingerprint of the speckle to load.</param>
		/// <returns>Returns the loaded speckle, or null if none is found.</returns>
		public static Speckle LoadTallySpeckleByUpnFingerprint( IStorehouseProvider storehouseProvider
			, Locator remoteUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				Speckle tallySpeckle = LoadTallySpeckleByUpnFingerprint( transaction, remoteUpnFingerprint );

				// Commit the transaction.
				transaction.Commit();

				return tallySpeckle;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the tally with the specified remote UPN fingerprint.
		/// </summary>
		/// <param name="transaction">The transaction with which to load the tally.</param>
		/// <param name="remoteUpnFingerprint">The remote UPN fingerprint of the tally to load.</param>
		/// <returns>Returns the loaded tally, or null if none is found.</returns>
		public static Tally LoadTallyByUpnFingerprint( IStorehouseTransaction transaction, Locator remoteUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( remoteUpnFingerprint, "remoteUpnFingerprint" );

			IDeserializer deserializer;

			SpeckleFilter tallyFilter = new SpeckleFilter( Tally.RemoteUpnFingerprintFilterIdentifier
				, remoteUpnFingerprint );

			return transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Tally ) )
				, new SpeckleFilters( tallyFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared
				, out deserializer ) as Tally;
		}

		/// <summary>
		/// Loads the tally with the specified remote UPN fingerprint.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="remoteUpnFingerprint">The remote UPN fingerprint of the tally to load.</param>
		/// <returns>Returns the loaded tally, or null if none is found.</returns>
		public static Tally LoadTallyByUpnFingerprint( IStorehouseProvider storehouseProvider, Locator remoteUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				Tally tally = LoadTallyByUpnFingerprint( transaction, remoteUpnFingerprint );

				// Commit the transaction.
				transaction.Commit();

				return tally;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the tally of the account with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="accountLocator">The locator of the account whose tally to load.</param>
		/// <returns>Returns the loaded tally, or null if none is found.</returns>
		/// <remarks>
		/// <para>The loaded sheafs are not locked.</para>
		/// </remarks>
		public static Tally LoadTallyOfAccount( IStorehouseProvider storehouseProvider, Locator accountLocator )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( accountLocator, "accountLocator" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the account speckle.
				Speckle accountSpeckle = transaction.LoadSpeckle( accountLocator, StorehouseRealms.Accounts, SheafLocks.Shared
					, null, new SpeckleFilterIdentifiers( Account.RemoteUpnFingerprintFilterIdentifier ) );
				if( accountSpeckle == null )
				{
					// Commit the transaction.
					transaction.Commit();

					return null;
				}

				#region Load the tally

				Locator remoteUpnFingerprint = accountSpeckle.GetLoadedSpeckleFilter<Locator>(
					Account.RemoteUpnFingerprintFilterIdentifier );
				SpeckleFilter tallyUpnFilter = new SpeckleFilter( Tally.RemoteUpnFingerprintFilterIdentifier
					, remoteUpnFingerprint );

				IDeserializer deserializer;
				Tally tally = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Tally ) )
					, new SpeckleFilters( tallyUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared
					, out deserializer ) as Tally;
				if( tally == null )
				{
					// Commit the transaction.
					transaction.Commit();

					return null;
				}

				#endregion Load the tally

				// Commit the transaction.
				transaction.Commit();

				return tally;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Checks if the storehouse with the specified provider contains at least 1 account.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="singleAccountLocator">Is set with the instance locator of the only account from the storehouse with the
		/// specified provider. This is null if there is either no account or there are at least 2 accounts.</param>
		/// <returns>Returns true if the storehouse contains at least 1 account.</returns>
		public static bool HasAccount( IStorehouseProvider storehouseProvider, out Locator singleAccountLocator )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Collect the locators of the first 2 accounts.
				CollectPaging paging = new CollectPaging( true, 0, 2 );
				Locators accountLocators = transaction.CollectKeys( PortableTypes.FindAllDerived( typeof( Account ) )
					, null, StorehouseRealms.Accounts, paging, null, null ) as Locators;

				singleAccountLocator = ((accountLocators != null) && (accountLocators.Count == 1))
					? accountLocators.GetFirstItem() : null;

				// Commit the transaction.
				transaction.Commit();

				return ListUtils.HasContent( accountLocators );
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Gets the local and the remote names of the account with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="accountLocator">The locator of the account whose local and the remote names to get.</param>
		/// <param name="localName">Is set with the name of the local peer of the account.</param>
		/// <param name="remoteName">Is set with the name of the remote peer of the account.</param>
		/// <returns>Returns the local and the remote names of the account.</returns>
		public static void GetAccountLocalAndRemoteEntityNames( IStorehouseProvider storehouseProvider, Locator accountLocator
			, out string localName, out string remoteName )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( accountLocator, "accountLocator" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the account speckle.
				Speckle accountSpeckle = transaction.LoadSpeckle( accountLocator, StorehouseRealms.Accounts, SheafLocks.Shared
					, null, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Account ) ) );
				ValidationUtils.ObjectHasContent( accountSpeckle, "accountSpeckle" );

				Locator localUpnFingerprint = accountSpeckle.GetLoadedSpeckleFilter<Locator>(
					Account.LocalUpnFingerprintFilterIdentifier );
				Locator remoteUpnFingerprint = accountSpeckle.GetLoadedSpeckleFilter<Locator>(
					Account.RemoteUpnFingerprintFilterIdentifier );

				// Load the tally speckle (for the remote entity name).
				SpeckleFilter tallyUpnFilter = new SpeckleFilter( Tally.RemoteUpnFingerprintFilterIdentifier
					, remoteUpnFingerprint );
				Speckle tallySpeckle = transaction.CollectFirstSpeckle( PortableTypes.FindAllDerived( typeof( Tally ) )
					, new SpeckleFilters( tallyUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared
					, null, new SpeckleFilterIdentifiers( Tally.EntityNameFilterIdentifier ) );
				ValidationUtils.ObjectHasContent( tallySpeckle, "tallySpeckle" );

				// Load the vault speckle (for the local entity name).
				SpeckleFilter vaultUpnFilter = new SpeckleFilter( Vault.LocalUpnFingerprintFilterIdentifier
					, localUpnFingerprint );
				Speckle vaultSpeckle = transaction.CollectFirstSpeckle( PortableTypes.FindAllDerived( typeof( Vault ) )
					, new SpeckleFilters( vaultUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared
					, null, new SpeckleFilterIdentifiers( Vault.EntityNameFilterIdentifier ) );
				ValidationUtils.ObjectHasContent( vaultSpeckle, "vaultSpeckle" );

				localName = vaultSpeckle.GetLoadedSpeckleFilter<string>( Vault.EntityNameFilterIdentifier );
				remoteName = tallySpeckle.GetLoadedSpeckleFilter<string>( Tally.EntityNameFilterIdentifier );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				if( storehouseProvider.IsTimeoutException( ex ) )
				{
					throw new Vela.Kernel.TimeoutException( "The account or one of its dependencies is being processed by a"
						+ " background operation. Try again later.", ex );
				}

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Gets the textual UPN of the account creation request (<see cref="CreateAccountAnonymousActionRequest"/>) with the
		/// specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="requestLocator">The locator of the account creation request whose textual UPN to get.</param>
		/// <returns>Returns the textual UPN of the account creation request.</returns>
		public static string GetAccountRequestRecipientTextualUpn( IStorehouseProvider storehouseProvider
			, Locator requestLocator )
		{
			Speckle requestSpeckle = null;

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the account creation request speckle.
				requestSpeckle = transaction.LoadSpeckle( requestLocator, StorehouseRealms.ActionRequestsToSend
					, SheafLocks.Shared, null
					, Sheaf.CollectSpeckleFilterIdentifiers( typeof( CreateAccountAnonymousActionRequest ) ) );
				ValidationUtils.ObjectHasContent( requestSpeckle, "requestSpeckle" );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			return requestSpeckle.GetLoadedSpeckleFilter<string>( ActionRequest.RecipientTextualUpnFilterIdentifier );
		}

		/// <summary>
		/// Loads the account speckles from the storehouse.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <returns>Returns the loaded account speckles.</returns>
		/// <remarks>
		/// <para>The returned account speckles are a union of the speckles of real accounts (<see cref="Account"/>) and of
		/// account creation requests (<see cref="CreateAccountAnonymousActionRequest"/>).</para>
		/// </remarks>
		public static AccountOrRequestSpeckles CollectAccountAndRequestSpeckles( IStorehouseProvider storehouseProvider )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the account speckles. Note that the local and remote names are not yet available, so the list can't be
				// sorted by them.
				Speckles accountSpeckles = transaction.CollectSpeckles( PortableTypes.FindAllDerived( typeof( Account ) ), null
					, StorehouseRealms.Accounts, null, null, typeof( AccountOrRequestSpeckle )
					, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Account ) ) ) as Speckles;

				// Load the tally speckles.
				SpeckleFilterIdentifiers tallyLoadFilterIdentifiers = new SpeckleFilterIdentifiers(
					Tally.RemoteUpnFingerprintFilterIdentifier, Tally.EntityNameFilterIdentifier
					, Tally.ApprovalFilterIdentifier );
				Speckles tallySpeckles = transaction.CollectSpeckles( PortableTypes.FindAllDerived( typeof( Tally ) ), null
					, StorehouseRealms.ProcessingIdentityContainers, null, null, null, tallyLoadFilterIdentifiers ) as Speckles;

				// Load the vault speckles.
				SpeckleFilterIdentifiers vaultLoadFilterIdentifiers = new SpeckleFilterIdentifiers(
					Vault.LocalUpnFingerprintFilterIdentifier, Vault.EntityNameFilterIdentifier );
				Speckles vaultSpeckles = transaction.CollectSpeckles( PortableTypes.FindAllDerived( typeof( Vault ) ), null
					, StorehouseRealms.ProcessingIdentityContainers, null, null, null, vaultLoadFilterIdentifiers ) as Speckles;

				// Collect the speckles keyed by the initiator UPN fingerprint.
				Dictionary<Locator, Speckle> seekableTallySpeckles = tallySpeckles.GetDictionaryByFilterIdentifier(
					Tally.RemoteUpnFingerprintFilterIdentifier );
				Dictionary<Locator, Speckle> seekableVaultSpeckles = vaultSpeckles.GetDictionaryByFilterIdentifier(
					Vault.LocalUpnFingerprintFilterIdentifier );

				AccountOrRequestSpeckles sortedAccountSpeckles = new AccountOrRequestSpeckles();

				#region Load the existing account

				// Set the information.
				foreach( AccountOrRequestSpeckle accountSpeckle in accountSpeckles )
				{
					Locator vaultUpnFingerprint = accountSpeckle.GetLoadedSpeckleFilter<Locator>(
						Account.LocalUpnFingerprintFilterIdentifier );
					Locator tallyUpnFingerprint = accountSpeckle.GetLoadedSpeckleFilter<Locator>(
						Account.RemoteUpnFingerprintFilterIdentifier );

					Speckle vaultSpeckle = seekableVaultSpeckles[ vaultUpnFingerprint ];
					Speckle tallySpeckle = seekableTallySpeckles[ tallyUpnFingerprint ];

					// Set the information. The progress is not required because the account was already created.
					accountSpeckle.LocalEntityName = vaultSpeckle.GetLoadedSpeckleFilter<string>(
						Vault.EntityNameFilterIdentifier );
					accountSpeckle.RemoteEntityName = tallySpeckle.GetLoadedSpeckleFilter<string>(
						Tally.EntityNameFilterIdentifier );
					accountSpeckle.Approval = tallySpeckle.GetLoadedSpeckleFilter<TallyApprovals>(
						Tally.ApprovalFilterIdentifier );

					// The speckle key has changed due to the new names, after being collected from the storehouse, so put the
					// speckle in a properly sorted list.
					sortedAccountSpeckles.Add( accountSpeckle );
				}

				#endregion Load the existing account

				#region Load the account creation requests

				// Load the speckles for account creation.
				Speckles createAccountSpeckles = transaction.CollectSpeckles(
					PortableTypes.FindAllDerived( typeof( CreateAccountAnonymousActionRequest ) ), null
					, StorehouseRealms.ActionRequestsToSend, null, null, typeof( AccountOrRequestSpeckle )
					, Sheaf.CollectSpeckleFilterIdentifiers( typeof( CreateAccountAnonymousActionRequest ) ) ) as Speckles;

				// Set the information.
				foreach( AccountOrRequestSpeckle accountSpeckle in createAccountSpeckles )
				{
					// Load the progress sheaf.
					IDeserializer deserializer;
					SpeckleFilter progressFilter = new SpeckleFilter(
						ActionClientProgress.ReferredActionRequestLocatorFilterIdentifier, accountSpeckle.StorehouseKey );
					CreateAccountClientProgress progress = transaction.CollectFirstSheaf(
						PortableTypes.FindAllDerived( typeof( CreateAccountClientProgress ) )
						, new SpeckleFilters( progressFilter ), StorehouseRealms.ActionProgresses, SheafLocks.Shared
						, out deserializer ) as CreateAccountClientProgress;
					ValidationUtils.ObjectHasContent( progress, "progress" );

					Locator vaultUpnFingerprint = accountSpeckle.GetLoadedSpeckleFilter<Locator>(
						ActionRequest.InitiatorUpnFingerprintFilterIdentifier );

					Speckle vaultSpeckle = seekableVaultSpeckles[ vaultUpnFingerprint ];

					// Set the information.
					accountSpeckle.LastProcessingAttempt = progress.LastProcessingAttempt;
					accountSpeckle.LocalEntityName = vaultSpeckle.GetLoadedSpeckleFilter<string>(
						Vault.EntityNameFilterIdentifier );
					accountSpeckle.RemoteEntityName = accountSpeckle.GetLoadedSpeckleFilter<string>(
						ActionRequest.RecipientTextualUpnFilterIdentifier );
					accountSpeckle.Approval = progress.Approval;

					// The speckle key has changed due to the new names, after being collected from the storehouse, so put the
					// speckle in a properly sorted list.
					sortedAccountSpeckles.Add( accountSpeckle );
				}

				#endregion Load the account creation requests

				// Commit the transaction.
				transaction.Commit();

				return sortedAccountSpeckles;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the speckles of the accounts associated to the specified remote UPN.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="remoteUpnFingerprint">The UPN fingerprint of the remote peer for which to load the associated
		/// accounts.</param>
		/// <returns>Returns the loaded account speckles.</returns>
		/// <remarks>
		/// <para>The returned account speckles may be either the speckles of real accounts (<see cref="Account"/>) or of
		/// account creation requests (<see cref="CreateAccountAnonymousActionRequest"/>).</para>
		/// </remarks>
		public static AccountOrRequestSpeckles CollectAccountAndRequestSpeckles( IStorehouseProvider storehouseProvider
			, Locator remoteUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( remoteUpnFingerprint, "remoteUpnFingerprint" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				AccountOrRequestSpeckles accountSpeckles = new AccountOrRequestSpeckles();

				#region Load the speckles of accounts

				SpeckleFilter accountRemoteUpnFilter = new SpeckleFilter( Account.RemoteUpnFingerprintFilterIdentifier
					, remoteUpnFingerprint );
				IKernelObservableList accountKeys = transaction.CollectKeys( new PortableTypes( typeof( Account ) )
					, new SpeckleFilters( accountRemoteUpnFilter ), StorehouseRealms.Accounts, null, null, null );

				foreach( Locator accountKey in accountKeys )
				{
					// Load the account speckle.
					AccountOrRequestSpeckle accountSpeckle = LoadAccountOrRequestSpeckle( transaction, accountKey );
					accountSpeckles.Add( accountSpeckle );
				}

				#endregion Load the speckles of accounts

				#region Load the speckles of account creation requests

				SpeckleFilter requestRemoteUpnFilter = new SpeckleFilter( ActionRequest.RecipientUpnFingerprintFilterIdentifier
					, remoteUpnFingerprint );
				IKernelObservableList requestKeys = transaction.CollectKeys(
					new PortableTypes( typeof( CreateAccountAnonymousActionRequest ) )
					, new SpeckleFilters( requestRemoteUpnFilter ), StorehouseRealms.ActionRequestsToSend, null, null, null );

				foreach( Locator requestKey in requestKeys )
				{
					// Load the account speckle.
					AccountOrRequestSpeckle requestSpeckle = LoadAccountOrRequestSpeckle( transaction, requestKey );
					accountSpeckles.Add( requestSpeckle );
				}

				#endregion Load the speckles of account creation requests

				// Commit the transaction.
				transaction.Commit();

				return accountSpeckles;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the account speckle with the specified storehouse key.
		/// </summary>
		/// <param name="transaction">The transaction with which to load the speckle.</param>
		/// <param name="accountLocator">The storehouse key of the account speckle to load.</param>
		/// <returns>Returns the loaded account speckle.</returns>
		/// <remarks>
		/// <para>The returned account speckle may be either the speckle of a real account (<see cref="Account"/>) or of a
		/// account creation request (<see cref="CreateAccountAnonymousActionRequest"/>).</para>
		/// </remarks>
		public static AccountOrRequestSpeckle LoadAccountOrRequestSpeckle( IStorehouseTransaction transaction
			, Locator accountLocator )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( accountLocator, "accountLocator" );

			Locator vaultUpnFingerprint;

			// Load the account speckle.
			AccountOrRequestSpeckle accountSpeckle = transaction.LoadSpeckle( accountLocator, StorehouseRealms.Accounts
				, SheafLocks.Shared, typeof( AccountOrRequestSpeckle )
				, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Account ) ) ) as AccountOrRequestSpeckle;

			bool isRealAccount = accountSpeckle != null;
			CreateAccountClientProgress progress = null;

			if( isRealAccount )
			{
				// A real account is being referenced.

				vaultUpnFingerprint = accountSpeckle.GetLoadedSpeckleFilter<Locator>(
					Account.LocalUpnFingerprintFilterIdentifier );
			}
			else
			{
				// An account creation request is being referenced.

				accountSpeckle = transaction.LoadSpeckle( accountLocator, StorehouseRealms.ActionRequestsToSend
						, SheafLocks.Shared, typeof( AccountOrRequestSpeckle )
						, Sheaf.CollectSpeckleFilterIdentifiers( typeof( CreateAccountAnonymousActionRequest ) ) )
					as AccountOrRequestSpeckle;
				ValidationUtils.ObjectHasContent( accountSpeckle, "accountSpeckle" );

				vaultUpnFingerprint = accountSpeckle.GetLoadedSpeckleFilter<Locator>(
					CreateAccountAnonymousActionRequest.InitiatorUpnFingerprintFilterIdentifier );

				// Load the progress sheaf.
				IDeserializer deserializer;
				SpeckleFilter progressFilter = new SpeckleFilter(
					ActionClientProgress.ReferredActionRequestLocatorFilterIdentifier, accountSpeckle.StorehouseKey );
				progress = transaction.CollectFirstSheaf(
					PortableTypes.FindAllDerived( typeof( CreateAccountClientProgress ) )
					, new SpeckleFilters( progressFilter ), StorehouseRealms.ActionProgresses, SheafLocks.Shared
					, out deserializer ) as CreateAccountClientProgress;
				ValidationUtils.ObjectHasContent( progress, "progress" );
			}

			Speckle vaultSpeckle = LoadVaultSpeckleByUpnFingerprint( transaction, vaultUpnFingerprint );
			ValidationUtils.ObjectHasContent( vaultSpeckle, "vaultSpeckle" );

			#region Set the information

			accountSpeckle.LocalEntityName = vaultSpeckle.GetLoadedSpeckleFilter<string>( Vault.EntityNameFilterIdentifier );

			if( isRealAccount )
			{
				Locator remoteUpnFingerprint = accountSpeckle.GetLoadedSpeckleFilter<Locator>(
					Account.RemoteUpnFingerprintFilterIdentifier );
				Speckle tallySpeckle = LoadTallySpeckleByUpnFingerprint( transaction, remoteUpnFingerprint );
				ValidationUtils.ObjectHasContent( tallySpeckle, "tallySpeckle" );

				accountSpeckle.RemoteEntityName = tallySpeckle.GetLoadedSpeckleFilter<string>(
					Tally.EntityNameFilterIdentifier );
				accountSpeckle.Approval = tallySpeckle.GetLoadedSpeckleFilter<TallyApprovals>( Tally.ApprovalFilterIdentifier );
			}
			else
			{
				accountSpeckle.LastProcessingAttempt = progress.LastProcessingAttempt;
				accountSpeckle.RemoteEntityName = accountSpeckle.GetLoadedSpeckleFilter<string>(
					ActionRequest.RecipientTextualUpnFilterIdentifier );
				accountSpeckle.Approval = progress.Approval;
			}

			#endregion Set the information

			return accountSpeckle;
		}

		/// <summary>
		/// Loads the account speckle with the specified storehouse key.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="accountLocator">The storehouse key of the account speckle to load.</param>
		/// <returns>Returns the loaded account speckle.</returns>
		/// <remarks>
		/// <para>The returned account speckle may be either the speckle of a real account (<see cref="Account"/>) or of a
		/// account creation request (<see cref="CreateAccountAnonymousActionRequest"/>).</para>
		/// </remarks>
		public static AccountOrRequestSpeckle LoadAccountOrRequestSpeckle( IStorehouseProvider storehouseProvider
			, Locator accountLocator )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( accountLocator, "accountLocator" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the account speckle.
				AccountOrRequestSpeckle accountSpeckle = LoadAccountOrRequestSpeckle( transaction, accountLocator );

				// Commit the transaction.
				transaction.Commit();

				return accountSpeckle;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the account creation request (<see cref="CreateAccountAnonymousActionRequest"/>) with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="requestLocator">The locator of the account creation request to load.</param>
		/// <param name="accountRequestProgress">Is set with the progress of the loaded account creation request.</param>
		/// <returns>Returns the loaded account creation request.</returns>
		public static CreateAccountAnonymousActionRequest LoadAccountRequestAndProgress( IStorehouseProvider storehouseProvider
			, Locator requestLocator, out CreateAccountClientProgress accountRequestProgress )
		{
			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the account creation request.
				IDeserializer deserializer;
				CreateAccountAnonymousActionRequest accountRequest = transaction.LoadSheaf( requestLocator
					, StorehouseRealms.ActionRequestsToSend, typeof( CreateAccountAnonymousActionRequest ), SheafLocks.Shared
					, out deserializer ) as CreateAccountAnonymousActionRequest;
				ValidationUtils.ObjectHasContent( accountRequest, "accountRequest" );

				#region Load the progress of the account creation request

				SpeckleFilter initiatorUpnFingerprintFilter = new SpeckleFilter(
					ActionProgress.InitiatorUpnFingerprintFilterIdentifier, accountRequest.InitiatorUpnFingerprint );
				SpeckleFilter recipientUpnFingerprintFilter = new SpeckleFilter(
					ActionProgress.RecipientUpnFingerprintFilterIdentifier, accountRequest.RecipientUpnFingerprint );
				SpeckleFilter typeNameFilter = new SpeckleFilter( ActionProgress.ReferredActionRequestTypeNameFilterIdentifier
					, typeof( CreateAccountAnonymousActionRequest ).FullName );

				accountRequestProgress = transaction.CollectFirstSheaf(
					PortableTypes.FindAllDerived( typeof( CreateAccountClientProgress ) )
					, new SpeckleFilters( initiatorUpnFingerprintFilter, recipientUpnFingerprintFilter, typeNameFilter )
					, StorehouseRealms.ActionProgresses, SheafLocks.Shared, out deserializer ) as CreateAccountClientProgress;
				ValidationUtils.ObjectHasContent( accountRequestProgress, "accountRequestProgress" );

				#endregion Load the progress of the account creation request

				// Commit the transaction.
				transaction.Commit();

				return accountRequest;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the speckle of the account associated to the specified local and remote UPN fingerprints.
		/// </summary>
		/// <param name="transaction">Transaction to use.</param>
		/// <param name="localUpnFingerprint">The UPN fingerprint of the local peer for which to load the associated
		/// account.</param>
		/// <param name="remoteUpnFingerprint">The UPN fingerprint of the remote peer for which to load the associated
		/// account.</param>
		/// <param name="loadFilterIdentifiers">The speckle filter identifiers whose values to load from the matching speckles.
		/// This may be null or empty.</param>
		/// <returns>Returns the loaded account speckle, or null if none is found.</returns>
		public static Speckle LoadAccountSpeckleByUpnFingerprints( IStorehouseTransaction transaction
			, Locator localUpnFingerprint, Locator remoteUpnFingerprint, SpeckleFilterIdentifiers loadFilterIdentifiers )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( localUpnFingerprint, "localUpnFingerprint" );
			ValidationUtils.ObjectHasContent( remoteUpnFingerprint, "remoteUpnFingerprint" );

			SpeckleFilter localUpnFilter = new SpeckleFilter( Account.LocalUpnFingerprintFilterIdentifier
				, localUpnFingerprint );
			SpeckleFilter remoteUpnFilter = new SpeckleFilter( Account.RemoteUpnFingerprintFilterIdentifier
				, remoteUpnFingerprint );

			return transaction.CollectFirstSpeckle( new PortableTypes( typeof( Account ) )
				, new SpeckleFilters( localUpnFilter, remoteUpnFilter ), StorehouseRealms.Accounts, SheafLocks.Shared, null
				, loadFilterIdentifiers );
		}

		/// <summary>
		/// Loads the account with the specified account locator, from the storehouse.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="accountLocator">The locator of the account to load.</param>
		/// <returns>Returns the loaded account, or null if none is found.</returns>
		public static Account LoadAccount( IStorehouseProvider storehouseProvider, Locator accountLocator )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( accountLocator, "accountLocator" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				IDeserializer deserializer;
				Account account = transaction.LoadSheaf( accountLocator, StorehouseRealms.Accounts, typeof( Account )
					, SheafLocks.Shared, out deserializer ) as Account;

				// Commit the transaction.
				transaction.Commit();

				return account;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Creates or updates an account using a presenter file.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="vaultLocator">The locator of the vault to which to link the account.</param>
		/// <returns>Returns the instance locator of the created account, or null if none was created.</returns>
		public static Locator AddAccountWithFile( IStorehouseProvider storehouseProvider, Locator vaultLocator )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( vaultLocator, "vaultLocator" );

			OpenFileDialog ofd = new OpenFileDialog();
			ofd.Title = "Load Presenter";
			ofd.AddExtension = true;
			ofd.DefaultExt = FileExtensions.Presenter;
			ofd.Filter = FileExtensions.PresenterFilter;
			ofd.CheckFileExists = true;
			ofd.CheckPathExists = true;

			if( !ControlsUtils.ShowDialog( ofd ) )
				return null;

			// Load the presenter.
			IDeserializer deserializer;
			Presenter presenter = SerializationUtils.LoadFromFile( ofd.FileName, typeof( Presenter ), out deserializer )
				as Presenter;
			if( presenter == null )
			{
				throw new ValidationException( "Failed to load the presenter.", null, null
					, new ExceptionDetail( "File name", ofd.FileName ) );
			}

			return ActionProcessorStorehouseUtils.AddAccountFromPresenter( storehouseProvider, vaultLocator, presenter, null
				, PeerDocumentsConsts.CreateAccountApprovalMin, null );
		}

		/// <summary>
		/// Deletes the account with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="accountLocator">The locator of the account to delete.</param>
		/// <param name="serviceManager">Action service manager which is used to stop processing action requests which have a
		/// multiframe action response.</param>
		/// <param name="clientManager">Action client manager which is used to stop processing action requests which have a
		/// multiframe action response.</param>
		/// <param name="lop">Long operation progress. This may be null.</param>
		public static void DeleteAccount( IStorehouseProvider storehouseProvider, Locator accountLocator
			, IActionServiceManager serviceManager, IActionClientManager clientManager, LongOperationProgress lop )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( accountLocator, "accountLocator" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				if( lop != null )
					lop.Progress( 0 );

				DeleteAccountSheafs( transaction, accountLocator, serviceManager, clientManager, lop );

				if( lop != null )
					lop.Progress( 1 );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				if( storehouseProvider.IsTimeoutException( ex ) )
				{
					throw new Vela.Kernel.TimeoutException( "The account or one of its dependencies is being processed by a"
						+ " background operation. Try again later.", ex );
				}

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Deletes the account creation request (<see cref="CreateAccountAnonymousActionRequest"/>) with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="requestLocator">The locator of the account creation request to delete.</param>
		/// <param name="serviceManager">Action service manager which is used to stop processing action requests which have a
		/// multiframe action response.</param>
		/// <param name="clientManager">Action client manager which is used to stop processing action requests which have a
		/// multiframe action response.</param>
		public static void DeleteAccountRequest( IStorehouseProvider storehouseProvider, Locator requestLocator
			, IActionServiceManager serviceManager, IActionClientManager clientManager )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( requestLocator, "requestLocator" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Stop the action processing in the action service and client.
				Locators requestLocators = new Locators( requestLocator );
				serviceManager.StopRequests( requestLocators );
				clientManager.StopRequests( requestLocators );

				// Delete the progress sheaf.
				SpeckleFilter referredLocatorFilter = new SpeckleFilter(
					ActionClientProgress.ReferredActionRequestLocatorFilterIdentifier, requestLocator );
				transaction.DeleteFirstSheaf( PortableTypes.FindAllDerived( typeof( ActionClientProgress ) )
					, new SpeckleFilters( referredLocatorFilter ), StorehouseRealms.ActionProgresses );

				// Delete the request sheaf.
				transaction.DeleteSheaf( requestLocator, StorehouseRealms.ActionRequestsToSend );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				if( storehouseProvider.IsTimeoutException( ex ) )
				{
					throw new Vela.Kernel.TimeoutException( "The account creation request or one of its dependencies is being"
						+ " processed by a background operation. Try again later.", ex );
				}

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Queues an action request for getting the presenter of the remote peer and creating an account for it.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="localIdentity">The private identity of the local peer for which to create an account.</param>
		/// <param name="remoteUpn">The UPN of the remote peer of the account to create.</param>
		/// <param name="tallyEntityName">The name of the identified entity, from the tally. This may be null or empty.</param>
		/// <param name="approval">The approval of the (tally of the) account to create.</param>
		/// <param name="alternativePeerAddress">The alternative peer address (of the remote peer). This may be null.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="hashSerializer">Hash serializer.</param>
		/// <returns>Returns the instance locator of the created action request.</returns>
		/// <remarks>
		/// <para>The action request is saved in the storehouse, it's not actually sent.</para>
		/// </remarks>
		public static Locator QueueCreateAccount( IStorehouseProvider storehouseProvider, PrivateIdentity localIdentity
			, Upn remoteUpn, string tallyEntityName, TallyApprovals approval, AlternativePeerAddress alternativePeerAddress
			, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg, IHashSerializer hashSerializer )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( localIdentity, "localIdentity" );
			ValidationUtils.ObjectHasContent( remoteUpn, "remoteUpn" );
			ValidationUtils.Text( tallyEntityName, -2, -1, TextValidations.EntityName, "tallyEntityName" );
			ValidationUtils.ObjectHasContent( macAlg, "macAlg" );
			ValidationUtils.ObjectHasContent( asymmSigningPaddingAlg, "asymmSigningPaddingAlg" );
			ValidationUtils.ObjectHasContent( hashSerializer, "hashSerializer" );

			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Check if the UPN is local.
				if( ActionProcessorStorehouseUtils.HasVaultWithUpn( transaction, remoteUpn ) )
				{
					throw new ValidationException( "Can't queue an action request to create an account because a vault with the"
						+ " same UPN exists.", null, null, new ExceptionDetail( "Remote UPN", remoteUpn.ToString() ) );
				}

				#region Check duplicate action request

				// Check if there is an action request to create an account for the same local UPN and the same remote UPN.

				SpeckleFilter initiatorUpnFilter = new SpeckleFilter( ActionRequest.InitiatorUpnFingerprintFilterIdentifier
					, localIdentity.Upn.Fingerprint );
				SpeckleFilter recipientUpnFilter = new SpeckleFilter( ActionRequest.RecipientUpnFingerprintFilterIdentifier
					, remoteUpn.Fingerprint );

				IDeserializer deserializer;
				ActionRequest existingActionRequest = transaction.CollectFirstSheaf(
					PortableTypes.FindAllDerived( typeof( CreateAccountAnonymousActionRequest ) )
					, new SpeckleFilters( initiatorUpnFilter, recipientUpnFilter ), StorehouseRealms.ActionRequestsToSend
					, SheafLocks.Shared, out deserializer ) as ActionRequest;

				// Accept the existing action request even if it's expired (because it's already stored locally).
				if( existingActionRequest != null )
				{
					throw new ValidationException( "Won't queue an action request to create an account because an action request"
							+ " for the same local UPN fingerprint and remote UPN exists.", null, null
						, new ExceptionDetail( "Local UPN fingerprint", localIdentity.Upn.Fingerprint )
						, new ExceptionDetail( "Remote UPN", remoteUpn.ToString() )
						, new ExceptionDetail( "Action request type", existingActionRequest.GetType().FullName ) );
				}

				#endregion Check duplicate action request

				#region Create the action request

				// Create the action request.
				SolitaryIdentityPublisher initiatorPublisher = CoreDocumentsUtils.GetSolitaryIdentity( localIdentity );
				CreateAccountAnonymousActionRequest request = new CreateAccountAnonymousActionRequest(
					CryptoUtils.GetLocator(), macAlg, asymmSigningPaddingAlg, hashSerializer, remoteUpn
					, PeerDocumentsConsts.ActionRequestValidity, 0, initiatorPublisher );

				// Sign the action request.
				request.Sign( localIdentity, null );

				// Save the action request.
				transaction.AddSheaf( request, StorehouseRealms.ActionRequestsToSend );

				#endregion Create the action request

				#region Create the action progress

				// Create the action progress.
				CreateAccountClientProgress progress = new CreateAccountClientProgress( CryptoUtils.GetLocator()
					, request.InitiatorUpnFingerprint, remoteUpn.Fingerprint, Locator.Dummy, request.GetType().FullName
					, request.InstanceLocator, request.DocumentExpirationTime, false, tallyEntityName, approval
					, alternativePeerAddress );

				// Save the action progress.
				transaction.AddSheaf( progress, StorehouseRealms.ActionProgresses );

				#endregion Create the action progress

				// Commit the transaction.
				transaction.Commit();

				return request.InstanceLocator;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		/// <summary>
		/// Queues an action request which pulls a remote presenter in order to update it on the local peer.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="accountLocator">The locator of the account to update.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="hashSerializer">Hash serializer.</param>
		/// <returns>Returns the instance locator of the created action request, or null if none was created.</returns>
		/// <remarks>
		/// <para>The action request is saved in the storehouse, it's not actually sent.</para>
		/// </remarks>
		public static Locator QueuePullUpdatePresenter( IStorehouseProvider storehouseProvider, Locator accountLocator
			, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg, IHashSerializer hashSerializer )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( accountLocator, "accountLocator" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				IDeserializer deserializer;

				Account account = transaction.LoadSheaf( accountLocator, StorehouseRealms.Accounts, typeof( Account )
					, SheafLocks.Shared, out deserializer ) as Account;
				ValidationUtils.ObjectHasContent( account, "account" );

				Vault vault = LoadVaultByUpnFingerprint( transaction, account.LocalUpnFingerprint );
				ValidationUtils.ObjectHasContent( vault, "vault" );

				Tally tally = LoadTallyByUpnFingerprint( transaction, account.RemoteUpnFingerprint );
				ValidationUtils.ObjectHasContent( tally, "tally" );

				Locator requestInstanceLocator = QueueCreateAccount( storehouseProvider, vault.ProcessingIdentity
					, account.RemoteUpn, null, tally.Approval, tally.AlternativePeerAddress, macAlg, asymmSigningPaddingAlg
					, hashSerializer );

				// Commit the transaction.
				transaction.Commit();

				return requestInstanceLocator;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Queues an action request which pushes an updated presenter in order to update it on a remote peer.
		/// </summary>
		/// <param name="transaction">Transaction to use.</param>
		/// <param name="vault">The vault of the initiator of the request.</param>
		/// <param name="remoteUpnFingerprint">The remote UPN fingerprint of the recipient of the request.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="hashSerializer">Hash serializer.</param>
		/// <param name="actionClientManager">Action client manager which is used to stop the processing of an existing update
		/// action request. Its content must be thread safe.</param>
		/// <returns>Returns the instance locator of the created action request, or null if the remote peer is blocked.</returns>
		public static Locator QueuePushUpdatePresenter( IStorehouseTransaction transaction, Vault vault
			, Locator remoteUpnFingerprint, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg
			, IHashSerializer hashSerializer, IActionClientManager actionClientManager )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( vault, "vault" );
			ValidationUtils.ObjectHasContent( remoteUpnFingerprint, "remoteUpnFingerprint" );
			ValidationUtils.ObjectHasContent( macAlg, "macAlg" );
			ValidationUtils.ObjectHasContent( asymmSigningPaddingAlg, "asymmSigningPaddingAlg" );
			ValidationUtils.ObjectHasContent( hashSerializer, "hashSerializer" );
			ValidationUtils.ObjectHasContent( actionClientManager, "actionClientManager" );

			IDeserializer deserializer;

			#region Create the action request

			// Load the tally.
			Tally tally = LoadTallyByUpnFingerprint( transaction, remoteUpnFingerprint );
			ValidationUtils.ObjectHasContent( tally, "tally" );

			if( PeerDocumentsUtils.IsBlockedPeer( tally.Approval ) )
				return null;

			Presenter presenter = ActionProcessorStorehouseUtils.GetPresenter( transaction, macAlg, asymmSigningPaddingAlg
				, vault, tally.Approval, hashSerializer );

			// Create the action request.
			PushUpdatePresenterActionRequest updateRequest = new PushUpdatePresenterActionRequest( CryptoUtils.GetLocator()
				, macAlg, asymmSigningPaddingAlg, hashSerializer, vault.LocalUpnFingerprint, tally.RemoteUpn
				, PeerDocumentsConsts.ActionRequestValidity, 0, presenter );

			// Sign the action request.
			updateRequest.Sign( vault.ProcessingIdentity, null );

			// Create the action progress.
			ActionClientProgress updateProgress = new ActionClientProgress( CryptoUtils.GetLocator()
				, vault.LocalUpnFingerprint, remoteUpnFingerprint, Locator.Dummy, updateRequest.GetType().FullName
				, updateRequest.InstanceLocator, updateRequest.DocumentExpirationTime, false );

			#endregion Create the action request

			#region Abort and delete any existing update request

			// Abort any existing update request for the same remote peer, as a preparation to delete the request before
			// adding the new one.

			// Load the existing update request.
			SpeckleFilter updateRequestRemoteUpnFilter = new SpeckleFilter(
				ActionRequest.RecipientUpnFingerprintFilterIdentifier, remoteUpnFingerprint );
			PushUpdatePresenterActionRequest existingUpdateRequest = transaction.CollectFirstSheaf(
				PortableTypes.FindAllDerived( typeof( PushUpdatePresenterActionRequest ) )
				, new SpeckleFilters( updateRequestRemoteUpnFilter ), StorehouseRealms.ActionRequestsToSend
				, SheafLocks.Exclusive, out deserializer ) as PushUpdatePresenterActionRequest;

			// Stop the action processing in the action client.
			if( existingUpdateRequest != null )
				actionClientManager.StopRequests( new Locators( existingUpdateRequest.InstanceLocator ) );

			// Delete the existing update request.
			if( existingUpdateRequest != null )
			{
				// Delete the progress sheaf.
				SpeckleFilter referredLocatorFilter = new SpeckleFilter(
					ActionClientProgress.ReferredActionRequestLocatorFilterIdentifier
					, existingUpdateRequest.InstanceLocator );
				transaction.DeleteFirstSheaf( PortableTypes.FindAllDerived( typeof( ActionClientProgress ) )
					, new SpeckleFilters( referredLocatorFilter ), StorehouseRealms.ActionProgresses );

				// Delete the request sheaf.
				transaction.DeleteSheaf( existingUpdateRequest.InstanceLocator, StorehouseRealms.ActionRequestsToSend );
			}

			#endregion Abort and delete any existing update request

			// Save the action request.
			transaction.AddSheaf( updateRequest, StorehouseRealms.ActionRequestsToSend );

			// Save the action progress.
			transaction.AddSheaf( updateProgress, StorehouseRealms.ActionProgresses );

			return updateRequest.InstanceLocator;
		}

		/// <summary>
		/// Counts the number of post speckles which have the specified flag.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="flag">The posts referred by the collected speckles must have this flag.</param>
		/// <param name="postAncestorType">The ancestor type of the posts to count. Only the post which have a type derived from
		/// this one will be counted.</param>
		/// <returns>Returns the number of post speckles which have the specified flag.</returns>
		public static int CountPostSpeckles( IStorehouseProvider storehouseProvider, string flag, Type postAncestorType )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.TextHasContent( flag, "flag" );
			ValidationUtils.ObjectHasContent( postAncestorType, "postAncestorType" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				SpeckleFilter flagFilter = new SpeckleFilter( Post.FlagFilterIdentifier, flag );
				int count = transaction.CountSpeckles( PortableTypes.FindAllDerived( postAncestorType )
					, new SpeckleFilters( flagFilter ), StorehouseRealms.Posts );

				// Commit the transaction.
				transaction.Commit();

				return count;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Counts the number of transient post speckles.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="postAncestorType">The ancestor type of the posts to count. Only the post which have a type derived from
		/// this one will be counted.</param>
		/// <returns>Returns the number of transient post speckles.</returns>
		public static int CountTransientPostSpeckles( IStorehouseProvider storehouseProvider, Type postAncestorType )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( postAncestorType, "postAncestorType" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				SpeckleFilter isTransientFilter = new SpeckleFilter( Post.IsTransientFilterIdentifier, true );
				int count = transaction.CountSpeckles( PortableTypes.FindAllDerived( postAncestorType )
					, new SpeckleFilters( isTransientFilter ), StorehouseRealms.Posts );

				// Commit the transaction.
				transaction.Commit();

				return count;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Updates the flag of the post with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="postLocator">The locator of the post whose flag to update.</param>
		/// <param name="flag">New flag.</param>
		/// <remarks>
		/// <para>The post will not be updated if its flag is already "flag".</para>
		/// </remarks>
		public static void UpdatePostFlag( IStorehouseProvider storehouseProvider, Locator postLocator, string flag )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( postLocator, "postLocator" );
			ValidationUtils.TextHasContent( flag, "flag" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the post.
				IDeserializer deserializer;
				Post post = transaction.LoadSheaf( postLocator, StorehouseRealms.Posts, typeof( Post ), SheafLocks.Update
					, out deserializer ) as Post;

				if( Post.EqualFlags( post.Flag, flag ) )
				{
					// Commit the transaction.
					transaction.Commit();

					return;
				}

				// Update the post flag.
				post.Flag = flag;

				transaction.UpdateSheaf( post, StorehouseRealms.Posts );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Updates the flags of the posts with the specified flag.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="flagToUpdate">The flag to update.</param>
		/// <param name="newFlag">The new flag.</param>
		/// <param name="postAncestorType">The ancestor type of the posts whose flags to update. Only the post which have a
		/// type derived from this one will have their flags updated.</param>
		/// <param name="lop">Long operation progress. This may be null.</param>
		/// <param name="updated">Is set with the number of posts whose flags were updated.</param>
		/// <remarks>
		/// <para>If "flagToUpdate" is equal with "newFlag", nothing happens.</para>
		/// 
		/// <para>The posts are processed in batches, each with its own transaction, so an abort will rollback only the changes
		/// from the current batch.</para>
		/// </remarks>
		public static void UpdatePostsFlag( IStorehouseProvider storehouseProvider, string flagToUpdate, string newFlag
			, Type postAncestorType, LongOperationProgress lop, out int updated )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.TextHasContent( flagToUpdate, "flagToUpdate" );
			ValidationUtils.TextHasContent( newFlag, "newFlag" );
			ValidationUtils.ObjectHasContent( postAncestorType, "postAncestorType" );

			updated = 0;

			if( Post.EqualFlags( flagToUpdate, newFlag ) )
				return;

			// Count the posts.
			int postsCount = CountPostSpeckles( storehouseProvider, flagToUpdate, postAncestorType );
			if( postsCount <= 0 )
				return;

			if( lop != null )
				lop.Progress( 0 );

			PortableTypes postTypes = PortableTypes.FindAllDerived( postAncestorType );

			// The flag speckle filter.
			SpeckleFilter flagFilter = new SpeckleFilter( Post.FlagFilterIdentifier, flagToUpdate );

			Locators postLocators = null;

			do
			{
				// Create a transaction for each batch.
				IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

				try
				{
					// Load the post locators.
					postLocators = transaction.CollectKeys( postTypes, new SpeckleFilters( flagFilter )
						, StorehouseRealms.Posts, new CollectPaging( true, 0, UpdatePostsFlagBatchSize ), null
						, null ) as Locators;

					int updatedPerBatch = 0;
					foreach( Locator postLocator in postLocators )
					{
						// Load the post.
						IDeserializer deserializer;
						Post post = transaction.LoadSheaf( postLocator, StorehouseRealms.Posts, typeof( Post ), SheafLocks.Update
							, out deserializer ) as Post;

						// Update the post flag.
						post.Flag = newFlag;

						transaction.UpdateSheaf( post, StorehouseRealms.Posts );

						updatedPerBatch++;

						if( lop != null )
							lop.Progress( (double)(updated + updatedPerBatch) / postsCount );
					}

					// Commit the transaction.
					transaction.Commit();

					updated += updatedPerBatch;
				}
				catch( Exception ex )
				{
					// Rolls back the transaction.
					transaction.Rollback();

					throw ExceptionUtils.PreserveStackTrace( ex );
				}
			}
			while( ListUtils.HasContent( postLocators ) );

			if( lop != null )
				lop.Progress( 1 );
		}

		/// <summary>
		/// Counts the number of letter speckles which have the specified flag.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="flag">The letters referred by the collected speckles must have this flag.</param>
		/// <returns>Returns the number of letter speckles which have the specified flag.</returns>
		public static int CountLetterSpeckles( IStorehouseProvider storehouseProvider, string flag )
		{
			return CountPostSpeckles( storehouseProvider, flag, typeof( Letter ) );
		}

		/// <summary>
		/// Counts the number of transient letter speckles.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <returns>Returns the number of transient letter speckles.</returns>
		public static int CountTransientLetterSpeckles( IStorehouseProvider storehouseProvider )
		{
			return CountTransientPostSpeckles( storehouseProvider, typeof( Letter ) );
		}

		/// <summary>
		/// Checks if the storehouse with the specified provider contains at least 1 letter with the
		/// <see cref="PostFlags.Main"/> flag.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="singleAccountLocator">Is set with the instance locator of the only account from the storehouse with the
		/// specified provider. This is null if there is either no account or there are at least 2 accounts.</param>
		/// <returns>Returns true if the storehouse contains at least 1 letter with the <see cref="PostFlags.Main"/> flag.
		/// </returns>
		public static bool HasLetterWithMainFlag( IStorehouseProvider storehouseProvider, out Locator singleAccountLocator )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				SpeckleFilter flagFilter = new SpeckleFilter( Post.FlagFilterIdentifier, PostFlags.Main.ToString() );
				Speckle speckle = transaction.CollectFirstSpeckle( PortableTypes.FindAllDerived( typeof( Letter ) )
					, new SpeckleFilters( flagFilter ), StorehouseRealms.Posts, SheafLocks.Shared, null, null );

				// Collect the locators of the first 2 accounts.
				CollectPaging paging = new CollectPaging( true, 0, 2 );
				Locators accountLocators = transaction.CollectKeys( PortableTypes.FindAllDerived( typeof( Account ) )
					, null, StorehouseRealms.Accounts, paging, null, null ) as Locators;

				singleAccountLocator = ((accountLocators != null) && (accountLocators.Count == 1))
					? accountLocators.GetFirstItem() : null;

				// Commit the transaction.
				transaction.Commit();

				return speckle != null;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the progress of the specified letter speckle, and sets its information.
		/// </summary>
		/// <param name="letterSpeckle">Letter speckle whose progress to load and whose information to set.</param>
		/// <param name="transaction">The transaction with which to load the documents.</param>
		public static void LoadLetterSpeckleProgressAndSetInformation( LetterSpeckle letterSpeckle
			, IStorehouseTransaction transaction )
		{
			ValidationUtils.ObjectHasContent( letterSpeckle, "letterSpeckle" );
			ValidationUtils.ObjectHasContent( transaction, "transaction" );

			IDeserializer deserializer;

			#region Load the progress

			SpeckleFilter referredSheafFilter = new SpeckleFilter( ActionProgress.ReferredSheafLocatorFilterIdentifier
				, letterSpeckle.StorehouseKey );
			SpeckleFilter typeNameFilter = new SpeckleFilter( ActionProgress.ReferredActionRequestTypeNameFilterIdentifier
				, typeof( SendLetterActionRequest ).FullName );

			ActionProgress progress = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( ActionProgress ) )
				, new SpeckleFilters( referredSheafFilter, typeNameFilter ), StorehouseRealms.ActionProgresses
				, SheafLocks.Shared, out deserializer ) as ActionProgress;
			ValidationUtils.ObjectHasContent( progress, "progress" );

			ActionClientProgress clientProgress = progress as ActionClientProgress;

			ActionProgress attachedFilesProgress = null;
			if( (progress.ProcessingState != ActionProcessingStates.RecipientImported) && letterSpeckle.HasAttachedFiles )
			{
				// Load the progress sheaf of the attached files only when the letter was sent (and so the sender was a client
				// from the point of view of the main progress), because that's when the sender has acted as a service and
				// transferred the attached files.
				if( clientProgress != null )
				{
					typeNameFilter = new SpeckleFilter( ActionProgress.ReferredActionRequestTypeNameFilterIdentifier
						, typeof( DownloadFilesActionRequest ).FullName );

					attachedFilesProgress = transaction.CollectFirstSheaf(
						PortableTypes.FindAllDerived( typeof( DownloadFilesActionServiceProgress ) )
						, new SpeckleFilters( referredSheafFilter, typeNameFilter ), StorehouseRealms.ActionProgresses
						, SheafLocks.Shared, out deserializer ) as DownloadFilesActionServiceProgress;
				}

				// Load the progress sheaf of the attached files only when the letter was received (and so the recipient was a
				// service from the point of view of the main progress), because that's when the recipient has acted as a client
				// and downloaded the attached files.
				if( progress as ActionServiceProgress != null )
				{
					typeNameFilter = new SpeckleFilter( ActionProgress.ReferredActionRequestTypeNameFilterIdentifier
						, typeof( DownloadFilesActionRequest ).FullName );

					attachedFilesProgress = transaction.CollectFirstSheaf(
						PortableTypes.FindAllDerived( typeof( DownloadFilesActionClientProgress ) )
						, new SpeckleFilters( referredSheafFilter, typeNameFilter ), StorehouseRealms.ActionProgresses
						, SheafLocks.Shared, out deserializer ) as DownloadFilesActionClientProgress;
				}
			}

			#endregion Load the progress

			#region Set the information

			letterSpeckle.LastProcessingAttempt = progress.LastProcessingAttempt;

			// Don't show the expiration time when the request is pending because that could confuse the user.
			if( (clientProgress != null) && (clientProgress.ProcessingState == ActionProcessingStates.Retry) )
			{
				letterSpeckle.CarrierActionRequestDocumentExpirationTime
					= clientProgress.ReferredActionRequestDocumentExpirationTime;
			}

			if( attachedFilesProgress != null )
			{
				letterSpeckle.LastProcessingAttemptOfAttachedFiles = attachedFilesProgress.LastProcessingAttempt;

				// Don't show the expiration time when the request is pending because that could confuse the user.
				DownloadFilesActionClientProgress downloadFilesActionClientProgress = attachedFilesProgress
					as DownloadFilesActionClientProgress;
				if( (downloadFilesActionClientProgress != null)
					&& (attachedFilesProgress.ProcessingState == ActionProcessingStates.Retry) )
				{
					letterSpeckle.CarrierActionRequestDocumentExpirationTimeOfAttachedFiles
						= downloadFilesActionClientProgress.ReferredActionRequestDocumentExpirationTime;
				}
			}

			#endregion Set the information

			// Ensure that after the information is loaded, it's not loaded again.
			letterSpeckle.TimeConsumingLoadingIsComplete = true;
		}

		/// <summary>
		/// Loads the progress of the specified letter speckles, and sets their information.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="letterSpeckles">Letter speckles whose progress to load and whose information to set.</param>
		public static void LoadLetterSpeckleProgressAndSetInformation( IStorehouseProvider storehouseProvider
			, LetterSpeckles letterSpeckles )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( letterSpeckles, "letterSpeckles" );

			if( ListUtils.NoContent( letterSpeckles ) )
				return;

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Set the information.
				foreach( LetterSpeckle letterSpeckle in letterSpeckles )
					LoadLetterSpeckleProgressAndSetInformation( letterSpeckle, transaction );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the letter speckles which have the specified flag.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="flag">The letters referred by the collected speckles must have this flag.</param>
		/// <param name="loadTimeConsumingInformation">If this is true then it loads the information which takes a significant
		/// amount of time to load, else it's the responsibility of the caller to load it afterwards (preferably just for a
		/// small part at a time).</param>
		/// <param name="paging">Information about how to process the documents in a paged manner. If this is null, no paging is
		/// used.</param>
		/// <param name="batchProcessAll">If this is true then all the letters are processed in batches (the size of a page).
		/// Each batch has its own transaction, so if a batch is interrupted, only the last batch is rolled back.</param>
		/// <param name="emoticonManager">Emoticon manager. If this is null then the emoticon full-sized and thumb-sized image
		/// file names are not loaded.</param>
		/// <param name="beforeBatchAction">An action to be executed before each batch is processed. This may be null.</param>
		/// <param name="afterBatchAction">An action to be executed after each batch is processed. This may be null.</param>
		/// <param name="afterLetterAction">An action to be executed after each letter is loaded.</param>
		public static void CollectLetterSpeckles( IStorehouseProvider storehouseProvider, string flag
			, bool loadTimeConsumingInformation, CollectPaging paging, bool batchProcessAll, EmoticonManager emoticonManager
			, Action<IStorehouseTransaction> beforeBatchAction, Action<IStorehouseTransaction> afterBatchAction
			, Action<IStorehouseTransaction, LetterSpeckle> afterLetterAction )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.TextHasContent( flag, "flag" );
			ValidationUtils.ObjectHasContent( paging, "paging" );
			ValidationUtils.ObjectHasContent( afterLetterAction, "letterAction" );

			#region Get the tallies and vaults

			// This is done on a separate transaction because the processing of the letters may be made in many batches and
			// the tallies and the vaults only need to be loaded once.

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			Speckles accountSpeckles;
			Dictionary<Locator, Speckle> seekableTallySpeckles;
			Dictionary<Locator, Speckle> seekableVaultSpeckles;

			try
			{
				// Load the account speckles (for the account alias).
				SpeckleFilterIdentifiers accountLoadFilterIdentifiers = new SpeckleFilterIdentifiers(
					Account.LocalUpnFingerprintFilterIdentifier, Account.RemoteUpnFingerprintFilterIdentifier
					, Account.AliasFilterIdentifier );
				accountSpeckles = transaction.CollectSpeckles( PortableTypes.FindAllDerived( typeof( Account ) ), null
					, StorehouseRealms.Accounts, null, null, null, accountLoadFilterIdentifiers ) as Speckles;

				// Load the tally speckles.
				SpeckleFilterIdentifiers tallyLoadFilterIdentifiers = new SpeckleFilterIdentifiers(
					Tally.RemoteUpnFingerprintFilterIdentifier, Tally.EntityNameFilterIdentifier );
				Speckles tallySpeckles = transaction.CollectSpeckles( PortableTypes.FindAllDerived( typeof( Tally ) ), null
					, StorehouseRealms.ProcessingIdentityContainers, null, null, null, tallyLoadFilterIdentifiers ) as Speckles;

				// Load the vault speckles.
				SpeckleFilterIdentifiers vaultLoadFilterIdentifiers = new SpeckleFilterIdentifiers(
					Vault.LocalUpnFingerprintFilterIdentifier, Vault.EntityNameFilterIdentifier );
				Speckles vaultSpeckles = transaction.CollectSpeckles( PortableTypes.FindAllDerived( typeof( Vault ) ), null
					, StorehouseRealms.ProcessingIdentityContainers, null, null, null, vaultLoadFilterIdentifiers ) as Speckles;

				// Collect the speckles keyed by the initiator UPN fingerprint.
				seekableTallySpeckles = tallySpeckles.GetDictionaryByFilterIdentifier(
					Tally.RemoteUpnFingerprintFilterIdentifier );
				seekableVaultSpeckles = vaultSpeckles.GetDictionaryByFilterIdentifier(
					Vault.LocalUpnFingerprintFilterIdentifier );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			#endregion Get the tallies and vaults

			while( true )
			{
				// Create a transaction.
				transaction = storehouseProvider.CreateTransaction();

				if( beforeBatchAction != null )
					beforeBatchAction( transaction );

				int batchLoaded = 0;

				try
				{
					// Load the letter speckles. The list is ordered because the items come sorted from the storehouse.
					LetterSpeckles letterSpeckles = new LetterSpeckles();
					SpeckleFilter flagFilter = new SpeckleFilter( Post.FlagFilterIdentifier, flag );
					transaction.CollectSpeckles( PortableTypes.FindAllDerived( typeof( Letter ) )
						, new SpeckleFilters( flagFilter ), StorehouseRealms.Posts, paging, letterSpeckles
						, typeof( LetterSpeckle ), Sheaf.CollectSpeckleFilterIdentifiers( typeof( Letter ) ) );

					batchLoaded = letterSpeckles.Count;
					paging.ModifyOffset( batchLoaded );

					// Set the information.
					foreach( LetterSpeckle letterSpeckle in letterSpeckles )
					{
						#region Load the initiator and recipient speckles (/ vault or tally)

						Locator initiatorUpnFingerprint = letterSpeckle.GetLoadedSpeckleFilter<Locator>(
							Post.InitiatorUpnFingerprintFilterIdentifier );
						Locator recipientUpnFingerprint = letterSpeckle.GetLoadedSpeckleFilter<Locator>(
							Letter.RecipientUpnFingerprintFilterIdentifier );

						// By trying to load the initiator and recipient speckles from both tally and vault speckles, it's
						// ensured that the imported letters work even for third parties (who don't have the vault of either the
						// initiator or of the recipient).
						//
						// Normally, one should be a tally and the other a vault speckle, but both would be tally speckles if the
						// letter is imported by a third party.

						Speckle initiatorTallySpeckle, recipientTallySpeckle;
						seekableTallySpeckles.TryGetValue( initiatorUpnFingerprint, out initiatorTallySpeckle );
						seekableTallySpeckles.TryGetValue( recipientUpnFingerprint, out recipientTallySpeckle );

						Speckle initiatorVaultSpeckle, recipientVaultSpeckle;
						seekableVaultSpeckles.TryGetValue( initiatorUpnFingerprint, out initiatorVaultSpeckle );
						seekableVaultSpeckles.TryGetValue( recipientUpnFingerprint, out recipientVaultSpeckle );

						Speckle initiatorSpeckle = (initiatorTallySpeckle != null) ? initiatorTallySpeckle
							: initiatorVaultSpeckle;
						Speckle recipientSpeckle = (recipientTallySpeckle != null) ? recipientTallySpeckle
							: recipientVaultSpeckle;

						ValidationUtils.ObjectHasContent( initiatorSpeckle, "initiatorSpeckle" );
						ValidationUtils.ObjectHasContent( recipientSpeckle, "recipientSpeckle" );

						string initiatorEntityNameFilterIdentifier = (initiatorTallySpeckle != null)
							? Tally.EntityNameFilterIdentifier : Vault.EntityNameFilterIdentifier;
						string recipientEntityNameFilterIdentifier = (recipientTallySpeckle != null)
							? Tally.EntityNameFilterIdentifier : Vault.EntityNameFilterIdentifier;

						// The letter could also have been imported; the last processing attempt is required to make an exact
						// determination.
						letterSpeckle.LocalPeerDirection = (initiatorTallySpeckle != null) ? LocalPeerDirections.Recipient
							: LocalPeerDirections.Initiator;

						#endregion Load the initiator and recipient speckles (/ vault or tally)

						// Loading the progress of the letters here significantly decreases the performance. It's faster to
						// load the information later in a timer, just for the letter speckles which are visible in the list.
						if( loadTimeConsumingInformation )
							LoadLetterSpeckleProgressAndSetInformation( letterSpeckle, transaction );

						SetLetterSpeckleInformation( letterSpeckle, initiatorEntityNameFilterIdentifier
							, recipientEntityNameFilterIdentifier, initiatorSpeckle, recipientSpeckle, emoticonManager );

						#region Set the account alias

						Locator localUpnFingerprint = (initiatorTallySpeckle != null)
							? recipientUpnFingerprint // The initiator is the remote peer.
							: initiatorUpnFingerprint; // The initiator is the local peer.
						Locator remoteUpnFingerprint = (initiatorTallySpeckle != null)
							? initiatorUpnFingerprint // The initiator is the remote peer.
							: recipientUpnFingerprint; // The initiator is the local peer.

						Speckle accountSpeckle = PeerDocumentsUtils.FindAccountByInitiatorAndRecipientUpnFingerprints(
							accountSpeckles, localUpnFingerprint, remoteUpnFingerprint );
						if( accountSpeckle != null )
						{
							letterSpeckle.AccountAliasTextualForm = accountSpeckle.TryGetLoadedSpeckleFilter<string>(
								Account.AliasFilterIdentifier );
						}

						#endregion Set the account alias

						afterLetterAction( transaction, letterSpeckle );
					}

					// Commit the transaction.
					transaction.Commit();
				}
				catch( Exception ex )
				{
					// Rolls back the transaction.
					transaction.Rollback();

					throw ExceptionUtils.PreserveStackTrace( ex );
				}

				if( afterBatchAction != null )
					afterBatchAction( transaction );

				if( !batchProcessAll || (batchLoaded <= 0) )
					return;
			}
		}

		/// <summary>
		/// Loads the letter speckles which have the specified flag.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="flag">The letters referred by the collected speckles must have this flag.</param>
		/// <param name="loadTimeConsumingInformation">If this is true then it loads the information which takes a significant
		/// amount of time to load, else it's the responsibility of the caller to load it afterwards (preferably just for a
		/// small part at a time).</param>
		/// <param name="paging">Information about how to process the documents in a paged manner. If this is null, no paging is
		/// used.</param>
		/// <param name="emoticonManager">Emoticon manager. If this is null then the emoticon full-sized and thumb-sized image
		/// file names are not loaded.</param>
		/// <returns>Returns the loaded letter speckles. This may be empty, but not null.</returns>
		public static LetterSpeckles CollectLetterSpeckles( IStorehouseProvider storehouseProvider, string flag
			, bool loadTimeConsumingInformation, CollectPaging paging, EmoticonManager emoticonManager )
		{
			// The speckle keys will change due to the new names, after being collected from the storehouse, so put the speckles
			// in a properly sorted list.
			LetterSpeckles sortedLetterSpeckles = new LetterSpeckles();

			CollectLetterSpeckles( storehouseProvider, flag, loadTimeConsumingInformation, paging, false, emoticonManager, null
				, null, ( transaction, letterSpeckle ) => { sortedLetterSpeckles.Add( letterSpeckle ); } );

			return sortedLetterSpeckles;
		}

		/// <summary>
		/// Loads the letter speckle with the specified storehouse key.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="letterLocator">The storehouse key of the letter speckle to load.</param>
		/// <param name="emoticonManager">Emoticon manager. If this is null then the emoticon full-sized and thumb-sized image
		/// file names are not loaded.</param>
		/// <returns>Returns the loaded letter speckle.</returns>
		/// <remarks>
		/// <para>The loaded sheafs are not locked.</para>
		/// </remarks>
		public static LetterSpeckle LoadLetterSpeckle( IStorehouseProvider storehouseProvider, Locator letterLocator
			, EmoticonManager emoticonManager )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( letterLocator, "letterLocator" );
			ValidationUtils.ObjectHasContent( emoticonManager, "emoticonManager" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the letter speckle.
				LetterSpeckle letterSpeckle = transaction.LoadSpeckle( letterLocator, StorehouseRealms.Posts, SheafLocks.Shared
					, typeof( LetterSpeckle ), Sheaf.CollectSpeckleFilterIdentifiers( typeof( Letter ) ) ) as LetterSpeckle;
				ValidationUtils.ObjectHasContent( letterSpeckle, "letterSpeckle" );

				#region Load the initiator and recipient speckles (/ vault or tally)

				Locator initiatorUpnFingerprint = letterSpeckle.GetLoadedSpeckleFilter<Locator>(
					Post.InitiatorUpnFingerprintFilterIdentifier );
				Locator recipientUpnFingerprint = letterSpeckle.GetLoadedSpeckleFilter<Locator>(
					Letter.RecipientUpnFingerprintFilterIdentifier );

				// By trying to load the initiator and recipient speckles from both tally and vault speckles, it's ensured
				// that the imported letters work even for third parties (who don't have the vault of either the initiator or
				// of the recipient).
				//
				// Normally, one should be a tally and the other a vault speckle, but it's possible that both are tally
				// speckles if the letter was imported.

				Speckle recipientSpeckle = null;
				string initiatorEntityNameFilterIdentifier, recipientEntityNameFilterIdentifier;
				Locator localUpnFingerprint, remoteUpnFingerprint;

				Speckle initiatorSpeckle = LoadTallySpeckleByUpnFingerprint( transaction, initiatorUpnFingerprint );

				if( initiatorSpeckle != null )
				{
					// The initiator is the remote peer.

					// The letter could also have been imported; the last processing attempt is required to make an exact
					// determination.
					letterSpeckle.LocalPeerDirection = LocalPeerDirections.Recipient;

					recipientSpeckle = LoadVaultSpeckleByUpnFingerprint( transaction, recipientUpnFingerprint );

					initiatorEntityNameFilterIdentifier = Tally.EntityNameFilterIdentifier;
					recipientEntityNameFilterIdentifier = Vault.EntityNameFilterIdentifier;

					localUpnFingerprint = recipientUpnFingerprint;
					remoteUpnFingerprint = initiatorUpnFingerprint;
				}
				else
				{
					// The initiator is the local peer.

					letterSpeckle.LocalPeerDirection = LocalPeerDirections.Initiator;

					initiatorSpeckle = LoadVaultSpeckleByUpnFingerprint( transaction, initiatorUpnFingerprint );

					initiatorEntityNameFilterIdentifier = Vault.EntityNameFilterIdentifier;
					recipientEntityNameFilterIdentifier = Tally.EntityNameFilterIdentifier;

					localUpnFingerprint = initiatorUpnFingerprint;
					remoteUpnFingerprint = recipientUpnFingerprint;
				}

				if( recipientSpeckle == null )
					recipientSpeckle = LoadTallySpeckleByUpnFingerprint( transaction, recipientUpnFingerprint );

				ValidationUtils.ObjectHasContent( initiatorSpeckle, "initiatorSpeckle" );
				ValidationUtils.ObjectHasContent( recipientSpeckle, "recipientSpeckle" );

				#endregion Load the initiator and recipient speckles (/ vault or tally)

				LoadLetterSpeckleProgressAndSetInformation( letterSpeckle, transaction );

				SetLetterSpeckleInformation( letterSpeckle, initiatorEntityNameFilterIdentifier
					, recipientEntityNameFilterIdentifier, initiatorSpeckle, recipientSpeckle, emoticonManager );

				#region Set the account alias

				SpeckleFilterIdentifiers accountLoadFilterIdentifiers = new SpeckleFilterIdentifiers(
					Account.AliasFilterIdentifier );

				Speckle accountSpeckle = LoadAccountSpeckleByUpnFingerprints( transaction, localUpnFingerprint
					, remoteUpnFingerprint, accountLoadFilterIdentifiers );
				if( accountSpeckle != null )
				{
					letterSpeckle.AccountAliasTextualForm = accountSpeckle.TryGetLoadedSpeckleFilter<string>(
						Account.AliasFilterIdentifier );
				}

				#endregion Set the account alias

				// Commit the transaction.
				transaction.Commit();

				return letterSpeckle;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the message summary and the emoticon symbol of the letter with the specified storehouse key.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="letterLocator">The storehouse key of the letter whose message summary to load.</param>
		/// <param name="messageSummary">Is set with the message summary of the loaded letter. This may be null or empty. If the
		/// letter doesn't exist, this is null.</param>
		/// <param name="emoticonSymbol">Is set with the emoticon symbol of the loaded letter. This may be null or empty. If the
		/// letter doesn't exist, this is null.</param>
		/// <remarks>
		/// <para>The loaded sheafs are not locked.</para>
		/// </remarks>
		public static void LoadLetterMessageSummaryAndEmoticonSymbol( IStorehouseProvider storehouseProvider
			, Locator letterLocator, out string messageSummary, out string emoticonSymbol )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( letterLocator, "letterLocator" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				Speckle letterSpeckle = transaction.LoadSpeckle( letterLocator, StorehouseRealms.Posts, SheafLocks.Shared, null
					, new SpeckleFilterIdentifiers( Post.MessageSummaryFilterIdentifier, Post.EmoticonSymbolFilterIdentifier ) );

				// Commit the transaction.
				transaction.Commit();

				if( letterSpeckle != null )
				{
					messageSummary = letterSpeckle.TryGetLoadedSpeckleFilter<string>( Post.MessageSummaryFilterIdentifier );
					emoticonSymbol = letterSpeckle.TryGetLoadedSpeckleFilter<string>( Post.EmoticonSymbolFilterIdentifier );
				}
				else
				{
					messageSummary = null;
					emoticonSymbol = null;
				}
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the letter with the specified instance locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="letterLocator">The instance locator of the letter to load.</param>
		/// <returns>Returns the loaded letter, or null if none is found.</returns>
		public static Letter LoadLetter( IStorehouseProvider storehouseProvider, Locator letterLocator )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( letterLocator, "letterLocator" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the letter.
				IDeserializer deserializer;
				Letter letter = transaction.LoadSheaf( letterLocator, StorehouseRealms.Posts, typeof( Letter ), SheafLocks.Shared
					, out deserializer ) as Letter;

				// Commit the transaction.
				transaction.Commit();

				return letter;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Deletes the letter with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="letterLocator">The locator of the letter to delete.</param>
		/// <param name="serviceManager">Action service manager which is used to stop processing action requests which have a
		/// multiframe action response.</param>
		/// <param name="clientManager">Action client manager which is used to stop processing action requests which have a
		/// multiframe action response.</param>
		/// <returns>Returns true if the letter with the specified locator was deleted.</returns>
		public static bool DeleteLetter( IStorehouseProvider storehouseProvider, Locator letterLocator
			, IActionServiceManager serviceManager, IActionClientManager clientManager )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( letterLocator, "letterLocator" );
			ValidationUtils.ObjectHasContent( serviceManager, "serviceManager" );
			ValidationUtils.ObjectHasContent( clientManager, "clientManager" );

			// Ask the user.
			string message = "Do you want to delete the letter?"
				+ "\n\n"
				+ "If you have sent the letter with attached files and they are currently being transferred, the transfer will"
				+ " not be interrupted (but the resume feature will not be supported).";
			if( MessageWindow.ShowDialog( message, MessageWindowModes.Query ) != QueryMessageResults.Yes )
				return false;

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Delete the letter.
				DeleteLetterSheafs( transaction, letterLocator, serviceManager, clientManager );

				// Commit the transaction.
				transaction.Commit();

				return true;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				if( storehouseProvider.IsTimeoutException( ex ) )
				{
					throw new Vela.Kernel.TimeoutException( "The letter or one of its dependencies is being processed by a"
						+ " background operation. Try again later.", ex );
				}

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Deletes the letters with the specified flag.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="flag">The letters to delete must have this flag.</param>
		/// <param name="afterDeleteAction">An action to be executed after each letter is deleted. This may be null.</param>
		/// <param name="lop">Long operation progress. This may be null.</param>
		/// <param name="serviceManager">Action service manager which is used to stop processing action requests which have a
		/// multiframe action response.</param>
		/// <param name="clientManager">Action client manager which is used to stop processing action requests which have a
		/// multiframe action response.</param>
		/// <param name="deleted">Is set with the number of deleted letters.</param>
		/// <para>If "flag" is equal with <see cref="PostFlags.Main"/>, an exception is thrown.</para>
		/// 
		/// <para>The letters are processed in batches, each with its own transaction, so an abort will rollback only the changes
		/// from the current batch.</para>
		public static void DeleteLetters( IStorehouseProvider storehouseProvider, string flag
			, Action<Locator> afterDeleteAction, LongOperationProgress lop, IActionServiceManager serviceManager
			, IActionClientManager clientManager, out int deleted )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.TextHasContent( flag, "flag" );
			ValidationUtils.ObjectHasContent( serviceManager, "serviceManager" );
			ValidationUtils.ObjectHasContent( clientManager, "clientManager" );

			deleted = 0;

			if( Post.EqualFlags( flag, PostFlags.Main.ToString() ) )
			{
				throw new ValidationException( string.Format( "The letters with the flag \"{0}\" can't be deleted because"
					+ " letters which are received after the deletion starts might also be deleted (and you would not get a"
					+ " chance to see them).", flag ) );
			}

			// Count the letters.
			int lettersCount = CountLetterSpeckles( storehouseProvider, flag );
			if( lettersCount <= 0 )
				return;

			if( lop != null )
				lop.Progress( 0 );

			PortableTypes letterTypes = PortableTypes.FindAllDerived( typeof( Letter ) );

			// The flag speckle filter.
			SpeckleFilter flagFilter = new SpeckleFilter( Post.FlagFilterIdentifier, flag );

			Locators letterLocators = null;

			do
			{
				// Create a transaction for each batch.
				IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

				try
				{
					// Load the letter locators.
					letterLocators = transaction.CollectKeys( letterTypes, new SpeckleFilters( flagFilter )
						, StorehouseRealms.Posts, new CollectPaging( true, 0, DeleteLettersBatchSize ), null, null ) as Locators;

					int deletedPerBatch = 0;
					foreach( Locator letterLocator in letterLocators )
					{
						DeleteLetterSheafs( transaction, letterLocator, serviceManager, clientManager );

						deletedPerBatch++;

						if( afterDeleteAction != null )
							afterDeleteAction( letterLocator );

						if( lop != null )
							lop.Progress( (double)(deleted + deletedPerBatch) / lettersCount );
					}

					// Commit the transaction.
					transaction.Commit();

					deleted += deletedPerBatch;
				}
				catch( Exception ex )
				{
					// Rolls back the transaction.
					transaction.Rollback();

					throw ExceptionUtils.PreserveStackTrace( ex );
				}
			}
			while( ListUtils.HasContent( letterLocators ) );

			if( lop != null )
				lop.Progress( 1 );
		}

		/// <summary>
		/// Deletes all the transient letters.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="lop">Long operation progress. This may be null.</param>
		/// <param name="serviceManager">Action service manager which is used to stop processing action requests which have a
		/// multiframe action response. This may be null.</param>
		/// <param name="clientManager">Action client manager which is used to stop processing action requests which have a
		/// multiframe action response. This may be null.</param>
		/// <param name="deleted">Is set with the number of deleted letters.</param>
		/// <para>The letters are processed in batches, each with its own transaction, so an abort will rollback only the changes
		/// from the current batch.</para>
		public static void DeleteTransientLetters( IStorehouseProvider storehouseProvider, LongOperationProgress lop
			, IActionServiceManager serviceManager, IActionClientManager clientManager, out int deleted )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			deleted = 0;

			// Count the letters.
			int lettersCount = CountTransientLetterSpeckles( storehouseProvider );
			if( lettersCount <= 0 )
				return;

			if( lop != null )
				lop.Progress( 0 );

			PortableTypes letterTypes = PortableTypes.FindAllDerived( typeof( Letter ) );

			// The speckle filter.
			SpeckleFilter isTransientFilter = new SpeckleFilter( Post.IsTransientFilterIdentifier, true );

			Locators letterLocators = null;

			do
			{
				// Create a transaction for each batch.
				IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

				try
				{
					// Load the letter locators.
					letterLocators = transaction.CollectKeys( letterTypes, new SpeckleFilters( isTransientFilter )
						, StorehouseRealms.Posts, new CollectPaging( true, 0, DeleteLettersBatchSize ), null, null ) as Locators;

					int deletedPerBatch = 0;
					foreach( Locator letterLocator in letterLocators )
					{
						DeleteLetterSheafs( transaction, letterLocator, serviceManager, clientManager );

						deletedPerBatch++;

						if( lop != null )
							lop.Progress( (double)(deleted + deletedPerBatch) / lettersCount );
					}

					// Commit the transaction.
					transaction.Commit();

					deleted += deletedPerBatch;
				}
				catch( Exception ex )
				{
					// Rolls back the transaction.
					transaction.Rollback();

					throw ExceptionUtils.PreserveStackTrace( ex );
				}
			}
			while( ListUtils.HasContent( letterLocators ) );

			if( lop != null )
				lop.Progress( 1 );
		}

		/// <summary>
		/// Gets all the letters with the specified flag as a text.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="flag">The flag of the letters to process.</param>
		/// <param name="accounts">List of accounts from which to get the account aliases.</param>
		/// <param name="lop">Long operation progress. This may be null.</param>
		/// <param name="lettersAsText">Is set with the resulted text. This may be null or empty.</param>
		/// <param name="processed">Is set with the number of processed letters.</param>
		public static void GetLettersAsText( IStorehouseProvider storehouseProvider, string flag
			, AccountOrRequestSpeckles accounts, LongOperationProgress lop, out string lettersAsText, out int processed )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.TextHasContent( flag, "flag" );
			ValidationUtils.ObjectHasContent( accounts, "accounts" );

			lettersAsText = null;
			processed = 0;

			StringBuilder sb = new StringBuilder();
			int totalProcessed = 0;

			// Count the letters.
			int lettersCount = CountLetterSpeckles( storehouseProvider, flag );
			if( lettersCount <= 0 )
				return;

			try
			{
				if( lop != null )
					lop.Progress( 0 );

				CollectPaging paging = new CollectPaging( false, 0, GetLettersAsTextFlagBatchSize );

				int letterCounter = 1;
				int processedPerBatch = 0;

				// The speckle keys will change due to the new names, after being collected from the storehouse, so put the
				// speckles in a properly sorted list.
				LetterSpeckles sortedLetterSpeckles = new LetterSpeckles();
				Dictionary<Locator, string> letterMessages = new Dictionary<Locator, string>();

				CollectLetterSpeckles( storehouseProvider, flag, true, paging, true, null
					, ( transaction ) => { processedPerBatch = 0; }
					, ( transaction ) => { totalProcessed += processedPerBatch; }
					, ( transaction, letterSpeckle ) =>
						{
							// Load the letter.
							IDeserializer deserializer;
							Letter letter = transaction.LoadSheaf( letterSpeckle.StorehouseKey, StorehouseRealms.Posts
								, typeof( Letter ), SheafLocks.Shared, out deserializer ) as Letter;

							letterMessages.Add( letterSpeckle.StorehouseKey, letter.Message.Trim() );

							sortedLetterSpeckles.Add( letterSpeckle );

							letterCounter++;
							processedPerBatch++;

							if( lop != null )
								lop.Progress( (double)(totalProcessed + processedPerBatch) / lettersCount );
						}
				);

				foreach( LetterSpeckle letterSpeckle in sortedLetterSpeckles )
				{
					AccountOrRequestSpeckle account = accounts.FindByUpnFingerprints( letterSpeckle.LocalUpnFingerprint
						, letterSpeckle.RemoteUpnFingerprint );

					string accountAlias = (account != null) ? account.AliasTextualForm : null;

					string letterText = string.Format( GetLettersAsTextTemplate, letterCounter
						, letterSpeckle.LastProcessingAttempt.Time, letterSpeckle.InitiatorEntityName
						, letterSpeckle.RecipientEntityName, accountAlias, letterMessages[ letterSpeckle.StorehouseKey ] );

					sb.Append( letterText );
				}

				if( lop != null )
					lop.Progress( 1 );
			}
			finally
			{
				lettersAsText = sb.ToString();
				processed = totalProcessed;
			}
		}

		/// <summary>
		/// Queues an action request which contains a letter with the specified message for sending to the remote peer of the
		/// account with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="localUpnFingerprint">The UPN fingerprint of the local peer from which to send a letter.</param>
		/// <param name="remoteUpnFingerprint">The UPN fingerprint of the remote peer to which to send a letter.</param>
		/// <param name="localIdentity">The private identity of the local peer for which to send a letter.</param>
		/// <param name="remoteIdentity">The identity publisher of the remote peer to which to send a letter.</param>
		/// <param name="actionRequestValidity">The validity timespan (in minutes) of the created action request.
		/// <see cref="ActionRequest.DocumentExpirationTime"/> is set with <see cref="Sheaf.CreationTime"/> +
		/// "actionRequestValidity". If this is 0 or a negative number, the action request is valid indefinitely.</param>
		/// <param name="message">The message of the letter to send.</param>
		/// <param name="emoticonSymbol">The emoticon symbol of the letter to send. This may be null or empty.</param>
		/// <param name="attachedFileNames">The list of names of the attached files. This may be null or empty.</param>
		/// <param name="isTransient">If this is true then the letter is transient.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="serializer">Serializer.</param>
		/// <param name="letterLocator">Is set with the instance locator of the queued letter.</param>
		/// <returns>Returns the instance locator of the created action request.</returns>
		/// <remarks>
		/// <para>The letter is put in an action request and saved in the storehouse, it's not actually sent.</para>
		/// </remarks>
		public static Locator QueueSendLetter( IStorehouseProvider storehouseProvider, Locator localUpnFingerprint
			, Locator remoteUpnFingerprint, PrivateIdentity localIdentity, IIdentityPublisher remoteIdentity
			, int actionRequestValidity, string message, string emoticonSymbol, AttachedFileNames attachedFileNames
			, bool isTransient, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg, ISerializer serializer
			, out Locator letterLocator )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( localUpnFingerprint, "localUpnFingerprint" );
			ValidationUtils.ObjectHasContent( remoteUpnFingerprint, "remoteUpnFingerprint" );
			ValidationUtils.ObjectHasContent( localIdentity, "localIdentity" );
			ValidationUtils.ObjectHasContent( remoteIdentity, "remoteIdentity" );
			ValidationUtils.TextHasContent( message, "message" );
			ValidationUtils.ObjectHasContent( macAlg, "macAlg" );
			ValidationUtils.ObjectHasContent( asymmSigningPaddingAlg, "asymmSigningPaddingAlg" );
			ValidationUtils.ObjectHasContent( serializer, "serializer" );

			if( !localUpnFingerprint.Equals( localIdentity.Upn.Fingerprint ) )
			{
				throw new ValidationException( "The UPN fingerprint of the sender is different than the UPN fingerprint of the"
						+ " identity of the local peer.", null, null
					, new ExceptionDetail( "Sender UPN fingerprint", localUpnFingerprint )
					, new ExceptionDetail( "Local identity UPN fingerprint", localIdentity.Upn.Fingerprint ) );
			}

			foreach( AttachedFileName attachedFileName in attachedFileNames )
			{
				if( Directory.Exists( attachedFileName.AbsoluteFileName ) )
				{
					throw new ValidationException( "Folders may not be attached to a letter.", null, null
						, new ExceptionDetail( "File name", attachedFileName.AbsoluteFileName ) );
				}
			}

			bool hasAttachedMediaFiles = AttachedFilesUtils.HasMediaFiles( attachedFileNames );
			bool hasImagePreview = AttachedFilesUtils.HasImageFile( attachedFileNames );

			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				IHashSerializer hashSerializer = serializer.GetHashSerializer( false );
				letterLocator = CryptoUtils.GetLocator();

				#region Create the letter

				// Create the letter.
				Letter letter = new Letter( letterLocator, macAlg, asymmSigningPaddingAlg, hashSerializer, localUpnFingerprint
					, message, emoticonSymbol, isTransient, remoteUpnFingerprint, new AttachedFiles( attachedFileNames )
					, hasAttachedMediaFiles, hasImagePreview );

				// Sign the letter.
				if( !letter.IsTransient )
					letter.Sign( localIdentity, null );

				// Save the letter.
				transaction.AddSheaf( letter, StorehouseRealms.Posts );

				#endregion Create the letter

				#region Create the action request

				// Create the action request.
				SolitaryIdentityPublisher remoteSolitaryIdentity = CoreDocumentsUtils.GetSolitaryIdentity( remoteIdentity );
				SendLetterActionRequest request = new SendLetterActionRequest( CryptoUtils.GetLocator()
					, macAlg, asymmSigningPaddingAlg, hashSerializer, localUpnFingerprint, remoteSolitaryIdentity.Identity.Upn
					, actionRequestValidity, 0, letter );

				// Sign the action request.
				request.Sign( localIdentity, null );

				// Save the action request.
				transaction.AddSheaf( request, StorehouseRealms.ActionRequestsToSend );

				#endregion Create the action request

				#region Create the action progress

				// Create the action progress.
				ActionClientProgress progress = new ActionClientProgress( CryptoUtils.GetLocator()
					, request.InitiatorUpnFingerprint, request.RecipientUpnFingerprint, letter.InstanceLocator
					, request.GetType().FullName, request.InstanceLocator, request.DocumentExpirationTime
					, request.ClientProgressIsPermanent );

				// Save the action progress.
				transaction.AddSheaf( progress, StorehouseRealms.ActionProgresses );

				#endregion Create the action progress

				// Commit the transaction.
				transaction.Commit();

				return request.InstanceLocator;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		/// <summary>
		/// Exports the letter with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="algorithmPool">Algorithm pool.</param>
		/// <param name="letterLocator">The locator of the letter to export.</param>
		/// <param name="exportObfuscated">If this is true then the letter is exported in obfuscated form, like action requests
		/// are when they are sent to their recipient.</param>
		/// <returns>Returns true if the letter was exported.</returns>
		public static bool ExportLetter( IStorehouseProvider storehouseProvider, AlgorithmPool algorithmPool
			, Locator letterLocator, bool exportObfuscated )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( algorithmPool, "algorithmPool" );
			ValidationUtils.ObjectHasContent( letterLocator, "letterLocator" );

			string selectedFileName;

			if( exportObfuscated )
			{
				if( !ControlsUtils.ShowSaveFileDialog( "Export Obfuscated Letter", null, FileExtensions.AllFilesFilter, null
					, null, true, out selectedFileName ) )
				{
					return false;
				}
			}
			else
			{
				DateTime now = DateTime.UtcNow;
				string initialFileName = string.Format( ExportedLetterFileNameTemplate
					, now.ToLocalTime().ToString( KernelConsts.TimeCompactFormat )
					, TextUtils.ArrayToText32( letterLocator.Value ) );

				if( !ControlsUtils.ShowSaveFileDialog( "Export Letter", FileExtensions.AsymmEncryptedDocument
					, FileExtensions.AsymmEncryptedDocumentFilter, null, initialFileName, true, out selectedFileName ) )
				{
					return false;
				}
			}

			string selectedFileExtension = Path.GetExtension( selectedFileName );
			if( exportObfuscated && TextUtils.Equals( selectedFileExtension, FileExtensions.AsymmEncryptedDocument ) )
			{
				throw new ValidationException( string.Format( "A letter to export in obfuscated form may not have the standard"
					+ " file extension \"{0}\".", FileExtensions.AsymmEncryptedDocument ) );
			}

			#region Load the data

			Letter letter = LoadLetter( storehouseProvider, letterLocator );
			ValidationUtils.ObjectHasContent( letter, "letter" );

			if( letter.HasAttachedFiles )
				throw new ValidationException( "Can't export the letter because it has attached files." );

			Tally tally = LoadTallyByUpnFingerprint( storehouseProvider, letter.RecipientUpnFingerprint );
			if( tally == null )
				throw new ValidationException( "Can't export the letter because you are not the sender." );

			// Don't use expired documents.
			KernelUtils.HasExpired( letter );
			KernelUtils.HasExpired( tally );

			#endregion Load the data

			MacAlg macAlg = algorithmPool.GetMacAlg();
			CompressionAlg compressionAlg = algorithmPool.GetCompressionAlg();
			SymmAlg symmAlg = algorithmPool.GetSymmAlg();
			IAsymmEncryptionPaddingAlg asymmEncryptionPaddingAlg = algorithmPool.GetAsymmEncryptionPadding();

			#region Save the encrypted data

			// Create the asymmetrically encrypted envelope.
			AsymmEncryptedEnvelope asymmEnvelope = new AsymmEncryptedEnvelope( CryptoUtils.GetLocator(), macAlg
				, compressionAlg, symmAlg, asymmEncryptionPaddingAlg );
			asymmEnvelope.EncryptedDocumentBlockSize = algorithmPool.GetExportedDocumentEnvelopeEncryptedDocumentBlockSize();
			asymmEnvelope.EncryptableDocument = letter;

			ISerializer serializer = algorithmPool.GetSerializerForExport();

			if( exportObfuscated )
			{
				Upn remoteUpn = tally.RemoteUpn;
				SteganographicAlg steganographicAlg = remoteUpn.UpnAlg.SteganographicAlg;

				byte[] encryptedLetter = asymmEnvelope.SaveToBuffer( null, tally.ProcessingIdentity, serializer );

				// Obfuscate the encrypted envelope of the letter.
				byte[] obfuscatedLetter = steganographicAlg.Obfuscate( encryptedLetter, remoteUpn.SymmEncryptionKey
					, remoteUpn.Fingerprint.Value, algorithmPool.GetExportedDocumentSteganographicDisalignmentSizeZenith() );

				// Save the encrypted and obfuscated envelope of the letter.
				File.WriteAllBytes( selectedFileName, obfuscatedLetter );
			}
			else
			{
				// Save the letter in its encrypted envelope.
				asymmEnvelope.SaveToFile( selectedFileName, null, tally.ProcessingIdentity, serializer );
			}

			#endregion Save the encrypted data

			if( exportObfuscated )
				TooltipWindow.ShowTooltip( "The letter was exported in OBFUSCATED form." );
			else
				TooltipWindow.ShowTooltip();

			return true;
		}

		/// <summary>
		/// Reveals the raw document from the specified obfuscated document, using the recipient UPNs.
		/// </summary>
		/// <param name="vaults">Vaults which may be the recipients of the obfuscated document.</param>
		/// <param name="obfuscatedDocument">Obfuscated document.</param>
		/// <param name="algorithmPool">Algorithm pool.</param>
		/// <param name="recipientUpn">Is set with the UPN of the recipient for which the document is intended.</param>
		/// <returns>Returns the raw document.</returns>
		public static byte[] SteganographicReveal( Sheafs vaults, byte[] obfuscatedDocument, AlgorithmPool algorithmPool
			, out Upn recipientUpn )
		{
			ValidationUtils.ObjectHasContent( obfuscatedDocument, "obfuscatedDocument" );
			ValidationUtils.ObjectHasContent( algorithmPool, "algorithmPool" );

			if( (vaults == null) || (vaults.Count <= 0) )
			{
				throw new ValidationException( "The document can't be steganographically revealed because there are no recipient"
					+ " vaults." );
			}

			foreach( Vault vault in vaults )
			{
				Upns upns = vault.GenerateUpns();

				foreach( Upn upn in upns )
				{
					byte[] rawDocument = SteganographicAlg.Reveal( obfuscatedDocument, upn
						, algorithmPool.GetExportedDocumentSteganographicDisalignmentSizeZenith() );

					if( rawDocument != null )
					{
						recipientUpn = upn;

						return rawDocument;
					}
				}
			}

			throw new ValidationException( "The document can't be steganographically revealed." );
		}

		/// <summary>
		/// Imports the letter from the file with the specified name.
		/// </summary>
		/// <param name="transaction">Transaction to use.</param>
		/// <param name="vaults">Vaults which may be the recipients of the obfuscated document.</param>
		/// <param name="fileName">The name of the file from which to import the letter.</param>
		/// <param name="algorithmPool">Algorithm pool.</param>
		/// <returns>Returns the locator of the imported letter.</returns>
		/// <remarks>
		/// <para>If a letter with the same instance locator is already in the storehouse, nothing happens.</para>
		/// </remarks>
		public static Locator ImportLetter( IStorehouseTransaction transaction, Sheafs vaults, string fileName
			, AlgorithmPool algorithmPool )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.TextHasContent( fileName, "fileName" );
			ValidationUtils.ObjectHasContent( algorithmPool, "algorithmPool" );

			#region Load the data

			string fileExtension = Path.GetExtension( fileName );
			bool isObfuscated = !TextUtils.Equals( fileExtension, FileExtensions.AsymmEncryptedDocument );

			Letter letter;

			if( isObfuscated )
			{
				byte[] obfuscatedDocument = FileUtils.Read( fileName, true );

				// Reveal the document.
				Upn recipientUpn;
				byte[] rawDocument = SteganographicReveal( vaults, obfuscatedDocument, algorithmPool, out recipientUpn );

				letter = LoadAsymmEncryptedDocument( transaction, rawDocument, fileName, true, null, typeof( Letter ) )
					as Letter;
			}
			else
			{
				letter = LoadAsymmEncryptedDocument( transaction, fileName, true, null, typeof( Letter ) ) as Letter;
			}

			Tally tally = LoadTallyByUpnFingerprint( transaction, letter.InitiatorUpnFingerprint );
			if( tally == null )
			{
				throw new ValidationException( "Can't import the letter because the tally is missing. It's possible that you are"
						+ " not the intended recipient.", null, null
					, new ExceptionDetail( "Sender UPN fingerprint", letter.InitiatorUpnFingerprint ) );
			}

			// Don't use expired documents.
			KernelUtils.HasExpired( letter );
			KernelUtils.HasExpired( tally );

			#endregion Load the data

			// Verify the letter's signature.
			if( !letter.IsTransient )
				letter.VerifySignature( tally.ProcessingIdentity );

			// Update the letter's local information.
			letter.Flag = PostFlags.Main.ToString();

			// Check if the letter already exists.
			SheafExistenceModes letterExistence = transaction.SheafExists( letter, StorehouseRealms.Posts );
			switch( letterExistence )
			{
				case SheafExistenceModes.DoesntExist:
					// Nothing to do.
					break;
				case SheafExistenceModes.IsIdentical:
					// The letter already exists, so it may have already been imported.
					return letter.InstanceLocator;
				case SheafExistenceModes.IsDifferent:
					throw new FatalCrossServiceException( "A different sheaf with the letter's locator already exists.", null
						, null, new ExceptionDetail( "Letter locator", letter.InstanceLocator ) );
				default:
					throw ValidationUtils.Unhandled( letterExistence, "letterExistence" );
			}

			// Set the processing state.
			ActionServiceProgress actionProgress = new ActionServiceProgress( CryptoUtils.GetLocator()
				, letter.InitiatorUpnFingerprint, letter.RecipientUpnFingerprint, letter.InstanceLocator
				, typeof( SendLetterActionRequest ).FullName, null );
			ActionLastProcessingAttempt lastProcessingAttempt = new ActionLastProcessingAttempt(
				ActionProcessingStates.RecipientImported, null, null );
			actionProgress.UpdateLastProcessingAttempt( lastProcessingAttempt );

			// Save the sheafs.
			transaction.AddSheaf( letter, StorehouseRealms.Posts );
			transaction.AddSheaf( actionProgress, StorehouseRealms.ActionProgresses );

			return letter.InstanceLocator;
		}

		/// <summary>
		/// Imports the letter from the file with the specified name.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="fileName">The name of the file from which to import the letter.</param>
		/// <param name="algorithmPool">Algorithm pool.</param>
		/// <returns>Returns the locator of the imported letter.</returns>
		/// <remarks>
		/// <para>If a letter with the same instance locator is already in the storehouse, nothing happens.</para>
		/// </remarks>
		public static Locator ImportLetter( IStorehouseProvider storehouseProvider, string fileName
			, AlgorithmPool algorithmPool )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				Sheafs vaults = CollectVaults( transaction );

				Locator letterLocator = ImportLetter( transaction, vaults, fileName, algorithmPool );

				// Commit the transaction.
				transaction.Commit();

				return letterLocator;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Imports the letters from the files with the specified names.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="fileNames">The names of the files from which to import letters. This may be empty, but not null.</param>
		/// <param name="lop">Long operation progress. This may be null.</param>
		/// <param name="algorithmPool">Algorithm pool.</param>
		/// <param name="importedCount">Is set with the number of imported letters.</param>
		/// <para>The letters are processed in batches, each with its own transaction, so an abort will rollback only the changes
		/// from the current batch.</para>
		/// <remarks>
		/// <para>The files are sorted by name being being imported.</para>
		/// 
		/// <para>If a letter with the same instance locator is already in the storehouse, nothing happens (for that letter).
		/// </para>
		/// </remarks>
		public static void ImportLetters( IStorehouseProvider storehouseProvider, string[] fileNames, LongOperationProgress lop
			, AlgorithmPool algorithmPool, out int importedCount )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( fileNames, "fileNames" );

			if( lop != null )
				lop.Progress( 0 );

			IStorehouseTransaction transaction = null;

			try
			{
				importedCount = 0;

				Sheafs vaults = CollectVaults( storehouseProvider );

				Strings sortedFileName = new Strings( fileNames );
				int lettersCount = sortedFileName.Count;
				int letterIndex = 0;
				int importedPerBatch = 0;

				foreach( string fileName in sortedFileName )
				{
					// Create a transaction for each batch.
					if( letterIndex % ImportLettersBatchSize == 0 )
					{
						transaction = storehouseProvider.CreateTransaction();
						importedPerBatch = 0;
					}

					ImportLetter( transaction, vaults, fileName, algorithmPool );

					letterIndex++;
					importedPerBatch++;

					if( letterIndex % ImportLettersBatchSize == 0 )
					{
						// Commit the transaction.
						transaction.Commit();
						transaction = null;

						importedCount += importedPerBatch;

						if( lop != null )
							lop.Progress( (double)importedCount / lettersCount );
					}
				}

				if( transaction != null )
				{
					// Commit the transaction.
					transaction.Commit();

					importedCount += importedPerBatch;
				}
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				if( transaction != null )
					transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			if( lop != null )
				lop.Progress( 1 );
		}

		#endregion Public static methods
	}
}
