////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Microsoft.Win32;
using System;
using System.IO;
using System.Text;
using System.Windows;
using Vela.Controls;
using Vela.Crypto;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.StarkDocuments;

namespace Vela.ApplicationBackend
{
	/// <summary>
	/// Application backend utilities.
	/// </summary>
	public static class ApplicationBackendUtils
	{
		#region Public static methods

		/// <summary>
		/// Gets the tooltip for the specified application notification.
		/// </summary>
		/// <param name="applicationNotification">Application notification for which to get the tooltip.</param>
		/// <param name="isGlobalPnrpCloudAvailable">If this is true then the global PNRP cloud is available (and so the local
		/// peer is visible on the Internet).</param>
		/// <returns>Returns the tooltip for the specified application notification. This may be null or empty.</returns>
		public static string GetApplicationNotificationTooltip( ApplicationNotifications applicationNotification
			, bool isGlobalPnrpCloudAvailable = true )
		{
			string applicationNotificationTooltip;

			switch( applicationNotification )
			{
				case ApplicationNotifications.None:
					applicationNotificationTooltip = "No UPN / service is started";
					break;
				case ApplicationNotifications.OneNotStarted:
					applicationNotificationTooltip = "At least 1 UPN / service is not started";
					break;
				case ApplicationNotifications.AllServicesStarted:
					applicationNotificationTooltip = "All services are started (PNRP is unused)";
					break;
				case ApplicationNotifications.AllStarted:
					applicationNotificationTooltip = "All UPNs and services are started";
					break;
				case ApplicationNotifications.NewLetter:
					applicationNotificationTooltip = "You have a new letter";
					break;
				default:
					throw ValidationUtils.Unhandled( applicationNotification, "applicationNotification" );
			}

			bool addPnrpAvailability = applicationNotification != ApplicationNotifications.AllServicesStarted;
			string globalPnrpCloudAvailability = (isGlobalPnrpCloudAvailable || !addPnrpAvailability) ? ""
				: " (No Internet PNRP)";

			return applicationNotificationTooltip + globalPnrpCloudAvailability;
		}

		/// <summary>
		/// Visits the URL extracted from the specified text.
		/// </summary>
		/// <param name="text">Text from which to extract the URL to open. This may be null or empty.</param>
		/// <param name="position">Position in "text" from which to extract the URL to open.</param>
		public static void VisitUrl( string text, int position )
		{
			string url = TextUtils.ExtractUrl( text, position );
			if( TextUtils.NoContent( url ) )
			{
				TooltipWindow.ShowTooltip( "There is no URL to visit.", TooltipWindowModes.Warning );

				return;
			}

			FileUtils.RunFile( url );
		}

		/// <summary>
		/// Runs the specified short command, using as command parameter the text extracted from the specified text.
		/// </summary>
		/// <param name="shortCommand">Short command to run. This may be null or empty.</param>
		/// <param name="allText">Text from which to extract the command parameter. This may be null or empty.</param>
		/// <param name="useAllText">If this is true then "allText" is used as command parameter. This has priority over
		/// "selectedText".</param>
		/// <param name="selectedText">Text selected within "allText". If this has content then it's used as command parameter.
		/// This has priority over "position".</param>
		/// <param name="position">Position in "allText" from which to extract the command parameter.</param>
		/// <remarks>
		/// <para>In the short command, the placeholder <see cref="ShortCommand.RawParameterPlaceholder"/> is replaced with the
		/// raw command parameter, and the placeholder <see cref="ShortCommand.UrlEscapedParameterPlaceholder"/> is replaced
		/// with the URL-escaped command parameter.</para>
		/// </remarks>
		public static void RunShortCommand( string shortCommand, string allText, bool useAllText, string selectedText
			, int position )
		{
			if( TextUtils.NoContent( shortCommand ) )
			{
				TooltipWindow.ShowTooltip( "There is no short command to run.", TooltipWindowModes.Warning );

				return;
			}

			bool hasParameterPlaceholder = TextUtils.Contains( shortCommand, ShortCommand.RawParameterPlaceholder )
				|| TextUtils.Contains( shortCommand, ShortCommand.UrlEscapedParameterPlaceholder );
			string commandParameter;

			if( useAllText )
			{
				commandParameter = allText;
			}
			else if( TextUtils.HasContent( selectedText ) )
			{
				commandParameter = selectedText;
			}
			else
			{
				commandParameter = TextUtils.ExtractWord( allText, position );
			}

			commandParameter = commandParameter.Trim();
			if( hasParameterPlaceholder && TextUtils.NoContent( commandParameter ) )
			{
				TooltipWindow.ShowTooltip( "There is no command parameter. Only the selected text or the letters from the"
					+ " caret position can be used as command parameter.", TooltipWindowModes.Warning );

				return;
			}

			shortCommand = TextUtils.Replace( shortCommand, ShortCommand.RawParameterPlaceholder, commandParameter );
			shortCommand = TextUtils.Replace( shortCommand, ShortCommand.UrlEscapedParameterPlaceholder
				, TextUtils.EscapeUrlParameter( commandParameter ) );

			FileUtils.RunSystemCommandWithParameters( shortCommand );
		}

		/// <summary>
		/// Attaches the specified file names to the specified list.
		/// </summary>
		/// <param name="fileNames">The names of the files to attach. This may be null or empty.</param>
		/// <param name="attachedFileNames">List of attached file names. The attached file names are added to this, in here. This
		/// may be empty, but not null.</param>
		/// <param name="attachedFilesSize">The size of the attached files (from the specified list). The size of the attached
		/// files is added to this, in here.</param>
		/// <returns>Returns the names of the files which were attached to the list. This may be null or empty.</returns>
		public static AttachedFileNames AttachFileNamesToPost( string[] fileNames, AttachedFileNames attachedFileNames
			, ref Int64 attachedFilesSize )
		{
			ValidationUtils.ObjectHasContent( attachedFileNames, "attachedFileNames" );

			if( ArrayUtils.NoContent( fileNames ) )
				return null;

			AttachedFileNames fileNamesToAttach = new AttachedFileNames();
			Int64 actualAttachedFilesSize = attachedFilesSize;

			// Collect all the file names from the attached folders.
			Strings childFileNames = new Strings();
			foreach( string fileName in fileNames )
			{
				childFileNames.Clear();

				if( Directory.Exists( fileName ) )
					FileUtils.CollectFileNamesFromFolder( fileName, childFileNames );
				else
					childFileNames.Add( fileName );

				string rootFolderName = Path.GetDirectoryName( fileName );

				foreach( string childFileName in childFileNames )
				{
					AttachedFileName attachedFileName = new AttachedFileName( rootFolderName, childFileName );
					AttachedFileName existingAttachedFileName = attachedFileNames[ attachedFileName.SortKey ];

					if( existingAttachedFileName != null )
					{
						throw new UserInteractionException( "The file may not be attached because its name conflicts with the"
							+ " name of an already attached file. The titles (= short names) of the attached files must all be"
							+ " different.", null
							, new ExceptionDetail( "Existing file name", existingAttachedFileName.AbsoluteFileName )
							, new ExceptionDetail( "New file name", childFileName ) );
					}

					string invalidReason = ValidationUtils.Text( childFileName, -1, -1, TextValidations.FileName
						, "childFileName", null, null, false );
					if( invalidReason != null )
					{
						throw new UserInteractionException( "The file name is invalid.", null
							, new ExceptionDetail( "File name", childFileName )
							, new ExceptionDetail( "Reason", invalidReason ) );
					}

					Int64 childFileSize = FileUtils.GetFileSize( childFileName, true );
					actualAttachedFilesSize += childFileSize;

					if( childFileSize > 0 )
						fileNamesToAttach.Add( attachedFileName );
				}
			}

			// Verify if the files are readable. Since this is a slow operation, only do it for a limited number of files.
			int i = 0;
			foreach( AttachedFileName attachedFileName in fileNamesToAttach )
			{
				if( i <= ApplicationBackendConsts.VerifyReadableFilesToAttachLimit )
					FileUtils.VerifyIsReadable( attachedFileName.AbsoluteFileName );

				i++;
			}

			// Clear the cached image previews because the files may have changed.
			ImagePreviewsCache.RemoveCached( childFileNames, false );
			ImagePreviewsCache.RemoveCached( childFileNames, true );

			attachedFilesSize = actualAttachedFilesSize;
			attachedFileNames.AddOnlyNew( fileNamesToAttach );

			return fileNamesToAttach;
		}

		/// <summary>
		/// Attaches file names to the specified list.
		/// </summary>
		/// <param name="allowMultiple">If this is true then multiple files may be attached.</param>
		/// <param name="attachedFileNames">List of attached file names. The attached file names are added to this, in here. This
		/// may be empty, but not null.</param>
		/// <param name="attachedFilesSize">The size of the attached files (from the specified list). The size of the attached
		/// files is added to this, in here.</param>
		/// <returns>Returns the names of the files which were attached to the list. This may be null or empty.</returns>
		public static AttachedFileNames AttachFileNamesToPost( bool allowMultiple, AttachedFileNames attachedFileNames
			, ref Int64 attachedFilesSize )
		{
			ValidationUtils.ObjectHasContent( attachedFileNames, "attachedFileNames" );

			OpenFileDialog ofd = new OpenFileDialog();
			ofd.Title = "Attach Files";
			ofd.AddExtension = false;
			ofd.Filter = FileExtensions.AllFilesFilter;
			ofd.Multiselect = allowMultiple;
			ofd.CheckFileExists = true;
			ofd.CheckPathExists = true;

			if( !ControlsUtils.ShowDialog( ofd ) )
				return null;

			AttachedFileNames fileNamesToAttach = AttachFileNamesToPost( ofd.FileNames, attachedFileNames
				, ref attachedFilesSize );

			return fileNamesToAttach;
		}

		/// <summary>
		/// Attaches the image from the clipboard to the specified list.
		/// </summary>
		/// <param name="attachedFileNames">List of attached file names. The attached file name is added to this, in here. This
		/// may be empty, but not null.</param>
		/// <param name="attachedFilesSize">The size of the attached files (from the specified list). The size of the attached
		/// file is added to this, in here.</param>
		/// <returns>Returns the names of the files which were attached to the list. This may be null or empty.</returns>
		public static AttachedFileNames AttachClipboardImageToPost( AttachedFileNames attachedFileNames
			, ref Int64 attachedFilesSize )
		{
			if( !Clipboard.ContainsImage() )
				return null;

			// Use a random file name in order to be able to attach several clipboard images to a single letter.
			string randomFileName = TextUtils.ArrayToText32( CryptoUtils.GetLocator().Value );
			string fileName = AssemblyUtils.GetApplicationTempFileName( ApplicationBackendConsts.AttachClipboardImageFolderName
				, randomFileName + ApplicationBackendConsts.AttachClipboardImageFileExtension );

			GraphicUtils.SaveImage( GraphicUtils.GetClipboardImage( ApplicationBackendConsts.AttachClipboardImageQuality )
				, fileName, ApplicationBackendConsts.AttachClipboardImageQuality );

			return AttachFileNamesToPost( new string[] { fileName }, attachedFileNames, ref attachedFilesSize );
		}

		/// <summary>
		/// Removes all the attached files from the specified list.
		/// </summary>
		/// <param name="attachedFileNames">List of attached file names. This is modified in here. This may be empty, but not
		/// null.</param>
		/// <param name="attachedFilesSize">The size of the attached files (from the specified list). This is modified in here.
		/// </param>
		public static void RemoveAttachedFilesFromPost( AttachedFileNames attachedFileNames, ref Int64 attachedFilesSize )
		{
			ValidationUtils.ObjectHasContent( attachedFileNames, "attachedFileNames" );

			string message = "Are you sure that you want to remove all the attached files?";
			if( MessageWindow.ShowDialog( message, MessageWindowModes.Query ) != QueryMessageResults.Yes )
				return;

			attachedFileNames.Clear();
			attachedFilesSize = 0;
		}

		/// <summary>
		/// Gets the specified file names as the content of a plain-text playlist.
		/// </summary>
		/// <param name="fileNames">Files name to get as the content of a playlist. This may be empty, but not null.</param>
		/// <returns>Returns the content of the playlist. This may be empty, but not null.</returns>
		public static string GetPlaylistContent( Strings fileNames )
		{
			ValidationUtils.ObjectHasContent( fileNames, "fileNames" );

			StringBuilder sb = new StringBuilder();

			foreach( string fileName in fileNames )
				sb.AppendLine( fileName );

			return sb.ToString();
		}

		/// <summary>
		/// Checks if the specified group image preview contains all the specified files.
		/// </summary>
		/// <param name="groupImagePreview">Group image preview. This may be null or empty.</param>
		/// <param name="localAttachedFiles">Local attached files to check if they are contained in the group image preview.
		/// This may be null.</param>
		/// <returns>Returns true if the specified group image preview contains all the specified files. If "localAttachedFiles"
		/// is null or empty then true is returned; if "localAttachedFiles" has content but "groupImagePreview" is null or empty
		/// then false is returned.</returns>
		public static bool GroupImagePreviewContainsAllFiles( GroupImagePreview groupImagePreview
			, LocalAttachedFiles localAttachedFiles )
		{
			if( ListUtils.NoContent( localAttachedFiles ) )
				return true;

			if( ListUtils.NoContent( groupImagePreview ) )
				return false;

			foreach( LocalAttachedFile localAttachedFile in localAttachedFiles )
			{
				if( !groupImagePreview.ContainsKey( localAttachedFile.LocalFileName ) )
					return false;
			}

			return true;
		}

		#endregion Public static methods
	}
}
