﻿////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Diagnostics;
using Vela.Kernel;

namespace Vela.ApplicationBackend
{
	/// <summary>
	/// Trace listener for binding errors, from XAML, which are normally captured by .Net.
	/// </summary>
	/// <remarks>
	/// <para>Call the <see cref="Initialize"/> method (at the start of the application) so that .Net can trace the binding
	/// errors.</para>
	/// 
	/// <para>The tracing of binding errors is activated only if <see cref="LogManager.LogDebugMessages"/> is true.</para>
	/// </remarks>
	public class BindingErrorsListener : TraceListener
	{
		#region Static constructors

		/// <summary>
		/// Initializes the current class.
		/// </summary>
		static BindingErrorsListener()
		{
			if( !LogManager.LogDebugMessages )
				return;

			PresentationTraceSources.Refresh();
		}

		#endregion Static constructors

		#region Public static methods

		/// <summary>
		/// Initializes the tracing of binding errors.
		/// </summary>
		/// <remarks>
		/// <para>Call this method (at the start of the application) so that .Net can trace the binding errors.</para>
		/// 
		/// <para>The tracing of binding errors is activated only if <see cref="LogManager.LogDebugMessages"/> is true.</para>
		/// </remarks>
		public static void Initialize()
		{
			if( !LogManager.LogDebugMessages )
				return;

			PresentationTraceSources.DataBindingSource.Switch.Level = SourceLevels.Error;
			PresentationTraceSources.DataBindingSource.Listeners.Add( new BindingErrorsListener() );
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Writes a message to the trace listener.
		/// </summary>
		/// <param name="message">Message to write.</param>
		public override void Write( string message )
		{
			// Nothing to do. The underlying system will write the message to the default trace listener.
		}

		/// <summary>
		/// Writes a message and a line break to the trace listener.
		/// </summary>
		/// <param name="message">Message to write.</param>
		public override void WriteLine( string message )
		{
			// Nothing to do. The underlying system will write the message to the default trace listener.
		}

		#endregion Public methods
	}
}
