////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.PeerDocuments;

namespace Vela.ApplicationBackend
{
	/// <summary>
	/// Cache for asynchronously loaded letter information items.
	/// </summary>
	public static class LetterAsyncItemsCache
	{
		#region Public constants

		/// <summary>
		/// Limit from which items are removed from the list before an item is added. This must be at least 1.
		/// </summary>
		public const int Limit = 10000;

		#endregion Public constants

		#region Private static fields

		/// <summary>
		/// The synchronization object for the <see cref="CachedItems"/> field. READ-ONLY.
		/// </summary>
		private static readonly object CachedItemsLock = new object();

		/// <summary>
		/// The cached items.
		/// </summary>
		private static LetterAsyncItems CachedItems = new LetterAsyncItems();

		#endregion Private static fields

		#region Public static methods

		/// <summary>
		/// Gets the cached item with the specified letter locator.
		/// </summary>
		/// <param name="letterLocator">The letter locator of the item to get.</param>
		/// <returns>Returns the cached item, or null if one isn't found.</returns>
		/// <remarks>
		/// <para>The <see cref="CachedItems"/> field is thread safe within this method.</para>
		/// </remarks>
		public static LetterAsyncItem Get( Locator letterLocator )
		{
			lock( CachedItemsLock )
			{
				return CachedItems[ letterLocator ];
			}
		}

		/// <summary>
		/// Adds or updates an existing item in the list, regardless of the value of the
		/// <see cref="KernelKeyedList{ TKey, TItem }.UpdateIfSameKey"/> property. If the list doesn't contain an item with the
		/// same key, it's added; if the list already contains "limit" items, (approximately) half of the existing items are
		/// removed from the list before the specified item is added. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="letterLocator">The letter locator of the item to add.</param>
		/// <param name="potentialImagePreviewFiles">The local attached files which can potentially be used for image previews.
		/// This may be null or empty.</param>
		/// <remarks>
		/// <para>This method will make room in the list of items, by moving (approximately) half of the existing items to keep
		/// into a new list.</para>
		/// </remarks>
		public static void AddOrUpdateWithLimit( Locator letterLocator, LocalAttachedFiles potentialImagePreviewFiles )
		{
			// Even if "potentialImagePreviewFiles" is null or empty, an item must still be added to indicate that the letter's
			// asynchronous information was loaded.

			LetterAsyncItem letterAsyncItem = new LetterAsyncItem( letterLocator, potentialImagePreviewFiles );

			lock( CachedItemsLock )
			{
				CachedItems.AddOrUpdateWithLimit( letterAsyncItem, Limit );
			}
		}

		/// <summary>
		/// Removes the cached item with the specified letter locator.
		/// </summary>
		/// <param name="letterLocator">The letter locator of the item to remove.</param>
		/// <returns>Returns the removed item, or null if none was removed.</returns>
		/// <remarks>
		/// <para>The <see cref="CachedItems"/> field is thread safe within this method.</para>
		/// </remarks>
		public static LetterAsyncItem Remove( Locator letterLocator )
		{
			lock( CachedItemsLock )
			{
				return CachedItems.Remove( letterLocator );
			}
		}

		/// <summary>
		/// Clears the cache.
		/// </summary>
		/// <remarks>
		/// <para>The <see cref="CachedItems"/> field is thread safe within this method.</para>
		/// </remarks>
		public static void Clear()
		{
			lock( CachedItemsLock )
			{
				CachedItems.Clear();
			}
		}

		#endregion Public static methods
	}
}
