////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Threading;
using Vela.ActionProcessor;
using Vela.ApplicationDocuments;
using Vela.Cloning;
using Vela.Controls;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Storehouse;

namespace Vela.ApplicationBackend
{
	/// <summary>
	/// Loader for the group image preview for a letter speckle (<see cref="LetterSpeckle"/>).
	/// </summary>
	/// <remarks>
	/// <para>When a page with letters is displayed, the speckles (with most information) are loaded synchronously. This allows
	/// the height of each list item to be synchronously determined (because a speckle contains the letter message and
	/// information about image previews).</para>
	/// 
	/// <para>Then, this loader loads synchronously the cached preview images, but if these are not available, a no-preview
	/// placeholder image is loaded. In the meantime, the loader starts a background thread to asynchronously the group image
	/// preview. When the thread finishes, it forces a rebind of the binding control.</para>
	/// </remarks>
	public class LetterGroupImagePreviewLoader
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected LetterGroupImagePreviewLoader()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		/// <param name="showPreviewsForAttachedImageFilesCount"> The number of attached image files for which image previews
		/// can be shown.</param>
		/// <param name="loadFullSize">If this is true then the full-sized image previews are loaded, else the thumb-sized image
		/// previews are loaded.</param>
		/// <param name="letterSpeckle">The letter speckle for which to load the group image preview.</param>
		/// <param name="bindingControl">The control which displays the group image preview.</param>
		/// <param name="dispatcher">The dispatcher used to synchronize cross-thread actions.</param>
		public LetterGroupImagePreviewLoader( IStorehouseProvider storehouseProvider, string rootSaveFolderName
			, int showPreviewsForAttachedImageFilesCount, bool loadFullSize, LetterSpeckle letterSpeckle
			, FrameworkElement bindingControl, Dispatcher dispatcher )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.TextHasContent( rootSaveFolderName, "rootSaveFolderName" );
			ValidationUtils.Integer( showPreviewsForAttachedImageFilesCount, 1
				, PeerDocumentsConsts.ShowPreviewsForAttachedImageFilesCountMax, "showPreviewsForAttachedImageFilesCount" );
			ValidationUtils.ObjectHasContent( letterSpeckle, "letterSpeckle" );
			ValidationUtils.ObjectHasContent( bindingControl, "bindingControl" );
			ValidationUtils.ObjectHasContent( dispatcher, "dispatcher" );

			StorehouseProvider = storehouseProvider;
			RootSaveFolderName = rootSaveFolderName;
			ShowPreviewsForAttachedImageFilesCount = showPreviewsForAttachedImageFilesCount;
			LoadFullSize = loadFullSize;
			LetterSpeckle = letterSpeckle;
			BindingControl = bindingControl;
			Dispatcher = dispatcher;
		}

		#endregion Constructors

		#region Private static fields

		/// <summary>
		/// Storehouse provider from which to create a storehouse transaction.
		/// </summary>
		private IStorehouseProvider StorehouseProvider;

		/// <summary>
		/// The name of the root folder where to save the attached files.
		/// </summary>
		private string RootSaveFolderName;

		/// <summary>
		/// The number of attached image files for which image previews can be shown.
		/// </summary>
		private int ShowPreviewsForAttachedImageFilesCount;

		/// <summary>
		/// If this is true then the full-sized image previews are loaded, else the thumb-sized image previews are loaded.
		/// </summary>
		private bool LoadFullSize;

		/// <summary>
		/// The speckle of the letter for which to load the information.
		/// </summary>
		private LetterSpeckle LetterSpeckle;

		/// <summary>
		/// The control which displays the letter.
		/// </summary>
		private FrameworkElement BindingControl;

		/// <summary>
		/// The dispatcher used to synchronize cross-thread actions.
		/// </summary>
		private Dispatcher Dispatcher;

		#endregion Private static fields

		#region Private static methods

		/// <summary>
		/// Gets the no-preview placeholder group image preview.
		/// </summary>
		/// <param name="loadFullSize">If this is true then the full-sized group image preview is loaded, else the thumb-sized
		/// group image preview is loaded.</param>
		/// <returns>Returns the no-preview placeholder group image preview.</returns>
		private static GroupImagePreview GetNoGroupImagePreview( bool loadFullSize )
		{
			ImagePreview noPreview = ImagePreviewsCache.GetNoPreview( loadFullSize );

			return new GroupImagePreview( noPreview );
		}

		/// <summary>
		/// Loads and caches the group image preview of the image files attached to the specified letter speckle.
		/// </summary>
		/// <param name="letterSpeckle">The letter speckle for which to load the group image preview.</param>
		/// <param name="potentialFilesForImagePreview">The local attached files which can potentially be used for image
		/// previews. This may be null or empty.</param>
		/// <param name="localAttachedFilesForPreviews">The attached files which can actually be used for image previews. This
		/// may be null or empty.</param>
		/// <param name="loadFullSize">If this is true then the full-sized image previews are loaded, else the thumb-sized image
		/// previews are loaded.</param>
		/// <returns>Returns the loaded group image preview.</returns>
		private static GroupImagePreview CacheGroupImagePreview( LetterSpeckle letterSpeckle
			, LocalAttachedFiles potentialFilesForImagePreview, LocalAttachedFiles localAttachedFilesForPreviews
			, bool loadFullSize )
		{
			ValidationUtils.ObjectHasContent( letterSpeckle, "letterSpeckle" );

			if( ListUtils.NoContent( potentialFilesForImagePreview ) )
				return GetNoGroupImagePreview( loadFullSize );

			string imageGroupIdentifier = ImagePreviewsCache.GetImageGroupIdentifier( letterSpeckle.StorehouseKey );
			GroupImagePreview groupImagePreview = new GroupImagePreview();

			// Cache all the usable files with no-preview image previews.
			if( ListUtils.HasContent( localAttachedFilesForPreviews ) )
			{
				try
				{
					// It's possible that the image previews cache is automatically cleared while these image previews are being
					// loaded, so lock this behavior. This is required so that when the UI bind is forced (after the loading),
					// all the image previews are available to be displayed; a partial display is unwanted.
					ImagePreviewsCache.LockAutomaticClear();

					foreach( LocalAttachedFile filesForPreview in localAttachedFilesForPreviews )
					{
						// Cache the image preview.
						ImagePreview imagePreview = ImagePreviewsCache.GetCachedOrLoad( filesForPreview.LocalFileName
							, imageGroupIdentifier, loadFullSize );

						groupImagePreview.Add( imagePreview );
					}
				}
				finally
				{
					ImagePreviewsCache.UnlockAutomaticClear();
				}
			}

			// Cache the remaining potential files with no-preview image previews. This ensures that the loader doesn't call
			// again the caching mechanism (for the same potential files), unless the no-preview cache items are manually removed
			// (for example, when new files are fully downloaded).
			try
			{
				// It's possible that the image previews cache is automatically cleared while these image previews are being
				// loaded, so lock this behavior. This is required so that when the UI bind is forced (after the loading), all
				// the image previews are available to be displayed; a partial display is unwanted.
				ImagePreviewsCache.LockAutomaticClear();

				foreach( LocalAttachedFile filesForPreview in potentialFilesForImagePreview )
				{
					// If the image preview is already cached, skip it.
					ImagePreview imagePreview = ImagePreviewsCache.GetCached( filesForPreview.LocalFileName
						, imageGroupIdentifier, loadFullSize );
					if( imagePreview != null )
						continue;

					ImagePreviewsCache.CacheNoPreview( filesForPreview.LocalFileName, imageGroupIdentifier, loadFullSize );
				}
			}
			finally
			{
				ImagePreviewsCache.UnlockAutomaticClear();
			}

			return (groupImagePreview.Count > 0) ? groupImagePreview : GetNoGroupImagePreview( loadFullSize );
		}

		/// <summary>
		/// Loads the specified letter speckle with all the information.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		/// <param name="showPreviewsForAttachedImageFilesCount"> The number of attached image files for which image previews
		/// can be shown.</param>
		/// <param name="loadFullSize">If this is true then the full-sized image previews are loaded, else the thumb-sized image
		/// previews are loaded.</param>
		/// <param name="letterSpeckle">The letter speckle for which to load the group image preview.</param>
		/// <param name="bindingControl">The control which displays the group image preview.</param>
		/// <param name="dispatcher">The dispatcher used to synchronize cross-thread actions.</param>
		private static void Load( IStorehouseProvider storehouseProvider, string rootSaveFolderName
			, int showPreviewsForAttachedImageFilesCount, bool loadFullSize, LetterSpeckle letterSpeckle
			, FrameworkElement bindingControl, Dispatcher dispatcher )
		{
			try
			{
				ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
				ValidationUtils.TextHasContent( rootSaveFolderName, "rootSaveFolderName" );
				ValidationUtils.Integer( showPreviewsForAttachedImageFilesCount, 1
					, PeerDocumentsConsts.ShowPreviewsForAttachedImageFilesCountMax, "showPreviewsForAttachedImageFilesCount" );
				ValidationUtils.ObjectHasContent( letterSpeckle, "letterSpeckle" );
				ValidationUtils.ObjectHasContent( bindingControl, "bindingControl" );
				ValidationUtils.ObjectHasContent( dispatcher, "dispatcher" );

				Letter letter = ApplicationBackendStorehouse.LoadLetter( storehouseProvider, letterSpeckle.StorehouseKey );
				if( letter == null )
					return;

				// Get the attached files which can potentially be used for image previews.
				LocalAttachedFiles potentialFilesForImagePreview = AttachedFilesUtils.GetFilesForImagePreview(
					letter.AttachedFiles, rootSaveFolderName, showPreviewsForAttachedImageFilesCount, letterSpeckle.StorehouseKey
					, letterSpeckle.CreationTime, letterSpeckle.LocalPeerDirection, false );

				// Cache the attached files which can potentially be used for image previews.
				LetterAsyncItemsCache.AddOrUpdateWithLimit( letterSpeckle.StorehouseKey, potentialFilesForImagePreview );

				// Get the attached files which actually can be used for image previews.
				LocalAttachedFiles localAttachedFilesForPreviews = AttachedFilesUtils.GetFilesForImagePreview(
					letter.AttachedFiles, rootSaveFolderName, showPreviewsForAttachedImageFilesCount, letterSpeckle.StorehouseKey
					, letterSpeckle.CreationTime, letterSpeckle.LocalPeerDirection, true );

				// Cache the group image preview so that it can be found at the forced UI bind which is triggered below.
				CacheGroupImagePreview( letterSpeckle, potentialFilesForImagePreview, localAttachedFilesForPreviews
					, loadFullSize );

				// Force a UI bind in order to display the cached group image preview. Process on the UI thread.
				dispatcher.Invoke( () => ControlsUtils.UpdateBindingTargets( bindingControl ), DispatcherPriority.Background );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				// Capture all exceptions because this method is running on background thread, so there is nowhere for them to
				// go.
				LogManager.Add( ex );
			}
		}

		#endregion Private static methods

		#region Public methods

		/// <summary>
		/// Gets the cached group image preview. If one is not available, it starts its asynchronous loading.
		/// </summary>
		/// <returns>Returns the cached group image preview. This may be null or empty.</returns>
		public GroupImagePreview Load()
		{
			if( !LetterSpeckle.HasImagePreview )
				return null;

			// Get the cached attached files which can potentially be used for image previews.
			LetterAsyncItem letterAsyncItem = LetterAsyncItemsCache.Get( LetterSpeckle.StorehouseKey );

			// Get the cached image previews (including the ones with no-preview image previews).
			string imageGroupIdentifier = ImagePreviewsCache.GetImageGroupIdentifier( LetterSpeckle.StorehouseKey );
			GroupImagePreview groupImagePreview = ImagePreviewsCache.GetCachedGroupImage( imageGroupIdentifier, true
				, LoadFullSize );

			// Check if all the potential image files have cached image previews.
			if( (letterAsyncItem != null) /* If a letter async item was cached. */
				&& ApplicationBackendUtils.GroupImagePreviewContainsAllFiles( groupImagePreview
					, letterAsyncItem.PotentialImagePreviewFiles ) )
			{
				return groupImagePreview;
			}

			IStorehouseProvider storehouseProvider = Cloner.Deep<IStorehouseProvider>( StorehouseProvider );
			LetterSpeckle letterSpeckle = Cloner.Deep<LetterSpeckle>( LetterSpeckle );

			#region Load the information on a different thread

			Action loaderAction = new Action(
				() =>
				{
					try
					{
						Load( storehouseProvider, RootSaveFolderName, ShowPreviewsForAttachedImageFilesCount, LoadFullSize
							, letterSpeckle, BindingControl, Dispatcher );
					}
					catch( TaskCanceledException )
					{
						// Nothing to do.
					}
				} );

			Task taskLoad = Task.Run( loaderAction );

			#endregion Load the information on a different thread

			// Return the existing group image preview, for synchronous binding.
			return ListUtils.HasContent( groupImagePreview ) ? groupImagePreview : GetNoGroupImagePreview( LoadFullSize );
		}

		#endregion Public methods
	}
}
