////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Drawing;
using System.Reflection;
using System.Windows;
using System.Windows.Forms;
using System.Windows.Shell;
using Vela.Controls;
using Vela.Kernel;

namespace Vela.ApplicationBackend
{
	/// <summary>
	/// Application notification manager.
	/// </summary>
	public class ApplicationNotificationManager
	{
		#region Public constants

		/// <summary>
		/// Template used to create the name of a resource file which contain an icon used for application notifications.
		/// Placeholders: 0 - resource prefix, 1 - notification.
		/// </summary>
		public const string ResourceNameTemplate = "{0}{1}.ico";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="resourceAssembly">The assembly which contains the resource files which contain the icons used for
		/// application notifications.</param>
		/// <param name="resourcePrefix">The prefix used to get the resource files which contain the icons used for application
		/// notifications.</param>
		public ApplicationNotificationManager( Assembly resourceAssembly, string resourcePrefix )
		{
			ResourceAssembly = resourceAssembly;
			ResourcePrefix = resourcePrefix;

			NotificationIcon = new NotifyIcon();
			NotificationIcon.MouseClick += new MouseEventHandler( NotificationIcon_MouseClick );
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The notification icon.
		/// </summary>
		private NotifyIcon NotificationIcon;

		/// <summary>
		/// A type in the resource assembly which is used to get a resource icon.
		/// </summary>
		private Type TypeInResourceAssembly;

		/// <summary>
		/// The time when the last notification was set.
		/// </summary>
		private DateTime LastNotificationTime = DateTime.MinValue;

		#endregion Private fields

		#region Public non-serializable properties

		private Assembly _ResourceAssembly;
		/// <summary>
		/// The assembly which contains the resource files which contain the icons used for application notifications. READ-ONLY.
		/// </summary>
		public Assembly ResourceAssembly
		{
			get { return _ResourceAssembly; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ResourceAssembly" );

				_ResourceAssembly = value;

				#region Set the "TypeInResourceAssembly"

				Type[] types = _ResourceAssembly.GetTypes();

				if( ArrayUtils.NoContent( types ) )
				{
					throw new ValidationException( string.Format( "The resource assembly \"{0}\" contains no types."
						, _ResourceAssembly.FullName ) );
				}

				TypeInResourceAssembly = types[ 0 ];

				#endregion Set the "TypeInResourceAssembly"
			}
		}

		private string _ResourcePrefix;
		/// <summary>
		/// The prefix used to get the resource files which contain the icons used for application notifications. READ-ONLY.
		/// </summary>
		public string ResourcePrefix
		{
			get { return _ResourcePrefix; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "ResourcePrefix" );

				_ResourcePrefix = value;
			}
		}

		private ApplicationNotifications _Notification = ApplicationNotifications.None;
		/// <summary>
		/// The current application notification. READ-ONLY.
		/// </summary>
		public ApplicationNotifications Notification
		{
			get { return _Notification; }
			protected set { _Notification = value; }
		}

		private string _Tooltip;
		/// <summary>
		/// The tooltip for the current application notification. This may be null or empty. READ-ONLY.
		/// </summary>
		public string Tooltip
		{
			get { return _Tooltip; }
			protected set { _Tooltip = value; }
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void NotificationIcon_MouseClick( object sender, MouseEventArgs e )
		{
			ControlsUtils.SetApplicationVisibility( ApplicationVisibilities.Normal );
		}

		#endregion Event handlers

		#region Private static methods

		/// <summary>
		/// Sets the specified notification in the application's button in the system's taskbar.
		/// </summary>
		/// <param name="notification">Notification to set.</param>
		private static void NotifyInTaskbar( ApplicationNotifications notification )
		{
			if( (notification == ApplicationNotifications.NewLetter)
				&& (ControlsUtils.GetApplicationVisibility() == ApplicationVisibilities.Hidden) )
			{
				ControlsUtils.SetApplicationVisibility( ApplicationVisibilities.Minimized );
			}

			// Now there should be a main window.

			Window mainWindow = ControlsUtils.GetAppMainWindow();
			if( (mainWindow == null) || (mainWindow.TaskbarItemInfo == null) ) return;

			mainWindow.TaskbarItemInfo.ProgressState = (notification == ApplicationNotifications.NewLetter)
				? TaskbarItemProgressState.Paused : TaskbarItemProgressState.None;
			mainWindow.TaskbarItemInfo.ProgressValue = 1;
		}

		#endregion Private static methods

		#region Public methods

		/// <summary>
		/// Closes the current tray icon.
		/// </summary>
		public void Close()
		{
			NotificationIcon.Dispose();
			NotificationIcon = null;
		}

		/// <summary>
		/// Sets the notification of the application.
		/// </summary>
		/// <param name="notification">Notification to set. If this has the same value as the current notification
		/// (<see cref="Notification"/>), nothing happens.</param>
		/// <param name="notificationTime">The time when the specified notification has occurred.</param>
		/// <param name="tooltip">The tooltip for the current application notification. This may be null or empty.</param>
		public void Notify( ApplicationNotifications notification, DateTime notificationTime, string tooltip )
		{
			if( (Notification == notification) && (LastNotificationTime >= notificationTime) ) return;

			Notification = notification;
			LastNotificationTime = notificationTime;
			Tooltip = tooltip;

			NotifyInTaskbar( Notification );

			if( NotificationIcon == null ) return;

			Icon previousIcon = NotificationIcon.Icon;

			if( Notification == ApplicationNotifications.None )
			{
				NotificationIcon.Icon = null;
				NotificationIcon.Visible = false;
			}
			else
			{
				string resourceName = string.Format( ResourceNameTemplate, ResourcePrefix, Notification );

				NotificationIcon.Icon = new Icon( TypeInResourceAssembly, resourceName );
				NotificationIcon.Text = Tooltip;
				NotificationIcon.Visible = true;
			}

			if( previousIcon != null ) previousIcon.Dispose();
		}

		/// <summary>
		/// Ensures that the <see cref="ApplicationNotifications.NewLetter"/> notification is shown in the application's button
		/// in the system's taskbar.
		/// </summary>
		public void EnsureNewLetterNotificationInTaskbar()
		{
			if( ControlsUtils.GetApplicationVisibility() != ApplicationVisibilities.Hidden ) return;

			ControlsUtils.SetApplicationVisibility( ApplicationVisibilities.Minimized );

			// Now there should be a main window.

			Window mainWindow = ControlsUtils.GetAppMainWindow();
			if( (mainWindow == null) || (mainWindow.TaskbarItemInfo == null) ) return;

			mainWindow.TaskbarItemInfo.ProgressState = TaskbarItemProgressState.Paused;
			mainWindow.TaskbarItemInfo.ProgressValue = 1;
		}

		#endregion Public methods
	}
}
