////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using System.Windows.Media;
using Vela.Kernel;

namespace Vela.ApplicationBackend
{
	/// <summary>
	/// Audio notification manager.
	/// </summary>
	public static class AudioNotificationManager
	{
		#region Public constants

		/// <summary>
		/// An audio notification is stopped only if this number of seconds has passed after the sound has started playing.
		/// </summary>
		public const int MustPlayBeforeStop = 3;

		#endregion Public constants

		#region Private static fields

		/// <summary>
		/// The synchronization object for the <see cref="MediaPlayer"/> property. READ-ONLY.
		/// </summary>
		private static readonly object MediaPlayerLock = new object();

		/// <summary>
		/// If this is true then the media player is playing.
		/// </summary>
		private static bool IsPlaying;

		/// <summary>
		/// The time (in UTC format) when the playing of the audio notification has started.
		/// </summary>
		private static DateTime PlayingStartTime = DateTime.MinValue;

		#endregion Private static fields

		#region Private static properties

		private static MediaPlayer _MediaPlayer;
		/// <summary>
		/// The global media player used for audio notifications. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// 
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		private static MediaPlayer MediaPlayer
		{
			get
			{
				lock( MediaPlayerLock )
				{
					if( _MediaPlayer != null ) return _MediaPlayer;

					_MediaPlayer = new MediaPlayer();
					_MediaPlayer.MediaEnded += MediaPlayer_MediaEnded;

					return _MediaPlayer;
				}
			}
		}

		#endregion Private static properties

		#region Event handlers

		private static void MediaPlayer_MediaEnded( object sender, EventArgs e )
		{
			IsPlaying = false;
			PlayingStartTime = DateTime.MinValue;
		}

		#endregion Event handlers

		#region Public static methods

		/// <summary>
		/// Plays an audio notification using the sound from the specified file name.
		/// </summary>
		/// <param name="fileName">The name of the file which contains the sound to play. This may be null or empty.</param>
		/// <param name="throwIfError">If this is true and there is an error playing the file, an exception is thrown.</param>
		public static void Play( string fileName, bool throwIfError )
		{
			if( TextUtils.NoContent( fileName ) || !File.Exists( fileName ) ) return;

			try
			{
				MediaPlayer.Close();
				MediaPlayer.Open( new Uri( fileName ) );

				IsPlaying = true;
				PlayingStartTime = DateTime.UtcNow;

				MediaPlayer.Play();
			}
			catch( Exception ex )
			{
				IsPlaying = false;
				PlayingStartTime = DateTime.MinValue;

				if( throwIfError ) throw ExceptionUtils.PreserveStackTrace( ex );
				else LogManager.Add( ex );
			}
		}

		/// <summary>
		/// Stops playing the current notification.
		/// </summary>
		/// <param name="force">If this is true then the audio notification is stopped immediately, else it's stopped only
		/// it at least <see cref="MustPlayBeforeStop"/> seconds have passed (after start).</param>
		/// <remarks>
		/// <para>An audio notification stops automatically. Call this method only if you want to manually stop the current
		/// notification.</para>
		/// </remarks>
		public static void Stop( bool force )
		{
			if( !IsPlaying ) return;

			DateTime now = DateTime.UtcNow;
			if( !force && ((now >= PlayingStartTime) && (now < PlayingStartTime.AddSeconds( MustPlayBeforeStop ))) ) return;

			IsPlaying = false;
			PlayingStartTime = DateTime.MinValue;

			MediaPlayer.Close();
		}

		#endregion Public static methods
	}
}
