////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using System.Windows;
using Vela.ActionProcessor;
using Vela.Controls;
using Vela.Kernel;
using Vela.PeerDocuments;

namespace Vela.ApplicationBackend
{
	/// <summary>
	/// Application wallpaper utilities.
	/// </summary>
	public static class ApplicationWallpaperUtils
	{
		#region Public constants

		/// <summary>
		/// If the wallpaper image scaled with the "fit" mode to the screen occupies at least this fraction of the screen, the
		/// wallpaper is set and the scaling mode is "fill".
		/// </summary>
		public const double FillScalingModeCoverageMin = 0.80;

		/// <summary>
		/// The compression quality of the resized image. This must be at least 1.
		/// </summary>
		public const int WallpaperQuality = 90;

		/// <summary>
		/// The title of the file which contains the wallpaper set by the application.
		/// </summary>
		public const string WallpaperFileTitle = "Wallpaper.jpg";

		#endregion Public constants

		#region Public static methods

		/// <summary>
		/// Gets the text to write on a wallpaper from the specified letter message summary.
		/// </summary>
		/// <param name="letterMessageSummary">The letter message summary from which to delete the specified emoticon symbol and
		/// return. This may be null or empty.</param>
		/// <param name="emoticonSymbol">The emoticon symbol to delete. This may be null or empty.</param>
		/// <param name="attachedFiles">The files attached to the letter, files whose names to delete from the returned text.
		/// This may be null or empty.</param>
		/// <returns>Returns the text to write on a wallpaper.</returns>
		public static string GetTextToWriteOnImage( string letterMessageSummary, string emoticonSymbol
			, AttachedFiles attachedFiles )
		{
			if( TextUtils.NoContent( letterMessageSummary ) )
				return letterMessageSummary;

			string textToWriteOnImage = letterMessageSummary;

			// Delete the emoticon symbol.
			if( TextUtils.HasContent( emoticonSymbol ) )
				textToWriteOnImage = TextUtils.Replace( textToWriteOnImage, emoticonSymbol, "" );

			if( ListUtils.NoContent( attachedFiles ) )
				return textToWriteOnImage;

			// Delete the file names.
			foreach( AttachedFile attachedFile in attachedFiles )
			{
				string fileName = Path.GetFileName( attachedFile.RelativeFileName );
				textToWriteOnImage = TextUtils.Replace( textToWriteOnImage, fileName, "" );
			}

			// Replace double spaces with a single one.
			textToWriteOnImage = textToWriteOnImage.Replace( "  ", " " );

			return textToWriteOnImage.Trim();
		}

		/// <summary>
		/// Sets the system's wallpaper with the first appropriate attached file from the specified list.
		/// </summary>
		/// <param name="attachedFiles">Attached files. If this is null or empty, nothing happens.</param>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		/// <param name="accessGranterLocator">The locator of the document which has offered the attached files.</param>
		/// <param name="accessGranterCreationTime">The creation time of the document which has offered the attached files.
		/// </param>
		/// <param name="localPeerDirection">The direction of the peer who has received the attached files.</param>
		/// <param name="textToWriteOnImage">The text to write on the wallpaper image. This may be null or empty.</param>
		/// <param name="textOnImageConfiguration">Configuration for the text which is to be written on the resized image.
		/// </param>
		/// <returns>Returns true if the wallpaper was set.</returns>
		/// <remarks>
		/// <para>The used file is the first file whose extension is in the <see cref="FileExtensions.ImageFile"/> list.
		/// </para>
		/// </remarks>
		public static bool Set( AttachedFiles attachedFiles, string rootSaveFolderName, Locator accessGranterLocator
			, DateTime accessGranterCreationTime, LocalPeerDirections localPeerDirection, string textToWriteOnImage
			, TextOnImageConfiguration textOnImageConfiguration )
		{
			LocalAttachedFiles localAttachedFiles = AttachedFilesUtils.GetFilesForImagePreview( attachedFiles
				, rootSaveFolderName, 1, accessGranterLocator, accessGranterCreationTime, localPeerDirection, true );

			string fileName = LocalAttachedFiles.GetFirstFileName( localAttachedFiles );
			if( TextUtils.NoContent( fileName ) )
				return false;

			string wallpaperFileName = AssemblyUtils.GetApplicationTempFileName( null, WallpaperFileTitle );

			GraphicUtils.SaveResizedImage( fileName, (int)SystemParameters.PrimaryScreenWidth
				, (int)SystemParameters.PrimaryScreenHeight, FillScalingModeCoverageMin, WallpaperQuality, textToWriteOnImage
				, textOnImageConfiguration, SystemParameters.WorkArea, wallpaperFileName );

			SystemWallpaperUtils.Set( wallpaperFileName );

			return true;
		}

		#endregion Public static methods
	}
}
