////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows;
using Vela.Controls;
using Vela.Kernel;
using Vela.StarkDocuments;
using Tuple = Vela.Kernel.Tuple;

namespace Vela.ApplicationBackend
{
	/// <summary>
	/// Window for editing a short command (<see cref="ShortCommand"/>).
	/// </summary>
	public partial class EditShortCommandWindow : ViewWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected EditShortCommandWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="shortCommand">Short command to edit.</param>
		/// <param name="originalKey">The list key of the short command to edit. This is null if the short command is being
		/// added.</param>
		/// <param name="shortCommands">Short commands whose display names to avoid. This may be empty, but not null.</param>
		public EditShortCommandWindow( ShortCommand shortCommand, Tuple originalKey, ShortCommands shortCommands )
		{
			ShortCommand = shortCommand;
			OriginalKey = originalKey;
			ShortCommands = shortCommands;

			InitializeComponent();

			LoadUi();
		}

		#endregion Constructors

		#region Public non-serializable properties

		private ShortCommand _ShortCommand;
		/// <summary>
		/// The short command to edit. Notifies property observers. READ-ONLY.
		/// </summary>
		public ShortCommand ShortCommand
		{
			get { return _ShortCommand; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ShortCommand" );

				_ShortCommand = value;

				// Notify property observers.
				OnPropertyChanged( "ShortCommand" );
			}
		}

		private Tuple _OriginalKey;
		/// <summary>
		/// The list key of the short command to edit. This is null if the short command is being added. Notifies property
		/// observers. READ-ONLY.
		/// </summary>
		public Tuple OriginalKey
		{
			get { return _OriginalKey; }
			protected set
			{
				_OriginalKey = value;

				// Notify property observers.
				OnPropertyChanged( "OriginalKey" );
			}
		}

		private ShortCommands _ShortCommands;
		/// <summary>
		/// The short commands whose display names to avoid. This may be empty, but not null. Notifies property observers.
		/// READ-ONLY.
		/// </summary>
		public ShortCommands ShortCommands
		{
			get { return _ShortCommands; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ShortCommands" );

				_ShortCommands = value;

				// Notify property observers.
				OnPropertyChanged( "ShortCommands" );
			}
		}

		private DisplayShortCommandShortcutKeys _DisplayShortcutKeys = DisplayShortCommandShortcutKeys.GetInitializedInstance();
		/// <summary>
		/// The display shortcut keys. The backing field is initialized by default. Notifies property observers. READ-ONLY.
		/// </summary>
		public DisplayShortCommandShortcutKeys DisplayShortcutKeys
		{
			get { return _DisplayShortcutKeys; }
			protected set
			{
				ValidationUtils.ListHasContent( value, "DisplayShortcutKeys" );

				_DisplayShortcutKeys = value;

				// Notify property observers.
				OnPropertyChanged( "DisplayShortcutKeys" );
				OnPropertyChanged( "CurrentDisplayShortcutKey" );
			}
		}

		private DisplayShortCommandShortcutKey _CurrentDisplayShortcutKey;
		/// <summary>
		/// The current display shortcut key. This may be null. Notifies property observers.
		/// </summary>
		public DisplayShortCommandShortcutKey CurrentDisplayShortcutKey
		{
			get { return _CurrentDisplayShortcutKey; }
			set
			{
				_CurrentDisplayShortcutKey = value;

				// Notify property observers.
				OnPropertyChanged( "CurrentDisplayShortcutKey" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void ViewWindow_Loaded( object sender, RoutedEventArgs e )
		{
			ControlsUtils.FocusTextBox( DisplayNameTextBox, true );
		}

		private void OkayButton_Click( object sender, RoutedEventArgs e )
		{
			try
			{
				ValidateUi( ErrorPresenter );
				SaveUi( ErrorPresenter );

				DialogResult = true;
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) ) throw ExceptionUtils.PreserveStackTrace( ex );

				ControlsUtils.ShowException( ex );
			}
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Validates the input data.
		/// </summary>
		/// <param name="errorPresenter">The control which shows that there is a validation error. This may be null.</param>
		/// <remarks>
		/// <para>If you don't validate of all the editable fields in here, it's possible that some bindings are commited even
		/// if the user cancels the edit operation. This can happen if the user edits some fields, clicks the "Okay" button, the
		/// validation fails (during binding commit) inside some properties (which are not validated in here), and then the user
		/// clicks the "Cancel" button. This is likely to matter only if those (successfully) bound properties are used outside
		/// the view.</para>
		/// </remarks>
		protected override void ValidateUi( FrameworkElement errorPresenter )
		{
			base.ValidateUi( errorPresenter );

			string displayName = DisplayNameTextBox.Text;
			string category = CategoryTextBox.Text;

			ValidationUtils.Text( displayName, -1, -1, TextValidations.EntityName, "Display name", DisplayNameTextBox
				, errorPresenter );
			ValidationUtils.Text( category, -1, -1, TextValidations.EntityName, "Category", CategoryTextBox, errorPresenter );
			ValidationUtils.Text( CommandTextBox.Text, -1, -1, TextValidations.ShortCommand, "Command", CommandTextBox
				, errorPresenter );

			Tuple key = ShortCommand.GetSortKey( displayName, category );
			ShortCommand existingShortCommand = ShortCommands[ key ];
			if( (existingShortCommand != null) && !existingShortCommand.SortKey.Equals( OriginalKey ) )
			{
				throw new PresentedValidationException( "The [Display name - Category] is already used.", null, "Display name"
					, DisplayNameTextBox, errorPresenter );
			}
		}

		/// <summary>
		/// Loads the user interface with the data to edit.
		/// </summary>
		protected override void LoadUi()
		{
			base.LoadUi();

			CurrentDisplayShortcutKey = DisplayShortcutKeys.FindByShortcutKey( ShortCommand.ShortcutKey );
		}

		/// <summary>
		/// Saves the user interface data into the data to edit.
		/// </summary>
		/// <param name="errorPresenter">The control which shows that there is a validation error. This may be null.</param>
		protected override void SaveUi( FrameworkElement errorPresenter )
		{
			base.SaveUi( errorPresenter );

			ShortCommand.ShortcutKey = CurrentDisplayShortcutKey.ShortcutKey;
		}

		#endregion Private methods
	}
}
