////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using Vela.Controls;
using Vela.Kernel;
using Vela.StarkDocuments;

namespace Vela.ApplicationBackend
{
	/// <summary>
	/// Window for viewing the available emoticons.
	/// </summary>
	public partial class SelectEmoticonWindow : ViewWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected SelectEmoticonWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="configuration">The emoticon manager configuration.</param>
		/// <param name="showSensualDisplayCategory">If this is true then the display category with sensual emoticons is shown
		/// to the user.</param>
		/// <param name="updateWithCustomEmoticons">Event raised when the emoticon configuration has to be updated with the
		/// custom emoticons specified by the user. This may be null.</param>
		public SelectEmoticonWindow( EmoticonManagerConfiguration configuration, bool showSensualDisplayCategory
			, Action<string[]> updateWithCustomEmoticons )
		{
			ShowSensualDisplayCategory = showSensualDisplayCategory; // This must be set before "Configuration".
			Configuration = configuration;
			UpdateWithCustomEmoticons = updateWithCustomEmoticons;

			InitializeComponent();

			LoadUi();
		}

		#endregion Constructors

		#region Private static fields

		/// <summary>
		/// Configuration for the layout of the window.
		/// </summary>
		private static WindowLayoutConfiguration WindowLayoutConfiguration;

		/// <summary>
		/// The current display category last viewed when the window was closed.
		/// </summary>
		private static string LastCurrentDisplayCategory;

		#endregion Private static fields

		#region Private fields

		/// <summary>
		/// Event raised when the emoticon configuration has to be updated with the custom emoticons specified by the user. This
		/// may be null.
		/// </summary>
		private Action<string[]> UpdateWithCustomEmoticons;

		#endregion Private fields

		#region Public non-serializable properties

		private EmoticonManagerConfiguration _Configuration;
		/// <summary>
		/// The emoticon manager configuration. Notifies property observers. READ-ONLY.
		/// </summary>
		public EmoticonManagerConfiguration Configuration
		{
			get { return _Configuration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Configuration" );

				_Configuration = value;

				string displayCategoryToHide = ShowSensualDisplayCategory ? null : Configuration.SensualDisplayCategory;
				DisplayCategories = Configuration.SymbolDescriptors.CollectDisplayCategories( displayCategoryToHide );

				CurrentDisplayCategory = TextUtils.HasContent( LastCurrentDisplayCategory )
					? LastCurrentDisplayCategory : Configuration.DisplayCategoryToShowFirst;

				// Notify property observers.
				OnPropertyChanged( "Configuration" );
			}
		}

		private bool _ShowSensualDisplayCategory;
		/// <summary>
		/// If this is true then the display category with sensual emoticons is shown to the user. READ-ONLY.
		/// </summary>
		public bool ShowSensualDisplayCategory
		{
			get { return _ShowSensualDisplayCategory; }
			protected set { _ShowSensualDisplayCategory = value; }
		}

		private EmoticonSymbolDescriptors _DisplaySymbolDescriptors;
		/// <summary>
		/// The emoticon symbol descriptors which are to be displayed. This may be null or empty. Notifies property observers.
		/// READ-ONLY.
		/// </summary>
		public EmoticonSymbolDescriptors DisplaySymbolDescriptors
		{
			get { return _DisplaySymbolDescriptors; }
			protected set
			{
				_DisplaySymbolDescriptors = value;

				// Notify property observers.
				OnPropertyChanged( "DisplaySymbolDescriptors" );
			}
		}

		private Strings _DisplayCategories;
		/// <summary>
		/// The display categories of the emoticon images. This may be null or empty. Notifies property observers. READ-ONLY.
		/// </summary>
		public Strings DisplayCategories
		{
			get { return _DisplayCategories; }
			protected set
			{
				_DisplayCategories = value;

				// Notify property observers.
				OnPropertyChanged( "DisplayCategories" );
			}
		}

		private string _CurrentDisplayCategory;
		/// <summary>
		/// The current display category of the emoticon images to show.
		/// </summary>
		public string CurrentDisplayCategory
		{
			get { return _CurrentDisplayCategory; }
			set
			{
				ValidationUtils.TextHasContent( value, "CurrentDisplayCategory" );

				if( TextUtils.Equals( CurrentDisplayCategory, value ) )
					return;

				_CurrentDisplayCategory = value;

				// Notify property observers.
				OnPropertyChanged( "CurrentDisplayCategory" );

				DisplaySymbolDescriptors = Configuration.SymbolDescriptors.CollectByDisplayCategory( CurrentDisplayCategory );

				ControlsUtils.FocusListBoxItem( EmoticonsListBox );
			}
		}

		private string _SelectedSymbol;
		/// <summary>
		/// The selected emoticon symbol. This may be null or empty. READ-ONLY.
		/// </summary>
		public string SelectedSymbol
		{
			get { return _SelectedSymbol; }
			protected set { _SelectedSymbol = value; }
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void ViewWindow_Loaded( object sender, System.Windows.RoutedEventArgs e )
		{
			// Once the size has been set, ignore future resizes. This is because the first category contains the most images, so
			// the window has its greatest height when it's first shown.
			SizeToContent = System.Windows.SizeToContent.Manual;

			if( WindowLayoutConfiguration == null )
				ControlsUtils.CenterWindowOnScreen( this );
			WindowLayoutConfiguration = WindowLayoutConfiguration.LoadUi( this, WindowLayoutConfiguration );

			ControlsUtils.FocusListBoxItem( EmoticonsListBox );
		}

		private void ViewWindow_Closing( object sender, System.ComponentModel.CancelEventArgs e )
		{
			if( WindowLayoutConfiguration != null )
				WindowLayoutConfiguration.SaveUi();
			LastCurrentDisplayCategory = CurrentDisplayCategory;
		}

		private void ViewWindow_PreviewKeyDown( object sender, KeyEventArgs e )
		{
			if( Keyboard.Modifiers == ModifierKeys.None )
			{
				// Accept the focused emoticon.
				if( e.Key == Key.Enter )
				{ AcceptFocusedEmoticon(); e.Handled = true; return; }
			}

			if( Keyboard.Modifiers == ModifierKeys.Control )
			{
				// Show the next display category.
				if( e.Key == Key.Tab )
				{ ShowNextDisplayCategory(); e.Handled = true; return; }
				else if( e.Key == Key.E )
				{ UpdateEmoticons( null ); e.Handled = true; return; }
			}

			if( Keyboard.Modifiers == (ModifierKeys.Control | ModifierKeys.Shift) )
			{
				// Show the previous display category.
				if( e.Key == Key.Tab )
				{ ShowPreviousDisplayCategory(); e.Handled = true; return; }
			}
		}

		private void EmoticonsListBox_MouseDoubleClick( object sender, MouseButtonEventArgs e )
		{
			// The user must click either an emoticon image or its symbol.
			if( (e.OriginalSource as Image == null) && (e.OriginalSource as TextBlock == null) )
				return;

			AcceptFocusedEmoticon();
		}

		private void EmoticonsListBox_PreviewDragEnter( object sender, DragEventArgs e )
		{
			if( MayDropFileNames( e.Data ) )
				e.Effects = DragDropEffects.Copy;
		}

		private void EmoticonsListBox_PreviewDragOver( object sender, DragEventArgs e )
		{
			e.Handled = true;
		}

		private void EmoticonsListBox_Drop( object sender, DragEventArgs e )
		{
			try
			{
				DropFileNames( e.Data );
			}
			catch( Exception ex )
			{
				// The operating system swallows drag-and-drop exceptions, so show them before that happens.
				ControlsUtils.ShowException( ex );
			}
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Shows the next display category.
		/// </summary>
		private void ShowNextDisplayCategory()
		{
			CurrentDisplayCategory = DisplayCategories.FindNext( CurrentDisplayCategory, true );
		}

		/// <summary>
		/// Shows the previous display category.
		/// </summary>
		private void ShowPreviousDisplayCategory()
		{
			CurrentDisplayCategory = DisplayCategories.FindPrevious( CurrentDisplayCategory, true );
		}

		/// <summary>
		/// Accepts the focused emoticon as the selected one and closes the window.
		/// </summary>
		private void AcceptFocusedEmoticon()
		{
			EmoticonSymbolDescriptor item = EmoticonsListBox.SelectedItem as EmoticonSymbolDescriptor;
			if( item == null )
				return;

			SelectedSymbol = item.Symbol;

			DialogResult = true;
		}

		/// <summary>
		/// Updates the emoticons with the custom ones.
		/// </summary>
		/// <param name="fileNames">The name of the files of the custom emoticons. If this contains folder names, the content of
		/// those folders is processed recursively, maintaining the folder structure. This may be null or empty.</param>
		private void UpdateEmoticons( string[] fileNames )
		{
			if( UpdateWithCustomEmoticons == null )
				return;

			UpdateWithCustomEmoticons( fileNames );

			// The display categories may have changed.
			string displayCategoryToHide = ShowSensualDisplayCategory ? null : Configuration.SensualDisplayCategory;
			DisplayCategories = Configuration.SymbolDescriptors.CollectDisplayCategories( displayCategoryToHide );

			// The emoticons from the current display category may have changed.
			DisplaySymbolDescriptors = Configuration.SymbolDescriptors.CollectByDisplayCategory( CurrentDisplayCategory );

			ControlsUtils.FocusListBoxItem( DisplayCategoriesListBox );
			ControlsUtils.FocusListBoxItem( EmoticonsListBox );
		}

		/// <summary>
		/// Checks if the specified file names may be dropped.
		/// </summary>
		/// <param name="dropData">The drop data. This may be null or empty.</param>
		private bool MayDropFileNames( IDataObject dropData )
		{
			return (dropData != null) && dropData.GetDataPresent( DataFormats.FileDrop );
		}

		/// <summary>
		/// Drops the file names.
		/// </summary>
		/// <param name="dropData">The drop data. This may be null or empty.</param>
		private void DropFileNames( IDataObject dropData )
		{
			if( !MayDropFileNames( dropData ) )
				return;

			string[] droppedFiles = (string[])dropData.GetData( DataFormats.FileDrop );

			if( ArrayUtils.HasContent( droppedFiles )
				&& FileUtils.FileCountExceeds( droppedFiles, EmoticonManagerConfiguration.CustomEmoticonsCountMax ) )
			{
				string message = string.Format( "You are attempting to use more than {0} custom emoticons."
					, EmoticonManagerConfiguration.CustomEmoticonsCountMax );
				throw new ValidationException( message );
			}

			UpdateEmoticons( droppedFiles );
		}

		#endregion Private methods
	}
}
