////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using Vela.ActionClient;
using Vela.ActionProcessor;
using Vela.ApplicationBackend;
using Vela.Controls;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.StarkDocuments;
using Vela.Storehouse;

namespace Vela.ImBackend
{
	/// <summary>
	/// Client for processing action responses (<see cref="ActionResponse"/>).
	/// </summary>
	public class ImClient : ActionClient.ActionClient
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ImClient()
		{
		}

		#endregion Constructors

		#region Private static methods

		/// <summary>
		/// Changes the wallpaper using the files attached to the current letter.
		/// </summary>
		/// <param name="actionRequest">The action request which was processed.</param>
		/// <param name="responseContext">Action response context.</param>
		/// <param name="messageSummary">The message summary of the received letter. A text longer than a message summary should
		/// not be written on a wallpaper image.</param>
		/// <param name="emoticonSymbol">The emoticon symbol of the received letter. This may be null or empty.</param>
		private static void ChangeWallpaper( DownloadFilesActionRequest actionRequest, ActionResponseContext responseContext
			, string messageSummary, string emoticonSymbol )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( responseContext, "responseContext" );
			ValidationUtils.TextHasContent( messageSummary, "messageSummary" );

			if( responseContext.ClientContext.Tally.Approval < ApplicationBackendConsts.AllowChangeWallpaperApprovalMin )
				return;

			ActionClientHost actionClientHost = (ActionClientHost)responseContext.ClientContext.ActionClientHost;

			EmoticonSymbolDescriptor emoticonSymbolDescriptor = actionClientHost.GetEmoticonDescriptorFromSymbol(
				emoticonSymbol );
			if( (emoticonSymbolDescriptor == null) || (emoticonSymbolDescriptor.EmoticonCommand != EmoticonCommands.Wallpaper) )
				return;

			string rootSaveFolderName = responseContext.ClientContext.ProcessorConfiguration.RootSaveFolderName;
			TextOnImageConfiguration textOnImageConfiguration = actionClientHost.GetTextOnImageConfiguration();

			string textToWriteOnImage = ApplicationWallpaperUtils.GetTextToWriteOnImage( messageSummary, emoticonSymbol
				, actionRequest.AttachedFiles );

			// Set the wallpaper (of the recipient of the transferred files).
			ApplicationWallpaperUtils.Set( actionRequest.AttachedFiles, rootSaveFolderName, actionRequest.AccessGranterLocator
				, actionRequest.AccessGranterCreationTime, LocalPeerDirections.Recipient, textToWriteOnImage
				, textOnImageConfiguration );
		}

		/// <summary>
		/// Updates the emoticons with the custom ones from the files attached to the current letter.
		/// </summary>
		/// <param name="actionRequest">The action request which was processed.</param>
		/// <param name="responseContext">Action response context.</param>
		/// <param name="emoticonSymbol">The emoticon symbol of the received letter. This may be null or empty.</param>
		private static void UpdateWithCustomEmoticons( DownloadFilesActionRequest actionRequest
			, ActionResponseContext responseContext, string emoticonSymbol )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( responseContext, "responseContext" );

			if( responseContext.ClientContext.Tally.Approval
				< ApplicationBackendConsts.AllowUpdateWithCustomEmoticonsApprovalMin )
			{
				return;
			}

			ActionClientHost actionClientHost = (ActionClientHost)responseContext.ClientContext.ActionClientHost;

			EmoticonSymbolDescriptor emoticonSymbolDescriptor = actionClientHost.GetEmoticonDescriptorFromSymbol(
				emoticonSymbol );
			if( (emoticonSymbolDescriptor == null) || (emoticonSymbolDescriptor.EmoticonCommand != EmoticonCommands.Emoticons) )
				return;

			string rootSaveFolderName = responseContext.ClientContext.ProcessorConfiguration.RootSaveFolderName;

			// Copy the attached files to the local folder with custom emoticons.
			string receivedFilesFolder = AttachedFilesUtils.GetSaveFolderName( rootSaveFolderName
				, actionRequest.AccessGranterLocator, actionRequest.AccessGranterCreationTime );
			FileUtils.CopyFolderContent( receivedFilesFolder, AlgorithmPool.GetCustomEmoticonsFolderName(), false
				, FileExtensions.ImageFile );

			// Update the emoticons with the custom ones.
			Strings fileNames = new Strings();
			FileUtils.CollectFileNamesFromFolder( AlgorithmPool.GetCustomEmoticonsFolderName(), fileNames );
			actionClientHost.UpdateWithCustomEmoticons( fileNames );
		}

		#endregion Private static methods

		#region Public methods

		#region Update action requests

		/// <summary>
		/// Updates the specified action request to download files.
		/// </summary>
		/// <param name="actionRequest">Action request to update.</param>
		/// <param name="clientContext">Action client context.</param>
		/// <returns>Returns the true if the action request must be sent to the service, and false if it must never be sent.
		/// </returns>
		public bool UpdateDownloadFiles( DownloadFilesActionRequest actionRequest, ActionClientContext clientContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( clientContext, "clientContext" );

			string rootSaveFolderName = clientContext.ProcessorConfiguration.RootSaveFolderName;

			// Create the folder where to save the downloaded files.
			string saveFolderName = AttachedFilesUtils.GetSaveFolderName( rootSaveFolderName, actionRequest.AccessGranterLocator
				, actionRequest.AccessGranterCreationTime );
			FileUtils.CreateFolder( saveFolderName );

			AttachedFilesUtils.ValidateFileNames( actionRequest.AttachedFiles, rootSaveFolderName
				, actionRequest.AccessGranterLocator, actionRequest.AccessGranterCreationTime );
			AttachedFilesUtils.ValidateAvailableSpaceWithSavedFiles( actionRequest.AttachedFiles, rootSaveFolderName
				, actionRequest.AccessGranterLocator, actionRequest.AccessGranterCreationTime, 0 );

			Int64 requiredSize, downloadedSize;
			bool sendToService = AttachedFilesUtils.UpdateSizesFromSavedFiles( actionRequest.AttachedFiles, rootSaveFolderName
				, actionRequest.AccessGranterLocator, actionRequest.AccessGranterCreationTime, out requiredSize
				, out downloadedSize );

			return sendToService;
		}

		#endregion Update action requests

		#region Process action responses

		/// <summary>
		/// Processes the specified action response received for a request to create an account with a UPN.
		/// </summary>
		/// <param name="successResponse">Success action response to process. This may be null.</param>
		/// <param name="errorResponse">Error action response to process. Because an error action response is not derived from
		/// "successResponse", it must be sent as a separate parameter. This may be null. If this is not null, it has priority
		/// over "actionResponse".</param>
		/// <param name="responseContext">Action response context.</param>
		/// <remarks>
		/// <para>"actionResponse" and "errorResponse" may not be null at the same time. There is no need to validate this
		/// because the action client which calls this method ensures of this.</para>
		/// </remarks>
		public void ProcessCreateAccountWithUpn( CreateAccountAnonymousActionResponse successResponse
			, ErrorActionResponse errorResponse, ActionResponseContext responseContext )
		{
			ValidationUtils.ObjectHasContent( responseContext, "responseContext" );

			if( successResponse == null )
				return;

			ActionClientContext clientContext = responseContext.ClientContext;

			#region Load the vault speckle of the initiator

			CreateAccountAnonymousActionRequest request = (CreateAccountAnonymousActionRequest)clientContext.ActionRequest;
			Locator localUpnFingerprint = request.InitiatorUpnFingerprint;
			SpeckleFilter vaultUpnFilter = new SpeckleFilter( Vault.LocalUpnFingerprintFilterIdentifier
				, localUpnFingerprint );

			Speckle vaultSpeckle = clientContext.Transaction.CollectFirstSpeckle(
				PortableTypes.FindAllDerived( typeof( Vault ) ), new SpeckleFilters( vaultUpnFilter )
				, StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared, null, null );

			#endregion Load the vault speckle of the initiator

			CreateAccountClientProgress progress = (CreateAccountClientProgress)responseContext.ClientContext.ActionProgress;

			ActionProcessorStorehouseUtils.AddAccountFromPresenter( clientContext.Transaction, vaultSpeckle.StorehouseKey
				, successResponse.Presenter, progress.TallyEntityName, progress.Approval, progress.AlternativePeerAddress );
		}

		/// <summary>
		/// Processes the specified action response received for a request which sends a local presenter to update it on a remote
		/// peer.
		/// </summary>
		/// <param name="successResponse">Success action response to process. This may be null.</param>
		/// <param name="errorResponse">Error action response to process. Because an error action response is not derived from
		/// "successResponse", it must be sent as a separate parameter. This may be null. If this is not null, it has priority
		/// over "actionResponse".</param>
		/// <param name="responseContext">Action response context.</param>
		/// <remarks>
		/// <para>"actionResponse" and "errorResponse" may not be null at the same time. There is no need to validate this
		/// because the action client which calls this method ensures of this.</para>
		/// </remarks>
		public void ProcessPushUpdatePresenter( PushUpdatePresenterActionResponse successResponse
			, ErrorActionResponse errorResponse, ActionResponseContext responseContext )
		{
			ValidationUtils.ObjectHasContent( responseContext, "responseContext" );

			if( successResponse == null )
				return;
		}

		/// <summary>
		/// Processes the specified action response received for a request to send a letter.
		/// </summary>
		/// <param name="successResponse">Success action response to process. This may be null.</param>
		/// <param name="errorResponse">Error action response to process. Because an error action response is not derived from
		/// "successResponse", it must be sent as a separate parameter. This may be null. If this is not null, it has priority
		/// over "actionResponse".</param>
		/// <param name="responseContext">Action response context.</param>
		/// <remarks>
		/// <para>"actionResponse" and "errorResponse" may not be null at the same time. There is no need to validate this
		/// because the action client which calls this method ensures of this.</para>
		/// </remarks>
		public void ProcessSendLetter( SendLetterActionResponse successResponse, ErrorActionResponse errorResponse
			, ActionResponseContext responseContext )
		{
			ValidationUtils.ObjectHasContent( responseContext, "responseContext" );

			if( successResponse == null )
				return;
		}

		/// <summary>
		/// Processes the specified action response received for a request to download files.
		/// </summary>
		/// <param name="successResponse">Success action response to process. This may be null.</param>
		/// <param name="errorResponse">Error action response to process. Because an error action response is not derived from
		/// "successResponse", it must be sent as a separate parameter. This may be null. If this is not null, it has priority
		/// over "actionResponse".</param>
		/// <param name="responseContext">Action response context.</param>
		/// <remarks>
		/// <para>"actionResponse" and "errorResponse" may not be null at the same time. There is no need to validate this
		/// because the action client which calls this method ensures of this.</para>
		/// </remarks>
		public void ProcessDownloadFiles( DownloadFilesActionResponse successResponse, ErrorActionResponse errorResponse
			, ActionResponseContext responseContext )
		{
			ValidationUtils.ObjectHasContent( responseContext, "responseContext" );

			if( successResponse == null )
				return;

			DownloadFilesActionRequest actionRequest = (DownloadFilesActionRequest)responseContext.ClientContext.ActionRequest;

			string rootSaveFolderName = responseContext.ClientContext.ProcessorConfiguration.RootSaveFolderName;
			string localFileName = AttachedFilesUtils.GetSaveFileName( successResponse.FileName.RelativeFileName
				, rootSaveFolderName, actionRequest.AccessGranterLocator, actionRequest.AccessGranterCreationTime );

			int filePartSize = successResponse.FilePart.Length;

			// Check if the file is in the request.
			AttachedFile attachedFile = actionRequest.AttachedFiles[ successResponse.FileName ];
			if( attachedFile == null )
			{
				throw new ValidationException( string.Format( "The action client has received {0} bytes for the position {1} of"
							+ " the file. However, the action request doesn't refer this file."
						, filePartSize, successResponse.FromPosition ), null, null
					, new ExceptionDetail( "File name", localFileName )
					, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator ) );
			}

			// Save the received file part at the end of the existing file.
			FileUtils.CreateFolder( Path.GetDirectoryName( localFileName ) );
			using( FileStream fs = new FileStream( localFileName, FileMode.Append, FileAccess.Write ) )
			{
				if( fs.Length != successResponse.FromPosition )
				{
					throw new ValidationException( string.Format( "The action client has received {0} bytes for the position {1}"
								+ " of the file. However, the file currently has {2} bytes, so the position is invalid."
							, filePartSize, successResponse.FromPosition, fs.Length ), null, null
						, new ExceptionDetail( "File name", localFileName )
						, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator ) );
				}

				// Check if the file size exceeds the size specified in the attached file.
				if( fs.Length + filePartSize > attachedFile.Size )
				{
					throw new ValidationException( string.Format( "The action client has received {0} bytes for the position {1}"
								+ " of the file. However, the file currently has {2} bytes and its expected size is {3}, so the"
								+ " received position or size is invalid."
							, filePartSize, successResponse.FromPosition, fs.Length, attachedFile.Size ), null, null
						, new ExceptionDetail( "File name", localFileName )
						, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator ) );
				}

				LogManager.DebugAddFormat( "The action client has received {0} bytes for the file \"{1}\", for position {2}."
					, filePartSize, localFileName, successResponse.FromPosition );

				fs.Write( successResponse.FilePart, 0, filePartSize );
			}

			// The progress of the file transfer is kept inside the action request.
			attachedFile.CurrentSize += filePartSize;
		}

		#endregion Process action responses

		#region Process last frame action responses

		/// <summary>
		/// Processes the specified last frame action response received for a request to download files.
		/// </summary>
		/// <param name="actionRequest">The action request which was processed.</param>
		/// <param name="actionResponse">The last frame action response to process.</param>
		/// <param name="responseContext">Action response context.</param>
		public void ProcessLastFrameDownloadFiles( DownloadFilesActionRequest actionRequest
			, LastFrameActionResponse actionResponse, ActionResponseContext responseContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( actionResponse, "actionResponse" );
			ValidationUtils.ObjectHasContent( responseContext, "responseContext" );

			// Load the message summary and the emoticon symbol of the letter. A text longer than a message summary should not be
			// written on a wallpaper image.
			string messageSummary, emoticonSymbol;
			ApplicationBackendStorehouse.LoadLetterMessageSummaryAndEmoticonSymbol(
				responseContext.ClientContext.StorehouseProvider, actionRequest.AccessGranterLocator, out messageSummary
				, out emoticonSymbol );

			ChangeWallpaper( actionRequest, responseContext, messageSummary, emoticonSymbol );
			UpdateWithCustomEmoticons( actionRequest, responseContext, emoticonSymbol );
		}

		#endregion Process last frame action responses

		#endregion Public methods
	}
}
