﻿////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.ComponentModel;
using Vela.ActionProcessor;
using Vela.ApplicationBackend;
using Vela.Cloning;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.ImBackend
{
	/// <summary>
	/// Engine to queue action requests which push an updated presenter after its parent vault is edited, for each local account.
	/// </summary>
	/// <remarks>
	/// <para>This is a best-effort attempt. Since the queueing is performed on a separate thread, the user will not know
	/// if anything went wrong.</para>
	/// </remarks>
	public class QueuePushUpdatePresenterEngine
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected QueuePushUpdatePresenterEngine()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction. This is cloned
		/// in here.</param>
		/// <param name="vaultLocator">The locator of the edited vault.</param>
		/// <param name="cryptoAlgorithmPool">Pool of cryptographic algorithms. Its content must be thread safe.</param>
		/// <param name="actionClientManager">Action client manager which is notified that action requests which were added to
		/// the storehouse. Its content must be thread safe.</param>
		public QueuePushUpdatePresenterEngine( IStorehouseProvider storehouseProvider, Locator vaultLocator
			, ICryptoAlgorithmPool cryptoAlgorithmPool, IActionClientManager actionClientManager )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( vaultLocator, "vaultLocator" );
			ValidationUtils.ObjectHasContent( cryptoAlgorithmPool, "cryptoAlgorithmPool" );
			ValidationUtils.ObjectHasContent( actionClientManager, "actionClientManager" );

			StorehouseProvider = Cloner.Deep( storehouseProvider );
			VaultLocator = vaultLocator;
			CryptoAlgorithmPool = cryptoAlgorithmPool;
			ActionClientManager = actionClientManager;

			StartProcessing();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// Storehouse provider from which to create a storehouse transaction.
		/// </summary>
		private IStorehouseProvider StorehouseProvider;

		/// <summary>
		/// The locator of the edited vault.
		/// </summary>
		private Locator VaultLocator;

		/// <summary>
		/// Pool of cryptographic algorithms. Its content must be thread safe.
		/// </summary>
		private ICryptoAlgorithmPool CryptoAlgorithmPool;

		/// <summary>
		/// Action client manager which is notified that action requests which were added to the storehouse. Its content must be
		/// thread safe.
		/// </summary>
		private IActionClientManager ActionClientManager;

		/// <summary>
		/// The worker thread.
		/// </summary>
		private BackgroundWorker BackgroundWorker;

		#endregion Private fields

		#region Event handlers

		private void BackgroundWorker_DoWork( object sender, DoWorkEventArgs e )
		{
			Execute();
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Queues the action requests.
		/// </summary>
		private void QueueActionRequests()
		{
			Locators updateRequestLocators = new Locators();

			// Create a transaction.
			IStorehouseTransaction transaction = StorehouseProvider.CreateTransaction();

			try
			{
				IDeserializer deserializer;
				Vault vault = transaction.LoadSheaf( VaultLocator, StorehouseRealms.ProcessingIdentityContainers
					, typeof( Vault ), SheafLocks.Shared, out deserializer ) as Vault;

				SpeckleFilter accountLocalUpnFilter = new SpeckleFilter( Account.LocalUpnFingerprintFilterIdentifier
					, vault.LocalUpnFingerprint );
				SpeckleFilterIdentifiers accountLoadFilterIdentifiers = new SpeckleFilterIdentifiers(
					Account.RemoteUpnFingerprintFilterIdentifier );

				Speckles accountSpeckles = transaction.CollectSpeckles( PortableTypes.FindAllDerived( typeof( Account ) )
					, new SpeckleFilters( accountLocalUpnFilter ), StorehouseRealms.Accounts, null, null, null
					, accountLoadFilterIdentifiers ) as Speckles;

				HashSet<Locator> usedRemoteUpnFingerprints = new HashSet<Locator>();

				MacAlg macAlg = CryptoAlgorithmPool.GetMacAlg();
				IAsymmSigningPaddingAlg asymmSigningPaddingAlg = CryptoAlgorithmPool.GetAsymmSigningPadding();
				IHashSerializer hashSerializer = StorehouseProvider.Serializer.GetHashSerializer( false );

				foreach( Speckle accountSpeckle in accountSpeckles )
				{
					Locator remoteUpnFingerprint = accountSpeckle.GetLoadedSpeckleFilter<Locator>(
						Account.RemoteUpnFingerprintFilterIdentifier );

					// Don't send the presenter multiple times to the same remote peer / tally.
					if( usedRemoteUpnFingerprints.Contains( remoteUpnFingerprint ) )
						continue;

					Locator updateRequestLocator = ApplicationBackendStorehouse.QueuePushUpdatePresenter( transaction, vault
						, remoteUpnFingerprint, macAlg, asymmSigningPaddingAlg, hashSerializer, ActionClientManager );

					if( updateRequestLocator != null )
					{
						usedRemoteUpnFingerprints.Add( remoteUpnFingerprint );
						updateRequestLocators.Add( updateRequestLocator );
					}
				}

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			// Notify the action client manager that action requests were added to the storehouse.
			ActionClientManager.Add( updateRequestLocators, true );
		}

		/// <summary>
		/// Executes the queueing.
		/// </summary>
		private void Execute()
		{
			try
			{
				QueueActionRequests();
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				if( ExceptionUtils.IsAbort( ex ) )
				{
					LogManager.DebugAddFormat( "The queueing of action requests which push an updated presenter has been"
						+ " aborted." );
				}
				else
				{
					LogManager.Add( ex );
				}
			}
		}

		/// <summary>
		/// Starts processing on a new thread.
		/// </summary>
		private void StartProcessing()
		{
			BackgroundWorker = new BackgroundWorker();
			BackgroundWorker.WorkerSupportsCancellation = false;
			BackgroundWorker.WorkerReportsProgress = false;
			BackgroundWorker.DoWork += BackgroundWorker_DoWork;

			BackgroundWorker.RunWorkerAsync();
		}

		#endregion Private methods
	}
}
