////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows.Threading;
using Vela.ActionProcessor;
using Vela.Controls;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.StarkDocuments;

namespace Vela.ImBackend
{
	/// <summary>
	/// Action client host.
	/// </summary>
	public class ActionClientHost : ActionProcessorHost, IActionClientHost
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionClientHost()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">The application model.</param>
		public ActionClientHost( ApplicationModel applicationModel )
			: base( applicationModel )
		{
		}

		#endregion Constructors

		#region Private methods

		/// <summary>
		/// Handles a revoked service vault error occurred in an action processor.
		/// </summary>
		/// <param name="serviceUpn">The UPN of the revoked service vault.</param>
		/// <remarks>
		/// <para>This is already running on the dispatcher of the application model.</para>
		/// </remarks>
		private void UpdateUiRevokedServiceVaultOnUiThread( Upn serviceUpn )
		{
			try
			{
				if( (ApplicationModel.StorehouseModel != null) && (ApplicationModel.StorehouseModel.AccountsModel != null) )
					ApplicationModel.StorehouseModel.AccountsModel.ReloadAccountsForUpn( serviceUpn.Fingerprint );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				// This exception may be swallowed by the system, so log it here.
				LogManager.Add( ex );

				throw new BackgroundOperationException( ex );
			}
		}

		/// <summary>
		/// Updates the UI while an action request is being processed by an action client.
		/// </summary>
		/// <param name="updateUiContext">Host context for updating the user interface of an action client application after an
		/// action request and response are processed.</param>
		/// <remarks>
		/// <para>This is already running on the dispatcher of the application model.</para>
		/// </remarks>
		private void UpdateUiOnUiThread( ActionClientHostUpdateUiContext updateUiContext )
		{
			try
			{
				ValidationUtils.ObjectHasContent( updateUiContext, "updateUiContext" );

				Type actionRequestType = PortableTypes.FindGlobal(
					updateUiContext.ActionProgress.ReferredActionRequestTypeName );
				if( actionRequestType == null )
				{
					throw new ValidationException( "The type of the action request is not registered globally.", null, null
						, new ExceptionDetail( "Request type", updateUiContext.ActionProgress.ReferredActionRequestTypeName )
						, new ExceptionDetail( "Request locator"
							, updateUiContext.ActionProgress.ReferredActionRequestLocator ) );
				}

				StorehouseModel sm = ApplicationModel.StorehouseModel;

				if( sm != null )
				{
					if( actionRequestType == typeof( CreateAccountAnonymousActionRequest ) )
					{
						if( sm.AccountsModel != null )
							sm.AccountsModel.UpdateClientUiOnCreateAccountWithUpn( updateUiContext );

						return;
					}

					if( actionRequestType == typeof( SendLetterActionRequest ) )
					{
						if( sm.LettersModel != null )
							sm.LettersModel.UpdateClientUiOnSendLetter( updateUiContext );

						return;
					}

					if( actionRequestType == typeof( DownloadFilesActionRequest ) )
					{
						if( sm.LettersModel != null )
							sm.LettersModel.UpdateClientUiOnDownloadFiles( updateUiContext );

						return;
					}
				}

				// Only log other action request types.
				LogManager.DebugAddFormat( "The type \"{0}\" of the action request with the locator \"{1}\" was not"
						+ " processed by the action client host."
					, updateUiContext.ActionProgress.ReferredActionRequestTypeName
					, updateUiContext.ActionProgress.ReferredActionRequestLocator );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				// This exception may be swallowed by the system, so log it here.
				LogManager.Add( ex );

				throw new BackgroundOperationException( ex );
			}
		}

		/// <summary>
		/// Gets the descriptor of the specified emoticon symbol.
		/// </summary>
		/// <param name="emoticonSymbol">Emoticon symbol whose descriptor to get. This may be null or empty.</param>
		/// <returns>Returns the descriptor of the specified emoticon symbol. This may be null.</returns>
		/// <remarks>
		/// <para>Do not modify the returned object because it's not thread safe.</para>
		/// </remarks>
		private EmoticonSymbolDescriptor GetEmoticonDescriptorFromSymbolOnUiThread( string emoticonSymbol )
		{
			try
			{
				if( TextUtils.NoContent( emoticonSymbol ) || (ApplicationModel.EmoticonManager == null) )
					return null;

				return ApplicationModel.EmoticonManager.GetEmoticonDescriptorFromSymbol( emoticonSymbol );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				// This exception may be swallowed by the system, so log it here.
				LogManager.Add( ex );

				throw new BackgroundOperationException( ex );
			}
		}

		/// <summary>
		/// Updates the emoticon configuration with the custom emoticons.
		/// </summary>
		/// <param name="fileNames">The name of the files of the custom emoticons. If this contains folder names, the content of
		/// those folders is processed recursively. This may be null or empty.</param>
		private void UpdateWithCustomEmoticonsOnUiThread( Strings fileNames )
		{
			try
			{
				ApplicationModel.UpdateWithCustomEmoticons( fileNames );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				// This exception may be swallowed by the system, so log it here.
				LogManager.Add( ex );

				throw new BackgroundOperationException( ex );
			}
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Updates the UI while an action request is being processed by an action client, and the service vault is revoked.
		/// </summary>
		/// <param name="serviceUpn">The UPN of the revoked service vault.</param>
		public void UpdateUiRevokedServiceVault( Upn serviceUpn )
		{
			// Process on the UI thread.
			ApplicationModel.Dispatcher.Invoke( () => UpdateUiRevokedServiceVaultOnUiThread( serviceUpn )
				, DispatcherPriority.Background );
		}

		/// <summary>
		/// Updates the UI while an action request is being processed by an action client.
		/// </summary>
		/// <param name="updateUiContext">Host context for updating the user interface of an action client application after an
		/// action request and response are processed.</param>
		public void UpdateUi( ActionClientHostUpdateUiContext updateUiContext )
		{
			// Process on the UI thread.
			ApplicationModel.Dispatcher.Invoke( () => UpdateUiOnUiThread( updateUiContext ), DispatcherPriority.Background );
		}

		/// <summary>
		/// Gets the descriptor of the specified emoticon symbol.
		/// </summary>
		/// <param name="emoticonSymbol">Emoticon symbol whose descriptor to get. This may be null or empty.</param>
		/// <returns>Returns the descriptor of the specified emoticon symbol. This may be null.</returns>
		/// <remarks>
		/// <para>Do not modify the returned object because it's not thread safe.</para>
		/// </remarks>
		public EmoticonSymbolDescriptor GetEmoticonDescriptorFromSymbol( string emoticonSymbol )
		{
			if( TextUtils.NoContent( emoticonSymbol ) )
				return null;

			// Process on the UI thread.
			object emoticonSymbolDescriptor = ApplicationModel.Dispatcher.Invoke(
				new Func<EmoticonSymbolDescriptor>(
					delegate
					{ return GetEmoticonDescriptorFromSymbolOnUiThread( emoticonSymbol ); } )
				, DispatcherPriority.Background );

			return emoticonSymbolDescriptor as EmoticonSymbolDescriptor;
		}

		/// <summary>
		/// Updates the emoticon configuration with the custom emoticons.
		/// </summary>
		/// <param name="fileNames">The name of the files of the custom emoticons. If this contains folder names, the content of
		/// those folders is processed recursively. This may be null or empty.</param>
		public void UpdateWithCustomEmoticons( Strings fileNames )
		{
			// Process on the UI thread.
			ApplicationModel.Dispatcher.Invoke( () => UpdateWithCustomEmoticonsOnUiThread( fileNames )
				, DispatcherPriority.Background );
		}

		/// <summary>
		/// Gets the configuration for a text which is to be written on an image.
		/// </summary>
		/// <returns>Returns the configuration for a text which is to be written on an image.</returns>
		/// <remarks>
		/// <para>Do not modify the returned object because it's not thread safe.</para>
		/// </remarks>
		public TextOnImageConfiguration GetTextOnImageConfiguration()
		{
			return ApplicationModel.Configuration.TextOnImageConfiguration;
		}

		#endregion Public methods
	}
}
