////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows.Threading;
using Vela.ActionProcessor;
using Vela.Kernel;

namespace Vela.ImBackend
{
	/// <summary>
	/// Contract for an action processor host.
	/// </summary>
	public abstract class ActionProcessorHost : IActionProcessorHost
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionProcessorHost()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">The application model.</param>
		public ActionProcessorHost( ApplicationModel applicationModel )
		{
			ApplicationModel = applicationModel;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private ApplicationModel _ApplicationModel;
		/// <summary>
		/// The application model. READ-ONLY.
		/// </summary>
		public ApplicationModel ApplicationModel
		{
			get { return _ApplicationModel; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ApplicationModel" );

				_ApplicationModel = value;
			}
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Checks if the action processor is under maintenance.
		/// </summary>
		/// <remarks>
		/// <para>This is already running on the dispatcher of the application model.</para>
		/// </remarks>
		private bool CheckMaintenanceOnUiThread()
		{
			try
			{
				return ApplicationModel.CheckMaintenance();
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				// This exception may be swallowed by the system, so log it here.
				LogManager.Add( ex );

				throw new BackgroundOperationException( ex );
			}
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Checks if the action processor is under maintenance.
		/// </summary>
		public bool CheckMaintenance()
		{
			// Process on the UI thread.
			object isMaintenance = ApplicationModel.Dispatcher.Invoke(
				new Func<bool>( delegate
				{ return CheckMaintenanceOnUiThread(); } ), DispatcherPriority.Background );

			return (isMaintenance != null) ? (bool)isMaintenance : false;
		}

		#endregion Public methods
	}
}
