////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows.Threading;
using Vela.ActionProcessor;
using Vela.Kernel;

namespace Vela.ImBackend
{
	/// <summary>
	/// Action service host.
	/// </summary>
	public class ActionServiceHost : ActionProcessorHost, IActionServiceHost
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionServiceHost()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">The application model.</param>
		public ActionServiceHost( ApplicationModel applicationModel )
			: base( applicationModel )
		{
		}

		#endregion Constructors

		#region Private methods

		/// <summary>
		/// Updates the UI after an action request was successfully processed by an action service, and a response was sent to
		/// the client.
		/// </summary>
		/// <param name="updateUiContext">Host context for updating the user interface of an action service application after an
		/// action request and response are processed.</param>
		/// <remarks>
		/// <para>This is already running on the dispatcher of the application model.</para>
		/// </remarks>
		private void UpdateUiOnUiThread( ActionServiceHostUpdateUiContext updateUiContext )
		{
			try
			{
				ValidationUtils.ObjectHasContent( updateUiContext, "updateUiContext" );

				StorehouseModel sm = ApplicationModel.StorehouseModel;

				if( sm != null )
				{
					if( updateUiContext.ActionRequest as PushUpdatePresenterActionRequest != null )
					{
						if( sm.AccountsModel != null )
							sm.AccountsModel.UpdateServiceUiOnPushUpdatePresenter( updateUiContext );

						return;
					}

					if( updateUiContext.ActionRequest as SendLetterActionRequest != null )
					{
						if( sm.LettersModel != null )
							sm.LettersModel.UpdateServiceUiOnSendLetter( updateUiContext );

						return;
					}

					if( updateUiContext.ActionRequest as DownloadFilesActionRequest != null )
					{
						if( sm.LettersModel != null )
							sm.LettersModel.UpdateServiceUiOnDownloadFiles( updateUiContext );

						return;
					}
				}

				// Only log other action request types.
				LogManager.DebugAddFormat( "The type \"{0}\" of the action request with the locator \"{1}\" was not"
						+ " processed by the action service host."
					, updateUiContext.ActionRequest.GetType().FullName, updateUiContext.ActionRequest.InstanceLocator );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				// This exception may be swallowed by the system, so log it here.
				LogManager.Add( ex );

				throw new BackgroundOperationException( ex );
			}
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Updates the UI after an action request was successfully processed by an action service, and a response was sent to
		/// the client.
		/// </summary>
		/// <param name="updateUiContext">Host context for updating the user interface of an action service application after an
		/// action request and response are processed.</param>
		public void UpdateUi( ActionServiceHostUpdateUiContext updateUiContext )
		{
			// Process on the UI thread.
			ApplicationModel.Dispatcher.Invoke( () => UpdateUiOnUiThread( updateUiContext ), DispatcherPriority.Background );
		}

		#endregion Public methods
	}
}
