////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using Vela.ActionProcessor;
using Vela.ApplicationBackend;
using Vela.ApplicationDocuments;
using Vela.Controls;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Crypto.Symmetric;
using Vela.ImDocuments;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.StarkDocuments;
using Vela.Storehouse;

namespace Vela.ImBackend
{
	/// <summary>
	/// Storehouse utilities for the application.
	/// </summary>
	public static class ImBackendStorehouse
	{
		#region Public constants

		/// <summary>
		/// The template of each line which shows a signing entity name and time. Placeholders: 0 - signing entity name,
		/// 1 - signing time (for a group identity, it's the newest signing time).
		/// </summary>
		public const string SigningEntityNameTimeLineTemplate = "    * {0} ({1})";

		/// <summary>
		/// The separator between consecutive lines which show a signing entity name and time.
		/// </summary>
		public const string SigningEntityNameTimeLineSeparator = "\n";

		#endregion Public constants

		#region Public static methods

		/// <summary>
		/// Creates a storehouse.
		/// </summary>
		/// <param name="algorithmPool">Algorithm pool.</param>
		/// <param name="storehouseAccess">The storehouse access.</param>
		/// <param name="serviceListenerConfiguration">The action service listener configuration.</param>
		/// <param name="storehouseFileName">The (initial) storehouse file name. This may be null or empty.</param>
		/// <param name="currentStorehouseFileName">The current storehouse file name. This is used to not allow the user to
		/// overwrite the current storehouse. This may be null or empty.</param>
		/// <returns>Returns the storehouse provider of the created storehouse, or null if none was created.</returns>
		public static IStorehouseProvider CreateStorehouse( AlgorithmPool algorithmPool, IStorehouseAccess storehouseAccess
			, ActionServiceListenerConfiguration serviceListenerConfiguration, string storehouseFileName
			, string currentStorehouseFileName )
		{
			ValidationUtils.ObjectHasContent( algorithmPool, "algorithmPool" );
			ValidationUtils.ObjectHasContent( serviceListenerConfiguration, "serviceListenerConfiguration" );

			ActionServiceListenerConfigurations serviceListenerConfigurations = new ActionServiceListenerConfigurations();
			serviceListenerConfigurations.Add( serviceListenerConfiguration );

			ImStorehouseConfiguration storehouseConfiguration = new ImStorehouseConfiguration( serviceListenerConfigurations );

			return ApplicationBackendStorehouse.CreateStorehouse( algorithmPool, storehouseAccess, storehouseConfiguration
				, storehouseFileName, currentStorehouseFileName );
		}

		/// <summary>
		/// Opens an existing storehouse.
		/// </summary>
		/// <param name="storehouseAccess">The storehouse access.</param>
		/// <param name="storehouseFileName">The (initial) storehouse file name. This may be null or empty.</param>
		/// <param name="browseStorehouseFile">If this is true then the dialog to open a storehouse file is shown.</param>
		/// <param name="storehousePassword">If this has content, the storehouse is automatically opened with this password; no
		/// password dialog is displayed.</param>
		/// <param name="serializer">The serializer of the sheafs.</param>
		/// <returns>Returns the storehouse provider of the opened storehouse, or null if none was opened.</returns>
		/// <remarks>
		/// <para>The storehouse's entropy state (<see cref="Vela.Crypto.EntropyPool"/>) is restored in the application's
		/// entropy pool.</para>
		/// </remarks>
		public static IStorehouseProvider OpenStorehouse( IStorehouseAccess storehouseAccess, string storehouseFileName
			, bool browseStorehouseFile, string storehousePassword, ISerializer serializer )
		{
			return ApplicationBackendStorehouse.OpenStorehouse( storehouseAccess, storehouseFileName, browseStorehouseFile
				, storehousePassword, serializer );
		}

		/// <summary>
		/// Creates a vault.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="algorithmPool">Algorithm pool.</param>
		/// <param name="algorithmSecurityLevel">The security level for the cryptographic algorithms to use.</param>
		/// <returns>Returns the vault speckle of the created vault, or null if none was created.</returns>
		public static VaultSpeckle AddVault( IStorehouseProvider storehouseProvider, AlgorithmPool algorithmPool
			, AlgorithmSecurityLevel algorithmSecurityLevel )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( algorithmPool, "algorithmPool" );

			#region Load the data

			ImStorehouseConfiguration storehouseConfiguration =
				StorehouseUtils.LoadConfiguration<ImStorehouseConfiguration>( storehouseProvider );

			ActionServiceListenerConfiguration listenerConfiguration
				= storehouseConfiguration.ActionServiceListeners.GetMostRecent();
			UpnAlg upnAlg = algorithmPool.GetUpnAlg( algorithmSecurityLevel );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				if( !ApplicationBackendStorehouse.SameUpnAlgorithmsPerPort( transaction, listenerConfiguration.ProcessingPort
					, upnAlg.CommunicationFrameAlg ) )
				{
					throw new ValidationException( "The communication frame algorithm can't be used for the processing port"
							+ " because a different one already exists for the same port. You can't mix vaults with different"
							+ " communication frame algorithms for the same port.", null, null
						, new ExceptionDetail( "Communication frame algorithm"
							, RegistrableAlgorithmUtils.GetAlgorithmName( upnAlg.CommunicationFrameAlg ) )
						, new ExceptionDetail( "Processing port", listenerConfiguration.ProcessingPort.ToString() ) );
				}

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			#endregion Load the data

			#region Generate the identity

			// Generate the identity.
			GenerateIdentityWindow identityWindow = new GenerateIdentityWindow( upnAlg, listenerConfiguration.ProcessingPort
				, algorithmPool.GetAsymmAlg( algorithmSecurityLevel ), algorithmPool.GetAsymmAlg( algorithmSecurityLevel )
				, algorithmSecurityLevel );
			if( !identityWindow.ShowProgressDialog() )
				return null;

			#endregion Generate the identity

			#region Edit the vault

			// The peer virtual identity.
			PersonVirtualIdentity virtualIdentity = new PersonVirtualIdentity();

			// The peer physical address.
			PhysicalAddress physicalAddress = new PhysicalAddress();

			// The appendix.
			Appendix appendix = new Appendix();
			appendix.VirtualIdentity = virtualIdentity;
			appendix.PhysicalAddress = physicalAddress;

			// The vault.
			Vault vault = new Vault( CryptoUtils.GetLocator(), 0, identityWindow.GeneratedIdentity, appendix );

			// The photo.
			Photo photo = new Photo( CryptoUtils.GetLocator(), vault.LocalUpnFingerprint );

			// Edit the vault.
			EditVaultWindow vaultWindow = new EditVaultWindow( vault, photo );
			if( !ControlsUtils.ShowDialog( vaultWindow ) )
				return null;

			#endregion Edit the vault

			#region Save the data

			// Create a transaction.
			transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Check if the UPN is local.
				Upn upn = identityWindow.GeneratedIdentity.Upn;
				if( ActionProcessorStorehouseUtils.HasVaultWithUpn( transaction, upn ) )
				{
					throw new ValidationException( "Can't add a vault with the generated identity because a vault with the same"
						+ "UPN already exists. Try again.", null, null, new ExceptionDetail( "UPN", upn.ToString() ) );
				}

				// Reserve (a sheaf with) the UPN fingerprint.
				transaction.AddSpeckle( vault.LocalUpnFingerprint, StorehouseRealms.UpnFingerprintReservations
					, typeof( ReservedUpnFingerprintSheaf ) );

				// Save the vault and photo.
				transaction.AddSheaf( vaultWindow.Vault, StorehouseRealms.ProcessingIdentityContainers );
				if( vaultWindow.Photo.Image != null )
					transaction.AddSheaf( vaultWindow.Photo, StorehouseRealms.Photos );

				// Get the actual added speckle.
				VaultSpeckle vaultSpeckle = transaction.LoadSpeckle( vault.InstanceLocator
					, StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared, typeof( VaultSpeckle )
					, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Vault ) ) ) as VaultSpeckle;
				ValidationUtils.ObjectHasContent( vaultSpeckle, "vaultSpeckle" );

				// Commit the transaction.
				transaction.Commit();

				return vaultSpeckle;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			#endregion Save the data
		}

		/// <summary>
		/// Edits the vault with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="vaultLocator">The locator of the vault to edit.</param>
		/// <param name="upnsHaveChanged">Is set with true if the (automated) UPNs of the edited vault have changed.</param>
		/// <returns>Returns the vault speckle of the edited vault, or null if it was not edited.</returns>
		public static VaultSpeckle EditVault( IStorehouseProvider storehouseProvider, Locator vaultLocator
			, out bool upnsHaveChanged )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( vaultLocator, "vaultLocator" );

			#region Load the data

			Vault vault;
			Photo photo;
			bool hasInitialPhoto;
			Locator initialPhotoLocator;

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the vault.
				IDeserializer deserializer;
				vault = transaction.LoadSheaf( vaultLocator, StorehouseRealms.ProcessingIdentityContainers, typeof( Vault ),
					SheafLocks.Update, out deserializer ) as Vault;
				if( vault == null )
				{
					throw new ValidationException( "The vault to edit is missing.", null, null
						, new ExceptionDetail( "Vault locator", vaultLocator ) );
				}

				#region Get the photo

				// Load the photo.
				SpeckleFilter photoUpnFilter = new SpeckleFilter( Photo.CreatorUpnFingerprintFilterIdentifier
					, vault.LocalUpnFingerprint );
				photo = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Photo ) )
					, new SpeckleFilters( photoUpnFilter ), StorehouseRealms.Photos, SheafLocks.Update
					, out deserializer ) as Photo;
				hasInitialPhoto = photo != null;
				initialPhotoLocator = hasInitialPhoto ? photo.InstanceLocator : null;

				// Create the photo.
				if( !hasInitialPhoto )
					photo = new Photo( CryptoUtils.GetLocator(), vault.LocalUpnFingerprint );

				#endregion Get the photo

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			#endregion Load the data

			#region Edit the vault

			EditVaultWindow vaultWindow = new EditVaultWindow( vault, photo );
			if( !ControlsUtils.ShowDialog( vaultWindow ) )
			{
				upnsHaveChanged = false;

				return null;
			}

			upnsHaveChanged = vaultWindow.UpnsHaveChanged;

			#endregion Edit the vault

			#region Save the data

			// Create a transaction.
			transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Save the vault.
				transaction.UpdateSheaf( vaultWindow.Vault, StorehouseRealms.ProcessingIdentityContainers );

				// Save the photo.
				if( vaultWindow.Photo.Image != null )
				{
					if( hasInitialPhoto )
						transaction.UpdateSheaf( vaultWindow.Photo, StorehouseRealms.Photos );
					else
						transaction.AddSheaf( vaultWindow.Photo, StorehouseRealms.Photos );
				}
				else
				{
					if( hasInitialPhoto )
						transaction.DeleteSheaf( initialPhotoLocator, StorehouseRealms.Photos );
				}

				// Get the actual edited speckle.
				VaultSpeckle vaultSpeckle = transaction.LoadSpeckle( vaultLocator
					, StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared, typeof( VaultSpeckle )
					, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Vault ) ) ) as VaultSpeckle;
				ValidationUtils.ObjectHasContent( vaultSpeckle, "vaultSpeckle" );

				// Commit the transaction.
				transaction.Commit();

				return vaultSpeckle;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			#endregion Save the data
		}

		/// <summary>
		/// Queues an action request for getting the presenter of the remote peer and creating an account for it, using a UPN.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="vaultLocator">The locator of the vault of the local peer for which to create an account.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="hashSerializer">Hash serializer.</param>
		/// <returns>Returns the instance locator of the created action request, or null if none was created.</returns>
		/// <remarks>
		/// <para>The action request is saved in the storehouse, it's not actually sent.</para>
		/// </remarks>
		public static Locator QueueCreateAccountWithUpn( IStorehouseProvider storehouseProvider, Locator vaultLocator
			, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg, IHashSerializer hashSerializer )
		{
			QueueCreateAccountWithUpnWindow upnWindow = new QueueCreateAccountWithUpnWindow();
			if( !ControlsUtils.ShowDialog( upnWindow ) )
				return null;

			Vault vault = ApplicationBackendStorehouse.LoadVault( storehouseProvider, vaultLocator );
			ValidationUtils.ObjectHasContent( vault, "vault" );

			// If the UPN algorithm of the remote peer is different than the local one, it's possible that the remote peer's Vela
			// doesn't know the UPN algorithm of the local peer, so any communication attempt would result in a remote error
			// which doesn't reach the local peer, which means that the local peer would not know what happened and would
			// constantly see a "Retry" status (indicating a timeout) for his action requests.
			IHashSerializer hashSerializerForEquality = hashSerializer.GetHashSerializer( true );
			if( !hashSerializerForEquality.EqualDocuments( vault.LocalUpn.UpnAlg, upnWindow.Upn.UpnAlg ) )
			{
				MessageWindow.ShowDialog( "Make sure that you and your contacts are using the newest version of the application."
					, MessageWindowModes.Info );
			}

			return ApplicationBackendStorehouse.QueueCreateAccount( storehouseProvider, vault.ProcessingIdentity
				, upnWindow.Upn, upnWindow.TallyEntityName, upnWindow.Approval, upnWindow.AlternativePeerAddress, macAlg
				, asymmSigningPaddingAlg, hashSerializer );
		}

		/// <summary>
		/// Edits the account with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="accountLocator">The locator of the account to edit.</param>
		/// <param name="recipientAccessStatesManager">Manager for the recipient access states, used by an action client to
		/// process action requests.</param>
		/// <param name="changedApproval">Is set with true if the approval of the remote peer is changed.</param>
		/// <param name="changedAlternativePeerAddress">Is set with true if the alternative peer address of the remote peer is
		/// changed.</param>
		/// <returns>Returns the speckle of the edited account, or null if it was not edited.</returns>
		public static AccountOrRequestSpeckle EditAccount( IStorehouseProvider storehouseProvider, Locator accountLocator
			, IRecipientAccessStatesManager recipientAccessStatesManager, out bool changedApproval
			, out bool changedAlternativePeerAddress )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( accountLocator, "accountLocator" );
			ValidationUtils.ObjectHasContent( recipientAccessStatesManager, "recipientAccessStatesManager" );

			#region Load the data

			AccountContext accountContext;

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				accountContext = ApplicationBackendStorehouse.LoadAccountContext( transaction, accountLocator, SheafLocks.Update
					, SheafLocks.Update, true );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			#endregion Load the data

			#region Edit the account

			EditAccountWindow accountWindow = new EditAccountWindow( accountContext );
			if( !ControlsUtils.ShowDialog( accountWindow ) )
			{
				changedApproval = false;
				changedAlternativePeerAddress = false;

				return null;
			}

			changedApproval = accountWindow.ChangedApproval;
			changedAlternativePeerAddress = accountWindow.ChangedAlternativePeerAddress;

			#endregion Edit the account

			#region Save the data

			// Create a transaction.
			transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Save the tally.
				transaction.UpdateSheaf( accountWindow.AccountContext.Tally, StorehouseRealms.ProcessingIdentityContainers );

				// Save the account.
				transaction.UpdateSheaf( accountWindow.AccountContext.Account, StorehouseRealms.Accounts );

				AccountOrRequestSpeckle accountSpeckle = ApplicationBackendStorehouse.LoadAccountOrRequestSpeckle(
					transaction, accountLocator );
				ValidationUtils.ObjectHasContent( accountSpeckle, "accountSpeckle" );

				// Commit the transaction.
				transaction.Commit();

				return accountSpeckle;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			#endregion Save the data
		}

		/// <summary>
		/// Creates an attesting document and signs it with the vault with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="algorithmPool">Algorithm pool.</param>
		/// <param name="vaultLocator">The locator of the vault with which to sign the file.</param>
		/// <returns>Returns the name of the file which contains the attesting document, or null if one was not created.
		/// </returns>
		/// <remarks>
		/// <para>Expired documents throw an <see cref="HasExpiredException"/> exception.</para>
		/// </remarks>
		public static string CreateAttestingDocument( IStorehouseProvider storehouseProvider, AlgorithmPool algorithmPool
			, Locator vaultLocator )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( algorithmPool, "algorithmPool" );
			ValidationUtils.ObjectHasContent( vaultLocator, "vaultLocator" );

			#region Load the data

			Vault vault = ApplicationBackendStorehouse.LoadVault( storehouseProvider, vaultLocator );
			if( vault == null )
			{
				throw new ValidationException( "The vault is missing.", null, null
					, new ExceptionDetail( "Vault locator", vaultLocator ) );
			}

			// Don't use expired documents.
			KernelUtils.HasExpired( vault );

			// File to attest.
			OpenFileDialog attestedOfd = new OpenFileDialog();
			attestedOfd.Title = "Open File to Attest";
			attestedOfd.AddExtension = false;
			attestedOfd.Filter = FileExtensions.AllFilesFilter;
			attestedOfd.CheckFileExists = true;
			attestedOfd.CheckPathExists = true;
			if( !ControlsUtils.ShowDialog( attestedOfd ) )
				return null;

			// Attesting document.
			string attestingFileName;
			if( !ControlsUtils.ShowSaveFileDialog( "Save Attesting Document", FileExtensions.AttestingDocument
				, FileExtensions.AttestingDocumentFilter, FileUtils.GetExistingFolderName( attestedOfd.FileName )
				, Path.GetFileName( attestedOfd.FileName ) + FileExtensions.AttestingDocument
				, false // Existing files are not overwritten, just the signature is updated.
				, out attestingFileName ) )
			{
				return null;
			}

			#endregion Load the data

			MacAlg macAlg = algorithmPool.GetMacAlg();

			#region Load / create the attesting document

			// Load the attesting document.
			IDeserializer deserializer;
			AttestingDocument attestingDocument = SerializationUtils.LoadFromFile( attestingFileName, typeof( AttestingDocument )
				, out deserializer ) as AttestingDocument;

			if( attestingDocument == null )
			{
				// Create the attesting document.
				attestingDocument = new AttestingDocument( CryptoUtils.GetLocator(), macAlg
					, algorithmPool.GetAsymmSigningPadding(), algorithmPool.GetHashSerializer( false )
					, vault.LocalUpnFingerprint );
			}
			else
			{
				// Check if the attesting document was already signed and that the signature is valid. This is done because
				// several entities may have signed the document, so the document can't be simply replaced.
				if( attestingDocument.Signatures.Find( vault.ProcessingIdentity.Upn.Fingerprint ) != null )
				{
					try
					{
						attestingDocument.VerifySignature( vault.ProcessingIdentity );
					}
					catch( Exception ex )
					{
						// The loop must continue (if possible).

						if( ExceptionUtils.IsAbort( ex ) )
							throw ExceptionUtils.PreserveStackTrace( ex );

						string message = string.Format( "The attesting document \"{0}\" was already signed but the signature is"
							+ " invalid. Do you want to overwrite it?", attestingFileName );
						if( MessageWindow.ShowDialog( message, MessageWindowModes.Query ) != QueryMessageResults.Yes )
							return null;
					}
				}
			}

			// Don't use expired documents.
			KernelUtils.HasExpired( attestingDocument );

			#endregion Load / create the attesting document

			// Compute the MAC of the file to attest.
			ProcessAttestingDocumentWindow processWindow = new ProcessAttestingDocumentWindow( "Create attesting document"
				, attestedOfd.FileName, attestingDocument );
			if( !processWindow.ShowProgressDialog() )
				return null;

			// If the attesting document was created now, set its attested MAC.
			if( BufferUtils.NoContent( attestingDocument.AttestedMacValue ) )
				attestingDocument.AttestedMacValue = processWindow.MacValue;

			// If the computed MAC and the existing one are different, don't allow the user to update the signature.
			if( !BufferUtils.Equals( processWindow.MacValue, attestingDocument.AttestedMacValue ) )
			{
				throw new ValidationException( "The computed MAC and the existing one are different for the attesting document."
						+ " If you are sure that the file has changed, delete the attesting file and start again.", null, null
					, new ExceptionDetail( "File name", attestingFileName ) );
			}

			#region Create the attesting document

			// Sign the attesting document.
			attestingDocument.Sign( vault.ProcessingIdentity, null );

			// Save the attesting document.
			ISerializer serializer = algorithmPool.GetSerializerForExport();
			serializer.SaveToFile( attestingDocument, null, attestingFileName );

			#endregion Create the attesting document

			MessageWindow.ShowDialog( "The attesting document was created.", MessageWindowModes.Info );

			return attestingFileName;
		}

		/// <summary>
		/// Verifies the (signature of the) attesting document of a file with one of the application's root presenters and
		/// with the sheafs derived from <see cref="ILocallyNamedPublicProcessingIdentitySheaf"/>.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction. This may
		/// be null if "includeSheafs" is false.</param>
		/// <param name="includeRootPresenters">If this is true then the root presenters are also used to verify the
		/// signatures.</param>
		/// <param name="assembly">The assembly from where to get the root presenters. If this is null, the application's
		/// assembly is used.</param>
		/// <param name="rootPresentersResourcePrefix">The prefix used to get the resource files which contain the root peer
		/// presenters. This may be null or empty if "includeRootPresenters" is false.</param>
		/// <param name="includeSheafs">If this is true then the sheafs derived from
		/// <see cref="ILocallyNamedPublicProcessingIdentitySheaf"/> are also used to verify the signatures.</param>
		/// <param name="signatureVerifierSheaf">Sheaf with whose identities to verify the signatures of the
		/// attesting document. This may be null if "includeRootPresenters" is true. This must be null if
		/// "includeSheafs" is true.</param>
		/// <remarks>
		/// <para>Expired documents are ignored if this doesn't affect the security, or throw an
		/// <see cref="HasExpiredException"/> exception.</para>
		/// </remarks>
		public static void VerifyAttestingDocument( IStorehouseProvider storehouseProvider, bool includeRootPresenters
			, Assembly assembly, string rootPresentersResourcePrefix, bool includeSheafs
			, ILocallyNamedPublicProcessingIdentitySheaf signatureVerifierSheaf )
		{
			if( includeSheafs )
				ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			#region Load the data

			// Attesting document.
			OpenFileDialog attestingOfd = new OpenFileDialog();
			attestingOfd.Title = "Open Attesting Document";
			attestingOfd.AddExtension = true;
			attestingOfd.DefaultExt = FileExtensions.AttestingDocument;
			attestingOfd.Filter = FileExtensions.AttestingDocumentFilter;
			attestingOfd.CheckFileExists = true;
			attestingOfd.CheckPathExists = true;
			if( !ControlsUtils.ShowDialog( attestingOfd ) )
				return;

			// File to attest.
			OpenFileDialog attestedOfd = new OpenFileDialog();
			attestedOfd.Title = "Open Attested File";
			attestedOfd.AddExtension = false;
			attestedOfd.Filter = FileExtensions.AllFilesFilter;
			attestedOfd.InitialDirectory = FileUtils.GetExistingFolderName( attestingOfd.FileName );
			attestedOfd.FileName = Path.GetFileNameWithoutExtension( attestingOfd.FileName );
			attestedOfd.CheckFileExists = true;
			attestedOfd.CheckPathExists = true;
			if( !ControlsUtils.ShowDialog( attestedOfd ) )
				return;

			IDeserializer deserializer;
			AttestingDocument attestingDocument = SerializationUtils.LoadFromFile( attestingOfd.FileName
				, typeof( AttestingDocument ), out deserializer ) as AttestingDocument;
			if( attestingDocument == null )
			{
				throw new ValidationException( "Failed to load the attesting document.", null, null
					, new ExceptionDetail( "File name", attestingOfd.FileName ) );
			}

			// Don't use expired documents.
			KernelUtils.HasExpired( attestingDocument );

			#region Check if there is any potential signer

			Sheafs signingSheafs;
			Dictionary<Locator, IIdentityPublisher> identityPublishers;
			Dictionary<Locator, PublicIdentity> identities;
			bool hasSigners = ApplicationBackendStorehouse.GetPotentialSigningIdentities( storehouseProvider, attestingDocument
				, includeRootPresenters, assembly, rootPresentersResourcePrefix, includeSheafs, signatureVerifierSheaf
				, out signingSheafs, out identityPublishers, out identities );

			if( !hasSigners )
			{
				throw new ValidationException( "No potential signers of the attested file were found. If you have attested the"
						+ " file, use your vault to verify the attesting document.", null, null
					, new ExceptionDetail( "File", attestedOfd.FileName ) );
			}

			#endregion Check if there is any potential signer

			#endregion Load the data

			// Compute the MAC of the file to attest.
			ProcessAttestingDocumentWindow processWindow = new ProcessAttestingDocumentWindow( "Verify attesting document"
				, attestedOfd.FileName, attestingDocument );
			if( !processWindow.ShowProgressDialog() )
				return;

			#region Verify the signature

			string entityNames = ApplicationBackendStorehouse.GetSigningEntityNames( storehouseProvider, attestingDocument
				, processWindow.MacValue, includeRootPresenters, assembly, rootPresentersResourcePrefix
				, includeSheafs, signatureVerifierSheaf, SigningEntityNameTimeLineTemplate, SigningEntityNameTimeLineSeparator );

			bool hasSignatories = TextUtils.HasContent( entityNames );
			string message = hasSignatories
				? string.Format( "The attested file was signed by:\n{0}\n\nFile \"{1}\".", entityNames
					, attestedOfd.FileName )
				: string.Format( "Nobody has signed the file. If you have attested the file, use your vault to verify"
					+ " the attesting document.\n\nFile \"{0}\".", attestedOfd.FileName );
			MessageWindow.ShowDialog( message, hasSignatories ? MessageWindowModes.Info : MessageWindowModes.Warning );

			#endregion Verify the signature
		}

		/// <summary>
		/// Verifies the (signature of the) attesting document of a file with the identity of the vault with the specified
		/// locator. An <see cref="Vela.Crypto.InvalidSignatureException"/> (among others) is thrown if the signature is
		/// invalid.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="vaultLocator">The locator of the vault with which to verify the signature.</param>
		/// <remarks>
		/// <para>Expired documents are ignored if this doesn't affect the security, or throw an
		/// <see cref="HasExpiredException"/> exception.</para>
		/// </remarks>
		public static void VerifyAttestingDocumentWithVault( IStorehouseProvider storehouseProvider, Locator vaultLocator )
		{
			ValidationUtils.ObjectHasContent( vaultLocator, "vaultLocator" );

			Vault vault = ApplicationBackendStorehouse.LoadVault( storehouseProvider, vaultLocator );
			if( vault == null )
			{
				throw new ValidationException( "The vault is missing.", null, null
					, new ExceptionDetail( "Vault locator", vaultLocator ) );
			}

			VerifyAttestingDocument( storehouseProvider, false, null, null, false, vault );
		}

		/// <summary>
		/// Verifies the (signature of the) attesting document of a file with the identity of the local vault of the account with
		/// the specified locator. An <see cref="Vela.Crypto.InvalidSignatureException"/> (among others) is thrown if the
		/// signature is invalid.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="accountLocator">The locator of the account with whose remote identity to verify the signature.
		/// </param>
		/// <remarks>
		/// <para>Expired documents are ignored if this doesn't affect the security, or throw an
		/// <see cref="HasExpiredException"/> exception.</para>
		/// </remarks>
		public static void VerifyAttestingDocumentWithAccount( IStorehouseProvider storehouseProvider
			, Locator accountLocator )
		{
			Tally tally = ApplicationBackendStorehouse.LoadTallyOfAccount( storehouseProvider, accountLocator );
			if( tally == null )
			{
				throw new ValidationException( "The tally is missing.", null, null
					, new ExceptionDetail( "Account locator", accountLocator ) );
			}

			VerifyAttestingDocument( storehouseProvider, false, null, null, false, tally );
		}

		/// <summary>
		/// Creates a secure document, signs a file with the local vault of the account with the specified locator, and encrypts
		/// it with the identity of the remote peer of the account with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="algorithmPool">Algorithm pool.</param>
		/// <param name="accountLocator">The locator of the account for which the securing file is meant.</param>
		/// <returns>Returns the name of the file which contains the securing document, or null if one was not created.
		/// </returns>
		/// <remarks>
		/// <para>Expired documents throw an <see cref="HasExpiredException"/> exception.</para>
		/// </remarks>
		public static string CreateSecuringDocument( IStorehouseProvider storehouseProvider, AlgorithmPool algorithmPool
			, Locator accountLocator )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( algorithmPool, "algorithmPool" );
			ValidationUtils.ObjectHasContent( accountLocator, "accountLocator" );

			#region Load the data

			Vault vault;
			Tally tally;
			ApplicationBackendStorehouse.LoadVaultTally( storehouseProvider, accountLocator, out vault, out tally );

			// Don't use expired documents.
			KernelUtils.HasExpired( vault );
			KernelUtils.HasExpired( tally );

			// File to secure.
			OpenFileDialog securedOfd = new OpenFileDialog();
			securedOfd.Title = "Open File to Secure";
			securedOfd.AddExtension = false;
			securedOfd.Filter = FileExtensions.AllFilesFilter;
			securedOfd.CheckFileExists = true;
			securedOfd.CheckPathExists = true;
			if( !ControlsUtils.ShowDialog( securedOfd ) )
				return null;

			// Securing document.
			string securingFileName;
			if( !ControlsUtils.ShowSaveFileDialog( "Save Securing Document", FileExtensions.SecuringDocument
				, FileExtensions.SecuringDocumentFilter, FileUtils.GetExistingFolderName( securedOfd.FileName )
				, Path.GetFileName( securedOfd.FileName ) + FileExtensions.SecuringDocument, true
				, out securingFileName ) )
			{
				return null;
			}

			// Secure / destination file.
			string secureFileName;
			if( !ControlsUtils.ShowSaveFileDialog( "Save Secure / Destination File", FileExtensions.SecureFile
				, FileExtensions.SecureFileFilter, FileUtils.GetExistingFolderName( securingFileName )
				, Path.GetFileNameWithoutExtension( securingFileName ) + FileExtensions.SecureFile, true
				, out secureFileName ) )
			{
				return null;
			}

			#endregion Load the data

			MacAlg macAlg = algorithmPool.GetMacAlg();
			CompressionAlg compressionAlg = algorithmPool.GetCompressionAlg();
			SymmAlg symmAlg = algorithmPool.GetSymmAlg();
			IAsymmEncryptionPaddingAlg asymmEncryptionPaddingAlg = algorithmPool.GetAsymmEncryptionPadding();
			IAsymmSigningPaddingAlg asymmSigningPaddingAlg = algorithmPool.GetAsymmSigningPadding();
			int encryptedDocumentBlockSize = algorithmPool.GetSecuredDocumentBlockSize();

			#region Create the securing document

			// Unlike an attesting document, a securing document is simply replaced because it can't be decrypted in order to
			// update the signature.

			// Generate the salt and the encryption key.
			byte[] salt = EntropyPool.GetBytes( symmAlg.AlgSizeBytes );
			byte[] symmEncryptionKey = EntropyPool.GetBytes( symmAlg.AlgSizeBytes );

			// Create the securing document.
			SecuringDocument securingDocument = new SecuringDocument( CryptoUtils.GetLocator(), macAlg
				, asymmSigningPaddingAlg, algorithmPool.GetHashSerializer( false ), vault.LocalUpnFingerprint
				, tally.RemoteUpnFingerprint, symmAlg, salt, symmEncryptionKey, encryptedDocumentBlockSize );

			// Don't use expired documents.
			KernelUtils.HasExpired( securingDocument );

			#endregion Create the securing document

			// Encrypt the file.
			ProcessSecuringDocumentWindow processWindow = new ProcessSecuringDocumentWindow( "Create securing document"
				, securedOfd.FileName, secureFileName, securingDocument, true );
			if( !processWindow.ShowProgressDialog() )
				return null;

			// Update and sign the securing document.
			securingDocument.AttestedMacValue = processWindow.MacValue;
			securingDocument.Sign( vault.ProcessingIdentity, null );

			#region Save the securing document

			// Create the asymmetrically encrypted envelope.
			AsymmEncryptedEnvelope asymmEnvelope = new AsymmEncryptedEnvelope( CryptoUtils.GetLocator(), macAlg, compressionAlg
				, symmAlg, asymmEncryptionPaddingAlg );
			asymmEnvelope.EncryptableDocument = securingDocument;

			// Save the securing document in its encrypted envelope.
			ISerializer serializer = algorithmPool.GetSerializerForExport();
			asymmEnvelope.SaveToFile( securingFileName, null, tally.ProcessingIdentity, serializer );

			#endregion Save the securing document

			MessageWindow.ShowDialog( "The securing document was created.", MessageWindowModes.Info );

			return securingFileName;
		}

		/// <summary>
		/// Unpacks a securing document with the sheafs derived from <see cref="ILocallyNamedPublicProcessingIdentitySheaf"/>,
		/// by decrypting it and verifying its signature.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="includeSheafs">If this is true then the sheafs derived from
		/// <see cref="ILocallyNamedPrivateProcessingIdentitySheaf"/> are used for decryption, and those derived from
		/// <see cref="ILocallyNamedPublicProcessingIdentitySheaf"/> are also used to verify the signatures.</param>
		/// <param name="signatureVerifierSheaf">Sheaf with whose identities to verify the signatures of the
		/// securing document. This must be null if "includeSheafs" is true.</param>
		/// <param name="decryptorSheaf">Sheaf with whose identities to decrypt the securing document. This
		/// must be null if "includeSheafs" is true.</param>
		/// <remarks>
		/// <para>Expired documents are ignored if this doesn't affect the security, or throw an
		/// <see cref="HasExpiredException"/> exception.</para>
		/// </remarks>
		public static void UnpackSecuringDocument( IStorehouseProvider storehouseProvider, bool includeSheafs
			, ILocallyNamedPrivateProcessingIdentitySheaf decryptorSheaf
			, ILocallyNamedPublicProcessingIdentitySheaf signatureVerifierSheaf )
		{
			#region Validate data

			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			if( !includeSheafs )
			{
				ValidationUtils.ObjectHasContent( signatureVerifierSheaf, "signatureVerifierSheaf" );
				ValidationUtils.ObjectHasContent( decryptorSheaf, "decryptorSheaf" );
			}

			if( includeSheafs && ((signatureVerifierSheaf != null) || (decryptorSheaf != null)) )
			{
				throw new ValidationException( "If the sheafs are used to get the encrypting and signing documents, the"
					+ " individual sheafs must be missing." );
			}

			#endregion Validate data

			#region Load the data

			// Securing document.
			OpenFileDialog securingOfd = new OpenFileDialog();
			securingOfd.Title = "Open Securing Document";
			securingOfd.AddExtension = true;
			securingOfd.DefaultExt = FileExtensions.SecuringDocument;
			securingOfd.Filter = FileExtensions.SecuringDocumentFilter;
			securingOfd.CheckFileExists = true;
			securingOfd.CheckPathExists = true;
			if( !ControlsUtils.ShowDialog( securingOfd ) )
				return;

			// Secure file.
			OpenFileDialog secureOfd = new OpenFileDialog();
			secureOfd.Title = "Open Secure File";
			secureOfd.AddExtension = true;
			secureOfd.DefaultExt = FileExtensions.SecureFile;
			secureOfd.Filter = FileExtensions.SecureFileFilter;
			secureOfd.InitialDirectory = FileUtils.GetExistingFolderName( securingOfd.FileName );
			secureOfd.FileName = Path.GetFileNameWithoutExtension( securingOfd.FileName ) + FileExtensions.SecureFile;
			secureOfd.CheckFileExists = true;
			secureOfd.CheckPathExists = true;
			if( !ControlsUtils.ShowDialog( secureOfd ) )
				return;

			// Unpacked / destination file.
			string destinationFileName;
			if( !ControlsUtils.ShowSaveFileDialog( "Save Unpacked / Destination File", null, FileExtensions.AllFilesFilter
				, FileUtils.GetExistingFolderName( secureOfd.FileName ), Path.GetFileNameWithoutExtension( secureOfd.FileName )
				, true, out destinationFileName ) )
			{
				return;
			}

			#endregion Load the data

			SecuringDocument securingDocument = ApplicationBackendStorehouse.LoadAsymmEncryptedDocument( storehouseProvider
				, securingOfd.FileName, includeSheafs, decryptorSheaf, typeof( SecuringDocument ) ) as SecuringDocument;

			// Don't use expired documents.
			KernelUtils.HasExpired( securingDocument );

			// Decrypt the file.
			ProcessSecuringDocumentWindow processWindow = new ProcessSecuringDocumentWindow( "Unpack securing document"
				, secureOfd.FileName, destinationFileName, securingDocument, false );
			if( !processWindow.ShowProgressDialog() )
				return;

			#region Verify the signature

			string entityNames = ApplicationBackendStorehouse.GetSigningEntityNames( storehouseProvider, securingDocument
				, processWindow.MacValue, false, null, null, includeSheafs, signatureVerifierSheaf
				, SigningEntityNameTimeLineTemplate, SigningEntityNameTimeLineSeparator );

			bool hasSignatories = TextUtils.HasContent( entityNames );
			string message = hasSignatories
				? string.Format( "The unpacked file was signed by:\n{0}\n\nFile \"{1}\".", entityNames, destinationFileName )
				: string.Format( "Nobody has signed the file.\n\nFile \"{0}\".", destinationFileName );
			MessageWindow.ShowDialog( message, hasSignatories ? MessageWindowModes.Info : MessageWindowModes.Warning );

			#endregion Verify the signature
		}

		/// <summary>
		/// Unpacks a secure document, decrypts it with the local vault, and verifies its signature with the identity of the
		/// remote peer of the current account.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="accountLocator">The locator of the account with which to unpack the securing file.</param>
		/// <remarks>
		/// <para>Expired documents are ignored if this doesn't affect the security, or throw an
		/// <see cref="HasExpiredException"/> exception.</para>
		/// </remarks>
		public static void UnpackSecuringDocument( IStorehouseProvider storehouseProvider, Locator accountLocator )
		{
			Vault vault;
			Tally tally;
			ApplicationBackendStorehouse.LoadVaultTally( storehouseProvider, accountLocator, out vault, out tally );

			UnpackSecuringDocument( storehouseProvider, false, vault, tally );
		}

		/// <summary>
		/// Saves the message of the draft letter.
		/// </summary>
		/// <param name="storehouseProvider">The storehouse provider whose configuration to update.</param>
		/// <param name="message">The message of the draft letter. This may be null or empty.</param>
		/// <remarks>
		/// <para>The configuration is first loaded from the storehouse, the message is set in it, and then the configuration
		/// is updated in the storehouse.</para>
		/// 
		/// <para>During the execution of this method, the configuration is locked with <see cref="SheafLocks.Update"/>.</para>
		/// 
		/// <para>Because a transaction is created in here, if you call this method while another transaction is active, it's
		/// possible that a lock timeout will occur.</para>
		/// </remarks>
		public static void SaveDraftLetterMessage( IStorehouseProvider storehouseProvider, string message )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the configuration.
				ImStorehouseConfiguration configuration
					= StorehouseUtils.LoadConfiguration<ImStorehouseConfiguration>( transaction, SheafLocks.Update );

				// Set the message.
				configuration.DraftLetterMessage = message;

				// Update the configuration.
				transaction.UpdateConfiguration( configuration );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		#endregion Public static methods
	}
}
