////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.ActionProcessor;
using Vela.ApplicationBackend;
using Vela.Controls;
using Vela.DatabaseStorehouse;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.StarkDocuments;
using Vela.Storehouse;

namespace Vela.ImBackend
{
	/// <summary>
	/// Application configuration.
	/// </summary>
	public class ImConfiguration : StorableConfiguration
	{
		#region Public constants

		/// <summary>
		/// The title of the file where the configuration is stored.
		/// </summary>
		public const string FileTitle = "App" + FileExtensions.StorableConfiguration;

		/// <summary>
		/// The default value for <see cref="LogDebugMessages"/>.
		/// </summary>
		/// <remarks>
		/// <para>Setting this with true may have dramatic performance implications because it activates writing the log
		/// messages to the system trace, and the usage of the <see cref="LogManager.Added"/> event (event which in turn may
		/// update the UI through the "Dispatcher.Invoke" method).</para>
		/// </remarks>
		public const bool LogDebugMessagesDef = false;

		/// <summary>
		/// The default value for <see cref="SaveDebugMessages"/>.
		/// </summary>
		public const bool SaveDebugMessagesDef = false;

		/// <summary>
		/// The type name of the storehouse provider to instantiate.
		/// </summary>
		/// <remarks>
		/// <para>For SQL Server CE, use "Vela.MsSqlCeDatabase.MsSqlCeDatabaseProvider".</para>
		/// 
		/// <para>For SQL Server FE, use "Vela.MsSqlFeDatabase.MsSqlFeDatabaseProvider".</para>
		/// </remarks>
		public const string StorehouseProviderTypeName = "Vela.MsSqlCeDatabase.MsSqlCeDatabaseProvider";

		/// <summary>
		/// Template for the database connection string. Placeholders:
		/// <see cref="StorehouseConsts.StorehouseFileNamePlaceholder"/> - name of file attached to the storehouse,
		/// <see cref="StorehouseConsts.ServerStorehouseNamePlaceholder"/> - server storehouse name,
		/// <see cref="StorehouseConsts.StorehouseAttachmentsPasswordPlaceholder"/> - storehouse attachments password.
		/// </summary>
		/// <remarks>
		/// <para>For SQL Server CE, use "Data source='{StorehouseFileName}'; Max database size=4091; File mode=Read Write;
		/// Encryption mode=Engine Default; Persist security info=False; Password={StorehouseAttachmentsPassword}; Enlist=False;
		/// Default lock timeout=5000; Autoshrink threshold=100;".</para>
		/// 
		/// <para>For SQL Server FE, use "Server=(local); Database={ServerStorehouseName}; Persist security info=False;
		/// User Id={ServerStorehouseName}; Password={StorehouseAttachmentsPassword}; Enlist=False;".</para>
		/// </remarks>
		public const string DatabaseConnectionTemplate = "Data source='{StorehouseFileName}'; Max database size=4091;"
			+ " File mode=Read Write; Encryption mode=Engine Default; Persist security info=False;"
			+ " Password={StorehouseAttachmentsPassword}; Enlist=False; Default lock timeout=5000; Autoshrink threshold=100;";

		/// <summary>
		/// The default value for <see cref="AutoStartApplicationWithSystem"/>.
		/// </summary>
		public const bool AutoStartApplicationWithSystemDef = true;
		
		/// <summary>
		/// The default value for <see cref="AutoStartStorehouseWithApplication"/>.
		/// </summary>
		public const bool AutoStartStorehouseWithApplicationDef = true;

		/// <summary>
		/// The default value for <see cref="SendDraftLetterWithPlainEnter"/>.
		/// </summary>
		public const bool SendDraftLetterWithPlainEnterDef = true;

		/// <summary>
		/// The default value for <see cref="MandatoryAccountAliasWhenSendingLetter"/>.
		/// </summary>
		public const bool MandatoryAccountAliasWhenSendingLetterDef = false;

		/// <summary>
		/// The default value for <see cref="ShowStateTimeInLetters"/>.
		/// </summary>
		public const bool ShowStateTimeInLettersDef = true;

		/// <summary>
		/// The default value for <see cref="ShowLettersSeparator"/>.
		/// </summary>
		public const bool ShowLettersSeparatorDef = false;

		/// <summary>
		/// The default value for <see cref="ShowPeerDirectionIndicatorInLetters"/>.
		/// </summary>
		public const bool ShowPeerDirectionIndicatorInLettersDef = true;

		/// <summary>
		/// The default value for <see cref="ShowFullSizedEmoticonsInLetters"/>.
		/// </summary>
		public const bool ShowFullSizedEmoticonsInLettersDef = false;

		/// <summary>
		/// The default value for <see cref="ShowPreviewsForAttachedImageFilesCount"/>.
		/// </summary>
		public const int ShowPreviewsForAttachedImageFilesCountDef = 3;

		/// <summary>
		/// The default value for <see cref="ShowLetterMessageSummaryInDesktopNotification"/>.
		/// </summary>
		public const bool ShowLetterMessageSummaryInDesktopNotificationDef = false;

		/// <summary>
		/// The default value for <see cref="AllowAudioNotification"/>.
		/// </summary>
		public const bool AllowAudioNotificationDef = true;

		/// <summary>
		/// The default value for <see cref="BellAudioNotificationFileName"/>.
		/// </summary>
		public const string BellAudioNotificationFileNameDef = @"Media\Ring02.wav";

		/// <summary>
		/// The default value for <see cref="HornAudioNotificationFileName"/>.
		/// </summary>
		public const string HornAudioNotificationFileNameDef = @"Media\Ring08.wav";

		#region Legacy files

		/// <summary>
		/// The default value for <see cref="BellAudioNotificationFileName"/>, for Windows 7.
		/// </summary>
		public const string BellAudioNotificationFileNameDefWindows7 = @"Media\Ringout.wav";

		/// <summary>
		/// The default value for <see cref="HornAudioNotificationFileName"/>, for Windows 7.
		/// </summary>
		public const string HornAudioNotificationFileNameDefWindows7 = @"Media\Notify.wav";

		#endregion Legacy files

		/// <summary>
		/// The default value for <see cref="PlayVideoFilesCommandLine"/>.
		/// </summary>
		public const string PlayVideoFilesCommandLineDef = "\"C:\\Program Files (x86)\\VideoLAN\\VLC\\vlc.exe\" --one-instance"
			+ " --no-qt-video-autoresize --playlist-enqueue --playlist-tree --playlist-autostart --no-video-title-show"
			+ " \"{PlaylistFileName}\"";

		/// <summary>
		/// The default value for <see cref="PlayAudioFilesCommandLine"/>.
		/// </summary>
		public const string PlayAudioFilesCommandLineDef = PlayVideoFilesCommandLineDef;

		/// <summary>
		/// The default value for <see cref="PlayImageFilesCommandLine"/>.
		/// </summary>
		public const string PlayImageFilesCommandLineDef = PlayVideoFilesCommandLineDef;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public ImConfiguration()
		{
			#region Audio notifications

			string osPath = Environment.GetFolderPath( Environment.SpecialFolder.Windows );

			string bellAudioNotificationFileName;
			string hornAudioNotificationFileName;

			if( SystemUtils.IsOsAtLeastWindows8() )
			{
				bellAudioNotificationFileName = BellAudioNotificationFileNameDef;
				hornAudioNotificationFileName = HornAudioNotificationFileNameDef;
			}
			else
			{
				bellAudioNotificationFileName = BellAudioNotificationFileNameDefWindows7;
				hornAudioNotificationFileName = HornAudioNotificationFileNameDefWindows7;
			}

			BellAudioNotificationFileName = FileUtils.MakeFullName( osPath, bellAudioNotificationFileName );
			HornAudioNotificationFileName = FileUtils.MakeFullName( osPath, hornAudioNotificationFileName );

			#endregion Audio notifications
		}

		#endregion Constructors

		#region Public serializable properties

		private bool _LogDebugMessages = LogDebugMessagesDef;
		/// <summary>
		/// <see cref="LogManager.LogDebugMessages"/> is set with this. Notifies property observers.
		/// </summary>
		[SerializeAttribute]
		public bool LogDebugMessages
		{
			get { return _LogDebugMessages; }
			set
			{
				_LogDebugMessages = value;

				// Notify property observers.
				OnPropertyChanged( "LogDebugMessages" );
			}
		}

		private bool _SaveDebugMessages = SaveDebugMessagesDef;
		/// <summary>
		/// <see cref="LogManager.SaveDebugMessages"/> is set with this. Notifies property observers.
		/// </summary>
		[SerializeAttribute]
		public bool SaveDebugMessages
		{
			get { return _SaveDebugMessages; }
			set
			{
				_SaveDebugMessages = value;

				// Notify property observers.
				OnPropertyChanged( "SaveDebugMessages" );
			}
		}

		private string _LogMessagesWhitelist;
		/// <summary>
		/// <see cref="LogManager.LogMessagesWhitelist"/> is set with this. This may be null or empty. Notifies property
		/// observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string LogMessagesWhitelist
		{
			get { return _LogMessagesWhitelist; }
			set
			{
				_LogMessagesWhitelist = value;

				// Notify property observers.
				OnPropertyChanged( "LogMessagesWhitelist" );
			}
		}

		private string _LogMessagesBlacklist;
		/// <summary>
		/// <see cref="LogManager.LogMessagesBlacklist"/> is set with this. This may be null or empty. Notifies property
		/// observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string LogMessagesBlacklist
		{
			get { return _LogMessagesBlacklist; }
			set
			{
				_LogMessagesBlacklist = value;

				// Notify property observers.
				OnPropertyChanged( "LogMessagesBlacklist" );
			}
		}

		private bool _AutoStartApplicationWithSystem = AutoStartApplicationWithSystemDef;
		/// <summary>
		/// If this is true then the application will start when the system starts. Notifies property observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public bool AutoStartApplicationWithSystem
		{
			get { return _AutoStartApplicationWithSystem; }
			set
			{
				_AutoStartApplicationWithSystem = value;

				AutoStartUtils.Set( AutoStartApplicationWithSystem );

				// Notify property observers.
				OnPropertyChanged( "AutoStartApplicationWithSystem" );
			}
		}

		private bool _AutoStartStorehouseWithApplication = AutoStartStorehouseWithApplicationDef;
		/// <summary>
		/// If this is true then the last opened storehouse (<see cref="LastOpenedStorehouseFileName"/>) is automatically opened
		/// when the application starts; the password of the storehouse must be
		/// <see cref="ImBackendConsts.StorehousePasswordDef"/>. Notifies property observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public bool AutoStartStorehouseWithApplication
		{
			get { return _AutoStartStorehouseWithApplication; }
			set
			{
				_AutoStartStorehouseWithApplication = value;

				// Notify property observers.
				OnPropertyChanged( "AutoStartStorehouseWithApplication" );
			}
		}

		private IStorehouseAccess _StorehouseAccess;
		/// <summary>
		/// The storehouse access.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		[SerializeAttribute]
		public IStorehouseAccess StorehouseAccess
		{
			get
			{
				if( _StorehouseAccess != null )
					return _StorehouseAccess;

				_StorehouseAccess = new DatabaseAccess( StorehouseProviderTypeName, DatabaseConnectionTemplate );

				return _StorehouseAccess;
			}
			set
			{
				ValidationUtils.ObjectHasContent( value, "StorehouseAccess" );

				_StorehouseAccess = value;
			}
		}

		private ActionProcessorConfiguration _ActionProcessor;
		/// <summary>		
		/// The action processor configuration.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		[SerializeAttribute]
		public ActionProcessorConfiguration ActionProcessor
		{
			get
			{
				if( _ActionProcessor != null )
					return _ActionProcessor;

				_ActionProcessor = AlgorithmPool.Instance.GetActionProcessorConfiguration();

				return _ActionProcessor;
			}
			set
			{
				ValidationUtils.ObjectHasContent( value, "ActionProcessor" );

				_ActionProcessor = value;
			}
		}

		private ActionServiceListenerConfiguration _ActionServiceListenerTemplate;
		/// <summary>
		/// Template for an action service listener configuration.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		[SerializeAttribute]
		public ActionServiceListenerConfiguration ActionServiceListenerTemplate
		{
			get
			{
				if( _ActionServiceListenerTemplate != null )
					return _ActionServiceListenerTemplate;

				_ActionServiceListenerTemplate = AlgorithmPool.Instance.GetActionServiceListenerConfiguration();

				return _ActionServiceListenerTemplate;
			}
			set
			{
				ValidationUtils.ObjectHasContent( value, "ActionServiceListenerTemplate" );

				_ActionServiceListenerTemplate = value;
			}
		}

		private bool _SendDraftLetterWithPlainEnter = SendDraftLetterWithPlainEnterDef;
		/// <summary>
		/// If this is true then the draft letter can be sent by pressing the "Enter" key, else "Ctrl + Enter". Notifies property
		/// observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public bool SendDraftLetterWithPlainEnter
		{
			get { return _SendDraftLetterWithPlainEnter; }
			set
			{
				_SendDraftLetterWithPlainEnter = value;

				// Notify property observers.
				OnPropertyChanged( "SendDraftLetterWithPlainEnter" );
			}
		}

		private bool _MandatoryAccountAliasWhenSendingLetter = MandatoryAccountAliasWhenSendingLetterDef;
		/// <summary>
		/// If this is true then the draft letter may be sent only if it's prefixed with an account alias. Notifies property
		/// observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public bool MandatoryAccountAliasWhenSendingLetter
		{
			get { return _MandatoryAccountAliasWhenSendingLetter; }
			set
			{
				_MandatoryAccountAliasWhenSendingLetter = value;

				// Notify property observers.
				OnPropertyChanged( "MandatoryAccountAliasWhenSendingLetter" );
			}
		}

		private bool _ShowStateTimeInLetters = ShowStateTimeInLettersDef;
		/// <summary>
		/// If this is true then the current state time is shown in the "Letters" tab. Notifies property observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public bool ShowStateTimeInLetters
		{
			get { return _ShowStateTimeInLetters; }
			set
			{
				_ShowStateTimeInLetters = value;

				// Notify property observers.
				OnPropertyChanged( "ShowStateTimeInLetters" );
			}
		}

		private bool _ShowLettersSeparator = ShowLettersSeparatorDef;
		/// <summary>
		/// If this is true then the letters shown in the "Letters" tab are separated by a horizontal line. Notifies property
		/// observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public bool ShowLettersSeparator
		{
			get { return _ShowLettersSeparator; }
			set
			{
				_ShowLettersSeparator = value;

				// Notify property observers.
				OnPropertyChanged( "ShowLettersSeparator" );
			}
		}

		private bool _ShowPeerDirectionIndicatorInLetters = ShowPeerDirectionIndicatorInLettersDef;
		/// <summary>
		/// If this is true then the letters shown in the "Letters" tab have a vertical bar on the left, bar which indicates the
		/// peer direction. Notifies property observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public bool ShowPeerDirectionIndicatorInLetters
		{
			get { return _ShowPeerDirectionIndicatorInLetters; }
			set
			{
				_ShowPeerDirectionIndicatorInLetters = value;

				// Notify property observers.
				OnPropertyChanged( "ShowPeerDirectionIndicatorInLetters" );
			}
		}

		private bool _ShowFullSizedEmoticonsInLetters = ShowFullSizedEmoticonsInLettersDef;
		/// <summary>
		/// If this is true then the full-sized emoticons are shown in the "Letters" tab. Notifies property observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public bool ShowFullSizedEmoticonsInLetters
		{
			get { return _ShowFullSizedEmoticonsInLetters; }
			set
			{
				_ShowFullSizedEmoticonsInLetters = value;

				// Notify property observers.
				OnPropertyChanged( "ShowFullSizedEmoticonsInLetters" );
			}
		}

		private int _ShowPreviewsForAttachedImageFilesCount = ShowPreviewsForAttachedImageFilesCountDef;
		/// <summary>
		/// The number of attached image files for which image previews can be shown.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public int ShowPreviewsForAttachedImageFilesCount
		{
			get { return _ShowPreviewsForAttachedImageFilesCount; }
			set
			{
				ValidationUtils.Integer( value, 1, PeerDocumentsConsts.ShowPreviewsForAttachedImageFilesCountMax
					, "ShowPreviewsForAttachedImageFilesCount" );

				_ShowPreviewsForAttachedImageFilesCount = value;

				// Notify property observers.
				OnPropertyChanged( "ShowPreviewsForAttachedImageFilesCount" );
			}
		}

		private bool _ShowLetterMessageSummaryInDesktopNotification = ShowLetterMessageSummaryInDesktopNotificationDef;
		/// <summary>
		/// If this is true then the message summary of the received letters is shown in the desktop notifications. Notifies
		/// property observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public bool ShowLetterMessageSummaryInDesktopNotification
		{
			get { return _ShowLetterMessageSummaryInDesktopNotification; }
			set
			{
				_ShowLetterMessageSummaryInDesktopNotification = value;

				// Notify property observers.
				OnPropertyChanged( "ShowLetterMessageSummaryInDesktopNotification" );
			}
		}

		private ShortCommands _ShortCommands;
		/// <summary>
		/// The URLs to use when a search is performed.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		[SerializeAttribute]
		public ShortCommands ShortCommands
		{
			get
			{
				if( _ShortCommands != null )
					return _ShortCommands;

				_ShortCommands = ShortCommands.GetInitializedInstance();

				return _ShortCommands;
			}
			set
			{
				ValidationUtils.ObjectHasContent( value, "ShortCommands" );

				_ShortCommands = value;
			}
		}

		private bool _AllowAudioNotification = AllowAudioNotificationDef;
		/// <summary>
		/// If this is true then audio notifications are allowed. The default value is true. Notifies property observers.
		/// </summary>
		[SerializeAttribute]
		public bool AllowAudioNotification
		{
			get { return _AllowAudioNotification; }
			set
			{
				_AllowAudioNotification = value;

				// Notify property observers.
				OnPropertyChanged( "AllowAudioNotification" );
			}
		}

		private string _BellAudioNotificationFileName;
		/// <summary>
		/// The path of the file which contains the <see cref="Vela.StarkDocuments.EmoticonCommands.Bell"/> audio notification.
		/// This may be null or empty. Notifies property observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string BellAudioNotificationFileName
		{
			get { return _BellAudioNotificationFileName; }
			set
			{
				_BellAudioNotificationFileName = value;

				// Notify property observers.
				OnPropertyChanged( "BellAudioNotificationFileName" );
			}
		}

		private string _HornAudioNotificationFileName;
		/// <summary>
		/// The path of the file which contains the <see cref="Vela.StarkDocuments.EmoticonCommands.Horn"/> audio notification.
		/// This may be null or empty. Notifies property observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string HornAudioNotificationFileName
		{
			get { return _HornAudioNotificationFileName; }
			set
			{
				_HornAudioNotificationFileName = value;

				// Notify property observers.
				OnPropertyChanged( "HornAudioNotificationFileName" );
			}
		}

		private string _PlayVideoFilesCommandLine = PlayVideoFilesCommandLineDef;
		/// <summary>
		/// The command line to execute when playing the video files attached to a letter. This may be null or empty. Notifies
		/// property observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string PlayVideoFilesCommandLine
		{
			get { return _PlayVideoFilesCommandLine; }
			set
			{
				_PlayVideoFilesCommandLine = value;

				// Notify property observers.
				OnPropertyChanged( "PlayVideoFilesCommandLine" );
			}
		}

		private string _PlayAudioFilesCommandLine = PlayAudioFilesCommandLineDef;
		/// <summary>
		/// The command line to execute when playing the audio files attached to a letter. This may be null or empty. Notifies
		/// property observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string PlayAudioFilesCommandLine
		{
			get { return _PlayAudioFilesCommandLine; }
			set
			{
				_PlayAudioFilesCommandLine = value;

				// Notify property observers.
				OnPropertyChanged( "PlayAudioFilesCommandLine" );
			}
		}

		private string _PlayImageFilesCommandLine = PlayImageFilesCommandLineDef;
		/// <summary>
		/// The command line to execute when playing the image files attached to a letter. This may be null or empty. Notifies
		/// property observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string PlayImageFilesCommandLine
		{
			get { return _PlayImageFilesCommandLine; }
			set
			{
				_PlayImageFilesCommandLine = value;

				// Notify property observers.
				OnPropertyChanged( "PlayImageFilesCommandLine" );
			}
		}

		private TextOnImageConfiguration _TextOnImageConfiguration = new TextOnImageConfiguration();
		/// <summary>
		/// Configuration for a text which is to be written on an image. The backing field is initialized by default. This may be
		/// null.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public TextOnImageConfiguration TextOnImageConfiguration
		{
			get { return _TextOnImageConfiguration; }
			set { _TextOnImageConfiguration = value; }
		}

		private WindowLayoutConfiguration _MainWindowLayoutConfiguration;
		/// <summary>
		/// Configuration for the layout of the main window. This may be null.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public WindowLayoutConfiguration MainWindowLayoutConfiguration
		{
			get { return _MainWindowLayoutConfiguration; }
			set { _MainWindowLayoutConfiguration = value; }
		}

		private string _LastOpenedStorehouseFileName;
		/// <summary>
		/// The name of the file of the storehouse which was last opened. This may be null or empty. Notifies property observers.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string LastOpenedStorehouseFileName
		{
			get { return _LastOpenedStorehouseFileName; }
			set
			{
				_LastOpenedStorehouseFileName = value;

				// Notify property observers.
				OnPropertyChanged( "LastOpenedStorehouseFileName" );
			}
		}

		#endregion Public serializable properties

		#region Public static methods

		/// <summary>
		/// Loads the configuration.
		/// </summary>
		/// <returns>Returns the loaded configuration object. If the configuration file doesn't exists then a new configuration
		/// object is returned.</returns>
		/// <remarks>
		/// <para>This method implements a recovery mechanism following a crash during a call of the <see cref="Save"/>
		/// method.</para>
		/// </remarks>
		public static ImConfiguration Load()
		{
			string fileName = AssemblyUtils.GetApplicationDataFileName( null, FileTitle );

			IDeserializer deserializer;

			ImConfiguration configuration = (ImConfiguration)Load( fileName, typeof( ImConfiguration )
				, out deserializer );

			LogManager.LogDebugMessages = configuration.LogDebugMessages;
			LogManager.SaveDebugMessages = configuration.SaveDebugMessages;
			LogManager.LogMessagesWhitelist = configuration.LogMessagesWhitelist;
			LogManager.LogMessagesBlacklist = configuration.LogMessagesBlacklist;

			return configuration;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Saves the configuration to the associated file.
		/// </summary>
		/// <remarks>
		/// <para>The saving is crash-safe. If the application crashes when the file is being saved, at least 1 of the
		/// following files is still available: a file with the specified name, a file with a backup name, a file with a
		/// temporary name. If the backup file exists, it's the valid one and you should implement your own recovery mechanism
		/// following a crash.</para>
		/// </remarks>
		public void Save()
		{
			Save( AlgorithmPool.Instance.GetXmlSerializer() );
		}

		#endregion Public methods
	}
}
