////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows;
using System.Windows.Media.Imaging;
using Vela.ApplicationBackend;
using Vela.Controls;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.PeerDocuments;
using Vela.Storehouse;

namespace Vela.ImBackend
{
	/// <summary>
	/// Account model.
	/// </summary>
	public class AccountModel : BaseModel
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AccountModel()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		protected AccountModel( ApplicationModel applicationModel )
			: base( applicationModel )
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		/// <param name="accountContext">The account context.</param>
		protected AccountModel( ApplicationModel applicationModel, AccountContext accountContext )
			: base( applicationModel )
		{
			StorehouseProvider = ApplicationModel.StorehouseProvider;
			AccountContext = accountContext;
		}

		#endregion Constructors

		#region Public static properties

		/// <summary>
		/// The default file name of the missing photo of a peer. This may be null or empty.
		/// </summary>
		public static string MissingPeerPhotoFileName;

		#endregion Public static properties

		#region Public non-serializable properties

		/// <summary>
		/// The account. READ-ONLY.
		/// </summary>
		public Account Account
		{
			get { return AccountContext.Account; }
		}

		/// <summary>
		/// The account locator. READ-ONLY.
		/// </summary>
		public Locator AccountLocator
		{
			get { return Account.InstanceLocator; }
		}

		/// <summary>
		/// The local UPN fingerprint who has sought (over the network) and found the contained presenter. READ-ONLY.
		/// </summary>
		public Locator LocalUpnFingerprint
		{
			get { return Account.LocalUpnFingerprint; }
		}

		/// <summary>
		/// The remote UPN fingerprint associated to the referred presenter. READ-ONLY.
		/// </summary>
		public Locator RemoteUpnFingerprint
		{
			get { return Account.RemoteUpnFingerprint; }
		}

		/// <summary>
		/// The vault of the local peer. READ-ONLY.
		/// </summary>
		public Vault Vault
		{
			get { return AccountContext.Vault; }
		}

		/// <summary>
		/// The vault locator. READ-ONLY.
		/// </summary>
		public Locator VaultLocator
		{
			get { return Vault.InstanceLocator; }
		}

		/// <summary>
		/// The entity name of the local peer. READ-ONLY.
		/// </summary>
		public string LocalEntityName
		{
			get { return Vault.Appendix.EntityName; }
		}

		/// <summary>
		/// The photo of the local peer. This may be null. READ-ONLY.
		/// </summary>
		public Photo LocalPhoto
		{
			get { return AccountContext.LocalPhoto; }
		}

		private BitmapSource _LocalSystemImage;
		/// <summary>
		/// The thumb-sized system image of the local peer's photo. This is null if <see cref="LocalPhoto"/> is null, or if its
		/// <see cref="Photo.Image"/> property is null. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public BitmapSource LocalSystemImage
		{
			get
			{
				if( _LocalSystemImage != null )
					return _LocalSystemImage;

				if( LocalPhoto != null )
				{
					PeerImage peerImage = LocalPhoto.Image as PeerImage;

					if( (peerImage != null) && (peerImage.SystemImage != null) )
					{
						_LocalSystemImage = GraphicUtils.ResizeImage( peerImage.SystemImage
							, PeerDocumentsConsts.ThumbSizedPhotoWidthInt, PeerDocumentsConsts.ThumbSizedPhotoHeightInt, true );
					}
				}

				if( (_LocalSystemImage == null) && TextUtils.HasContent( MissingPeerPhotoFileName ) )
					_LocalSystemImage = ImagePreviewsCache.GetCachedOrLoadFirstFrame( MissingPeerPhotoFileName, false );

				return _LocalSystemImage;
			}
		}

		/// <summary>
		/// The visibility of the local photo. READ-ONLY.
		/// </summary>
		public Visibility LocalPhotoVisibility
		{
			get { return (LocalSystemImage != null) ? Visibility.Visible : Visibility.Collapsed; }
		}

		/// <summary>
		/// The presenter of the remote peer. READ-ONLY.
		/// </summary>
		public Presenter Presenter
		{
			get { return AccountContext.Presenter; }
		}

		/// <summary>
		/// The presenter locator. READ-ONLY.
		/// </summary>
		public Locator PresenterLocator
		{
			get { return Presenter.InstanceLocator; }
		}

		/// <summary>
		/// The tally. READ-ONLY.
		/// </summary>
		public Tally Tally
		{
			get { return AccountContext.Tally; }
		}

		/// <summary>
		/// The tally locator. READ-ONLY.
		/// </summary>
		public Locator TallyLocator
		{
			get { return Tally.InstanceLocator; }
		}

		/// <summary>
		/// The entity name of the remote peer. READ-ONLY.
		/// </summary>
		public string RemoteEntityName
		{
			get { return Tally.EntityName; }
		}

		/// <summary>
		/// The photo of the remote peer. This may be null. READ-ONLY.
		/// </summary>
		public Photo RemotePhoto
		{
			get { return AccountContext.RemotePhoto; }
		}

		private BitmapSource _RemoteSystemImage;
		/// <summary>
		/// The thumb-sized system image of the remote peer's photo. This is null if <see cref="RemotePhoto"/> is null, or if its
		/// <see cref="Photo.Image"/> property is null. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public BitmapSource RemoteSystemImage
		{
			get
			{
				if( _RemoteSystemImage != null )
					return _RemoteSystemImage;

				if( RemotePhoto != null )
				{
					PeerImage peerImage = RemotePhoto.Image as PeerImage;

					if( (peerImage != null) && (peerImage.SystemImage != null) )
					{
						_RemoteSystemImage = GraphicUtils.ResizeImage( peerImage.SystemImage
							, PeerDocumentsConsts.ThumbSizedPhotoWidthInt, PeerDocumentsConsts.ThumbSizedPhotoHeightInt, true );
					}
				}

				if( (_RemoteSystemImage == null) && TextUtils.HasContent( MissingPeerPhotoFileName ) )
					_RemoteSystemImage = ImagePreviewsCache.GetCachedOrLoadFirstFrame( MissingPeerPhotoFileName, false );

				return _RemoteSystemImage;
			}
		}

		/// <summary>
		/// The visibility of the remote photo. READ-ONLY.
		/// </summary>
		public Visibility RemotePhotoVisibility
		{
			get { return (RemoteSystemImage != null) ? Visibility.Visible : Visibility.Collapsed; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private IStorehouseProvider _StorehouseProvider;
		/// <summary>
		/// The storehouse provider. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public IStorehouseProvider StorehouseProvider
		{
			get { return _StorehouseProvider; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "StorehouseProvider" );

				_StorehouseProvider = value;
			}
		}

		private AccountContext _AccountContext;
		/// <summary>
		/// The account context. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public AccountContext AccountContext
		{
			get { return _AccountContext; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "AccountContext" );

				_AccountContext = value;
			}
		}

		#endregion Public serializable properties

		#region Public static methods

		/// <summary>
		/// Loads the account model based on the specified account locator. If the account doesn't exist and "throwIfMissing" is
		/// true, an exception is thrown.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		/// <param name="accountLocator">The account locator.</param>
		/// <param name="throwIfMissing">If this is true and the account with the specified locator doesn't exist, an exception
		/// is thrown.</param>
		/// <returns>Returns the account model based on the specified account locator. This is null if the account doesn't exist
		/// and "throwIfMissing" is false.</returns>
		public static AccountModel Load( ApplicationModel applicationModel, Locator accountLocator, bool throwIfMissing )
		{
			ValidationUtils.ObjectHasContent( applicationModel, "applicationModel" );
			ValidationUtils.ObjectHasContent( accountLocator, "accountLocator" );

			// Create a transaction.
			IStorehouseTransaction transaction = applicationModel.StorehouseProvider.CreateTransaction();

			try
			{
				AccountContext accountContext = ApplicationBackendStorehouse.LoadAccountContext( transaction, accountLocator
					, SheafLocks.Shared, SheafLocks.Shared, throwIfMissing );
				if( (accountContext == null) && !throwIfMissing )
				{
					// Commit the transaction.
					transaction.Commit();

					return null;
				}

				AccountModel accountModel = new AccountModel( applicationModel, accountContext );

				// Commit the transaction.
				transaction.Commit();

				return accountModel;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the account model based on the specified local and remote UPN fingerprints. If the account doesn't exist and
		/// "throwIfMissing" is true, an exception is thrown.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		/// <param name="accountLocalUpnFingerprint">The local UPN fingerprint of the account.</param>
		/// <param name="accountRemoteUpnFingerprint">The remote UPN fingerprint of the account.</param>
		/// <param name="throwIfMissing">If this is true and the account with the specified UPN fingerprint doesn't exist, an
		/// exception is thrown.</param>
		/// <returns>Returns the account model based on the specified local and remote UPN fingerprints. This is null if the
		/// account doesn't exist and "throwIfMissing" is false.</returns>
		public static AccountModel Load( ApplicationModel applicationModel, Locator accountLocalUpnFingerprint
			, Locator accountRemoteUpnFingerprint, bool throwIfMissing )
		{
			ValidationUtils.ObjectHasContent( applicationModel, "applicationModel" );
			ValidationUtils.ObjectHasContent( accountLocalUpnFingerprint, "accountLocalUpnFingerprint" );
			ValidationUtils.ObjectHasContent( accountRemoteUpnFingerprint, "accountRemoteUpnFingerprint" );

			// Create a transaction.
			IStorehouseTransaction transaction = applicationModel.StorehouseProvider.CreateTransaction();

			try
			{
				#region Load the account

				SpeckleFilter accountLocalUpnSeekFilter = new SpeckleFilter( Account.LocalUpnFingerprintFilterIdentifier
					, accountLocalUpnFingerprint );
				SpeckleFilter accountRemoteUpnSeekFilter = new SpeckleFilter( Account.RemoteUpnFingerprintFilterIdentifier
					, accountRemoteUpnFingerprint );

				Speckle accountSpeckle = transaction.CollectFirstSpeckle( new PortableTypes( typeof( Account ) )
					, new SpeckleFilters( accountLocalUpnSeekFilter, accountRemoteUpnSeekFilter ), StorehouseRealms.Accounts
					, SheafLocks.Shared, null, null );

				if( accountSpeckle == null )
				{
					if( throwIfMissing )
					{
						throw new ValidationException( "The account is missing.", null, null
							, new ExceptionDetail( "Local UPN fingerprint", accountLocalUpnFingerprint )
							, new ExceptionDetail( "Remote UPN fingerprint", accountRemoteUpnFingerprint ) );
					}
					else
					{
						// Commit the transaction.
						transaction.Commit();

						return null;
					}
				}

				Locator accountLocator = accountSpeckle.StorehouseKey;

				#endregion Load the account

				AccountContext accountContext = ApplicationBackendStorehouse.LoadAccountContext( transaction, accountLocator
					, SheafLocks.Shared, SheafLocks.Shared, throwIfMissing );

				AccountModel accountModel = new AccountModel( applicationModel, accountContext );

				// Commit the transaction.
				transaction.Commit();

				return accountModel;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Checks if the sensual emoticons may be shown to the user.
		/// </summary>
		/// <returns>Returns true if sensual emoticons may be shown to the user.</returns>
		public bool MayShowSensualEmoticons()
		{
			DateTime birthDate = Vault.Appendix.VirtualIdentity.BirthDate;
			DateTime now = DateTime.UtcNow;

			return (birthDate > KernelConsts.DateMin)
				&& (birthDate.AddYears( PeerDocumentsConsts.MayShowSensualEmoticonsAge ) <= now);
		}

		#endregion Public methods
	}
}
