////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.ActionProcessor;
using Vela.ApplicationBackend;
using Vela.ApplicationDocuments;
using Vela.Controls;
using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.ImBackend
{
	/// <summary>
	/// Accounts model.
	/// </summary>
	public class AccountsModel : BaseModel
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AccountsModel()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		public AccountsModel( ApplicationModel applicationModel )
			: base( applicationModel )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		private AccountOrRequestSpeckles _Accounts;
		/// <summary>
		/// The speckles with the accounts. This may be empty, but not null. Notifies property observers. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public AccountOrRequestSpeckles Accounts
		{
			get
			{
				if( _Accounts != null )
					return _Accounts;

				CollectAccounts();

				return _Accounts;
			}
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Accounts" );

				_Accounts = value;

				// Notify property observers.
				OnPropertyChanged( "Accounts" );

				CurrentAccount = null;
			}
		}

		private AccountOrRequestSpeckle _CurrentAccount;
		/// <summary>
		/// The current account speckle. This may be null or empty. Notifies property observers.
		/// </summary>
		public AccountOrRequestSpeckle CurrentAccount
		{
			get { return _CurrentAccount; }
			set
			{
				_CurrentAccount = value;

				// Notify property observers.
				OnPropertyChanged( "CurrentAccount" );
				OnPropertyChanged( "HasCurrentAccount" );
				OnPropertyChanged( "HasCurrentRealAccount" );
				OnPropertyChanged( "HasCurrentAccountRequest" );
			}
		}

		/// <summary>
		/// This is true if an account speckle is selected. READ-ONLY.
		/// </summary>
		public bool HasCurrentAccount
		{
			get { return CurrentAccount != null; }
		}

		/// <summary>
		/// This is true if an account speckle is selected, and it refers a real account
		/// (<see cref="Vela.PeerDocuments.Account"/>), rather than an account creation request
		/// (<see cref="Vela.ActionProcessor.CreateAccountAnonymousActionRequest"/>). READ-ONLY.
		/// </summary>
		public bool HasCurrentRealAccount
		{
			get { return HasCurrentAccount && CurrentAccount.RefersRealAccount(); }
		}

		/// <summary>
		/// This is true if an account speckle is selected, and it refers an account creation request
		/// (<see cref="Vela.ActionProcessor.CreateAccountAnonymousActionRequest"/>), rather than a real account
		/// (<see cref="Vela.PeerDocuments.Account"/>). READ-ONLY.
		/// </summary>
		public bool HasCurrentAccountRequest
		{
			get { return HasCurrentAccount && CurrentAccount.RefersAccountRequest(); }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Queues an action request which pushes an updated presenter after an associated account is edited.
		/// </summary>
		/// <param name="accountSpeckle">The speckle of the edited account.</param>
		private void QueuePushUpdatePresenter( AccountOrRequestSpeckle accountSpeckle )
		{
			ValidationUtils.ObjectHasContent( accountSpeckle, "accountSpeckle" );

			StorehouseModel sm = ApplicationModel.StorehouseModel;

			Locator localUpnFingerprint = accountSpeckle.LocalUpnFingerprint;
			Locator remoteUpnFingerprint = accountSpeckle.RemoteUpnFingerprint;
			MacAlg macAlg = ApplicationModel.AlgorithmPool.GetMacAlg();
			IAsymmSigningPaddingAlg asymmSigningPaddingAlg = ApplicationModel.AlgorithmPool.GetAsymmSigningPadding();
			IHashSerializer hashSerializer = sm.StorehouseProvider.Serializer.GetHashSerializer( false );

			Locator updateRequestLocator;

			// Create a transaction.
			IStorehouseTransaction transaction = sm.StorehouseProvider.CreateTransaction();

			try
			{
				IDeserializer deserializer;
				SpeckleFilter vaultFilter = new SpeckleFilter( Vault.LocalUpnFingerprintFilterIdentifier, localUpnFingerprint );
				Vault vault = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Vault ) )
					, new SpeckleFilters( vaultFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared
					, out deserializer ) as Vault;

				updateRequestLocator = ApplicationBackendStorehouse.QueuePushUpdatePresenter( transaction, vault
					, remoteUpnFingerprint, macAlg, asymmSigningPaddingAlg, hashSerializer, sm.ActionClientManager );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			// Notify the action client manager that an action request was added to the storehouse.
			if( updateRequestLocator != null )
				sm.ActionClientManager.Add( updateRequestLocator, true );
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Collects all the accounts.
		/// </summary>
		public void CollectAccounts()
		{
			// Get the locator of the selected speckle.
			Locator currentAccountLocator = (CurrentAccount != null) ? CurrentAccount.StorehouseKey : null;

			Accounts = ApplicationBackendStorehouse.CollectAccountAndRequestSpeckles(
				ApplicationModel.StorehouseModel.StorehouseProvider );

			// Update the selected speckle.
			if( currentAccountLocator != null )
				CurrentAccount = Accounts.FindByStorehouseKey( currentAccountLocator );
		}

		/// <summary>
		/// Reloads all the accounts associated to the specified remote UPN.
		/// </summary>
		/// <param name="remoteUpnFingerprint">The UPN fingerprint of the remote peer for which to reload the associated
		/// accounts.</param>
		public void ReloadAccountsForUpn( Locator remoteUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( remoteUpnFingerprint, "remoteUpnFingerprint" );

			StorehouseModel sm = ApplicationModel.StorehouseModel;

			// Get the locator of the selected speckle.
			Locator currentAccountLocator = (CurrentAccount != null) ? CurrentAccount.StorehouseKey : null;

			AccountOrRequestSpeckles updatedSpeckles = ApplicationBackendStorehouse.CollectAccountAndRequestSpeckles(
				sm.StorehouseProvider, remoteUpnFingerprint );

			foreach( AccountOrRequestSpeckle updatedSpeckle in updatedSpeckles )
			{
				// Update the list of speckles. Use "Replace" because the key may have changed.
				AccountOrRequestSpeckle oldSpeckle = Accounts.FindByStorehouseKey( updatedSpeckle.StorehouseKey );
				Accounts.AddOrReplace( oldSpeckle, updatedSpeckle );
			}

			if( currentAccountLocator != null )
			{
				// Update the selected speckle.
				CurrentAccount = Accounts.FindByStorehouseKey( currentAccountLocator );

				// Update the account model.
				sm.UpdateAccountModelForAccountLocator( currentAccountLocator, false );
			}
		}

		/// <summary>
		/// Sets the account model with the current account, and shows the "Letters" view.
		/// </summary>
		public void SetAccountModel()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasCurrentRealAccount )
				return;

			sm.SetAccountModel( CurrentAccount.StorehouseKey, true );
		}

		/// <summary>
		/// Edits the current account.
		/// </summary>
		public void EditAccount()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasCurrentAccount )
				return;

			AccountOrRequestSpeckle accountToEdit = CurrentAccount;
			bool refersRealAccount = accountToEdit.RefersRealAccount();

			if( refersRealAccount )
			{
				bool changedApproval, changedAlternativePeerAddress;

				AccountOrRequestSpeckle speckle = ImBackendStorehouse.EditAccount( sm.StorehouseProvider
					, accountToEdit.StorehouseKey, sm.ActionClientManager.RecipientAccessStatesManager, out changedApproval
					, out changedAlternativePeerAddress );
				if( speckle == null )
					return;

				// Update the list of speckles. Use "Replace" because the key may have changed.
				Accounts.AddOrReplace( accountToEdit, speckle );

				// Update the selected speckle.
				CurrentAccount = speckle;

				// Update the lists of accounts and letters (because the account alias and the tally entity name may have
				// changed).
				CollectAccounts();
				sm.LettersModel.CollectLettersWithCurrentFlag( false );

				// Update the account model.
				sm.UpdateAccountModelForAccountLocator( speckle.StorehouseKey, false );

				// No attempt is made to restart the action service or peer registration because no relevant information can be
				// changed by editing an account.

				if( changedApproval )
					QueuePushUpdatePresenter( speckle );

				if( changedAlternativePeerAddress )
				{
					sm.ActionClientManager.RecipientAccessStatesManager.RetrySendingRequestsAsSoonAsPossible(
						speckle.RecipientUpnFingerprint );
				}
			}
			else
			{
				CreateAccountClientProgress accountRequestProgress;
				CreateAccountAnonymousActionRequest accountRequest = ApplicationBackendStorehouse.LoadAccountRequestAndProgress(
					sm.StorehouseProvider, accountToEdit.StorehouseKey, out accountRequestProgress );

				QueueCreateAccountWithUpnWindow accountRequestWindow = new QueueCreateAccountWithUpnWindow( accountRequest
					, accountRequestProgress );
				ControlsUtils.ShowDialog( accountRequestWindow );
			}
		}

		/// <summary>
		/// Deletes the current account.
		/// </summary>
		public void DeleteAccount()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasCurrentAccount )
				return;

			AccountOrRequestSpeckle accountToDelete = CurrentAccount;
			bool refersRealAccount = accountToDelete.RefersRealAccount();

			#region Ask the user

			string message;
			string confirmationTextPart;

			if( refersRealAccount )
			{
				string localName, remoteName;
				ApplicationBackendStorehouse.GetAccountLocalAndRemoteEntityNames( sm.StorehouseProvider
					, accountToDelete.StorehouseKey, out localName, out remoteName );

				confirmationTextPart = remoteName;

				message = string.Format( "Do you want to delete the account with the remote name \"{0}\" and local name"
					+ " \"{1}\"?", remoteName, localName );
			}
			else
			{
				string recipientTextualUpn = ApplicationBackendStorehouse.GetAccountRequestRecipientTextualUpn(
					sm.StorehouseProvider, accountToDelete.StorehouseKey );

				confirmationTextPart = recipientTextualUpn.Substring( 0, TextUtils.BufferMoteSize );

				message = string.Format( "Do you want to delete the account creation request with the remote UPN \"{0}\"?"
					, recipientTextualUpn );
			}

			// Ask the user.
			if( MessageWindow.ShowDialog( message, MessageWindowModes.Query ) != QueryMessageResults.Yes )
				return;

			#region To continue, the user must type a text

			string confirmationText = string.Format( ImBackendConsts.DeleteAccountConfirmationTemplate
				, confirmationTextPart.ToLowerInvariant() );
			string ensureMessage = string.Format( "Deleting the account will also delete all the associated letters."
				+ " Type \"{0}\" if you still want to delete it.", confirmationText );
			string userText = EditSingleLineNoteWindow.ShowDialog( null, ensureMessage, null, false
				, TextValidations.ShortNote, true, KernelConsts.ShortNoteSizeMax );

			if( (userText == null) || !TextUtils.Equals( userText, confirmationText ) )
				return;

			#endregion To continue, the user must type a text

			#endregion Ask the user

			// Find the next speckle.
			AccountOrRequestSpeckle nextSpeckle = Accounts.FindNext( accountToDelete, false );

			if( refersRealAccount )
			{
				DeleteAccountWindow deleteWindow = new DeleteAccountWindow( ApplicationModel, accountToDelete.StorehouseKey
					, sm.ActionServiceManager, sm.ActionClientManager );
				if( !deleteWindow.ShowProgressDialog() )
					return;
			}
			else
			{
				ApplicationBackendStorehouse.DeleteAccountRequest( sm.StorehouseProvider, accountToDelete.StorehouseKey
					, sm.ActionServiceManager, sm.ActionClientManager );
			}

			// Update the list of speckles.
			Accounts.RemoveItem( accountToDelete );

			// Update list of letters.
			if( sm.LettersModel != null )
				sm.LettersModel.CollectLettersWithCurrentFlag( true );

			// Update the account model.
			sm.UpdateAccountModelForAccountLocator( accountToDelete.StorehouseKey, false );

			// Update the selected speckle.
			CurrentAccount = nextSpeckle;

			TooltipWindow.ShowTooltip( "The account has been deleted." );
		}

		/// <summary>
		/// Queues an action request which pulls a remote presenter in order to update it on the local peer, for the current
		/// account.
		/// </summary>
		public void QueuePullUpdatePresenter()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;
			AlgorithmPool algorithmPool = ApplicationModel.AlgorithmPool;

			if( !HasCurrentRealAccount )
				return;

			Locator requestInstanceLocator = ApplicationBackendStorehouse.QueuePullUpdatePresenter( sm.StorehouseProvider
				, CurrentAccount.StorehouseKey, algorithmPool.GetMacAlg(), algorithmPool.GetAsymmSigningPadding()
				, algorithmPool.GetHashSerializer( false ) );
			if( requestInstanceLocator == null )
				return;

			// Update the UI with the account speckle.
			AccountOrRequestSpeckle accountSpeckle = ApplicationBackendStorehouse.LoadAccountOrRequestSpeckle(
				sm.StorehouseProvider, requestInstanceLocator );
			Accounts.AddOrUpdate( accountSpeckle );

			string message = string.Format( "The account between the remote peer \"{0}\" and the local peer \"{1}\" was queued"
				+ " to be updated.", accountSpeckle.RemoteEntityName, accountSpeckle.LocalEntityName );
			TooltipWindow.ShowTooltip( message );

			// Notify the action client manager that a request was added to the storehouse.
			sm.ActionClientManager.Add( requestInstanceLocator, true );
		}

		/// <summary>
		/// Verifies an attesting document with the identity of the remote peer of the current account.
		/// </summary>
		public void VerifyAttestingDocument()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasCurrentRealAccount )
				return;

			ImBackendStorehouse.VerifyAttestingDocumentWithAccount( sm.StorehouseProvider
				, CurrentAccount.StorehouseKey );
		}

		/// <summary>
		/// Creates a secure document, signs a file with the local vault of the current account, and encrypts it with the
		/// identity of the remote peer of the current account.
		/// </summary>
		public void CreateSecuringDocument()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasCurrentRealAccount )
				return;

			ImBackendStorehouse.CreateSecuringDocument( sm.StorehouseProvider, ApplicationModel.AlgorithmPool
				, CurrentAccount.StorehouseKey );
		}

		/// <summary>
		/// Unpacks a secure document, decrypts it with the local vault, and verifies its signature with the identity of the
		/// remote peer of the current account.
		/// </summary>
		public void UnpackSecuringDocument()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasCurrentRealAccount )
				return;

			ImBackendStorehouse.UnpackSecuringDocument( sm.StorehouseProvider, CurrentAccount.StorehouseKey );
		}

		/// <summary>
		/// Handles UI updates for an action client which is processing an action request of type
		/// <see cref="CreateAccountAnonymousActionResponse"/>.
		/// </summary>
		/// <param name="updateUiContext">Host context for updating the user interface of an action client application after an
		/// action request and response are processed.</param>
		public void UpdateClientUiOnCreateAccountWithUpn( ActionClientHostUpdateUiContext updateUiContext )
		{
			ValidationUtils.ObjectHasContent( updateUiContext, "updateUiContext" );

			Locator accountLocator = updateUiContext.ActionProgress.ReferredActionRequestLocator;

			if( updateUiContext.ActionProgress.ProcessingState == ActionProcessingStates.Error )
			{
				// An account can't be created, so remove the account speckle from the UI.

				Accounts.RemoveItem( Accounts.FindByStorehouseKey( accountLocator ) );
			}
			else
			{
				// An account was created / updated locally.

				StorehouseModel sm = ApplicationModel.StorehouseModel;

				if( (updateUiContext.ActionProgress.ProcessingState == ActionProcessingStates.Pending)
					|| (updateUiContext.ActionProgress.ProcessingState == ActionProcessingStates.Retry) )
				{
					// Only the progress has changed, so reload the account speckle which identifies an account creation request.
					AccountOrRequestSpeckle speckle = ApplicationBackendStorehouse.LoadAccountOrRequestSpeckle(
						sm.StorehouseProvider, accountLocator );
					Accounts.AddOrReplace( Accounts.FindByStorehouseKey( accountLocator ), speckle );
				}
				else
				{
					// Remove from the UI the speckle of the account creation request.
					Accounts.RemoveItem( Accounts.FindByStorehouseKey( accountLocator ) );

					// Some information displayed in the accounts list may have changed, so reload all the accounts (because
					// there may be several which refer the same remote UPN).
					ReloadAccountsForUpn( updateUiContext.ActionProgress.RecipientUpnFingerprint );
				}

				// Update the list of letters (because several of them may have new sender / recipient names).
				if( updateUiContext.ActionProgress.ProcessingState == ActionProcessingStates.Sent )
					sm.LettersModel.CollectLettersWithCurrentFlag( false );
			}
		}

		/// <summary>
		/// Handles UI updates for an action service which has received an action request of type
		/// <see cref="PushUpdatePresenterActionRequest"/>.
		/// </summary>
		/// <param name="updateUiContext">Host context for updating the user interface of an action service application after an
		/// action request and response are processed.</param>
		public void UpdateServiceUiOnPushUpdatePresenter( ActionServiceHostUpdateUiContext updateUiContext )
		{
			ValidationUtils.ObjectHasContent( updateUiContext, "updateUiContext" );

			StorehouseModel sm = ApplicationModel.StorehouseModel;
			if( sm == null )
				return;

			PushUpdatePresenterActionRequest request = (PushUpdatePresenterActionRequest)updateUiContext.ActionRequest;

			sm.UpdateAccountModelForRemoteUpnFingerprint( request.InitiatorUpnFingerprint, true );
		}

		#endregion Public methods
	}
}
