////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using System.Windows;
using System.Windows.Threading;
using Vela.ActionProcessor;
using Vela.ApplicationBackend;
using Vela.Controls;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Pnrp;
using Vela.StarkDocuments;
using Vela.Storehouse;

namespace Vela.ImBackend
{
	/// <summary>
	/// Application model.
	/// </summary>
	/// <remarks>
	/// <para>The application model is a supermodel (/ supervising model). All the other (normal) models contain a reference to
	/// the application model. It is meant to be created / set only once per application (else UI bindings to it may not be
	/// automatically updated), and should contain all the other models (at least those used by the main window).</para>
	/// 
	/// <para>Don't bind a view's UI to properties (from that view) whose values depend on this model. If you were to do so, the
	/// UI would not be properly updated. Instead, create properties in this model, notify the property observers, and bind to
	/// them.</para>
	/// </remarks>
	public class ApplicationModel : NotifyPropertyChanged, IModel
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ApplicationModel()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="dispatcher">The dispatcher used to synchronize cross-thread actions.</param>
		/// <param name="notificationManager">Application notification manager.</param>
		/// <param name="commandLineParameters">The command line parameters of the application. This may be null or empty.
		/// </param>
		/// <remarks>
		/// <para>The <see cref="StorehouseModel"/> property is set in here.</para>
		/// </remarks>
		public ApplicationModel( Dispatcher dispatcher, ApplicationNotificationManager notificationManager
			, string[] commandLineParameters )
		{
			Dispatcher = dispatcher;
			NotificationManager = notificationManager;
			CommandLineParameters = commandLineParameters;

			Configuration = ImConfiguration.Load();

			LogRecordsModel = new LogRecordsModel( this );
			ViewsModel = new FunctionViewsModel( this );
			ActionClientHost = new ActionClientHost( this );
			ActionServiceHost = new ActionServiceHost( this );

			LogManager.Added += LogManager_Added;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private Dispatcher _Dispatcher;
		/// <summary>
		/// The dispatcher used to synchronize cross-thread actions. READ-ONLY.
		/// </summary>
		public Dispatcher Dispatcher
		{
			get { return _Dispatcher; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Dispatcher" );

				_Dispatcher = value;
			}
		}

		private ApplicationNotificationManager _NotificationManager;
		/// <summary>
		/// Application notification manager. READ-ONLY.
		/// </summary>
		public ApplicationNotificationManager NotificationManager
		{
			get { return _NotificationManager; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "NotificationManager" );

				_NotificationManager = value;
			}
		}

		private string[] _CommandLineParameters;
		/// <summary>
		/// The command line parameters of the application. This may be null or empty. READ-ONLY.
		/// </summary>
		public string[] CommandLineParameters
		{
			get { return _CommandLineParameters; }
			protected set { _CommandLineParameters = value; }
		}

		private EmoticonManager _EmoticonManager;
		/// <summary>
		/// Emoticon manager. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public EmoticonManager EmoticonManager
		{
			get
			{
				if( _EmoticonManager != null )
					return _EmoticonManager;

				_EmoticonManager = new EmoticonManager( Properties.Resources.EmoticonManagerConfiguration
					, AlgorithmPool.GetCustomEmoticonsFolderName() );

				return _EmoticonManager;
			}
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "EmoticonManager" );

				_EmoticonManager = value;
			}
		}

		private ImConfiguration _Configuration;
		/// <summary>
		/// The application configuration. READ-ONLY.
		/// </summary>
		public ImConfiguration Configuration
		{
			get { return _Configuration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Configuration" );

				_Configuration = value;
			}
		}

		/// <summary>
		/// Algorithm pool. READ-ONLY.
		/// </summary>
		public AlgorithmPool AlgorithmPool
		{
			get { return AlgorithmPool.Instance; }
		}

		private LogRecordsModel _LogRecordsModel;
		/// <summary>
		/// The log records model. Notifies property observers. READ-ONLY.
		/// </summary>
		public LogRecordsModel LogRecordsModel
		{
			get { return _LogRecordsModel; }
			protected set
			{
				_LogRecordsModel = value;

				// Notify property observers.
				OnPropertyChanged( "LogRecordsModel" );
			}
		}

		private FunctionViewsModel _ViewsModel;
		/// <summary>
		/// The function views model. Notifies property observers. READ-ONLY.
		/// </summary>
		public FunctionViewsModel ViewsModel
		{
			get { return _ViewsModel; }
			protected set
			{
				_ViewsModel = value;

				// Notify property observers.
				OnPropertyChanged( "ViewsModel" );
			}
		}

		private IStorehouseAccess _StorehouseAccess;
		/// <summary>
		/// The storehouse access. Notifies property observers. READ-ONLY.
		/// </summary>
		public IStorehouseAccess StorehouseAccess
		{
			get { return _StorehouseAccess; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "StorehouseAccess" );

				_StorehouseAccess = value;

				// Notify property observers.
				OnPropertyChanged( "StorehouseAccess" );
			}
		}

		private StorehouseModel _StorehouseModel;
		/// <summary>
		/// The storehouse model. This may be null. Notifies property observers. READ-ONLY.
		/// </summary>
		public StorehouseModel StorehouseModel
		{
			get { return _StorehouseModel; }
			protected set
			{
				// Close the previous storehouse model.
				if( _StorehouseModel != null )
					_StorehouseModel.Close();

				_StorehouseModel = value;

				// Notify property observers.
				OnPropertyChanged( "StorehouseModel" );
				OnPropertyChanged( "StorehouseProvider" );
				OnPropertyChanged( "StorehouseExists" );
				OnPropertyChanged( "HasStorehouseModelVisibility" );
				OnPropertyChanged( "MissingStorehouseModelVisibility" );
			}
		}

		/// <summary>
		/// The storehouse provider. This may be null. Notifies property observers. READ-ONLY.
		/// </summary>
		public IStorehouseProvider StorehouseProvider
		{
			get { return (StorehouseModel != null) ? StorehouseModel.StorehouseProvider : null; }
		}

		/// <summary>
		/// Is true if a storehouse exists. Notifies property observers. READ-ONLY.
		/// </summary>
		public bool StorehouseExists
		{
			get { return (StorehouseModel != null) ? StorehouseModel.StorehouseExists : false; }
		}

		/// <summary>
		/// This is visible if <see cref="StorehouseModel"/> is not null. Notifies property observers. READ-ONLY.
		/// </summary>
		public Visibility HasStorehouseModelVisibility
		{
			get { return (StorehouseModel != null) ? Visibility.Visible : Visibility.Collapsed; }
		}

		/// <summary>
		/// This is visible if <see cref="StorehouseModel"/> is null. Notifies property observers. READ-ONLY.
		/// </summary>
		public Visibility MissingStorehouseModelVisibility
		{
			get { return (StorehouseModel == null) ? Visibility.Visible : Visibility.Collapsed; }
		}

		private ActionClientHost _ActionClientHost;
		/// <summary>
		/// Action client host. READ-ONLY.
		/// </summary>
		public ActionClientHost ActionClientHost
		{
			get { return _ActionClientHost; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionClientHost" );

				_ActionClientHost = value;
			}
		}

		private ActionServiceHost _ActionServiceHost;
		/// <summary>
		/// Action service host. READ-ONLY.
		/// </summary>
		public ActionServiceHost ActionServiceHost
		{
			get { return _ActionServiceHost; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionServiceHost" );

				_ActionServiceHost = value;
			}
		}

		private IAppViewWindowsManager _AppViewWindowsManager;
		/// <summary>
		/// Manager of application specific window views.
		/// </summary>
		/// <remarks>
		/// <para>This must be set after the application model is created because it needs the application model, and should be
		/// created after the application's main window is created.</para>
		/// </remarks>
		public IAppViewWindowsManager AppViewWindowsManager
		{
			get { return _AppViewWindowsManager; }
			set
			{
				ValidationUtils.ObjectHasContent( value, "AppViewWindowsManager" );

				_AppViewWindowsManager = value;
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		/// <summary>
		/// Event raised when a log record is added to the log manager.
		/// </summary>
		/// <param name="logRecord">Log record added to the log manager. This may be null.</param>
		private void LogManager_Added( LogRecord logRecord )
		{
			if( (LogRecordsModel == null) || (logRecord == null) )
				return;

			// Process on the UI thread. It's important to use "BeginInvoke" because this method may be called from a background
			// thread. If "Invoke" were used and if the UI thread were to call back into the background thread (for the state of
			// an operation), a deadlock would occur. The UI thread can call back into the background thread because another
			// operation may have a higher processing priority than this operation.
			Dispatcher.BeginInvoke( new Action( delegate
			{ AddLogRecordOnUiThread( logRecord ); } )
				, DispatcherPriority.Background );
		}

		#endregion Event handlers

		#region Private static methods

		/// <summary>
		/// Gets the message which will be shown to the user when the application is closing.
		/// </summary>
		/// <returns>Returns the message which will be shown to the user when the application is closing.</returns>
		private static string GetClosingMessage()
		{
			string messageTemplate = "Are you sure that you want to close the application?"
				+ "\n\nThe application has to run if you want to receive letters. The applications of your contacts who try to"
				+ " send you letters while this application is closed, will retry to send them for {0} days."
				+ "\n\nThe application has to run if you want it to retry sending the letters which were not yet sent."
				+ "\n\nAll the transient letters will be deleted, so any such letter which was not yet sent will be lost.";

			// The validity timespan is the default value since the local peer doesn't know for how long a remote peer will retry
			// to send action requests.
			int validityTimespan = PeerDocumentsConsts.ActionRequestValidityInDays;

			return string.Format( messageTemplate, validityTimespan );
		}

		#endregion Private static methods

		#region Private methods

		/// <summary>
		/// Handles a log record which was added to the log manager.
		/// </summary>
		/// <param name="logRecord">Log record added to the log manager. This may be null.</param>
		/// <remarks>
		/// <para>This is already running on the dispatcher of the application model.</para>
		/// </remarks>
		private void AddLogRecordOnUiThread( LogRecord logRecord )
		{
			if( LogRecordsModel == null )
				return;

			try
			{
				LogRecordsModel.AddLogRecord( logRecord );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				// This exception may be swallowed by the system, so log it here.
				LogManager.Add( ex );

				throw new BackgroundOperationException( ex );
			}
		}

		#endregion Private methods

		#region Public static methods

		/// <summary>
		/// Sets the application notification according to all dependencies.
		/// </summary>
		/// <param name="storehouseModel">Storehouse model. If this is null, nothing happens.</param>
		/// <param name="notificationManager">Application notification manager. If this is null, nothing happens.</param>
		public static void SetNotification( StorehouseModel storehouseModel, ApplicationNotificationManager notificationManager )
		{
			if( (storehouseModel == null) || (notificationManager == null) )
				return;

			bool isGlobalCloudAvailable = (storehouseModel.ActionServiceManager != null)
				&& storehouseModel.ActionServiceManager.PeerIsOnline
				&& storehouseModel.ApplicationModel.Configuration.ActionProcessor.UseUpnResolvers
				&& PnrpUtils.IsGlobalCloudAvailable();

			// "NewLetter" notification. This has priority.
			if( (storehouseModel.LettersModel != null) && storehouseModel.LettersModel.HasReceivedLetter )
			{
				string tooltip = ApplicationBackendUtils.GetApplicationNotificationTooltip(
					ApplicationNotifications.NewLetter, isGlobalCloudAvailable );
				notificationManager.Notify( ApplicationNotifications.NewLetter
					, storehouseModel.LettersModel.LastReceivedLetterTime, tooltip );

				return;
			}

			ActionServiceManagerStates serviceManagerState = (storehouseModel.ActionServiceManager != null)
				? storehouseModel.ActionServiceManager.AreAllRegisteredAndListening()
				: ActionServiceManagerStates.OneNotRegisteredOrListening;

			if( serviceManagerState == ActionServiceManagerStates.AllRegisteredAndListening )
			{
				// "AllStarted" notification.

				ApplicationNotifications tooltipNotification = storehouseModel.ActionServiceManager.UseUpnResolvers
					? ApplicationNotifications.AllStarted : ApplicationNotifications.AllServicesStarted;
				string tooltip = ApplicationBackendUtils.GetApplicationNotificationTooltip( tooltipNotification
					, isGlobalCloudAvailable );

				ApplicationNotifications notification = (storehouseModel.ActionServiceManager.UseUpnResolvers
						&& isGlobalCloudAvailable)
					? ApplicationNotifications.AllStarted : ApplicationNotifications.AllServicesStarted;

				notificationManager.Notify( notification, DateTime.UtcNow, tooltip );
			}
			else if( serviceManagerState == ActionServiceManagerStates.OneNotRegisteredOrListening )
			{
				// "OneNotStarted" notification.

				string tooltip = ApplicationBackendUtils.GetApplicationNotificationTooltip(
					ApplicationNotifications.OneNotStarted, isGlobalCloudAvailable );
				notificationManager.Notify( ApplicationNotifications.OneNotStarted, DateTime.UtcNow, tooltip );
			}
			else if( serviceManagerState == ActionServiceManagerStates.Unknown )
			{
				// Do nothing because the state can't be retrieved at the moment.
			}
			else
			{
				throw ValidationUtils.Unhandled( serviceManagerState, "serviceManagerState" );
			}
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Checks if the action processor is under maintenance.
		/// </summary>
		public bool CheckMaintenance()
		{
			return false;
		}

		/// <summary>
		/// Closes all the resources of the application model.
		/// </summary>
		public void Close()
		{
			if( StorehouseModel != null )
				StorehouseModel.Close();

			if( Configuration.MainWindowLayoutConfiguration != null )
				Configuration.MainWindowLayoutConfiguration.SaveUi();
			Configuration.Save();

			NotificationManager.Close();

			// Clean up the current view.
			ViewsModel.CurrentRegistration = null;

			LogManager.Added -= LogManager_Added;
		}

		/// <summary>
		/// Closes all the resources of the application model, or minimizes the application, depending on the close mode.
		/// </summary>
		/// <param name="applicationCloseMode">The current close mode of the application.</param>
		/// <param name="timePassedFromLastReceivedLetterTimeToClearIt">The time (in seconds) which must have passed since the
		/// time when the last letter was received (<see cref="LettersModel.LastReceivedLetterTime"/>), in order to clear the
		/// time (when the last letter was received).</param>
		/// <returns>Returns the new close mode of the application.</returns>
		public ApplicationCloseModes CloseOrMinimize( ApplicationCloseModes applicationCloseMode
			, double timePassedFromLastReceivedLetterTimeToClearIt )
		{
			if( applicationCloseMode == ApplicationCloseModes.Force )
			{
				// Don't perform cleanup operations.

				return ApplicationCloseModes.Force; // The application must now close without performing cleanup operations.
			}

			if( applicationCloseMode == ApplicationCloseModes.CleanUp )
			{
				Close();

				return ApplicationCloseModes.Force; // The application must now close without performing cleanup operations.
			}

			if( applicationCloseMode == ApplicationCloseModes.UserPermission )
			{
				if( MessageWindow.ShowDialog( GetClosingMessage(), MessageWindowModes.Query ) == QueryMessageResults.Yes )
				{
					Close();

					return ApplicationCloseModes.Force; // The application must now close without performing cleanup operations.
				}
				else
				{
					return ApplicationCloseModes.Minimize;
				}
			}

			if( (StorehouseModel != null) && (StorehouseModel.LettersModel != null) )
			{
				StorehouseModel.LettersModel.ClearLastReceivedLetterTime( timePassedFromLastReceivedLetterTimeToClearIt );
			}

			ControlsUtils.SetApplicationVisibility( ApplicationVisibilities.Hidden );

			return ApplicationCloseModes.Minimize;
		}

		/// <summary>
		/// Sets the application notification according to all dependencies.
		/// </summary>
		public void SetNotification()
		{
			SetNotification( StorehouseModel, NotificationManager );
		}

		/// <summary>
		/// Sets the application notification according to all dependencies.
		/// </summary>
		/// <param name="storehouseModel">Storehouse model. If this is null, nothing happens.</param>
		public void SetNotification( StorehouseModel storehouseModel )
		{
			SetNotification( storehouseModel, NotificationManager );
		}

		/// <summary>
		/// Sets the notification of the application on its start.
		/// </summary>
		public void NotifyOnStart()
		{
			if( NotificationManager == null )
				return;

			NotificationManager.Notify( ApplicationNotifications.OneNotStarted, DateTime.UtcNow
				, ApplicationBackendUtils.GetApplicationNotificationTooltip( ApplicationNotifications.OneNotStarted ) );
		}

		/// <summary>
		/// Ensures that the <see cref="ApplicationNotifications.NewLetter"/> notification is shown in the application's button
		/// in the system's taskbar.
		/// </summary>
		public void EnsureNewLetterNotificationInTaskbar()
		{
			if( (StorehouseModel == null) || (NotificationManager == null) || (StorehouseModel.LettersModel == null)
				|| !StorehouseModel.LettersModel.HasReceivedLetter )
			{
				return;
			}

			NotificationManager.EnsureNewLetterNotificationInTaskbar();
		}

		/// <summary>
		/// Creates a new storehouse.
		/// </summary>
		/// <returns>Returns true if the storehouse was created.</returns>
		public bool CreateStorehouse()
		{
			string currentStorehouseFileName = (StorehouseProvider != null) ? StorehouseProvider.FileName : null;

			IStorehouseProvider storehouseProvider = ImBackendStorehouse.CreateStorehouse( AlgorithmPool
				, Configuration.StorehouseAccess, Configuration.ActionServiceListenerTemplate
				, Path.GetDirectoryName( Configuration.LastOpenedStorehouseFileName ), currentStorehouseFileName );

			if( storehouseProvider != null )
			{
				StorehouseModel = null; // First close the previous storehouse.
				StorehouseModel = new StorehouseModel( this, storehouseProvider );

				Configuration.LastOpenedStorehouseFileName = storehouseProvider.FileName;
			}

			return storehouseProvider != null;
		}

		/// <summary>
		/// Opens an existing storehouse.
		/// </summary>
		/// <param name="browseStorehouseFile">If this is true then the dialog to open a storehouse file is shown.</param>
		/// <param name="storehousePassword">If this has content, the last opened storehouse is automatically opened with this
		/// password; no password dialog is displayed.</param>
		/// <returns>Returns true if the storehouse was opened.</returns>
		/// <remarks>
		/// <para>The storehouse's entropy state (<see cref="Vela.Crypto.EntropyPool"/>) is restored in the application's
		/// entropy pool.</para>
		/// </remarks>
		public bool OpenStorehouse( bool browseStorehouseFile, string storehousePassword )
		{
			IStorehouseProvider storehouseProvider = ImBackendStorehouse.OpenStorehouse( Configuration.StorehouseAccess
				, Configuration.LastOpenedStorehouseFileName, browseStorehouseFile, storehousePassword
				, AlgorithmPool.GetSerializer() );

			if( storehouseProvider != null )
			{
				StorehouseModel = null; // First close the previous storehouse.
				StorehouseModel = new StorehouseModel( this, storehouseProvider );

				Configuration.LastOpenedStorehouseFileName = storehouseProvider.FileName;
			}

			return storehouseProvider != null;
		}

		/// <summary>
		/// Updates the emoticon configuration with the custom emoticons.
		/// </summary>
		/// <param name="fileNames">The name of the files of the custom emoticons. If this contains folder names, the content of
		/// those folders is processed recursively. This may be null or empty.</param>
		public void UpdateWithCustomEmoticons( Strings fileNames )
		{
			if( EmoticonManager == null )
				return;

			EmoticonManager.Configuration.UpdateWithCustomEmoticons( fileNames );

			// Clear the image cache.
			ImagePreviewsCache.Clear();

			// Refresh the list of letters (to show the updated emoticon images).
			if( (StorehouseModel != null) && (StorehouseModel.LettersModel != null) )
				StorehouseModel.LettersModel.NotifyRefreshList();
		}

		/// <summary>
		/// Updates the emoticon configuration with the custom emoticons.
		/// </summary>
		/// <param name="fileNames">The name of the files of the custom emoticons. If this contains folder names, the content of
		/// those folders is processed recursively, maintaining the folder structure. This may be null or empty.</param>
		/// <remarks>
		/// <para>Only the files whose extensions are in the <see cref="FileExtensions.ImageFile"/> list are copied.</para>
		/// </remarks>
		public void UpdateWithCustomEmoticons( string[] fileNames )
		{
			if( EmoticonManager == null )
				return;

			FileUtils.CopyFolderContent( fileNames, AlgorithmPool.GetCustomEmoticonsFolderName(), FileExtensions.ImageFile );

			// Update the emoticons with the custom ones.
			Strings allFileNames = new Strings();
			FileUtils.CollectFileNamesFromFolder( AlgorithmPool.GetCustomEmoticonsFolderName(), allFileNames );
			UpdateWithCustomEmoticons( allFileNames );

			TooltipWindow.ShowTooltip();
		}

		#endregion Public methods
	}
}
