////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Controls;

namespace Vela.ImBackend
{
	/// <summary>
	/// Function views model.
	/// </summary>
	public class FunctionViewsModel : BaseModel
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected FunctionViewsModel()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		public FunctionViewsModel( ApplicationModel applicationModel )
			: base( applicationModel )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		private ViewRegistrations<FunctionViewRegistrationAttribute> _Registrations;
		/// <summary>
		/// The function views registrations. Notifies property observers. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>These views are shown when the user selects an item from the listbox from the left side of the window.</para>
		/// 
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public ViewRegistrations<FunctionViewRegistrationAttribute> Registrations
		{
			get
			{
				if( _Registrations != null ) return _Registrations;

				_Registrations = new ViewRegistrations<FunctionViewRegistrationAttribute>();
				_Registrations.RegisterAssemblies();

				// Notify property observers.
				OnPropertyChanged( "Registrations" );

				return _Registrations;
			}
		}

		private ViewRegistration _CurrentRegistration;
		/// <summary>
		/// Current function view registration. This may be null. Notifies property observers.
		/// </summary>
		public ViewRegistration CurrentRegistration
		{
			get { return _CurrentRegistration; }
			set
			{
				if( _CurrentRegistration == value ) return;

				_CurrentRegistration = value;

				// Notify property observers.
				OnPropertyChanged( "CurrentRegistration" );

				CurrentControl = null;
			}
		}

		private AppViewUserControl _CurrentControl;
		/// <summary>
		/// Current function view control. This may be null. Notifies property observers. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public AppViewUserControl CurrentControl
		{
			get
			{
				if( _CurrentControl != null ) return _CurrentControl;

				if( CurrentRegistration != null )
				{
					_CurrentControl = CurrentRegistration.GetViewUserControl( ApplicationModel ) as AppViewUserControl;

					// Notify property observers. Avoid stack overflows when notifying about null values.
					if( _CurrentControl != null ) OnPropertyChanged( "CurrentControl" );
				}

				return _CurrentControl;
			}
			protected set
			{
				if( _CurrentControl != value )
				{
					// The control is being unloaded from view.
					if( _CurrentControl != null ) _CurrentControl.ExitingUi();

					AppViewUserControl previousCurrentControl = _CurrentControl;
					_CurrentControl = value;

					// The control is being loaded into view.
					if( _CurrentControl != null ) _CurrentControl.EnteringUi();
				}

				// Notify property observers.
				OnPropertyChanged( "CurrentControl" );
			}
		}

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Navigates (= sets as current) to the function view of the specified type.
		/// </summary>
		/// <param name="viewType">View type to set as current.</param>
		public void Navigate( Type viewType )
		{
			ValidationUtils.ObjectHasContent( viewType, "viewType" );

			ViewRegistration registration = Registrations.FindByType( viewType );
			if( registration == null ) return;

			CurrentRegistration = registration;
		}

		#endregion Public methods
	}
}
