////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Microsoft.Win32;
using System;
using System.Collections.Specialized;
using System.IO;
using System.Windows;
using Vela.ActionProcessor;
using Vela.ApplicationBackend;
using Vela.ApplicationDocuments;
using Vela.Controls;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.StarkDocuments;
using Vela.Storehouse;

namespace Vela.ImBackend
{
	/// <summary>
	/// Letters model.
	/// </summary>
	/// <remarks>
	/// <para>The "draft letter" is the letter which is currently being written (in the "Letters" tab).</para>
	/// </remarks>
	public class LettersModel : BaseModel
	{
		#region Private constants

		// Tooltip shown there are no files to attach to the draft letter.
		private const string DraftNoFileNamesToAttachTooltip = "There are no files to attach";

		// Tooltip shown after file names are attached to the draft letter.
		private const string DraftAttachFileNamesTooltip = "The attached files must not be modified until they are transferred";

		#endregion Private constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected LettersModel()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		public LettersModel( ApplicationModel applicationModel )
			: base( applicationModel )
		{
			Letters.CollectionChanged += Letters_CollectionChanged;
			DraftAttachedFileNames.CollectionChanged += DraftAttachedFileNames_CollectionChanged;

			// The storehouse model may be empty at this point, so the letters with the current flag can't be loaded.
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// If this is true then letters were loaded for the first time.
		/// </summary>
		private bool LettersLoadedFirstTime;

		#endregion Private fields

		#region Public non-serializable properties

		private Strings _PredefinedPostFlags;
		/// <summary>
		/// The predefined post flags. Notifies property observers.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public Strings PredefinedPostFlags
		{
			get
			{
				if( _PredefinedPostFlags != null )
					return _PredefinedPostFlags;

				PredefinedPostFlags = new Strings( Enum.GetNames( typeof( PostFlags ) ) );

				return _PredefinedPostFlags;
			}
			set
			{
				ValidationUtils.ListHasContent( value, "PredefinedPostFlags" );

				_PredefinedPostFlags = value;

				// Setting "FlagFilter" here would reload the letters when this property is read for the first time.

				// Notify property observers.
				OnPropertyChanged( "PredefinedPostFlags" );
			}
		}

		private Strings _PostFlags;
		/// <summary>
		/// The available post flags. This may be empty, but not null. Notifies property observers.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public Strings PostFlags
		{
			get
			{
				if( _PostFlags != null )
					return _PostFlags;

				PostFlags = ApplicationModel.StorehouseModel.Configuration.PostFlags;

				return _PostFlags;
			}
			set
			{
				ValidationUtils.ObjectHasContent( value, "PostFlags" );

				_PostFlags = value;

				// Setting "FlagFilter" here would reload the letters when this property is read for the first time.

				// Notify property observers.
				OnPropertyChanged( "PostFlags" );
			}
		}

		// The private field must be initialized (because it may be used before it's set from outside).
		private string _FlagFilter = PeerDocuments.PostFlags.Main.ToString();
		/// <summary>
		/// The flag filter of the letters to show. Notifies property observers.
		/// </summary>
		public string FlagFilter
		{
			get { return _FlagFilter; }
			set
			{
				ValidationUtils.TextHasContent( value, "FlagFilter" );

				_FlagFilter = value;

				CollectLettersWithCurrentFlag( true );

				// Notify property observers.
				OnPropertyChanged( "FlagFilter" );
				OnPropertyChanged( "MayUpdateLettersFlagWithTrash" );
				OnPropertyChanged( "MayDeleteLetters" );
			}
		}

		private int _PageCount;
		/// <summary>
		/// The number of pages which may be shown. This must be at least 0. Notifies property observers. READ-ONLY.
		/// </summary>
		public int PageCount
		{
			get { return _PageCount; }
			protected set
			{
				ValidationUtils.Integer( value, 0, int.MaxValue, "PageCount" );

				_PageCount = value;

				// Notify property observers.
				OnPropertyChanged( "PageCount" );
				OnPropertyChanged( "ShownPageIndex" );
				OnPropertyChanged( "HasOlderPage" );
				OnPropertyChanged( "HasNewerPage" );
			}
		}

		private int _ShownPageIndex;
		/// <summary>
		/// The index of the page which is currently shown. This is a zero-based shift from the newest page toward the oldest.
		/// Notifies property observers. READ-ONLY.
		/// </summary>
		public int ShownPageIndex
		{
			get { return _ShownPageIndex; }
			protected set
			{
				_ShownPageIndex = GetConstrainedPageIndexToShow( value );

				// Notify property observers.
				OnPropertyChanged( "ShownPageIndex" );
				OnPropertyChanged( "HasOlderPage" );
				OnPropertyChanged( "HasNewerPage" );
			}
		}

		/// <summary>
		/// If this is true then a page older than the currently shown one may be shown. READ-ONLY.
		/// </summary>
		public bool HasOlderPage
		{
			get { return ShownPageIndex < PageCount - 1; }
		}

		/// <summary>
		/// If this is true then a page newer than the currently shown one may be shown. READ-ONLY.
		/// </summary>
		public bool HasNewerPage
		{
			get { return ShownPageIndex > 0; }
		}

		private LetterSpeckles _Letters = new LetterSpeckles();
		/// <summary>
		/// The speckles with the letters. The backing field is initialized by default. This may be empty, but not null.
		/// READ-ONLY.
		/// </summary>
		public LetterSpeckles Letters
		{
			get { return _Letters; }
		}

		/// <summary>
		/// This is true if there is at least 1 letter speckle in the list. READ-ONLY.
		/// </summary>
		public bool HasLetters
		{
			get { return (Letters != null) && (Letters.Count > 0); }
		}

		private LetterSpeckle _CurrentLetter;
		/// <summary>
		/// The current letter speckle. This may be null or empty. Notifies property observers.
		/// </summary>
		public LetterSpeckle CurrentLetter
		{
			get { return _CurrentLetter; }
			set
			{
				_CurrentLetter = value;

				// Notify property observers.
				OnPropertyChanged( "CurrentLetter" );
				OnPropertyChanged( "HasCurrentLetter" );
				OnPropertyChanged( "HasNextProcessingAttempt" );
				OnPropertyChanged( "HasAttachedMediaFiles" );
				OnPropertyChanged( "HasImagePreview" );
				OnPropertyChanged( "MayExploreAttachedFiles" );
				OnPropertyChanged( "MayPlayAttachedMediaFiles" );
				OnPropertyChanged( "MayViewAttachedFileNames" );
				OnPropertyChanged( "MayExportLetter" );
			}
		}

		/// <summary>
		/// This is true if a letter speckle is selected. READ-ONLY.
		/// </summary>
		public bool HasCurrentLetter
		{
			get { return CurrentLetter != null; }
		}

		/// <summary>
		/// If this is true then a new attempt to send the current letter will be made in the future. READ-ONLY.
		/// </summary>
		public bool HasNextProcessingAttempt
		{
			get { return HasCurrentLetter && CurrentLetter.HasNextProcessingAttempt; }
		}

		/// <summary>
		/// If this is true then some of the files attached to the current letter are media files. READ-ONLY.
		/// </summary>
		public bool HasAttachedMediaFiles
		{
			get { return HasCurrentLetter && CurrentLetter.HasAttachedMediaFiles; }
		}

		/// <summary>
		/// If this is true then the current letter has an image preview, that is, it has an attached image file. READ-ONLY.
		/// </summary>
		public bool HasImagePreview
		{
			get { return HasCurrentLetter && CurrentLetter.HasImagePreview; }
		}

		/// <summary>
		/// This is true if the files attached to the current letter may be explored. READ-ONLY.
		/// </summary>
		public bool MayExploreAttachedFiles
		{
			get { return HasCurrentLetter && CurrentLetter.MayExploreAttachedFiles; }
		}

		/// <summary>
		/// This is true if the files attached to the current letter may be played. READ-ONLY.
		/// </summary>
		public bool MayPlayAttachedMediaFiles
		{
			get { return HasCurrentLetter && CurrentLetter.MayPlayAttachedMediaFiles; }
		}

		/// <summary>
		/// This is true if the names of the files attached to the current letter may be seen. READ-ONLY.
		/// </summary>
		public bool MayViewAttachedFileNames
		{
			get { return HasCurrentLetter && CurrentLetter.HasAttachedFiles; }
		}

		/// <summary>
		/// This is true if the current letter may be exported. READ-ONLY.
		/// </summary>
		public bool MayExportLetter
		{
			get { return HasCurrentLetter && CurrentLetter.MayExportLetter; }
		}

		/// <summary>
		/// This is true if <see cref="FlagFilter"/> is not with <see cref="PeerDocuments.PostFlags.Trash"/>. READ-ONLY.
		/// </summary>
		public bool MayUpdateLettersFlagWithTrash
		{
			get { return HasLetters && !Post.EqualFlags( FlagFilter, PeerDocuments.PostFlags.Trash.ToString() ); }
		}

		/// <summary>
		/// This is true if <see cref="FlagFilter"/> is not with <see cref="PeerDocuments.PostFlags.Main"/>. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The letters with the <see cref="PeerDocuments.PostFlags.Main"/> flag may not be deleted because the letters
		/// received while the deletion is in progress would be affected (since their flags are set with it).</para>
		/// </remarks>
		public bool MayDeleteLetters
		{
			get { return HasLetters && !Post.EqualFlags( FlagFilter, PeerDocuments.PostFlags.Main.ToString() ); }
		}

		private DateTime _LastReceivedLetterTime = DateTime.MinValue;
		/// <summary>
		/// The time when the last letter was received. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is set with a value higher than <see cref="DateTime.MinValue"/>, the application notification is set in
		/// here.</para>
		/// </remarks>
		public DateTime LastReceivedLetterTime
		{
			get { return _LastReceivedLetterTime; }
			protected set
			{
				_LastReceivedLetterTime = value;

				// Set the application notification.
				if( HasReceivedLetter )
					ApplicationModel.SetNotification();
			}
		}

		/// <summary>
		/// This is true if a letter was received (after the last time when the <see cref="ClearLastReceivedLetterTime"/> method
		/// was called). READ-ONLY.
		/// </summary>
		public bool HasReceivedLetter
		{
			get { return LastReceivedLetterTime > DateTime.MinValue; }
		}

		private AttachedFileNames _DraftAttachedFileNames = new AttachedFileNames();
		/// <summary>
		/// The list of names of the files attached to the draft letter. The backing field is initialized by default. This may be
		/// empty, but not null. Notifies property observers. READ-ONLY.
		/// </summary>
		public AttachedFileNames DraftAttachedFileNames
		{
			get { return _DraftAttachedFileNames; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "DraftAttachedFileNames" );

				_DraftAttachedFileNames = value;

				// Notify property observers.
				OnPropertyChanged( "DraftAttachedFileNames" );
				OnPropertyChanged( "DraftHasAttachedFiles" );
				OnPropertyChanged( "DraftAttachedFilesVisibility" );
				OnPropertyChanged( "DraftAttachedFilesSize" );
			}
		}

		/// <summary>
		/// This is true if the draft letter has attached files. READ-ONLY.
		/// </summary>
		public bool DraftHasAttachedFiles
		{
			get { return ListUtils.HasContent( DraftAttachedFileNames ); }
		}

		/// <summary>
		/// The visibility of the attached files. Notifies property observers. READ-ONLY.
		/// </summary>
		public Visibility DraftAttachedFilesVisibility
		{
			get { return ListUtils.HasContent( DraftAttachedFileNames ) ? Visibility.Visible : Visibility.Collapsed; }
		}

		/// <summary>
		/// The number of the files attached (from <see cref="DraftAttachedFileNames"/>) to the draft letter. Notifies property
		/// observers. READ-ONLY.
		/// </summary>
		public int DraftAttachedFilesCount
		{
			get { return DraftAttachedFileNames.Count; }
		}

		private Int64 _DraftAttachedFilesSize;
		/// <summary>
		/// The size of the files attached (from <see cref="DraftAttachedFileNames"/>) to the draft letter. Notifies property
		/// observers. READ-ONLY.
		/// </summary>
		public Int64 DraftAttachedFilesSize
		{
			get { return _DraftAttachedFilesSize; }
			protected set
			{
				_DraftAttachedFilesSize = value;

				// Notify property observers.
				OnPropertyChanged( "DraftAttachedFilesSize" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void Letters_CollectionChanged( object sender, NotifyCollectionChangedEventArgs e )
		{
			// Notify property observers.
			OnPropertyChanged( "HasLetters" );
			OnPropertyChanged( "MayUpdateLettersFlagWithTrash" );
			OnPropertyChanged( "MayDeleteLetters" );
		}

		private void DraftAttachedFileNames_CollectionChanged( object sender, NotifyCollectionChangedEventArgs e )
		{
			// Notify property observers.
			OnPropertyChanged( "DraftHasAttachedFiles" );
			OnPropertyChanged( "DraftAttachedFilesVisibility" );
			OnPropertyChanged( "DraftAttachedFilesCount" );
			OnPropertyChanged( "DraftAttachedFilesSize" );
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Replaces the content of the current list of letters with the specified letters.
		/// </summary>
		/// <param name="letters">Letters with which to replace the content of the current list of letters. This may be null or
		/// empty.</param>
		/// <remarks>
		/// <para>Notifies property observers.</para>
		/// </remarks>
		private void ReplaceLetterListContent( LetterSpeckles letters )
		{
			try
			{
				try
				{
					Letters.FreezeCollectionChangedNotifications( true );

					Letters.Clear();

					if( ListUtils.HasContent( letters ) )
						Letters.Add( letters );
				}
				finally
				{
					Letters.FreezeCollectionChangedNotifications( false );
				}
			}
			finally
			{
				// Notify property observers.
				OnPropertyChanged( "Letters" );
				OnPropertyChanged( "HasLetters" );
				OnPropertyChanged( "HasCurrentLetter" );
				OnPropertyChanged( "HasNextProcessingAttempt" );
				OnPropertyChanged( "HasAttachedMediaFiles" );
				OnPropertyChanged( "HasImagePreview" );
				OnPropertyChanged( "MayExploreAttachedFiles" );
				OnPropertyChanged( "MayPlayAttachedMediaFiles" );
				OnPropertyChanged( "MayViewAttachedFileNames" );
				OnPropertyChanged( "MayExportLetter" );
				OnPropertyChanged( "MayUpdateLettersFlagWithTrash" );
				OnPropertyChanged( "MayDeleteLetters" );
			}
		}

		/// <summary>
		/// Gets the page index to show constrained to the valid bounds.
		/// </summary>
		/// <param name="pageIndex">Page index to show.</param>
		/// <returns>Returns the constrained page index.</returns>
		private int GetConstrainedPageIndexToShow( int pageIndex )
		{
			if( pageIndex < 0 )
				return 0;
			else if( pageIndex >= PageCount )
				return Math.Max( 0, PageCount - 1 );
			else
				return pageIndex;
		}

		/// <summary>
		/// Reloads the paging mechanism.
		/// </summary>
		/// <param name="resetShownPageIndex">If this is true then <see cref="ShownPageIndex"/> is reset.</param>
		private void ReloadPagingMechanism( bool resetShownPageIndex )
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;
			if( sm == null )
			{
				PageCount = 0;
				ShownPageIndex = 0;

				return;
			}

			int lettersCount = ApplicationBackendStorehouse.CountLetterSpeckles( sm.StorehouseProvider, FlagFilter );

			PageCount = MathUtils.RoundUp( lettersCount, ImBackendConsts.ShowLetterCount );

			if( resetShownPageIndex )
				ShownPageIndex = 0;
			else
				ShownPageIndex = GetConstrainedPageIndexToShow( ShownPageIndex );
		}

		/// <summary>
		/// Updates the current page and reloads the paging mechanism after a letter was added to the list.
		/// </summary>
		/// <remarks>
		/// <para>Call this method in order to avoid to fully reload the current page (after a letter was added to the list).
		/// </para>
		/// 
		/// <para>This method may reset the current letter.</para>
		/// </remarks>
		private void UpdatePageAfterAddingLetter()
		{
			if( ShownPageIndex <= 0 )
			{
				// The newest page is shown.

				// If there are fewer letters in the list than the maximum allowed, do nothing.
				if( Letters.Count <= ImBackendConsts.ShowLetterCount )
					return;

				Letters.RemoveAt( 0 );
				ReloadPagingMechanism( false );
			}
			else
			{
				CollectLettersWithCurrentFlag( true );
			}
		}

		/// <summary>
		/// Updates the current page and reloads the paging mechanism after a letter was deleted from the list.
		/// </summary>
		/// <remarks>
		/// <para>Call this method in order to avoid to fully reload the current page (after a letter was deleted from the list).
		/// </para>
		/// 
		/// <para>This method may reset the current letter.</para>
		/// </remarks>
		private void UpdatePageAfterDeletingLetter()
		{
			// If the oldest page is shown and there are still letters in it, do nothing.
			if( (ShownPageIndex >= PageCount - 1) && (Letters.Count > 0) )
				return;

			CollectLettersWithCurrentFlag( true );
		}

		/// <summary>
		/// Shows the page with the specified index.
		/// </summary>
		/// <param name="pageIndex">The index of the page to show.</param>
		private void ShowPage( int pageIndex )
		{
			ShownPageIndex = pageIndex;

			CollectLettersWithCurrentFlag( false );
		}

		/// <summary>
		/// Checks if a newly received letter may be in the list.
		/// </summary>
		public bool MayContainNewlyReceivedLetter()
		{
			return (ShownPageIndex == 0) && Post.EqualFlags( FlagFilter, PeerDocuments.PostFlags.Main.ToString() );
		}

		/// <summary>
		/// Updates the list of letters after a letter was sent, received or imported.
		/// </summary>
		/// <param name="letterSpeckle">The speckle of the letter which was sent, received or imported.</param>
		private void UpdateUiWithSentReceivedOrImportedLetter( LetterSpeckle letterSpeckle )
		{
			ValidationUtils.ObjectHasContent( letterSpeckle, "letterSpeckle" );

			// Check if the letter can be displayed in the list.
			if( !letterSpeckle.HasFlag( FlagFilter ) )
				return;

			// If the page with the newest letters is not displayed, don't reload the list because that's too slow for certain
			// situations. Reloading can be done with: "CollectLettersWithCurrentFlag( true )".
			if( ShownPageIndex > 0 )
				return;

			// If the letter is already in the list, don't do anything.
			if( Letters.FindByStorehouseKey( letterSpeckle.StorehouseKey ) != null )
				return;

			// This may wrongly add the letter speckle to the list if the letter was already in the storehouse (with the
			// same flag) and is old enough to be displayed on a previous page; the wrong list item would disappear when the
			// page or flag would be changed. The alternative would be to always reload the list, which is too slow for certain
			// situations.
			Letters.Add( letterSpeckle );

			UpdatePageAfterAddingLetter();
		}

		/// <summary>
		/// Clears the attached files.
		/// </summary>
		private void ClearAttachedFiles()
		{
			DraftAttachedFileNames.Clear();
			DraftAttachedFilesSize = 0;
		}

		/// <summary>
		/// Shows the notification associated to the emoticon command of the emoticon symbol contained by the specified letter
		/// associated.
		/// </summary>
		/// <param name="initiatorTally">The tally of the initiator of the letter.</param>
		/// <param name="initiatorAccount">The account of the initiator of the letter.</param>
		/// <param name="letter">Letter for which to show the notification.</param>
		private void ShowNotificationFromEmoticonCommand( Tally initiatorTally, Account initiatorAccount, Letter letter )
		{
			ValidationUtils.ObjectHasContent( letter, "letter" );

			EmoticonSymbolDescriptor emoticonSymbolDescriptor = ApplicationModel.EmoticonManager.GetEmoticonDescriptorFromSymbol(
				letter.EmoticonSymbol );
			if( emoticonSymbolDescriptor == null )
				return;

			switch( emoticonSymbolDescriptor.EmoticonCommand )
			{
				case EmoticonCommands.Bell:
					if( ApplicationModel.Configuration.AllowAudioNotification )
						AudioNotificationManager.Play( ApplicationModel.Configuration.BellAudioNotificationFileName, false );
					break;
				case EmoticonCommands.Horn:
					if( ApplicationModel.Configuration.AllowAudioNotification )
						AudioNotificationManager.Play( ApplicationModel.Configuration.HornAudioNotificationFileName, false );
					break;
				default:
					break;
			}

			switch( emoticonSymbolDescriptor.EmoticonCommand )
			{
				case EmoticonCommands.Bell:
				case EmoticonCommands.Horn:
				case EmoticonCommands.PersistentDesktopNotification:
					LetterDesktopNotificationWindow.ShowNotification( ApplicationModel, initiatorTally, initiatorAccount
						, letter );
					break;
				default:
					break;
			}
		}

		/// <summary>
		/// Checks if the specified letter belongs to the current account.
		/// </summary>
		/// <param name="letter">Letter to check if it belongs to the current account. This may be null.</param>
		/// <returns>Returns true if the specified letter belongs to the current account.</returns>
		private bool BelongsToCurrentAccount( Letter letter )
		{
			if( letter == null )
				return false;

			StorehouseModel sm = ApplicationModel.StorehouseModel;
			if( sm == null )
				return false;

			AccountModel am = sm.AccountModel;
			if( am == null )
				return false;

			return
				(
					am.LocalUpnFingerprint.Equals( letter.InitiatorUpnFingerprint )
					&& am.RemoteUpnFingerprint.Equals( letter.RecipientUpnFingerprint )
				)
				|| (
					am.LocalUpnFingerprint.Equals( letter.RecipientUpnFingerprint )
					&& am.RemoteUpnFingerprint.Equals( letter.InitiatorUpnFingerprint )
				);
		}

		/// <summary>
		/// Creates a plain-text playlist which contains the specified content.
		/// </summary>
		/// <param name="accessGranterLocator">The locator of the document which has offered the attached files.</param>
		/// <param name="playlistContent">Playlist content.</param>
		/// <returns>Returns the name of the created playlist file.</returns>
		private string CreatePlaylist( Locator accessGranterLocator, string playlistContent )
		{
			ValidationUtils.ObjectHasContent( accessGranterLocator, "accessGranterLocator" );
			ValidationUtils.TextHasContent( playlistContent, "playlistContent" );

			string playlistFileTitle = TextUtils.ArrayToText32( accessGranterLocator.Value );
			string playlistFileName = AssemblyUtils.GetApplicationTempFileName( ImBackendConsts.PlaylistFolderName
				, playlistFileTitle + ImBackendConsts.PlaylistFileExtension );

			File.WriteAllText( playlistFileName, playlistContent );

			return playlistFileName;
		}

		/// <summary>
		/// Imports the letter from the file with the specified name.
		/// </summary>
		/// <param name="fileName">The name of the file from which to import the letter.</param>
		private void ImportLetter( string fileName )
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			Locator letterLocator = ApplicationBackendStorehouse.ImportLetter( sm.StorehouseProvider, fileName
				, ApplicationModel.AlgorithmPool );

			LetterSpeckle letterSpeckle = ApplicationBackendStorehouse.LoadLetterSpeckle( sm.StorehouseProvider, letterLocator
				, ApplicationModel.EmoticonManager );

			UpdateUiWithSentReceivedOrImportedLetter( letterSpeckle );

			TooltipWindow.ShowTooltip( string.Format( "The letter from \"{0}\" has been imported. If you don't see it, it may"
				+ " have been previously received / imported.", letterSpeckle.InitiatorEntityName ) );
		}

		/// <summary>
		/// Validates that the current account is uniquely identified by the account alias from the specified post message,
		/// and gets the message without the account alias.
		/// </summary>
		/// <param name="message">Post message. This may be null or empty.</param>
		/// <param name="accountAlias">Is set with the account alias extracted from the specified message. This may be null.
		/// </param>
		/// <param name="aliasUniquelyIdentifiesAccount">Is set with true if the account alias extracted from the specified
		/// message uniquely identifies an account.</param>
		/// <param name="messageToSend">Is set with the message from which the unique account alias was removed. This may be
		/// null or empty.</param>
		/// <returns>Returns null if the validation was successful, or the reason why the validation has failed.</returns>
		private string ValidateAccountAliasAndGetMessageParts( string message, out AccountAlias accountAlias
			, out bool aliasUniquelyIdentifiesAccount, out string messageToSend )
		{
			accountAlias = null;
			aliasUniquelyIdentifiesAccount = false;
			messageToSend = null;

			if( TextUtils.NoContent( message ) )
				return null;

			bool mandatoryAccountAlias = ApplicationModel.Configuration.MandatoryAccountAliasWhenSendingLetter;
			AccountOrRequestSpeckles accountSpeckles = ApplicationModel.StorehouseModel.AccountsModel.Accounts;
			AccountContext accountContext = ApplicationModel.StorehouseModel.AccountModel.AccountContext;
			Account account = accountContext.Account;
			string localEntityName = accountContext.Vault.EntityName;
			string remoteEntityName = accountContext.Tally.EntityName;

			string errorMessageForActiveAccount = TextUtils.HasContent( account.AliasTextualForm )
				? string.Format( "If you want to send a letter from \"{0}\" to \"{1}\", start the message with \"{2}\""
						+ " (without quotes), then a whitespace, then the message."
					, localEntityName, remoteEntityName, account.AliasTextualForm )
				: "The active account has no alias.";

			string messageWithoutAlias;
			accountAlias = AccountAlias.FromPostMessage( message, out messageWithoutAlias );

			if( mandatoryAccountAlias && AccountAlias.NoContent( accountAlias ) )
			{
				return string.Format( "You have to start the draft message with an account alias.\n\n{0}"
					, errorMessageForActiveAccount );
			}

			Locator aliasedAccountLocator = accountSpeckles.FindStorehouseKeyOfAccountUniquelyIdentifiedByAlias( accountAlias );
			aliasUniquelyIdentifiesAccount = aliasedAccountLocator != null;

			// If the account alias is not mandatory and a unique account alias is not found, the entire draft message has to
			// be sent.
			messageToSend = (mandatoryAccountAlias || aliasUniquelyIdentifiesAccount) ? messageWithoutAlias : message;

			if( mandatoryAccountAlias && !aliasUniquelyIdentifiesAccount )
			{
				return string.Format( "You have to start the draft message with a unique account alias. \"{0}\" doesn't"
					+ " uniquely identify an account.\n\nGo to your accounts and modify the \"{0}\" alias to be different"
					+ " / unique for each account which currently has it.", accountAlias.TextualForm );
			}

			if( (aliasedAccountLocator != null) && !aliasedAccountLocator.Equals( account.InstanceLocator ) )
			{
				return string.Format( "The draft message doesn't start with the account alias of the active account. It starts"
						+ " with \"{0}\" instead of \"{1}\".\n\n{2}"
					, accountAlias.TextualForm, account.AliasTextualForm, errorMessageForActiveAccount );
			}

			if( TextUtils.NoContent( messageToSend ) )
			{
				if( aliasUniquelyIdentifiesAccount )
				{
					return "The draft message contains a unique account alias, but not a message to send.";
				}
				else
				{
					return "There is no message to send.";
				}
			}

			return null;
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Notifies the observers that the letters list UI has to be refreshed. Notifies property observers. Notifies list
		/// observers.
		/// </summary>
		/// <remarks>
		/// <para>The list of letters is not loaded if it's empty (since they don't have to be refreshed if they have not yet
		/// been loaded).</para>
		/// 
		/// <para>Call this, for example, when the letters list has to be refreshed because the colors have to be changed.</para>
		/// </remarks>
		public void NotifyRefreshList()
		{
			// Ensure that the letters are not loaded here since they don't have to be refreshed if they have not yet been
			// loaded.
			if( !HasLetters )
				return;

			Letters.NotifyRefreshList();
		}

		/// <summary>
		/// Clears the time when the last letter was received.
		/// </summary>
		/// <param name="timePassedFromLastReceivedLetterTime">The time (in seconds) which must have passed since the time when
		/// the last letter was received (<see cref="LastReceivedLetterTime"/>), in order to clear the time (when the last letter
		/// was received).</param>
		/// <returns>Returns true if the time when the last letter was received is cleared.</returns>
		/// <remarks>
		/// <para>The time is cleared only if the newest page with the letters with the
		/// <see cref="PeerDocuments.PostFlags.Main"/> flag is shown.</para>
		/// </remarks>
		public bool ClearLastReceivedLetterTime( double timePassedFromLastReceivedLetterTime )
		{
			if( ((DateTime.UtcNow - LastReceivedLetterTime).TotalSeconds < timePassedFromLastReceivedLetterTime)
				|| (ShownPageIndex > 0) || !Post.EqualFlags( FlagFilter, PeerDocuments.PostFlags.Main.ToString() ) )
			{
				return false;
			}

			LastReceivedLetterTime = DateTime.MinValue;

			return true;
		}

		/// <summary>
		/// Collects all the letters which have the current flag.
		/// </summary>
		/// <param name="resetPaging">If this is true then the paging mechanism (= the page count and the index of the shown
		/// page) is reset.</param>
		public void CollectLettersWithCurrentFlag( bool resetPaging )
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;
			if( sm == null )
			{
				ReplaceLetterListContent( null );

				return;
			}

			// Get the locator of the selected speckle.
			Locator currentLetterLocator = (CurrentLetter != null) ? CurrentLetter.StorehouseKey : null;

			if( resetPaging )
				ReloadPagingMechanism( false );

			CollectPaging paging = new CollectPaging( true, ShownPageIndex * ImBackendConsts.ShowLetterCount
				, ImBackendConsts.ShowLetterCount );

			LetterSpeckles letters = ApplicationBackendStorehouse.CollectLetterSpeckles( sm.StorehouseProvider, FlagFilter
				, ImBackendConsts.LoadTimeConsumingInformationWhenCollectingLetters, paging, ApplicationModel.EmoticonManager );

			ReplaceLetterListContent( letters );

			// Update the selected speckle.
			if( currentLetterLocator != null )
				CurrentLetter = Letters.FindByStorehouseKey( currentLetterLocator );
		}

		/// <summary>
		/// Loads the letters for the first time.
		/// </summary>
		public void LoadLettersFirstTime()
		{
			if( LettersLoadedFirstTime )
				return;

			FlagFilter = PeerDocuments.PostFlags.Main.ToString();
			LettersLoadedFirstTime = true;
		}

		/// <summary>
		/// Loads the letter speckles with the list indexes within the specified values with the information which takes a
		/// significant amount of time to load.
		/// </summary>
		/// <param name="firstItemIndex">Is set with the index of the first item whose information to load.</param>
		/// <param name="lastItemIndex">Is set with the index of the last item whose information to load</param>
		public void LoadTimeConsumingInformationOfLetters( int firstItemIndex, int lastItemIndex )
		{
			if( (firstItemIndex < 0) || (lastItemIndex < firstItemIndex) )
				return;

			StorehouseModel sm = ApplicationModel.StorehouseModel;
			if( sm == null )
				return;

			LetterSpeckles letterSpecklesToUpdate = new LetterSpeckles();

			for( int i = firstItemIndex; i <= lastItemIndex; i++ )
			{
				if( !Letters[ i ].TimeConsumingLoadingIsComplete )
					letterSpecklesToUpdate.Add( Letters[ i ] );
			}

			if( ListUtils.NoContent( letterSpecklesToUpdate ) )
				return;

			ApplicationBackendStorehouse.LoadLetterSpeckleProgressAndSetInformation( sm.StorehouseProvider
				, letterSpecklesToUpdate );
		}

		/// <summary>
		/// Sets the account model with the account associated with the current letter.
		/// </summary>
		/// <returns>Returns true if the account model is set.</returns>
		public bool SetAccountModel()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( (sm == null) || !HasCurrentLetter )
				return false;

			if( (CurrentLetter.LocalUpnFingerprint == null) || (CurrentLetter.RemoteUpnFingerprint == null) )
				throw new ValidationException( "The local or remote UPN fingerprint for the account model to set is missing." );

			return sm.SetAccountModel( CurrentLetter.LocalUpnFingerprint, CurrentLetter.RemoteUpnFingerprint, true );
		}

		/// <summary>
		/// Shows (in the file explorer) the attached files which have been saved (during the file download).
		/// </summary>
		public void ExploreAttachedFiles()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !MayExploreAttachedFiles )
				return;

			Letter letter = ApplicationBackendStorehouse.LoadLetter( sm.StorehouseProvider, CurrentLetter.StorehouseKey );
			ValidationUtils.ObjectHasContent( letter, "letter" );

			if( CurrentLetter.LocalPeerDirection == LocalPeerDirections.Initiator )
			{
				string fileName = letter.AttachedFiles.GetFirstItem().AbsoluteFileName;

				FileUtils.ExploreFile( fileName );
			}
			else if( CurrentLetter.LocalPeerDirection == LocalPeerDirections.Recipient )
			{
				string rootSaveFolderName = ApplicationModel.Configuration.ActionProcessor.RootSaveFolderName;
				bool hasFolder = AttachedFilesUtils.HasFolder( letter.AttachedFiles );
				string imageFileName = null;

				if( !hasFolder )
				{
					LocalAttachedFiles localAttachedFiles = AttachedFilesUtils.GetFilesForImagePreview( letter.AttachedFiles
						, rootSaveFolderName, 1, letter.InstanceLocator, letter.CreationTime, LocalPeerDirections.Recipient
						, false );

					imageFileName = LocalAttachedFiles.GetFirstFileName( localAttachedFiles );
				}

				if( TextUtils.HasContent( imageFileName ) )
				{
					FileUtils.ExploreFile( imageFileName );
				}
				else
				{
					string folderName = AttachedFilesUtils.GetSaveFolderName( rootSaveFolderName, letter.InstanceLocator
						, letter.CreationTime );

					FileUtils.ExploreFile( folderName );
				}
			}
			else
			{
				throw new ValidationException( string.Format( "Can't explore the files attached to a letter whose local peer"
					+ " direction is \"{0}\".", CurrentLetter.LocalPeerDirection ) );
			}
		}

		/// <summary>
		/// Plays the attached media files which have been saved (during the file download).
		/// </summary>
		public void PlayAttachedMediaFiles()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !MayPlayAttachedMediaFiles )
				return;

			Letter letter = ApplicationBackendStorehouse.LoadLetter( sm.StorehouseProvider, CurrentLetter.StorehouseKey );
			ValidationUtils.ObjectHasContent( letter, "letter" );

			string rootSaveFolderName = ApplicationModel.Configuration.ActionProcessor.RootSaveFolderName;

			Strings videoFileNames, audioFileNames, imageFileNames;
			AttachedFilesUtils.CollectMediaFiles( letter.AttachedFiles, rootSaveFolderName, letter.InstanceLocator
				, letter.CreationTime, CurrentLetter.LocalPeerDirection, out videoFileNames, out audioFileNames
				, out imageFileNames );

			string commandLine = null;
			string playlistContent = null;

			if( videoFileNames.Count > 0 )
			{
				commandLine = ApplicationModel.Configuration.PlayVideoFilesCommandLine;
				playlistContent = ApplicationBackendUtils.GetPlaylistContent( videoFileNames );
			}
			else if( audioFileNames.Count > 0 )
			{
				commandLine = ApplicationModel.Configuration.PlayAudioFilesCommandLine;
				playlistContent = ApplicationBackendUtils.GetPlaylistContent( audioFileNames );
			}
			else if( imageFileNames.Count > 0 )
			{
				commandLine = ApplicationModel.Configuration.PlayImageFilesCommandLine;
				playlistContent = ApplicationBackendUtils.GetPlaylistContent( imageFileNames );
			}

			if( TextUtils.NoContent( commandLine ) )
				throw new ValidationException( "There is no configured command line to play the attached media files." );

			if( !TextUtils.Contains( commandLine, ImBackendConsts.PlaylistFileNamePlaceholder ) )
			{
				throw new ValidationException( "The command line doesn't contain a placeholder which identifies the files to"
					+ " play.", null, null, new ExceptionDetail( "Command line", commandLine ) );
			}

			if( TextUtils.NoContent( playlistContent ) )
			{
				TooltipWindow.ShowTooltip( "There are no attached media files to play." );

				return;
			}

			string playlistFileName = CreatePlaylist( letter.InstanceLocator, playlistContent );

			commandLine = TextUtils.Replace( commandLine, ImBackendConsts.PlaylistFileNamePlaceholder, playlistFileName );
			FileUtils.RunSystemCommandWithParameters( commandLine );
		}

		/// <summary>
		/// Views the names of the files attached to the current letter.
		/// </summary>
		public void ViewAttachedFilesNames()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !MayViewAttachedFileNames )
				return;

			Letter letter = ApplicationBackendStorehouse.LoadLetter( sm.StorehouseProvider, CurrentLetter.StorehouseKey );
			ValidationUtils.ObjectHasContent( letter, "letter" );

			string rootSaveFolderName = ApplicationModel.Configuration.ActionProcessor.RootSaveFolderName;

			string plainTextList = AttachedFilesUtils.GetPlainTextList( letter.AttachedFiles, rootSaveFolderName
				, letter.InstanceLocator, letter.CreationTime, CurrentLetter.LocalPeerDirection, false );

			EditMultiLineNoteWindow.ShowDialog( "Attached File Names", null, plainTextList, true, TextValidations.None
				, false, int.MaxValue );
		}

		/// <summary>
		/// Changes the wallpaper using the files attached to the current letter.
		/// </summary>
		public void ChangeWallpaper()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasImagePreview || (CurrentLetter.LocalPeerDirection == LocalPeerDirections.RecipientImported) )
			{
				TooltipWindow.ShowTooltip( "No attached file can be used as wallpaper." );

				return;
			}

			// Load the attached files.
			Letter letter = ApplicationBackendStorehouse.LoadLetter( sm.StorehouseProvider, CurrentLetter.StorehouseKey );
			ValidationUtils.ObjectHasContent( letter, "letter" );
			AttachedFiles attachedFiles = letter.AttachedFiles;

			string rootSaveFolderName = ApplicationModel.Configuration.ActionProcessor.RootSaveFolderName;
			TextOnImageConfiguration textOnImageConfiguration = ApplicationModel.Configuration.TextOnImageConfiguration;

			string textToWriteOnImage = ApplicationWallpaperUtils.GetTextToWriteOnImage( CurrentLetter.MessageSummary
				, CurrentLetter.EmoticonSymbol, attachedFiles );

			bool wasChanged = ApplicationWallpaperUtils.Set( attachedFiles, rootSaveFolderName, CurrentLetter.StorehouseKey
				, CurrentLetter.CreationTime, CurrentLetter.LocalPeerDirection, textToWriteOnImage, textOnImageConfiguration );

			if( !wasChanged )
				TooltipWindow.ShowTooltip( "No attached file can be used as wallpaper." );
		}

		/// <summary>
		/// Updates the emoticons with the custom ones from the files attached to the current letter.
		/// </summary>
		public void UpdateWithCustomEmoticons()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasImagePreview || (CurrentLetter.LocalPeerDirection == LocalPeerDirections.RecipientImported) )
			{
				TooltipWindow.ShowTooltip( "No attached file can be used as custom emoticon." );

				return;
			}

			string message = "Do you want to update the emoticons with the ones contained by the image files attached to the"
				+ " letter?";
			if( MessageWindow.ShowDialog( message, MessageWindowModes.Query ) != QueryMessageResults.Yes )
				return;

			// Load the attached files.
			Letter letter = ApplicationBackendStorehouse.LoadLetter( sm.StorehouseProvider, CurrentLetter.StorehouseKey );
			ValidationUtils.ObjectHasContent( letter, "letter" );
			AttachedFiles attachedFiles = letter.AttachedFiles;

			string rootSaveFolderName = ApplicationModel.Configuration.ActionProcessor.RootSaveFolderName;

			// Copy the attached image files to the local folder with custom emoticons.
			bool hasCopied = AttachedFilesUtils.CopyImageFiles( attachedFiles, rootSaveFolderName, CurrentLetter.StorehouseKey
				, CurrentLetter.CreationTime, CurrentLetter.LocalPeerDirection, AlgorithmPool.GetCustomEmoticonsFolderName() );

			// Update the emoticons with the custom ones.
			Strings fileNames = new Strings();
			FileUtils.CollectFileNamesFromFolder( AlgorithmPool.GetCustomEmoticonsFolderName(), fileNames );
			ApplicationModel.ActionClientHost.UpdateWithCustomEmoticons( fileNames );

			if( hasCopied )
				TooltipWindow.ShowTooltip();
			else
				TooltipWindow.ShowTooltip( "No attached file can be used as custom emoticon." );
		}

		/// <summary>
		/// Retries to send the current letter as soon as possible.
		/// </summary>
		public void RetrySendLetter()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasNextProcessingAttempt )
				return;

			IRecipientAccessStatesManager recipientAccessStatesManager = sm.ActionClientManager.RecipientAccessStatesManager;
			recipientAccessStatesManager.RetrySendingRequestsAsSoonAsPossible( CurrentLetter.RecipientUpnFingerprint );

			TooltipWindow.ShowTooltip( "Retrying soon" );
		}

		/// <summary>
		/// Views the current letter in a separate window.
		/// </summary>
		public void ViewLetterInWindow()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasCurrentLetter )
				return;

			Letter letter = ApplicationBackendStorehouse.LoadLetter( sm.StorehouseProvider, CurrentLetter.StorehouseKey );
			ValidationUtils.ObjectHasContent( letter, "letter" );

			EditMultiLineNoteWindow window = new EditMultiLineNoteWindow( "Letter", null, letter.Message, true
				, TextValidations.LongNote, true, PeerDocumentsConsts.PostMessageSizeMax );
			ControlsUtils.ShowDialog( window );
		}

		/// <summary>
		/// Deletes the current letter.
		/// </summary>
		/// <returns>Returns true if the letter with the specified locator was deleted.</returns>
		public bool DeleteLetter()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasCurrentLetter )
				return false;

			bool deleted = false;

			LetterSpeckle letterToDelete = CurrentLetter;

			// Find the next letter.
			LetterSpeckle nextLetter = Letters.FindNext( letterToDelete, false );

			try
			{
				deleted = ApplicationBackendStorehouse.DeleteLetter( sm.StorehouseProvider, letterToDelete.StorehouseKey
					, sm.ActionServiceManager, sm.ActionClientManager );

				if( deleted )
				{

					// Update the list of letters.
					Letters.RemoveItem( letterToDelete );

					// Update the selected letter.
					CurrentLetter = nextLetter;

					ImagePreviewsCache.RemoveCachedFromGroup( letterToDelete.StorehouseKey, false );
				}
			}
			finally
			{
				if( deleted )
					UpdatePageAfterDeletingLetter();
			}

			return deleted;
		}

		/// <summary>
		/// Deletes all the letters with the current flag filter.
		/// </summary>
		/// <returns>Returns the number of deleted letters.</returns>
		public int DeleteLetters()
		{
			if( !MayDeleteLetters )
			{
				throw new ValidationException( string.Format( "The letters with the flag \"{0}\" can't be deleted because"
					+ " letters which are received after the deletion starts might also be deleted (and you would not get a"
					+ " chance to see them).", FlagFilter ) );
			}

			StorehouseModel sm = ApplicationModel.StorehouseModel;

			string flagToDelete = FlagFilter;

			#region Ask the user

			int lettersCount = ApplicationBackendStorehouse.CountLetterSpeckles( sm.StorehouseProvider, flagToDelete );
			if( lettersCount <= 0 )
				return 0;

			string messageTemplate = "Do you want to delete {0} letters with the \"{1}\" flag?"
				+ "\n\n"
				+ "If among these are letters which you have sent with attached files and they are currently being transferred,"
				+ " the transfers will not be interrupted (but the resume feature will not be supported).";
			string message = string.Format( messageTemplate, lettersCount, flagToDelete );
			if( MessageWindow.ShowDialog( message, MessageWindowModes.Query ) != QueryMessageResults.Yes )
				return 0;

			// Ensure that the user understands what happens (to delete, the user must type a text).

			string confirmationText = string.Format( ImBackendConsts.DeleteLettersConfirmationTemplate
				, flagToDelete.ToLowerInvariant() );
			string ensureMessage = string.Format( "Type \"{0}\" if you still want to delete the letters.", confirmationText );
			string userText = EditSingleLineNoteWindow.ShowDialog( null, ensureMessage, null, false, TextValidations.ShortNote
				, true, KernelConsts.ShortNoteSizeMax );

			if( (userText == null) || !TextUtils.Equals( userText, confirmationText ) )
				return 0;

			#endregion Ask the user

			int deletedCount = 0;
			try
			{
				DeleteLettersWindow deleteWindow = new DeleteLettersWindow( ApplicationModel, flagToDelete
					, ( letterLocator ) => ImagePreviewsCache.RemoveCachedFromGroup( letterLocator, false ) );

				if( !deleteWindow.ShowProgressDialog() )
					return 0;

				deletedCount = deleteWindow.DeletedCount;
			}
			finally
			{
				if( deletedCount > 0 )
				{
					CollectLettersWithCurrentFlag( true );

					TooltipWindow.ShowTooltip( string.Format( "{0} letters with the \"{1}\" flag have been deleted."
						, deletedCount, flagToDelete ) );
				}
			}

			return deletedCount;
		}

		/// <summary>
		/// Shows the oldest page.
		/// </summary>
		public void ShowOldestPage()
		{
			ShowPage( PageCount - 1 );
		}

		/// <summary>
		/// Shows the page older than the currently shown one.
		/// </summary>
		public void ShowOlderPage()
		{
			ShowPage( ShownPageIndex + 1 );
		}

		/// <summary>
		/// Shows the page newer than the currently shown one.
		/// </summary>
		public void ShowNewerPage()
		{
			ShowPage( ShownPageIndex - 1 );
		}

		/// <summary>
		/// Shows the newest page.
		/// </summary>
		public void ShowNewestPage()
		{
			ShowPage( 0 );
		}

		/// <summary>
		/// Updates the flag of the current letter with the specified flag.
		/// </summary>
		/// <param name="flag">New flag.</param>
		public void UpdateLetterFlag( string flag )
		{
			ValidationUtils.TextHasContent( flag, "flag" );

			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasCurrentLetter || Post.EqualFlags( flag, FlagFilter ) )
				return;

			LetterSpeckle letterToFlag = CurrentLetter;

			// Find the next letter.
			LetterSpeckle nextLetter = Letters.FindNext( letterToFlag, false );

			try
			{
				ApplicationBackendStorehouse.UpdatePostFlag( sm.StorehouseProvider, letterToFlag.StorehouseKey, flag );

				// Update the list of letters.
				Letters.RemoveItem( letterToFlag );

				// Update the selected letter.
				CurrentLetter = nextLetter;
			}
			finally
			{
				UpdatePageAfterDeletingLetter();
			}
		}

		/// <summary>
		/// Updates the flag of the current letter with the specified flag.
		/// </summary>
		/// <param name="flag">New flag.</param>
		public void UpdateLetterFlag( PostFlags flag )
		{
			UpdateLetterFlag( flag.ToString() );
		}

		/// <summary>
		/// Updates the flags of the letters with the current flag filter, with the specified flag.
		/// </summary>
		/// <param name="newFlag">New flag.</param>
		/// <param name="listOutOfSightTime">The time (in UTC format) when the list of letters last became out of sight. For
		/// example, this is the time when the context menu of the list of letters was last opened. This is used to ensure that
		/// the letters which arrive after the list became out of sight are not affected.</param>
		/// <returns>Returns the number of letters whose flags were updated.</returns>
		public int UpdateLettersFlag( string newFlag, DateTime listOutOfSightTime )
		{
			ValidationUtils.TextHasContent( newFlag, "newFlag" );

			if( !HasLetters || Post.EqualFlags( newFlag, FlagFilter ) )
				return 0;

			StorehouseModel sm = ApplicationModel.StorehouseModel;

			string flagToUpdate = FlagFilter;

			#region Ask the user

			int lettersCount = ApplicationBackendStorehouse.CountLetterSpeckles( sm.StorehouseProvider, flagToUpdate );
			if( lettersCount <= 0 )
				return 0;

			string message = string.Format( "Do you want to update the flags of {0} letters from the \"{1}\" flag to the"
				+ " \"{2}\" flag?", lettersCount, flagToUpdate, newFlag );
			if( MessageWindow.ShowDialog( message, MessageWindowModes.Query ) != QueryMessageResults.Yes )
				return 0;

			#endregion Ask the user

			if( (LastReceivedLetterTime > DateTime.MinValue) && (listOutOfSightTime <= LastReceivedLetterTime)
				&& Post.EqualFlags( FlagFilter, PeerDocuments.PostFlags.Main.ToString() ) )
			{
				throw new UserInteractionException( "A new letter which you may have not seen has been received. The flags can't"
					+ " be updated because the flag of the new letter would also be updated." );
			}

			int updatedCount = 0;
			try
			{
				UpdateLettersFlagWindow updateWindow = new UpdateLettersFlagWindow( ApplicationModel, flagToUpdate, newFlag );
				if( !updateWindow.ShowProgressDialog() )
					return 0;

				updatedCount = updateWindow.UpdatedCount;
			}
			finally
			{
				if( updatedCount > 0 )
				{
					CollectLettersWithCurrentFlag( true );

					TooltipWindow.ShowTooltip( string.Format( "The flags of {0} letters have been updated to \"{1}\"."
						, updatedCount, newFlag ) );
				}
			}

			if( (LastReceivedLetterTime > DateTime.MinValue) && (listOutOfSightTime <= LastReceivedLetterTime)
				&& Post.EqualFlags( FlagFilter, PeerDocuments.PostFlags.Main.ToString() ) )
			{
				string warning = string.Format( "A new letter which you may have not seen has been received. Its flag may have"
					+ " been updated, so make sure to change the flag filter to \"{0}\" in order to check if any new letter"
					+ " was moved there.", newFlag );
				MessageWindow.ShowDialog( message, MessageWindowModes.Warning );
			}

			return updatedCount;
		}

		/// <summary>
		/// Copies to the clipboard all the letters with the current flag filter.
		/// </summary>
		public void ClipAllLetters()
		{
			if( !HasLetters )
				return;

			StorehouseModel sm = ApplicationModel.StorehouseModel;

			#region Ask the user

			int lettersCount = ApplicationBackendStorehouse.CountLetterSpeckles( sm.StorehouseProvider, FlagFilter );
			if( lettersCount <= 0 )
				return;

			string message = string.Format( "Do you want to copy to the clipboard {0} letters with the \"{1}\" flag?"
				, lettersCount, FlagFilter );
			if( MessageWindow.ShowDialog( message, MessageWindowModes.Query ) != QueryMessageResults.Yes )
				return;

			#endregion Ask the user

			AccountOrRequestSpeckles accounts = sm.AccountsModel.Accounts;

			GetLettersAsTextWindow window = new GetLettersAsTextWindow( ApplicationModel, FlagFilter, accounts );
			if( !window.ShowProgressDialog() || (window.ProcessedCount <= 0) )
				return;

			Clipboard.SetText( window.LettersAsText );

			TooltipWindow.ShowTooltip( string.Format( "{0} letters have been copied to the clipboard."
				, window.ProcessedCount ) );
		}

		/// <summary>
		/// Copies to the clipboard the message form the secondary processing state of the focused letter.
		/// </summary>
		public void ClipCurrentLetterSecondaryProcessingStateMessage()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;
			if( (sm == null) || (sm.LettersModel == null) || !sm.LettersModel.HasCurrentLetter )
				return;

			SecondaryProcessingState state = sm.LettersModel.CurrentLetter.LastProcessingAttempt.SecondaryProcessingState;
			if( (state == null) || TextUtils.NoContent( state.Message ) )
				return;

			Clipboard.SetText( state.Message );

			TooltipWindow.ShowTooltip( "The secondary processing state of the focused letter has been copied to the"
				+ " clipboard." );
		}

		/// <summary>
		/// Exports the current letter.
		/// </summary>
		/// <param name="exportObfuscated">If this is true then the letter is exported in obfuscated form, like action requests
		/// are when they are sent to their recipient.</param>
		public void ExportLetter( bool exportObfuscated )
		{
			if( !HasCurrentLetter )
				throw new ValidationException( "There is no letter to export." );

			if( !CurrentLetter.MayExportLetter )
				throw new ValidationException( "Only letters to send may be exported." );

			if( CurrentLetter.HasAttachedFiles )
				throw new ValidationException( "Can't export the letter because it has attached files." );

			StorehouseModel sm = ApplicationModel.StorehouseModel;
			AlgorithmPool algorithmPool = ApplicationModel.AlgorithmPool;

			if( !ApplicationBackendStorehouse.ExportLetter( sm.StorehouseProvider, algorithmPool, CurrentLetter.StorehouseKey
				, exportObfuscated ) )
			{
				return;
			}
		}

		/// <summary>
		/// Imports letters.
		/// </summary>
		/// <returns>Returns the number of imported letters.</returns>
		public int ImportLetters()
		{
			#region Ask the user

			OpenFileDialog lettersOfd = new OpenFileDialog();
			lettersOfd.Title = "Import Letters";
			lettersOfd.Multiselect = true;
			lettersOfd.AddExtension = true;
			lettersOfd.DefaultExt = FileExtensions.AsymmEncryptedDocument;
			lettersOfd.Filter = FileExtensions.AsymmEncryptedAndPossiblyObfuscatedDocumentFilter;
			lettersOfd.CheckFileExists = true;
			lettersOfd.CheckPathExists = true;

			if( !ControlsUtils.ShowDialog( lettersOfd ) || (lettersOfd.FileNames.Length <= 0) )
				return 0;

			#endregion Ask the user

			if( lettersOfd.FileNames.Length == 1 )
			{
				ImportLetter( lettersOfd.FileNames[ 0 ] );

				return 1;
			}
			else
			{
				int importedCount = 0;

				try
				{
					ImportLettersWindow importWindow = new ImportLettersWindow( ApplicationModel, lettersOfd.FileNames );
					if( !importWindow.ShowProgressDialog() )
						return 0;

					importedCount = importWindow.ImportedCount;
				}
				finally
				{
					bool reloadLetters = Post.EqualFlags( FlagFilter, PeerDocuments.PostFlags.Main.ToString() )
						&& (importedCount > 0);

					if( reloadLetters )
						CollectLettersWithCurrentFlag( true );

					TooltipWindow.ShowTooltip( string.Format( "{0} letters have been imported. If you don't see them, they"
						+ " may have been previously received / imported.", importedCount ) );
				}

				return importedCount;
			}
		}

		/// <summary>
		/// Attaches file names to the draft letter.
		/// </summary>
		/// <returns>Returns the names of the files which were attached. This may be null or empty.</returns>
		public AttachedFileNames DraftAttachFileNames()
		{
			AttachedFileNames fileNamesToAttach = ApplicationBackendUtils.AttachFileNamesToPost( true, DraftAttachedFileNames
				, ref _DraftAttachedFilesSize );

			if( ListUtils.NoContent( fileNamesToAttach ) )
			{
				TooltipWindow.ShowTooltip( DraftNoFileNamesToAttachTooltip, TooltipWindowModes.Warning );

				return null;
			}

			// Notify property observers.
			OnPropertyChanged( "DraftAttachedFilesCount" );
			OnPropertyChanged( "DraftAttachedFilesSize" );

			string imageFileName = AttachedFilesUtils.GetFirstImageFileName( fileNamesToAttach );

			if( TextUtils.HasContent( imageFileName ) )
				TooltipWindow.ShowTooltip( imageFileName, TooltipWindowModes.Image );
			else
				TooltipWindow.ShowTooltip( DraftAttachFileNamesTooltip );

			return fileNamesToAttach;
		}

		/// <summary>
		/// Attaches file names to the draft letter.
		/// </summary>
		/// <param name="droppedFiles">The dropped files. This may be null or empty.</param>
		/// <returns>Returns the names of the files which were attached. This may be null or empty.</returns>
		public AttachedFileNames DraftAttachDroppedFileNames( string[] droppedFiles )
		{
			if( ArrayUtils.NoContent( droppedFiles ) )
			{
				TooltipWindow.ShowTooltip( DraftNoFileNamesToAttachTooltip, TooltipWindowModes.Warning );

				return null;
			}

			AttachedFileNames fileNamesToAttach = ApplicationBackendUtils.AttachFileNamesToPost( droppedFiles
				, DraftAttachedFileNames, ref _DraftAttachedFilesSize );

			if( ListUtils.NoContent( fileNamesToAttach ) )
			{
				TooltipWindow.ShowTooltip( DraftNoFileNamesToAttachTooltip, TooltipWindowModes.Warning );

				return null;
			}

			// Notify property observers.
			OnPropertyChanged( "DraftAttachedFilesCount" );
			OnPropertyChanged( "DraftAttachedFilesSize" );

			string imageFileName = AttachedFilesUtils.GetFirstImageFileName( fileNamesToAttach );

			if( TextUtils.HasContent( imageFileName ) )
				TooltipWindow.ShowTooltip( imageFileName, TooltipWindowModes.Image );
			else
				TooltipWindow.ShowTooltip( DraftAttachFileNamesTooltip );

			return fileNamesToAttach;
		}

		/// <summary>
		/// Attaches the image from the clipboard to the draft letter.
		/// </summary>
		/// <returns>Returns the names of the files which were attached. This may be null or empty.</returns>
		public AttachedFileNames DraftAttachClipboardImage()
		{
			AttachedFileNames fileNamesToAttach = ApplicationBackendUtils.AttachClipboardImageToPost( DraftAttachedFileNames
				, ref _DraftAttachedFilesSize );

			if( ListUtils.NoContent( fileNamesToAttach ) )
				return null;

			// Notify property observers.
			OnPropertyChanged( "DraftAttachedFilesCount" );
			OnPropertyChanged( "DraftAttachedFilesSize" );

			string imageFileName = AttachedFilesUtils.GetFirstImageFileName( fileNamesToAttach );

			if( TextUtils.HasContent( imageFileName ) )
				TooltipWindow.ShowTooltip( imageFileName, TooltipWindowModes.Image );
			else
				TooltipWindow.ShowTooltip( DraftAttachFileNamesTooltip );

			return fileNamesToAttach;
		}

		/// <summary>
		/// Attaches the (local peer's) custom emoticons to the draft letter.
		/// </summary>
		/// <returns>Returns the names of the files which were attached. This may be null or empty.</returns>
		public AttachedFileNames DraftAttachCustomEmoticons()
		{
			Strings files = FileUtils.CollectFileNamesFromFolderNoRecursing( AlgorithmPool.GetCustomEmoticonsFolderName() );

			return DraftAttachDroppedFileNames( files.ToArray() );
		}

		/// <summary>
		/// Removes all the files attached to the draft letter.
		/// </summary>
		public void DraftRemoveAttachedFiles()
		{
			ApplicationBackendUtils.RemoveAttachedFilesFromPost( DraftAttachedFileNames, ref _DraftAttachedFilesSize );

			// Notify property observers.
			OnPropertyChanged( "DraftAttachedFilesCount" );
			OnPropertyChanged( "DraftAttachedFilesSize" );
		}

		/// <summary>
		/// Views the names of the files attached to the draft letter.
		/// </summary>
		public void ViewDraftAttachedFilesNames()
		{
			if( !DraftHasAttachedFiles )
				return;

			string plainTextList = AttachedFilesUtils.GetPlainTextList( DraftAttachedFileNames );

			EditMultiLineNoteWindow.ShowDialog( "Attached File Names", null, plainTextList, true, TextValidations.None
				, false, int.MaxValue );
		}

		/// <summary>
		/// Queues an action request which contains a letter with the specified message for sending to the remote peer of the
		/// current account.
		/// </summary>
		/// <param name="message">Message to send. If this is null or empty, no letter is sent.</param>
		/// <param name="emoticonSymbol">The emoticon symbol of the letter to send.</param>
		/// <param name="isTransient">If this is true then the letter is transient.</param>
		/// <returns>Returns true if a letter with the specified message was queued for sending.</returns>
		public bool DraftQueueSendLetter( string message, string emoticonSymbol, bool isTransient )
		{
			if( TextUtils.NoContent( message ) )
				return false;

			ValidationUtils.Text( message, -1, PeerDocumentsConsts.PostMessageSizeMax, TextValidations.LongNote, "message" );

			StorehouseModel sm = ApplicationModel.StorehouseModel;
			AlgorithmPool algorithmPool = ApplicationModel.AlgorithmPool;

			if( !sm.HasAccountModel )
				return false;

			AccountAlias accountAlias;
			bool aliasUniquelyIdentifiesAccount;
			string messageToSend;
			string failedValidationMessage = ValidateAccountAliasAndGetMessageParts( message, out accountAlias
				, out aliasUniquelyIdentifiesAccount, out messageToSend );

			if( TextUtils.HasContent( failedValidationMessage ) )
			{
				TooltipWindow.ShowTooltip( failedValidationMessage, TooltipWindowModes.Warning );

				return false;
			}

			Locator letterLocator;
			Locator requestInstanceLocator = ApplicationBackendStorehouse.QueueSendLetter( sm.StorehouseProvider
				, sm.AccountModel.LocalUpnFingerprint, sm.AccountModel.RemoteUpnFingerprint
				, sm.AccountModel.Vault.ProcessingIdentity, sm.AccountModel.Tally.ProcessingIdentity
				, sm.AccountModel.Tally.ActionRequestValidity, messageToSend, emoticonSymbol, DraftAttachedFileNames
				, isTransient, algorithmPool.GetMacAlg(), algorithmPool.GetAsymmSigningPadding(), algorithmPool.GetSerializer()
				, out letterLocator );

			LetterSpeckle letterSpeckle = ApplicationBackendStorehouse.LoadLetterSpeckle( sm.StorehouseProvider, letterLocator
				, ApplicationModel.EmoticonManager );

			UpdateUiWithSentReceivedOrImportedLetter( letterSpeckle );

			ClearAttachedFiles();

			// Notify the action client manager that a request was added to the storehouse.
			sm.ActionClientManager.Add( requestInstanceLocator, true );

			TooltipWindow.CloseTooltip();

			return true;
		}

		/// <summary>
		/// Handles UI updates for an action client which is processing an action request of type
		/// <see cref="SendLetterActionRequest"/>.
		/// </summary>
		/// <param name="updateUiContext">Host context for updating the user interface of an action client application after an
		/// action request and response are processed.</param>
		/// <remarks>
		/// <para>The emoticon is not updated in the UI because the sent letter may be an older one which was not successfully
		/// sent when it was created, so another letter (with its own emoticon) may have been sent or received in the mean time.
		/// </para>
		/// </remarks>
		public void UpdateClientUiOnSendLetter( ActionClientHostUpdateUiContext updateUiContext )
		{
			ValidationUtils.ObjectHasContent( updateUiContext, "updateUiContext" );

			StorehouseModel sm = ApplicationModel.StorehouseModel;
			if( sm == null )
				return;

			#region Update the UI with the letter speckle

			Locator letterLocator = updateUiContext.ActionProgress.ReferredSheafLocator;
			LetterSpeckle letterSpeckle = ApplicationBackendStorehouse.LoadLetterSpeckle( sm.StorehouseProvider, letterLocator
				, ApplicationModel.EmoticonManager );

			// Update the UI with the progress of the letter speckle. This will show a progress (of 0%) in the UI.
			letterSpeckle.AttachedFilesProgress = double.Epsilon;

			// Update the letter (but don't add it, it must already have been added).
			LetterSpeckle existingSpeckle = Letters.FindByStorehouseKey( letterLocator );
			if( existingSpeckle != null )
				Letters.AddOrReplace( existingSpeckle, letterSpeckle );

			#endregion Update the UI with the letter speckle

			if( (updateUiContext.ActionResponse != null)
				&& TextUtils.HasContent( updateUiContext.ActionResponse.SecondaryProcessingOutput ) )
			{
				ApplicationModel.AppViewWindowsManager.ShowCommandLinesOutput( letterSpeckle.CreationTime
					, letterSpeckle.MessageSummary, updateUiContext.ActionResponse.SecondaryProcessingOutput );
			}
		}

		/// <summary>
		/// Handles UI updates for an action client which is processing an action request of type
		/// <see cref="DownloadFilesActionRequest"/>.
		/// </summary>
		/// <param name="updateUiContext">Host context for updating the user interface of an action client application after an
		/// action request and response are processed.</param>
		public void UpdateClientUiOnDownloadFiles( ActionClientHostUpdateUiContext updateUiContext )
		{
			ValidationUtils.ObjectHasContent( updateUiContext, "updateUiContext" );

			StorehouseModel sm = ApplicationModel.StorehouseModel;
			if( sm == null )
				return;

			// Get the existing letter speckle.
			Locator letterLocator = updateUiContext.ActionProgress.ReferredSheafLocator;
			LetterSpeckle letterSpeckle = Letters.FindByStorehouseKey( letterLocator );
			if( letterSpeckle == null )
				return; // The letter must have already been added, but was not found.

			// Update the UI with the progress of the letter speckle.
			DownloadFilesActionRequest actionRequest = updateUiContext.ActionRequest as DownloadFilesActionRequest;
			if( actionRequest != null )
			{
				if( updateUiContext.Exception == null )
				{
					// Checking the actual size of the downloaded file parts would have a severe impact on the performance (when
					// there are many attached files), so use a value cached in the update context.

					// Update the UI with the progress of the letter speckle.
					letterSpeckle.AttachedFilesProgress = AttachedFilesUtils.GetProgress( actionRequest.AttachedFiles );
				}
				else
				{
					letterSpeckle.AttachedFilesProgress = 0;
				}
			}

			// Remove the existing cached image preview in order to force a rebind. This has to be done because when the
			// initial bind was done, when the letter was received, no image file had been downloaded, so the cached image
			// preview is (at this point) the "No preview" image (and this has to be removed from the cache).
			LastFrameActionResponse lastFrameActionResponse = updateUiContext.ActionResponse as LastFrameActionResponse;
			if( (actionRequest != null) || (lastFrameActionResponse != null) )
			{
				if( letterSpeckle.HasImagePreview )
				{
					string rootSaveFolderName = ApplicationModel.Configuration.ActionProcessor.RootSaveFolderName;

					// Get the attached files for which image previews can actually be loaded.
					LocalAttachedFiles localAttachedFiles = AttachedFilesUtils.GetFilesForImagePreview(
						actionRequest.AttachedFiles, rootSaveFolderName
						, ApplicationModel.Configuration.ShowPreviewsForAttachedImageFilesCount, letterSpeckle.StorehouseKey
						, letterSpeckle.CreationTime, letterSpeckle.LocalPeerDirection, true );

					// Get the cached image previews (excluding the ones with no-preview image previews).
					string imageGroupIdentifier = ImagePreviewsCache.GetImageGroupIdentifier( letterSpeckle.StorehouseKey );
					GroupImagePreview groupImagePreview = ImagePreviewsCache.GetCachedGroupImage( imageGroupIdentifier, false
						, ImBackendConsts.ShowFullSizedGroupImagePreview );

					// Check if all the loadable image files have cached image previews. If not, force a UI rebind (for the
					// entire letter speckle).
					if( !ApplicationBackendUtils.GroupImagePreviewContainsAllFiles( groupImagePreview, localAttachedFiles ) )
					{
						ImagePreviewsCache.RemoveCachedFromGroup( imageGroupIdentifier, true );

						// Don't force a UI rebind if the entire letter speckle will be updated again (soon) after this.
						if( !updateUiContext.MustFullyUpdateUi )
							letterSpeckle.ForceRebind = true;
					}
				}
			}

			// Update the entire letter speckle (in particular, the processing state).
			if( updateUiContext.MustFullyUpdateUi )
			{
				// Get the new letter speckle.
				LetterSpeckle newSpeckle = ApplicationBackendStorehouse.LoadLetterSpeckle( sm.StorehouseProvider
					, letterLocator, ApplicationModel.EmoticonManager );

				// Update the UI with the progress of the letter speckle. This will force a rebind of the entire letter speckle.
				Letters.AddOrReplace( letterSpeckle, newSpeckle );

				// Prepare to update the progress.
				letterSpeckle = newSpeckle;
			}
		}

		/// <summary>
		/// Handles UI updates for an action service which has received an action request of type
		/// <see cref="SendLetterActionRequest"/>.
		/// </summary>
		/// <param name="updateUiContext">Host context for updating the user interface of an action service application after an
		/// action request and response are processed.</param>
		public void UpdateServiceUiOnSendLetter( ActionServiceHostUpdateUiContext updateUiContext )
		{
			ValidationUtils.ObjectHasContent( updateUiContext, "updateUiContext" );

			StorehouseModel sm = ApplicationModel.StorehouseModel;
			if( sm == null )
				return;

			Letter letter = ((SendLetterActionRequest)updateUiContext.ActionRequest).Letter;

			LetterSpeckle letterSpeckle = ApplicationBackendStorehouse.LoadLetterSpeckle( sm.StorehouseProvider
				, letter.InstanceLocator, ApplicationModel.EmoticonManager );

			UpdateUiWithSentReceivedOrImportedLetter( letterSpeckle );

			// Set the application notification.
			LastReceivedLetterTime = DateTime.UtcNow;

			ShowNotificationFromEmoticonCommand( updateUiContext.InitiatorTally, updateUiContext.InitiatorAccount, letter );
		}

		/// <summary>
		/// Handles UI updates for an action service which is processing an action request of type
		/// <see cref="DownloadFilesActionRequest"/>.
		/// </summary>
		/// <param name="updateUiContext">Host context for updating the user interface of an action service application after an
		/// action request and response are processed.</param>
		public void UpdateServiceUiOnDownloadFiles( ActionServiceHostUpdateUiContext updateUiContext )
		{
			ValidationUtils.ObjectHasContent( updateUiContext, "updateUiContext" );

			StorehouseModel sm = ApplicationModel.StorehouseModel;
			if( sm == null )
				return;

			// Get the action request.
			DownloadFilesActionRequest actionRequest = updateUiContext.ActionRequest as DownloadFilesActionRequest;
			if( actionRequest == null )
				return;

			// Get the existing letter speckle.
			Locator letterLocator = actionRequest.AccessGranterLocator;
			LetterSpeckle letterSpeckle = Letters.FindByStorehouseKey( letterLocator );
			if( letterSpeckle == null )
				return;

			// Update the entire letter speckle (in particular, the processing state).
			if( updateUiContext.MustFullyUpdateUi )
			{
				// Get the new letter speckle.
				LetterSpeckle newSpeckle = ApplicationBackendStorehouse.LoadLetterSpeckle( sm.StorehouseProvider
					, letterLocator, ApplicationModel.EmoticonManager );

				// Update the UI with the progress of the letter speckle. This will force a rebind of the entire letter speckle.
				Letters.AddOrReplace( letterSpeckle, newSpeckle );

				// Prepare to update the progress.
				letterSpeckle = newSpeckle;
			}

			// Update the UI with the progress of the letter speckle.
			letterSpeckle.AttachedFilesProgress = AttachedFilesUtils.GetProgress( actionRequest.AttachedFiles );
		}

		#endregion Public methods
	}
}
