////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Specialized;
using Vela.Controls;
using Vela.Kernel;

namespace Vela.ImBackend
{
	/// <summary>
	/// Log records model.
	/// </summary>
	public class LogRecordsModel : BaseModel
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected LogRecordsModel()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		public LogRecordsModel( ApplicationModel applicationModel )
			: base( applicationModel )
		{
			LogRecords.CollectionChanged += LogRecords_CollectionChanged;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private LogRecords _LogRecords = new LogRecords();
		/// <summary>
		/// The log records for the flow of the application. The backing field is initialized by default. This may be empty, but
		/// not null. Notifies property observers. READ-ONLY.
		/// </summary>
		public LogRecords LogRecords
		{
			get { return _LogRecords; }
			protected set
			{
				_LogRecords = value;

				// Notify property observers.
				OnPropertyChanged( "LogRecords" );

				CurrentLogRecord = null;
			}
		}

		/// <summary>
		/// This is true if there is at least 1 letter speckle in the list. READ-ONLY.
		/// </summary>
		public bool HasLogRecords
		{
			get { return (LogRecords != null) && (LogRecords.Count > 0); }
		}

		private LogRecord _CurrentLogRecord;
		/// <summary>
		/// The current vault speckle. This may be null or empty. Notifies property observers.
		/// </summary>
		public LogRecord CurrentLogRecord
		{
			get { return _CurrentLogRecord; }
			set
			{
				_CurrentLogRecord = value;

				// Notify property observers.
				OnPropertyChanged( "CurrentLogRecord" );
				OnPropertyChanged( "HasCurrentLogRecord" );
			}
		}

		/// <summary>
		/// This is true if a vault speckle is selected. READ-ONLY.
		/// </summary>
		public bool HasCurrentLogRecord
		{
			get { return CurrentLogRecord != null; }
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void LogRecords_CollectionChanged( object sender, NotifyCollectionChangedEventArgs e )
		{
			// Notify property observers.
			OnPropertyChanged( "HasLogRecords" );
		}

		#endregion Event handlers

		#region Public methods

		/// <summary>
		/// Adds a new log record.
		/// </summary>
		/// <param name="logRecord">Log record to add. This may be null.</param>
		public void AddLogRecord( LogRecord logRecord )
		{
			if( logRecord == null )
				return;

			LogRecords.Add( logRecord );
		}

		/// <summary>
		/// Views the current log record in a separate window.
		/// </summary>
		public void ViewLogRecordInWindow()
		{
			LogRecord logRecord = ApplicationModel.LogRecordsModel.CurrentLogRecord;
			if( logRecord == null )
				return;

			string message = string.Format( "{0}\n\n{1}", logRecord.GetMessageForLogList(), logRecord.GetLogMessage() );

			EditMultiLineNoteWindow window = new EditMultiLineNoteWindow( "Log Message", null, message, true
				, TextValidations.None, false, int.MaxValue );
			ControlsUtils.ShowDialog( window );
		}

		/// <summary>
		/// Clears the log records.
		/// </summary>
		public void ClearLogRecords()
		{
			LogRecords.Clear();

			// Profit from the opportunity and do some cleaning.
			GC.Collect();
		}

		#endregion Public methods
	}
}
