////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Reflection;
using System.Windows;
using Vela.ActionClient;
using Vela.ActionProcessor;
using Vela.ActionService;
using Vela.ApplicationBackend;
using Vela.Controls;
using Vela.ImDocuments;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Storehouse;

namespace Vela.ImBackend
{
	/// <summary>
	/// Storehouse model.
	/// </summary>
	/// <remarks>
	/// <para>The storehouse model is meant to be created / set multiple times per application.</para>
	/// 
	/// <para>Don't bind a view's UI to properties (from that view) whose values depend on this model. If you were to do so, the
	/// UI would not be properly updated. Instead, create properties in this model, notify the property observers, and bind to
	/// them.</para>
	/// </remarks>
	public class StorehouseModel : BaseModel
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected StorehouseModel()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">The application model.</param>
		/// <param name="storehouseProvider">Storehouse provider.</param>
		public StorehouseModel( ApplicationModel applicationModel, IStorehouseProvider storehouseProvider )
			: base( applicationModel )
		{
			StorehouseProvider = storehouseProvider;

			Open();
		}

		#endregion Constructors

		#region Public non-serializable properties

		private IStorehouseProvider _StorehouseProvider;
		/// <summary>
		/// The storehouse provider. Notifies property observers. READ-ONLY.
		/// </summary>
		public IStorehouseProvider StorehouseProvider
		{
			get { return _StorehouseProvider; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "StorehouseProvider" );

				_StorehouseProvider = value;

				// Notify property observers.
				OnPropertyChanged( "StorehouseProvider" );
				OnPropertyChanged( "StorehouseExists" );
				OnPropertyChanged( "AppTitle" );
			}
		}

		/// <summary>
		/// Is true if a storehouse exists. READ-ONLY.
		/// </summary>
		public bool StorehouseExists
		{
			get { return StorehouseProvider.StorehouseExists; }
		}

		private ImStorehouseConfiguration _Configuration;
		/// <summary>
		/// The storehouse configuration. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>Do not modify the content of this property.</para>
		/// </remarks>
		public ImStorehouseConfiguration Configuration
		{
			get { return _Configuration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Configuration" );

				_Configuration = value;
			}
		}

		/// <summary>
		/// The application title, based on the storehouse name. READ-ONLY.
		/// </summary>
		public string AppTitle
		{
			get
			{
				return string.Format( ImBackendConsts.AppTitleTemplateWithStorehouseName
					, ImBuildInfo.ProductDisplayNameDef, StorehouseProvider.ServerStorehouseName );
			}
		}

		private IActionServiceManager _ActionServiceManager;
		/// <summary>
		/// Action service manager. This may be null. Notifies property observers. READ-ONLY.
		/// </summary>
		public IActionServiceManager ActionServiceManager
		{
			get { return _ActionServiceManager; }
			protected set
			{
				_ActionServiceManager = value;

				// Notify property observers.
				OnPropertyChanged( "ActionServiceManager" );
			}
		}

		private IActionClientManager _ActionClientManager;
		/// <summary>
		/// Action client manager. This may be null. Notifies property observers. READ-ONLY.
		/// </summary>
		public IActionClientManager ActionClientManager
		{
			get { return _ActionClientManager; }
			protected set
			{
				_ActionClientManager = value;

				// Notify property observers.
				OnPropertyChanged( "ActionClientManager" );
			}
		}

		private VaultsModel _VaultsModel;
		/// <summary>
		/// The vaults model. This may be set with null. Notifies property observers.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public VaultsModel VaultsModel
		{
			get
			{
				if( _VaultsModel != null )
					return _VaultsModel;

				_VaultsModel = new VaultsModel( ApplicationModel );

				// Notify property observers. Avoid stack overflows when notifying about null values.
				if( _VaultsModel != null )
					OnPropertyChanged( "VaultsModel" );

				return _VaultsModel;
			}
			set
			{
				_VaultsModel = value;

				// Notify property observers.
				OnPropertyChanged( "VaultsModel" );
			}
		}

		private AccountsModel _AccountsModel;
		/// <summary>
		/// The accounts model. This may be set with null. Notifies property observers.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public AccountsModel AccountsModel
		{
			get
			{
				if( _AccountsModel != null )
					return _AccountsModel;

				_AccountsModel = new AccountsModel( ApplicationModel );

				// Notify property observers. Avoid stack overflows when notifying about null values.
				if( _AccountModel != null )
					OnPropertyChanged( "AccountsModel" );

				return _AccountsModel;
			}
			set
			{
				_AccountsModel = value;

				// Notify property observers.
				OnPropertyChanged( "AccountsModel" );
			}
		}

		private LettersModel _LettersModel;
		/// <summary>
		/// The letters model. This may be set with null. Notifies property observers.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public LettersModel LettersModel
		{
			get
			{
				if( _LettersModel != null )
					return _LettersModel;

				_LettersModel = new LettersModel( ApplicationModel );

				// Notify property observers. Avoid stack overflows when notifying about null values.
				if( _LettersModel != null )
					OnPropertyChanged( "LettersModel" );

				return _LettersModel;
			}
			set
			{
				_LettersModel = value;

				// Notify property observers.
				OnPropertyChanged( "LettersModel" );
			}
		}

		private AccountModel _AccountModel;
		/// <summary>
		/// The account model. This may be null. Notifies property observers.
		/// </summary>
		public AccountModel AccountModel
		{
			get { return _AccountModel; }
			set
			{
				_AccountModel = value;

				// Notify property observers.
				OnPropertyChanged( "AccountModel" );
				OnPropertyChanged( "HasAccountModel" );
				OnPropertyChanged( "HasAccountModelVisibility" );
				OnPropertyChanged( "MissingAccountModelVisibility" );

				// Refresh the letters list because the colors have to be changed.
				LettersModel.NotifyRefreshList();
			}
		}

		/// <summary>
		/// This is true if <see cref="AccountModel"/> is not null. READ-ONLY.
		/// </summary>
		public bool HasAccountModel
		{
			get { return AccountModel != null; }
		}

		/// <summary>
		/// This is visible if <see cref="AccountModel"/> is not null. READ-ONLY.
		/// </summary>
		public Visibility HasAccountModelVisibility
		{
			get { return (AccountModel != null) ? Visibility.Visible : Visibility.Collapsed; }
		}

		/// <summary>
		/// This is visible if <see cref="AccountModel"/> is null. READ-ONLY.
		/// </summary>
		public Visibility MissingAccountModelVisibility
		{
			get { return (AccountModel == null) ? Visibility.Visible : Visibility.Collapsed; }
		}

		private string _DraftLetterMessage;
		/// <summary>
		/// The message of the draft letter. This may be null or empty. Notifies property observers.
		/// </summary>
		/// <remarks>
		/// <para>This exists in order to save the message when the application is closed, and restore it when it's opened.
		/// </para>
		/// 
		/// <para>Because this is saved / loaded only (once) together with the storehouse, it's not part of the
		/// <see cref="LettersModel"/> (because this may be reloaded any number of times for the same the storehouse, to
		/// refresh it).</para>
		/// </remarks>
		public string DraftLetterMessage
		{
			get { return _DraftLetterMessage; }
			set
			{
				_DraftLetterMessage = value;

				// Notify property observers.
				OnPropertyChanged( "DraftLetterMessage" );
			}
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Saves the message of the draft letter.
		/// </summary>
		private void SaveDraftLetterMessage()
		{
			ImBackendStorehouse.SaveDraftLetterMessage( StorehouseProvider, DraftLetterMessage );
		}

		/// <summary>
		/// Loads the used configuration information, which includes the message of the draft letter.
		/// </summary>
		/// <param name="fullStart">If this is true then the storehouse resources are started fully, so all the configuration
		/// settings are loaded, else a restart is in progress, so not all the configuration settings are reloaded.</param>
		private void LoadConfiguration( bool fullStart )
		{
			Configuration = StorehouseUtils.LoadConfiguration<ImStorehouseConfiguration>( StorehouseProvider );

			if( !fullStart )
				return;

			DraftLetterMessage = Configuration.DraftLetterMessage;
		}

		/// <summary>
		/// Starts the action service.
		/// </summary>
		/// <param name="restartClient">If this is true then the action client is also restarted.</param>
		/// <remarks>
		/// <para>The action client is always started before the action service because the client is passed as parameter to the
		/// service.</para>
		/// </remarks>
		private void StartActionService( bool restartClient )
		{
			StopActionService( restartClient );

			try
			{
				if( restartClient )
				{
					Strings clientTypeNames = new Strings();
					clientTypeNames.Add( typeof( ImClient ).FullName );

					if( ApplicationModel.Configuration.ActionProcessor.PeerIsOnline )
					{
						ActionClientManager = new ActionClientManager( StorehouseProvider, clientTypeNames
							, ApplicationModel.AlgorithmPool, ApplicationModel.AlgorithmPool
							, ApplicationModel.Configuration.ActionProcessor, ApplicationModel.ActionClientHost );
					}
					else
					{
						LogManager.Add( "The action client manager is offline." );

						ActionClientManager = new DummyActionClientManager();
					}
				}

				Strings serviceTypeNames = new Strings();
				serviceTypeNames.Add( typeof( ImService ).FullName );

				if( ApplicationModel.Configuration.ActionProcessor.PeerIsOnline )
				{
					ActionServiceManager = new ActionServiceManager( StorehouseProvider, serviceTypeNames
						, ApplicationModel.Configuration.ActionProcessor, ApplicationModel.ActionServiceHost
						, ActionClientManager );
				}
				else
				{
					LogManager.Add( "The action service manager is offline." );

					ActionServiceManager = new DummyActionServiceManager();
				}
			}
			catch( Exception ex )
			{
				StopActionService( restartClient );

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Stops the action service.
		/// </summary>
		/// <param name="stopClient">If this is true then the action client is also stopped.</param>
		private void StopActionService( bool stopClient )
		{
			if( stopClient && (ActionClientManager != null) )
			{
				ActionClientManager.Stop();
				ActionClientManager = null;
			}

			if( ActionServiceManager != null )
			{
				ActionServiceManager.Stop();
				ActionServiceManager = null;
			}
		}

		/// <summary>
		/// Opens all the resources of the storehouse model.
		/// </summary>
		private void Open()
		{
			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				try
				{
					LoadConfiguration( true );

					StartActionService( true );
				}
				finally
				{
					ApplicationModel.SetNotification( this );
				}
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				Close();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}

			string message = string.Format( "The storehouse \"{0}\" was opened.", StorehouseProvider.FileName );
			LogManager.Add( message );
		}

		/// <summary>
		/// Deletes the transient letters from the storehouse.
		/// </summary>
		private void DeleteTransientLetters()
		{
			if( ApplicationBackendStorehouse.CountTransientLetterSpeckles( StorehouseProvider ) <= 0 )
				return;

			DeleteTransientLettersWindow deleteWindow = new DeleteTransientLettersWindow( ApplicationModel );
			deleteWindow.ShowProgressDialog();
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Closes all the resources of the storehouse model.
		/// </summary>
		public void Close()
		{
			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				try
				{
					StopActionService( true );

					SaveDraftLetterMessage();
				}
				finally
				{
					// Save the entropy.
					StorehouseUtils.UpdateEntropy( StorehouseProvider );

					ApplicationModel.SetNotification( this );
				}

				DeleteTransientLetters();
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );

				string message = string.Format( "The storehouse \"{0}\" was closed.", StorehouseProvider.FileName );
				LogManager.Add( message );
			}
		}

		/// <summary>
		/// Restarts the action service (= the peer registration manager and the service listeners) and the client.
		/// </summary>
		/// <param name="restartClient">If this is true then the action client is restarted.</param>
		/// <remarks>
		/// <para>Call this method, for example, when a vault was added to or deleted from the storehouse. Do not attempt to
		/// optimize by using a shortcut (= restart just for that vault) because it's possible that a vault is imported, then
		/// removed, then imported again, so your shortcut would fail.</para>
		/// </remarks>
		public void Restart( bool restartClient )
		{
			try
			{
				StopActionService( restartClient );

				LoadConfiguration( false );

				StartActionService( restartClient );
			}
			finally
			{
				ApplicationModel.SetNotification( this );
			}
		}

		/// <summary>
		/// Changes the password of the storehouse.
		/// </summary>
		public void ChangePassword()
		{
			// Ensure that the user knows the current storehouse password.
			PasswordWindow passwordWindow = new PasswordWindow( "Change Storehouse Password", null
				, StorehouseProvider.HeaderUserPassword, true );
			if( !ControlsUtils.ShowDialog( passwordWindow ) )
				return;

			try
			{
				// Because the storehouse provider contains the header password and is cloned by the action service and
				// client, restart the service and the client.

				StopActionService( true );

				StorehouseProvider.ChangeHeaderUserPassword( passwordWindow.Password );

				// There is no need to reload the configuration.

				StartActionService( true );
			}
			finally
			{
				ApplicationModel.SetNotification( this );
			}

			TooltipWindow.ShowTooltip( "The password was changed." );
		}

		/// <summary>
		/// Adds entropy to the storehouse.
		/// </summary>
		/// <returns>Returns true if entropy was added to the storehouse.</returns>
		public bool AddEntropy()
		{
			// Add new (global) entropy (from the user).
			EntropyWindow entropyWindow = new EntropyWindow();
			if( !ControlsUtils.ShowDialog( entropyWindow ) )
				return false;

			StorehouseUtils.UpdateEntropy( StorehouseProvider );

			TooltipWindow.ShowTooltip( "Entropy was added to the storehouse." );

			return true;
		}

		/// <summary>
		/// Verifies the (signature of the) attesting document of a file with the sheafs derived from
		/// <see cref="Vela.CoreDocuments.ILocallyNamedPublicProcessingIdentitySheaf"/>.
		/// </summary>
		public void VerifyAttestingDocument()
		{
			ImBackendStorehouse.VerifyAttestingDocument( StorehouseProvider, false, null, null, true, null );
		}

		/// <summary>
		/// Unpacks a securing document with the sheafs derived from
		/// <see cref="Vela.CoreDocuments.ILocallyNamedPublicProcessingIdentitySheaf"/>, by decrypting it and verifying its
		/// signature.
		/// </summary>
		public void UnpackSecuringDocument()
		{
			ImBackendStorehouse.UnpackSecuringDocument( StorehouseProvider, true, null, null );
		}

		/// <summary>
		/// Copies to the clipboard the password used for the file attachments of the storehouse.
		/// </summary>
		public void ClipStorehouseAttachmentsPassword()
		{
			// Ensure that the user wants to clip the password.
			string message = "Do you want to copy to the clipboard the password used for the file attachments of the storehouse?"
				+ " You can use the password in a storehouse manager application to manually manage the storehouse.";
			if( MessageWindow.ShowDialog( message, MessageWindowModes.Query ) != QueryMessageResults.Yes )
				return;

			// Ensure that the user knows the current storehouse password.
			PasswordWindow passwordWindow = new PasswordWindow( "Clip Storehouse Attachments Password", null
				, StorehouseProvider.HeaderUserPassword, false );
			if( !ControlsUtils.ShowDialog( passwordWindow ) )
				return;

			Clipboard.SetText( StorehouseProvider.GetStorehouseAttachmentsPassword() );

			TooltipWindow.ShowTooltip();
		}

		/// <summary>
		/// Sets the account model based on the account with the specified instance locator. If the account doesn't exist and
		/// "throwIfMissing" is true, an exception is thrown; if "throwIfMissing" is false, the account model is cleared.
		/// </summary>
		/// <param name="accountLocator">The instance locator of the account based on which to set the account model.</param>
		/// <param name="throwIfMissing">If this is true and the account with the specified locator doesn't exist, an exception
		/// is thrown.</param>
		public void SetAccountModel( Locator accountLocator, bool throwIfMissing )
		{
			AccountModel = AccountModel.Load( ApplicationModel, accountLocator, throwIfMissing );
		}

		/// <summary>
		/// Sets the account model based on the account with the specified alias, but only if the alias uniquely identifies an
		/// account.
		/// </summary>
		/// <param name="accountAlias">The account alias based on which to set the account model. This may be null.</param>
		public void SetAccountModel( AccountAlias accountAlias )
		{
			if( AccountAlias.NoContent( accountAlias ) )
				return;

			Locator accountLocator = AccountsModel.Accounts.FindStorehouseKeyOfAccountUniquelyIdentifiedByAlias( accountAlias );

			// To change the account model, the account must be uniquely identified by the specified alias.
			if( accountLocator != null )
			{
				if( !HasAccountModel || !accountLocator.Equals( AccountModel.Account.InstanceLocator ) )
					SetAccountModel( accountLocator, true );
			}
		}

		/// <summary>
		/// Sets the account model based on the account with the specified initiator and recipient UPN fingerprints. If the
		/// account doesn't exist and "throwIfMissing" is true, an exception is thrown; if "throwIfMissing" is false, the account
		/// model is cleared.
		/// </summary>
		/// <param name="accountLocalUpnFingerprint">The local UPN fingerprint of the account.</param>
		/// <param name="accountRemoteUpnFingerprint">The remote UPN fingerprint of the account.</param>
		/// <param name="throwIfMissing">If this is true and the account with the specified local and remote UPN fingerprints
		/// doesn't exist, an exception is thrown.</param>
		/// <returns>Returns true if the account model is set.</returns>
		public bool SetAccountModel( Locator accountLocalUpnFingerprint, Locator accountRemoteUpnFingerprint
			, bool throwIfMissing )
		{
			if( (AccountModel != null)
				&& AccountModel.Account.LocalUpnFingerprint.Equals( accountLocalUpnFingerprint )
				&& AccountModel.Account.RemoteUpnFingerprint.Equals( accountRemoteUpnFingerprint ) )
			{
				return false;
			}

			AccountModel = AccountModel.Load( ApplicationModel, accountLocalUpnFingerprint, accountRemoteUpnFingerprint
				, throwIfMissing );

			return true;
		}

		/// <summary>
		/// Updates the account model based on the current account model, if the specified account locator is equal with the
		/// locator of the account from the current account model. If the account with the current locator doesn't exist and
		/// "throwIfMissing" is true, an exception is thrown; if "throwIfMissing" is false, the account model is cleared.
		/// </summary>
		/// <param name="accountLocator">The update is performed only this this account locator is equal with the locator of the
		/// account from the current account model.</param>
		/// <param name="throwIfMissing">If this is true and the account with the current locator doesn't exist, an exception
		/// is thrown.</param>
		public void UpdateAccountModelForAccountLocator( Locator accountLocator, bool throwIfMissing )
		{
			ValidationUtils.ObjectHasContent( accountLocator, "accountLocator" );

			if( AccountModel == null )
				return;

			if( accountLocator.Equals( AccountModel.AccountLocator ) )
				SetAccountModel( AccountModel.AccountLocator, throwIfMissing );
		}

		/// <summary>
		/// Updates the account model based on the current account model, if the specified UPN fingerprint is equal with the
		/// remote UPN fingerprint from the current account model. If the account with the current locator doesn't exist and
		/// "throwIfMissing" is true, an exception is thrown; if "throwIfMissing" is false, the account model is cleared.
		/// </summary>
		/// <param name="accountRemoteUpnFingerprint">The update is performed only this this UPN fingerprint is equal with the
		/// remote UPN fingerprint from the current account model.</param>
		/// <param name="throwIfMissing">If this is true and the account with the current locator doesn't exist, an exception
		/// is thrown.</param>
		public void UpdateAccountModelForRemoteUpnFingerprint( Locator accountRemoteUpnFingerprint, bool throwIfMissing )
		{
			ValidationUtils.ObjectHasContent( accountRemoteUpnFingerprint, "accountRemoteUpnFingerprint" );

			if( AccountModel == null )
				return;

			if( accountRemoteUpnFingerprint.Equals( AccountModel.RemoteUpnFingerprint ) )
				SetAccountModel( AccountModel.AccountLocator, throwIfMissing );
		}

		/// <summary>
		/// Updates the account model based on the current account model, if the specified vault locator is equal with the
		/// locator of the vault from the current account model. If the account with the current locator doesn't exist and
		/// "throwIfMissing" is true, an exception is thrown; if "throwIfMissing" is false, the account model is cleared.
		/// </summary>
		/// <param name="vaultLocator">The update is performed only this this vault locator is equal with the locator of the
		/// vault from the current account model.</param>
		/// <param name="throwIfMissing">If this is true and the account with the current locator doesn't exist, an exception
		/// is thrown.</param>
		public void UpdateAccountModelForVaultLocator( Locator vaultLocator, bool throwIfMissing )
		{
			ValidationUtils.ObjectHasContent( vaultLocator, "vaultLocator" );

			if( AccountModel == null )
				return;

			if( vaultLocator.Equals( AccountModel.Vault.InstanceLocator ) )
				SetAccountModel( AccountModel.AccountLocator, throwIfMissing );
		}

		#endregion Public methods
	}
}
