////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.ApplicationBackend;
using Vela.ApplicationDocuments;
using Vela.Controls;
using Vela.Kernel;

namespace Vela.ImBackend
{
	/// <summary>
	/// Vaults model.
	/// </summary>
	public class VaultsModel : BaseModel
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected VaultsModel()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		public VaultsModel( ApplicationModel applicationModel )
			: base( applicationModel )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		private VaultSpeckles _Vaults;
		/// <summary>
		/// The speckles with the vaults. This may be empty, but not null. Notifies property observers. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public VaultSpeckles Vaults
		{
			get
			{
				if( _Vaults != null )
					return _Vaults;

				CollectVaults();

				return _Vaults;
			}
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Vaults" );

				_Vaults = value;

				// Notify property observers.
				OnPropertyChanged( "Vaults" );

				CurrentVault = null;
			}
		}

		private VaultSpeckle _CurrentVault;
		/// <summary>
		/// The current vault speckle. This may be null or empty. Notifies property observers.
		/// </summary>
		public VaultSpeckle CurrentVault
		{
			get { return _CurrentVault; }
			set
			{
				_CurrentVault = value;

				// Notify property observers.
				OnPropertyChanged( "CurrentVault" );
				OnPropertyChanged( "HasCurrentVault" );
				OnPropertyChanged( "VaultName" );
			}
		}

		/// <summary>
		/// This is true if a vault speckle is selected. READ-ONLY.
		/// </summary>
		public bool HasCurrentVault
		{
			get { return CurrentVault != null; }
		}

		/// <summary>
		/// The entity name of the current vault speckle. This is null if <see cref="CurrentVault"/> is null. READ-ONLY.
		/// </summary>
		public string EntityName
		{
			get { return (CurrentVault != null) ? (CurrentVault.EntityName as string) : null; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Queues action requests which push an updated presenter after its parent vault (with the specified locator) is edited,
		/// for each local account.
		/// </summary>
		/// <param name="vaultLocator">The instance locator of the edited vault.</param>
		/// <remarks>
		/// <para>This is a best-effort attempt. Since the queueing is performed on a separate thread, the user will not know
		/// if anything went wrong.</para>
		/// </remarks>
		private void QueuePushUpdatePresenter( Locator vaultLocator )
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			QueuePushUpdatePresenterEngine engine = new QueuePushUpdatePresenterEngine( sm.StorehouseProvider, vaultLocator
				, ApplicationModel.AlgorithmPool, sm.ActionClientManager );
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Collects all the vaults.
		/// </summary>
		public void CollectVaults()
		{
			// Get the locator of the selected speckle.
			Locator currentVaultLocator = (CurrentVault != null) ? CurrentVault.StorehouseKey : null;

			Vaults = ApplicationBackendStorehouse.CollectVaultSpeckles( ApplicationModel.StorehouseModel.StorehouseProvider );

			// Update the selected speckle.
			if( currentVaultLocator != null )
				CurrentVault = Vaults.FindByStorehouseKey( currentVaultLocator );
		}

		/// <summary>
		/// Adds a new vault.
		/// </summary>
		/// <param name="algorithmSecurityLevel">The security level for the cryptographic algorithms to use.</param>
		public void AddVault( AlgorithmSecurityLevel algorithmSecurityLevel )
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			VaultSpeckle speckle = ImBackendStorehouse.AddVault( sm.StorehouseProvider, ApplicationModel.AlgorithmPool
				, algorithmSecurityLevel );
			if( speckle == null )
				return;

			// Update the list of speckles.
			Vaults.Add( speckle );

			// Update the selected speckle.
			CurrentVault = speckle;

			sm.Restart( false );
		}

		/// <summary>
		/// Edits the current vault.
		/// </summary>
		public void EditVault()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasCurrentVault )
				return;

			bool upnsHaveChanged;
			VaultSpeckle speckle = ImBackendStorehouse.EditVault( sm.StorehouseProvider, CurrentVault.StorehouseKey
				, out upnsHaveChanged );
			if( speckle == null )
				return;

			// Update the list of speckles. Use "Replace" because the key may have changed.
			Vaults.AddOrReplace( CurrentVault, speckle );

			// Update the selected speckle.
			CurrentVault = speckle;

			// Update the lists of accounts and letters (because the entity name may have changed).
			sm.AccountsModel.CollectAccounts();
			sm.LettersModel.CollectLettersWithCurrentFlag( false );

			// Update the account model.
			sm.UpdateAccountModelForVaultLocator( speckle.StorehouseKey, false );

			// Restart the peer registration (because the automated UPNs of the appendix may have changed).
			if( upnsHaveChanged )
				sm.Restart( false );

			QueuePushUpdatePresenter( speckle.StorehouseKey );
		}

		/// <summary>
		/// Deletes the current vault.
		/// </summary>
		public void DeleteVault()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasCurrentVault )
				return;

			VaultSpeckle vaultToDelete = CurrentVault;

			#region Ask the user

			string entityName = ApplicationBackendStorehouse.GetVaultEntityName( sm.StorehouseProvider,
				vaultToDelete.StorehouseKey );

			// Ask the user.
			string message = string.Format( "Do you want to delete the vault with the name \"{0}\"?", entityName );
			if( MessageWindow.ShowDialog( message, MessageWindowModes.Query ) != QueryMessageResults.Yes )
				return;

			#region To continue, the user must type a text

			string confirmationText = string.Format( ImBackendConsts.DeleteVaultConfirmationTemplate
				, entityName.ToLowerInvariant() );
			string ensureMessage = string.Format( "Deleting the vault will make it impossible for you to talk"
					+ " to anyone to whom you have given its UPN. Type \"{0}\" if you still want to delete it."
				, confirmationText );
			string userText = EditSingleLineNoteWindow.ShowDialog( null, ensureMessage, null, false
				, TextValidations.ShortNote, true, KernelConsts.ShortNoteSizeMax );

			if( (userText == null) || !TextUtils.Equals( userText, confirmationText ) )
				return;

			#endregion To continue, the user must type a text

			#endregion Ask the user

			// Find the next speckle.
			VaultSpeckle nextSpeckle = Vaults.FindNext( vaultToDelete, false );

			DeleteVaultWindow deleteWindow = new DeleteVaultWindow( ApplicationModel, vaultToDelete.StorehouseKey
				, sm.ActionServiceManager, sm.ActionClientManager );
			if( !deleteWindow.ShowProgressDialog() )
				return;

			// Update the list of speckles.
			Vaults.RemoveItem( vaultToDelete );

			// Update the lists of accounts and letters.
			if( sm.AccountsModel != null )
				sm.AccountsModel.CollectAccounts();
			if( sm.LettersModel != null )
				sm.LettersModel.CollectLettersWithCurrentFlag( true );

			// Update the account model.
			sm.UpdateAccountModelForVaultLocator( vaultToDelete.StorehouseKey, false );

			// Update the selected speckle.
			CurrentVault = nextSpeckle;

			sm.Restart( false );

			TooltipWindow.ShowTooltip( string.Format( "The vault with the name \"{0}\" has been deleted.", entityName ) );
		}

		/// <summary>
		/// Exports the current presenter.
		/// </summary>
		public void ExportPresenter()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;
			AlgorithmPool algorithmPool = ApplicationModel.AlgorithmPool;

			if( !HasCurrentVault )
				return;

			ApplicationBackendStorehouse.ExportPresenter( sm.StorehouseProvider, CurrentVault.StorehouseKey
				, algorithmPool.GetMacAlg(), algorithmPool.GetAsymmSigningPadding(), algorithmPool.GetHashSerializer( false )
				, algorithmPool.GetSerializerForExport() );
		}

		/// <summary>
		/// Queues an action request for getting the presenter of the remote peer and creating an account for it, using a UPN.
		/// </summary>
		public void QueueCreateAccountWithUpn()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;
			AlgorithmPool algorithmPool = ApplicationModel.AlgorithmPool;

			if( !HasCurrentVault )
				return;

			Locator requestInstanceLocator = ImBackendStorehouse.QueueCreateAccountWithUpn( sm.StorehouseProvider
				, CurrentVault.StorehouseKey, algorithmPool.GetMacAlg(), algorithmPool.GetAsymmSigningPadding()
				, algorithmPool.GetHashSerializer( false ) );
			if( requestInstanceLocator == null )
				return;

			// Update the UI with the account speckle.
			AccountOrRequestSpeckle accountSpeckle = ApplicationBackendStorehouse.LoadAccountOrRequestSpeckle(
				sm.StorehouseProvider, requestInstanceLocator );
			sm.AccountsModel.Accounts.AddOrUpdate( accountSpeckle );

			string message = string.Format( "An account was queued to be created between the remote peer \"{0}\" and the local"
					+ " peer \"{1}\", with the \"{2}\" approval."
				, accountSpeckle.RemoteEntityName, accountSpeckle.LocalEntityName, accountSpeckle.Approval );
			TooltipWindow.ShowTooltip( message );

			// Notify the action client manager that a request was added to the storehouse.
			sm.ActionClientManager.Add( requestInstanceLocator, true );
		}

		/// <summary>
		/// Adds a new account using a presenter file.
		/// </summary>
		public void AddAccountWithFile()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasCurrentVault )
				return;

			Locator accountLocator = ApplicationBackendStorehouse.AddAccountWithFile( sm.StorehouseProvider
				, CurrentVault.StorehouseKey );
			if( accountLocator == null )
				return;

			// Update the UI with the account speckle.
			AccountOrRequestSpeckle accountSpeckle = ApplicationBackendStorehouse.LoadAccountOrRequestSpeckle(
				sm.StorehouseProvider, accountLocator );
			sm.AccountsModel.Accounts.AddOrUpdate( accountSpeckle );

			string message = string.Format( "An account was created between the remote peer \"{0}\" and the local peer"
					+ " \"{1}\", with the \"{2}\" approval."
				, accountSpeckle.RemoteEntityName, accountSpeckle.LocalEntityName, accountSpeckle.Approval );
			TooltipWindow.ShowTooltip( message );
		}

		/// <summary>
		/// Creates an attesting document with the current vault.
		/// </summary>
		public void CreateAttestingDocument()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasCurrentVault )
				return;

			ImBackendStorehouse.CreateAttestingDocument( sm.StorehouseProvider, ApplicationModel.AlgorithmPool
				, CurrentVault.StorehouseKey );
		}

		/// <summary>
		/// Verifies an attesting document with the current vault.
		/// </summary>
		public void VerifyAttestingDocument()
		{
			StorehouseModel sm = ApplicationModel.StorehouseModel;

			if( !HasCurrentVault )
				return;

			ImBackendStorehouse.VerifyAttestingDocumentWithVault( sm.StorehouseProvider, CurrentVault.StorehouseKey );
		}

		#endregion Public methods
	}
}
