////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.ComponentModel;
using System.Windows;
using Vela.Controls;
using Vela.Kernel;

namespace Vela.ImBackend
{
	/// <summary>
	/// Contract for a window view used to show the progress of an operation which is running on a background thread.
	/// </summary>
	/// <remarks>
	/// <para>When the background operation is finished, the window closes automatically.</para>
	/// </remarks>
	public abstract class BackgroundOperationWindow : AppViewWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected BackgroundOperationWindow()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		public BackgroundOperationWindow( ApplicationModel applicationModel )
			: base( applicationModel )
		{
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// Handler for the progress of long operations.
		/// </summary>
		protected LongOperationProgress LongOperationProgress;

		/// <summary>
		/// The worker thread.
		/// </summary>
		private BackgroundWorker BackgroundWorker;

		/// <summary>
		/// The error occurred during the background operation. This may be null.
		/// </summary>
		private Exception Error;

		#endregion Private fields

		#region Public non-serializable properties

		private double _ProgressPercentage;
		/// <summary>
		/// The progress (in percentages) of the background operation. Notifies property observers. READ-ONLY.
		/// </summary>
		public double ProgressPercentage
		{
			get { return _ProgressPercentage; }
			protected set
			{
				_ProgressPercentage = value;

				// Notify property observers.
				OnPropertyChanged( "ProgressPercentage" );
			}
		}

		private TimeSpan _ElapsedTime;
		/// <summary>
		/// The time span elapsed from the beginning of the background operation. Notifies property observers. READ-ONLY.
		/// </summary>
		public TimeSpan ElapsedTime
		{
			get { return _ElapsedTime; }
			protected set
			{
				_ElapsedTime = value;

				// Notify property observers.
				OnPropertyChanged( "ElapsedTime" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		/// <summary>
		/// Method called when the window loaded event is raised.
		/// </summary>
		/// <param name="sender">Sender.</param>
		/// <param name="e">Event data.</param>
		protected void ViewWindow_Loaded( object sender, RoutedEventArgs e )
		{
			Start();
		}

		/// <summary>
		/// Method called when the window's "Cancel" button is clicked.
		/// </summary>
		/// <param name="sender">Sender.</param>
		/// <param name="e">Event data.</param>
		protected void CancelButton_Click( object sender, RoutedEventArgs e )
		{
			try
			{
				Stop();

				DialogResult = false;
			}
			finally
			{
				BackgroundWorker.Dispose();
			}
		}

		private void BackgroundWorker_DoWork( object sender, DoWorkEventArgs e )
		{
			e.Result = Execute();
		}

		private void BackgroundWorker_RunWorkerCompleted( object sender, RunWorkerCompletedEventArgs e )
		{
			if( e.Error != null )
			{
				Error = e.Error;
				DialogResult = false;

				return;
			}

			BackgroundOperationSuccessful( e );

			if( IsVisible )
				DialogResult = true;
		}

		private void BackgroundWorker_ProgressChanged( object sender, ProgressChangedEventArgs e )
		{
			ProgressPercentage = e.ProgressPercentage / KernelConsts.PrecisionProgressToPercentageDivisor;
			ElapsedTime = (TimeSpan)e.UserState;
		}

		/// <summary>
		/// Handles the progress during the background operation.
		/// </summary>
		/// <param name="lop">The long operation progress object.</param>
		private void Lop_Progress( LongOperationProgress lop )
		{
			BackgroundWorker.ReportProgress( (int)(KernelConsts.PrecisionProgressMax * lop.LastProgress), lop.ElapsedTimeSpan );
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Method called when the background operation was successful.
		/// </summary>
		/// <param name="e">Information about the state of the background operation.</param>
		protected abstract void BackgroundOperationSuccessful( RunWorkerCompletedEventArgs e );

		/// <summary>
		/// Executes the core of the background operation.
		/// </summary>
		/// <returns>Returns the result of the background operation.</returns>
		protected abstract object ExecuteCore();

		/// <summary>
		/// Executes the background operation.
		/// </summary>
		/// <returns>Returns the result of the background operation.</returns>
		private object Execute()
		{
			LongOperationProgress = new LongOperationProgress( Lop_Progress );
			LongOperationProgress.Start();

			try
			{
				return ExecuteCore();
			}
			catch( AbortLongOperationException )
			{
				// Nothing to do.
			}

			return null;
		}

		/// <summary>
		/// Starts the background operation on a new thread.
		/// </summary>
		private void Start()
		{
			BackgroundWorker = new BackgroundWorker();
			BackgroundWorker.WorkerSupportsCancellation = true;
			BackgroundWorker.WorkerReportsProgress = true;
			BackgroundWorker.DoWork += BackgroundWorker_DoWork;
			BackgroundWorker.RunWorkerCompleted += BackgroundWorker_RunWorkerCompleted;
			BackgroundWorker.ProgressChanged += BackgroundWorker_ProgressChanged;

			BackgroundWorker.RunWorkerAsync();
		}

		/// <summary>
		/// Stops the background operation.
		/// </summary>
		private void Stop()
		{
			if( LongOperationProgress != null )
				LongOperationProgress.IsAbort = true;
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Shows the window in a modal manner.
		/// </summary>
		/// <returns>Returns true if the background operation has finished successfully, and false if the user has cancelled the
		/// background operation or an error has occurred during the background operation.</returns>
		/// <remarks>
		/// <para>If an exception has occurred during the background operation, it is rethrown after the window is closed.</para>
		/// </remarks>
		public bool ShowProgressDialog()
		{
			bool okay = ControlsUtils.ShowDialog( this );

			if( Error != null )
				throw Error;

			return okay;
		}

		#endregion Public methods
	}
}
