////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Text;
using System.Windows;
using System.Windows.Input;
using Vela.Controls;
using Vela.Kernel;

namespace Vela.ImBackend
{
	/// <summary>
	/// A window which displays the output of command lines.
	/// </summary>
	public partial class CommandLinesOutputWindow : AppViewWindow, IFloatingWindow
	{
		#region Public constants

		/// <summary>
		/// The separator between consecutive command lines output.
		/// </summary>
		public static readonly string CommandLinesSeparator = new string( '=', 80 );

		/// <summary>
		/// The separator between the body and the output of a command line.
		/// </summary>
		public static readonly string CommandLineBodyOutputSeparator = new string( '-', 80 );

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected CommandLinesOutputWindow()
		{
			InitializeComponent();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		public CommandLinesOutputWindow( ApplicationModel applicationModel )
			: base( applicationModel )
		{
			InitializeComponent();
		}

		#endregion Constructors

		#region Event handlers

		private void AppViewWindow_Loaded( object sender, RoutedEventArgs e )
		{
			OutputTextBox.Focus();
		}

		private void AppViewWindow_Closing( object sender, System.ComponentModel.CancelEventArgs e )
		{
			// Don't close the window.
			e.Cancel = true;

			// Hide the window.
			ControlsUtils.SetWindowVisibility( this, ApplicationVisibilities.Hidden );
		}

		private void AppViewWindow_KeyDown( object sender, System.Windows.Input.KeyEventArgs e )
		{
			if( Keyboard.Modifiers == ModifierKeys.None )
			{
				if( e.Key == Key.Escape )
				{
					// Hide the window.
					ControlsUtils.SetApplicationVisibility( ApplicationVisibilities.Hidden );

					e.Handled = true;

					return;
				}
			}
		}

		#endregion Event handlers

		#region Private static fields

		/// <summary>
		/// The global window for the output of command lines.
		/// </summary>
		private static CommandLinesOutputWindow Instance;

		#endregion Private static fields

		#region Public methods

		/// <summary>
		/// Appends the specified text to the existing one, and displays the window.
		/// </summary>
		public static void Show( ApplicationModel applicationModel, string text )
		{
			if( ControlsUtils.GetAppDispatcher() == null )
				return;

			if( Instance == null )
			{
				Instance = new CommandLinesOutputWindow( applicationModel );
				ControlsUtils.SetWindowOwnerToMainWindow( Instance );
			}

			ControlsUtils.SetWindowVisibility( Instance, ApplicationVisibilities.Normal );

			ControlsUtils.AppendText( Instance.OutputTextBox, text );
			ControlsUtils.FocusTextBox( Instance.OutputTextBox, true );
		}

		/// <summary>
		/// Appends the specified command line output to the existing one, and displays the window.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		/// <param name="creationTime">The creation time (in UTC format) of the command line.</param>
		/// <param name="commandLine">Command line to append. This may be null or empty.</param>
		/// <param name="commandLineOutput">Command line output to append. This may be null or empty.</param>
		public static void Show( ApplicationModel applicationModel, DateTime creationTime, string commandLine
			, string commandLineOutput )
		{
			if( TextUtils.HasContent( commandLineOutput ) )
			{
				StringBuilder sb = new StringBuilder();

				sb.AppendLine( creationTime.ToLocalTime().ToString( KernelConsts.TimeUiFormat ) );
				sb.AppendLine( commandLine );
				sb.AppendLine( CommandLineBodyOutputSeparator );
				sb.AppendLine( commandLineOutput.Trim() );
				sb.AppendLine( CommandLinesSeparator );

				Show( applicationModel, sb.ToString() );
			}
			else
			{
				Show( applicationModel, null );
			}
		}

		#endregion Public methods
	}
}
