////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.ComponentModel;
using Vela.ApplicationBackend;
using Vela.Kernel;

namespace Vela.ImBackend
{
	/// <summary>
	/// Window for deleting all the letters which have a specified flag.
	/// </summary>
	public partial class DeleteLettersWindow : BackgroundOperationWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected DeleteLettersWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		/// <param name="flagToDelete">The flag of the letters to delete.</param>
		/// <param name="afterDeleteAction">An action to be executed after each letter is deleted. This may be null.</param>
		public DeleteLettersWindow( ApplicationModel applicationModel, string flagToDelete, Action<Locator> afterDeleteAction )
			: base( applicationModel )
		{
			FlagToDelete = flagToDelete;
			AfterDeleteAction = afterDeleteAction;

			InitializeComponent();

			LoadUi();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// An action to be executed after each letter is deleted. This may be null.
		/// </summary>
		private Action<Locator> AfterDeleteAction;

		#endregion Private fields

		#region Public non-serializable properties

		private string _FlagToDelete;
		/// <summary>
		/// The flag of the letters to delete. Notifies property observers. READ-ONLY.
		/// </summary>
		public string FlagToDelete
		{
			get { return _FlagToDelete; }
			protected set
			{
				_FlagToDelete = value;

				// Notify property observers.
				OnPropertyChanged( "FlagToDelete" );
			}
		}

		private int _DeletedCount;
		/// <summary>
		/// The number of deleted letters. Notifies property observers. READ-ONLY.
		/// </summary>
		public int DeletedCount
		{
			get { return _DeletedCount; }
			protected set
			{
				_DeletedCount = value;

				// Notify property observers.
				OnPropertyChanged( "DeletedCount" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		/// <summary>
		/// Method called when the background operation was successful.
		/// </summary>
		/// <param name="e">Information about the state of the background operation.</param>
		protected override void BackgroundOperationSuccessful( RunWorkerCompletedEventArgs e )
		{
			ValidationUtils.ObjectHasContent( e, "e" );

			if( e.Result == null )
				return;

			DeletedCount = (int)e.Result;
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Executes the core of the background operation.
		/// </summary>
		/// <returns>Returns the number of deleted letters.</returns>
		protected override object ExecuteCore()
		{
			int deleted = 0;

			try
			{
				StorehouseModel sm = StorehouseModel;

				ApplicationBackendStorehouse.DeleteLetters( sm.StorehouseProvider, FlagToDelete, AfterDeleteAction
					, LongOperationProgress, sm.ActionServiceManager, sm.ActionClientManager, out deleted );
			}
			catch( AbortLongOperationException )
			{
				// Nothing to do.
			}

			return deleted;
		}

		#endregion Private methods
	}
}
