////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.ComponentModel;
using Vela.ApplicationBackend;
using Vela.Kernel;

namespace Vela.ImBackend
{
	/// <summary>
	/// Window for deleting all the transient letters.
	/// </summary>
	public partial class DeleteTransientLettersWindow : BackgroundOperationWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected DeleteTransientLettersWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		public DeleteTransientLettersWindow( ApplicationModel applicationModel )
			: base( applicationModel )
		{
			InitializeComponent();

			LoadUi();
		}

		#endregion Constructors

		#region Public non-serializable properties

		private bool _DeletingLetters = true;
		/// <summary>
		/// If this is true then the letters are being deleted and the progress can be shown. Notifies property observers.
		/// READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This has to be false while the storehouse is being compacted.</para>
		/// </remarks>
		public bool DeletingLetters
		{
			get { return _DeletingLetters; }
			protected set
			{
				_DeletingLetters = value;

				// Notify property observers.
				OnPropertyChanged( "DeletingLetters" );
			}
		}

		private int _DeletedCount;
		/// <summary>
		/// The number of deleted letters. Notifies property observers. READ-ONLY.
		/// </summary>
		public int DeletedCount
		{
			get { return _DeletedCount; }
			protected set
			{
				_DeletedCount = value;

				// Notify property observers.
				OnPropertyChanged( "DeletedCount" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		/// <summary>
		/// Method called when the background operation was successful.
		/// </summary>
		/// <param name="e">Information about the state of the background operation.</param>
		protected override void BackgroundOperationSuccessful( RunWorkerCompletedEventArgs e )
		{
			ValidationUtils.ObjectHasContent( e, "e" );

			if( e.Result == null ) return;

			DeletedCount = (int)e.Result;
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Executes the core of the background operation.
		/// </summary>
		/// <returns>Returns the number of deleted letters.</returns>
		protected override object ExecuteCore()
		{
			int deleted = 0;

			try
			{
				StorehouseModel sm = StorehouseModel;

				ApplicationBackendStorehouse.DeleteTransientLetters( sm.StorehouseProvider, LongOperationProgress
					, sm.ActionServiceManager, sm.ActionClientManager, out deleted );

				// Compact the storehouse.
				if( deleted > 0 )
				{
					DeletingLetters = false;

					sm.StorehouseProvider.CompactStorehouse( null, false );
				}
			}
			catch( AbortLongOperationException )
			{
				// Nothing to do.
			}

			return deleted;
		}

		#endregion Private methods
	}
}
