////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows;
using Vela.Controls;
using Vela.Kernel;
using Vela.PeerDocuments;

namespace Vela.ImBackend
{
	/// <summary>
	/// Window for editing an account.
	/// </summary>
	public partial class EditAccountWindow : ViewWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected EditAccountWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="accountContext">The account context to edit. This is modified in here.</param>
		public EditAccountWindow( AccountContext accountContext )
		{
			AccountContext = accountContext;

			InitialApproval = AccountContext.Tally.Approval;
			InitialAlternativePeerAddress = AccountContext.Tally.AlternativePeerAddress.Address;
			InitialAlternativeProcessingPort = AccountContext.Tally.AlternativePeerAddress.ProcessingPort;

			InitializeComponent();

			LoadUi();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The approval of the remote peer when the current window opens.
		/// </summary>
		private TallyApprovals InitialApproval;

		/// <summary>
		/// The address of the alternative peer address of the remote peer when the current window opens.
		/// </summary>
		private string InitialAlternativePeerAddress;

		/// <summary>
		/// The processing port of the alternative peer address of the remote peer when the current window opens.
		/// </summary>
		private UInt16 InitialAlternativeProcessingPort;

		#endregion Private fields

		#region Public non-serializable properties

		private AccountContext _AccountContext;
		/// <summary>
		/// The account context to edit. Notifies property observers. READ-ONLY.
		/// </summary>
		public AccountContext AccountContext
		{
			get { return _AccountContext; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "AccountContext" );

				_AccountContext = value;

				// Notify property observers.
				OnPropertyChanged( "AccountContext" );
				OnPropertyChanged( "PhotoVisibility" );
			}
		}

		/// <summary>
		/// The visibility of the photo. READ-ONLY.
		/// </summary>
		public Visibility PhotoVisibility
		{
			get
			{
				if( (AccountContext.RemotePhoto != null) && (AccountContext.RemotePhoto.Image != null)
					&& (((PeerImage)AccountContext.RemotePhoto.Image).SystemImage != null) )
				{
					return Visibility.Visible;
				}
				else
					return Visibility.Collapsed;
			}
		}

		private bool _ChangedApproval;
		/// <summary>
		/// This is true if the approval of the remote peer is different when the current window closes. READ-ONLY.
		/// </summary>
		public bool ChangedApproval
		{
			get { return _ChangedApproval; }
			private set { _ChangedApproval = value; }
		}

		private bool _ChangedAlternativePeerAddress;
		/// <summary>
		/// This is true if the alternative peer address of the remote peer is different when the current window closes.
		/// READ-ONLY.
		/// </summary>
		public bool ChangedAlternativePeerAddress
		{
			get { return _ChangedAlternativePeerAddress; }
			private set { _ChangedAlternativePeerAddress = value; }
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void ViewWindow_Loaded( object sender, RoutedEventArgs e )
		{
			ControlsUtils.FocusTextBox( AccountAliasTextBox, true );
		}

		private void ViewWindow_SizeChanged( object sender, SizeChangedEventArgs e )
		{
			ControlsUtils.CenterWindowOnScreen( sender as Window, -(int)PhotoContainer.ActualWidth / 2, 0 );
		}

		private void ClipLocalUpnButton_Click( object sender, RoutedEventArgs e )
		{
			Clipboard.SetText( AccountContext.Vault.LocalUpn.ToString() );

			TooltipWindow.ShowTooltip();
		}

		private void ClipRemoteUpnButton_Click( object sender, RoutedEventArgs e )
		{
			Clipboard.SetText( AccountContext.Presenter.CreatorUpn.ToString() );

			TooltipWindow.ShowTooltip();
		}

		private void EditLocalNoteButtonClick( object sender, RoutedEventArgs e )
		{
			EditMultiLineNoteWindow window = new EditMultiLineNoteWindow( "Local Note", null, AccountContext.Tally.Note, false
				, TextValidations.LongNote, false, KernelConsts.LongNoteSizeMax );
			if( !ControlsUtils.ShowDialog( window ) )
				return;

			AccountContext.Tally.Note = window.Note;
		}

		private void ViewIntimateNoteButtonClick( object sender, RoutedEventArgs e )
		{
			string note = (AccountContext.Presenter.Appendix.VirtualIdentity != null)
				? AccountContext.Presenter.Appendix.VirtualIdentity.IntimateNote : null;

			EditMultiLineNoteWindow window = new EditMultiLineNoteWindow( "Intimate Note", null, note, true
				, TextValidations.LongNote, false, KernelConsts.LongNoteSizeMax );
			ControlsUtils.ShowDialog( window );
		}

		private void ViewPublicNoteButtonClick( object sender, RoutedEventArgs e )
		{
			EditMultiLineNoteWindow window = new EditMultiLineNoteWindow( "Public Note", null
				, AccountContext.Presenter.Appendix.PublicNote, true, TextValidations.LongNote, false
				, KernelConsts.LongNoteSizeMax );
			ControlsUtils.ShowDialog( window );
		}

		private void OkayButton_Click( object sender, RoutedEventArgs e )
		{
			try
			{
				if( RemotePeerMayRunLocalCommandLinesCheckBox.IsChecked == true )
				{
					string message = string.Format( "The remote peer is allowed to run command lines on the local computer"
						+ " (by sending letters whose messages are used as command lines)."
						+ "\n\n"
						+ "Only if you trust the remote peer with all the information from your computer, you may continue." );

					if( ChoiceWindow.ShowDialog( message, "I accept", "I refuse" ) != ChoiceMessageResults.Left )
						return;
				}

				ValidateUi( ErrorPresenter );
				SaveUi( ErrorPresenter );

				ChangedApproval = InitialApproval != AccountContext.Tally.Approval;
				ChangedAlternativePeerAddress =
					(InitialAlternativePeerAddress != AccountContext.Tally.AlternativePeerAddress.Address)
					|| (InitialAlternativeProcessingPort != AccountContext.Tally.AlternativePeerAddress.ProcessingPort);

				DialogResult = true;
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				ControlsUtils.ShowException( ex );
			}
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Validates the input data.
		/// </summary>
		/// <param name="errorPresenter">The control which shows that there is a validation error. This may be null.</param>
		/// <remarks>
		/// <para>If you don't validate of all the editable fields in here, it's possible that some bindings are commited even
		/// if the user cancels the edit operation. This can happen if the user edits some fields, clicks the "Okay" button, the
		/// validation fails (during binding commit) inside some properties (which are not validated in here), and then the user
		/// clicks the "Cancel" button. This is likely to matter only if those (successfully) bound properties are used outside
		/// the view.</para>
		/// </remarks>
		protected override void ValidateUi( FrameworkElement errorPresenter )
		{
			base.ValidateUi( errorPresenter );

			ValidationUtils.Text( TallyEntityNameTextBox.Text, -1, -1, TextValidations.EntityName, "Tally entity name"
				, TallyEntityNameTextBox, errorPresenter );

			string autoAccountAlias = AccountAlias.FromLocalAndRemoteEntityNames( AccountContext.Vault.EntityName
				, TallyEntityNameTextBox.Text ).TextualForm;
			ValidationUtils.Text( AccountAliasTextBox.Text, -2, -1, TextValidations.Alias
				, string.Format( "Account alias (suggested \"{0}\")", autoAccountAlias ), AccountAliasTextBox, errorPresenter );

			ValidationUtils.Text( AlternativePeerAddressTextBox.Text, -2, -1, TextValidations.AlternativePeerAddress
				, "Alternative peer address", AlternativePeerAddressTextBox, errorPresenter );
			ValidationUtils.Integer( OverridingProcessingPortTextBox.Text, 0, UInt16.MaxValue, "Overriding processing port"
				, OverridingProcessingPortTextBox, errorPresenter );

			TallyApprovals tallyApproval = (TallyApprovals)TallyApprovalComboBox.SelectedItem;
			if( RemotePeerMayRunLocalCommandLinesCheckBox.IsChecked == true
				&& !PeerDocumentsUtils.RemotePeerMayRunLocalCommandLines( tallyApproval ) )
			{
				string message = string.Format( "In order for the remote peer to be allowed to run command lines on the local"
						+ " computer (by sending letters whose messages are used as command lines), its tally approval must be"
						+ " at least \"{0}\"."
					, PeerDocumentsConsts.RemotePeerMayRunLocalCommandLinesTallyApprovalMin );

				throw new ValidationException( message );
			}
		}

		#endregion Private methods
	}
}
