////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Microsoft.Win32;
using System;
using System.Windows;
using Vela.Controls;
using Vela.Kernel;
using Vela.PeerDocuments;

namespace Vela.ImBackend
{
	/// <summary>
	/// Window for editing a vault (<see cref="Vault"/>).
	/// </summary>
	public partial class EditVaultWindow : ViewWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected EditVaultWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="vault">The vault to edit.</param>
		/// <param name="photo">The photo to edit.</param>
		public EditVaultWindow( Vault vault, Photo photo )
		{
			Vault = vault;
			Photo = photo;

			InitializeComponent();

			LoadUi();
		}

		#endregion Constructors

		#region Public non-serializable properties

		private Vault _Vault;
		/// <summary>
		/// The vault to edit. Notifies property observers. READ-ONLY.
		/// </summary>
		public Vault Vault
		{
			get { return _Vault; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Vault" );

				_Vault = value;

				// Notify property observers.
				OnPropertyChanged( "Vault" );
			}
		}

		private bool _UpnsHaveChanged;
		/// <summary>
		/// This is true if the (automated) UPNs of the edited vault have changed. Notifies property observers. READ-ONLY.
		/// </summary>
		public bool UpnsHaveChanged
		{
			get { return _UpnsHaveChanged; }
			protected set
			{
				_UpnsHaveChanged = value;

				// Notify property observers.
				OnPropertyChanged( "UpnsHaveChanged" );
			}
		}

		private Photo _Photo;
		/// <summary>
		/// The photo to edit. Notifies property observers. READ-ONLY.
		/// </summary>
		public Photo Photo
		{
			get { return _Photo; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Photo" );

				_Photo = value;

				// Notify property observers.
				OnPropertyChanged( "Photo" );
				OnPropertyChanged( "PeerImage" );
				OnPropertyChanged( "PhotoVisibility" );
			}
		}

		/// <summary>
		/// The peer image. This is null if <see cref="Photo"/> is null, or if its
		/// <see cref="Vela.PeerDocuments.Photo.Image"/> property is null. Notifies property observers. READ-ONLY.
		/// </summary>
		public IPeerImage PeerImage
		{
			get { return (Photo != null) ? Photo.Image : null; }
			protected set
			{
				Photo.Image = (Photo != null) ? value : null;

				// Notify property observers.
				OnPropertyChanged( "Photo" );
				OnPropertyChanged( "PeerImage" );
				OnPropertyChanged( "PhotoVisibility" );
			}
		}

		/// <summary>
		/// The visibility of the photo. READ-ONLY.
		/// </summary>
		public Visibility PhotoVisibility
		{
			get { return (PeerImage != null) ? Visibility.Visible : Visibility.Collapsed; }
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void ViewWindow_Loaded( object sender, RoutedEventArgs e )
		{
			ControlsUtils.FocusTextBox( PeerEntityNameTextBox, true );
		}

		private void ViewWindow_SizeChanged( object sender, SizeChangedEventArgs e )
		{
			ControlsUtils.CenterWindowOnScreen( sender as Window, -(int)PhotoContainer.ActualWidth / 2, 0 );
		}

		private void ClipUpnButton_Click( object sender, RoutedEventArgs e )
		{
			Clipboard.SetText( Vault.LocalUpn.ToString() );

			TooltipWindow.ShowTooltip();
		}

		private void ClipAlternativePeerAddressInfoButton_Click( object sender, RoutedEventArgs e )
		{
			string alternativePeerAddressInfo = ImBackendUtils.GetAlternativePeerAddressInfo( Vault.LocalUpn.LongHostLabel
				, Vault.LocalUpn.ShortHostLabel );

			Clipboard.SetText( alternativePeerAddressInfo );

			TooltipWindow.ShowTooltip();
		}

		private void LoadPhotoButton_Click( object sender, RoutedEventArgs e )
		{
			LoadPhoto();
		}

		private void DeletePhotoButton_Click( object sender, RoutedEventArgs e )
		{
			PeerImage = null;
		}

		private void EditPrivateNoteButtonClick( object sender, RoutedEventArgs e )
		{
			EditMultiLineNoteWindow window = new EditMultiLineNoteWindow( "Private Note", null, Vault.PrivateNote, false
				, TextValidations.LongNote, false, KernelConsts.LongNoteSizeMax );
			if( !ControlsUtils.ShowDialog( window ) )
				return;

			Vault.PrivateNote = window.Note;
		}

		private void EditIntimateNoteButtonClick( object sender, RoutedEventArgs e )
		{
			EditMultiLineNoteWindow window = new EditMultiLineNoteWindow( "Intimate Note", null
				, Vault.Appendix.VirtualIdentity.IntimateNote, false, TextValidations.LongNote, false
				, KernelConsts.LongNoteSizeMax );
			if( !ControlsUtils.ShowDialog( window ) )
				return;

			Vault.Appendix.VirtualIdentity.IntimateNote = window.Note;
		}

		private void EditPublicNoteButtonClick( object sender, RoutedEventArgs e )
		{
			EditMultiLineNoteWindow window = new EditMultiLineNoteWindow( "Public Note", null, Vault.Appendix.PublicNote, false
				, TextValidations.LongNote, false, KernelConsts.LongNoteSizeMax );
			if( !ControlsUtils.ShowDialog( window ) )
				return;

			Vault.Appendix.PublicNote = window.Note;
		}

		private void OkayButton_Click( object sender, RoutedEventArgs e )
		{
			try
			{
				ValidateUi( ErrorPresenter );
				SaveUi( ErrorPresenter );

				DialogResult = true;
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				ControlsUtils.ShowException( ex );
			}
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Loads a new photo.
		/// </summary>
		private void LoadPhoto()
		{
			OpenFileDialog ofd = new OpenFileDialog();
			ofd.Title = "Load Photo";
			ofd.AddExtension = true;
			ofd.Filter = FileExtensions.ImageFilter;
			ofd.CheckFileExists = true;

			if( !ControlsUtils.ShowDialog( ofd ) )
				return;

			// Load the photo.
			byte[] imageData = GraphicUtils.LoadResizeImageBuffer( ofd.FileName, (int)PeerDocumentsConsts.FullSizedPhotoWidth
				, (int)PeerDocumentsConsts.FullSizedPhotoHeight, true, PeerDocumentsConsts.PhotoQuality );

			// Update the photo.
			PeerImage = new PeerImage( null, imageData );
		}

		/// <summary>
		/// Validates the input data.
		/// </summary>
		/// <param name="errorPresenter">The control which shows that there is a validation error. This may be null.</param>
		/// <remarks>
		/// <para>If you don't validate of all the editable fields in here, it's possible that some bindings are commited even
		/// if the user cancels the edit operation. This can happen if the user edits some fields, clicks the "Okay" button, the
		/// validation fails (during binding commit) inside some properties (which are not validated in here), and then the user
		/// clicks the "Cancel" button. This is likely to matter only if those (successfully) bound properties are used outside
		/// the view.</para>
		/// </remarks>
		protected override void ValidateUi( FrameworkElement errorPresenter )
		{
			base.ValidateUi( errorPresenter );

			ValidationUtils.Text( PeerEntityNameTextBox.Text, -1, -1, TextValidations.EntityName, "Entity name"
				, PeerEntityNameTextBox, errorPresenter );
			ValidationUtils.Date( BirthDateTextBox.Text, KernelConsts.DateMin, DateTime.MaxValue, "Birth date"
				, BirthDateTextBox, errorPresenter );
			ValidationUtils.Text( EmailAddressTextBox.Text, -2, -1, TextValidations.EmailAddress, "Email address"
				, EmailAddressTextBox, errorPresenter );
			ValidationUtils.Text( PhoneNumberTextBox.Text, -2, -1, TextValidations.PhoneNumber, "Phone number"
				, PhoneNumberTextBox, errorPresenter );
			ValidationUtils.Text( WebsiteUrlTextBox.Text, -2, -1, TextValidations.Url, "Website URL"
				, WebsiteUrlTextBox, errorPresenter );
		}

		#endregion Private methods
	}
}
