////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.ComponentModel;
using System.Windows;
using Vela.Controls;
using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Kernel;
using Vela.PeerDocuments;

namespace Vela.ImBackend
{
	/// <summary>
	/// Window for generating an identity for a vault (<see cref="Vault"/>).
	/// </summary>
	public partial class GenerateIdentityWindow : BackgroundOperationWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected GenerateIdentityWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="upnAlg">UPN algorithm.</param>
		/// <param name="processingPort">The network port of a service where action requests are received from clients.</param>
		/// <param name="encryptionAca">Asymmetric encryption algorithm.</param>
		/// <param name="signingAca">Asymmetric signing algorithm.</param>
		/// <param name="algorithmSecurityLevel">The security level for the cryptographic algorithms to use.</param>
		public GenerateIdentityWindow( UpnAlg upnAlg, UInt16 processingPort, AsymmAlg encryptionAca, AsymmAlg signingAca
			, AlgorithmSecurityLevel algorithmSecurityLevel )
		{
			UpnAlg = upnAlg;
			ProcessingPort = processingPort;
			EncryptionAca = encryptionAca;
			SigningAca = signingAca;
			AlgorithmSecurityLevel = algorithmSecurityLevel;

			InitializeComponent();

			LoadUi();
		}

		#endregion Constructors

		#region Public non-serializable properties

		private UpnAlg _UpnAlg;
		/// <summary>
		/// UPN algorithm. Notifies property observers. READ-ONLY.
		/// </summary>
		public UpnAlg UpnAlg
		{
			get { return _UpnAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "UpnAlg" );

				_UpnAlg = value;

				// Notify property observers.
				OnPropertyChanged( "UpnAlg" );
			}
		}

		private UInt16 _ProcessingPort;
		/// <summary>
		/// The network port of a service where action requests are received from clients. This must be at least 1. Notifies
		/// property observers. READ-ONLY.
		/// </summary>
		public UInt16 ProcessingPort
		{
			get { return _ProcessingPort; }
			protected set
			{
				ValidationUtils.Integer( value, 1, UInt16.MaxValue, "ProcessingPort" );

				_ProcessingPort = value;

				// Notify property observers.
				OnPropertyChanged( "ProcessingPort" );
			}
		}

		private AsymmAlg _EncryptionAca;
		/// <summary>
		/// Asymmetric encryption algorithm. Notifies property observers. READ-ONLY.
		/// </summary>
		public AsymmAlg EncryptionAca
		{
			get { return _EncryptionAca; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "EncryptionAca" );

				_EncryptionAca = value;

				// Notify property observers.
				OnPropertyChanged( "EncryptionAca" );
			}
		}

		private AsymmAlg _SigningAca;
		/// <summary>
		/// Asymmetric signing algorithm. Notifies property observers. READ-ONLY.
		/// </summary>
		public AsymmAlg SigningAca
		{
			get { return _SigningAca; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "SigningAca" );

				_SigningAca = value;

				// Notify property observers.
				OnPropertyChanged( "SigningAca" );
			}
		}

		private AlgorithmSecurityLevel _AlgorithmSecurityLevel;
		/// <summary>
		/// The security level for the cryptographic algorithms to use. Notifies property observers. READ-ONLY.
		/// </summary>
		public AlgorithmSecurityLevel AlgorithmSecurityLevel
		{
			get { return _AlgorithmSecurityLevel; }
			protected set
			{
				_AlgorithmSecurityLevel = value;

				// Notify property observers.
				OnPropertyChanged( "AlgorithmSecurityLevel" );
				OnPropertyChanged( "AlgorithmSecurityHint" );
			}
		}

		/// <summary>
		/// The algorithm security hint. READ-ONLY.
		/// </summary>
		public string AlgorithmSecurityHint
		{
			get
			{
				switch( AlgorithmSecurityLevel )
				{
					case AlgorithmSecurityLevel.Normal:
						return "(The normal algorithms are used, so this may take a few seconds.)";
					case AlgorithmSecurityLevel.Strong:
						return "(The strong algorithms are used, so this may take a few minutes.)";
					case AlgorithmSecurityLevel.Strongest:
						return "(The strongest algorithms are used, so this may take a few tens of minutes.)";
					default:
						throw ValidationUtils.Unhandled( AlgorithmSecurityLevel, "AlgorithmSecurityLevel" );
				}
			}
		}

		private PrivateIdentity _GeneratedIdentity;
		/// <summary>
		/// The generated identity. Notifies property observers. READ-ONLY.
		/// </summary>
		public PrivateIdentity GeneratedIdentity
		{
			get { return _GeneratedIdentity; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "GeneratedIdentity" );

				_GeneratedIdentity = value;

				// Notify property observers.
				OnPropertyChanged( "GeneratedIdentity" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		/// <summary>
		/// Method called when the background operation was successful.
		/// </summary>
		/// <param name="e">Information about the state of the background operation.</param>
		protected override void BackgroundOperationSuccessful( RunWorkerCompletedEventArgs e )
		{
			ValidationUtils.ObjectHasContent( e, "e" );

			if( e.Result == null ) return;

			GeneratedIdentity = e.Result as PrivateIdentity;
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Executes the core of the background operation.
		/// </summary>
		/// <returns>Returns the generated identity.</returns>
		protected override object ExecuteCore()
		{
			// Generate the asymmetric keys.
			EncryptionAca.GenerateKeyPair( LongOperationProgress );
			SigningAca.GenerateKeyPair( LongOperationProgress );

			PrivateIdentity identity = new PrivateIdentity( UpnAlg, ProcessingPort, EncryptionAca, SigningAca );

			string message = string.Format( "Identity generated in {0:0.0} seconds.", LongOperationProgress.ElapsedTime );
			LogManager.Add( message );

			return identity;
		}

		#endregion Private methods
	}
}
