////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.ComponentModel;
using Vela.ApplicationBackend;
using Vela.ApplicationDocuments;
using Vela.Kernel;

namespace Vela.ImBackend
{
	/// <summary>
	/// Window for getting as a text all the letters which have a specified flag.
	/// </summary>
	public partial class GetLettersAsTextWindow : BackgroundOperationWindow
	{
		#region Inner classes

		/// <summary>
		/// The result of the background worker.
		/// </summary>
		private class BackgroundWorkerResult
		{
			public BackgroundWorkerResult( int processedCount, string lettersAsText )
			{
				ProcessedCount = processedCount;
				LettersAsText = lettersAsText;
			}

			public int ProcessedCount;
			public string LettersAsText;
		}

		#endregion Inner classes

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected GetLettersAsTextWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		/// <param name="flag">The flag of the letters to get as a text.</param>
		/// <param name="accounts">List of accounts from which to get the account aliases.</param>
		public GetLettersAsTextWindow( ApplicationModel applicationModel, string flag, AccountOrRequestSpeckles accounts )
			: base( applicationModel )
		{
			ValidationUtils.ObjectHasContent( accounts, "accounts" );

			Flag = flag;
			Accounts = accounts;

			InitializeComponent();

			LoadUi();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// List of accounts from which to get the account aliases.
		/// </summary>
		private AccountOrRequestSpeckles Accounts;

		#endregion Private fields

		#region Public non-serializable properties

		private string _Flag;
		/// <summary>
		/// The flag of the letters to get as a text. Notifies property observers. READ-ONLY.
		/// </summary>
		public string Flag
		{
			get { return _Flag; }
			protected set
			{
				_Flag = value;

				// Notify property observers.
				OnPropertyChanged( "Flag" );
			}
		}

		private int _ProcessedCount;
		/// <summary>
		/// The number of letters which were retrieved as a text. READ-ONLY.
		/// </summary>
		public int ProcessedCount
		{
			get { return _ProcessedCount; }
			protected set { _ProcessedCount = value; }
		}

		private string _LettersAsText;
		/// <summary>
		/// The letters retrieved as a text. This may be null or empty. READ-ONLY.
		/// </summary>
		public string LettersAsText
		{
			get { return _LettersAsText; }
			protected set { _LettersAsText = value; }
		}

		#endregion Public non-serializable properties

		#region Event handlers

		/// <summary>
		/// Method called when the background operation was successful.
		/// </summary>
		/// <param name="e">Information about the state of the background operation.</param>
		protected override void BackgroundOperationSuccessful( RunWorkerCompletedEventArgs e )
		{
			ValidationUtils.ObjectHasContent( e, "e" );

			if( e.Result == null )
				return;

			BackgroundWorkerResult result = (BackgroundWorkerResult)e.Result;
			ProcessedCount = result.ProcessedCount;
			LettersAsText = result.LettersAsText;
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Executes the core of the background operation.
		/// </summary>
		/// <returns>Returns the number of letters retrieved as a text.</returns>
		protected override object ExecuteCore()
		{
			string lettersAsText = null;
			int processed = 0;

			try
			{
				StorehouseModel sm = StorehouseModel;

				ApplicationBackendStorehouse.GetLettersAsText( sm.StorehouseProvider, Flag, Accounts, LongOperationProgress
					, out lettersAsText, out processed );
			}
			catch( AbortLongOperationException )
			{
				// Nothing to do.
			}

			return new BackgroundWorkerResult( processed, lettersAsText );
		}

		#endregion Private methods
	}
}
