////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.ComponentModel;
using Vela.ApplicationBackend;
using Vela.Kernel;

namespace Vela.ImBackend
{
	/// <summary>
	/// Window for importing letters.
	/// </summary>
	public partial class ImportLettersWindow : BackgroundOperationWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ImportLettersWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		/// <param name="fileNames">The names of the files from which to import letters. This may be empty, but not null.</param>
		public ImportLettersWindow( ApplicationModel applicationModel, string[] fileNames )
			: base( applicationModel )
		{
			FileNames = fileNames;

			InitializeComponent();

			LoadUi();
		}

		#endregion Constructors

		#region Public non-serializable properties

		private string[] _FileNames;
		/// <summary>
		/// The number of imported letters. This may be empty, but not null. Notifies property observers. READ-ONLY.
		/// </summary>
		public string[] FileNames
		{
			get { return _FileNames; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "FileNames" );

				_FileNames = value;

				// Notify property observers.
				OnPropertyChanged( "FileNames" );
			}
		}

		private int _ImportedCount;
		/// <summary>
		/// The number of imported letters. Notifies property observers. READ-ONLY.
		/// </summary>
		public int ImportedCount
		{
			get { return _ImportedCount; }
			protected set
			{
				_ImportedCount = value;

				// Notify property observers.
				OnPropertyChanged( "ImportedCount" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		/// <summary>
		/// Method called when the background operation was successful.
		/// </summary>
		/// <param name="e">Information about the state of the background operation.</param>
		protected override void BackgroundOperationSuccessful( RunWorkerCompletedEventArgs e )
		{
			ValidationUtils.ObjectHasContent( e, "e" );

			if( e.Result == null )
				return;

			ImportedCount = (int)e.Result;
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Executes the core of the background operation.
		/// </summary>
		/// <returns>Returns the number of imported letters.</returns>
		protected override object ExecuteCore()
		{
			int importedCount = 0;

			try
			{
				StorehouseModel sm = StorehouseModel;

				ApplicationBackendStorehouse.ImportLetters( sm.StorehouseProvider, FileNames, LongOperationProgress
					, ApplicationModel.AlgorithmPool, out importedCount );
			}
			catch( AbortLongOperationException )
			{
				// Nothing to do.
			}

			return importedCount;
		}

		#endregion Private methods
	}
}
