////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Windows;
using Vela.ApplicationBackend;
using Vela.Controls;
using Vela.Crypto;
using Vela.Kernel;
using Vela.PeerDocuments;

namespace Vela.ImBackend
{
	/// <summary>
	/// A desktop-level notification shown for a received letter.
	/// </summary>
	public partial class LetterDesktopNotificationWindow : AppViewWindow, IFloatingWindow
	{
		#region Private constants

		/// <summary>
		/// Funny texts to show when the message summary (of the received letters) may not be shown, or there is none to show.
		/// </summary>
		private static readonly string[] FunnyTexts = new string[]
		{
			"Hellooo?!... Hellooo?!... Anybody there?... Anybody there?... Echooo... Echooo..."
			, "Ayyy, caraaamba! Some people just won't let you sleep!"
			, "It seems that only smoke signs can draw your attention on important matters. So here you go!"
			, "Holy shmoly! Important things are going down and you are the last one to find out!"
			, "Seriously? This is what things have come to? You can't just answer normal letters? You have to be summoned?"
			, "In the name of Jupiter, the planets have aligned. Check the letter for the treasure map!"
			, "Is that your boss looking for you, by any chance?"
			, "Holmes, I think your powers of observation are failing you! You didn't see this letter coming, did you?"
			, "Come Watson, come! The game is afoot. Quickly man, read the telegram from our client!"
			, "Sacrebleu! Hastings, I have been a fool! The letter was in front of me all this time!"
			, "From the mysterious depths of the blue abyss, a letter in a bottle has been washed ashore, at your feet."
			, "Finally, the aliens have arrived and you're the first they've contacted."
			, "There may be cookies waiting for you on the other side. Touch this to be transported there!"
			, "You seem to be an important person. People that matter are looking for you!"
			, "People who don't know how busy you are, are pestering you! It's your chance to lecture them about bad timing!"
			, "Push the red button! Come one, push it! You know to want to!"
		};

		/// <summary>
		/// The margin (in pixels) to have around the notification window.
		/// </summary>
		private const int MarginAroundWindow = 10;

		/// <summary>
		/// The maximum size (in characters) of the message summary (of the received letters) to show.
		/// </summary>
		private const int MessageSummarySizeMax = 100;

		/// <summary>
		/// The maximum number of lines of the message summary (of the received letters) to show.
		/// </summary>
		private const int MessageSummaryLineCountMax = 2;

		#endregion Private constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected LetterDesktopNotificationWindow()
		{
			InitializeComponent();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		/// <param name="initiatorTally">The tally of the initiator of the letter.</param>
		/// <param name="initiatorAccount">The account of the initiator of the letter.</param>
		/// <param name="letter">Letter for which to show the notification associated to the emoticon command of the contained
		/// emoticon symbol.</param>
		protected LetterDesktopNotificationWindow( ApplicationModel applicationModel, Tally initiatorTally
				, Account initiatorAccount, Letter letter )
			: base( applicationModel )
		{
			ValidationUtils.ObjectHasContent( initiatorTally, "initiatorTally" );
			ValidationUtils.ObjectHasContent( initiatorAccount, "initiatorAccount" );
			ValidationUtils.ObjectHasContent( letter, "letter" );

			InitiatorTally = initiatorTally;
			InitiatorAccount = initiatorAccount;
			Letter = letter;

			RemoteEntityName = initiatorTally.EntityName;

			#region Prepare the message to show

			// Delete the emoticon symbol from the message to show.
			string message = TextUtils.HasContent( Letter.EmoticonSymbol )
				? TextUtils.Replace( Letter.MessageSummary, Letter.EmoticonSymbol, "" ) : Letter.MessageSummary;

			if( ApplicationModel.Configuration.ShowLetterMessageSummaryInDesktopNotification && TextUtils.HasContent( message ) )
			{
				Message = TextUtils.GetBeginning( message, MessageSummarySizeMax, true, MessageSummaryLineCountMax
					, PeerDocumentsConsts.PostMessageHasMoreContentSuffix );
			}
			else
			{
				int r = EntropyPool.GetInt( 0, FunnyTexts.Length );
				Message = FunnyTexts[ r ];
			}

			#endregion Prepare the message to show

			InitializeComponent();
		}

		#endregion Constructors

		#region Private static fields

		/// <summary>
		/// The global instance window.
		/// </summary>
		private static LetterDesktopNotificationWindow Instance;

		#endregion Private static fields

		#region Private fields

		/// <summary>
		/// The tally of the initiator of the letter.
		/// </summary>
		private Tally InitiatorTally;

		/// <summary>
		/// The account of the initiator of the letter.
		/// </summary>
		private Account InitiatorAccount;

		/// <summary>
		/// Letter for which to show the notification associated to the emoticon command of the contained emoticon symbol.
		/// </summary>
		private Letter Letter;

		#endregion Private fields

		#region Public non-serializable properties

		private string _RemoteEntityName;
		/// <summary>
		/// The entity name of the remote peer. This may be null or empty. Notifies property observers. READ-ONLY.
		/// </summary>
		public string RemoteEntityName
		{
			get { return _RemoteEntityName; }
			protected set
			{
				_RemoteEntityName = value;

				// Notify property observers.
				OnPropertyChanged( "RemoteEntityName" );
			}
		}

		private string _Message;
		/// <summary>
		/// Message summary to show. This may be null or empty. Notifies property observers. READ-ONLY.
		/// </summary>
		public string Message
		{
			get { return _Message; }
			protected set
			{
				_Message = value;

				// Notify property observers.
				OnPropertyChanged( "Message" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void ViewWindow_Loaded( object sender, RoutedEventArgs e )
		{
			ControlsUtils.RepositionBottomRightToScreen( this, MarginAroundWindow, MarginAroundWindow );
		}

		private void ViewWindow_PreviewMouseDown( object sender, System.Windows.Input.MouseButtonEventArgs e )
		{
			AudioNotificationManager.Stop( true );

			CloseNotification();

			if( StorehouseModel != null )
				StorehouseModel.SetAccountModel( InitiatorAccount.InstanceLocator, false );

			// Show the main window.
			ControlsUtils.SetApplicationVisibility( ApplicationVisibilities.Normal );
		}

		#endregion Event handlers

		#region Public static methods

		/// <summary>
		/// Shows the window using the global instance (<see cref="Instance"/>).
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		/// <param name="initiatorTally">The tally of the initiator of the letter.</param>
		/// <param name="initiatorAccount">The account of the initiator of the letter.</param>
		/// <param name="letter">Letter for which to show the notification associated to the emoticon command of the contained
		/// emoticon symbol.</param>
		public static void ShowNotification( ApplicationModel applicationModel, Tally initiatorTally, Account initiatorAccount
			, Letter letter )
		{
			if( ControlsUtils.GetAppDispatcher() == null )
				return;

			if( Instance != null )
				return; // Keep the existing notification.

			Instance = new LetterDesktopNotificationWindow( applicationModel, initiatorTally, initiatorAccount, letter );
			Instance.Show();
		}

		/// <summary>
		/// Closes the global instance (<see cref="Instance"/>) of the window (if it's visible).
		/// </summary>
		public static void CloseNotification()
		{
			if( ControlsUtils.GetAppDispatcher() == null )
				return;

			if( Instance != null )
				Instance.Close();

			Instance = null;
		}

		#endregion Public static methods
	}
}
