////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.ComponentModel;
using System.IO;
using Vela.Kernel;
using Vela.StarkDocuments;

namespace Vela.ImBackend
{
	/// <summary>
	/// Window for processing an attesting document.
	/// </summary>
	public partial class ProcessAttestingDocumentWindow : BackgroundOperationWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ProcessAttestingDocumentWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="actionName">The name of the executed action.</param>
		/// <param name="fileName">The name of the file to process.</param>
		/// <param name="attestingDocument">The associated attesting document.</param>
		public ProcessAttestingDocumentWindow( string actionName, string fileName, AttestingDocument attestingDocument )
		{
			ActionName = actionName;
			FileName = fileName;
			AttestingDocument = attestingDocument;

			InitializeComponent();

			LoadUi();
		}

		#endregion Constructors

		#region Public non-serializable properties

		private string _ActionName;
		/// <summary>
		/// The name of the executed action. Notifies property observers. READ-ONLY.
		/// </summary>
		public string ActionName
		{
			get { return _ActionName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "ActionName" );

				_ActionName = value;

				// Notify property observers.
				OnPropertyChanged( "ActionName" );
			}
		}

		private string _FileName;
		/// <summary>
		/// The name of the file to process. Notifies property observers. READ-ONLY.
		/// </summary>
		public string FileName
		{
			get { return _FileName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "FileName" );

				_FileName = value;

				// Notify property observers.
				OnPropertyChanged( "FileName" );
			}
		}

		private AttestingDocument _AttestingDocument;
		/// <summary>
		/// The associated attesting document. READ-ONLY.
		/// </summary>
		public AttestingDocument AttestingDocument
		{
			get { return _AttestingDocument; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "AttestingDocument" );

				_AttestingDocument = value;
			}
		}

		private byte[] _MacValue;
		/// <summary>
		/// The computed MAC. Notifies property observers. READ-ONLY.
		/// </summary>
		public byte[] MacValue
		{
			get { return _MacValue; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "MacValue" );

				_MacValue = value;

				// Notify property observers.
				OnPropertyChanged( "MacValue" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		/// <summary>
		/// Method called when the background operation was successful.
		/// </summary>
		/// <param name="e">Information about the state of the background operation.</param>
		protected override void BackgroundOperationSuccessful( RunWorkerCompletedEventArgs e )
		{
			ValidationUtils.ObjectHasContent( e, "e" );

			if( e.Result == null ) return;

			MacValue = e.Result as byte[];
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Executes the core of the background operation.
		/// </summary>
		/// <returns>Returns the computed MAC.</returns>
		protected override object ExecuteCore()
		{
			if( !File.Exists( FileName ) || (FileUtils.GetFileSize( FileName ) <= 0) )
			{
				throw new ValidationException( "Can't process the file because it's missing or is empty.", null, null
					, new ExceptionDetail( "File name", FileName ) );
			}

			byte[] mac;

			using( FileStream fs = new FileStream( FileName, FileMode.Open, FileAccess.Read, FileShare.Read
				, KernelConsts.FileBufferSize, FileOptions.SequentialScan ) )
			{
				mac = AttestingDocument.Process( fs, LongOperationProgress );
			}

			string message = string.Format( "{0}: File \"{1}\" of size {2} was processed in {3:0.0} seconds.", ActionName
				, FileName, FileUtils.GetFileSize( FileName ), LongOperationProgress.ElapsedTime );
			LogManager.Add( message );

			return mac;
		}

		#endregion Private methods
	}
}
