////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.ComponentModel;
using System.IO;
using Vela.Kernel;
using Vela.StarkDocuments;

namespace Vela.ImBackend
{
	/// <summary>
	/// Window for processing a securing document.
	/// </summary>
	public partial class ProcessSecuringDocumentWindow : BackgroundOperationWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ProcessSecuringDocumentWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="actionName">The name of the executed action.</param>
		/// <param name="sourceFileName">The name of the file to process.</param>
		/// <param name="destinationFileName">The name of the file resulted after processing.</param>
		/// <param name="securingDocument">The associated securing document.</param>
		/// <param name="isEncrypt">If this is true then the file is encrypted, else it's decrypted.</param>
		public ProcessSecuringDocumentWindow( string actionName, string sourceFileName, string destinationFileName
			, SecuringDocument securingDocument, bool isEncrypt )
		{
			ActionName = actionName;
			SourceFileName = sourceFileName;
			DestinationFileName = destinationFileName;
			SecuringDocument = securingDocument;
			IsEncrypt = isEncrypt;

			InitializeComponent();

			LoadUi();
		}

		#endregion Constructors

		#region Public non-serializable properties

		private string _ActionName;
		/// <summary>
		/// The name of the executed action. Notifies property observers. READ-ONLY.
		/// </summary>
		public string ActionName
		{
			get { return _ActionName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "ActionName" );

				_ActionName = value;

				// Notify property observers.
				OnPropertyChanged( "ActionName" );
			}
		}

		private string _SourceFileName;
		/// <summary>
		/// The name of the file to process. Notifies property observers. READ-ONLY.
		/// </summary>
		public string SourceFileName
		{
			get { return _SourceFileName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "SourceFileName" );

				_SourceFileName = value;

				// Notify property observers.
				OnPropertyChanged( "SourceFileName" );
			}
		}

		private string _DestinationFileName;
		/// <summary>
		/// The name of the file resulted after processing. Notifies property observers. READ-ONLY.
		/// </summary>
		public string DestinationFileName
		{
			get { return _DestinationFileName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "DestinationFileName" );

				_DestinationFileName = value;

				// Notify property observers.
				OnPropertyChanged( "DestinationFileName" );
			}
		}

		private SecuringDocument _SecuringDocument;
		/// <summary>
		/// The associated securing document. READ-ONLY.
		/// </summary>
		public SecuringDocument SecuringDocument
		{
			get { return _SecuringDocument; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "SecuringDocument" );

				_SecuringDocument = value;
			}
		}

		private bool _IsEncrypt;
		/// <summary>
		/// If this is true then the file is encrypted, else it's decrypted. READ-ONLY.
		/// </summary>
		public bool IsEncrypt
		{
			get { return _IsEncrypt; }
			protected set { _IsEncrypt = value; }
		}

		private byte[] _MacValue;
		/// <summary>
		/// The computed MAC. Notifies property observers. READ-ONLY.
		/// </summary>
		public byte[] MacValue
		{
			get { return _MacValue; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "MacValue" );

				_MacValue = value;

				// Notify property observers.
				OnPropertyChanged( "MacValue" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		/// <summary>
		/// Method called when the background operation was successful.
		/// </summary>
		/// <param name="e">Information about the state of the background operation.</param>
		protected override void BackgroundOperationSuccessful( RunWorkerCompletedEventArgs e )
		{
			ValidationUtils.ObjectHasContent( e, "e" );

			if( e.Result == null ) return;

			MacValue = e.Result as byte[];
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Executes the core of the background operation.
		/// </summary>
		/// <returns>Returns the computed MAC.</returns>
		protected override object ExecuteCore()
		{
			if( !File.Exists( SourceFileName ) || (FileUtils.GetFileSize( SourceFileName ) <= 0) )
			{
				throw new ValidationException( "Can't process the file because it's missing or is empty.", null, null
					, new ExceptionDetail( "File name", SourceFileName ) );
			}

			byte[] mac;

			using( FileStream sfs = new FileStream( SourceFileName, FileMode.Open, FileAccess.Read, FileShare.Read
				, KernelConsts.FileBufferSize, FileOptions.SequentialScan ) )
			using( FileStream dfs = new FileStream( DestinationFileName, FileMode.Create, FileAccess.Write, FileShare.None
				, KernelConsts.FileBufferSize, FileOptions.SequentialScan ) )
			{
				mac = SecuringDocument.Process( sfs, dfs, IsEncrypt, LongOperationProgress );
			}

			string message = string.Format( "{0}: File \"{1}\" of size {2} was processed in {3:0.0} seconds.", ActionName
				, SourceFileName, FileUtils.GetFileSize( SourceFileName ), LongOperationProgress.ElapsedTime );
			LogManager.Add( message );

			return mac;
		}

		#endregion Private methods
	}
}
