////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows;
using Vela.ActionProcessor;
using Vela.Controls;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.PeerDocuments;

namespace Vela.ImBackend
{
	/// <summary>
	/// Window for adding an account using the UPN of another peer.
	/// </summary>
	public partial class QueueCreateAccountWithUpnWindow : ViewWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public QueueCreateAccountWithUpnWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="accountRequestToView">Account creation request to view.</param>
		/// <param name="accountRequestProgress">The progress of the account creation request.</param>
		public QueueCreateAccountWithUpnWindow( CreateAccountAnonymousActionRequest accountRequestToView
			, CreateAccountClientProgress accountRequestProgress )
		{
			ValidationUtils.ObjectHasContent( accountRequestToView, "accountRequestToView" );
			ValidationUtils.ObjectHasContent( accountRequestProgress, "accountRequestProgress" );

			ReadOnly = true;

			UpnText = accountRequestToView.RecipientTextualUpn;
			TallyEntityName = accountRequestProgress.TallyEntityName;
			Approval = accountRequestProgress.Approval;
			AlternativePeerAddress = accountRequestProgress.AlternativePeerAddress;

			InitializeComponent();

			LoadUi();

			ControlsUtils.MakeReadOnly( this );
		}

		#endregion Constructors

		#region Public non-serializable properties

		private bool _ReadOnly = false;
		/// <summary>
		/// If this is true the information is shown (in the window) in read-only mode. READ-ONLY.
		/// </summary>
		public bool ReadOnly
		{
			get { return _ReadOnly; }
			private set { _ReadOnly = value; }
		}

		private string _UpnText;
		/// <summary>
		/// The textual UPN of the remote peer to add as account. Notifies property observers.
		/// </summary>
		public string UpnText
		{
			get { return _UpnText; }
			set
			{
				ValidationUtils.Text( value, -1, -1, TextValidations.Upn, "UpnText" );

				_UpnText = value;

				// Notify property observers.
				OnPropertyChanged( "UpnText" );
			}
		}

		private Upn _Upn;
		/// <summary>
		/// The UPN of the remote peer to add as account. Notifies property observers. READ-ONLY.
		/// </summary>
		public Upn Upn
		{
			get { return _Upn; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Upn" );

				_Upn = value;

				// Notify property observers.
				OnPropertyChanged( "Upn" );
			}
		}

		private string _TallyEntityName;
		/// <summary>
		/// The name of the identified entity, from the tally. This may be null or empty. Notifies property observers.
		/// </summary>
		/// <remarks>
		/// <para>If a tally for the remote peer already exists, this is ignored and its entity name remains as is.</para>
		/// </remarks>
		public string TallyEntityName
		{
			get { return _TallyEntityName; }
			set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.EntityName, "TallyEntityName" );

				_TallyEntityName = value;

				// Notify property observers.
				OnPropertyChanged( "TallyEntityName" );
			}
		}

		private TallyApprovals _Approval = PeerDocumentsConsts.CreateAccountApprovalMin;
		/// <summary>
		/// The approval of the tally. Notifies property observers.
		/// </summary>
		/// <remarks>
		/// <para>If a tally for the remote peer already exists, this is ignored and its approval remains as is.</para>
		/// </remarks>
		public TallyApprovals Approval
		{
			get { return _Approval; }
			set
			{
				_Approval = value;

				// Notify property observers.
				OnPropertyChanged( "Approval" );
			}
		}

		private AlternativePeerAddress _AlternativePeerAddress = new AlternativePeerAddress();
		/// <summary>
		/// The alternative peer address (of the remote peer). Notifies property observers. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If a tally for the remote peer already exists, this is ignored if it's null, and the tally's alternative peer
		/// address remains as is.</para>
		/// </remarks>
		public AlternativePeerAddress AlternativePeerAddress
		{
			get { return _AlternativePeerAddress; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "AlternativePeerAddress" );

				_AlternativePeerAddress = value;

				// Notify property observers.
				OnPropertyChanged( "AlternativePeerAddress" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void ViewWindow_Loaded( object sender, RoutedEventArgs e )
		{
			ControlsUtils.FocusTextBox( UpnOfContactTextBox, !ReadOnly );
		}

		private void OkayButton_Click( object sender, RoutedEventArgs e )
		{
			try
			{
				ValidateUi( ErrorPresenter );
				SaveUi( ErrorPresenter );

				DialogResult = true;
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				ControlsUtils.ShowException( ex );
			}
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Validates the input data.
		/// </summary>
		/// <param name="errorPresenter">The control which shows that there is a validation error. This may be null.</param>
		/// <remarks>
		/// <para>If you don't validate of all the editable fields in here, it's possible that some bindings are commited even
		/// if the user cancels the edit operation. This can happen if the user edits some fields, clicks the "Okay" button, the
		/// validation fails (during binding commit) inside some properties (which are not validated in here), and then the user
		/// clicks the "Cancel" button. This is likely to matter only if those (successfully) bound properties are used outside
		/// the view.</para>
		/// </remarks>
		protected override void ValidateUi( FrameworkElement errorPresenter )
		{
			base.ValidateUi( errorPresenter );

			ValidationUtils.Text( UpnOfContactTextBox.Text, -1, -1, TextValidations.Upn, "UPN of contact", UpnOfContactTextBox
				, errorPresenter );
			ValidationUtils.Text( TallyEntityNameTextBox.Text, -2, -1, TextValidations.EntityName, "Tally entity name"
				, TallyEntityNameTextBox, errorPresenter );
			ValidationUtils.Text( AlternativePeerAddressTextBox.Text, -2, -1, TextValidations.AlternativePeerAddress
				, "Alternative peer address", AlternativePeerAddressTextBox, errorPresenter );
			ValidationUtils.Integer( OverridingProcessingPortTextBox.Text, 0, UInt16.MaxValue, "OverridingProcessingPort" );
		}

		/// <summary>
		/// Saves the user interface data into the data to edit.
		/// </summary>
		/// <param name="errorPresenter">The control which shows that there is a validation error. This may be null.</param>
		protected override void SaveUi( FrameworkElement errorPresenter )
		{
			base.SaveUi( errorPresenter );

			try
			{
				UpnAlg upnAlg = CoreDocumentsUtils.ExtractUpnAlg( UpnOfContactTextBox.Text );
				Upn = upnAlg.Parse( UpnOfContactTextBox.Text );
			}
			catch( Exception ex )
			{
				throw new ValidationException( "The UPN is invalid. Make sure that you and your contacts are using the"
					+ " newest version of the application.", ex );
			}
		}

		#endregion Private methods
	}
}
