////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.ComponentModel;
using Vela.ApplicationBackend;
using Vela.Kernel;
using Vela.PeerDocuments;

namespace Vela.ImBackend
{
	/// <summary>
	/// Window for updating all the letters which have a specified flag.
	/// </summary>
	public partial class UpdateLettersFlagWindow : BackgroundOperationWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected UpdateLettersFlagWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		/// <param name="flagToUpdate">The flag to update.</param>
		/// <param name="newFlag">The new flag.</param>
		public UpdateLettersFlagWindow( ApplicationModel applicationModel, string flagToUpdate, string newFlag )
			: base( applicationModel )
		{
			FlagToUpdate = flagToUpdate;
			NewFlag = newFlag;

			InitializeComponent();

			LoadUi();
		}

		#endregion Constructors

		#region Public non-serializable properties

		private string _FlagToUpdate;
		/// <summary>
		/// The flag to update. Notifies property observers. READ-ONLY.
		/// </summary>
		public string FlagToUpdate
		{
			get { return _FlagToUpdate; }
			protected set
			{
				_FlagToUpdate = value;

				// Notify property observers.
				OnPropertyChanged( "FlagToUpdate" );
			}
		}

		private string _NewFlag;
		/// <summary>
		/// The new flag. Notifies property observers. READ-ONLY.
		/// </summary>
		public string NewFlag
		{
			get { return _NewFlag; }
			protected set
			{
				_NewFlag = value;

				// Notify property observers.
				OnPropertyChanged( "NewFlag" );
			}
		}

		private int _UpdatedCount;
		/// <summary>
		/// The number of letters whose flags were updated. Notifies property observers. READ-ONLY.
		/// </summary>
		public int UpdatedCount
		{
			get { return _UpdatedCount; }
			protected set
			{
				_UpdatedCount = value;

				// Notify property observers.
				OnPropertyChanged( "UpdatedCount" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		/// <summary>
		/// Method called when the background operation was successful.
		/// </summary>
		/// <param name="e">Information about the state of the background operation.</param>
		protected override void BackgroundOperationSuccessful( RunWorkerCompletedEventArgs e )
		{
			ValidationUtils.ObjectHasContent( e, "e" );

			if( e.Result == null ) return;

			UpdatedCount = (int)e.Result;
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Executes the core of the background operation.
		/// </summary>
		/// <returns>Returns the number of letters whose flags were updated.</returns>
		protected override object ExecuteCore()
		{
			int updated = 0;

			try
			{
				StorehouseModel sm = StorehouseModel;

				ApplicationBackendStorehouse.UpdatePostsFlag( sm.StorehouseProvider, FlagToUpdate, NewFlag
					, typeof( Letter ), LongOperationProgress, out updated );
			}
			catch( AbortLongOperationException )
			{
				// Nothing to do.
			}

			return updated;
		}

		#endregion Private methods
	}
}
