////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Animation;
using Vela.Kernel;

namespace Vela.Controls
{
	/// <summary>
	/// An animated image.
	/// </summary>
	/// <remarks>
	/// <para>All the frames of an animated image should have the same size; otherwise, if the image control doesn't have a
	/// fixed size, the layout would change at every frame.</para>
	/// </remarks>
	public class AnimatedImage : Image
	{
		#region Public constants

		/// <summary>
		/// The name of the dependency property for the <see cref="FullSizedPreview"/> property.
		/// </summary>
		public const string FullSizedPreviewPropertyName = "FullSizedPreview";

		/// <summary>
		/// The name of the dependency property for the <see cref="SourceUrl"/> property.
		/// </summary>
		public const string SourceUrlPropertyName = "SourceUrl";

		/// <summary>
		/// The name of the dependency property for the <see cref="SourceImage"/> property.
		/// </summary>
		public const string SourceImagePropertyName = "SourceImage";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public AnimatedImage()
		{
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The animation manager.
		/// </summary>
		private ObjectAnimationUsingKeyFrames Animation;

		#endregion Private fields

		#region Dependency properties

		/// <summary>
		/// Dependency property for the <see cref="FullSizedPreview"/> property.
		/// </summary>
		public static readonly DependencyProperty FullSizedPreviewProperty = DependencyProperty.Register(
			FullSizedPreviewPropertyName, typeof( bool ), typeof( AnimatedImage )
			, new FrameworkPropertyMetadata( false, FrameworkPropertyMetadataOptions.AffectsRender
				| FrameworkPropertyMetadataOptions.AffectsMeasure ) );

		/// <summary>
		/// Dependency property for the <see cref="SourceUrl"/> property.
		/// </summary>
		public static readonly DependencyProperty SourceUrlProperty = DependencyProperty.Register( SourceUrlPropertyName
			, typeof( string ), typeof( AnimatedImage )
			, new FrameworkPropertyMetadata( null, FrameworkPropertyMetadataOptions.AffectsRender
				| FrameworkPropertyMetadataOptions.AffectsMeasure, OnSourceUrlChanged ) );

		/// <summary>
		/// Dependency property for the <see cref="SourceImage"/> property.
		/// </summary>
		public static readonly DependencyProperty SourceImageProperty = DependencyProperty.Register( SourceImagePropertyName
			, typeof( ObjectAnimationUsingKeyFrames ), typeof( AnimatedImage )
			, new FrameworkPropertyMetadata( null, FrameworkPropertyMetadataOptions.AffectsRender
				| FrameworkPropertyMetadataOptions.AffectsMeasure, OnSourceImageChanged ) );

		#endregion Dependency properties

		#region Public non-serializable properties

		/// <summary>
		/// The full-sized preview indicator of the animated image.
		/// </summary>
		public bool FullSizedPreview
		{
			get { return (bool)GetValue( FullSizedPreviewProperty ); }
			set { SetValue( FullSizedPreviewProperty, value ); }
		}

		/// <summary>
		/// The source URL of the animated image.
		/// </summary>
		public string SourceUrl
		{
			get { return (string)GetValue( SourceUrlProperty ); }
			set { SetValue( SourceUrlProperty, value ); }
		}

		/// <summary>
		/// The source image of the animated image.
		/// </summary>
		public ObjectAnimationUsingKeyFrames SourceImage
		{
			get { return (ObjectAnimationUsingKeyFrames)GetValue( SourceImageProperty ); }
			set { SetValue( SourceImageProperty, value ); }
		}

		#endregion Public non-serializable properties

		#region Private static methods

		/// <summary>
		/// Event raised when the animation's source URL has changed.
		/// </summary>
		/// <param name="dp">The dependency object for which the value of the associated dependency property has changed. This
		/// may be null.</param>
		/// <param name="e">Event data.</param>
		private static void OnSourceUrlChanged( DependencyObject dp, DependencyPropertyChangedEventArgs e )
		{
			ValidationUtils.ObjectHasContent( e, "e" );

			AnimatedImage ai = dp as AnimatedImage;
			if( ai == null )
				return;

			if( (e.NewValue == null) || (e.NewValue is string) )
				ai.PrepareAnimation( e.NewValue as string );
		}

		/// <summary>
		/// Event raised when the animation's source image has changed.
		/// </summary>
		/// <param name="dp">The dependency object for which the value of the associated dependency property has changed. This
		/// may be null.</param>
		/// <param name="e">Event data.</param>
		private static void OnSourceImageChanged( DependencyObject dp, DependencyPropertyChangedEventArgs e )
		{
			ValidationUtils.ObjectHasContent( e, "e" );

			AnimatedImage ai = dp as AnimatedImage;
			if( ai == null )
				return;

			if( (e.NewValue == null) || (e.NewValue is ObjectAnimationUsingKeyFrames) )
				ai.PrepareAnimation( e.NewValue as ObjectAnimationUsingKeyFrames );
		}

		#endregion Private static methods

		#region Private methods

		/// <summary>
		/// Clears the animation.
		/// </summary>
		private void ClearAnimation()
		{
			// Stop the animation.
			if( Animation != null )
				BeginAnimation( Image.SourceProperty, null );

			Animation = null;
			base.Source = null;
		}

		/// <summary>
		/// Prepares the animation.
		/// </summary>
		/// <param name="fileName">The name of the file of the animated image. This may be null or empty.</param>
		private void PrepareAnimation( string fileName )
		{
			ClearAnimation();

			if( TextUtils.NoContent( fileName ) )
				return;

			ImagePreview imagePreview = ImagePreviewsCache.GetCachedOrLoad( fileName, fileName, FullSizedPreview );
			if( imagePreview == null )
			{
				ClearAnimation();

				return;
			}

			Animation = imagePreview.Image;

			if( (Animation == null) || (Animation.KeyFrames.Count <= 0) )
			{
				ClearAnimation();

				return;
			}

			base.Source = (ImageSource)Animation.KeyFrames[ 0 ].Value;

			if( Animation.KeyFrames.Count > 1 )
				BeginAnimation( Image.SourceProperty, Animation );
		}

		/// <summary>
		/// Prepares the animation.
		/// </summary>
		/// <param name="image">The animated image. This may be null.</param>
		private void PrepareAnimation( ObjectAnimationUsingKeyFrames image )
		{
			ClearAnimation();

			if( (image == null) || (image.KeyFrames.Count <= 0) )
				return;

			Animation = image;

			base.Source = (ImageSource)Animation.KeyFrames[ 0 ].Value;

			if( Animation.KeyFrames.Count > 1 )
				BeginAnimation( Image.SourceProperty, Animation );
		}

		#endregion Private methods
	}
}
