////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Microsoft.Win32;
using System;
using System.Collections;
using System.Globalization;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Markup;
using System.Windows.Media;
using System.Windows.Threading;
using Vela.Crypto;
using Vela.Kernel;
using Vela.Resources;

namespace Vela.Controls
{
	/// <summary>
	/// Utilities for working with controls.
	/// </summary>
	public static class ControlsUtils
	{
		#region Public constants

		/// <summary>
		/// If up to this number of pixels from the bottom of a list-box is visible to the user, it's considered that the
		/// list-box is scrolled near the bottom.
		/// </summary>
		public const int ScrolledNearBottomPixelsCount = 100;

		#endregion Public constants

		#region Private static methods

		/// <summary>
		/// Gets the color from the specified text. No exception is thrown if a color can't be retrieved from the specified text.
		/// </summary>
		/// <param name="colorText">Text from which to get the color.</param>
		/// <param name="defaultColor">Color to return if no color could be retrieved from the specified text.</param>
		/// <returns>Returns the color retrieved from the specified text.</returns>
		public static Color GetColorFromText( string colorText, Color defaultColor )
		{
			if( TextUtils.NoContent( colorText ) )
				return defaultColor;

			object color = ColorConverter.ConvertFromString( colorText );

			return (color != null) ? (Color)color : defaultColor;
		}

		/// <summary>
		/// Gets the color from the specified text.
		/// </summary>
		/// <param name="colorText">Text from which to get the color.</param>
		/// <returns>Returns the color retrieved from the specified text.</returns>
		public static Color GetColorFromText( string colorText )
		{
			ValidationUtils.TextHasContent( colorText, "colorText" );

			object color = ColorConverter.ConvertFromString( colorText );
			ValidationUtils.ObjectHasContent( color, "colorText" );

			return (Color)color;
		}

		/// <summary>
		/// Makes all the children of the specified control read-only.
		/// </summary>
		/// <param name="control">Control whose children to make read-only. This may be null or empty.</param>
		public static void MakeReadOnly( FrameworkElement control )
		{
			if( control == null )
				return;

			// Make the control read-only.
			if( control as TextBoxBase != null )
			{
				((TextBoxBase)control).IsReadOnly = true;
			}
			else if( (control is ComboBox) || (control is CheckBox) || (control is DatePicker) )
			{
				control.IsHitTestVisible = false;
				control.Focusable = false;
			}

			// Iterate through children.
			IEnumerable children = LogicalTreeHelper.GetChildren( control );
			if( children == null )
				return;
			foreach( object child in children )
			{
				// Recurse children.
				if( child as FrameworkElement != null )
					MakeReadOnly( child as FrameworkElement );
			}
		}

		/// <summary>
		/// Gets the dependency property of the writable content of the specified control.
		/// </summary>
		/// <param name="control">The control for which to get the dependency property.</param>
		/// <returns>Returns the content dependency property.</returns>
		private static DependencyProperty GetWritableDependencyProperty( FrameworkElement control )
		{
			ValidationUtils.ObjectHasContent( control, "control" );

			// First check the derived types, then the root ones.
			if( control is TextBox )
				return TextBox.TextProperty;
			else if( control is ComboBox )
				return Selector.SelectedItemProperty;
			else if( control is CheckBox )
				return ToggleButton.IsCheckedProperty;
			else if( control is DatePicker )
				return DatePicker.SelectedDateProperty;
			else
				return null;
		}

		/// <summary>
		/// Gets the dependency property of the read-only content of the specified control.
		/// </summary>
		/// <param name="control">The control for which to get the dependency property.</param>
		/// <returns>Returns the content dependency property.</returns>
		private static DependencyProperty GetReadOnlyDependencyProperty( FrameworkElement control )
		{
			ValidationUtils.ObjectHasContent( control, "control" );

			// First check the derived types, then the root ones.
			if( control is TextBlock )
				return TextBlock.TextProperty;
			else if( control is Label )
				return ContentControl.ContentProperty;
			else if( control is ItemsControl )
				return ItemsControl.ItemsSourceProperty;
			else if( control is AnimatedImage )
				return AnimatedImage.SourceImageProperty;
			else if( control is Image )
				return Image.SourceProperty;
			else
				return null;
		}

		/// <summary>
		/// Marks as invalid the binding of the error control from the specified exception.
		/// </summary>
		/// <param name="exception">Exception.</param>
		private static void MarkInvalidBinding( PresentedValidationException exception )
		{
			ValidationUtils.ObjectHasContent( exception, "exception" );

			FrameworkElement errorControl = exception.ErrorControl as FrameworkElement;
			if( errorControl == null )
				return;

			DependencyProperty dp = GetWritableDependencyProperty( errorControl );
			if( dp == null )
				return;

			BindingExpression be = errorControl.GetBindingExpression( dp );
			if( (be == null) || (be.ParentBinding == null) )
				return;

			ExceptionValidationRule vr = new ExceptionValidationRule();
			ValidationError ve = new ValidationError( vr, be.ParentBinding );
			ve.Exception = exception;

			// Mark invalid binding.
			Validation.MarkInvalid( be, ve );
		}

		/// <summary>
		/// Clears the specified binding of errors.
		/// </summary>
		/// <param name="bindingExpression">Binding expression to clear.</param>
		private static void ClearInvalidBinding( BindingExpression bindingExpression )
		{
			ValidationUtils.ObjectHasContent( bindingExpression, "bindingExpression" );

			Validation.ClearInvalid( bindingExpression );
		}

		/// <summary>
		/// Clears the specified control of binding errors.
		/// </summary>
		/// <param name="control">The control whose bindings to clear of errors. This may be null.</param>
		private static void ClearInvalidBinding( FrameworkElement control )
		{
			if( control == null )
				return;

			DependencyProperty dp = GetWritableDependencyProperty( control );
			if( dp == null )
				return;

			BindingExpression be = control.GetBindingExpression( dp );
			if( (be == null) || (be.ParentBinding == null) )
				return;

			Validation.ClearInvalid( be );
		}

		/// <summary>
		/// Clears the error bindings of all the children of the specified control.
		/// </summary>
		/// <param name="control">The control whose bindings to clear of errors. The children of the control are recursed. This
		/// may be null.</param>
		private static void ClearInvalidBindings( FrameworkElement control )
		{
			if( control == null )
				return;

			// Clear the binding error.
			ClearInvalidBinding( control );

			// Iterate through children.
			IEnumerable children = LogicalTreeHelper.GetChildren( control );
			if( children == null )
				return;
			foreach( object child in children )
			{
				// Recurse children.
				if( child as FrameworkElement != null )
					ClearInvalidBindings( child as FrameworkElement );
			}
		}

		/// <summary>
		/// Updates the source of the content binding of the specified control.
		/// </summary>
		/// <param name="control">The control whose binding source to update. This may be null.</param>
		/// <param name="errorPresenter">The control which shows that there is a validation error.</param>
		private static void UpdateBindingSource( FrameworkElement control, FrameworkElement errorPresenter )
		{
			ValidationUtils.ObjectHasContent( errorPresenter, "errorPresenter" );

			if( control == null )
				return;

			DependencyProperty dp = GetWritableDependencyProperty( control );
			if( dp == null )
				return;

			BindingExpression be = control.GetBindingExpression( dp );
			if( (be == null) || (be.ParentBinding == null) )
				return;

			// Check "UpdateSourceTrigger".
			if( (be.ParentBinding.UpdateSourceTrigger != UpdateSourceTrigger.Explicit)
				&& (be.ParentBinding.Mode == BindingMode.TwoWay) )
			{
				control.Focus();

				string message = string.Format( "The binding with path \"{0}\" has its \"UpdateSourceTrigger\" property set to"
					+ " a value different than \"Explicit\" and \"OneWay\" and will therefore update its data source before"
					+ " the user clicks the \"Commit\" button.", be.ParentBinding.Path.Path );

				throw new ValidationException( message );
			}

			if( be.ParentBinding.Mode != BindingMode.TwoWay )
				return;

			// Check "ValidatesOnExceptions".
			if( !be.ParentBinding.ValidatesOnExceptions )
			{
				control.Focus();

				string message = string.Format( "The binding with path \"{0}\" has its \"ValidatesOnExceptions\" property set"
						+ " to false and will therefore swallow any exception which may occur while its data source is updated."
					, be.ParentBinding.Path.Path );

				throw new ValidationException( message );
			}

			// Update the source.
			be.UpdateSource();

			// Check for errors.
			if( be.HasError && (be.ValidationError != null) )
			{
				// The default error presentation mechanism from WPF is (also) used. This consists from an error border around
				// the controls with validation errors.

				// Repackage exception.
				throw new PresentedValidationException( be.ValidationError.Exception.Message, be.ValidationError.Exception
					, be.ParentBinding.Path.Path, control, errorPresenter );
			}
		}

		/// <summary>
		/// Updates the target of the content binding of the specified control.
		/// </summary>
		/// <param name="control">The control whose binding target to update. This may be null.</param>
		private static void UpdateBindingTarget( FrameworkElement control )
		{
			if( control == null )
				return;

			DependencyProperty dp = GetReadOnlyDependencyProperty( control );
			if( dp == null )
				return;

			BindingExpressionBase be = control.GetBindingExpression( dp );
			if( be == null )
			{
				be = BindingOperations.GetMultiBindingExpression( control, dp );

				if( be == null )
					return;
			}

			// Update the target.
			be.UpdateTarget();
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Adds entropy from the specified mouse data.
		/// </summary>
		/// <param name="mouseData">Mouse data from which to add entropy.</param>
		public static void AddMouseEntropy( MouseEventArgs mouseData )
		{
			ValidationUtils.ObjectHasContent( mouseData, "mouseData" );
			ValidationUtils.ObjectType<UIElement>( mouseData.Source, "mouseData.Source" );

			UIElement eventSource = mouseData.Source as UIElement;
			Point eventSourcePosition = mouseData.GetPosition( eventSource );
			Point screenPosition = eventSource.PointToScreen( eventSourcePosition );

			byte[] buffer = BitConverter.GetBytes( unchecked((int)screenPosition.X + ((int)screenPosition.Y << 16)) );

			EntropyPool.Add( buffer );
		}

		/// <summary>
		/// Adds entropy from the specified keyboard data.
		/// </summary>
		/// <param name="keyboardData">Keyboard data from which to add entropy.</param>
		public static void AddKeyboardEntropy( KeyEventArgs keyboardData )
		{
			ValidationUtils.ObjectHasContent( keyboardData, "keyboardData" );

			byte[] buffer = BitConverter.GetBytes( (int)keyboardData.Key );

			EntropyPool.Add( buffer );
		}

		/// <summary>
		/// Converts the specified keyboard key to an alphanumeric character.
		/// </summary>
		/// <param name="key">Key to convert.</param>
		/// <param name="systemKey">The system key to convert.</param>
		/// <returns>Returns the resulted character.</returns>
		public static char ConvertKeyToAlpha( Key key, Key systemKey )
		{
			if( key == Key.None )
				return char.MinValue;

			if( key == Key.System )
				key = systemKey;

			if( (key >= Key.A) && (key <= Key.Z) )
				return (char)('A' + (key - Key.A));
			if( (key >= Key.D0) && (key <= Key.D9) )
				return (char)('0' + (key - Key.D0));

			return char.MinValue;
		}

		/// <summary>
		/// Sets the application's global UI settings.
		/// </summary>
		public static void SetApplicationGlobalUiSettings()
		{
			// Language.
			XmlLanguage language = XmlLanguage.GetLanguage( CultureInfo.CurrentCulture.IetfLanguageTag );
			FrameworkPropertyMetadata metadata = new FrameworkPropertyMetadata( language );
			FrameworkElement.LanguageProperty.OverrideMetadata( typeof( FrameworkElement ), metadata );

			// Tooltip. List-box items probably use the same tooltip because "TooltipBetweenShowDelay" and "TooltipPlacement"
			// don't (appear to) work.
			ToolTipService.InitialShowDelayProperty.OverrideMetadata( typeof( DependencyObject )
				, new FrameworkPropertyMetadata( ControlsConsts.TooltipInitialShowDelay ) );
			ToolTipService.BetweenShowDelayProperty.OverrideMetadata( typeof( DependencyObject )
				, new FrameworkPropertyMetadata( ControlsConsts.TooltipBetweenShowDelay ) );
			ToolTipService.ShowDurationProperty.OverrideMetadata( typeof( DependencyObject )
				, new FrameworkPropertyMetadata( ControlsConsts.TooltipShowDuration ) );
			ToolTipService.PlacementProperty.OverrideMetadata( typeof( DependencyObject )
				, new FrameworkPropertyMetadata( ControlsConsts.TooltipPlacement ) );
		}

		/// <summary>
		/// Sets the application instance's global settings.
		/// </summary>
		public static void SetApplicationInstanceGlobalSettings( Application appInstance )
		{
			FontFamily normalFontFamily = (FontFamily)appInstance.FindResource( ResourcesConsts.NormalFontFamilyResourceKey );
			double normalFontSize = (double)appInstance.FindResource( ResourcesConsts.NormalFontSizeResourceKey );

			TextElement.FontFamilyProperty.OverrideMetadata( typeof( TextElement )
				, new FrameworkPropertyMetadata( normalFontFamily ) );
			TextElement.FontSizeProperty.OverrideMetadata( typeof( TextElement )
				, new FrameworkPropertyMetadata( normalFontSize ) );

			TextBlock.FontFamilyProperty.OverrideMetadata( typeof( TextBlock )
				, new FrameworkPropertyMetadata( normalFontFamily ) );
			TextBlock.FontSizeProperty.OverrideMetadata( typeof( TextBlock )
				, new FrameworkPropertyMetadata( normalFontSize ) );
		}

		/// <summary>
		/// Gets the application's dispatcher.
		/// </summary>
		/// <returns>Returns the application's dispatcher. This may be null.</returns>
		public static Dispatcher GetAppDispatcher()
		{
			Application app = Application.Current;
			if( app == null )
				return null;

			return app.Dispatcher;
		}

		/// <summary>
		/// Processes the UI events of the application.
		/// </summary>
		public static void ProcessAppUiEvents()
		{
			Dispatcher appDispatcher = GetAppDispatcher();
			if( appDispatcher == null )
				return;

			appDispatcher.Invoke( () => { }, DispatcherPriority.Background );
		}

		/// <summary>
		/// Gets the application's main window.
		/// </summary>
		/// <returns>Returns the application's main window. This may be null.</returns>
		public static Window GetAppMainWindow()
		{
			Application app = Application.Current;
			if( app == null )
				return null;

			return app.MainWindow;
		}

		/// <summary>
		/// Activates a window which is not <see cref="IFloatingWindow"/>, preferably the last used one, or the application's
		/// main window.
		/// </summary>
		/// <returns>Returns the activated window. This may be null.</returns>
		public static Window ActivateLastUsedWindow()
		{
			Application app = Application.Current;
			if( (app == null) || ListUtils.NoContent( app.Windows ) )
				return null;

			// First check if there is a child window.
			int count = app.Windows.Count;
			for( int i = count - 1; i >= 0; i-- )
			{
				Window window = app.Windows[ i ];

				if( window.IsVisible && (window as IFloatingWindow == null) )
				{
					window.Activate();

					return window;
				}
			}

			// Get the application's main window.
			Window mainWindow = GetAppMainWindow();
			if( (mainWindow == null) || !mainWindow.IsVisible )
				return null;

			mainWindow.Activate();

			return mainWindow;
		}

		/// <summary>
		/// Gets the application's visibility.
		/// </summary>
		public static ApplicationVisibilities GetApplicationVisibility()
		{
			Window mainWindow = GetAppMainWindow();
			if( mainWindow == null )
				return ApplicationVisibilities.Hidden;

			if( mainWindow.WindowState == WindowState.Minimized )
				return ApplicationVisibilities.Minimized;

			if( !mainWindow.IsVisible )
				return ApplicationVisibilities.Hidden;

			return ApplicationVisibilities.Normal;
		}

		/// <summary>
		/// Sets the application's visibility with the specified value.
		/// </summary>
		/// <param name="visibility">Application visibility to set.</param>
		public static void SetApplicationVisibility( ApplicationVisibilities visibility )
		{
			Window mainWindow = GetAppMainWindow();
			if( mainWindow == null )
				return;

			try
			{
				switch( visibility )
				{
					case ApplicationVisibilities.Normal:
						mainWindow.WindowState = WindowState.Normal;
						mainWindow.Show();
						mainWindow.ShowInTaskbar = true; // If you put this first, the window will flicker.

						ActivateLastUsedWindow();

						return;
					case ApplicationVisibilities.Minimized:
						// Do not change the order! (If you do, when the tray icon is clicked, the taskbar button is not shown.)
						mainWindow.ShowActivated = false; // Avoid stealing the focus.
						mainWindow.ShowInTaskbar = true;
						mainWindow.WindowState = WindowState.Minimized;
						mainWindow.Show();
						mainWindow.ShowActivated = true;

						TooltipWindow.CloseTooltip();

						return;
					case ApplicationVisibilities.Hidden:
						mainWindow.Hide();
						mainWindow.ShowInTaskbar = false; // If you put this first, the window will flicker.

						TooltipWindow.CloseTooltip();

						return;
					default:
						throw ValidationUtils.Unhandled( visibility, "visibility" );
				}
			}
			catch( InvalidOperationException )
			{
				// Catch "System.InvalidOperationException: Cannot set Visibility to Visible or call Show, ShowDialog, Close, or
				// WindowInteropHelper.EnsureHandle while a Window is closing.".
				//
				// This exception is thrown if the user tries to close the application, and while the confirmation dialog is
				// shown, the user clicks on the application notification icon.
			}
		}

		/// <summary>
		/// Closes the application.
		/// </summary>
		public static void CloseApplication()
		{
			Window mainWindow = GetAppMainWindow();
			if( mainWindow == null )
				return;

			mainWindow.Close();
		}

		/// <summary>
		/// Sets the window's visibility with the specified value.
		/// </summary>
		/// <param name="window">Window whose visibility to set.</param>
		/// <param name="visibility">Window visibility to set.</param>
		public static void SetWindowVisibility( Window window, ApplicationVisibilities visibility )
		{
			if( window == null )
				return;

			try
			{
				switch( visibility )
				{
					case ApplicationVisibilities.Normal:
						window.WindowState = WindowState.Normal;
						window.Show();
						window.ShowInTaskbar = true; // If you put this first, the window will flicker.
						window.Activate();

						return;
					case ApplicationVisibilities.Minimized:
						// Do not change the order! (If you do, when the tray icon is clicked, the taskbar button is not shown.)
						window.ShowActivated = false; // Avoid stealing the focus.
						window.ShowInTaskbar = true;
						window.WindowState = WindowState.Minimized;
						window.Show();
						window.ShowActivated = true;

						return;
					case ApplicationVisibilities.Hidden:
						window.Hide();
						window.ShowInTaskbar = false; // If you put this first, the window will flicker.

						return;
					default:
						throw ValidationUtils.Unhandled( visibility, "visibility" );
				}
			}
			catch( InvalidOperationException )
			{
				// Catch "System.InvalidOperationException: Cannot set Visibility to Visible or call Show, ShowDialog, Close, or
				// WindowInteropHelper.EnsureHandle while a Window is closing.".
			}
		}

		/// <summary>
		/// Centers the specified window on the screen.
		/// </summary>
		/// <param name="window">Window to center. This may be null.</param>
		/// <param name="horizontalShift">The amount to shift the window horizontally.</param>
		/// <param name="verticalShift">The amount to shift the window vertically.</param>
		public static void CenterWindowOnScreen( Window window, int horizontalShift = 0, int verticalShift = 0 )
		{
			if( window == null )
				return;

			Rect workArea = SystemParameters.WorkArea;
			int screenLeft = (int)workArea.Left;
			int screenTop = (int)workArea.Top;
			int screenWidth = (int)workArea.Width;
			int screenHeight = (int)workArea.Height;

			int width = (int)window.ActualWidth;
			int height = (int)window.ActualHeight;

			int left = screenLeft + (screenWidth - width) / 2 + horizontalShift;
			if( left + width >= screenWidth )
				left = screenWidth - width;
			if( left < 0 )
				left = 0;

			int top = screenTop + (screenHeight - height) / 2 + verticalShift;
			if( top + height >= screenHeight )
				top = screenHeight - height;
			if( top < 0 )
				top = 0;

			window.WindowStartupLocation = WindowStartupLocation.Manual;
			window.Left = left;
			window.Top = top;
		}

		/// <summary>
		/// Repositions the specified window's bottom-right corner at the bottom-right corner of the screen.
		/// </summary>
		/// <param name="window">Window to reposition. This may be null.</param>
		/// <param name="horizontalShift">The amount to shift the window horizontally.</param>
		/// <param name="verticalShift">The amount to shift the window vertically.</param>
		public static void RepositionBottomRightToScreen( Window window, int horizontalShift = 0, int verticalShift = 0 )
		{
			if( window == null )
				return;

			Rect workArea = SystemParameters.WorkArea;
			int screenLeft = (int)workArea.Left;
			int screenTop = (int)workArea.Top;
			int screenWidth = (int)workArea.Width;
			int screenHeight = (int)workArea.Height;

			int screenRight = screenLeft + screenWidth - 1;
			int screenBottom = screenTop + screenHeight - 1;

			int width = (int)window.ActualWidth;
			int height = (int)window.ActualHeight;

			int left = screenRight - width - horizontalShift;
			if( left < 0 )
				left = 0;

			int top = screenBottom - height - verticalShift;
			if( top < 0 )
				top = 0;

			window.WindowStartupLocation = WindowStartupLocation.Manual;
			window.Left = left;
			window.Top = top;
		}

		/// <summary>
		/// Sets the owner of the specified window with the application's main window.
		/// </summary>
		/// <param name="window">Windows whose owner to set.</param>
		public static void SetWindowOwnerToMainWindow( Window window )
		{
			ValidationUtils.ObjectHasContent( window, "window" );

			Window appMainWindow = GetAppMainWindow();
			if( (appMainWindow == null) || (appMainWindow == window) )
				return;

			window.Owner = appMainWindow;
		}

		/// <summary>
		/// Repositions the specified window's top-right corner at the top-right corner of the owner. If the owner is missing,
		/// nothing happens.
		/// </summary>
		/// <param name="window">Window to reposition.</param>
		public static void RepositionTopRightToOwner( Window window )
		{
			ValidationUtils.ObjectHasContent( window, "window" );

			if( window.Owner == null )
				return;

			FrameworkElement ownerContent = window.Owner.Content as FrameworkElement;
			if( ownerContent == null )
				return;

			Point relativePoint = new Point( ownerContent.ActualWidth, 0 );
			Point absolutePoint = ownerContent.PointToScreen( relativePoint );
			window.Left = absolutePoint.X - window.Width;
			window.Top = absolutePoint.Y;
		}

		/// <summary>
		/// Shows the specified window as a modal dialog.
		/// </summary>
		/// <param name="window">Window to show.</param>
		/// <returns>Returns true if the user has chosen the "Okay" button.</returns>
		public static bool ShowDialog( Window window )
		{
			ValidationUtils.ObjectHasContent( window, "window" );

			Window appMainWindow = GetAppMainWindow();

			if( (appMainWindow != null) && (appMainWindow != window) )
				window.Owner = appMainWindow;

			return window.ShowDialog() == true;
		}

		/// <summary>
		/// Shows the specified dialog.
		/// </summary>
		/// <param name="dialog">Dialog to show.</param>
		/// <returns>Returns true if the user has chosen the "Okay" button.</returns>
		public static bool ShowDialog( CommonDialog dialog )
		{
			ValidationUtils.ObjectHasContent( dialog, "dialog" );

			return dialog.ShowDialog( GetAppMainWindow() ) == true;
		}

		/// <summary>
		/// Shows a file dialog to save a file.
		/// </summary>
		/// <param name="dialogTitle">The title of the dialog. This may be null or empty.</param>
		/// <param name="defaultFileExtension">The default file extension. This may be null or empty.</param>
		/// <param name="fileNameFilter">The file name filter.</param>
		/// <param name="initialFolder">The initial folder. This may be null or empty.</param>
		/// <param name="initialFileName">The initial file name. This may be null or empty.</param>
		/// <param name="showPromptToOverwrite">If this is true and the selected file already exists, a dialog asks the user
		/// whether he wants to overwrite the file or not.</param>
		/// <param name="selectedFileName">Is set with the selected file name. This may be null or empty.</param>
		/// <returns>Returns true if the user has chosen the "Okay" button.</returns>
		public static bool ShowSaveFileDialog( string dialogTitle, string defaultFileExtension, string fileNameFilter
			, string initialFolder, string initialFileName, bool showPromptToOverwrite, out string selectedFileName )
		{
			ValidationUtils.TextHasContent( fileNameFilter, "fileNameFilter" );

			SaveFileDialog sfd = new SaveFileDialog();

			// If "sfd.DefaultExt" has content, or the selected filter has an extension, the dialog always (= even if
			// "sfd.AddExtension" is false) adds the selected filter extension to the selected file name.

			sfd.Title = dialogTitle;
			sfd.AddExtension = true;
			sfd.DefaultExt = defaultFileExtension;
			sfd.Filter = fileNameFilter;
			sfd.InitialDirectory = initialFolder;
			sfd.FileName = initialFileName;
			sfd.CheckPathExists = true;
			sfd.OverwritePrompt = showPromptToOverwrite;

			if( ShowDialog( sfd ) )
			{
				selectedFileName = sfd.FileName;

				return true;
			}
			else
			{
				selectedFileName = null;

				return false;
			}
		}

		/// <summary>
		/// Clears the exception presenter control.
		/// </summary>
		/// <param name="control">The control to which the error presenter (and the bound controls) is associated, like a user
		/// control or a window. The children of the control are recursed.</param>
		/// <param name="errorPresenter">The control which shows that there is a validation error.</param>
		public static void ClearErrorPresenter( FrameworkElement control, FrameworkElement errorPresenter )
		{
			ValidationUtils.ObjectHasContent( control, "control" );
			ValidationUtils.ObjectHasContent( errorPresenter, "errorPresenter" );

			// Change the appearance of the error presenter control.
			errorPresenter.Tag = null;
			errorPresenter.ToolTip = null;
			errorPresenter.Visibility = Visibility.Hidden;

			// Clear the default error presentation of all children.
			ClearInvalidBindings( control );
		}

		/// <summary>
		/// Shows the specified message and exception.
		/// </summary>
		/// <param name="message">Message to show. This may be null or empty.</param>
		/// <param name="exception">Exception to show. This may be null.</param>
		/// <returns>Returns true if the exception was shown.</returns>
		public static bool ShowException( string message, Exception exception )
		{
			if( !LogManager.Add( message, exception, LogRecordModes.Error, true ) )
				return false;

			if( !ExceptionUtils.ShouldShow( exception ) )
				return false;

			// If there is no application dispatcher, message can't be shown.
			if( GetAppDispatcher() == null )
				return false;

			string messageToShow = ExceptionUtils.GetPresenterMessage( message, exception, false );

			PresentedValidationException pve = exception as PresentedValidationException;
			if( pve != null )
			{
				// Focus the error control.
				FrameworkElement errorControl = pve.ErrorControl as FrameworkElement;
				if( errorControl != null )
					errorControl.Focus();

				// Change the appearance of the error presenter control.
				FrameworkElement errorPresenter = pve.ErrorPresenter as FrameworkElement;
				if( errorPresenter == null )
				{
					throw new ValidationException( string.Format( "The error presenter of the \"{0}\" type is not a \"{1}\""
						+ " type.", typeof( PresentedValidationException ).FullName, typeof( FrameworkElement ).FullName ) );
				}

				errorPresenter.Tag = exception;
				errorPresenter.ToolTip = messageToShow;
				errorPresenter.Visibility = Visibility.Visible;

				// The default error presentation mechanism from WPF is (also) used. This consists from an error border around
				// the controls with validation errors.
				MarkInvalidBinding( pve );
			}
			else
			{
				MessageWindow.ShowDialog( messageToShow, MessageWindowModes.Error );
			}

			return true;
		}

		/// <summary>
		/// Shows the specified exception.
		/// </summary>
		/// <param name="exception">Exception to show.</param>
		/// <returns>Returns true if the exception was shown.</returns>
		public static bool ShowException( Exception exception )
		{
			return ShowException( null, exception );
		}

		/// <summary>
		/// Updates the source of the content binding of all the children of the specified control.
		/// </summary>
		/// <param name="control">The control whose binding source to update. The children of the control are recursed. This may
		/// be null.</param>
		/// <param name="errorPresenter">The control which shows that there is a validation error.</param>
		public static void UpdateBindingSources( FrameworkElement control, FrameworkElement errorPresenter )
		{
			ValidationUtils.ObjectHasContent( errorPresenter, "errorPresenter" );

			if( control == null )
				return;

			// Update the binding source.
			UpdateBindingSource( control, errorPresenter );

			// Iterate through children.
			IEnumerable children = LogicalTreeHelper.GetChildren( control );
			if( children == null )
				return;
			foreach( object child in children )
			{
				// Recurse children.
				if( child as FrameworkElement != null )
					UpdateBindingSources( child as FrameworkElement, errorPresenter );
			}
		}

		/// <summary>
		/// Updates the target of the content binding of all the children of the specified control.
		/// </summary>
		/// <param name="control">The control whose binding target to update. The children of the control are recursed. This may
		/// be null.</param>
		public static void UpdateBindingTargets( FrameworkElement control )
		{
			if( control == null )
				return;

			// Update the binding target.
			UpdateBindingTarget( control );

			// Iterate through children.
			IEnumerable children = LogicalTreeHelper.GetChildren( control );
			if( children == null )
				return;
			foreach( object child in children )
			{
				// Recurse children.
				if( child as FrameworkElement != null )
					UpdateBindingTargets( child as FrameworkElement );
			}
		}

		/// <summary>
		/// Gets the first visual parent with the specified type, of the specified control.
		/// </summary>
		/// <typeparam name="TVisual">The type of the visual parent to find.</typeparam>
		/// <param name="control">Control whose parent to find. This may be null.</param>
		/// <returns>Returns the first visual parent with the specified type. This may be null.</returns>
		public static TVisual FindVisualParent<TVisual>( DependencyObject control )
			where TVisual : Visual
		{
			if( control == null )
				return null;

			DependencyObject visual = VisualTreeHelper.GetParent( control );
			while( visual != null )
			{
				visual = VisualTreeHelper.GetParent( visual );

				TVisual parent = visual as TVisual;
				if( parent != null )
					return parent;
			}

			return null;
		}

		/// <summary>
		/// Gets the first visual child with the specified type, of the specified control.
		/// </summary>
		/// <typeparam name="TVisual">The type of the visual child to find.</typeparam>
		/// <param name="control">Control whose child to find. This may be null.</param>
		/// <returns>Returns the first visual child with the specified type. This may be null.</returns>
		public static TVisual FindVisualChild<TVisual>( Visual control )
			where TVisual : Visual
		{
			if( control == null )
				return null;

			int count = VisualTreeHelper.GetChildrenCount( control );

			for( int i = 0; i < count; i++ )
			{
				Visual visual = VisualTreeHelper.GetChild( control, i ) as Visual;
				if( visual == null )
					continue;

				TVisual child = visual as TVisual;
				if( child == null )
					child = FindVisualChild<TVisual>( visual );

				if( child != null )
					return child;
			}

			return null;
		}

		/// <summary>
		/// Checks if the specified control is visible to the user, even if only partially visible.
		/// </summary>
		/// <param name="control">Control to check if it's visible to the user. This may be null.</param>
		/// <param name="scrollViewer">The scroll viewer in which the specified control is. This may be null.</param>
		/// <returns>Returns true if the specified control is visible to the user.</returns>
		public static bool IsVisibleToTheUser( FrameworkElement control, ScrollViewer scrollViewer )
		{
			if( (control == null) || !control.IsVisible || (scrollViewer == null) || !scrollViewer.IsVisible )
				return false;

			// This method doesn't handle parent controls which fit their content.
			if( (scrollViewer.ActualWidth <= 0) || (scrollViewer.ActualHeight <= 0) )
				return false;

			Rect controlRect = new Rect( 0.0, 0.0, control.ActualWidth, control.ActualHeight );
			Rect relativeToScrollViewerRect = control.TransformToAncestor( scrollViewer ).TransformBounds( controlRect );
			Rect scrollViewerRect = new Rect( 0.0, 0.0, scrollViewer.ActualWidth, scrollViewer.ActualHeight );

			return scrollViewerRect.IntersectsWith( relativeToScrollViewerRect );
		}

		/// <summary>
		/// Passes the specified mouse-down event which has occurred in the specified control, to the control's parent.
		/// </summary>
		/// <typeparam name="TParent">The type of the parent to which to pass the event.</typeparam>
		/// <param name="control">Control in which the specified event has occurred. This may be null.</param>
		/// <param name="e">The event to pass. This may be null.</param>
		public static void PassMouseDownEventToParent<TParent>( DependencyObject control, MouseButtonEventArgs e )
			where TParent : UIElement
		{
			if( (control == null) || (e == null) )
				return;

			TParent parent = FindVisualParent<TParent>( control );
			if( parent == null )
				return;

			MouseButtonEventArgs mbea = new MouseButtonEventArgs( e.MouseDevice, e.Timestamp, e.ChangedButton, e.StylusDevice );
			mbea.RoutedEvent = Mouse.MouseDownEvent;

			parent.RaiseEvent( mbea );
		}

		/// <summary>
		/// Passes the specified key-down event which has occurred in the specified control, to the control's parent.
		/// </summary>
		/// <typeparam name="TParent">The type of the parent to which to pass the event.</typeparam>
		/// <param name="control">Control in which the specified event has occurred. This may be null.</param>
		/// <param name="e">The event to pass. This may be null.</param>
		public static void PassKeyDownEventToParent<TParent>( DependencyObject control, KeyEventArgs e )
			where TParent : UIElement
		{
			if( (control == null) || (e == null) )
				return;

			TParent parent = FindVisualParent<TParent>( control );
			if( parent == null )
				return;

			KeyEventArgs kea = new KeyEventArgs( e.KeyboardDevice, e.InputSource, e.Timestamp, e.Key );
			kea.RoutedEvent = Keyboard.KeyDownEvent;

			parent.RaiseEvent( kea );
		}

		/// <summary>
		/// Focuses the specified text-box and moves the caret at the end of the text.
		/// </summary>
		/// <param name="textBox">Text-box to focus. This may be null.</param>
		/// <param name="moveCaretAtEnd">If this is true then the caret is moved at the end of the text-box.</param>
		public static void FocusTextBox( TextBox textBox, bool moveCaretAtEnd )
		{
			if( (textBox == null) || !textBox.IsEnabled || !textBox.IsVisible )
				return;

			textBox.Focus();

			if( moveCaretAtEnd )
			{
				textBox.CaretIndex = textBox.Text.Length;
				textBox.ScrollToEnd();
			}
		}

		/// <summary>
		/// Checks if adding the specified text in the specified text-box should be blocked, and shows a message if this is true.
		/// </summary>
		/// <param name="textBox">Text-box in which text is about to be added.</param>
		/// <param name="textToAdd">Text to add to the text-box. This may be null or empty.</param>
		/// <param name="selectionWillBeReplaced">If this is true then the selected text will be replaced, so its size must be
		/// subtracted.</param>
		/// <returns>Returns true if the adding should be blocked.</returns>
		public static bool CheckBlockAddingTextIfTooLong( TextBox textBox, string textToAdd, bool selectionWillBeReplaced )
		{
			ValidationUtils.ObjectHasContent( textBox, "textBox" );

			if( TextUtils.NoContent( textToAdd ) )
				return false;

			int selectedTextSize = selectionWillBeReplaced ? textBox.SelectedText.Length : 0;
			int newTextSize = (textBox.Text.Length - selectedTextSize) + textToAdd.Length;

			if( (textBox.MaxLength != 0) && (newTextSize > textBox.MaxLength) )
			{
				TooltipWindow.ShowTooltip( string.Format( "The text-box can contain maximum {0} characters, not {1}."
					, textBox.MaxLength, newTextSize ) );

				return true;
			}

			return false;
		}

		/// <summary>
		/// Appends the specified text at the end of the (text from the) specified text-box.
		/// </summary>
		/// <param name="textBox">Text-box in which to append text.</param>
		/// <param name="textToAppend">Text to append. If this is null or empty, nothing happens.</param>
		/// <returns>Returns true if the appending of the text was blocked.</returns>
		/// <remarks>
		/// <para>The caret remains at its initial position.</para>
		/// </remarks>
		public static void AppendText( TextBox textBox, string textToAppend )
		{
			ValidationUtils.ObjectHasContent( textBox, "textBox" );

			if( TextUtils.NoContent( textToAppend ) || CheckBlockAddingTextIfTooLong( textBox, textToAppend, false ) )
				return;

			textBox.AppendText( textToAppend );
		}

		/// <summary>
		/// Inserts the specified text in the specified text-box.
		/// </summary>
		/// <param name="textBox">Text-box in which to insert text.</param>
		/// <param name="textToInsert">Text to insert. If this is null or empty, nothing happens.</param>
		/// <returns>Returns true if the inserting of the text was blocked.</returns>
		/// <remarks>
		/// <para>The caret is moved at the end of the inserted text.</para>
		/// </remarks>
		public static void InsertText( TextBox textBox, string textToInsert )
		{
			ValidationUtils.ObjectHasContent( textBox, "textBox" );

			if( TextUtils.NoContent( textToInsert ) || CheckBlockAddingTextIfTooLong( textBox, textToInsert, true ) )
				return;

			textBox.SelectedText = textToInsert;
			textBox.CaretIndex += textToInsert.Length;
			textBox.SelectionLength = 0;
		}

		/// <summary>
		/// Blocks the current paste event if the specified text-box can't contain all the text to paste.
		/// </summary>
		/// <param name="textBox">Text-box in which text is about to be pasted.</param>
		/// <param name="e">Event data.</param>
		/// <returns>Returns true if the pasting of the text was blocked.</returns>
		public static bool BlockPasteIfTooLong( TextBox textBox, DataObjectPastingEventArgs e )
		{
			ValidationUtils.ObjectHasContent( textBox, "textBox" );
			ValidationUtils.ObjectHasContent( e, "eventArg" );

			string textToPaste = e.DataObject.GetData( typeof( string ) ) as string;

			bool blockAddingTextIfTooLong = CheckBlockAddingTextIfTooLong( textBox, textToPaste, true );

			if( blockAddingTextIfTooLong )
				e.CancelCommand();

			return blockAddingTextIfTooLong;
		}

		/// <summary>
		/// Sets the position of the caret in the specified text-box, at the position of the mouse from the specified mouse
		/// event data, if there is no text selected in the text-box.
		/// </summary>
		/// <param name="textBox">Text-box in which to set the caret position. If this is null, nothing happens.</param>
		/// <param name="e">Event data. If this is null, nothing happens.</param>
		/// <remarks>
		/// <para>If the <see cref="MouseButtonEventArgs.ClickCount"/> is not 1, nothing happens. This is due to an issue with
		/// WPF's event previewing. For example, if the text-box is double-clicked and a window is opened, WPF raises a
		/// single-click event before the window is opened, and a double-click event after the window is closed. However, if at
		/// this point this method is called (in the event handler), the <see cref="TextBox.GetCharacterIndexFromPoint"/> method
		/// raises an exception with the message "The specified Visual is not a descendant of this Visual".</para>
		/// </remarks>
		public static void SetCaretAtClick( TextBox textBox, MouseButtonEventArgs e )
		{
			if( (textBox == null) || (e == null) || (textBox.SelectionLength > 0) || (e.ClickCount != 1) )
				return;

			Point mousePoint = e.GetPosition( textBox );
			int index = textBox.GetCharacterIndexFromPoint( mousePoint, true );
			textBox.CaretIndex = index;
		}

		/// <summary>
		/// Focuses the specified list-box item and manually synchronizes the focused item with the selected item (because the
		/// focused item can't be bound as can the selected item), and scrolls the item into view.
		/// </summary>
		/// <param name="listBox">List-box in which to perform the focus synchronization. This may be null.</param>
		/// <param name="listBoxItem">List-box item to focus. If this is null, the currently selected item is focused.</param>
		/// <remarks>
		/// <para>If the list-box is not rendered, this method will not have any effect.</para>
		/// </remarks>
		public static void FocusListBoxItem( ListBox listBox, object listBoxItem = null )
		{
			if( listBox == null )
				return;

			if( listBoxItem == null )
				listBoxItem = listBox.SelectedItem;
			if( (listBoxItem == null) && (listBox.Items.Count > 0) )
				listBoxItem = listBox.Items[ 0 ];

			if( listBoxItem as ListBoxItem == null )
			{
				listBox.UpdateLayout();
				listBox.ScrollIntoView( listBoxItem );
			}

			listBox.Focus();

			if( listBox.ItemContainerGenerator.Status != GeneratorStatus.ContainersGenerated )
				return;

			ListBoxItem item = (listBoxItem is ListBoxItem) ? (ListBoxItem)listBoxItem
				: (listBox.ItemContainerGenerator.ContainerFromItem( listBoxItem ) as ListBoxItem);

			// Synchronize the focused item with the selected item.
			if( item != null )
			{
				item.IsSelected = true;
				item.Focus();
			}
		}

		/// <summary>
		/// Focuses the list-box item which is the parent of the specified child, and manually synchronizes the focused item with
		/// the selected item (because the focused item can't be bound as can the selected item).
		/// </summary>
		/// <param name="listBox">List-box in which to perform the focus synchronization. This may be null.</param>
		/// <param name="listBoxItemChild">A child of the list-box item to focus. This may be null.</param>
		public static void FocusParentListBoxItem( ListBox listBox, DependencyObject listBoxItemChild )
		{
			if( listBoxItemChild == null )
				return;

			Visual listBoxItem = FindVisualParent<ListBoxItem>( listBoxItemChild );

			FocusListBoxItem( listBox, listBoxItem );
		}

		/// <summary>
		/// Checks if the specified list-box is scrolled down near the bottom.
		/// </summary>
		/// <param name="listBox">List-box whose scroll to check.</param>
		/// <returns>Returns true if the specified list-box is scrolled down near the bottom.</returns>
		/// <remarks>
		/// <para>A list-box which is not rendered will always return true.</para>
		/// </remarks>
		public static bool IsListBoxScrolledNearBottom( ListBox listBox )
		{
			ValidationUtils.ObjectHasContent( listBox, "listBox" );

			if( listBox.Items.Count <= 0 )
				return true;

			ScrollViewer scrollViewer = FindVisualChild<ScrollViewer>( listBox ) as ScrollViewer;
			if( scrollViewer == null )
				return true;

			return scrollViewer.VerticalOffset >= scrollViewer.ScrollableHeight - ScrolledNearBottomPixelsCount;
		}

		/// <summary>
		/// Scrolls into view the top of the specified list-box.
		/// </summary>
		/// <param name="listBox">List-box to scroll.</param>
		public static void ScrollListBoxTopIntoView( ListBox listBox )
		{
			ValidationUtils.ObjectHasContent( listBox, "listBox" );

			if( listBox.Items.Count <= 0 )
				return;

			// When the list-box is not rendered, "ScrollIntoView" may fail, so try to scroll directly.
			ScrollViewer scrollViewer = FindVisualChild<ScrollViewer>( listBox ) as ScrollViewer;

			if( scrollViewer != null )
			{
				scrollViewer.UpdateLayout();
				scrollViewer.ScrollToTop();
			}
			else
				listBox.ScrollIntoView( listBox.Items[ 0 ] );
		}

		/// <summary>
		/// Scrolls into view the bottom of the specified list-box.
		/// </summary>
		/// <param name="listBox">List-box to scroll.</param>
		public static void ScrollListBoxBottomIntoView( ListBox listBox )
		{
			ValidationUtils.ObjectHasContent( listBox, "listBox" );

			if( listBox.Items.Count <= 0 )
				return;

			// When the list-box is not rendered, "ScrollIntoView" may fail, so try to scroll directly.
			ScrollViewer scrollViewer = FindVisualChild<ScrollViewer>( listBox ) as ScrollViewer;
			if( scrollViewer != null )
			{
				scrollViewer.UpdateLayout();
				scrollViewer.ScrollToBottom();
			}

			// Without this being always called, sometimes, when the list of letters is automatically scrolled to the bottom, the
			// scroll stops at the item before the last.
			listBox.ScrollIntoView( listBox.Items[ listBox.Items.Count - 1 ] );
		}

		/// <summary>
		/// Performs an action and maintains the (before-after) scroll in the specified list-box (if the action result is true).
		/// </summary>
		/// <param name="action">Action to perform. If this returns false, the list-box scroll is not affected. This may be null.
		/// </param>
		/// <param name="actionParameter">Action parameter. This may be null.</param>
		/// <param name="listBox">List-box in which to perform the scroll and focus synchronization. This may be null.</param>
		/// <returns>Returns the result of the action.</returns>
		public static bool ExecuteAndMaintainListBoxScrollIfActionSuccessful<T>( Func<T, bool> action, T actionParameter
			, ListBox listBox )
		{
			if( action == null )
				return false;

			ScrollViewer scrollViewer = null;
			double verticalScroll = 0;

			if( listBox != null )
			{
				scrollViewer = FindVisualChild<ScrollViewer>( listBox ) as ScrollViewer;
				if( scrollViewer != null )
					verticalScroll = scrollViewer.VerticalOffset;
			}

			if( !action( actionParameter ) )
				return false;

			if( listBox != null )
			{
				if( scrollViewer != null )
				{
					scrollViewer.UpdateLayout();
					scrollViewer.ScrollToVerticalOffset( verticalScroll );
				}

				FocusListBoxItem( listBox, null );
			}

			return true;
		}

		/// <summary>
		/// Performs an action and maintains the (before-after) scroll in the specified list-box.
		/// </summary>
		/// <param name="action">Action to perform. This may be null.</param>
		/// <param name="actionParameter">Action parameter. This may be null.</param>
		/// <param name="listBox">List-box in which to perform the scroll and focus synchronization. This may be null.</param>
		public static void ExecuteAndMaintainListBoxScroll<T>( Action<T> action, T actionParameter, ListBox listBox )
		{
			ExecuteAndMaintainListBoxScrollIfActionSuccessful<T>( ( ap ) => { action( ap ); return true; }
				, actionParameter, listBox );
		}

		/// <summary>
		/// Gets the control with the specified type, which is a child of the specified list item from the specified list-box.
		/// This may be null.
		/// </summary>
		/// <typeparam name="TChild">The type of the child control to find.</typeparam>
		/// <param name="listBox">List-box in which to look for the child control. This may be null.</param>
		/// <param name="listItem">The list item whose UI child control to find. This may be null.</param>
		public static TChild GetListBoxItemChild<TChild>( ListBox listBox, object listItem )
			where TChild : Visual
		{
			if( (listBox == null) || (listItem == null) )
				return null;

			Visual itemControl = listBox.ItemContainerGenerator.ContainerFromItem( listItem ) as Visual;
			if( itemControl == null )
				return null;

			return FindVisualChild<TChild>( itemControl );
		}

		/// <summary>
		/// Gets the control with the specified type, which is a child of the focused item of the specified list-box.
		/// This may be null.
		/// </summary>
		/// <typeparam name="TChild">The type of the child to find.</typeparam>
		/// <param name="listBox">List-box in which to look for the child. This may be null.</param>
		public static TChild GetListBoxFocusedItemChild<TChild>( ListBox listBox )
			where TChild : Visual
		{
			if( (listBox == null) || (listBox.SelectedItem == null) )
				return null;

			Visual itemControl = listBox.ItemContainerGenerator.ContainerFromItem( listBox.SelectedItem ) as Visual;
			if( itemControl == null )
				return null;

			return FindVisualChild<TChild>( itemControl );
		}

		/// <summary>
		/// Gets the indexes of the first and last visible items from the specified list-box.
		/// </summary>
		/// <param name="listBox">List-box from which to get the indexes of the visible items. This may be null.</param>
		/// <param name="firstVisibleItemIndex">Is set with the index of the first visible item.</param>
		/// <param name="lastVisibleItemIndex">Is set with the index of the last visible item.</param>
		/// <returns>Returns the number of visible items from the specified list-box.</returns>
		public static int GetVisibleItems( ListBox listBox, out int firstVisibleItemIndex, out int lastVisibleItemIndex )
		{
			firstVisibleItemIndex = -1;
			lastVisibleItemIndex = -1;

			if( listBox == null )
				return 0;

			if( listBox.ItemContainerGenerator.Status != GeneratorStatus.ContainersGenerated )
				return 0;

			ScrollViewer scrollViewer = FindVisualChild<ScrollViewer>( listBox ) as ScrollViewer;
			if( scrollViewer == null )
				return 0;

			int visibleItemsCount = 0;

			for( int i = 0; i < listBox.Items.Count; i++ )
			{
				ListBoxItem item = listBox.ItemContainerGenerator.ContainerFromIndex( i ) as ListBoxItem;

				if( item != null )
				{
					if( IsVisibleToTheUser( item, scrollViewer ) )
					{
						if( visibleItemsCount <= 0 )
							firstVisibleItemIndex = i;

						visibleItemsCount++;
					}
					else if( visibleItemsCount > 0 )
					{
						// The last visible item was encountered.
						break;
					}
				}
			}

			if( visibleItemsCount > 0 )
				lastVisibleItemIndex = firstVisibleItemIndex + visibleItemsCount - 1;

			return visibleItemsCount;
		}

		#endregion Public static methods
	}
}
