////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows.Media.Animation;
using Vela.Kernel;

namespace Vela.Controls
{
	/// <summary>
	/// Image preview.
	/// </summary>
	/// <remarks>
	/// <para>An image preview is a small version of an image.</para>
	/// 
	/// <para>An image preview can be either full-sized or thumb-sized.</para>
	/// 
	/// <para>An image preview may have multiple frames.</para>
	/// </remarks>
	public class ImagePreview
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ImagePreview()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="fileName">The name of the file from which the preview was loaded.</param>
		/// <param name="groupIdentifier">The group identifier of the image.</param>
		/// <param name="image">The image.</param>
		/// <param name="refersNoPreview">If this is true then the current image preview represents an image from which a preview
		/// could not be loaded, so "image" refers the no-preview image preview; if this is false then "image" refers the actual
		/// image preview.</param>
		public ImagePreview( string fileName, string groupIdentifier, ObjectAnimationUsingKeyFrames image
			, bool refersNoPreview )
		{
			FileName = fileName;
			GroupIdentifiers.Add( groupIdentifier );
			Image = image;
			RefersNoPreview = refersNoPreview;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private DateTime _CreationTime = DateTime.UtcNow;
		/// <summary>
		/// The time (in UTC format) when the current object was created. This is automatically set with the time of the creation
		/// of the current object (and you should not modify it). READ-ONLY.
		/// </summary>
		public DateTime CreationTime
		{
			get { return _CreationTime; }
			protected set { _CreationTime = value; }
		}

		private string _FileName;
		/// <summary>
		/// The name of the file from which the preview was loaded. READ-ONLY.
		/// </summary>
		public string FileName
		{
			get { return _FileName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "FileName" );

				_FileName = value;
			}
		}

		private Strings _GroupIdentifiers = new Strings();
		/// <summary>
		/// The group identifiers of the image. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This can be used to indicate that the image preview belongs to several groups of images, in an image cache,
		/// like those attached to letters.</para>
		/// </remarks>
		public Strings GroupIdentifiers
		{
			get { return _GroupIdentifiers; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "GroupIdentifiers" );

				_GroupIdentifiers = value;
			}
		}

		private ObjectAnimationUsingKeyFrames _Image;
		/// <summary>
		/// The image. READ-ONLY.
		/// </summary>
		public ObjectAnimationUsingKeyFrames Image
		{
			get { return _Image; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Image" );

				_Image = value;
			}
		}

		private bool _RefersNoPreview;
		/// <summary>
		/// If this is true then the current image preview represents an image from which a preview could not be loaded, so
		/// <see cref="Image"/> refers the no-preview image preview; if this is false then <see cref="Image"/> refers the
		/// actual image preview. READ-ONLY.
		/// </summary>
		public bool RefersNoPreview
		{
			get { return _RefersNoPreview; }
			protected set { _RefersNoPreview = value; }
		}

		#endregion Public non-serializable properties
	}
}
