////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Windows.Media.Animation;
using System.Windows.Media.Imaging;
using Vela.Kernel;

namespace Vela.Controls
{
	/// <summary>
	/// Cache for image previews.
	/// </summary>
	public static class ImagePreviewsCache
	{
		#region Public constants

		/// <summary>
		/// The maximum application memory size (in bytes) above which the image preview cache is cleared in order to make room
		/// for new cached images.
		/// </summary>
		public const int CleanUpMemoryThreshold = 1 << 30;

		#endregion Public constants

		#region Private static fields

		/// <summary>
		/// The synchronization object for the <see cref="FullSizedCachedPreviews"/> and
		/// <see cref="ThumbSizedCachedPreviews"/> fields. READ-ONLY.
		/// </summary>
		private static readonly object CachedPreviewsLock = new object();

		/// <summary>
		/// If this is 0 (or a negative number), the cache may be automatically cleared in order to free memory (see
		/// <see cref="MakeRoom"/>).
		/// </summary>
		private static int AutomaticClearLockIndex;

		/// <summary>
		/// The cached full-sized image previews.
		/// </summary>
		private static ImagePreviews FullSizedCachedPreviews = new ImagePreviews();

		/// <summary>
		/// The cached thumb-sized image previews.
		/// </summary>
		private static ImagePreviews ThumbSizedCachedPreviews = new ImagePreviews();

		#endregion Private static fields

		#region Private static methods

		/// <summary>
		/// Frees the memory taken by the cache, if the cache occupies too much memory.
		/// </summary>
		private static void MakeRoom()
		{
			lock( CachedPreviewsLock )
			{
				if( AutomaticClearLockIndex > 0 )
					return;

				Int64 appMemory = Process.GetCurrentProcess().PrivateMemorySize64;

				if( appMemory <= CleanUpMemoryThreshold )
					return;

				Clear();

				GC.Collect( 2, GCCollectionMode.Forced, false );
			}
		}

		/// <summary>
		/// Loads the image preview from the file with the specified name.
		/// </summary>
		/// <param name="fileName">The name of the file from which to load the image.</param>
		/// <param name="targetWidth">The width to which to resize the loaded image. This must be at least 1.</param>
		/// <param name="targetHeight">The height to which to resize the loaded image. This must be at least 1.</param>
		/// <returns>Returns the loaded image preview. This may be null.</returns>
		private static ObjectAnimationUsingKeyFrames Load( string fileName, int targetWidth, int targetHeight )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );
			ValidationUtils.Integer( targetWidth, 1, int.MaxValue, "targetWidth" );
			ValidationUtils.Integer( targetHeight, 1, int.MaxValue, "targetHeight" );

			Uri uri = new Uri( fileName );
			if( uri.IsFile )
			{
				FileInfo fi = new FileInfo( fileName );
				if( !fi.Exists )
					return null;
			}

			return GraphicUtils.LoadAndResizeMultiFrameImage( fileName, targetWidth, targetHeight, false );
		}

		/// <summary>
		/// Loads the image preview from the file with the specified name.
		/// </summary>
		/// <param name="fileName">The name of the file from which to load the image.</param>
		/// <param name="loadFullSize">If this is true then the full-sized preview is returned, else the thumb-sized preview
		/// is returned.</param>
		/// <returns>Returns the loaded image preview. This may be null.</returns>
		private static ObjectAnimationUsingKeyFrames Load( string fileName, bool loadFullSize )
		{
			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				if( loadFullSize )
				{
					ObjectAnimationUsingKeyFrames image = Load( fileName, KernelConsts.FullSizedImagePreviewWidthInt
						, KernelConsts.FullSizedImagePreviewHeightInt );
					if( image != null )
						return image;
				}
				else
				{
					ObjectAnimationUsingKeyFrames image = Load( fileName, KernelConsts.ThumbSizedImagePreviewWidthInt
						, KernelConsts.ThumbSizedImagePreviewHeightInt );
					if( image != null )
						return image;
				}

				return null;
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		/// <summary>
		/// Removes all the cached image previews of the group with the specified identifier, from the specified cache.
		/// </summary>
		/// <param name="groupIdentifier">The group identifier of the image previews to remove.</param>
		/// <param name="onlyNoPreviews">If this is true then only the cached image previews which refer the no-preview image
		/// preview (<see cref="ImagePreview.RefersNoPreview"/> is true) are removed.</param>
		/// <param name="cachedPreviews">The cache from which to remove the image previews. This may be empty.</param>
		/// <returns>Returns true if an image preview was removed from the cache.</returns>
		/// <remarks>
		/// <para>The <see cref="FullSizedCachedPreviews"/> and <see cref="ThumbSizedCachedPreviews"/> properties are
		/// thread safe within this method.</para>
		/// </remarks>
		private static bool RemoveCachedFromGroup( string groupIdentifier, bool onlyNoPreviews, ImagePreviews cachedPreviews )
		{
			ValidationUtils.TextHasContent( groupIdentifier, "groupIdentifier" );
			ValidationUtils.ObjectHasContent( cachedPreviews, "cachedPreviews" );

			bool removed = false;

			List<string> fileNamesToRemove = new List<string>();

			foreach( ImagePreview cachedPreview in cachedPreviews )
			{
				if( onlyNoPreviews && !cachedPreview.RefersNoPreview )
					continue;

				if( !cachedPreview.GroupIdentifiers.ContainsKey( groupIdentifier ) )
					continue;

				cachedPreview.GroupIdentifiers.Remove( groupIdentifier );

				if( cachedPreview.GroupIdentifiers.Count <= 0 )
					fileNamesToRemove.Add( cachedPreview.FileName );

				removed = true;
			}

			foreach( string fileNameToRemove in fileNamesToRemove )
				cachedPreviews.Remove( fileNameToRemove );

			return removed;
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Locks the automatic clearing of the cache which is done to free memory, when required (see <see cref="MakeRoom"/>).
		/// </summary>
		public static void LockAutomaticClear()
		{
			lock( CachedPreviewsLock )
			{
				AutomaticClearLockIndex++;
			}
		}

		/// <summary>
		/// Unlocks the automatic clearing of the cache which is done to free memory, when required (see <see cref="MakeRoom"/>).
		/// </summary>
		public static void UnlockAutomaticClear()
		{
			lock( CachedPreviewsLock )
			{
				if( AutomaticClearLockIndex > 0 )
					AutomaticClearLockIndex--;
			}
		}

		/// <summary>
		/// Clears the cache.
		/// </summary>
		/// <remarks>
		/// <para>The <see cref="FullSizedCachedPreviews"/> and <see cref="ThumbSizedCachedPreviews"/> properties are
		/// thread safe within this method.</para>
		/// </remarks>
		public static void Clear()
		{
			lock( CachedPreviewsLock )
			{
				FullSizedCachedPreviews.Clear();
				ThumbSizedCachedPreviews.Clear();
			}
		}

		/// <summary>
		/// Gets the image group identifier for the specified locator.
		/// </summary>
		/// <param name="locator">Locator for which to get the image group identifier.</param>
		/// <returns>Returns the image group identifier for the specified locator.</returns>
		public static string GetImageGroupIdentifier( Locator locator )
		{
			ValidationUtils.ObjectHasContent( locator, "locator" );

			return TextUtils.ArrayToText16( locator.Value );
		}

		/// <summary>
		/// Gets the cached image preview of the file with the specified name.
		/// </summary>
		/// <param name="fileName">The file name of the image preview to get.</param>
		/// <param name="groupIdentifier">The group identifier of the image preview to get. If the file name is found but this
		/// group identifier is not, it's added to the found file.</param>
		/// <param name="loadFullSize">If this is true then the full-sized preview is returned, else the thumb-sized preview
		/// is returned.</param>
		/// <returns>Returns the cached image preview. This may be null.</returns>
		/// <remarks>
		/// <para>The <see cref="FullSizedCachedPreviews"/> and <see cref="ThumbSizedCachedPreviews"/> properties are
		/// thread safe within this method.</para>
		/// </remarks>
		public static ImagePreview GetCached( string fileName, string groupIdentifier, bool loadFullSize )
		{
			ValidationUtils.ObjectHasContent( fileName, "fileName" );
			ValidationUtils.ObjectHasContent( groupIdentifier, "groupIdentifier" );

			ImagePreviews cachedPreviews = loadFullSize ? FullSizedCachedPreviews : ThumbSizedCachedPreviews;

			lock( CachedPreviewsLock )
			{
				ImagePreview cachedPreview = cachedPreviews[ fileName ];

				// If the image preview is cached, return it.
				if( cachedPreview != null )
				{
					cachedPreview.GroupIdentifiers.Add( groupIdentifier, true );

					return cachedPreview;
				}
			}

			return null;
		}

		/// <summary>
		/// Gets a group image with all the cached image previews with the specified group identifier.
		/// </summary>
		/// <param name="groupIdentifier">The group identifier for which to get the cached image previews.</param>
		/// <param name="includeRefersNoPreviews">If this is true then the image previews which refer the no-preview image
		/// preview (<see cref="ImagePreview.RefersNoPreview"/> is true) are also included in the returned group image preview,
		/// else they are not.</param>
		/// <param name="loadFullSize">If this is true then the full-sized previews are returned, else the thumb-sized previews
		/// are returned.</param>
		/// <returns>Returns the cached group image.</returns>
		/// <remarks>
		/// <para>The <see cref="FullSizedCachedPreviews"/> and <see cref="ThumbSizedCachedPreviews"/> properties are
		/// thread safe within this method.</para>
		/// </remarks>
		public static GroupImagePreview GetCachedGroupImage( string groupIdentifier, bool includeRefersNoPreviews
			, bool loadFullSize )
		{
			ValidationUtils.ObjectHasContent( groupIdentifier, "groupIdentifier" );

			ImagePreviews cachedPreviews = loadFullSize ? FullSizedCachedPreviews : ThumbSizedCachedPreviews;
			GroupImagePreview groupImage = new GroupImagePreview();

			lock( CachedPreviewsLock )
			{
				foreach( ImagePreview cachedPreview in cachedPreviews )
				{
					if( cachedPreview == null || !cachedPreview.GroupIdentifiers.ContainsKey( groupIdentifier ) )
						continue;

					if( cachedPreview.RefersNoPreview && !includeRefersNoPreviews )
						continue;

					groupImage.Add( cachedPreview );
				}
			}

			return groupImage;
		}

		/// <summary>
		/// Gets the no-preview placeholder image preview.
		/// </summary>
		/// <param name="loadFullSize">If this is true then the full-sized image preview is loaded, else the thumb-sized image
		/// preview is loaded.</param>
		/// <returns>Returns the no-preview placeholder image preview.</returns>
		public static ImagePreview GetNoPreview( bool loadFullSize )
		{
			string noPreviewUrl = loadFullSize ? ControlsConsts.FullSizedNoPreviewFileName
				: ControlsConsts.ThumbSizedNoPreviewFileName;

			return GetCachedOrLoad( noPreviewUrl, noPreviewUrl, loadFullSize, true );
		}

		/// <summary>
		/// Caches a no-preview image preview for the file with the specified name.
		/// </summary>
		/// <param name="fileName">The name of the file for which to cache a no-preview image preview.</param>
		/// <param name="groupIdentifier">The group identifier of the image file with the specified name.</param>
		/// <param name="loadFullSize">If this is true then the full-sized preview is returned, else the thumb-sized preview
		/// is returned.</param>
		/// <remarks>
		/// <para>The <see cref="FullSizedCachedPreviews"/> and <see cref="ThumbSizedCachedPreviews"/> properties are
		/// thread safe within this method.</para>
		/// </remarks>
		public static void CacheNoPreview( string fileName, string groupIdentifier, bool loadFullSize )
		{
			lock( CachedPreviewsLock )
			{
				ImagePreview existingImagePreview = GetCached( fileName, groupIdentifier, loadFullSize );
				if( existingImagePreview != null )
					return;

				ImagePreviews cachedPreviews = loadFullSize ? FullSizedCachedPreviews : ThumbSizedCachedPreviews;

				ObjectAnimationUsingKeyFrames image = GetNoPreview( loadFullSize ).Image;
				ImagePreview imagePreview = new ImagePreview( fileName, groupIdentifier, image, true );

				cachedPreviews.Add( imagePreview, true );
			}
		}

		/// <summary>
		/// Loads and caches the image preview from the file with the specified name.
		/// </summary>
		/// <param name="fileName">The name of the file from which to load the image.</param>
		/// <param name="groupIdentifier">The group identifier of the image file with the specified name.</param>
		/// <param name="loadFullSize">If this is true then the full-sized preview is returned, else the thumb-sized preview
		/// is returned.</param>
		/// <param name="throwIfMissing">If this is true then an exception is thrown if an image preview can't be loaded.</param>
		/// <returns>Returns the loaded image preview.</returns>
		/// <remarks>
		/// <para>If an image preview can't be loaded, if "throwIfMissing" is false, an <see cref="ImagePreview"/> item is
		/// added to the cache to indicate that no further attempts should be made to load the file. The image of this item
		/// refers to the no-preview image preview.</para>
		/// 
		/// <para>The <see cref="FullSizedCachedPreviews"/> and <see cref="ThumbSizedCachedPreviews"/> properties are
		/// thread safe within this method.</para>
		/// </remarks>
		public static ImagePreview GetCachedOrLoad( string fileName, string groupIdentifier, bool loadFullSize
			, bool throwIfMissing = false )
		{
			ImagePreview existingImagePreview = GetCached( fileName, groupIdentifier, loadFullSize );
			if( existingImagePreview != null )
				return existingImagePreview;

			ObjectAnimationUsingKeyFrames image = Load( fileName, loadFullSize );
			if( (image == null) && throwIfMissing )
				throw new FileNotFoundException( string.Format( "The file \"{0}\" was not found", fileName ) );

			lock( CachedPreviewsLock )
			{
				MakeRoom();

				ImagePreviews cachedPreviews = loadFullSize ? FullSizedCachedPreviews : ThumbSizedCachedPreviews;

				// The image preview may have been cached between the beginning of the current method and the lock, so try to
				// get it again from the cache in order to merge the group identifiers.
				existingImagePreview = cachedPreviews[ fileName ];

				if( existingImagePreview != null )
				{
					// Merge the group identifiers.
					existingImagePreview.GroupIdentifiers.Add( groupIdentifier, true );

					return existingImagePreview;
				}
				else
				{
					ImagePreview imagePreview = (image != null)
						? new ImagePreview( fileName, groupIdentifier, image, false )
						: new ImagePreview( fileName, groupIdentifier, GetNoPreview( loadFullSize ).Image, true );

					cachedPreviews.Add( imagePreview, true );

					return imagePreview;
				}
			}
		}

		/// <summary>
		/// Loads and caches the image preview from the file with the specified name.
		/// </summary>
		/// <param name="fileName">The name of the file from which to load the image.</param>
		/// <param name="groupIdentifier">The group identifier of the image file with the specified name.</param>
		/// <param name="loadFullSize">If this is true then the full-sized preview is returned, else the thumb-sized preview
		/// is returned.</param>
		/// <returns>Returns the loaded image preview. This may be null.</returns>
		/// <remarks>
		/// <para>The <see cref="FullSizedCachedPreviews"/> and <see cref="ThumbSizedCachedPreviews"/> properties are
		/// thread safe within this method.</para>
		/// </remarks>
		public static BitmapSource GetCachedOrLoadFirstFrame( string fileName, string groupIdentifier, bool loadFullSize )
		{
			ImagePreview imagePreview = GetCachedOrLoad( fileName, groupIdentifier, loadFullSize );
			if( imagePreview == null )
				return null;

			ObjectAnimationUsingKeyFrames image = imagePreview.Image;
			if( (image == null) || (image.KeyFrames.Count <= 0) )
				return null;

			return image.KeyFrames[ 0 ].Value as BitmapSource;
		}

		/// <summary>
		/// Loads and caches the image preview from the file with the specified name.
		/// </summary>
		/// <param name="fileName">The name of the file from which to load the image.</param>
		/// <param name="loadFullSize">If this is true then the full-sized preview is returned, else the thumb-sized preview
		/// is returned.</param>
		/// <returns>Returns the loaded image preview. This may be null.</returns>
		/// <remarks>
		/// <para>The <see cref="FullSizedCachedPreviews"/> and <see cref="ThumbSizedCachedPreviews"/> properties are
		/// thread safe within this method.</para>
		/// </remarks>
		public static BitmapSource GetCachedOrLoadFirstFrame( string fileName, bool loadFullSize )
		{
			return GetCachedOrLoadFirstFrame( fileName, fileName, loadFullSize );
		}

		/// <summary>
		/// Removes the cached image preview (= full-sized or thumb-sized) of the file with the specified name.
		/// </summary>
		/// <param name="fileName">The file name of the image to remove.</param>
		/// <param name="removeFullSize">If this is true then the full-sized preview is removed, else the thumb-sized preview
		/// is removed.</param>
		/// <returns>Returns true if the image preview was removed from the cache.</returns>
		/// <remarks>
		/// <para>The <see cref="FullSizedCachedPreviews"/> and <see cref="ThumbSizedCachedPreviews"/> properties are
		/// thread safe within this method.</para>
		/// 
		/// <para>Be careful with this method because it removes the preview for all the group identifiers!</para>
		/// </remarks>
		public static bool RemoveCached( string fileName, bool removeFullSize )
		{
			ValidationUtils.ObjectHasContent( fileName, "fileName" );

			ImagePreviews cachedPreviews = removeFullSize ? FullSizedCachedPreviews : ThumbSizedCachedPreviews;

			lock( CachedPreviewsLock )
			{
				ImagePreview cachedPreview = cachedPreviews.Remove( fileName );

				return cachedPreview != null;
			}
		}

		/// <summary>
		/// Removes the cached image previews (= full-sized or thumb-sized) of the files with the specified names.
		/// </summary>
		/// <param name="fileNames">The file names of the images to remove.</param>
		/// <param name="removeFullSize">If this is true then the full-sized previews are removed, else the thumb-sized previews
		/// are removed.</param>
		/// <remarks>
		/// <para>The <see cref="FullSizedCachedPreviews"/> and <see cref="ThumbSizedCachedPreviews"/> properties are
		/// thread safe within this method.</para>
		/// 
		/// <para>Be careful with this method because it removes the previews for all the group identifiers!</para>
		/// </remarks>
		public static void RemoveCached( Strings fileNames, bool removeFullSize )
		{
			ValidationUtils.ObjectHasContent( fileNames, "fileNames" );

			ImagePreviews cachedPreviews = removeFullSize ? FullSizedCachedPreviews : ThumbSizedCachedPreviews;

			lock( CachedPreviewsLock )
			{
				foreach( string fileName in fileNames )
					cachedPreviews.Remove( fileName );
			}
		}

		/// <summary>
		/// Removes all the cached image previews (= full-sized and thumb-sized) of the group with the specified identifier.
		/// </summary>
		/// <param name="groupIdentifier">The group identifier of the image previews to remove.</param>
		/// <param name="onlyNoPreviews">If this is true then only the cached image previews which refer the no-preview image
		/// preview (<see cref="ImagePreview.RefersNoPreview"/> is true) are removed.</param>
		/// <returns>Returns true if an image preview was removed from the cache.</returns>
		/// <remarks>
		/// <para>The <see cref="FullSizedCachedPreviews"/> and <see cref="ThumbSizedCachedPreviews"/> properties are
		/// thread safe within this method.</para>
		/// </remarks>
		public static bool RemoveCachedFromGroup( string groupIdentifier, bool onlyNoPreviews )
		{
			ValidationUtils.ObjectHasContent( groupIdentifier, "groupIdentifier" );

			bool removedFullSized = false;
			bool removedThumbSized = false;

			lock( CachedPreviewsLock )
			{
				removedFullSized = RemoveCachedFromGroup( groupIdentifier, onlyNoPreviews, FullSizedCachedPreviews );
				removedThumbSized = RemoveCachedFromGroup( groupIdentifier, onlyNoPreviews, ThumbSizedCachedPreviews );
			}

			return removedFullSized || removedThumbSized;
		}

		/// <summary>
		/// Removes all the cached image previews (= full-sized and thumb-sized) of the group with the specified identifier.
		/// </summary>
		/// <param name="groupIdentifier">The group identifier of the image previews to remove.</param>
		/// <param name="onlyNoPreviews">If this is true then only the cached image previews which refer the no-preview image
		/// preview (<see cref="ImagePreview.RefersNoPreview"/> is true) are removed.</param>
		/// <returns>Returns true if an image preview was removed from the cache.</returns>
		/// <remarks>
		/// <para>The <see cref="FullSizedCachedPreviews"/> and <see cref="ThumbSizedCachedPreviews"/> properties are
		/// thread safe within this method.</para>
		/// </remarks>
		public static bool RemoveCachedFromGroup( Locator groupIdentifier, bool onlyNoPreviews )
		{
			ValidationUtils.ObjectHasContent( groupIdentifier, "groupIdentifier" );

			string imageGroupIdentifier = GetImageGroupIdentifier( groupIdentifier );

			return RemoveCachedFromGroup( imageGroupIdentifier, onlyNoPreviews );
		}

		#endregion Public static methods
	}
}
