////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using System.Windows.Media.Imaging;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.Controls
{
	/// <summary>
	/// Peer image.
	/// </summary>
	public class PeerImage : NotifyPropertyChanged, IPeerImage
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected PeerImage()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="description">Image description. This may be null or empty.</param>
		/// <param name="imageData">Image data.</param>
		public PeerImage( string description, byte[] imageData )
		{
			Description = description;
			ImageData = imageData;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The width (in pixels). READ-ONLY.
		/// </summary>
		public int Width
		{
			get { return (SystemImage != null) ? SystemImage.PixelWidth : 0; }
		}

		/// <summary>
		/// The height (in pixels). READ-ONLY.
		/// </summary>
		public int Height
		{
			get { return (SystemImage != null) ? SystemImage.PixelHeight : 0; }
		}

		/// <summary>
		/// The date when the image was taken. READ-ONLY.
		/// </summary>
		public string DateTaken
		{
			get
			{
				if( (SystemImage == null) || (SystemImage.Metadata as BitmapMetadata == null) )
					return "Unknown";

				BitmapMetadata metadata = SystemImage.Metadata as BitmapMetadata;

				string dateTakenText = metadata.DateTaken;
				if( TextUtils.NoContent( metadata.DateTaken ) )
					return "Unknown";

				DateTime dateTaken;
				if( DateTime.TryParse( dateTakenText, out dateTaken ) )
					return dateTaken.ToLocalTime().ToString( KernelConsts.TimeUiFormat );

				return dateTakenText;
			}
		}

		private BitmapFrame _SystemImage;
		/// <summary>
		/// The underlying system image. This may be null. Notifies property observers. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public BitmapFrame SystemImage
		{
			get
			{
				if( _SystemImage != null )
					return _SystemImage;

				if( ImageData != null )
				{
					// Load the image from the stream.
					using( MemoryStream imageStream = new MemoryStream( ImageData ) )
					{
						_SystemImage = GraphicUtils.LoadImage( imageStream );

						// Notify property observers. Avoid stack overflows when notifying about null values.
						if( _SystemImage != null )
						{
							OnPropertyChanged( "Width" );
							OnPropertyChanged( "Height" );
							OnPropertyChanged( "DateTaken" );
							OnPropertyChanged( "SystemImage" );
						}
					}
				}

				return _SystemImage;
			}
			protected set
			{
				if( value != null )
				{
					ValidationUtils.Integer( (int)value.Width, 1, KernelConsts.ImageWidthMax, "SystemImage.Width" );
					ValidationUtils.Integer( (int)value.Height, 1, KernelConsts.ImageHeightMax, "SystemImage.Height" );
				}

				_SystemImage = value;

				// Notify property observers.
				OnPropertyChanged( "Width" );
				OnPropertyChanged( "Height" );
				OnPropertyChanged( "DateTaken" );
				OnPropertyChanged( "SystemImage" );
			}
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private string _Description;
		/// <summary>
		/// Image description. This may be null or empty. Notifies property observers. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string Description
		{
			get { return _Description; }
			protected set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.ShortNote, "Description" );

				_Description = value;

				// Notify property observers.
				OnPropertyChanged( "Description" );
			}
		}

		private byte[] _ImageData;
		/// <summary>
		/// Image data. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] ImageData
		{
			get { return _ImageData; }
			protected set
			{
				ValidationUtils.Array( value, 1, KernelConsts.PeerImageSizeMax, "ImageData" );

				_ImageData = value;

				// Recreate the underlying system image.
				_SystemImage = null;
			}
		}

		#endregion Public serializable properties
	}
}
