////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Runtime.InteropServices;
using System.Windows;
using System.Windows.Input;
using Vela.Kernel;

namespace Vela.Controls
{
	/// <summary>
	/// Global manager for recent user input.
	/// </summary>
	public class RecentUserInputGlobalManager : IDisposable
	{
		#region Private classes

		/// <summary>
		/// Native API utilities.
		/// </summary>
		private class NativeApi
		{
			[DllImport( "user32.dll", CharSet = CharSet.Unicode )]
			private static extern bool GetKeyboardState( [Out] byte[] keyStates );

			/// <summary>
			/// Checks if any key (from the keyboard or the mouse) is pressed.
			/// </summary>
			/// <returns>Returns true if any key (from the keyboard or the mouse) is pressed.</returns>
			public static bool IsAnyKeyPressed()
			{
				byte[] keyStates = new byte[ 256 ];

				if( !GetKeyboardState( keyStates ) )
					return false;

				foreach( byte keyState in keyStates )
				{
					if( (keyState & 0x80) != 0 )
						return true;
				}

				return false;
			}
		}

		#endregion Private classes

		#region Private constants

		/// <summary>
		/// The timespan (in seconds) within which to consider that the user input is (still) recent.
		/// </summary>
		private const double RecentUserInputTimespan = 0.5;

		#endregion Private constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="window">Window for which to add global handlers for keyboard and mouse activity.</param>
		public RecentUserInputGlobalManager( Window window )
		{
			ValidationUtils.ObjectHasContent( window, "window" );

			Window = window;

			Keyboard.AddPreviewKeyDownHandler( Window, PreviewKeyDownHandler );
			Keyboard.AddPreviewKeyUpHandler( Window, PreviewKeyUpHandler );
			Mouse.AddPreviewMouseDownHandler( Window, PreviewMouseDownHandler );
			Mouse.AddPreviewMouseUpHandler( Window, PreviewMouseUpHandler );
		}

		/// <summary>
		/// Cleans up the current instance.
		/// </summary>
		~RecentUserInputGlobalManager()
		{
			Dispose( false );
		}

		#endregion Constructors

		#region Private static fields

		/// <summary>
		/// The synchronization object for the <see cref="Instance"/> property. READ-ONLY.
		/// </summary>
		private static readonly object InstanceLock = new object();

		#endregion Private static fields

		#region Private fields

		/// <summary>
		/// If this is true then the current instance was disposed.
		/// </summary>
		private bool WasDisposed;

		/// <summary>
		/// Window for which to add global handlers for keyboard and mouse activity.
		/// </summary>
		private Window Window;

		/// <summary>
		/// The last time (in UTC format) when active (= key down or up) user input was detected.
		/// </summary>
		private DateTime UserInputLastTime;

		#endregion Private fields

		#region Private static properties

		private static RecentUserInputGlobalManager _Instance;
		/// <summary>
		/// The global instance for the global manager for recent user input.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// </remarks>
		public static RecentUserInputGlobalManager Instance
		{
			get
			{
				lock( InstanceLock )
				{
					if( _Instance == null )
					{
						throw new ValidationException( "The global instance for the global manager for recent user input should"
							+ " be set when the application is started, before this property is read." );
					}

					return _Instance;
				}
			}
			set
			{
				ValidationUtils.ObjectHasContent( value, "Instance" );

				lock( InstanceLock )
				{
					_Instance = value;
				}
			}
		}

		#endregion Private static properties

		#region Event handlers

		private void PreviewKeyDownHandler( object sender, KeyEventArgs e )
		{
			UserInputLastTime = DateTime.UtcNow;
		}

		private void PreviewKeyUpHandler( object sender, KeyEventArgs e )
		{
			UserInputLastTime = DateTime.UtcNow;
		}

		private void PreviewMouseDownHandler( object sender, MouseButtonEventArgs e )
		{
			UserInputLastTime = DateTime.UtcNow;
		}

		private void PreviewMouseUpHandler( object sender, MouseButtonEventArgs e )
		{
			UserInputLastTime = DateTime.UtcNow;
		}

		#endregion Event handlers

		#region Public methods

		/// <summary>
		/// Disposes the current instance.
		/// </summary>
		/// <param name="calledFromDisposeNotFinalize">If this is true then the method is being called from
		/// <see cref="Dispose()"/> not from the finalizer.</param>
		public void Dispose( bool calledFromDisposeNotFinalize )
		{
			if( WasDisposed )
				return;

			if( calledFromDisposeNotFinalize )
			{
				Keyboard.RemovePreviewKeyDownHandler( Window, PreviewKeyDownHandler );
				Keyboard.RemovePreviewKeyUpHandler( Window, PreviewKeyUpHandler );
				Mouse.RemovePreviewMouseDownHandler( Window, PreviewMouseDownHandler );
				Mouse.RemovePreviewMouseUpHandler( Window, PreviewMouseUpHandler );
			}

			WasDisposed = true;
		}

		/// <summary>
		/// Disposes the current instance.
		/// </summary>
		public void Dispose()
		{
			Dispose( true );
			GC.SuppressFinalize( this );
		}

		/// <summary>
		/// Checks if there was recent user input through the keyboard or the mouse.
		/// </summary>
		/// <returns>Returns true if there was recent user input through the keyboard or the mouse.</returns>
		public bool HasRecentUserInput()
		{
			bool isAnyKeyPressed = NativeApi.IsAnyKeyPressed();

			DateTime now = DateTime.UtcNow;

			if( isAnyKeyPressed )
				UserInputLastTime = now;

			return isAnyKeyPressed || ((now - UserInputLastTime).TotalSeconds <= RecentUserInputTimespan);
		}

		#endregion Public methods
	}
}
