////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Windows.Media;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.Controls
{
	/// <summary>
	/// Configuration for a text which is to be written on an image.
	/// </summary>
	public class TextOnImageConfiguration : IDocument
	{
		#region Public constants

		/// <summary>
		/// The default value for <see cref="WriteTextIfSizeMin"/>. This must be at least 1.
		/// </summary>
		public const int WriteTextIfSizeMinDef = 1;

		/// <summary>
		/// The default value for <see cref="HorizontalMarginFraction"/> and <see cref="VerticalMarginFraction"/>. This must be
		/// at least 0.
		/// </summary>
		public const double MarginFractionDef = 0.03;

		/// <summary>
		/// The minimum value for <see cref="FrameWidthFractionMax"/>. This must be greater than 0.
		/// </summary>
		public const double FrameWidthFractionMaxMin = 0.1;

		/// <summary>
		/// The default value for <see cref="FrameWidthFractionMax"/>. This must be greater than or equal with
		/// "FrameWidthFractionMaxMin".
		/// </summary>
		public const double FrameWidthFractionMaxDef = 0.6;

		/// <summary>
		/// The default value for <see cref="FrameAlignment"/>.
		/// </summary>
		public const TextOnImageAlignments FrameAlignmentDef = TextOnImageAlignments.BottomCenter;

		/// <summary>
		/// The minimum value for <see cref="FontSizeFraction"/>. This must be greater than 0.
		/// </summary>
		public const double FontSizeFractionMin = 0.005;

		/// <summary>
		/// The default value for <see cref="FontSizeFraction"/>.
		/// </summary>
		public const double FontSizeFractionDef = 0.04;

		/// <summary>
		/// The minimum value for <see cref="OutlineThicknessFraction"/>. This must be greater than 0.
		/// </summary>
		public const double OutlineThicknessFractionMin = 0.01;

		/// <summary>
		/// The default value for <see cref="OutlineThicknessFraction"/>.
		/// </summary>
		public const double OutlineThicknessFractionDef = 0.05;

		/// <summary>
		/// The default value for <see cref="FontFamilyName"/>.
		/// </summary>
		public const string FontFamilyNameDef = "Arial";

		/// <summary>
		/// The default value for <see cref="IsBold"/>.
		/// </summary>
		public const bool IsBoldDef = true;

		/// <summary>
		/// The default value for <see cref="IsItalic"/>.
		/// </summary>
		public const bool IsItalicDef = true;

		/// <summary>
		/// The default value for <see cref="FontColor"/>.
		/// </summary>
		public static readonly Color FontColorDef = Colors.White;

		/// <summary>
		/// The default value for <see cref="OutlineColor"/>.
		/// </summary>
		public static readonly Color OutlineColorDef = Colors.Black;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public TextOnImageConfiguration()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="writeTextIfSizeMin">The minimum number of characters that the text to write on the image must have in
		/// order for it to be written. This must be at least 1.</param>
		/// <param name="horizontalMarginFraction">The horizontal margin around the rectangle which bounds the text to write on
		/// the image, as a fraction from the shorter side of the image. This must be at least 0.</param>
		/// <param name="verticalMarginFraction">The vertical margin around the rectangle which bounds the text to write on the
		/// image, as a fraction from the shorter side of the image. This must be at least 0.</param>
		/// <param name="textWidthMaxFraction">The maximum width of the rectangle which bounds the text to write on the image, as
		/// a fraction from the width of the image. This must be greater than 0.</param>
		/// <param name="fontSizeFraction">The size of the font with which to write the text on the image, as a fraction from the
		/// shorter side of the image. This must be greater than 0.</param>
		/// <param name="outlineThicknessFraction">The thickness of the outline of the font with which to write the text on the
		/// image, as a fraction from the font size. This must be greater than 0.</param>
		/// <param name="fontFamilyName">The family name of the font with which to write the text on the image.</param>
		/// <param name="isBold">If this is true then the font with which to write the text on the image is bold.</param>
		/// <param name="isItalic">If this is true then the font with which to write the text on the image is italic.</param>
		/// <param name="fontColor">The color of the font with which to write the text on the image.</param>
		/// <param name="outlineColor">The color of the outline of the font with which to write the text on the image.</param>
		public TextOnImageConfiguration( int writeTextIfSizeMin, double horizontalMarginFraction, double verticalMarginFraction
			, double textWidthMaxFraction, double fontSizeFraction, double outlineThicknessFraction, string fontFamilyName
			, bool isBold, bool isItalic, Color fontColor, Color outlineColor )
		{
			WriteTextIfSizeMin = writeTextIfSizeMin;
			HorizontalMarginFraction = horizontalMarginFraction;
			VerticalMarginFraction = verticalMarginFraction;
			FrameWidthFractionMax = textWidthMaxFraction;
			FontSizeFraction = fontSizeFraction;
			OutlineThicknessFraction = outlineThicknessFraction;
			FontFamilyName = fontFamilyName;
			IsBold = isBold;
			IsItalic = isItalic;
			FontColor = fontColor;
			OutlineColor = outlineColor;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private Color _FontColor = FontColorDef;
		/// <summary>
		/// The color of the font with which to write the text on the image. READ-ONLY.
		/// </summary>
		public Color FontColor
		{
			get { return _FontColor; }
			protected set { _FontColor = value; }
		}

		private Color _OutlineColor = OutlineColorDef;
		/// <summary>
		/// The color of the outline of the font with which to write the text on the image. READ-ONLY.
		/// </summary>
		public Color OutlineColor
		{
			get { return _OutlineColor; }
			protected set { _OutlineColor = value; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private int _WriteTextIfSizeMin = WriteTextIfSizeMinDef;
		/// <summary>
		/// The minimum number of characters that the text to write on the image must have in order for it to be written. This
		/// must be at least 1. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public int WriteTextIfSizeMin
		{
			get { return _WriteTextIfSizeMin; }
			protected set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "WriteTextIfSizeMin" );

				_WriteTextIfSizeMin = value;
			}
		}

		private double _HorizontalMarginFraction = MarginFractionDef;
		/// <summary>
		/// The horizontal margin around the rectangle which bounds the text to write on the image, as a fraction from the
		/// shorter side of the image. This must be at least 0. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>Since this is a new property, it must have "MustBeDeserialized = false".</para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false )]
		public double HorizontalMarginFraction
		{
			get { return _HorizontalMarginFraction; }
			protected set
			{
				ValidationUtils.Real( value, 0, 1, "HorizontalMarginFraction" );

				_HorizontalMarginFraction = value;
			}
		}

		private double _VerticalMarginFraction = MarginFractionDef;
		/// <summary>
		/// The vertical margin around the rectangle which bounds the text to write on the image, as a fraction from the shorter
		/// side of the image. This must be at least 0. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>Since this is a new property, it must have "MustBeDeserialized = false".</para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false )]
		public double VerticalMarginFraction
		{
			get { return _VerticalMarginFraction; }
			protected set
			{
				ValidationUtils.Real( value, 0, 1, "VerticalMarginFraction" );

				_VerticalMarginFraction = value;
			}
		}

		private double _FrameWidthFractionMax = FrameWidthFractionMaxDef;
		/// <summary>
		/// The maximum width of the rectangle which bounds the text to write on the image, as a fraction from the width of the
		/// image. This must be greater than 0. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public double FrameWidthFractionMax
		{
			get { return _FrameWidthFractionMax; }
			protected set
			{
				ValidationUtils.Real( value, FrameWidthFractionMaxMin, 1, "FrameWidthFractionMax" );

				_FrameWidthFractionMax = value;
			}
		}

		private TextOnImageAlignments _FrameAlignment = FrameAlignmentDef;
		/// <summary>
		/// The alignment of the frame which bounds the text to write on the image. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public TextOnImageAlignments FrameAlignment
		{
			get { return _FrameAlignment; }
			protected set { _FrameAlignment = value; }
		}

		private double _FontSizeFraction = FontSizeFractionDef;
		/// <summary>
		/// The size of the font with which to write the text on the image, as a fraction from the shorter side of the image.
		/// This must be greater than 0. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public double FontSizeFraction
		{
			get { return _FontSizeFraction; }
			protected set
			{
				ValidationUtils.Real( value, FontSizeFractionMin, 1, "FontSizeFraction" );

				_FontSizeFraction = value;
			}
		}

		private double _OutlineThicknessFraction = OutlineThicknessFractionDef;
		/// <summary>
		/// The thickness of the outline of the font with which to write the text on the image, as a fraction from the font size.
		/// This must be greater than 0. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public double OutlineThicknessFraction
		{
			get { return _OutlineThicknessFraction; }
			protected set
			{
				ValidationUtils.Real( value, OutlineThicknessFractionMin, 1, "OutlineThicknessFraction" );

				_OutlineThicknessFraction = value;
			}
		}

		private string _FontFamilyName = FontFamilyNameDef;
		/// <summary>
		/// The family name of the font with which to write the text on the image. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string FontFamilyName
		{
			get { return _FontFamilyName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "FontFamilyName" );

				_FontFamilyName = value;
			}
		}

		private bool _IsBold = IsBoldDef;
		/// <summary>
		/// If this is true then the font with which to write the text on the image is bold. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public bool IsBold
		{
			get { return _IsBold; }
			protected set { _IsBold = value; }
		}

		private bool _IsItalic = IsItalicDef;
		/// <summary>
		/// If this is true then the font with which to write the text on the image is italic. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public bool IsItalic
		{
			get { return _IsItalic; }
			protected set { _IsItalic = value; }
		}

		/// <summary>
		/// The color of the font with which to write the text on the image. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string FontColorText
		{
			get { return FontColor.ToString(); }
			protected set
			{
				ValidationUtils.TextHasContent( value, "FontColorText" );

				_FontColor = ControlsUtils.GetColorFromText( value );
			}
		}

		/// <summary>
		/// The color of the outline of the font with which to write the text on the image. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string OutlineColorText
		{
			get { return OutlineColor.ToString(); }
			protected set
			{
				ValidationUtils.TextHasContent( value, "OutlineColorText" );

				_OutlineColor = ControlsUtils.GetColorFromText( value );
			}
		}

		#endregion Public serializable properties
	}
}
