////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Globalization;
using System.IO;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Media.Imaging;
using Vela.Kernel;

namespace Vela.Controls
{
	/// <summary>
	/// Graphics utilities.
	/// </summary>
	public static class GraphicUtils
	{
		#region Private constants

		/// <summary>
		/// The metadata query for an image width.
		/// </summary>
		private const string ImageWidthMetadataQuery = "/logscrdesc/Width";

		/// <summary>
		/// The metadata query for an image height.
		/// </summary>
		private const string ImageHeightMetadataQuery = "/logscrdesc/Height";

		/// <summary>
		/// The metadata query for a frame's left corner shift.
		/// </summary>
		private const string FrameLeftMetadataQuery = "/imgdesc/Left";

		/// <summary>
		/// The metadata query for a frame's top corner shift.
		/// </summary>
		private const string FrameTopMetadataQuery = "/imgdesc/Top";

		/// <summary>
		/// The metadata query for a frame width.
		/// </summary>
		private const string FrameWidthMetadataQuery = "/imgdesc/Width";

		/// <summary>
		/// The metadata query for a frame height.
		/// </summary>
		private const string FrameHeightMetadataQuery = "/imgdesc/Height";

		/// <summary>
		/// The metadata query for the delay (in 1 / 100 seconds) between consecutive frames of a multi-frame image.
		/// </summary>
		private const string FrameDelayMetadataQuery = "/grctlext/Delay";

		/// <summary>
		/// The default frame delay (in 1 / 100 seconds) between consecutive frames of a multi-frame image.
		/// </summary>
		private const int FrameDelayDef = 10;

		/// <summary>
		/// The metadata query for the disposal of a frame of a multi-frame image.
		/// </summary>
		private const string FrameDisposalMetadataQuery = "/grctlext/Disposal";

		#endregion Private constants

		#region Private static methods

		/// <summary>
		/// Copies the safe metadata from the source to the destination image.
		/// </summary>
		/// <param name="source">Image from which to copy the metadata.</param>
		/// <param name="metadataContainerFormat">The format of the image which will contain the metadata, like "jpg".</param>
		/// <returns>Returns the safe metadata, or null if the source image has no metadata.</returns>
		private static BitmapMetadata GetSafeMetadata( BitmapFrame source, string metadataContainerFormat )
		{
			ValidationUtils.ObjectHasContent( source, "source" );
			ValidationUtils.ObjectHasContent( metadataContainerFormat, "metadataContainerFormat" );

			BitmapMetadata sourceMetadata = source.Metadata as BitmapMetadata;
			if( sourceMetadata == null ) return null;

			BitmapMetadata destMetadata = new BitmapMetadata( metadataContainerFormat );

			if( TextUtils.HasContent( sourceMetadata.DateTaken ) ) destMetadata.DateTaken = sourceMetadata.DateTaken;

			destMetadata.Freeze();

			return destMetadata;
		}

		/// <summary>
		/// Gets the size of the specified image from its metadata.
		/// </summary>
		/// <param name="image">Image whose size to get from its metadata.</param>
		/// <param name="width">Is set with the image width.</param>
		/// <param name="height">Is set with the image height.</param>
		private static void GetMetadataImageSize( BitmapDecoder image, out int width, out int height )
		{
			ValidationUtils.ObjectHasContent( image, "image" );

			BitmapMetadata metadata = image.Metadata;
			if( (metadata == null) || (image.Frames.Count <= 0) )
			{
				if( image.Frames.Count <= 0 )
				{
					width = 0;
					height = 0;
				}
				else
				{
					width = image.Frames[ 0 ].PixelWidth;
					height = image.Frames[ 0 ].PixelHeight;
				}

				return;
			}

			object widthObj = metadata.GetQuery( ImageWidthMetadataQuery );
			object heightObj = metadata.GetQuery( ImageHeightMetadataQuery );

			width = (widthObj != null) ? (UInt16)widthObj : image.Frames[ 0 ].PixelWidth;
			height = (heightObj != null) ? (UInt16)heightObj : image.Frames[ 0 ].PixelHeight;
		}

		/// <summary>
		/// Gets the top-left corner shift and size of a frame of a multi-frame image.
		/// </summary>
		/// <param name="metadata">Metadata from which to get the information. This may be null.</param>
		/// <param name="left">Is set with the frame's left corner shift.</param>
		/// <param name="top">Is set with the frame's top corner shift.</param>
		/// <param name="width">Is set with the frame width.</param>
		/// <param name="height">Is set with the frame height.</param>
		private static void GetMetadataFrameSize( BitmapMetadata metadata, out int left, out int top, out int width
			, out int height )
		{
			if( metadata == null )
			{
				left = 0;
				top = 0;
				width = 0;
				height = 0;

				return;
			}

			object leftObj = metadata.GetQuery( FrameLeftMetadataQuery );
			object topObj = metadata.GetQuery( FrameTopMetadataQuery );
			object widthObj = metadata.GetQuery( FrameWidthMetadataQuery );
			object heightObj = metadata.GetQuery( FrameHeightMetadataQuery );

			left = (leftObj != null) ? (UInt16)leftObj : 0;
			top = (topObj != null) ? (UInt16)topObj : 0;
			width = (widthObj != null) ? (UInt16)widthObj : 0;
			height = (heightObj != null) ? (UInt16)heightObj : 0;
		}

		/// <summary>
		/// Gets the delay (in 1 / 100 seconds) between consecutive frames of a multi-frame image.
		/// </summary>
		/// <param name="metadata">Metadata from which to get the delay. This may be null.</param>
		/// <returns>Returns the frame delay.</returns>
		private static int GetMetadataFrameDelay( BitmapMetadata metadata )
		{
			if( (metadata == null) || !metadata.ContainsQuery( FrameDelayMetadataQuery ) ) return FrameDelayDef;

			object delayObj = metadata.GetQuery( FrameDelayMetadataQuery );
			if( delayObj == null ) return FrameDelayDef;

			int delay = (UInt16)delayObj;
			if( delay <= 1 ) return FrameDelayDef;

			return delay;
		}

		/// <summary>
		/// Gets the disposal of a frame of a multi-frame image.
		/// </summary>
		/// <param name="metadata">Metadata from which to get the disposal. This may be null.</param>
		/// <returns>Returns the frame disposal.</returns>
		/// <remarks>
		/// <para>Disposals: 0 - none, 1 - do not dispose, 2 - restore background color, 3 - reuse previous.</para>
		/// </remarks>
		private static int GetMetadataFrameDisposal( BitmapMetadata metadata )
		{
			if( (metadata == null) || !metadata.ContainsQuery( FrameDisposalMetadataQuery ) ) return 0;

			object disposalObj = metadata.GetQuery( FrameDisposalMetadataQuery );
			if( disposalObj == null ) return 0;

			int disposal = (byte)disposalObj;
			if( disposal <= 0 ) return 0;

			return disposal;
		}

		/// <summary>
		/// Draws the specified top frame over the specified bottom frame.
		/// </summary>
		/// <param name="imageWidth">The image's width.</param>
		/// <param name="imageHeight">The image's height.</param>
		/// <param name="bottomFrame">The frame over which to draw. This may be null.</param>
		/// <param name="topFrame">The frame to draw on top.</param>
		/// <param name="topFrameLeft">The top frame's left corner shift.</param>
		/// <param name="topFrameTop">The top frame's top corner shift.</param>
		/// <param name="topFrameWidth">The top frame's width.</param>
		/// <param name="topFrameHeight">The top frame's height.</param>
		/// <returns>Returns the blended frame.</returns>
		private static BitmapFrame DrawOver( int imageWidth, int imageHeight, BitmapSource bottomFrame, BitmapSource topFrame
			, int topFrameLeft, int topFrameTop, int topFrameWidth, int topFrameHeight )
		{
			ValidationUtils.Integer( imageWidth, 1, int.MaxValue, "imageWidth" );
			ValidationUtils.Integer( imageHeight, 1, int.MaxValue, "imageHeight" );
			ValidationUtils.ObjectHasContent( topFrame, "topFrame" );
			ValidationUtils.Integer( topFrameWidth, 1, int.MaxValue, "topFrameWidth" );
			ValidationUtils.Integer( topFrameHeight, 1, int.MaxValue, "topFrameHeight" );

			DrawingVisual drawingVisual = new DrawingVisual();
			using( DrawingContext drawingContext = drawingVisual.RenderOpen() )
			{
				if( bottomFrame != null )
					drawingContext.DrawImage( bottomFrame, new Rect( 0, 0, bottomFrame.PixelWidth, bottomFrame.PixelHeight ) );

				drawingContext.DrawImage( topFrame, new Rect( topFrameLeft, topFrameTop, topFrameWidth, topFrameHeight ) );
			}

			RenderTargetBitmap blendedFrame = new RenderTargetBitmap( imageWidth, imageHeight, 96, 96, PixelFormats.Pbgra32 );

			blendedFrame.Render( drawingVisual );
			blendedFrame.Freeze();

			BitmapFrame bitmapFrame = BitmapFrame.Create( blendedFrame );
			bitmapFrame.Freeze();

			return bitmapFrame;
		}

		/// <summary>
		/// Draws the specified top frame over an image background.
		/// </summary>
		/// <param name="imageWidth">The image's width.</param>
		/// <param name="imageHeight">The image's height.</param>
		/// <param name="topFrame">The frame to draw on top.</param>
		/// <param name="topFrameLeft">The top frame's left corner shift.</param>
		/// <param name="topFrameTop">The top frame's top corner shift.</param>
		/// <param name="topFrameWidth">The top frame's width.</param>
		/// <param name="topFrameHeight">The top frame's height.</param>
		/// <returns>Returns the blended frame.</returns>
		private static BitmapFrame DrawOverBackground( int imageWidth, int imageHeight, BitmapSource topFrame, int topFrameLeft
			, int topFrameTop, int topFrameWidth, int topFrameHeight )
		{
			ValidationUtils.Integer( imageWidth, 1, int.MaxValue, "imageWidth" );
			ValidationUtils.Integer( imageHeight, 1, int.MaxValue, "imageHeight" );
			ValidationUtils.ObjectHasContent( topFrame, "topFrame" );
			ValidationUtils.Integer( topFrameWidth, 1, int.MaxValue, "topFrameWidth" );
			ValidationUtils.Integer( topFrameHeight, 1, int.MaxValue, "topFrameHeight" );

			DrawingVisual drawingVisual = new DrawingVisual();
			using( DrawingContext drawingContext = drawingVisual.RenderOpen() )
			{
				Rect imageRect = new Rect( 0, 0, imageWidth, imageHeight );
				RectangleGeometry imageRectGeometry = new RectangleGeometry( imageRect );
				imageRectGeometry.Freeze();

				Rect frameRect = new Rect( topFrameLeft, topFrameTop, topFrameWidth, topFrameHeight );
				RectangleGeometry frameRectGeometry = new RectangleGeometry( frameRect );
				frameRectGeometry.Freeze();

				PathGeometry clipGeometry = Geometry.Combine( imageRectGeometry, frameRectGeometry, GeometryCombineMode.Exclude
					, null );
				clipGeometry.Freeze();

				drawingContext.PushClip( clipGeometry );
				drawingContext.DrawImage( topFrame, imageRect );
			}

			RenderTargetBitmap blendedFrame = new RenderTargetBitmap( imageWidth, imageHeight, 96, 96, PixelFormats.Pbgra32 );

			blendedFrame.Render( drawingVisual );
			blendedFrame.Freeze();

			BitmapFrame bitmapFrame = BitmapFrame.Create( blendedFrame );
			bitmapFrame.Freeze();

			return bitmapFrame;
		}

		/// <summary>
		/// Draws the specified text on the specified image.
		/// </summary>
		/// <param name="textToWriteOnImage">The text to write on the image. This may be null or empty.</param>
		/// <param name="textOnImageConfiguration">Configuration for the text which is to be written on the resized image.
		/// </param>
		/// <param name="workArea">The rectangle in which the text must be written.</param>
		/// <param name="drawingContext">The image in which to draw the text.</param>
		/// <param name="destinationWidth">The width to which the image will be resized later. This must be at least 1.</param>
		/// <param name="destinationHeight">The height to which the image will be resized later. This must be at least 1.</param>
		/// <param name="targetWidth">The width to which the image will be resized later. This must be at least 1.</param>
		/// <remarks>
		/// <para>The text is right-aligned relative to the bottom-right corner of the image.</para>
		/// </remarks>
		private static void DrawTextOnImage( string textToWriteOnImage, TextOnImageConfiguration textOnImageConfiguration
			, Rect workArea, DrawingContext drawingContext, int destinationWidth, int destinationHeight, int targetWidth )
		{
			ValidationUtils.ObjectHasContent( textOnImageConfiguration, "textOnImageConfiguration" );
			ValidationUtils.ObjectHasContent( drawingContext, "drawingContext" );
			ValidationUtils.Integer( destinationWidth, 1, int.MaxValue, "destinationWidth" );
			ValidationUtils.Integer( destinationHeight, 1, int.MaxValue, "destinationHeight" );
			ValidationUtils.Integer( targetWidth, 1, int.MaxValue, "targetWidth" );

			textToWriteOnImage = (textToWriteOnImage != null) ? textToWriteOnImage.Trim() : null;

			if( TextUtils.NoContent( textToWriteOnImage )
				|| (textToWriteOnImage.Length < textOnImageConfiguration.WriteTextIfSizeMin) )
			{
				return;
			}

			CultureInfo cultureInfo = CultureInfo.CurrentUICulture;
			bool isRightToLeft = CultureInfo.CurrentUICulture.TextInfo.IsRightToLeft;
			FlowDirection flowDirection = isRightToLeft ? FlowDirection.RightToLeft : FlowDirection.LeftToRight;

			TextOnImageConfiguration toic = textOnImageConfiguration;
			int sizeReference = Math.Min( destinationWidth, destinationHeight );

			// The fewer characters there are in the text, the greater the font size will be.
			double amplifierX = textOnImageConfiguration.WriteTextIfSizeMin / (double)textToWriteOnImage.Length;
			double fontSizeAmplifier = 1 + Math.Pow( amplifierX, 0.5 );

			int horizontalMargin = (int)(toic.HorizontalMarginFraction * sizeReference);
			int verticalMargin = (int)(toic.VerticalMarginFraction * sizeReference);
			int frameWidthMax = Math.Min( destinationWidth, (int)(toic.FrameWidthFractionMax * targetWidth) );
			int fontSize = Math.Max( 1, (int)(toic.FontSizeFraction * sizeReference * fontSizeAmplifier) );
			double outlineThickness = Math.Max( 1, toic.OutlineThicknessFraction * fontSize );

			SolidColorBrush fontBrush = new SolidColorBrush( toic.FontColor );
			fontBrush.Freeze();

			SolidColorBrush outlinePenBrush = new SolidColorBrush( toic.OutlineColor );
			outlinePenBrush.Freeze();
			Pen outlinePen = new Pen( outlinePenBrush, outlineThickness );
			outlinePen.Freeze();

			FontFamily fontFamily = new FontFamily( toic.FontFamilyName );
			Typeface typeface = new Typeface( fontFamily, toic.IsItalic ? FontStyles.Italic : FontStyles.Normal
				, toic.IsBold ? FontWeights.Bold : FontWeights.Normal, FontStretches.Normal );

			FormattedText formattedText = new FormattedText( textToWriteOnImage, cultureInfo, flowDirection, typeface
				, fontSize, fontBrush );
			formattedText.Trimming = TextTrimming.None;
			formattedText.MaxTextWidth = frameWidthMax;
			formattedText.TextAlignment = (toic.FrameAlignment == TextOnImageAlignments.BottomCenter)
				? TextAlignment.Center : TextAlignment.Right;

			int textFrameRight = Math.Min( (int)workArea.Right, destinationWidth );
			int textFrameBottom = Math.Min( (int)workArea.Bottom, destinationHeight );

			int textLeftCorner = (toic.FrameAlignment == TextOnImageAlignments.BottomCenter)
				? ((textFrameRight - frameWidthMax) / 2) : (textFrameRight - horizontalMargin - frameWidthMax);

			Point textLeftTopCorner = new Point( textLeftCorner, textFrameBottom - verticalMargin - (int)formattedText.Height );

			Geometry textGeometry = formattedText.BuildGeometry( textLeftTopCorner );
			textGeometry.Freeze();
			drawingContext.DrawGeometry( fontBrush, outlinePen, textGeometry );
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Normalizes the size of an image.
		/// </summary>
		/// <param name="originalWidth">Original width of image. This must be at least 1.</param>
		/// <param name="originalHeight">Original height of image. This must be at least 1.</param>
		/// <param name="normalizeToWidth">Width to which to normalize the image size. This must be at least 1.</param>
		/// <param name="normalizeToHeight">Height to which to normalize the image size. This must be at least 1.</param>
		/// <param name="stretch">If this is true then the image must be stretched, that is, the aspect ratio may be modified
		/// during the resize.</param>
		/// <param name="stretchSkewMax">If the aspect ratio of the stretched image would exceed this skew, the image would not
		/// be streched. This is used only if "stretch" is true.</param>
		/// <param name="resizeSmaller">If this is true then an image which is smaller than the size to which it must be
		/// normalized, is resized. This is ignored if "stretch" is true.</param>
		/// <param name="sizeToResizeMin">The image is resized or streched only if its size
		/// (= width and height) is at least this value. If this is 0 or a negative number, it's ignored.</param>
		/// <param name="newWidth">Is set with the width of the resized image.</param>
		/// <param name="newHeight">Is set with the height of the resized image.</param>
		/// <param name="skew">Is set with the skew of the resized image.</param>
		/// <returns>Returns true if the image is stretched (instead of resized).</returns>
		public static bool NormalizeImageSize( int originalWidth, int originalHeight, int normalizeToWidth
			, int normalizeToHeight, bool stretch, double stretchSkewMax, bool resizeSmaller, int sizeToResizeMin
			, out int newWidth, out int newHeight, out double skew )
		{
			ValidationUtils.Integer( originalWidth, 1, int.MaxValue, "originalWidth" );
			ValidationUtils.Integer( originalHeight, 1, int.MaxValue, "originalHeight" );
			ValidationUtils.Integer( normalizeToWidth, 1, int.MaxValue, "normalizeToWidth" );
			ValidationUtils.Integer( normalizeToHeight, 1, int.MaxValue, "normalizeToHeight" );

			newWidth = 0;
			newHeight = 0;
			skew = 1;

			if( (originalWidth <= 0) || (originalHeight <= 0) ) return false;

			if( sizeToResizeMin > 0 )
			{
				bool isMinSize = (originalWidth >= sizeToResizeMin) && (originalHeight >= sizeToResizeMin);

				stretch = stretch && isMinSize;
				resizeSmaller = resizeSmaller && isMinSize;
			}

			// Image aspect ratio and skew.
			double imageAr = (double)originalWidth / originalHeight;
			double ownerAr = (double)normalizeToWidth / normalizeToHeight;
			double actualSkew = imageAr / ownerAr;
			if( actualSkew < 1 ) actualSkew = 1 / actualSkew;
			skew = actualSkew;

			// If the maximum allowed skew is exceeded, the image is not streched.
			if( stretch && ((actualSkew <= stretchSkewMax) || (stretchSkewMax <= 0)) )
			{
				newWidth = normalizeToWidth;
				newHeight = normalizeToHeight;

				return true;
			}

			newWidth = resizeSmaller ? normalizeToWidth : Math.Min( normalizeToWidth, originalWidth );
			newHeight = (newWidth != originalWidth) ? (int)((double)originalHeight * newWidth / originalWidth)
				: originalHeight;

			if( newHeight > normalizeToHeight )
			{
				newHeight = resizeSmaller ? normalizeToHeight : Math.Min( normalizeToHeight, originalHeight );
				newWidth = (newHeight != originalHeight) ? (int)((double)originalWidth * newHeight / originalHeight)
					: originalWidth;
			}

			return false;
		}

		/// <summary>
		/// Gets the ratio between the normalized (to the screen) size of the image with the specified size and the screen size.
		/// </summary>
		/// <param name="originalWidth">Original width of image. This must be at least 1.</param>
		/// <param name="originalHeight">Original height of image. This must be at least 1.</param>
		/// <returns>Returns the occupied space ratio.</returns>
		public static double GetOccupiedSpaceRatioAfterNormalizeImageSizeToScreen( int originalWidth, int originalHeight )
		{
			ValidationUtils.Integer( originalWidth, 1, int.MaxValue, "originalWidth" );
			ValidationUtils.Integer( originalHeight, 1, int.MaxValue, "originalHeight" );

			int screenWidth = (int)SystemParameters.PrimaryScreenWidth;
			int screenHeight = (int)SystemParameters.PrimaryScreenHeight;

			int newWidth, newHeight;
			double skew;

			NormalizeImageSize( originalWidth, originalHeight, screenWidth, screenHeight, false, 0, true, 0, out newWidth
				, out newHeight, out skew );

			return (double)(newWidth * newHeight) / (screenWidth * screenHeight);
		}

		/// <summary>
		/// Loads the image from the specified file name.
		/// </summary>
		/// <param name="fileName">The name of the file from which to load the image.</param>
		/// <param name="widthMax">The maximum width that the loaded image may have.</param>
		/// <param name="heightMax">The maximum height that the loaded image may have.</param>
		/// <returns>Returns the loaded image. It's guaranteed that the returned image has at least 1 pixel on each axis.
		/// </returns>
		public static BitmapFrame LoadImage( string fileName, int widthMax = KernelConsts.ImageWidthMax
			, int heightMax = KernelConsts.ImageHeightMax )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );
			ValidationUtils.Integer( widthMax, 1, int.MaxValue, "widthMax" );
			ValidationUtils.Integer( heightMax, 1, int.MaxValue, "heightMax" );

			// DotNet locks the image files (until the application is closed) if they are loaded with
			// <see cref="BitmapCacheOption.None"/>, and caches the entire image files when they are loaded with
			// <see cref="BitmapCacheOption.OnLoad"/>, and so first load the image data in a memory stream.

			BitmapFrame bitmapFrame;

			Uri fileUri = new Uri( fileName );
			if( fileUri.IsFile )
			{
				using( FileStream fs = new FileStream( fileName, FileMode.Open, FileAccess.Read, FileShare.ReadWrite ) )
				{
					byte[] fileBuffer = new byte[ fs.Length ];
					fs.Read( fileBuffer, 0, (int)fs.Length );

					MemoryStream ms = new MemoryStream( fileBuffer );
					bitmapFrame = BitmapFrame.Create( ms, BitmapCreateOptions.PreservePixelFormat
						, BitmapCacheOption.None );
				}
			}
			else
			{
				bitmapFrame = BitmapFrame.Create( fileUri, BitmapCreateOptions.PreservePixelFormat
					, BitmapCacheOption.None );
			}

			bitmapFrame.Freeze();

			ValidationUtils.Integer( (int)bitmapFrame.PixelWidth, 1, widthMax, "bf.PixelWidth" );
			ValidationUtils.Integer( (int)bitmapFrame.PixelHeight, 1, heightMax, "bf.PixelHeight" );

			return bitmapFrame;
		}

		/// <summary>
		/// Loads the image from the specified stream.
		/// </summary>
		/// <param name="imageStream">Stream from which to load the image.</param>
		/// <param name="widthMax">The maximum width that the loaded image may have.</param>
		/// <param name="heightMax">The maximum height that the loaded image may have.</param>
		/// <returns>Returns the loaded image. It's guaranteed that the returned image has at least 1 pixel on each axis.
		/// </returns>
		public static BitmapFrame LoadImage( Stream imageStream, int widthMax = KernelConsts.ImageWidthMax
			, int heightMax = KernelConsts.ImageHeightMax )
		{
			ValidationUtils.ObjectHasContent( imageStream, "imageStream" );
			ValidationUtils.Integer( widthMax, 1, int.MaxValue, "widthMax" );
			ValidationUtils.Integer( heightMax, 1, int.MaxValue, "heightMax" );

			// Since the stream may soon be disposed, the image must be loaded now.
			BitmapFrame bitmapFrame = BitmapFrame.Create( imageStream, BitmapCreateOptions.PreservePixelFormat
				, BitmapCacheOption.OnLoad );
			bitmapFrame.Freeze();

			ValidationUtils.Integer( (int)bitmapFrame.PixelWidth, 1, widthMax, "bitmapFrame.PixelWidth" );
			ValidationUtils.Integer( (int)bitmapFrame.PixelHeight, 1, heightMax, "bitmapFrame.PixelHeight" );

			return bitmapFrame;
		}

		/// <summary>
		/// Loads the image from the specified file name.
		/// </summary>
		/// <param name="fileName">The name of the file from which to load the image.</param>
		/// <param name="widthMax">The maximum width that the loaded image may have.</param>
		/// <param name="heightMax">The maximum height that the loaded image may have.</param>
		/// <returns>Returns the loaded image. It's guaranteed that the returned image has at least 1 pixel on each axis.
		/// </returns>
		/// <remarks>
		/// <para>The loaded image may have multiple frames.</para>
		/// </remarks>
		public static BitmapDecoder LoadMultiFrameImage( string fileName, int widthMax = KernelConsts.ImageWidthMax
			, int heightMax = KernelConsts.ImageHeightMax )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );
			ValidationUtils.Integer( widthMax, 1, int.MaxValue, "widthMax" );
			ValidationUtils.Integer( heightMax, 1, int.MaxValue, "heightMax" );

			// DotNet locks the image files (until the application is closed) if they are loaded with
			// <see cref="BitmapCacheOption.None"/>, and caches the entire image files when they are loaded with
			// <see cref="BitmapCacheOption.OnLoad"/>, and so first load the image data in a memory stream.

			BitmapDecoder bitmapDecoder;

			Uri fileUri = new Uri( fileName );
			if( fileUri.IsFile )
			{
				using( FileStream fs = new FileStream( fileName, FileMode.Open, FileAccess.Read, FileShare.ReadWrite ) )
				{
					byte[] fileBuffer = new byte[ fs.Length ];
					fs.Read( fileBuffer, 0, (int)fs.Length );

					MemoryStream ms = new MemoryStream( fileBuffer );
					bitmapDecoder = BitmapDecoder.Create( ms, BitmapCreateOptions.PreservePixelFormat
						, BitmapCacheOption.None );
				}
			}
			else
			{
				bitmapDecoder = BitmapDecoder.Create( fileUri, BitmapCreateOptions.PreservePixelFormat
					, BitmapCacheOption.None );
			}

			foreach( BitmapFrame frame in bitmapDecoder.Frames )
			{
				ValidationUtils.Integer( (int)frame.PixelWidth, 1, widthMax, "frame.PixelWidth" );
				ValidationUtils.Integer( (int)frame.PixelHeight, 1, heightMax, "frame.PixelHeight" );

				frame.Freeze();
			}

			return bitmapDecoder;
		}

		/// <summary>
		/// Gets the size of the image from the specified file.
		/// </summary>
		/// <param name="fileName">Name of file from which to load the image.</param>
		/// <param name="width">Is set with the width of the loaded image.</param>
		/// <param name="height">Is set with the height of the loaded image.</param>
		public static void GetImageSize( string fileName, out int width, out int height )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			BitmapFrame bitmapFrame = LoadImage( fileName );

			width = bitmapFrame.PixelWidth;
			height = bitmapFrame.PixelHeight;
		}

		/// <summary>
		/// Resizes the specified image.
		/// </summary>
		/// <param name="image">The image to resize.</param>
		/// <param name="targetWidth">The width to which to resize the image. This must be at least 1.</param>
		/// <param name="targetHeight">The height to which to resize the image. This must be at least 1.</param>
		/// <param name="copySafeMetadata">If this is true then the safe metadata is copied in the resized image, else no
		/// metadata is copied.</param>
		/// <returns>Returns the resized image.</returns>
		public static BitmapFrame ResizeImage( BitmapFrame image, int targetWidth, int targetHeight, bool copySafeMetadata )
		{
			ValidationUtils.ObjectHasContent( image, "image" );
			ValidationUtils.Integer( targetWidth, 1, int.MaxValue, "targetWidth" );
			ValidationUtils.Integer( targetHeight, 1, int.MaxValue, "targetHeight" );

			// Get the safe metadata.
			BitmapMetadata metadata = copySafeMetadata ? GetSafeMetadata( image, "jpg" ) : null;

			// Get the new image size.
			int newWidth, newHeight;
			double skew;
			NormalizeImageSize( image.PixelWidth, image.PixelHeight, targetWidth, targetHeight, false, 0, false, 0
				, out newWidth, out newHeight, out skew );

			if( (image.PixelWidth == newWidth) && (image.PixelHeight == newHeight) ) return image;

			#region Resize image

			// This scaling method has a quality between "Linear" and "HighQuality".

			ScaleTransform scaleTransform = new ScaleTransform( (double)newWidth / image.PixelWidth
				, (double)newHeight / image.PixelHeight, 0, 0 );
			scaleTransform.Freeze();

			TransformedBitmap transformedBitmap = new TransformedBitmap( image, scaleTransform );
			transformedBitmap.Freeze();

			BitmapFrame resizedFrame = BitmapFrame.Create( transformedBitmap, null, metadata, null );
			resizedFrame.Freeze();

			return resizedFrame;

			#endregion Resize image
		}

		/// <summary>
		/// Loads and resizes the image from the specified file.
		/// </summary>
		/// <param name="fileName">Name of file from which to load the image.</param>
		/// <param name="targetWidth">The width to which to resize the loaded image. This must be at least 1.</param>
		/// <param name="targetHeight">The height to which to resize the loaded image. This must be at least 1.</param>
		/// <param name="copySafeMetadata">If this is true then the safe metadata is copied in the resized image, else no
		/// metadata is copied.</param>
		/// <returns>Returns the resized image.</returns>
		public static BitmapFrame LoadResizeImage( string fileName, int targetWidth, int targetHeight, bool copySafeMetadata )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );
			ValidationUtils.Integer( targetWidth, 1, int.MaxValue, "targetWidth" );
			ValidationUtils.Integer( targetHeight, 1, int.MaxValue, "targetHeight" );

			// Load the image from the file.
			BitmapFrame bitmapFrame = LoadImage( fileName );

			return ResizeImage( bitmapFrame, targetWidth, targetHeight, copySafeMetadata );
		}

		/// <summary>
		/// Loads and resizes the image from the specified file.
		/// </summary>
		/// <param name="fileName">Name of file from which to load the image.</param>
		/// <param name="targetWidth">The width to which to resize the loaded image. This must be at least 1.</param>
		/// <param name="targetHeight">The height to which to resize the loaded image. This must be at least 1.</param>
		/// <param name="copySafeMetadata">If this is true then the safe metadata is copied in the resized image, else no
		/// metadata is copied.</param>
		/// <param name="quality">The compression quality of the resized image. This must be at least 1.</param>
		/// <returns>Returns the buffer of the image.</returns>
		public static byte[] LoadResizeImageBuffer( string fileName, int targetWidth, int targetHeight, bool copySafeMetadata
			, int quality )
		{
			ValidationUtils.Integer( quality, 1, 100, "quality" );

			BitmapFrame resizedBitmapFrame = LoadResizeImage( fileName, targetWidth, targetHeight, copySafeMetadata );

			using( MemoryStream saveStream = new MemoryStream() )
			{
				JpegBitmapEncoder bitmapEncoder = new JpegBitmapEncoder();
				bitmapEncoder.Frames.Add( resizedBitmapFrame );
				bitmapEncoder.QualityLevel = quality;
				bitmapEncoder.Save( saveStream );

				return saveStream.ToArray();
			}
		}

		/// <summary>
		/// Loads and resizes the image from the specified file.
		/// </summary>
		/// <param name="fileName">Name of file from which to load the image.</param>
		/// <param name="targetWidth">The width to which to resize the loaded image. This must be at least 1.</param>
		/// <param name="targetHeight">The height to which to resize the loaded image. This must be at least 1.</param>
		/// <param name="copySafeMetadata">If this is true then the safe metadata is copied in the resized image, else no
		/// metadata is copied.</param>
		/// <returns>Returns the resized image.</returns>
		/// <remarks>
		/// <para>The loaded image may have multiple frames.</para>
		/// </remarks>
		public static ObjectAnimationUsingKeyFrames LoadAndResizeMultiFrameImage( string fileName, int targetWidth
			, int targetHeight, bool copySafeMetadata )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );
			ValidationUtils.Integer( targetWidth, 1, int.MaxValue, "targetWidth" );
			ValidationUtils.Integer( targetHeight, 1, int.MaxValue, "targetHeight" );

			BitmapDecoder bitmapDecoder = LoadMultiFrameImage( fileName );

			int imageWidth, imageHeight;
			GetMetadataImageSize( bitmapDecoder, out imageWidth, out imageHeight );

			BitmapSource baseFrame = null;
			ObjectAnimationUsingKeyFrames image = new ObjectAnimationUsingKeyFrames();
			TimeSpan keyTime = TimeSpan.Zero;
			TimeSpan totalTimespan = TimeSpan.Zero;

			foreach( BitmapFrame loadedFrame in bitmapDecoder.Frames )
			{
				BitmapMetadata frameMetadata = loadedFrame.Metadata as BitmapMetadata;

				int metadataLeft, metadataTop, metadataWidth, metadataHeight;

				if( bitmapDecoder.Frames.Count > 1 )
				{
					GetMetadataFrameSize( frameMetadata, out metadataLeft, out metadataTop, out metadataWidth
						, out metadataHeight );
				}
				else
				{
					metadataLeft = 0;
					metadataTop = 0;
					metadataWidth = loadedFrame.PixelWidth;
					metadataHeight = loadedFrame.PixelHeight;
				}

				bool isFullFrame = (metadataLeft == 0) && (metadataTop == 0) && (imageWidth == metadataWidth)
					&& (imageHeight == metadataHeight);
				bool showFrame = (metadataLeft >= 0) && (metadataTop >= 0) && (metadataWidth >= 1) && (metadataHeight >= 1);

				if( showFrame )
				{
					// Draw a partial frame in a temporary image-sized frame.
					BitmapFrame blendedFrame = (isFullFrame && (baseFrame == null)) ? loadedFrame
						: DrawOver( imageWidth, imageHeight, baseFrame, loadedFrame, metadataLeft, metadataTop, metadataWidth
							, metadataHeight );

					BitmapFrame resizedFrame = ResizeImage( blendedFrame, targetWidth, targetHeight, copySafeMetadata );

					int disposal = GetMetadataFrameDisposal( frameMetadata );
					switch( disposal )
					{
						case 0: // None.
						case 1: // Do not dispose.
							baseFrame = blendedFrame;
							break;
						case 2: // Restore background color.
							baseFrame = isFullFrame ? null
								: DrawOverBackground( imageWidth, imageHeight, blendedFrame, metadataLeft, metadataTop
									, metadataWidth, metadataHeight );
							break;
						case 3: // Reuse previous.
						default:
							break;
					}

					DiscreteObjectKeyFrame keyFrame = new DiscreteObjectKeyFrame( resizedFrame, keyTime );
					keyFrame.Freeze();

					image.KeyFrames.Add( keyFrame );
				}
				else if( baseFrame != null )
				{
					DiscreteObjectKeyFrame keyFrame = new DiscreteObjectKeyFrame( baseFrame, keyTime );
					keyFrame.Freeze();

					image.KeyFrames.Add( keyFrame );
				}
				else continue;

				// Get the frame rate timespan.
				int delay = GraphicUtils.GetMetadataFrameDelay( frameMetadata );
				TimeSpan frameTimespan = TimeSpan.FromMilliseconds( 10 * delay );

				keyTime += frameTimespan;
				totalTimespan += frameTimespan;
			}

			if( image.KeyFrames.Count > 1 )
			{
				image.Duration = new Duration( totalTimespan );
				image.RepeatBehavior = RepeatBehavior.Forever;
			}

			image.KeyFrames.Freeze();
			image.Freeze();

			return image;
		}

		/// <summary>
		/// Loads, resizes and saves (as JPEG) the image from the specified file.
		/// </summary>
		/// <param name="fileName">Name of file from which to load the image.</param>
		/// <param name="targetWidth">The width to which to resize the loaded image. This must be at least 1.</param>
		/// <param name="targetHeight">The height to which to resize the loaded image. This must be at least 1.</param>
		/// <param name="fillScalingModeCoverageMin">If the image scaled with the "fit" mode to the specified size occupies at
		/// least this fraction of the specified size, the used scaling mode is "fill" (else it remains "fit").</param>
		/// <param name="quality">The compression quality of the resized image. This must be at least 1.</param>
		/// <param name="textToWriteOnImage">The text to write on the image. This may be null or empty.</param>
		/// <param name="textOnImageConfiguration">Configuration for the text which is to be written on the resized image.
		/// </param>
		/// <param name="workArea">The rectangle in which the text must be written.</param>
		/// <param name="saveFileName">The name of the file in which to save the resized image. If this is null or empty, the
		/// resized image is not saved to a file.</param>
		/// <remarks>
		/// <para>The text is right-aligned relative to the bottom-right corner of the image.</para>
		/// </remarks>
		public static void SaveResizedImage( string fileName, int targetWidth, int targetHeight
			, double fillScalingModeCoverageMin, int quality, string textToWriteOnImage
			, TextOnImageConfiguration textOnImageConfiguration, Rect workArea, string saveFileName )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );
			ValidationUtils.Integer( targetWidth, 1, int.MaxValue, "targetWidth" );
			ValidationUtils.Integer( targetHeight, 1, int.MaxValue, "targetHeight" );
			ValidationUtils.Real( fillScalingModeCoverageMin, 0, 1, "fillScalingModeCoverageMin" );
			ValidationUtils.Integer( quality, 1, 100, "quality" );
			ValidationUtils.TextHasContent( saveFileName, "saveFileName" );

			// Load the image from the file.
			BitmapFrame bitmapFrame = LoadImage( fileName );

			#region Determine new image size

			// Get the new image size.
			int newImageWidth, newImageHeight;
			double skew;
			NormalizeImageSize( bitmapFrame.PixelWidth, bitmapFrame.PixelHeight, targetWidth, targetHeight, false, 0, true, 0
				, out newImageWidth, out newImageHeight, out skew );

			int destinationX = 0, destinationY = 0, destinationWidth = newImageWidth, destinationHeight = newImageHeight;

			double occupiedSpaceRatio = (double)(newImageWidth * newImageHeight) / (targetWidth * targetHeight);
			if( occupiedSpaceRatio >= fillScalingModeCoverageMin )
			{
				if( newImageWidth == targetWidth )
				{
					int fillHeight = targetHeight;
					int fillWidth = (targetWidth * targetHeight) / newImageHeight;

					newImageWidth = fillWidth;
					newImageHeight = targetHeight;

					destinationX = (targetWidth - fillWidth) / 2;
				}
				else
				{
					int fillWidth = targetWidth;
					int fillHeight = (targetHeight * targetWidth) / newImageWidth;

					newImageWidth = targetWidth;
					newImageHeight = fillHeight;

					destinationY = (targetHeight - fillHeight) / 2;
				}

				destinationWidth = targetWidth;
				destinationHeight = targetHeight;
			}

			#endregion Determine new image size

			#region Resize image

			BitmapFrame resizedFrame;

			Rect destinationRect = new Rect( destinationX, destinationY, newImageWidth, newImageHeight );
			ImageDrawing imageDrawing = new ImageDrawing( bitmapFrame, destinationRect );
			imageDrawing.Freeze();

			DrawingGroup group = new DrawingGroup();
			RenderOptions.SetBitmapScalingMode( group, BitmapScalingMode.HighQuality );
			group.Children.Add( imageDrawing );
			group.Freeze();

			DrawingVisual drawingVisual = new DrawingVisual();
			using( DrawingContext drawingContext = drawingVisual.RenderOpen() )
			{
				drawingContext.DrawDrawing( group );

				DrawTextOnImage( textToWriteOnImage, textOnImageConfiguration, workArea, drawingContext, destinationWidth
					, destinationHeight, targetWidth );
			}

			RenderTargetBitmap transformedBitmap = new RenderTargetBitmap( destinationWidth, destinationHeight, 96, 96
				, PixelFormats.Pbgra32 );
			transformedBitmap.Render( drawingVisual );
			transformedBitmap.Freeze();

			resizedFrame = BitmapFrame.Create( transformedBitmap );
			resizedFrame.Freeze();

			#endregion Resize image

			using( FileStream saveStream = new FileStream( saveFileName, FileMode.Create ) )
			{
				JpegBitmapEncoder bitmapEncoder = new JpegBitmapEncoder();
				bitmapEncoder.Frames.Add( resizedFrame );
				bitmapEncoder.QualityLevel = quality;
				bitmapEncoder.Save( saveStream );
			}
		}

		/// <summary>
		/// Saves the specified image in the specified file.
		/// </summary>
		/// <param name="image">The image to save.</param>
		/// <param name="fileName">The name of the file in which to save the image.</param>
		/// <param name="quality">The compression quality of the saved image. If this is 0 or a negative number, the image is
		/// saved in a lossless format.</param>
		public static void SaveImage( BitmapSource image, string fileName, int quality )
		{
			ValidationUtils.ObjectHasContent( image, "image" );
			ValidationUtils.TextHasContent( fileName, "fileName" );

			BitmapFrame bitmapFrame = BitmapFrame.Create( image );
			bitmapFrame.Freeze();

			if( quality > 0 )
			{
				ValidationUtils.Integer( quality, 1, 100, "quality" );

				using( FileStream saveStream = new FileStream( fileName, FileMode.Create, FileAccess.Write, FileShare.None ) )
				{
					JpegBitmapEncoder bitmapEncoder = new JpegBitmapEncoder();
					bitmapEncoder.Frames.Add( bitmapFrame );
					bitmapEncoder.QualityLevel = quality;
					bitmapEncoder.Save( saveStream );
				}
			}
			else
			{
				using( FileStream saveStream = new FileStream( fileName, FileMode.Create, FileAccess.Write, FileShare.None ) )
				{
					PngBitmapEncoder bitmapEncoder = new PngBitmapEncoder();
					bitmapEncoder.Interlace = PngInterlaceOption.Off;
					bitmapEncoder.Frames.Add( bitmapFrame );
					bitmapEncoder.Save( saveStream );
				}
			}
		}

		/// <summary>
		/// Gets the image from the clipboard.
		/// </summary>
		/// <param name="quality">The compression quality which will be used to save the image. If this is 0 or a negative
		/// number, the image will be saved in a lossless format.</param>
		/// <returns>Returns the image from the clipboard, or null if none exists.</returns>
		/// <remarks>
		/// <para>This is a workaround for a .Net bug where an image which is retrieved from the clipboard is saved blank /
		/// garbled in a PNG format, if it was first processed in another application.</para>
		/// 
		/// <para>For details, see http://connect.microsoft.com/VisualStudio/feedback/details/469419/pngbitmapencoder-corrupts-bgra32-images .</para>
		/// </remarks>
		public static BitmapSource GetClipboardImage( int quality )
		{
			BitmapSource clipboardImage = Clipboard.GetImage();
			if( (clipboardImage == null) || (quality > 0) ) return clipboardImage;

			PixelFormat format = clipboardImage.Format;

			if( (format != PixelFormats.Bgra32) && (format != PixelFormats.Pbgra32) && (format != PixelFormats.Prgba64)
				&& (format != PixelFormats.Rgba128Float) && (format != PixelFormats.Prgba128Float) )
			{
				return clipboardImage;
			}

			FormatConvertedBitmap convertedClipboardImage = new FormatConvertedBitmap( clipboardImage, PixelFormats.Rgb24, null
				, 0 );
			convertedClipboardImage.Freeze();

			return convertedClipboardImage;
		}

		#endregion Public static methods
	}
}
