////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;

namespace Vela.Controls
{
	/// <summary>
	/// A view registration.
	/// </summary>
	public class ViewRegistration
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ViewRegistration()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="sortKey">Sorting key.</param>
		/// <param name="title">Title.</param>
		/// <param name="tooltip">Tooltip. This may be null or empty.</param>
		/// <param name="viewType">View type.</param>
		/// <param name="model">The model of the view. This may be null.</param>
		public ViewRegistration( string sortKey, string title, string tooltip, Type viewType, IModel model )
		{
			SortKey = sortKey;
			Title = title;
			Tooltip = tooltip;
			ViewType = viewType;
			Model = model;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private string _SortKey;
		/// <summary>
		/// Sorting key. READ-ONLY.
		/// </summary>
		public string SortKey
		{
			get { return _SortKey; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "SortKey" );

				_SortKey = value;
			}
		}

		private string _Title;
		/// <summary>
		/// Title. READ-ONLY.
		/// </summary>
		public string Title
		{
			get { return _Title; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "Title" );

				_Title = value;
			}
		}

		private string _Tooltip;
		/// <summary>
		/// Tooltip. This may be null or empty. READ-ONLY.
		/// </summary>
		public string Tooltip
		{
			get { return _Tooltip; }
			protected set { _Tooltip = value; }
		}

		private Type _ViewType;
		/// <summary>
		/// View type. READ-ONLY.
		/// </summary>
		public Type ViewType
		{
			get { return _ViewType; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ViewType" );

				_ViewType = value;
				_ViewUserControl = null;
			}
		}

		private IModel _Model;
		/// <summary>
		/// The model of the view. This may be null. READ-ONLY.
		/// </summary>
		public IModel Model
		{
			get { return _Model; }
			protected set { _Model = value; }
		}

		private ViewUserControl _ViewUserControl;
		/// <summary>
		/// User control view. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public ViewUserControl ViewUserControl
		{
			get
			{
				if( _ViewUserControl != null ) return _ViewUserControl;

				_ViewUserControl = Activator.CreateInstance( ViewType, Model ) as ViewUserControl;

				return _ViewUserControl;
			}
		}

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Gets the user control view.
		/// </summary>
		/// <param name="model">The model of the view. This may be null.</param>
		/// <returns>Returns the user control view.</returns>
		public ViewUserControl GetViewUserControl( IModel model )
		{
			Model = model;

			return ViewUserControl;
		}

		#endregion Public methods
	}
}
