////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Reflection;
using Vela.Kernel;

namespace Vela.Controls
{
	/// <summary>
	/// List with all the application views.
	/// </summary>
	public class ViewRegistrations<TAttribute> : KernelSortedList<string, ViewRegistration>
		where TAttribute : ViewRegistrationAttribute
	{
		#region Public constants

		/// <summary>
		/// Property name for the title of the view. The property must be public and static.
		/// </summary>
		public const string TitlePropertyName = "Title";

		/// <summary>
		/// Property name for the tooltip of the view. The property must be public and static.
		/// </summary>
		public const string TooltipPropertyName = "Tooltip";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public ViewRegistrations()
			: base( StringInsensitiveComparer.Instance )
		{
		}

		#endregion Constructors

		#region Private static methods

		/// <summary>
		/// Checks if the specified type may be registered as a view.
		/// </summary>
		/// <param name="type">Type to check if it may be registered as a view.</param>
		/// <param name="attr">Is set with the found registration attribute, or null if none is found.</param>
		/// <returns>Returns true if the specified type may be registered as a view.</returns>
		private static bool IsRegistrable( Type type, out ViewRegistrationAttribute attr )
		{
			ValidationUtils.ObjectHasContent( type, "type" );

			attr = null;

			// The class must have an instance constructor with no parameters.
			if( !type.IsVisible || !type.IsClass || type.IsAbstract || !typeof( ViewUserControl ).IsAssignableFrom( type )
				|| (
					type.GetConstructor( BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance, null
						, Type.EmptyTypes, null ) == null
				) )
			{
				return false;
			}

			// Check the "ViewRegistrationAttribute" attribute.
			object[] attributes = type.GetCustomAttributes( typeof( TAttribute ), false );

			if( attributes != null )
			{
				foreach( object attribute in attributes )
				{
					ViewRegistrationAttribute attrTyped = attribute as ViewRegistrationAttribute;

					// In order to be registrable, the attribute type must be identical.
					if( (attrTyped != null) && (attrTyped.GetType() == typeof( TAttribute )) )
					{
						// Register the view.
						attr = attrTyped;

						return true;
					}
				}
			}

			return false;
		}

		/// <summary>
		/// Creates a view registration from the specified type.
		/// </summary>
		/// <param name="sortKey">Sorting key.</param>
		/// <param name="type">Type from which to create a view registration.</param>
		/// <returns>Returns the newly created view registration</returns>
		private static ViewRegistration CreateViewRegistration( string sortKey, Type type )
		{
			string title = null, tooltip = null;

			#region Get the title

			PropertyInfo pi = type.GetProperty( TitlePropertyName, BindingFlags.Public | BindingFlags.Static );
			ValidationUtils.ObjectHasContent( pi, "pi" );

			title = pi.GetValue( type, null ) as string;

			#endregion Get the title

			#region Get the tooltip

			pi = type.GetProperty( TooltipPropertyName, BindingFlags.Public | BindingFlags.Static );
			ValidationUtils.ObjectHasContent( pi, "pi" );

			tooltip = pi.GetValue( type, null ) as string;

			#endregion Get the tooltip

			return new ViewRegistration( sortKey, title, tooltip, type, null );
		}

		#endregion Private static methods

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para>The sorting key <see cref="ViewRegistration.SortKey"/> of the item is used as key.</para>
		/// </remarks>
		public override string GetKey( ViewRegistration item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.SortKey;
		}

		/// <summary>
		/// Registers all the views from the specified assembly.
		/// </summary>
		/// <param name="assembly">The assembly in which to look for views.</param>
		public void Register( Assembly assembly )
		{
			ValidationUtils.ObjectHasContent( assembly, "assembly" );

			// Get all the types of the assembly.
			Type[] assemblyTypes = assembly.GetTypes();

			foreach( Type type in assemblyTypes )
			{
				ViewRegistrationAttribute attr;
				if( IsRegistrable( type, out attr ) )
				{
					// Register the view.
					Add( CreateViewRegistration( attr.SortKey, type ) );
				}
			}
		}

		/// <summary>
		/// Registers all the views (= the classes with the <see cref="ViewRegistrationAttribute"/> attribute) from the
		/// application. If a view with the same sort key already exists, an exception is thrown.
		/// </summary>
		public void RegisterAssemblies()
		{
			Dictionary<string, Assembly> assemblies = AssemblyUtils.GetApplicationAssemblies();

			// Register the views from all the assemblies.
			foreach( KeyValuePair<string, Assembly> kvp in assemblies ) Register( kvp.Value );
		}

		/// <summary>
		/// Finds the view registration with the specified view type.
		/// </summary>
		/// <param name="viewType">The view type of the view registration to find.</param>
		/// <returns>Returns the found view registration, or null if none is found.</returns>
		public ViewRegistration FindByType( Type viewType )
		{
			ValidationUtils.ObjectHasContent( viewType, "viewType" );

			foreach( ViewRegistration registration in this )
			{
				if( registration.ViewType == viewType ) return registration;
			}

			return null;
		}

		#endregion Public methods
	}
}
