////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.ComponentModel;
using System.Windows;

namespace Vela.Controls
{
	/// <summary>
	/// Contract for a window view.
	/// </summary>
	public abstract class ViewWindow : Window, INotifyPropertyChanged
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public ViewWindow()
		{
			DataContext = this;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="model">The model of the view. This may be null.</param>
		public ViewWindow( IModel model )
		{
			DataContext = this;
			Model = model;
		}

		#endregion Constructors

		#region Events

		/// <summary>
		/// Event raised when a property of the current class changes. This may be null.
		/// </summary>
		public event PropertyChangedEventHandler PropertyChanged;

		#endregion Events

		#region Public non-serializable properties

		private bool _IsChanged = false;
		/// <summary>
		/// This is true if the current object has been changed and must be saved. READ-ONLY.
		/// </summary>
		public bool IsChanged
		{
			get { return _IsChanged; }
			protected set { _IsChanged = value; }
		}

		private IModel _Model;
		/// <summary>
		/// The model of the view. This may be null. Notifies property observers. READ-ONLY.
		/// </summary>
		public IModel Model
		{
			get { return _Model; }
			protected set
			{
				_Model = value;

				// Notify property observers.
				OnPropertyChanged( "Model" );
			}
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Validates the input data.
		/// </summary>
		/// <param name="errorPresenter">The control which shows that there is a validation error. This may be null.</param>
		/// <remarks>
		/// <para>If you don't validate of all the editable fields in here, it's possible that some bindings are commited even
		/// if the user cancels the edit operation. This can happen if the user edits some fields, clicks the "Okay" button, the
		/// validation fails (during binding commit) inside some properties (which are not validated in here), and then the user
		/// clicks the "Cancel" button. This is likely to matter only if those (successfully) bound properties are used outside
		/// the view.</para>
		/// </remarks>
		protected virtual void ValidateUi( FrameworkElement errorPresenter )
		{
			ClearErrorPresenter( errorPresenter );
		}

		/// <summary>
		/// Loads the user interface with the data to edit.
		/// </summary>
		protected virtual void LoadUi()
		{
			IsChanged = false;
		}

		/// <summary>
		/// Saves the user interface data into the data to edit.
		/// </summary>
		/// <param name="errorPresenter">The control which shows that there is a validation error. This may be null.</param>
		protected virtual void SaveUi( FrameworkElement errorPresenter )
		{
			UpdateContentBindingSources( errorPresenter );

			IsChanged = false;
		}

		/// <summary>
		/// Notifies property observers when a property was changed.
		/// </summary>
		/// <param name="propertyName">The name of the property which has changed. Validation is not performed.</param>
		protected virtual void OnPropertyChanged( string propertyName )
		{
			IsChanged = true;

			if( PropertyChanged != null )
			{
				PropertyChangedEventArgs eventArgs = new PropertyChangedEventArgs( propertyName );

				PropertyChanged( this, eventArgs );
			}
		}

		/// <summary>
		/// Clears the exception presenter control.
		/// </summary>
		/// <param name="errorPresenter">The control which shows that there is a validation error.</param>
		public void ClearErrorPresenter( FrameworkElement errorPresenter )
		{
			ControlsUtils.ClearErrorPresenter( this, errorPresenter );
		}

		/// <summary>
		/// Updates the source of the binding of the content of all the controls from the current view.
		/// </summary>
		/// <param name="errorPresenter">The control which shows that there is a validation error. This may be null.</param>
		protected void UpdateContentBindingSources( FrameworkElement errorPresenter )
		{
			ControlsUtils.UpdateBindingSources( this, errorPresenter );
		}

		#endregion Private methods
	}
}
