////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.Controls
{
	/// <summary>
	/// Configuration for a window's layout (position and size).
	/// </summary>
	public class WindowLayoutConfiguration : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected WindowLayoutConfiguration()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="window">Window whose layout to copy in the current configuration.</param>
		public WindowLayoutConfiguration( Window window )
		{
			ValidationUtils.ObjectHasContent( window, "window" );

			TargetWindow = window;

			SetLayout();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// Window whose layout is stored.
		/// </summary>
		private Window TargetWindow;

		#endregion Private fields

		#region Public serializable properties

		private bool _IsMaximized;
		/// <summary>
		/// If this is true then the window is maximized. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public bool IsMaximized
		{
			get { return _IsMaximized; }
			protected set { _IsMaximized = value; }
		}

		private double _WindowLeft;
		/// <summary>
		/// Window left position on the screen. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public double WindowLeft
		{
			get { return _WindowLeft; }
			protected set { _WindowLeft = value; }
		}

		private double _WindowTop;
		/// <summary>
		/// Window top position on the screen. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public double WindowTop
		{
			get { return _WindowTop; }
			protected set { _WindowTop = value; }
		}

		private double _WindowWidth;
		/// <summary>
		/// Window width on the screen. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public double WindowWidth
		{
			get { return _WindowWidth; }
			protected set { _WindowWidth = value; }
		}

		private double _WindowHeight;
		/// <summary>
		/// Window height on the screen. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public double WindowHeight
		{
			get { return _WindowHeight; }
			protected set { _WindowHeight = value; }
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Returns true if the window is sizeable (= its size can be modified by the user).
		/// </summary>
		private bool IsSizable()
		{
			return (TargetWindow.ResizeMode == ResizeMode.CanResize)
				|| (TargetWindow.ResizeMode == ResizeMode.CanResizeWithGrip);
		}

		/// <summary>
		/// Copies the layout of the target window into the current configuration.
		/// </summary>
		private void SetLayout()
		{
			IsMaximized = TargetWindow.WindowState == WindowState.Maximized;

			WindowLeft = double.IsNaN( TargetWindow.Left ) ? 0 : TargetWindow.Left;
			WindowTop = double.IsNaN( TargetWindow.Top ) ? 0 : TargetWindow.Top;
			WindowWidth = TargetWindow.Width;
			WindowHeight = TargetWindow.Height;
		}

		/// <summary>
		/// Shapes-up the current object.
		/// </summary>
		private void Shapeup()
		{
			Rect workArea = SystemParameters.WorkArea;
			int screenLeft = (int)workArea.Left;
			int screenTop = (int)workArea.Top;
			int screenWidth = (int)workArea.Width;
			int screenHeight = (int)workArea.Height;

			// Shape-up bounds.
			if( WindowLeft < screenLeft ) WindowLeft = screenLeft;
			if( WindowTop < screenTop ) WindowTop = screenTop;
			if( WindowWidth > screenWidth ) WindowWidth = screenWidth;
			if( WindowHeight > screenHeight ) WindowHeight = screenHeight;

			// Shape-up size.
			if( WindowLeft + WindowWidth > screenWidth )
				WindowLeft = Math.Max( screenLeft, screenWidth - WindowWidth );
			if( WindowTop + WindowHeight > screenHeight )
				WindowTop = Math.Max( screenTop, screenHeight - WindowHeight );
		}

		#endregion Private methods

		#region Public static methods

		/// <summary>
		/// Loads the layout of the specified window from the specified configuration. If the configuration don't exist yet,
		/// they are created and loaded with the layout of the specified window.
		/// </summary>
		/// <param name="window">Window whose layout is to be loaded.</param>
		/// <param name="layout">Layout to load in the specified window. This may be null.</param>
		/// <returns>Returns the configuration loaded with the layout of the specified window. This should be the same property
		/// as "layout" because it's possible that the window layout is changed in here (in particular when "layout" is null).
		/// </returns>
		public static WindowLayoutConfiguration LoadUi( Window window, WindowLayoutConfiguration layout )
		{
			if( layout == null ) layout = new WindowLayoutConfiguration( window );
			else layout.LoadUi( window );

			return layout;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Loads the layout of the specified window from the current configuration.
		/// </summary>
		/// <param name="window">Window whose layout is to be loaded.</param>
		public void LoadUi( Window window )
		{
			ValidationUtils.ObjectHasContent( window, "window" );

			Shapeup();

			TargetWindow = window;
			TargetWindow.SizeToContent = SizeToContent.Manual;

			TargetWindow.Left = WindowLeft;
			TargetWindow.Top = WindowTop;
			TargetWindow.Width = WindowWidth;
			TargetWindow.Height = WindowHeight;

			if( IsMaximized ) TargetWindow.WindowState = WindowState.Maximized;
		}

		/// <summary>
		/// Saves the layout of the referenced window into the current configuration.
		/// </summary>
		public void SaveUi()
		{
			// It's possible that after deserialization, "LoadUi" was not called.
			if( TargetWindow == null ) return;

			SetLayout();
		}

		#endregion Public methods
	}
}
