////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Windows;
using System.Windows.Media.Imaging;
using Vela.Kernel;

namespace Vela.Controls
{
	/// <summary>
	/// A window which shows messages.
	/// </summary>
	public partial class ChoiceWindow : ViewWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ChoiceWindow()
		{
			InitializeComponent();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="message">Message to show. This may be null or empty.</param>
		/// <param name="leftChoiceMessage">The message for the left choice.</param>
		/// <param name="rightChoiceMessage">The message for the right choice.</param>
		public ChoiceWindow( string message, string leftChoiceMessage, string rightChoiceMessage )
		{
			Message = message;
			LeftChoiceMessage = leftChoiceMessage;
			RightChoiceMessage = rightChoiceMessage;

			InitializeComponent();
		}

		#endregion Constructors

		#region Public non-serializable properties

		private string _Message;
		/// <summary>
		/// Message to show. This may be null or empty. Notifies property observers. READ-ONLY.
		/// </summary>
		public string Message
		{
			get { return _Message; }
			protected set
			{
				_Message = value;

				// Notify property observers.
				OnPropertyChanged( "Message" );
			}
		}

		private string _LeftChoiceMessage;
		/// <summary>
		/// The message for the left choice. Notifies property observers. READ-ONLY.
		/// </summary>
		public string LeftChoiceMessage
		{
			get { return _LeftChoiceMessage; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "LeftChoiceMessage" );

				_LeftChoiceMessage = value;

				// Notify property observers.
				OnPropertyChanged( "LeftChoiceMessage" );
			}
		}

		private string _RightChoiceMessage;
		/// <summary>
		/// The message for the right choice. Notifies property observers. READ-ONLY.
		/// </summary>
		public string RightChoiceMessage
		{
			get { return _RightChoiceMessage; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "RightChoiceMessage" );

				_RightChoiceMessage = value;

				// Notify property observers.
				OnPropertyChanged( "RightChoiceMessage" );
			}
		}

		private BitmapSource _ModeImage;
		/// <summary>
		/// Mode image. This may be null. Notifies property observers. READ-ONLY.
		/// </summary>
		public BitmapSource ModeImage
		{
			get
			{
				if( _ModeImage != null )
					return _ModeImage;

				string fileName = string.Format( ControlsConsts.ModeImageFileNameTemplate, MessageWindowModes.Query );
				_ModeImage = ImagePreviewsCache.GetCachedOrLoadFirstFrame( fileName, fileName, false );

				// Notify property observers. Avoid stack overflows when notifying about null values.
				if( _ModeImage != null )
					OnPropertyChanged( "ModeImage" );

				return _ModeImage;
			}
			protected set
			{
				_ModeImage = value;

				// Notify property observers.
				OnPropertyChanged( "ModeImage" );
			}
		}

		private ChoiceMessageResults _Result = ChoiceMessageResults.Cancel;
		/// <summary>
		/// The message window result. Notifies property observers. READ-ONLY.
		/// </summary>
		public ChoiceMessageResults Result
		{
			get { return _Result; }
			protected set
			{
				_Result = value;

				// Notify property observers.
				OnPropertyChanged( "Result" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void LeftButton_Click( object sender, RoutedEventArgs e )
		{
			Result = ChoiceMessageResults.Left;
			DialogResult = true;
		}

		private void RightButton_Click( object sender, RoutedEventArgs e )
		{
			Result = ChoiceMessageResults.Right;
			DialogResult = false;
		}

		private void CancelButton_Click( object sender, RoutedEventArgs e )
		{
			Result = ChoiceMessageResults.Cancel;
			DialogResult = false;
		}

		#endregion Event handlers

		#region Public static methods

		/// <summary>
		/// Shows the specified message in a modal manner.
		/// </summary>
		/// <param name="message">Message to show. This may be null or empty.</param>
		/// <param name="leftChoiceMessage">The message for the left choice.</param>
		/// <param name="rightChoiceMessage">The message for the right choice.</param>
		/// <returns>Returns which button was pressed.</returns>
		public static ChoiceMessageResults ShowDialog( string message, string leftChoiceMessage, string rightChoiceMessage )
		{
			if( ControlsUtils.GetAppDispatcher() == null )
			{
				LogManager.Add( message, null, LogRecordModes.Warning, false );

				return ChoiceMessageResults.Cancel;
			}

			ChoiceWindow window = new ChoiceWindow( message, leftChoiceMessage, rightChoiceMessage );

			ControlsUtils.ShowDialog( window );

			return window.Result;
		}

		#endregion Public static methods
	}
}
