////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows;
using System.Windows.Controls;
using Vela.Kernel;

namespace Vela.Controls
{
	/// <summary>
	/// A window which allows the user to edit a note.
	/// </summary>
	public abstract class EditNoteWindow : ViewWindow
	{
		#region Public constants

		/// <summary>
		/// The default value for <see cref="CustomTitle"/>.
		/// </summary>
		public const string CustomTitleDef = "Note";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected EditNoteWindow()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="customTitle">The window title. This may be null or empty.</param>
		/// <param name="comment">Comment to show above the note text-box. This may be null or empty.</param>
		/// <param name="note">Note to edit. This may be null or empty.</param>
		/// <param name="readOnly">If this is true then the note can't be edited.</param>
		/// <param name="contentType">The type of content of the note to edit.</param>
		/// <param name="mustHaveContent">If this is true then the note must have content.</param>
		/// <param name="noteSizeMax">The maximum size of the note to edit. This must be at least 1.</param>
		public EditNoteWindow( string customTitle, string comment, string note, bool readOnly, TextValidations contentType
			, bool mustHaveContent, int noteSizeMax )
		{
			CustomTitle = customTitle;
			Comment = comment;
			Note = note;
			MustHaveContent = mustHaveContent;
			NoteSizeMax = noteSizeMax;
			ReadOnly = readOnly;
		}

		#endregion Constructors

		#region Private non-serializable properties

		/// <summary>
		/// The editable text-box.
		/// </summary>
		protected abstract TextBox OverriddenTextBox { get; }

		/// <summary>
		/// The error presenter.
		/// </summary>
		protected abstract FrameworkElement OverriddenErrorPresenter { get; }

		#endregion Private non-serializable properties

		#region Public non-serializable properties

		private string _CustomTitle;
		/// <summary>
		/// The window title. If this is null or empty, the getter returns the default value. Notifies property observers.
		/// READ-ONLY.
		/// </summary>
		public string CustomTitle
		{
			get { return TextUtils.HasContent( _CustomTitle ) ? _CustomTitle : CustomTitleDef; }
			protected set
			{
				_CustomTitle = value;

				// Notify property observers.
				OnPropertyChanged( "CustomTitle" );
			}
		}

		private string _Comment;
		/// <summary>
		/// Comment to show above the note text-box. This may be null or empty. Notifies property observers. READ-ONLY.
		/// </summary>
		public string Comment
		{
			get { return _Comment; }
			protected set
			{
				_Comment = value;

				// Notify property observers.
				OnPropertyChanged( "Comment" );
			}
		}

		private string _Note;
		/// <summary>
		/// Note to edit. This may be null or empty. Notifies property observers.
		/// </summary>
		public string Note
		{
			get { return _Note; }
			set
			{
				_Note = value;

				// Notify property observers.
				OnPropertyChanged( "Note" );
			}
		}

		private bool _ReadOnly;
		/// <summary>
		/// If this is true then the note can't be edited. Notifies property observers. READ-ONLY.
		/// </summary>
		public bool ReadOnly
		{
			get { return _ReadOnly; }
			protected set
			{
				_ReadOnly = value;

				// Notify property observers.
				OnPropertyChanged( "ReadOnly" );
			}
		}

		private TextValidations _ContentType;
		/// <summary>
		/// The type of content of the note to edit. Notifies property observers. READ-ONLY.
		/// </summary>
		public TextValidations ContentType
		{
			get { return _ContentType; }
			protected set
			{
				_ContentType = value;

				// Notify property observers.
				OnPropertyChanged( "ContentType" );
			}
		}

		private bool _MustHaveContent;
		/// <summary>
		/// If this is true then the note must have content. Notifies property observers. READ-ONLY.
		/// </summary>
		public bool MustHaveContent
		{
			get { return _MustHaveContent; }
			protected set
			{
				_MustHaveContent = value;

				// Notify property observers.
				OnPropertyChanged( "MustHaveContent" );
			}
		}

		private int _NoteSizeMax;
		/// <summary>
		/// The maximum size of the note to edit. This must be at least 1. Notifies property observers. READ-ONLY.
		/// </summary>
		public int NoteSizeMax
		{
			get { return _NoteSizeMax; }
			protected set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "NoteSizeMax" );

				_NoteSizeMax = value;

				// Notify property observers.
				OnPropertyChanged( "NoteSizeMax" );
			}
		}

		/// <summary>
		/// The visibility of the comment. READ-ONLY.
		/// </summary>
		public Visibility CommentVisibility
		{
			get { return TextUtils.HasContent( Comment ) ? Visibility.Visible : Visibility.Collapsed; }
		}

		/// <summary>
		/// The visibility of the "Okay" and "Cancel" buttons. READ-ONLY.
		/// </summary>
		public Visibility OkayCancelButtonsVisibility
		{
			get { return ReadOnly ? Visibility.Collapsed : Visibility.Visible; }
		}

		/// <summary>
		/// The visibility of the "Close" button. READ-ONLY.
		/// </summary>
		public Visibility CloseButtonVisibility
		{
			get { return ReadOnly ? Visibility.Visible : Visibility.Collapsed; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Validates the input data.
		/// </summary>
		/// <param name="errorPresenter">The control which shows that there is a validation error. This may be null.</param>
		/// <remarks>
		/// <para>If you don't validate of all the editable fields in here, it's possible that some bindings are commited even
		/// if the user cancels the edit operation. This can happen if the user edits some fields, clicks the "Okay" button, the
		/// validation fails (during binding commit) inside some properties (which are not validated in here), and then the user
		/// clicks the "Cancel" button. This is likely to matter only if those (successfully) bound properties are used outside
		/// the view.</para>
		/// </remarks>
		protected override void ValidateUi( FrameworkElement errorPresenter )
		{
			base.ValidateUi( errorPresenter );

			ValidationUtils.Text( OverriddenTextBox.Text, MustHaveContent ? -1 : -2, NoteSizeMax, ContentType, "Note"
				, OverriddenTextBox, errorPresenter );
		}

		/// <summary>
		/// Accepts the changes made by the user.
		/// </summary>
		protected void AcceptChanges()
		{
			try
			{
				ValidateUi( OverriddenErrorPresenter );
				SaveUi( OverriddenErrorPresenter );

				DialogResult = true;
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				ControlsUtils.ShowException( ex );
			}
		}

		#endregion Private methods
	}
}
