////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Windows;
using System.Windows.Controls;
using Vela.Kernel;

namespace Vela.Controls
{
	/// <summary>
	/// A window which allows the user to edit a single-line note.
	/// </summary>
	public partial class EditSingleLineNoteWindow : EditNoteWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected EditSingleLineNoteWindow()
		{
			InitializeComponent();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="customTitle">The window title. This may be null or empty.</param>
		/// <param name="comment">Comment to show above the note text-box. This may be null or empty.</param>
		/// <param name="note">Note to edit. This may be null or empty.</param>
		/// <param name="readOnly">If this is true then the note can't be edited.</param>
		/// <param name="contentType">The type of content of the note to edit.</param>
		/// <param name="mustHaveContent">If this is true then the note must have content.</param>
		/// <param name="noteSizeMax">The maximum size of the note to edit. This must be at least 1.</param>
		public EditSingleLineNoteWindow( string customTitle, string comment, string note, bool readOnly
				, TextValidations contentType, bool mustHaveContent, int noteSizeMax )
			: base( customTitle, comment, note, readOnly, contentType, mustHaveContent, noteSizeMax )
		{
			InitializeComponent();
		}

		#endregion Constructors

		#region Private non-serializable properties

		/// <summary>
		/// The editable text-box.
		/// </summary>
		protected override TextBox OverriddenTextBox
		{
			get { return SingleLineNoteTextBox; }
		}

		/// <summary>
		/// The error presenter.
		/// </summary>
		protected override FrameworkElement OverriddenErrorPresenter
		{
			get { return ErrorPresenter; }
		}

		#endregion Private non-serializable properties

		#region Event handlers

		private void ViewWindow_Loaded( object sender, RoutedEventArgs e )
		{
			ControlsUtils.FocusTextBox( SingleLineNoteTextBox, true );
		}

		private void SingleLineNoteTextBox_Pasting( object sender, DataObjectPastingEventArgs e )
		{
			ControlsUtils.BlockPasteIfTooLong( (TextBox)sender, e );
		}

		private void OkayButton_Click( object sender, RoutedEventArgs e )
		{
			AcceptChanges();
		}

		#endregion Event handlers

		#region Public static methods

		/// <summary>
		/// Shows the specified message in a modal manner.
		/// </summary>
		/// <param name="customTitle">The window title. This may be null or empty.</param>
		/// <param name="comment">Comment to show above the note text-box. This may be null or empty.</param>
		/// <param name="note">Note to edit. This may be null or empty.</param>
		/// <param name="readOnly">If this is true then the note can't be edited.</param>
		/// <param name="contentType">The type of content of the note to edit.</param>
		/// <param name="mustHaveContent">If this is true then the note must have content.</param>
		/// <param name="noteSizeMax">The maximum size of the note to edit.</param>
		/// <returns>Returns the edited note. This may be null or empty.</returns>
		public static string ShowDialog( string customTitle, string comment, string note, bool readOnly
			, TextValidations contentType, bool mustHaveContent, int noteSizeMax )
		{
			EditSingleLineNoteWindow window = new EditSingleLineNoteWindow( customTitle, comment, note, readOnly, contentType
				, mustHaveContent, noteSizeMax );

			bool okay = ControlsUtils.ShowDialog( window );

			return okay ? window.Note : null;
		}

		#endregion Public static methods
	}
}
