////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows;
using System.Windows.Input;
using Vela.Kernel;

namespace Vela.Controls
{
	/// <summary>
	/// A window which allows the user to generate entropy.
	/// </summary>
	/// <remarks>
	/// <para>The generated entropy is automatically added to the entropy pool.</para>
	/// </remarks>
	public partial class EntropyWindow : ViewWindow
	{
		#region Public constants

		/// <summary>
		/// The minimum value for <see cref="RequiredEntropyEventCount"/>.
		/// </summary>
		public const int RequiredEntropyEventCountMin = 10;

		/// <summary>
		/// The default value for <see cref="RequiredEntropyEventCount"/>.
		/// </summary>
		public const int RequiredEntropyEventCountDef = 50;

		/// <summary>
		/// The timespan (in seconds) which must pass between two consecutive user generated entropy events in order to accept
		/// the second one.
		/// </summary>
		public const double TimespanBetweenAcceptedEvents = 0.3;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public EntropyWindow()
		{
			InitializeComponent();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="requiredEntropyEventCount">The number of entropy events which must be generated by the user. If this
		/// is 0 or a negative number, the default is used.</param>
		public EntropyWindow( int requiredEntropyEventCount )
		{
			if( requiredEntropyEventCount > 0 ) RequiredEntropyEventCount = requiredEntropyEventCount;

			InitializeComponent();
		}

		#endregion Constructors

		#region Private fields

		private int _RequiredEntropyEventCount = RequiredEntropyEventCountDef;
		/// <summary>
		/// The number of entropy events which must be generated by the user. This must be at least
		/// <see cref="RequiredEntropyEventCountMin"/>.
		/// </summary>
		private int RequiredEntropyEventCount
		{
			get { return _RequiredEntropyEventCount; }
			set
			{
				ValidationUtils.Integer( value, RequiredEntropyEventCountMin, int.MaxValue, "RequiredEntropyEventCount" );

				_RequiredEntropyEventCount = value;

				// Notify property observers.
				OnPropertyChanged( "RequiredEntropyEventCount" );
			}
		}

		/// <summary>
		/// The time when the last user generated entropy event was accepted.
		/// </summary>
		private DateTime TimeOfLastAcceptedEvent = DateTime.UtcNow;

		#endregion Private fields

		#region Public non-serializable properties

		private int _EntropyEventCount;
		/// <summary>
		/// The number of entropy events generated by the user. Notifies property observers.
		/// </summary>
		public int EntropyEventCount
		{
			get { return _EntropyEventCount; }
			set
			{
				ValidationUtils.Integer( value, 0, int.MaxValue, "EntropyEventCount" );

				_EntropyEventCount = value;

				MayAccept = EntropyEventCount >= RequiredEntropyEventCount;
				Progress = (100 * EntropyEventCount) / RequiredEntropyEventCount;

				// Notify property observers.
				OnPropertyChanged( "EntropyEventCount" );
			}
		}

		private bool _MayAccept = false;
		/// <summary>
		/// If this is true then the "Okay" button is enabled. Notifies property observers.
		/// </summary>
		public bool MayAccept
		{
			get { return _MayAccept; }
			protected set
			{
				_MayAccept = value;

				// Notify property observers.
				OnPropertyChanged( "MayAccept" );
			}
		}

		private int _Progress;
		/// <summary>
		/// The entropy generation progress. Notifies property observers. READ-ONLY.
		/// </summary>
		public int Progress
		{
			get { return _Progress; }
			protected set
			{
				_Progress = value;

				// Notify property observers.
				OnPropertyChanged( "Progress" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void ViewWindow_PreviewMouseMove( object sender, MouseEventArgs e )
		{
			if( MayAcceptEntropyEvent() )
			{
				ControlsUtils.AddMouseEntropy( e );
				EntropyEventWasGenerated();
			}
		}

		private void ViewWindow_PreviewMouseDown( object sender, MouseButtonEventArgs e )
		{
			if( MayAcceptEntropyEvent() )
			{
				ControlsUtils.AddMouseEntropy( e );
				EntropyEventWasGenerated();
			}
		}

		private void ViewWindow_PreviewKeyDown( object sender, KeyEventArgs e )
		{
			if( MayAcceptEntropyEvent() )
			{
				ControlsUtils.AddKeyboardEntropy( e );
				EntropyEventWasGenerated();
			}
		}

		private void OkayButton_Click( object sender, RoutedEventArgs e )
		{
			try
			{
				ValidateUi( ErrorPresenter );
				SaveUi( ErrorPresenter );

				DialogResult = true;
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) ) throw ExceptionUtils.PreserveStackTrace( ex );

				ControlsUtils.ShowException( ex );
			}
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Checks if the current user generated entropy event may be accepted.
		/// </summary>
		/// <returns>Returns true if the current user generated entropy event may be accepted.</returns>
		private bool MayAcceptEntropyEvent()
		{
			return DateTime.UtcNow > TimeOfLastAcceptedEvent.AddSeconds( TimespanBetweenAcceptedEvents );
		}

		/// <summary>
		/// Handles the last accepted user generated entropy event.
		/// </summary>
		private void EntropyEventWasGenerated()
		{
			TimeOfLastAcceptedEvent = DateTime.UtcNow;
			EntropyEventCount++;
		}

		/// <summary>
		/// Validates the input data.
		/// </summary>
		/// <param name="errorPresenter">The control which shows that there is a validation error. This may be null.</param>
		/// <remarks>
		/// <para>If you don't validate of all the editable fields in here, it's possible that some bindings are commited even
		/// if the user cancels the edit operation. This can happen if the user edits some fields, clicks the "Okay" button, the
		/// validation fails (during binding commit) inside some properties (which are not validated in here), and then the user
		/// clicks the "Cancel" button. This is likely to matter only if those (successfully) bound properties are used outside
		/// the view.</para>
		/// </remarks>
		protected override void ValidateUi( FrameworkElement errorPresenter )
		{
			base.ValidateUi( errorPresenter );

			ValidationUtils.Integer( EntropyEventCount, RequiredEntropyEventCount, int.MaxValue, "EntropyEventCount"
				, errorPresenter );
		}

		#endregion Private methods
	}
}
