////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows;
using System.Windows.Media.Imaging;
using Vela.Kernel;

namespace Vela.Controls
{
	/// <summary>
	/// A window which shows messages.
	/// </summary>
	public partial class MessageWindow : ViewWindow
	{
		#region Public constants

		/// <summary>
		/// Template for the window title.
		/// </summary>
		public const string TitleTemplate = "{0} - {1}";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected MessageWindow()
		{
			InitializeComponent();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="message">Message to show. This may be null or empty.</param>
		/// <param name="mode">Window mode.</param>
		public MessageWindow( string message, MessageWindowModes mode )
		{
			Message = message;
			Mode = mode;

			InitializeComponent();
		}

		#endregion Constructors

		#region Public non-serializable properties

		private string _Message;
		/// <summary>
		/// Message to show. This may be null or empty. Notifies property observers. READ-ONLY.
		/// </summary>
		public string Message
		{
			get { return _Message; }
			protected set
			{
				_Message = value;

				// Notify property observers.
				OnPropertyChanged( "Message" );
			}
		}

		private MessageWindowModes _Mode;
		/// <summary>
		/// The window mode. Notifies property observers. READ-ONLY.
		/// </summary>
		public MessageWindowModes Mode
		{
			get { return _Mode; }
			protected set
			{
				_Mode = value;
				_ModeImage = null;

				// Notify property observers.
				OnPropertyChanged( "Mode" );
				OnPropertyChanged( "WindowTitle" );
				OnPropertyChanged( "ModeImage" );
				OnPropertyChanged( "YesNoButtonsVisibility" );
				OnPropertyChanged( "OkayButtonVisibility" );
			}
		}

		/// <summary>
		/// Window title. This may be null or empty. READ-ONLY.
		/// </summary>
		public string WindowTitle
		{
			get
			{
				switch( Mode )
				{
					case MessageWindowModes.Error:
						return string.Format( TitleTemplate, "Error", AssemblyUtils.ApplicationBuildInfo.ProductDisplayName );
					case MessageWindowModes.Info:
						return string.Format( TitleTemplate, "Info", AssemblyUtils.ApplicationBuildInfo.ProductDisplayName );
					case MessageWindowModes.Query:
						return string.Format( TitleTemplate, "Query", AssemblyUtils.ApplicationBuildInfo.ProductDisplayName );
					case MessageWindowModes.Warning:
						return string.Format( TitleTemplate, "Warning", AssemblyUtils.ApplicationBuildInfo.ProductDisplayName );
					default:
						throw ValidationUtils.Unhandled( Mode, "Mode" );
				}
			}
		}

		private BitmapSource _ModeImage;
		/// <summary>
		/// Mode image. This may be null. Notifies property observers. READ-ONLY.
		/// </summary>
		public BitmapSource ModeImage
		{
			get
			{
				if( _ModeImage != null )
					return _ModeImage;

				switch( Mode )
				{
					case MessageWindowModes.Error:
					case MessageWindowModes.Info:
					case MessageWindowModes.Query:
					case MessageWindowModes.Warning:
						string fileName = string.Format( ControlsConsts.ModeImageFileNameTemplate, Mode );
						_ModeImage = ImagePreviewsCache.GetCachedOrLoadFirstFrame( fileName, fileName, false );
						break;
					default:
						throw ValidationUtils.Unhandled( Mode, "Mode" );
				}

				// Notify property observers. Avoid stack overflows when notifying about null values.
				if( _ModeImage != null )
					OnPropertyChanged( "ModeImage" );

				return _ModeImage;
			}
			protected set
			{
				_ModeImage = value;

				// Notify property observers.
				OnPropertyChanged( "ModeImage" );
			}
		}

		/// <summary>
		/// The visibility of the "Yes" and "No" buttons. READ-ONLY.
		/// </summary>
		public Visibility YesNoButtonsVisibility
		{
			get
			{
				switch( Mode )
				{
					case MessageWindowModes.Error:
					case MessageWindowModes.Info:
					case MessageWindowModes.Warning:
						return Visibility.Collapsed;
					case MessageWindowModes.Query:
						return Visibility.Visible;
					default:
						throw ValidationUtils.Unhandled( YesNoButtonsVisibility, "YesNoButtonsVisibility" );
				}
			}
		}

		/// <summary>
		/// The visibility of the "Okay" button. READ-ONLY.
		/// </summary>
		public Visibility OkayButtonVisibility
		{
			get
			{
				switch( Mode )
				{
					case MessageWindowModes.Error:
					case MessageWindowModes.Info:
					case MessageWindowModes.Warning:
						return Visibility.Visible;
					case MessageWindowModes.Query:
						return Visibility.Collapsed;
					default:
						throw ValidationUtils.Unhandled( OkayButtonVisibility, "OkayButtonVisibility" );
				}
			}
		}

		private QueryMessageResults _Result = QueryMessageResults.Okay;
		/// <summary>
		/// The message window result. Notifies property observers. READ-ONLY.
		/// </summary>
		public QueryMessageResults Result
		{
			get { return _Result; }
			protected set
			{
				_Result = value;

				// Notify property observers.
				OnPropertyChanged( "Result" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void YesButton_Click( object sender, RoutedEventArgs e )
		{
			Result = QueryMessageResults.Yes;
			DialogResult = true;
		}

		private void NoButton_Click( object sender, RoutedEventArgs e )
		{
			Result = QueryMessageResults.No;
			DialogResult = false;
		}

		private void OkayButton_Click( object sender, RoutedEventArgs e )
		{
			Result = QueryMessageResults.Okay;
			DialogResult = false;
		}

		#endregion Event handlers

		#region Public static methods

		/// <summary>
		/// Shows the specified message in a modal manner.
		/// </summary>
		/// <param name="message">Message to show. This may be null or empty.</param>
		/// <param name="mode">Window mode.</param>
		/// <returns>Returns which button was pressed.</returns>
		public static QueryMessageResults ShowDialog( string message, MessageWindowModes mode )
		{
			if( ControlsUtils.GetAppDispatcher() == null )
			{
				LogManager.Add( message, null, LogRecordModes.Warning, false );

				return QueryMessageResults.Okay;
			}

			try
			{
				MessageWindow window = new MessageWindow( message, mode );

				ControlsUtils.ShowDialog( window );

				return window.Result;
			}
			catch( Exception ex )
			{
				// Log even critical exceptions.
				LogManager.Add( ex );

				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				// Use the system's message-box in case of nasty exceptions.
				MessageBox.Show( ex.ToString() );

				return QueryMessageResults.Okay;
			}
		}

		#endregion Public static methods
	}
}
