////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using Vela.Kernel;

namespace Vela.Controls
{
	/// <summary>
	/// A window where to type a password.
	/// </summary>
	public partial class PasswordWindow : ViewWindow
	{
		#region Public constants

		/// <summary>
		/// The default value for <see cref="CustomTitle"/>.
		/// </summary>
		public const string CustomTitleDef = "Password";

		/// <summary>
		/// The display character for a password character with an even index, when the password character with the same index
		/// from the other password is matching.
		/// </summary>
		public const char MatchingPasswordsEvenDisplayCharacter = '\x52';

		/// <summary>
		/// The display character for a password character with an odd index, when the password character with the same index
		/// from the other password is matching.
		/// </summary>
		public const char MatchingPasswordsOddDisplayCharacter = '\x54';

		/// <summary>
		/// The display character for a password character when the password character with the same index from the other
		/// password is different.
		/// </summary>
		public const char DifferentPasswordsDisplayCharacter = '\x7C';

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected PasswordWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="customTitle">The window title. This may be null or empty.</param>
		/// <param name="storehouseFileName">The storehouse file name. This may be null or empty.</param>
		/// <param name="currentPasswordToMatch">The user must type this value in the current password text-box. If this is null
		/// or empty, it's ignored.</param>
		/// <param name="requiresConfirmation">If this is true, the user must confirm the password by typing it twice.</param>
		public PasswordWindow( string customTitle, string storehouseFileName, string currentPasswordToMatch
			, bool requiresConfirmation )
		{
			CustomTitle = customTitle;
			StorehouseFileName = storehouseFileName;
			CurrentPasswordToMatch = currentPasswordToMatch;
			RequiresConfirmation = requiresConfirmation;

			InitializeComponent();

			LoadUi();
		}

		#endregion Constructors

		#region Public non-serializable properties

		private string _CustomTitle;
		/// <summary>
		/// The window title. If this is null or empty, the getter returns the default value. Notifies property observers.
		/// READ-ONLY.
		/// </summary>
		public string CustomTitle
		{
			get { return TextUtils.HasContent( _CustomTitle ) ? _CustomTitle : CustomTitleDef; }
			protected set
			{
				_CustomTitle = value;

				// Notify property observers.
				OnPropertyChanged( "CustomTitle" );
			}
		}

		private string _StorehouseFileName;
		/// <summary>
		/// The storehouse file name. This may be null or empty. Notifies property observers. READ-ONLY.
		/// </summary>
		public string StorehouseFileName
		{
			get { return _StorehouseFileName; }
			protected set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.FileName, "StorehouseFileName" );

				_StorehouseFileName = value;

				// Notify property observers.
				OnPropertyChanged( "StorehouseFileName" );
				OnPropertyChanged( "StorehouseFileNameVisibility" );
			}
		}

		/// <summary>
		/// The visibility of the storehouse file name. READ-ONLY.
		/// </summary>
		public Visibility StorehouseFileNameVisibility
		{
			get { return TextUtils.HasContent( StorehouseFileName ) ? Visibility.Visible : Visibility.Collapsed; }
		}

		private string _CurrentPasswordToMatch;
		/// <summary>
		/// The user must type this value in the current password text-box. If this is null or empty, it's ignored. Notifies
		/// property observers. READ-ONLY.
		/// </summary>
		public string CurrentPasswordToMatch
		{
			get { return _CurrentPasswordToMatch; }
			protected set
			{
				if( TextUtils.HasContent( value ) ) ValidationUtils.Password( value, "CurrentPasswordToMatch" );

				_CurrentPasswordToMatch = value;

				// Notify property observers.
				OnPropertyChanged( "CurrentPasswordToMatch" );
				OnPropertyChanged( "CurrentPasswordVisibility" );
				OnPropertyChanged( "PasswordVisibility" );
			}
		}

		/// <summary>
		/// The visibility of the current password. READ-ONLY.
		/// </summary>
		public Visibility CurrentPasswordVisibility
		{
			get { return TextUtils.HasContent( CurrentPasswordToMatch ) ? Visibility.Visible : Visibility.Collapsed; }
		}

		/// <summary>
		/// The visibility of the password. READ-ONLY.
		/// </summary>
		public Visibility PasswordVisibility
		{
			get
			{
				return (TextUtils.NoContent( CurrentPasswordToMatch ) || RequiresConfirmation)
					? Visibility.Visible : Visibility.Collapsed;
			}
		}

		private bool _RequiresConfirmation = true;
		/// <summary>
		/// If this is true, the user must confirm the password by typing it twice. Notifies property observers. READ-ONLY.
		/// </summary>
		public bool RequiresConfirmation
		{
			get { return _RequiresConfirmation; }
			protected set
			{
				_RequiresConfirmation = value;

				// Notify property observers.
				OnPropertyChanged( "RequiresConfirmation" );
				OnPropertyChanged( "PasswordVisibility" );
				OnPropertyChanged( "ConfirmationPasswordVisibility" );
			}
		}

		/// <summary>
		/// The visibility of the confirmation password. READ-ONLY.
		/// </summary>
		public Visibility ConfirmationPasswordVisibility
		{
			get { return RequiresConfirmation ? Visibility.Visible : Visibility.Collapsed; }
		}

		#region Current password

		private StringBuilder _DisplayCurrentPasswordBuilder = new StringBuilder();
		/// <summary>
		/// The display current password typed by the user. The backing field is initialized by default. READ-ONLY.
		/// </summary>
		public StringBuilder DisplayCurrentPasswordBuilder
		{
			get { return _DisplayCurrentPasswordBuilder; }
		}

		private StringBuilder _CurrentPasswordBuilder = new StringBuilder();
		/// <summary>
		/// The current password typed by the user. The backing field is initialized by default. READ-ONLY.
		/// </summary>
		public StringBuilder CurrentPasswordBuilder
		{
			get { return _CurrentPasswordBuilder; }
		}

		/// <summary>
		/// The current password typed by the user. READ-ONLY.
		/// </summary>
		public string CurrentPassword
		{
			get { return CurrentPasswordBuilder.ToString().ToLowerInvariant(); }
		}

		#endregion Current password

		#region Password

		private StringBuilder _DisplayPasswordBuilder = new StringBuilder();
		/// <summary>
		/// The display password typed by the user. The backing field is initialized by default. READ-ONLY.
		/// </summary>
		public StringBuilder DisplayPasswordBuilder
		{
			get { return _DisplayPasswordBuilder; }
		}

		private StringBuilder _PasswordBuilder = new StringBuilder();
		/// <summary>
		/// The password typed by the user. The backing field is initialized by default. READ-ONLY.
		/// </summary>
		public StringBuilder PasswordBuilder
		{
			get { return _PasswordBuilder; }
		}

		/// <summary>
		/// The password typed by the user. READ-ONLY.
		/// </summary>
		public string Password
		{
			get { return PasswordBuilder.ToString().ToLowerInvariant(); }
		}

		#endregion Password

		#region Confirmation password

		private StringBuilder _DisplayConfirmationPasswordBuilder = new StringBuilder();
		/// <summary>
		/// The display confirmation password typed by the user. The backing field is initialized by default. READ-ONLY.
		/// </summary>
		public StringBuilder DisplayConfirmationPasswordBuilder
		{
			get { return _DisplayConfirmationPasswordBuilder; }
		}

		private StringBuilder _ConfirmationPasswordBuilder = new StringBuilder();
		/// <summary>
		/// The confirmation password typed by the user. The backing field is initialized by default. READ-ONLY.
		/// </summary>
		public StringBuilder ConfirmationPasswordBuilder
		{
			get { return _ConfirmationPasswordBuilder; }
		}

		/// <summary>
		/// The confirmation password typed by the user. READ-ONLY.
		/// </summary>
		public string ConfirmationPassword
		{
			get { return ConfirmationPasswordBuilder.ToString().ToLowerInvariant(); }
		}

		#endregion Confirmation password

		#endregion Public non-serializable properties

		#region Event handlers

		private void ViewWindow_Loaded( object sender, RoutedEventArgs e )
		{
			if( TextUtils.HasContent( CurrentPasswordToMatch ) ) CurrentPasswordTextBox.Focus();
			else PasswordTextBox.Focus();
		}

		private void PasswordTextBox_PreviewKeyDown( object sender, KeyEventArgs e )
		{
			e.Handled = HandleKey( sender as TextBox, e.Key );

			if( sender == PasswordTextBox ) ControlsUtils.AddKeyboardEntropy( e );
		}

		private void PasswordTextBox_PreviewTextInput( object sender, TextCompositionEventArgs e )
		{
			InsertText( sender as TextBox, e.TextComposition.Text );
			e.Handled = true;
		}

		private void PasswordTextBox_Pasting( object sender, DataObjectPastingEventArgs e )
		{
			// The password edit-boxes have no size limit and the pasting is not blocked (but the password size is validated when
			// the user clicks "Okay") because the user can't see how many characters the edit-boxes accept (since there is no
			// error message if he continues to type beyond the size limit of a text-box).

			InsertText( sender as TextBox, e.DataObject.GetData( typeof( string ) ) as string );
			e.Handled = true;
			e.CancelCommand();
		}

		private void OkayButton_Click( object sender, RoutedEventArgs e )
		{
			try
			{
				ValidateUi( ErrorPresenter );
				SaveUi( ErrorPresenter );

				DialogResult = true;
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) ) throw ExceptionUtils.PreserveStackTrace( ex );

				ControlsUtils.ShowException( ex );
			}
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Gets the display character for the password builder and character.
		/// </summary>
		/// <param name="passwordBuilder">Password builder in which the specified character is to be added.</param>
		/// <param name="newChar">Character which is to be added in the specified password builder.</param>
		/// <returns>Returns the display character.</returns>
		private char GetDisplayChar( StringBuilder passwordBuilder, char newChar )
		{
			ValidationUtils.ObjectHasContent( passwordBuilder, "passwordBuilder" );

			int i = passwordBuilder.Length;

			// Only the confirmation password builder has a visual confirmation for different characters.
			if( (passwordBuilder == CurrentPasswordBuilder) || (passwordBuilder == PasswordBuilder) )
				return (i % 2 == 0) ? MatchingPasswordsEvenDisplayCharacter : MatchingPasswordsOddDisplayCharacter;

			bool matching = (i < PasswordBuilder.Length) && (newChar == PasswordBuilder[ i ]);

			if( matching )
			{
				return (i % 2 == 0) ? MatchingPasswordsEvenDisplayCharacter : MatchingPasswordsOddDisplayCharacter;
			}
			else return DifferentPasswordsDisplayCharacter;
		}

		/// <summary>
		/// Handles the specified key into the specified text-box. Notifies property observers.
		/// </summary>
		/// <param name="textBox">Text-box where to handle the key.</param>
		/// <param name="key">Key to handle.</param>
		/// <returns>Returns true if the key was handled and the underlying infrastructure must ignore it.</returns>
		private bool HandleKey( TextBox textBox, Key key )
		{
			ValidationUtils.ObjectHasContent( textBox, "textBox" );

			// This key has to be processed by the underlying infrastructure.
			if( (key == Key.Escape) || (key == Key.Enter) || (key == Key.Tab) ) return false;

			// The key was handled and the underlying infrastructure must ignore it.
			if( (key == Key.Insert) || (key == Key.Delete) || (key == Key.Home) || (key == Key.End) || (key == Key.PageUp)
				|| (key == Key.PageDown) || (key == Key.Left) || (key == Key.Right) || (key == Key.Up) || (key == Key.Down)
				|| (key == Key.Space) )
			{
				return true;
			}

			StringBuilder displayBuilder = null;
			StringBuilder passwordBuilder = null;
			string propertyName = null;

			if( textBox == CurrentPasswordTextBox )
			{
				displayBuilder = DisplayCurrentPasswordBuilder;
				passwordBuilder = CurrentPasswordBuilder;
				propertyName = "DisplayCurrentPasswordBuilder";
			}
			else if( textBox == PasswordTextBox )
			{
				displayBuilder = DisplayPasswordBuilder;
				passwordBuilder = PasswordBuilder;
				propertyName = "DisplayPasswordBuilder";
			}
			else if( textBox == ConfirmationPasswordTextBox )
			{
				displayBuilder = DisplayConfirmationPasswordBuilder;
				passwordBuilder = ConfirmationPasswordBuilder;
				propertyName = "DisplayConfirmationPasswordBuilder";
			}
			else throw ValidationUtils.Unhandled( textBox.GetType().FullName, "textBox" );

			if( key == Key.Back )
			{
				if( displayBuilder.Length >= 1 )
				{
					displayBuilder.Length -= 1;
					passwordBuilder.Length -= 1;

					// Notify property observers.
					OnPropertyChanged( propertyName );

					// Move the caret.
					textBox.CaretIndex = displayBuilder.Length;
				}

				// The key was handled and the underlying infrastructure must ignore it.
				return true;
			}

			// This key has to be processed by the underlying infrastructure, in order to reach the "InsertText" method.
			return false;
		}

		/// <summary>
		/// Inserts the specified text into the specified text-box. Notifies property observers.
		/// </summary>
		/// <param name="textBox">Text-box where to insert the text.</param>
		/// <param name="text">Text to insert. This may be null or empty.</param>
		/// <returns>Returns true if the text is inserted.</returns>
		private void InsertText( TextBox textBox, string text )
		{
			ValidationUtils.ObjectHasContent( textBox, "textBox" );

			if( TextUtils.NoContent( text ) ) return;

			StringBuilder displayBuilder = null;
			StringBuilder passwordBuilder = null;
			string propertyName = null;

			if( textBox == CurrentPasswordTextBox )
			{
				displayBuilder = DisplayCurrentPasswordBuilder;
				passwordBuilder = CurrentPasswordBuilder;
				propertyName = "DisplayCurrentPasswordBuilder";
			}
			else if( textBox == PasswordTextBox )
			{
				displayBuilder = DisplayPasswordBuilder;
				passwordBuilder = PasswordBuilder;
				propertyName = "DisplayPasswordBuilder";
			}
			else if( textBox == ConfirmationPasswordTextBox )
			{
				displayBuilder = DisplayConfirmationPasswordBuilder;
				passwordBuilder = ConfirmationPasswordBuilder;
				propertyName = "DisplayConfirmationPasswordBuilder";
			}
			else throw ValidationUtils.Unhandled( textBox.GetType().FullName, "textBox" );

			if( displayBuilder.Length >= KernelConsts.PasswordSizeMax ) return;

			foreach( char ch in text )
			{
				// Display a character.
				char displayChar = GetDisplayChar( passwordBuilder, ch );
				displayBuilder.Append( displayChar );

				passwordBuilder.Append( ch );
			}

			// Notify property observers.
			OnPropertyChanged( propertyName );

			// Move the caret.
			textBox.CaretIndex = displayBuilder.Length;
		}

		/// <summary>
		/// Validates the input data.
		/// </summary>
		/// <param name="errorPresenter">The control which shows that there is a validation error. This may be null.</param>
		/// <remarks>
		/// <para>If you don't validate of all the editable fields in here, it's possible that some bindings are commited even
		/// if the user cancels the edit operation. This can happen if the user edits some fields, clicks the "Okay" button, the
		/// validation fails (during binding commit) inside some properties (which are not validated in here), and then the user
		/// clicks the "Cancel" button. This is likely to matter only if those (successfully) bound properties are used outside
		/// the view.</para>
		/// </remarks>
		protected override void ValidateUi( FrameworkElement errorPresenter )
		{
			base.ValidateUi( errorPresenter );

			if( TextUtils.HasContent( StorehouseFileNameTextBox.Text ) )
				ValidationUtils.Text( StorehouseFileNameTextBox.Text, -2, -1, TextValidations.FileName, "Storehouse" );

			if( CurrentPasswordVisibility == Visibility.Visible )
			{
				ValidationUtils.Password( CurrentPassword, "CurrentPassword", CurrentPasswordTextBox, errorPresenter );

				if( CurrentPassword != CurrentPasswordToMatch )
				{
					throw new PresentedValidationException( "The current password doesn't match with the required one."
						, null, "Current password", CurrentPasswordTextBox, errorPresenter );
				}
			}

			if( PasswordVisibility == Visibility.Visible )
			{
				ValidationUtils.Password( Password, "Password", PasswordTextBox, errorPresenter );
			}

			if( ConfirmationPasswordVisibility == Visibility.Visible )
			{
				ValidationUtils.Password( ConfirmationPassword, "ConfirmationPassword", ConfirmationPasswordTextBox
					, errorPresenter );

				ValidationUtils.PasswordConfirm( Password, ConfirmationPassword, "ConfirmationPassword"
					, ConfirmationPasswordTextBox, errorPresenter, true );
			}
		}

		#endregion Private methods
	}
}
