////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows;
using System.Windows.Media.Imaging;
using System.Windows.Threading;
using Vela.Kernel;

namespace Vela.Controls
{
	/// <summary>
	/// A window which shows that an operation is done / over.
	/// </summary>
	public partial class TooltipWindow : ViewWindow, IFloatingWindow
	{
		#region Public constants

		/// <summary>
		/// The default value for <see cref="Message"/>.
		/// </summary>
		public const string MessageDef = "Done";

		/// <summary>
		/// The maximum width that the message label may have.
		/// </summary>
		public const int MessageLabelWidthMaxDef = 400;

		/// <summary>
		/// The timespan (in seconds) for which to show the window when there is no message.
		/// </summary>
		public const double NoMessageAutoCloseTimespan = 1.5;

		/// <summary>
		/// The timespan (in seconds) for which to show the window when there is a message.
		/// </summary>
		public const double WithMessageAutoCloseTimespan = 7;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected TooltipWindow()
		{
			InitializeComponent();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="message">Message to show. If "mode" is <see cref="TooltipWindowModes.Image"/> or
		/// <see cref="TooltipWindowModes.CachedImage"/>, this must contain the name of the file to show. This may be null or
		/// empty.</param>
		/// <param name="mode">Window mode.</param>
		public TooltipWindow( string message, TooltipWindowModes mode )
		{
			if( TextUtils.HasContent( message ) )
				Message = message;
			Mode = mode;

			InitializeComponent();
		}

		#endregion Constructors

		#region Private static fields

		/// <summary>
		/// The global tooltip window.
		/// </summary>
		private static TooltipWindow Instance;

		#endregion Private static fields

		#region Private fields

		/// <summary>
		/// Timer used to automatically close the window.
		/// </summary>
		private DispatcherTimer AutoCloseTimer;

		#endregion Private fields

		#region Public non-serializable properties

		private string _Message = MessageDef;
		/// <summary>
		/// Message to show. If <see cref="Mode"/> is <see cref="TooltipWindowModes.Image"/> or
		/// <see cref="TooltipWindowModes.CachedImage"/>, this must contain the name of the file to show. This may be null or
		/// empty. Notifies property observers. READ-ONLY.
		/// </summary>
		public string Message
		{
			get { return _Message; }
			protected set
			{
				_Message = value;

				// Notify property observers.
				OnPropertyChanged( "Message" );
			}
		}

		/// <summary>
		/// The width (in pixels) of the message to show. READ-ONLY.
		/// </summary>
		public int MessageLabelWidthMax
		{
			get { return MessageLabelWidthMaxDef; }
		}

		private TooltipWindowModes _Mode;
		/// <summary>
		/// The message window mode. If this is <see cref="TooltipWindowModes.Image"/> or
		/// <see cref="TooltipWindowModes.CachedImage"/>, <see cref="Message"/> must contain the name of the file to show.
		/// Notifies property observers. READ-ONLY.
		/// </summary>
		public TooltipWindowModes Mode
		{
			get { return _Mode; }
			protected set
			{
				_Mode = value;
				_ModeImage = null;

				// Notify property observers.
				OnPropertyChanged( "Mode" );
				OnPropertyChanged( "ModeImage" );
			}
		}

		private BitmapSource _ModeImage;
		/// <summary>
		/// Mode image. This may be null. Notifies property observers. READ-ONLY.
		/// </summary>
		public BitmapSource ModeImage
		{
			get
			{
				if( _ModeImage != null )
					return _ModeImage;

				switch( Mode )
				{
					case TooltipWindowModes.Done:
					case TooltipWindowModes.Info:
					case TooltipWindowModes.Warning:
						string fileName = string.Format( ControlsConsts.ModeImageFileNameTemplate, Mode );
						_ModeImage = ImagePreviewsCache.GetCachedOrLoadFirstFrame( fileName, fileName, false );
						break;
					case TooltipWindowModes.Image:
						_ModeImage = GraphicUtils.LoadResizeImage( Message, KernelConsts.FullSizedImagePreviewWidthInt
							, KernelConsts.FullSizedImagePreviewHeightInt, false );
						break;
					case TooltipWindowModes.CachedImage:
						_ModeImage = ImagePreviewsCache.GetCachedOrLoadFirstFrame( Message, Message, true );
						break;
					default:
						throw ValidationUtils.Unhandled( Mode, "Mode" );
				}

				// Notify property observers. Avoid stack overflows when notifying about null values.
				if( _ModeImage != null )
					OnPropertyChanged( "ModeImage" );

				return _ModeImage;
			}
			protected set
			{
				_ModeImage = value;

				// Notify property observers.
				OnPropertyChanged( "ModeImage" );
			}
		}

		/// <summary>
		/// The timespan (in seconds) for which to show the window. READ-ONLY.
		/// </summary>
		public double AutoCloseTimespan
		{
			get { return TextUtils.NoContent( Message ) ? NoMessageAutoCloseTimespan : WithMessageAutoCloseTimespan; }
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void ViewWindow_Loaded( object sender, RoutedEventArgs e )
		{
			ControlsUtils.RepositionTopRightToOwner( this );

			// Prepare to automatically close the window.
			if( AutoCloseTimespan > 0 )
			{
				AutoCloseTimer = new DispatcherTimer();
				AutoCloseTimer.Interval = TimeSpan.FromSeconds( AutoCloseTimespan );
				AutoCloseTimer.Tick += new EventHandler( AutoCloseTimer_Tick );
				AutoCloseTimer.IsEnabled = true;
			}
		}

		private void ViewWindow_Closing( object sender, System.ComponentModel.CancelEventArgs e )
		{
			if( AutoCloseTimer != null )
				AutoCloseTimer.Stop();
		}

		private void ViewWindow_PreviewMouseDown( object sender, System.Windows.Input.MouseButtonEventArgs e )
		{
			Close();
		}

		private void AutoCloseTimer_Tick( object sender, EventArgs e )
		{
			Close();
		}

		#endregion Event handlers

		#region Public static methods

		/// <summary>
		/// Shows the window as a tooltip, using the global instance (<see cref="Instance"/>).
		/// </summary>
		/// <param name="message">Message to show. If "mode" is <see cref="TooltipWindowModes.Image"/>, this must contain the
		/// name of the file to show. This may be null or empty.</param>
		/// <param name="mode">Window mode.</param>
		public static void ShowTooltip( string message = null, TooltipWindowModes mode = TooltipWindowModes.Done )
		{
			if( ControlsUtils.GetAppDispatcher() == null )
				return;

			CloseTooltip();

			Instance = new TooltipWindow( message, mode );
			ControlsUtils.SetWindowOwnerToMainWindow( Instance );
			Instance.Show();
		}

		/// <summary>
		/// Closes the global instance (<see cref="Instance"/>) of the tooltip window (if it's visible).
		/// </summary>
		public static void CloseTooltip()
		{
			if( ControlsUtils.GetAppDispatcher() == null )
				return;

			if( Instance != null )
				Instance.Close();

			Instance = null;
		}

		#endregion Public static methods
	}
}
