////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.ActionProcessor;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.ActionClient
{
	/// <summary>
	/// Contract for an action client.
	/// </summary>
	/// <remarks>
	/// <para>Expired action requests are not updated. Expired action responses are not processed.</para>
	/// 
	/// <para>All the sheafs are locked with <see cref="SheafLocks.Shared"/>, unless they will be saved (in which case they are
	/// locked with <see cref="SheafLocks.Update"/>).</para>
	/// </remarks>
	public abstract class ActionClient : IActionClient
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionClient()
		{
		}

		#endregion Constructors

		#region Private static methods

		/// <summary>
		/// Throws a <see cref="RevokedVaultException"/> exception if the specified error response indicates that the service
		/// vault is revoked.
		/// </summary>
		/// <param name="errorResponse">Error response. This may be null.</param>
		private static void ThrowIfRevokedVault( ErrorActionResponse errorResponse )
		{
			if( (errorResponse == null) || TextUtils.NoContent( errorResponse.ErrorTypeName ) )
				return;

			Type errorType = PortableTypes.GetGlobal( errorResponse.ErrorTypeName, typeof( RevokedVaultException ), false );
			if( errorType == null )
				return;

			throw (RevokedVaultException)Activator.CreateInstance( errorType, errorResponse.Message );
		}

		#endregion Private static methods

		#region Public methods

		/// <summary>
		/// Updates the specified action request before it's sent to the service.
		/// </summary>
		/// <param name="actionRequest">Action request to update. This may be modified in here if the request's
		/// <see cref="Vela.ActionProcessor.ActionRequest.NeedsUpdateBeforeSending"/> is true; it's also automatically
		/// signed afterwards.</param>
		/// <param name="clientContext">The action client context.</param>
		/// <param name="transaction">The transaction with which to load documents. This is separate from the client context
		/// in order to ensure that it's not used by the derived action client to make changes in the storehouse and expect
		/// them to be rolled back if the action response can't be processed.</param>
		/// <param name="updateRequestMethodInfo">The information of the method which can update the action request.</param>
		/// <returns>Returns the true if the action request must be sent to the service, and false if it must never be sent.
		/// </returns>
		/// <remarks>
		/// <para>See <see cref="Vela.ActionProcessor.ActionRequest.NeedsUpdateBeforeSending"/> for details.
		/// </para>
		/// </remarks>
		public bool UpdateActionRequest( ActionRequest actionRequest, ActionClientContext clientContext
			, IStorehouseTransaction transaction, UpdateRequestMethodInfo updateRequestMethodInfo )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( clientContext, "clientContext" );
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( updateRequestMethodInfo, "updateRequestMethodInfo" );

			if( !actionRequest.NeedsUpdateBeforeSending )
			{
				throw new ValidationException( string.Format( "The action request with the locator \"{0}\" doesn't require"
					+ " to be updated.", actionRequest.InstanceLocator ) );
			}

			// Don't update expired action requests.
			KernelUtils.HasExpired( actionRequest );

			LogManager.DebugAddFormat( "Updating the action request with the locator \"{0}\" and the type \"{1}\", in order"
					+ " to send it to the service with the UPN \"{2}\"."
				, actionRequest.InstanceLocator, actionRequest.GetType().FullName, clientContext.RemoteUpn );

			// Update the action request.
			object updateResult = updateRequestMethodInfo.MethodInfo.Invoke( this, new object[] { actionRequest
					, clientContext } );

			if( updateResult == null )
			{
				throw new ValidationException( string.Format( "The method which has updated the action request with the"
					+ " locator \"{0}\", has not returned anything.", actionRequest.InstanceLocator ) );
			}

			ValidationUtils.ObjectType<bool>( updateResult, "updateResult" );
			bool requestUpdated = (bool)updateResult;

			if( requestUpdated )
			{
				// Automatically sign the modified request.

				IDeserializer deserializer;
				SpeckleFilter vaultFilter = new SpeckleFilter( Vault.LocalUpnFingerprintFilterIdentifier
					, actionRequest.InitiatorUpnFingerprint );
				Vault clientVault = transaction.CollectFirstSheaf( new PortableTypes( typeof( Vault ) )
					, new SpeckleFilters( vaultFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared
					, out deserializer ) as Vault;
				ValidationUtils.ObjectHasContent( clientVault, "clientVault" );

				actionRequest.Sign( clientVault.ProcessingIdentity, null );
			}

			return requestUpdated;
		}

		/// <summary>
		/// Processes the specified action response.
		/// </summary>
		/// <param name="actionResponse">The action response to process.</param>
		/// <param name="clientContext">The action client context.</param>
		/// <param name="processResponseMethodInfo">The information of the method which can process the action response.</param>
		public void ProcessResponse( ActionResponse actionResponse, ActionClientContext clientContext
			, ProcessResponseMethodInfo processResponseMethodInfo )
		{
			ValidationUtils.ObjectHasContent( actionResponse, "actionResponse" );
			ValidationUtils.ObjectHasContent( clientContext, "clientContext" );
			ValidationUtils.ObjectHasContent( processResponseMethodInfo, "processResponseMethodInfo" );

			IStorehouseTransaction transaction = clientContext.Transaction;

			// Don't process expired action responses.
			KernelUtils.HasExpired( actionResponse );

			LogManager.DebugAddFormat( "The action response with the locator \"{0}\" was received from the service with"
				+ " the UPN \"{1}\".", actionResponse.InstanceLocator, clientContext.RemoteUpn );

			// Because error action responses are not derived from type of the "actionResponse", they must be sent as a
			// separate parameter. The success and error response can't be both null or non-null.
			ErrorActionResponse errorResponse = actionResponse as ErrorActionResponse;
			ActionResponse successResponse = (errorResponse == null) ? actionResponse : null;

			ThrowIfRevokedVault( errorResponse );

			// Create the action response context.
			ActionResponseContext actionResponseContext = new ActionResponseContext( clientContext );

			// Process the action response.
			processResponseMethodInfo.MethodInfo.Invoke( this, new object[] { successResponse, errorResponse
				, actionResponseContext } );
		}

		/// <summary>
		/// Processes the specified last frame action response.
		/// </summary>
		/// <param name="actionRequest">The action request which was processed.</param>
		/// <param name="actionResponse">The last frame action response to process.</param>
		/// <param name="clientContext">The action client context.</param>
		/// <param name="processLastFrameResponseMethodInfo">The information of the method which can process the last frame
		/// action response. This may be null.</param>
		public void ProcessLastFrameResponse( ActionRequest actionRequest, LastFrameActionResponse actionResponse
			, ActionClientContext clientContext, ProcessLastFrameResponseMethodInfo processLastFrameResponseMethodInfo )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( actionResponse, "actionResponse" );
			ValidationUtils.ObjectHasContent( clientContext, "clientContext" );

			LogManager.DebugAddFormat( "The last frame action response with the locator \"{0}\" and the type \"{1}\" was"
					+ " received from the service with the UPN \"{2}\"."
				, actionResponse.InstanceLocator, actionResponse.GetType().FullName, clientContext.RemoteUpn );

			if( processLastFrameResponseMethodInfo == null )
				return;

			IStorehouseTransaction transaction = clientContext.Transaction;

			// Don't process expired action responses.
			KernelUtils.HasExpired( actionResponse );

			// Create the action response context.
			ActionResponseContext actionResponseContext = new ActionResponseContext( clientContext );

			// Process the action response.
			processLastFrameResponseMethodInfo.MethodInfo.Invoke( this, new object[] { actionRequest, actionResponse
				, actionResponseContext } );
		}

		#endregion Public methods
	}
}
