////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Net;
using System.Net.PeerToPeer;
using System.Net.Sockets;
using System.Reflection;
using System.Threading;
using Vela.ActionProcessor;
using Vela.Cloning;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Networking;
using Vela.PeerDocuments;
using Vela.Pnrp;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.ActionClient
{
	/// <summary>
	/// A client channel which sends an action request to a service, receives an action response from the service and dispatches
	/// it to the appropriate client.
	/// </summary>
	/// <remarks>
	/// <para>The action request is sent on a background thread, and the action response is received on the same thread.</para>
	/// 
	/// <para>Expired action requests and responses are not processed.</para>
	/// 
	/// <para>All the sheafs are locked with <see cref="SheafLocks.Shared"/>, unless they will be saved (in which case they are
	/// locked with <see cref="SheafLocks.Update"/>).</para>
	/// </remarks>
	public class ActionClientChannel
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionClientChannel()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="channelContext">Action client channel context. This is cloned in here, so it doesn't have to be
		/// locked.</param>
		/// <param name="pendingRequest">The pending manager request request which has to be processed by the action client
		/// channel.</param>
		/// <param name="channelFinished">Event raised when an action client channel finishes processing an action request.
		/// </param>
		public ActionClientChannel( ActionClientChannelContext channelContext, PendingRequest pendingRequest
			, ChannelFinishedDelegate channelFinished )
		{
			ValidationUtils.ObjectHasContent( pendingRequest, "pendingRequest" );
			ValidationUtils.ObjectHasContent( channelFinished, "channelFinished" );

			ChannelContext = Cloner.Deep( channelContext );
			PendingRequest = pendingRequest;
			ChannelFinished = channelFinished;

			Start();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// Action client channel context.
		/// </summary>
		private ActionClientChannelContext ChannelContext;

		/// <summary>
		/// The pending manager request request which has to be processed by the action client channel.
		/// </summary>
		private PendingRequest PendingRequest;

		/// <summary>
		/// Event raised when an action client channel finishes processing an action request.
		/// </summary>
		private ChannelFinishedDelegate ChannelFinished;

		/// <summary>
		/// The thread where the client channel processes action requests and responses.
		/// </summary>
		private Thread ChannelThread;

		/// <summary>
		/// The synchronization object for the <see cref="StopChannelThread"/> property.
		/// </summary>
		private readonly object StopChannelThreadLock = new object();

		private bool _StopChannelThread = false;
		/// <summary>
		/// If this is true then the channel thread must stop as soon as possible.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// </remarks>
		private bool StopChannelThread
		{
			get
			{
				lock( StopChannelThreadLock )
				{
					return _StopChannelThread;
				}
			}
			set
			{
				lock( StopChannelThreadLock )
				{
					_StopChannelThread = value;
				}
			}
		}

		#endregion Private fields

		#region Public non-serializable properties

		/// <summary>
		/// The instance locator of the action request to process. READ-ONLY.
		/// </summary>
		public Locator RequestLocator
		{
			get { return PendingRequest.RequestLocator; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Instantiate an action client which can process the specified action response.
		/// </summary>
		/// <param name="actionRequestType">The type of action request which is being processed.</param>
		/// <param name="actionResponseType">The type of action response to process.</param>
		/// <param name="updateRequestMethodInfo">Is set with the information of the method which can update the action request
		/// which is being processed. This may be null.</param>
		/// <param name="processResponseMethodInfo">Is set with the information of the method which can process the received
		/// action response.</param>
		/// <param name="processLastFrameResponseMethodInfo">Is set with the information of the method which can process the
		/// received last frame action response.</param>
		/// <returns>Returns the action client.</returns>
		private IActionClient InstantiateClient( Type actionRequestType, Type actionResponseType
			, out UpdateRequestMethodInfo updateRequestMethodInfo, out ProcessResponseMethodInfo processResponseMethodInfo
			, out ProcessLastFrameResponseMethodInfo processLastFrameResponseMethodInfo )
		{
			ValidationUtils.ObjectHasContent( actionRequestType, "actionRequestType" );
			ValidationUtils.ObjectHasContent( actionResponseType, "actionResponseType" );

			// Find the cached method information for updating an action request.
			string actionRequestTypeName = actionRequestType.FullName;
			updateRequestMethodInfo = ChannelContext.UpdateRequestMethodInfos.Find( actionRequestTypeName )
				as UpdateRequestMethodInfo;

			// Find the cached method information for processing an action response.
			string actionResponseTypeName = actionResponseType.FullName;
			processResponseMethodInfo = ChannelContext.ProcessResponseMethodInfos.Find( actionResponseTypeName );
			if( processResponseMethodInfo == null )
			{
				throw new ValidationException( string.Format( "The action client can't process action responses with the type"
					+ " \"{0}\".", actionResponseTypeName ) );
			}

			// Find the cached method information for processing a last frame action response.
			processLastFrameResponseMethodInfo = ChannelContext.ProcessLastFrameResponseMethodInfos.Find( actionRequestTypeName )
				as ProcessLastFrameResponseMethodInfo;

			// Instantiate the action client.
			Type clientType = PortableTypes.FindGlobal( processResponseMethodInfo.ClientTypeName );
			IActionClient client = (clientType != null) ? (Activator.CreateInstance( clientType, true ) as IActionClient) : null;
			if( client == null )
			{
				throw new ValidationException( string.Format( "An action client with the type \"{0}\" could not be"
					+ " instantiated.", processResponseMethodInfo.ClientTypeName ) );
			}

			return client;
		}

		/// <summary>
		/// Validates the specified action response. If the action response is invalid, an exception is thrown.
		/// </summary>
		/// <param name="actionResponse">Action response to validate.</param>
		/// <param name="isFirstFrame">This is true if "actionResponse" is the first frame of the action response.</param>
		/// <param name="requiredActionResponseType">The required type of the action response associated to the current action
		/// request. An action service must return an action response derived either from this type or from
		/// <see cref="ErrorActionResponse"/>.</param>
		/// <param name="tally">Tally. This may be null if the action request is derived from
		/// <see cref="AnonymousActionRequest"/>.</param>
		/// <param name="remoteUpn">Remote UPN.</param>
		/// <param name="remoteIpAddress">Remote IP address.</param>
		private void ValidateActionResponse( ActionResponse actionResponse, bool isFirstFrame, Type requiredActionResponseType
			, Tally tally, Upn remoteUpn, byte[] remoteIpAddress )
		{
			ValidationUtils.ObjectHasContent( actionResponse, "actionResponse" );
			ValidationUtils.ObjectHasContent( requiredActionResponseType, "requiredActionResponseType" );
			ValidationUtils.ObjectHasContent( remoteUpn, "remoteUpn" );
			ValidationUtils.ArrayHasContent( remoteIpAddress, "remoteIpAddress" );

			// Verify the type.
			Type actualActionResponseType = actionResponse.GetType();
			if( !requiredActionResponseType.IsAssignableFrom( actualActionResponseType )
				&& !typeof( LastFrameActionResponse ).IsAssignableFrom( actualActionResponseType )
				&& !typeof( ErrorActionResponse ).IsAssignableFrom( actualActionResponseType ) )
			{
				throw new ValidationException( string.Format( "The action response with the locator \"{0}\" has the type \"{1}\""
						+ " instead of either \"{2}\", \"{3}\" or \"{4}\"."
					, actionResponse.InstanceLocator, actualActionResponseType.FullName, requiredActionResponseType.FullName
					, typeof( LastFrameActionResponse ).FullName, typeof( ErrorActionResponse ).FullName ) );
			}

			if( actionResponse as ErrorActionResponse != null )
			{
				// Check that the error type may be cross service. Otherwise, if the client were to throw that exception, it
				// could interfere with the normal flow, like in the case of an abort exception.

				ErrorActionResponse errorResponse = (ErrorActionResponse)actionResponse;
				if( TextUtils.HasContent( errorResponse.ErrorTypeName ) )
					PortableTypes.GetGlobal( errorResponse.ErrorTypeName, typeof( CrossServiceException ), true );
			}

			IIdentityPublisher remoteProcessingIdentity;

			if( actionResponse as ActionResponseWithInitiator != null )
			{
				ValidationUtils.ObjectHasContent( tally, "tally" );

				remoteProcessingIdentity = tally.ProcessingIdentity;
			}
			else if( actionResponse as AnonymousActionResponse != null )
			{
				// Error responses are anonymous.

				remoteProcessingIdentity = ((AnonymousActionResponse)actionResponse).RemoteProcessingIdentity;

				#region Verify the UPN of the identity from the response

				// This must be done in order to ensure that criminals don't register (on multiple computers) the expected
				// service UPN (but with fake identities).

				bool foundCorrectUpn = false;
				foreach( SolitaryIdentityPublisher identityPublisher in remoteProcessingIdentity.Identities )
				{
					if( identityPublisher.Identity.Upn.Equals( remoteUpn ) )
					{
						foundCorrectUpn = true;

						break;
					}
				}

				if( !foundCorrectUpn )
				{
					throw new ValidationException( string.Format( "The identity of the peer from the IP address \"{0}\" doesn't"
							+ " have the expected UPN \"{1}\"."
						, NetworkingUtils.GetTextualIpAddress( remoteIpAddress ), remoteUpn ) );
				}

				#endregion Verify the UPN of the identity from the response
			}
			else
			{
				throw new ValidationException( string.Format( "An action client may process action responses derived from"
						+ " either \"{0}\" or \"{1}\", not of type \"{2}\"."
					, typeof( ActionResponseWithInitiator ), typeof( AnonymousActionResponse ).FullName
					, actionResponse.GetType().FullName ) );
			}

			// Verify the signature. The signature of the secondary frames of a multiframe action response are not verified if
			// they are derived from "IMultiframeActionResponse".
			if( isFirstFrame || (actionResponse as IMultiframeActionResponse == null) )
				actionResponse.VerifySignature( remoteProcessingIdentity );
		}

		/// <summary>
		/// Updates the action progress with the specified locator, according to the specified exception.
		/// </summary>
		/// <param name="actionProgressLocator">The instance locator of the action progress to update.</param>
		/// <param name="actionRequest">The locator of the action request which is currently being processed. This may be null.
		/// </param>
		/// <param name="exception">Exception with which to perform the update.</param>
		/// <remarks>
		/// <para>The progress is updated only if it exists in the storehouse and the existing processing state is
		/// <see cref="ActionProcessingStates.Pending"/> or <see cref="ActionProcessingStates.Retry"/>. In any other case it
		/// means that the action request was successfully processed, but an error has occurred afterwards.</para>
		/// 
		/// <para>The update is performed on a new transaction.</para>
		/// 
		/// <para>Although unlikely, it's still possible that the attempt index of the recipient access state (whose recipient
		/// UPN fingerprint is that of the action progress with the specified locator) is incremented twice while an action
		/// request is processed. This can happen if an action response is received but an error occurrs afterwards (and this
		/// method is called). However, this is not a problem since this issue would merely increase the timespan until the next
		/// processing attempt.</para>
		/// </remarks>
		private void UpdateActionProgressWhenException( Locator actionProgressLocator, ActionRequest actionRequest
			, Exception exception )
		{
			ValidationUtils.ObjectHasContent( actionProgressLocator, "actionProgressLocator" );
			ValidationUtils.ObjectHasContent( exception, "exception" );

			if( exception as SkipProcessingActionRequestException != null )
				return;

			ActionClientProgress progress;

			// Create a transaction.
			IStorehouseTransaction transaction = ChannelContext.StorehouseProvider.CreateTransaction();

			try
			{
				IDeserializer deserializer;

				// Load the progress.
				progress = transaction.LoadSheaf( actionProgressLocator, StorehouseRealms.ActionProgresses
					, typeof( ActionClientProgress ), SheafLocks.Update, out deserializer ) as ActionClientProgress;

				// Update the progress.
				if( (progress != null)
					&& (
						(progress.ProcessingState == ActionProcessingStates.Pending)
						|| (progress.ProcessingState == ActionProcessingStates.Retry)
					) )
				{
					Upn remoteUpn = (actionRequest != null) ? actionRequest.RecipientUpn : null;

					ChannelContext.RecipientAccessStatesManager.UpdateProgressToRetryOrError( progress, exception
						, progress.LastProcessingAttempt.SecondaryProcessingState );
					ActionClientUtils.AfterUpdateProcessingState( progress, null, remoteUpn, transaction );
				}

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			// Notify the action processor host to update the UI. Do this outside the main transaction because another one may be
			// created and may try to load the sheafs which are locked by the main transaction.
			if( progress != null )
			{
				ActionClientHostUpdateUiContext updateUiContext = new ActionClientHostUpdateUiContext(
					ChannelContext.ProcessorConfiguration, progress, actionRequest, null, null, exception, true );
				ChannelContext.ActionClientHost.UpdateUi( updateUiContext );
			}
		}

		#region Resolve the remote IP address

		/// <summary>
		/// Gets the existing remote IP address.
		/// </summary>
		/// <param name="remoteUpnFingerprint">The UPN fingerprint of the remote peer. This may be null.</param>
		/// <param name="tally">Tally. This may be null.</param>
		/// <param name="actionProgress">The progress of the action request to process.</param>
		/// <param name="alternativePeerAddress">Is set with the existing alternative peer address. This may be null.</param>
		/// <returns>Returns the existing remote IP address. This may be null or empty.</returns>
		/// <remarks>
		/// <para>The alternative IP address has priority over the one from the ephemerid.</para>
		/// </remarks>
		private byte[] GetExistingRemoteIpAddress( Locator remoteUpnFingerprint, Tally tally, ActionClientProgress actionProgress
			, out AlternativePeerAddress alternativePeerAddress )
		{
			ValidationUtils.ObjectHasContent( actionProgress, "actionProgress" );

			byte[] remoteIpAddress = null;

			IAlternativePeerAddressContainer alternativeContainer = actionProgress as IAlternativePeerAddressContainer;

			if( alternativeContainer != null )
			{
				alternativePeerAddress = alternativeContainer.AlternativePeerAddress;
				if( alternativePeerAddress != null )
					remoteIpAddress = alternativePeerAddress.GetIpAddressBuffer();
			}
			else
				alternativePeerAddress = null;

			if( ArrayUtils.NoContent( remoteIpAddress ) && (tally != null) )
			{
				alternativePeerAddress = tally.AlternativePeerAddress;
				if( alternativePeerAddress != null )
					remoteIpAddress = alternativePeerAddress.GetIpAddressBuffer();
			}

			if( ArrayUtils.HasContent( remoteIpAddress ) )
				return remoteIpAddress;

			// Load the ephemerid (on a separate transaction so as to not affect the scalability of the action processor).
			if( remoteUpnFingerprint != null )
			{
				Ephemerid ephemerid = ActionProcessorStorehouseUtils.LoadEphemerid( ChannelContext.StorehouseProvider
					, remoteUpnFingerprint );

				// Get the IP address of the service.
				if( ephemerid != null )
					return ephemerid.GetIpAddressBuffer();
			}

			return null;
		}

		/// <summary>
		/// Resolves the classifier of the specified UPN, using the underlying infrastructure.
		/// </summary>
		/// <param name="remoteUpn">UPN whose classifier to resolve.</param>
		/// <returns>Returns the found peer records. This may be null or empty.</returns>
		private PeerNameRecordCollection SystemResolvePeerClassifier( Upn remoteUpn )
		{
			ValidationUtils.ObjectHasContent( remoteUpn, "remoteUpn" );

			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				PeerName peerName = new PeerName( remoteUpn.Classifier, PeerNameType.Unsecured );

				PeerNameResolver resolver = new PeerNameResolver();
				PeerNameRecordCollection peerRecords = resolver.Resolve( peerName, PnrpConsts.Cloud );

				LogManager.DebugAddFormat( "The underlying infrastructure has resolved the remote UPN \"{0}\" (classifier"
						+ " \"{1}\") to {2} IP addresses."
					, remoteUpn, remoteUpn.Classifier, (peerRecords != null) ? peerRecords.Count : 0 );

				return peerRecords;
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		/// <summary>
		/// Sends an action request and expects an action response. If a valid action response is not received, an exception is
		/// thrown; for the action response to be valid, its signature must be verifiable with the received identity publisher.
		/// </summary>
		/// <param name="initiatorUpnFingerprint">The fingerprint of the UPN of the initiator of the current action request.
		/// </param>
		/// <param name="remoteUpn">Remote UPN.</param>
		/// <param name="overridingProcessingPort">The overriding processing port where the local peer must send action requests
		/// to the remote peer. If this is 0 or a negative number, it's ignored.</param>
		/// <param name="remoteIpAddress">IP address of the service where to send the action request.</param>
		private void GetResolveUpnResponse( Locator initiatorUpnFingerprint, Upn remoteUpn, UInt16 overridingProcessingPort
			, byte[] remoteIpAddress )
		{
			ValidationUtils.ObjectHasContent( initiatorUpnFingerprint, "initiatorUpnFingerprint" );
			ValidationUtils.ObjectHasContent( remoteUpn, "remoteUpn" );
			ValidationUtils.ArrayHasContent( remoteIpAddress, "remoteIpAddress" );

			LogManager.DebugAddFormat( "Attempting to authenticate the remote peer with the UPN \"{0}\" to the IP address"
				+ " \"{1}\".", remoteUpn, NetworkingUtils.GetTextualIpAddress( remoteIpAddress ) );

			try
			{
				ISerializer serializer = ChannelContext.StorehouseProvider.Serializer;
				ActionProcessorConfiguration processorConfiguration = ChannelContext.ProcessorConfiguration;
				SteganographicAlg steganographicAlg = remoteUpn.UpnAlg.SteganographicAlg;
				ICommunicationFrameAlg communicationFrameAlg = remoteUpn.UpnAlg.CommunicationFrameAlg;

				#region Create the action request

				ICryptoAlgorithmPool cap = ChannelContext.CryptoAlgorithmPool;
				ResolveUpnAnonymousActionRequest actionRequest = new ResolveUpnAnonymousActionRequest( CryptoUtils.GetLocator()
					, cap.GetMacAlg(), cap.GetAsymmSigningPadding(), serializer.GetHashSerializer( false )
					, initiatorUpnFingerprint, remoteUpn, PeerDocumentsConsts.ActionRequestValidity, 0 );

				#endregion Create the action request

				#region Encrypt the action request

				// Put the action request in an envelope symmetrically encrypted with a UPN.
				ICoreAlgorithmPool coreap = ChannelContext.CoreAlgorithmPool;
				UpnEncryptedEnvelope requestEnvelope = coreap.GetUpnEncryptedEnvelope( remoteUpn.Fingerprint );

				requestEnvelope.EncryptedDocumentBlockSize = coreap.GetActionRequestEnvelopeEncryptedDocumentBlockSize();
				requestEnvelope.EncryptableDocument = actionRequest;

				byte[] encryptedRequest = requestEnvelope.SaveToBuffer( null, remoteUpn.SymmEncryptionKey, serializer );

				#endregion Encrypt the action request

				#region Send the action request (and receive the action response)

				// Obfuscate the action request.
				byte[] obfuscatedRequest = steganographicAlg.Obfuscate( encryptedRequest, remoteUpn.SymmEncryptionKey
					, remoteUpn.Fingerprint.Value );

				CommunicationFrame communicationFrame = new CommunicationFrame( communicationFrameAlg
					, processorConfiguration.CommunicationTimeoutPerBlob, processorConfiguration.CommunicationFrameSizeMax );

				UInt16 remotePort = (overridingProcessingPort > 0) ? overridingProcessingPort : remoteUpn.Port;
				byte[] obfuscatedResponse = communicationFrame.SendRequestAndReceiveResponse( remoteIpAddress, remotePort
					, obfuscatedRequest );

				#endregion Send the action request (and receive the action response)

				#region Receive the action response

				// Reveal the action response.
				byte[] revealedResponse = steganographicAlg.Reveal( obfuscatedResponse, actionRequest.ResponseEncryptionKey
					, remoteUpn.Fingerprint.Value );

				// Decrypt the action response.
				IDeserializer deserializer;
				SymmEncryptedEnvelope responseEnvelope = SymmEncryptedEnvelope.LoadFromBuffer( revealedResponse
					, actionRequest.ResponseEncryptionKey, typeof( ActionResponse ), out deserializer );
				ValidationUtils.ObjectHasContent( responseEnvelope, "responseEnvelope" );

				if( responseEnvelope as UpnEncryptedEnvelope != null )
				{
					throw new ValidationException( string.Format( "An action response may not be encrypted in an envelope with"
						+ " the \"{0}\" type.", responseEnvelope.GetType().FullName ) );
				}

				ActionResponse actionResponse = (ActionResponse)responseEnvelope.EncryptableDocument;

				ValidateActionResponse( actionResponse, true, actionRequest.ActionResponseType, null, remoteUpn
					, remoteIpAddress );

				#endregion Receive the action response

				string message = string.Format( "The remote peer with the UPN \"{0}\" has been authenticated; its IP address is"
					+ " \"{1}\".", remoteUpn, NetworkingUtils.GetTextualIpAddress( remoteIpAddress ) );
				LogManager.Add( message );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				LogManager.DebugAddFormat( "Failed to authenticate the remote peer with the UPN \"{0}\".", remoteUpn );

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Resolves the remote IP address of the specified UPN. An exception is thrown if the UPN can't be resolved.
		/// </summary>
		/// <param name="initiatorUpnFingerprint">The fingerprint of the UPN of the initiator of the current action request.
		/// </param>
		/// <param name="remoteUpn">Remote UPN to resolve to an IP address.</param>
		/// <param name="overridingProcessingPort">The overriding processing port where the local peer must send action requests
		/// to the remote peer. If this is 0 or a negative number, it's ignored.</param>
		/// <returns>Returns the remote IP address.</returns>
		private byte[] ResolveRemoteIpAddressWithPnrp( Locator initiatorUpnFingerprint, Upn remoteUpn
			, UInt16 overridingProcessingPort )
		{
			ValidationUtils.ObjectHasContent( initiatorUpnFingerprint, "initiatorUpnFingerprint" );
			ValidationUtils.ObjectHasContent( remoteUpn, "remoteUpn" );

			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			string securelyErrorMessage = string.Format( "Failed to securely resolve the UPN \"{0}\" (classifier \"{1}\")."
				+ " It's possible that the recipient peer's application is not running.", remoteUpn, remoteUpn.Classifier );
			Exception lastException = null;

			try
			{
				PeerNameRecordCollection peerRecords = SystemResolvePeerClassifier( remoteUpn );
				if( ListUtils.NoContent( peerRecords ) )
				{
					string errorMessage = string.Format( "The underlying PNRP peer resolver has failed to resolve the UPN"
						+ " \"{0}\" (classifier \"{1}\"). It's possible that the recipient peer's application is not"
						+ " running.", remoteUpn, remoteUpn.Classifier );
					throw new ValidationException( errorMessage );
				}

				ChannelContext.RecipientAccessStatesManager.ResetOnSuccessfulUpnResolution( remoteUpn.Fingerprint );

				if( StopChannelThread )
					throw new AbortLongOperationException();

				List<IPAddress> ipAddresses = NetworkingUtils.GetPeerIpAddressesByPriority( peerRecords );

				// Iterate through the found IP addresses and send an action request to each; stop when a signed answer comes
				// back.
				foreach( IPAddress ipAddress in ipAddresses )
				{
					try
					{
						byte[] ipAddressBuffer = ipAddress.GetAddressBytes();

						// This will throw an exception if it can't get a valid response from the service with the specified
						// IP address.
						GetResolveUpnResponse( initiatorUpnFingerprint, remoteUpn, overridingProcessingPort, ipAddressBuffer );

						// "RecipientAccessStatesManager.RecipientIsAccessibleAndProcessing" should be called here, but if that
						// were to happen and if the recipient action service were to throw a "RetryCrossServiceException"
						// exception (for example, if the account between the peers doesn't exist yet), the action client would
						// retry to send the action request as often as possible, fact which would uselessly waste network
						// resources.

						return ipAddressBuffer;
					}
					catch( Exception ex )
					{
						// The loop must continue (if possible).

						if( ExceptionUtils.IsAbort( ex ) )
							throw ExceptionUtils.PreserveStackTrace( ex );

						lastException = ExceptionUtils.PreserveStackTrace( ex );
					}

					if( StopChannelThread )
						throw new AbortLongOperationException();
				}
			}
			catch( Exception ex )
			{
				throw new ValidationException( securelyErrorMessage, ex );
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}

			throw new ValidationException( securelyErrorMessage, lastException );
		}

		/// <summary>
		/// Resolves the remote IP address of the specified UPN. An exception is thrown if the UPN can't be resolved.
		/// </summary>
		/// <param name="remoteUpn">Remote UPN to resolve.</param>
		/// <param name="remoteEntityName">The entity name of the remote peer. This may be null or empty.</param>
		/// <param name="actionRequest">Action request to process.</param>
		/// <param name="overridingProcessingPort">The overriding processing port where the local peer must send action requests
		/// to the remote peer. If this is 0 or a negative number, it's ignored.</param>
		/// <returns>Returns the remote IP address.</returns>
		private byte[] ResolveRemoteIpAddressWithPnrp( Upn remoteUpn, string remoteEntityName, ActionRequest actionRequest
			, UInt16 overridingProcessingPort )
		{
			ValidationUtils.ObjectHasContent( remoteUpn, "remoteUpn" );
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );

			if( (actionRequest as ActionRequestWithInitiator == null) && (actionRequest as AnonymousActionRequest == null) )
			{
				throw new ValidationException( string.Format( "An action client may process only action requests derived"
						+ " from \"{0}\" or from \"{1}\", but it's type is \"{2}\"."
					, typeof( ActionRequestWithInitiator ).FullName, typeof( AnonymousActionRequest ).FullName
					, actionRequest.GetType().FullName ) );
			}

			// If UPN resolvers (like the PNRP) may not be used, don't resolve the remote UPN.
			if( !ChannelContext.ProcessorConfiguration.UseUpnResolvers )
			{
				string message = string.Format( "The action request with the locator \"{0}\" can't be sent to the action service"
					+ " with the UPN \"{1}\". The use of UPN resolvers (like the PNRP) is forbidden, so either ensure that the"
					+ " account with the entity name \"{2}\" (= the remote peer) has a valid alternative peer address, or enable"
					+ " the use of UPN resolvers.", actionRequest.InstanceLocator, remoteUpn, remoteEntityName );
				throw new ValidationException( message );
			}

			ChannelContext.UpnResolutionLocker.Starting( remoteUpn );

			try
			{
				byte[] remoteIpAddress = ResolveRemoteIpAddressWithPnrp( actionRequest.InitiatorUpnFingerprint, remoteUpn
					, overridingProcessingPort );

				// "RecipientAccessStatesManager.RecipientIsAccessibleAndProcessing" should be called here, but if that were to
				// happen and if the recipient action service were to throw a "RetryCrossServiceException" exception (for
				// example, if the account between the peers doesn't exist yet), the action client would retry to send the action
				// request as often as possible, fact which would uselessly waste network resources.

				// Update the ephemerid (on a separate transaction so as to not affect the scalability of the action processor).
				ActionProcessorStorehouseUtils.UpdateEphemerid( ChannelContext.StorehouseProvider, remoteUpn.Fingerprint
					, remoteIpAddress, actionRequest );

				return remoteIpAddress;
			}
			finally
			{
				ChannelContext.UpnResolutionLocker.Ended( remoteUpn );
			}
		}

		/// <summary>
		/// Resolves the remote IP address of the specified UPN. An exception is thrown if the UPN can't be resolved.
		/// </summary>
		/// <param name="initiatorUpnFingerprint">The fingerprint of the UPN of the initiator of the current action request.
		/// </param>
		/// <param name="remoteUpn">Remote UPN to resolve to an IP address.</param>
		/// <param name="remoteHostName">The host name of the remote peer to resolve.</param>
		/// <param name="overridingProcessingPort">The overriding processing port where the local peer must send action requests
		/// to the remote peer. If this is 0 or a negative number, it's ignored.</param>
		/// <returns>Returns the remote IP address.</returns>
		private byte[] ResolveRemoteIpAddressWithDns( Locator initiatorUpnFingerprint, Upn remoteUpn, string remoteHostName
			, UInt16 overridingProcessingPort )
		{
			ValidationUtils.ObjectHasContent( initiatorUpnFingerprint, "initiatorUpnFingerprint" );
			ValidationUtils.ObjectHasContent( remoteUpn, "remoteUpn" );
			ValidationUtils.TextHasContent( remoteHostName, "remoteHostName" );

			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			string securelyErrorMessage = string.Format( "Failed to securely resolve the UPN \"{0}\" (classifier \"{1}\"),"
					+ " using the host name \"{2}\". It's possible that the recipient peer's application is not running."
				, remoteUpn, remoteUpn.Classifier, remoteHostName );
			Exception lastException = null;

			try
			{
				IPAddress[] remoteIpAddresses = NetworkingUtils.ResolveUrl( remoteHostName );
				if( ListUtils.NoContent( remoteIpAddresses ) )
				{
					string errorMessage = string.Format( "The underlying DNS peer resolver has failed to resolve the UPN \"{0}\""
						+ " (classifier \"{1}\"), using the host name \"{2}\". It's possible that the recipient peer's"
						+ " application is not running.", remoteUpn, remoteUpn.Classifier, remoteHostName );
					throw new ValidationException( errorMessage );
				}

				ChannelContext.RecipientAccessStatesManager.ResetOnSuccessfulUpnResolution( remoteUpn.Fingerprint );

				if( StopChannelThread )
					throw new AbortLongOperationException();

				List<IPAddress> ipAddresses = NetworkingUtils.GetPeerIpAddressesByPriority( remoteIpAddresses );

				LogManager.DebugAddFormat( "The underlying infrastructure has resolved the remote UPN \"{0}\" (classifier"
						+ " \"{1}\") to {2} IP addresses, using the host name \"{3}\"."
					, remoteUpn, remoteUpn.Classifier, (ipAddresses != null) ? ipAddresses.Count : 0, remoteHostName );

				// Iterate through the found IP addresses and send an action request to each; stop when a signed answer comes
				// back.
				foreach( IPAddress ipAddress in ipAddresses )
				{
					try
					{
						byte[] ipAddressBuffer = ipAddress.GetAddressBytes();

						// This will throw an exception if it can't get a valid response from the service with the specified
						// IP address.
						GetResolveUpnResponse( initiatorUpnFingerprint, remoteUpn, overridingProcessingPort, ipAddressBuffer );

						// "RecipientAccessStatesManager.RecipientIsAccessibleAndProcessing" should be called here, but if that
						// were to happen and if the recipient action service were to throw a "RetryCrossServiceException"
						// exception (for example, if the account between the peers doesn't exist yet), the action client would
						// retry to send the action request as often as possible, fact which would uselessly waste network
						// resources.

						return ipAddressBuffer;
					}
					catch( Exception ex )
					{
						// The loop must continue (if possible).

						if( ExceptionUtils.IsAbort( ex ) )
							throw ExceptionUtils.PreserveStackTrace( ex );

						lastException = ExceptionUtils.PreserveStackTrace( ex );
					}

					if( StopChannelThread )
						throw new AbortLongOperationException();
				}
			}
			catch( Exception ex )
			{
				throw new ValidationException( securelyErrorMessage, ex );
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}

			throw new ValidationException( securelyErrorMessage, lastException );
		}

		/// <summary>
		/// Resolves the remote IP address of the specified host name. An exception is thrown if the UPN can't be resolved.
		/// </summary>
		/// <param name="remoteUpn">Remote UPN to resolve.</param>
		/// <param name="actionRequest">Action request to process.</param>
		/// <param name="alternativePeerAddress">The alternative peer address (of the remote peer) to use for the alternative
		/// peer resolution. This may be null or empty.</param>
		/// <param name="overridingProcessingPort">The overriding processing port where the local peer must send action requests
		/// to the remote peer. If this is 0 or a negative number, it's ignored.</param>
		/// <param name="exception">Inner exception to use if an exception is thrown. This may be null.</param>
		/// <returns>Returns the remote IP address.</returns>
		private byte[] ResolveRemoteIpAddressWithDns( Upn remoteUpn, ActionRequest actionRequest, string alternativePeerAddress
			, UInt16 overridingProcessingPort, Exception exception )
		{
			ValidationUtils.ObjectHasContent( remoteUpn, "remoteUpn" );
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );

			if( (actionRequest as ActionRequestWithInitiator == null) && (actionRequest as AnonymousActionRequest == null) )
			{
				throw new ValidationException( string.Format( "An action client may process only action requests derived"
						+ " from \"{0}\" or from \"{1}\", but it's type is \"{2}\"."
					, typeof( ActionRequestWithInitiator ).FullName, typeof( AnonymousActionRequest ).FullName
					, actionRequest.GetType().FullName ) );
			}

			ChannelContext.UpnResolutionLocker.Starting( remoteUpn );

			try
			{
				string remoteHostName = AlternativePeerAddress.GetFormattedHostName( alternativePeerAddress
					, remoteUpn.LongHostLabel, remoteUpn.ShortHostLabel );

				byte[] remoteIpAddress = ResolveRemoteIpAddressWithDns( actionRequest.InitiatorUpnFingerprint, remoteUpn
					, remoteHostName, overridingProcessingPort );

				// "RecipientAccessStatesManager.RecipientIsAccessibleAndProcessing" should be called here, but if that were to
				// happen and if the recipient action service were to throw a "RetryCrossServiceException" exception (for
				// example, if the account between the peers doesn't exist yet), the action client would retry to send the action
				// request as often as possible, fact which would uselessly waste network resources.

				// Update the ephemerid (on a separate transaction so as to not affect the scalability of the action processor).
				ActionProcessorStorehouseUtils.UpdateEphemerid( ChannelContext.StorehouseProvider, remoteUpn.Fingerprint
					, remoteIpAddress, actionRequest );

				return remoteIpAddress;
			}
			finally
			{
				ChannelContext.UpnResolutionLocker.Ended( remoteUpn );
			}
		}

		#endregion Resolve the remote IP address

		/// <summary>
		/// Receives an action response frame.
		/// </summary>
		/// <param name="actionRequest">Action request which is being processed.</param>
		/// <param name="isFirstFrame">This is true if the received message is the first frame of the action response.</param>
		/// <param name="remoteUpn">Remote UPN.</param>
		/// <param name="remoteIpAddress">Remote IP address.</param>
		/// <param name="communicationFrame">Communication frame to use for receiving the response.</param>
		/// <param name="socket">Socket to use for receiving the response.</param>
		/// <param name="tally">Tally. This may be null if the action request is derived from
		/// <see cref="AnonymousActionRequest"/>.</param>
		/// <param name="client">The action client.</param>
		/// <param name="clientContext">The action client context.</param>
		/// <returns>Returns the received action response frame.</returns>
		private ActionResponse ReceiveFrameResponse( ActionRequest actionRequest, bool isFirstFrame, Upn remoteUpn
			, byte[] remoteIpAddress, CommunicationFrame communicationFrame, Socket socket, Tally tally, IActionClient client
			, ActionClientContext clientContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( remoteUpn, "remoteUpn" );
			ValidationUtils.ArrayHasContent( remoteIpAddress, "remoteIpAddress" );
			ValidationUtils.ObjectHasContent( communicationFrame, "communicationFrame" );
			ValidationUtils.ObjectHasContent( socket, "socket" );
			ValidationUtils.ObjectHasContent( client, "client" );
			ValidationUtils.ObjectHasContent( clientContext, "clientContext" );

			SteganographicAlg steganographicAlg = remoteUpn.UpnAlg.SteganographicAlg;

			// Receive the action response.
			byte[] obfuscatedResponse = communicationFrame.ReceiveMessage( socket );

			ActionResponse actionResponse;

			// Reveal the action response.
			byte[] revealedResponse = steganographicAlg.Reveal( obfuscatedResponse, actionRequest.ResponseEncryptionKey
				, remoteUpn.Fingerprint.Value );

			#region Get the action response

			// Deserialize the response.
			IDeserializer deserializer;
			IDocument responseDocument = SerializationUtils.LoadFromBuffer( revealedResponse, typeof( IDocument )
				, out deserializer );

			Locator responseLocator = (responseDocument as ILocatable != null) ? ((ILocatable)responseDocument).InstanceLocator
				: null;

			if( responseDocument as SymmEncryptedEnvelope != null )
			{
				SymmEncryptedEnvelope responseEnvelope = responseDocument as SymmEncryptedEnvelope;

				if( responseEnvelope == null )
				{
					throw new ValidationException( string.Format( "The action response with the locator \"{0}\" must be derived"
							+ " from the \"{1}\" type, not from \"{2}\"."
						, responseLocator, typeof( SymmEncryptedEnvelope ).FullName, responseDocument.GetType().FullName ) );
				}

				if( responseEnvelope as UpnEncryptedEnvelope != null )
				{
					throw new ValidationException( string.Format( "The action response with the locator \"{0}\" may not be"
							+ " encrypted in an envelope with the \"{1}\" type."
						, responseLocator, responseEnvelope.GetType().FullName ) );
				}

				// Decrypt the action response.
				IDeserializer encryptedContentDeserializer;
				EncryptedEnvelope.LoadFromEnvelope( responseEnvelope, actionRequest.ResponseEncryptionKey
					, typeof( ActionResponse ), out encryptedContentDeserializer );
				ValidationUtils.ObjectHasContent( responseEnvelope, "responseEnvelope" );

				actionResponse = (ActionResponse)responseEnvelope.EncryptableDocument;
			}
			else if( responseDocument as ActionResponse != null )
			{
				// Action responses may be sent unencrypted because the response frames are encrypted / obfuscated with the
				// response encryption key (see above the revelation step).

				actionResponse = responseDocument as ActionResponse;

				if( actionResponse == null )
				{
					throw new ValidationException( string.Format( "The action response with the locator \"{0}\" must be derived"
							+ " from the \"{1}\" type, not from \"{2}\"."
						, responseLocator, typeof( ActionResponse ).FullName, responseDocument.GetType().FullName ) );
				}
			}
			else
			{
				throw new ValidationException( string.Format( "The action response with the locator \"{0}\" has the type \"{1}\""
						+ " instead of being derived from either \"{2}\" or \"{3}\"."
					, responseLocator, responseDocument.GetType().FullName, typeof( SymmEncryptedEnvelope ).FullName
					, typeof( ActionResponse ).FullName ) );
			}

			#endregion Get the action response

			ValidateActionResponse( actionResponse, isFirstFrame, actionRequest.ActionResponseType, tally, remoteUpn
				, remoteIpAddress );

			// Add new entropy data to the entropy algorithm. The data isn't known to the attackers who physically monitor the
			// action responses received by the client, because they are sent symmetrically encrypted with the encryption key
			// from the action requests (while the data is obtained after decryption).
			//
			// An action request with initiator can't be encrypted in envelopes with the "UpnEncryptedEnvelope" type, so an
			// attacker who knows the UPN of the recipient, can't know the content of the request, and consequently the content
			// of the response.
			if( actionRequest as ActionRequestWithInitiator != null )
				EntropyPool.Add( actionResponse.InstanceLocator.Value );

			return actionResponse;
		}

		/// <summary>
		/// Send the specified action request and receives a uniframe action response.
		/// </summary>
		/// <param name="actionRequest">Action request which is being processed.</param>
		/// <param name="obfuscatedRequest">The obfuscated request to send.</param>
		/// <param name="remoteUpn">Remote UPN.</param>
		/// <param name="remoteIpAddress">Remote IP address.</param>
		/// <param name="communicationFrame">Communication frame to use for receiving the response.</param>
		/// <param name="socket">Socket to use for receiving the response.</param>
		/// <param name="tally">Tally. This may be null if the action request is derived from
		/// <see cref="AnonymousActionRequest"/>.</param>
		/// <param name="processResponseMethodInfo">The information of the method which can process the action response.</param>
		/// <param name="client">The action client.</param>
		/// <param name="clientContext">The action client context.</param>
		/// <remarks>
		/// <para>The transaction (of the client context) is committed in here.</para>
		/// </remarks>
		private void SendRequestAndReceiveUniframeResponse( ActionRequest actionRequest, byte[] obfuscatedRequest
			, Upn remoteUpn, byte[] remoteIpAddress, CommunicationFrame communicationFrame, Socket socket, Tally tally
			, ProcessResponseMethodInfo processResponseMethodInfo, IActionClient client, ActionClientContext clientContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ArrayHasContent( obfuscatedRequest, "obfuscatedRequest" );
			ValidationUtils.ObjectHasContent( remoteUpn, "remoteUpn" );
			ValidationUtils.ArrayHasContent( remoteIpAddress, "remoteIpAddress" );
			ValidationUtils.ObjectHasContent( communicationFrame, "communicationFrame" );
			ValidationUtils.ObjectHasContent( socket, "socket" );
			ValidationUtils.ObjectHasContent( client, "client" );
			ValidationUtils.ObjectHasContent( clientContext, "clientContext" );
			ValidationUtils.ObjectHasContent( processResponseMethodInfo, "processResponseMethodInfo" );

			// Send the action request.
			communicationFrame.SendMessage( socket, obfuscatedRequest );

			// Receive the action response.
			ActionResponse actionResponse = ReceiveFrameResponse( actionRequest, true, remoteUpn, remoteIpAddress
				, communicationFrame, socket, tally, client, clientContext );

			ActionClientProgress progress = clientContext.ActionProgress;

			// Create a transaction.
			IStorehouseTransaction transaction = ChannelContext.StorehouseProvider.CreateTransaction();

			try
			{
				// This transaction is available only while the action request is updated.
				clientContext.Transaction = transaction;

				// Process the action response.
				client.ProcessResponse( actionResponse, clientContext, processResponseMethodInfo );

				// Update the processing state of the action progress.
				ErrorActionResponse errorResponse = actionResponse as ErrorActionResponse;
				ChannelContext.RecipientAccessStatesManager.UpdateProgress( progress, errorResponse == null, true, errorResponse
					, true, actionResponse.SecondaryProcessingState );
				ActionClientUtils.AfterUpdateProcessingState( progress, errorResponse, clientContext.RemoteUpn, transaction );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
			finally
			{
				clientContext.Transaction = null;
			}

			// Notify the action processor host to update the UI. Do this outside the main transaction because
			// another one may be created and may try to load the sheafs which are locked by the main transaction.
			ActionClientHostUpdateUiContext updateUiContext = new ActionClientHostUpdateUiContext(
				ChannelContext.ProcessorConfiguration, progress, actionRequest, clientContext.UpdateRequestContext
				, actionResponse, null, true );
			clientContext.ActionClientHost.UpdateUi( updateUiContext );
		}

		/// <summary>
		/// Send the specified action request and receives a multiframe action response.
		/// </summary>
		/// <param name="actionRequest">Action request which is being processed.</param>
		/// <param name="obfuscatedRequest">The obfuscated request to send.</param>
		/// <param name="remoteUpn">Remote UPN.</param>
		/// <param name="remoteIpAddress">Remote IP address.</param>
		/// <param name="communicationFrame">Communication frame to use for receiving the response.</param>
		/// <param name="socket">Socket to use for receiving the response.</param>
		/// <param name="tally">Tally. This may be null if the action request is derived from
		/// <see cref="AnonymousActionRequest"/>.</param>
		/// <param name="processResponseMethodInfo">The information of the method which can process the action response.</param>
		/// <param name="processLastFrameResponseMethodInfo">The information of the method which can process the last frame
		/// action response. This may be null.</param>
		/// <param name="client">The action client.</param>
		/// <param name="clientContext">The action client context.</param>
		/// <remarks>
		/// <para>The transaction (of the client context) is committed in here.</para>
		/// </remarks>
		private void SendRequestAndReceiveMultiframeResponse( ActionRequest actionRequest, byte[] obfuscatedRequest
			, Upn remoteUpn, byte[] remoteIpAddress, CommunicationFrame communicationFrame, Socket socket, Tally tally
			, ProcessResponseMethodInfo processResponseMethodInfo
			, ProcessLastFrameResponseMethodInfo processLastFrameResponseMethodInfo, IActionClient client
			, ActionClientContext clientContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ArrayHasContent( obfuscatedRequest, "obfuscatedRequest" );
			ValidationUtils.ObjectHasContent( remoteUpn, "remoteUpn" );
			ValidationUtils.ArrayHasContent( remoteIpAddress, "remoteIpAddress" );
			ValidationUtils.ObjectHasContent( communicationFrame, "communicationFrame" );
			ValidationUtils.ObjectHasContent( socket, "socket" );
			ValidationUtils.ObjectHasContent( client, "client" );
			ValidationUtils.ObjectHasContent( clientContext, "clientContext" );
			ValidationUtils.ObjectHasContent( processResponseMethodInfo, "processResponseMethodInfo" );

			Chronometer chrone = new Chronometer();
			chrone.Start();

			// Send the action request.
			communicationFrame.SendMessage( socket, obfuscatedRequest );

			SteganographicAlg steganographicAlg = remoteUpn.UpnAlg.SteganographicAlg;
			ActionClientProgress progress = clientContext.ActionProgress;

			bool isFirstFrame = true;
			bool continueReceiving = true;
			while( continueReceiving )
			{
				// Receive the action response.
				ActionResponse actionResponse = ReceiveFrameResponse( actionRequest, isFirstFrame, remoteUpn, remoteIpAddress
					, communicationFrame, socket, tally, client, clientContext );

				bool isMultiframeResponse = actionResponse as IMultiframeActionResponse != null;

				// Process the action response. An action request which has a multiframe response must be stopped if a uniframe
				// response is received. This is also trigerred for a response of type "ErrorActionResponse".
				if( !isMultiframeResponse || (actionResponse as ErrorActionResponse != null) )
					continueReceiving = false;

				// The last frame action response is processed in a separate manner (since it's a single type).
				if( actionResponse as LastFrameActionResponse == null )
				{
					if( isFirstFrame )
					{
						// Create a transaction. Only the first frame of a multiframe response has access to the transaction.
						IStorehouseTransaction transaction = ChannelContext.StorehouseProvider.CreateTransaction();

						try
						{
							// This transaction is available only while the (first frame of the) action request is processed.
							clientContext.Transaction = transaction;
							client.ProcessResponse( actionResponse, clientContext, processResponseMethodInfo );

							// Commit the transaction.
							transaction.Commit();
						}
						catch( Exception ex )
						{
							// Rolls back the transaction.
							transaction.Rollback();

							throw ExceptionUtils.PreserveStackTrace( ex );
						}
						finally
						{
							clientContext.Transaction = null;
						}
					}
					else
					{
						client.ProcessResponse( actionResponse, clientContext, processResponseMethodInfo );
					}
				}
				else
				{
					client.ProcessLastFrameResponse( actionRequest, (LastFrameActionResponse)actionResponse, clientContext
						, processLastFrameResponseMethodInfo );
				}

				// Check if the channel must stop processing. Do this before the progress is updated because it may have been
				// deleted (from the storehouse).
				if( StopChannelThread )
				{
					if( isMultiframeResponse )
					{
						LogManager.DebugAddFormat( "The action client channel has stopped processing the multiframe response"
								+ " for the action request with the locator \"{0}\" and the type \"{1}\"."
							, actionRequest.InstanceLocator, actionRequest.GetType().FullName );
					}

					throw new AbortLongOperationException();
				}

				// Update the processing state of the action progress. The main transaction is not available anymore.
				//
				// The processing state of a request which has a multiframe response may be set with "Sent" only at the end of
				// the response, else the ability to send the request again later would be broken (for example, to resume a file
				// download).
				//
				// Setting the "Sent" processing state here instead of after the "ActionClient.UpdateActionRequest" method
				// (which is called for action requests whose "NeedsUpdateBeforeSending" is true) returns false, may lead to
				// subtle errors (because the service may prematurely and wrongly signal the end of the multiframe response). In
				// such a case, it's necessary to fix the service.
				if( isFirstFrame || !continueReceiving )
				{
					ErrorActionResponse errorResponse = actionResponse as ErrorActionResponse;
					ChannelContext.RecipientAccessStatesManager.UpdateProgress( progress, errorResponse == null, true
						, errorResponse, !continueReceiving, actionResponse.SecondaryProcessingState );
					ActionClientUtils.AfterUpdateProcessingState( progress, errorResponse, clientContext.RemoteUpn
						, ChannelContext.StorehouseProvider );
				}

				// The host UI must not be updated too often because that would cause significant performance degradation.
				//
				// Perform a full UI update for the first frame in order to ensure that the UI is up to date when the
				// multiframe response is resumed (after an interruption). Perform a full UI update for the last frame.
				bool mustFullyUpdateUi = isFirstFrame || !continueReceiving;
				if( mustFullyUpdateUi || chrone.PassedSeconds( ActionProcessorConsts.MultiframeActionResponseUpdateUiTimespan ) )
				{
					// Notify the action processor host to update the UI. Do this outside the main transaction because
					// another one may be created and may try to load the sheafs which are locked by the main transaction.
					ActionClientHostUpdateUiContext updateUiContext = new ActionClientHostUpdateUiContext(
						ChannelContext.ProcessorConfiguration, progress, actionRequest, clientContext.UpdateRequestContext
						, actionResponse, null, mustFullyUpdateUi );
					clientContext.ActionClientHost.UpdateUi( updateUiContext );
				}

				isFirstFrame = false;
			}
		}

		/// <summary>
		/// Updates the specified action request.
		/// </summary>
		/// <param name="actionRequest">Action request to update and process. This may be modified in here if the request's
		/// <see cref="Vela.ActionProcessor.ActionRequest.NeedsUpdateBeforeSending"/> is true; it's also
		/// automatically signed afterwards.</param>
		/// <param name="client">The action client.</param>
		/// <param name="clientContext">The action client context.</param>
		/// <param name="updateRequestMethodInfo">The information of the method which can update the action request. This may
		/// be null if the request's <see cref="Vela.ActionProcessor.ActionRequest.NeedsUpdateBeforeSending"/>
		/// is false.</param>
		/// <returns>Returns the true if the action request must be sent to the service, and false if it must never be sent.
		/// </returns>
		private bool UpdateActionRequest( ActionRequest actionRequest, IActionClient client, ActionClientContext clientContext
			, UpdateRequestMethodInfo updateRequestMethodInfo )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( client, "client" );
			ValidationUtils.ObjectHasContent( clientContext, "clientContext" );

			if( !actionRequest.NeedsUpdateBeforeSending )
				return true;

			ValidationUtils.ObjectHasContent( updateRequestMethodInfo, "updateRequestMethodInfo" );

			ActionClientProgress progress = clientContext.ActionProgress;

			// Create a transaction.
			IStorehouseTransaction transaction = ChannelContext.StorehouseProvider.CreateTransaction();

			try
			{
				if( client.UpdateActionRequest( actionRequest, clientContext, transaction, updateRequestMethodInfo ) )
				{
					// Commit the transaction.
					transaction.Commit();

					return true;
				}

				// If the action request is not to be sent to the service, update the state to indicate the successful
				// termination of the request.
				ChannelContext.RecipientAccessStatesManager.UpdateProgress( progress, true, false, null, true
					, progress.LastProcessingAttempt.SecondaryProcessingState );
				ActionClientUtils.AfterUpdateProcessingState( progress, null, clientContext.RemoteUpn, transaction );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			// Notify the action processor host to update the UI. Do this outside the main transaction because another
			// one may be created and may try to load the sheafs which are locked by the main transaction.
			ActionClientHostUpdateUiContext updateUiContext = new ActionClientHostUpdateUiContext(
				ChannelContext.ProcessorConfiguration, progress, actionRequest, clientContext.UpdateRequestContext, null
				, null, true );
			clientContext.ActionClientHost.UpdateUi( updateUiContext );

			return false;
		}

		/// <summary>
		/// Updates and processes the specified action request. The action request may be updated and processed multiple times.
		/// </summary>
		/// <param name="actionRequest">Action request to update and process. This may be modified in here if the request's
		/// <see cref="Vela.ActionProcessor.ActionRequest.NeedsUpdateBeforeSending"/> is true; it's also
		/// automatically signed afterwards.</param>
		/// <param name="remoteUpn">Remote UPN.</param>
		/// <param name="remoteIpAddress">Remote IP address.</param>
		/// <param name="tally">Tally. This may be null if the action request is derived from
		/// <see cref="AnonymousActionRequest"/>.</param>
		/// <param name="serializer">Serializer using for encryption.</param>
		/// <param name="client">The action client.</param>
		/// <param name="clientContext">The action client context.</param>
		/// <param name="updateRequestMethodInfo">The information of the method which can update the action request. This may
		/// be null if the request's <see cref="Vela.ActionProcessor.ActionRequest.NeedsUpdateBeforeSending"/>
		/// is false.</param>
		/// <param name="processResponseMethodInfo">The information of the method which can process the action response.</param>
		/// <param name="processLastFrameResponseMethodInfo">The information of the method which can process the last frame
		/// action response. This may be null.</param>
		private void UpdateAndProcessRequest( ActionRequest actionRequest, Upn remoteUpn, byte[] remoteIpAddress
			, Tally tally, ISerializer serializer, IActionClient client, ActionClientContext clientContext
			, UpdateRequestMethodInfo updateRequestMethodInfo, ProcessResponseMethodInfo processResponseMethodInfo
			, ProcessLastFrameResponseMethodInfo processLastFrameResponseMethodInfo )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( remoteUpn, "remoteUpn" );
			ValidationUtils.ArrayHasContent( remoteIpAddress, "remoteIpAddress" );
			ValidationUtils.ObjectHasContent( serializer, "serializer" );
			ValidationUtils.ObjectHasContent( client, "client" );
			ValidationUtils.ObjectHasContent( clientContext, "clientContext" );
			ValidationUtils.ObjectHasContent( processResponseMethodInfo, "processResponseMethodInfo" );

			Socket socket = null;

			try
			{
				if( !UpdateActionRequest( actionRequest, client, clientContext, updateRequestMethodInfo ) )
					return;

				#region Encrypt the action request

				ICoreAlgorithmPool coreap = ChannelContext.CoreAlgorithmPool;

				EncryptedEnvelope requestEnvelope = (tally != null) ? (EncryptedEnvelope)coreap.GetAsymmEncryptedEnvelope()
					: (EncryptedEnvelope)coreap.GetUpnEncryptedEnvelope( remoteUpn.Fingerprint );

				requestEnvelope.EncryptedDocumentBlockSize = coreap.GetActionRequestEnvelopeEncryptedDocumentBlockSize();
				requestEnvelope.EncryptableDocument = actionRequest;

				object password = (tally != null) ? (object)tally.ProcessingIdentity : (object)remoteUpn.SymmEncryptionKey;
				byte[] encryptedRequest = requestEnvelope.SaveToBuffer( null, password, serializer );

				#endregion Encrypt the action request

				ActionProcessorConfiguration processorConfiguration = ChannelContext.ProcessorConfiguration;
				SteganographicAlg steganographicAlg = remoteUpn.UpnAlg.SteganographicAlg;
				ICommunicationFrameAlg communicationFrameAlg = remoteUpn.UpnAlg.CommunicationFrameAlg;

				// Obfuscate the action request.
				byte[] obfuscatedRequest = steganographicAlg.Obfuscate( encryptedRequest, remoteUpn.SymmEncryptionKey
					, remoteUpn.Fingerprint.Value );

				CommunicationFrame communicationFrame = new CommunicationFrame( communicationFrameAlg
					, processorConfiguration.CommunicationTimeoutPerBlob, processorConfiguration.CommunicationFrameSizeMax );

				UInt16 overridingProcessingPort = (tally != null) ? tally.AlternativePeerAddress.ProcessingPort : (UInt16)0;
				UInt16 remotePort = (overridingProcessingPort > 0) ? overridingProcessingPort : remoteUpn.Port;

				// A single connection is used to send the action request and receive all the frames of the action response.
				socket = communicationFrame.CreateSocketForSendMessage( remoteIpAddress, remotePort );

				// Set the socket buffer size for receiving the action response (frames).
				if( actionRequest.ActionResponseRequiresLargeCommunicationCacheSize
					&& (processorConfiguration.LargeCommunicationCacheSize > 0) )
				{
					socket.ReceiveBufferSize = processorConfiguration.LargeCommunicationCacheSize;
				}

				if( actionRequest as IActionRequestWithMultiframeResponse == null )
				{
					SendRequestAndReceiveUniframeResponse( actionRequest, obfuscatedRequest, remoteUpn, remoteIpAddress
						, communicationFrame, socket, tally, processResponseMethodInfo, client, clientContext );
				}
				else
				{
					SendRequestAndReceiveMultiframeResponse( actionRequest, obfuscatedRequest, remoteUpn, remoteIpAddress
						, communicationFrame, socket, tally, processResponseMethodInfo, processLastFrameResponseMethodInfo
						, client, clientContext );
				}

				LogManager.DebugAddFormat( "The action response associated to the action request with the locator \"{0}\" and"
						+ " the type \"{1}\", received from the service with the UPN \"{2}\", has been successfully processed."
					, actionRequest.InstanceLocator, actionRequest.GetType().FullName, remoteUpn );
			}
			catch( Exception ex )
			{
				string message = string.Format( "Failed to process the action response associated to the action request with the"
						+ " locator \"{0}\" and the type \"{1}\", received from the service with the UPN \"{2}\"."
					, actionRequest.InstanceLocator, actionRequest.GetType().FullName, remoteUpn );

				throw new ValidationException( message, ExceptionUtils.PreserveStackTrace( ex ) );
			}
			finally
			{
				NetworkingUtils.CloseSocket( socket );
			}
		}

		/// <summary>
		/// Processes the specified action request.
		/// </summary>
		/// <param name="actionRequest">Action request to process. This may be modified in here if the request's
		/// <see cref="Vela.ActionProcessor.ActionRequest.NeedsUpdateBeforeSending"/> is true; it's also
		/// automatically signed afterwards.</param>
		/// <param name="actionProgress">The progress of the action request to process.</param>
		/// <param name="remoteUpn">Remote UPN.</param>
		/// <param name="remoteIpAddress">Remote IP address.</param>
		/// <param name="tally">Tally. This may be null if the action request is derived from
		/// <see cref="AnonymousActionRequest"/>.</param>
		/// <param name="serializer">Serializer using for encryption.</param>
		private void ProcessRequest( ActionRequest actionRequest, ActionClientProgress actionProgress, Upn remoteUpn
			, byte[] remoteIpAddress, Tally tally, ISerializer serializer )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( actionProgress, "actionProgress" );
			ValidationUtils.ObjectHasContent( remoteUpn, "remoteUpn" );
			ValidationUtils.ArrayHasContent( remoteIpAddress, "remoteIpAddress" );
			ValidationUtils.ObjectHasContent( serializer, "serializer" );

			if( actionRequest as ActionRequestWithInitiator != null )
			{
				ValidationUtils.ObjectHasContent( tally, "tally" );
			}
			else if( actionRequest as AnonymousActionRequest == null )
			{
				throw new ValidationException( string.Format( "An action client may process only action requests derived"
						+ " from \"{0}\" or from \"{1}\", but it's type is \"{2}\"."
					, typeof( ActionRequestWithInitiator ).FullName, typeof( AnonymousActionRequest ).FullName
					, actionRequest.GetType().FullName ) );
			}

			LogManager.DebugAddFormat( "Attempting to send the action request with the locator \"{0}\" and the type"
					+ " \"{1}\", to the service with the UPN \"{2}\", to the IP address \"{3}\"."
				, actionRequest.InstanceLocator, actionRequest.GetType().FullName, remoteUpn
				, NetworkingUtils.GetTextualIpAddress( remoteIpAddress ) );

			#region Instantiate the client

			ActionClientContext clientContext = new ActionClientContext( ChannelContext.StorehouseProvider
				, ChannelContext.ProcessorConfiguration, ChannelContext.StorehouseProvider.Serializer
				, (ActionRequest)actionRequest, actionProgress, tally, remoteUpn, remoteIpAddress
				, actionRequest.ResponseEncryptionKey, ChannelContext.ActionClientHost );

			UpdateRequestMethodInfo updateRequestMethodInfo;
			ProcessResponseMethodInfo processResponseMethodInfo;
			ProcessLastFrameResponseMethodInfo processLastFrameResponseMethodInfo;
			IActionClient client = InstantiateClient( actionRequest.GetType(), actionRequest.ActionResponseType
				, out updateRequestMethodInfo, out processResponseMethodInfo, out processLastFrameResponseMethodInfo );

			#endregion Instantiate the client

			UpdateAndProcessRequest( actionRequest, remoteUpn, remoteIpAddress, tally, serializer, client, clientContext
				, updateRequestMethodInfo, processResponseMethodInfo, processLastFrameResponseMethodInfo );

			// The transaction must have been closed by now.
			if( (clientContext.Transaction != null) && clientContext.Transaction.IsOpen )
			{
				throw new ValidationException( "The action request was processed, but the transaction is still open. Storehouse"
					+ " timeouts may occur." );
			}
		}

		/// <summary>
		/// Resolves the UPN of the recipient of the specified action request if it's necessary, and processes the specified
		/// action request.
		/// </summary>
		/// <param name="actionRequest">Action request to process. This may be modified in here if the request's
		/// <see cref="Vela.ActionProcessor.ActionRequest.NeedsUpdateBeforeSending"/> is true; it's also
		/// automatically signed afterwards.</param>
		/// <param name="actionProgress">The progress of the action request to process.</param>
		/// <param name="tally">Tally.</param>
		private void ResolveUpnAndProcessRequestWithInitiator( ActionRequestWithInitiator actionRequest
			, ActionClientProgress actionProgress, Tally tally )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( actionProgress, "actionProgress" );
			ValidationUtils.ObjectHasContent( tally, "tally" );

			Upn remoteUpn = actionRequest.RecipientUpn;

			// Don't process expired action requests.
			KernelUtils.HasExpired( actionRequest, tally.ActionRequestValidity, true );

			#region Resolve the recipient UPN

			AlternativePeerAddress alternativePeerAddress;
			byte[] remoteIpAddress = GetExistingRemoteIpAddress( remoteUpn.Fingerprint, tally, actionProgress
				, out alternativePeerAddress );

			UInt16 overridingProcessingPort = (alternativePeerAddress != null) ? alternativePeerAddress.ProcessingPort
				: (UInt16)0;

			bool remoteUpnWasResolved = false;

			// Resolve the IP address of the service.
			if( ArrayUtils.NoContent( remoteIpAddress ) )
			{
				// The alternative host name has priority.
				if( (alternativePeerAddress != null) && alternativePeerAddress.HasHostName() )
				{
					remoteIpAddress = ResolveRemoteIpAddressWithDns( remoteUpn, actionRequest
						, alternativePeerAddress.GetHostName(), overridingProcessingPort, null );
				}
				else
				{
					remoteIpAddress = ResolveRemoteIpAddressWithPnrp( remoteUpn, tally.EntityName, actionRequest
						, overridingProcessingPort );
				}

				// At this point, the remote UPN has been resolved; else, an exception was thrown.
				remoteUpnWasResolved = true;
			}

			// At this point, there is a remote IP address either because it was previously saved (and loaded) or the remote UPN
			// was resolved; else, an exception was thrown.

			#endregion Resolve the recipient UPN

			bool retryToProcessActionRequest = false;

			try
			{
				// Process the action request.
				ProcessRequest( actionRequest, actionProgress, remoteUpn, remoteIpAddress, tally
					, ChannelContext.StorehouseProvider.Serializer );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsAbort( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				// Don't resolve the remote UPN again because an algorithm / document has expired.
				if( ExceptionUtils.HasException<HasExpiredException>( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				// Don't resolve the remote UPN again because a "CrossServiceException" exception was received (in an error
				// response), so the remote IP address is correct.
				if( ExceptionUtils.HasException<CrossServiceException>( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				// Don't resolve the remote UPN again because it was just resolved.
				if( remoteUpnWasResolved )
					throw ExceptionUtils.PreserveStackTrace( ex );

				// Don't resolve the remote UPN again because there is an alternative peer address.
				if( (alternativePeerAddress != null) && alternativePeerAddress.HasIpAddress() )
				{
					string message = string.Format( "The action client could not successfully send the action request with the"
							+ " locator \"{0}\" to the action service with the UPN \"{1}\". The alternative peer resolution is"
							+ " active for the remote peer, so ensure that the account with the entity name \"{2}\" (= the"
							+ " remote peer) has a valid alternative peer address. You could also try to remove the alternative"
							+ " peer address from the account in order to allow the PNRP to resolve the UPN."
						, actionRequest.InstanceLocator, remoteUpn, tally.EntityName );

					throw new ValidationException( message, ex );
				}

				if( !ExceptionUtils.HasException<SocketException>( ex ) )
					LogManager.Add( ex );

				// The communication was interrupted because, presumably, the remote IP address doesn't belong to the remote
				// peer, so try to resolve the remote UPN again. There is no guarantee that the exception is a "SocketException",
				// so do this for all remaining exceptions.
				retryToProcessActionRequest = true;
			}

			if( retryToProcessActionRequest )
			{
				// The first processing attempt has failed because, presumably, the remote IP address doesn't belong to the
				// remote peer. In such a case we want to process the action request again, right now, in case the remote peer is
				// online and the remote UPN can be resolved immediately. This way, the processing state will not be changed to
				// "Retry" every time the remote IP address changes, and the user will not have to wait for the next scheduled
				// processing attempt.
				//
				// If this would not be done, the existing remote IP address would have to be removed from the ephemerid so
				// that it could be resolved at the next scheduled processing attempt. However, the remote IP address might
				// actually be correct, but the remote peer might be offline, so the remote IP address should be preserved.

				// Try to resolve the remote UPN. The alternative host name has priority.
				if( (alternativePeerAddress != null) && alternativePeerAddress.HasHostName() )
				{
					remoteIpAddress = ResolveRemoteIpAddressWithDns( remoteUpn, actionRequest
						, alternativePeerAddress.GetHostName(), overridingProcessingPort, null );
				}
				else
				{
					remoteIpAddress = ResolveRemoteIpAddressWithPnrp( remoteUpn, tally.EntityName, actionRequest
						, overridingProcessingPort );
				}

				// At this point, the remote UPN has been resolved; else, an exception was thrown.

				// Process the action request.
				ProcessRequest( actionRequest, actionProgress, remoteUpn, remoteIpAddress, tally
					, ChannelContext.StorehouseProvider.Serializer );
			}
		}

		/// <summary>
		/// Resolves the UPN of the recipient of the specified action request, and processes the specified action request.
		/// </summary>
		/// <param name="actionRequest">Action request to process. This may be modified in here if the request's
		/// <see cref="Vela.ActionProcessor.ActionRequest.NeedsUpdateBeforeSending"/> is true; it's also
		/// automatically signed afterwards.</param>
		/// <param name="actionProgress">The progress of the action request to process.</param>
		private void ResolveUpnAndProcessAnonymousRequest( AnonymousActionRequest actionRequest
			, ActionClientProgress actionProgress )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( actionProgress, "actionProgress" );

			// Don't process expired action requests.
			KernelUtils.HasExpired( actionRequest, PeerDocumentsConsts.ActionRequestValidity, true );

			Upn remoteUpn = actionRequest.RecipientUpn;

			#region Resolve the recipient UPN

			AlternativePeerAddress alternativePeerAddress;
			byte[] remoteIpAddress = GetExistingRemoteIpAddress( null, null, actionProgress, out alternativePeerAddress );

			// Resolve the IP address of the service.
			if( ArrayUtils.NoContent( remoteIpAddress ) )
			{
				UInt16 overridingProcessingPort = (alternativePeerAddress != null) ? alternativePeerAddress.ProcessingPort
					: (UInt16)0;

				// The alternative host name has priority.
				if( (alternativePeerAddress != null) && alternativePeerAddress.HasHostName() )
				{
					remoteIpAddress = ResolveRemoteIpAddressWithDns( remoteUpn, actionRequest
						, alternativePeerAddress.GetHostName(), overridingProcessingPort, null );
				}
				else
				{
					remoteIpAddress = ResolveRemoteIpAddressWithPnrp( remoteUpn, null, actionRequest, overridingProcessingPort );
				}
			}

			#endregion Resolve the recipient UPN

			ProcessRequest( actionRequest, actionProgress, remoteUpn, remoteIpAddress, null
				, ChannelContext.StorehouseProvider.Serializer );
		}

		/// <summary>
		/// Resolves the UPN of the recipient of the action request with the current request locator
		/// (<see cref="RequestLocator"/>), and processes the action request.
		/// </summary>
		private void ResolveUpnAndProcessRequest()
		{
			ActionRequest actionRequest = null;
			ActionClientProgress actionProgress = null;
			Tally tally = null;

			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				// Create a transaction.
				IStorehouseTransaction transaction = ChannelContext.StorehouseProvider.CreateTransaction();

				try
				{
					IDeserializer deserializer;

					#region Handle the action progress

					// Load the action progress.
					SpeckleFilter requestLocatorFilter = new SpeckleFilter(
						ActionClientProgress.ReferredActionRequestLocatorFilterIdentifier, RequestLocator );
					actionProgress = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived(
						typeof( ActionClientProgress ) ), new SpeckleFilters( requestLocatorFilter )
						, StorehouseRealms.ActionProgresses, SheafLocks.Shared, out deserializer ) as ActionClientProgress;
					ValidationUtils.ObjectHasContent( actionProgress, "progress" );

					// Don't process expired documents.
					KernelUtils.HasExpired( actionProgress );

					// Check if the action request may be processed.
					if( (actionProgress.ProcessingState != ActionProcessingStates.Pending)
						&& (actionProgress.ProcessingState != ActionProcessingStates.Retry) )
					{
						// Commit the transaction.
						transaction.Commit();

						return;
					}

					// Check if the scheduler allows the action request to be processed.
					if( !PendingRequest.IgnoreScheduler
						&& !ChannelContext.RecipientAccessStatesManager.ScheduleAllowsProcessing( actionProgress ) )
					{
						// Commit the transaction.
						transaction.Commit();

						LogManager.DebugAddFormat( "The retry schedules don't allow the action client to process the action"
								+ " request with the type \"{0}\" and the locator \"{1}\" associated to the document with the"
								+ " locator \"{2}\". A processing attempt will be made later."
							, actionProgress.ReferredActionRequestTypeName, actionProgress.ReferredActionRequestLocator
							, actionProgress.ReferredSheafLocator );

						return;
					}

					#endregion Handle the action progress

					// Load the action request.
					actionRequest = transaction.LoadSheaf( RequestLocator, StorehouseRealms.ActionRequestsToSend
						, typeof( ActionRequest ), SheafLocks.Shared, out deserializer ) as ActionRequest;
					ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );

					// Load the tally.
					if( actionRequest as ActionRequestWithInitiator != null )
					{
						SpeckleFilter remoteUpnFilter = new SpeckleFilter( Tally.RemoteUpnFingerprintFilterIdentifier
							, actionRequest.RecipientUpnFingerprint );
						tally = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Tally ) )
							, new SpeckleFilters( remoteUpnFilter ), StorehouseRealms.ProcessingIdentityContainers
							, SheafLocks.Shared, out deserializer ) as Tally;
						ValidationUtils.ObjectHasContent( tally, "tally" );
					}

					// Commit the transaction.
					transaction.Commit();
				}
				catch( Exception ex )
				{
					// Rolls back the transaction.
					transaction.Rollback();

					throw ExceptionUtils.PreserveStackTrace( ex );
				}

				// Do the actual processing outside the transaction because the UPN resolution could take a long time, and the
				// progress should not be locked for update during all this time.
				if( actionRequest as ActionRequestWithInitiator != null )
				{
					ActionRequestWithInitiator requestWithInitiator = (ActionRequestWithInitiator)actionRequest;
					ResolveUpnAndProcessRequestWithInitiator( requestWithInitiator, actionProgress, tally );
				}
				else if( actionRequest as AnonymousActionRequest != null )
				{
					AnonymousActionRequest anonymousActionRequest = (AnonymousActionRequest)actionRequest;
					ResolveUpnAndProcessAnonymousRequest( anonymousActionRequest, actionProgress );
				}
				else
				{
					throw new ValidationException( string.Format( "An action client may process only action requests derived"
							+ " from \"{0}\" or from \"{1}\", but it's type is \"{2}\"."
						, typeof( ActionRequestWithInitiator ).FullName, typeof( AnonymousActionRequest ).FullName
						, actionRequest.GetType().FullName ) );
				}
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsAbort( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				if( actionProgress != null )
				{
					if( actionProgress.InstanceLocator != null )
						UpdateActionProgressWhenException( actionProgress.InstanceLocator, actionRequest, ex );

					if( (actionRequest != null)
						&& ActionProcessorStorehouseUtils.UpdateTallyIfRevokedServiceVault( ChannelContext.StorehouseProvider
							, actionRequest, ex ) )
					{
						// Notify the action processor host to update the UI. Do this outside the main transaction because
						// another one may be created and may try to load the sheafs which are locked by the main transaction.
						ChannelContext.ActionClientHost.UpdateUiRevokedServiceVault( actionRequest.RecipientUpn );
					}
				}

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		/// <summary>
		/// Handler for the thread (<see cref="ChannelThread"/>) which processes action requests and responses.
		/// </summary>
		private void ChannelThreadHandler()
		{
			try
			{
				LogManager.DebugAddFormat( "An action client channel was started for the action request locator \"{0}\"."
					, RequestLocator );

				ResolveUpnAndProcessRequest();

				LogManager.DebugAddFormat( "The action client channel for the action request locator \"{0}\" has finished"
					+ " processing.", RequestLocator );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				if( ExceptionUtils.IsAbort( ex ) )
				{
					LogManager.DebugAddFormat( "The action client channel for the action request locator \"{0}\" has been"
						+ " aborted.", RequestLocator );
				}
				else
				{
					string message = string.Format( "The action client channel has failed to process the action request with the"
						+ " locator \"{0}\".", RequestLocator );
					LogManager.Add( message, ex, LogRecordModes.Error, true );
				}
			}
			finally
			{
				try
				{
					ChannelFinished( RequestLocator );
				}
				catch( Exception ex )
				{
					if( ExceptionUtils.IsCritical( ex ) )
						throw ExceptionUtils.PreserveStackTrace( ex );

					LogManager.Add( ex );
				}
			}
		}

		/// <summary>
		/// Starts processing the current action request on a new thread.
		/// </summary>
		private void Start()
		{
			ChannelThread = new Thread( ChannelThreadHandler );

			// The thread must terminate in order for the application to terminate.
			ChannelThread.IsBackground = false;

			ChannelThread.Start();
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Stops processing the current action request as soon as possible.
		/// </summary>
		public void Stop()
		{
			StopChannelThread = true;
		}

		#endregion Public methods
	}
}
