////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Threading;
using Vela.ActionProcessor;
using Vela.Cloning;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Kernel;
using Vela.Storehouse;

namespace Vela.ActionClient
{
	/// <summary>
	/// Action client manager.
	/// </summary>
	/// <remarks>
	/// <para>Action requests are sent to the remote peers on background threads, and the actions responses are received and
	/// processed on those threads.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Processing queue</b></para>
	/// 
	/// <para>The number of (queued) action requests which may be processed in parallel is limited.</para>
	/// 
	/// <para>Action requests can be queued manually with the <see cref="Add( Locator, bool )"/> or
	/// <see cref="Add( Locators, bool )"/> methods. Requests are also automatically queued from the storehouse, every
	/// <see cref="ActionClientManager.ClientStorehouseLoadTimespan"/> seconds.</para>
	/// 
	/// <para>The size of the queue (with action request locators from which action client channels are created) is not limited
	/// when manually queueing is done. This allows the client manager to process the entire queue as soon as possible, that is,
	/// without waiting for action requests to be loaded from the storehouse.</para>
	/// 
	/// <para>It's technically possible to continuously process action requests from the storehouse until there isn't any left.
	/// However, this can only be done if the progress state of each action request was handled; if this were not done, the
	/// client manager would overload the computer by continuously trying to process action requests which have failed (so
	/// far).</para>
	/// </remarks>
	public class ActionClientManager : IActionClientManager
	{
		#region Public constants

		/// <summary>
		/// The time (in milliseconds) to sleep between consecutive attempts to check (in memory) for new action requests.
		/// </summary>
		public const int ProcessingSleep = 10;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionClientManager()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction. This is cloned in
		/// here, so it doesn't have to be locked.</param>
		/// <param name="actionClientTypeNames">The type names of the action clients which can process action responses.</param>
		/// <param name="cryptoAlgorithmPool">Pool of cryptographic algorithms. Its content must be thread safe.</param>
		/// <param name="coreAlgorithmPool">Pool of core algorithms. Its content must be thread safe.</param>
		/// <param name="processorConfiguration">The action processor configuration.</param>
		/// <param name="actionClientHost">Host used for updating the user interface of an action client application after an
		/// action request and response are processed. Its content must be thread safe.</param>
		public ActionClientManager( IStorehouseProvider storehouseProvider, Strings actionClientTypeNames
			, ICryptoAlgorithmPool cryptoAlgorithmPool, ICoreAlgorithmPool coreAlgorithmPool
			, ActionProcessorConfiguration processorConfiguration, IActionClientHost actionClientHost )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( actionClientTypeNames, "actionClientTypeNames" );

			// Cache the method information for the action clients, for updating action requests.
			UpdateRequestMethodInfos updateRequestMethodInfos = new UpdateRequestMethodInfos();
			updateRequestMethodInfos.Add( actionClientTypeNames );

			// Cache the method information for the action clients, for processing action responses.
			ProcessResponseMethodInfos processResponseMethodInfos = new ProcessResponseMethodInfos();
			processResponseMethodInfos.Add( actionClientTypeNames );

			// Cache the method information for the action clients, for processing last frame action responses.
			ProcessLastFrameResponseMethodInfos processLastFrameResponseMethodInfos = new ProcessLastFrameResponseMethodInfos();
			processLastFrameResponseMethodInfos.Add( actionClientTypeNames );

			RecipientAccessStatesManager recipientAccessStatesManager = new RecipientAccessStatesManager(
				processorConfiguration.ActionClient.RecipientAccessStatesManager );

			UpnResolutionLocker upnResolutionLocker = new UpnResolutionLocker();

			ActionClientChannelContext channelContext = new ActionClientChannelContext( storehouseProvider
				, processorConfiguration, updateRequestMethodInfos, processResponseMethodInfos
				, processLastFrameResponseMethodInfos, cryptoAlgorithmPool, coreAlgorithmPool, recipientAccessStatesManager
				, upnResolutionLocker, actionClientHost );

			ChannelContext = Cloner.Deep( channelContext );

			Start();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The synchronization object for the <see cref="ChannelContext"/> property.
		/// </summary>
		private readonly object ChannelContextLock = new object();

		/// <summary>
		/// Action client channel context.
		/// </summary>
		private ActionClientChannelContext ChannelContext;

		/// <summary>
		/// The number of physical and virtual processor cores of the system.
		/// </summary>
		private int ProcessorCoresCount = SystemUtils.GetProcessorCoresCount();

		/// <summary>
		/// The synchronization object for the <see cref="ClientChannels"/> property.
		/// </summary>
		private readonly object ClientChannelsLock = new object();

		/// <summary>
		/// The action client channels. The backing field is initialized by default. This may be empty, but not null.
		/// </summary>
		private ActionClientChannels ClientChannels = new ActionClientChannels();

		/// <summary>
		/// The synchronization object for the <see cref="PendingRequests"/> property.
		/// </summary>
		private readonly object PendingRequestsLock = new object();

		/// <summary>
		/// The requests which should be processed as soon as possible (the maximum delay is <see cref="ProcessingSleep"/>). The
		/// backing field is initialized by default. This may be empty, but not null.
		/// </summary>
		private PendingRequests PendingRequests = new PendingRequests();

		/// <summary>
		/// The thread where the client processing happens.
		/// </summary>
		private Thread ProcessingThread;

		/// <summary>
		/// The synchronization object for the <see cref="StopProcessingThread"/> property.
		/// </summary>
		private readonly object StopProcessingThreadLock = new object();

		private bool _StopProcessingThread = false;
		/// <summary>
		/// If this is true then the processing thread must stop as soon as possible.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// </remarks>
		private bool StopProcessingThread
		{
			get
			{
				lock( StopProcessingThreadLock )
				{
					return _StopProcessingThread;
				}
			}
			set
			{
				lock( StopProcessingThreadLock )
				{
					_StopProcessingThread = value;
				}
			}
		}

		/// <summary>
		/// The chronometer used to the automatically call the <see cref="AddFromStorehouse"/> method in the processing thread.
		/// </summary>
		private Chronometer AddFromStorehouseChrono = new Chronometer();

		/// <summary>
		/// The default value must be 0 or a negative value so that a random value can be generated the first time this property
		/// is read.
		/// </summary>
		private double _ClientStorehouseLoadTimespan = 0;
		/// <summary>
		/// The timespan (in seconds) after which an action client automatically loads action requests (from the storehouse) to
		/// process in the processing thread.
		/// </summary>
		/// <remarks>
		/// <para>This is random in order to protect from statistical analysis. Since the action requests are loaded at random
		/// moments, they are also processed at random moments, and their action responses are sent at random moments.</para>
		/// </remarks>
		private double ClientStorehouseLoadTimespan
		{
			get
			{
				if( _ClientStorehouseLoadTimespan <= 0 )
				{
					_ClientStorehouseLoadTimespan = ActionProcessorConsts.ClientStorehouseLoadTimespanMax
						* EntropyPool.GetReal();
				}

				return _ClientStorehouseLoadTimespan;
			}
			set { _ClientStorehouseLoadTimespan = value; }
		}

		/// <summary>
		/// The offset used for paging when adding action requests from the storehouse.
		/// </summary>
		private int AddFromStorehousePagingOffset;

		#endregion Private fields

		#region Public non-serializable properties

		/// <summary>
		/// The action processor configuration. READ-ONLY.
		/// </summary>
		public ActionProcessorConfiguration ProcessorConfiguration
		{
			get { return ChannelContext.ProcessorConfiguration; }
		}

		/// <summary>
		/// Manager for the recipient access states. Its content is thread safe. READ-ONLY.
		/// </summary>
		public IRecipientAccessStatesManager RecipientAccessStatesManager
		{
			get { return ChannelContext.RecipientAccessStatesManager; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Event raised when an action client channel finishes processing an action request.
		/// </summary>
		/// <param name="requestLocator">The instance locator of the action request which was processed.</param>
		/// <remarks>
		/// <para>The <see cref="ClientChannels"/> property is thread safe within this method.</para>
		/// 
		/// <para>This is called on the client channel thread.</para>
		/// </remarks>
		private void ChannelFinished( Locator requestLocator )
		{
			ValidationUtils.ObjectHasContent( requestLocator, "requestLocator" );

			lock( ClientChannelsLock )
			{
				ClientChannels.Remove( requestLocator );
			}
		}

		/// <summary>
		/// Creates a client channel for the action request with the specified locator.
		/// </summary>
		/// <param name="pendingRequest">The pending request for which to create a client channel.</param>
		/// <remarks>
		/// <para>If there already is a client channel for the specified locator, nothing happens.</para>
		/// 
		/// <para>The <see cref="ChannelContext"/> property is thread safe within this method.</para>
		/// 
		/// <para>The <see cref="ClientChannels"/> property is thread safe within this method.</para>
		/// </remarks>
		private void CreateChannel( PendingRequest pendingRequest )
		{
			ValidationUtils.ObjectHasContent( pendingRequest, "pendingRequest" );

			// Create a client channel to process the action request.
			lock( ChannelContextLock )
			{
				lock( ClientChannelsLock )
				{
					// Check if the locator is already queued for processing.
					if( ClientChannels.ContainsKey( pendingRequest.RequestLocator ) )
						return;

					// Create the client channel.
					ActionClientChannel clientChannel = new ActionClientChannel( ChannelContext, pendingRequest
						, ChannelFinished );

					ClientChannels.Add( clientChannel );

					LogManager.DebugAddFormat( "An action client channel was added for the action request locator \"{0}\"."
						, pendingRequest.RequestLocator );
				}
			}
		}

		/// <summary>
		/// Creates the client channels for the action request locators queued in <see cref="PendingRequests"/>.
		/// </summary>
		/// <remarks>
		/// <para>If the maximum number of client channels which may be running in parallel has been reached, nothing happens.
		/// </para>
		/// 
		/// <para>If there already is a client channel for a locator, the locator is ignored.</para>
		/// 
		/// <para>The <see cref="ChannelContext"/> property is thread safe within this method.</para>
		/// 
		/// <para>The <see cref="ClientChannels"/> property is thread safe within this method.</para>
		/// 
		/// <para>The <see cref="PendingRequests"/> property is thread safe within this method.</para>
		/// </remarks>
		private void CreateChannels()
		{
			lock( ChannelContextLock )
			{
				lock( ClientChannelsLock )
				{
					lock( PendingRequestsLock )
					{
						if( PendingRequests.Count <= 0 )
							return;

						// Don't create a new client channel if too many are already running.
						int requestsMax = (ProcessorConfiguration.ParallelRequestsMax > 0)
							? ProcessorConfiguration.ParallelRequestsMax : ProcessorCoresCount;

						Locators usedLocators = new Locators();

						// Create client channels to process the action requests with the queued locators.
						foreach( PendingRequest pendingRequest in PendingRequests )
						{
							if( ClientChannels.Count < requestsMax )
							{
								CreateChannel( pendingRequest );
								usedLocators.Add( pendingRequest.RequestLocator );
							}
						}

						// Remove from the queue the locators for which client channels were created.
						foreach( Locator usedLocator in usedLocators )
						{
							int index = PendingRequests.FindIndexByRequestLocator( usedLocator );

							if( index >= 0 )
								PendingRequests.RemoveAt( index );
						}
					}
				}
			}
		}

		/// <summary>
		/// Adds a batch of action requests from the storehouse, to the list of locators of action requests to process.
		/// </summary>
		/// <returns>Returns true if any action request has been added from the storehouse.</returns>
		/// <remarks>
		/// <para>If the maximum number of client channels which may be running in parallel has been reached, nothing happens.
		/// </para>
		/// 
		/// <para>The <see cref="ChannelContext"/> property is thread safe within this method.</para>
		/// 
		/// <para>The <see cref="ClientChannels"/> property is thread safe within this method.</para>
		/// 
		/// <para>The <see cref="PendingRequests"/> property is thread safe within this method.</para>
		/// </remarks>
		private bool AddFromStorehouse()
		{
			lock( ChannelContextLock )
			{
				lock( ClientChannelsLock )
				{
					lock( PendingRequestsLock )
					{
						// Establish batch size.
						int requestsMax = (ProcessorConfiguration.ParallelRequestsMax > 0)
							? ProcessorConfiguration.ParallelRequestsMax : ProcessorCoresCount;
						int batchSize = requestsMax - ClientChannels.Count;
						if( batchSize <= 0 )
						{
							LogManager.DebugAddFormat( "The action client manager is already processing the maximum number"
								+ " ({0}) of action requests.", requestsMax );

							return false;
						}

						KernelObservableList<Locator> requestKeys = null;

						// Create a transaction.
						IStorehouseTransaction transaction = ChannelContext.StorehouseProvider.CreateTransaction();

						try
						{
							// When action requests are loaded from the storehouse, the loading must be paged because otherwise
							// if some of them are not sent, the ones queued after them would not get sent either. The loading
							// offset must be incremented at each load.

							PortableTypes actionRequestTypes = PortableTypes.FindAllDerived( typeof( ActionRequest ) );

							int requestsCount = transaction.CountSpeckles( actionRequestTypes, null
								, StorehouseRealms.ActionRequestsToSend );

							// There must be at least 1 action request to load (through the paging mechanism).
							if( requestsCount <= AddFromStorehousePagingOffset )
								AddFromStorehousePagingOffset = 0;

							if( requestsCount > 0 )
							{
								// Load the oldest requests first.
								CollectPaging collectPaging = new CollectPaging( false, AddFromStorehousePagingOffset
									, batchSize );
								requestKeys = transaction.CollectKeys( actionRequestTypes, null
									, StorehouseRealms.ActionRequestsToSend, collectPaging, null
									, null ) as KernelObservableList<Locator>;

								if( requestKeys.Count <= 0 )
									AddFromStorehousePagingOffset = 0;
								else
									AddFromStorehousePagingOffset += requestKeys.Count;

								LogManager.DebugAddFormat( "The action client manager has loaded for processing {0} action"
									+ " request locators from the storehouse.", requestKeys.Count );
							}
							else
							{
								LogManager.DebugAdd( "The action client manager has no action request locators to load"
									+ " for processing from the storehouse.", null );
							}

							// Commit the transaction.
							transaction.Commit();
						}
						catch( Exception ex )
						{
							// Rolls back the transaction.
							transaction.Rollback();

							throw ExceptionUtils.PreserveStackTrace( ex );
						}

						if( ListUtils.NoContent( requestKeys ) )
							return false;

						// Queue for processing only new locators.
						foreach( Locator requestKey in requestKeys )
						{
							if( PendingRequests.FindItemByRequestLocator( requestKey ) == null )
								PendingRequests.Add( new PendingRequest( requestKey, false ) );
						}

						return true;
					}
				}
			}
		}

		/// <summary>
		/// Stops the action client channels.
		/// </summary>
		/// <remarks>
		/// <para>The <see cref="ClientChannels"/> property is thread safe within this method.</para>
		/// </remarks>
		private void StopClientChannels()
		{
			lock( ClientChannelsLock )
			{
				foreach( ActionClientChannel clientChannel in ClientChannels )
				{
					clientChannel.Stop();

					LogManager.DebugAddFormat( "The action client channel for the action request locator \"{0}\" will stop."
						, clientChannel.RequestLocator );
				}

				ClientChannels.Clear();
			}
		}

		/// <summary>
		/// Handler for the thread (<see cref="ProcessingThread"/>) which processes action requests and responses.
		/// </summary>
		/// <remarks>
		/// <para>The <see cref="PendingRequests"/> property is thread safe within this method.</para>
		/// </remarks>
		private void ProcessingThreadHandler()
		{
			try
			{
				LogManager.Add( "The action client manager was started." );

				AddFromStorehouseChrono.Start();

				// Periodically check for new action requests.
				while( true )
				{
					try
					{
						bool isMaintenance;
						lock( ChannelContextLock )
						{
							isMaintenance = ChannelContext.ActionClientHost.CheckMaintenance();
						}

						if( !isMaintenance )
						{
							// From time to time, load action request locators from the storehouse. Since the timespan is random,
							// only the action requests from the storehouse are processed at random moments, with random delays.
							// The action requests wich are manually added to "PendingRequests" are not affected by the random
							// timespan.
							if( AddFromStorehouseChrono.PassedSeconds( ClientStorehouseLoadTimespan ) )
							{
								// Reset the timespan in order to force a new random value to be generated.
								ClientStorehouseLoadTimespan = 0;

								AddFromStorehouse();
							}

							CreateChannels();
						}
					}
					catch( Exception ex )
					{
						// The loop must continue (if possible).

						if( ExceptionUtils.IsAbort( ex ) )
							throw ExceptionUtils.PreserveStackTrace( ex );

						// Log exception.
						LogManager.Add( ex );
					}

					if( StopProcessingThread )
						throw new AbortLongOperationException();

					TimeUtils.SleepMilliseconds( ProcessingSleep );
				}
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				if( ExceptionUtils.IsAbort( ex ) )
				{
					LogManager.Add( "The action client manager has been aborted." );
				}
				else
				{
					LogManager.Add( "The action client manager has failed processing.", ex, LogRecordModes.Error, true );
				}
			}
			finally
			{
				StopClientChannels();
			}
		}

		/// <summary>
		/// Starts processing requests on a new thread.
		/// </summary>
		private void StartProcessingThread()
		{
			ProcessingThread = new Thread( ProcessingThreadHandler );

			// The thread must terminate in order for the application to terminate.
			ProcessingThread.IsBackground = false;

			ProcessingThread.Start();
		}

		/// <summary>
		/// Starts the action client manager.
		/// </summary>
		private void Start()
		{
			StartProcessingThread();
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Adds the specified locator to the list of locators of action requests to process.
		/// </summary>
		/// <param name="requestLocator">The instance locator of the action request to process.</param>
		/// <param name="ignoreScheduler">If this is true then the action request with the specified locator must be processed
		/// even if the retry schedules don't allow it.</param>
		/// <remarks>
		/// <para>If the specified locator is already queued for processing, nothing happens.</para>
		/// 
		/// <para>The <see cref="PendingRequests"/> property is thread safe within this method.</para>
		/// </remarks>
		public void Add( Locator requestLocator, bool ignoreScheduler )
		{
			ValidationUtils.ObjectHasContent( requestLocator, "requestLocator" );

			// Create a client channel to process the action request.
			lock( PendingRequestsLock )
			{
				// Queue for processing only new locators.
				if( PendingRequests.FindItemByRequestLocator( requestLocator ) == null )
					PendingRequests.Add( new PendingRequest( requestLocator, ignoreScheduler ) );
			}
		}

		/// <summary>
		/// Adds the specified locators to the list of locators of action requests to process.
		/// </summary>
		/// <param name="requestLocators">The instance locators of the action requests to process. This may be empty, but not
		/// null.</param>
		/// <param name="ignoreScheduler">If this is true then the action requests with the specified locators must be processed
		/// even if the retry schedules don't allow it.</param>
		/// <remarks>
		/// <para>If the specified locators are already queued for processing, nothing happens.</para>
		/// 
		/// <para>The <see cref="PendingRequests"/> property is thread safe within this method.</para>
		/// </remarks>
		public void Add( Locators requestLocators, bool ignoreScheduler )
		{
			ValidationUtils.ObjectHasContent( requestLocators, "requestLocators" );

			// Create a client channel to process the action request.
			lock( PendingRequestsLock )
			{
				// Queue for processing only new locators.
				foreach( Locator requestLocator in requestLocators )
				{
					if( PendingRequests.FindItemByRequestLocator( requestLocator ) == null )
						PendingRequests.Add( new PendingRequest( requestLocator, ignoreScheduler ) );
				}
			}
		}

		/// <summary>
		/// Stops the action client manager as soon as possible.
		/// </summary>
		public void Stop()
		{
			StopProcessingThread = true;
		}

		/// <summary>
		/// Stops processing the action requests with the specified locators as soon as possible.
		/// </summary>
		/// <param name="requestLocators">The instance locators of the action requests whose processing to stop. If a locator
		/// refers to an action request which is not being processed, nothing happens (with that locator). This may be null or
		/// empty. This must be thread safe.</param>
		/// <remarks>
		/// <para>The <see cref="ClientChannels"/> property is thread safe within this method.</para>
		/// </remarks>
		public void StopRequests( Locators requestLocators )
		{
			if( ListUtils.NoContent( requestLocators ) )
				return;

			lock( ClientChannelsLock )
			{
				foreach( Locator requestLocator in requestLocators )
				{
					ActionClientChannel clientChannel = ClientChannels[ requestLocator ];
					if( clientChannel == null )
						continue;

					clientChannel.Stop();
				}
			}
		}

		#endregion Public methods
	}
}
