////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.ActionProcessor;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Storehouse;

namespace Vela.ActionClient
{
	/// <summary>
	/// Action processor utilities.
	/// </summary>
	public static class ActionClientUtils
	{
		#region Public static methods

		/// <summary>
		/// Performs various operations after the processing state of the specified action progress was updated.
		/// </summary>
		/// <param name="progress">Progress whose processing state was just updated.</param>
		/// <param name="errorResponse">Error action response received while processing the referred action request. This may be
		/// null.</param>
		/// <param name="remoteUpn">Remote UPN where the referred action request was sent for processing. This may be null.
		/// </param>
		/// <param name="transaction">Current storehouse transaction.</param>
		public static void AfterUpdateProcessingState( ActionClientProgress progress, ErrorActionResponse errorResponse
			, Upn remoteUpn, IStorehouseTransaction transaction )
		{
			ValidationUtils.ObjectHasContent( progress, "progress" );
			ValidationUtils.ObjectHasContent( transaction, "transaction" );

			SheafLocks progressLock = progress.StorehouseLock;

			try
			{
				// Since the progress is not locked (the entire time) while an action request is being processed, set the lock
				// now in order to be able to update it.
				progress.SetLock( SheafLocks.Update );

				// Update or delete the action progress.
				if( progress.ProgressIsPermanent || (progress.ProcessingState == ActionProcessingStates.Pending)
					|| (progress.ProcessingState == ActionProcessingStates.Retry) )
				{
					transaction.UpdateSheaf( progress, StorehouseRealms.ActionProgresses );
				}
				else transaction.DeleteSheaf( progress.InstanceLocator, StorehouseRealms.ActionProgresses );

				LogRecordModes logRecordMode =
					((errorResponse == null) && (progress.ProcessingState != ActionProcessingStates.Error))
						? LogRecordModes.Info : LogRecordModes.Error;
				bool saveMessages = (errorResponse != null) || (progress.ProcessingState == ActionProcessingStates.Error);

				string progressMessage = string.Format( "The progress of the action request with the type \"{0}\" and the"
						+ " locator \"{1}\", which was sent to the service with the UPN \"{2}\", has the \"{3}\" processing"
						+ " state."
					, progress.ReferredActionRequestTypeName, progress.ReferredActionRequestLocator, remoteUpn
					, progress.ProcessingState );
				LogManager.DebugAdd( progressMessage, null, logRecordMode, saveMessages );

				// Delete the action request if no more processing attempts may be made.
				if( (progress.ProcessingState != ActionProcessingStates.Pending)
					&& (progress.ProcessingState != ActionProcessingStates.Retry) )
				{
					transaction.DeleteSheaf( progress.ReferredActionRequestLocator, StorehouseRealms.ActionRequestsToSend );

					string requestMessage = string.Format( "The action request with the type \"{0}\" and the locator \"{1}\","
							+ " which was sent to the service with the UPN \"{2}\", has been deleted because no more processing"
							+ " attempts may be made. The processing state is \"{3}\"."
						, progress.ReferredActionRequestTypeName, progress.ReferredActionRequestLocator, remoteUpn
						, progress.ProcessingState );
					LogManager.DebugAdd( requestMessage, null, logRecordMode, saveMessages );
				}
			}
			finally
			{
				progress.SetLock( progressLock );
			}
		}

		/// <summary>
		/// Performs various operations after the processing state of the specified action progress was updated.
		/// </summary>
		/// <param name="progress">Progress whose processing state was just updated.</param>
		/// <param name="errorResponse">Error action response received while processing the referred action request. This may be
		/// null.</param>
		/// <param name="remoteUpn">Remote UPN where the referred action request was sent for processing. This may be null.
		/// </param>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		public static void AfterUpdateProcessingState( ActionClientProgress progress, ErrorActionResponse errorResponse
			, Upn remoteUpn, IStorehouseProvider storehouseProvider )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				AfterUpdateProcessingState( progress, errorResponse, remoteUpn, transaction );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		#endregion Public static methods
	}
}
