////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Reflection;
using Vela.ActionProcessor;
using Vela.Kernel;

namespace Vela.ActionClient
{
	/// <summary>
	/// List of method informations which allow an action client to process last frame action responses.
	/// </summary>
	public class ProcessLastFrameResponseMethodInfos : HandleRequestMethodInfos
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public ProcessLastFrameResponseMethodInfos()
		{
		}

		#endregion Constructors

		#region Private static methods

		/// <summary>
		/// Checks if the specified types are properly derived to form the prototype of a method which processes an action
		/// response (with the same type as "paramType1").
		/// </summary>
		/// <param name="paramType1">The type of the first parameter of the method which processss an action response; this is
		/// the type of the action request which was processed.</param>
		/// <param name="paramType2">The type of the first parameter of the method which processss an action response; this is
		/// the type of the action response to process. This must be derived from <see cref="LastFrameActionResponse"/>.</param>
		/// <param name="paramType3">The type of the third parameter of the method which processes an action response; this is
		/// the type of the response context.</param>
		/// <returns>Returns true if the specified types are properly derived.</returns>
		private static bool CheckBaseTypeDerivation( Type paramType1, Type paramType2, Type paramType3 )
		{
			ValidationUtils.ObjectHasContent( paramType1, "paramType1" );
			ValidationUtils.ObjectHasContent( paramType1, "paramType2" );
			ValidationUtils.ObjectHasContent( paramType2, "paramType3" );

			// The action request.
			if( !typeof( ActionRequest ).IsAssignableFrom( paramType1 ) ) return false;

			// The action response.
			if( !typeof( LastFrameActionResponse ).IsAssignableFrom( paramType2 ) ) return false;

			// The context.
			if( !typeof( ActionResponseContext ).IsAssignableFrom( paramType3 ) ) return false;

			return true;
		}

		#endregion Private static methods

		#region Private methods

		/// <summary>
		/// Adds to the list the method informations which can handle action requests, from the action client with the
		/// specified type name.
		/// </summary>
		/// <param name="actionClientTypeName">The type name of the action client whose method informations to add to the
		/// list. The type must be globally registered.</param>
		protected override void AddMethodInfos( string actionClientTypeName )
		{
			ValidationUtils.TextHasContent( actionClientTypeName, "actionClientTypeName" );

			Type clientType = PortableTypes.FindGlobal( actionClientTypeName );

			if( (clientType == null) || !typeof( IActionClient ).IsAssignableFrom( clientType ) )
			{
				throw new ValidationException( string.Format( "The action client type must derive from \"{0}\"."
					, typeof( IActionClient ).FullName ) );
			}

			// Go through all the public methods.
			MethodInfo[] methodInfos = clientType.GetMethods( BindingFlags.Public | BindingFlags.Instance );
			if( methodInfos != null )
			{
				foreach( MethodInfo mi in methodInfos )
				{
					ParameterInfo[] parameterInfos = mi.GetParameters();
					if( (parameterInfos == null) || (parameterInfos.Length != 3) ) continue;

					Type paramType1 = parameterInfos[ 0 ].ParameterType;
					Type paramType2 = parameterInfos[ 1 ].ParameterType;
					Type paramType3 = parameterInfos[ 2 ].ParameterType;

					if( CheckBaseTypeDerivation( paramType1, paramType2, paramType3 ) )
					{
						// Found an accepted type for an action request.
						ProcessLastFrameResponseMethodInfo requestMethodInfo = new ProcessLastFrameResponseMethodInfo(
							paramType1.FullName, mi );
						Add( requestMethodInfo );
					}
				}
			}
		}

		#endregion Private methods
	}
}
