////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Reflection;
using Vela.Kernel;
using Vela.Kernel.Cloning;

namespace Vela.ActionClient
{
	/// <summary>
	/// Method information which allows an action client to process an action response.
	/// </summary>
	public class ProcessResponseMethodInfo : IUnclonable
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ProcessResponseMethodInfo()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="responseTypeName">The type name of the action response to be processed by the "methodInfo" method.
		/// </param>
		/// <param name="clientTypeName">The type name of the action client to instantiate when an action response with the
		/// type name "responseTypeName" has to be processed by the client. The type must be globally registered.</param>
		/// <param name="methodInfo">The information of the method which must be processed when an action response with the type
		/// name "responseTypeName" has to be processed by the client.</param>
		public ProcessResponseMethodInfo( string responseTypeName, string clientTypeName, MethodInfo methodInfo )
		{
			ResponseTypeName = responseTypeName;
			ClientTypeName = clientTypeName;
			MethodInfo = methodInfo;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private string _ResponseTypeName;
		/// <summary>
		/// The type name of the action response to be processed by the <see cref="MethodInfo"/> method. READ-ONLY.
		/// </summary>
		public string ResponseTypeName
		{
			get { return _ResponseTypeName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "ResponseTypeName" );

				_ResponseTypeName = value;
			}
		}

		private string _ClientTypeName;
		/// <summary>
		/// The type name of the action client to instantiate when an action response with the type name
		/// <see cref="ResponseTypeName"/> has to be processed by the client. The type must be globally registered. READ-ONLY.
		/// </summary>
		public string ClientTypeName
		{
			get { return _ClientTypeName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "ClientTypeName" );

				_ClientTypeName = value;
			}
		}

		[UnclonableAttribute]
		private MethodInfo _MethodInfo;
		/// <summary>
		/// The information of the method which must be processed when an action response with the type name
		/// <see cref="ResponseTypeName"/> has to be processed by the client. READ-ONLY.
		/// </summary>
		public MethodInfo MethodInfo
		{
			get { return _MethodInfo; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "MethodInfo" );

				_MethodInfo = value;
			}
		}

		#endregion Public non-serializable properties
	}
}
