////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Reflection;
using Vela.ActionProcessor;
using Vela.Kernel;
using Vela.Kernel.Cloning;

namespace Vela.ActionClient
{
	/// <summary>
	/// List of method informations which allow an action client to process action responses.
	/// </summary>
	public class ProcessResponseMethodInfos : KernelSortedList<string, ProcessResponseMethodInfo>, IUnclonable
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public ProcessResponseMethodInfos()
			: base( StringInsensitiveComparer.Instance )
		{
		}

		#endregion Constructors

		#region Private static methods

		/// <summary>
		/// Checks if the specified types are properly derived to form the prototype of a method which processes an action
		/// response (with the same type as "paramType1").
		/// </summary>
		/// <param name="paramType1">The type of the first parameter of the method which processss an action response; this is
		/// the type of the action response to process.</param>
		/// <param name="paramType2">The type of the second parameter of the method which processss an action response; this is
		/// the type of the error action response to process. Because error action responses are not derived from "paramType1",
		/// they must be sent as a separate parameter ("paramType2").</param>
		/// <param name="paramType3">The type of the third parameter of the method which processes an action response; this is
		/// the type of the response context.</param>
		/// <returns>Returns true if the specified types are properly derived.</returns>
		private static bool CheckBaseTypeDerivation( Type paramType1, Type paramType2, Type paramType3 )
		{
			ValidationUtils.ObjectHasContent( paramType1, "paramType1" );
			ValidationUtils.ObjectHasContent( paramType2, "paramType2" );
			ValidationUtils.ObjectHasContent( paramType3, "paramType3" );

			int responseContextTypeMode, contextTypeMode;

			// The (error) action response.
			if( !typeof( ActionResponse ).IsAssignableFrom( paramType1 )
				|| !typeof( ErrorActionResponse ).IsAssignableFrom( paramType2 ) )
			{
				return false;
			}

			// The response context.
			if( typeof( ActionResponse ).IsAssignableFrom( paramType1 ) )
				responseContextTypeMode = 0;
			else
				return false;

			// The context.
			if( typeof( ActionResponseContext ).IsAssignableFrom( paramType3 ) )
				contextTypeMode = 0;
			else
				return false;

			return responseContextTypeMode == contextTypeMode;
		}

		#endregion Private static methods

		#region Private methods

		/// <summary>
		/// Adds to the list the method informations which can process action responses, from the action client with the
		/// specified type name.
		/// </summary>
		/// <param name="actionClientTypeName">The type name of the action client whose method informations to add to the
		/// list. The type must be globally registered.</param>
		private void AddMethodInfos( string actionClientTypeName )
		{
			ValidationUtils.TextHasContent( actionClientTypeName, "actionClientTypeName" );

			Type clientType = PortableTypes.FindGlobal( actionClientTypeName );

			if( (clientType == null) || !typeof( IActionClient ).IsAssignableFrom( clientType ) )
			{
				throw new ValidationException( string.Format( "The action client type must derive from \"{0}\"."
					, typeof( IActionClient ).FullName ) );
			}

			// Go through all the public methods.
			MethodInfo[] methodInfos = clientType.GetMethods( BindingFlags.Public | BindingFlags.Instance );
			if( methodInfos != null )
			{
				foreach( MethodInfo mi in methodInfos )
				{
					ParameterInfo[] parameterInfos = mi.GetParameters();
					if( (parameterInfos == null) || (parameterInfos.Length != 3) )
						continue;

					Type paramType1 = parameterInfos[ 0 ].ParameterType;
					Type paramType2 = parameterInfos[ 1 ].ParameterType;
					Type paramType3 = parameterInfos[ 2 ].ParameterType;

					if( CheckBaseTypeDerivation( paramType1, paramType2, paramType3 ) )
					{
						// Found an accepted type for an action response.
						ProcessResponseMethodInfo processResponseMethodInfo = new ProcessResponseMethodInfo( paramType1.FullName
							, actionClientTypeName, mi );
						Add( processResponseMethodInfo );
					}
				}
			}
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para><see cref="ProcessResponseMethodInfo.ResponseTypeName"/> is used as key.</para>
		/// </remarks>
		public override string GetKey( ProcessResponseMethodInfo item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.ResponseTypeName;
		}

		/// <summary>
		/// Adds to the list the method informations which can process action responses, from the action client with the
		/// specified type names.
		/// </summary>
		/// <param name="actionClientTypeNames">The type names of the action clients whose method informations to add to the
		/// list. The types must be globally registered.</param>
		public void Add( Strings actionClientTypeNames )
		{
			ValidationUtils.ListHasContent( actionClientTypeNames, "actionClientTypeNames" );

			foreach( string actionClientTypeName in actionClientTypeNames )
				AddMethodInfos( actionClientTypeName );
		}

		/// <summary>
		/// Adds to the list the method informations which can process action responses, from the action client with the
		/// specified type name.
		/// </summary>
		/// <param name="actionClientTypeName">The type name of the action client whose method informations to add to the
		/// list. The type must be globally registered.</param>
		public void Add( string actionClientTypeName )
		{
			AddMethodInfos( actionClientTypeName );
		}

		#endregion Public methods
	}
}
