////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Reflection;
using Vela.ActionProcessor;
using Vela.Kernel;

namespace Vela.ActionClient
{
	/// <summary>
	/// List of method informations which allow an action client to update action requests before sending them to an action
	/// service.
	/// </summary>
	public class UpdateRequestMethodInfos : HandleRequestMethodInfos
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public UpdateRequestMethodInfos()
		{
		}

		#endregion Constructors

		#region Private static methods

		/// <summary>
		/// Checks if the specified types are properly derived to form the prototype of a method which updates an action
		/// request (with the same type as "paramType1").
		/// </summary>
		/// <param name="returnType">The type of the returned value of the method which updates an action request.</param>
		/// <param name="paramType1">The type of the first parameter of the method which updates an action request; this is the
		/// type of the action request to update.</param>
		/// <param name="paramType2">The type of the second parameter of the method which updates an action request; this is
		/// the type of the request context.</param>
		/// <returns>Returns true if the specified types are properly derived.</returns>
		private static bool CheckBaseTypeDerivation( Type returnType, Type paramType1, Type paramType2 )
		{
			if( returnType == null ) return false;

			ValidationUtils.ObjectHasContent( returnType, "returnType" );
			ValidationUtils.ObjectHasContent( paramType1, "paramType1" );
			ValidationUtils.ObjectHasContent( paramType2, "paramType2" );

			int requestContextTypeMode, contextTypeMode;

			// The returned value.
			if( returnType != typeof( bool ) ) return false;

			// The request context.
			if( typeof( ActionRequest ).IsAssignableFrom( paramType1 ) ) requestContextTypeMode = 0;
			else return false;

			// The context.
			if( typeof( ActionClientContext ).IsAssignableFrom( paramType2 ) ) contextTypeMode = 0;
			else return false;

			return requestContextTypeMode == contextTypeMode;
		}

		#endregion Private static methods

		#region Private methods

		/// <summary>
		/// Adds to the list the method informations which can handle action requests, from the action client with the
		/// specified type name.
		/// </summary>
		/// <param name="actionClientTypeName">The type name of the action client whose method informations to add to the
		/// list. The type must be globally registered.</param>
		protected override void AddMethodInfos( string actionClientTypeName )
		{
			ValidationUtils.TextHasContent( actionClientTypeName, "actionClientTypeName" );

			Type clientType = PortableTypes.FindGlobal( actionClientTypeName );

			if( (clientType == null) || !typeof( IActionClient ).IsAssignableFrom( clientType ) )
			{
				throw new ValidationException( string.Format( "The action client type must derive from \"{0}\"."
					, typeof( IActionClient ).FullName ) );
			}

			// Go through all the public methods.
			MethodInfo[] methodInfos = clientType.GetMethods( BindingFlags.Public | BindingFlags.Instance );
			if( methodInfos != null )
			{
				foreach( MethodInfo mi in methodInfos )
				{
					ParameterInfo[] parameterInfos = mi.GetParameters();
					if( (parameterInfos == null) || (parameterInfos.Length != 2) ) continue;

					Type returnType = mi.ReturnType;
					Type paramType1 = parameterInfos[ 0 ].ParameterType;
					Type paramType2 = parameterInfos[ 1 ].ParameterType;

					if( CheckBaseTypeDerivation( returnType, paramType1, paramType2 ) )
					{
						// Found an accepted type for an action request.
						UpdateRequestMethodInfo requestMethodInfo = new UpdateRequestMethodInfo( paramType1.FullName, mi );
						Add( requestMethodInfo );
					}
				}
			}
		}

		#endregion Private methods
	}
}
