////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.ActionProcessor;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Kernel;
using Vela.Kernel.Cloning;
using Vela.Storehouse;

namespace Vela.ActionClient
{
	/// <summary>
	/// Action client channel context.
	/// </summary>
	public class ActionClientChannelContext
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionClientChannelContext()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseProvider">The storehouse that the client channel can use while processing action responses.
		/// </param>
		/// <param name="processorConfiguration">The action processor configuration.</param>
		/// <param name="updateRequestMethodInfos">List of method informations which can update action requests. Do not modify
		/// the content of this object while the processor is running. This may be empty, but not null.</param>
		/// <param name="processResponseMethodInfos">List of method informations which can process action responses. Do not
		/// modify the content of this object while the processor is running.</param>
		/// <param name="processLastFrameResponseMethodInfos">List of method informations which can process last frame action
		/// responses. Do not modify the content of this object while the processor is running. This may be empty, but not
		/// null.</param>
		/// <param name="cryptoAlgorithmPool">Pool of cryptographic algorithms. Its content must be thread safe.</param>
		/// <param name="coreAlgorithmPool">Pool of core algorithms. Its content must be thread safe.</param>
		/// <param name="recipientAccessStatesManager">Manager for the recipient access states. Its content must be thread
		/// safe.</param>
		/// <param name="upnResolutionLocker">Locker for the resolution of UPNs by an action client.</param>
		/// <param name="actionClientHost">Host used for updating the user interface of an action client application after an
		/// action request and response are processed. Its content must be thread safe.</param>
		public ActionClientChannelContext( IStorehouseProvider storehouseProvider
			, ActionProcessorConfiguration processorConfiguration, UpdateRequestMethodInfos updateRequestMethodInfos
			, ProcessResponseMethodInfos processResponseMethodInfos
			, ProcessLastFrameResponseMethodInfos processLastFrameResponseMethodInfos, ICryptoAlgorithmPool cryptoAlgorithmPool
			, ICoreAlgorithmPool coreAlgorithmPool, RecipientAccessStatesManager recipientAccessStatesManager
			, UpnResolutionLocker upnResolutionLocker, IActionClientHost actionClientHost )
		{
			StorehouseProvider = storehouseProvider;
			ProcessorConfiguration = processorConfiguration;
			UpdateRequestMethodInfos = updateRequestMethodInfos;
			ProcessResponseMethodInfos = processResponseMethodInfos;
			ProcessLastFrameResponseMethodInfos = processLastFrameResponseMethodInfos;
			CryptoAlgorithmPool = cryptoAlgorithmPool;
			CoreAlgorithmPool = coreAlgorithmPool;
			RecipientAccessStatesManager = recipientAccessStatesManager;
			UpnResolutionLocker = upnResolutionLocker;
			ActionClientHost = actionClientHost;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private IStorehouseProvider _StorehouseProvider;
		/// <summary>
		/// The storehouse provider that the client channel can use while processing action responses. READ-ONLY.
		/// </summary>
		public IStorehouseProvider StorehouseProvider
		{
			get { return _StorehouseProvider; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "StorehouseProvider" );

				_StorehouseProvider = value;
			}
		}

		private ActionProcessorConfiguration _ProcessorConfiguration;
		/// <summary>		
		/// The action processor configuration. READ-ONLY.
		/// </summary>
		public ActionProcessorConfiguration ProcessorConfiguration
		{
			get { return _ProcessorConfiguration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ProcessorConfiguration" );

				_ProcessorConfiguration = value;
			}
		}

		[UnclonableAttribute]
		private UpdateRequestMethodInfos _UpdateRequestMethodInfos;
		/// <summary>
		/// List of method informations which can update action requests. Do not modify the content of this object while the
		/// processor is running. This may be empty, but not null. READ-ONLY.
		/// </summary>
		public UpdateRequestMethodInfos UpdateRequestMethodInfos
		{
			get { return _UpdateRequestMethodInfos; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "UpdateRequestMethodInfos" );

				_UpdateRequestMethodInfos = value;
			}
		}

		[UnclonableAttribute]
		private ProcessResponseMethodInfos _ProcessResponseMethodInfos;
		/// <summary>
		/// List of method informations which can process action responses. Do not modify the content of this object while the
		/// processor is running. READ-ONLY.
		/// </summary>
		public ProcessResponseMethodInfos ProcessResponseMethodInfos
		{
			get { return _ProcessResponseMethodInfos; }
			protected set
			{
				ValidationUtils.ListHasContent( value, "ProcessResponseMethodInfos" );

				_ProcessResponseMethodInfos = value;
			}
		}

		[UnclonableAttribute]
		private ProcessLastFrameResponseMethodInfos _ProcessLastFrameResponseMethodInfos;
		/// <summary>
		/// List of method informations which can process last frame action responses. Do not modify the content of this object
		/// while the processor is running. This may be empty, but not null. READ-ONLY.
		/// </summary>
		public ProcessLastFrameResponseMethodInfos ProcessLastFrameResponseMethodInfos
		{
			get { return _ProcessLastFrameResponseMethodInfos; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ProcessLastFrameResponseMethodInfos" );

				_ProcessLastFrameResponseMethodInfos = value;
			}
		}

		[UnclonableAttribute]
		private ICryptoAlgorithmPool _CryptoAlgorithmPool;
		/// <summary>
		/// Pool of cryptographic algorithms. Its content must be thread safe. READ-ONLY.
		/// </summary>
		public ICryptoAlgorithmPool CryptoAlgorithmPool
		{
			get { return _CryptoAlgorithmPool; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "CryptoAlgorithmPool" );

				_CryptoAlgorithmPool = value;
			}
		}

		[UnclonableAttribute]
		private ICoreAlgorithmPool _CoreAlgorithmPool;
		/// <summary>
		/// Pool of core algorithms. Its content must be thread safe. READ-ONLY.
		/// </summary>
		public ICoreAlgorithmPool CoreAlgorithmPool
		{
			get { return _CoreAlgorithmPool; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "CoreAlgorithmPool" );

				_CoreAlgorithmPool = value;
			}
		}

		[UnclonableAttribute]
		private RecipientAccessStatesManager _RecipientAccessStatesManager;
		/// <summary>
		/// Manager for the recipient access states. Its content must be thread safe. READ-ONLY.
		/// </summary>
		public RecipientAccessStatesManager RecipientAccessStatesManager
		{
			get { return _RecipientAccessStatesManager; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RecipientAccessStatesManager" );

				_RecipientAccessStatesManager = value;
			}
		}

		[UnclonableAttribute]
		private UpnResolutionLocker _UpnResolutionLocker;
		/// <summary>
		/// Locker for the resolution of UPNs by an action client. Its content must be thread safe. READ-ONLY.
		/// </summary>
		public UpnResolutionLocker UpnResolutionLocker
		{
			get { return _UpnResolutionLocker; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "UpnResolutionLocker" );

				_UpnResolutionLocker = value;
			}
		}

		[UnclonableAttribute]
		private IActionClientHost _ActionClientHost;
		/// <summary>
		/// Host used for updating the user interface of an action client application after an action request and response
		/// are processed. Its content must be thread safe. READ-ONLY.
		/// </summary>
		public IActionClientHost ActionClientHost
		{
			get { return _ActionClientHost; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionClientHost" );

				_ActionClientHost = value;
			}
		}

		#endregion Public non-serializable properties
	}
}
