////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.ActionProcessor;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Kernel.Cloning;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.ActionClient
{
	/// <summary>
	/// Contract for a context for a service which processes action requests.
	/// </summary>
	public class ActionClientContext
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionClientContext()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseProvider">The storehouse that the client channel can use while processing action responses.
		/// </param>
		/// <param name="processorConfiguration">The action processor configuration.</param>
		/// <param name="serializer">The service serializer.</param>
		/// <param name="actionRequest">Action request which is currently being processed.</param>
		/// <param name="actionProgress">The progress of the action request which is currently being processed.</param>
		/// <param name="tally">The tally of the recipient of the action request which is currently being processed. This may be
		/// null if the action request is derived from <see cref="AnonymousActionRequest"/>.</param>
		/// <param name="remoteUpn">The UPN of the service which is the recipient of the action request.</param>
		/// <param name="remoteIpAddress">Remote IP address.</param>
		/// <param name="responseEncryptionKey">The encryption key of the response.</param>
		/// <param name="actionClientHost">Host used for updating the user interface of an action client application after an
		/// action request and response are processed. Its content must be thread safe.</param>
		public ActionClientContext( IStorehouseProvider storehouseProvider, ActionProcessorConfiguration processorConfiguration
			, ISerializer serializer, ActionRequest actionRequest, ActionClientProgress actionProgress, Tally tally
			, Upn remoteUpn, byte[] remoteIpAddress, byte[] responseEncryptionKey, IActionClientHost actionClientHost )
		{
			if( (tally == null) && (actionRequest as AnonymousActionRequest == null) )
			{
				throw new ValidationException( string.Format( "The tally of the recipient of the action request which is being"
						+ " processed may be null only if the action request is derived from \"{0}\"."
					, typeof( AnonymousActionRequest ) ) );
			}

			StorehouseProvider = storehouseProvider;
			ProcessorConfiguration = processorConfiguration;
			Serializer = serializer;
			ActionRequest = actionRequest;
			ActionProgress = actionProgress;
			Tally = tally;
			RemoteUpn = remoteUpn;
			RemoteIpAddress = remoteIpAddress;
			ResponseEncryptionKey = responseEncryptionKey;
			ActionClientHost = actionClientHost;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private IStorehouseProvider _StorehouseProvider;
		/// <summary>
		/// The storehouse provider that the client channel can use while processing action responses. READ-ONLY.
		/// </summary>
		public IStorehouseProvider StorehouseProvider
		{
			get { return _StorehouseProvider; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "StorehouseProvider" );

				_StorehouseProvider = value;
			}
		}

		private ActionProcessorConfiguration _ProcessorConfiguration;
		/// <summary>		
		/// The action processor configuration. READ-ONLY.
		/// </summary>
		public ActionProcessorConfiguration ProcessorConfiguration
		{
			get { return _ProcessorConfiguration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ProcessorConfiguration" );

				_ProcessorConfiguration = value;
			}
		}

		private ISerializer _Serializer;
		/// <summary>
		/// The service serializer. READ-ONLY.
		/// </summary>
		public ISerializer Serializer
		{
			get { return _Serializer; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Serializer" );

				_Serializer = value;
			}
		}

		private ActionRequest _ActionRequest;
		/// <summary>
		/// Action request which is currently being processed. READ-ONLY.
		/// </summary>
		public ActionRequest ActionRequest
		{
			get { return _ActionRequest; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionRequest" );

				_ActionRequest = value;
			}
		}

		private ActionClientProgress _ActionProgress;
		/// <summary>
		/// The progress of the action request which is currently being processed. This is not updated with the progress from
		/// the end of the action processing. READ-ONLY.
		/// </summary>
		public ActionClientProgress ActionProgress
		{
			get { return _ActionProgress; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionProgress" );

				_ActionProgress = value;
			}
		}

		private Tally _Tally;
		/// <summary>
		/// The tally of the recipient of the action request which is currently being processed. This may be null if the action
		/// request is derived from <see cref="AnonymousActionRequest"/>. READ-ONLY.
		/// </summary>
		public Tally Tally
		{
			get { return _Tally; }
			protected set { _Tally = value; }
		}

		private Upn _RemoteUpn;
		/// <summary>
		/// The UPN of the service which is the recipient of the action request. READ-ONLY.
		/// </summary>
		public Upn RemoteUpn
		{
			get { return _RemoteUpn; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RemoteUpn" );

				_RemoteUpn = value;
			}
		}

		private byte[] _RemoteIpAddress;
		/// <summary>
		/// The IP address of the service of the action request. This may be null or empty. READ-ONLY.
		/// </summary>
		public byte[] RemoteIpAddress
		{
			get { return _RemoteIpAddress; }
			protected set { _RemoteIpAddress = value; }
		}

		private byte[] _ResponseEncryptionKey;
		/// <summary>
		/// The encryption key of the response (<see cref="Vela.ActionProcessor.ActionRequest.ResponseEncryptionKey"/>).
		/// READ-ONLY.
		/// </summary>
		public byte[] ResponseEncryptionKey
		{
			get { return _ResponseEncryptionKey; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "ResponseEncryptionKey" );

				_ResponseEncryptionKey = value;
			}
		}

		[UnclonableAttribute]
		private IActionClientHost _ActionClientHost;
		/// <summary>
		/// Host used for updating the user interface of an action client application after an action request and response
		/// are processed. Its content must be thread safe. READ-ONLY.
		/// </summary>
		public IActionClientHost ActionClientHost
		{
			get { return _ActionClientHost; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionClientHost" );

				_ActionClientHost = value;
			}
		}

		[UnclonableAttribute]
		private IStorehouseTransaction _Transaction;
		/// <summary>
		/// Current storehouse transaction. This may be null.
		/// </summary>
		/// <remarks>
		/// <para>This is committed (by an action client) even if an error response is received.</para>
		/// 
		/// <para>If an action request with a multiframe response is being processed, this is available only to the first frame
		/// (in the method which processes the action response).</para>
		/// </remarks>
		public IStorehouseTransaction Transaction
		{
			get { return _Transaction; }
			set { _Transaction = value; }
		}

		[UnclonableAttribute]
		private UpdateActionRequestContext _UpdateRequestContext;
		/// <summary>
		/// Context for an action request which is updated (before it's sent to the service for processing). This may be null.
		/// </summary>
		public UpdateActionRequestContext UpdateRequestContext
		{
			get { return _UpdateRequestContext; }
			set { _UpdateRequestContext = value; }
		}

		#endregion Public non-serializable properties
	}
}
