////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.ActionProcessor;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Storehouse;

namespace Vela.ActionClient
{
	/// <summary>
	/// Contract for an action client.
	/// </summary>
	/// <remarks>
	/// <para>An action client processes action responses (<see cref="Vela.ActionProcessor.ActionResponse"/>),
	/// and responds with action responses (<see cref="Vela.ActionProcessor.ActionResponse"/>).</para>
	/// 
	/// <para>If an error occurs while an action client processes an action request (and response), a new attempt will be made
	/// later only if the error is not fatal; for details, see the
	/// <see cref="Vela.ActionProcessor.ErrorActionResponse.HasFatalCrossServiceException"/> method.</para>
	/// 
	/// <para>All the sheafs are locked with <see cref="SheafLocks.Shared"/>, unless they will be saved (in which case they are
	/// locked with <see cref="SheafLocks.Update"/>).</para>
	/// </remarks>
	public interface IActionClient : IPortableType
	{
		#region Public methods

		/// <summary>
		/// Updates the specified action request before it's sent to the service.
		/// </summary>
		/// <param name="actionRequest">Action request to update. This may be modified in here if the request's
		/// <see cref="Vela.ActionProcessor.ActionRequest.NeedsUpdateBeforeSending"/> is true; it's also automatically
		/// signed afterwards.</param>
		/// <param name="clientContext">The action client context.</param>
		/// <param name="transaction">The transaction with which to load documents. This is separate from the client context
		/// in order to ensure that it's not used by the derived action client to make changes in the storehouse and expect
		/// them to be rolled back if the action response can't be processed.</param>
		/// <param name="updateRequestMethodInfo">The information of the method which can update the action request.</param>
		/// <returns>Returns the true if the action request must be sent to the service, and false if it must never be sent.
		/// </returns>
		/// <remarks>
		/// <para>See <see cref="Vela.ActionProcessor.ActionRequest.NeedsUpdateBeforeSending"/> for details.
		/// </para>
		/// </remarks>
		bool UpdateActionRequest( ActionRequest actionRequest, ActionClientContext clientContext
			, IStorehouseTransaction transaction, UpdateRequestMethodInfo updateRequestMethodInfo );

		/// <summary>
		/// Processes the specified action response.
		/// </summary>
		/// <param name="actionResponse">The action response to process.</param>
		/// <param name="clientContext">The action client context.</param>
		/// <param name="processResponseMethodInfo">The information of the method which can process the action response.</param>
		void ProcessResponse( ActionResponse actionResponse, ActionClientContext clientContext
			, ProcessResponseMethodInfo processResponseMethodInfo );

		/// <summary>
		/// Processes the specified last frame action response.
		/// </summary>
		/// <param name="actionRequest">The action request which was processed.</param>
		/// <param name="actionResponse">The last frame action response to process.</param>
		/// <param name="clientContext">The action client context.</param>
		/// <param name="processLastFrameResponseMethodInfo">The information of the method which can process the last frame
		/// action response. This may be null.</param>
		void ProcessLastFrameResponse( ActionRequest actionRequest, LastFrameActionResponse actionResponse
			, ActionClientContext clientContext, ProcessLastFrameResponseMethodInfo processLastFrameResponseMethodInfo );

		#endregion Public methods
	}
}
