////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

namespace Vela.ActionClient
{
	/// <summary>
	/// <para>An action client receives action responses and dispatches them to the appropriate processing module.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Code flow</b></para>
	/// 
	/// <para>An action client channel sends an action request to a service and receives an encrypted envelope with an action
	/// response, as a buffer. This is deserialized and is dispatched to the appropriate action client based on the type of the
	/// action response.</para>
	/// 
	/// <para>A storehouse transaction is created only to load the action request and progress. This is because the UPN
	/// resolution could take a long time, and the progress should not be locked for update during all this time.</para>
	/// 
	/// <para>The action progress is loaded. The retry schedules are checked to see if the action request may be processed. All
	/// the action requests which are queued in (and loaded from) the storehouse honor the retry schedules. However, there are
	/// cases when the retry schedules are ignored, for example when a letter is queued to be sent.</para>
	/// 
	/// <para>The action request is loaded.</para>
	/// 
	/// <para>Expired action requests are not processed.</para>
	/// 
	/// <para>The tally and the ephemerid of the recipient / service of the action request to process are loaded.</para>
	/// 
	/// <para>If no IP address for the recipient / service is found, an attempt is made to resolve to service UPN to an IP
	/// address. If the UPN is resolved, the ephemerid is saved (with the new IP address).</para>
	/// 
	/// <para>For action requests whose <see cref="Vela.ActionProcessor.ActionRequest.NeedsUpdateBeforeSending"/>
	/// is true, the derived action client classes must contain public instance update methods which have one parameter derived
	/// from <see cref="Vela.ActionProcessor.ActionRequest"/> and one parameter derived from
	/// <see cref="ActionClientContext"/>, and a boolean return a value. These methods are executed by the client just before the
	/// action request is sent to the service for processing; the method whose action request parameter has the same type (as the
	/// action request which is being processed) is executed.</para>
	/// 
	/// <para>If an action request's <see cref="Vela.ActionProcessor.ActionRequest.NeedsUpdateBeforeSending"/> is true, the
	/// request is updated before it's sent to the service; it's also automatically signed afterwards. The action client must
	/// contain a method to update the action request type.</para>
	/// 
	/// <para>The action request is sent to the service and an action response is received from the service. This happens except
	/// when the action request's <see cref="Vela.ActionProcessor.ActionRequest.NeedsUpdateBeforeSending"/> is
	/// true and the request update method returns false.</para>
	/// 
	/// <para>The action client decrypts the encrypted envelope and extracts the action response from it.</para>
	/// 
	/// <para>Expired action responses are not processed.</para>
	/// 
	/// <para>The signatures of the secondary frames of a multiframe action response are not verified if they are derived from
	/// <see cref="Vela.ActionProcessor.IMultiframeActionResponse"/>.</para>
	/// 
	/// <para>Entropy is added to the entropy algorithm of the client, using only private data from the action response.</para>
	/// 
	/// <para>A storehouse transaction is created. If the action request has a multiframe action response, the transaction is
	/// available only for the first frame (because the action request may take a long time to be processed), but a new one can
	/// be created from the available storehouse provider; make sure to close the created transaction.</para>
	/// 
	/// <para>An action client is instantiated (every time an encrypted envelope is received), so don't perform time consuming
	/// tasks in a client's constructor. This also means that static objects must be synchronized for multi-threading.</para>
	/// 
	/// <para>The derived action client classes must contain public instance processing methods which have one parameter
	/// derived from <see cref="Vela.ActionProcessor.ActionResponse"/>, one parameter derived from
	/// <see cref="Vela.ActionProcessor.ErrorActionResponse"/>, and one parameter derived from
	/// <see cref="ActionResponseContext"/>; the first two parameters may be null, but not at the same time; if the second
	/// parameter is not null, it has priority over the first one. These methods are executed by the client when an action
	/// response is received from a service; the method whose action response parameter has the same type (as the received action
	/// response) is executed.</para>
	/// 
	/// <para>Because error action responses are not derived from the type of the first parameter, they must be sent as a
	/// separate parameter (= the second parameter). Otherwise it would not be possible for the action client to dynamically find
	/// the correct method to call.</para>
	/// 
	/// <para>If the action request has a multiframe action response, a method which processes the last frame action response is
	/// called, if it exists, from the action client. This method must be public instance and must have one parameter derived
	/// from <see cref="Vela.ActionProcessor.ActionRequest"/>, one derived from
	/// <see cref="Vela.ActionProcessor.LastFrameActionResponse"/>, and one parameter derived from
	/// <see cref="ActionResponseContext"/>. The method whose action request parameter has the same type (as the action request
	/// which is being processed) is executed.</para>
	/// 
	/// <para>In the processing method, the action response must be processed.</para>
	/// 
	/// <para>In the processing method, documents can be added / updated in the storehouse, or deleted from the storehouse,
	/// using the storehouse transaction.</para>
	/// 
	/// <para>In the processing method, do not modify the vault and do not try to save it in the storehouse. If some data
	/// associated to the client must be saved, do so in another document (which must first be loaded from the storehouse).
	/// Still, note that this may severely impact the scalability of the client.</para>
	/// 
	/// <para>If any error occurs in the processing method, throw an exception. In such a case, the storehouse transaction will
	/// be rolled back.</para>
	/// 
	/// <para>If the action request has a multiframe action response, after all the frames of the multiframe response are
	/// received, a response with the type <see cref="Vela.ActionProcessor.LastFrameActionResponse"/> is received. This allows
	/// the client to perform specific operations after the entire multiframe response is received successfully.</para>
	/// 
	/// <para>It's considered that there is a fatal error if the received response is an error response and it's
	/// <see cref="Vela.ActionProcessor.ErrorActionResponse.HasFatalCrossServiceException"/> method returns true. If the error
	/// isn't fatal, a new attempt to process the action request will be made later.</para>
	/// 
	/// <para>If the action request was processed successfully or a fatal error has occurred, the action request is deleted from
	/// the storehouse. Also, the action progress is deleted / updated from / in the storehouse. The progress is deleted only if
	/// its <see cref="Vela.ActionProcessor.ActionClientProgress.ProgressIsPermanent"/> (which is copied from the referred
	/// action request's <see cref="Vela.ActionProcessor.ActionRequestWithInitiator.ClientProgressIsPermanent"/>) is
	/// false.</para>
	/// 
	/// <para>The storehouse transaction is committed.</para>
	/// 
	/// <para>The client application's UI is updated.</para>
	/// 
	/// <para>If an error has occurred, an attempt is made to resolve to service UPN to an IP address. If the UPN is resolved,
	/// the ephemerid is saved (with the new IP address).</para>
	/// 
	/// <para>If an error has occurred, the action progress is updated.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Client scalability</b></para>
	/// 
	/// <para>The scalability of the client depends on the scalability of the used storehouse.</para>
	/// 
	/// <para>If a client runs on multiple computers, where each of them sends action requests and receives action responses,
	/// the client is scalable up to the storehouse's capacity.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Multi-threading synchronization</b></para>
	/// 
	/// <para>All non-centralized used resources must be locked while processing each action response. The centralized resources
	/// should not be modified while the response is running; an action response receives a copy of the vault, so it can change
	/// its properties, but this should be avoided.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Various details</b></para>
	/// 
	/// <para>The action client doesn't lock the storehouse for its thread, it clones it. Therefore, if the properties of the
	/// storehouse must be changed from other threads, stop the client, change the properties and start the client.</para>
	/// 
	/// <para>The client relies on the storehouse for transactionality and fault-tolerance.</para>
	/// 
	/// <para>Don't add the IP address of the action response to the entropy algorithm because if action responses come in
	/// through traffic redirecting services, the IP would always be the same (or one of very few possible IP addresses).</para>
	/// </summary>
	class NamespaceDoc
	{
		// Only for documentation.
	}
}
