////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.ActionProcessor;
using Vela.Kernel;

namespace Vela.ActionClient
{
	/// <summary>
	/// The access state of the recipient of an action request processed by an action client.
	/// </summary>
	public class RecipientAccessState
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected RecipientAccessState()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="recipientUpnFingerprint">The fingerprint of the UPN whose access state is represented by the current
		/// object.</param>
		public RecipientAccessState( Locator recipientUpnFingerprint )
		{
			RecipientUpnFingerprint = recipientUpnFingerprint;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private Locator _RecipientUpnFingerprint;
		/// <summary>
		/// The fingerprint of the UPN whose access state is represented by the current object. READ-ONLY.
		/// </summary>
		public Locator RecipientUpnFingerprint
		{
			get { return _RecipientUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RecipientUpnFingerprint" );

				_RecipientUpnFingerprint = value;
			}
		}

		private DateTime _CreationTime = DateTime.UtcNow;
		/// <summary>
		/// The time (in UTC format) when the current object was created. This is automatically set with the time of the creation
		/// of the current object (and you should not modify it). READ-ONLY.
		/// </summary>
		public DateTime CreationTime
		{
			get { return _CreationTime; }
			protected set { _CreationTime = value; }
		}

		private DateTime _LastAttemptTime = DateTime.UtcNow;
		/// <summary>
		/// The time (in UTC format) when the last attempt to (whose recipient UPN fingerprint is
		/// <see cref="RecipientUpnFingerprint"/>) was made. This is automatically set with the time of the creation of the
		/// current object. READ-ONLY.
		/// </summary>
		public DateTime LastAttemptTime
		{
			get { return _LastAttemptTime; }
			protected set { _LastAttemptTime = value; }
		}

		private DateTime _NextAttemptTime = DateTime.UtcNow;
		/// <summary>
		/// The time (in UTC format) when the next attempt to process an action request (whose recipient UPN fingerprint is
		/// <see cref="RecipientUpnFingerprint"/>) should be made. This is automatically set with the time of the creation of the
		/// current object. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is strictly associated to <see cref="RetryIndex"/>. When <see cref="RetryIndex"/> changes, this must also
		/// change.</para>
		/// </remarks>
		public DateTime NextAttemptTime
		{
			get { return _NextAttemptTime; }
			protected set { _NextAttemptTime = value; }
		}

		private int _AttemptIndex;
		/// <summary>
		/// The index of the last attempt to process an action request whose recipient UPN fingerprint is
		/// <see cref="RecipientUpnFingerprint"/>. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>Initially, if this is 0 or a negative value, it means that no attempt has been made to process the action
		/// request. If this is 1, it means that an attempt has been made to process the action request, but no retry has been
		/// made.</para>
		/// </remarks>
		public int AttemptIndex
		{
			get { return _AttemptIndex; }
			protected set { _AttemptIndex = value; }
		}

		/// <summary>
		/// The index of the last retry to process the referred action request. READ-ONLY.
		/// </summary>
		public int RetryIndex
		{
			get { return AttemptIndex - 1; }
		}

		private bool _WasResetOnSuccessfulUpnResolution = false;
		/// <summary>
		/// If this is true then the current access state was reset when this was false and an attempt to resolve the
		/// <see cref="RecipientUpnFingerprint"/> UPN fingerprint was successfully made. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para><see cref="RecipientAccessStatesManager.RecipientIsAccessibleAndProcessing"/> should be called every time when
		/// an attempt to resolve the <see cref="RecipientUpnFingerprint"/> UPN fingerprint is successfully made. However, if
		/// that were to happen before the action service successfully processes an action request, and before the action client
		/// receives a signed action response which indicates the success, the action client would retry to send the action
		/// request as often as possible, fact which would uselessly waste network resources (because it's possible that the
		/// service would throw a <see cref="RetryCrossServiceException"/> exception, for example, if the account between the
		/// peers doesn't exist yet).</para>
		/// 
		/// <para>However, not calling <see cref="RecipientAccessStatesManager.RecipientIsAccessibleAndProcessing"/> can create
		/// time gaps between consecutive retries to send an action request to an actionservice.</para>
		/// 
		/// <para>For example, if the recipient peer of an action request comes online a day after the request was first sent,
		/// and has a IP address different than the one which the action client already knows, which is the usual case, the
		/// client already has a large time gap between consecutive retries. Since the UPN resolution is performed after an
		/// action request could not be sent (for performance reasons it can't be performed before), if the UPN resolution
		/// succeeds then the next attempt (which would normally be successful if done immediately) would be after a long time.
		/// </para>
		/// 
		/// <para>To avoid both issues, <see cref="ResetOnSuccessfulUpnResolution"/> must be called every time the UPN resolution
		/// is successful (but not yet securely made). In that case, <see cref="AttemptIndex"/> is reset so that the client
		/// retries to send the request as soon as possible, and this property is set to true so that the reset isn't performed
		/// again (because if it were, the client would again uselessly waste network resources by resending the request as often
		/// as possible).</para>
		/// </remarks>
		public bool WasResetOnSuccessfulUpnResolution
		{
			get { return _WasResetOnSuccessfulUpnResolution; }
			protected set { _WasResetOnSuccessfulUpnResolution = value; }
		}

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Updates the current access state.
		/// </summary>
		/// <param name="actionClientRetryScheduleAlg">Algorithm which generates the retry timespan for the processing of action
		/// requests by an action client.</param>
		/// <remarks>
		/// <para>This method computes the time (in UTC format) when the next processing attempt of an action request (whose
		/// recipient UPN fingerprint is <see cref="RecipientUpnFingerprint"/>) should to be made.</para>
		/// </remarks>
		public void Update( IActionClientRetryScheduleAlg actionClientRetryScheduleAlg )
		{
			LastAttemptTime = DateTime.UtcNow;
			AttemptIndex++;

			double nextAttemptTimespan = actionClientRetryScheduleAlg.GenerateTimespan( RetryIndex );
			NextAttemptTime = LastAttemptTime.AddSeconds( nextAttemptTimespan );
		}

		/// <summary>
		/// Resets the current access state when <see cref="WasResetOnSuccessfulUpnResolution"/> is false.
		/// </summary>
		/// <remarks>
		/// <para>Call this method only after an attempt to resolve the <see cref="RecipientUpnFingerprint"/> UPN fingerprint is
		/// successfully (but not yet securely) made. For a more detailed discussion, see
		/// <see cref="WasResetOnSuccessfulUpnResolution"/>.</para>
		/// </remarks>
		public void ResetOnSuccessfulUpnResolution()
		{
			// The access state may not be reset if "WasResetOnSuccessfulUpnResolution" is true for reasons explained in the
			// property's comments.
			if( WasResetOnSuccessfulUpnResolution )
				return;

			LastAttemptTime = DateTime.UtcNow;
			AttemptIndex = 0;
			WasResetOnSuccessfulUpnResolution = true;

			NextAttemptTime = LastAttemptTime;
		}

		#endregion Public methods
	}
}
