////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.ActionProcessor;
using Vela.Kernel;

namespace Vela.ActionClient
{
	/// <summary>
	/// Manager for the recipient access states.
	/// </summary>
	/// <remarks>
	/// <para>The <see cref="RecipientAccessStatesManager.States"/> list contains only the states for action progresses
	/// whose processing state is <see cref="ActionProcessingStates.Retry"/>.</para>
	/// 
	/// <para>If an action client may no longer process an action request, it must remove from the list (of states) the state
	/// with the same recipient UPN fingerprint.</para>
	/// 
	/// <para>If the state for a given UPN fingerprint is not found in the list (of states), an action client must attempt to
	/// process an action request with that recipient UPN fingerprint as soon as the retry schedules can possibly allow it.
	/// </para>
	/// 
	/// <para>To exemplify, let's say that Alice has sent several letters to Bob, but Bob's computer was off for a while. The
	/// current wait timespan of the retry schedules has reached 1 hour. Once a retry attempt successfully sends one of the
	/// letters, the rest should be sent as soon as possible, not (each) wait for 1 more hour.</para>
	/// 
	/// <para>An instance of this class is thread safe.</para>
	/// </remarks>
	public class RecipientAccessStatesManager : IRecipientAccessStatesManager
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected RecipientAccessStatesManager()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="configuration">The recipient access states manager configuration.</param>
		public RecipientAccessStatesManager( RecipientAccessStatesManagerConfiguration configuration )
		{
			ValidationUtils.ObjectHasContent( configuration, "configuration" );

			Configuration = configuration;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>		
		/// The recipient access states manager configuration.
		/// </summary>
		private RecipientAccessStatesManagerConfiguration Configuration;

		/// <summary>
		/// The root synchronization object.
		/// </summary>
		private readonly object RootLock = new object();

		/// <summary>
		/// List of recipient access states.
		/// </summary>
		private RecipientAccessStates States = new RecipientAccessStates();

		#endregion Private fields

		#region Private methods

		/// <summary>
		/// Adds or updates a recipient access state with the recipient UPN fingerprint of the specified action progress.
		/// </summary>
		/// <param name="actionProgress">The progress of an action request whose recipient UPN fingerprint is
		/// <see cref="RecipientAccessState.RecipientUpnFingerprint"/>.</param>
		/// <param name="updateExisting">If this is true then the existing recipient access state is updated.</param>
		/// <returns>Returns the added or updated state.</returns>
		/// <remarks>
		/// <para>The <see cref="States"/> property is thread safe within this method.</para>
		/// </remarks>
		private RecipientAccessState AddOrUpdate( ActionClientProgress actionProgress, bool updateExisting )
		{
			ValidationUtils.ObjectHasContent( actionProgress, "actionProgress" );

			lock( RootLock )
			{
				Locator recipientUpnFingerprint = actionProgress.RecipientUpnFingerprint;
				RecipientAccessState existingState = States[ recipientUpnFingerprint ];

				if( existingState == null )
				{
					// The recipient state doesn't exist.

					if( actionProgress.ProcessingState == ActionProcessingStates.Retry )
					{
						existingState = new RecipientAccessState( recipientUpnFingerprint );

						if( updateExisting )
							existingState.Update( Configuration.ActionClientRetryScheduleAlg );

						States.AddOrUpdateWithLimit( existingState, Configuration.RecordsCountMax );
					}
				}
				else
				{
					// The recipient state already exists.

					if( actionProgress.ProcessingState == ActionProcessingStates.Retry )
					{
						if( updateExisting )
							existingState.Update( Configuration.ActionClientRetryScheduleAlg );
					}
					else
					{
						// All the states other than "Retry" are changed into another state when this method is called.

						States.RemoveItem( existingState );
					}
				}

				return existingState;
			}
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Finds a recipient access state with the specified recipient UPN fingerprint.
		/// </summary>
		/// <param name="recipientUpnFingerprint">The recipient UPN fingerprint whose state to find.</param>
		/// <returns>Returns the found state, or null if none is found.</returns>
		/// <remarks>
		/// <para>The <see cref="States"/> property is thread safe within this method.</para>
		/// </remarks>
		public RecipientAccessState Find( Locator recipientUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( recipientUpnFingerprint, "recipientUpnFingerprint" );

			lock( RootLock )
			{
				return States[ recipientUpnFingerprint ];
			}
		}

		/// <summary>
		/// Resets the access state of the specified recipient UPN fingerprint after an attempt to resolve the specified
		/// recipient UPN fingerprint was successfully made.
		/// </summary>
		/// <param name="recipientUpnFingerprint">The recipient UPN fingerprint whose state to reset.</param>
		/// <remarks>
		/// <para>The <see cref="States"/> property is thread safe within this method.</para>
		/// 
		/// <para>Call this method only after an attempt to resolve the recipient UPN fingerprint is successfully (but not yet
		/// securely) made. For a more detailed discussion, see
		/// <see cref="RecipientAccessState.WasResetOnSuccessfulUpnResolution"/>.</para>
		/// </remarks>
		public void ResetOnSuccessfulUpnResolution( Locator recipientUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( recipientUpnFingerprint, "recipientUpnFingerprint" );

			lock( RootLock )
			{
				RecipientAccessState existingState = States[ recipientUpnFingerprint ];
				if( existingState != null )
					existingState.ResetOnSuccessfulUpnResolution();
			}
		}

		/// <summary>
		/// Marks the access state of the specified recipient UPN fingerprint as being accessible, by removing it from the list
		/// of states.
		/// </summary>
		/// <param name="recipientUpnFingerprint">The UPN fingerprint of the recipient whose access state to mark as being
		/// accessible.</param>
		/// <remarks>
		/// <para>The <see cref="States"/> property is thread safe within this method.</para>
		/// 
		/// <para>Call this method only after the recipient action service successfully processes an action request, and the
		/// action client receives a signed action response which indicates the success. For a more detailed discussion, see
		/// <see cref="RecipientAccessState.WasResetOnSuccessfulUpnResolution"/>.</para>
		/// </remarks>
		public void RecipientIsAccessibleAndProcessing( Locator recipientUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( recipientUpnFingerprint, "recipientUpnFingerprint" );

			lock( RootLock )
			{
				RecipientAccessState existingState = States[ recipientUpnFingerprint ];
				if( existingState != null )
					States.RemoveItem( existingState );
			}
		}

		/// <summary>
		/// Marks the access state of the specified recipient UPN fingerprint to be processed as soon as possible, by removing it
		/// from the list of states.
		/// </summary>
		/// <param name="recipientUpnFingerprint">The UPN fingerprint of the recipient whose access state to mark to be
		/// processed.</param>
		/// <remarks>
		/// <para>The <see cref="States"/> property is thread safe within this method.</para>
		/// </remarks>
		public void RetrySendingRequestsAsSoonAsPossible( Locator recipientUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( recipientUpnFingerprint, "recipientUpnFingerprint" );

			lock( RootLock )
			{
				RecipientAccessState existingState = States[ recipientUpnFingerprint ];
				if( existingState != null )
					States.RemoveItem( existingState );
			}
		}

		/// <summary>
		/// Gets information about when the next processing attempt of an action request whose recipient UPN fingerprint is
		/// specified will be made.
		/// </summary>
		/// <param name="recipientUpnFingerprint">The fingerprint of the UPN of the recipient for which to get the information.
		/// </param>
		/// <param name="processingState">The processing state of the action request.</param>
		/// <returns>Returns the information about the next processing attempt, or null if there may / will be no retry.
		/// </returns>
		/// <remarks>
		/// <para>The <see cref="States"/> property is thread safe within this method.</para>
		/// </remarks>
		public ActionNextProcessingAttempt GetNextProcessingAttempt( Locator recipientUpnFingerprint
			, ActionProcessingStates processingState )
		{
			ValidationUtils.ObjectHasContent( recipientUpnFingerprint, "recipientUpnFingerprint" );

			// Don't show the next processing attempt when the request is pending because that could confuse the user.
			if( processingState != ActionProcessingStates.Retry )
				return null;

			lock( RootLock )
			{
				RecipientAccessState accessState = Find( recipientUpnFingerprint );
				if( accessState == null )
					return null;

				DateTime now = DateTime.UtcNow;

				// Necessary for time changes.
				if( now < accessState.LastAttemptTime )
					return new ActionNextProcessingAttempt( now );

				DateTime nextTime = TimeUtils.Max( accessState.NextAttemptTime, now );

				return new ActionNextProcessingAttempt( nextTime );
			}
		}

		/// <summary>
		/// Gets information about when the next processing attempt of an action request whose recipient UPN fingerprint is
		/// <see cref="ActionProgress.RecipientUpnFingerprint"/> will be made.
		/// </summary>
		/// <param name="actionProgress">The action request progress for which to get the information.</param>
		/// <returns>Returns the information about the next processing attempt, or null if there may / will be no retry.
		/// </returns>
		public ActionNextProcessingAttempt GetNextProcessingAttempt( ActionClientProgress actionProgress )
		{
			ValidationUtils.ObjectHasContent( actionProgress, "actionProgress" );

			return GetNextProcessingAttempt( actionProgress.RecipientUpnFingerprint, actionProgress.ProcessingState );
		}

		/// <summary>
		/// Checks if the scheduled <see cref="RecipientAccessState.NextAttemptTime"/> allows the referred action request to be
		/// processed.
		/// </summary>
		/// <param name="actionProgress">The progress of an action request whose recipient UPN fingerprint is
		/// <see cref="RecipientAccessState.RecipientUpnFingerprint"/>.</param>
		/// <returns>Returns true if the schedules allow the referred action request to be processed.</returns>
		/// <remarks>
		/// <para>Call this method <b>before</b> an attempt to process the referred action request is made.</para>
		/// 
		/// <para>The <see cref="States"/> property is thread safe within this method.</para>
		/// </remarks>
		public bool ScheduleAllowsProcessing( ActionClientProgress actionProgress )
		{
			ValidationUtils.ObjectHasContent( actionProgress, "actionProgress" );

			lock( RootLock )
			{
				switch( actionProgress.ProcessingState )
				{
					case ActionProcessingStates.Pending:
						return true;
					case ActionProcessingStates.Retry:
						RecipientAccessState accessState = Find( actionProgress.RecipientUpnFingerprint );
						if( accessState == null )
							return true;

						DateTime now = DateTime.UtcNow;

						return (now >= accessState.NextAttemptTime)
							|| (now < accessState.LastAttemptTime) /* Necessary for time changes. */;
					case ActionProcessingStates.Error:
					case ActionProcessingStates.Sent:
					case ActionProcessingStates.Received:
					case ActionProcessingStates.RecipientImported:
						return false;
					default:
						throw ValidationUtils.Unhandled( actionProgress.ProcessingState, "actionProgress.ProcessingState" );
				}
			}
		}

		/// <summary>
		/// Updates the processing state of the specified action progress to either retry or error, if an exception has occurred.
		/// </summary>
		/// <param name="actionProgress">The progress of an action request whose recipient UPN fingerprint is
		/// <see cref="RecipientAccessState.RecipientUpnFingerprint"/>.</param>
		/// <param name="exception">The exception which has occurred. If this contains a <see cref="FatalCrossServiceException"/>
		/// exception, that is used instead (of this) and the processing state is set to error.</param>
		/// <param name="secondaryProcessingState">Secondary processing state. This may be null.</param>
		/// <remarks>
		/// <para>Since the processing state is updated for the <b>next</b> attempt to process the referred action, don't call
		/// this method <b>before</b> an attempt to process the referred action request is made (if you also verify the
		/// processing state before the attempt to process is made).</para>
		/// 
		/// <para>The <see cref="States"/> property is thread safe within this method.</para>
		/// </remarks>
		public void UpdateProgressToRetryOrError( ActionClientProgress actionProgress, Exception exception
			, SecondaryProcessingState secondaryProcessingState )
		{
			ValidationUtils.ObjectHasContent( actionProgress, "actionProgress" );
			ValidationUtils.ObjectHasContent( exception, "exception" );

			lock( RootLock )
			{
				ActionLastProcessingAttempt lastProcessingAttempt = actionProgress.LastProcessingAttempt;

				bool maySetRetry = (lastProcessingAttempt.ProcessingState == ActionProcessingStates.Pending)
					|| (lastProcessingAttempt.ProcessingState == ActionProcessingStates.Retry);

				Exception fcsException = ExceptionUtils.FindFatalCrossServiceException( exception );

				if( !maySetRetry || (fcsException != null) )
				{
					string exceptionMessage = ExceptionUtils.GetPresenterMessage( null, fcsException, false );

					string errorMessage = string.Format( "No more retries to process the action request may be made.\n\n{0}"
						, exceptionMessage );
					lastProcessingAttempt = new ActionLastProcessingAttempt( ActionProcessingStates.Error, errorMessage
						, secondaryProcessingState );

					errorMessage = string.Format( "No more retries to process the action request with the locator \"{0}\""
						+ " may be made.\n\n{1}", actionProgress.ReferredActionRequestLocator, exceptionMessage );
					LogManager.Add( errorMessage, null, LogRecordModes.Error, true );
				}
				else
				{
					string exceptionMessage = ExceptionUtils.GetPresenterMessage( null, exception, false );

					string errorMessage = string.Format( "The action client has failed to process the action request. A"
						+ " processing attempt will be made later.\n\n{0}", exceptionMessage );
					lastProcessingAttempt = new ActionLastProcessingAttempt( ActionProcessingStates.Retry, errorMessage
						, secondaryProcessingState );
				}

				actionProgress.UpdateLastProcessingAttempt( lastProcessingAttempt );

				AddOrUpdate( actionProgress, true );
			}
		}

		/// <summary>
		/// Updates the processing state of the specified action progress immediately <b>after</b> an attempt to process the
		/// referred action request was made, to either retry or error.
		/// </summary>
		/// <param name="actionProgress">The progress of an action request whose recipient UPN fingerprint is
		/// <see cref="RecipientAccessState.RecipientUpnFingerprint"/>.</param>
		/// <param name="errorResponse">Error action response received while processing the referred action request. This may
		/// be null.</param>
		/// <param name="secondaryProcessingState">Secondary processing state. This may be null.</param>
		/// <remarks>
		/// <para>The processing state is set to error if "errorResponse" is not null.</para>
		/// 
		/// <para>Since the processing state is updated for the <b>next</b> attempt to process the referred action, don't call
		/// this method <b>before</b> an attempt to process the referred action request is made (if you also verify the
		/// processing state before the attempt to process is made).</para>
		/// 
		/// <para>The <see cref="States"/> property is thread safe within this method.</para>
		/// </remarks>
		public void UpdateProgressToRetryOrError( ActionClientProgress actionProgress, ErrorActionResponse errorResponse
			, SecondaryProcessingState secondaryProcessingState )
		{
			ValidationUtils.ObjectHasContent( actionProgress, "actionProgress" );

			lock( RootLock )
			{
				ActionLastProcessingAttempt lastProcessingAttempt = actionProgress.LastProcessingAttempt;

				bool maySetRetry = (lastProcessingAttempt.ProcessingState == ActionProcessingStates.Pending)
					|| (lastProcessingAttempt.ProcessingState == ActionProcessingStates.Retry);

				bool isFatalException = (errorResponse != null) ? errorResponse.HasFatalCrossServiceException() : false;

				string responseError = (errorResponse != null) ? errorResponse.Message : null;

				if( !maySetRetry || isFatalException )
				{
					string errorMessage = string.Format( "No more retries to process the action request may be made.\n\n{0}"
						, responseError );
					lastProcessingAttempt = new ActionLastProcessingAttempt( ActionProcessingStates.Error, errorMessage
						, secondaryProcessingState );

					errorMessage = string.Format( "No more retries to process the action request with the locator \"{0}\""
						+ " may be made.\n\n{1}", actionProgress.ReferredActionRequestLocator, responseError );
					LogManager.Add( errorMessage, null, LogRecordModes.Error, true );
				}
				else
				{
					string errorMessage = string.Format( "The action client has failed to process the action request. A"
						+ " processing attempt will be made later.\n\n{0}", responseError );
					lastProcessingAttempt = new ActionLastProcessingAttempt( ActionProcessingStates.Retry, errorMessage
						, secondaryProcessingState );
				}

				actionProgress.UpdateLastProcessingAttempt( lastProcessingAttempt );

				AddOrUpdate( actionProgress, true );
			}
		}

		/// <summary>
		/// Updates the processing state of the specified action progress immediately <b>after</b> an attempt to process the
		/// referred action request was made.
		/// </summary>
		/// <param name="actionProgress">The progress of an action request whose recipient UPN fingerprint is
		/// <see cref="RecipientAccessState.RecipientUpnFingerprint"/>.</param>
		/// <param name="success">If this is true the referred action request was successfully processed, else there was an
		/// error.</param>
		/// <param name="resetRecipientAccessStateOnSuccess">If this is true then the recipient access state is reset if
		/// "success" is true.</param>
		/// <param name="errorResponse">Error action response received while processing the referred action request. This may
		/// be null.</param>
		/// <param name="setProcessingStateToSentOnSuccess">If "success" is true: if this is true then the processing state is
		/// set with <see cref="ActionProcessingStates.Sent"/>, else with <see cref="ActionProcessingStates.Pending"/>
		/// (because a multiframe action response is currently being processed).</param>
		/// <param name="secondaryProcessingState">Secondary processing state. This may be null.</param>
		/// <remarks>
		/// <para>The processing state is set to error if either "errorResponse" or "lastErrorMessage" have content.</para>
		/// 
		/// <para>Since the processing state is updated for the <b>next</b> attempt to process the referred action, don't call
		/// this method <b>before</b> an attempt to process the referred action request is made if you also verify the processing
		/// state before the attempt to process is made.</para>
		/// 
		/// <para>The <see cref="States"/> property is thread safe within this method.</para>
		/// </remarks>
		public void UpdateProgress( ActionClientProgress actionProgress, bool success, bool resetRecipientAccessStateOnSuccess
			, ErrorActionResponse errorResponse, bool setProcessingStateToSentOnSuccess
			, SecondaryProcessingState secondaryProcessingState )
		{
			ValidationUtils.ObjectHasContent( actionProgress, "actionProgress" );

			if( success && (errorResponse != null) )
			{
				throw new ValidationException( "When updating the processing state of an action request, \"errorResponse\""
					+ " must be null if \"success\" is true." );
			}

			lock( RootLock )
			{
				if( success )
				{
					ActionProcessingStates processingState = setProcessingStateToSentOnSuccess ? ActionProcessingStates.Sent
						: ActionProcessingStates.Pending;

					ActionLastProcessingAttempt lastProcessingAttempt = new ActionLastProcessingAttempt( processingState, null
						, secondaryProcessingState );

					actionProgress.UpdateLastProcessingAttempt( lastProcessingAttempt );

					if( resetRecipientAccessStateOnSuccess )
						RecipientIsAccessibleAndProcessing( actionProgress.RecipientUpnFingerprint );
				}
				else
				{
					UpdateProgressToRetryOrError( actionProgress, errorResponse, secondaryProcessingState );
				}
			}
		}

		#endregion Public methods
	}
}
