////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Kernel;

namespace Vela.ActionClient
{
	/// <summary>
	/// Locker for the resolution of UPNs by an action client.
	/// </summary>
	/// <remarks>
	/// <para>There is no point in trying to resolve a specific UPN multiple times simultaneously, so a lock is used for each
	/// UPN.</para>
	/// 
	/// <para>If a UPN resolution is in progress on a thread, no other thread is allowed to start the resolution of the same UPN
	/// and an exception is thrown, so that a retry is made later.</para>
	/// 
	/// <para>An instance of this class is thread safe.</para>
	/// </remarks>
	public class UpnResolutionLocker
	{
		#region Private fields

		/// <summary>
		/// The synchronization object for the <see cref="Records"/> property. READ-ONLY.
		/// </summary>
		private readonly object RecordsLock = new object();

		#endregion Private fields

		#region Private properties

		private readonly UpnResolutionLockerRecords _Records = new UpnResolutionLockerRecords();
		/// <summary>
		/// The UPN resolution locker records. The backing field is initialized by default. This may be empty, but not null.
		/// </summary>
		private UpnResolutionLockerRecords Records
		{
			get { return _Records; }
		}

		#endregion Private properties

		#region Public methods

		/// <summary>
		/// Marks that the resolution of the specified UPN is about to start. If the resolution for the specified UPN was
		/// already started, an exception is thrown. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="upn">The UPN for which the resolution is about to start.</param>
		/// <remarks>
		/// <para>Call this method just before the resolution of a UPN is about to start.</para>
		/// 
		/// <para>The <see cref="Records"/> property is thread safe within this method.</para>
		/// </remarks>
		public void Starting( Upn upn )
		{
			ValidationUtils.ObjectHasContent( upn, "upn" );

			lock( RecordsLock )
			{
				UpnResolutionLockerRecord record = Records[ upn.Fingerprint ];

				if( record == null )
				{
					// New UPN. Add a record to the list.

					record = new UpnResolutionLockerRecord( upn );

					Records.AddOrUpdate( record );
				}
				else
				{
					throw new UpnResolutionInProgressException( "The resolution of the UPN is already in progress.", null
						, new ExceptionDetail( "UPN", upn.ToString() ) );
				}
			}
		}

		/// <summary>
		/// Marks that the resolution of the specified UPN has ended. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="upn">The UPN for which the resolution is about to start. If the resolution of the UPN was not started,
		/// nothing happens.</param>
		/// <remarks>
		/// <para>Call this method just after the resolution of a UPN has ended.</para>
		/// 
		/// <para>The <see cref="Records"/> property is thread safe within this method.</para>
		/// </remarks>
		public void Ended( Upn upn )
		{
			ValidationUtils.ObjectHasContent( upn, "upn" );

			lock( RecordsLock )
			{
				Records.Remove( upn.Fingerprint );
			}
		}

		#endregion Public methods
	}
}
