////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Algorithm which generates the retry timespan for the processing of action requests by an action client.
	/// </summary>
	public class ActionClientRetryScheduleAlg : IActionClientRetryScheduleAlg
	{
		#region Public constants

		/// <summary>
		/// The minimum timespan (in seconds).
		/// </summary>
		public const int TimespanMin = 30;

		/// <summary>
		/// The maximum timespan (in seconds).
		/// </summary>
		public const int TimespanMax = 6000;

		/// <summary>
		/// The minimum value for the random range for a timespan.
		/// </summary>
		public const double TimespanRandomRangeMin = 0.3;

		/// <summary>
		/// The maximum value for the random range for a timespan.
		/// </summary>
		public const double TimespanRandomRangeMax = 0.8;

		/// <summary>
		/// The cap for the retry index. If the retry index is higher than this value, this value is used instead.
		/// </summary>
		public const int RetryIndexCap = 100;

		/// <summary>
		/// The coverage range of the curve (generated with the <see cref="GenerateCurve"/> method), relative to the inflection
		/// point.
		/// </summary>
		/// <remarks>
		/// <para>This must be higher than 0.</para>
		/// 
		/// <para>This value is also equivalent with the steepness of the curve around the inflection point.</para>
		/// </remarks>
		public const double CurveCoverageRange = 5;

		/// <summary>
		/// The shift of the inflection point of the curve (generated with the <see cref="GenerateCurve"/> method), relative to
		/// the middle of the coverage range.
		/// </summary>
		/// <remarks>
		/// <para>This must be within [-1...+1]. The value 0 corresponds to the inflection point being in the middle of the
		/// coverage range. A negative value corresponds to the inflection point being shifted to the left of the middle of the
		/// coverage range.</para>
		/// </remarks>
		public const double CurveInflectionShift = 0.5;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public ActionClientRetryScheduleAlg()
		{
		}

		#endregion Constructors

		#region Private static methods

		/// <summary>
		/// Generates the Y point for the specified X coordinate.
		/// </summary>
		/// <param name="x">X coordinate.</param>
		/// <returns>Returns the generated the Y point.</returns>
		private static double GenerateCurve( double x )
		{
			double leftCoverageRange = CurveCoverageRange * (1 + CurveInflectionShift) / 2;
			double rightCoverageRange = CurveCoverageRange - leftCoverageRange;

			double shiftedX = CurveCoverageRange * x - leftCoverageRange;

			double yMin = Math.Atan( -leftCoverageRange );
			double yMax = Math.Atan( rightCoverageRange );

			double curve = (Math.Atan( shiftedX ) - yMin) / (yMax - yMin);

			return curve;
		}

		#endregion Private static methods

		#region Public methods

		/// <summary>
		/// Generates the timespan (in seconds) corresponding to the specified retry index.
		/// </summary>
		/// <param name="retryIndex">The index of the last retry to process an action request. If this is a negative value, it
		/// means that no retry has been made to process the action request, so 0 is returned.</param>
		/// <returns>Returns the timespan (in seconds) corresponding to the specified retry index.</returns>
		/// <remarks>
		/// <para>An action client has to wait this timespan between consecutive attempts to send an action request to an
		/// action service.</para>
		/// 
		/// <para>The returned timespan is (partly) random.</para>
		/// 
		/// <para>Generally, the higher the retry index is, the higher the timespan is, but there is no guarantee. For any given
		/// retry index, the generated timespan can be smaller or greater than the timespan generated for the previous retry
		/// index.</para>
		/// 
		/// <para>The generated timespans must be random so that an attacker can't see a pattern in the distribution of
		/// timespans. A pattern would be a sign that this program is being used.</para>
		/// 
		/// <para>Statistical results for the generated timespans:</para>
		/// <list type="bullet">
		///		<item>Retry index[ 0...5 ]: average = 53 s, total = 319 s.</item>
		///		<item>Retry index[ 0...10 ]: average = 78 s, total = 868 s.</item>
		///		<item>Retry index[ 0...20 ]: average = 139 s, total = 2923 s.</item>
		///		<item>Retry index[ 0...30 ]: average = 211 s, total = 6546 s.</item>
		///		<item>Retry index[ 0...40 ]: average = 306 s, total = 12568 s.</item>
		///		<item>Retry index[ 0...50 ]: average = 429 s, total = 21882 s.</item>
		///		<item>Retry index[ 0...60 ]: average = 602 s, total = 36724 s.</item>
		///		<item>Retry index[ 0...70 ]: average = 853 s, total = 60587 s.</item>
		///		<item>Retry index[ 0...80 ]: average = 1193 s, total = 96684 s.</item>
		///		<item>Retry index[ 0...90 ]: average = 1591 s, total = 144808 s.</item>
		///		<item>Retry index[ 0...100 ]: average = 1988 s, total = 200808 s.</item>
		///		<item>Retry index[ 100...100 ]: average = 5826 s, total = 5826 s.</item>
		/// </list>
		/// </remarks>
		public double GenerateTimespan( int retryIndex )
		{
			if( retryIndex < 0 )
				return 0;

			// The minimum is 0, the maximum is 1.
			double x = Math.Min( (double)retryIndex / RetryIndexCap, 1 );

			// The minimum is 0, the maximum is 1.
			double curve = GenerateCurve( x );

			// The minimum is "TimespanMin", the maximum is "TimespanMax".
			double t = TimespanMin + (TimespanMax - TimespanMin) * curve;

			// The minimum is 0, the maximum is 1.
			double r = EntropyPool.GetReal( false );

			// This makes the timespan get more and more random and the retry index increases.
			double timespanRandomRange = TimespanRandomRangeMin + (TimespanRandomRangeMax - TimespanRandomRangeMin) * x;

			double timespan = t * (1 + timespanRandomRange * (2 * r - 1));

			return timespan;
		}

		#endregion Public methods
	}
}
