////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Information about the last processing attempt of an action request, by an action client or service.
	/// </summary>
	public class ActionLastProcessingAttempt : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionLastProcessingAttempt()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="processingState">The processing state of the referred action request.</param>
		/// <param name="errorMessage">The message of the error which has occurred the last attempt when the referred action
		/// request was processed. This may be null or empty.</param>
		/// <param name="secondaryProcessingState">Secondary processing state. This may be null.</param>
		public ActionLastProcessingAttempt( ActionProcessingStates processingState, string errorMessage
			, SecondaryProcessingState secondaryProcessingState )
		{
			ProcessingState = processingState;
			ErrorMessage = errorMessage;
			SecondaryProcessingState = secondaryProcessingState;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// This is true if the processing state (of the referred action request) is
		/// <see cref="ActionProcessingStates.Pending"/>. READ-ONLY.
		/// </summary>
		public bool ProcessingStateIsPending
		{
			get
			{
				return ProcessingState == ActionProcessingStates.Pending;
			}
		}

		/// <summary>
		/// This is true if the processing state (of the referred action request) is <see cref="ActionProcessingStates.Retry"/>
		/// or <see cref="ActionProcessingStates.Error"/>. READ-ONLY.
		/// </summary>
		public bool ProcessingStateIsRetryOrError
		{
			get
			{
				return (ProcessingState == ActionProcessingStates.Retry) || (ProcessingState == ActionProcessingStates.Error);
			}
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private ActionProcessingStates _ProcessingState = ActionProcessingStates.Pending;
		/// <summary>
		/// The processing state of the referred action request. The backing field is initialized by default. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public ActionProcessingStates ProcessingState
		{
			get { return _ProcessingState; }
			protected set
			{
				_ProcessingState = value;
			}
		}

		private string _ErrorMessage;
		/// <summary>
		/// The message of the error which has occurred the last attempt when the referred action request was processed. This may
		/// be null or empty. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string ErrorMessage
		{
			get { return _ErrorMessage; }
			protected set { _ErrorMessage = value; }
		}

		private SecondaryProcessingState _SecondaryProcessingState;
		/// <summary>
		/// Secondary processing state. This may be null. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public SecondaryProcessingState SecondaryProcessingState
		{
			get { return _SecondaryProcessingState; }
			protected set { _SecondaryProcessingState = value; }
		}

		private DateTime _Time = DateTime.UtcNow;
		/// <summary>
		/// The time (in UTC format) when the last attempt to process the referred action request was made. This is automatically
		/// set with the time of the creation of the current object. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public DateTime Time
		{
			get { return _Time; }
			protected set { _Time = value; }
		}

		#endregion Public serializable properties
	}
}
