////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Host context used for updating the user interface of an action client application after an action request and response
	/// are processed.
	/// </summary>
	public class ActionClientHostUpdateUiContext
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionClientHostUpdateUiContext()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="processorConfiguration">The action processor configuration.</param>
		/// <param name="actionProgress">Action progress.</param>
		/// <param name="actionRequest">Action request to send to the service. This may be null.</param>
		/// <param name="updateRequestContext">Context for an action request which is updated (before it's sent to the service
		/// for processing). This may be null.</param>
		/// <param name="actionResponse">Action response received from the service. This may be null.</param>
		/// <param name="exception">Exception due to which the UI has to be updated. This may be null.</param>
		/// <param name="mustFullyUpdateUi">If this is true then the host must fully update its UI. If this is false then the
		/// action client is processing a frame of a multiframe action response, so the host only has to perform a partial UI
		/// update (for example just the progress of a file transfer).</param>
		public ActionClientHostUpdateUiContext( ActionProcessorConfiguration processorConfiguration
			, ActionClientProgress actionProgress, ActionRequest actionRequest, UpdateActionRequestContext updateRequestContext
			, ActionResponse actionResponse, Exception exception, bool mustFullyUpdateUi )
		{
			ProcessorConfiguration = processorConfiguration;
			ActionProgress = actionProgress;
			ActionRequest = actionRequest;
			UpdateRequestContext = updateRequestContext;
			ActionResponse = actionResponse;
			Exception = exception;
			MustFullyUpdateUi = mustFullyUpdateUi;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private ActionProcessorConfiguration _ProcessorConfiguration;
		/// <summary>		
		/// The action processor configuration. READ-ONLY.
		/// </summary>
		public ActionProcessorConfiguration ProcessorConfiguration
		{
			get { return _ProcessorConfiguration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ProcessorConfiguration" );

				_ProcessorConfiguration = value;
			}
		}

		private ActionClientProgress _ActionProgress;
		/// <summary>
		/// Action progress. READ-ONLY.
		/// </summary>
		public ActionClientProgress ActionProgress
		{
			get { return _ActionProgress; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionProgress" );

				_ActionProgress = value;
			}
		}

		private ActionRequest _ActionRequest;
		/// <summary>
		/// Action request to send to the service. This may be null. READ-ONLY.
		/// </summary>
		public ActionRequest ActionRequest
		{
			get { return _ActionRequest; }
			protected set { _ActionRequest = value; }
		}

		private UpdateActionRequestContext _UpdateRequestContext;
		/// <summary>
		/// Context for an action request which is updated (before it's sent to the service for processing). This may be null.
		/// </summary>
		public UpdateActionRequestContext UpdateRequestContext
		{
			get { return _UpdateRequestContext; }
			set { _UpdateRequestContext = value; }
		}

		private ActionResponse _ActionResponse;
		/// <summary>
		/// Action response received from the service. This may be null. READ-ONLY.
		/// </summary>
		public ActionResponse ActionResponse
		{
			get { return _ActionResponse; }
			protected set { _ActionResponse = value; }
		}

		private Exception _Exception;
		/// <summary>
		/// Exception due to which the UI has to be updated. This may be null. READ-ONLY.
		/// </summary>
		public Exception Exception
		{
			get { return _Exception; }
			protected set { _Exception = value; }
		}

		private bool _MustFullyUpdateUi;
		/// <summary>
		/// If this is true then the host must fully update its UI. If this is false then the action client is processing a frame
		/// of a multiframe action response, so the host only has to perform a partial UI update (for example just the progress
		/// of a file transfer). READ-ONLY.
		/// </summary>
		public bool MustFullyUpdateUi
		{
			get { return _MustFullyUpdateUi; }
			protected set { _MustFullyUpdateUi = value; }
		}

		#endregion Public non-serializable properties
	}
}
