////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Cloning;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Contract for an action client manager.
	/// </summary>
	public interface IActionClientManager : IUnclonable
	{
		#region Public non-serializable properties

		/// <summary>
		/// The recipient access states manager. READ-ONLY.
		/// </summary>
		IRecipientAccessStatesManager RecipientAccessStatesManager { get; }

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Adds the specified locator to the list of locators of action requests to process.
		/// </summary>
		/// <param name="requestLocator">The instance locator of the action request to process.</param>
		/// <param name="ignoreScheduler">If this is true then the action request with the specified locator must be processed
		/// even if the retry schedules don't allow it.</param>
		/// <remarks>
		/// <para>If the specified locator is already queued for processing, nothing happens.</para>
		/// </remarks>
		void Add( Locator requestLocator, bool ignoreScheduler );

		/// <summary>
		/// Adds the specified locators to the list of locators of action requests to process.
		/// </summary>
		/// <param name="requestLocators">The instance locators of the action requests to process. This may be empty, but not
		/// null.</param>
		/// <param name="ignoreScheduler">If this is true then the action requests with the specified locators must be processed
		/// even if the retry schedules don't allow it.</param>
		/// <remarks>
		/// <para>If the specified locators are already queued for processing, nothing happens.</para>
		/// </remarks>
		void Add( Locators requestLocators, bool ignoreScheduler );

		/// <summary>
		/// Stops the action client manager as soon as possible.
		/// </summary>
		void Stop();

		/// <summary>
		/// Stops processing the action requests with the specified locators as soon as possible.
		/// </summary>
		/// <param name="requestLocators">The instance locators of the action requests whose processing to stop. If a locator
		/// refers to an action request which is not being processed, nothing happens (with that locator). This may be null or
		/// empty. This must be thread safe.</param>
		void StopRequests( Locators requestLocators );

		#endregion Public methods
	}
}
