////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Progress for the processing of an action request by an action client.
	/// </summary>
	/// <remarks>
	/// <para>A progress document is created (and saved) when an action request is created (and saved), and is updated while
	/// the action request is being processed.</para>
	/// </remarks>
	public class ActionClientProgress : ActionProgress, IExpirableDocument
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="ReferredActionRequestLocator"/> property.
		/// </summary>
		public const string ReferredActionRequestLocatorFilterIdentifier = "ActionClientProgress.ReferredActionRequestLocator#9";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="DocumentExpirationTime"/>
		/// property.
		/// </summary>
		public const string DocumentExpirationTimeFilterIdentifier = "ActionClientProgress.DocumentExpirationTime#10";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionClientProgress()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpnFingerprint">The UPN fingerprint of the recipient of the current object.</param>
		/// <param name="referredSheafLocator">The locator of the referred sheaf (which is referred by the referred
		/// action request).</param>
		/// <param name="referredActionRequestTypeName">The name of the type of the referred action request.</param>
		/// <param name="referredActionRequestLocator">The locator of the referred action request.</param>
		/// <param name="referredActionRequestDocumentExpirationTime">The document expiration time (in UTC format) of the
		/// referred action request. The current progress is valid before this time; from and after this time, it should not be
		/// used in processing.</param>
		/// <param name="progressIsPermanent">If this is true then after an action client processes the referred action request,
		/// the current progress is kept in the storehouse (else it's deleted).</param>
		public ActionClientProgress( Locator instanceLocator, Locator initiatorUpnFingerprint
				, Locator recipientUpnFingerprint, Locator referredSheafLocator, string referredActionRequestTypeName
				, Locator referredActionRequestLocator, DateTime referredActionRequestDocumentExpirationTime
				, bool progressIsPermanent )
			: base( instanceLocator, initiatorUpnFingerprint, recipientUpnFingerprint, referredSheafLocator
				, referredActionRequestTypeName )
		{
			ReferredActionRequestLocator = referredActionRequestLocator;
			DocumentExpirationTime = referredActionRequestDocumentExpirationTime;
			ProgressIsPermanent = progressIsPermanent;

			LastProcessingAttempt = new ActionLastProcessingAttempt( ActionProcessingStates.Pending, null, null );
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( base.ExpirationTime, DocumentExpirationTime ); }
		}

		/// <summary>
		/// The document expiration time (in UTC format) of the referred action request. The current progress is valid before
		/// this time; from and after this time, it should not be used in processing. READ-ONLY.
		/// </summary>
		public DateTime ReferredActionRequestDocumentExpirationTime
		{
			get { return _DocumentExpirationTime; }
			protected set { _DocumentExpirationTime = value; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Locator _ReferredActionRequestLocator;
		/// <summary>
		/// The locator of the referred action request. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( ReferredActionRequestLocatorFilterIdentifier )]
		public Locator ReferredActionRequestLocator
		{
			get { return _ReferredActionRequestLocator; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ReferredActionRequestLocator" );

				_ReferredActionRequestLocator = value;
			}
		}

		private DateTime _DocumentExpirationTime = DateTime.MaxValue;
		/// <summary>
		/// The expiration time (in UTC format) of the document. The current document is valid before this time; from and after
		/// this time, it should not be used in processing. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( DocumentExpirationTimeFilterIdentifier )]
		public DateTime DocumentExpirationTime
		{
			get { return _DocumentExpirationTime; }
			protected set { _DocumentExpirationTime = value; }
		}

		private bool _ProgressIsPermanent;
		/// <summary>
		/// If this is true then after an action client processes the referred action request, the current progress is kept in
		/// the storehouse (else it's deleted). READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public bool ProgressIsPermanent
		{
			get { return _ProgressIsPermanent; }
			protected set { _ProgressIsPermanent = value; }
		}

		#endregion Public serializable properties
	}
}
