////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.PeerDocuments;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Client-side progress for processing a <see cref="CreateAccountAnonymousActionRequest"/> action request.
	/// </summary>
	public class CreateAccountClientProgress : ActionClientProgress, IAlternativePeerAddressContainer
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected CreateAccountClientProgress()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpnFingerprint">The UPN fingerprint of the recipient of the current object.</param>
		/// <param name="referredSheafLocator">The locator of the referred sheaf (which is referred by the referred
		/// action request).</param>
		/// <param name="referredActionRequestTypeName">The name of the type of the referred action request.</param>
		/// <param name="referredActionRequestLocator">The locator of the referred action request.</param>
		/// <param name="referredActionRequestDocumentExpirationTime">The document expiration time (in UTC format) of the
		/// referred action request. The current progress is valid before this time; from and after this time, it should not be
		/// used in processing.</param>
		/// <param name="progressIsPermanent">If this is true then after an action client processes the referred action request,
		/// the current progress is kept in the storehouse (else it's deleted).</param>
		/// <param name="tallyEntityName">The name of the identified entity, from the tally. This may be null or empty.</param>
		/// <param name="approval">The approval of the (tally of the) account to create.</param>
		/// <param name="alternativePeerAddress">The alternative peer address (of the remote peer). This may be null.</param>
		public CreateAccountClientProgress( Locator instanceLocator, Locator initiatorUpnFingerprint
				, Locator recipientUpnFingerprint, Locator referredSheafLocator, string referredActionRequestTypeName
				, Locator referredActionRequestLocator, DateTime referredActionRequestDocumentExpirationTime
				, bool progressIsPermanent, string tallyEntityName
				, TallyApprovals approval, AlternativePeerAddress alternativePeerAddress )
			: base( instanceLocator, initiatorUpnFingerprint, recipientUpnFingerprint, referredSheafLocator
				, referredActionRequestTypeName, referredActionRequestLocator, referredActionRequestDocumentExpirationTime
				, progressIsPermanent )
		{
			TallyEntityName = tallyEntityName;
			Approval = approval;
			AlternativePeerAddress = alternativePeerAddress;
		}

		#endregion Constructors

		#region Public serializable properties

		private string _TallyEntityName;
		/// <summary>
		/// The name of the identified entity, from the tally. This may be null or empty. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If a tally for the remote peer already exists, this is ignored and its entity name remains as is.</para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string TallyEntityName
		{
			get { return _TallyEntityName; }
			protected set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.EntityName, "TallyEntityName" );

				_TallyEntityName = value;
			}
		}

		private TallyApprovals _Approval = PeerDocumentsConsts.CreateAccountApprovalMin;
		/// <summary>
		/// The approval of the (tally of the) account to create. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If a tally for the remote peer already exists, this is ignored and its approval remains as is.</para>
		/// </remarks>
		[SerializeAttribute]
		public TallyApprovals Approval
		{
			get { return _Approval; }
			protected set { _Approval = value; }
		}

		private AlternativePeerAddress _AlternativePeerAddress;
		/// <summary>
		/// The alternative peer address (of the remote peer). This may be null.
		/// </summary>
		/// <remarks>
		/// <para>If a tally for the remote peer already exists, this is ignored if it's null, and the tally's alternative peer
		/// address remains as is.</para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false )]
		public AlternativePeerAddress AlternativePeerAddress
		{
			get { return _AlternativePeerAddress; }
			set { _AlternativePeerAddress = value; }
		}

		#endregion Public serializable properties
	}
}
