////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Recipient access states manager configuration for an action client.
	/// </summary>
	public class RecipientAccessStatesManagerConfiguration : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected RecipientAccessStatesManagerConfiguration()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="actionClientRetryScheduleAlg">The algorithm which generates the retry timespan for the processing of
		/// action requests by an action client.</param>
		public RecipientAccessStatesManagerConfiguration( IActionClientRetryScheduleAlg actionClientRetryScheduleAlg )
		{
			ActionClientRetryScheduleAlg = actionClientRetryScheduleAlg;
		}

		#endregion Constructors

		#region Public serializable properties

		private IActionClientRetryScheduleAlg _ActionClientRetryScheduleAlg;
		/// <summary>		
		/// The algorithm which generates the retry timespan for the processing of action requests by an action client. For
		/// legacy reasons, this may be null, but may not be set with null. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public IActionClientRetryScheduleAlg ActionClientRetryScheduleAlg
		{
			get { return _ActionClientRetryScheduleAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionClientRetryScheduleAlg" );

				_ActionClientRetryScheduleAlg = value;
			}
		}

		private int _RecordsCountMax = ActionProcessorConsts.RecipientAccessStatesManagerRecordsCountMax;
		/// <summary>
		/// The maximum number of records kept by a recipient access states manager. This must be at least 1.
		/// </summary>
		public int RecordsCountMax
		{
			get { return _RecordsCountMax; }
			set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "RecordsCountMax" );

				_RecordsCountMax = value;
			}
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// This method is automatically called after the current object has been deserialized.
		/// </summary>
		protected virtual void Deserialized()
		{
			if( ActionClientRetryScheduleAlg == null )
				ActionClientRetryScheduleAlg = new ActionClientRetryScheduleAlg();
		}

		#endregion Private methods
	}
}
