////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Constants used for an action processor.
	/// </summary>
	public static class ActionProcessorConsts
	{
		#region Public constants

		/// <summary>
		/// The maximum timespan (in seconds) after which an action client automatically loads action requests (from the
		/// storehouse) to process in the processing thread. This must be at least 1.
		/// </summary>
		/// <remarks>
		/// <para>This is only the maximum timespan. The actual timespan is random in order to protect from statistical analysis.
		/// Since the action requests are loaded at random moments, they are also processed at random moments, and their action
		/// responses are sent at random moments.</para>
		/// </remarks>
		public const int ClientStorehouseLoadTimespanMax = 60;

		/// <summary>
		/// The minimum network port where service requests are to be sent by a client. This must be at least 1.
		/// </summary>
		public const UInt16 ServiceProcessingPortMin = 49152;

		/// <summary>
		/// The maximum network port where service requests are to be sent by a client. This must be at least 1.
		/// </summary>
		public const UInt16 ServiceProcessingPortMax = 65535;

		/// <summary>
		/// The default network port where service requests are to be sent by a client. This must be at least 1.
		/// </summary>
		/// <remarks>
		/// <para>The default port is 8676, which is the concatenation of the decimal codes of the "VL" characters). Currently,
		/// the port is unasigned by IANA.</para>
		/// 
		/// <para>IANA (http://www.iana.org/assignments/port-numbers) defines:</para>
		/// <list type="bullet">
		///		<item>The "Well Known Ports" are 0...1023.</item>
		///		<item>The "Registered Ports" are 1024...49151.</item>
		///		<item>The "Dynamic / Private Ports" are 49152...65535.</item>
		/// </list>
		/// </remarks>
		public const UInt16 ServiceProcessingPortDef = 8676;

		/// <summary>
		/// The maximum number of connections which may be queued while waiting for processing. This must be at least 1.
		/// </summary>
		public const int QueuedConnectionsMax = 20;

		/// <summary>
		/// The maximum size (in bytes) of the cache used for sending and receiving data through the network, for large data
		/// transfers. This must be at least 0.
		/// </summary>
		public const int LargeCommunicationCacheSizeMax = 1 << 24;

		/// <summary>
		/// The default size (in bytes) of the cache used for sending and receiving data through the network, for large data
		/// transfers. This must be at least 0.
		/// </summary>
		public const int LargeCommunicationCacheSizeDef = DownloadFilePartSizeMax;

		/// <summary>
		/// The timeout (in seconds) used for sending and receiving data through the network, per blob, per frame. This must be
		/// at least 1.
		/// </summary>
		public const int CommunicationTimeoutPerBlob = 30;

		/// <summary>
		/// The maximum size (in bytes) of a communication frame which a service may process (in a single send or receive). This
		/// must be at least 1.
		/// </summary>
		public const int CommunicationFrameSizeMax = 1 << 20;

		/// <summary>
		/// The maximum number of action requests which a processor may process at the same time. If this is 0 or a negative
		/// number, <see cref="Vela.Kernel.SystemUtils.GetProcessorCoresCount"/> is used instead.
		/// </summary>
		/// <remarks>
		/// <para>For action services, this limit is per service listener.</para>
		/// </remarks>
		public const int ParallelRequestsMax = 10;

		/// <summary>
		/// The maximum number of records kept by a recipient access states manager. This must be at least 1.
		/// </summary>
		public const int RecipientAccessStatesManagerRecordsCountMax = 10000;

		/// <summary>
		/// The maximum number of records kept by an action traffic filter. This must be at least 1.
		/// </summary>
		public const int TrafficFilterRecordsCountMax = 1000000;

		/// <summary>
		/// The timespan (in hours) during which the network traffic of a peer is limited by an action service. This must be at
		/// least 1.
		/// </summary>
		public const int TrafficFilterRecordResetTimespan = 24;

		/// <summary>
		/// The maximum number of network requests which may be accepted by an action service from an anonymous peer, during the
		/// <see cref="TrafficFilterRecordResetTimespan"/> timespan. This must be at least 1.
		/// </summary>
		/// <remarks>
		/// <para>Action requests may be preceded by a request to resolve the UPN, so this may also increase the number of
		/// incoming requests.</para>
		/// 
		/// <para>Some network requests may be cut in the middle, but they are still counted, so you should use a value higher
		/// than 1.</para>
		/// 
		/// <para>This value should be higher than the (maximum) number of retries made by a remote peer to create an account
		/// with the local peer (see the action client retry schedules).</para>
		/// 
		/// <para>The default value is very high so that an attacker can't easily deduce that this program is being used by the
		/// attacked target; the disadvantage is that the attacked target can't be protected from denial of service attacks.
		/// </para>
		/// </remarks>
		public const int TrafficFilterAnonymousPeerIncomingRequestsCountMax = 2000000000;

		/// <summary>
		/// The maximum size (in bytes) of the data which may be processed by an action service for a known (= non-anonymous)
		/// peer, during the <see cref="TrafficFilterRecordResetTimespan"/> timespan. This must be at least
		/// <see cref="CommunicationFrameSizeMax"/>.
		/// </summary>
		/// <remarks>
		/// <para>Keep in mind that a large part of the data coming from known peers may be saved in the storehouse, so don't
		/// set this value too high.</para>
		/// </remarks>
		public const int TrafficFilterKnownPeerProcessableSizeMax = 50000000;

		/// <summary>
		/// The minimum rounded timespan (in seconds) up to which the service must wait as a form of protection against
		/// timing attacks.
		/// </summary>
		public const double TimingAttackProtectionRoundoffMin = 0.010;

		/// <summary>
		/// The maximum rounded timespan (in seconds) up to which the service must wait as a form of protection against
		/// timing attacks.
		/// </summary>
		public const double TimingAttackProtectionRoundoffMax = 1.000;

		/// <summary>
		/// The default rounded timespan (in seconds) up to which the service must wait as a form of protection against
		/// timing attacks.
		/// </summary>
		public const double TimingAttackProtectionRoundoffDef = 0.020;

		/// <summary>
		/// The timespan (in seconds) after which an action service / client notifies the host to update the UI, while processing
		/// an action request which has a multiframe action response.
		/// </summary>
		public const double MultiframeActionResponseUpdateUiTimespan = 2;

		/// <summary>
		/// The minimum size (in bytes) that the part of a file to download must have. This must be at least 1.
		/// </summary>
		/// <remarks>
		/// <para>The file part size matters to some extent. The reason is that when the file download is performed in the local
		/// area network, the processing (= encryption) is the bottleneck, while when the file download is performed over the
		/// Internet, the network speed is the bottleneck.</para>
		/// 
		/// <para>A small file part size is preferred because it's less demanding about the garbage collector.</para>
		/// 
		/// <para>Tests were performed with a small and a large file part, for 100 seconds:</para>
		/// <list type="bullet">
		///		<item>When the file part size was 50'000 bytes, the transfer speed was 3.3 MB / s.</item>
		///		<item>When the file part size was 1'000'000 bytes, the transfer speed was 5.7 MB / s.</item>
		/// </list>
		/// 
		/// <para>While the improvement is significant, it's not relevant because the majority of file downloads are expected to
		/// be performed over the Internet.</para>
		/// 
		/// <para>While the garbage collector has done a good job of maintaining the memory of the application under control in
		/// both cases, when the file part size was large, there were significant swings in the memory footprint, though not huge
		/// ones.</para>
		/// 
		/// <para>Note that the tests included both the client and service applications on the same computer, so the (processing)
		/// performance was basically halved.</para>
		/// </remarks>
		public const int DownloadFilePartSizeMin = 50000;

		/// <summary>
		/// The maximum size (in bytes) that the part of a file to download must have. This must be at least
		/// <see cref="DownloadFilePartSizeMin"/>.
		/// </summary>
		public const int DownloadFilePartSizeMax = 200000;

		/// <summary>
		/// The maximum value for the download progress of files.
		/// </summary>
		public const int DownloadFilesProgressMax = 100;

		#endregion Public constants
	}
}
