////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.ComponentModel;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// The available states for an action service manager (<see cref="IActionServiceManager"/>).
	/// </summary>
	public enum ActionServiceManagerStates
	{
		/// <summary>
		/// All the UPNs of all the service vaults are registered, and all the listeners are listening for client requests.
		/// </summary>
		AllRegisteredAndListening,

		/// <summary>
		/// At least 1 UPN of a service vault is not registered, or one listener is not listening for client requests.
		/// </summary>
		OneNotRegisteredOrListening,

		/// <summary>
		/// The state can't be determined at the moment, try again later.
		/// </summary>
		Unknown
	}

	/// <summary>
	/// The available states of processing of an action request (<see cref="ActionRequest"/>).
	/// </summary>
	public enum ActionProcessingStates
	{
		/// <summary>
		/// The action request will be sent to the service, but no attempt has yet been made (to send it).
		/// </summary>
		Pending,

		/// <summary>
		/// At attempt was made to send the action request to the service but it has failed. A new attempt will be made later.
		/// </summary>
		Retry,

		/// <summary>
		/// Attempts were made to send the action request to the service but they have all failed. No more attempts will be made.
		/// </summary>
		Error,

		/// <summary>
		/// The action request was successfully sent to the service (and processed) and an action response was successfully
		/// received by the client (and processed).
		/// </summary>
		Sent,

		/// <summary>
		/// The action request was successfully received (and processed). It's possible that the action response was NOT
		/// successfully received by the client.
		/// </summary>
		Received,

		/// <summary>
		/// The action request or the document which it refers was imported. This is meant to be used only to import documents
		/// whose recipient is the local peer, that is, when the local peer has the vault which can decrypt the envelope of the
		/// document; this should be treated similar to <see cref="Received"/>.
		/// </summary>
		[Description( "Imported" )]
		RecipientImported
	}

	/// <summary>
	/// The available directions of the local peer.
	/// </summary>
	public enum LocalPeerDirections
	{
		/// <summary>
		/// The local peer is the initiator (of the sheaf).
		/// </summary>
		Initiator,

		/// <summary>
		/// The local peer is the recipient (of the sheaf).
		/// </summary>
		Recipient,

		/// <summary>
		/// The sheaf was imported. This is meant to be used only to import documents whose recipient is the local peer, that is,
		/// when the local peer has the vault which can decrypt the envelope of the document; this has to be treated similar to
		/// <see cref="Recipient"/>.
		/// </summary>
		[Description( "Imported" )]
		RecipientImported
	}
}
