////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Utilities for working with the storehouse.
	/// </summary>
	public static class ActionProcessorStorehouseUtils
	{
		#region Public static methods

		/// <summary>
		/// Checks if there is a vault with the specified UPN.
		/// </summary>
		/// <param name="transaction">The transaction with which to check if a vault exists.</param>
		/// <param name="upn">The UPN of the vault to check if it exists.</param>
		/// <returns>Returns true if a vault exists.</returns>
		public static bool HasVaultWithUpn( IStorehouseTransaction transaction, Upn upn )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( upn, "upn" );

			SpeckleFilter vaultUpnFilter = new SpeckleFilter( Vault.LocalUpnFingerprintFilterIdentifier, upn.Fingerprint );

			Speckle vaultSpeckle = transaction.CollectFirstSpeckle( PortableTypes.FindAllDerived( typeof( Vault ) )
				, new SpeckleFilters( vaultUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared, null
				, null );

			return vaultSpeckle != null;
		}

		/// <summary>
		/// Gets the presenter of the specified vault.
		/// </summary>
		/// <param name="transaction">The transaction with which to load the documents.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="vault">Vault for which to get the presenter.</param>
		/// <param name="tallyApproval">The approval of the tally of the initiator of the action request to process.</param>
		/// <param name="hashSerializer">Hash serializer.</param>
		/// <returns>Returns the presenter of the specified vault.</returns>
		public static Presenter GetPresenter( IStorehouseTransaction transaction, MacAlg macAlg
			, IAsymmSigningPaddingAlg asymmSigningPaddingAlg, Vault vault, TallyApprovals tallyApproval
			, IHashSerializer hashSerializer )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( vault, "vault" );
			ValidationUtils.ObjectHasContent( hashSerializer, "hashSerializer" );

			if( PeerDocumentsUtils.IsBlockedPeer( tallyApproval ) )
			{
				throw new FatalCrossServiceException( "Can't get the presenter for the vault because the approval indicates that"
						+ " the remote peer is blocked.", null
					, new ExceptionDetail( "Vault locator", vault.InstanceLocator ) );
			}

			// Include the photo only if allowed.
			Photo photo = null;
			if( tallyApproval >= PeerDocumentsConsts.SendPhotoTallyApprovalMin )
			{
				// Load the photo.
				IDeserializer deserializer;
				SpeckleFilter localUpnFilter = new SpeckleFilter( Photo.CreatorUpnFingerprintFilterIdentifier
					, vault.LocalUpnFingerprint );
				photo = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Photo ) )
					, new SpeckleFilters( localUpnFilter ), StorehouseRealms.Photos, SheafLocks.Shared
					, out deserializer ) as Photo;
			}

			// Include the intimate information only if allowed.
			bool sendIntimateInfo = tallyApproval >= PeerDocumentsConsts.SendIntimateInfoTallyApprovalMin;

			return vault.GetPresenter( CryptoUtils.GetLocator(), macAlg, asymmSigningPaddingAlg, photo, sendIntimateInfo
				, hashSerializer );
		}

		/// <summary>
		/// Creates or updates an account using the specified presenter.
		/// </summary>
		/// <param name="transaction">Transaction to use.</param>
		/// <param name="vaultLocator">The locator of the vault to which to link the account.</param>
		/// <param name="presenter">Presenter from which to create an account.</param>
		/// <param name="tallyEntityName">The name of the identified entity, from the tally. If a tally for the remote peer
		/// already exists, this is ignored and its tally entity name remains as is. This may be null or empty.</param>
		/// <param name="approval">The approval of the (tally of the) account to create. If a tally for the remote peer already
		/// exists, this is ignored and its approval remains as is.</param>
		/// <param name="alternativePeerAddress">The alternative peer address (of the remote peer). If a tally for the remote
		/// peer already exists, this is ignored if it's null, and the tally's alternative peer address remains as is. This may
		/// be null.</param>
		/// <returns>Returns the instance locator of the created account.</returns>
		/// <remarks>
		/// <para>The specified presenter must be signed by the processing and communication identities that it contains.
		/// </para>
		/// 
		/// <para>In order to be able to update an existing presenter, the new presenter must be signed by the
		/// processing and communication identities of the existing presenter.</para>
		/// 
		/// <para>Expired documents are accepted in order to allow the creation of the account. However, further operations
		/// (with the documents of the account) will verify the expiration time.</para>
		/// </remarks>
		public static Locator AddAccountFromPresenter( IStorehouseTransaction transaction, Locator vaultLocator
			, Presenter presenter, string tallyEntityName, TallyApprovals approval
			, AlternativePeerAddress alternativePeerAddress )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( vaultLocator, "vaultLocator" );
			ValidationUtils.ObjectHasContent( presenter, "presenter" );
			ValidationUtils.Text( tallyEntityName, -2, -1, TextValidations.EntityName, "tallyEntityName" );

			// Verify the presenter self-signature.
			presenter.VerifySelfSignature();

			// Check if the UPN is local.
			foreach( SolitaryIdentityPublisher identity in presenter.ProcessingIdentity.Identities )
			{
				if( HasVaultWithUpn( transaction, identity.Identity.Upn ) )
				{
					throw new ValidationException( "Can't add an account because a vault with the same UPN exists.", null, null
						, new ExceptionDetail( "UPN", identity.Identity.Upn.ToString() ) );
				}
			}

			IDeserializer deserializer;

			// Load the vault.
			Vault vault = transaction.LoadSheaf( vaultLocator, StorehouseRealms.ProcessingIdentityContainers, typeof( Vault )
				, SheafLocks.Shared, out deserializer ) as Vault;
			if( vault == null )
			{
				throw new ValidationException( "The vault is missing.", null, null
					, new ExceptionDetail( "Vault locator", vaultLocator ) );
			}

			Locator remoteUpnFingerprint = presenter.CreatorUpnFingerprint;

			#region Check if the documents already exist

			// Check if the remote UPN is a local one as well.
			if( vault.LocalUpnFingerprint.Equals( remoteUpnFingerprint ) )
			{
				throw new ValidationException( "The UPN fingerprint is both local and remote, so you are probably trying to"
						+ " create an account between a presenter and the vault from which the presenter was exported."
					, null, null, new ExceptionDetail( "UPN fingerprint", remoteUpnFingerprint ) );
			}

			// Check if the new presenter was already imported.
			SheafExistenceModes presenterExistence = transaction.SheafExists( presenter
				, StorehouseRealms.ProcessingIdentityContainers );
			switch( presenterExistence )
			{
				case SheafExistenceModes.DoesntExist:
					// Nothing to do.
					break;
				case SheafExistenceModes.IsIdentical:
					throw new FatalCrossServiceException( string.Format( "The presenter with the locator \"{0}\" already exists,"
						+ " so it may have already been used to create an account.", presenter.InstanceLocator ) );
				case SheafExistenceModes.IsDifferent:
					throw new FatalCrossServiceException( "A different sheaf with the presenter's locator already exists.", null
						, null, new ExceptionDetail( "Presenter locator", presenter.InstanceLocator ) );
				default:
					throw ValidationUtils.Unhandled( presenterExistence, "presenterExistence" );
			}

			// Check if the remote UPN already exists.
			Speckle remoteUpnSpeckle = transaction.LoadSpeckle( remoteUpnFingerprint
				, StorehouseRealms.UpnFingerprintReservations, SheafLocks.Shared, null, null );

			#region Check if the presenter already exists

			SpeckleFilter presenterUpnFilter = new SpeckleFilter( Presenter.CreatorUpnFingerprintFilterIdentifier
				, remoteUpnFingerprint );
			Presenter existingPresenter = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Presenter ) )
				, new SpeckleFilters( presenterUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Update
				, out deserializer ) as Presenter;

			// Verify that the new presenter belongs to the same entity as the existing presenter.
			if( existingPresenter != null )
				presenter.VerifySignature( existingPresenter.ProcessingIdentity );

			#endregion Check if the presenter already exists

			#region Check if the photo already exists

			SpeckleFilter photoUpnFilter = new SpeckleFilter( Photo.CreatorUpnFingerprintFilterIdentifier
				, remoteUpnFingerprint );
			Speckle existingPhotoSpeckle = transaction.CollectFirstSpeckle( PortableTypes.FindAllDerived( typeof( Photo ) )
				, new SpeckleFilters( photoUpnFilter ), StorehouseRealms.Photos, SheafLocks.Update, null, null );

			Locator photoLocator = (existingPhotoSpeckle == null) ? CryptoUtils.GetLocator()
				: existingPhotoSpeckle.StorehouseKey;

			#endregion Check if the photo already exists

			// Check if the tally already exists.
			SpeckleFilter tallyUpnFilter = new SpeckleFilter( Tally.RemoteUpnFingerprintFilterIdentifier
				, remoteUpnFingerprint );
			Speckle existingTallySpeckle = transaction.CollectFirstSpeckle( PortableTypes.FindAllDerived( typeof( Tally ) )
				, new SpeckleFilters( tallyUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Update, null
				, new SpeckleFilterIdentifiers( Tally.EntityNameFilterIdentifier ) ) as Speckle;

			#region Check if the account already exists

			SpeckleFilter accountLocalUpnFilter = new SpeckleFilter( Account.LocalUpnFingerprintFilterIdentifier
				, vault.LocalUpnFingerprint );
			SpeckleFilter accountRemoteUpnFilter = new SpeckleFilter( Account.RemoteUpnFingerprintFilterIdentifier
				, remoteUpnFingerprint );

			Speckle existingAccountSpeckle = transaction.CollectFirstSpeckle( PortableTypes.FindAllDerived( typeof( Account ) )
				, new SpeckleFilters( accountLocalUpnFilter, accountRemoteUpnFilter ), StorehouseRealms.Accounts
				, SheafLocks.Update, null, null ) as Speckle;

			Locator accountLocator = (existingAccountSpeckle == null) ? CryptoUtils.GetLocator()
				: existingAccountSpeckle.StorehouseKey;

			#endregion Check if the account already exists

			#endregion Check if the documents already exist

			#region Save documents

			// Reserve (a sheaf with) the UPN fingerprint.
			if( remoteUpnSpeckle == null )
			{
				transaction.AddSpeckle( remoteUpnFingerprint, StorehouseRealms.UpnFingerprintReservations
					, typeof( ReservedUpnFingerprintSheaf ) );
			}

			// Add the presenter. The instance locator is different, so add the new presenter and delete the old one.
			transaction.AddSheaf( presenter, StorehouseRealms.ProcessingIdentityContainers );
			if( existingPresenter != null )
				transaction.DeleteSheaf( existingPresenter.InstanceLocator, StorehouseRealms.ProcessingIdentityContainers );

			// Add or update the photo. If the new presenter has no photo, keep the existing one.
			if( (presenter.Photo != null) && (presenter.Photo.Image != null) )
			{
				Photo photo = new Photo( photoLocator, remoteUpnFingerprint, presenter.Photo.Image );
				if( existingPhotoSpeckle == null )
				{
					// Add the photo.
					transaction.AddSheaf( photo, StorehouseRealms.Photos );
				}
				else
				{
					// Update the photo.
					photo.PrepareToUpdate( existingPhotoSpeckle );
					transaction.UpdateSheaf( photo, StorehouseRealms.Photos );
				}
			}

			string newTallyEntityName = (existingTallySpeckle != null)
				? existingTallySpeckle.GetLoadedSpeckleFilter<string>( Tally.EntityNameFilterIdentifier )
				: (TextUtils.HasContent( tallyEntityName ) ? tallyEntityName : presenter.EntityName);

			// Add or update the tally.
			if( existingTallySpeckle == null )
			{
				// Add the tally.

				Tally tally = new Tally( CryptoUtils.GetLocator(), presenter.ProcessingIdentity
					, presenter.Appendix.ActionRequestValidity, newTallyEntityName );

				// Set the information.
				tally.Approval = approval;
				tally.AlternativePeerAddress = (alternativePeerAddress != null) ? alternativePeerAddress
					: new AlternativePeerAddress();

				transaction.AddSheaf( tally, StorehouseRealms.ProcessingIdentityContainers );
			}
			else
			{
				// Update the tally.

				Tally tally = transaction.LoadSheaf( existingTallySpeckle.StorehouseKey
					, StorehouseRealms.ProcessingIdentityContainers, typeof( Tally ), SheafLocks.Update
					, out deserializer ) as Tally;

				// Update the information which could have changed. Most of the information which is (also) set locally remains
				// as is. The processing identity was used to check for the signature of the new presenter, so it remains as is.
				tally.ActionRequestValidity = presenter.Appendix.ActionRequestValidity;
				tally.AlternativePeerAddress.UpdateOverridingData( alternativePeerAddress );

				transaction.UpdateSheaf( tally, StorehouseRealms.ProcessingIdentityContainers );
			}

			// Add the account.
			if( existingAccountSpeckle == null )
			{
				AccountAlias accountAlias = AccountAlias.FromLocalAndRemoteEntityNames( vault.EntityName, newTallyEntityName );
				Account account = new Account( accountLocator, vault.LocalUpn, presenter.CreatorUpn, accountAlias );

				transaction.AddSheaf( account, StorehouseRealms.Accounts );
			}

			#endregion Save documents

			return accountLocator;
		}

		/// <summary>
		/// Creates or updates an account using the specified presenter.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="vaultLocator">The locator of the vault to which to link the account.</param>
		/// <param name="presenter">Presenter from which to create an account.</param>
		/// <param name="tallyEntityName">The name of the identified entity, from the tally. If a tally for the remote peer
		/// already exists, this is ignored and its tally entity name remains as is. This may be null or empty.</param>
		/// <param name="approval">The approval of the (tally of the) account to create.</param>
		/// <param name="alternativePeerAddress">The alternative peer address (of the remote peer). This may be null.</param>
		/// <returns>Returns the instance locator of the created account.</returns>
		/// <remarks>
		/// <para>The specified presenter must be signed by the processing and communication identities that it contains.
		/// </para>
		/// 
		/// <para>In order to be able to update an existing presenter, the new presenter must be signed by the
		/// processing and communication identities of the existing presenter.</para>
		/// 
		/// <para>Expired documents are accepted in order to allow the creation of the account. However, further operations
		/// (with the documents of the account) will verify the expiration time.</para>
		/// </remarks>
		public static Locator AddAccountFromPresenter( IStorehouseProvider storehouseProvider, Locator vaultLocator
			, Presenter presenter, string tallyEntityName, TallyApprovals approval
			, AlternativePeerAddress alternativePeerAddress )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( vaultLocator, "vaultLocator" );
			ValidationUtils.ObjectHasContent( presenter, "presenter" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				Locator accountLocator = AddAccountFromPresenter( transaction, vaultLocator, presenter, tallyEntityName, approval
					, alternativePeerAddress );

				// Commit the transaction.
				transaction.Commit();

				return accountLocator;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the ephemerid with the specified initiator UPN fingerprint.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="remoteUpnFingerprint">The remote UPN fingerprint for which to load the ephemerid.</param>
		/// <returns>Returns the loaded ephemerid. This may be null.</returns>
		/// <remarks>
		/// <para>The loading is performed on a separate transaction so as to not affect the scalability of the action
		/// processor.</para>
		/// </remarks>
		public static Ephemerid LoadEphemerid( IStorehouseProvider storehouseProvider, Locator remoteUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( remoteUpnFingerprint, "initiatorUpnFingerprint" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				IDeserializer deserializer;

				// Load the ephemerid.
				SpeckleFilter remoteUpnFilter = new SpeckleFilter( Ephemerid.RemoteUpnFingerprintFilterIdentifier
					, remoteUpnFingerprint );
				Ephemerid ephemerid = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Ephemerid ) )
					, new SpeckleFilters( remoteUpnFilter ), StorehouseRealms.Ephemerids, SheafLocks.Shared
					, out deserializer ) as Ephemerid;

				// Commit the transaction.
				transaction.Commit();

				return ephemerid;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Updates the ephemerid with the specified initiator UPN fingerprint.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="remoteUpnFingerprint">The remote UPN fingerprint for which to update the ephemerid.</param>
		/// <param name="ipAddress">The (last known) IP addresses of the remote peer. This may be null or empty.</param>
		/// <param name="actionRequest">Action request which is currently being processed.</param>
		/// <remarks>
		/// <para>The update is performed on a separate transaction so as to not affect the scalability of the action
		/// processor.</para>
		/// 
		/// <para>The update is performed only if something has changed in the ephemerid. If the update is performed often,
		/// it may reduce the scalability of the action processor.</para>
		/// 
		/// <para>The update is performed only if the specified action request has an initiator, or requires mutual knowledge.
		/// This is done in order to minimize the cases when an ephemerid would remain in the storehouse as an orphan sheaf.
		/// </para>
		/// 
		/// <para>Orphan ephemerids would occur, for example, if an action client were to try to create an account associated
		/// to a remote peer, but fails and abandons any new attempt. This possibility exists because saving the ephemerids in
		/// the storehouse before ensuring that an account exists or is created, helps greatly for future attempts to resolve
		/// the remote UPN (since the IP address is already in the storehouse).</para>
		/// </remarks>
		public static void UpdateEphemerid( IStorehouseProvider storehouseProvider, Locator remoteUpnFingerprint
			, byte[] ipAddress, ActionRequest actionRequest )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( remoteUpnFingerprint, "initiatorUpnFingerprint" );
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );

			if( (actionRequest as ActionRequestWithInitiator == null)
				&& (actionRequest as AnonymousActionRequestWithMutualKnowledge == null) )
			{
				return;
			}

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				IDeserializer deserializer;

				// Load the ephemerid.
				SpeckleFilter ephemeridFilter = new SpeckleFilter( Ephemerid.RemoteUpnFingerprintFilterIdentifier
					, remoteUpnFingerprint );
				Ephemerid ephemerid = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Ephemerid ) )
					, new SpeckleFilters( ephemeridFilter ), StorehouseRealms.Ephemerids, SheafLocks.Update
					, out deserializer ) as Ephemerid;

				bool newEphemerid = ephemerid == null;
				if( newEphemerid )
					ephemerid = new Ephemerid( CryptoUtils.GetLocator(), remoteUpnFingerprint );

				PeerIpAddress oldIpAddress = ephemerid.IpAddress;

				// Update the information.
				if( ArrayUtils.HasContent( ipAddress ) )
					ephemerid.IpAddress = new PeerIpAddress( ipAddress );

				// Save the ephemerid.
				if( !PeerIpAddress.Equals( oldIpAddress, ephemerid.IpAddress ) )
				{
					if( newEphemerid )
						transaction.AddSheaf( ephemerid, StorehouseRealms.Ephemerids );
					else
						transaction.UpdateSheaf( ephemerid, StorehouseRealms.Ephemerids );
				}

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Updates the tally with the specified instance locator if the specified exception is derived from
		/// <see cref="RevokedVaultException"/>.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="actionRequest">The locator of the action request which is currently being processed.</param>
		/// <param name="exception">Exception to check if it's derived from <see cref="RevokedVaultException"/>.</param>
		/// <returns>Returns true if the tally is updated.</returns>
		public static bool UpdateTallyIfRevokedServiceVault( IStorehouseProvider storehouseProvider, ActionRequest actionRequest
			, Exception exception )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( exception, "exception" );

			if( !ExceptionUtils.HasException<RevokedVaultException>( exception ) )
				return false;

			ActionRequestWithInitiator actionRequestWithInitiator = actionRequest as ActionRequestWithInitiator;
			if( actionRequestWithInitiator == null )
				return false;

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				#region Load the tally

				Locator remoteUpnFingerprint = actionRequestWithInitiator.RecipientUpnFingerprint;

				IDeserializer deserializer;

				SpeckleFilter remoteUpnFilter = new SpeckleFilter( Tally.RemoteUpnFingerprintFilterIdentifier
					, remoteUpnFingerprint );
				Tally tally = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Tally ) )
					, new SpeckleFilters( remoteUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Update
					, out deserializer ) as Tally;
				ValidationUtils.ObjectHasContent( tally, "tally" );

				#endregion Load the tally

				// Don't process expired action requests.
				KernelUtils.HasExpired( actionRequest, tally.ActionRequestValidity, true );

				// Update the tally approval.
				tally.Approval = TallyApprovals.Revoked;

				// Update the tally.
				transaction.UpdateSheaf( tally, StorehouseRealms.ProcessingIdentityContainers );

				// Commit the transaction.
				transaction.Commit();

				return true;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		#endregion Public static methods
	}
}
