////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Kernel;
using Vela.Storehouse;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Action processor utilities.
	/// </summary>
	public static class ActionProcessorUtils
	{
		#region Public static methods

		/// <summary>
		/// Gets the direction of the local peer, for a sheaf with the specified processing state.
		/// </summary>
		/// <param name="processingState">The processing state of the sheaf for which to get the peer direction.</param>
		/// <returns>Returns the direction of the local peer.</returns>
		public static LocalPeerDirections GetLocalPeerDirection( ActionProcessingStates processingState )
		{
			if( processingState == ActionProcessingStates.Pending )
				return LocalPeerDirections.Initiator;
			else if( processingState == ActionProcessingStates.Retry )
				return LocalPeerDirections.Initiator;
			else if( processingState == ActionProcessingStates.Error )
				return LocalPeerDirections.Initiator;
			else if( processingState == ActionProcessingStates.Sent )
				return LocalPeerDirections.Initiator;
			else if( processingState == ActionProcessingStates.Received )
				return LocalPeerDirections.Recipient;
			else if( processingState == ActionProcessingStates.RecipientImported )
				return LocalPeerDirections.RecipientImported;

			throw ValidationUtils.Unhandled( processingState, "processingState" );
		}

		/// <summary>
		/// Checks if the specified port is already used in "configurations" as a service processing port.
		/// </summary>
		/// <param name="port">Port to check if it's used. This must be at least 1.</param>
		/// <param name="configurations">Action service configurations whose ports to avoid. This may be empty, but not null.
		/// </param>
		/// <returns>Returns true if the port is used.</returns>
		public static bool IsUsedServiceProcessingPort( UInt16 port, ActionServiceListenerConfigurations configurations )
		{
			ValidationUtils.Integer( port, 1, UInt16.MaxValue, "port" );
			ValidationUtils.ObjectHasContent( configurations, "configurations" );

			return configurations[ port ] != null;
		}

		/// <summary>
		/// Generates a random service processing port (to use in <see cref="ActionServiceConfiguration"/>).
		/// </summary>
		/// <returns>Returns the randomly generated port.</returns>
		public static UInt16 GetRandomServiceProcessingPort()
		{
			return (UInt16)EntropyPool.GetInt( ActionProcessorConsts.ServiceProcessingPortMin
				, ActionProcessorConsts.ServiceProcessingPortMax );
		}

		/// <summary>
		/// Generates a random service processing port (to use in <see cref="ActionServiceConfiguration"/>) which is not
		/// already used in "configurations".
		/// </summary>
		/// <param name="configurations">Action service configurations whose ports to avoid. This may be empty, but not null.
		/// </param>
		/// <returns>Returns the randomly generated port.</returns>
		public static UInt16 GetRandomServiceProcessingPort( ActionServiceListenerConfigurations configurations )
		{
			ValidationUtils.ObjectHasContent( configurations, "configurations" );

			while( true )
			{
				UInt16 port = GetRandomServiceProcessingPort();

				// If the port is not already used, return.
				if( configurations[ port ] == null )
					return port;
			}
		}

		/// <summary>
		/// Generates a random service processing port (to use in <see cref="ActionServiceConfiguration"/>) which is not
		/// already used in "configuration".
		/// </summary>
		/// <param name="configuration">Action service listener configuration whose port to avoid.</param>
		/// <returns>Returns the randomly generated port.</returns>
		public static UInt16 GetRandomServiceProcessingPort( ActionServiceListenerConfiguration configuration )
		{
			ValidationUtils.ObjectHasContent( configuration, "configuration" );

			while( true )
			{
				UInt16 port = GetRandomServiceProcessingPort();

				// If the port is not already used, return.
				if( port != configuration.ProcessingPort )
					return port;
			}
		}

		/// <summary>
		/// Loads the configuration from a storehouse. An exception is thrown if the configuration is missing.
		/// </summary>
		/// <param name="transaction">The transaction with which to load the configuration.</param>
		/// <returns>Returns the loaded configuration.</returns>
		/// <remarks>
		/// <para>During the execution of this method, the configuration is locked with <see cref="SheafLocks.Shared"/>.</para>
		/// </remarks>
		public static ServiceListenersStorehouseConfiguration LoadStorehouseConfiguration( IStorehouseTransaction transaction )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );

			ServiceListenersStorehouseConfiguration storehouseConfiguration = transaction.LoadConfiguration( SheafLocks.Shared )
				as ServiceListenersStorehouseConfiguration;
			if( storehouseConfiguration == null )
			{
				throw new ValidationException( string.Format( "A storehouse configuration used by an action processor"
					+ " must be derived from \"{0}\".", typeof( ServiceListenersStorehouseConfiguration ).FullName ) );
			}

			return storehouseConfiguration;
		}

		/// <summary>
		/// Loads the configuration of the specified storehouse. An exception is thrown if the configuration is missing.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <returns>Returns the loaded configuration.</returns>
		/// <remarks>
		/// <para>During the execution of this method, the configuration is locked with <see cref="SheafLocks.Shared"/>.</para>
		/// 
		/// <para>Because a transaction is created in here, if you call this method while another transaction is active, it's
		/// possible that a lock timeout will occur.</para>
		/// </remarks>
		public static ServiceListenersStorehouseConfiguration LoadStorehouseConfiguration(
			IStorehouseProvider storehouseProvider )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the configuration.
				ServiceListenersStorehouseConfiguration configuration = LoadStorehouseConfiguration( transaction );

				// Commit the transaction.
				transaction.Commit();

				return configuration;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the action service listener configurations from a storehouse.
		/// </summary>
		/// <param name="transaction">The transaction with which to load the configurations.</param>
		/// <returns>Returns the loaded configuration.</returns>
		/// <remarks>
		/// <para>During the execution of this method, the configuration is locked with <see cref="SheafLocks.Shared"/>.</para>
		/// </remarks>
		public static ActionServiceListenerConfigurations LoadServiceListenerConfigurations( IStorehouseTransaction transaction )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );

			ServiceListenersStorehouseConfiguration storehouseConfiguration = transaction.LoadConfiguration( SheafLocks.Shared )
				as ServiceListenersStorehouseConfiguration;
			if( storehouseConfiguration == null )
			{
				throw new ValidationException( string.Format( "A storehouse configuration used by an action service must be"
					+ " derived from \"{0}\".", typeof( ServiceListenersStorehouseConfiguration ).FullName ) );
			}

			return storehouseConfiguration.ActionServiceListeners;
		}

		/// <summary>
		/// Loads the action service configurations from a storehouse.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <returns>Returns the loaded configuration.</returns>
		/// <remarks>
		/// <para>During the execution of this method, the configuration is locked with <see cref="SheafLocks.Shared"/>.</para>
		/// 
		/// <para>Because a transaction is created in here, if you call this method while another transaction is active, it's
		/// possible that a lock timeout will occur.</para>
		/// </remarks>
		public static ActionServiceListenerConfigurations LoadServiceConfigurations( IStorehouseProvider storehouseProvider )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				ActionServiceListenerConfigurations configurations = LoadServiceListenerConfigurations( transaction );

				// Commit the transaction.
				transaction.Commit();

				return configurations;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		#endregion Public static methods
	}
}
