////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Contract for a progress for the processing of an action request by an action client or service.
	/// </summary>
	/// <remarks>
	/// <para>A progress document is created (and saved) when an action request is created (and saved) by a client, or may be
	/// created (and saved) when an action request is successfully processed by an action service.</para>
	/// </remarks>
	public abstract class ActionProgress : Sheaf, IInitiatorUpnFingerprintContainer, IRecipientUpnFingerprintContainer
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="InitiatorUpnFingerprint"/> property.
		/// </summary>
		public const string InitiatorUpnFingerprintFilterIdentifier = "ActionProgress.InitiatorUpnFingerprint#4";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="RecipientUpnFingerprint"/> property.
		/// </summary>
		public const string RecipientUpnFingerprintFilterIdentifier = "ActionProgress.RecipientUpnFingerprint#5";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="ReferredSheafLocator"/> property.
		/// </summary>
		public const string ReferredSheafLocatorFilterIdentifier = "ActionProgress.ReferredSheafLocator#6";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="ReferredActionRequestTypeName"/> property.
		/// </summary>
		public const string ReferredActionRequestTypeNameFilterIdentifier = "ActionProgress.ReferredActionRequestTypeName#7";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="ProcessingState"/> property.
		/// </summary>
		public const string ProcessingStateFilterIdentifier = "ActionProgress.ProcessingState#8";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionProgress()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpnFingerprint">The UPN fingerprint of the recipient of the current object.</param>
		/// <param name="referredSheafLocator">The locator of the referred sheaf (which is referred by the referred
		/// action request).</param>
		/// <param name="referredActionRequestTypeName">The name of the type of the referred action request.</param>
		public ActionProgress( Locator instanceLocator, Locator initiatorUpnFingerprint, Locator recipientUpnFingerprint
				, Locator referredSheafLocator, string referredActionRequestTypeName )
			: base( instanceLocator )
		{
			InitiatorUpnFingerprint = initiatorUpnFingerprint;
			RecipientUpnFingerprint = recipientUpnFingerprint;
			ReferredSheafLocator = referredSheafLocator;
			ReferredActionRequestTypeName = referredActionRequestTypeName;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The processing state of the referred action request. READ-ONLY.
		/// </summary>
		[SpeckleFilterAttribute( ProcessingStateFilterIdentifier )]
		public ActionProcessingStates ProcessingState
		{
			get { return LastProcessingAttempt.ProcessingState; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Locator _InitiatorUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the initiator of the current object. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( InitiatorUpnFingerprintFilterIdentifier )]
		public Locator InitiatorUpnFingerprint
		{
			get { return _InitiatorUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "InitiatorUpnFingerprint" );

				_InitiatorUpnFingerprint = value;
			}
		}

		private Locator _RecipientUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the recipient of the current object. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( RecipientUpnFingerprintFilterIdentifier )]
		public Locator RecipientUpnFingerprint
		{
			get { return _RecipientUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RecipientUpnFingerprint" );

				_RecipientUpnFingerprint = value;
			}
		}

		private Locator _ReferredSheafLocator;
		/// <summary>
		/// The locator of the referred sheaf (which is referred by the referred action request). READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>In the cases when there is no referred sheaf, set this with any value (for example,
		/// <see cref="Locator.Dummy"/>). Since most action requests have referred sheafs, it's better to not allow this to be
		/// null and force the developer to use a non-null value.</para>
		/// 
		/// <para>The current action progress should be deleted when the sheaf with this instance locator is deleted.</para>
		/// </remarks>
		[SerializeAttribute]
		[SpeckleFilterAttribute( ReferredSheafLocatorFilterIdentifier )]
		public Locator ReferredSheafLocator
		{
			get { return _ReferredSheafLocator; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ReferredSheafLocator" );

				_ReferredSheafLocator = value;
			}
		}

		private string _ReferredActionRequestTypeName;
		/// <summary>
		/// The name of the type of the referred action request. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( ReferredActionRequestTypeNameFilterIdentifier )]
		public string ReferredActionRequestTypeName
		{
			get { return _ReferredActionRequestTypeName; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ReferredActionRequestTypeName" );

				_ReferredActionRequestTypeName = value;
			}
		}

		private ActionLastProcessingAttempt _LastProcessingAttempt;
		/// <summary>
		/// Information about the last processing attempt of the referred action request. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This has to be set from the inheritors.</para>
		/// </remarks>
		[SerializeAttribute]
		public ActionLastProcessingAttempt LastProcessingAttempt
		{
			get { return _LastProcessingAttempt; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "LastProcessingAttempt" );

				_LastProcessingAttempt = value;
			}
		}

		#endregion Public serializable properties

		#region Public static methods

		/// <summary>
		/// Updates the information about the last processing attempt of the referred action request.
		/// </summary>
		/// <param name="lastProcessingAttempt">Information about the last processing attempt of the referred action request.
		/// </param>
		public void UpdateLastProcessingAttempt( ActionLastProcessingAttempt lastProcessingAttempt )
		{
			LastProcessingAttempt = lastProcessingAttempt;
		}

		#endregion Public static methods
	}
}
