////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Contract for an action request which is processed in the name of an initiator account and for the benefit a beneficiary
	/// account.
	/// </summary>
	public abstract class ActionRequestWithBeneficiary : ActionRequestWithInitiator
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="BeneficiaryUpnFingerprint"/> property.
		/// </summary>
		public const string BeneficiaryUpnFingerprintFilterIdentifier
			= "ActionRequestWithBeneficiary.BeneficiaryUpnFingerprint#9";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionRequestWithBeneficiary()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpn">The UPN of the recipient of the current action request.</param>
		/// <param name="actionRequestValidity">The validity timespan (in minutes) of the current action request.
		/// <see cref="ActionRequest.DocumentExpirationTime"/> is set with <see cref="Sheaf.CreationTime"/> +
		/// "actionRequestValidity". If this is 0 or a negative number, the action request is valid indefinitely.</param>
		/// <param name="responseEncryptionKeySize">The size (in bytes) that the response's symmetric encryption key must
		/// have. If this is 0 or a negative number, the default <see cref="CoreDocumentsConsts.SymmEncryptionKeySizeDef"/> is
		/// used.</param>
		/// <param name="beneficiaryUpn">The UPN of the beneficiary of the action request.</param>
		/// <remarks>
		/// <para>The symmetric encryption key of the response is automatically generated in here.</para>
		/// </remarks>
		public ActionRequestWithBeneficiary( Locator instanceLocator, MacAlg macAlg
				, IAsymmSigningPaddingAlg asymmSigningPaddingAlg, IHashSerializer signatureHashSerializer
				, Locator initiatorUpnFingerprint, Upn recipientUpn, int actionRequestValidity, int responseEncryptionKeySize
				, Upn beneficiaryUpn )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer, initiatorUpnFingerprint
				, recipientUpn, actionRequestValidity, responseEncryptionKeySize )
		{
			BeneficiaryUpn = beneficiaryUpn;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// If this is true then the tally of the beneficiary is automatically saved if the action request if processed
		/// successfully (by the service which received it). The value is false. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is true then an action service will lock the tally for update, so it will no longer be able to process
		/// multiple action requests in parallel for the same tally.</para>
		/// 
		/// <para>Don't change this value while an action request is processed because the storehouse locking of the tally is
		/// optimized.</para>
		/// </remarks>
		public virtual bool ServiceSavesBeneficiaryTally
		{
			get { return false; }
		}

		/// <summary>
		/// If this is true then the account of the beneficiary is automatically saved if the action request if processed
		/// successfully (by the service which received it). The value is false. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is true then an action service will lock the account for update, so it will no longer be able to
		/// process multiple action requests in parallel for the same account.</para>
		/// 
		/// <para>Don't change this value while an action request is processed because the storehouse locking of the account is
		/// optimized.</para>
		/// </remarks>
		public virtual bool ServiceSavesBeneficiaryAccount
		{
			get { return false; }
		}

		/// <summary>
		/// The fingerprint of the UPN of the beneficiary of the action request. READ-ONLY.
		/// </summary>
		[SpeckleFilterAttribute( BeneficiaryUpnFingerprintFilterIdentifier )]
		public Locator BeneficiaryUpnFingerprint
		{
			get { return BeneficiaryUpn.Fingerprint; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Upn _BeneficiaryUpn;
		/// <summary>
		/// The UPN of the beneficiary of the action request. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is used to prove with cryptographic certainty who is the owner of the account for the benefit of which
		/// the action request is to be processed.</para>
		/// </remarks>
		[SerializeAttribute]
		public Upn BeneficiaryUpn
		{
			get { return _BeneficiaryUpn; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "BeneficiaryUpn" );

				_BeneficiaryUpn = value;
			}
		}

		#endregion Public serializable properties
	}
}
