////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Contract for an anonymous action request, that is, one without an initiator account but with an initiator identity.
	/// </summary>
	/// <remarks>
	/// <para>An anonymous action request with mutual knowledge is used when its recipient either has an account for the
	/// initiator (and the request is signed with the identity from that account), or its recipient has a progress
	/// (<see cref="CreateAccountClientProgress"/>) for a "CreateAccountAnonymousActionRequest" action request queued to be sent
	/// to the initiator (and the action request is self-signed with the identity from which the UPN from the progress was
	/// generated).</para>
	/// 
	/// <para>An action request of this type must be signed with the initiator identity
	/// (<see cref="AnonymousActionRequestWithMutualKnowledge.InitiatorIdentity"/>).</para>
	/// </remarks>
	public abstract class AnonymousActionRequestWithMutualKnowledge : AnonymousActionRequest
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AnonymousActionRequestWithMutualKnowledge()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpn">The UPN of the recipient of the current action request.</param>
		/// <param name="actionRequestValidity">The validity timespan (in minutes) of the current action request.
		/// <see cref="ActionRequest.DocumentExpirationTime"/> is set with <see cref="Sheaf.CreationTime"/> +
		/// "actionRequestValidity". If this is 0 or a negative number, the action request is valid indefinitely.</param>
		/// <param name="responseEncryptionKeySize">The size (in bytes) that the response's symmetric encryption key must
		/// have. If this is 0 or a negative number, the default <see cref="CoreDocumentsConsts.SymmEncryptionKeySizeDef"/> is
		/// used.</param>
		/// <param name="initiatorIdentity">The processing identity publisher of the initiator.</param>
		/// <remarks>
		/// <para>The symmetric encryption key of the response is automatically generated in here.</para>
		/// </remarks>
		public AnonymousActionRequestWithMutualKnowledge( Locator instanceLocator, MacAlg macAlg
				, IAsymmSigningPaddingAlg asymmSigningPaddingAlg, IHashSerializer signatureHashSerializer
				, Locator initiatorUpnFingerprint, Upn recipientUpn, int actionRequestValidity, int responseEncryptionKeySize
				, IIdentityPublisher initiatorIdentity )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer, initiatorUpnFingerprint
				, recipientUpn, actionRequestValidity, responseEncryptionKeySize )
		{
			InitiatorIdentity = initiatorIdentity;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( base.ExpirationTime, InitiatorIdentity.ExpirationTime ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private IIdentityPublisher _InitiatorIdentity;
		/// <summary>
		/// The processing identity publisher of the initiator. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public IIdentityPublisher InitiatorIdentity
		{
			get { return _InitiatorIdentity; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "InitiatorIdentity" );

				_InitiatorIdentity = value;
			}
		}

		#endregion Public serializable properties
	}
}
