////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Contract for an action request. An action request describes an action which may be processed by a service, perhaps in the
	/// name of an initiator, and perhaps for the benefit a beneficiary.
	/// </summary>
	/// <remarks>
	/// <para>Action requests are meant to be relatively small documents, in the order of (tens of) kilobytes. However, the
	/// processing is efficient even for the (normal) maximum size of a sheaf. Beyond this, the information should be split in
	/// multiple action request, or should be transferred through a separate communication protocol.</para>
	/// 
	/// <para>An action request is a single message which is sent by an action client and received by an action service, on
	/// network connections which are independent for each request.</para>
	/// 
	/// <para>When processed by an action client, an action request is signed, is asymmetrically encrypted (in a
	/// <see cref="Vela.CoreDocuments.AsymmEncryptedEnvelope"/>) envelope or symmetrically encrypted (in a
	/// <see cref="Vela.CoreDocuments.UpnEncryptedEnvelope"/>) envelope, and requires a response from the recipient service.
	/// </para>
	/// 
	/// <para>When processed by an action service, the signature of an action request is verified, random data from a request is
	/// added to the entropy pool, and the recipient service loads associated storehouse information (like account information)
	/// while it process it.</para>
	/// 
	/// <para>An action request may be of the following types: <see cref="ActionRequestWithInitiator"/> and
	/// <see cref="AnonymousActionRequest"/>. All the other requests must be derived from these types.</para>
	/// 
	/// <para><see cref="ActionRequest.TransportedSheafLocator"/> may not be null if
	/// <see cref="ActionRequest.ServiceSavesProgress"/> is true.</para>
	/// 
	/// <para>Examples of actions:</para>
	/// <list type="bullet">
	///		<item>Get presenter.</item>
	///		<item>Make payment.</item>
	/// </list>
	/// </remarks>
	public abstract class ActionRequest : SignedDocument, IInitiatorUpnFingerprintContainer, IRecipientUpnFingerprintContainer
		, IExpirableDocument
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="InitiatorUpnFingerprint"/> property.
		/// </summary>
		public const string InitiatorUpnFingerprintFilterIdentifier = "ActionRequest.InitiatorUpnFingerprint#4";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="RecipientUpnFingerprint"/> property.
		/// </summary>
		public const string RecipientUpnFingerprintFilterIdentifier = "ActionRequest.RecipientUpnFingerprint#5";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="RecipientTextualUpn"/> property.
		/// </summary>
		public const string RecipientTextualUpnFilterIdentifier = "ActionRequest.RecipientTextualUpn#6";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="TransportedSheafLocator"/> property.
		/// </summary>
		public const string TransportedSheafLocatorFilterIdentifier = "ActionRequest.TransportedSheafLocator#7";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="DocumentExpirationTime"/> property.
		/// </summary>
		public const string DocumentExpirationTimeFilterIdentifier = "ActionRequest.DocumentExpirationTime#8";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionRequest()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpn">The UPN of the recipient of the current action request.</param>
		/// <param name="actionRequestValidity">The validity timespan (in minutes) of the current action request.
		/// <see cref="ActionRequest.DocumentExpirationTime"/> is set with <see cref="Sheaf.CreationTime"/> +
		/// "actionRequestValidity". If this is 0 or a negative number, the action request is valid indefinitely.</param>
		/// <param name="responseEncryptionKeySize">The size (in bytes) that the response's symmetric encryption key must
		/// have. If this is 0 or a negative number, the default <see cref="CoreDocumentsConsts.SymmEncryptionKeySizeDef"/> is
		/// used.</param>
		/// <remarks>
		/// <para>The symmetric encryption key of the response is automatically generated in here.</para>
		/// </remarks>
		public ActionRequest( Locator instanceLocator, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg
				, IHashSerializer signatureHashSerializer, Locator initiatorUpnFingerprint, Upn recipientUpn
				, int actionRequestValidity, int responseEncryptionKeySize )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer )
		{
			InitiatorUpnFingerprint = initiatorUpnFingerprint;
			RecipientUpn = recipientUpn;

			if( actionRequestValidity > 0 )
				DocumentExpirationTime = CreationTime.AddMinutes( actionRequestValidity );

			if( responseEncryptionKeySize <= 0 )
				responseEncryptionKeySize = CoreDocumentsConsts.SymmEncryptionKeySizeDef;
			ResponseEncryptionKey = EntropyPool.GetBytes( responseEncryptionKeySize );
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The type of the action response associated to the current action request. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>An action service must return an action response derived either from this type or from
		/// <see cref="ErrorActionResponse"/>.</para>
		/// 
		/// <para>This type must have a constructor with the following parameters: the instance locator of the action response,
		/// the action request to which the current action response is associated. This isn't enforced.</para>
		/// </remarks>
		public abstract Type ActionResponseType { get; }

		/// <summary>
		/// If this is true then the current action request must be updated before it's sent to a service for processing. The
		/// value is false. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this property is true then the action request is called "updatable (action) request".</para>
		/// 
		/// <para>If the action request is updated, it's automatically (re)signed by the action client.</para>
		/// </remarks>
		public virtual bool NeedsUpdateBeforeSending
		{
			get { return false; }
		}

		/// <summary>
		/// If this is true then action request is automatically saved if it was processed successfully (by the service which
		/// received it). The value is false. READ-ONLY.
		/// </summary>
		public virtual bool ServiceSavesReceivedActionRequest
		{
			get { return false; }
		}

		/// <summary>
		/// If this is true then an action service creates and saves an action progress (<see cref="ActionServiceProgress"/>)
		/// after it successfully processes the current action request. The value is false. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is true, the associated action progress may NOT miss from the storehouse, so ensure that one is created
		/// when the document (which is transported by the current action request) is saved in the storehouse.</para>
		/// 
		/// <para>If this is true, the <see cref="TransportedSheaf"/> property may not miss.</para>
		/// </remarks>
		public virtual bool ServiceSavesProgress
		{
			get { return false; }
		}

		/// <summary>
		/// If this is true and the storehouse already contains the <see cref="TransportedSheaf"/> sheaf then the service ignores
		/// the current action request and returns a basic answer (with the type <see cref="ActionResponseType"/>). The default
		/// value is true. READ-ONLY.
		/// </summary>
		public virtual bool ServiceIgnoresActionRequestIfTransportedSheafExists
		{
			get { return true; }
		}

		/// <summary>
		/// If this is true then the action response is automatically saved if the action request was processed successfully (by
		/// the service which received it). The value is false. READ-ONLY.
		/// </summary>
		public virtual bool ServiceSavesSentActionResponse
		{
			get { return false; }
		}

		/// <summary>
		/// If this is true then the associated action response can be created (dynamically, with <see cref="Activator"/>) in a
		/// consistent state with the standard constructor. The value is false. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The value of this property is important when the sheaf transported by the action request already exists in the
		/// service's storehouse, and <see cref="ServiceIgnoresActionRequestIfTransportedSheafExists"/> is true, but the
		/// associated action responses are not saved because <see cref="ServiceSavesSentActionResponse"/> is false. In such a
		/// case, if this and the <see cref="ServiceSavesProgress"/> properties are true then an action response can be created
		/// and sent to the client, but if either of these properties are false then a
		/// <see cref="TransportedSheafExistsException"/> fatal exception is thrown.</para>
		/// 
		/// <para>The standard constructor for an action response has the following parameters: instance locator, the associated
		/// action request, the secondary processing state (normally, taken from the action service progress).</para>
		/// 
		/// <para>To avoid mistakes, this must be false by default and must be set with true only for individual action requests.
		/// </para>
		/// </remarks>
		public virtual bool ServiceResponseIsConsistentWithStandardConstructor
		{
			get { return false; }
		}

		/// <summary>
		/// If this is true then an action service should compress the associated action response before sending it to the
		/// initiator of the current action request. The value is true. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is ignored if <see cref="IActionRequestWithMultiframeResponse.ServiceMustEncryptResponsesInFrames"/> is
		/// false because compression can only be performed when using encrypted envelopes (because they store the compression
		/// algorithm).</para>
		/// </remarks>
		public virtual bool ServiceShouldCompressResponse
		{
			get { return true; }
		}

		/// <summary>
		/// If this is true and an exception occurs while the service is processing the current action request, the service may
		/// update the application's UI; if this is false, the UI may not be updated. The default value is false. READ-ONLY.
		/// </summary>
		public virtual bool ServiceMayUpdateUiOnException
		{
			get { return false; }
		}

		/// <summary>
		/// If this is true then a large cache is required for sending and receiving data through the network when processing the
		/// action response associated to the current action request (because it's used for large data transfers). The default
		/// value is false. READ-ONLY.
		/// </summary>
		public virtual bool ActionResponseRequiresLargeCommunicationCacheSize
		{
			get { return false; }
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( base.ExpirationTime, DocumentExpirationTime, RecipientUpn.ExpirationTime ); }
		}

		/// <summary>
		/// The UPN fingerprint of the recipient of the current object. READ-ONLY.
		/// </summary>
		[SpeckleFilterAttribute( RecipientUpnFingerprintFilterIdentifier )]
		public Locator RecipientUpnFingerprint
		{
			get { return RecipientUpn.Fingerprint; }
		}

		/// <summary>
		/// The textual UPN of the recipient of the current action request. READ-ONLY.
		/// </summary>
		[SpeckleFilterAttribute( RecipientTextualUpnFilterIdentifier )]
		public string RecipientTextualUpn
		{
			get { return RecipientUpn.ToString(); }
		}

		/// <summary>
		/// The sheaf which is transported by the current action request, and also the referred sheaf from the associated action
		/// progress. This may be null. The default value is null. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The current action request should be deleted when this sheaf is deleted.</para>
		/// </remarks>
		public virtual ISheaf TransportedSheaf
		{
			get { return null; }
		}

		/// <summary>
		/// The locator of the sheaf which is transported by the current action request, and also of the referred sheaf from the
		/// associated action progress. This may be null. The default value is null. READ-ONLY.
		/// </summary>
		[SpeckleFilterAttribute( TransportedSheafLocatorFilterIdentifier )]
		public Locator TransportedSheafLocator
		{
			get { return (TransportedSheaf != null) ? TransportedSheaf.InstanceLocator : null; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Locator _InitiatorUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the initiator of the current object. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( InitiatorUpnFingerprintFilterIdentifier )]
		public Locator InitiatorUpnFingerprint
		{
			get { return _InitiatorUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "InitiatorUpnFingerprint" );

				_InitiatorUpnFingerprint = value;
			}
		}

		private Upn _RecipientUpn;
		/// <summary>
		/// The UPN of the recipient of the current action request. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Upn RecipientUpn
		{
			get { return _RecipientUpn; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RecipientUpn" );

				_RecipientUpn = value;
			}
		}

		private DateTime _DocumentExpirationTime = DateTime.MaxValue;
		/// <summary>
		/// The expiration time (in UTC format) of the document. The current document is valid before this time; from and after
		/// this time, it should not be used in processing. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( DocumentExpirationTimeFilterIdentifier )]
		public DateTime DocumentExpirationTime
		{
			get { return _DocumentExpirationTime; }
			protected set { _DocumentExpirationTime = value; }
		}

		private byte[] _ResponseEncryptionKey;
		/// <summary>
		/// The symmetric encryption key used to encrypt the action response, randomly generated. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] ResponseEncryptionKey
		{
			get { return _ResponseEncryptionKey; }
			protected set
			{
				ValidationUtils.Array( value, CoreDocumentsConsts.SymmEncryptionKeySizeMin
					, CoreDocumentsConsts.SymmEncryptionKeySizeMax, "ResponseEncryptionKey" );

				_ResponseEncryptionKey = value;
			}
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Validates the current object.
		/// </summary>
		protected virtual void Validate()
		{
			if( ServiceSavesProgress && (TransportedSheafLocator == null) )
			{
				throw new ValidationException( "The progress associated to the action request is saved, so the locator of"
						+ " the transported sheaf must exist.", null, null
					, new ExceptionDetail( "Action request type", GetType().FullName )
					, new ExceptionDetail( "Action request locator", InstanceLocator ) );
			}
		}

		/// <summary>
		/// This method is automatically called before the current object is serialized.
		/// </summary>
		protected virtual void Serializing()
		{
			Validate();
		}

		/// <summary>
		/// This method is automatically called after the current object has been deserialized.
		/// </summary>
		protected virtual void Deserialized()
		{
			Validate();
		}

		#endregion Private methods
	}
}
