////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Contract for an action request which is processed in the name of an initiator account.
	/// </summary>
	/// <remarks>
	/// <para>An action request with initiator must be signed by the processing identity of the initiator.</para>
	/// 
	/// <para>An action request with initiator must be sent to a service encrypted in an
	/// <see cref="AsymmEncryptedEnvelope"/> envelope.</para>
	/// </remarks>
	public abstract class ActionRequestWithInitiator : ActionRequest
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionRequestWithInitiator()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpn">The UPN of the recipient of the current action request.</param>
		/// <param name="actionRequestValidity">The validity timespan (in minutes) of the current action request.
		/// <see cref="ActionRequest.DocumentExpirationTime"/> is set with <see cref="Sheaf.CreationTime"/> +
		/// "actionRequestValidity". If this is 0 or a negative number, the action request is valid indefinitely.</param>
		/// <param name="responseEncryptionKeySize">The size (in bytes) that the response's symmetric encryption key must
		/// have. If this is 0 or a negative number, the default <see cref="CoreDocumentsConsts.SymmEncryptionKeySizeDef"/> is
		/// used.</param>
		/// <remarks>
		/// <para>The symmetric encryption key of the response is automatically generated in here.</para>
		/// </remarks>
		public ActionRequestWithInitiator( Locator instanceLocator, MacAlg macAlg
				, IAsymmSigningPaddingAlg asymmSigningPaddingAlg, IHashSerializer signatureHashSerializer
				, Locator initiatorUpnFingerprint, Upn recipientUpn, int actionRequestValidity, int responseEncryptionKeySize )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer, initiatorUpnFingerprint
				, recipientUpn, actionRequestValidity, responseEncryptionKeySize )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// If this is true then after an action client processes the current action request, the associated action request
		/// progress (<see cref="ActionClientProgress"/>) is kept in the storehouse, else it's deleted. The value is false.
		/// READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is true, the associated action progress may NOT miss from the storehouse, so ensure that one is created
		/// when the document (which is transported by the current action request) is saved in the storehouse.</para>
		/// </remarks>
		public virtual bool ClientProgressIsPermanent
		{
			get { return false; }
		}

		/// <summary>
		/// If this is true then the tally of the initiator is automatically saved if the action request is processed
		/// successfully (by the service which received it). The value is false. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is true then an action service will lock the tally for update, so it will no longer be able to process
		/// multiple action requests in parallel for the same tally.</para>
		/// 
		/// <para>Don't change this value while an action request is processed because the storehouse locking of the tally is
		/// optimized.</para>
		/// </remarks>
		public virtual bool ServiceSavesInitiatorTally
		{
			get { return false; }
		}

		/// <summary>
		/// If this is true then the account of the initiator is automatically saved if the action request is processed
		/// successfully (by the service which received it). The value is false. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is true then an action service will lock the account for update, so it will no longer be able to
		/// process multiple action requests in parallel for the same account.</para>
		/// 
		/// <para>Don't change this value while an action request is processed because the storehouse locking of the account is
		/// optimized.</para>
		/// </remarks>
		public virtual bool ServiceSavesInitiatorAccount
		{
			get { return false; }
		}

		#endregion Public non-serializable properties
	}
}
