////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Contract for an anonymous action request which the recipient processes even if it has no knowledge of the initiator.
	/// </summary>
	/// <remarks>
	/// <para>An anonymous action request is used when its recipient has no knowledge of its initiator (specifically, when its
	/// initiator doesn't have an account on the remote peer, and the remote peer has no "CreateAccountAnonymousActionRequest"
	/// action request queued to be sent to the initiator).</para>
	/// 
	/// <para>This action request can normally be used when its initiator doesn't have an account <b>on</b> the remote peer, but
	/// does have an account <b>for</b> the remote peer.</para>
	/// 
	/// <para>"Anonymous" doesn't mean unlinkable / untraceable.</para>
	/// 
	/// <para>An anonymous action request must be sent to a service encrypted either in an
	/// <see cref="AsymmEncryptedEnvelope"/> or in an <see cref="UpnEncryptedEnvelope"/> envelope.</para>
	/// </remarks>
	public abstract class AnonymousActionRequest : ActionRequest
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AnonymousActionRequest()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpn">The UPN of the recipient of the current action request.</param>
		/// <param name="actionRequestValidity">The validity timespan (in minutes) of the current action request.
		/// <see cref="ActionRequest.DocumentExpirationTime"/> is set with <see cref="Sheaf.CreationTime"/> +
		/// "actionRequestValidity". If this is 0 or a negative number, the action request is valid indefinitely.</param>
		/// <param name="responseEncryptionKeySize">The size (in bytes) that the response's symmetric encryption key must
		/// have. If this is 0 or a negative number, the default <see cref="CoreDocumentsConsts.SymmEncryptionKeySizeDef"/> is
		/// used.</param>
		/// <remarks>
		/// <para>The symmetric encryption key of the response is automatically generated in here.</para>
		/// </remarks>
		public AnonymousActionRequest( Locator instanceLocator, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg
				, IHashSerializer signatureHashSerializer, Locator initiatorUpnFingerprint, Upn recipientUpn
				, int actionRequestValidity, int responseEncryptionKeySize )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer, initiatorUpnFingerprint
				, recipientUpn, actionRequestValidity, responseEncryptionKeySize )
		{
		}

		#endregion Constructors
	}
}
