////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// An anonymous action request to get a presenter and create an account.
	/// </summary>
	public class CreateAccountAnonymousActionRequest : AnonymousActionRequestWithMutualKnowledge
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected CreateAccountAnonymousActionRequest()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="recipientUpn">The UPN of the recipient of the current action request.</param>
		/// <param name="actionRequestValidity">The validity timespan (in minutes) of the current action request.
		/// <see cref="ActionRequest.DocumentExpirationTime"/> is set with <see cref="Sheaf.CreationTime"/> +
		/// "actionRequestValidity". If this is 0 or a negative number, the action request is valid indefinitely.</param>
		/// <param name="responseEncryptionKeySize">The size (in bytes) that the response's symmetric encryption key must
		/// have. If this is 0 or a negative number, the default <see cref="CoreDocumentsConsts.SymmEncryptionKeySizeDef"/> is
		/// used.</param>
		/// <param name="initiatorIdentity">The processing identity publisher of the initiator.</param>
		/// <remarks>
		/// <para>The symmetric encryption key of the response is automatically generated in here.</para>
		/// </remarks>
		public CreateAccountAnonymousActionRequest( Locator instanceLocator, MacAlg macAlg
				, IAsymmSigningPaddingAlg asymmSigningPaddingAlg, IHashSerializer signatureHashSerializer, Upn recipientUpn
				, int actionRequestValidity, int responseEncryptionKeySize, IIdentityPublisher initiatorIdentity )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer, initiatorIdentity.Upn.Fingerprint
				, recipientUpn, actionRequestValidity, responseEncryptionKeySize, initiatorIdentity )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The type of the action response associated to the current action request. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>An action service must return an action response derived either from this type or from
		/// <see cref="ErrorActionResponse"/>.</para>
		/// 
		/// <para>This type must have a constructor with the following parameters: the instance locator of the action response,
		/// the action request to which the current action response is associated. This isn't enforced.</para>
		/// </remarks>
		public override Type ActionResponseType
		{
			get { return typeof( CreateAccountAnonymousActionResponse ); }
		}

		#endregion Public non-serializable properties
	}
}
