////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.PeerDocuments;
using Vela.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// An action request to download the files which were offered by the remote peer through a
	/// <see cref="SendLetterActionRequest"/> action request.
	/// </summary>
	/// <remarks>
	/// <para>An action service which receives such an action request, sends back to the local action service a multiframe
	/// response (<see cref="DownloadFilesActionResponse"/>) which contains the requested files.</para>
	/// </remarks>
	public class DownloadFilesActionRequest : ActionRequestWithInitiator, IActionRequestWithMultiframeResponse
		, IAccessGranterLocatorContainer
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="AccessGranterLocator"/> property.
		/// </summary>
		public const string AccessGranterLocatorFilterIdentifier = "DownloadFilesActionRequest.AccessGranterLocator#9";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected DownloadFilesActionRequest()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpn">The UPN of the recipient of the current action request.</param>
		/// <param name="actionRequestValidity">The validity timespan (in minutes) of the current action request.
		/// <see cref="ActionRequest.DocumentExpirationTime"/> is set with <see cref="Sheaf.CreationTime"/> +
		/// "actionRequestValidity". If this is 0 or a negative number, the action request is valid indefinitely.</param>
		/// <param name="responseEncryptionKeySize">The size (in bytes) that the response's symmetric encryption key must
		/// have. If this is 0 or a negative number, the default <see cref="CoreDocumentsConsts.SymmEncryptionKeySizeDef"/> is
		/// used.</param>
		/// <param name="accessGranterLocator">The locator of the document which has offered the attached files.</param>
		/// <param name="accessGranterCreationTime">The creation time of the document which has offered the attached files.
		/// </param>
		/// <param name="attachedFiles">Files that the local peer may download. This may be empty, but not null.</param>
		/// <remarks>
		/// <para>The symmetric encryption key of the response is automatically generated in here.</para>
		/// </remarks>
		public DownloadFilesActionRequest( Locator instanceLocator, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg
				, IHashSerializer signatureHashSerializer, Locator initiatorUpnFingerprint, Upn recipientUpn
				, int actionRequestValidity, int responseEncryptionKeySize, Locator accessGranterLocator
				, DateTime accessGranterCreationTime, AttachedFiles attachedFiles )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer, initiatorUpnFingerprint
				, recipientUpn, actionRequestValidity, responseEncryptionKeySize )
		{
			AccessGranterLocator = accessGranterLocator;
			AccessGranterCreationTime = accessGranterCreationTime;
			AttachedFiles = attachedFiles;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The type of the action response associated to the current action request. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>An action service must return an action response derived either from this type or from
		/// <see cref="ErrorActionResponse"/>.</para>
		/// 
		/// <para>This type must have a constructor with the following parameters: the instance locator of the action response,
		/// the action request to which the current action response is associated. This isn't enforced.</para>
		/// </remarks>
		public override Type ActionResponseType
		{
			get { return typeof( DownloadFilesActionResponse ); }
		}

		/// <summary>
		/// If this is true then the current action request must be updated before it's sent to a service for processing. The
		/// value is true. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this property is true then the action request is called "updatable (action) request".</para>
		/// 
		/// <para>This allows an action client, for example, to download a file.</para>
		/// 
		/// <para>If the action request is updated, it's automatically (re)signed by the action client.</para>
		/// </remarks>
		public sealed override bool NeedsUpdateBeforeSending
		{
			get { return true; }
		}

		/// <summary>
		/// If this is true then after an action client processes the current action request, the associated action request
		/// progress (<see cref="ActionClientProgress"/>) is kept in the storehouse, else it's deleted. The value is true.
		/// READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is true, the associated action progress may NOT miss from the storehouse, so ensure that one is created
		/// when the document (which is transported by the current action request) is saved in the storehouse.</para>
		/// </remarks>
		public override bool ClientProgressIsPermanent
		{
			get { return true; }
		}

		/// <summary>
		/// If this is true then an action service should compress the associated action response before sending it to the
		/// initiator of the current action request. The value is true. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>There is no point in compressing a response of the current type because it's most likely used to transfer a
		/// media file (which is already compressed).</para>
		/// 
		/// <para>This is ignored if <see cref="IActionRequestWithMultiframeResponse.ServiceMustEncryptResponsesInFrames"/> is
		/// false because compression can only be performed when using encrypted envelopes (because they store the compression
		/// algorithm).</para>
		/// </remarks>
		public override bool ServiceShouldCompressResponse
		{
			get { return false; }
		}

		/// <summary>
		/// If this is true and an exception occurs while the service is processing the current action request, the service may
		/// update the application's UI; if this is false, the UI may not be updated. The value is true. READ-ONLY.
		/// </summary>
		public override bool ServiceMayUpdateUiOnException
		{
			get { return true; }
		}

		/// <summary>
		/// If this is true then a large cache is required for sending and receiving data through the network when processing the
		/// action response associated to the current action request (because it's used for large data transfers). The value is
		/// true. READ-ONLY.
		/// </summary>
		public override bool ActionResponseRequiresLargeCommunicationCacheSize
		{
			get { return true; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Locator _AccessGranterLocator;
		/// <summary>
		/// The locator of the document which has offered the attached files. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The current action request must be deleted when the document with this instance locator is deleted.</para>
		/// </remarks>
		[SerializeAttribute]
		[SpeckleFilterAttribute( AccessGranterLocatorFilterIdentifier )]
		public Locator AccessGranterLocator
		{
			get { return _AccessGranterLocator; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "AccessGranterLocator" );

				_AccessGranterLocator = value;
			}
		}

		private DateTime _AccessGranterCreationTime;
		/// <summary>
		/// The creation time (in UTC format) of the document which has offered the attached files. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public DateTime AccessGranterCreationTime
		{
			get { return _AccessGranterCreationTime; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "AccessGranterCreationTime" );

				_AccessGranterCreationTime = value;
			}
		}

		private AttachedFiles _AttachedFiles;
		/// <summary>
		/// Files that the local peer wants to download. This may be empty, but not null. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The size of the files must contain the current (= already downloaded) size of the files.</para>
		/// </remarks>
		[SerializeAttribute]
		public AttachedFiles AttachedFiles
		{
			get { return _AttachedFiles; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "AttachedFiles" );

				_AttachedFiles = value;
			}
		}

		private bool _ServiceMustEncryptResponsesInFrames = true;
		/// <summary>
		/// If this is true then an action service must encrypt the action responses from the frames of a multiframe response.
		/// This may be false because the frames are (further) encrypted / obfuscated with the response encryption key.
		/// </summary>
		/// <remarks>
		/// <para>Since this is a new property, it must have "MustBeDeserialized = false".</para>
		/// 
		/// <para>If this is false, <see cref="ActionRequest.ServiceShouldCompressResponse"/> is ignored because compression is
		/// only performed when using encrypted envelopes.</para>
		/// 
		/// <para>If this is false, the size of the content of the action response frames is not aligned, so you should either
		/// manually align the size, or send a fixed or random amount of information in each frame.</para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false, DontHashSerializeValue = true )]
		public bool ServiceMustEncryptResponsesInFrames
		{
			get { return _ServiceMustEncryptResponsesInFrames; }
			set { _ServiceMustEncryptResponsesInFrames = value; }
		}

		#endregion Public serializable properties
	}
}
