////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.PeerDocuments;
using Vela.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// An action request which sends a local presenter to update it on a remote peer.
	/// </summary>
	public class PushUpdatePresenterActionRequest : ActionRequestWithInitiator
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected PushUpdatePresenterActionRequest()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpn">The UPN of the recipient of the current action request.</param>
		/// <param name="actionRequestValidity">The validity timespan (in minutes) of the current action request.
		/// <see cref="ActionRequest.DocumentExpirationTime"/> is set with <see cref="Sheaf.CreationTime"/> +
		/// "actionRequestValidity". If this is 0 or a negative number, the action request is valid indefinitely.</param>
		/// <param name="responseEncryptionKeySize">The size (in bytes) that the response's symmetric encryption key must
		/// have. If this is 0 or a negative number, the default <see cref="CoreDocumentsConsts.SymmEncryptionKeySizeDef"/> is
		/// used.</param>
		/// <param name="presenter">The updated presenter to push.</param>
		/// <remarks>
		/// <para>The symmetric encryption key of the response is automatically generated in here.</para>
		/// </remarks>
		public PushUpdatePresenterActionRequest( Locator instanceLocator, MacAlg macAlg
				, IAsymmSigningPaddingAlg asymmSigningPaddingAlg, IHashSerializer signatureHashSerializer
				, Locator initiatorUpnFingerprint, Upn recipientUpn, int actionRequestValidity, int responseEncryptionKeySize
				, Presenter presenter )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer, initiatorUpnFingerprint
				, recipientUpn, actionRequestValidity, responseEncryptionKeySize )
		{
			Presenter = presenter;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The type of the action response associated to the current action request. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>An action service must return an action response derived either from this type or from
		/// <see cref="ErrorActionResponse"/>.</para>
		/// 
		/// <para>This type must have a constructor with the following parameters: the instance locator of the action response,
		/// the action request to which the current action response is associated. This isn't enforced.</para>
		/// </remarks>
		public override Type ActionResponseType
		{
			get { return typeof( PushUpdatePresenterActionResponse ); }
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( base.ExpirationTime, Presenter.ExpirationTime ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Presenter _Presenter;
		/// <summary>
		/// The presenter. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This must be self-signed.</para>
		/// </remarks>
		[SerializeAttribute]
		public Presenter Presenter
		{
			get { return _Presenter; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Presenter" );

				_Presenter = value;
			}
		}

		#endregion Public serializable properties
	}
}
