////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// An anonymous action request to resolve a UPN.
	/// </summary>
	/// <remarks>
	/// <para>This is used to guarantee that the PNRP has resolved a UPN to the correct peer. Basically, the initiator must go
	/// through all the found peers to verify the signature and compare the resolved UPN with the UPN of the identity of each
	/// found peer.</para>
	/// 
	/// <para>This can't be derived from <see cref="AnonymousActionRequestWithMutualKnowledge"/> because the UPNs must be
	/// resolvable even for (other) action requests derived from <see cref="AnonymousActionRequest"/> (like
	/// <see cref="GetFileAnonymousActionRequest"/>). This doesn't create a privacy leak since the PNRP itself publishes the UPNs
	/// for anyone to see / resolve.</para>
	/// </remarks>
	public class ResolveUpnAnonymousActionRequest : AnonymousActionRequest
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ResolveUpnAnonymousActionRequest()
		{
		}


		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpn">The UPN of the recipient of the current action request.</param>
		/// <param name="actionRequestValidity">The validity timespan (in minutes) of the current action request.
		/// <see cref="ActionRequest.DocumentExpirationTime"/> is set with <see cref="Sheaf.CreationTime"/> +
		/// "actionRequestValidity". If this is 0 or a negative number, the action request is valid indefinitely.</param>
		/// <param name="responseEncryptionKeySize">The size (in bytes) that the response's symmetric encryption key must
		/// have. If this is 0 or a negative number, the default <see cref="CoreDocumentsConsts.SymmEncryptionKeySizeDef"/> is
		/// used.</param>
		/// <remarks>
		/// <para>The symmetric encryption key of the response is automatically generated in here.</para>
		/// </remarks>
		public ResolveUpnAnonymousActionRequest( Locator instanceLocator, MacAlg macAlg
				, IAsymmSigningPaddingAlg asymmSigningPaddingAlg, IHashSerializer signatureHashSerializer
				, Locator initiatorUpnFingerprint, Upn recipientUpn, int actionRequestValidity, int responseEncryptionKeySize )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer, initiatorUpnFingerprint
				, recipientUpn, actionRequestValidity, responseEncryptionKeySize )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The type of the action response associated to the current action request. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>An action service must return an action response derived either from this type or from
		/// <see cref="ErrorActionResponse"/>.</para>
		/// 
		/// <para>This type must have a constructor with the following parameters: the instance locator of the action response,
		/// the action request to which the current action response is associated. This isn't enforced.</para>
		/// </remarks>
		public override Type ActionResponseType
		{
			get { return typeof( ResolveUpnAnonymousActionResponse ); }
		}

		#endregion Public non-serializable properties
	}
}
