////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.PeerDocuments;
using Vela.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// An action request to send a letter.
	/// </summary>
	public class SendLetterActionRequest : ActionRequestWithInitiator
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected SendLetterActionRequest()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpn">The UPN of the recipient of the current action request.</param>
		/// <param name="actionRequestValidity">The validity timespan (in minutes) of the current action request.
		/// <see cref="ActionRequest.DocumentExpirationTime"/> is set with <see cref="Sheaf.CreationTime"/> +
		/// "actionRequestValidity". If this is 0 or a negative number, the action request is valid indefinitely.</param>
		/// <param name="responseEncryptionKeySize">The size (in bytes) that the response's symmetric encryption key must
		/// have. If this is 0 or a negative number, the default <see cref="CoreDocumentsConsts.SymmEncryptionKeySizeDef"/> is
		/// used.</param>
		/// <param name="letter">The letter to send.</param>
		/// <remarks>
		/// <para>The symmetric encryption key of the response is automatically generated in here.</para>
		/// </remarks>
		public SendLetterActionRequest( Locator instanceLocator, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg
				, IHashSerializer signatureHashSerializer, Locator initiatorUpnFingerprint, Upn recipientUpn
				, int actionRequestValidity, int responseEncryptionKeySize, Letter letter )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer, initiatorUpnFingerprint
				, recipientUpn, actionRequestValidity, responseEncryptionKeySize )
		{
			Letter = letter;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The type of the action response associated to the current action request. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>An action service must return an action response derived either from this type or from
		/// <see cref="ErrorActionResponse"/>.</para>
		/// 
		/// <para>This type must have a constructor with the following parameters: the instance locator of the action response,
		/// the action request to which the current action response is associated. This isn't enforced.</para>
		/// </remarks>
		public override Type ActionResponseType
		{
			get { return typeof( SendLetterActionResponse ); }
		}

		/// <summary>
		/// If this is true then after an action client processes the current action request, the associated action request
		/// progress (<see cref="ActionClientProgress"/>) is kept in the storehouse, else it's deleted. The value is true.
		/// READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is true, the associated action progress may NOT miss from the storehouse, so ensure that one is created
		/// when the document (which is transported by the current action request) is saved in the storehouse.</para>
		/// </remarks>
		public override bool ClientProgressIsPermanent
		{
			get { return true; }
		}

		/// <summary>
		/// If this is true then an action service creates and saves an action progress (<see cref="ActionServiceProgress"/>)
		/// after it successfully processes the current action request. The value is true. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is true, the associated action progress may NOT miss from the storehouse, so ensure that one is created
		/// when the document (which is transported by the current action request) is saved in the storehouse.</para>
		/// </remarks>
		public override bool ServiceSavesProgress
		{
			get { return true; }
		}

		/// <summary>
		/// If this is true then the associated action response can be created (dynamically, with <see cref="Activator"/>) in a
		/// consistent state with the standard constructor. The value is false. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The value of this property is important when the sheaf transported by the action request already exists in the
		/// service's storehouse, and <see cref="ActionRequest.ServiceIgnoresActionRequestIfTransportedSheafExists"/> is true,
		/// but the associated action responses are not saved because <see cref="ActionRequest.ServiceSavesSentActionResponse"/>
		/// is false. In such a case, if this and the <see cref="ServiceSavesProgress"/> properties are true then an action
		/// response can be created and sent to the client, but if either of these properties are false then a
		/// <see cref="TransportedSheafExistsException"/> fatal exception is thrown.</para>
		/// 
		/// <para>The standard constructor for an action response has the following parameters: instance locator, the associated
		/// action request, the secondary processing state (normally, taken from the action service progress).</para>
		/// 
		/// <para>To avoid mistakes, this must be false by default and must be set with true only for individual action requests.
		/// </para>
		/// </remarks>
		public override bool ServiceResponseIsConsistentWithStandardConstructor
		{
			get { return true; }
		}

		/// <summary>
		/// The sheaf which is transported by the current action request, and also the referred sheaf from the associated action
		/// progress. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The current action request should be deleted when this sheaf is deleted.</para>
		/// </remarks>
		public override ISheaf TransportedSheaf
		{
			get { return Letter; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Letter _Letter;
		/// <summary>
		/// Letter. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Letter Letter
		{
			get { return _Letter; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Letter" );

				_Letter = value;
			}
		}

		#endregion Public serializable properties
	}
}
