////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Contract for an action response associated to an action request which was processed in the name of an initiator account
	/// and for the benefit a beneficiary account.
	/// </summary>
	public abstract class ActionResponseWithBeneficiary : ActionResponseWithInitiator
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="BeneficiaryUpnFingerprint"/> property.
		/// </summary>
		public const string BeneficiaryUpnFingerprintFilterIdentifier
			= "ActionResponseWithBeneficiary.BeneficiaryUpnFingerprint#8";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionResponseWithBeneficiary()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="actionRequest">The action request to which the current action response is associated.</param>
		/// <param name="secondaryProcessingState">Secondary processing state. This may be null.</param>
		/// <remarks>
		/// <para>This is the standard constructor for an action response, and is mandatory if the associated action request has
		/// its <see cref="ActionRequest.ServiceResponseIsConsistentWithStandardConstructor"/> property true, because it may be
		/// invoked dynamically (if the request was already processed by the action service).</para>
		/// </remarks>
		public ActionResponseWithBeneficiary( Locator instanceLocator, ActionRequestWithBeneficiary actionRequest
				, SecondaryProcessingState secondaryProcessingState )
			: base( instanceLocator, actionRequest, secondaryProcessingState )
		{
			BeneficiaryUpnFingerprint = actionRequest.BeneficiaryUpnFingerprint;
		}

		#endregion Constructors

		#region Public serializable properties

		private Locator _BeneficiaryUpnFingerprint;
		/// <summary>
		/// The fingerprint of the UPN of the beneficiary of the associated action request. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( BeneficiaryUpnFingerprintFilterIdentifier )]
		public Locator BeneficiaryUpnFingerprint
		{
			get { return _BeneficiaryUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "BeneficiaryUpnFingerprint" );

				_BeneficiaryUpnFingerprint = value;
			}
		}

		#endregion Public serializable properties
	}
}
