////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// An action response for a request which had an error during processing.
	/// </summary>
	public class ErrorActionResponse : AnonymousActionResponse
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ErrorActionResponse()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="actionRequest">The action request to which the current action response is associated.</param>
		/// <param name="secondaryProcessingState">Secondary processing state. This may be null.</param>
		/// <param name="remoteProcessingIdentity">Remote processing identity publisher.</param>
		/// <param name="exception">The exception. This may be null.</param>
		/// <param name="isTimeoutException">If this is true then the exception represents a timeout. This is used only for
		/// exceptions which are not derived from <see cref="CrossServiceException"/> and are not fatal exceptions
		/// (see the <see cref="ExceptionUtils.FindFatalCrossServiceException"/> method for details).</param>
		/// <remarks>
		/// <para>The error action response is meant to be sent to an action client.</para>
		/// </remarks>
		public ErrorActionResponse( Locator instanceLocator, ActionRequest actionRequest
				, SecondaryProcessingState secondaryProcessingState, IIdentityPublisher remoteProcessingIdentity
				, Exception exception, bool isTimeoutException )
			: base( instanceLocator, actionRequest, secondaryProcessingState, remoteProcessingIdentity )
		{
			RemoteProcessingIdentity = remoteProcessingIdentity;

			CrossServiceException crossServiceException = ExceptionUtils.FindException<CrossServiceException>( exception );
			if( crossServiceException != null )
			{
				ErrorTypeName = crossServiceException.GetType().FullName;
				Message = ExceptionUtils.GetPresenterMessage( null, crossServiceException, true );
			}
			else
			{
				Exception fatalCrossServiceException = ExceptionUtils.FindFatalCrossServiceException( exception );
				if( fatalCrossServiceException != null )
				{
					// Generic fatal error. The exception's message is not sent to the client because it may contain sensitive
					// details.

					ErrorTypeName = typeof( FatalCrossServiceException ).GetType().FullName;
					Message = "A fatal error has occurred. Details can be found in the log of the recipient peer's application.";
				}
				else
				{
					ErrorTypeName = isTimeoutException ? typeof( Vela.Kernel.TimeoutException ).FullName : null;
					Message = isTimeoutException
						? "A timeout has occurred. Details can be found in the log of the recipient peer's application."
						: "An error has occurred. Details can be found in the log of the recipient peer's application.";
				}
			}
		}

		#endregion Constructors

		#region Public serializable properties

		private string _ErrorTypeName;
		/// <summary>
		/// The full name of the error's type. This may be null or empty. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string ErrorTypeName
		{
			get { return _ErrorTypeName; }
			protected set { _ErrorTypeName = value; }
		}

		private string _Message;
		/// <summary>
		/// The error message. This may be null or empty. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string Message
		{
			get { return _Message; }
			protected set { _Message = value; }
		}

		/// <summary>
		/// Legacy property which may be deleted after most users upgrade to version 1.3.2.0 or a newer one. READ-ONLY.
		/// </summary>
		[Obsolete]
		[SerializeAttribute( MustBeDeserialized = false )]
		public bool ShouldAvoidRetryActionRequest
		{
			get { return HasFatalCrossServiceException(); }
			protected set {; }
		}

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Checks if <see cref="ErrorTypeName"/> identifies a fatal exception.
		/// </summary>
		/// <returns>Returns true if <see cref="ErrorTypeName"/> identifies a fatal exception.</returns>
		public bool HasFatalCrossServiceException()
		{
			return ExceptionUtils.IsFatalCrossServiceException( ErrorTypeName );
		}

		/// <summary>
		/// Checks if <see cref="ErrorTypeName"/> identifies an exception derived from the specified exception.
		/// </summary>
		/// <param name="exception">Exception from which <see cref="ErrorTypeName"/> may derive.</param>
		/// <returns>Returns true if <see cref="ErrorTypeName"/> identifies an exception derived from the specified exception.
		/// </returns>
		public bool HasException( Type exception )
		{
			ValidationUtils.ObjectHasContent( exception, "exception" );

			return ExceptionUtils.IsDerivedFromException( ErrorTypeName, exception );
		}

		#endregion Public methods
	}
}
