////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Contract for an action response. An action response is sent to the initiator of an action request, as a response to the
	/// request.
	/// </summary>
	/// <remarks>
	/// <para>An action response must be signed with the processing identity of the service, so that it can be used as proof that
	/// the action has been processed, and symmetrically encrypted (in an envelope of type <see cref="SymmEncryptedEnvelope"/>,
	/// but not <see cref="UpnEncryptedEnvelope"/>) with the <see cref="ActionRequest.ResponseEncryptionKey"/> key from the
	/// action request.</para>
	/// 
	/// <para>The secondary frames of a multiframe action response may be unsigned if they are derived from
	/// <see cref="IMultiframeActionResponse"/>.</para>
	/// 
	/// <para>An action response may be of the following types: <see cref="ActionResponseWithInitiator"/> and
	/// <see cref="AnonymousActionResponse"/>. All the other responses must be derived from these types.</para>
	/// 
	/// <para>An action response may be multiframe (see <see cref="IActionRequestWithMultiframeResponse"/> and
	/// <see cref="IMultiframeActionResponse"/> for details).</para>
	/// </remarks>
	public abstract class ActionResponse : SignedDocument, IInitiatorUpnFingerprintContainer, IRecipientUpnFingerprintContainer
		, IExpirableDocument
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="InitiatorUpnFingerprint"/> property.
		/// </summary>
		public const string InitiatorUpnFingerprintFilterIdentifier = "ActionResponse.InitiatorUpnFingerprint#4";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="RecipientUpnFingerprint"/> property.
		/// </summary>
		public const string RecipientUpnFingerprintFilterIdentifier = "ActionResponse.RecipientUpnFingerprint#5";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="ActionRequestLocator"/> property.
		/// </summary>
		public const string ActionRequestLocatorFilterIdentifier = "ActionResponse.ActionRequestLocator#6";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="DocumentExpirationTime"/> property.
		/// </summary>
		public const string DocumentExpirationTimeFilterIdentifier = "ActionResponse.DocumentExpirationTime#7";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionResponse()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="actionRequest">The action request to which the current action response is associated.</param>
		/// <param name="secondaryProcessingState">Secondary processing state. This may be null.</param>
		/// <remarks>
		/// <para>This is the standard constructor for an action response, and is mandatory if the associated action request has
		/// its <see cref="ActionRequest.ServiceResponseIsConsistentWithStandardConstructor"/> property true, because it may be
		/// invoked dynamically (if the request was already processed by the action service).</para>
		/// </remarks>
		public ActionResponse( Locator instanceLocator, ActionRequest actionRequest
				, SecondaryProcessingState secondaryProcessingState )
			: base( instanceLocator, actionRequest.MacAlg, actionRequest.AsymmSigningPaddingAlg
				, actionRequest.SignatureHashSerializer )
		{
			ActionRequestLocator = actionRequest.InstanceLocator;
			InitiatorUpnFingerprint = actionRequest.InitiatorUpnFingerprint;
			RecipientUpnFingerprint = actionRequest.RecipientUpnFingerprint;
			DocumentExpirationTime = actionRequest.ExpirationTime;
			SecondaryProcessingState = secondaryProcessingState;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( base.ExpirationTime, DocumentExpirationTime ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Locator _ActionRequestLocator;
		/// <summary>
		/// The locator of the associated action request. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( ActionRequestLocatorFilterIdentifier )]
		public Locator ActionRequestLocator
		{
			get { return _ActionRequestLocator; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionRequestLocator" );

				_ActionRequestLocator = value;
			}
		}

		private Locator _InitiatorUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the initiator of the current object. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( InitiatorUpnFingerprintFilterIdentifier )]
		public Locator InitiatorUpnFingerprint
		{
			get { return _InitiatorUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "InitiatorUpnFingerprint" );

				_InitiatorUpnFingerprint = value;
			}
		}

		private Locator _RecipientUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the recipient of the current object. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( RecipientUpnFingerprintFilterIdentifier )]
		public Locator RecipientUpnFingerprint
		{
			get { return _RecipientUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RecipientUpnFingerprint" );

				_RecipientUpnFingerprint = value;
			}
		}

		private DateTime _DocumentExpirationTime = DateTime.MaxValue;
		/// <summary>
		/// The expiration time (in UTC format) of the document. The current document is valid before this time; from and after
		/// this time, it should not be used in processing. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( DocumentExpirationTimeFilterIdentifier )]
		public DateTime DocumentExpirationTime
		{
			get { return _DocumentExpirationTime; }
			protected set { _DocumentExpirationTime = value; }
		}

		private SecondaryProcessingState _SecondaryProcessingState;
		/// <summary>
		/// Secondary processing state. This may be null.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public SecondaryProcessingState SecondaryProcessingState
		{
			get { return _SecondaryProcessingState; }
			set { _SecondaryProcessingState = value; }
		}

		private string _SecondaryProcessingOutput;
		/// <summary>
		/// Is set with the output of the secondary processing. This may be null or empty.
		/// </summary>
		/// <remarks>
		/// <para>This may be used to send back to the client a large amount of data which should not be stored in an action
		/// progress (where the secondary processing state is stored).</para>
		/// 
		/// <para>It's highly recommended to treat this information as ephemeral (in terms of storage requirements).</para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string SecondaryProcessingOutput
		{
			get { return _SecondaryProcessingOutput; }
			set { _SecondaryProcessingOutput = value; }
		}

		#endregion Public serializable properties
	}
}
