////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Contract for an action response for an anonymous action request, that is, one without an initiator account.
	/// </summary>
	/// <remarks>
	/// <para>An action response derived from this class can't have the
	/// <see cref="ActionRequest.ServiceResponseIsConsistentWithStandardConstructor"/> property of its associated action request
	/// set with true because the constructor of the action response can't be the standard constructor.</para>
	/// </remarks>
	public abstract class AnonymousActionResponse : ActionResponse
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AnonymousActionResponse()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="actionRequest">The action request to which the current action response is associated.</param>
		/// <param name="secondaryProcessingState">Secondary processing state. This may be null.</param>
		/// <param name="remoteProcessingIdentity">Remote processing identity publisher.</param>
		protected AnonymousActionResponse( Locator instanceLocator, ActionRequest actionRequest
				, SecondaryProcessingState secondaryProcessingState, IIdentityPublisher remoteProcessingIdentity )
			: base( instanceLocator, actionRequest, secondaryProcessingState )
		{
			RemoteProcessingIdentity = remoteProcessingIdentity;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( base.ExpirationTime, RemoteProcessingIdentity.ExpirationTime ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private IIdentityPublisher _RemoteProcessingIdentity;
		/// <summary>
		/// Remote processing identity publisher. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This must exist because the initiator of the associated action request must be able verify the signature of
		/// the response. This is done in order to ensure that, when the action request was encrypted with the UPN, the real
		/// owner of the UPN has answered, and not someone else who knows the UPN (and who was able to decrypt the request and
		/// retrieve the encryption key for the response).</para>
		/// </remarks>
		[SerializeAttribute]
		public IIdentityPublisher RemoteProcessingIdentity
		{
			get { return _RemoteProcessingIdentity; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RemoteProcessingIdentity" );

				_RemoteProcessingIdentity = value;
			}
		}

		#endregion Public serializable properties
	}
}
