////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.PeerDocuments;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// An action response for a request to download files which were offered by the remote peer through a
	/// <see cref="SendLetterActionRequest"/> action request.
	/// </summary>
	public class DownloadFilesActionResponse : ActionResponseWithInitiator, IMultiframeActionResponse
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected DownloadFilesActionResponse()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="actionRequest">The action request to which the current action response is associated.</param>
		/// <param name="secondaryProcessingState">Secondary processing state. This may be null.</param>
		/// <param name="fileName">The name of the file to download.</param>
		/// <param name="fromPosition">The position (in the file to download) from which to download data.</param>
		/// <param name="filePart">The file part to download.</param>
		/// <param name="disalignmentData">Data used to randomize the size of the current response. This may be null or empty.
		/// </param>
		public DownloadFilesActionResponse( Locator instanceLocator, ActionRequestWithInitiator actionRequest
				, SecondaryProcessingState secondaryProcessingState, AttachedFileName fileName, Int64 fromPosition
				, byte[] filePart, byte[] disalignmentData )
			: base( instanceLocator, actionRequest, secondaryProcessingState )
		{
			FileName = fileName;
			FromPosition = fromPosition;
			FilePart = filePart;
			DisalignmentData = disalignmentData;
		}

		#endregion Constructors

		#region Public serializable properties

		private AttachedFileName _FileName;
		/// <summary>
		/// The name of the file to download. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public AttachedFileName FileName
		{
			get { return _FileName; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "FileName" );

				_FileName = value;
			}
		}

		private Int64 _FromPosition;
		/// <summary>
		/// The position (in the file to download) from which to download data. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>Before the current request is sent to the service, the client checks to see how much of the file has already
		/// been saved, and requests the download to resume from the next byte.</para>
		/// </remarks>
		[SerializeAttribute]
		public Int64 FromPosition
		{
			get { return _FromPosition; }
			protected set
			{
				ValidationUtils.Integer( value, 0, Int64.MaxValue - 1, "FromPosition" );

				_FromPosition = value;
			}
		}

		private byte[] _FilePart;
		/// <summary>
		/// The file part to download. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] FilePart
		{
			get { return _FilePart; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "FilePart" );

				_FilePart = value;
			}
		}

		private byte[] _DisalignmentData;
		/// <summary>
		/// Data used to randomize the size of the current response. This may be null or empty. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>Since the current action response may contain a part of only a single file (this is done in order to reduce the
		/// complexity, not because it's otherwise impossible), and since the last part of a file would normally be smaller than
		/// the minimum size (<see cref="ActionServiceConfiguration.DownloadFilePartSizeMin"/>) of the current response, whithout
		/// this buffer the last response frame would reveal the end of the currently transferred file. Also, files smaller than
		/// the minimum size would be exposed in the same way.</para>
		/// 
		/// <para>This buffer is used to fill the current response up to the usual random size.</para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false )]
		public byte[] DisalignmentData
		{
			get { return _DisalignmentData; }
			protected set { _DisalignmentData = value; }
		}

		#endregion Public serializable properties
	}
}
