////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Action service configuration.
	/// </summary>
	public class ActionServiceConfiguration : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public ActionServiceConfiguration()
		{
		}

		#endregion Constructors

		#region Public serializable properties

		private int _QueuedConnectionsMax = ActionProcessorConsts.QueuedConnectionsMax;
		/// <summary>
		/// The maximum number of connections which may be queued while waiting for processing. This must be at least 1.
		/// </summary>
		[SerializeAttribute]
		public int QueuedConnectionsMax
		{
			get { return _QueuedConnectionsMax; }
			set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "QueuedConnectionsMax" );

				_QueuedConnectionsMax = value;
			}
		}

		private TrafficFilterConfiguration _TrafficFilterConfiguration = new TrafficFilterConfiguration();
		/// <summary>
		/// The traffic filter configuration. The backing field is initialized by default. For legacy reasons, this may be null,
		/// but may not be set with null.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public TrafficFilterConfiguration TrafficFilterConfiguration
		{
			get { return _TrafficFilterConfiguration; }
			set
			{
				ValidationUtils.ObjectHasContent( value, "TrafficFilterConfiguration" );

				_TrafficFilterConfiguration = value;
			}
		}

		private double _TimingAttackProtectionRoundoff = ActionProcessorConsts.TimingAttackProtectionRoundoffDef;
		/// <summary>
		/// The rounded timespan (in seconds) up to which the service must wait as a form of protection against timing
		/// attacks. This must be at least <see cref="ActionProcessorConsts.TimingAttackProtectionRoundoffMin"/>.
		/// </summary>
		[SerializeAttribute]
		public double TimingAttackProtectionRoundoff
		{
			get { return _TimingAttackProtectionRoundoff; }
			set
			{
				ValidationUtils.Real( value, ActionProcessorConsts.TimingAttackProtectionRoundoffMin
					, ActionProcessorConsts.TimingAttackProtectionRoundoffMax, "TimingAttackProtectionRoundoff" );

				_TimingAttackProtectionRoundoff = value;
			}
		}

		private int _DownloadFilePartSizeMin = ActionProcessorConsts.DownloadFilePartSizeMin;
		/// <summary>
		/// The minimum size (in bytes) that the part of a file to download must have. This must be at least 1.
		/// </summary>
		[SerializeAttribute]
		public int DownloadFilePartSizeMin
		{
			get { return _DownloadFilePartSizeMin; }
			set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "DownloadFilePartSizeMin" );

				_DownloadFilePartSizeMin = value;
			}
		}

		private int _DownloadFilePartSizeMax = ActionProcessorConsts.DownloadFilePartSizeMax;
		/// <summary>
		/// The maximum size (in bytes) that the part of a file to download must have. This must be at least
		/// <see cref="DownloadFilePartSizeMin"/>.
		/// </summary>
		[SerializeAttribute]
		public int DownloadFilePartSizeMax
		{
			get { return _DownloadFilePartSizeMax; }
			set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "DownloadFilePartSizeMax" );

				_DownloadFilePartSizeMax = value;
			}
		}

		private FillerUpnClassifiersConfiguration _FillerUpnClassifiers = new FillerUpnClassifiersConfiguration();
		/// <summary>
		/// Filler UPN classifiers configuration. For legacy reasons, this may be null, but may not be set with null.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public FillerUpnClassifiersConfiguration FillerUpnClassifiers
		{
			get { return _FillerUpnClassifiers; }
			set
			{
				ValidationUtils.ObjectHasContent( value, "FillerUpnClassifiers" );

				_FillerUpnClassifiers = value;
			}
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Validates the current object.
		/// </summary>
		protected virtual void Validate()
		{
			ValidationUtils.Integer( DownloadFilePartSizeMax, DownloadFilePartSizeMin, int.MaxValue, "DownloadFilePartSizeMax" );
		}

		/// <summary>
		/// This method is automatically called before the current object is serialized.
		/// </summary>
		protected virtual void Serializing()
		{
			Validate();
		}

		/// <summary>
		/// This method is automatically called after the current object has been deserialized.
		/// </summary>
		protected virtual void Deserialized()
		{
			Validate();
		}

		#endregion Private methods
	}
}
