////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.PeerDocuments;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Host context used for updating the user interface of an action service application after an action request and response
	/// are processed.
	/// </summary>
	public class ActionServiceHostUpdateUiContext
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionServiceHostUpdateUiContext()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="configuration">Configuration for the action service.</param>
		/// <param name="serviceVault">The service vault.</param>
		/// <param name="initiatorTally">The tally of the initiator of the action request which is currently being processed by
		/// the service. This may be null.</param>
		/// <param name="initiatorAccount">The account of the initiator of the action request which is currently being processed
		/// by the service. This may be null.</param>
		/// <param name="actionRequest">Action request which is currently being processed by the service.</param>
		/// <param name="actionResponse">Action response to send to the client.</param>
		/// <param name="mustFullyUpdateUi">If this is true then the host must fully update its UI. If this is false then the
		/// action service is processing a frame of a multiframe action response, so the host only has to perform a partial UI
		/// update (for example just the progress of a file transfer).</param>
		public ActionServiceHostUpdateUiContext( ActionServiceConfiguration configuration, Vault serviceVault
			, Tally initiatorTally, Account initiatorAccount, ActionRequest actionRequest, ActionResponse actionResponse
			, bool mustFullyUpdateUi )
		{
			Configuration = configuration;
			ServiceVault = serviceVault;
			InitiatorTally = initiatorTally;
			InitiatorAccount = initiatorAccount;
			ActionRequest = actionRequest;
			ActionResponse = actionResponse;
			MustFullyUpdateUi = mustFullyUpdateUi;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private ActionServiceConfiguration _Configuration;
		/// <summary>
		/// The action service configuration. READ-ONLY.
		/// </summary>
		public ActionServiceConfiguration Configuration
		{
			get { return _Configuration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Configuration" );

				_Configuration = value;
			}
		}

		private Vault _ServiceVault;
		/// <summary>
		/// The service vault. READ-ONLY.
		/// </summary>
		public Vault ServiceVault
		{
			get { return _ServiceVault; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ServiceVault" );

				_ServiceVault = value;
			}
		}

		private Tally _InitiatorTally;
		/// <summary>
		/// The tally of the initiator of the action request which is currently being processed by the service. This may be null.
		/// READ-ONLY.
		/// </summary>
		public Tally InitiatorTally
		{
			get { return _InitiatorTally; }
			protected set { _InitiatorTally = value; }
		}

		private Account _InitiatorAccount;
		/// <summary>
		/// The account of the initiator of the action request which is currently being processed by the service. This may be
		/// null. READ-ONLY.
		/// </summary>
		public Account InitiatorAccount
		{
			get { return _InitiatorAccount; }
			protected set { _InitiatorAccount = value; }
		}

		private ActionRequest _ActionRequest;
		/// <summary>
		/// Action request which is currently being processed by the service. READ-ONLY.
		/// </summary>
		public ActionRequest ActionRequest
		{
			get { return _ActionRequest; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionRequest" );

				_ActionRequest = value;
			}
		}

		private ActionResponse _ActionResponse;
		/// <summary>
		/// Action response to send to the client. READ-ONLY.
		/// </summary>
		public ActionResponse ActionResponse
		{
			get { return _ActionResponse; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionResponse" );

				_ActionResponse = value;
			}
		}

		private bool _MustFullyUpdateUi;
		/// <summary>
		/// If this is true then the host must fully update its UI. If this is false then the action service is processing a
		/// frame of a multiframe action response, so the host only has to perform a partial UI update (for example just the
		/// progress of a file transfer). READ-ONLY.
		/// </summary>
		public bool MustFullyUpdateUi
		{
			get { return _MustFullyUpdateUi; }
			protected set { _MustFullyUpdateUi = value; }
		}

		#endregion Public non-serializable properties
	}
}
