////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Cloning;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Action service configuration.
	/// </summary>
	public class ActionServiceListenerConfiguration : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionServiceListenerConfiguration()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="communicationFrameAlg">The communication frame algorithm of a service.</param>
		/// <param name="processingPort">The network port of a service where action requests are to be sent by a client.
		/// </param>
		public ActionServiceListenerConfiguration( ICommunicationFrameAlg communicationFrameAlg, UInt16 processingPort )
		{
			CommunicationFrameAlg = communicationFrameAlg;
			ProcessingPort = processingPort;
		}

		#endregion Constructors

		#region Public serializable properties

		private DateTime _CreationTime = DateTime.UtcNow;
		/// <summary>
		/// The time (in UTC format) when the current object was created. This is automatically set with the time of the
		/// creation of the current object (and you should not modify it). READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public DateTime CreationTime
		{
			get { return _CreationTime; }
			protected set { _CreationTime = value; }
		}

		private ICommunicationFrameAlg _CommunicationFrameAlg;
		/// <summary>
		/// The communication frame algorithm which a client must use to send action requests. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This algorithm is copied in each appendix from the storehouse which contains the current action service
		/// configuration.</para>
		/// 
		/// <para>Once a storehouse is created with the current configuration, if you were to change this algorithm after you
		/// gave to other people some UPNs of the vaults from the storehouse, those people would no longer be able to
		/// communicate with the respective vaults.</para>
		/// </remarks>
		[SerializeAttribute]
		public ICommunicationFrameAlg CommunicationFrameAlg
		{
			get { return _CommunicationFrameAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "CommunicationFrameAlg" );

				_CommunicationFrameAlg = value;
			}
		}

		private UInt16 _ProcessingPort = ActionProcessorConsts.ServiceProcessingPortDef;
		/// <summary>
		/// The network port of a service where action requests are to be sent by a client. This must be at least 1.
		/// </summary>
		/// <remarks>
		/// <para>When a storehouse is being created, this port may be randomly generated.</para>
		/// 
		/// <para>This port is copied in each appendix from the storehouse which contains the current action service
		/// configuration.</para>
		/// 
		/// <para>Once a storehouse is created with the current configuration, if you were to change this port after you gave
		/// to other people some UPNs of the vaults from the storehouse, those people would no longer be able to communicate with
		/// the respective vaults.</para>
		/// </remarks>
		[SerializeAttribute]
		public UInt16 ProcessingPort
		{
			get { return _ProcessingPort; }
			set
			{
				ValidationUtils.Integer( value, 1, UInt16.MaxValue, "ProcessingPort" );

				_ProcessingPort = value;
			}
		}

		#endregion Public serializable properties

		#region Public static methods

		/// <summary>
		/// Creates a new instance of the current class, by cloning the specified object and port, setting the
		/// <see cref="ActionServiceListenerConfiguration.CreationTime"/> to minimum (because this configuration must not have
		/// priority over the non-default ports, when a configuration is chosen for a new vault).
		/// </summary>
		/// <param name="configuration">The configuration from which to create the new one.</param>
		/// <param name="processingPort">The network port of a service where action requests are to be sent by a client.</param>
		/// <returns>Returns a new instance of the current class.</returns>
		public static ActionServiceListenerConfiguration CreateWithPort( ActionServiceListenerConfiguration configuration
			, UInt16 processingPort )
		{
			ActionServiceListenerConfiguration newConfiguration = Cloner.Deep( configuration );

			newConfiguration.CreationTime = DateTime.UtcNow;
			newConfiguration.ProcessingPort = processingPort;

			return newConfiguration;
		}

		#endregion Public static methods
	}
}
