////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// List of action service listener configurations (<see cref="ActionServiceListenerConfiguration"/>).
	/// </summary>
	public class ActionServiceListenerConfigurations : KernelSortedList<int, ActionServiceListenerConfiguration>
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public ActionServiceListenerConfigurations()
			: base( IntComparer.Instance )
		{
		}

		#endregion Constructors

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para><see cref="ActionServiceListenerConfiguration.ProcessingPort"/> is used as key.</para>
		/// </remarks>
		public override int GetKey( ActionServiceListenerConfiguration item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.ProcessingPort;
		}

		/// <summary>
		/// Gets the item from the list which was created most recently.
		/// </summary>
		/// <returns>Returns the item from the list which was created most recently. This is null if the list is empty.</returns>
		public ActionServiceListenerConfiguration GetMostRecent()
		{
			DateTime maxTime = DateTime.MinValue;
			ActionServiceListenerConfiguration recentConfiguration = null;

			foreach( ActionServiceListenerConfiguration configuration in this )
			{
				if( configuration.CreationTime >= maxTime )
				{
					maxTime = configuration.CreationTime;
					recentConfiguration = configuration;
				}
			}

			return recentConfiguration;
		}

		/// <summary>
		/// Adds or updates a list item for the specified port.
		/// </summary>
		/// <param name="processingPort">Processing port for which to add a new list item. This must be at least 1.</param>
		/// <param name="removeOlderProcessingPorts">If this is true then all the processing ports other than "processingPort"
		/// are removed from the current storehouse configuration.</param>
		/// <returns>Returns the added / updated / existing list item.</returns>
		/// <remarks>
		/// <para>If a list item with the same processing port already exists, it's updated to be the most recent item.</para>
		/// </remarks>
		public ActionServiceListenerConfiguration AddOrUpdateAsMostRecent( UInt16 processingPort
			, bool removeOlderProcessingPorts )
		{
			ValidationUtils.Integer( processingPort, 1, UInt16.MaxValue, "processingPort" );

			ActionServiceListenerConfiguration newConfiguration;

			ActionServiceListenerConfiguration configuration = this[ processingPort ];
			if( configuration != null )
			{
				if( configuration != GetMostRecent() )
				{
					if( removeOlderProcessingPorts )
						Clear();

					newConfiguration = ActionServiceListenerConfiguration.CreateWithPort( configuration, processingPort );
					AddOrReplace( configuration, newConfiguration );

					return newConfiguration;
				}
				else
				{
					if( removeOlderProcessingPorts )
					{
						Clear();

						Add( configuration );
					}

					return configuration;
				}
			}
			else
			{
				configuration = GetMostRecent();

				if( removeOlderProcessingPorts )
					Clear();

				newConfiguration = ActionServiceListenerConfiguration.CreateWithPort( configuration, processingPort );
				Add( newConfiguration );

				return newConfiguration;
			}
		}

		#endregion Public methods
	}
}
