////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Configuration for the network registration of the filler UPN classifiers.
	/// </summary>
	/// <remarks>
	/// <para>The filler UPN classifiers are used strictly to create a larger peer network because the PNRP has problems
	/// resolving remote peer names when just one local peer name is registered. If the PNRP still can't find peers, increase
	/// the <see cref="FillerUpnClassifiersConfiguration.ClassifiersCount"/> value; it's useful if the remote peers also
	/// increase this value.</para>
	/// 
	/// <para>Also, because the filler classifiers have the same size and structure as normal UPN classifiers, and because a
	/// fixed number of them is registered on the network, an attacker can't know how many vaults a peer has. However, if the
	/// peer creates a new vault, the attacker might see a change in the registered classifiers.</para>
	/// </remarks>
	public class FillerUpnClassifiersConfiguration : IDocument
	{
		#region Public constants

		/// <summary>
		/// The minimum value of <see cref="ClassifiersCount"/>. This must be at least 0.
		/// </summary>
		public const int ClassifiersCountMin = 0;

		/// <summary>
		/// The maximum value of <see cref="ClassifiersCount"/>. This must be at least <see cref="ClassifiersCountMin"/>.
		/// </summary>
		public const int ClassifiersCountMax = 500;

		/// <summary>
		/// The minimum initial value of <see cref="ClassifiersCount"/>.
		/// </summary>
		public const int ClassifiersCountInitialMin = 5;

		/// <summary>
		/// The maximum initial value of <see cref="ClassifiersCount"/>.
		/// </summary>
		public const int ClassifiersCountInitialMax = 10;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public FillerUpnClassifiersConfiguration()
		{
			ClassifiersCount = EntropyPool.GetInt( ClassifiersCountInitialMin, ClassifiersCountInitialMax, false );

			FillClassifiersList();
		}

		#endregion Constructors

		#region Public serializable properties

		private int _ClassifiersCount;
		/// <summary>
		/// The number of classifiers to register.
		/// </summary>
		/// <remarks>
		/// <para>The filler UPN classifiers will be registered on the network up to (at least) this value.</para>
		/// </remarks>
		[SerializeAttribute]
		public int ClassifiersCount
		{
			get { return _ClassifiersCount; }
			set
			{
				ValidationUtils.Integer( value, ClassifiersCountMin, ClassifiersCountMax, "ClassifiersCount" );

				_ClassifiersCount = value;
			}
		}

		private Strings _Classifiers = new Strings();
		/// <summary>
		/// The filler UPN classifiers to register on the network. The backing field is initialized by default.
		/// </summary>
		[SerializeAttribute]
		public Strings Classifiers
		{
			get { return _Classifiers; }
			set
			{
				ValidationUtils.ObjectHasContent( value, "Classifiers" );

				_Classifiers = value;
			}
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Generates a unique random UPN classifier.
		/// </summary>
		/// <returns>Returns a unique random UPN classifier.</returns>
		private string GenerateRandomClassifier()
		{
			while( true )
			{
				// The filler classifiers should have the same size and structure as the normal UPN classifiers. See the
				// comment of the "CoreDocumentsConsts.UpnClassifierSizeMax" constant to understand why.
				string classifier = EntropyPool.GetText16( CoreDocumentsConsts.UpnClassifierSizeMax, false );

				if( !Classifiers.ContainsKey( classifier ) )
					return classifier;
			}
		}

		/// <summary>
		/// This method is automatically called before the current object is serialized.
		/// </summary>
		protected virtual void Serializing()
		{
			FillClassifiersList();
		}

		/// <summary>
		/// This method is automatically called after the current object has been deserialized.
		/// </summary>
		protected virtual void Deserialized()
		{
			FillClassifiersList();
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Fills the list of classifiers up to the <see cref="ClassifiersCount"/> limit.
		/// </summary>
		public void FillClassifiersList()
		{
			int countRequired = ClassifiersCount - Classifiers.Count;

			if( countRequired <= 0 )
				return;

			for( int i = 0; i < countRequired; i++ )
				Classifiers.Add( GenerateRandomClassifier() );
		}

		#endregion Public methods
	}
}
