////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Traffic filter configuration for an action service.
	/// </summary>
	public class TrafficFilterConfiguration : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public TrafficFilterConfiguration()
		{
		}

		#endregion Constructors

		#region Public serializable properties

		private int _RecordsCountMax = ActionProcessorConsts.TrafficFilterRecordsCountMax;
		/// <summary>
		/// The maximum number of records kept by an action traffic filter. This must be at least 1.
		/// </summary>
		public int RecordsCountMax
		{
			get { return _RecordsCountMax; }
			set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "RecordsCountMax" );

				_RecordsCountMax = value;
			}
		}

		private int _RecordResetTimespan = ActionProcessorConsts.TrafficFilterRecordResetTimespan;
		/// <summary>
		/// The timespan (in hours) during which the network traffic of a peer is limited by an action service. This must be at
		/// least 1.
		/// </summary>
		[SerializeAttribute]
		public int RecordResetTimespan
		{
			get { return _RecordResetTimespan; }
			set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "RecordResetTimespan" );

				_RecordResetTimespan = value;
			}
		}

		private UInt64 _AnonymousPeerIncomingRequestsCountMax
			= ActionProcessorConsts.TrafficFilterAnonymousPeerIncomingRequestsCountMax;
		/// <summary>
		/// The maximum number of network requests which may be accepted by an action service from an anonymous peer, during the
		/// <see cref="RecordResetTimespan"/> timespan. This must be at least 1.
		/// </summary>
		/// <remarks>
		/// <para>Action requests may be preceded by a request to resolve the UPN, so this may also increase the number of
		/// incoming requests.</para>
		/// 
		/// <para>Some network requests may be cut in the middle, but they are still counted, so you should use a value higher
		/// than 1.</para>
		/// 
		/// <para>This value should be higher than the number of retries made by a remote peer to create an account with the
		/// local peer.</para>
		/// </remarks>
		[SerializeAttribute]
		public UInt64 AnonymousPeerIncomingRequestsCountMax
		{
			get { return _AnonymousPeerIncomingRequestsCountMax; }
			set
			{
				ValidationUtils.Integer( value, 1, UInt64.MaxValue, "AnonymousPeerIncomingRequestsCountMax" );

				_AnonymousPeerIncomingRequestsCountMax = value;
			}
		}

		private UInt64 _KnownPeerProcessableSizeMax = ActionProcessorConsts.TrafficFilterKnownPeerProcessableSizeMax;
		/// <summary>
		/// The maximum size (in bytes) of the data which may be processed by an action service for a known (= non-anonymous)
		/// peer, during the <see cref="RecordResetTimespan"/> timespan. This must be at least
		/// <see cref="ActionProcessorConfiguration.CommunicationFrameSizeMax"/>.
		/// </summary>
		/// <remarks>
		/// <para>Keep in mind that a large part of the data coming from known peers may be saved in the storehouse, so don't
		/// set this value too high.</para>
		/// </remarks>
		[SerializeAttribute]
		public UInt64 KnownPeerProcessableSizeMax
		{
			get { return _KnownPeerProcessableSizeMax; }
			set
			{
				ValidationUtils.Integer( value, 1, UInt64.MaxValue, "KnownPeerProcessableSizeMax" );

				_KnownPeerProcessableSizeMax = value;
			}
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Validates the current object.
		/// </summary>
		protected virtual void Validate()
		{
			ValidationUtils.Integer( AnonymousPeerIncomingRequestsCountMax, 1, UInt64.MaxValue
				, "AnonymousPeerIncomingRequestsCountMax" );
			ValidationUtils.Integer( KnownPeerProcessableSizeMax, 1, UInt64.MaxValue, "KnownPeerProcessableSizeMax" );
		}

		/// <summary>
		/// This method is automatically called before the current object is serialized.
		/// </summary>
		protected virtual void Serializing()
		{
			Validate();
		}

		/// <summary>
		/// This method is automatically called after the current object has been deserialized.
		/// </summary>
		protected virtual void Deserialized()
		{
			Validate();
		}

		#endregion Private methods
	}
}
