////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using Vela.Kernel;
using Vela.PeerDocuments;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Utilities for attached files.
	/// </summary>
	public static class AttachedFilesUtils
	{
		#region Public constants

		/// <summary>
		/// The storage space (in bytes) which must remain free after attached files are saved.
		/// </summary>
		public const Int64 SaveFileBufferSpace = 2000000000;

		/// <summary>
		/// Template for the title of the folder, on the local peer, where to save an attached file. Placeholders: 0 - the
		/// creation time of the document which has offered the attached files, 1 - the locator of the document which has offered
		/// the attached files.
		/// </summary>
		public const string SaveFolderTitleTemplate = "{0:yyyyMMddHHmmss}-{1}";

		/// <summary>
		/// The suffix for the name of an unsafe file extension.
		/// </summary>
		/// <remarks>
		/// <para>The save name of an attached file whose extension isn't in the <see cref="SafeFileExtensions"/> list, is
		/// suffixed with this text.</para>
		/// </remarks>
		public const string SaveFileNameSuffixForUnsafeFile = "_";

		/// <summary>
		/// File extensions for attached files which are considered safe to be opened by the recipient without being scanned for
		/// malicious code.
		/// </summary>
		/// <remarks>
		/// <para><see cref="FileExtensions.VelaFile"/> and <see cref="FileExtensions.MediaFile"/> are merged into this.</para>
		/// 
		/// <para>The save name of an attached file whose extension isn't in this list, is suffixed with
		/// <see cref="SaveFileNameSuffixForUnsafeFile"/>.</para>
		/// </remarks>
		public static readonly HashSet<string> SafeFileExtensions = new HashSet<string> { ".7z", ".bin", ".cpp", ".cs", ".dat"
			, ".dll", ".h", ".hpp", ".pas", ".txt", ".xml", ".zip" };

		#endregion Public constants

		#region Static constructors

		/// <summary>
		/// Initializes the current class.
		/// </summary>
		static AttachedFilesUtils()
		{
			SafeFileExtensions.UnionWith( FileExtensions.VelaFile );
			SafeFileExtensions.UnionWith( FileExtensions.MediaFile );
		}

		#endregion Static constructors

		#region Public static methods

		/// <summary>
		/// Gets the name of the folder, on the local peer, where to save the files attached to the document with the specified
		/// locator.
		/// </summary>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		/// <param name="accessGranterLocator">The locator of the document which has offered the attached files.</param>
		/// <param name="accessGranterCreationTime">The creation time of the document which has offered the attached files.
		/// </param>
		/// <returns>Returns the name of the folder, on the local peer, where to save the files.</returns>
		public static string GetSaveFolderName( string rootSaveFolderName, Locator accessGranterLocator
			, DateTime accessGranterCreationTime )
		{
			ValidationUtils.TextHasContent( rootSaveFolderName, "rootSaveFolderName" );
			ValidationUtils.ObjectHasContent( accessGranterLocator, "accessGranterLocator" );

			string accessGranterLocatorText = TextUtils.ArrayToText32( accessGranterLocator.Value );

			string saveFolderTitle = string.Format( SaveFolderTitleTemplate, accessGranterCreationTime.ToLocalTime()
				, accessGranterLocatorText );

			return FileUtils.MakeFullName( rootSaveFolderName, saveFolderTitle );
		}

		/// <summary>
		/// Gets the name of the file, on the local peer, where to save the specified attached file.
		/// </summary>
		/// <param name="relativeFileName">The relative name of the attached file for which to get the save file name.</param>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		/// <param name="accessGranterLocator">The locator of the document which has offered the attached files.</param>
		/// <param name="accessGranterCreationTime">The creation time of the document which has offered the attached files.
		/// </param>
		/// <returns>Returns the name of the file, on the local peer, where to save the specified attached file.</returns>
		public static string GetSaveFileName( string relativeFileName, string rootSaveFolderName, Locator accessGranterLocator
			, DateTime accessGranterCreationTime )
		{
			ValidationUtils.TextHasContent( relativeFileName, "relativeFileName" );
			ValidationUtils.TextHasContent( rootSaveFolderName, "rootSaveFolderName" );
			ValidationUtils.ObjectHasContent( accessGranterLocator, "accessGranterLocator" );

			string fileExtension = Path.GetExtension( relativeFileName );
			if( TextUtils.HasContent( fileExtension ) )
				fileExtension = fileExtension.ToLowerInvariant();

			string nameSuffix =
				(
					SafeFileExtensions.Contains( fileExtension )
					|| fileExtension.EndsWith( SaveFileNameSuffixForUnsafeFile, StringComparison.OrdinalIgnoreCase )
				) ? "" : SaveFileNameSuffixForUnsafeFile;

			string saveFolderName = GetSaveFolderName( rootSaveFolderName, accessGranterLocator, accessGranterCreationTime );

			return FileUtils.MakeFullName( saveFolderName, relativeFileName + nameSuffix );
		}

		/// <summary>
		/// Checks if any of the same files from the specified list has changed (or doesn't exist).
		/// </summary>
		/// <param name="attachedFiles">The attached files. This may be empty, but not null.</param>
		/// <returns>Returns the name of the file which has changed (or doesn't exist), or null if none has changed.</returns>
		public static string CheckIfChangedOriginalFile( AttachedFiles attachedFiles )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );

			foreach( AttachedFile attachedFile in attachedFiles )
			{
				FileInfo fi = new FileInfo( attachedFile.AbsoluteFileName );

				if( !fi.Exists || (attachedFile.Size != fi.Length) || (attachedFile.LastChangeTime != fi.LastWriteTimeUtc) )
					return attachedFile.AbsoluteFileName;
			}

			return null;
		}

		/// <summary>
		/// Checks if the file with the specified name has changed (or doesn't exist).
		/// </summary>
		/// <param name="attachedFiles">The attached files. This may be empty, but not null.</param>
		/// <param name="fileName">The name of the file to check if it has changed (or doesn't exist).</param>
		/// <returns>Returns true if the file with the specified name has changed (or doesn't exist).</returns>
		public static bool CheckIfChangedOriginalFile( AttachedFiles attachedFiles, AttachedFileName fileName )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );
			ValidationUtils.ObjectHasContent( fileName, "fileName" );

			AttachedFile attachedFile = attachedFiles[ fileName ];
			if( attachedFile == null )
				return true;

			FileInfo fi = new FileInfo( fileName.AbsoluteFileName );

			return !fi.Exists || (attachedFile.Size != fi.Length) || (attachedFile.LastChangeTime != fi.LastWriteTimeUtc);
		}

		/// <summary>
		/// Gets the attached file to send to the initiator of the file transfer.
		/// </summary>
		/// <param name="attachedFiles">The attached files. This may be empty, but not null.</param>
		/// <returns>Returns the attached file. If this is null, no file has to be transferred.</returns>
		/// <remarks>
		/// <para>The returned attached file is the first one whose current size is smaller than what it should be.</para>
		/// 
		/// <para>The first <see cref="PeerDocumentsConsts.ShowPreviewsForAttachedImageFilesCountMax"/> image files have
		/// priority, so that they can be downloaded first, so that image previews can be shown to the recipient of the attached
		/// files as soon as possible.</para>
		/// </remarks>
		public static AttachedFile GetAttachedFileToSend( AttachedFiles attachedFiles )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );

			#region Prioritize the image files which can be used for image previews.

			AttachedFiles potentialFilesForImagePreview = GetPotentialFilesForImagePreview( attachedFiles
				, PeerDocumentsConsts.ShowPreviewsForAttachedImageFilesCountMax );

			foreach( AttachedFile attachedFile in potentialFilesForImagePreview )
			{
				// "attachedFiles" is coming from the recipient (= the action client) with the same paths as those which were
				// sent from the sender (= the current action service). This means that the "attachedFile.SortKey" matches
				// correctly.
				AttachedFile attachedFileToSend = attachedFiles[ attachedFile.SortKey ];

				// If the current size of the file is less than what it should be, transfer the file.
				if( (attachedFileToSend != null) && (attachedFileToSend.CurrentSize < attachedFileToSend.Size) )
					return attachedFileToSend;
			}

			#endregion Prioritize the image files which can be used for image previews.

			foreach( AttachedFile attachedFile in attachedFiles )
			{
				// If the current size of the file is less than what it should be, transfer the file.
				if( attachedFile.CurrentSize < attachedFile.Size )
					return attachedFile;
			}

			// There is no attached file.
			return null;
		}

		/// <summary>
		/// Gets the attached file whose transfer to initiate.
		/// </summary>
		/// <param name="attachedFiles">The attached files. This may be empty, but not null.</param>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		/// <param name="accessGranterLocator">The locator of the document which has offered the attached files.</param>
		/// <param name="accessGranterCreationTime">The creation time of the document which has offered the attached files.
		/// </param>
		/// <param name="fromPosition">Is set with the position in the file from which data has to be transfer.</param>
		/// <returns>Returns the attached file. If this is null, no file has to be transferred.</returns>
		/// <remarks>
		/// <para>The returned attached file is the first one whose current size is smaller than what it should be.</para>
		/// </remarks>
		public static AttachedFile GetAttachedFileToSave( AttachedFiles attachedFiles, string rootSaveFolderName
			, Locator accessGranterLocator, DateTime accessGranterCreationTime, out Int64 fromPosition )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );
			ValidationUtils.TextHasContent( rootSaveFolderName, "rootSaveFolderName" );
			ValidationUtils.ObjectHasContent( accessGranterLocator, "accessGranterLocator" );

			fromPosition = 0;

			foreach( AttachedFile attachedFile in attachedFiles )
			{
				string saveFileName = GetSaveFileName( attachedFile.RelativeFileName, rootSaveFolderName, accessGranterLocator
					, accessGranterCreationTime );

				// If the save file doesn't exist yet, transfer it.
				if( !File.Exists( saveFileName ) )
					return attachedFile;

				// If the current size of the save file is equal with (or greater than) what it should be, skip the file.
				Int64 currentFileSize = FileUtils.GetFileSize( saveFileName );
				if( currentFileSize >= attachedFile.Size )
					continue;

				fromPosition = currentFileSize;

				return attachedFile;
			}

			// There is no attached file.
			return null;
		}

		/// <summary>
		/// Updates the sizes of the specified attached files with the current (= already transferred) size.
		/// </summary>
		/// <param name="attachedFiles">The attached files. This may be empty, but not null.</param>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		/// <param name="accessGranterLocator">The locator of the document which has offered the attached files.</param>
		/// <param name="accessGranterCreationTime">The creation time of the document which has offered the attached files.
		/// </param>
		/// <param name="requiredSize">The total size that the files to transfer have.</param>
		/// <param name="occupiedSize">Is set with the size of the file parts which were already transferred.</param>
		/// <returns>Returns true if the updated size is smaller than the (specified) initial size, that is, if the files are
		/// not yet fully transferred.</returns>
		public static bool UpdateSizesFromSavedFiles( AttachedFiles attachedFiles, string rootSaveFolderName
			, Locator accessGranterLocator, DateTime accessGranterCreationTime, out Int64 requiredSize
			, out Int64 occupiedSize )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );
			ValidationUtils.TextHasContent( rootSaveFolderName, "rootSaveFolderName" );
			ValidationUtils.ObjectHasContent( accessGranterLocator, "accessGranterLocator" );

			requiredSize = 0;
			occupiedSize = 0;

			foreach( AttachedFile attachedFile in attachedFiles )
			{
				requiredSize += attachedFile.Size;

				string saveFileName = GetSaveFileName( attachedFile.RelativeFileName, rootSaveFolderName, accessGranterLocator
					, accessGranterCreationTime );
				FileInfo fi = new FileInfo( saveFileName );

				if( fi.Exists )
				{
					occupiedSize += fi.Length;
					attachedFile.CurrentSize = fi.Length;
				}
				else
				{
					attachedFile.CurrentSize = 0;
				}
			}

			return occupiedSize < requiredSize;
		}

		/// <summary>
		/// Gets the total physical (= as stored) size of the files from the specified list.
		/// </summary>
		/// <param name="attachedFiles">List of files whose size to get. This may be empty, but not null.</param>
		/// <returns>Returns the total size of the files.</returns>
		/// <remarks>
		/// <para>The files which don't exist, don't contribute to the returned size.</para>
		/// </remarks>
		public static Int64 GetTotalPhysicalSize( AttachedFiles attachedFiles )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );

			Int64 totalSize = 0;

			foreach( AttachedFile attachedFile in attachedFiles )
			{
				FileInfo fi = new FileInfo( attachedFile.AbsoluteFileName );
				if( fi.Exists )
					totalSize += fi.Length;
			}

			return totalSize;
		}

		/// <summary>
		/// Gets the space which is required to save the specified attached files.
		/// </summary>
		/// <param name="attachedFiles">Files which have to be saved. This may be empty, but not null.</param>
		public static Int64 GetRequiredSize( AttachedFiles attachedFiles )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );

			Int64 totalSize = 0;

			foreach( AttachedFile attachedFile in attachedFiles )
				totalSize += attachedFile.Size;

			return totalSize;
		}

		/// <summary>
		/// Gets the currently transferred size of the specified attached files.
		/// </summary>
		/// <param name="attachedFiles">Files which have to be saved. This may be empty, but not null.</param>
		public static Int64 GetCurrentSize( AttachedFiles attachedFiles )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );

			Int64 totalCurrentSize = 0;

			foreach( AttachedFile attachedFile in attachedFiles )
				totalCurrentSize += attachedFile.CurrentSize;

			return totalCurrentSize;
		}

		/// <summary>
		/// Gets the space which is occupied by the specified saved files.
		/// </summary>
		/// <param name="attachedFiles">Files which have to be (or have been) saved. This may be empty, but not null.
		/// </param>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		/// <param name="accessGranterLocator">The locator of the document which has offered the attached files.</param>
		/// <param name="accessGranterCreationTime">The creation time of the document which has offered the attached files.
		/// </param>
		public static Int64 GetSpaceOfSavedFiles( AttachedFiles attachedFiles, string rootSaveFolderName
			, Locator accessGranterLocator, DateTime accessGranterCreationTime )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );
			ValidationUtils.TextHasContent( rootSaveFolderName, "rootSaveFolderName" );
			ValidationUtils.ObjectHasContent( accessGranterLocator, "accessGranterLocator" );

			Int64 totalSize = 0;

			foreach( AttachedFile attachedFile in attachedFiles )
			{
				string saveFileName = GetSaveFileName( attachedFile.RelativeFileName, rootSaveFolderName, accessGranterLocator
					, accessGranterCreationTime );
				FileInfo fi = new FileInfo( saveFileName );

				if( fi.Exists )
					totalSize += fi.Length;
			}

			return totalSize;
		}

		/// <summary>
		/// Gets the transfer progress (in percents).
		/// </summary>
		/// <param name="requiredSize">The total size that the files to transfer have.</param>
		/// <param name="occupiedSize">The size of the file parts which were already transferred.</param>
		/// <returns>Returns the transfer progress.</returns>
		public static double GetProgress( Int64 requiredSize, Int64 occupiedSize )
		{
			return (requiredSize > 0) ? (ActionProcessorConsts.DownloadFilesProgressMax * (double)occupiedSize / requiredSize)
				: 0;
		}

		/// <summary>
		/// Gets the transfer progress (in percents) of the specified attached files.
		/// </summary>
		/// <param name="attachedFiles">Files which have to be saved. This may be empty, but not null.</param>
		/// <returns>Returns the transfer progress.</returns>
		public static double GetProgress( AttachedFiles attachedFiles )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );

			Int64 requiredSize = GetRequiredSize( attachedFiles );
			Int64 currentSize = GetCurrentSize( attachedFiles );

			return GetProgress( requiredSize, currentSize );
		}

		/// <summary>
		/// Gets the transfer progress (in percents) of the specified saved files.
		/// </summary>
		/// <param name="attachedFiles">Files which have to be (or have been) saved. This may be empty, but not null.
		/// </param>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		/// <param name="accessGranterLocator">The locator of the document which has offered the attached files.</param>
		/// <param name="accessGranterCreationTime">The creation time of the document which has offered the attached files.
		/// </param>
		/// <returns>Returns the transfer progress.</returns>
		public static double GetProgressOfSavedFiles( AttachedFiles attachedFiles, string rootSaveFolderName
			, Locator accessGranterLocator, DateTime accessGranterCreationTime )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );
			ValidationUtils.TextHasContent( rootSaveFolderName, "rootSaveFolderName" );
			ValidationUtils.ObjectHasContent( accessGranterLocator, "accessGranterLocator" );

			Int64 requiredSize = GetRequiredSize( attachedFiles );
			Int64 occupiedSize = GetSpaceOfSavedFiles( attachedFiles, rootSaveFolderName, accessGranterLocator
				, accessGranterCreationTime );

			return GetProgress( requiredSize, occupiedSize );
		}

		/// <summary>
		/// Checks if the names of the specified files to save are valid, including their sizes. An exception is thrown if at
		/// least one file name is invalid.
		/// </summary>
		/// <param name="attachedFiles">Files which have to be saved. This may be empty, but not null.</param>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		/// <param name="accessGranterLocator">The locator of the document which has offered the attached files.</param>
		/// <param name="accessGranterCreationTime">The creation time of the document which has offered the attached files.
		/// </param>
		public static void ValidateFileNames( AttachedFiles attachedFiles, string rootSaveFolderName
			, Locator accessGranterLocator, DateTime accessGranterCreationTime )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );
			ValidationUtils.TextHasContent( rootSaveFolderName, "rootSaveFolderName" );
			ValidationUtils.ObjectHasContent( accessGranterLocator, "accessGranterLocator" );

			foreach( AttachedFile attachedFile in attachedFiles )
			{
				try
				{
					string saveFileName = GetSaveFileName( attachedFile.RelativeFileName, rootSaveFolderName
						, accessGranterLocator, accessGranterCreationTime );

					string invalidReason = ValidationUtils.Text( saveFileName, -1, -1, TextValidations.FileName, "saveFileName"
						, null, null, false );
					if( invalidReason != null )
					{
						throw new FatalCrossServiceException( "The name of the attached file to save is invalid.", null
							, new ExceptionDetail( "File name", saveFileName )
							, new ExceptionDetail( "Reason", invalidReason ) );
					}
				}
				catch( PathTooLongException ex )
				{
					throw new FatalCrossServiceException( "The name of the attached file to save is invalid.", null
						, new ExceptionDetail( "File name", attachedFile.RelativeFileName )
						, new ExceptionDetail( "Reason", ex.Message ) );
				}
			}
		}

		/// <summary>
		/// Checks if there is enough available space to save the specified files. An exception is thrown if there isn't enough
		/// space.
		/// </summary>
		/// <param name="attachedFiles">Files which have to be saved. This may be empty, but not null.</param>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		public static void ValidateAvailableSpace( AttachedFiles attachedFiles, string rootSaveFolderName )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );
			ValidationUtils.TextHasContent( rootSaveFolderName, "rootSaveFolderName" );

			Int64 requiredSize = GetRequiredSize( attachedFiles );
			Int64 availableSize = FileUtils.GetDriveAvailableFreeSpace( rootSaveFolderName );

			if( requiredSize > availableSize - SaveFileBufferSpace )
			{
				throw new FatalCrossServiceException( string.Format( "There is not enough space to save the attached files."
						+ " There must be at least {0} bytes of free space on the drive of the root save folder."
					, TextUtils.NumberToMoteText( SaveFileBufferSpace ) ) );
			}
		}

		/// <summary>
		/// Checks if there is enough available space to save the specified files. An exception is thrown if there isn't enough
		/// space.
		/// </summary>
		/// <param name="attachedFiles">Files which have to be saved. This may be empty, but not null.</param>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		/// <param name="accessGranterLocator">The locator of the document which has offered the attached files.</param>
		/// <param name="accessGranterCreationTime">The creation time of the document which has offered the attached files.
		/// </param>
		/// <param name="occupiedExtraSpace">Extra space which has to be considered occupied. This must be at least 0.</param>
		public static void ValidateAvailableSpaceWithSavedFiles( AttachedFiles attachedFiles, string rootSaveFolderName
			, Locator accessGranterLocator, DateTime accessGranterCreationTime, int occupiedExtraSpace )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );
			ValidationUtils.TextHasContent( rootSaveFolderName, "rootSaveFolderName" );
			ValidationUtils.ObjectHasContent( accessGranterLocator, "accessGranterLocator" );
			ValidationUtils.Integer( occupiedExtraSpace, 0, int.MaxValue, "occupiedExtraSpace" );

			Int64 requiredSize = GetRequiredSize( attachedFiles );
			Int64 availableSize = FileUtils.GetDriveAvailableFreeSpace( rootSaveFolderName );
			Int64 occupiedSize = GetSpaceOfSavedFiles( attachedFiles, rootSaveFolderName, accessGranterLocator
				, accessGranterCreationTime );

			if( requiredSize - occupiedSize > availableSize - SaveFileBufferSpace - occupiedExtraSpace )
			{
				throw new FatalCrossServiceException( string.Format( "There is not enough space to save the attached files."
						+ " There must be at least {0} bytes of free space on the drive of the root save folder."
					, TextUtils.NumberToMoteText( SaveFileBufferSpace ) ) );
			}
		}

		/// <summary>
		/// Gets the specified attached files as a plain-text list.
		/// </summary>
		/// <param name="attachedFiles">Attached files to get as list. This may be empty, but not null.</param>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		/// <param name="accessGranterLocator">The locator of the document which has offered the attached files.</param>
		/// <param name="accessGranterCreationTime">The creation time of the document which has offered the attached files.
		/// </param>
		/// <param name="localPeerDirection">The direction of the peer who has received the attached files.</param>
		/// <param name="fullPathForRecipient">If this is true then the returned file names contain the full path for the peer
		/// who has received the attached files, else the file names are relative to the peer's root save folder name
		/// ("rootSaveFolderName").</param>
		/// <returns>Returns the attached files as a plain-text list. This may be empty, but not null.</returns>
		public static string GetPlainTextList( AttachedFiles attachedFiles, string rootSaveFolderName
			, Locator accessGranterLocator, DateTime accessGranterCreationTime, LocalPeerDirections localPeerDirection
			, bool fullPathForRecipient )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );
			ValidationUtils.TextHasContent( rootSaveFolderName, "rootSaveFolderName" );
			ValidationUtils.ObjectHasContent( accessGranterLocator, "accessGranterLocator" );

			StringBuilder sb = new StringBuilder();

			foreach( AttachedFile attachedFile in attachedFiles )
			{
				string fileName;
				if( localPeerDirection == LocalPeerDirections.Initiator )
				{
					fileName = attachedFile.AbsoluteFileName;
				}
				else if( localPeerDirection == LocalPeerDirections.Recipient )
				{
					fileName = fullPathForRecipient
						? GetSaveFileName( attachedFile.RelativeFileName, rootSaveFolderName, accessGranterLocator
							, accessGranterCreationTime )
						: attachedFile.RelativeFileName;
				}
				else
					throw ValidationUtils.Unhandled( localPeerDirection, "localPeerDirection" );

				sb.AppendLine( fileName );
			}

			return sb.ToString();
		}

		/// <summary>
		/// Gets the specified attached file names as a plain-text list.
		/// </summary>
		/// <param name="attachedFileNames">Attached file names to get as list. This may be empty, but not null.</param>
		/// <returns>Returns the attached file names as a plain-text list. This may be empty, but not null.</returns>
		public static string GetPlainTextList( AttachedFileNames attachedFileNames )
		{
			ValidationUtils.ObjectHasContent( attachedFileNames, "attachedFileNames" );

			StringBuilder sb = new StringBuilder();

			foreach( AttachedFileName attachedFile in attachedFileNames )
				sb.AppendLine( attachedFile.AbsoluteFileName );

			return sb.ToString();
		}

		/// <summary>
		/// Gets an indicator for the draft message which is meant to show what files have been attached to the draft letter.
		/// </summary>
		/// <param name="attachedFileNames">The names of the files which were attached. This may be null or empty.</param>
		/// <returns>Returns an indicator for the draft message which is meant to show what files have been attached to the draft
		/// letter. This may be null or empty if there are no attached files.</returns>
		public static string GetDraftMessageIndicatorToShowAttachedFiles( AttachedFileNames attachedFileNames )
		{
			if( ListUtils.NoContent( attachedFileNames ) )
				return null;

			string fileName = attachedFileNames.GetFirstItem().RelativeFileName;
			string indicator = null;

			if( attachedFileNames.Count > 1 )
				indicator = FileUtils.GetRootParentFolderName( Path.GetDirectoryName( fileName ) );

			if( TextUtils.NoContent( indicator ) )
				indicator = Path.GetFileName( fileName );

			return indicator;
		}

		/// <summary>
		/// Checks if the specified attached files contain at least one file in a folder.
		/// </summary>
		/// <param name="attachedFiles">Attached files in which to look. If this is null or empty, nothing happens.</param>
		/// <returns>Returns true if the specified attached files contain at least one (file in a) folder.</returns>
		public static bool HasFolder( AttachedFiles attachedFiles )
		{
			if( ListUtils.NoContent( attachedFiles ) )
				return false;

			foreach( AttachedFile attachedFile in attachedFiles )
			{
				if( TextUtils.HasContent( Path.GetDirectoryName( attachedFile.RelativeFileName ) ) )
					return true;
			}

			return false;
		}

		/// <summary>
		/// Checks if the specified attached file names refer media files.
		/// </summary>
		/// <param name="attachedFileNames">Attached file names in which to look. This may be empty, but not null.</param>
		/// <returns>Returns true if the specified attached file names refer media files.</returns>
		public static bool HasMediaFiles( AttachedFileNames attachedFileNames )
		{
			ValidationUtils.ObjectHasContent( attachedFileNames, "attachedFileNames" );

			foreach( AttachedFileName attachedFileName in attachedFileNames )
			{
				if( FileExtensions.IsMediaFile( attachedFileName.RelativeFileName ) )
					return true;
			}

			return false;
		}

		/// <summary>
		/// Checks if the specified attached file names refer at least 1 image file.
		/// </summary>
		/// <param name="attachedFileNames">Attached file names in which to look. This may be empty, but not null.</param>
		/// <returns>Returns true if the specified attached file names refer at least 1 image file.</returns>
		public static bool HasImageFile( AttachedFileNames attachedFileNames )
		{
			ValidationUtils.ObjectHasContent( attachedFileNames, "attachedFileNames" );

			foreach( AttachedFileName attachedFileName in attachedFileNames )
			{
				if( FileExtensions.IsImageFile( attachedFileName.RelativeFileName ) )
					return true;
			}

			return false;
		}

		/// <summary>
		/// Collects the media files from the specified attached files.
		/// </summary>
		/// <param name="attachedFiles">Files which have to be saved. This may be empty, but not null.</param>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		/// <param name="accessGranterLocator">The locator of the document which has offered the attached files.</param>
		/// <param name="accessGranterCreationTime">The creation time of the document which has offered the attached files.
		/// </param>
		/// <param name="localPeerDirection">The direction of the peer who has received the attached files.</param>
		/// <param name="videoFileNames">Is set with the collected names of video files. This may be empty, but not null.</param>
		/// <param name="audioFileNames">Is set with the collected names of audio files. This may be empty, but not null.</param>
		/// <param name="imageFileNames">Is set with the collected names of image files. This may be empty, but not null.</param>
		public static void CollectMediaFiles( AttachedFiles attachedFiles, string rootSaveFolderName
			, Locator accessGranterLocator, DateTime accessGranterCreationTime, LocalPeerDirections localPeerDirection
			, out Strings videoFileNames, out Strings audioFileNames, out Strings imageFileNames )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );
			ValidationUtils.TextHasContent( rootSaveFolderName, "rootSaveFolderName" );
			ValidationUtils.ObjectHasContent( accessGranterLocator, "accessGranterLocator" );

			videoFileNames = new Strings();
			audioFileNames = new Strings();
			imageFileNames = new Strings();

			foreach( AttachedFile attachedFile in attachedFiles )
			{
				if( !FileExtensions.IsMediaFile( attachedFile.RelativeFileName ) )
					continue;

				string fileName;
				if( localPeerDirection == LocalPeerDirections.Initiator )
				{
					fileName = attachedFile.AbsoluteFileName;
				}
				else if( localPeerDirection == LocalPeerDirections.Recipient )
				{
					fileName = GetSaveFileName( attachedFile.RelativeFileName, rootSaveFolderName, accessGranterLocator
						, accessGranterCreationTime );
				}
				else
					throw ValidationUtils.Unhandled( localPeerDirection, "localPeerDirection" );

				if( FileExtensions.IsVideoFile( attachedFile.RelativeFileName ) )
					videoFileNames.Add( fileName );
				else if( FileExtensions.IsAudioFile( attachedFile.RelativeFileName ) )
					audioFileNames.Add( fileName );
				else if( FileExtensions.IsImageFile( attachedFile.RelativeFileName ) )
					imageFileNames.Add( fileName );
			}
		}

		/// <summary>
		/// Gets the attached files which can be used for image previews, from the specified attached files.
		/// </summary>
		/// <param name="attachedFiles">Attached files in which to look for image files. If this is null or empty, nothing
		/// happens.</param>
		/// <param name="showPreviewsForAttachedImageFilesCount">The number of attached image files for which image previews can
		/// be shown.</param>
		/// <returns>Returns the attached files which can be used for image previews. This may be null or empty.</returns>
		/// <remarks>
		/// <para>The returned attached files are the first "showPreviewsForAttachedImageFilesCount" files (from the list of
		/// attached files) whose extensions are in the <see cref="FileExtensions.ImageFile"/> list, and whose size is maximum
		/// <see cref="PeerDocumentsConsts.LetterImagePreviewSizeMax"/>. For performance purposes, the search doesn't continue
		/// after the first "showPreviewsForAttachedImageFilesCount" image files; this is because, at some point, the code would
		/// check the size of these files and there may be thousands of files, none of which may be fully downloaded yet.</para>
		/// 
		/// <para>In order to ensure optimum UI performance (regarding the order of the downloaded and shown image previews),
		/// this and the <see cref="GetFilesForImagePreview"/> methods have to return lists which are sorted in the same way
		/// (so <see cref="AttachedFiles"/> and <see cref="LocalAttachedFiles"/> have to be sorted in the same way).</para>
		/// </remarks>
		public static AttachedFiles GetPotentialFilesForImagePreview( AttachedFiles attachedFiles
			, int showPreviewsForAttachedImageFilesCount )
		{
			ValidationUtils.Integer( showPreviewsForAttachedImageFilesCount, 1
				, PeerDocumentsConsts.ShowPreviewsForAttachedImageFilesCountMax, "showPreviewsForAttachedImageFilesCount" );

			if( ListUtils.NoContent( attachedFiles ) )
				return null;

			AttachedFiles potentialFilesForImagePreview = new AttachedFiles();
			int verifiedImageFilesCount = 0;

			foreach( AttachedFile attachedFile in attachedFiles )
			{
				if( !FileExtensions.IsImageFile( attachedFile.RelativeFileName ) )
					continue;

				verifiedImageFilesCount++;

				if( attachedFile.Size > PeerDocumentsConsts.LetterImagePreviewSizeMax )
					continue;

				potentialFilesForImagePreview.Add( attachedFile );

				if( Math.Max( verifiedImageFilesCount, potentialFilesForImagePreview.Count )
					>= showPreviewsForAttachedImageFilesCount )
				{
					break;
				}
			}

			return potentialFilesForImagePreview;
		}

		/// <summary>
		/// Gets the local attached files which can be used for image previews, from the specified attached files.
		/// </summary>
		/// <param name="attachedFiles">Attached files in which to look for image files. If this is null or empty, nothing
		/// happens.</param>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		/// <param name="showPreviewsForAttachedImageFilesCount">The number of attached image files for which image previews can
		/// be shown.</param>
		/// <param name="accessGranterLocator">The locator of the document which has offered the attached files.</param>
		/// <param name="accessGranterCreationTime">The creation time of the document which has offered the attached files.
		/// </param>
		/// <param name="localPeerDirection">The direction of the peer who has received the attached files.</param>
		/// <param name="onlyCompleteLocalFiles">If this is true then only the files whose actual size is the one from the
		/// specified attached files are returned (some image files may be missing or may have to downloaded, so they are not yet
		/// available for image previews).</param>
		/// <returns>Returns the local attached files which can be used for image previews. This may be null or empty.
		/// </returns>
		/// <remarks>
		/// <para>The returned attached files are the first "showPreviewsForAttachedImageFilesCount" files (from the list of
		/// attached files) whose extensions are in the <see cref="FileExtensions.ImageFile"/> list, (if "onlyCompleteFiles" is
		/// true) whose actual size is the one from the specified attached files (some image files may be missing or may have to
		/// downloaded, so they are not yet available for image previews), and whose size is maximum
		/// <see cref="PeerDocumentsConsts.LetterImagePreviewSizeMax"/>. For performance purposes, the search doesn't continue
		/// after the first "showPreviewsForAttachedImageFilesCount" image files; this is because, at some point, the code would
		/// check the size of these files and there may be thousands of files, none of which may be fully downloaded yet.</para>
		/// 
		/// <para>In order to ensure optimum UI performance (regarding the order of the downloaded and shown image previews),
		/// this and the <see cref="GetPotentialFilesForImagePreview"/> methods have to return lists which are sorted in the same
		/// way (so <see cref="AttachedFiles"/> and <see cref="LocalAttachedFiles"/> have to be sorted in the same way).</para>
		/// </remarks>
		public static LocalAttachedFiles GetFilesForImagePreview( AttachedFiles attachedFiles, string rootSaveFolderName
			, int showPreviewsForAttachedImageFilesCount, Locator accessGranterLocator, DateTime accessGranterCreationTime
			, LocalPeerDirections localPeerDirection, bool onlyCompleteLocalFiles )
		{
			ValidationUtils.TextHasContent( rootSaveFolderName, "rootSaveFolderName" );
			ValidationUtils.Integer( showPreviewsForAttachedImageFilesCount, 1
				, PeerDocumentsConsts.ShowPreviewsForAttachedImageFilesCountMax, "showPreviewsForAttachedImageFilesCount" );
			ValidationUtils.ObjectHasContent( accessGranterLocator, "accessGranterLocator" );

			if( ListUtils.NoContent( attachedFiles ) || (localPeerDirection == LocalPeerDirections.RecipientImported) )
				return null;

			LocalAttachedFiles localAttachedFiles = new LocalAttachedFiles();
			int verifiedImageFilesCount = 0;

			foreach( AttachedFile attachedFile in attachedFiles )
			{
				if( !FileExtensions.IsImageFile( attachedFile.RelativeFileName ) )
					continue;

				string localFileName = null;
				if( localPeerDirection == LocalPeerDirections.Initiator )
				{
					localFileName = attachedFile.AbsoluteFileName;
				}
				else if( localPeerDirection == LocalPeerDirections.Recipient )
				{
					localFileName = GetSaveFileName( attachedFile.RelativeFileName, rootSaveFolderName, accessGranterLocator
						, accessGranterCreationTime );
				}
				else
					throw ValidationUtils.Unhandled( localPeerDirection, "localPeerDirection" );

				verifiedImageFilesCount++;

				if( attachedFile.Size > PeerDocumentsConsts.LetterImagePreviewSizeMax )
					continue;

				// Check that the file is completely available (for loading).
				if( onlyCompleteLocalFiles )
				{
					FileInfo fi = new FileInfo( localFileName );

					if( !fi.Exists )
						continue;

					// Only received files can be partially available.
					if( (localPeerDirection == LocalPeerDirections.Recipient) && (fi.Length != attachedFile.Size) )
						continue;
				}

				localAttachedFiles.Add( attachedFile, localFileName );

				if( Math.Max( verifiedImageFilesCount, localAttachedFiles.Count ) >= showPreviewsForAttachedImageFilesCount )
					break;
			}

			return localAttachedFiles;
		}

		/// <summary>
		/// Gets the name of the first image file from the specified list.
		/// </summary>
		/// <param name="attachedFileNames">The names of the attached files in which to look. This may be null or empty.</param>
		/// <returns>Returns the name of the first image file from the specified list. This may be null or empty.</returns>
		public static string GetFirstImageFileName( AttachedFileNames attachedFileNames )
		{
			if( ListUtils.NoContent( attachedFileNames ) )
				return null;

			foreach( AttachedFileName attachedFileName in attachedFileNames )
			{
				if( FileExtensions.IsImageFile( attachedFileName.RelativeFileName ) )
					return attachedFileName.AbsoluteFileName;
			}

			return null;
		}

		/// <summary>
		/// Copies the attached image files to the local folder with custom emoticons.
		/// </summary>
		/// <param name="attachedFiles">Attached files to copy. If this is null or empty, nothing happens.</param>
		/// <param name="rootSaveFolderName">The name of the root folder where to save the attached files.</param>
		/// <param name="accessGranterLocator">The locator of the document which has offered the attached files.</param>
		/// <param name="accessGranterCreationTime">The creation time of the document which has offered the attached files.
		/// </param>
		/// <param name="localPeerDirection">The direction of the peer who has received the attached files.</param>
		/// <param name="customEmoticonsFolderName">The name of the folder in which to copy the files.</param>
		/// <returns>Returns true if at least 1 file was copied.</returns>
		/// <remarks>
		/// <para>Only the files whose extensions are in the <see cref="FileExtensions.ImageFile"/> list are copied.</para>
		/// 
		/// <para>A file is copied only if it's available in its entirety on the local system.</para>
		/// </remarks>
		public static bool CopyImageFiles( AttachedFiles attachedFiles, string rootSaveFolderName, Locator accessGranterLocator
			, DateTime accessGranterCreationTime, LocalPeerDirections localPeerDirection, string customEmoticonsFolderName )
		{
			ValidationUtils.TextHasContent( rootSaveFolderName, "rootSaveFolderName" );
			ValidationUtils.ObjectHasContent( accessGranterLocator, "accessGranterLocator" );

			if( ListUtils.NoContent( attachedFiles ) || (localPeerDirection == LocalPeerDirections.RecipientImported) )
				return false;

			bool hasCopied = false;
			foreach( AttachedFile attachedFile in attachedFiles )
			{
				if( !FileExtensions.IsImageFile( attachedFile.RelativeFileName ) )
					continue;

				string fileName = null;
				if( localPeerDirection == LocalPeerDirections.Initiator )
				{
					fileName = attachedFile.AbsoluteFileName;
				}
				else if( localPeerDirection == LocalPeerDirections.Recipient )
				{
					fileName = GetSaveFileName( attachedFile.RelativeFileName, rootSaveFolderName, accessGranterLocator
						, accessGranterCreationTime );
				}
				else
					throw ValidationUtils.Unhandled( localPeerDirection, "localPeerDirection" );

				FileInfo fi = new FileInfo( fileName );
				if( !fi.Exists || (fi.Length != attachedFile.Size) )
					continue;

				string destinationFileName = FileUtils.MakeFullName( customEmoticonsFolderName, attachedFile.RelativeFileName );
				FileUtils.CreateFolder( Path.GetDirectoryName( destinationFileName ) );

				File.Copy( fileName, destinationFileName, true );
				hasCopied = true;
			}

			return hasCopied;
		}

		#endregion Public static methods
	}
}
