////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Cloning;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// List of method informations which allow an action processor to handle action requests.
	/// </summary>
	public abstract class HandleRequestMethodInfos : KernelSortedList<string, HandleRequestMethodInfo>, IUnclonable
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public HandleRequestMethodInfos()
			: base( StringInsensitiveComparer.Instance )
		{
		}

		#endregion Constructors

		#region Private methods

		/// <summary>
		/// Adds to the list the method informations which can handle action requests, from the action processor with the
		/// specified type name.
		/// </summary>
		/// <param name="actionProcessorTypeName">The type name of the action processor whose method informations to add to the
		/// list. The type must be globally registered.</param>
		protected abstract void AddMethodInfos( string actionProcessorTypeName );

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para><see cref="HandleRequestMethodInfo.RequestTypeName"/> is used as key.</para>
		/// </remarks>
		public override string GetKey( HandleRequestMethodInfo item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.RequestTypeName;
		}

		/// <summary>
		/// Adds to the list the method informations which can handle action requests, from the action processors with the
		/// specified type names.
		/// </summary>
		/// <param name="actionProcessorTypeNames">The type names of the action processors whose method informations to add to
		/// the list. The types must be globally registered.</param>
		public void Add( Strings actionProcessorTypeNames )
		{
			ValidationUtils.ListHasContent( actionProcessorTypeNames, "actionProcessorTypeNames" );

			foreach( string actionProcessorTypeName in actionProcessorTypeNames )
				AddMethodInfos( actionProcessorTypeName );
		}

		/// <summary>
		/// Adds to the list the method informations which can handle action requests, from the action processor with the
		/// specified type name.
		/// </summary>
		/// <param name="actionProcessorTypeName">The type name of the action processor whose method informations to add to the
		/// list. The type must be globally registered.</param>
		public void Add( string actionProcessorTypeName )
		{
			AddMethodInfos( actionProcessorTypeName );
		}

		#endregion Public methods
	}
}
