////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Action processor configuration.
	/// </summary>
	/// <remarks>
	/// <para>Unlike <see cref="ActionProcessorConfiguration"/>, a single instance of this type is used hold some configuration
	/// about all the action processors which may be created for a given storehouse.</para>
	/// </remarks>
	public class ActionProcessorConfiguration : NotifyPropertyChanged, IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionProcessorConfiguration()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="useUpnResolvers">If this is true then the UPN resolvers (like the PNRP) will be used to register UPNs,
		/// and resolve UPNs to IP addresses. If this is false then the UPN resolvers will not be used, so the alternative peer
		/// resolution must be used instead.</param>
		/// <param name="forwardPorts">If this is true then the action service ports will be forwarded; this is important for the
		/// peers whose computers are connected to routers, and who have no IPv6 network connectivity.</param>
		/// <param name="rootSaveFolderName">The name of the folder where the downloaded files are saved.</param>
		/// <param name="actionClient">The action client configuration.</param>
		/// <param name="actionService">The action service configuration.</param>
		public ActionProcessorConfiguration( bool useUpnResolvers, bool forwardPorts, string rootSaveFolderName
			, ActionClientConfiguration actionClient, ActionServiceConfiguration actionService )
		{
			UseUpnResolvers = useUpnResolvers;
			ForwardPorts = forwardPorts;
			RootSaveFolderName = rootSaveFolderName;
			ActionClient = actionClient;
			ActionService = actionService;
		}

		#endregion Constructors

		#region Public serializable properties

		private bool _PeerIsOnline = true;
		/// <summary>
		/// If this is true then the current peer is normally online, else it's always offline and therefore no action client and
		/// service managers are created. The default is true. Notifies property observers.
		/// </summary>
		[SerializeAttribute]
		public bool PeerIsOnline
		{
			get { return _PeerIsOnline; }
			set
			{
				_PeerIsOnline = value;

				// Notify property observers.
				OnPropertyChanged( "PeerIsOnline" );
			}
		}

		private bool _UseUpnResolvers = true;
		/// <summary>
		/// If this is true then the UPN resolvers (like the PNRP) will be used to register UPNs, and resolve UPNs to IP
		/// addresses. If this is false then the UPN resolvers will not be used, so the alternative peer resolution must be used
		/// instead. The default is true. Notifies property observers.
		/// </summary>
		[SerializeAttribute]
		public bool UseUpnResolvers
		{
			get { return _UseUpnResolvers; }
			set
			{
				_UseUpnResolvers = value;

				// Notify property observers.
				OnPropertyChanged( "UseUpnResolvers" );
			}
		}

		private bool _ForwardPorts = true;
		/// <summary>
		/// If this is true then the action service ports will be forwarded; this is important for the peers whose computers are
		/// connected to routers, and who have no IPv6 network connectivity. The default is true. Notifies property observers.
		/// </summary>
		[SerializeAttribute]
		public bool ForwardPorts
		{
			get { return _ForwardPorts; }
			set
			{
				_ForwardPorts = value;

				// Notify property observers.
				OnPropertyChanged( "ForwardPorts" );
			}
		}

		private string _RootSaveFolderName;
		/// <summary>		
		/// The name of the folder where the downloaded files are saved. Notifies property observers.
		/// </summary>
		[SerializeAttribute]
		public string RootSaveFolderName
		{
			get { return _RootSaveFolderName; }
			set
			{
				ValidationUtils.TextHasContent( value, "RootSaveFolderName" );

				_RootSaveFolderName = value;

				// Notify property observers.
				OnPropertyChanged( "RootSaveFolderName" );
			}
		}

		private UpnResolverAddresses _UpnResolverAddresses = new UpnResolverAddresses();
		/// <summary>		
		/// List of addresses for UPN resolver services. The backing field is initialized by default. This may be empty, but not
		/// null.
		/// </summary>
		[SerializeAttribute]
		public UpnResolverAddresses UpnResolverAddresses
		{
			get { return _UpnResolverAddresses; }
			set { _UpnResolverAddresses = value; }
		}

		private int _LargeCommunicationCacheSize = ActionProcessorConsts.LargeCommunicationCacheSizeDef;
		/// <summary>
		/// The size (in bytes) of the cache used for sending and receiving data through the network, for large data transfers.
		/// If this is 0 or a negative value, it's ignored. Notifies property observers.
		/// </summary>
		/// <remarks>
		/// <para>Since this is a new property, it must have "MustBeDeserialized = false".</para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false )]
		public int LargeCommunicationCacheSize
		{
			get { return _LargeCommunicationCacheSize; }
			set
			{
				ValidationUtils.Integer( value, 0, ActionProcessorConsts.LargeCommunicationCacheSizeMax
					, "LargeCommunicationCacheSize" );

				_LargeCommunicationCacheSize = value;

				// Notify property observers.
				OnPropertyChanged( "LargeCommunicationCacheSize" );
			}
		}

		private int _CommunicationTimeoutPerBlob = ActionProcessorConsts.CommunicationTimeoutPerBlob;
		/// <summary>
		/// The timeout (in seconds) used for sending and receiving data through the network, per blob, per frame. This must be
		/// at least 1. Notifies property observers.
		/// </summary>
		[SerializeAttribute]
		public int CommunicationTimeoutPerBlob
		{
			get { return _CommunicationTimeoutPerBlob; }
			set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "CommunicationTimeoutPerBlob" );

				_CommunicationTimeoutPerBlob = value;

				// Notify property observers.
				OnPropertyChanged( "CommunicationTimeoutPerBlob" );
			}
		}

		private int _CommunicationFrameSizeMax = ActionProcessorConsts.CommunicationFrameSizeMax;
		/// <summary>
		/// The maximum size (in bytes) of a communication frame which a processor may process (in a single send or receive).
		/// This must be at least 1. Notifies property observers.
		/// </summary>
		[SerializeAttribute]
		public int CommunicationFrameSizeMax
		{
			get { return _CommunicationFrameSizeMax; }
			set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "CommunicationFrameSizeMax" );

				_CommunicationFrameSizeMax = value;

				// Notify property observers.
				OnPropertyChanged( "CommunicationFrameSizeMax" );
			}
		}

		private int _ParallelRequestsMax = ActionProcessorConsts.ParallelRequestsMax;
		/// <summary>
		/// The maximum number of action requests which a processor may process at the same time. If this is 0 or a negative
		/// number, <see cref="SystemUtils.GetProcessorCoresCount"/> is used instead. Notifies property observers.
		/// </summary>
		/// <remarks>
		/// <para>For action services, this limit is per service listener.</para>
		/// </remarks>
		[SerializeAttribute]
		public int ParallelRequestsMax
		{
			get { return _ParallelRequestsMax; }
			set
			{
				_ParallelRequestsMax = value;

				// Notify property observers.
				OnPropertyChanged( "ParallelRequestsMax" );
			}
		}

		private ActionClientConfiguration _ActionClient;
		/// <summary>
		/// The action client configuration. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public ActionClientConfiguration ActionClient
		{
			get { return _ActionClient; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionClient" );

				_ActionClient = value;
			}
		}

		private ActionServiceConfiguration _ActionService;
		/// <summary>
		/// The action service configuration. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public ActionServiceConfiguration ActionService
		{
			get { return _ActionService; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionService" );

				_ActionService = value;
			}
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Validates the current object.
		/// </summary>
		protected virtual void Validate()
		{
			ValidationUtils.Integer( (UInt64)CommunicationFrameSizeMax, 1
				, ActionService.TrafficFilterConfiguration.KnownPeerProcessableSizeMax, "CommunicationFrameSizeMax" );
		}

		/// <summary>
		/// This method is automatically called before the current object is serialized.
		/// </summary>
		protected virtual void Serializing()
		{
			Validate();
		}

		/// <summary>
		/// This method is automatically called after the current object has been deserialized.
		/// </summary>
		protected virtual void Deserialized()
		{
			Validate();
		}

		#endregion Private methods
	}
}
