////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Text;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.PeerDocuments;

namespace Vela.ActionProcessor
{
	/// <summary>
	/// Secondary processing state.
	/// </summary>
	/// <remarks>
	/// <para>This contains information regarding the state of the secondary processing sequence which is performed after the
	/// main processing sequence of an action request. The secondary processing sequence is not critical and if its fails, no
	/// exception is thrown in either the action service or the action client. Errors during the main processing sequence
	/// are indicated by <see cref="ErrorActionResponse"/>.</para>
	/// 
	/// <para>This is stored in the action progress (<see cref="ActionProgress"/>), on both the service and the client.</para>
	/// 
	/// <para>This is meant for the user to see and decide what to do, not for the action client to take further automatic
	/// actions.</para>
	/// 
	/// <para>This may, for example, contain an autoresponse (from the service), or the execution state of a command line.</para>
	/// </remarks>
	public class SecondaryProcessingState : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected SecondaryProcessingState()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="anErrorHasOccurred">If this is true then there was an error during the secondary processing sequence of
		/// the associated action request.</param>
		/// <param name="subject">Subject.</param>
		/// <param name="state">State. This may be null or empty.</param>
		/// <param name="message">Message. This may be null or empty.</param>
		/// <param name="cutTextsToSizeMax">If this is true and any text is longer than its maximum size, only the first part of
		/// the text is used. If this is false and any text is longer than its maximum size, an exception is thrown.</param>
		public SecondaryProcessingState( bool anErrorHasOccurred, string subject, string state, string message
			, bool cutTextsToSizeMax )
		{
			AnErrorHasOccurred = anErrorHasOccurred;

			if( cutTextsToSizeMax )
			{
				Subject = TextUtils.GetBeginning( subject, KernelConsts.TinyNoteSizeMax, false, 1
					, PeerDocumentsConsts.TextWasTruncatedSuffix );
				State = TextUtils.GetBeginning( state, KernelConsts.TinyNoteSizeMax, false, 1
					, PeerDocumentsConsts.TextWasTruncatedSuffix );
				Message = TextUtils.GetBeginning( message, KernelConsts.MediumNoteSizeMax, false, 0
					, PeerDocumentsConsts.TextWasTruncatedSuffix );
			}
			else
			{
				Subject = subject;
				State = state;
				Message = message;
			}
		}

		#endregion Constructors

		#region Public serializable properties

		private bool _AnErrorHasOccurred;
		/// <summary>
		/// If this is true then there was an error during the secondary processing sequence of the associated action request.
		/// READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is true, the client should not retry to send the action request to the service (unless the main
		/// processing sequence of the action request has failed with a <see cref="RetryCrossServiceException"/> exception).
		/// </para>
		/// </remarks>
		[SerializeAttribute]
		public bool AnErrorHasOccurred
		{
			get { return _AnErrorHasOccurred; }
			protected set { _AnErrorHasOccurred = value; }
		}

		private string _Subject;
		/// <summary>
		/// Subject. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string Subject
		{
			get { return _Subject; }
			protected set
			{
				ValidationUtils.Text( value, -1, -1, TextValidations.TinyNote, "Subject" );

				_Subject = value;
			}
		}

		private string _State;
		/// <summary>
		/// State. This may be null or empty. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string State
		{
			get { return _State; }
			protected set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.TinyNote, "State" );

				_State = value;
			}
		}

		private string _Message;
		/// <summary>
		/// Message. This may be null or empty. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string Message
		{
			get { return _Message; }
			protected set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.MediumNote, "Message" );

				_Message = value;
			}
		}

		#endregion Public serializable properties

		#region Public static methods

		/// <summary>
		/// Merges the specified secondary processing state.
		/// </summary>
		/// <param name="secondaryProcessingStates">Secondary processing states to merge. This may be null or empty.</param>
		/// <returns>Returns the merged secondary processing states. This is null if "states" is null or empty, or all its items
		/// are null.</returns>
		/// <remarks>
		/// <para>If a single specified state has its <see cref="AnErrorHasOccurred"/> property true, the merged state will
		/// also have it true.</para>
		/// 
		/// <para>The first input state has priority, so the merged state will contain its "Subject" and "State".</para>
		/// </remarks>
		public static SecondaryProcessingState Merge( params SecondaryProcessingState[] secondaryProcessingStates )
		{
			if( ListUtils.NoContent( secondaryProcessingStates ) )
				return null;

			bool allInStatesNull = true;
			StringBuilder mergedMessageSb = new StringBuilder();
			bool mergedAnErrorHasOccurred = false;
			string mergedSubject = null;
			string mergedState = null;

			// Merge "Message" from all the states. There is no point in merging "Subject" and "State" because these texts should
			// be as short as possible.
			foreach( SecondaryProcessingState secondaryProcessingState in secondaryProcessingStates )
			{
				if( secondaryProcessingState == null )
					continue;

				allInStatesNull = false;

				// If a single specified state has its "AnErrorHasOccurred" true, keep the state (with top priority).
				if( secondaryProcessingState.AnErrorHasOccurred )
				{
					mergedAnErrorHasOccurred = true;
					mergedSubject = secondaryProcessingState.Subject;
					mergedState = secondaryProcessingState.State;
				}
				else if( TextUtils.NoContent( mergedSubject ) && !mergedAnErrorHasOccurred )
				{
					// Keep only the first "Subject" and "State".
					mergedSubject = secondaryProcessingState.Subject;
					mergedState = secondaryProcessingState.State;
				}

				// Concatenate "Message".
				if( mergedMessageSb.Length < KernelConsts.MediumNoteSizeMax )
				{
					mergedMessageSb.Append( secondaryProcessingState.Subject );

					if( TextUtils.HasContent( secondaryProcessingState.State ) )
						mergedMessageSb.AppendFormat( " {0}", secondaryProcessingState.State );

					if( TextUtils.HasContent( secondaryProcessingState.Message ) )
						mergedMessageSb.AppendFormat( ": {0}", secondaryProcessingState.Message );

					mergedMessageSb.AppendLine( "\n" );
				}
			}

			if( allInStatesNull )
				return null;

			string mergedMessage = TextUtils.GetBeginning( mergedMessageSb.ToString(), KernelConsts.MediumNoteSizeMax, false, 0
				, PeerDocumentsConsts.TextWasTruncatedSuffix );

			SecondaryProcessingState mergedSecondaryProcessingState = new SecondaryProcessingState( mergedAnErrorHasOccurred
				, mergedSubject, mergedState, mergedMessage, true );

			return mergedSecondaryProcessingState;
		}

		#endregion Public static methods
	}
}
