////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.ActionProcessor;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.ActionService
{
	/// <summary>
	/// Contract for an action service.
	/// </summary>
	/// <remarks>
	/// <para>Expired action requests are not processed.</para>
	/// 
	/// <para>All the sheafs are locked with <see cref="SheafLocks.Shared"/>, unless they will be saved (in which case they are
	/// locked with <see cref="SheafLocks.Update"/>).</para>
	/// </remarks>
	public abstract class ActionService : IActionService
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionService()
		{
		}

		#endregion Constructors

		#region Private static methods

		/// <summary>
		/// Creates the properly derived action request context.
		/// </summary>
		/// <param name="actionRequestType">The type of the action request which must be processed.</param>
		/// <param name="serviceContext">The service context.</param>
		/// <param name="anonymousKnowledge">The sheaf which proves that the local peer has knowledge of the initiator of the
		/// current action request. This may be null.</param>
		/// <param name="actionResponseLocator">The locator of the action response associated to the current action
		/// request.</param>
		/// <param name="initiatorAccountContext">The account context of the initiator of the action request for which to
		/// process the specified action. This may be null if the action request has no initiator.</param>
		/// <param name="beneficiaryAccountContext">The account context of the beneficiary of the action request. This may be
		/// null if the action request has no beneficiary.</param>
		/// <returns>Returns the properly derived action request context.</returns>
		private static ActionRequestContext CreateDerivedRequestContext( Type actionRequestType
			, ActionServiceContext serviceContext, Sheaf anonymousKnowledge, Locator actionResponseLocator
			, ActionAccountContext initiatorAccountContext, ActionAccountContext beneficiaryAccountContext )
		{
			ValidationUtils.ObjectHasContent( actionRequestType, "actionRequestType" );

			if( actionRequestType.IsSubclassOf( typeof( ActionRequestWithBeneficiary ) ) )
			{
				return new ActionRequestWithBeneficiaryContext( serviceContext, actionResponseLocator, initiatorAccountContext
					, beneficiaryAccountContext );
			}
			else if( actionRequestType.IsSubclassOf( typeof( ActionRequestWithInitiator ) ) )
			{
				return new ActionRequestWithInitiatorContext( serviceContext, actionResponseLocator, initiatorAccountContext );
			}
			else if( actionRequestType.IsSubclassOf( typeof( ActionRequest ) ) )
			{
				return new ActionRequestContext( serviceContext, anonymousKnowledge, actionResponseLocator );
			}
			else
			{
				throw ValidationUtils.Unhandled( actionRequestType.FullName, "actionRequestType" );
			}
		}

		/// <summary>
		/// Gets the action account context for the specified action initiator and recipient UPN fingerprints.
		/// </summary>
		/// <param name="initiatorUpnFingerprint">Initiator UPN fingerprint.</param>
		/// <param name="recipientUpnFingerprint">Recipient UPN fingerprint.</param>
		/// <param name="serviceSavesTally">If this is true then the tally is automatically saved if the action request is
		/// processed successfully.</param>
		/// <param name="serviceSavesAccount">If this is true then the account is automatically saved if the action request is
		/// processed successfully.</param>
		/// <param name="transaction">The transaction with which to load the documents.</param>
		/// <returns>Returns the loaded action account context.</returns>
		private static ActionAccountContext LoadActionAccountContext( Locator initiatorUpnFingerprint
			, Locator recipientUpnFingerprint, bool serviceSavesTally, bool serviceSavesAccount
			, IStorehouseTransaction transaction )
		{
			ValidationUtils.ObjectHasContent( initiatorUpnFingerprint, "initiatorUpnFingerprint" );
			ValidationUtils.ObjectHasContent( recipientUpnFingerprint, "recipientUpnFingerprint" );
			ValidationUtils.ObjectHasContent( transaction, "transaction" );

			IDeserializer deserializer;

			#region Load the tally

			SpeckleFilter tallyUpnFilter = new SpeckleFilter( Tally.RemoteUpnFingerprintFilterIdentifier
				, initiatorUpnFingerprint );

			Tally tally = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Tally ) )
				, new SpeckleFilters( tallyUpnFilter ), StorehouseRealms.ProcessingIdentityContainers
				, serviceSavesTally ? SheafLocks.Update : SheafLocks.Shared, out deserializer ) as Tally;

			if( tally == null )
			{
				// Allow the action client to retry sending the request because this error can sometimes occur when accounts are
				// being created, but one peer is faster and is trying to send a message to the other.
				throw new RetryCrossServiceException( "The tally of the remote peer is missing. It's likely that the remote peer"
						+ " has not (yet) added an account for your UPN."
					, null, new ExceptionDetail( "Remote UPN fingerprint", initiatorUpnFingerprint ) );
			}

			#endregion Load the tally

			#region Load the account

			SpeckleFilter localUpnFilter = new SpeckleFilter( Account.LocalUpnFingerprintFilterIdentifier
				, recipientUpnFingerprint );
			SpeckleFilter remoteUpnFilter = new SpeckleFilter( Account.RemoteUpnFingerprintFilterIdentifier
				, initiatorUpnFingerprint );

			Account account = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Account ) )
				, new SpeckleFilters( localUpnFilter, remoteUpnFilter ), StorehouseRealms.Accounts
				, serviceSavesAccount ? SheafLocks.Update : SheafLocks.Shared, out deserializer ) as Account;

			if( account == null )
			{
				// Allow the action client to retry sending the request because this error can sometimes occur when accounts are
				// being created, but one peer is faster and is trying to send a message to the other.
				throw new RetryCrossServiceException( string.Format( "The account with the remote UPN fingerprint \"{0}\" and"
					+ " local UPN fingerprint \"{1}\" is missing.", initiatorUpnFingerprint, recipientUpnFingerprint ) );
			}

			#endregion Load the account

			return new ActionAccountContext( tally, account );
		}

		/// <summary>
		/// Validates if the specified action request may be anonymous. An exception is thrown if the action request may not be
		/// anonymous.
		/// </summary>
		/// <param name="actionRequest">Action request to validate.</param>
		/// <param name="serviceContext">The action service context.</param>
		/// <returns>Returns the sheaf which proves that the local peer has knowledge of the initiator of the current action
		/// request. This may be null.</returns>
		/// <remarks>
		/// <para>An anonymous action request may be processed when its recipient requires no knowledge about its initiator, or
		/// its recipient has an account for the initiator (and the request is signed with the identity from that account), or
		/// its recipient has a progress (<see cref="CreateAccountClientProgress"/>) for a "CreateAccountAnonymousActionRequest"
		/// action request queued to be sent to the initiator (and the action request is self-signed with the identity from which
		/// the UPN from the progress was generated).</para>
		/// </remarks>
		private static Sheaf ValidateAllowedAnonymousAction( ActionRequest actionRequest, ActionServiceContext serviceContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( serviceContext, "serviceContext" );

			// Check if the recipient may processes the action request even if it has no knowledge of the initiator.
			AnonymousActionRequestWithMutualKnowledge requestWmk = actionRequest as AnonymousActionRequestWithMutualKnowledge;
			if( requestWmk == null )
				return null;

			// The action request (is anonymous and) requires mutual knowledge.

			// Verify if the action request was signed by the initiator.
			requestWmk.VerifySignature( requestWmk.InitiatorIdentity );

			Sheaf knowledge = ActionServiceUtils.GetAnonymousActionRequestKnowledge( requestWmk, serviceContext.Transaction );

			if( knowledge == null )
			{
				// The client should retry to send the action request because the knowledge sheaf may exist later; this is
				// essential when processing a "CreateAccountAnonymousActionRequest" request.
				throw new RetryCrossServiceException( "The anonymous action request may not be processed because no mutual"
						+ " knowledge exists. It's likely that the remote peer has not (yet) added an account for your UPN."
					, null, new ExceptionDetail( "Action request type", actionRequest.GetType().FullName )
					, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator ) );
			}

			Tally tally = knowledge as Tally;
			if( (tally != null) && PeerDocumentsUtils.IsBlockedPeer( tally.Approval ) )
			{
				throw new FatalCrossServiceException( "Can't process the action request for the tally because its approval"
						+ " indicates that the remote peer is blocked.", null
					, new ExceptionDetail( "Action request type", actionRequest.GetType().FullName )
					, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator )
					, new ExceptionDetail( "Tally locator", tally.InstanceLocator ) );
			}

			CreateAccountClientProgress actionProgress = knowledge as CreateAccountClientProgress;
			if( (actionProgress != null) && PeerDocumentsUtils.IsBlockedPeer( actionProgress.Approval ) )
			{
				throw new FatalCrossServiceException( "Can't process the action request because the approval of the action"
						+ " progress which represents the mutual knowledge indicates that the remote peer is blocked.", null
					, new ExceptionDetail( "Action request type", actionRequest.GetType().FullName )
					, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator )
					, new ExceptionDetail( "Action progress locator", actionProgress.InstanceLocator ) );
			}

			if( (tally != null) || (actionProgress != null) )
				return knowledge;

			throw new FatalCrossServiceException( "The anonymous action request may not be processed because no mutual knowledge"
					+ " exists. It's likely that the remote peer has not (yet) added an account for your UPN.", null
				, new ExceptionDetail( "Action request type", actionRequest.GetType().FullName )
				, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator ) );
		}

		/// <summary>
		/// Loads the existing action response associated to the specified action request.
		/// </summary>
		/// <param name="actionRequest">Action request to process.</param>
		/// <param name="serviceContext">The action service context.</param>
		/// <returns>Returns the existing action response, or null if none exists.</returns>
		private static ActionResponse LoadExistingActionResponse( ActionRequest actionRequest
			, ActionServiceContext serviceContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( serviceContext, "serviceContext" );

			IDeserializer existingResponseDeserializer;
			SpeckleFilter actionRequestFilter = new SpeckleFilter( ActionResponse.ActionRequestLocatorFilterIdentifier
				, actionRequest.InstanceLocator );

			ActionResponse actionResponse = serviceContext.Transaction.CollectFirstSheaf(
				PortableTypes.FindAllDerived( typeof( ActionResponse ) ), new SpeckleFilters( actionRequestFilter )
				, StorehouseRealms.ServiceSentActionResponses, SheafLocks.Shared
				, out existingResponseDeserializer ) as ActionResponse;

			if( actionResponse == null )
				return null;

			// If an action request has already been processed, send back the existing response (even if it's expired).

			// Verify that the existing action response is associated to the current action request, so that nobody else could
			// retrieve the response.
			if( !actionRequest.InstanceLocator.Equals( actionResponse.ActionRequestLocator )
				|| !actionRequest.InitiatorUpnFingerprint.Equals( actionResponse.RecipientUpnFingerprint )
				|| (actionRequest.ActionResponseType != actionResponse.GetType()) )
			{
				throw new FatalCrossServiceException( string.Format( "The existing action response (with the locator"
						+ " \"{0}\") is not associated to the current action request (with the locator \"{1}\")."
					, actionResponse.ActionRequestLocator, actionRequest.InstanceLocator ) );
			}

			// An anonymous action response is created for an action request which is (likely) encrypted with the recipient's
			// UPN. This means that anyone who has that UPN may decrypt the action request and resend it (or one with the same
			// instance locator) to the service in order to get the same response (which may contain sensitive information).
			if( (actionRequest as AnonymousActionRequest != null) || (actionResponse as AnonymousActionResponse != null) )
			{
				throw new FatalCrossServiceException( string.Format( "The existing action response (with the locator"
						+ " \"{0}\") can't be sent as a response to the current action request (with the locator \"{1}\")"
						+ " because one of them is anonymous."
						+ "\n\n"
						+ "An anonymous action response is created for an action request which is (likely) encrypted with the"
						+ " recipient's UPN. This means that anyone who has that UPN may decrypt the action request and resend"
						+ " it (or one with the same instance locator) to the service in order to get the same response (which"
						+ " may contain sensitive information)."
					, actionResponse.ActionRequestLocator, actionRequest.InstanceLocator ) );
			}

			serviceContext.ActionProcessedBefore = true;

			return actionResponse;
		}

		/// <summary>
		/// Checks if the sheaf transported by the specified action request already exists in the storehouse.
		/// </summary>
		/// <param name="transaction">Transaction to use.</param>
		/// <param name="actionRequest">Action request which contains the transported sheaf to check if it exists.</param>
		/// <returns>Returns true if the sheaf transported by the action request already exists in the storehouse.</returns>
		/// <remarks>
		/// <para><see cref="ActionRequest.TransportedSheaf"/> may miss for "actionRequest".</para>
		/// </remarks>
		private static bool CheckIfTransportedSheafExists( IStorehouseTransaction transaction, ActionRequest actionRequest )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );

			ISheaf transportedSheaf = actionRequest.TransportedSheaf;
			if( transportedSheaf == null )
				return false;

			SheafExistenceModes sheafExistence = transaction.SheafExists( transportedSheaf, StorehouseRealms.Posts );
			switch( sheafExistence )
			{
				case SheafExistenceModes.DoesntExist:
					return false;
				case SheafExistenceModes.IsIdentical:
					return true;
				case SheafExistenceModes.IsDifferent:
					throw new FatalCrossServiceException( "A different sheaf with the same locator as the sheaf transported by"
							+ " the action request already exists.", null
						, new ExceptionDetail( "Sheaf locator", transportedSheaf.InstanceLocator )
						, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator ) );
				default:
					throw ValidationUtils.Unhandled( sheafExistence, "sheafExistence" );
			}
		}

		/// <summary>
		/// Gets the action service progress, from the storehouse, associated to the specified action request.
		/// </summary>
		/// <param name="actionRequest">Action request for which to get the action service progress.</param>
		/// <param name="transaction">The transaction with which to load the documents.</param>
		/// <returns>Returns the loaded action service progress.</returns>
		/// <remarks>
		/// <para><see cref="ActionRequest.TransportedSheaf"/> may not miss for "actionRequest".</para>
		/// </remarks>
		private static ActionServiceProgress GetMandatoryActionServiceProgress( ActionRequest actionRequest
			, IStorehouseTransaction transaction )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( transaction, "transaction" );

			if( actionRequest.TransportedSheafLocator == null )
			{
				throw new ValidationException( string.Format( "The action service progress can't be loaded because there is no"
						+ " transported sheaf locator for the action request with the locator \"{0}\"."
					, actionRequest.InstanceLocator ) );
			}

			SpeckleFilter referredSheafFilter = new SpeckleFilter( ActionProgress.ReferredSheafLocatorFilterIdentifier
				, actionRequest.TransportedSheafLocator );

			IDeserializer deserializer;
			ActionServiceProgress actionServiceProgress = transaction.CollectFirstSheaf(
				PortableTypes.FindAllDerived( typeof( ActionServiceProgress ) )
				, new SpeckleFilters( referredSheafFilter ), StorehouseRealms.ActionProgresses
				, SheafLocks.Shared, out deserializer ) as ActionServiceProgress;

			if( actionServiceProgress == null )
			{
				throw new ValidationException( string.Format( "The action service progress is missing for the action request"
					+ " with the locator \"{0}\".", actionRequest.InstanceLocator ) );
			}

			return actionServiceProgress;
		}

		/// <summary>
		/// Checks if the transported sheaf of the specified action request already exists in the storehouse, and creates a basic
		/// action response for when action responses are not saved.
		/// </summary>
		/// <param name="actionRequest">Action request to process.</param>
		/// <param name="serviceContext">The action service context.</param>
		/// <returns>Returns the created basic action response, or null if none was created.</returns>
		private static ActionResponse CheckIfTransportedSheafExistsAndCreateBasicActionResponse( ActionRequest actionRequest
			, ActionServiceContext serviceContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( serviceContext, "serviceContext" );

			// Check if the transported sheaf already exists in the storehouse.
			bool transportedSheafExists = CheckIfTransportedSheafExists( serviceContext.Transaction, actionRequest );
			if( !transportedSheafExists )
				return null;

			if( !actionRequest.ServiceResponseIsConsistentWithStandardConstructor || !actionRequest.ServiceSavesProgress )
			{
				throw new TransportedSheafExistsException( "The transported sheaf already exists because the action"
						+ " request was already processed, but an action response wasn't saved and one can't be"
						+ " created dynamically from the action progress."
					, null
					, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator )
					, new ExceptionDetail( "Transported sheaf locator"
						, actionRequest.TransportedSheaf.InstanceLocator ) );
			}

			// Get the mandatory action service progress.
			ActionServiceProgress progress = GetMandatoryActionServiceProgress( actionRequest, serviceContext.Transaction );

			ActionResponse actionResponse = Activator.CreateInstance( actionRequest.ActionResponseType, CryptoUtils.GetLocator()
				, actionRequest, progress.LastProcessingAttempt.SecondaryProcessingState ) as ActionResponse;

			// Sign the action response.
			actionResponse.Sign( serviceContext.Vault.ProcessingIdentity, null );

			serviceContext.ActionProcessedBefore = true;

			return actionResponse;
		}

		/// <summary>
		/// Loads the account context of the initiator of the specified action request.
		/// </summary>
		/// <param name="actionRequest">Action request to process.</param>
		/// <param name="serviceContext">The action service context.</param>
		/// <returns>Returns the loaded initiator account context. This may be null if the action request has no initiator.
		/// </returns>
		/// <remarks>
		/// <para>The signature of the specified action request is verified with the identity of the loaded account.</para>
		/// </remarks>
		private static ActionAccountContext LoadInitiatorAccountContext( ActionRequest actionRequest
			, ActionServiceContext serviceContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( serviceContext, "serviceContext" );

			Vault vault = serviceContext.Vault;

			Locator initiatorUpnFingerprint = null;
			ActionAccountContext initiatorAccountContext = null;

			if( actionRequest as ActionRequestWithInitiator == null )
				return null;

			ActionRequestWithInitiator actionRequestWithInitiator = (ActionRequestWithInitiator)actionRequest;
			initiatorUpnFingerprint = actionRequestWithInitiator.InitiatorUpnFingerprint;

			// Verify the recipient UPN of the action request.
			if( !actionRequestWithInitiator.RecipientUpnFingerprint.Equals( vault.LocalUpnFingerprint ) )
			{
				throw new FatalCrossServiceException( string.Format( "The vault with the UPN fingerprint \"{0}\" is not the"
						+ " recipient of the action request with the locator \"{1}\"."
					, vault.LocalUpnFingerprint, actionRequest.InstanceLocator ) );
			}

			initiatorAccountContext = LoadActionAccountContext( initiatorUpnFingerprint
				, actionRequestWithInitiator.RecipientUpnFingerprint, actionRequestWithInitiator.ServiceSavesInitiatorTally
				, actionRequestWithInitiator.ServiceSavesInitiatorAccount, serviceContext.Transaction );

			// Don't use expired documents.
			KernelUtils.HasExpired( initiatorAccountContext );

			Tally tally = initiatorAccountContext.Tally;

			if( PeerDocumentsUtils.IsBlockedPeer( tally.Approval ) )
			{
				throw new FatalCrossServiceException( "Can't process the action request for the tally because its approval"
						+ " indicates that the remote peer is blocked.", null
					, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator )
					, new ExceptionDetail( "Tally locator", tally.InstanceLocator )
					, new ExceptionDetail( "Account locator", initiatorAccountContext.AccountLocator ) );
			}

			if( (tally.Approval <= PeerDocumentsConsts.ProcessNonAnonymousActionRequestsTallyApprovalMin)
				&& (actionRequest as AnonymousActionRequest == null) )
			{
				string message = string.Format( "Can't process the action request for the initiator account because its"
							+ " approval (\"{0}\") is \"{1}\" or lower and the action request is not anonymous."
						, tally.Approval, PeerDocumentsConsts.ProcessNonAnonymousActionRequestsTallyApprovalMin );

				throw new FatalCrossServiceException( message, null
					, new ExceptionDetail( "Action request type", actionRequest.GetType().FullName )
					, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator )
					, new ExceptionDetail( "Tally locator", tally.InstanceLocator )
					, new ExceptionDetail( "Account locator", initiatorAccountContext.AccountLocator ) );
			}

			// Verify the initiator UPN of the action request.
			if( !actionRequestWithInitiator.InitiatorUpnFingerprint.Equals( tally.RemoteUpnFingerprint ) )
			{
				// The client should retry to send the action request because the tally may exist later; this is essential when
				// there is a race between processing "CreateAccountAnonymousActionRequest" requests and sending a letter from
				// the peer which has created an account first.
				throw new RetryCrossServiceException( "The tally is not the initiator of the action request.", null
					, new ExceptionDetail( "Remote UPN fingerprint", tally.RemoteUpnFingerprint )
					, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator ) );
			}

			// Verify if the action request was signed by the initiator.
			actionRequest.VerifySignature( tally.ProcessingIdentity );

			serviceContext.RemoteUpnFingerprint = initiatorAccountContext.RemoteUpnFingerprint;

			return initiatorAccountContext;
		}

		/// <summary>
		/// Loads the account context of the beneficiary of the specified action request.
		/// </summary>
		/// <param name="actionRequest">Action request to process.</param>
		/// <param name="serviceContext">The action service context.</param>
		/// <param name="initiatorAccountContext">Initiator account context.</param>
		/// <returns>Returns the loaded beneficiary account context. This may be null if the action requst has no beneficiary.
		/// </returns>
		private static ActionAccountContext LoadBeneficiaryAccountContext( ActionRequest actionRequest
			, ActionServiceContext serviceContext, ActionAccountContext initiatorAccountContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( serviceContext, "serviceContext" );
			ValidationUtils.ObjectHasContent( initiatorAccountContext, "initiatorAccountContext" );

			if( actionRequest as ActionRequestWithBeneficiary == null )
				return null;

			ActionRequestWithBeneficiary actionRequestWithBeneficiary = (ActionRequestWithBeneficiary)actionRequest;
			Locator beneficiaryUpnFingerprint = actionRequestWithBeneficiary.BeneficiaryUpnFingerprint;
			Upn beneficiaryUpn = actionRequestWithBeneficiary.BeneficiaryUpn;

			ActionAccountContext beneficiaryAccountContext = LoadActionAccountContext( beneficiaryUpnFingerprint
				, actionRequestWithBeneficiary.RecipientUpnFingerprint, actionRequestWithBeneficiary.ServiceSavesBeneficiaryTally
				, actionRequestWithBeneficiary.ServiceSavesBeneficiaryAccount, serviceContext.Transaction );

			// Don't use expired documents.
			KernelUtils.HasExpired( beneficiaryAccountContext );

			// The initiator and the beneficiary may not be the same.
			if( initiatorAccountContext.AccountLocator.Equals( beneficiaryAccountContext.AccountLocator ) )
			{
				throw new FatalCrossServiceException( string.Format( "The account locator (\"{0}\") of the initiator of"
						+ " the action request and the account locator (\"{1}\") of the beneficiary are the same."
					, initiatorAccountContext.AccountLocator, beneficiaryAccountContext.AccountLocator ) );
			}

			if( PeerDocumentsUtils.IsBlockedPeer( beneficiaryAccountContext.Tally.Approval ) )
			{
				throw new FatalCrossServiceException( string.Format( "Can't process any action request in the beneficiary"
						+ " account with the locator \"{0}\" because its approval indicates that the remote peer is blocked."
					, beneficiaryAccountContext.AccountLocator ) );
			}

			if( (beneficiaryAccountContext.Tally.Approval
					<= PeerDocumentsConsts.ProcessNonAnonymousActionRequestsTallyApprovalMin)
				&& (actionRequest as AnonymousActionRequest == null) )
			{
				throw new FatalCrossServiceException( string.Format( "Can't process an action request of type \"{0}\" in"
						+ " the beneficiary account with the locator \"{1}\" because its approval (\"{2}\") is \"{3}\" or lower"
						+ " and the action request is not anonymous."
					, actionRequest.GetType().FullName, beneficiaryAccountContext.AccountLocator
					, beneficiaryAccountContext.Tally.Approval
					, PeerDocumentsConsts.ProcessNonAnonymousActionRequestsTallyApprovalMin ) );
			}

			#region Verify the fingerprint of the beneficiary identity

			Upn processingIdentityUpn = beneficiaryAccountContext.Tally.ProcessingIdentity.Upn;

			if( !processingIdentityUpn.Equals( beneficiaryUpn ) )
			{
				throw new FatalCrossServiceException( string.Format( "The UPN \"{0}\" of the processing identity of the"
						+ " beneficiary account is different than the beneficiary UPN \"{1}\" from the action request."
					, processingIdentityUpn, beneficiaryUpn ) );
			}

			#endregion Verify the fingerprint of the beneficiary identity

			return beneficiaryAccountContext;
		}

		#endregion Private static methods

		#region Public methods

		/// <summary>
		/// Processes the specified action request (as a first frame of a request which has a multiframe response).
		/// </summary>
		/// <param name="actionRequest">The action request to process.</param>
		/// <param name="serviceContext">The action service context.</param>
		/// <param name="processRequestMethodInfo">The information of the method which can process the action request.</param>
		/// <param name="actionRequestContext">Is set with the created action request context.</param>
		/// <returns>Returns the action response which is to be sent back to the initiator of the action request.</returns>
		/// <remarks>
		/// <para>Expired action requests are not processed.</para>
		/// </remarks>
		public ActionResponse ProcessFirstRequestFrame( ActionRequest actionRequest
			, ActionServiceContext serviceContext, ProcessRequestMethodInfo processRequestMethodInfo
			, out ActionRequestContext actionRequestContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( serviceContext, "serviceContext" );
			ValidationUtils.ObjectHasContent( processRequestMethodInfo, "processRequestMethodInfo" );

			Vault serviceVault = serviceContext.Vault;

			// Don't process expired action requests. For anonymous action requests, the action request validity is always
			// the same because the client doesn't have the service's presenter from which to get the validity.
			int actionRequestValidity = (actionRequest as AnonymousActionRequest != null)
				? PeerDocumentsConsts.ActionRequestValidity : serviceVault.Appendix.ActionRequestValidity;
			KernelUtils.HasExpired( actionRequest, actionRequestValidity, true );

			Sheaf anonymousKnowledge = ValidateAllowedAnonymousAction( actionRequest, serviceContext );

			LogManager.DebugAddFormat( "The action request with the locator \"{0}\" and the type \"{1}\" was received on port"
				+ " {2}.", actionRequest.InstanceLocator, actionRequest.GetType().FullName, serviceContext.ProcessingPort );

			#region Process the action request

			// Load the initiator account first (= before checking if an associated action response exists) because if the
			// same action request is received multiple times (simultaneously), only one thread will succeed in locking the
			// initiator account, while the others will wait (on the locked resource) and when they are unlocked, they will
			// find the response in the storehouse (instead of being rejected later because the creation time of the action
			// request is now before the last action time from the initiator account).
			ActionAccountContext initiatorAccountContext = LoadInitiatorAccountContext( actionRequest, serviceContext );
			serviceContext.InitiatorAccountContext = initiatorAccountContext;

			// Pass through the traffic filter.
			if( initiatorAccountContext != null )
			{
				TrafficFilter.UpdateKnownPeer( serviceContext.InitiatorIpAddress, initiatorAccountContext.Tally.IsKnownPeer() );
			}
			else if( anonymousKnowledge != null )
			{
				CreateAccountClientProgress progressTyped = anonymousKnowledge as CreateAccountClientProgress;
				if( progressTyped != null )
				{
					TrafficFilter.UpdateKnownPeer( serviceContext.InitiatorIpAddress
						, PeerDocumentsUtils.IsKnownPeer( progressTyped.Approval ) );
				}
			}

			ActionResponse actionResponse = null;
			Locator actionResponseLocator;

			// Load the existing action response.
			if( actionRequest.ServiceSavesSentActionResponse )
			{
				actionResponse = LoadExistingActionResponse( actionRequest, serviceContext );
			}
			else if( actionRequest.ServiceIgnoresActionRequestIfTransportedSheafExists )
			{
				actionResponse = CheckIfTransportedSheafExistsAndCreateBasicActionResponse( actionRequest, serviceContext );
			}

			if( actionResponse != null )
			{
				// The action request was already processed.

				actionResponseLocator = actionResponse.InstanceLocator;
				actionRequestContext = null;
			}
			else
			{
				// There is no existing action response, so the action request is new.

				ActionAccountContext beneficiaryAccountContext = (initiatorAccountContext != null)
					? LoadBeneficiaryAccountContext( actionRequest, serviceContext, initiatorAccountContext ) : null;

				// Allocate the action response locator.
				actionResponseLocator = CryptoUtils.GetLocator();

				#region Process the action request

				// Create the action request context.
				actionRequestContext = CreateDerivedRequestContext( actionRequest.GetType(), serviceContext
					, anonymousKnowledge, actionResponseLocator, initiatorAccountContext, beneficiaryAccountContext );

				// Process the action request. The action response is then sent by the service to the initiator of an action
				// request.
				actionResponse = processRequestMethodInfo.MethodInfo.Invoke( this
					, new object[] { actionRequest, actionRequestContext } ) as ActionResponse;

				// If the action response is null, the service must end processing the current action request and must return a
				// "LastFrameActionResponse" action response.
				if( actionResponse == null )
				{
					if( actionRequest as IActionRequestWithMultiframeResponse == null )
					{
						throw new FatalCrossServiceException( string.Format( "The action response which is to be sent back to"
								+ " the initiator of the action request with the type \"{0}\" is missing."
							, actionRequest.GetType().FullName ) );
					}

					actionResponse = new LastFrameActionResponse( CryptoUtils.GetLocator(), actionRequest, null
						, serviceVault.GetSolitaryIdentity() );
				}

				#endregion Process the action request

				// Sign the action response.
				actionResponse.Sign( serviceVault.ProcessingIdentity, null );

				// Create the action progress.
				ActionServiceProgress actionProgress = null;
				if( actionRequest.ServiceSavesProgress )
				{
					actionProgress = new ActionServiceProgress( CryptoUtils.GetLocator()
						, actionRequest.InitiatorUpnFingerprint, actionRequest.RecipientUpnFingerprint
						, actionRequest.TransportedSheafLocator, actionRequest.GetType().FullName
						, actionResponse.SecondaryProcessingState );
				}

				#region Save documents

				// Save the action request.
				if( actionRequest.ServiceSavesReceivedActionRequest )
					serviceContext.Transaction.AddSheaf( actionRequest, StorehouseRealms.ServiceReceivedActionRequests );

				// Save the action response.
				if( actionRequest.ServiceSavesSentActionResponse && (actionResponse != null) )
					serviceContext.Transaction.AddSheaf( actionResponse, StorehouseRealms.ServiceSentActionResponses );

				// Save the tally and the account of the initiator of the action request.
				if( initiatorAccountContext != null )
				{
					ActionRequestWithInitiator actionRequestWithInitiator = (ActionRequestWithInitiator)actionRequest;

					if( actionRequestWithInitiator.ServiceSavesInitiatorTally )
					{
						serviceContext.Transaction.UpdateSheaf( initiatorAccountContext.Tally
							, StorehouseRealms.ProcessingIdentityContainers );
					}

					if( actionRequestWithInitiator.ServiceSavesInitiatorAccount )
						serviceContext.Transaction.UpdateSheaf( initiatorAccountContext.Account, StorehouseRealms.Accounts );
				}

				// Save the tally and the account of the beneficiary of the action request.
				if( beneficiaryAccountContext != null )
				{
					ActionRequestWithBeneficiary arWithBeneficiary = (ActionRequestWithBeneficiary)actionRequest;

					if( arWithBeneficiary.ServiceSavesBeneficiaryTally )
					{
						serviceContext.Transaction.UpdateSheaf( beneficiaryAccountContext.Tally
							, StorehouseRealms.ProcessingIdentityContainers );
					}

					if( arWithBeneficiary.ServiceSavesBeneficiaryAccount )
						serviceContext.Transaction.UpdateSheaf( beneficiaryAccountContext.Account, StorehouseRealms.Accounts );
				}

				// Save the action progress.
				if( actionProgress != null )
					serviceContext.Transaction.AddSheaf( actionProgress, StorehouseRealms.ActionProgresses );

				#endregion Save documents
			}

			// Commit the transaction.
			serviceContext.Transaction.Commit();

			#endregion Process the action request

			LogManager.DebugAddFormat( "The action request with the locator \"{0}\" and the type \"{1}\" was successfully"
					+ " processed on port {2}."
				, actionRequest.InstanceLocator, actionRequest.GetType().FullName, serviceContext.ProcessingPort );

			return actionResponse;
		}

		/// <summary>
		/// Processes the specified action request (as a secondary frame of a request which has a multiframe response).
		/// </summary>
		/// <param name="actionRequest">The action request to process.</param>
		/// <param name="serviceContext">The action service context.</param>
		/// <param name="processRequestMethodInfo">The information of the method which can process the action request.</param>
		/// <param name="actionRequestContext">The action request context created when the first frame was processed.</param>
		/// <returns>Returns the action response which is to be sent back to the initiator of the action request.</returns>
		public ActionResponse ProcessSecondaryRequestFrame( ActionRequest actionRequest
			, ActionServiceContext serviceContext, ActionRequestContext actionRequestContext
			, ProcessRequestMethodInfo processRequestMethodInfo )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( serviceContext, "serviceContext" );
			ValidationUtils.ObjectHasContent( actionRequestContext, "actionRequestContext" );
			ValidationUtils.ObjectHasContent( processRequestMethodInfo, "processRequestMethodInfo" );

			Vault serviceVault = serviceContext.Vault;

			#region Process the action request

			// Process the action request. The action response is then sent by the service to the initiator of an action
			// request.
			ActionResponse actionResponse = processRequestMethodInfo.MethodInfo.Invoke( this
				, new object[] { actionRequest, actionRequestContext } ) as ActionResponse;

			// If the action response is null, the service must end processing the current action request and must return a
			// "LastFrameActionResponse" action response.
			if( actionResponse == null )
			{
				if( actionRequest as IActionRequestWithMultiframeResponse == null )
				{
					throw new FatalCrossServiceException( string.Format( "The action response which is to be sent back to"
							+ " the initiator of the action request with the type \"{0}\" is missing."
						, actionRequest.GetType().FullName ) );
				}

				actionResponse = new LastFrameActionResponse( CryptoUtils.GetLocator(), actionRequest, null
					, serviceVault.GetSolitaryIdentity() );
			}

			#endregion Process the action request

			// Sign the action response. The secondary frames of a multiframe action response are not signed if they are
			// derived from "IMultiframeActionResponse".
			if( actionResponse as IMultiframeActionResponse == null )
				actionResponse.Sign( serviceVault.ProcessingIdentity, null );

			return actionResponse;
		}

		#endregion Public methods
	}
}
