////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using Vela.Cloning;
using Vela.Kernel;
using Vela.Networking;

namespace Vela.ActionService
{
	/// <summary>
	/// A service listener which receives action requests, dispatches them to the appropriate service, and finally sends back the
	/// action responses.
	/// </summary>
	/// <remarks>
	/// <para>Action requests are received on background threads.</para>
	/// 
	/// <para>Expired vaults are not used.</para>
	/// </remarks>
	public class ActionServiceListener
	{
		#region Public constants

		/// <summary>
		/// The time (in seconds) to sleep between consecutive attempts to bind the service socket.
		/// </summary>
		public const int BindServiceSocketRetrySleep = 10;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionServiceListener()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="listenerContext">Service listener context. This is cloned in here, so it doesn't have to be locked.
		/// </param>
		public ActionServiceListener( ActionServiceListenerContext listenerContext )
		{
			ListenerContext = Cloner.Deep( listenerContext );

			Start();
		}

		#endregion Constructors

		#region Private static fields

		/// <summary>
		/// The last unique identifier of an instance of a listener worker.
		/// </summary>
		private static Int64 LastListenerWorkerInstanceId;

		#endregion Private static fields

		#region Private fields

		/// <summary>
		/// Action service listener context.
		/// </summary>
		private ActionServiceListenerContext ListenerContext;

		/// <summary>
		/// The number of physical and virtual processor cores of the system.
		/// </summary>
		private int ProcessorCoresCount = SystemUtils.GetProcessorCoresCount();

		/// <summary>
		/// The synchronization object for the <see cref="ListenerWorkers"/> property.
		/// </summary>
		private readonly object ListenerWorkersLock = new object();

		/// <summary>
		/// The action service listener workers. The backing field is initialized by default. This may be empty, but not null.
		/// </summary>
		private ActionServiceListenerWorkers ListenerWorkers = new ActionServiceListenerWorkers();

		/// <summary>
		/// The thread where the service listener listens for client connections.
		/// </summary>
		private Thread ProcessingThread;

		/// <summary>
		/// The synchronization object for the <see cref="StopProcessingThread"/> property.
		/// </summary>
		private readonly object StopProcessingThreadLock = new object();

		private bool _StopProcessingThread = false;
		/// <summary>
		/// If this is true then the channel thread must stop as soon as possible.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// </remarks>
		private bool StopProcessingThread
		{
			get
			{
				lock( StopProcessingThreadLock )
				{
					return _StopProcessingThread;
				}
			}
			set
			{
				lock( StopProcessingThreadLock )
				{
					_StopProcessingThread = value;
				}
			}
		}

		/// <summary>
		/// The synchronization object for the <see cref="RequestsBeingProcessed"/> property.
		/// </summary>
		private readonly object RequestsBeingProcessedLock = new object();

		private int _RequestsBeingProcessed = 0;
		/// <summary>
		/// The number of requests which are being processed by all the threads.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// </remarks>
		private int RequestsBeingProcessed
		{
			get
			{
				lock( RequestsBeingProcessedLock )
				{
					return _RequestsBeingProcessed;
				}
			}
			set
			{
				lock( RequestsBeingProcessedLock )
				{
					_RequestsBeingProcessed = value;
				}
			}
		}

		/// <summary>
		/// The synchronization object for the <see cref="IsListening"/> property.
		/// </summary>
		private readonly object IsListeningLock = new object();

		#endregion Private fields

		#region Public non-serializable properties

		/// <summary>
		/// The network port of a service where action requests are to be sent by a client. READ-ONLY.
		/// </summary>
		public UInt16 ProcessingPort
		{
			get { return ListenerContext.Configuration.ProcessingPort; }
		}

		private bool _IsListening;
		/// <summary>
		/// If this is true then the listener is listening for client requests. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// </remarks>
		public bool IsListening
		{
			get
			{
				lock( IsListeningLock )
				{
					return _IsListening;
				}
			}
			protected set
			{
				lock( IsListeningLock )
				{
					_IsListening = value;
				}
			}
		}

		#endregion Public non-serializable properties

		#region Private static methods

		/// <summary>
		/// Creates a new unique identifier for an instance of a listener worker.
		/// </summary>
		/// <returns>Returns a new unique identifier.</returns>
		/// <remarks>
		/// <para><see cref="LastListenerWorkerInstanceId"/> is thread safe.</para>
		/// 
		/// <para><see cref="LastListenerWorkerInstanceId"/> is incremented here.</para>
		/// </remarks>
		private static Int64 CreateListenerWorkerInstanceId()
		{
			return Interlocked.Increment( ref LastListenerWorkerInstanceId );
		}

		#endregion Private static methods

		#region Private methods

		/// <summary>
		/// Event raised when an action service listener worker finishes processing an action request.
		/// </summary>
		/// <param name="instanceId">The unique identifier of the listener worker. This must be at least 0.</param>
		/// <remarks>
		/// <para>The <see cref="ListenerWorkers"/> property is thread safe within this method.</para>
		/// 
		/// <para>This is called on the listener worker thread.</para>
		/// </remarks>
		private void ListenerWorkerFinished( Int64 instanceId )
		{
			ValidationUtils.Integer( instanceId, 0, Int64.MaxValue, "instanceId" );

			lock( ListenerWorkersLock )
			{
				ListenerWorkers.Remove( instanceId );
			}
		}

		/// <summary>
		/// Binds the specified service socket to the specified local endpoint. The bind is retried until it's successful.
		/// </summary>
		/// <param name="serviceSocket">Service socket to bind.</param>
		/// <param name="localEndpoint">Local endpoint to which to bind the socket.</param>
		private void BindServiceSocket( Socket serviceSocket, IPEndPoint localEndpoint )
		{
			ValidationUtils.ObjectHasContent( serviceSocket, "serviceSocket" );
			ValidationUtils.ObjectHasContent( localEndpoint, "localEndpoint" );

			while( true )
			{
				try
				{
					serviceSocket.Bind( localEndpoint );

					// Log success.
					string message = string.Format( "The action service listener has bound its service socket for the port {0}."
						, ProcessingPort );
					LogManager.Add( message );

					return;
				}
				catch( SocketException ex )
				{
					if( ExceptionUtils.IsAbort( ex ) )
						throw ExceptionUtils.PreserveStackTrace( ex );

					// Log exception.
					string message = string.Format( "The action service listener has failed to bind its service socket for the"
						+ " port {0}. A retry will be made in {1} seconds. It's possible that this port is being used by another"
						+ " application, so close that application if you know which it is (use a TCP viewer to find it), or"
						+ " restart your computer.", ProcessingPort, BindServiceSocketRetrySleep );
					LogManager.Add( message, ex, LogRecordModes.Warning, false );

					TimeUtils.SleepSeconds( BindServiceSocketRetrySleep, () => { return StopProcessingThread; } );
				}

				if( StopProcessingThread )
					throw new AbortLongOperationException();
			}
		}

		/// <summary>
		/// Handler for a thread which processes requests.
		/// </summary>
		/// <param name="threadContext">Thread context. This must have the type <see cref="ActionServiceListenerWorker"/>.
		/// </param>
		/// <remarks>
		/// <para>All exceptions are caught in here, and therefore an exception which occurs while an action request is
		/// processed can't bring down the entire service.</para>
		/// </remarks>
		private void RequestThreadHandler( object threadContext )
		{
			ValidationUtils.ObjectType<ActionServiceListenerWorker>( threadContext, "threadContext" );

			try
			{
				ActionServiceListenerWorker listenerWorker = (ActionServiceListenerWorker)threadContext;
				listenerWorker.ProcessRequest();
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				if( ExceptionUtils.IsAbort( ex ) )
				{
					string message = string.Format( "The request handler of the action service listener has been aborted while"
						+ " processing an action request received on port {0}.", ProcessingPort );
					LogManager.DebugAdd( message, null );
				}
				else
				{
					string message = string.Format( "The request handler of the action service listener has failed to process an"
						+ " action request received on port {0}.", ProcessingPort );
					LogManager.Add( message, ex, LogRecordModes.Error, true );
				}
			}
			finally
			{
				lock( RequestsBeingProcessedLock )
				{
					RequestsBeingProcessed--;
				}

				LogManager.DebugAddFormat( "The action service listener has finished processing an action request received on"
					+ " port {0}.", ProcessingPort );
			}
		}

		/// <summary>
		/// Processes the received service requests and sends back a response to the initiator of the request.
		/// </summary>
		/// <remarks>
		/// <para>The service socket is closed in here.</para>
		/// 
		/// <para>The <see cref="ListenerWorkers"/> property is thread safe within this method.</para>
		/// </remarks>
		private void ProcessRequests()
		{
			// The socket which listens for incoming service request.
			Socket serviceSocket = null;

			try
			{
				int communicationTimeoutPerData = NetworkingUtils.GetCommunicationTimeoutPerData(
					ListenerContext.ProcessorConfiguration.CommunicationTimeoutPerBlob );

				// Create the socket which listens for clients. The socket must be reusable because a service may be opened,
				// closed and opened again in a time shorter than that in which a socket would normally become reusable.
				// A dual-stack socket is used in order to accept both IPv6 and IPv4 traffic.
				serviceSocket = new Socket( AddressFamily.InterNetworkV6, KernelConsts.SocketPacketType
					, KernelConsts.SocketProtocolType );
				serviceSocket.SetIPProtectionLevel( IPProtectionLevel.Unrestricted ); // Allow NAT traversal.
				serviceSocket.SetSocketOption( SocketOptionLevel.IPv6, SocketOptionName.IPv6Only, 0 ); // Allow IPv4 traffic.
				serviceSocket.ReceiveTimeout = TimeUtils.SecondsToMilli( communicationTimeoutPerData );
				serviceSocket.SendTimeout = serviceSocket.ReceiveTimeout;

				// Create the service listener endpoint.
				IPEndPoint localEndpoint = new IPEndPoint( IPAddress.IPv6Any, ProcessingPort );

				// Start listening for clients.
				BindServiceSocket( serviceSocket, localEndpoint );
				serviceSocket.Listen( ListenerContext.ProcessorConfiguration.ActionService.QueuedConnectionsMax );

				IsListening = true;

				string message = string.Format( "An action service listener was started for the port {0}.", ProcessingPort );
				LogManager.Add( message );

				#region The listening loop

				while( true )
				{
					Socket requestSocket = null;

					try
					{
						// Wait for a request.
						while( !serviceSocket.Poll( NetworkingConsts.SleepWaitingConnection, SelectMode.SelectRead ) )
						{
							if( StopProcessingThread )
								throw new AbortLongOperationException();
						}

						// Service requests are processed on separate threads.
						lock( RequestsBeingProcessedLock )
						{
							int requestsMax = (ListenerContext.ProcessorConfiguration.ParallelRequestsMax > 0)
								? ListenerContext.ProcessorConfiguration.ParallelRequestsMax : ProcessorCoresCount;

							if( RequestsBeingProcessed < requestsMax )
							{
								// Pass through the traffic filter.
								byte[] initiatorIpAddress = TrafficFilter.AcceptRequest( serviceSocket
									, ListenerContext.ProcessorConfiguration.ActionService, out requestSocket );

								// Create the thread which processes the current request.
								Thread requestThread = new Thread( RequestThreadHandler );

								// The thread must terminate in order for the application to terminate.
								requestThread.IsBackground = false;

								// Create the request processor. Clone the (necessary parts of the) listener context.
								ActionServiceListenerWorker listenerWorker = new ActionServiceListenerWorker(
									CreateListenerWorkerInstanceId(), ListenerContext, requestSocket, initiatorIpAddress
									, ListenerWorkerFinished );

								lock( ListenerWorkersLock )
								{
									ListenerWorkers.Add( listenerWorker );
								}

								// Start the thread which processes the current request.
								requestThread.Start( listenerWorker );

								RequestsBeingProcessed++;
							}
							else
							{
								// A connection is available, so accept it and close it immediately.
								requestSocket = serviceSocket.Accept();

								// Since the exception is thrown before the action request is loaded, it's not possible to send
								// back an action response, so there is no need for the exception type to be
								// "RetryCrossServiceException".
								throw new ValidationException( "The action service listener is processing too many requests." );
							}
						}
					}
					catch( Exception ex )
					{
						// The loop must continue (if possible).

						if( ExceptionUtils.IsCritical( ex ) )
							throw ExceptionUtils.PreserveStackTrace( ex );

						// Log exception.
						if( !ExceptionUtils.IsAbort( ex ) )
							LogManager.Add( ex );

						NetworkingUtils.CloseSocket( requestSocket );

						if( ExceptionUtils.IsAbort( ex ) )
							throw ExceptionUtils.PreserveStackTrace( ex );
					}
				}

				#endregion The listening loop
			}
			finally
			{
				IsListening = false;

				NetworkingUtils.CloseSocket( serviceSocket );
			}
		}

		/// <summary>
		/// Stops the listener workers.
		/// </summary>
		private void StopListenerWorkers()
		{
			lock( ListenerWorkersLock )
			{
				foreach( ActionServiceListenerWorker listenerWorker in ListenerWorkers )
					listenerWorker.Stop();
			}
		}

		/// <summary>
		/// Handler for the thread (<see cref="ProcessingThread"/>) which listens for client connections.
		/// </summary>
		/// <remarks>
		/// <para>The <see cref="ListenerWorkers"/> property is thread safe within this method.</para>
		/// </remarks>
		private void ListenerThreadHandler()
		{
			try
			{
				ProcessRequests();
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				if( ExceptionUtils.IsAbort( ex ) )
				{
					string message = string.Format( "The action service listener has aborted processing for the port {0}."
						, ProcessingPort );
					LogManager.Add( message );
				}
				else
				{
					string message = string.Format( "The action service listener has failed processing for the port {0}."
						, ProcessingPort );
					LogManager.Add( message, ex, LogRecordModes.Error, true );
				}
			}
			finally
			{
				StopListenerWorkers();
			}
		}

		/// <summary>
		/// Starts processing action requests on a new thread.
		/// </summary>
		private void Start()
		{
			ProcessingThread = new Thread( ListenerThreadHandler );

			// The thread must terminate in order for the application to terminate.
			ProcessingThread.IsBackground = false;

			ProcessingThread.Start();
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Stops listening for new action requests, and the processing of all the current action requests, as soon as possible.
		/// </summary>
		public void Stop()
		{
			StopProcessingThread = true;
		}

		/// <summary>
		/// Stops processing the action requests with the specified locators as soon as possible.
		/// </summary>
		/// <param name="requestLocators">The instance locators of the action requests whose processing to stop. If a locator
		/// refers to an action request which is not being processed, nothing happens (with that locator). This may be null or
		/// empty. This must be thread safe.</param>
		/// <remarks>
		/// <para>The <see cref="ListenerWorkers"/> property is thread safe within this method.</para>
		/// </remarks>
		public void StopRequests( Locators requestLocators )
		{
			lock( ListenerWorkersLock )
			{
				foreach( ActionServiceListenerWorker listenerWorker in ListenerWorkers )
					listenerWorker.StopRequests( requestLocators );
			}
		}

		#endregion Public methods
	}
}
