////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Net.Sockets;
using System.Reflection;
using System.Threading;
using Vela.ActionProcessor;
using Vela.Cloning;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Kernel;
using Vela.Networking;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.ActionService
{
	/// <summary>
	/// A worker for an action service listener.
	/// </summary>
	/// <remarks>
	/// <para>Expired vaults are not used.</para>
	/// </remarks>
	public class ActionServiceListenerWorker
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionServiceListenerWorker()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceId">The unique identifier of the current object. This must be at least 0.</param>
		/// <param name="listenerContext">Service listener context.</param>
		/// <param name="requestSocket">Request socket.</param>
		/// <param name="initiatorIpAddress">The IP address of the initiator of the request.</param>
		/// <param name="listenerWorkerFinished">Event raised when an action service listener worker finishes processing an
		/// action request.</param>
		public ActionServiceListenerWorker( Int64 instanceId, ActionServiceListenerContext listenerContext, Socket requestSocket
			, byte[] initiatorIpAddress, ListenerWorkerFinishedDelegate listenerWorkerFinished )
		{
			InstanceId = instanceId;

			ActionServiceListenerWorkerContext workerContext = new ActionServiceListenerWorkerContext( listenerContext
				, requestSocket, initiatorIpAddress );
			WorkerContext = Cloner.Deep( workerContext );

			ListenerWorkerFinished = listenerWorkerFinished;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// Action service listener worker context.
		/// </summary>
		private ActionServiceListenerWorkerContext WorkerContext;

		/// <summary>
		/// Event raised when an action service listener worker finishes processing an action request.
		/// </summary>
		private ListenerWorkerFinishedDelegate ListenerWorkerFinished;

		/// <summary>
		/// The synchronization object for the <see cref="StopProcessing"/> property.
		/// </summary>
		private readonly object StopProcessingLock = new object();

		private bool _StopProcessing = false;
		/// <summary>
		/// If this is true then the worker must stop as soon as possible.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// </remarks>
		private bool StopProcessing
		{
			get
			{
				lock( StopProcessingLock )
				{
					return _StopProcessing;
				}
			}
			set
			{
				lock( StopProcessingLock )
				{
					_StopProcessing = value;
				}
			}
		}

		/// <summary>
		/// The synchronization object for the <see cref="CurrentRequestLocator"/> property.
		/// </summary>
		private readonly object CurrentRequestLocatorLock = new object();

		private Locator _CurrentRequestLocator;
		/// <summary>
		/// The locator which is currently being processed.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// </remarks>
		private Locator CurrentRequestLocator
		{
			get
			{
				lock( CurrentRequestLocatorLock )
				{
					return _CurrentRequestLocator;
				}
			}
			set
			{
				lock( CurrentRequestLocatorLock )
				{
					_CurrentRequestLocator = value;
				}
			}
		}

		#endregion Private fields

		#region Public non-serializable properties

		private Int64 _InstanceId;
		/// <summary>
		/// The unique identifier of the current object. This must be at least 0. READ-ONLY.
		/// </summary>
		public Int64 InstanceId
		{
			get { return _InstanceId; }
			protected set
			{
				ValidationUtils.Integer( value, 0, Int64.MaxValue, "InstanceId" );

				_InstanceId = value;
			}
		}

		#endregion Public non-serializable properties

		#region Private static methods

		/// <summary>
		/// Protects against timing attacks.
		/// </summary>
		/// <param name="processingStartTicks">The ticks of the high resolution timer when the processing of the current action
		/// request has started.</param>
		/// <param name="timingAttackProtectionRoundoff">The rounded timespan (in seconds) up to which the service must wait as
		/// a form of protection against timing attacks.</param>
		/// <remarks>
		/// <para>To protect against timing attacks, the service waits up to a rounded timespan, after it processes an action
		/// request and before it sends back a response.</para>
		/// </remarks>
		private static void TimingAttackProtection( Int64 processingStartTicks, double timingAttackProtectionRoundoff )
		{
			ValidationUtils.Real( timingAttackProtectionRoundoff, ActionProcessorConsts.TimingAttackProtectionRoundoffMin
				, ActionProcessorConsts.TimingAttackProtectionRoundoffMax, "timingAttackProtectionRoundoff" );

			Int64 processingEndTicks = HighResolutionTimer.GetTicks();
			if( processingEndTicks <= processingStartTicks )
				return;

			Int64 actualTimespanTicks = processingEndTicks - processingStartTicks;
			double actualTimespan = HighResolutionTimer.TicksToSeconds( actualTimespanTicks );

			// Randomly add a roundoff unit.
			if( EntropyPool.GetBool() )
				actualTimespan += timingAttackProtectionRoundoff;

			double roundedTimespan = MathUtils.AlignUp( actualTimespan, timingAttackProtectionRoundoff );
			Int64 roundedTimespanTicks = HighResolutionTimer.SecondsToTicks( roundedTimespan );

			while( true )
			{
				Int64 currentTimespanTicks = (processingStartTicks + roundedTimespanTicks) - HighResolutionTimer.GetTicks();
				double currentTimespan = HighResolutionTimer.TicksToSeconds( currentTimespanTicks );

				// In order to keep time jumps under control, limit the timespan in both directions.
				if( (currentTimespan <= 0) || (currentTimespan > 1) )
					break;
			}
		}

		/// <summary>
		/// Instantiate an action service which can process the specified action request.
		/// </summary>
		/// <param name="actionRequest">Action request to process.</param>
		/// <param name="serviceVault">Service vault.</param>
		/// <param name="processRequestMethodInfos">List of method informations which can process action requests.</param>
		/// <param name="processRequestMethodInfo">Is set with the information of the method which can process the specified
		/// action request.</param>
		/// <returns>Returns the action service.</returns>
		private static IActionService InstantiateService( ActionRequest actionRequest, Vault serviceVault
			, ProcessRequestMethodInfos processRequestMethodInfos, out ProcessRequestMethodInfo processRequestMethodInfo )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( serviceVault, "vault" );

			// Find the cached method information.
			string actionRequestTypeName = actionRequest.GetType().FullName;
			processRequestMethodInfo = processRequestMethodInfos.Find( actionRequestTypeName );
			if( processRequestMethodInfo == null )
			{
				throw new FatalCrossServiceException( string.Format( "The action service can't process action requests with the"
					+ " type \"{0}\".", actionRequestTypeName ) );
			}

			// Check if the service vault is allowed to use the action service.
			Strings vaultActionServiceTypeNames = serviceVault.Appendix.ActionServiceTypeNames;
			if( (processRequestMethodInfo.IsEmbeddedService || serviceVault.Appendix.UseOnlyEmbeddedActionServices)
				&& !vaultActionServiceTypeNames.ContainsKey( processRequestMethodInfo.ServiceTypeName ) )
			{
				throw new FatalCrossServiceException( string.Format( "The action service \"{0}\" can't process action requests"
						+ " with the type \"{1}\" because the vault with the locator \"{2}\" is not allowed to use the service."
					, processRequestMethodInfo.ServiceTypeName, actionRequestTypeName, serviceVault.InstanceLocator ) );
			}

			// Instantiate the action service.
			Type serviceType = PortableTypes.FindGlobal( processRequestMethodInfo.ServiceTypeName );
			IActionService service = (serviceType != null)
				? (Activator.CreateInstance( serviceType, true ) as IActionService) : null;
			if( service == null )
			{
				throw new FatalCrossServiceException( string.Format( "An action service with the type \"{0}\" could not be"
					+ " instantiated.", processRequestMethodInfo.ServiceTypeName ) );
			}

			return service;
		}

		#endregion Private static methods

		#region Private methods

		/// <summary>
		/// Reveals the raw action request from the specified obfuscated action request, using the recipient UPNs.
		/// </summary>
		/// <param name="obfuscatedRequest">Obfuscated action request.</param>
		/// <param name="processingPort">Service processing port.</param>
		/// <param name="recipientUpn">Is set with the UPN of the recipient for which the action request is intended.</param>
		/// <returns>Returns the raw action request.</returns>
		private byte[] SteganographicReveal( byte[] obfuscatedRequest, UInt16 processingPort, out Upn recipientUpn )
		{
			ValidationUtils.ObjectHasContent( obfuscatedRequest, "obfuscatedRequest" );

			if( WorkerContext.ListenerContext.ActionServiceVaults.Count <= 0 )
			{
				throw new ValidationException( string.Format( "The action request received on port {0} can't be"
					+ " steganographically revealed because there are no service vaults.", processingPort ) );
			}

			foreach( ActionServiceVault actionServiceVault in WorkerContext.ListenerContext.ActionServiceVaults )
			{
				foreach( Upn upn in actionServiceVault.Upns )
				{
					byte[] rawRequest = SteganographicAlg.Reveal( obfuscatedRequest, upn );

					if( rawRequest != null )
					{
						recipientUpn = upn;

						return rawRequest;
					}
				}
			}

			throw new ValidationException( string.Format( "The action request received on port {0} can't be steganographically"
				+ " revealed.", processingPort ) );
		}

		/// <summary>
		/// Finds the target service vault of the specified action request.
		/// </summary>
		/// <param name="request">Action request.</param>
		/// <param name="recipientUpn">The UPN which is the recipient of the action request.</param>
		/// <param name="requestEnvelope">Is set with the encrypted envelope of the action request.</param>
		/// <returns>Returns the vault for which the action request is meant.</returns>
		private Vault FindServiceVault( byte[] request, Upn recipientUpn, out EncryptedEnvelope requestEnvelope )
		{
			ValidationUtils.ArrayHasContent( request, "request" );
			ValidationUtils.ObjectHasContent( recipientUpn, "recipientUpn" );

			// Find the target service which must receive the current action request.

			// Extract the action request envelope.
			IDeserializer deserializer;
			requestEnvelope = SerializationUtils.LoadFromBuffer( request, typeof( EncryptedEnvelope ), out deserializer )
				as EncryptedEnvelope;

			if( (requestEnvelope as AsymmEncryptedEnvelope == null) && (requestEnvelope as UpnEncryptedEnvelope == null) )
			{
				throw new ValidationException( string.Format( "An action service may process only encrypted envelopes of"
						+ " the types \"{0}\" or \"{1}\", not \"{2}\"."
					, typeof( AsymmEncryptedEnvelope ).FullName, typeof( UpnEncryptedEnvelope ).FullName
					, requestEnvelope.GetType().FullName ) );
			}

			ActionServiceListenerContext listenerContext = WorkerContext.ListenerContext;
			ActionServiceVault actionServiceVault = null;

			// Find the target service vault.
			if( requestEnvelope as AsymmEncryptedEnvelope != null )
			{
				AsymmRecipientSessionKeys rsks = ((AsymmEncryptedEnvelope)requestEnvelope).RecipientSessionKeys;

				foreach( AsymmRecipientSessionKey rsk in rsks )
				{
					Locator upnFingerprint = rsk.SolitarySessionKey.UpnFingerprint;
					actionServiceVault = listenerContext.ActionServiceVaults.Find( upnFingerprint );

					if( actionServiceVault != null )
					{
						ActionServiceVault upnActionServiceVault = listenerContext.ActionServiceVaults.FindByAnyUpnFingerprint(
							recipientUpn.Fingerprint );
						if( upnActionServiceVault != actionServiceVault )
						{
							throw new ValidationException( string.Format( "The revealed recipient UPN fingerprint identifies"
									+ " the vault with the locator \"{0}\", but the asymmetric recipient session key identifies"
									+ " vault with the locator \"{1}\"."
								, upnActionServiceVault.Vault.InstanceLocator, actionServiceVault.Vault.InstanceLocator ) );
						}

						break; // Service vault found.
					}
				}
			}
			else if( requestEnvelope as UpnEncryptedEnvelope != null )
			{
				// This is an action request for a public service.

				// Identify the intended service with the UPN fingerprint.
				Locator recipientUpnFingerprint = ((UpnEncryptedEnvelope)requestEnvelope).RecipientUpnFingerprint;
				if( !recipientUpn.Fingerprint.Equals( recipientUpnFingerprint ) )
				{
					throw new ValidationException( string.Format( "The revealed recipient UPN fingerprint \"{0}\" is"
							+ " different than the one from the action request \"{1}\"."
						, recipientUpn, recipientUpnFingerprint ) );
				}

				actionServiceVault = listenerContext.ActionServiceVaults.FindByAnyUpnFingerprint( recipientUpnFingerprint );
			}
			else
			{
				throw new ValidationException( string.Format( "The action service listener can't process encrypted"
					+ " envelopes of type \"{0}\".", requestEnvelope.GetType().FullName ) );
			}

			if( actionServiceVault == null )
			{
				throw new ValidationException( string.Format( "A vault with the processing identity with the same UPN"
					+ " fingerprint as the recipient UPN fingerprint of the action request with the locator \"{0}\" was not"
					+ " found.", requestEnvelope.InstanceLocator ) );
			}

			return actionServiceVault.Vault;
		}

		/// <summary>
		/// Decrypts the specified action request envelope.
		/// </summary>
		/// <param name="actionRequestEnvelope">Action request envelope to decrypt.</param>
		/// <param name="vault">Vault with which to decrypt the action request envelope.</param>
		/// <param name="recipientUpn">The UPN which is the recipient of the action request.</param>
		/// <returns>Returns the action request from the envelope.</returns>
		private ActionRequest DecryptRequestEnvelope( EncryptedEnvelope actionRequestEnvelope, Vault vault, Upn recipientUpn )
		{
			ValidationUtils.ObjectHasContent( actionRequestEnvelope, "actionRequestEnvelope" );
			ValidationUtils.ObjectHasContent( vault, "vault" );
			ValidationUtils.ObjectHasContent( recipientUpn, "recipientUpn" );

			#region Decrypt the action request

			// Get the password.
			object password;
			if( actionRequestEnvelope as AsymmEncryptedEnvelope != null )
			{
				// Verify the recipient UPN of the encrypted envelope. This is a double-check because the envelope of the
				// action request was encrypted with the UPN of the recipient.
				AsymmEncryptedEnvelope requestEnvelope = (AsymmEncryptedEnvelope)actionRequestEnvelope;
				if( !requestEnvelope.RecipientSessionKeys.ContainsKey( recipientUpn.Fingerprint ) )
				{
					throw new ValidationException( string.Format( "The UPN \"{0}\" is not the recipient of the asymmetric"
							+ " encrypted envelope with the locator \"{1}\"."
						, recipientUpn, actionRequestEnvelope.InstanceLocator ) );
				}

				password = new AsymmDecryptionPassword( vault.ProcessingIdentity );
			}
			else if( actionRequestEnvelope as UpnEncryptedEnvelope != null )
			{
				// Verify the recipient UPN of the encrypted envelope. This is a double-check because the envelope of the
				// action request was encrypted with the UPN of the recipient.
				Locator recipientUpnFingerprint = ((UpnEncryptedEnvelope)actionRequestEnvelope).RecipientUpnFingerprint;
				if( !recipientUpn.Equals( recipientUpnFingerprint ) )
				{
					throw new ValidationException( string.Format( "The fingerprint of the UPN \"{0}\" of the action service is"
						+ " not the same as the recipient UPN fingerprint \"{1}\" of the action request envelope with the"
						+ " locator \"{2}\".", recipientUpn, recipientUpnFingerprint, actionRequestEnvelope.InstanceLocator ) );
				}

				password = recipientUpn.SymmEncryptionKey;
			}
			else
			{
				throw new ValidationException( string.Format( "An action service may process only encrypted envelopes of"
						+ " the types \"{0}\" or \"{1}\", not \"{2}\"."
					, typeof( AsymmEncryptedEnvelope ).FullName, typeof( UpnEncryptedEnvelope ).FullName
					, actionRequestEnvelope.GetType().FullName ) );
			}

			// Load the action request.
			IDeserializer actionRequestDeserializer;
			EncryptedEnvelope.LoadFromEnvelope( actionRequestEnvelope, password, typeof( ActionRequest )
				, out actionRequestDeserializer );
			ActionRequest actionRequest = (ActionRequest)actionRequestEnvelope.EncryptableDocument;

			#endregion Decrypt the action request

			return actionRequest;
		}

		/// <summary>
		/// Extracts the action request from the specified obfuscated request.
		/// </summary>
		/// <param name="obfuscatedRequest">Obfuscated request from which to extract the action request.</param>
		/// <param name="requestEnvelope">Is set with the extracted encrypted envelope of the action request.</param>
		/// <param name="serviceVault">Is set with the service vault which is the recipient of the action request.</param>
		/// <returns>Returns the extracted action request.</returns>
		private ActionRequest ExtractActionRequest( byte[] obfuscatedRequest, out EncryptedEnvelope requestEnvelope
			, out Vault serviceVault )
		{
			ActionServiceListenerConfiguration serviceConfiguration = WorkerContext.ListenerContext.Configuration;

			// Reveal the action request.
			Upn recipientUpn;
			byte[] rawRequest = SteganographicReveal( obfuscatedRequest, serviceConfiguration.ProcessingPort, out recipientUpn );

			#region Decrypt the request envelope

			// Find the target service vault.
			Vault targetVault = FindServiceVault( rawRequest, recipientUpn, out requestEnvelope );

			// Clone the service vault (= the algorithms inside it) so that it doesn't have to be locked.
			serviceVault = Cloner.Deep( targetVault );

			// Don't use expired documents. This includes the recipient UPN and steganographic algorithm.
			KernelUtils.HasExpired( serviceVault );

			ActionRequest actionRequest = DecryptRequestEnvelope( requestEnvelope, serviceVault, recipientUpn );

			// Verify the recipient UPN of the action request. This is a double-check because the envelope of the
			// action request was encrypted with the UPN of the recipient.
			if( !recipientUpn.Equals( actionRequest.RecipientUpn ) )
			{
				throw new ValidationException( string.Format( "The UPN \"{0}\" of the action service is not the same as the"
						+ " recipient UPN \"{1}\" of the action request with the locator \"{2}\"."
					, recipientUpn, actionRequest.RecipientUpn, actionRequest.InstanceLocator ) );
			}

			#endregion Decrypt the request envelope

			return actionRequest;
		}

		/// <summary>
		/// Processes the received service request and sends back a response to the initiator of the request.
		/// </summary>
		/// <param name="actionRequest">The action request to process.</param>
		/// <param name="requestEnvelope">The request's encrypted envelope.</param>
		/// <param name="serviceVault">The recipient service vault.</param>
		/// <param name="communicationFrame">Communication frame (with which the request was received, and through which the
		/// response has to be sent).</param>
		/// <param name="notifyActionClientManagerAboutRequests">Is set with the action request locators about which the action
		/// client manager associated to the current service is notified to process. This may be null or empty.</param>
		private void ProcessRequest( ActionRequest actionRequest, EncryptedEnvelope requestEnvelope, Vault serviceVault
			, CommunicationFrame communicationFrame, out Locators notifyActionClientManagerAboutRequests )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( requestEnvelope, "requestEnvelope" );
			ValidationUtils.ObjectHasContent( serviceVault, "serviceVault" );
			ValidationUtils.ObjectHasContent( communicationFrame, "communicationFrame" );

			try
			{
				IStorehouseTransaction transaction = null;
				ActionResponse actionResponse = null;

				ActionServiceListenerContext listenerContext = WorkerContext.ListenerContext;
				ActionProcessorConfiguration processorConfiguration = listenerContext.ProcessorConfiguration;
				ActionServiceConfiguration serviceConfiguration = processorConfiguration.ActionService;
				ActionServiceListenerConfiguration listenerConfiguration = listenerContext.Configuration;
				UInt16 processingPort = listenerConfiguration.ProcessingPort;
				Upn recipientUpn = actionRequest.RecipientUpn;

				// The ticks of the high resolution timer when the processing of the current action request has started. This
				// is used to protect against timing attacks.
				Thread.BeginThreadAffinity(); // Fix a bug on some computers where the high resolution timer is unreliable.
				Int64 processingStartTicks = HighResolutionTimer.GetTicks();

				bool mayUpdateUiOnException = false;

				try
				{
					if( listenerContext.ActionServiceHost.CheckMaintenance() )
						throw new ServiceInMaintenanceException( "The action service is under maintenance." );

					// Inform the client that the service vault is revoked, but only if the client isn't trying to resolve the
					// service's UPN.
					if( (actionRequest as ResolveUpnAnonymousActionRequest == null) && serviceVault.Appendix.IsRevoked )
					{
						throw new RevokedVaultException( string.Format( "The service vault with the UPN \"{0}\" is revoked."
							, recipientUpn ) );
					}

					// Set the socket buffer size for sending the action response (frames).
					if( actionRequest.ActionResponseRequiresLargeCommunicationCacheSize
						&& (processorConfiguration.LargeCommunicationCacheSize > 0) )
					{
						WorkerContext.RequestSocket.SendBufferSize = processorConfiguration.LargeCommunicationCacheSize;
					}

					// Create a transaction.
					transaction = listenerContext.StorehouseProvider.CreateTransaction();

					ActionServiceContext serviceContext = new ActionServiceContext( listenerContext.StorehouseProvider
						, processorConfiguration, listenerConfiguration, WorkerContext.InitiatorIpAddress
						, listenerContext.StorehouseProvider.Serializer, serviceVault, recipientUpn, processingPort
						, transaction, listenerContext.ActionServiceHost, listenerContext.ActionClientManager );

					ProcessRequestMethodInfo processRequestMethodInfo;
					IActionService service = InstantiateService( actionRequest, serviceVault
						, listenerContext.ProcessRequestMethodInfos, out processRequestMethodInfo );

					ActionRequestContext actionRequestContext = null;

					Chronometer chronometer = new Chronometer();
					chronometer.Start();

					mayUpdateUiOnException = true;

					#region Send response frames

					bool continueSending = true;
					while( continueSending )
					{
						if( actionRequest as IActionRequestWithMultiframeResponse == null )
							continueSending = false;

						if( serviceContext.IsFirstFrame )
						{
							actionResponse = service.ProcessFirstRequestFrame( actionRequest, serviceContext
								, processRequestMethodInfo, out actionRequestContext );
						}
						else
						{
							actionResponse = service.ProcessSecondaryRequestFrame( actionRequest, serviceContext
								, actionRequestContext, processRequestMethodInfo );
						}

						// The transaction must have been closed by now.
						if( transaction.IsOpen )
						{
							throw new ValidationException( "The first frame of the multiframe action response was"
								+ " processed, but the transaction is still open. Storehouse timeouts may occur." );
						}

						bool isMultiframeResponse = actionResponse as IMultiframeActionResponse != null;

						// Stop sending responses if the current response is not a multiframe response, or if the request was
						// processed before (= "actionRequestContext" is null).
						if( !isMultiframeResponse || (actionRequestContext == null) )
							continueSending = false;

						// Create the encrypted action response.
						byte[] obfuscatedResponse = ActionServiceUtils.PrepareActionResponse( requestEnvelope, actionRequest
							, actionResponse, recipientUpn, listenerContext.StorehouseProvider.Serializer );

						if( serviceContext.IsFirstFrame )
							TimingAttackProtection( processingStartTicks, serviceConfiguration.TimingAttackProtectionRoundoff );

						// The host UI must not be updated too often because that would cause significant performance
						// degradation. Also, don't notify the action service host of an action which was already processed
						// (otherwise you could end up with the same letter being shown multiple times in the UI, even though
						// it's only one in the storehouse).
						//
						// Perform a full UI update for the first frame in order to ensure that the UI is up to date when the
						// multiframe response is resumed (after an interruption). Perform a full UI update for the last frame.
						bool mustFullyUpdateUi = serviceContext.IsFirstFrame || !continueSending;
						if( !serviceContext.ActionProcessedBefore
							&& (
								mustFullyUpdateUi
								|| chronometer.PassedSeconds( ActionProcessorConsts.MultiframeActionResponseUpdateUiTimespan )
							) )
						{
							Tally initiatorTally = (serviceContext.InitiatorAccountContext != null)
								? serviceContext.InitiatorAccountContext.Tally : null;
							Account initiatorAccount = (serviceContext.InitiatorAccountContext != null)
								? serviceContext.InitiatorAccountContext.Account : null;

							// Notify the action processor host to update the UI. Do this outside the main transaction because
							// another one may be created and may try to load the sheafs which are locked by the main
							// transaction.
							ActionServiceHostUpdateUiContext updateUiContext = new ActionServiceHostUpdateUiContext(
								serviceConfiguration, serviceVault, initiatorTally, initiatorAccount, actionRequest
								, actionResponse, mustFullyUpdateUi );
							WorkerContext.ListenerContext.ActionServiceHost.UpdateUi( updateUiContext );
						}

						// Send back the response.
						communicationFrame.SendMessage( WorkerContext.RequestSocket, obfuscatedResponse );

						serviceContext.IsFirstFrame = false;

						// Check if the worker must stop processing.
						if( StopProcessing )
						{
							if( isMultiframeResponse )
							{
								LogManager.DebugAddFormat( "The action service listener has stopped processing the"
										+ " multiframe response for the action request with the locator \"{0}\" and the type"
										+ " \"{1}\"."
									, actionRequest.InstanceLocator, actionRequest.GetType().FullName );
							}

							throw new AbortLongOperationException();
						}
					}

					#endregion Send response frames

					// Add new entropy data to the entropy algorithm. The data isn't known to the attackers who physically
					// monitor the action requests received by the service, because they are sent asymmetrically
					// encrypted (while the data is obtained after decryption).
					//
					// An action request with initiator can't be encrypted in envelopes with the "UpnEncryptedEnvelope" type,
					// so an attacker who knows the UPN of the recipient, can't know the content of the request.
					//
					// The data should be unique, so the instance locator of the transported sheaf is used.
					if( (actionRequest as ActionRequestWithInitiator != null)
						&& (actionRequest.TransportedSheafLocator != null)
						&& !serviceContext.ActionProcessedBefore )
					{
						EntropyPool.Add( actionRequest.TransportedSheafLocator.Value );
					}

					notifyActionClientManagerAboutRequests = serviceContext.ActionProcessedBefore
						? null : serviceContext.NotifyActionClientManagerAboutRequests;
				}
				catch( Exception ex )
				{
					// Rolls back the transaction.
					if( transaction != null )
						transaction.Rollback();

					notifyActionClientManagerAboutRequests = null;

					if( ExceptionUtils.IsAbort( ex ) )
						throw ExceptionUtils.PreserveStackTrace( ex );

					// Log exception.
					string message = string.Format( "Failed to process the action request with the locator \"{0}\" and the"
						+ " type \"{1}\".", actionRequest.InstanceLocator, actionRequest.GetType().FullName );
					LogManager.Add( message, ex, LogRecordModes.Error, true );

					// An error response is allowed because the request was steganographically revealed (so the sender knows
					// the UPN of the recipient), and the action request was properly loaded.

					// Try to send back a gracious error.
					actionResponse = ActionServiceUtils.CreateErrorResponse( actionRequest, serviceVault
						, ex, listenerContext.StorehouseProvider.IsTimeoutException( ex ) );

					try
					{
						// Notify the action processor host to update the UI. Do this outside the main transaction because
						// another one may be created and may try to load the sheafs which are locked by the main transaction.
						if( mayUpdateUiOnException && actionRequest.ServiceMayUpdateUiOnException )
						{
							ActionServiceHostUpdateUiContext updateUiContext = new ActionServiceHostUpdateUiContext(
								serviceConfiguration, serviceVault, null, null, actionRequest, actionResponse, true );
							WorkerContext.ListenerContext.ActionServiceHost.UpdateUi( updateUiContext );
						}
					}
					finally
					{

						// Create the encrypted action response.
						byte[] obfuscatedResponse = ActionServiceUtils.PrepareActionResponse( requestEnvelope, actionRequest
							, actionResponse, recipientUpn, listenerContext.StorehouseProvider.Serializer );

						// Send back the response.
						TimingAttackProtection( processingStartTicks, serviceConfiguration.TimingAttackProtectionRoundoff );
						communicationFrame.SendMessage( WorkerContext.RequestSocket, obfuscatedResponse );
					}
				}
			}
			catch( Exception ex )
			{
				string message = string.Format( "Failed to process the action request with the locator \"{0}\" and the"
					+ " type \"{1}\".", actionRequest.InstanceLocator, actionRequest.GetType().FullName );
				throw new ValidationException( message, ExceptionUtils.PreserveStackTrace( ex ) );
			}
			finally
			{
				Thread.EndThreadAffinity();
			}
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Processes the received service request and sends back a response to the initiator of the request.
		/// </summary>
		/// <remarks>
		/// <para>The request socket is closed in here.</para>
		/// </remarks>
		public void ProcessRequest()
		{
			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				ActionServiceListenerContext listenerContext = WorkerContext.ListenerContext;
				ActionProcessorConfiguration processorConfiguration = listenerContext.ProcessorConfiguration;
				ActionServiceConfiguration serviceConfiguration = listenerContext.ProcessorConfiguration.ActionService;
				ActionServiceListenerConfiguration listenerConfiguration = WorkerContext.ListenerContext.Configuration;

				// The communication frame algorithm must be the same for all the UPNs which have the same processing port.
				CommunicationFrame communicationFrame = new CommunicationFrame( listenerConfiguration.CommunicationFrameAlg
					, processorConfiguration.CommunicationTimeoutPerBlob, processorConfiguration.CommunicationFrameSizeMax );

				// Wait for the action request. The data can't be read here if the communication frame algorithm of the current
				// listener worker is different then the communication frame algorithm of the recipient UPN; this can happen when
				// the port of the current listener worker is a sender-overridden port, not the port from the recipient's UPN,
				// and the two ports have a different communication frame algorithm.
				byte[] obfuscatedRequest = communicationFrame.ReceiveMessage( WorkerContext.RequestSocket );

				EncryptedEnvelope requestEnvelope;
				Vault serviceVault;
				ActionRequest actionRequest = ExtractActionRequest( obfuscatedRequest, out requestEnvelope, out serviceVault );
				CurrentRequestLocator = actionRequest.InstanceLocator;

				// Pass through the traffic filter. The size of the downloaded files isn't filtered (because that is processed
				// in a multiframe request, in the "ProcessRequest" call below).
				TrafficFilter.UpdateDataSize( WorkerContext.InitiatorIpAddress, obfuscatedRequest.Length );

				// Process the action request.
				Locators notifyActionClientManagerAboutRequests;
				ProcessRequest( actionRequest, requestEnvelope, serviceVault, communicationFrame
					, out notifyActionClientManagerAboutRequests );

				listenerContext.ActionClientManager.RecipientAccessStatesManager.RecipientIsAccessibleAndProcessing(
					actionRequest.InitiatorUpnFingerprint );

				try
				{
					// Update the ephemerid (on a separate transaction so as to not affect the scalability of the action
					// processor).
					ActionProcessorStorehouseUtils.UpdateEphemerid( listenerContext.StorehouseProvider
						, actionRequest.InitiatorUpnFingerprint, WorkerContext.InitiatorIpAddress, actionRequest );
				}
				finally
				{
					// Notify the action client manager that action requests were added to the storehouse.
					if( notifyActionClientManagerAboutRequests != null )
						listenerContext.ActionClientManager.Add( notifyActionClientManagerAboutRequests, true );
				}
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );

				try
				{
					NetworkingUtils.CloseSocket( WorkerContext.RequestSocket );
				}
				finally
				{
					try
					{
						ListenerWorkerFinished( InstanceId );
					}
					finally
					{
						LogManager.DebugAddFormat( "The action service listener worker for the action request locator \"{0}\""
							+ " has finished.", CurrentRequestLocator );
					}
				}
			}
		}

		/// <summary>
		/// Stops processing of the current action request as soon as possible.
		/// </summary>
		public void Stop()
		{
			StopProcessing = true;
		}

		/// <summary>
		/// Stops processing the current action request if it's found in the specified locators, as soon as possible.
		/// </summary>
		/// <param name="requestLocators">The instance locators in which to find the locator of the current action request
		/// whose processing to stop. If the locator of the current action request is not in the list, nothing happens. This may
		/// be null or empty. This must be thread safe.</param>
		public void StopRequests( Locators requestLocators )
		{
			if( ListUtils.NoContent( requestLocators ) || (CurrentRequestLocator == null) )
				return;

			if( requestLocators[ CurrentRequestLocator ] != null )
				StopProcessing = true;
		}

		#endregion Public methods
	}
}
