////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Threading;
using Vela.ActionProcessor;
using Vela.Cloning;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Networking;
using Vela.PeerDocuments;
using Vela.Pnrp;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.ActionService
{
	/// <summary>
	/// Action service manager.
	/// </summary>
	/// <remarks>
	/// <para>All the sheafs are locked with <see cref="SheafLocks.Shared"/>, unless they will be saved (in which case they are
	/// locked with <see cref="SheafLocks.Update"/>).</para>
	/// </remarks>
	public class ActionServiceManager : IActionServiceManager
	{
		#region Private constants

		/// <summary>
		/// The maximum number of times to attempt to forward the ports.
		/// </summary>
		private const int ForwardPortsAttemptCountMax = 10;

		/// <summary>
		/// The initial timespan (in seconds) between consecutive attempts to forward the ports.
		/// </summary>
		private const int ForwardPortsSleepInitialTimespan = 10;

		/// <summary>
		/// The time (in milliseconds) to sleep between consecutive checks to see if the processing thread must stop.
		/// </summary>
		private const int WaitStopRequestTimespan = 10;

		#endregion Private constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionServiceManager()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="actionServiceTypeNames">The type names of the action services which can process action requests, and
		/// which are available to all the vaults.</param>
		/// <param name="processorConfiguration">The action processor configuration.</param>
		/// <param name="actionServiceHost">Host used for updating the user interface of an action service application after an
		/// action request and response are processed. Its content must be thread safe.</param>
		/// <param name="actionClientManager">Action client manager which is notified by the action service about (local) action
		/// requests which were added to the (local) storehouse by the service, and which have to be processed after the current
		/// (remote) action request is processed by the service. Its content must be thread safe.</param>
		public ActionServiceManager( IStorehouseProvider storehouseProvider, Strings actionServiceTypeNames
			, ActionProcessorConfiguration processorConfiguration, IActionServiceHost actionServiceHost
			, IActionClientManager actionClientManager )
		{
			// Load the storehouse configuration.
			ServiceListenersStorehouseConfiguration configuration = ActionProcessorUtils.LoadStorehouseConfiguration(
				storehouseProvider );

			ActionServiceManagerContext serviceManagerContext = new ActionServiceManagerContext( storehouseProvider
				, processorConfiguration, configuration.ActionServiceListeners, actionServiceTypeNames, actionServiceHost
				, actionClientManager );

			ServiceManagerContext = Cloner.Deep( serviceManagerContext );

			Start();
		}

		#endregion Constructors

		#region Private static fields

		/// <summary>
		/// The synchronization object for the running state of the service manager.
		/// </summary>
		/// <remarks>
		/// <para>This is meant to ensure that no service manager instance may start while another instance is running. This way,
		/// the shared resources (like UPNs and forwarded ports) will not be mixedly created and cleaned up by the two
		/// instances.</para>
		/// </remarks>
		private static readonly object RunningLock = new object();

		#endregion Private static fields

		#region Private fields

		/// <summary>
		/// The main synchronization object.
		/// </summary>
		private readonly object MainLock = new object();

		/// <summary>
		/// Action service manager context.
		/// </summary>
		private ActionServiceManagerContext ServiceManagerContext;

		/// <summary>
		/// The vaults which can be the recipients of action requests. The backing field is initialized by default. This may be
		/// empty, but not null.
		/// </summary>
		private ActionServiceVaults ActionServiceVaults = new ActionServiceVaults();

		/// <summary>
		/// List of peer registrations. The backing field is initialized by default. This may be empty, but not null.
		/// </summary>
		private PeerRegistrations PeerRegistrations = new PeerRegistrations();

		/// <summary>
		/// The registration for the filler UPN classifiers.
		/// </summary>
		private PeerRegistration FillerUpnClassifiers;

		/// <summary>
		/// Port forwarder. The backing field is initialized by default.
		/// </summary>
		/// <remarks>
		/// <para>While working with this, it doesn't have to be locked with <see cref="MainLock"/> because it is never used
		/// from multiple threads relative to the action service listeners (which it uses). Due to the high timeouts involved,
		/// it's highly recommended to not lock this field.</para>
		/// </remarks>
		private PortForwarder PortForwarder = new PortForwarder();

		/// <summary>
		/// The action service listeners. The backing field is initialized by default. This may be empty, but not null.
		/// </summary>
		private ActionServiceListeners ServiceListeners = new ActionServiceListeners();

		/// <summary>
		/// The number of vaults which must have listeners.
		/// </summary>
		private int VaultCount;

		/// <summary>
		/// The thread where the processing happens.
		/// </summary>
		private Thread ProcessingThread;

		/// <summary>
		/// The synchronization object for the <see cref="StopProcessingThread"/> property.
		/// </summary>
		private readonly object StopProcessingThreadLock = new object();

		private bool _StopProcessingThread = false;
		/// <summary>
		/// If this is true then the processing thread must stop as soon as possible.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// </remarks>
		private bool StopProcessingThread
		{
			get
			{
				lock( StopProcessingThreadLock )
				{
					return _StopProcessingThread;
				}
			}
			set
			{
				lock( StopProcessingThreadLock )
				{
					_StopProcessingThread = value;
				}
			}
		}

		#endregion Private fields

		#region Public non-serializable properties

		/// <summary>
		/// If this is true then the current peer is normally online, else it's always offline and therefore no action client and
		/// service managers are created. READ-ONLY.
		/// </summary>
		public bool PeerIsOnline
		{
			get { return ServiceManagerContext.ProcessorConfiguration.PeerIsOnline; }
		}

		/// <summary>
		/// If this is true then the UPN resolvers (like the PNRP) will be used to register UPNs, and resolve UPNs to IP
		/// addresses. If this is false then the UPN resolvers will not be used, so the alternative peer resolution must be used
		/// instead. READ-ONLY.
		/// </summary>
		public bool UseUpnResolvers
		{
			get { return ServiceManagerContext.ProcessorConfiguration.UseUpnResolvers; }
		}

		/// <summary>
		/// If this is true then the action service ports will be forwarded; this is important for the peers whose computers are
		/// connected to routers, and who have no IPv6 network connectivity. READ-ONLY.
		/// </summary>
		public bool ForwardPorts
		{
			get { return ServiceManagerContext.ProcessorConfiguration.ForwardPorts; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Starts the filler UPN classifiers (= registers on the network the filler UPN classifiers).
		/// </summary>
		private void StartFillerUpnClassifierRegistrations()
		{
			if( FillerUpnClassifiers != null )
				FillerUpnClassifiers.Stop();

			if( !UseUpnResolvers || (PeerRegistrations.Count <= 0) )
				return;

			FillerUpnClassifiersConfiguration fillerUpnClassifiersConfiguration
				= ServiceManagerContext.ProcessorConfiguration.ActionService.FillerUpnClassifiers;

			// Ensure that all the necessary filler UPN classifiers are present.
			fillerUpnClassifiersConfiguration.FillClassifiersList();

			// Register filler classifiers up to (at least) the configured limit.
			int classifiersCount = fillerUpnClassifiersConfiguration.ClassifiersCount - PeerRegistrations.Count;

			FillerUpnClassifiers = new ClassifiersPeerRegistration( Locator.Dummy, true, classifiersCount
				, fillerUpnClassifiersConfiguration.Classifiers );

			FillerUpnClassifiers.Start();
		}

		/// <summary>
		/// Starts the peer registrations (= registers on the network all the vaults from the storehouse).
		/// </summary>
		/// <remarks>
		/// <para>The <see cref="ActionServiceVaults"/>, <see cref="PeerRegistrations"/> and <see cref="FillerUpnClassifiers"/>
		/// properties are thread safe within this method.</para>
		/// </remarks>
		private void StartPeerRegistrations()
		{
			lock( MainLock )
			{
				foreach( ActionServiceVault serviceVault in ActionServiceVaults )
				{
					PeerRegistration peerRegistration = PeerRegistrations[ serviceVault.Vault.InstanceLocator ];

					if( peerRegistration != null )
						peerRegistration.Stop();

					peerRegistration = new VaultPeerRegistration( serviceVault.Vault );
					peerRegistration.Start();

					PeerRegistrations.Add( peerRegistration );
				}

				StartFillerUpnClassifierRegistrations();
			}
		}

		/// <summary>
		/// Stops the peer registrations.
		/// </summary>
		/// <remarks>
		/// <para>The <see cref="PeerRegistrations"/> and <see cref="FillerUpnClassifiers"/> properties are thread safe within
		/// this method.</para>
		/// </remarks>
		private void StopPeerRegistrations()
		{
			Exception rethrowAbortEx = null;

			try
			{
				lock( MainLock )
				{
					if( PeerRegistrations != null )
					{
						foreach( PeerRegistration peerRegistration in PeerRegistrations )
						{
							try
							{
								peerRegistration.Stop();
							}
							catch( Exception ex )
							{
								// The loop must continue (if possible).

								if( ExceptionUtils.IsCritical( ex ) )
									throw ExceptionUtils.PreserveStackTrace( ex );

								if( ExceptionUtils.IsAbort( ex ) )
									rethrowAbortEx = ExceptionUtils.PreserveStackTrace( ex );

								LogManager.Add( ex );
							}
						}

						PeerRegistrations.Clear();
					}

					if( FillerUpnClassifiers != null )
						FillerUpnClassifiers.Stop();
				}
			}
			finally
			{
				if( rethrowAbortEx != null )
					throw ExceptionUtils.PreserveStackTrace( rethrowAbortEx );
			}
		}

		/// <summary>
		/// Starts the port forwardings.
		/// </summary>
		/// <remarks>
		/// <para>The <see cref="ServiceListeners"/> property is thread safe within this method.</para>
		/// </remarks>
		private void StartPortForwardings()
		{
			int remainingAttemptsCount = ForwardPortsAttemptCountMax;
			HashSet<int> registeredPorts = new HashSet<int>();

			bool hasNotDiscoveredService = true;
			bool hasUnregisteredPort = false;

			while( ForwardPorts && (remainingAttemptsCount > 0) )
			{
				try
				{
					if( !PortForwarder.HasDiscoveredService )
						PortForwarder.DiscoverService();

					hasNotDiscoveredService = false;

					foreach( ActionServiceListener serviceListener in ServiceListeners )
					{
						UInt16 port = serviceListener.ProcessingPort;

						// Some routers don't allow forwarding of certain ports.
						if( !PortForwarder.MayForward( port ) )
						{
							string message = string.Format( "A forwarding rule for the port {0} must be manually added in"
								+ " your router. Some ports may be critical to the underlying infrastructure, so check with"
								+ " someone who knows if this should be done.", port );
							LogManager.Add( message, null, LogRecordModes.Warning, false );

							continue;
						}

						if( registeredPorts.Contains( port ) )
							continue;

						try
						{
							// Don't use a description in order to be as stealthy as possible; the router may be compromised.
							PortForwarder.AddRule( null, port );

							registeredPorts.Add( port );

							string message = string.Format( "A forwarding rule for the port {0} was added.", port );
							LogManager.Add( message );
						}
						catch( Exception ex )
						{
							// The loop must continue (if possible).

							if( ExceptionUtils.IsAbort( ex ) )
								throw ExceptionUtils.PreserveStackTrace( ex );

							hasUnregisteredPort = true;

							string message = string.Format( "Failed to add a forwarding rule for the port {0}.", port );
							LogManager.Add( message, ex, LogRecordModes.Warning, false );
						}

						if( StopProcessingThread )
							throw new AbortLongOperationException();
					}

					if( !hasUnregisteredPort )
						remainingAttemptsCount = 0;
				}
				catch( Exception ex )
				{
					if( ExceptionUtils.IsAbort( ex ) )
						throw ExceptionUtils.PreserveStackTrace( ex );

					hasNotDiscoveredService = true;

					LogManager.Add( PortForwarder.NoServiceMessage, ex, LogRecordModes.Warning, false );
				}

				if( StopProcessingThread )
					throw new AbortLongOperationException();

				if( hasNotDiscoveredService || hasUnregisteredPort )
				{
					// Decrease the number of attempts to forward the ports.
					remainingAttemptsCount--;

					// Sleep between failed attempts.
					if( remainingAttemptsCount > 0 )
					{
						// Increase the sleep timespan with each attempt.
						int sleepTimespan = ForwardPortsSleepInitialTimespan
							+ 2 * (ForwardPortsAttemptCountMax - remainingAttemptsCount);

						TimeUtils.SleepSeconds( sleepTimespan, () => { return StopProcessingThread; } );

						if( StopProcessingThread )
							throw new AbortLongOperationException();
					}
					else
					{
						string message = "No more attempts will be made to add port forwarding rules. You have to either"
							+ " restart the application to automatically retry, or you have to manually add (in your router's"
							+ " configuration) forwarding rules for the processing ports of your vaults.";
						LogManager.Add( message, null, LogRecordModes.Error, false );
					}
				}
			}
		}

		/// <summary>
		/// Stops the port forwardings.
		/// </summary>
		/// <remarks>
		/// <para>As a security precaution, to ensure that no port forwarding remains orphan, the port forwardings are stopped
		/// even if <see cref="ActionProcessorConfiguration.ForwardPorts"/> is false.</para>
		/// 
		/// <para>The <see cref="ServiceListeners"/> property is thread safe within this method.</para>
		/// </remarks>
		private void StopPortForwardings()
		{
			Exception rethrowAbortEx = null;

			try
			{
				foreach( ActionServiceListener serviceListener in ServiceListeners )
				{
					UInt16 port = serviceListener.ProcessingPort;

					// Some routers don't allow forwarding of certain ports.
					if( !PortForwarder.MayForward( port ) )
						continue;

					try
					{
						PortForwarder.RemoveRule( port );

						string message = string.Format( "A forwarding rule for the port {0} was removed.", port );
						LogManager.Add( message );
					}
					catch( Exception ex )
					{
						// The loop must continue (if possible).

						if( ExceptionUtils.IsCritical( ex ) )
							throw ExceptionUtils.PreserveStackTrace( ex );

						if( ExceptionUtils.IsAbort( ex ) )
							rethrowAbortEx = ExceptionUtils.PreserveStackTrace( ex );

						string message = string.Format( "Failed to remove a forwarding rule for the port {0}.", port );
						LogManager.Add( message, ex, LogRecordModes.Error, false );
					}
				}
			}
			finally
			{
				if( rethrowAbortEx != null )
					throw ExceptionUtils.PreserveStackTrace( rethrowAbortEx );
			}
		}

		/// <summary>
		/// Adds a service listener for the specified port.
		/// </summary>
		/// <param name="actionServiceListenerConfigurations">The action service listener configurations.</param>
		/// <param name="processRequestMethodInfos">List of method informations which can process action requests. Do not modify
		/// the content of this object while the processor is running.</param>
		/// <param name="processingPort">Port for which to add a service listener.</param>
		/// <remarks>
		/// <para>If a service listener for the specified port already exists, nothing happens.</para>
		/// 
		/// <para>If an action service configuration doesn't exist for the specified port, one is created from the one which was
		/// created most recently. This ensures that there are service listeners even for the imported vaults (which may have
		/// processing ports which don't have associated service configurations in the storehouse).</para>
		/// 
		/// <para>The <see cref="ActionServiceVaults"/> and <see cref="ServiceListeners"/> properties are thread safe within this
		/// method.</para>
		/// </remarks>
		private void AddServiceListener( ActionServiceListenerConfigurations actionServiceListenerConfigurations
			, ProcessRequestMethodInfos processRequestMethodInfos, UInt16 processingPort )
		{
			ValidationUtils.ListHasContent( actionServiceListenerConfigurations, "actionServiceListenerConfigurations" );
			ValidationUtils.ListHasContent( processRequestMethodInfos, "processRequestMethodInfos" );
			ValidationUtils.Integer( processingPort, 1, UInt16.MaxValue, "ProcessingPort" );

			lock( MainLock )
			{
				// Get the existing (= previously created) service listener.
				ActionServiceListener serviceListener = ServiceListeners[ processingPort ];
				if( serviceListener != null )
					return; // The service listener was already created for this port.

				// Get the existing service configuration, or create a new one from the one which was created most recently.
				ActionServiceListenerConfiguration serviceConfiguration = actionServiceListenerConfigurations[ processingPort ];
				if( serviceConfiguration == null )
				{
					ActionServiceListenerConfiguration recentConfiguration = actionServiceListenerConfigurations.GetMostRecent();
					serviceConfiguration = ActionServiceListenerConfiguration.CreateWithPort( recentConfiguration
						, processingPort );
				}

				// Create the service listener.
				ActionServiceListenerContext listenerContext = new ActionServiceListenerContext(
					ServiceManagerContext.StorehouseProvider, ServiceManagerContext.ProcessorConfiguration
					, serviceConfiguration, ActionServiceVaults, processRequestMethodInfos
					, ServiceManagerContext.ActionServiceHost, ServiceManagerContext.ActionClientManager );
				serviceListener = new ActionServiceListener( listenerContext );

				ServiceListeners.Add( serviceListener );

				string message = string.Format( "An action service listener was added for the port {0}."
					, serviceListener.ProcessingPort );
				LogManager.Add( message );
			}
		}

		/// <summary>
		/// Starts the service listeners.
		/// </summary>
		/// <param name="vaultActionServiceTypeNames">The type names of the action services which can process action request
		/// whose recipient is the current vault; these are in addition to the default type names of the service application.
		/// This may be empty, but not null.</param>
		/// <param name="actionServiceListenerConfigurations">The action service listener configurations. This may be empty, but
		/// not null.</param>
		/// <param name="actionServiceVaults">The vaults which can be the recipients of action requests. This may be null or
		/// empty.</param>
		private void StartServiceListeners( Strings vaultActionServiceTypeNames
			, ActionServiceListenerConfigurations actionServiceListenerConfigurations, ActionServiceVaults actionServiceVaults )
		{
			ValidationUtils.ObjectHasContent( vaultActionServiceTypeNames, "vaultActionServiceTypeNames" );
			ValidationUtils.ObjectHasContent( actionServiceListenerConfigurations, "actionServiceListenerConfigurations" );

			// Caches the method information which can process action requests, for the action services.
			ProcessRequestMethodInfos processRequestMethodInfos = new ProcessRequestMethodInfos();
			processRequestMethodInfos.Add( ServiceManagerContext.ActionServiceTypeNames, false );
			if( vaultActionServiceTypeNames.Count > 0 )
				processRequestMethodInfos.Add( vaultActionServiceTypeNames, true );

			// Create the service listeners, one for each port of the service application. All vaults can process action requests
			// on all the ports.
			foreach( ActionServiceListenerConfiguration configuration in actionServiceListenerConfigurations )
			{
				AddServiceListener( actionServiceListenerConfigurations, processRequestMethodInfos
					, configuration.ProcessingPort );
			}

			// Create the service listeners, one for each vault. All vaults can process action requests on all the ports.
			foreach( ActionServiceVault actionServiceVault in actionServiceVaults )
			{
				AddServiceListener( actionServiceListenerConfigurations, processRequestMethodInfos
					, actionServiceVault.Vault.ProcessingIdentity.ProcessingPort );
			}
		}

		/// <summary>
		/// Stops the service listeners.
		/// </summary>
		/// <remarks>
		/// <para>The <see cref="ActionServiceVaults"/> and <see cref="ServiceListeners"/> properties are thread safe within this
		/// method.</para>
		/// </remarks>
		private void StopServiceListeners()
		{
			lock( MainLock )
			{
				foreach( ActionServiceListener serviceListener in ServiceListeners )
				{
					serviceListener.Stop();

					string message = string.Format( "The action service listener for the port {0} will stop."
						, serviceListener.ProcessingPort );
					LogManager.Add( message );
				}

				ServiceListeners.Clear();
				ActionServiceVaults.Clear();
			}
		}

		/// <summary>
		/// Starts the listeners and registers the UPNs.
		/// </summary>
		/// <remarks>
		/// <para>The UPNs of the items from <see cref="ActionServiceVaults"/> are computed in here in order to avoid having to
		/// later lock <see cref="ActionServiceVaults"/> or its items.</para>
		/// 
		/// <para>The <see cref="ActionServiceVaults"/>, <see cref="PeerRegistrations"/>, <see cref="FillerUpnClassifiers"/>
		/// and <see cref="ServiceListeners"/> properties are thread safe within this method.</para>
		/// </remarks>
		private void StartListenersAndUpns()
		{
			lock( MainLock )
			{
				// Collect the type names of the allowed action services.
				Strings vaultActionServiceTypeNames = new Strings();

				#region Initialize

				// Create a transaction.
				IStorehouseTransaction transaction = ServiceManagerContext.StorehouseProvider.CreateTransaction();

				try
				{
					// Load the vaults.
					IDeserializer deserializer;
					IKernelObservableList vaults = transaction.CollectSheafs( PortableTypes.FindAllDerived( typeof( Vault ) )
						, null, StorehouseRealms.ProcessingIdentityContainers, null, SheafLocks.Shared, null, out deserializer );

					VaultCount = vaults.Count;

					// Create the action service vaults.
					foreach( Vault vault in vaults )
					{
						try
						{
							// Collect the type names of the allowed action services.
							vaultActionServiceTypeNames.Add( vault.Appendix.ActionServiceTypeNames );

							// Create the service vault cache.
							ActionServiceVault actionServiceVault = new ActionServiceVault( vault );

							// Add the service vault.
							ActionServiceVaults.Add( actionServiceVault );
						}
						catch( Exception ex )
						{
							string message = string.Format( "The action service manager has failed to process the vault with the"
								+ " locator \"{0}\".", vault.InstanceLocator );
							throw new ValidationException( message, ExceptionUtils.PreserveStackTrace( ex ) );
						}
					}

					// Commit the transaction.
					transaction.Commit();
				}
				catch( Exception ex )
				{
					// Rolls back the transaction.
					transaction.Rollback();

					throw ExceptionUtils.PreserveStackTrace( ex );
				}

				#endregion Initialize

				ActionServiceUtils.VerifySameUpnAlgorithmsPerPort( ActionServiceVaults );

				// Compute all the UPNs (and their lazy loaded properties) in order to avoid having to later lock
				// "ActionServiceVaults" or its items.
				ActionServiceVaults.GenerateUpns();

				StartServiceListeners( vaultActionServiceTypeNames, ServiceManagerContext.ActionServices, ActionServiceVaults );

				if( UseUpnResolvers )
					StartPeerRegistrations();
			}
		}

		/// <summary>
		/// Starts the service manager.
		/// </summary>
		/// <remarks>
		/// <para>The <see cref="ActionServiceVaults"/>, <see cref="PeerRegistrations"/>, <see cref="FillerUpnClassifiers"/>
		/// and <see cref="ServiceListeners"/> properties are thread safe within this method.</para>
		/// </remarks>
		private void ProcessingThreadHandler()
		{
			try
			{
				lock( RunningLock )
				{
					try
					{
						StartListenersAndUpns();
						StartPortForwardings();

						while( true )
						{
							TimeUtils.SleepMilliseconds( WaitStopRequestTimespan );

							if( StopProcessingThread )
								throw new AbortLongOperationException();
						}
					}
					finally
					{
						StopPortForwardings();
						StopPeerRegistrations();
						StopServiceListeners();
					}
				}
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				if( ExceptionUtils.IsAbort( ex ) )
				{
					LogManager.Add( "The action service manager has been aborted." );
				}
				else
				{
					LogManager.Add( "The action service manager has failed processing.", ex, LogRecordModes.Error, true );
				}
			}
		}

		/// <summary>
		/// Starts the action service manager.
		/// </summary>
		private void Start()
		{
			ProcessingThread = new Thread( ProcessingThreadHandler );

			// The thread must terminate in order for the application to terminate.
			ProcessingThread.IsBackground = false;

			ProcessingThread.Start();
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Returns the state of the action service manager.
		/// </summary>
		/// <remarks>
		/// <para>The <see cref="ActionServiceVaults"/>, <see cref="PeerRegistrations"/> and <see cref="ServiceListeners"/>
		/// properties are thread safe within this method.</para>
		/// </remarks>
		public ActionServiceManagerStates AreAllRegisteredAndListening()
		{
			// If a lock was already acquired on "MainLock", the UI thread (which calls this method to show the service state)
			// would hang, so only TRY to acquire a lock.

			bool wasTaken = false;

			try
			{
				Monitor.TryEnter( MainLock, ref wasTaken );
				if( !wasTaken )
					return ActionServiceManagerStates.Unknown;

				if( UseUpnResolvers )
				{
					if( PeerRegistrations.Count < VaultCount )
						return ActionServiceManagerStates.OneNotRegisteredOrListening;

					foreach( PeerRegistration peerRegistration in PeerRegistrations )
					{
						if( !peerRegistration.AreAllRegistered() )
							return ActionServiceManagerStates.OneNotRegisteredOrListening;
					}
				}

				if( ServiceListeners.Count < ServiceManagerContext.ActionServices.Count )
					return ActionServiceManagerStates.OneNotRegisteredOrListening;

				return ServiceListeners.AreAllListening() ? ActionServiceManagerStates.AllRegisteredAndListening
					: ActionServiceManagerStates.OneNotRegisteredOrListening;
			}
			finally
			{
				if( wasTaken )
					Monitor.Exit( MainLock );
			}
		}

		/// <summary>
		/// Stops the action service manager as soon as possible.
		/// </summary>
		public void Stop()
		{
			StopProcessingThread = true;
		}

		/// <summary>
		/// Stops processing the action requests with the specified locators as soon as possible.
		/// </summary>
		/// <param name="requestLocators">The instance locators of the action requests whose processing to stop. If a locator
		/// refers to an action request which is not being processed, nothing happens (with that locator). This may be null or
		/// empty. This must be thread safe.</param>
		public void StopRequests( Locators requestLocators )
		{
			lock( MainLock )
			{
				foreach( ActionServiceListener serviceListener in ServiceListeners )
				{
					serviceListener.StopRequests( requestLocators );
				}
			}
		}

		#endregion Public methods
	}
}
