////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.ActionProcessor;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.ActionService
{
	/// <summary>
	/// Action service utilities.
	/// </summary>
	public static class ActionServiceUtils
	{
		#region Public static methods

		/// <summary>
		/// Verifies that the communication frame algorithms of all the vaults with each distinct processing port are the same.
		/// If there is a difference, an exception is thrown.
		/// </summary>
		/// <param name="serviceVaults">The vaults whose communication frame algorithms to check. This may be null or empty.
		/// </param>
		public static void VerifySameUpnAlgorithmsPerPort( ActionServiceVaults serviceVaults )
		{
			if( ListUtils.NoContent( serviceVaults ) )
				return;

			foreach( ActionServiceVault serviceVault1 in serviceVaults )
			{
				Upn upn1 = serviceVault1.Vault.ProcessingIdentity.Upn;
				ICommunicationFrameAlg communicationFrameAlg1 = upn1.UpnAlg.CommunicationFrameAlg;

				foreach( ActionServiceVault serviceVault2 in serviceVaults )
				{
					Upn upn2 = serviceVault2.Vault.ProcessingIdentity.Upn;
					ICommunicationFrameAlg communicationFrameAlg2 = upn2.UpnAlg.CommunicationFrameAlg;

					if( upn1.Port == upn2.Port )
					{
						string frameAlgName1 = RegistrableAlgorithmUtils.GetAlgorithmName( communicationFrameAlg1 );
						string frameAlgName2 = RegistrableAlgorithmUtils.GetAlgorithmName( communicationFrameAlg2 );

						// Immutable algorithms with identical names are identical.
						if( !TextUtils.Equals( frameAlgName1, frameAlgName2 ) )
						{
							throw new ValidationException( "The UPNs have the same processing port but have different"
									+ " communication frame algorithms.", null, null
								, new ExceptionDetail( "UPN 1", upn1.ToString() )
								, new ExceptionDetail( "UPN 2", upn2.ToString() )
								, new ExceptionDetail( "Processing port", upn1.Port.ToString() ) );
						}
					}
				}
			}
		}

		/// <summary>
		/// Gets the sheaf which proves that the local peer has knowledge of the initiator of the current action request.
		/// </summary>
		/// <param name="actionRequest">Anonymous action request.</param>
		/// <param name="transaction">The transaction with which to load the documents.</param>
		/// <returns>Returns the sheaf which proves that the local peer has knowledge of the initiator of the current action
		/// request. This may be null.</returns>
		public static Sheaf GetAnonymousActionRequestKnowledge( AnonymousActionRequestWithMutualKnowledge actionRequest
			, IStorehouseTransaction transaction )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( transaction, "transaction" );

			// The recipient may processes the action request only if it either has an account for the initiator (and the request
			// is signed with the identity from that account), or its recipient has a progress ("CreateAccountClientProgress")
			// for an action request queued to be sent to the initiator (and the action request is self-signed with the identity
			// from which the UPN from the progress was generated).

			IDeserializer deserializer;

			#region Check if there is an account for the initiator UPN fingerprint

			SpeckleFilter remoteUpnFilter = new SpeckleFilter( Tally.RemoteUpnFingerprintFilterIdentifier
				, actionRequest.InitiatorUpnFingerprint );

			Tally tally = transaction.CollectFirstSheaf( PortableTypes.FindAllDerived( typeof( Tally ) )
				, new SpeckleFilters( remoteUpnFilter ), StorehouseRealms.ProcessingIdentityContainers, SheafLocks.Shared
				, out deserializer ) as Tally;

			if( tally != null )
			{
				// Don't use expired documents.
				KernelUtils.HasExpired( tally );

				// Verify if the action request was signed by the anonymous peer.
				actionRequest.VerifySignature( tally.ProcessingIdentity );

				return tally;
			}

			#endregion Check if there is an account for the initiator UPN fingerprint

			#region Check if there is a "CreateAccountAnonymousActionRequest" action request for the initiator UPN fingerprint

			// Check if there is a progress ("CreateAccountClientProgress") for a "CreateAccountAnonymousActionRequest" action
			// request queued to be sent to the initiator (and the action request is self-signed with the identity from which the
			// UPN from the progress was generated).
			//
			// The progress is being checked instead of the "CreateAccountAnonymousActionRequest" action request because the
			// action request may still be in the storehouse (possibly due to a subtle bug) even though the account of the remote
			// peer was deleted; however, the progress is always current (since it's not permanent).

			SpeckleFilter recipientUpnFilter = new SpeckleFilter( ActionProgress.RecipientUpnFingerprintFilterIdentifier
				, actionRequest.InitiatorUpnFingerprint );

			CreateAccountClientProgress actionProgress = transaction.CollectFirstSheaf(
				PortableTypes.FindAllDerived( typeof( CreateAccountClientProgress ) ), new SpeckleFilters( recipientUpnFilter )
				, StorehouseRealms.ActionProgresses, SheafLocks.Shared, out deserializer ) as CreateAccountClientProgress;

			// Accept the existing action request even if it's expired (because it's already stored locally).
			if( actionProgress != null )
			{
				// Verify if the action request was signed by the anonymous peer.
				actionRequest.VerifySignature( actionRequest.InitiatorIdentity );

				if( actionRequest.InitiatorUpnFingerprint.Equals( actionProgress.RecipientUpnFingerprint ) )
					return actionProgress;
			}

			#endregion Check if there is a "CreateAccountAnonymousActionRequest" action request for the initiator UPN fingerprint

			return null;
		}

		/// <summary>
		/// Prepares the action response to send back to the initiator of the specified action request envelope.
		/// </summary>
		/// <param name="actionRequestEnvelope">Envelope of the action request for which the response is meant.</param>
		/// <param name="actionRequest">The action request which is being processed.</param>
		/// <param name="actionResponse">Action response to send.</param>
		/// <param name="recipientUpn">The UPN of the recipient of the action request.</param>
		/// <param name="serializer">Serializer used to serialize the action response.</param>
		/// <returns>Returns the obfuscated action response to send.</returns>
		public static byte[] PrepareActionResponse( EncryptedEnvelope actionRequestEnvelope, ActionRequest actionRequest
			, ActionResponse actionResponse, Upn recipientUpn, ISerializer serializer )
		{
			ValidationUtils.ObjectHasContent( actionRequestEnvelope, "actionRequestEnvelope" );
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( actionResponse, "actionResponse" );
			ValidationUtils.ObjectHasContent( recipientUpn, "recipientUpn" );
			ValidationUtils.ObjectHasContent( serializer, "serializer" );

			bool encryptResponsesInFrames = (actionRequest as IActionRequestWithMultiframeResponse != null)
				? ((IActionRequestWithMultiframeResponse)actionRequest).ServiceMustEncryptResponsesInFrames : true;

			if( (actionResponse as LastFrameActionResponse != null) || (actionResponse as ErrorActionResponse != null) )
				encryptResponsesInFrames = true;

			byte[] actionResponseBuffer;
			if( encryptResponsesInFrames )
			{
				// Compression can only be performed when using encrypted envelopes because they store the compression algorithm.
				CompressionAlg compressionAlg = actionRequest.ServiceShouldCompressResponse
					? actionRequestEnvelope.CompressionAlg : null;

				// Prepare to encrypt the action response.
				SymmEncryptedEnvelope responseEnvelope = new SymmEncryptedEnvelope( CryptoUtils.GetLocator()
					, actionRequestEnvelope.MacAlg, compressionAlg, actionRequestEnvelope.SymmAlg, null );

				responseEnvelope.EncryptableDocument = actionResponse;

				actionResponseBuffer = responseEnvelope.SaveToBuffer( null, actionRequest.ResponseEncryptionKey, serializer );
			}
			else
			{
				// Action responses may be sent unencrypted because the response frames are encrypted / obfuscated with the
				// response encryption key (see below the obfuscation step).

				if( recipientUpn.UpnAlg.SteganographicAlg as MacedSteganographicAlg == null )
				{
					throw new FatalCrossServiceException( "The recipient of the action request may send only encrypted action"
							+ " responses because its steganographic algorithm doesn't compute the MAC of communication frame"
							+ " bodies (which means that the integrity of the bodies is not guaranteed).", null
						, new ExceptionDetail( "Recipient UPN", actionRequest.RecipientUpn.ToString() )
						, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator ) );
				}

				actionResponseBuffer = serializer.SaveToBuffer( actionResponse, null );
			}

			// Obfuscate the action response.
			byte[] obfuscatedResponse = recipientUpn.UpnAlg.SteganographicAlg.Obfuscate( actionResponseBuffer
				, actionRequest.ResponseEncryptionKey, recipientUpn.Fingerprint.Value );

			return obfuscatedResponse;
		}

		/// <summary>
		/// Creates an action error response.
		/// </summary>
		/// <param name="actionRequest">The action request to process.</param>
		/// <param name="vault">The service vault.</param>
		/// <param name="exception">Exception for which to create the response.</param>
		/// <param name="isTimeoutException">If this is true then the exception represents a timeout. This is used only for
		/// exceptions which are not derived from <see cref="CrossServiceException"/>.</param>
		/// <returns>Returns the action response which is to be sent back to the initiator of the action request.</returns>
		public static ActionResponse CreateErrorResponse( ActionRequest actionRequest, Vault vault, Exception exception
			, bool isTimeoutException )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( vault, "vault" );
			ValidationUtils.ObjectHasContent( exception, "exception" );

			SolitaryIdentityPublisher identityPublisher = CoreDocumentsUtils.GetSolitaryIdentity( vault.ProcessingIdentity );

			// Create an error response.
			ErrorActionResponse errorResponse = new ErrorActionResponse( CryptoUtils.GetLocator(), actionRequest, null
				, identityPublisher, exception, isTimeoutException );

			// Sign the error response.
			errorResponse.Sign( vault.ProcessingIdentity, null );

			return errorResponse;
		}

		/// <summary>
		/// Updates the information about the last processing attempt of the action request which requests the transfer of the
		/// files attached to the letter with the specified locator.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="actionRequest">Action request which has requested the download of the files attached to the letter.
		/// </param>
		/// <param name="letterLocator">The locator of the letter to which the files to transfer are attached.</param>
		/// <param name="exception">The exception occurred while processing the current action response. If this is null, this
		/// method must be called only when the last frame of the current multiframe action response is being processed.</param>
		/// <param name="secondaryProcessingState">Secondary processing state. This may be null.</param>
		public static void UpdateLastProcessingAttemptOfAttachedFiles( IStorehouseProvider storehouseProvider
			, DownloadFilesActionRequest actionRequest, Locator letterLocator, Exception exception
			, SecondaryProcessingState secondaryProcessingState )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( letterLocator, "letterLocator" );

			#region Create the last processing attempt

			ActionLastProcessingAttempt lastProcessingAttempt = null;

			if( exception == null )
			{
				// Both the "Sent" and "Received" states could be used because the point of view can be the completion of either
				// the transfer or of the download action request.
				lastProcessingAttempt = new ActionLastProcessingAttempt( ActionProcessingStates.Sent, null
					, secondaryProcessingState );
			}
			else
			{
				Exception fatalCrossServiceException = ExceptionUtils.FindFatalCrossServiceException( exception );
				if( fatalCrossServiceException != null )
				{
					string message = ExceptionUtils.GetPresenterMessage( null, fatalCrossServiceException, true );
					lastProcessingAttempt = new ActionLastProcessingAttempt( ActionProcessingStates.Error, message
						, secondaryProcessingState );
				}
			}

			if( lastProcessingAttempt == null )
				return;

			#endregion Create the last processing attempt

			#region Create and save the progress

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// If the letter (= access granter) doesn't exist anymore, don't update the processing state because it's
				// displayed only together with the letter, and without the letter it would also remain orphan in the
				// storehouse.
				if( !transaction.SpeckleHasSheaf( letterLocator, StorehouseRealms.Posts ) )
				{
					// Commit the transaction.
					transaction.Commit();

					return;
				}

				// Load the existing progress (in order to overwrite it).
				IDeserializer deserializer;
				SpeckleFilter referredLetterSeekFilter = new SpeckleFilter( ActionProgress.ReferredSheafLocatorFilterIdentifier
					, letterLocator );
				DownloadFilesActionServiceProgress existingActionProgress = transaction.CollectFirstSheaf(
					new PortableTypes( typeof( DownloadFilesActionServiceProgress ) )
					, new SpeckleFilters( referredLetterSeekFilter ), StorehouseRealms.ActionProgresses, SheafLocks.Update
					, out deserializer ) as DownloadFilesActionServiceProgress;

				Locator actionProgressLocator = (existingActionProgress != null) ? existingActionProgress.InstanceLocator
					: CryptoUtils.GetLocator();

				DownloadFilesActionServiceProgress actionProgress = new DownloadFilesActionServiceProgress(
					actionProgressLocator, actionRequest.InitiatorUpnFingerprint, actionRequest.RecipientUpnFingerprint
					, letterLocator, actionRequest.GetType().FullName, secondaryProcessingState );

				actionProgress.UpdateLastProcessingAttempt( lastProcessingAttempt );

				// Prepare to overwrite the existing progress.
				if( existingActionProgress != null )
				{
					actionProgress.SetLockAndUpdateVersion( SheafLocks.Update, existingActionProgress.SheafUpdateVersion );

					transaction.UpdateSheaf( actionProgress, StorehouseRealms.ActionProgresses );
				}
				else
				{
					transaction.AddSheaf( actionProgress, StorehouseRealms.ActionProgresses );
				}

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			#endregion Create and save the progress
		}

		#endregion Public static methods
	}
}
