////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;

namespace Vela.ActionService
{
	/// <summary>
	/// List of vaults which can be the recipients of action requests
	/// (<see cref="Vela.ActionProcessor.ActionRequest"/>).
	/// </summary>
	/// <remarks>
	/// <para>This is used to cache all the service vaults (which can be the recipient of action requests) of a service
	/// application.</para>
	/// </remarks>
	public class ActionServiceVaults : KernelSortedList<Locator, ActionServiceVault>
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public ActionServiceVaults()
			: base( LocatorComparer.Instance )
		{
		}

		#endregion Constructors

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para>The UPN fingerprint of the processing identity of the service vault is used as key.</para>
		/// </remarks>
		public override Locator GetKey( ActionServiceVault item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.Vault.ProcessingIdentity.Upn.Fingerprint;
		}

		/// <summary>
		/// Generates the UPNs (and their lazy loaded properties) of all the vaults.
		/// </summary>
		public void GenerateUpns()
		{
			foreach( ActionServiceVault actionServiceVault in this )
				actionServiceVault.GenerateUpns();
		}

		/// <summary>
		/// Returns the item (from the list) which has the specified UPN fingerprint in any of its UPNs.
		/// </summary>
		/// <param name="upnFingerprint">UPN fingerprint to look for.</param>
		/// <returns>Returns the found item.</returns>
		public ActionServiceVault FindByAnyUpnFingerprint( Locator upnFingerprint )
		{
			ValidationUtils.ObjectHasContent( upnFingerprint, "upnFingerprint" );

			foreach( ActionServiceVault actionService in this )
			{
				if( actionService.Upns[ upnFingerprint ] != null )
					return actionService; // Item found.
			}

			return null; // Item not found.
		}

		#endregion Public methods
	}
}
