////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Reflection;
using Vela.ActionProcessor;
using Vela.Kernel;

namespace Vela.ActionService
{
	/// <summary>
	/// Method information which allows an action service to process an action request.
	/// </summary>
	public class ProcessRequestMethodInfo : HandleRequestMethodInfo
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ProcessRequestMethodInfo()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="requestTypeName">The type name of the action request to be handled by the "methodInfo" method.</param>
		/// <param name="methodInfo">The information of the method which must be executed when an action request with the type
		/// name "requestTypeName" has to be handled.</param>
		/// <param name="serviceTypeName">The type name of the action service to instantiate when an action request with the type
		/// name "requestTypeName" has to be processed by the service. The type must be globally registered.</param>
		/// <param name="isEmbeddedService">If this is true then "serviceTypeName" is the name of an action service which is
		/// embedded in a vault (in <see cref="Vela.PeerDocuments.Appendix.ActionServiceTypeNames"/>), and is not
		/// available to the vaults which don't have it embedded; if this is false then "serviceTypeName" is the name of an
		/// action service which is available to all the vaults.</param>
		public ProcessRequestMethodInfo( string requestTypeName, MethodInfo methodInfo, string serviceTypeName
				, bool isEmbeddedService )
			: base( requestTypeName, methodInfo )
		{
			ServiceTypeName = serviceTypeName;
			IsEmbeddedService = isEmbeddedService;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private string _ServiceTypeName;
		/// <summary>
		/// The type name of the action service to instantiate when an action request with the type name
		/// <see cref="HandleRequestMethodInfo.RequestTypeName"/> has to be processed by the service. The type must be globally
		/// registered. READ-ONLY.
		/// </summary>
		public string ServiceTypeName
		{
			get { return _ServiceTypeName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "ServiceTypeName" );

				_ServiceTypeName = value;
			}
		}

		private bool _IsEmbeddedService;
		/// <summary>
		/// If this is true then <see cref="ServiceTypeName"/> is the name of an action service which is embedded in a vault (in
		/// <see cref="Vela.PeerDocuments.Appendix.ActionServiceTypeNames"/>), and is not available to the vaults which
		/// don't have it embedded; if this is false then <see cref="ServiceTypeName"/> is the name of an action service which is
		/// available to all the vaults. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is used to know for which services a vault is allowed to process action requests.</para>
		/// </remarks>
		public bool IsEmbeddedService
		{
			get { return _IsEmbeddedService; }
			protected set { _IsEmbeddedService = value; }
		}

		#endregion Public non-serializable properties
	}
}
