////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Reflection;
using Vela.ActionProcessor;
using Vela.Kernel;
using Vela.Kernel.Cloning;

namespace Vela.ActionService
{
	/// <summary>
	/// List of method informations which allow an action service to process action requests.
	/// </summary>
	public class ProcessRequestMethodInfos : KernelSortedList<string, ProcessRequestMethodInfo>, IUnclonable
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public ProcessRequestMethodInfos()
			: base( StringInsensitiveComparer.Instance )
		{
		}

		#endregion Constructors

		#region Private methods

		/// <summary>
		/// Checks if the specified types are properly derived to form the prototype of a method which processes an action
		/// request (with the same type as "paramType1").
		/// </summary>
		/// <param name="returnType">The type of the returned value of the method which processes an action request.</param>
		/// <param name="paramType1">The type of the first parameter of the method which processes an action request; this is the
		/// type of the action request to process.</param>
		/// <param name="paramType2">The type of the second parameter of the method which processes an action request; this is
		/// the type of the request context.</param>
		/// <returns>Returns true if the specified types are properly derived.</returns>
		private bool CheckBaseTypeDerivation( Type returnType, Type paramType1, Type paramType2 )
		{
			if( returnType == null )
				return false;

			ValidationUtils.ObjectHasContent( returnType, "returnType" );
			ValidationUtils.ObjectHasContent( paramType1, "paramType1" );
			ValidationUtils.ObjectHasContent( paramType2, "paramType2" );

			int responseActionTypeMode, requestActionTypeMode, requestContextTypeMode, contextTypeMode;

			// The response.
			if( typeof( ActionResponseWithBeneficiary ).IsAssignableFrom( returnType ) )
				responseActionTypeMode = 3;
			else if( typeof( ActionResponseWithInitiator ).IsAssignableFrom( returnType ) )
				responseActionTypeMode = 2;
			else if( typeof( AnonymousActionResponse ).IsAssignableFrom( returnType ) )
				responseActionTypeMode = 1;
			else if( typeof( ActionResponse ).IsAssignableFrom( returnType ) )
				responseActionTypeMode = 0;
			else
				return false;

			// The request.
			if( typeof( ActionRequestWithBeneficiary ).IsAssignableFrom( paramType1 ) )
				requestActionTypeMode = 3;
			else if( typeof( ActionRequestWithInitiator ).IsAssignableFrom( paramType1 ) )
				requestActionTypeMode = 2;
			else if( typeof( AnonymousActionRequest ).IsAssignableFrom( paramType1 ) )
				requestActionTypeMode = 1;
			else if( typeof( ActionRequest ).IsAssignableFrom( paramType1 ) )
				requestActionTypeMode = 0;
			else
				return false;

			if( responseActionTypeMode != requestActionTypeMode )
				return false;

			// The request context.
			if( typeof( ActionRequestWithBeneficiary ).IsAssignableFrom( paramType1 ) )
				requestContextTypeMode = 2;
			else if( typeof( ActionRequestWithInitiator ).IsAssignableFrom( paramType1 ) )
				requestContextTypeMode = 1;
			else if( typeof( AnonymousActionRequest ).IsAssignableFrom( paramType1 ) )
				requestContextTypeMode = 0;
			else if( typeof( ActionRequest ).IsAssignableFrom( paramType1 ) )
				requestContextTypeMode = 0;
			else
				return false;

			// The context.
			if( typeof( ActionRequestWithBeneficiaryContext ).IsAssignableFrom( paramType2 ) )
				contextTypeMode = 2;
			else if( typeof( ActionRequestWithInitiatorContext ).IsAssignableFrom( paramType2 ) )
				contextTypeMode = 1;
			else if( typeof( ActionRequestContext ).IsAssignableFrom( paramType2 ) )
				contextTypeMode = 0;
			else
				return false;

			return requestContextTypeMode == contextTypeMode;
		}

		/// <summary>
		/// Adds to the list the method informations which can process action requests, from the action service with the
		/// specified type name.
		/// </summary>
		/// <param name="actionServiceTypeName">The type name of the action service whose method informations to add to the
		/// list. The type must be globally registered.</param>
		/// <param name="isEmbeddedService">If this is true then "actionServiceTypeName" is the name of an action service which
		/// is embedded in a vault (in <see cref="Vela.PeerDocuments.Appendix.ActionServiceTypeNames"/>), and is not
		/// available to the vaults which don't have it embedded; if this is false then "actionServiceTypeName" is the name of an
		/// action service which is available to all the vaults.</param>
		private void AddMethodInfos( string actionServiceTypeName, bool isEmbeddedService )
		{
			ValidationUtils.TextHasContent( actionServiceTypeName, "actionServiceTypeName" );

			Type serviceType = PortableTypes.FindGlobal( actionServiceTypeName );

			if( (serviceType == null) || !typeof( IActionService ).IsAssignableFrom( serviceType ) )
			{
				throw new ValidationException( string.Format( "The action service type must derive from \"{0}\"."
					, typeof( IActionService ).FullName ) );
			}

			// Go through all the public methods.
			MethodInfo[] methodInfos = serviceType.GetMethods( BindingFlags.Public | BindingFlags.Instance );
			if( methodInfos != null )
			{
				foreach( MethodInfo mi in methodInfos )
				{
					ParameterInfo[] parameterInfos = mi.GetParameters();
					if( (parameterInfos == null) || (parameterInfos.Length != 2) )
						continue;

					Type returnType = mi.ReturnType;
					Type paramType1 = parameterInfos[ 0 ].ParameterType;
					Type paramType2 = parameterInfos[ 1 ].ParameterType;

					if( CheckBaseTypeDerivation( returnType, paramType1, paramType2 ) )
					{
						// Found an accepted type for an action request.
						ProcessRequestMethodInfo processRequestMethodInfo = new ProcessRequestMethodInfo( paramType1.FullName
							, mi, actionServiceTypeName, isEmbeddedService );
						Add( processRequestMethodInfo );
					}
				}
			}
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para><see cref="HandleRequestMethodInfo.RequestTypeName"/> is used as key.</para>
		/// </remarks>
		public override string GetKey( ProcessRequestMethodInfo item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.RequestTypeName;
		}

		/// <summary>
		/// Adds to the list the method informations which can process action requests, from the action service with the
		/// specified type names.
		/// </summary>
		/// <param name="actionServiceTypeNames">The type names of the action services whose method informations to add to the
		/// list. The types must be globally registered.</param>
		/// <param name="areEmbeddedServices">If this is true then "actionServiceTypeNames" are the names of action services
		/// which are embedded in vaults (in <see cref="Vela.PeerDocuments.Appendix.ActionServiceTypeNames"/>), and are
		/// not available to the vaults which don't have it embedded; if this is false then "actionServiceTypeNames" are the
		/// names of action services which are available to all the vaults.</param>
		public void Add( Strings actionServiceTypeNames, bool areEmbeddedServices )
		{
			ValidationUtils.ListHasContent( actionServiceTypeNames, "actionServiceTypeNames" );

			foreach( string actionServiceTypeName in actionServiceTypeNames )
				AddMethodInfos( actionServiceTypeName, areEmbeddedServices );
		}

		/// <summary>
		/// Adds to the list the method informations which can process action requests, from the action service with the
		/// specified type name.
		/// </summary>
		/// <param name="actionServiceTypeName">The type name of the action service whose method informations to add to the
		/// list. The type must be globally registered.</param>
		/// <param name="isEmbeddedService">If this is true then "actionServiceTypeName" is the name of an action service which
		/// is embedded in a vault (in <see cref="Vela.PeerDocuments.Appendix.ActionServiceTypeNames"/>), and is not
		/// available to the vaults which don't have it embedded; if this is false then "actionServiceTypeName" is the name of an
		/// action service which is available to all the vaults.</param>
		public void Add( string actionServiceTypeName, bool isEmbeddedService )
		{
			AddMethodInfos( actionServiceTypeName, isEmbeddedService );
		}

		#endregion Public methods
	}
}
