﻿////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.ActionService
{
	/// <summary>
	/// Command line to run with <see cref="CommandLineProcessor"/>.
	/// </summary>
	public class CommandLine : IDocument
	{
		#region Public constants

		/// <summary>
		/// The default value for <see cref="ApplicationFileNameToRun"/>; this is the interpreter for the command lines.
		/// </summary>
		public const string ApplicationFileNameToRunDef = "cmd.exe";

		/// <summary>
		/// The default value for <see cref="WorkingFolder"/>.
		/// </summary>
		public const string WorkingFolderDef = "Cmd";

		/// <summary>
		/// The default value for <see cref="MayCreateWindow"/>.
		/// </summary>
		public const bool MayCreateWindowDef = false;

		/// <summary>
		/// The maximum value for <see cref="WaitFinishTimespan"/>.
		/// </summary>
		public const int WaitFinishSecondsMax = 20;

		/// <summary>
		/// The default value for <see cref="WaitFinishTimespan"/>.
		/// </summary>
		public const int WaitFinishTimespanDef = 10;

		/// <summary>
		/// The default value for <see cref="KillAfterWait"/>.
		/// </summary>
		public const bool KillAfterWaitDef = false;

		/// <summary>
		/// The default value for <see cref="RetrieveOutput"/>.
		/// </summary>
		public const bool RetrieveOutputDef = false;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public CommandLine()
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// Returns true if the current command line can be run the application <see cref="ApplicationFileNameToRun"/>, or open
		/// the file <see cref="FileNameToRun"/>.
		/// </summary>
		/// <returns>Returns true if the current command line can be run.</returns>
		public bool CanRunAppOrOpenFile
		{
			get
			{
				if( !TextUtils.Equals( ApplicationFileNameToRun, ApplicationFileNameToRunDef ) )
					return true;

				if( TextUtils.Equals( ApplicationFileNameToRun, ApplicationFileNameToRunDef )
					&& TextUtils.HasContent( ApplicationParameters ) )
				{
					return true;
				}

				if( TextUtils.HasContent( FileNameToRun ) )
					return true;

				return false;
			}
		}

		/// <summary>
		/// Returns true if the current command line has an application to run, and false if it has a file to open.
		/// </summary>
		/// <returns>Returns true if the current command line has an application to run, and false if it has a file to open.
		/// </returns>
		public bool HasApplicationToRun
		{
			get
			{
				if( !TextUtils.Equals( ApplicationFileNameToRun, ApplicationFileNameToRunDef ) )
					return true;

				if( TextUtils.Equals( ApplicationFileNameToRun, ApplicationFileNameToRunDef )
					&& TextUtils.HasContent( ApplicationParameters ) )
				{
					return true;
				}

				if( TextUtils.HasContent( FileNameToRun ) )
					return false;

				return false;
			}
		}

		/// <summary>
		/// Returns true if the current command line has a file to open, and false if it has an application to run.
		/// </summary>
		/// <returns>Returns true if the current command line has a file to open, and false if it has an application to run.
		/// </returns>
		public bool HasFileToOpen
		{
			get
			{
				return !HasApplicationToRun && TextUtils.HasContent( FileNameToRun );
			}
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private string _ApplicationFileNameToRun = ApplicationFileNameToRunDef;
		/// <summary>
		/// The file name of the application to run. If this and <see cref="FileNameToRun"/> are null or empty, the command line
		/// interpreter is used instead (for this property). This may be null or empty.
		/// </summary>
		/// <remarks>
		/// <para>If this has content which is not <see cref="ApplicationFileNameToRunDef"/>, the specified application is run.
		/// </para>
		/// 
		/// <para>If this and <see cref="FileNameToRun"/> are null or empty, <see cref="ApplicationFileNameToRunDef"/> is used
		/// instead.</para>
		/// </remarks>
		public string ApplicationFileNameToRun
		{
			get
			{
				if( TextUtils.NoContent( _ApplicationFileNameToRun ) )
					_ApplicationFileNameToRun = ApplicationFileNameToRunDef;

				return _ApplicationFileNameToRun;
			}
			set { _ApplicationFileNameToRun = value; }
		}

		private string _ApplicationParameters;
		/// <summary>
		/// Command line parameters for <see cref="ApplicationFileNameToRun"/>. This may be null or empty.
		/// </summary>
		/// <remarks>
		/// <para>If this has content, the application specified in <see cref="ApplicationFileNameToRun"/> (or, if it's null or
		/// empty, the default <see cref="ApplicationFileNameToRunDef"/>) is run.</para>
		/// </remarks>
		public string ApplicationParameters
		{
			get { return _ApplicationParameters; }
			set { _ApplicationParameters = value; }
		}

		private string _FileNameToRun;
		/// <summary>
		/// The name of the file to open (using the operating system's file associations). This may be null or empty.
		/// </summary>
		/// <remarks>
		/// <para> If this has a value and both <see cref="ApplicationFileNameToRun"/> and <see cref="ApplicationParameters"/>
		/// are null or empty, the specified file is opened, else the application specified in
		/// <see cref="ApplicationFileNameToRun"/> (or, if it's null or empty, the default
		/// <see cref="ApplicationFileNameToRunDef"/>) is run.</para>
		/// </remarks>
		public string FileNameToRun
		{
			get { return _FileNameToRun; }
			set { _FileNameToRun = value; }
		}

		private string _WorkingFolder;
		/// <summary>
		/// Working folder for the command line. This may be null or empty.
		/// </summary>
		/// <remarks>
		/// <para>If this is null or empty, the temporary folder of the application, subfolder <see cref="WorkingFolderDef"/>, is
		/// (created and) used instead.</para>
		/// </remarks>
		public string WorkingFolder
		{
			get
			{
				if( TextUtils.NoContent( _WorkingFolder ) )
					_WorkingFolder = AssemblyUtils.GetApplicationTempFolderName( WorkingFolderDef );

				return _WorkingFolder;
			}
			set
			{
				_WorkingFolder = value;
			}
		}

		private bool _MayCreateWindow = MayCreateWindowDef;
		/// <summary>
		/// If this is true then a window may be created when running the command line.
		/// </summary>
		/// <remarks>
		/// <para>This is always true if <see cref="HasFileToOpen"/> is true.</para>
		/// </remarks>
		public bool MayCreateWindow
		{
			get { return _MayCreateWindow; }
			set { _MayCreateWindow = value; }
		}

		private int _WaitFinishTimespan = WaitFinishTimespanDef;
		/// <summary>
		/// Time (in seconds) to wait for the started process to finish.
		/// </summary>
		/// <remarks>
		/// <para>If the command line doesn't finish within this time, it's forcefully terminated only if
		/// <see cref="KillAfterWait"/> is true. This may be between 0 and <see cref="WaitFinishSecondsMax"/>.</para>
		/// </remarks>
		public int WaitFinishTimespan
		{
			get { return _WaitFinishTimespan; }
			set
			{
				ValidationUtils.Integer( value, 0, WaitFinishSecondsMax, "WaitFinishTimespan" );

				_WaitFinishTimespan = value;
			}
		}

		private bool _KillAfterWait = KillAfterWaitDef;
		/// <summary>
		/// If this is true then the process started by the command line is forcefully terminated after waiting
		/// <see cref="WaitFinishTimespan"/> seconds for it to finish. This is ignored if <see cref="WaitFinishTimespan"/> is 0 or
		/// a negative number.
		/// </summary>
		public bool KillAfterWait
		{
			get { return _KillAfterWait; }
			set { _KillAfterWait = value; }
		}

		private bool _RetrieveOutput = RetrieveOutputDef;
		/// <summary>
		/// If this is true then the output of the command line is retrieved and displayed in the "Command lines" window.
		/// </summary>
		/// <remarks>
		/// <para>Even if you get a timeout, or an exit code which indicates an error, the output may still contain relevant
		/// information.</para>
		/// </remarks>
		public bool RetrieveOutput
		{
			get { return _RetrieveOutput; }
			set { _RetrieveOutput = value; }
		}

		#endregion Public methods
	}
}
