﻿////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;

namespace Vela.ActionService
{
	/// <summary>
	/// A processor which runs command lines.
	/// </summary>
	public class CommandLineParser
	{
		#region Public constants

		/// <summary>
		/// The name of the field used for the <see cref="CommandLine.ApplicationFileNameToRun"/> property.
		/// </summary>
		public const string ApplicationFileNameToRunFieldName = "app";

		/// <summary>
		/// The name of the field used for the <see cref="CommandLine.ApplicationParameters"/> property.
		/// </summary>
		public const string ApplicationParametersFieldName = "clp";

		/// <summary>
		/// The name of the field used for the <see cref="CommandLine.FileNameToRun"/> property.
		/// </summary>
		public const string FileNameToRunFieldName = "file";

		/// <summary>
		/// The name of the field used for the <see cref="CommandLine.WorkingFolder"/> property.
		/// </summary>
		public const string WorkingFolderFieldName = "folder";

		/// <summary>
		/// The name of the field used for the <see cref="CommandLine.MayCreateWindow"/> property.
		/// </summary>
		public const string MayCreateWindowFieldName = "window";

		/// <summary>
		/// The name of the field used for the <see cref="CommandLine.WaitFinishTimespan"/> property.
		/// </summary>
		public const string WaitFinishSecondsFieldName = "wait";

		/// <summary>
		/// The name of the field used for the <see cref="CommandLine.KillAfterWait"/> property.
		/// </summary>
		public const string KillAfterWaitFieldName = "kill";

		/// <summary>
		/// The name of the field used for the timeout. The timeout is a shortcut for setting
		/// <see cref="CommandLine.WaitFinishTimespan"/> with the timeout value and <see cref="CommandLine.KillAfterWait"/>
		/// with true. This has priority over <see cref="CommandLine.WaitFinishTimespan"/> and
		/// <see cref="CommandLine.KillAfterWait"/>.
		/// </summary>
		public const string TimeoutSecondsFieldName = "timeout";

		/// <summary>
		/// The name of the field used for the <see cref="CommandLine.RetrieveOutput"/> property.
		/// </summary>
		public const string RetrieveOutputFieldName = "output";

		/// <summary>
		/// The field name which ends a command line.
		/// </summary>
		/// <remarks>
		/// <para>When the <see cref="CommandLineParser"/> parser encounters a field with this name, it must stop parsing
		/// name-value fields.</para>
		/// 
		/// <para>This field name ends the information related to the command line and starts the normal message of the letter
		/// (if there is any). It's recommended to use this, for performance reasons, but it's not required.</para>
		/// </remarks>
		public const string CommandLineEndFieldName = "msg";

		#endregion Public constants

		#region Public static methods

		/// <summary>
		/// Parses the specified textual command line to get the command line.
		/// </summary>
		/// <param name="textualCommandLine">Textual command line to parse. This may be null or empty.</param>
		/// <returns>Returns the extracted command line.</returns>
		public static CommandLine Parse( string textualCommandLine )
		{
			CommandLine commandLine = new CommandLine();
			if( TextUtils.NoContent( textualCommandLine ) )
				return commandLine;

			NameValueParser parser = new NameValueParser( textualCommandLine, CommandLineEndFieldName );

			commandLine.ApplicationFileNameToRun = parser.GetTextValue( ApplicationFileNameToRunFieldName
				, CommandLine.ApplicationFileNameToRunDef );
			commandLine.ApplicationParameters = parser.GetTextValue( ApplicationParametersFieldName, null );
			commandLine.FileNameToRun = parser.GetTextValue( FileNameToRunFieldName, null );
			commandLine.WorkingFolder = parser.GetTextValue( WorkingFolderFieldName, null );
			commandLine.MayCreateWindow = parser.GetBooleanValue( MayCreateWindowFieldName, CommandLine.MayCreateWindowDef );
			commandLine.WaitFinishTimespan = parser.GetIntegerValue( WaitFinishSecondsFieldName, 0
				, CommandLine.WaitFinishSecondsMax, CommandLine.WaitFinishTimespanDef );
			commandLine.KillAfterWait = parser.GetBooleanValue( KillAfterWaitFieldName, CommandLine.KillAfterWaitDef );
			commandLine.RetrieveOutput = parser.GetBooleanValue( RetrieveOutputFieldName, CommandLine.RetrieveOutputDef );

			// The timeout has priority over "commandLine.WaitFinishTimespan" and "commandLine.KillAfterWait".
			if( parser.HasField( TimeoutSecondsFieldName ) )
			{
				commandLine.WaitFinishTimespan = parser.GetIntegerValue( TimeoutSecondsFieldName, 0
					, CommandLine.WaitFinishSecondsMax, CommandLine.WaitFinishTimespanDef );
				commandLine.KillAfterWait = true;
			}

			if( commandLine.RetrieveOutput && (commandLine.WaitFinishTimespan <= 0) )
				throw new ValidationException( "The command line output can be retrieved only if the waiting time is above 0." );

			// If the user appears to have wanted to specify a command line, but it contains nothing to run or open, generate an
			// error.
			if( !commandLine.HasApplicationToRun && !commandLine.HasFileToOpen
				&& (
					parser.HasField( ApplicationFileNameToRunFieldName )
					|| parser.HasField( ApplicationParametersFieldName )
					|| parser.HasField( FileNameToRunFieldName )
				) )
			{
				throw new ValidationException( "The command line has neither an application to run, nor a file to open." );
			}

			return commandLine;
		}

		#endregion Public methods
	}
}
