﻿////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Diagnostics;
using System.Text;
using System.Threading;
using Vela.ActionProcessor;
using Vela.Kernel;
using Vela.PeerDocuments;

namespace Vela.ActionService
{
	/// <summary>
	/// A processor which runs command lines.
	/// </summary>
	public class CommandLineProcessor
	{
		#region Public constants

		/// <summary>
		/// The subject of the secondary processing state.
		/// </summary>
		public const string SecondaryProcessingStateSubject = "Command";

		/// <summary>
		/// The maximum (read and returned) size for the output of the command line.
		/// </summary>
		public const int OutputSizeMax = 100000;

		/// <summary>
		/// Timeout (in milliseconds) to wait for the complete output data.
		/// </summary>
		public const int WaitCommandLineOutputTimeout = 3000;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="commandLine">Command line to run.</param>
		public CommandLineProcessor( CommandLine commandLine )
		{
			ValidationUtils.ObjectHasContent( commandLine, "commandLine" );

			CommandLine = commandLine;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// Command line to run.
		/// </summary>
		private CommandLine CommandLine;

		/// <summary>
		/// The synchronization object for the <see cref="CommandLineOutput"/> field. READ-ONLY.
		/// </summary>
		private static readonly object CommandLineOutputLock = new object();

		/// <summary>
		/// The output accumulator for the command line.
		/// </summary>
		private StringBuilder CommandLineOutput = new StringBuilder();

		/// <summary>
		/// Event which, when set, indicates that the error data of the run process has been fully received.
		/// </summary>
		private AutoResetEvent ErrorDataEndedAutoResetEvent;

		/// <summary>
		/// Event which, when set, indicates that the output data of the run process has been fully received.
		/// </summary>
		private AutoResetEvent OutputDataEndedAutoResetEvent;

		#endregion Private fields

		#region Private static methods

		/// <summary>
		/// Checks if the specified process waits for user input.
		/// </summary>
		/// <param name="process">Process to check if it waits for user input.</param>
		/// <returns>Returns true if the specified process waits for user input.</returns>
		private static bool IsWaitingForUserInput( Process process )
		{
			ValidationUtils.ObjectHasContent( process, "process" );

			foreach( ProcessThread thread in process.Threads )
			{
				if( (thread.ThreadState == System.Diagnostics.ThreadState.Wait)
					&& (thread.WaitReason == ThreadWaitReason.UserRequest) )
				{
					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Waits for the specified process to finish execution.
		/// </summary>
		/// <param name="process">Process for which to wait.</param>
		/// <param name="waitFinishTimespan">Time (in seconds) to wait for the started process to finish.</param>
		/// <param name="killAfterWait">If this is true then the specified process is forcefully terminated after waiting
		/// "waitFinishTimespan" seconds for it to finish.</param>
		/// <param name="finishedInTime">Is set with true if the process has finished in time (= after maximum
		/// "waitFinishTimespan" seconds from its start), or if "waitFinishTimespan" is 0 or a negative number.</param>
		private static void WaitProcessFinish( Process process, int waitFinishTimespan, bool killAfterWait
			, out bool finishedInTime )
		{
			ValidationUtils.ObjectHasContent( process, "process" );

			waitFinishTimespan = MathUtils.Limit( waitFinishTimespan, 0, CommandLine.WaitFinishSecondsMax );
			if( waitFinishTimespan <= 0 )
			{
				finishedInTime = true;

				return;
			}

			finishedInTime = process.WaitForExit( TimeUtils.SecondsToMilli( waitFinishTimespan ) );

			if( !finishedInTime && killAfterWait )
				process.Kill();
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Runs the specified textual command line.
		/// </summary>
		/// <param name="textualCommandLine">Textual command line to run. This may be null or empty.</param>
		/// <param name="commandLineOutput">Is set with the output of the command line. This may be null or empty.</param>
		/// <returns>Returns a secondary processing state which has its <see cref="SecondaryProcessingState.AnErrorHasOccurred"/>
		/// set with true if the exit code of the command line is not 0. This may be null.</returns>
		/// <remarks>
		/// <para>All non-critical exceptions are captured inside.</para>
		/// </remarks>
		public static SecondaryProcessingState RunForLetter( string textualCommandLine, out string commandLineOutput )
		{
			commandLineOutput = null;

			try
			{
				if( TextUtils.NoContent( textualCommandLine ) )
					return null;

				CommandLine commandLine = CommandLineParser.Parse( textualCommandLine );
				if( !commandLine.CanRunAppOrOpenFile )
					return null;

				CommandLineProcessor commandLineProcessor = new CommandLineProcessor( commandLine );

				bool finishedInTime;
				int exitCode;
				commandLineProcessor.Run( out finishedInTime, out exitCode, out commandLineOutput );

				string state = null;
				string message = null;

				if( !finishedInTime )
				{
					state = "Timeout";
				}
				else if( exitCode != 0 )
				{
					state = string.Format( "Exit code {0}", exitCode );
					message = SystemUtils.GetUserFriendlyApplicationExitCode( exitCode );
				}
				else
				{
					state = "OK";
				}

				return new SecondaryProcessingState( exitCode != 0, SecondaryProcessingStateSubject, state, message, true );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				// Since all exceptions are caught here, it's important to log them.
				LogManager.Add( ex );

				string message = ExceptionUtils.GetPresenterMessage( null, ex, true );
				if( TextUtils.NoContent( message ) )
				{
					// Use the actual message (rather than a generic one) of the exception because the remote peer can do nearly
					// everything on the local peer, so there is no point in protecting potentially sensitive information from
					// the actual message.
					message = ex.Message;
				}

				return new SecondaryProcessingState( true, SecondaryProcessingStateSubject, "Exception", message, true );
			}
		}

		#endregion Public static methods

		#region Event handlers

		/// <summary>
		/// Processes the error data from the running command line.
		/// </summary>
		/// <param name="sender">Sender.</param>
		/// <param name="e">Event data.</param>
		private void Process_ErrorDataReceived( object sender, DataReceivedEventArgs e )
		{
			AddCommandLineOutput( e.Data, ErrorDataEndedAutoResetEvent );
		}

		/// <summary>
		/// Processes the output data from the running command line.
		/// </summary>
		/// <param name="sender">Sender.</param>
		/// <param name="e">Event data.</param>
		private void Process_OutputDataReceived( object sender, DataReceivedEventArgs e )
		{
			AddCommandLineOutput( e.Data, OutputDataEndedAutoResetEvent );
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Processes the specified output data of a command line.
		/// </summary>
		/// <param name="output">Output data to process. This may be null or empty.</param>
		/// <param name="outputDataEndedAutoResetEvent">Auto reset event to set when there is no more output data to process.
		/// </param>
		private void AddCommandLineOutput( string output, AutoResetEvent outputDataEndedAutoResetEvent )
		{
			try
			{
				ValidationUtils.ObjectHasContent( outputDataEndedAutoResetEvent, "outputDataEndedAutoResetEvent" );

				// Allow empty lines.
				if( output == null )
				{
					outputDataEndedAutoResetEvent.Set();

					return;
				}

				lock( CommandLineOutputLock )
				{
					int currentSize = CommandLineOutput.Length;
					bool dataFitsInSizeMax = currentSize + output.Length <= OutputSizeMax;
					int sizeToRead = dataFitsInSizeMax ? output.Length : (OutputSizeMax - currentSize - 1);

					if( sizeToRead > 0 )
					{
						CommandLineOutput.Append( output, 0, sizeToRead );

						if( dataFitsInSizeMax )
							CommandLineOutput.AppendLine();
						else
							CommandLineOutput.Append( PeerDocumentsConsts.TextWasTruncatedSuffix );
					}
				}
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				// There is nowhere for this exception to go, so log it here.
				LogManager.Add( ex );
			}
		}

		/// <summary>
		/// Gets the command line output as a text. The text is suffixed with
		/// <see cref="PeerDocumentsConsts.TextWasTruncatedSuffix"/> if "outputIsComplete" is false.
		/// </summary>
		/// <param name="outputIsComplete">If this is true then the command line output is complete.</param>
		/// <returns>Returns the command line output.</returns>
		private string GetCommandLineOutput( bool outputIsComplete )
		{
			if( outputIsComplete )
				return CommandLineOutput.ToString();

			if( CommandLineOutput.Length + PeerDocumentsConsts.TextWasTruncatedSuffix.Length <= OutputSizeMax )
			{
				CommandLineOutput.Append( PeerDocumentsConsts.TextWasTruncatedSuffix );

				return CommandLineOutput.ToString();
			}

			CommandLineOutput.Length = CommandLineOutput.Length - PeerDocumentsConsts.TextWasTruncatedSuffix.Length;
			CommandLineOutput.Append( PeerDocumentsConsts.TextWasTruncatedSuffix );

			return CommandLineOutput.ToString();
		}

		/// <summary>
		/// Runs the command line.
		/// </summary>
		/// <param name="finishedInTime">Is set with true if the process has finished in time (= after maximum
		/// "waitFinishTimespan" seconds from its start), or if <see cref="Vela.ActionService.CommandLine.WaitFinishTimespan"/>
		/// is 0 or a negative number.</param>
		/// <param name="exitCode">Is set with the exit code of the command line. This is set with <see cref="int.MinValue"/> if
		/// the process started by the command line hasn't finished in time
		/// (<see cref="Vela.ActionService.CommandLine.WaitFinishTimespan"/>).</param>
		/// <param name="output">Is set with the output of the command line. This may be null or empty.</param>
		private void Run( out bool finishedInTime, out int exitCode, out string output )
		{
			try
			{
				using( Process process = new Process() )
				{
					if( CommandLine.HasApplicationToRun )
					{
						process.StartInfo.FileName = CommandLine.ApplicationFileNameToRun;
						process.StartInfo.Arguments = CommandLine.ApplicationParameters;
						process.StartInfo.UseShellExecute = false;
					}
					else if( CommandLine.HasFileToOpen )
					{
						process.StartInfo.FileName = CommandLine.FileNameToRun;
						process.StartInfo.UseShellExecute = true;
					}
					else
					{
						ValidationUtils.Unhandled( "The command line has neither an application to run, nor a file to open."
							, null );
					}

					process.StartInfo.WorkingDirectory = CommandLine.WorkingFolder;

					if( CommandLine.MayCreateWindow || process.StartInfo.UseShellExecute )
					{
						process.StartInfo.WindowStyle = ProcessWindowStyle.Normal;
						process.StartInfo.CreateNoWindow = false;
						process.StartInfo.ErrorDialog = process.StartInfo.UseShellExecute;
					}
					else
					{
						process.StartInfo.WindowStyle = ProcessWindowStyle.Hidden;
						process.StartInfo.CreateNoWindow = true;
						process.StartInfo.ErrorDialog = false;
					}

					bool canRetrieveOutput = CommandLine.RetrieveOutput && !process.StartInfo.UseShellExecute;
					if( canRetrieveOutput )
					{
						process.StartInfo.RedirectStandardError = true;
						process.StartInfo.RedirectStandardOutput = true;
						process.ErrorDataReceived += Process_ErrorDataReceived;
						process.OutputDataReceived += Process_OutputDataReceived;

						ErrorDataEndedAutoResetEvent = new AutoResetEvent( false );
						OutputDataEndedAutoResetEvent = new AutoResetEvent( false );
					}

					bool outputIsComplete;

					try
					{
						process.Start();

						if( canRetrieveOutput )
						{
							process.BeginErrorReadLine();
							process.BeginOutputReadLine();
						}

						WaitProcessFinish( process, CommandLine.WaitFinishTimespan, CommandLine.KillAfterWait
							, out finishedInTime );

						exitCode = (CommandLine.WaitFinishTimespan > 0)
							? (finishedInTime ? process.ExitCode : int.MinValue)
							: 0 /* Show no error if the user specifies not to wait for the process to finish. */;

						// "CommandLineOutput" might be read before all the data is written to it. This is because a process
						// could finish before the "ErrorDataReceived" and "OutputDataReceived" events can write all the data to
						// "CommandLineOutput". Because of this, it's necessary to wait for these events to finish.
						//
						// There is no official documentation about how to determine if all the data has been written (to
						// "CommandLineOutput"), so we are waiting for the "ErrorDataReceived" and "OutputDataReceived" events to
						// receive a null data and set an auto reset event.
						if( canRetrieveOutput )
						{
							bool errorDataIsComplete = ErrorDataEndedAutoResetEvent.WaitOne( WaitCommandLineOutputTimeout );
							bool outputDataIsComplete = OutputDataEndedAutoResetEvent.WaitOne( WaitCommandLineOutputTimeout );

							outputIsComplete = errorDataIsComplete && outputDataIsComplete;
						}
						else
						{
							outputIsComplete = true;
						}
					}
					finally
					{
						process.ErrorDataReceived -= Process_OutputDataReceived;
						process.OutputDataReceived -= Process_OutputDataReceived;
					}

					lock( CommandLineOutputLock )
					{
						output = GetCommandLineOutput( outputIsComplete );
					}
				}
			}
			finally
			{
				if( ErrorDataEndedAutoResetEvent != null )
					ErrorDataEndedAutoResetEvent.Dispose();
				if( OutputDataEndedAutoResetEvent != null )
					OutputDataEndedAutoResetEvent.Dispose();

				ErrorDataEndedAutoResetEvent = null;
				OutputDataEndedAutoResetEvent = null;
			}
		}
	}

	#endregion Private methods
}
