////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.PeerDocuments;

namespace Vela.ActionService
{
	/// <summary>
	/// Action account context.
	/// </summary>
	public class ActionAccountContext : IExpirableAlgorithm
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionAccountContext()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="tally">The tally.</param>
		/// <param name="account">The account.</param>
		public ActionAccountContext( Tally tally, Account account )
		{
			Tally = tally;
			Account = account;

			ValidateLocators();
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The fingerprint of the UPN of the (normal / static) local peer, the one who has sought (over the network) and found
		/// the contained presenter. READ-ONLY.
		/// </summary>
		public Locator LocalUpnFingerprint
		{
			get { return Account.LocalUpnFingerprint; }
		}

		/// <summary>
		/// The fingerprint of the UPN of the (normal / static) remote peer. READ-ONLY.
		/// </summary>
		public Locator RemoteUpnFingerprint
		{
			get { return Account.RemoteUpnFingerprint; }
		}

		/// <summary>
		/// The tally locator. READ-ONLY.
		/// </summary>
		public Locator TallyLocator
		{
			get { return Tally.InstanceLocator; }
		}

		/// <summary>
		/// The account locator. READ-ONLY.
		/// </summary>
		public Locator AccountLocator
		{
			get { return Account.InstanceLocator; }
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public virtual DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( Tally.ExpirationTime, Account.ExpirationTime ); }
		}

		private Tally _Tally;
		/// <summary>
		/// The tally. READ-ONLY.
		/// </summary>
		public Tally Tally
		{
			get { return _Tally; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Tally" );

				_Tally = value;
			}
		}

		private Account _Account;
		/// <summary>
		/// The account. READ-ONLY.
		/// </summary>
		public Account Account
		{
			get { return _Account; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Account" );

				_Account = value;
			}
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Validates the locators.
		/// </summary>
		private void ValidateLocators()
		{
			if( !Tally.RemoteUpnFingerprint.Equals( Account.RemoteUpnFingerprint ) )
			{
				throw new ValidationException( "Different tally-account UPN fingerprints."
					, null, "Tally.RemoteUpnFingerprint != Account.RemoteUpnFingerprint" );
			}
		}

		#endregion Private methods
	}
}
