////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.ActionProcessor;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.PeerDocuments;

namespace Vela.ActionService
{
	/// <summary>
	/// Context for an action request.
	/// </summary>
	public class ActionRequestContext
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionRequestContext()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="serviceContext">The service context.</param>
		/// <param name="anonymousKnowledge">The sheaf which proves that the local peer has knowledge of the initiator of the
		/// current action request. This may be null.</param>
		/// <param name="actionResponseLocator">The locator of the action response associated to the current action
		/// request.</param>
		public ActionRequestContext( ActionServiceContext serviceContext, Sheaf anonymousKnowledge
			, Locator actionResponseLocator )
		{
			ServiceContext = serviceContext;
			AnonymousKnowledge = anonymousKnowledge;
			ActionResponseLocator = actionResponseLocator;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private ActionServiceContext _ServiceContext;
		/// <summary>
		/// The service context. READ-ONLY.
		/// </summary>
		public ActionServiceContext ServiceContext
		{
			get { return _ServiceContext; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ServiceContext" );

				_ServiceContext = value;
			}
		}

		private Sheaf _AnonymousKnowledge;
		/// <summary>
		/// The sheaf which proves that the local peer has knowledge of the initiator of the current action request. This may be
		/// null. READ-ONLY.
		/// </summary>
		public Sheaf AnonymousKnowledge
		{
			get { return _AnonymousKnowledge; }
			protected set { _AnonymousKnowledge = value; }
		}

		private Locator _ActionResponseLocator;
		/// <summary>
		/// The locator of the action response associated to the current action request. READ-ONLY.
		/// </summary>
		public Locator ActionResponseLocator
		{
			get { return _ActionResponseLocator; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionResponseLocator" );

				_ActionResponseLocator = value;
			}
		}

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Gets the approval of the initiator of the current anonymous action request, using <see cref="AnonymousKnowledge"/>.
		/// </summary>
		/// <returns>Returns the approval of the initiator of the current anonymous action request, or
		/// <see cref="TallyApprovals.Black"/> if no approval is found.</returns>
		public TallyApprovals GetAnonymousKnowledgeApproval()
		{
			TallyApprovals approval = TallyApprovals.Black;

			if( AnonymousKnowledge as Tally != null )
			{
				approval = ((Tally)AnonymousKnowledge).Approval;
			}
			else if( AnonymousKnowledge as CreateAccountClientProgress != null )
			{
				approval = ((CreateAccountClientProgress)AnonymousKnowledge).Approval;
			}

			return approval;
		}

		#endregion Public methods
	}
}
