////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;

namespace Vela.ActionService
{
	/// <summary>
	/// Context for processing an action request with beneficiary.
	/// </summary>
	/// <remarks>
	/// <para>When an action request arrives at the service for processing, the beneficiary account context is loaded from the
	/// service's storehouse. This context is forwarded to the action processing method.</para>
	/// </remarks>
	public class ActionRequestWithBeneficiaryContext : ActionRequestWithInitiatorContext
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionRequestWithBeneficiaryContext()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="serviceContext">The service context.</param>
		/// <param name="actionResponseLocator">The locator of the action response associated to the current action
		/// request.</param>
		/// <param name="initiatorAccountContext">The account context of the initiator of the action request, for which to
		/// process the action request.</param>
		/// <param name="beneficiaryAccountContext">The account context of the beneficiary of the action request.</param>
		public ActionRequestWithBeneficiaryContext( ActionServiceContext serviceContext, Locator actionResponseLocator
				, ActionAccountContext initiatorAccountContext, ActionAccountContext beneficiaryAccountContext )
			: base( serviceContext, actionResponseLocator, initiatorAccountContext )
		{
			BeneficiaryAccountContext = beneficiaryAccountContext;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private ActionAccountContext _BeneficiaryAccountContext;
		/// <summary>
		/// The account context of the beneficiary of the action request. READ-ONLY.
		/// </summary>
		public ActionAccountContext BeneficiaryAccountContext
		{
			get { return _BeneficiaryAccountContext; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "BeneficiaryAccountContext" );

				_BeneficiaryAccountContext = value;
			}
		}

		#endregion Public non-serializable properties
	}
}
