////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.ActionProcessor;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Kernel.Cloning;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.ActionService
{
	/// <summary>
	/// Context for processing an action request by an action service.
	/// </summary>
	public class ActionServiceContext
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionServiceContext()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="processorConfiguration">The action processor configuration.</param>
		/// <param name="listenerConfiguration">The action service listener configuration.</param>
		/// <param name="initiatorIpAddress">The IP address of the initiator of the action request. This may be null or empty.
		/// </param>
		/// <param name="serializer">The service serializer.</param>
		/// <param name="vault">The service vault.</param>
		/// <param name="recipientUpn">The UPN of the service which is the recipient of the action request.</param>
		/// <param name="processingPort">The network port where the action request was received.</param>
		/// <param name="transaction">Current storehouse transaction.</param>
		/// <param name="actionServiceHost">Host used for updating the user interface of an action service application after an
		/// action request and response are processed. Its content must be thread safe.</param>
		/// <param name="actionClientManager">Action client manager which is notified by the action service about (local) action
		/// requests which were added to the (local) storehouse by the service, and which have to be processed after the current
		/// (remote) action request is processed by the service. Its content must be thread safe.</param>
		public ActionServiceContext( IStorehouseProvider storehouseProvider, ActionProcessorConfiguration processorConfiguration
			, ActionServiceListenerConfiguration listenerConfiguration, byte[] initiatorIpAddress, ISerializer serializer
			, Vault vault, Upn recipientUpn, UInt16 processingPort, IStorehouseTransaction transaction
			, IActionServiceHost actionServiceHost, IActionClientManager actionClientManager )
		{
			StorehouseProvider = storehouseProvider;
			ProcessorConfiguration = processorConfiguration;
			ListenerConfiguration = listenerConfiguration;
			InitiatorIpAddress = initiatorIpAddress;
			Serializer = serializer;
			Vault = vault;
			RecipientUpn = recipientUpn;
			ProcessingPort = processingPort;
			Transaction = transaction;
			ActionServiceHost = actionServiceHost;
			ActionClientManager = actionClientManager;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private IStorehouseProvider _StorehouseProvider;
		/// <summary>
		/// Storehouse provider from which to create a storehouse transaction. READ-ONLY.
		/// </summary>
		public IStorehouseProvider StorehouseProvider
		{
			get { return _StorehouseProvider; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "StorehouseProvider" );

				_StorehouseProvider = value;
			}
		}

		private ActionProcessorConfiguration _ProcessorConfiguration;
		/// <summary>		
		/// The action processor configuration. READ-ONLY.
		/// </summary>
		public ActionProcessorConfiguration ProcessorConfiguration
		{
			get { return _ProcessorConfiguration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ProcessorConfiguration" );

				_ProcessorConfiguration = value;
			}
		}

		private ActionServiceListenerConfiguration _ListenerConfiguration;
		/// <summary>
		/// The action service listener configuration. READ-ONLY.
		/// </summary>
		public ActionServiceListenerConfiguration ListenerConfiguration
		{
			get { return _ListenerConfiguration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ListenerConfiguration" );

				_ListenerConfiguration = value;
			}
		}

		private byte[] _InitiatorIpAddress;
		/// <summary>
		/// The IP address of the initiator of the action request. This may be null or empty. READ-ONLY.
		/// </summary>
		public byte[] InitiatorIpAddress
		{
			get { return _InitiatorIpAddress; }
			protected set { _InitiatorIpAddress = value; }
		}

		private ISerializer _Serializer;
		/// <summary>
		/// The service serializer. READ-ONLY.
		/// </summary>
		public ISerializer Serializer
		{
			get { return _Serializer; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Serializer" );

				_Serializer = value;
			}
		}

		private Vault _Vault;
		/// <summary>
		/// The service vault. READ-ONLY.
		/// </summary>
		public Vault Vault
		{
			get { return _Vault; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Vault" );

				_Vault = value;
			}
		}

		private Upn _RecipientUpn;
		/// <summary>
		/// The UPN of the service which is the recipient of the action request. READ-ONLY.
		/// </summary>
		public Upn RecipientUpn
		{
			get { return _RecipientUpn; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RecipientUpn" );

				_RecipientUpn = value;
			}
		}

		private UInt16 _ProcessingPort;
		/// <summary>
		/// The network port where the action request was received. This must be at least 1. READ-ONLY.
		/// </summary>
		public UInt16 ProcessingPort
		{
			get { return _ProcessingPort; }
			protected set
			{
				ValidationUtils.Integer( value, 1, UInt16.MaxValue, "ProcessingPort" );

				_ProcessingPort = value;
			}
		}

		[UnclonableAttribute]
		private IStorehouseTransaction _Transaction;
		/// <summary>
		/// Current storehouse transaction. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is not committed (by an action service) if an error response is sent.</para>
		/// 
		/// <para>If an action request with a multiframe response is being processed, this is available only to the first frame
		/// (in the method which processes the action response).</para>
		/// </remarks>
		public IStorehouseTransaction Transaction
		{
			get { return _Transaction; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Transaction" );

				_Transaction = value;
			}
		}

		[UnclonableAttribute]
		private IActionServiceHost _ActionServiceHost;
		/// <summary>
		/// Host used for updating the user interface of an action service application after an action request and response
		/// are processed. Its content must be thread safe. READ-ONLY.
		/// </summary>
		public IActionServiceHost ActionServiceHost
		{
			get { return _ActionServiceHost; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionServiceHost" );

				_ActionServiceHost = value;
			}
		}

		[UnclonableAttribute]
		private IActionClientManager _ActionClientManager;
		/// <summary>
		/// Action client manager which is notified by the action service about (local) action requests which were added to the
		/// (local) storehouse by the service, and which have to be processed after the current (remote) action request is
		/// processed by the service. Its content must be thread safe. READ-ONLY.
		/// </summary>
		public IActionClientManager ActionClientManager
		{
			get { return _ActionClientManager; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionClientManager" );

				_ActionClientManager = value;
			}
		}

		private bool _ActionProcessedBefore = false;
		/// <summary>
		/// While an action request is being processed, this is set with true if the action request has been processed before
		/// (the current request). If this is true, the action response is the previous action response.
		/// </summary>
		public bool ActionProcessedBefore
		{
			get { return _ActionProcessedBefore; }
			set { _ActionProcessedBefore = value; }
		}

		private Locator _RemoteUpnFingerprint;
		/// <summary>
		/// While an action request is being processed, this is set with the remote UPN fingerprint; this is taken from the
		/// initiator's account. This is null if there is no initiator account.
		/// </summary>
		public Locator RemoteUpnFingerprint
		{
			get { return _RemoteUpnFingerprint; }
			set { _RemoteUpnFingerprint = value; }
		}

		private ActionAccountContext _InitiatorAccountContext;
		/// <summary>
		/// While an action request is being processed, this is set with the initiator's account context. This is null if there
		/// is no initiator account.
		/// </summary>
		public ActionAccountContext InitiatorAccountContext
		{
			get { return _InitiatorAccountContext; }
			set { _InitiatorAccountContext = value; }
		}

		private bool _IsFirstFrame = true;
		/// <summary>
		/// If this is true then the service is processing the first frame of a multiframe action response. The default value is
		/// true.
		/// </summary>
		public bool IsFirstFrame
		{
			get { return _IsFirstFrame; }
			set { _IsFirstFrame = value; }
		}

		private Locators _NotifyActionClientManagerAboutRequests = new Locators();
		/// <summary>
		/// While an action request is being processed, the action client manager associated to the current service is notified
		/// to process the action request locators from this list. The backing field is initialized by default. This may be
		/// empty, but not null.
		/// </summary>
		public Locators NotifyActionClientManagerAboutRequests
		{
			get { return _NotifyActionClientManagerAboutRequests; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "NotifyActionClientManagerAboutRequests" );

				_NotifyActionClientManagerAboutRequests = value;
			}
		}

		#endregion Public non-serializable properties
	}
}
