////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.ActionProcessor;
using Vela.Kernel;
using Vela.Kernel.Cloning;
using Vela.Storehouse;

namespace Vela.ActionService
{
	/// <summary>
	/// Action service listener context.
	/// </summary>
	public class ActionServiceListenerContext
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionServiceListenerContext()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="processorConfiguration">The action processor configuration.</param>
		/// <param name="configuration">Configuration for the action service listener.</param>
		/// <param name="actionServiceVaults">The list of vaults which can be the recipients of action requests.</param>
		/// <param name="processRequestMethodInfos">List of method informations which can process action requests. Do not modify
		/// the content of this object while the processor is running.</param>
		/// <param name="actionServiceHost">Host used for updating the user interface of an action service application after an
		/// action request and response are processed. Its content must be thread safe.</param>
		/// <param name="actionClientManager">Action client manager which is notified by the action service about (local) action
		/// requests which were added to the (local) storehouse by the service, and which have to be processed after the current
		/// (remote) action request is processed by the service. Its content must be thread safe.</param>
		public ActionServiceListenerContext( IStorehouseProvider storehouseProvider
			, ActionProcessorConfiguration processorConfiguration, ActionServiceListenerConfiguration configuration
			, ActionServiceVaults actionServiceVaults, ProcessRequestMethodInfos processRequestMethodInfos
			, IActionServiceHost actionServiceHost, IActionClientManager actionClientManager )
		{
			StorehouseProvider = storehouseProvider;
			ProcessorConfiguration = processorConfiguration;
			Configuration = configuration;
			ActionServiceVaults = actionServiceVaults;
			ProcessRequestMethodInfos = processRequestMethodInfos;
			ActionServiceHost = actionServiceHost;
			ActionClientManager = actionClientManager;

			// The UPNs of the vaults (from "ActionServiceVaults") must have been generated (in order to avoid having to later
			// lock "ActionServiceVaults" or its items).
			foreach( ActionServiceVault actionServiceVault in ActionServiceVaults )
			{
				if( ListUtils.NoContent( actionServiceVault.Upns ) )
				{
					throw new ValidationException( string.Format( "The UPNs of the vault with the locator \"{0}\" have"
						+ " not been generated.", actionServiceVault.Vault.InstanceLocator ) );
				}
			}
		}

		#endregion Constructors

		#region Public non-serializable properties

		private IStorehouseProvider _StorehouseProvider;
		/// <summary>
		/// Storehouse provider from which to create a storehouse transaction. READ-ONLY.
		/// </summary>
		public IStorehouseProvider StorehouseProvider
		{
			get { return _StorehouseProvider; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "StorehouseProvider" );

				_StorehouseProvider = value;
			}
		}

		private ActionProcessorConfiguration _ProcessorConfiguration;
		/// <summary>		
		/// The action processor configuration. READ-ONLY.
		/// </summary>
		public ActionProcessorConfiguration ProcessorConfiguration
		{
			get { return _ProcessorConfiguration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ProcessorConfiguration" );

				_ProcessorConfiguration = value;
			}
		}

		private ActionServiceListenerConfiguration _Configuration;
		/// <summary>
		/// The action service listener configuration. READ-ONLY.
		/// </summary>
		public ActionServiceListenerConfiguration Configuration
		{
			get { return _Configuration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Configuration" );

				_Configuration = value;
			}
		}

		[UnclonableAttribute]
		private ActionServiceVaults _ActionServiceVaults;
		/// <summary>
		/// The vaults which can be the recipient of action requests. The UPNs of the vaults (from this object) must have been
		/// computed (in order to avoid having to later lock this object or its items). This may be empty, but not null.
		/// READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The used vault (from this object) is cloned, so this object doesn't have to be locked.</para>
		/// </remarks>
		public ActionServiceVaults ActionServiceVaults
		{
			get { return _ActionServiceVaults; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionServiceVaults" );

				_ActionServiceVaults = value;
			}
		}

		[UnclonableAttribute]
		private ProcessRequestMethodInfos _ProcessRequestMethodInfos;
		/// <summary>
		/// List of method informations which can process action requests. Do not modify the content of this object while the
		/// processor is running. READ-ONLY.
		/// </summary>
		public ProcessRequestMethodInfos ProcessRequestMethodInfos
		{
			get { return _ProcessRequestMethodInfos; }
			protected set
			{
				ValidationUtils.ListHasContent( value, "ProcessRequestMethodInfos" );

				_ProcessRequestMethodInfos = value;
			}
		}

		[UnclonableAttribute]
		private IActionServiceHost _ActionServiceHost;
		/// <summary>
		/// Host used for updating the user interface of an action service application after an action request and response
		/// are processed. Its content must be thread safe. READ-ONLY.
		/// </summary>
		public IActionServiceHost ActionServiceHost
		{
			get { return _ActionServiceHost; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionServiceHost" );

				_ActionServiceHost = value;
			}
		}

		[UnclonableAttribute]
		private IActionClientManager _ActionClientManager;
		/// <summary>
		/// Action client manager which is notified by the action service about (local) action requests which were added to the
		/// (local) storehouse by the service, and which have to be processed after the current (remote) action request is
		/// processed by the service. Its content must be thread safe. READ-ONLY.
		/// </summary>
		public IActionClientManager ActionClientManager
		{
			get { return _ActionClientManager; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionClientManager" );

				_ActionClientManager = value;
			}
		}

		#endregion Public non-serializable properties
	}
}
