////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.ActionProcessor;
using Vela.Kernel;
using Vela.Kernel.Cloning;
using Vela.Storehouse;

namespace Vela.ActionService
{
	/// <summary>
	/// Context for a service which processes action requests.
	/// </summary>
	public class ActionServiceManagerContext
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionServiceManagerContext()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="processorConfiguration">The action processor configuration.</param>
		/// <param name="actionServices">The action service configurations.</param>
		/// <param name="actionServiceTypeNames">The type names of the action services which can process action requests, and
		/// which are available to all the vaults. Do not modify the content of this object while the processor is running.
		/// </param>
		/// <param name="actionServiceHost">Host used for updating the user interface of an action service application after an
		/// action request and response are processed. Its content must be thread safe.</param>
		/// <param name="actionClientManager">Action client manager which is notified by the action service about (local) action
		/// requests which were added to the (local) storehouse by the service, and which have to be processed after the current
		/// (remote) action request is processed by the service. Its content must be thread safe.</param>
		public ActionServiceManagerContext( IStorehouseProvider storehouseProvider
			, ActionProcessorConfiguration processorConfiguration, ActionServiceListenerConfigurations actionServices
			, Strings actionServiceTypeNames, IActionServiceHost actionServiceHost, IActionClientManager actionClientManager )
		{
			StorehouseProvider = storehouseProvider;
			ProcessorConfiguration = processorConfiguration;
			ActionServices = actionServices;
			ActionServiceTypeNames = actionServiceTypeNames;
			ActionServiceHost = actionServiceHost;
			ActionClientManager = actionClientManager;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private IStorehouseProvider _StorehouseProvider;
		/// <summary>
		/// Storehouse provider from which to create a storehouse transaction. READ-ONLY.
		/// </summary>
		public IStorehouseProvider StorehouseProvider
		{
			get { return _StorehouseProvider; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "StorehouseProvider" );

				_StorehouseProvider = value;
			}
		}

		private ActionProcessorConfiguration _ProcessorConfiguration;
		/// <summary>		
		/// The action processor configuration. READ-ONLY.
		/// </summary>
		public ActionProcessorConfiguration ProcessorConfiguration
		{
			get { return _ProcessorConfiguration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ProcessorConfiguration" );

				_ProcessorConfiguration = value;
			}
		}

		private ActionServiceListenerConfigurations _ActionServices;
		/// <summary>
		/// The action service configurations. READ-ONLY.
		/// </summary>
		public ActionServiceListenerConfigurations ActionServices
		{
			get { return _ActionServices; }
			protected set
			{
				ValidationUtils.ListHasContent( value, "ActionServices" );

				_ActionServices = value;
			}
		}

		[UnclonableAttribute]
		private Strings _ActionServiceTypeNames;
		/// <summary>
		/// The type names of the action services which can process action requests, and which are available to all the vaults.
		/// Do not modify the content of this object while the processor is running. READ-ONLY.
		/// </summary>
		public Strings ActionServiceTypeNames
		{
			get { return _ActionServiceTypeNames; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionServiceTypeNames" );

				_ActionServiceTypeNames = value;
			}
		}

		[UnclonableAttribute]
		private IActionServiceHost _ActionServiceHost;
		/// <summary>
		/// Host used for updating the user interface of an action service application after an action request and response are
		/// processed. Its content must be thread safe. READ-ONLY.
		/// </summary>
		public IActionServiceHost ActionServiceHost
		{
			get { return _ActionServiceHost; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionServiceHost" );

				_ActionServiceHost = value;
			}
		}

		[UnclonableAttribute]
		private IActionClientManager _ActionClientManager;
		/// <summary>
		/// Action client manager which is notified by the action service about (local) action requests which were added to the
		/// (local) storehouse by the service, and which have to be processed after the current (remote) action request is
		/// processed by the service. Its content must be thread safe. READ-ONLY.
		/// </summary>
		public IActionClientManager ActionClientManager
		{
			get { return _ActionClientManager; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionClientManager" );

				_ActionClientManager = value;
			}
		}

		#endregion Public non-serializable properties
	}
}
