////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.ActionProcessor;
using Vela.Kernel;

namespace Vela.ActionService
{
	/// <summary>
	/// Contract for an action service.
	/// </summary>
	/// <remarks>
	/// <para>An action service processes action requests (<see cref="Vela.ActionProcessor.ActionRequest"/>), and
	/// responds with action responses (<see cref="Vela.ActionProcessor.ActionResponse"/>).</para>
	/// 
	/// <para>All the sheafs are locked with <see cref="Vela.CoreDocuments.SheafLocks.Shared"/>, unless they will be saved (in
	/// which case they are locked with <see cref="Vela.CoreDocuments.SheafLocks.Update"/>).</para>
	/// </remarks>
	public interface IActionService : IPortableType
	{
		#region Public methods

		/// <summary>
		/// Processes the specified action request (as a first frame of a request which has a multiframe response).
		/// </summary>
		/// <param name="actionRequest">The action request to process.</param>
		/// <param name="serviceContext">The action service context.</param>
		/// <param name="processRequestMethodInfo">The information of the method which can process the action request.</param>
		/// <param name="actionRequestContext">Is set with the created action request context.</param>
		/// <returns>Returns the action response which is to be sent back to the initiator of the action request.</returns>
		/// <remarks>
		/// <para>Expired action requests are not processed.</para>
		/// </remarks>
		ActionResponse ProcessFirstRequestFrame( ActionRequest actionRequest, ActionServiceContext serviceContext
			, ProcessRequestMethodInfo processRequestMethodInfo, out ActionRequestContext actionRequestContext );

		/// <summary>
		/// Processes the specified action request (as a secondary frame of a request which has a multiframe response).
		/// </summary>
		/// <param name="actionRequest">The action request to process.</param>
		/// <param name="serviceContext">The action service context.</param>
		/// <param name="processRequestMethodInfo">The information of the method which can process the action request.</param>
		/// <param name="actionRequestContext">The action request context created when the first frame was processed.</param>
		/// <returns>Returns the action response which is to be sent back to the initiator of the action request.</returns>
		ActionResponse ProcessSecondaryRequestFrame( ActionRequest actionRequest
			, ActionServiceContext serviceContext, ActionRequestContext actionRequestContext
			, ProcessRequestMethodInfo processRequestMethodInfo );

		#endregion Public methods
	}
}
