////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

namespace Vela.ActionService
{
	/// <summary>
	/// <para>An action service receives action requests, dispatches them to the appropriate processing module, and sends back
	/// the action responses.</para>
	/// 
	/// <para>An action response may be multiframe
	/// (see <see cref="Vela.ActionProcessor.IActionRequestWithMultiframeResponse"/> and
	/// <see cref="Vela.ActionProcessor.IMultiframeActionResponse"/> for details).</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Code flow</b></para>
	/// 
	/// <para>The action service listener receives an encrypted envelope with an action request, as a buffer. This is
	/// deserialized and is dispatched to the appropriate action service based on the equality between the UPN fingerprint of the
	/// processing identity of the service and the UPN fingerprint of one of the recipient session keys of the encrypted
	/// envelope.</para>
	/// 
	/// <para>The action service decrypts the encrypted envelope and extracts the action request from it.</para>
	/// 
	/// <para>A storehouse transaction is created. If the action request has a multiframe action response, the transaction is
	/// available only for the first frame (because the action request may take a long time to be processed), but a new one can
	/// be created from the available storehouse provider; make sure to close the created transaction.</para>
	/// 
	/// <para>If <see cref="Vela.ActionProcessor.ActionRequest.ServiceIgnoresActionRequestIfTransportedSheafExists"/> is true and
	/// the storehouse already contains the <see cref="Vela.ActionProcessor.ActionRequest.TransportedSheaf"/> sheaf then the
	/// service ignores the current action request and returns a basic uniframe response (with the type
	/// <see cref="Vela.ActionProcessor.ActionRequest.ActionResponseType"/>) constructed with the standard constructor of an
	/// action response.</para>
	/// 
	/// <para>An action service is instantiated (every time an encrypted envelope is received), so don't perform time consuming
	/// tasks in a service's constructor. This also means that static objects must be synchronized for multi-threading.</para>
	/// 
	/// <para>Expired action requests are not processed.</para>
	/// 
	/// <para>If the action request has an initiator, the tally and the account of the initiator are loaded. The signature of the
	/// initiator (on the action request) is verified.</para>
	/// 
	/// <para>If an action response was already sent (for the current action request), it is sent again. The storehouse
	/// transaction is committed.</para>
	/// 
	/// <para>If the action request has a beneficiary, the tally and the account of the beneficiary are loaded. The UPN
	/// fingerprint of the beneficiary (from the action request) is verified against that of the identity from the loaded
	/// tally.</para>
	/// 
	/// <para>An action response speckle is allocated in the storehouse, if action responses are to be saved by the
	/// service.</para>
	/// 
	/// <para>The derived action service classes must contain public instance processing methods which have one parameter
	/// derived from <see cref="Vela.ActionProcessor.ActionRequest"/> and one parameter derived from
	/// <see cref="ActionRequestContext"/>, and the return value derived from
	/// <see cref="Vela.ActionProcessor.ActionResponse"/>. These methods are executed by the service at a
	/// client's request; the method whose action request parameter has the same type (as the received action request) is
	/// executed. If the action request has an initiator or a beneficiary, the action request context and the action
	/// response must also have.</para>
	/// 
	/// <para>In the processing method, you must create the action response and set its properties.</para>
	/// 
	/// <para>In the processing method, you can add / update documents in the storehouse, or delete them from the storehouse,
	/// using the storehouse transaction.</para>
	/// 
	/// <para>In the processing method, do not modify the vault and do not try to save it in the storehouse. If you must
	/// absolutely save some data associated to the service, do so in another document (which you must first load from the
	/// storehouse). Still, note that this may severely impact the scalability of the service.</para>
	/// 
	/// <para>If the action request has a multiframe action response, the processing method is called until it returns null, or
	/// returns a response which is not derived from <see cref="Vela.ActionProcessor.IMultiframeActionResponse"/>). After all
	/// the frames of the multiframe response are sent, a response with the type
	/// <see cref="Vela.ActionProcessor.LastFrameActionResponse"/> is sent. This allows the service to perform specific
	/// operations after the entire multiframe response is sent successfully.</para>
	/// 
	/// <para>If any error occurs in the processing method, throw an exception. In such a case, the storehouse transaction will
	/// be rolled back, and an error response will be sent back to the initiator of the action request.</para>
	/// 
	/// <para>If you want the exception type and message to reach the initiator of the action request, you must throw an
	/// exception of type <see cref="Vela.Kernel.CrossServiceException"/> (or derived from it).</para>
	/// 
	/// <para>The secondary frames of a multiframe action response are not signed if they are derived from
	/// <see cref="Vela.ActionProcessor.IMultiframeActionResponse"/>.</para>
	/// 
	/// <para>The action request and response are saved, if they are to be saved by the service. The tallies and the accounts of
	/// the initiator and the beneficiary are automatically saved (if mandated so by the action request).</para>
	/// 
	/// <para>If <see cref="Vela.ActionProcessor.ActionRequest.ServiceSavesProgress"/> is true, an action progress
	/// (<see cref="Vela.ActionProcessor.ActionServiceProgress"/>) is created and saved in the storehouse.</para>
	/// 
	/// <para>The storehouse transaction is committed.</para>
	/// 
	/// <para>The action response is encrypted and serialized, and then sent back to the initiator of the action request. A
	/// stream may be attached after the action response; this allows a service to send a very large response.</para>
	/// 
	/// <para>Entropy is added to the entropy algorithm of the service, using only private data from the action request.</para>
	/// 
	/// <para>The ephemerid of the initiator of the action request is updated (with the new IP address of the initiator).</para>
	/// 
	/// <para>The service application's UI is updated.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Re-processing action request</b></para>
	/// 
	/// <para>Sometimes, a peer could send the same action request several times, to the service, because he can't get a visual
	/// confirmation that something has happened. However, it's usually crucial that the service doesn't process the same action
	/// multiple times, but only once.</para>
	/// 
	/// <para>For example, if a payment action is sent twice, the service must make the payment only once.</para>
	/// 
	/// <para>For this to happen, action responses must be saved (see
	/// <see cref="Vela.ActionProcessor.ActionRequest.ServiceSavesSentActionResponse"/>). This is because if an
	/// action request was previously processed, it's not executed again but the same action response is resent.</para>
	/// 
	/// <para>It may also be important to have a relatively small
	/// <see cref="Vela.PeerDocuments.Appendix.ActionRequestValidity"/> in order to be able to delete the expired action
	/// responses before they slow down the storehouse too much.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Service scalability</b></para>
	/// 
	/// <para>The scalability of the service depends on the scalability of the used storehouse.</para>
	/// 
	/// <para>If a service runs on multiple computers, where each of them receives action requests (through traffic
	/// redirectors), sequentially or randomly distributed, the service is scalable up to the storehouse's capacity.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Multi-threading synchronization</b></para>
	/// 
	/// <para>All non-centralized used resources must be locked while processing each action request. The centralized resources
	/// should not be modified while the service is running; an action service receives a copy of the vault, so it can change its
	/// properties, but this should be avoided.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Various details</b></para>
	/// 
	/// <para>The action service doesn't lock the storehouse for its thread. Therefore, you must not change the properties of
	/// the storehouse from other threads. If you want to do so, you must first stop the service.</para>
	/// 
	/// <para>The service relies on the storehouse for transactionality and fault-tolerance.</para>
	/// 
	/// <para>Don't add the IP address of the action request to the entropy algorithm because if action requests come in through
	/// traffic redirecting services, the IP would always be the same (or one of very few possible IP addresses).</para>
	/// </summary>
	class NamespaceDoc
	{
		// Only for documentation.
	}
}
