////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using Vela.ActionProcessor;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.ActionService
{
	/// <summary>
	/// Contract for a service which processes elementary action requests (<see cref="ActionRequest"/>).
	/// </summary>
	public abstract class StarkActionService : ActionService
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected StarkActionService()
		{
		}

		#endregion Constructors

		#region Private static methods

		/// <summary>
		/// Verifies if the specified action request to download files may be processed; if it may not, an exception is thrown.
		/// </summary>
		/// <param name="actionRequest">Action request to process.</param>
		/// <param name="requestContext">Action request context.</param>
		public void VerifyIfDownloadFilesIsAllowed( DownloadFilesActionRequest actionRequest
			, ActionRequestWithInitiatorContext requestContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( requestContext, "requestContext" );

			IStorehouseTransaction transaction = requestContext.ServiceContext.Transaction;

			if( actionRequest.AttachedFiles.Count <= 0 )
				throw new FatalCrossServiceException( "No files to transfer were specified." );

			// Load the granter.
			IDeserializer deserializer;
			Letter granter = transaction.LoadSheaf( actionRequest.AccessGranterLocator, StorehouseRealms.Posts, typeof( Letter )
				, SheafLocks.Shared, out deserializer ) as Letter;

			if( granter == null )
			{
				throw new FatalCrossServiceException( "The access granter doesn't exist, so no files can be transferred. The"
						+ " remote peer may have deleted the letter through which the files were sent to you.", null
					, new ExceptionDetail( "Access granter locator", actionRequest.AccessGranterLocator ) );
			}

			// Check if the initiator of the request is recipient of the granter.
			Locator initiatorUpnFingerprint = requestContext.InitiatorAccountContext.Tally.RemoteUpnFingerprint;
			if( !initiatorUpnFingerprint.Equals( granter.RecipientUpnFingerprint ) )
			{
				throw new FatalCrossServiceException( "The access granter indicates that the UPN fingerprint of the recipient of"
						+ " the files to transfer is different than the UPN fingerprint of the initiator of the action request"
						+ " which has requested the file transfer.", null
					, new ExceptionDetail( "Access granter locator", granter.InstanceLocator )
					, new ExceptionDetail( "Recipient UPN fingerprint", granter.RecipientUpnFingerprint )
					, new ExceptionDetail( "Initiator UPN fingerprint", initiatorUpnFingerprint )
					, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator ) );
			}

			// Check if the initiator requests only the offered files.
			if( !actionRequest.AttachedFiles.SameFiles( granter.AttachedFiles ) )
			{
				throw new FatalCrossServiceException( "The access granter doesn't refer at least 1 of the attached files"
						+ " requested by the action request which has requested the file transfer.", null
					, new ExceptionDetail( "Access granter locator", granter.InstanceLocator )
					, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator ) );
			}

			// Check if there is any changed file.
			string changedFileName = AttachedFilesUtils.CheckIfChangedOriginalFile( actionRequest.AttachedFiles );
			if( changedFileName != null )
			{
				throw new FatalCrossServiceException( "The file can't be transferred because it has been changed or deleted."
					, null, new ExceptionDetail( "File name", changedFileName ) );
			}
		}

		/// <summary>
		/// Creates a secondary processing state for a received letter with the specified message.
		/// </summary>
		/// <param name="remotePeerMayRunLocalCommandLines">If this is true then the remote peer may run command lines on the
		/// local computer (by sending letters whose messages are used as command lines).</param>
		/// <param name="tallyApproval">The tally approval for the initiator of the letter.</param>
		/// <param name="letterMessage">The message of the received letter.</param>
		/// <param name="downloadFilesFolderName">The name of the folder where files would / will be downloaded. This may be null
		/// or empty (if no files are to be downloaded).</param>
		/// <param name="secondaryProcessingOutput">Is set with the output of the secondary processing. This may be used to send
		/// back to the client a large amount of data which should not be stored in an action progress (where the secondary
		/// processing state is stored). It's highly recommended to treat this information as ephemeral (in terms of storage
		/// requirements). This may be null or empty.</param>
		/// <returns>Returns the created secondary processing state.</returns>
		private static SecondaryProcessingState CreateSecondaryProcessingStateForLetter(
			bool remotePeerMayRunLocalCommandLines, TallyApprovals tallyApproval, string letterMessage
			, string downloadFilesFolderName, out string secondaryProcessingOutput )
		{
			secondaryProcessingOutput = null;

			SecondaryProcessingState commandLineState = null;
			SecondaryProcessingState downloadFilesState = null;

			if( remotePeerMayRunLocalCommandLines && PeerDocumentsUtils.RemotePeerMayRunLocalCommandLines( tallyApproval ) )
			{
				commandLineState = CommandLineProcessor.RunForLetter( letterMessage, out secondaryProcessingOutput );

				// The folder where files would / will be downloaded is only returned if the initiator of the letter can run
				// command lines on the local computer.
				if( TextUtils.HasContent( downloadFilesFolderName ) )
				{
					downloadFilesState = new SecondaryProcessingState( false, "Download folder", null, downloadFilesFolderName
						, true );
				}
			}

			// The command line is first because it's more important for the user to see the state of the command.
			return SecondaryProcessingState.Merge( commandLineState, downloadFilesState );
		}

		#endregion Private static methods

		#region Public methods

		/// <summary>
		/// Processes an anonymous action request to get an identity.
		/// </summary>
		/// <param name="actionRequest">Action request to process.</param>
		/// <param name="requestContext">Action request context.</param>
		/// <returns>Returns the action response which is to be sent back to the initiator of the action request.</returns>
		public virtual ResolveUpnAnonymousActionResponse ProcessResolveUpnAnonymous(
			ResolveUpnAnonymousActionRequest actionRequest, ActionRequestContext requestContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( requestContext, "requestContext" );

			// There is no need to load the newest vault from the storehouse because the client is not supposed to store the
			// returned information (but only use it to resolve the service's UPN). Besides, this behavior optimizes performance
			// (in case there are many requests to resolve the service's UPN).
			Vault vault = requestContext.ServiceContext.Vault;

			#region Process the action request

			SolitaryIdentityPublisher identityPublisher = CoreDocumentsUtils.GetSolitaryIdentity( vault.ProcessingIdentity );

			#endregion Process the action request

			#region Prepare the action response

			ResolveUpnAnonymousActionResponse actionResponse = new ResolveUpnAnonymousActionResponse(
				requestContext.ActionResponseLocator, actionRequest, null, identityPublisher );

			#endregion Prepare the action response

			return actionResponse;
		}

		/// <summary>
		/// Processes an anonymous action request to get a presenter and create an account.
		/// </summary>
		/// <param name="actionRequest">Action request to process.</param>
		/// <param name="requestContext">Action request context.</param>
		/// <returns>Returns the action response which is to be sent back to the initiator of the action request.</returns>
		public virtual CreateAccountAnonymousActionResponse ProcessCreateAccountAnonymous(
			CreateAccountAnonymousActionRequest actionRequest, ActionRequestContext requestContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( requestContext, "requestContext" );

			IStorehouseTransaction transaction = requestContext.ServiceContext.Transaction;

			#region Process the action request

			Locator vaultLocator = requestContext.ServiceContext.Vault.InstanceLocator;

			// Get the newest vault. The one from "requestContext.ServiceContext.Vault" contain old information if the user
			// has edited the vault after the last time when the service was restarted.
			IDeserializer deserializer;
			Vault vault = transaction.LoadSheaf( vaultLocator, StorehouseRealms.ProcessingIdentityContainers, typeof( Vault )
				, SheafLocks.Shared, out deserializer ) as Vault;

			Presenter presenter = ActionProcessorStorehouseUtils.GetPresenter( transaction, actionRequest.MacAlg
				, actionRequest.AsymmSigningPaddingAlg, vault, requestContext.GetAnonymousKnowledgeApproval()
				, requestContext.ServiceContext.Serializer.GetHashSerializer( false ) );

			#endregion Process the action request

			#region Prepare the action response

			CreateAccountAnonymousActionResponse actionResponse = new CreateAccountAnonymousActionResponse(
				requestContext.ActionResponseLocator, actionRequest, null, presenter );

			#endregion Prepare the action response

			return actionResponse;
		}

		/// <summary>
		/// Processes an action request which sends a local presenter to update it on a remote peer.
		/// </summary>
		/// <param name="actionRequest">Action request to process.</param>
		/// <param name="requestContext">Action request context.</param>
		/// <returns>Returns the action response which is to be sent back to the initiator of the action request.</returns>
		public virtual PushUpdatePresenterActionResponse ProcessPushUpdatePresenter(
			PushUpdatePresenterActionRequest actionRequest, ActionRequestWithInitiatorContext requestContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( requestContext, "requestContext" );

			IStorehouseTransaction transaction = requestContext.ServiceContext.Transaction;

			#region Process the action request

			Locator vaultLocator = requestContext.ServiceContext.Vault.InstanceLocator;

			// Since the account already exists (at the remote peer), the approval is ignored by the following call, so there is
			// no need to load the real one from the tally. Also, action requests from blocked peers don't get this far.
			ActionProcessorStorehouseUtils.AddAccountFromPresenter( transaction, vaultLocator, actionRequest.Presenter, null
				, TallyApprovals.Black /* Ensure that a bug somewhere doesn't increase the approval. */, null );

			#endregion Process the action request

			#region Prepare the action response

			PushUpdatePresenterActionResponse actionResponse = new PushUpdatePresenterActionResponse(
				requestContext.ActionResponseLocator, actionRequest, null );

			#endregion Prepare the action response

			return actionResponse;
		}

		/// <summary>
		/// Processes an action request to send a letter.
		/// </summary>
		/// <param name="actionRequest">Action request to process.</param>
		/// <param name="requestContext">Action request context.</param>
		/// <returns>Returns the action response which is to be sent back to the initiator of the action request.</returns>
		public SendLetterActionResponse ProcessSendLetter( SendLetterActionRequest actionRequest
			, ActionRequestWithInitiatorContext requestContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( requestContext, "requestContext" );

			IStorehouseTransaction transaction = requestContext.ServiceContext.Transaction;
			Letter receivedLetter = actionRequest.Letter;

			Tally tally = requestContext.InitiatorAccountContext.Tally;
			Account account = requestContext.InitiatorAccountContext.Account;

			#region Process the action request

			// Verify the letter's signature.
			if( !receivedLetter.IsTransient )
				receivedLetter.VerifySignature( tally.ProcessingIdentity );

			// Update the letter's local information.
			receivedLetter.Flag = PostFlags.Main.ToString();

			// Save the received letter.
			transaction.AddSheaf( receivedLetter, StorehouseRealms.Posts );

			string downloadFilesFolderName = null;

			// Prepare to download the offered files. The request will be signed when it's updated.
			if( ListUtils.HasContent( receivedLetter.AttachedFiles ) )
			{
				if( tally.Approval < PeerDocumentsConsts.ReceiveFilesTallyApprovalMin )
				{
					string message = string.Format( "Can't receive files from the tally because its approval (\"{0}\")"
						+ " is lower than \"{1}\".", tally.Approval, PeerDocumentsConsts.ReceiveFilesTallyApprovalMin );

					throw new FatalCrossServiceException( message, null
						, new ExceptionDetail( "Action request type", actionRequest.GetType().FullName )
						, new ExceptionDetail( "Action request locator", actionRequest.InstanceLocator )
						, new ExceptionDetail( "Tally locator", tally.InstanceLocator ) );
				}

				string rootSaveFolderName = requestContext.ServiceContext.ProcessorConfiguration.RootSaveFolderName;
				AttachedFilesUtils.ValidateFileNames( receivedLetter.AttachedFiles, rootSaveFolderName
					, receivedLetter.InstanceLocator, receivedLetter.CreationTime );
				AttachedFilesUtils.ValidateAvailableSpace( receivedLetter.AttachedFiles, rootSaveFolderName );

				// Create an action request to download the offered files.
				DownloadFilesActionRequest downloadRequest = new DownloadFilesActionRequest( CryptoUtils.GetLocator()
					, actionRequest.MacAlg, actionRequest.AsymmSigningPaddingAlg, actionRequest.SignatureHashSerializer
					, account.LocalUpnFingerprint, account.RemoteUpn, tally.ActionRequestValidity, 0
					, receivedLetter.InstanceLocator, receivedLetter.CreationTime, receivedLetter.AttachedFiles );

				if( (account.RemoteUpn.UpnAlg.SteganographicAlg as MacedSteganographicAlg != null)
					&& (account.RemoteUpn.UpnAlg.SteganographicAlg as MacedSteganographicAlg != null) )
				{
					downloadRequest.ServiceMustEncryptResponsesInFrames = false;
				}

				// Create the action progress.
				DownloadFilesActionClientProgress downloadProgress = new DownloadFilesActionClientProgress(
					CryptoUtils.GetLocator(), downloadRequest.InitiatorUpnFingerprint, account.RemoteUpn.Fingerprint
					, receivedLetter.InstanceLocator, downloadRequest.GetType().FullName, downloadRequest.InstanceLocator
					, downloadRequest.DocumentExpirationTime, downloadRequest.ClientProgressIsPermanent );

				transaction.AddSheaf( downloadProgress, StorehouseRealms.ActionProgresses );
				transaction.AddSheaf( downloadRequest, StorehouseRealms.ActionRequestsToSend );

				// Queue a notification for the action client manager that an action request was added to the storehouse.
				requestContext.ServiceContext.NotifyActionClientManagerAboutRequests.Add( downloadRequest.InstanceLocator );

				downloadFilesFolderName = AttachedFilesUtils.GetSaveFolderName( rootSaveFolderName
					, receivedLetter.InstanceLocator, receivedLetter.CreationTime );
			}

			#endregion Process the action request

			string secondaryProcessingOutput;
			SecondaryProcessingState secondaryProcessingState = CreateSecondaryProcessingStateForLetter(
				account.RemotePeerMayRunLocalCommandLines, tally.Approval, receivedLetter.Message, downloadFilesFolderName
				, out secondaryProcessingOutput );

			return new SendLetterActionResponse( requestContext.ActionResponseLocator, actionRequest, secondaryProcessingState
				, secondaryProcessingOutput );
		}

		/// <summary>
		/// Processes an action request to download files.
		/// </summary>
		/// <param name="actionRequest">Action request to process.</param>
		/// <param name="requestContext">Action request context.</param>
		/// <returns>Returns the action response which is to be sent back to the initiator of the action request. If this is
		/// null, the service must end processing the current action request and must return a
		/// <see cref="LastFrameActionResponse"/> action response.</returns>
		public DownloadFilesActionResponse ProcessDownloadFiles( DownloadFilesActionRequest actionRequest
			, ActionRequestWithInitiatorContext requestContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( requestContext, "requestContext" );

			try
			{
				if( requestContext.ServiceContext.IsFirstFrame )
					VerifyIfDownloadFilesIsAllowed( actionRequest, requestContext );

				AttachedFile attachedFile = AttachedFilesUtils.GetAttachedFileToSend( actionRequest.AttachedFiles );
				if( attachedFile == null )
				{
					// This is the last frame of the current multiframe action response.

					ActionServiceUtils.UpdateLastProcessingAttemptOfAttachedFiles(
						requestContext.ServiceContext.StorehouseProvider, actionRequest, actionRequest.AccessGranterLocator
						, null, null );

					return null;
				}

				Int64 fromPosition = attachedFile.CurrentSize;

				// Check if the file has changed.
				if( AttachedFilesUtils.CheckIfChangedOriginalFile( actionRequest.AttachedFiles, attachedFile ) )
				{
					throw new FatalCrossServiceException( "The file can't be transferred because it has been changed or deleted."
						, null, new ExceptionDetail( "File name", attachedFile.AbsoluteFileName ) );
				}

				ActionServiceConfiguration serviceConfiguration
					= requestContext.ServiceContext.ProcessorConfiguration.ActionService;

				int partSize = EntropyPool.GetInt( serviceConfiguration.DownloadFilePartSizeMin
					, serviceConfiguration.DownloadFilePartSizeMax );
				byte[] filePart;

				try
				{
					filePart = FileUtils.Read( attachedFile.AbsoluteFileName, fromPosition, partSize );
				}
				catch( IOException ex )
				{
					// Ensure that the client sees the file access exceptions. The exception should not be fatal because the
					// client should not stop downloading a large file simply because an application has temporarily locked the
					// file.
					throw new RetryCrossServiceException( ex.Message );
				}

				// Fill the response size up to the usual random size in order to hide the end of the file.
				byte[] disalignmentData = EntropyPool.GetFillBytes( partSize - filePart.Length );

				DownloadFilesActionResponse response = new DownloadFilesActionResponse( CryptoUtils.GetLocator(), actionRequest
					, null, attachedFile, fromPosition, filePart, disalignmentData );

				// The progress of the file transfer is kept inside the action request.
				attachedFile.CurrentSize += filePart.Length;

				return response;
			}
			catch( Exception ex )
			{
				ActionServiceUtils.UpdateLastProcessingAttemptOfAttachedFiles( requestContext.ServiceContext.StorehouseProvider
					, actionRequest, actionRequest.AccessGranterLocator, ex, null );

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		#endregion Public methods
	}
}
