////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Net.Sockets;
using Vela.ActionProcessor;
using Vela.Kernel;
using Vela.Networking;

namespace Vela.ActionService
{
	/// <summary>
	/// The global traffic filter for the action requests received by an action service.
	/// </summary>
	/// <remarks>
	/// <para>The traffic filter is used to limit the data traffic coming from various IP addresses.</para>
	/// 
	/// <para>For unknown peers, only the number of the received action requests matters.</para>
	/// 
	/// <para>For known peers, only the total size of the received action requests matters.</para>
	/// </remarks>
	public static class TrafficFilter
	{
		#region Private static fields

		/// <summary>
		/// The synchronization object for the <see cref="Records"/> property. READ-ONLY.
		/// </summary>
		private static readonly object RecordsLock = new object();

		#endregion Private static fields

		#region Private static properties

		private static readonly TrafficFilterRecords _Records = new TrafficFilterRecords();
		/// <summary>
		/// The global action traffic filter records. The backing field is initialized by default. This may be empty, but not
		/// null.
		/// </summary>
		private static TrafficFilterRecords Records
		{
			get { return _Records; }
		}

		#endregion Private static properties

		#region Public static methods

		/// <summary>
		/// Accepts the incoming action request from the specified socket. If a traffic filter record with the IP address of the
		/// initiator doesn't exist, one is added to the list; if the list already contains the maximum allowed number of items,
		/// (approximately) half of the existing items are removed from the list before the specified item is added. If the
		/// record with the specified IP address has reached its processing limit, an exception is thrown. Notifies property
		/// observers. Notifies list observers.
		/// </summary>
		/// <param name="serviceSocket">Service socket from which to accept the request.</param>
		/// <param name="serviceConfiguration">Action service configuration.</param>
		/// <param name="requestSocket">Is set with the socket of the accepted request.</param>
		/// <returns>Returns the IP address of the initiator of the action request.</returns>
		/// <remarks>
		/// <para>The <see cref="Records"/> property is thread safe within this method.</para>
		/// </remarks>
		public static byte[] AcceptRequest( Socket serviceSocket, ActionServiceConfiguration serviceConfiguration
			, out Socket requestSocket )
		{
			ValidationUtils.ObjectHasContent( serviceSocket, "serviceSocket" );
			ValidationUtils.ObjectHasContent( serviceConfiguration, "serviceConfiguration" );

			// A connection is available, so accept it.
			requestSocket = serviceSocket.Accept();

			// Get the initiator's IP address.
			UInt16 remotePort;
			byte[] initiatorIpAddress = NetworkingUtils.GetRemoteIpAddress( requestSocket, out remotePort );

			lock( RecordsLock )
			{
				TrafficFilterConfiguration filterConfiguration = serviceConfiguration.TrafficFilterConfiguration;
				TrafficFilterRecord record = Records[ initiatorIpAddress ];

				if( record == null )
				{
					// New IP address. Add a record to the list.

					record = new TrafficFilterRecord( initiatorIpAddress );
					record.IncomingRequestsCount++;

					Records.AddOrUpdateWithLimit( record, filterConfiguration.RecordsCountMax );

					return initiatorIpAddress;
				}

				// Reset the ban of the IP address.
				DateTime now = DateTime.UtcNow;
				if( now > record.LastResetTime.AddHours( filterConfiguration.RecordResetTimespan ) ) record.Reset();

				record.IncomingRequestsCount++;

				#region Check the processing limit

				string message = null;

				if( record.IsKnownPeer )
				{
					if( record.ProcessedDataSize > filterConfiguration.KnownPeerProcessableSizeMax )
					{
						message = string.Format( "The known peer from the IP address \"{0}\" has reached its data size limit"
								+ " of {1} bytes."
							, NetworkingUtils.GetTextualIpAddress( initiatorIpAddress )
							, filterConfiguration.KnownPeerProcessableSizeMax );
					}
				}
				else
				{
					if( record.IncomingRequestsCount > filterConfiguration.AnonymousPeerIncomingRequestsCountMax )
					{
						message = string.Format( "The anonymous peer from the IP address \"{0}\" has reached its incoming"
								+ " requests limit of {1}. The remote peer may try to detect if you are using this program."
								+ " If this represents a serious threat to you, you should use the program in offline mode."
							, NetworkingUtils.GetTextualIpAddress( initiatorIpAddress )
							, filterConfiguration.AnonymousPeerIncomingRequestsCountMax );
					}
				}

				if( message != null )
				{
					// Since the exception is thrown before the action request is loaded, it's not possible to send back an
					// action response, so there is no need for the exception type to be "RetryCrossServiceException".
					ValidationException exception = new ValidationException( message );

					// Only log one time.
					if( record.WasLogged )
					{
						exception.DoLog = false;
						exception.ShowToUser = false;
					}
					else
					{
						record.WasLogged = true;
					}

					throw exception;
				}

				#endregion Check the processing limit
			}

			return initiatorIpAddress;
		}

		/// <summary>
		/// Updates the data size of an existing traffic filter record in the list.
		/// </summary>
		/// <param name="ipAddress">The IP address of the record to update.</param>
		/// <param name="dataSize">The data size to add to the record. This must be at least 1.</param>
		/// <returns>Returns the updated item.</returns>
		/// <remarks>
		/// <para>The <see cref="Records"/> property is thread safe within this method.</para>
		/// </remarks>
		public static TrafficFilterRecord UpdateDataSize( byte[] ipAddress, int dataSize )
		{
			ValidationUtils.ArrayHasContent( ipAddress, "ipAddress" );
			ValidationUtils.Integer( dataSize, 1, int.MaxValue, "dataSize" );

			lock( RecordsLock )
			{
				TrafficFilterRecord record = Records[ ipAddress ];
				ValidationUtils.ObjectHasContent( record, "record" );

				// Don't validate the data size limit. This will be done at the next request.
				record.ProcessedDataSize += (UInt64)dataSize;

				return record;
			}
		}

		/// <summary>
		/// Updates the "known peer" flag of an existing traffic filter record in the list.
		/// </summary>
		/// <param name="ipAddress">The IP address of the record to update.</param>
		/// <param name="isKnownPeer">If this is true then the peer from the specified IP address is known.</param>
		/// <returns>Returns the updated item.</returns>
		/// <remarks>
		/// <para>The <see cref="Records"/> property is thread safe within this method.</para>
		/// </remarks>
		public static TrafficFilterRecord UpdateKnownPeer( byte[] ipAddress, bool isKnownPeer )
		{
			ValidationUtils.ArrayHasContent( ipAddress, "ipAddress" );

			lock( RecordsLock )
			{
				TrafficFilterRecord record = Records[ ipAddress ];
				ValidationUtils.ObjectHasContent( record, "record" );

				record.IsKnownPeer = isKnownPeer;

				return record;
			}
		}

		#endregion Public static methods
	}
}
