////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.ActionService
{
	/// <summary>
	/// Action traffic filter record.
	/// </summary>
	public class TrafficFilterRecord : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected TrafficFilterRecord()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="ipAddress">The IP address.</param>
		public TrafficFilterRecord( byte[] ipAddress )
		{
			IpAddress = ipAddress;
		}

		#endregion Constructors

		#region Public serializable properties

		private byte[] _IpAddress;
		/// <summary>
		/// The IP address. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] IpAddress
		{
			get { return _IpAddress; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "IpAddress" );

				_IpAddress = value;
			}
		}

		private bool _IsKnownPeer;
		/// <summary>
		/// If this is true then the peer (from the associated IP address) is known, else it's anonymous.
		/// </summary>
		[SerializeAttribute]
		public bool IsKnownPeer
		{
			get { return _IsKnownPeer; }
			set { _IsKnownPeer = value; }
		}

		private bool _WasLogged;
		/// <summary>
		/// If this is true then the current IP address (<see cref="IpAddress"/>) has reached its processing limit and an
		/// exception was thrown (and should have been logged by the application).
		/// </summary>
		/// <remarks>
		/// <para>This exists so that an application only logs once when an anonymous peer has reached its processing limit.
		/// </para>
		/// </remarks>
		[SerializeAttribute]
		public bool WasLogged
		{
			get { return _WasLogged; }
			set { _WasLogged = value; }
		}

		private DateTime _LastResetTime = DateTime.UtcNow;
		/// <summary>
		/// The time (in UTC format) when the processing limit was last reset. This is automatically set with the time of the
		/// creation of the current object.
		/// </summary>
		[SerializeAttribute]
		public DateTime LastResetTime
		{
			get { return _LastResetTime; }
			set { _LastResetTime = value; }
		}

		private UInt64 _IncomingRequestsCount;
		/// <summary>
		/// The number of network requests which were accepted. This must be at least 0.
		/// </summary>
		[SerializeAttribute]
		public UInt64 IncomingRequestsCount
		{
			get { return _IncomingRequestsCount; }
			set
			{
				ValidationUtils.Integer( value, 0, UInt64.MaxValue, "IncomingRequestsCount" );

				_IncomingRequestsCount = value;
			}
		}

		private UInt64 _ProcessedDataSize;
		/// <summary>
		/// The size of the processed data. This must be at least 0.
		/// </summary>
		[SerializeAttribute]
		public UInt64 ProcessedDataSize
		{
			get { return _ProcessedDataSize; }
			set
			{
				ValidationUtils.Integer( value, 0, UInt64.MaxValue, "ProcessedDataSize" );

				_ProcessedDataSize = value;
			}
		}

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Resets the current record.
		/// </summary>
		public void Reset()
		{
			WasLogged = false;
			LastResetTime = DateTime.UtcNow;
			IncomingRequestsCount = 0;
			ProcessedDataSize = 0;
		}

		#endregion Public methods
	}
}
