////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.ActionProcessor;

namespace Vela.ApplicationDocuments
{
	/// <summary>
	/// Utilities for working with application documents.
	/// </summary>
	public static class ApplicationDocumentsUtils
	{
		#region Public static methods

		/// <summary>
		/// Gets the textual state of the transfer progress of the files attached to a letter. This may be null or empty.
		/// </summary>
		/// <param name="attachedFilesCount">The number of files attached to the current letter.</param>
		/// <param name="lastProcessingAttemptOfAttachedFiles">Information about the last processing attempt of the associated
		/// action request which requests the transfer of the files attached to the letter. This may be null.</param>
		/// <param name="attachedFilesProgress">The transfer progress of the files attached to the letter.</param>
		/// <returns>Returns the textual state of the transfer progress of the files attached to a letter</returns>
		public static string GetAttachedFilesState( int attachedFilesCount
			, ActionLastProcessingAttempt lastProcessingAttemptOfAttachedFiles, double attachedFilesProgress )
		{
			if( attachedFilesCount <= 0 )
				return null;

			if( (lastProcessingAttemptOfAttachedFiles != null)
				&& (lastProcessingAttemptOfAttachedFiles.ProcessingState == ActionProcessingStates.Error) )
			{
				return lastProcessingAttemptOfAttachedFiles.ProcessingState.ToString();
			}

			// If the file transfer is complete, show an indicator. Note that both the "Sent" and "Received" states are supported
			// because the point of view can be the completion of either the transfer or of the download action request.
			if( (lastProcessingAttemptOfAttachedFiles != null)
				&& (
					(lastProcessingAttemptOfAttachedFiles.ProcessingState == ActionProcessingStates.Sent)
					|| (lastProcessingAttemptOfAttachedFiles.ProcessingState == ActionProcessingStates.Received)
				) )
			{
				return string.Format( ApplicationDocumentsConsts.AttachedFilesTransferredStateTemplate
					, attachedFilesCount );
			}

			if( attachedFilesProgress > 0 )
			{
				return (attachedFilesProgress < ActionProcessorConsts.DownloadFilesProgressMax)
					? string.Format( ApplicationDocumentsConsts.AttachedFilesTransferringStateTemplate
						, attachedFilesProgress )
					: string.Format( ApplicationDocumentsConsts.AttachedFilesTransferredStateTemplate
						, attachedFilesCount );
			}

			if( (lastProcessingAttemptOfAttachedFiles != null)
				&& (
					(lastProcessingAttemptOfAttachedFiles.ProcessingState == ActionProcessingStates.Pending)
					|| (lastProcessingAttemptOfAttachedFiles.ProcessingState == ActionProcessingStates.Retry)
				) )
			{
				return lastProcessingAttemptOfAttachedFiles.ProcessingState.ToString();
			}

			return attachedFilesCount.ToString();
		}

		#endregion Public static methods
	}
}
