////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.ActionProcessor;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.PeerDocuments;
using Tuple = Vela.Kernel.Tuple;

namespace Vela.ApplicationDocuments
{
	/// <summary>
	/// A speckle for accounts.
	/// </summary>
	/// <remarks>
	/// <para>This may be either the speckle of a real account (<see cref="Account"/>) or of an account creation request
	/// (<see cref="CreateAccountAnonymousActionRequest"/>).</para>
	/// </remarks>
	public class AccountOrRequestSpeckle : PossibleActionRequestSpeckle, IRecipientUpnFingerprintContainer
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AccountOrRequestSpeckle()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the referred sheaf. This must be unique in the storehouse.
		/// </param>
		/// <param name="sheafTypeName">The type name of the stored sheaf.</param>
		public AccountOrRequestSpeckle( Locator storehouseKey, string sheafTypeName )
			: base( storehouseKey, sheafTypeName )
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the referred sheaf. This must be unique in the storehouse.
		/// </param>
		/// <param name="sheafTypeName">The type name of the stored sheaf.</param>
		/// <param name="loadedSpeckleFilters">The speckle filters of the speckle, loaded from the storehouse. This may
		/// be null or empty.</param>
		public AccountOrRequestSpeckle( Locator storehouseKey, string sheafTypeName, SpeckleFilters loadedSpeckleFilters )
			: base( storehouseKey, sheafTypeName, loadedSpeckleFilters )
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="sheaf">The sheaf to refer.</param>
		/// <remarks>
		/// <para>All the properties declared in this class are automatically set in here.</para>
		/// </remarks>
		public AccountOrRequestSpeckle( ISheaf sheaf )
			: base( sheaf )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The UPN fingerprint of the local peer. READ-ONLY.
		/// </summary>
		public Locator LocalUpnFingerprint
		{
			get { return GetLoadedSpeckleFilter<Locator>( Account.LocalUpnFingerprintFilterIdentifier ); }
		}

		/// <summary>
		/// The UPN fingerprint of the remote peer. READ-ONLY.
		/// </summary>
		public Locator RemoteUpnFingerprint
		{
			get { return GetLoadedSpeckleFilter<Locator>( Account.RemoteUpnFingerprintFilterIdentifier ); }
		}

		/// <summary>
		/// The textual form of the account alias. This may be null or empty. READ-ONLY.
		/// </summary>
		public string AliasTextualForm
		{
			get { return TryGetLoadedSpeckleFilter<string>( Account.AliasFilterIdentifier ); }
		}

		/// <summary>
		/// If this is true then the remote peer may run command lines on the local computer (by sending letters whose messages
		/// are used as command lines). READ-ONLY.
		/// </summary>
		public bool RemotePeerMayRunLocalCommandLines
		{
			get { return TryGetLoadedSpeckleFilter<bool>( Account.RemotePeerMayRunLocalCommandLinesFilterIdentifier, false ); }
		}

		private string _LocalEntityName;
		/// <summary>
		/// The local entity name. This is null after the speckle is loaded, and must be set from outside.
		/// </summary>
		public string LocalEntityName
		{
			get { return _LocalEntityName; }
			set
			{
				ValidationUtils.TextHasContent( value, "LocalEntityName" );

				_LocalEntityName = value;
			}
		}

		private string _RemoteEntityName;
		/// <summary>
		/// The remote entity name. This is null after the speckle is loaded, and must be set from outside.
		/// </summary>
		public string RemoteEntityName
		{
			get { return _RemoteEntityName; }
			set
			{
				ValidationUtils.TextHasContent( value, "RemoteEntityName" );

				_RemoteEntityName = value;
			}
		}

		/// <summary>
		/// The UPN fingerprint of the remote peer of the account, or of the recipient of the account creation request.
		/// READ-ONLY.
		/// </summary>
		public Locator RecipientUpnFingerprint
		{
			get
			{
				string compositeIdentifier = RefersAccountRequest()
					? ActionRequest.RecipientUpnFingerprintFilterIdentifier
					: Account.RemoteUpnFingerprintFilterIdentifier;

				return GetLoadedSpeckleFilter<Locator>( compositeIdentifier );
			}
		}

		private TallyApprovals _Approval = TallyApprovals.Black;
		/// <summary>
		/// The approval of the tally. This must be set from outside.
		/// </summary>
		public TallyApprovals Approval
		{
			get { return _Approval; }
			set { _Approval = value; }
		}

		/// <summary>
		/// The sort key. READ-ONLY.
		/// </summary>
		public Tuple SortKey
		{
			get { return new Tuple( LocalEntityName, RemoteEntityName, CreationTime, StorehouseKey ); }
		}

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Checks if the current speckle refers a real account (<see cref="Vela.PeerDocuments.Account"/>), rather than a request
		/// to create an account (<see cref="Vela.ActionProcessor.CreateAccountAnonymousActionRequest"/>).
		/// </summary>
		/// <returns>Returns true if the current speckle refers a real account.</returns>
		public bool RefersRealAccount()
		{
			return PortableTypes.GetGlobal( SheafTypeName, typeof( Account ), false ) != null;
		}

		/// <summary>
		/// Checks if the current speckle refers an account creation request
		/// (<see cref="Vela.ActionProcessor.CreateAccountAnonymousActionRequest"/>), rather than a real account
		/// (<see cref="Vela.PeerDocuments.Account"/>)
		/// </summary>
		/// <returns>Returns true if the current speckle refers an account creation request.</returns>
		public bool RefersAccountRequest()
		{
			return PortableTypes.GetGlobal( SheafTypeName, typeof( CreateAccountAnonymousActionRequest ), false ) != null;
		}

		#endregion Public methods
	}
}
