////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.PeerDocuments;

namespace Vela.ApplicationDocuments
{
	/// <summary>
	/// List of speckles for accounts (<see cref="AccountOrRequestSpeckle"/>).
	/// </summary>
	/// <remarks>
	/// <para>An item of this list may be either the speckle of a real account (<see cref="Vela.PeerDocuments.Account"/>)
	/// or of an account creation request (<see cref="Vela.ActionProcessor.CreateAccountAnonymousActionRequest"/>).</para>
	/// </remarks>
	public class AccountOrRequestSpeckles : KernelSortedList<Tuple, AccountOrRequestSpeckle>
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public AccountOrRequestSpeckles()
			: base( TupleComparer.Instance )
		{
		}

		#endregion Constructors

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para>The <see cref="AccountOrRequestSpeckle.SortKey"/> property is used as key.</para>
		/// </remarks>
		public override Tuple GetKey( AccountOrRequestSpeckle item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.SortKey;
		}

		/// <summary>
		/// Finds the speckle with the specified storehouse key.
		/// </summary>
		/// <param name="storehouseKey">Storehouse key to find.</param>
		/// <returns>Returns the found speckle, or null if none is found.</returns>
		/// <remarks>
		/// <para>The storehouse key is not the item key.</para>
		/// </remarks>
		public AccountOrRequestSpeckle FindByStorehouseKey( Locator storehouseKey )
		{
			ValidationUtils.ObjectHasContent( storehouseKey, "storehouseKey" );

			foreach( AccountOrRequestSpeckle speckle in this )
			{
				if( speckle.StorehouseKey.Equals( storehouseKey ) )
					return speckle;
			}

			return null;
		}

		/// <summary>
		/// Finds the speckle with the specified local and remote UPN fingerprints.
		/// </summary>
		/// <param name="localUpnFingerprint">The local UPN fingerprint of the account speckle to find.</param>
		/// <param name="remoteUpnFingerprint">The remote UPN fingerprint of the account speckle to find.</param>
		/// <returns>Returns the found speckle, or null if none is found.</returns>
		public AccountOrRequestSpeckle FindByUpnFingerprints( Locator localUpnFingerprint, Locator remoteUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( localUpnFingerprint, "localUpnFingerprint" );
			ValidationUtils.ObjectHasContent( remoteUpnFingerprint, "remoteUpnFingerprint" );

			foreach( AccountOrRequestSpeckle speckle in this )
			{
				if( speckle.LocalUpnFingerprint.Equals( localUpnFingerprint )
					&& speckle.RemoteUpnFingerprint.Equals( remoteUpnFingerprint ) )
				{
					return speckle;
				}
			}

			return null;
		}

		/// <summary>
		/// Finds the first speckle with the specified account alias.
		/// </summary>
		/// <param name="accountAlias">Account alias to find. This may be null.</param>
		/// <returns>Returns the storehouse key of the account speckle which is uniquely identified by the specified alias.
		/// This is null if the account alias to find is null, or if no account speckle can be uniquely identified.</returns>
		public Locator FindStorehouseKeyOfAccountUniquelyIdentifiedByAlias( AccountAlias accountAlias )
		{
			if( AccountAlias.NoContent( accountAlias ) )
				return null;

			Locator storehouseKey = null;
			int count = 0;

			foreach( AccountOrRequestSpeckle speckle in this )
			{
				if( TextUtils.Equals( speckle.AliasTextualForm, accountAlias.TextualForm ) )
				{
					storehouseKey = speckle.StorehouseKey;
					count++;
				}
			}

			return (count == 1) ? storehouseKey : null;
		}

		#endregion Public methods
	}
}
