////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using Vela.ActionProcessor;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.StarkDocuments;

namespace Vela.ApplicationDocuments
{
	/// <summary>
	/// A speckle for letters.
	/// </summary>
	public class LetterSpeckle : PossibleActionRequestSpeckle, IRecipientUpnFingerprintContainer
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected LetterSpeckle()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the referred sheaf. This must be unique in the storehouse.
		/// </param>
		/// <param name="sheafTypeName">The type name of the stored sheaf.</param>
		public LetterSpeckle( Locator storehouseKey, string sheafTypeName )
			: base( storehouseKey, sheafTypeName )
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the referred sheaf. This must be unique in the storehouse.
		/// </param>
		/// <param name="sheafTypeName">The type name of the stored sheaf.</param>
		/// <param name="loadedSpeckleFilters">The speckle filters of the speckle, loaded from the storehouse. This may
		/// be null or empty.</param>
		public LetterSpeckle( Locator storehouseKey, string sheafTypeName, SpeckleFilters loadedSpeckleFilters )
			: base( storehouseKey, sheafTypeName, loadedSpeckleFilters )
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="sheaf">The sheaf to refer.</param>
		/// <remarks>
		/// <para>All the properties declared in this class are automatically set in here.</para>
		/// </remarks>
		public LetterSpeckle( ISheaf sheaf )
			: base( sheaf )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// Information about the last processing attempt of the referred action request. This is null after the speckle is
		/// loaded, and must be set from outside. Notifies property observers.
		/// </summary>
		public override ActionLastProcessingAttempt LastProcessingAttempt
		{
			get
			{
				return base.LastProcessingAttempt;
			}
			set
			{
				base.LastProcessingAttempt = value;

				DetermineLocalPeerDirectionFromLastProcessingAttempt();
			}
		}

		private bool _TimeConsumingLoadingIsComplete;
		/// <summary>
		/// If this is true then the current letter speckle has been loaded with all the information which takes a significant
		/// amount of time load; this information is normally loaded with a timer or asynchronously.
		/// </summary>
		public bool TimeConsumingLoadingIsComplete
		{
			get { return _TimeConsumingLoadingIsComplete; }
			set { _TimeConsumingLoadingIsComplete = value; }
		}

		private LocalPeerDirections _LocalPeerDirection;
		/// <summary>
		/// The direction of the local peer. This is unknown after the speckle is loaded, and must be set from outside.
		/// Notifies property observers.
		/// </summary>
		public LocalPeerDirections LocalPeerDirection
		{
			get { return _LocalPeerDirection; }
			set
			{
				_LocalPeerDirection = value;

				// Notify property observers.
				OnPropertyChanged( "LocalPeerDirection" );
				OnPropertyChanged( "LocalUpnFingerprint" );
				OnPropertyChanged( "RemoteUpnFingerprint" );
				OnPropertyChanged( "MayExploreAttachedFiles" );
				OnPropertyChanged( "MayPlayAttachedMediaFiles" );
				OnPropertyChanged( "MayExportLetter" );
			}
		}

		/// <summary>
		/// The UPN fingerprint of the initiator of the letter. READ-ONLY.
		/// </summary>
		public Locator InitiatorUpnFingerprint
		{
			get { return GetLoadedSpeckleFilter<Locator>( Post.InitiatorUpnFingerprintFilterIdentifier ); }
		}

		private string _InitiatorEntityName;
		/// <summary>
		/// The initiator entity name. This is null after the speckle is loaded, and must be set from outside.
		/// </summary>
		public string InitiatorEntityName
		{
			get { return _InitiatorEntityName; }
			set
			{
				ValidationUtils.TextHasContent( value, "InitiatorEntityName" );

				_InitiatorEntityName = value;
			}
		}

		/// <summary>
		/// The UPN fingerprint of the recipient of the letter. READ-ONLY.
		/// </summary>
		public Locator RecipientUpnFingerprint
		{
			get { return GetLoadedSpeckleFilter<Locator>( Letter.RecipientUpnFingerprintFilterIdentifier ); }
		}

		private string _RecipientEntityName;
		/// <summary>
		/// The recipient entity name. This is null after the speckle is loaded, and must be set from outside.
		/// </summary>
		public string RecipientEntityName
		{
			get { return _RecipientEntityName; }
			set
			{
				ValidationUtils.TextHasContent( value, "RecipientEntityName" );

				_RecipientEntityName = value;
			}
		}

		/// <summary>
		/// The local UPN fingerprint of the letter. This may be null. READ-ONLY.
		/// </summary>
		public Locator LocalUpnFingerprint
		{
			get
			{
				if( LastProcessingAttempt == null )
					return null;

				string localUpnFingerprintFilterIdentifier = (LocalPeerDirection == LocalPeerDirections.Initiator)
					? Post.InitiatorUpnFingerprintFilterIdentifier : Letter.RecipientUpnFingerprintFilterIdentifier;

				return GetLoadedSpeckleFilter<Locator>( localUpnFingerprintFilterIdentifier );
			}
		}

		/// <summary>
		/// The remote UPN fingerprint of the letter. This may be null. READ-ONLY.
		/// </summary>
		public Locator RemoteUpnFingerprint
		{
			get
			{
				if( LastProcessingAttempt == null )
					return null;

				string remoteUpnFingerprintFilterIdentifier = (LocalPeerDirection == LocalPeerDirections.Initiator)
					? Letter.RecipientUpnFingerprintFilterIdentifier : Post.InitiatorUpnFingerprintFilterIdentifier;

				return GetLoadedSpeckleFilter<Locator>( remoteUpnFingerprintFilterIdentifier );
			}
		}

		/// <summary>
		/// Summary of the message of the letter. This may be null or empty. READ-ONLY.
		/// </summary>
		public string MessageSummary
		{
			get { return TryGetLoadedSpeckleFilter<string>( Post.MessageSummaryFilterIdentifier, null ); }
		}

		/// <summary>
		/// Emoticon symbol. This may be null or empty. READ-ONLY.
		/// </summary>
		public string EmoticonSymbol
		{
			get { return TryGetLoadedSpeckleFilter<string>( Post.EmoticonSymbolFilterIdentifier, null ); }
		}

		private List<EmoticonSymbolDescriptor> _EmoticonDescriptors;
		/// <summary>
		/// The emoticon descriptors extracted from the <see cref="MessageSummary"/>. This may be null or empty. This is null
		/// after the speckle is loaded, and must be set from outside.
		/// </summary>
		public List<EmoticonSymbolDescriptor> EmoticonDescriptors
		{
			get { return _EmoticonDescriptors; }
			set { _EmoticonDescriptors = value; }
		}

		/// <summary>
		/// The flag of the post. READ-ONLY.
		/// </summary>
		public string Flag
		{
			get { return GetLoadedSpeckleFilter<string>( Post.FlagFilterIdentifier ); }
		}

		/// <summary>
		/// The number of files attached to the current letter. READ-ONLY.
		/// </summary>
		public int AttachedFilesCount
		{
			get { return TryGetLoadedSpeckleFilter<int>( Letter.AttachedFilesCountFilterIdentifier, 0 ); }
		}

		/// <summary>
		/// If this is true then the current letter has files attached (, files which the remote peer may download). READ-ONLY.
		/// </summary>
		public bool HasAttachedFiles
		{
			get { return AttachedFilesCount > 0; }
		}

		/// <summary>
		/// If this is true then some of the attached files are media files. READ-ONLY.
		/// </summary>
		public bool HasAttachedMediaFiles
		{
			get { return TryGetLoadedSpeckleFilter<bool>( Letter.HasAttachedMediaFilesFilterIdentifier, false ); }
		}

		/// <summary>
		/// If this is true then the current letter has an image preview, that is, it has an attached image file. READ-ONLY.
		/// </summary>
		public bool HasImagePreview
		{
			get { return TryGetLoadedSpeckleFilter<bool>( Letter.HasImagePreviewFilterIdentifier, false ); }
		}

		/// <summary>
		/// If this is true then the current letter has an image preview (that is, it has an attached image file), or an emoticon
		/// descriptor. READ-ONLY.
		/// </summary>
		public bool HasImagePreviewOrEmoticonDescriptor
		{
			get { return HasImagePreview || ListUtils.HasContent( EmoticonDescriptors ); }
		}

		/// <summary>
		/// This is true if the attached files may be explored. READ-ONLY.
		/// </summary>
		public bool MayExploreAttachedFiles
		{
			get { return HasAttachedFiles && (LocalPeerDirection != LocalPeerDirections.RecipientImported); }
		}

		/// <summary>
		/// This is true if the attached media files may be played (with a media player). READ-ONLY.
		/// </summary>
		public bool MayPlayAttachedMediaFiles
		{
			get
			{
				if( LastProcessingAttempt == null )
					return false;

				return HasAttachedMediaFiles
					&& (
						(LocalPeerDirection == LocalPeerDirections.Initiator)
						|| (
							(LocalPeerDirection == LocalPeerDirections.Recipient)
							&& (LastProcessingAttemptOfAttachedFiles != null)
							// The recipient of the letter is the sender of the action request to download the attached files.
							&& (LastProcessingAttemptOfAttachedFiles.ProcessingState == ActionProcessingStates.Sent)
						)
					);
			}
		}

		private ActionLastProcessingAttempt _LastProcessingAttemptOfAttachedFiles;
		/// <summary>
		/// Information about the last processing attempt of the associated action request which requests the transfer of the
		/// files attached to the letter. This is null after the speckle is loaded, and must be set from outside. This may be /
		/// remain null (especially if <see cref="HasAttachedFiles"/> is false). Notifies property observers.
		/// </summary>
		public ActionLastProcessingAttempt LastProcessingAttemptOfAttachedFiles
		{
			get
			{
				return _LastProcessingAttemptOfAttachedFiles;
			}
			set
			{
				_LastProcessingAttemptOfAttachedFiles = value;

				// Notify property observers.
				OnPropertyChanged( "LastProcessingAttemptOfAttachedFiles" );
				OnPropertyChanged( "HasNextProcessingAttemptOfAttachedFiles" );
				OnPropertyChanged( "AttachedFilesStateText" );
				OnPropertyChanged( "MayPlayAttachedMediaFiles" );
			}
		}

		/// <summary>
		/// If this is true then a new attempt to transfer the files attached to the letter will be made.
		/// </summary>
		public bool HasNextProcessingAttemptOfAttachedFiles
		{
			get
			{
				return (LastProcessingAttemptOfAttachedFiles != null)
					? (LastProcessingAttemptOfAttachedFiles.ProcessingState == ActionProcessingStates.Retry) : false;
			}
		}

		/// <summary>
		/// If this is true then the current letter has files that the remote peer may download, and has a transfer progress
		/// which is above 0. READ-ONLY.
		/// </summary>
		public bool HasAttachedFilesProgress
		{
			get { return HasAttachedFiles && (AttachedFilesProgress > 0); }
		}

		private double _AttachedFilesProgress;
		/// <summary>
		/// The transfer progress of the files attached to the letter. This is 0 after the speckle is loaded, and must be set
		/// from outside. Notifies property observers.
		/// </summary>
		public double AttachedFilesProgress
		{
			get { return _AttachedFilesProgress; }
			set
			{
				_AttachedFilesProgress = value;

				// Notify property observers.
				OnPropertyChanged( "AttachedFilesProgress" );
				OnPropertyChanged( "HasAttachedFilesProgress" );
				OnPropertyChanged( "AttachedFilesStateText" );
			}
		}

		/// <summary>
		/// The textual state of the transfer progress of the files attached to the letter. This may be null or empty. READ-ONLY.
		/// </summary>
		public string AttachedFilesStateText
		{
			get
			{
				return ApplicationDocumentsUtils.GetAttachedFilesState( AttachedFilesCount, LastProcessingAttemptOfAttachedFiles
					, AttachedFilesProgress );
			}
		}

		/// <summary>
		/// If this is true then the current letter is transient. READ-ONLY.
		/// </summary>
		public bool IsTransient
		{
			get { return TryGetLoadedSpeckleFilter<bool>( Post.IsTransientFilterIdentifier, false ); }
		}

		/// <summary>
		/// This is true if the letter may be exported. READ-ONLY.
		/// </summary>
		public bool MayExportLetter
		{
			get { return (LocalPeerDirection == LocalPeerDirections.Initiator) && !HasAttachedFiles; }
		}

		private DateTime _CarrierActionRequestDocumentExpirationTime = DateTime.MinValue;
		/// <summary>
		/// The document expiration time (in UTC format) of the action request which contains the document referred by the
		/// current speckle. This is <see cref="DateTime.MinValue"/> after the speckle is loaded, and must be set from outside.
		/// </summary>
		public DateTime CarrierActionRequestDocumentExpirationTime
		{
			get { return _CarrierActionRequestDocumentExpirationTime; }
			set { _CarrierActionRequestDocumentExpirationTime = value; }
		}

		private DateTime _CarrierActionRequestDocumentExpirationTimeOfAttachedFiles = DateTime.MinValue;
		/// <summary>
		/// The document expiration time (in UTC format) of the action request which requests the transfer of the files attached
		/// to the letter. This is <see cref="DateTime.MinValue"/> after the speckle is loaded, and must be set from outside.
		/// </summary>
		public DateTime CarrierActionRequestDocumentExpirationTimeOfAttachedFiles
		{
			get { return _CarrierActionRequestDocumentExpirationTimeOfAttachedFiles; }
			set { _CarrierActionRequestDocumentExpirationTimeOfAttachedFiles = value; }
		}

		private string _AccountAliasTextualForm;
		/// <summary>
		/// The textual form of the account alias. This may be null or empty. This is null after the speckle is loaded, and must
		/// be set from outside. Notifies property observers.
		/// </summary>
		public string AccountAliasTextualForm
		{
			get { return _AccountAliasTextualForm; }
			set
			{
				_AccountAliasTextualForm = value;

				// Notify property observers.
				OnPropertyChanged( "AccountAliasTextualForm" );
			}
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Determines the direction of the local peer from <see cref="LastProcessingAttempt"/>.
		/// </summary>
		/// <remarks>
		/// <para>This can make the difference between <see cref="LocalPeerDirections.Recipient"/> and
		/// <see cref="LocalPeerDirections.RecipientImported"/>.</para>
		/// </remarks>
		private void DetermineLocalPeerDirectionFromLastProcessingAttempt()
		{
			ValidationUtils.ObjectHasContent( LastProcessingAttempt, "LastProcessingAttempt" );

			LocalPeerDirections localPeerDirection = ActionProcessorUtils.GetLocalPeerDirection(
				LastProcessingAttempt.ProcessingState );

			if( (localPeerDirection != LocalPeerDirection) && (localPeerDirection != LocalPeerDirections.RecipientImported) )
			{
				string message = string.Format( "The direction of the local peer determined from the last processing attempt"
					+ " (\"{0}\") is different than the direction of the local peer determined during the loading of the letter"
					+ " speckle (\"{1}\").", localPeerDirection, LocalPeerDirection );
				throw new ValidationException( message );
			}

			LocalPeerDirection = localPeerDirection;
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Checks if the current letter speckle has the specified flag.
		/// </summary>
		/// <param name="flag">Flag to check if the current letter speckle has.</param>
		/// <returns>Returns true if the current letter speckle has the specified flag.</returns>
		public bool HasFlag( string flag )
		{
			return Post.EqualFlags( Flag, flag );
		}

		#endregion Public methods
	}
}
