////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;

namespace Vela.ApplicationDocuments
{
	/// <summary>
	/// List of speckles for letters (<see cref="LetterSpeckle"/>).
	/// </summary>
	/// <remarks>
	/// <para>The items from this list should have unique speckle keys (<see cref="Vela.CoreDocuments.Speckle.StorehouseKey"/>).
	/// </para>
	/// 
	/// <para>This list must be unsorted (but ordered) because the collection of speckles from the storehouse doesn't support
	/// sorting by custom criteria (but only by the underlying database ID), and the letter speckles are paged. Therefore, if
	/// this list were sorted, there would appear inconsistencies in the order of items in a page of letter speckles.</para>
	/// </remarks>
	public class LetterSpeckles : KernelList<LetterSpeckle>
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public LetterSpeckles()
		{
		}

		#endregion Constructors

		#region Public methods

		/// <summary>
		/// Finds the speckle with the specified storehouse key.
		/// </summary>
		/// <param name="storehouseKey">Storehouse key to find.</param>
		/// <returns>Returns the found speckle, or null if none is found.</returns>
		public LetterSpeckle FindByStorehouseKey( Locator storehouseKey )
		{
			ValidationUtils.ObjectHasContent( storehouseKey, "storehouseKey" );

			foreach( LetterSpeckle speckle in this )
			{
				if( speckle.StorehouseKey.Equals( storehouseKey ) )
					return speckle;
			}

			return null;
		}

		#endregion Public methods
	}
}
