////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.ActionProcessor;
using Vela.CoreDocuments;
using Vela.Kernel;

namespace Vela.ApplicationDocuments
{
	/// <summary>
	/// Contract for a speckle which may represent either a non-action request sheaf of an action request.
	/// </summary>
	public abstract class PossibleActionRequestSpeckle : ViewSpeckle
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected PossibleActionRequestSpeckle()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the referred sheaf. This must be unique in the storehouse.
		/// </param>
		/// <param name="sheafTypeName">The type name of the stored sheaf.</param>
		public PossibleActionRequestSpeckle( Locator storehouseKey, string sheafTypeName )
			: base( storehouseKey, sheafTypeName )
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the referred sheaf. This must be unique in the storehouse.
		/// </param>
		/// <param name="sheafTypeName">The type name of the stored sheaf.</param>
		/// <param name="loadedSpeckleFilters">The speckle filters of the speckle, loaded from the storehouse. This may
		/// be null or empty.</param>
		public PossibleActionRequestSpeckle( Locator storehouseKey, string sheafTypeName, SpeckleFilters loadedSpeckleFilters )
			: base( storehouseKey, sheafTypeName, loadedSpeckleFilters )
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="sheaf">The sheaf to refer.</param>
		/// <remarks>
		/// <para>All the properties declared in this class are automatically set in here.</para>
		/// </remarks>
		public PossibleActionRequestSpeckle( ISheaf sheaf )
			: base( sheaf )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		private ActionLastProcessingAttempt _LastProcessingAttempt;
		/// <summary>
		/// Information about the last processing attempt of the referred action request. This is null after the speckle is
		/// loaded, and must be set from outside. Notifies property observers.
		/// </summary>
		public virtual ActionLastProcessingAttempt LastProcessingAttempt
		{
			get
			{
				return _LastProcessingAttempt;
			}
			set
			{
				_LastProcessingAttempt = value;

				// Notify property observers.
				OnPropertyChanged( "LastProcessingAttempt" );
				OnPropertyChanged( "HasNextProcessingAttempt" );
			}
		}

		/// <summary>
		/// If this is true then a new attempt to process the referred action request will be made in the future.
		/// </summary>
		public bool HasNextProcessingAttempt
		{
			get
			{
				return (LastProcessingAttempt != null)
					? (LastProcessingAttempt.ProcessingState == ActionProcessingStates.Retry) : false;
			}
		}

		/// <summary>
		/// The document expiration time (in UTC format). This is <see cref="DateTime.MinValue"/> if
		/// <see cref="Speckle.LoadedSpeckleFilters"/> is null or doesn't contain the
		/// <see cref="ActionRequest.DocumentExpirationTimeFilterIdentifier"/> speckle filter. The current document is valid
		/// before this time; from and after this time, it should not be used in processing. READ-ONLY.
		/// </summary>
		public DateTime ActionRequestDocumentExpirationTime
		{
			get
			{
				return TryGetLoadedSpeckleFilter<DateTime>( ActionRequest.DocumentExpirationTimeFilterIdentifier
					, DateTime.MinValue );
			}
		}

		#endregion Public non-serializable properties
	}
}
