////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.ComponentModel;
using Vela.CoreDocuments;
using Vela.Kernel;

namespace Vela.ApplicationDocuments
{
	/// <summary>
	/// A speckle meant to be shown in the UI.
	/// </summary>
	public class ViewSpeckle : Speckle, INotifyPropertyChanged
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ViewSpeckle()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the referred sheaf. This must be unique in the storehouse.
		/// </param>
		/// <param name="sheafTypeName">The type name of the stored sheaf.</param>
		public ViewSpeckle( Locator storehouseKey, string sheafTypeName )
		{
			StorehouseKey = storehouseKey;
			SheafTypeName = sheafTypeName;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the referred sheaf. This must be unique in the storehouse.</param>
		/// <param name="sheafTypeName">The type name of the stored sheaf.</param>
		/// <param name="loadedSpeckleFilters">The filters of the speckle, loaded from the storehouse. This may be null or
		/// empty.</param>
		public ViewSpeckle( Locator storehouseKey, string sheafTypeName, SpeckleFilters loadedSpeckleFilters )
			: this( storehouseKey, sheafTypeName )
		{
			LoadedSpeckleFilters = loadedSpeckleFilters;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="sheaf">The sheaf to refer.</param>
		/// <remarks>
		/// <para>All the properties declared in this class are automatically set in here.</para>
		/// </remarks>
		public ViewSpeckle( ISheaf sheaf )
		{
			ValidationUtils.ObjectHasContent( sheaf, "sheaf" );

			StorehouseLock = sheaf.StorehouseLock;
			SheafUpdateVersion = sheaf.SheafUpdateVersion;

			StorehouseKey = sheaf.InstanceLocator;
			SheafTypeName = sheaf.GetType().FullName;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private bool _ForceRebind;
		/// <summary>
		/// When this is set (with any value), it notifies property observers that a change has occurred. Notifies property
		/// observers.
		/// </summary>
		/// <remarks>
		/// <para>This can be bound in a UI converter in order to notify the converter of a change which would not otherwise be
		/// visible to the converter.</para>
		/// </remarks>
		public bool ForceRebind
		{
			get { return _ForceRebind; }
			set
			{
				// Change the value every time in order to ensure that nobody tries to optimize the bound value, as they have
				// done with the items from a list when the object reference doesn't change (and which therefore fail to accept
				// change notifications).
				_ForceRebind = !_ForceRebind;

				// Notify property observers.
				OnPropertyChanged( "ForceRebind" );
			}
		}

		#endregion Public non-serializable properties

		#region Events

		/// <summary>
		/// Event raised when a property of the current class changes. This may be null.
		/// </summary>
		public event PropertyChangedEventHandler PropertyChanged;

		#endregion Events

		#region Private methods

		/// <summary>
		/// Notifies property observers when a property was changed.
		/// </summary>
		/// <param name="propertyName">The name of the property which has changed. Validation is not performed.</param>
		protected virtual void OnPropertyChanged( string propertyName )
		{
			if( PropertyChanged != null )
			{
				PropertyChangedEventArgs eventArgs = new PropertyChangedEventArgs( propertyName );

				PropertyChanged( this, eventArgs );
			}
		}

		#endregion Private methods
	}
}
