////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections;
using System.Reflection;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.BinarySerialization
{
	/// <summary>
	/// A deserializer for documents. It can deserialize documents which were serialized with <see cref="BinarySerializer"/>.
	/// </summary>
	public class BinaryDeserializer : IDeserializer
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected BinaryDeserializer()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="textEncoding">Text encoding.</param>
		public BinaryDeserializer( TextEncoding textEncoding )
		{
			TextEncoding = textEncoding;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The document reader which contains all the nodes to be used to set the property values of the object to create.
		/// </summary>
		private BinarySerializerReader DocumentReader;

		#endregion Private fields

		#region Public non-serializable properties

		private int _CandidatePropCount;
		/// <summary>
		/// The number of properties which were candidates for deserialization, and had to be verified whether they are to be
		/// deserialized or not. This does not include the root node. READ-ONLY.
		/// </summary>
		public int CandidatePropCount
		{
			get { return _CandidatePropCount; }
			protected set { _CandidatePropCount = value; }
		}

		private int _ProcessedPropCount;
		/// <summary>
		/// The number of properties which were deserialized. This does not include the root node. READ-ONLY.
		/// </summary>
		public int ProcessedPropCount
		{
			get { return _ProcessedPropCount; }
			protected set { _ProcessedPropCount = value; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private TextEncoding _TextEncoding;
		/// <summary>
		/// Text encoding. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public TextEncoding TextEncoding
		{
			get { return _TextEncoding; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "TextEncoding" );

				_TextEncoding = value;
			}
		}

		#endregion Public serializable properties

		#region Private static methods

		/// <summary>
		/// Finds the root serialization node. An exception if thrown if this is not found.
		/// </summary>
		/// <param name="documentType">The root type of the document to load; the document to load may either have this type or
		/// a type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <param name="contentType">Is set with the content type of the found serialization node. By default, this is
		/// "declaredType".</param>
		/// <returns>Returns the found serialization node.</returns>
		private BinarySerializationNode FindRootSerializationNode( Type documentType, out Type contentType )
		{
			BinarySerializationNode foundNode = DocumentReader[ BinarySerializationConsts.RootSerializationName ];
			if( foundNode == null )
				throw new ValidationException( "The root serialization node was not found in the document to deserialize." );
			if( (foundNode.NodeKind != SerializationNodeKinds.Class) && (foundNode.NodeKind != SerializationNodeKinds.List) )
				throw new ValidationException( "The root serialization node must be a non-null class or list." );

			// The type specified in the document must be derived from the declared type and must be found in the list of global
			// portable types.
			contentType = PortableTypes.FindGlobal( foundNode.ValueTypeName );
			if( contentType == null )
			{
				throw new ValidationException( string.Format( "The type \"{0}\" of the root serialization node was not found in"
					+ " the list of global portable types.", foundNode.ValueTypeName ) );
			}

			// Check if the found type is derived from the document type.
			if( (documentType != null) && !documentType.IsAssignableFrom( contentType ) )
			{
				throw new ValidationException( string.Format( "The type \"{0}\" of the root serialization node is not derived"
					+ " from the declared type \"{1}\" of the property.", foundNode.ValueTypeName, documentType.FullName ) );
			}

			return foundNode;
		}

		/// <summary>
		/// Finds a serialization node whose name matches the specified serialization name.
		/// </summary>
		/// <param name="currentNode">The serialization node of the current object.</param>
		/// <param name="serializationName">The serialization name for which to find the serialization node.</param>
		/// <param name="declaredType">The declared type of the property for which to find a serialization node.</param>
		/// <param name="contentType">Is set with the content type of the found serialization node. By default, this is
		/// "declaredType".</param>
		/// <returns>Returns the found serialization node, or null if none is found.</returns>
		private BinarySerializationNode FindSerializationNode( BinarySerializationNode currentNode
			, string serializationName, Type declaredType, out Type contentType )
		{
			ValidationUtils.ObjectHasContent( currentNode, "currentNode" );
			ValidationUtils.TextHasContent( serializationName, "serializationName" );
			ValidationUtils.ObjectHasContent( declaredType, "declaredType" );

			string fullSerializationName = SerializationUtils.FormatSerializationName( currentNode.SerializationName
				, serializationName );

			BinarySerializationNode foundNode = DocumentReader[ fullSerializationName ];
			if( foundNode == null )
			{
				contentType = declaredType;

				return null;
			}

			// The type specified in the document must be derived from the declared type and must be found in the list of global
			// portable types.
			contentType = PortableTypes.FindGlobal( foundNode.ValueTypeName );
			if( contentType == null )
			{
				throw new ValidationException( string.Format( "The type \"{0}\" of the property \"{1}\" of the serialization"
						+ " node \"{2}\" was not found in the list of global portable types."
					, foundNode.ValueTypeName, serializationName, foundNode.SerializationName ) );
			}

			// Check if the found type is derived from the declared type. Don't check scalar types because scalar type
			// conversion may be implemented by the value deserialization method ("ReadScalar").
			if( !ScalarEncoder.IsScalar( contentType ) && !declaredType.IsAssignableFrom( contentType ) )
			{
				throw new ValidationException( string.Format( "The type \"{0}\" of the property \"{1}\" of the serialization"
						+ " node \"{2}\" is not derived from the declared type \"{3}\" of the property."
					, foundNode.ValueTypeName, serializationName, foundNode.SerializationName, declaredType.FullName ) );
			}

			return foundNode;
		}

		/// <summary>
		/// Reads a scalar object from the specified serialization node.
		/// </summary>
		/// <param name="currentNode">The serialization node of the object to read.</param>
		/// <param name="contentType">The type of the scalar object to read.</param>
		/// <param name="declaredType">Declared type of the property to which the converted value is to be assigned.</param>
		/// <param name="enumFallbackValue">The fallback value for enumerations. If the object to deserialize is an
		/// enumeration, if the deserialization of the enumeration fails, this value is assigned instead of an exception being
		/// thrown. This may be null.</param>
		/// <returns>Returns the read scalar object. This may be null.</returns>
		private static object ReadScalar( BinarySerializationNode currentNode, Type contentType, Type declaredType
			, object enumFallbackValue )
		{
			ValidationUtils.ObjectHasContent( currentNode, "objectNode" );

			if( !currentNode.IsScalar || !ScalarEncoder.IsScalar( contentType ) )
			{
				throw new ValidationException( string.Format( "The object to deserialize (\"{0}\") is not a scalar, but a"
					+ " \"{1}\".", currentNode.SerializationName, contentType.FullName ) );
			}

			if( currentNode.IsNullValue )
			{
				throw new ValidationException( string.Format( "The object to deserialize (\"{0}\") has no value."
					, currentNode.SerializationName ) );
			}

			return ScalarEncoder.ValueFromBuffer( currentNode.Value, contentType, declaredType, enumFallbackValue );
		}

		#endregion Private static methods

		#region Private methods

		/// <summary>
		/// Reads a scalar object from a serialization node with the specified name.
		/// </summary>
		/// <param name="serializationName">The serialization name of the scalar to geet.</param>
		/// <param name="contentType">The type of the scalar object to read.</param>
		/// <param name="declaredType">Declared type of the property to which the converted value is to be assigned.</param>
		/// <param name="enumFallbackValue">The fallback value for enumerations. If the object to deserialize is an
		/// enumeration, if the deserialization of the enumeration fails, this value is assigned instead of an exception being
		/// thrown. This may be null.</param>
		/// <returns>Returns the read scalar object. This may be null.</returns>
		private object GetScalar( string serializationName, Type contentType, Type declaredType, object enumFallbackValue )
		{
			ValidationUtils.ObjectHasContent( serializationName, "serializationName" );

			if( !ScalarEncoder.IsScalar( contentType ) )
			{
				throw new ValidationException( string.Format( "The object to deserialize (\"{0}\") is not a scalar, but a"
					+ " \"{1}\".", serializationName, contentType.FullName ) );
			}

			byte[] valueBuffer = DocumentReader.GetValue( serializationName, contentType );
			if( valueBuffer == null ) return null;

			return ScalarEncoder.ValueFromBuffer( valueBuffer, contentType, declaredType, enumFallbackValue );
		}

		/// <summary>
		/// Reads a scalar object from a serialization node with the specified name.
		/// </summary>
		/// <param name="serializationName">The serialization name of the scalar to geet.</param>
		/// <param name="contentType">The type of the scalar object to read.</param>
		/// <returns>Returns the read scalar object. This may be null.</returns>
		private object GetScalar( string serializationName, Type contentType )
		{
			return GetScalar( serializationName, contentType, contentType, null );
		}

		/// <summary>
		/// Checks if the specified serialization name is that of a list item from the list with the specified prefix name.
		/// </summary>
		/// <param name="serializationName">The serialization name of the list item to check. This is not validated.</param>
		/// <param name="prefixName">List prefix name. This is not validated.</param>
		/// <returns>Returns true if the specified serialization name is that of a list item from the list with the specified
		/// prefix name.</returns>
		private bool IsItemNameChildOfListNode( string serializationName, string prefixName )
		{
			return serializationName.StartsWith( prefixName, StringComparison.OrdinalIgnoreCase )
				&& (serializationName.IndexOf( SerializationConsts.FormattedSerializationNameSeparator, prefixName.Length
					, StringComparison.OrdinalIgnoreCase ) < 0);
		}

		/// <summary>
		/// Gets all the list item nodes whose parent is the specified list node.
		/// </summary>
		/// <param name="listNode">The node which is the parent of the nodes to get.</param>
		/// <returns>Returns all the list item nodes whose parent is the specified list node.</returns>
		private BinarySerializationNodes GetListItemNodes( BinarySerializationNode listNode )
		{
			ValidationUtils.ObjectHasContent( listNode, "listNode" );

			BinarySerializationNodes nodes = new BinarySerializationNodes();

			string prefixName = listNode.SerializationName + SerializationConsts.FormattedSerializationNameSeparator;

			foreach( BinarySerializationNode node in DocumentReader )
			{
				if( IsItemNameChildOfListNode( node.SerializationName, prefixName ) ) nodes.Add( node );
			}

			return nodes;
		}

		/// <summary>
		/// Processes a list (which may actually be a class derived from <see cref="IEnumerable"/>) with its items.
		/// </summary>
		/// <param name="currentNode">The serialization node of the current object.</param>
		/// <param name="listType">The type of the list object to read.</param>
		/// <param name="enumFallbackValue">The fallback value for enumerations. If the list item to deserialize is an
		/// enumeration, if the deserialization of the enumeration fails, this value is assigned instead of an exception being
		/// thrown. This may be null.</param>
		/// <returns>Returns the read list object.</returns>
		private object ProcessList( BinarySerializationNode currentNode, Type listType, object enumFallbackValue )
		{
			ValidationUtils.ObjectHasContent( currentNode, "currentNode" );
			ValidationUtils.ObjectHasContent( listType, "listType" );

			if( currentNode.NodeKind != SerializationNodeKinds.List )
			{
				throw new ValidationException( string.Format( "The type \"{0}\" is a list but the serialization node from"
					+ " which to read its content is not.", listType.FullName ) );
			}

			// The deserialized list object.
			object listObject;

			// The serialization nodes for the list items.
			BinarySerializationNodes listNodes = GetListItemNodes( currentNode );
			int itemsCount = listNodes.Count;

			// Prepare to call the method named "DeserializingMethodName" before the list is deserialized, so that the method is
			// called before it is called for its items.
			MethodInfo eventMethodInfo = listType.GetMethod( SerializationConsts.DeserializingMethodName
				, BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance, null, Type.EmptyTypes, null );

			Array array = null; // Used if the list is an array.
			MethodInfo addMethodInfo = null; // Used if the list isn't an array.
			Type itemDeclaredType = null;

			if( listType.IsArray )
			{
				// The type is an array.

				itemDeclaredType = listType.GetElementType();

				if( itemDeclaredType == null )
				{
					throw new ValidationException( string.Format( "The list of type \"{0}\" doesn't have a type for items."
						, listType.FullName ) );
				}

				// Create an instance of the specified list type.
				listObject = Array.CreateInstance( itemDeclaredType, itemsCount );

				array = (Array)listObject;

				// Multi-dimensional arrays can't be deserialized.
				if( array.Rank != 1 )
				{
					throw new ValidationException( string.Format( "Can't deserialize list of type \"{0}\" because it is"
						+ " multi-dimensional ({1}).", listType.FullName, array.Rank ) );
				}

				// Call the method "DeserializingMethodName" before the list is deserialized, so that the method is called
				// before it is called for its items.
				if( eventMethodInfo != null ) eventMethodInfo.Invoke( listObject, null );
			}
			else
			{
				// The type is a list.

				// Create an instance of the specified list type.
				listObject = Activator.CreateInstance( listType, true );

				// Call the method "DeserializingMethodName" before the list is deserialized, so that the method is called
				// before it is called for its items.
				if( eventMethodInfo != null ) eventMethodInfo.Invoke( listObject, null );

				// Get the type's instance method named "ListAddMethodName" with a single "object" parameter.
				addMethodInfo = listType.GetMethod( SerializationConsts.ListAddMethodName, BindingFlags.Public
					| BindingFlags.NonPublic | BindingFlags.Instance, null, new Type[] { typeof( object ) }, null );

				if( typeof( IItemTypeAwareList ).IsAssignableFrom( listType ) )
				{
					IItemTypeAwareList itaList = (IItemTypeAwareList)listObject;
					itemDeclaredType = itaList.ItemDeclaredType;

					if( itemDeclaredType == null )
					{
						throw new ValidationException( string.Format( "The list of type \"{0}\" doesn't have a type for items."
							, listType.FullName ) );
					}
				}
			}

			int i = 0;
			foreach( BinarySerializationNode itemNode in listNodes )
			{
				#region Check global type

				// Item-type-aware lists know the declared type of their items, so the types of the items from an
				// item-type-aware list don't need to be globally registered (unless those types are derived from the declared
				// type of the items).

				string itemContentTypeName = itemNode.ValueTypeName;

				Type itemType = null;
				bool checkGlobalType = true;

				if( itemDeclaredType != null )
				{
					// Check if the item's type is the declared type of the list's items.
					checkGlobalType = !TextUtils.Equals( itemDeclaredType.FullName, itemContentTypeName );

					if( !checkGlobalType ) itemType = itemDeclaredType;
				}

				if( checkGlobalType )
				{
					// The types of the items from a list which is not item-type-aware need to be globally registered.

					itemType = PortableTypes.FindGlobal( itemContentTypeName );
				}

				if( itemType == null )
				{
					throw new ValidationException( string.Format( "The type attribute \"{0}\" of the list of type \"{1}\" of"
							+ " the item's serialization node \"{2}\" is unknown."
						, itemContentTypeName, listType.FullName, itemNode.SerializationName ) );
				}

				#endregion Check global type

				// The deserialized list item.
				object listItem;

				if( ScalarEncoder.IsScalar( itemType ) )
				{
					listItem = ReadScalar( itemNode, itemType, itemDeclaredType, enumFallbackValue );
				}
				else if( SerializationUtils.IsList( itemType ) )
				{
					throw new ValidationException( string.Format( "Can't deserialize nested list of type \"{0}\"."
						, itemType.FullName ) );
				}
				else if( SerializationUtils.IsClass( itemType ) )
				{
					listItem = ProcessClass( itemNode, itemType );
				}
				else
				{
					throw new ValidationException( string.Format( "List items of type \"{0}\" can't be deserialized."
						, itemType.FullName ) );
				}

				// Deserialize only non-null items.
				if( listType.IsArray )
				{
					if( listItem != null )
					{
						array.SetValue( listItem, i );
						i++;
					}
				}
				else
				{
					if( listItem != null ) addMethodInfo.Invoke( listObject, new object[] { listItem } );
				}
			}

			// Call the method "DeserializedMethodName" after the class is deserialized, so that the method is called after it
			// was called for its items.
			eventMethodInfo = listType.GetMethod( SerializationConsts.DeserializedMethodName, BindingFlags.Public
				| BindingFlags.NonPublic | BindingFlags.Instance, null, Type.EmptyTypes, null );
			if( eventMethodInfo != null ) eventMethodInfo.Invoke( listObject, null );

			return listObject;
		}

		/// <summary>
		/// Processes all serializable members of the specified class type, and sets them with objects read from the specified
		/// serialization node, using serialization name matching.
		/// </summary>
		/// <param name="currentNode">The serialization node of the current object.</param>
		/// <param name="classType">The type of the class object to read.</param>
		/// <returns>Returns the read class object.</returns>
		private IDocument ProcessClass( BinarySerializationNode currentNode, Type classType )
		{
			ValidationUtils.ObjectHasContent( currentNode, "currentNode" );

			ConstructorInfo classConstructor = SerializationUtils.CheckSerializableClass( classType );

			// Create an instance of the specified class type. "Activator.CreateInstance" is not used for performance reasons.
			IDocument serializableData = (IDocument)classConstructor.Invoke( null );

			// Call the method "DeserializingMethodName" before the class is deserialized, so that the method is called before
			// it is called for child class properties.
			MethodInfo eventMethodInfo = classType.GetMethod( SerializationConsts.DeserializingMethodName
				, BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance, null, Type.EmptyTypes, null );
			if( eventMethodInfo != null ) eventMethodInfo.Invoke( serializableData, null );

			// Process the properties of the class.

			PropertyInfo[] propInfos = classType.GetProperties( BindingFlags.Public | BindingFlags.NonPublic
				| BindingFlags.Instance );

			foreach( PropertyInfo propInfo in propInfos )
			{
				// Serialize only members which can be read and written.
				if( propInfo.CanRead && propInfo.CanWrite )
				{
					string serializationName, listItemName;
					bool mustBeDeserialized;
					object enumFallbackValue;

					if( SerializationUtils.CheckDeserializeAttribute( propInfo, out mustBeDeserialized, out serializationName
						, out listItemName, out enumFallbackValue ) )
					{
						// If the value of the serialization node is null, it's name would be mapped to the declared
						// serialization name and type of the current property.

						Type contentType; // The content type.

						BinarySerializationNode thisNode = FindSerializationNode( currentNode, serializationName
							, propInfo.PropertyType, out contentType );

						if( thisNode != null )
						{
							// A node was found, so process the current property.
							ProcessObject( thisNode, propInfo, contentType, serializableData, enumFallbackValue );
						}

						if( thisNode == null )
						{
							if( mustBeDeserialized )
							{
								throw new ValidationException( string.Format( "The property \"{0}\" from class \"{1}\" must"
										+ " have a value in the document which is being deserialized, but it has none."
									, propInfo.Name, classType.FullName ) );
							}
							else
							{
								// Don't deserialize null objects.
							}
						}

						ProcessedPropCount++;
					}

					CandidatePropCount++;
				}
			}

			// Call the method "DeserializedMethodName" after the class is deserialized, so that the method is called after it
			// was called for child class properties.
			eventMethodInfo = classType.GetMethod( SerializationConsts.DeserializedMethodName, BindingFlags.Public
				| BindingFlags.NonPublic | BindingFlags.Instance, null, Type.EmptyTypes, null );
			if( eventMethodInfo != null ) eventMethodInfo.Invoke( serializableData, null );

			return serializableData;
		}

		/// <summary>
		/// Processes the specified object, according to the type of the object's content.
		/// </summary>
		/// <param name="currentNode">The serialization node of the current object.</param>
		/// <param name="propInfo">Information about the property which is to be set with the processed object.</param>
		/// <param name="contentType">The type of the object to read.</param>
		/// <param name="serializableData">The parent class object for the property which is to be set with the read
		/// object.</param>
		/// <param name="enumFallbackValue">The fallback value for enumerations. If the object to deserialize is an
		/// enumeration, if the deserialization of the enumeration fails, this value is assigned instead of an exception being
		/// thrown. This may be null.</param>
		private void ProcessObject( BinarySerializationNode currentNode, PropertyInfo propInfo, Type contentType
			, object serializableData, object enumFallbackValue )
		{
			ValidationUtils.ObjectHasContent( currentNode, "currentNode" );
			ValidationUtils.ObjectHasContent( propInfo, "propInfo" );
			ValidationUtils.ObjectHasContent( contentType, "contentType" );
			ValidationUtils.ObjectHasContent( serializableData, "serializableData" );

			// The deserialized document.
			object val;

			if( ScalarEncoder.IsScalar( contentType ) )
			{
				// Deserialize scalar.
				val = ReadScalar( currentNode, contentType, propInfo.PropertyType, enumFallbackValue );

				// Set the property content with the deserialized document.
				propInfo.SetValue( serializableData, val, null );
			}
			else
			{
				if( SerializationUtils.IsList( contentType ) )
				{
					// Process the items of the list.
					val = ProcessList( currentNode, contentType, enumFallbackValue );
				}
				else if( SerializationUtils.IsClass( contentType ) )
				{
					// Process class.
					val = ProcessClass( currentNode, contentType );
				}
				else
				{
					throw new ValidationException( string.Format( "Objects of type \"{0}\" can't be deserialized."
						, contentType.FullName ) );
				}

				// Set the property content with the deserialized document.
				propInfo.SetValue( serializableData, val, null );
			}
		}

		/// <summary>
		/// Loads the class object from the specified buffer.
		/// </summary>
		/// <param name="buffer">Buffer to deserialize. This may be null or empty.</param>
		/// <param name="documentType">The root type of the document to load; the document to load may either have this type or
		/// a type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <returns>Returns the deserialized document, or null if the input data is null or empty.</returns>
		private IDocument DeserializeCore( byte[] buffer, Type documentType )
		{
			if( BufferUtils.NoContent( buffer ) ) return null;

			DocumentReader = new BinarySerializerReader( buffer );

			// Get the serializer type name.
			string serializerTypeName = (string)GetScalar( BinarySerializationConsts.SerializerTypeNameNodeName
				, typeof( string ) );

			// Check the type of the serializer.
			if( !TextUtils.Equals( serializerTypeName, typeof( BinarySerializer ).FullName ) )
			{
				throw new ValidationException( string.Format( "The extracted serializer type name is \"{0}\" instead of the"
					+ " expected \"{1}\".", serializerTypeName, typeof( BinarySerializer ).FullName ) );
			}

			// Get the text encoding name.
			string textEncodingName = (string)GetScalar( BinarySerializationConsts.TextEncodingNameNodeName
				, typeof( string ) );

			// Get the text encoding.
			TextEncoding = (TextEncoding)RegistrableAlgorithms.InstantiateGlobal( textEncodingName, typeof( TextEncoding ) );

			// Get the root node.
			Type rootType;
			BinarySerializationNode rootNode = FindRootSerializationNode( documentType, out rootType );

			// Deserialize the specified class type.
			CandidatePropCount = 0;
			ProcessedPropCount = 0;

			IDocument serializableData;

			if( SerializationUtils.IsList( rootType ) ) serializableData = ProcessList( rootNode, rootType, null ) as IDocument;
			else serializableData = ProcessClass( rootNode, rootType );

			return serializableData;
		}

		#endregion Private methods

		#region Public static methods

		/// <summary>
		/// Checks if the current deserializer may be used to deserialize the specified data.
		/// </summary>
		/// <param name="buffer">Serialized data. This may be a binary encoded text.</param>
		/// <returns>Returns an instance of itself if it may parse the specified data, or null if it may not parse it.</returns>
		/// <remarks>
		/// <para>The text encoding of the deserializer is extracted from the specified data.</para>
		/// </remarks>
		public static BinaryDeserializer GetParsingAlgorithm( byte[] buffer )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );

			BinaryNodeStream bns = new BinaryNodeStream( sizeof( Int32 ), buffer );

			if( !bns.CanReadFirstNode( SerializationConsts.SerializerTypeNameSizeMax ) ) return null;

			#region Read the serializer type name

			BinarySerializationNode node = BinarySerializerReader.ReadNextNode( bns, typeof( string ) );
			if( (node == null) || !node.IsScalar ) return null;

			string serializerTypeName = (string)ScalarEncoder.ValueFromBuffer( node.Value, typeof( string ) );
			if( TextUtils.NoContent( serializerTypeName ) ) return null;

			if( !TextUtils.Equals( serializerTypeName, typeof( BinarySerializer ).FullName ) ) return null;

			#endregion Read the serializer type name

			#region Read the text encoding name

			node = BinarySerializerReader.ReadNextNode( bns, typeof( string ) );
			if( node == null ) return null;

			string textEncodingName = (string)ScalarEncoder.ValueFromBuffer( node.Value, typeof( string ) );
			if( TextUtils.NoContent( textEncodingName ) ) return null;

			TextEncoding textEncoding = RegistrableAlgorithms.InstantiateGlobal( textEncodingName, typeof( TextEncoding ) )
				as TextEncoding;

			#endregion Read the text encoding name

			if( TextUtils.Equals( serializerTypeName, typeof( BinarySerializer ).FullName ) )
				return new BinaryDeserializer( textEncoding );

			return null;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Returns a serializer with the configuration from the current serialization factory.
		/// </summary>
		/// <returns>Returns the serializer with the configuration from the current serialization factory.</returns>
		public ISerializer GetSerializer()
		{
			return new BinarySerializer( TextEncoding );
		}

		/// <summary>
		/// Returns a deserializer with the configuration from the current serialization factory.
		/// </summary>
		/// <returns>Returns the deserializer with the configuration from the current serialization factory.</returns>
		public IDeserializer GetDeserializer()
		{
			return new BinaryDeserializer( TextEncoding );
		}

		/// <summary>
		/// Returns a hash serializer with the configuration from the current serialization factory.
		/// </summary>
		/// <param name="serializeValueTypeName">If this is true then the type names of the serialized values are stored in the
		/// serialization output. If this is true, the compatibility level of different versions of a serialized document is
		/// limited (see the "Serialization versioning" section from the documentation of the <see cref="IHashSerializer"/>
		/// interface). This must be true for a hash serializer which will be used to test the equality of documents.</param>
		/// <returns>Returns the hash serializer with the configuration from the current serialization factory.</returns>
		public IHashSerializer GetHashSerializer( bool serializeValueTypeName )
		{
			return new BinaryHashSerializer( TextEncoding, serializeValueTypeName );
		}

		/// <summary>
		/// Creates a deep clone of the specified document by serializing and then deserializing it.
		/// </summary>
		/// <param name="source">Document to clone. This may be null.</param>
		/// <returns>Returns the cloned document. This is null if "source" is null.</returns>
		public IDocument CloneDocument( IDocument source )
		{
			return GetSerializer().CloneDocument( source );
		}

		/// <summary>
		/// Checks if the specified documents are equal by comparing the output of their serializations.
		/// </summary>
		/// <param name="document1">Document to compare. This may be null.</param>
		/// <param name="document2">Document to compare. This may be null.</param>
		/// <returns>Returns the true if the documents are equal.</returns>
		/// <remarks>
		/// <para>The documents to compare may have different types.</para>
		/// </remarks>
		public bool EqualDocuments( IDocument document1, IDocument document2 )
		{
			return GetHashSerializer( true ).EqualDocuments( document1, document2 );
		}

		/// <summary>
		/// Loads the class object from the specified buffer.
		/// </summary>
		/// <param name="buffer">Buffer to deserialize. This may be null or empty.</param>
		/// <param name="documentType">The root type of the document to load; the document to load may either have this type or
		/// a type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <returns>Returns the deserialized document, or null if the input data is null or empty.</returns>
		public IDocument LoadFromBuffer( byte[] buffer, Type documentType )
		{
			return DeserializeCore( buffer, documentType );
		}

		/// <summary>
		/// Loads the class object from the specified file.
		/// </summary>
		/// <param name="fileName">The name of the file to deserialize. The file may be missing and its content may be null or
		/// empty.</param>
		/// <param name="documentType">The root type of the document to load; the document to load may either have this type or
		/// a type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <returns>Returns the deserialized document, or null if the file is missing or its content is null or empty.</returns>
		/// <remarks>
		/// <para>This method implements a recovery mechanism following a crash during a call of the
		/// <see cref="ISerializer.SaveToFile"/> method.</para>
		/// </remarks>
		public IDocument LoadFromFile( string fileName, Type documentType )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			try
			{
				// Read the serialized data from the file.
				byte[] buffer = FileUtils.Read( fileName, false );

				// Deserialize.
				return LoadFromBuffer( buffer, documentType );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsAbort( ex ) ) throw ExceptionUtils.PreserveStackTrace( ex );

				throw new ValidationException( string.Format( "Failed to load the document from file \"{0}\".", fileName ), ex );
			}
		}

		#endregion Public methods
	}
}
