////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Serialization;

namespace Vela.BinarySerialization
{
	/// <summary>
	/// A binary hash serializer writer.
	/// </summary>
	/// <remarks>
	/// <para>The nodes with null or non-scalar values are not written to the output.</para>
	/// </remarks>
	public class BinaryHashSerializerWriter : KernelSortedList<string, BinarySerializationNode>, IBinarySerializerWriter
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="textEncoding">Text encoding. This may be null if the data to convert to a buffer is not a
		/// text.</param>
		/// <param name="serializeValueTypeName">If this is true then the type names of the serialized values are stored in the
		/// accumulator. If this is true, the compatibility level of different versions of a serialized document is
		/// limited (see the "Serialization versioning" section from the documentation of the
		/// <see cref="Vela.Serialization.IHashSerializer"/> interface). This must be true for a hash serializer which will be
		/// used to test the equality of documents.</param>
		/// <param name="accumulator">The accumulator where to write the output.</param>
		public BinaryHashSerializerWriter( TextEncoding textEncoding, bool serializeValueTypeName, BinaryNodeStream accumulator )
			: base( StringInsensitiveComparer.Instance )
		{
			ValidationUtils.ObjectHasContent( accumulator, "accumulator" );
			ValidationUtils.ObjectHasContent( textEncoding, "textEncoding" );

			TextEncoding = textEncoding;
			SerializeValueTypeName = serializeValueTypeName;
			OutputAccumulator = accumulator;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// Text encoding.
		/// </summary>
		private TextEncoding TextEncoding;

		/// <summary>
		/// If this is true then the type names of the serialized values are stored in the accumulator. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is true, the compatibility level of different versions of a serialized document is limited (see the
		/// "Serialization versioning" section from the documentation of the 
		/// <see cref="Vela.Serialization.IHashSerializer"/> interface).</para>
		/// 
		/// <para>This must be true for a hash serializer which will be used to test the equality of documents. This is because
		/// the parameterless constructor may set some non-serializable properties which can make two (otherwise identical) types
		/// different.</para>
		/// </remarks>
		private bool SerializeValueTypeName;

		/// <summary>
		/// The accumulator where to write the serializer's output.
		/// </summary>
		private BinaryNodeStream OutputAccumulator;

		#endregion Private fields

		#region Private methods

		/// <summary>
		/// Writes a node to the output of the serializer.
		/// </summary>
		/// <param name="node">The node which is to be written to the output of the serializer writer.</param>
		private void WriteNodeOutput( BinarySerializationNode node )
		{
			ValidationUtils.ObjectHasContent( node, "node" );

			// For maximum compatibility between different serialization versions, ignore null values.
			if( !node.IsScalar && !SerializeValueTypeName )
				return;

			// The serialization name.
			OutputAccumulator.WriteValue( node.SerializationName );

			// The value type name.
			if( SerializeValueTypeName )
				OutputAccumulator.WriteValue( node.ValueTypeName );

			// The value.
			if( node.IsScalar )
				OutputAccumulator.WriteValue( node.Value );

			// The end of node marker.
			OutputAccumulator.WriteValue( BinarySerializationConsts.EndOfNodeMarker );
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Gets the key from the specified node.
		/// </summary>
		/// <param name="node">Node to get the key from.</param>
		/// <returns>Returns the key from the specified node.</returns>
		/// <remarks>
		/// <para><see cref="BinarySerializationNode.SerializationName"/> is used as key.</para>
		/// </remarks>
		public override string GetKey( BinarySerializationNode node )
		{
			ValidationUtils.ObjectHasContent( node, "node" );

			return node.SerializationName;
		}

		/// <summary>
		/// Adds a new class type name node to the writer.
		/// </summary>
		/// <param name="serializationName">Serialization name.</param>
		/// <param name="type">Class type.</param>
		public void AddClassTypeName( string serializationName, Type type )
		{
			ValidationUtils.ObjectHasContent( type, "type" );

			SerializationUtils.ValidateSignableType( type );

			BinarySerializationNode node = BinarySerializationNode.GetClassNode( serializationName, type.FullName );

			Add( node );
		}

		/// <summary>
		/// Adds a new list type name node to the writer.
		/// </summary>
		/// <param name="serializationName">Serialization name.</param>
		/// <param name="type">List type.</param>
		public void AddListTypeName( string serializationName, Type type )
		{
			ValidationUtils.ObjectHasContent( type, "type" );

			SerializationUtils.ValidateSignableType( type );

			BinarySerializationNode node = BinarySerializationNode.GetListNode( serializationName, type.FullName );

			Add( node );
		}

		/// <summary>
		/// Adds the specified scalar to the writer.
		/// </summary>
		/// <param name="serializationName">Serialization name.</param>
		/// <param name="value">Value. This must be a scalar.</param>
		public void AddScalar( string serializationName, object value )
		{
			ValidationUtils.ObjectHasContent( value, "value" );

			string valueTypeName = value.GetType().FullName;
			byte[] valueBuffer = ScalarEncoder.ValueToBuffer( value, TextEncoding, true );

			BinarySerializationNode node = BinarySerializationNode.GetScalarNode( serializationName, valueTypeName
				, valueBuffer );

			Add( node );
		}

		/// <summary>
		/// Writes the output of the serializer.
		/// </summary>
		public void WriteOutput()
		{
			foreach( BinarySerializationNode node in this )
				WriteNodeOutput( node );
		}

		#endregion Public methods
	}
}
