////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Serialization;

namespace Vela.BinarySerialization
{
	/// <summary>
	/// A node for a binary serialization writer / reader.
	/// </summary>
	public class BinarySerializationNode : ISerializationNode
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="nodeKind">Node kind.</param>
		protected BinarySerializationNode( SerializationNodeKinds nodeKind )
		{
			NodeKind = nodeKind;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private SerializationNodeKinds _NodeKind;
		/// <summary>
		/// Node kind. READ-ONLY.
		/// </summary>
		public SerializationNodeKinds NodeKind
		{
			get { return _NodeKind; }
			protected set { _NodeKind = value; }
		}

		private string _SerializationName;
		/// <summary>
		/// Serialization name. READ-ONLY.
		/// </summary>
		public string SerializationName
		{
			get { return _SerializationName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "SerializationName" );

				_SerializationName = value;
			}
		}

		private string _ValueTypeName;
		/// <summary>
		/// Value type name. READ-ONLY.
		/// </summary>
		public string ValueTypeName
		{
			get { return _ValueTypeName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "ValueTypeName" );

				_ValueTypeName = value;
			}
		}

		private byte[] _Value;
		/// <summary>
		/// Value. This may be null or empty. READ-ONLY.
		/// </summary>
		public byte[] Value
		{
			get { return _Value; }
			protected set { _Value = value; }
		}

		/// <summary>
		/// If this is true then the node contains a scalar value. READ-ONLY.
		/// </summary>
		public bool IsScalar
		{
			get { return NodeKind == SerializationNodeKinds.Scalar; }
		}

		/// <summary>
		/// If this is true then the value of the node is null. READ-ONLY.
		/// </summary>
		public bool IsNullValue
		{
			get { return Value == null; }
		}

		#endregion Public non-serializable properties

		#region Public static methods

		/// <summary>
		/// Gets a node used for class objects.
		/// </summary>
		/// <param name="serializationName">Serialization name.</param>
		/// <param name="valueTypeName">Value type name. This has to be the content type name if the value is not null, or the
		/// declared type name if the value is null.</param>
		/// <returns>Returns a node used for class objects.</returns>
		public static BinarySerializationNode GetClassNode( string serializationName, string valueTypeName )
		{
			BinarySerializationNode node = new BinarySerializationNode( SerializationNodeKinds.Class );

			node.SerializationName = serializationName;
			node.ValueTypeName = valueTypeName;

			return node;
		}

		/// <summary>
		/// Gets a node used for list objects.
		/// </summary>
		/// <param name="serializationName">Serialization name.</param>
		/// <param name="valueTypeName">Value type name. This has to be the content type name if the value is not null, or the
		/// declared type name if the value is null.</param>
		/// <returns>Returns a node used for list objects.</returns>
		public static BinarySerializationNode GetListNode( string serializationName, string valueTypeName )
		{
			BinarySerializationNode node = new BinarySerializationNode( SerializationNodeKinds.List );

			node.SerializationName = serializationName;
			node.ValueTypeName = valueTypeName;

			return node;
		}

		/// <summary>
		/// Gets a node used for scalar objects.
		/// </summary>
		/// <param name="serializationName">Serialization name.</param>
		/// <param name="valueTypeName">Value type name.</param>
		/// <param name="value">Value.</param>
		/// <returns>Returns a node used for scalar objects.</returns>
		public static BinarySerializationNode GetScalarNode( string serializationName, string valueTypeName, byte[] value )
		{
			ValidationUtils.ObjectHasContent( value, "value" );

			BinarySerializationNode node = new BinarySerializationNode( SerializationNodeKinds.Scalar );

			node.SerializationName = serializationName;
			node.ValueTypeName = valueTypeName;
			node.Value = value;

			return node;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Gets the textual representation of all the items from the list.
		/// </summary>
		/// <returns>Returns the textual representation of all the items from the list.</returns>
		public override string ToString()
		{
			return SerializationName;
		}

		#endregion Public methods
	}
}
