////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.BinarySerialization
{
	/// <summary>
	/// Binary serializer. The serialization output can be deserialized with <see cref="BinaryDeserializer"/>.
	/// </summary>
	public class BinarySerializer : BinarySerializerBase
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected BinarySerializer()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="textEncoding">Text encoding.</param>
		public BinarySerializer( TextEncoding textEncoding )
			: base( textEncoding )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// This is false because the current serializer is a not hash serializer. READ-ONLY.
		/// </summary>
		public override bool IsHashSerializer
		{
			get { return false; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Creates the serializer's output writer.
		/// </summary>
		/// <param name="output">The output of the serializer is written here.</param>
		protected override void CreateOutputWriter( object output )
		{
			ValidationUtils.ObjectType<BinaryNodeStream>( output, "output" );

			SerializerWriter = new BinarySerializerWriter( (BinaryNodeStream)output, TextEncoding );
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Returns a serializer with the configuration from the current serialization factory.
		/// </summary>
		/// <returns>Returns the serializer with the configuration from the current serialization factory.</returns>
		public override ISerializer GetSerializer()
		{
			return new BinarySerializer( TextEncoding );
		}

		/// <summary>
		/// Returns a deserializer with the configuration from the current serialization factory.
		/// </summary>
		/// <returns>Returns the deserializer with the configuration from the current serialization factory.</returns>
		public override IDeserializer GetDeserializer()
		{
			return new BinaryDeserializer( TextEncoding );
		}

		/// <summary>
		/// Returns a hash serializer with the configuration from the current serialization factory.
		/// </summary>
		/// <param name="serializeValueTypeName">If this is true then the type names of the serialized values are stored in the
		/// serialization output. If this is true, the compatibility level of different versions of a serialized document is
		/// limited (see the "Serialization versioning" section from the documentation of the <see cref="IHashSerializer"/>
		/// interface). This must be true for a hash serializer which will be used to test the equality of documents.</param>
		/// <returns>Returns the hash serializer with the configuration from the current serialization factory.</returns>
		public override IHashSerializer GetHashSerializer( bool serializeValueTypeName )
		{
			return new BinaryHashSerializer( TextEncoding, serializeValueTypeName );
		}

		/// <summary>
		/// Creates a deep clone of the specified document by serializing and then deserializing it.
		/// </summary>
		/// <param name="source">Document to clone. This may be null.</param>
		/// <returns>Returns the cloned document. This is null if "source" is null.</returns>
		public override IDocument CloneDocument( IDocument source )
		{
			if( source == null ) return null;

			Type docType = source.GetType();

			// Serialize the source object.
			byte[] serialized = SaveToBuffer( source, null );

			// Deserialize the serialized source object.
			return GetDeserializer().LoadFromBuffer( serialized, docType );
		}

		/// <summary>
		/// Checks if the specified documents are equal by comparing the output of their serializations.
		/// </summary>
		/// <param name="document1">Document to compare. This may be null.</param>
		/// <param name="document2">Document to compare. This may be null.</param>
		/// <returns>Returns the true if the documents are equal.</returns>
		/// <remarks>
		/// <para>The documents to compare may have different types.</para>
		/// </remarks>
		public override bool EqualDocuments( IDocument document1, IDocument document2 )
		{
			return GetHashSerializer( true ).EqualDocuments( document1, document2 );
		}

		/// <summary>
		/// Saves the specified class object to a buffer.
		/// </summary>
		/// <param name="serializableData">The object to serialize. This has to be a class.</param>
		/// <param name="rootName">The name of the root class object which is serialized. This name will be used during
		/// deserialization to recognize the serialized type. If this is null then the name of the specified type is used
		/// instead.</param>
		/// <returns>Returns the serialized document.</returns>
		public override byte[] SaveToBuffer( IDocument serializableData, string rootName )
		{
			BinaryNodeStream bns = new BinaryNodeStream( 0, TextEncoding );

			WriteOutput( serializableData, bns );

			return bns.ToArray();
		}

		/// <summary>
		/// Saves the specified class object to the specified file. Existing files are overwritten.
		/// </summary>
		/// <param name="serializableData">The object to serialize. This has to be a class.</param>
		/// <param name="rootName">The name of the root class object which is serialized. This name will be used during
		/// deserialization to recognize the serialized type. If this is null then the name of the specified type is used
		/// instead.</param>
		/// <param name="fileName">The name of the file to be saved.</param>
		/// <remarks>
		/// <para>The saving is crash-safe. If the application crashes when the file is being saved, at least 1 of the
		/// following files is still available: a file with the specified name, a file with a backup name, a file with a
		/// temporary name. If the backup file exists, it's the valid one and you should implement your own recovery mechanism
		/// following a crash.</para>
		/// </remarks>
		public override void SaveToFile( IDocument serializableData, string rootName, string fileName )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			try
			{
				byte[] buffer = SaveToBuffer( serializableData, rootName );

				// Save the serialized data.
				FileUtils.Save( fileName, buffer );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsAbort( ex ) ) throw ExceptionUtils.PreserveStackTrace( ex );

				throw new ValidationException( string.Format( "Failed to save the document to file \"{0}\".", fileName ), ex );
			}
		}

		#endregion Public methods
	}
}
