////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections;
using System.Reflection;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.BinarySerialization
{
	/// <summary>
	/// Contract for a binary serializer.
	/// </summary>
	public abstract class BinarySerializerBase : ISerializer
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected BinarySerializerBase()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="textEncoding">Text encoding.</param>
		protected BinarySerializerBase( TextEncoding textEncoding )
		{
			TextEncoding = textEncoding;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The writer of the output from a serializer.
		/// </summary>
		protected IBinarySerializerWriter SerializerWriter;

		/// <summary>
		/// The object which is currently being serialized.
		/// </summary>
		private object RootSerializableData;

		#endregion Private fields

		#region Public non-serializable properties

		/// <summary>
		/// This is true if the current serializer is a hash serializer. Otherwise, an object can be deserialized with
		/// <see cref="BinaryDeserializer"/>. READ-ONLY.
		/// </summary>
		public abstract bool IsHashSerializer { get; }

		private int _CandidatePropCount;
		/// <summary>
		/// The number of properties which were candidates for serialization, and had to be verified whether they are to be
		/// serialized or not. This does not include the root node. READ-ONLY.
		/// </summary>
		public int CandidatePropCount
		{
			get { return _CandidatePropCount; }
			protected set { _CandidatePropCount = value; }
		}

		private int _ProcessedPropCount;
		/// <summary>
		/// The number of properties which were serialized. This does not include the root node. READ-ONLY.
		/// </summary>
		public int ProcessedPropCount
		{
			get { return _ProcessedPropCount; }
			protected set { _ProcessedPropCount = value; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private TextEncoding _TextEncoding;
		/// <summary>
		/// Text encoding. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public TextEncoding TextEncoding
		{
			get { return _TextEncoding; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "TextEncoding" );

				_TextEncoding = value;
			}
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Creates the serializer's output writer.
		/// </summary>
		/// <param name="output">The output of the serializer is written here.</param>
		protected abstract void CreateOutputWriter( object output );

		/// <summary>
		/// Processes a list (which may actually be a <b>class</b> derived from <see cref="IEnumerable"/>) with its items. List
		/// items with a null value are ignored.
		/// </summary>
		/// <param name="serializationName">The serialization name of the specified list.</param>
		/// <param name="listObject">The list object to process.</param>
		/// <param name="listItemName">The serialization name of list items. If this is null or empty and the serialized
		/// property is a list, the default name (<see cref="SerializationConsts.ListItemName"/>) is used.</param>
		private void ProcessList( string serializationName, object listObject, string listItemName )
		{
			ValidationUtils.TextHasContent( serializationName, "serializationName" );
			ValidationUtils.ObjectHasContent( listObject, "listObject" );

			if( TextUtils.NoContent( listItemName ) )
				listItemName = SerializationConsts.ListItemName;

			Type listType = listObject.GetType();

			if( listType.IsArray )
			{
				// Multi-dimensional arrays can't be serialized.

				Array mda = (Array)listObject;
				if( mda.Rank != 1 )
				{
					throw new ValidationException( string.Format( "Can't serialize list \"{0}\" because it is multi-dimensional"
						+ " ({1}).", serializationName, mda.Rank ) );
				}
			}

			// Call the method "SerializingMethodName" before the list is serialized, so that the method is called before it is
			// called for its items.
			MethodInfo eventMethodInfo = listType.GetMethod( SerializationConsts.SerializingMethodName
				, BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance, null, Type.EmptyTypes, null );
			if( eventMethodInfo != null )
				eventMethodInfo.Invoke( listObject, null );

			IEnumerable typedListObject = (IEnumerable)listObject;

			Type itemDeclaredType = null;

			if( listType.IsArray )
			{
				// The type is an array.

				itemDeclaredType = listType.GetElementType();

				if( itemDeclaredType == null )
				{
					throw new ValidationException( string.Format( "The list of type \"{0}\" doesn't have a type for items."
						, listType.FullName ) );
				}
			}
			else
			{
				// The type is a list.

				if( typeof( IItemTypeAwareList ).IsAssignableFrom( listType ) )
				{
					IItemTypeAwareList itaList = (IItemTypeAwareList)typedListObject;
					itemDeclaredType = itaList.ItemDeclaredType;

					if( itemDeclaredType == null )
					{
						throw new ValidationException( string.Format( "The list of type \"{0}\" doesn't have a type for items."
							, listType.FullName ) );
					}
				}
			}

			int count = 0;
			foreach( object listItem in typedListObject )
			{
				// Don't serialize null objects.
				if( listItem == null )
					continue;

				Type itemType = listItem.GetType();

				if( itemType == null )
				{
					throw new ValidationException( string.Format( "The list item type for list \"{0}\" is missing."
						, serializationName ) );
				}

				#region Check global type

				// Item-type-aware lists know the declared type of their items, so the types of the items from an
				// item-type-aware list don't need to be globally registered (unless those types are derived from the declared
				// type of the items).

				bool checkGlobalType = true;

				if( itemDeclaredType != null )
				{
					// Check if the item's type is the declared type of the list's items.
					checkGlobalType = itemType != itemDeclaredType;
				}

				if( checkGlobalType )
				{
					// The types of the items from a list which is not item-type-aware need to be globally registered. We check
					// the type during serialization so that we don't have surprises during deserialiation.

					if( PortableTypes.FindGlobal( itemType ) == null )
					{
						throw new ValidationException( string.Format( "The list item with the content type \"{0}\", from the"
								+ " list property \"{1}\", can't be serialized because the type was not declared as"
								+ " serializable."
							, itemType.FullName, serializationName ) );
					}
				}

				#endregion Check global type

				string itemNameWithCounter = SerializationUtils.FormatSerializationListItemName( serializationName, listItemName
					, count );

				if( ScalarEncoder.IsScalar( itemType ) )
				{
					SerializerWriter.AddScalar( itemNameWithCounter, listItem );
				}
				else
				{
					if( SerializationUtils.IsList( itemType ) )
					{
						throw new ValidationException( string.Format( "Can't serialize nested list \"{0}\"."
							, serializationName ) );
					}
					else if( SerializationUtils.IsClass( itemType ) )
					{
						SerializerWriter.AddClassTypeName( itemNameWithCounter, itemType );

						ProcessClass( itemNameWithCounter, listItem );
					}
					else
					{
						throw new ValidationException( string.Format( "List items of type \"{0}\" can't be serialized."
							, itemType.FullName ) );
					}
				}

				count++;
			}

			// Call the method "SerializedMethodName" after the list is serialized, so that the method is called after it was
			// called for its items.
			eventMethodInfo = listType.GetMethod( SerializationConsts.SerializedMethodName, BindingFlags.Public
				| BindingFlags.NonPublic | BindingFlags.Instance, null, Type.EmptyTypes, null );
			if( eventMethodInfo != null )
				eventMethodInfo.Invoke( typedListObject, null );
		}

		/// <summary>
		/// Processes all serializable members of the specified class object.
		/// </summary>
		/// <param name="serializationName">The serialization name of the specified class object.</param>
		/// <param name="serializableData">Class object to process.</param>
		private void ProcessClass( string serializationName, object serializableData )
		{
			ValidationUtils.TextHasContent( serializationName, "serializationName" );
			ValidationUtils.ObjectHasContent( serializableData, "serializableData" );

			Type classType = serializableData.GetType();

			SerializationUtils.CheckSerializableClass( classType );

			// Call the method "SerializingMethodName" before the class is serialized, so that the method is called before it is
			// called for child class properties.
			MethodInfo eventMethodInfo = classType.GetMethod( SerializationConsts.SerializingMethodName
				, BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance, null, Type.EmptyTypes, null );
			if( eventMethodInfo != null )
				eventMethodInfo.Invoke( serializableData, null );

			PropertyInfo[] propInfos = classType.GetProperties( BindingFlags.Public | BindingFlags.NonPublic
				| BindingFlags.Instance );

			// Process the properties of the class.
			foreach( PropertyInfo propInfo in propInfos )
			{
				// Serialize only members which can be read and written.
				if( propInfo.CanRead && propInfo.CanWrite )
				{
					string propName, listItemName;
					object objVal;

					if( SerializationUtils.CheckSerializeAttribute( propInfo, serializableData, RootSerializableData
						, IsHashSerializer, out propName, out listItemName, out objVal ) )
					{
						string objectName = SerializationUtils.FormatSerializationName( serializationName, propName );
						ProcessObject( objectName, objVal, listItemName );

						ProcessedPropCount++;
					}

					CandidatePropCount++;
				}
			}

			// Call the method "SerializedMethodName" after the class is serialized, so that the method is called after it was
			// called for child class properties.
			eventMethodInfo = classType.GetMethod( SerializationConsts.SerializedMethodName, BindingFlags.Public
				| BindingFlags.NonPublic | BindingFlags.Instance, null, Type.EmptyTypes, null );
			if( eventMethodInfo != null )
				eventMethodInfo.Invoke( serializableData, null );
		}

		/// <summary>
		/// Processes the specified object, according to the type of the object's content.
		/// </summary>
		/// <param name="serializationName">The serialization name of the specified object.</param>
		/// <param name="objectToProcess">The object to process. This may be null, in which case a null serialization node
		/// attribute will be written.</param>
		/// <param name="listItemName">The serialization name of list items. If this is null or empty and the serialized
		/// property is a list, the default name (<see cref="SerializationConsts.ListItemName"/>) is used.</param>
		private void ProcessObject( string serializationName, object objectToProcess, string listItemName )
		{
			ValidationUtils.TextHasContent( serializationName, "serializationName" );

			if( objectToProcess != null )
			{
				Type objectType = objectToProcess.GetType();

				if( ScalarEncoder.IsScalar( objectType ) )
				{
					// Serialize scalar.
					SerializerWriter.AddScalar( serializationName, objectToProcess );
				}
				else
				{
					if( SerializationUtils.IsList( objectType ) )
					{
						SerializerWriter.AddListTypeName( serializationName, objectType );

						// Process the items of the list.
						ProcessList( serializationName, objectToProcess, listItemName );
					}
					else if( SerializationUtils.IsClass( objectType ) )
					{
						SerializerWriter.AddClassTypeName( serializationName, objectType );

						// Process class.
						ProcessClass( serializationName, objectToProcess );
					}
					else
					{
						throw new ValidationException( string.Format( "Objects of type \"{0}\" can't be serialized."
							, objectType.FullName ) );
					}
				}
			}
			else
			{
				// Don't serialize null objects.
			}
		}

		/// <summary>
		/// Writes the serialization output of the specified class object to the specified output object.
		/// </summary>
		/// <param name="serializableData">The object to serialize. This has to be a class.</param>
		/// <param name="output">The output of the serializer is written here.</param>
		protected void WriteOutput( object serializableData, object output )
		{
			ValidationUtils.ObjectHasContent( serializableData, "serializableData" );
			ValidationUtils.ObjectHasContent( output, "output" );

			Type objectType = serializableData.GetType();
			if( ScalarEncoder.IsScalar( objectType ) || !SerializationUtils.IsClass( objectType ) )
			{
				throw new ValidationException( string.Format( "The specified object of type \"{0}\" is not a class type."
					, objectType.FullName ) );
			}

			CreateOutputWriter( output );
			ValidationUtils.ObjectHasContent( SerializerWriter, "SerializerWriter" );

			CandidatePropCount = 0;
			ProcessedPropCount = 0;
			RootSerializableData = serializableData;

			// Add the serializer type name.
			SerializerWriter.AddScalar( BinarySerializationConsts.SerializerTypeNameNodeName, GetType().FullName );

			// Add the text encoding name.
			SerializerWriter.AddScalar( BinarySerializationConsts.TextEncodingNameNodeName
				, RegistrableAlgorithmUtils.GetAlgorithmName( TextEncoding ) );

			if( SerializationUtils.IsList( objectType ) )
			{
				// Add the root type.
				SerializerWriter.AddListTypeName( BinarySerializationConsts.RootSerializationName, objectType );

				// Serialize the specified list object.
				ProcessList( BinarySerializationConsts.RootSerializationName, serializableData, null );
			}
			else
			{
				// Add the root type.
				SerializerWriter.AddClassTypeName( BinarySerializationConsts.RootSerializationName, objectType );

				// Serialize the specified class object.
				ProcessClass( BinarySerializationConsts.RootSerializationName, serializableData );
			}

			SerializerWriter.WriteOutput();
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Returns a serializer with the configuration from the current serialization factory.
		/// </summary>
		/// <returns>Returns the serializer with the configuration from the current serialization factory.</returns>
		public abstract ISerializer GetSerializer();

		/// <summary>
		/// Returns a deserializer with the configuration from the current serialization factory.
		/// </summary>
		/// <returns>Returns the deserializer with the configuration from the current serialization factory.</returns>
		public abstract IDeserializer GetDeserializer();

		/// <summary>
		/// Returns a hash serializer with the configuration from the current serialization factory.
		/// </summary>
		/// <param name="serializeValueTypeName">If this is true then the type names of the serialized values are stored in the
		/// serialization output. If this is true, the compatibility level of different versions of a serialized document is
		/// limited (see the "Serialization versioning" section from the documentation of the <see cref="IHashSerializer"/>
		/// interface). This must be true for a hash serializer which will be used to test the equality of documents.</param>
		/// <returns>Returns the hash serializer with the configuration from the current serialization factory.</returns>
		public abstract IHashSerializer GetHashSerializer( bool serializeValueTypeName );

		/// <summary>
		/// Creates a deep clone of the specified document by serializing and then deserializing it.
		/// </summary>
		/// <param name="source">Document to clone. This may be null.</param>
		/// <returns>Returns the cloned document. This is null if "source" is null.</returns>
		public abstract IDocument CloneDocument( IDocument source );

		/// <summary>
		/// Checks if the specified documents are equal by comparing the output of their serializations.
		/// </summary>
		/// <param name="document1">Document to compare. This may be null.</param>
		/// <param name="document2">Document to compare. This may be null.</param>
		/// <returns>Returns the true if the documents are equal.</returns>
		/// <remarks>
		/// <para>The documents to compare may have different types.</para>
		/// </remarks>
		public abstract bool EqualDocuments( IDocument document1, IDocument document2 );

		/// <summary>
		/// Saves the specified class object to a buffer.
		/// </summary>
		/// <param name="serializableData">The object to serialize. This has to be a class.</param>
		/// <param name="rootName">The name of the root class object which is serialized. This name will be used during
		/// deserialization to recognize the serialized type. If this is null then the name of the specified type is used
		/// instead.</param>
		/// <returns>Returns the serialized document.</returns>
		public abstract byte[] SaveToBuffer( IDocument serializableData, string rootName );

		/// <summary>
		/// Saves the specified class object to the specified file. Existing files are overwritten.
		/// </summary>
		/// <param name="serializableData">The object to serialize. This has to be a class.</param>
		/// <param name="rootName">The name of the root class object which is serialized. This name will be used during
		/// deserialization to recognize the serialized type. If this is null then the name of the specified type is used
		/// instead.</param>
		/// <param name="fileName">The name of the file to be saved.</param>
		public abstract void SaveToFile( IDocument serializableData, string rootName, string fileName );

		#endregion Public methods
	}
}
