////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Serialization;

namespace Vela.BinarySerialization
{
	/// <summary>
	/// A binary serializer reader.
	/// </summary>
	public class BinarySerializerReader : BinarySerializationNodes
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="buffer">The data from where to read the nodes.</param>
		public BinarySerializerReader( byte[] buffer )
		{
			LoadNodes( buffer );
		}

		#endregion Constructors

		#region Private methods

		/// <summary>
		/// Loads all the nodes from the specified buffer
		/// </summary>
		/// <param name="buffer">The data from where to read the nodes.</param>
		private void LoadNodes( byte[] buffer )
		{
			BinaryNodeStream bns = new BinaryNodeStream( 0, buffer );

			BinarySerializationNode node;
			while( (node = ReadNextNode( bns )) != null )
			{
				Add( node );
			}
		}

		#endregion Private methods

		#region Public static methods

		/// <summary>
		/// Reads the next node from the specified stream.
		/// </summary>
		/// <param name="bns">Stream from which to read the next node.</param>
		/// <param name="valueType">The value must have this type. If this is null, it's ignored.</param>
		/// <returns>Returns the read node, or null if no more nodes are found.</returns>
		public static BinarySerializationNode ReadNextNode( BinaryNodeStream bns, Type valueType = null )
		{
			ValidationUtils.ObjectHasContent( bns, "bns" );

			BinarySerializationNode node = null;

			if( !bns.HasNextNode() )
				return null;

			// The node kind.
			SerializationNodeKinds nodeKind = (SerializationNodeKinds)bns.ReadValue( typeof( byte ) );

			// The serialization name.
			string serializationName = (string)bns.ReadValue( typeof( string ) );

			// The value type name.
			string valueTypeName = (string)bns.ReadValue( typeof( string ) );

			if( (valueType != null) && !TextUtils.Equals( valueTypeName, valueType.FullName ) )
			{
				throw new ValidationException( string.Format( "The read value type is \"{0}\" instead of \"{1}\"."
					, valueTypeName, valueType.FullName ) );
			}

			if( nodeKind == SerializationNodeKinds.Class )
			{
				node = BinarySerializationNode.GetClassNode( serializationName, valueTypeName );
			}
			else if( nodeKind == SerializationNodeKinds.List )
			{
				node = BinarySerializationNode.GetListNode( serializationName, valueTypeName );
			}
			else if( nodeKind == SerializationNodeKinds.Scalar )
			{
				// The value.
				byte[] value = bns.ReadBuffer();

				node = BinarySerializationNode.GetScalarNode( serializationName, valueTypeName, value );
			}
			else
			{
				throw ValidationUtils.Unhandled( nodeKind, "nodeKind" );
			}

			// The end of node marker.
			byte[] eonMarker = (byte[])bns.ReadValue( typeof( byte[] ) );
			if( !BufferUtils.Equals( eonMarker, BinarySerializationConsts.EndOfNodeMarker ) )
				throw new ValidationException( "The end of node marker read by the binary deserializer is invalid." );

			return node;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Gets the value of the node with the specified serialization name. An exception is thrown if a node with the
		/// specified serialization name and value type is not found.
		/// </summary>
		/// <param name="serializationName">The serialization name of the node whose value to get.</param>
		/// <param name="valueType">The type of the value to get.</param>
		/// <returns>Returns the value of the node with the specified serialization name. This may be null.</returns>
		public byte[] GetValue( string serializationName, Type valueType )
		{
			BinarySerializationNode node = Items[ serializationName ];

			if( node == null )
			{
				throw new ValidationException( string.Format( "A binary serializer reader node named \"{0}\" was not found."
					, serializationName ) );
			}

			if( !TextUtils.Equals( node.ValueTypeName, valueType.FullName ) )
			{
				throw new ValidationException( string.Format( "A binary serializer reader node named \"{0}\" was found with"
						+ " the \"{1}\" type instead of the requested \"{2}\"."
					, serializationName, node.ValueTypeName, valueType.FullName ) );
			}

			return node.Value;
		}

		#endregion Public methods
	}
}
