////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;

namespace Vela.BinarySerialization
{
	/// <summary>
	/// A binary serializer writer.
	/// </summary>
	public class BinarySerializerWriter : BinarySerializationNodes, IBinarySerializerWriter
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="accumulator">The accumulator where to write the output.</param>
		/// <param name="textEncoding">Text encoding. This may be null if the data to convert to a buffer is not a
		/// text.</param>
		public BinarySerializerWriter( BinaryNodeStream accumulator, TextEncoding textEncoding )
		{
			ValidationUtils.ObjectHasContent( accumulator, "accumulator" );
			ValidationUtils.ObjectHasContent( textEncoding, "textEncoding" );

			OutputAccumulator = accumulator;
			TextEncoding = textEncoding;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The accumulator where to write the serializer's output.
		/// </summary>
		private BinaryNodeStream OutputAccumulator;

		/// <summary>
		/// Text encoding.
		/// </summary>
		private TextEncoding TextEncoding;

		#endregion Private fields

		#region Private methods

		/// <summary>
		/// Writes a node to the output of the serializer.
		/// </summary>
		/// <param name="node">The node which is to be written to the output of the serializer writer.</param>
		private void WriteNodeOutput( BinarySerializationNode node )
		{
			ValidationUtils.ObjectHasContent( node, "node" );

			// The node kind.
			OutputAccumulator.WriteValue( (byte)node.NodeKind );

			// The serialization name.
			OutputAccumulator.WriteValue( node.SerializationName );

			// The value type name.
			OutputAccumulator.WriteValue( node.ValueTypeName );

			// The value.
			if( node.IsScalar )
				OutputAccumulator.WriteValue( node.Value );

			// The end of node marker.
			OutputAccumulator.WriteValue( BinarySerializationConsts.EndOfNodeMarker );
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Adds a new class type name node to the writer.
		/// </summary>
		/// <param name="serializationName">Serialization name.</param>
		/// <param name="type">Class type.</param>
		public void AddClassTypeName( string serializationName, Type type )
		{
			ValidationUtils.ObjectHasContent( type, "type" );

			BinarySerializationNode node = BinarySerializationNode.GetClassNode( serializationName, type.FullName );

			Add( node );
		}

		/// <summary>
		/// Adds a new list type name node to the writer.
		/// </summary>
		/// <param name="serializationName">Serialization name.</param>
		/// <param name="type">List type.</param>
		public void AddListTypeName( string serializationName, Type type )
		{
			ValidationUtils.ObjectHasContent( type, "type" );

			BinarySerializationNode node = BinarySerializationNode.GetListNode( serializationName, type.FullName );

			Add( node );
		}

		/// <summary>
		/// Adds a new scalar node to the writer.
		/// </summary>
		/// <param name="serializationName">Serialization name.</param>
		/// <param name="value">Value. This must be a scalar.</param>
		public void AddScalar( string serializationName, object value )
		{
			ValidationUtils.ObjectHasContent( value, "value" );

			string valueTypeName = value.GetType().FullName;
			byte[] valueBuffer = ScalarEncoder.ValueToBuffer( value, TextEncoding, true );

			BinarySerializationNode node = BinarySerializationNode.GetScalarNode( serializationName, valueTypeName
				, valueBuffer );

			Add( node );
		}

		/// <summary>
		/// Writes the output of the serializer.
		/// </summary>
		public void WriteOutput()
		{
			#region Write the header

			// The order of these fields is critical because "BinaryDeserializer.GetParsingAlgorithm" tries to read them in this
			// order, so that it doesn't have to read the entire serialized object just to get to them.

			// Write the serializer type name.
			BinarySerializationNode serializerTypeNameNode = this[ BinarySerializationConsts.SerializerTypeNameNodeName ];
			ValidationUtils.ObjectHasContent( serializerTypeNameNode, "serializerTypeNameNode" );
			WriteNodeOutput( serializerTypeNameNode );

			// Write the text encoding name.
			BinarySerializationNode textEncodingNameNode = this[ BinarySerializationConsts.TextEncodingNameNodeName ];
			ValidationUtils.ObjectHasContent( textEncodingNameNode, "textEncodingNameNode" );
			WriteNodeOutput( textEncodingNameNode );

			#endregion Write the header

			foreach( BinarySerializationNode node in this )
			{
				// Skip the header fields (which were written above).
				if( (node == serializerTypeNameNode) || (node == textEncodingNameNode) )
					continue;

				WriteNodeOutput( node );
			}
		}

		#endregion Public methods
	}
}
