﻿////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Reflection;
using System.Windows.Threading;
using Vela.Kernel;
using Vela.Kernel.Cloning;

namespace Vela.Cloning
{
	/// <summary>
	/// Cloner.
	/// </summary>
	/// <remarks>
	/// <para>This cloner deep clones all the instance (= non-static) fields of a class.</para>
	/// 
	/// <para>Properties are not cloned through their setters because the source value may be invalid, so the setter may throw an
	/// exception, and because the setter might change other fields.</para>
	/// 
	/// <para>The fields which have the <see cref="UnclonableAttribute"/> attribute on them are not clonable.</para>
	/// 
	/// <para>The following types are not clonable: value types (= primitives, structures, enumerations), types derived from
	/// <see cref="IUnclonable"/>, <see cref="Pointer"/>, <see cref="Delegate"/>, <see cref="EventHandler"/>,
	/// <see cref="Dispatcher"/>, <see cref="DispatcherObject"/>.</para>
	/// 
	/// <para>A type is not clonable if one of its following properties is true: <see cref="Type.IsCOMObject"/>,
	/// <see cref="Type.IsImport"/>, <see cref="Type.IsMarshalByRef"/> true, <see cref="Type.IsPointer"/>.</para>
	/// 
	/// <para>Fields of unclonable types are shallow cloned with the <see cref="object.MemberwiseClone"/> method when the class
	/// in which they are declared is cloned.</para>
	/// </remarks>
	public class Cloner
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public Cloner()
		{
		}

		#endregion Constructors

		#region Private static fields

		/// <summary>
		/// The information of the method to call in order to perform a shallow clone on an object.
		/// </summary>
		private static readonly MethodInfo MemberwiseCloneMethod = typeof( object ).GetMethod( "MemberwiseClone"
			, BindingFlags.NonPublic | BindingFlags.Instance );

		#endregion Private static fields

		#region Private fields

		/// <summary>
		/// List of fields which were already (deep) cloned.
		/// </summary>
		/// <remarks>
		/// <para>This is necessary because (in the object to clone) there may be several fields which refer the same object "A".
		/// "A" must be (deep) cloned a single time.</para>
		/// </remarks>
		private Dictionary<object, object> ClonedFields = new Dictionary<object, object>( ReferenceComparer.Instance );

		#endregion Private fields

		#region Private static methods

		/// <summary>
		/// Checks if the specified type may be (deep) cloned.
		/// </summary>
		/// <param name="type">The type to check. This is not validated.</param>
		/// <returns>Returns true if the specified type may be (deep) cloned, and false if it may NOT be (deep) cloned.</returns>
		private static bool MayClone( Type type )
		{
			return !type.IsValueType && (type != typeof( string ))
				&& !type.IsCOMObject && !type.IsImport && !type.IsMarshalByRef && !type.IsPointer
				&& !typeof( IUnclonable ).IsAssignableFrom( type )
				&& !typeof( Pointer ).IsAssignableFrom( type )
				&& !typeof( Delegate ).IsAssignableFrom( type )
				&& !typeof( EventHandler ).IsAssignableFrom( type )
				&& !typeof( Dispatcher ).IsAssignableFrom( type )
				&& !typeof( DispatcherObject ).IsAssignableFrom( type );
		}

		/// <summary>
		/// Checks if the specified field must be (deep) cloned.
		/// </summary>
		/// <param name="fieldInfo">Information about the field to check. This is not validated.</param>
		/// <returns>Returns true if the specified field must be deep cloned, and false if it must NOT be deep cloned.</returns>
		private static bool MustClone( FieldInfo fieldInfo )
		{
			if( !MayClone( fieldInfo.FieldType ) ) return false;

			object[] deepCloneAttributes = fieldInfo.GetCustomAttributes( typeof( UnclonableAttribute ), false );

			// The specified field may be deep cloned if it doesn't have the "UnclonableAttribute" attribute.
			return ArrayUtils.NoContent( deepCloneAttributes );
		}

		/// <summary>
		/// Computes the indices to access the next array item.
		/// </summary>
		/// <param name="maxIndices">The maximum values for the indices. This is not validated.</param>
		/// <param name="currentIndices">The current values for the indices. This is modified in here. This is not validated.
		/// </param>
		/// <returns>Returns true if the current indices refer a valid array item (after the method ends), and false if there are
		/// no more items in the array.</returns>
		private static bool NextArrayItem( int[] maxIndices, int[] currentIndices )
		{
			int indicesCount = currentIndices.Length;
			for( int i = 0; i < indicesCount; i++ )
			{
				if( currentIndices[ i ] < maxIndices[ i ] )
				{
					currentIndices[ i ]++;

					return true;
				}
				else if( i < indicesCount - 1 )
				{
					currentIndices[ i ] = 0;
					currentIndices[ i + 1 ]++;
				}
			}

			// There are no more items in the array.
			return false;
		}

		#endregion Private static methods

		#region Private methods

		/// <summary>
		/// Deep clones the fields of the specified object.
		/// </summary>
		/// <param name="source">Object whose fields to (deep) clone. This is not validated.</param>
		/// <param name="shallowClone">Object which was already shallow cloned. This is modified in here. This is not
		/// validated.</param>
		/// <param name="typeToClone">The type of the object to clone. This may be null.</param>
		private void DeepCloneFields( object source, object shallowClone, Type typeToClone )
		{
			if( typeToClone == null ) return;

			BindingFlags bindingFlags = BindingFlags.Instance | BindingFlags.DeclaredOnly | BindingFlags.NonPublic
				| BindingFlags.Public;

			FieldInfo[] fieldInfos = typeToClone.GetFields( bindingFlags );

			foreach( FieldInfo fieldInfo in fieldInfos )
			{
				if( !MustClone( fieldInfo ) ) continue;

				object sourceFieldValue = fieldInfo.GetValue( source );

				if( sourceFieldValue != null ) fieldInfo.SetValue( shallowClone, InternalDeep( sourceFieldValue ) );
			}

			// Recurse the base type. This is necessary because the "Type.GetFields" method doesn't return the private fields
			// from the base type.
			DeepCloneFields( source, shallowClone, typeToClone.BaseType );
		}

		/// <summary>
		/// Deep clones the items of the specified array.
		/// </summary>
		/// <param name="shallowClone">Array which was already shallow cloned. This is modified in here. This is not validated.
		/// </param>
		/// <param name="typeToClone">The type of the array to clone. This is not validated.</param>
		private void DeepCloneArrayItems( Array shallowClone, Type typeToClone )
		{
			if( !MayClone( typeToClone.GetElementType() ) || (shallowClone.LongLength <= 0) ) return;

			int[] maxIndices = new int[ shallowClone.Rank ];
			for( int i = 0; i < shallowClone.Rank; i++ ) maxIndices[ i ] = shallowClone.GetLength( i ) - 1;

			int[] currentIndices = new int[ shallowClone.Rank ];

			do
			{
				object clonedItem = InternalDeep( shallowClone.GetValue( currentIndices ) );

				shallowClone.SetValue( clonedItem, currentIndices );
			}
			while( NextArrayItem( maxIndices, currentIndices ) );
		}

		/// <summary>
		/// Deep clones the specified object.
		/// </summary>
		/// <param name="source">Object to (deep) clone. This may be null.</param>
		private object InternalDeep( object source )
		{
			if( source == null ) return null;

			if( ClonedFields.ContainsKey( source ) ) return ClonedFields[ source ];

			Type typeToClone = source.GetType();
			if( !MayClone( typeToClone ) ) return source;

			// Shallow clone.
			object clone = MemberwiseCloneMethod.Invoke( source, null );
			ClonedFields.Add( source, clone );

			// Deep clone.
			if( typeToClone.IsArray ) DeepCloneArrayItems( (Array)clone, typeToClone );
			else DeepCloneFields( source, clone, typeToClone );

			return clone;
		}

		#endregion Private methods

		#region Public static methods

		/// <summary>
		/// Creates a deep clone of the specified object.
		/// </summary>
		/// <param name="source">Object to clone. This may be null.</param>
		/// <returns>Returns the clone object. This is null if "source" is null.</returns>
		public static object Deep( object source )
		{
			if( source == null ) return null;

			Cloner cloner = new Cloner();

			return cloner.InternalDeep( source );
		}

		/// <summary>
		/// Creates a deep clone of the specified object.
		/// </summary>
		/// <typeparam name="TSource">The type of the object to clone.</typeparam>
		/// <param name="source">Object to clone. This may be null.</param>
		/// <returns>Returns the clone object. This is null if "source" is null.</returns>
		public static TSource Deep<TSource>( TSource source )
			where TSource : class
		{
			if( source == null ) return null;

			Cloner cloner = new Cloner();

			return (TSource)cloner.InternalDeep( source );
		}

		#endregion Public static methods
	}
}
