////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// "Communication Frame Algorithm" communication frame algorithm.
	/// </summary>
	/// <remarks>
	/// <para>Communication frame header (4 bytes):</para>
	/// <list type="bullet">
	///		<item>Body size (4 bytes).</item>
	/// </list>
	/// </remarks>
	public class Cofa : ICommunicationFrameAlg
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public Cofa()
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The size of the header. READ-ONLY.
		/// </summary>
		public int HeaderSize
		{
			get { return 4; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Sets up the header of the communication frame.
		/// </summary>
		/// <param name="body">The body of the communication frame.</param>
		/// <returns>Returns the created header.</returns>
		public byte[] SetupHeader( byte[] body )
		{
			ValidationUtils.ArrayHasContent( body, "body" );

			byte[] header = new byte[ HeaderSize ];
			int headerOffset = 0;

			// The body size.
			byte[] bodySizeBuffer = BitConverter.GetBytes( body.Length );
			if( bodySizeBuffer.Length != sizeof( int ) )
				throw new ValidationException( "The size of the body size buffer must be the size of an integer." );
			Buffer.BlockCopy( bodySizeBuffer, 0, header, headerOffset, bodySizeBuffer.Length );
			headerOffset += 4;

			return header;
		}

		/// <summary>
		/// Encodes the specified message into the body of the communication frame.
		/// </summary>
		/// <param name="message">The message to encode.</param>
		/// <returns>Returns the resulted body.</returns>
		public byte[] EncodeBody( byte[] message )
		{
			ValidationUtils.ArrayHasContent( message, "message" );

			return message;
		}

		/// <summary>
		/// Decodes the specified body of the communication frame into a message.
		/// </summary>
		/// <param name="body">The body to decode.</param>
		/// <returns>Returns the resulted message.</returns>
		public byte[] DecodeBody( byte[] body )
		{
			ValidationUtils.ArrayHasContent( body, "body" );

			return body;
		}

		#endregion Private methods
	}
}
