////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

namespace Vela.CoreDocuments
{
	/// <summary>
	/// The available sheaf locks.
	/// </summary>
	/// <remarks>
	/// <para>The sheaf locking system is designed to work together with the "read committed" database isolation level. Its
	/// purpose is to support the safety level required by financial operations.</para>
	/// 
	/// <para>For example, an account sheaf which is supposed to be loaded from the storehouse at the beginning of a transaction,
	/// modified and then saved back in the storehouse, must be locked with <see cref="SheafLocks.Update"/> for the entire
	/// transaction. No other thread / transaction will then be able load the same sheaf during this time, if it intends to
	/// update it (but will be able load it if it doesn't intend to update it).</para>
	/// 
	/// <para>If there is no intent to update a sheaf, the sheaf should be loaded with a read / shared lock
	/// (<see cref="SheafLocks.Shared"/>) because this doesn't keep it locked for the entire transaction.</para>
	/// 
	/// <para>Sheafs also use update versions (= database row versioning), so they may be safely loaded during one transaction
	/// and saved during another.</para>
	/// </remarks>
	public enum SheafLocks
	{
		/// <summary>
		/// A shared lock. This type of lock is not kept for the entire transaction.
		/// </summary>
		/// <remarks>
		/// <para>A transaction can acquire a <see cref="Shared"/> lock only if no other transaction has an
		/// <see cref="Exclusive"/> lock on the same resource.</para>
		/// </remarks>
		Shared,

		/// <summary>
		/// An update lock. This type of lock is kept for the entire transaction, from the moment when it's acquired.
		/// </summary>
		/// <remarks>
		/// <para>A transaction can acquire an <see cref="Update"/> lock only if no other transaction has an <see cref="Update"/>
		/// or <see cref="Exclusive"/> lock on the same resource.</para>
		/// 
		/// <para>If the transaction holding this lock writes / changes the locked sheaf, this lock is converted to an
		/// <see cref="Exclusive"/> lock.</para>
		/// </remarks>
		Update,

		/// <summary>
		/// An exclusive lock. This type of lock is kept for the entire transaction, from the moment when it's acquired.
		/// </summary>
		/// <remarks>
		/// <para>A transaction can acquire an <see cref="Exclusive"/> lock only if no other transaction has an
		/// <see cref="Update"/> or <see cref="Exclusive"/> lock on the same resource.</para>
		/// </remarks>
		Exclusive
	}
}
