////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Collections.Generic;
using Vela.Kernel;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Utilities for working with core documents.
	/// </summary>
	public static class CoreDocumentsUtils
	{
		#region Public static methods

		/// <summary>
		/// Extracts the UPN algorithm associated to the specified data. An exception is thrown if a UPN algorithm can't be
		/// extracted.
		/// </summary>
		/// <param name="data">UPN data. This may be either a buffer or a text.</param>
		/// <returns>Returns the extracted UPN algorithm.</returns>
		/// <remarks>
		/// <para>The text encoding of the UPN algorithm is also extracted.</para>
		/// </remarks>
		public static UpnAlg ExtractUpnAlg( object data )
		{
			return RegistrableAlgorithmUtils.GetParser( data, typeof( UpnAlg ) ) as UpnAlg;
		}

		/// <summary>
		/// Collects the identities from the specified sheafs.
		/// </summary>
		/// <param name="sheafs">The sheafs from which to collect the identities. Only the sheafs derived from
		/// <see cref="ILocallyNamedPublicProcessingIdentitySheaf"/> are processed. This may be null or empty.</param>
		/// <param name="includeRemotelyNamed">If this is true then the remotely named sheafs are also included in the search.
		/// </param>
		/// <param name="identityPublishers">List where to collect the identity publishers. The dictionary key is the instance
		/// locator of the current sheaf. This may be empty, but not null.</param>
		/// <param name="identities">List where to collect the identities. The dictionary key is the instance locator of the
		/// current sheaf. This may be empty, but not null.</param>
		public static void CollectIdentities( Sheafs sheafs, bool includeRemotelyNamed
			, Dictionary<Locator, IIdentityPublisher> identityPublishers, Dictionary<Locator, PublicIdentity> identities )
		{
			ValidationUtils.ObjectHasContent( identityPublishers, "identityPublishers" );
			ValidationUtils.ObjectHasContent( identities, "identities" );

			if( ListUtils.NoContent( sheafs ) ) return;

			if( includeRemotelyNamed )
			{
				foreach( ISheaf sheaf in sheafs )
				{
					if( sheaf as IPublicProcessingIdentitySheaf != null )
						((IPublicProcessingIdentitySheaf)sheaf).CollectIdentities( identityPublishers, identities );
				}
			}
			else
			{
				foreach( ISheaf sheaf in sheafs )
				{
					if( sheaf as ILocallyNamedPublicProcessingIdentitySheaf != null )
						((ILocallyNamedPublicProcessingIdentitySheaf)sheaf).CollectIdentities( identityPublishers, identities );
				}
			}
		}

		/// <summary>
		/// Collects the private identities from the specified sheafs.
		/// </summary>
		/// <param name="sheafs">The sheafs from which to collect the identities. Only the sheafs derived from
		/// <see cref="ILocallyNamedPrivateProcessingIdentitySheaf"/> are processed. This may be null or empty.</param>
		/// <param name="identities">List where to collect the identities. The dictionary key is the instance locator of the
		/// current sheaf. This may be empty, but not null.</param>
		public static void CollectPrivateIdentities( Sheafs sheafs, Dictionary<Locator, PrivateIdentity> identities )
		{
			ValidationUtils.ObjectHasContent( identities, "identities" );

			if( ListUtils.NoContent( sheafs ) ) return;

			foreach( ISheaf sheaf in sheafs )
			{
				if( sheaf as ILocallyNamedPrivateProcessingIdentitySheaf != null )
					((ILocallyNamedPrivateProcessingIdentitySheaf)sheaf).CollectPrivateIdentities( identities );
			}
		}

		/// <summary>
		/// Checks if the specified sheaf contains an identity with the specified UPN fingerprint.
		/// </summary>
		/// <param name="sheaf">Sheaf whose identities to check.</param>
		/// <param name="upnFingerprints">UPN fingerprints to check. This may be empty, but not null.</param>
		/// <returns>Returns true if the specified sheaf contains an identity with the specified UPN fingerprint.</returns>
		public static bool HasIdentityWithUpnFingerprint( ILocallyNamedPublicProcessingIdentitySheaf sheaf
			, Locators upnFingerprints )
		{
			ValidationUtils.ObjectHasContent( sheaf, "sheaf" );
			ValidationUtils.ObjectHasContent( upnFingerprints, "upnFingerprints" );

			Dictionary<Locator, IIdentityPublisher> identityPublishers = new Dictionary<Locator, IIdentityPublisher>();
			Dictionary<Locator, PublicIdentity> identities = new Dictionary<Locator, PublicIdentity>();

			sheaf.CollectIdentities( identityPublishers, identities );

			foreach( Locator upnFingerprint in upnFingerprints )
			{
				foreach( KeyValuePair<Locator, IIdentityPublisher> kvp in identityPublishers )
				{
					if( kvp.Value.Identities.ContainsKey( upnFingerprint ) ) return true;
				}

				foreach( KeyValuePair<Locator, PublicIdentity> kvp in identities )
				{
					if( kvp.Value.Upn.Fingerprint.Equals( upnFingerprint ) ) return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Gets a solitary identity publisher for the specified identity.
		/// </summary>
		/// <param name="identity">Identity whose identity publisher to get.</param>
		/// <returns>Returns a solitary identity publisher for the specified identity.</returns>
		public static SolitaryIdentityPublisher GetSolitaryIdentity( PrivateIdentity identity )
		{
			ValidationUtils.ObjectHasContent( identity, "identity" );

			return new SolitaryIdentityPublisher( identity.GetPublicIdentity() );
		}

		/// <summary>
		/// Gets the solitary identity publisher of the specified identity publisher.
		/// </summary>
		/// <param name="identityPublisher">Identity publisher whose identity publisher to get.</param>
		/// <returns>Returns the solitary identity publisher of the specified identity publisher.</returns>
		public static SolitaryIdentityPublisher GetSolitaryIdentity( IIdentityPublisher identityPublisher )
		{
			ValidationUtils.ObjectHasContent( identityPublisher, "identityPublisher" );

			if( identityPublisher as SolitaryIdentityPublisher != null ) return (SolitaryIdentityPublisher)identityPublisher;
			else if( identityPublisher.IdentityCount == 1 ) return identityPublisher.Identities.GetFirstItem();
			else
			{
				throw new ValidationException( "The identity publisher must contain one (and only one) solitary identity"
						+ " publisher.", null, null
					, new ExceptionDetail( "Identity publisher UPN", identityPublisher.Upn.ToString() ) );
			}
		}

		#endregion Public static methods
	}
}
