////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// An asymmetric decryption password structure, used by the <see cref="AsymmEncryptedEnvelope"/> class.
	/// </summary>
	public class AsymmDecryptionPassword
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AsymmDecryptionPassword()
		{
			IsDecryptDocument = true;
			IgnoreRecipientIdentity = true;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="recipientIdentity">The private identity which will be used during the decryption of the
		/// document.</param>
		public AsymmDecryptionPassword( PrivateIdentity recipientIdentity )
		{
			IsDecryptDocument = true;
			IgnoreRecipientIdentity = false;
			RecipientIdentity = recipientIdentity;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="passwords">The asymm decryption passwords to use for decryption.</param>
		public AsymmDecryptionPassword( params AsymmDecryptionPassword[] passwords )
		{
			IsDecryptDocument = true;
			IgnoreRecipientIdentity = true;

			MergeSessionKeys( passwords );
		}

		#endregion Constructors

		#region Public non-serializable properties

		private bool _IsDecryptDocument = true;
		/// <summary>
		/// Specifies whether the document has to be decrypted. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is false, when the document to decrypt is loaded, it's not decrypted, but the session keys intended for
		/// the current recipient identity (<see cref="RecipientIdentity"/>)are decrypted and added to the
		/// <see cref="SessionKeys"/> list.</para>
		/// </remarks>
		public bool IsDecryptDocument
		{
			get { return _IsDecryptDocument; }
			protected set { _IsDecryptDocument = value; }
		}

		private bool _IgnoreRecipientIdentity = false;
		/// <summary>
		/// If this is true then <see cref="RecipientIdentity"/> is ignored during the decryption of the document. READ-ONLY.
		/// </summary>
		public bool IgnoreRecipientIdentity
		{
			get { return _IgnoreRecipientIdentity; }
			protected set { _IgnoreRecipientIdentity = value; }
		}

		private PrivateIdentity _RecipientIdentity;
		/// <summary>
		/// The private identity of the recipient of the encrypted document
		/// (which is the complement of one of the public identities which were used to encrypt the document). READ-ONLY.
		/// </summary>
		public PrivateIdentity RecipientIdentity
		{
			get { return _RecipientIdentity; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RecipientIdentity" );

				_RecipientIdentity = value;
			}
		}

		private AsymmSessionKeys _SessionKeys = new AsymmSessionKeys();
		/// <summary>
		/// The list of session keys which were created in order to asymmetrically encrypt the document to a group of
		/// identities. These session keys need to be only the session keys of other recipients from the group. The backing
		/// field is initialized by default. READ-ONLY.
		/// </summary>
		public AsymmSessionKeys SessionKeys
		{
			get { return _SessionKeys; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "SessionKeys" );

				_SessionKeys = value;
			}
		}

		#endregion Public non-serializable properties

		#region Public static methods

		/// <summary>
		/// Prepares to collect the session keys from the document to decrypt. When the document to decrypt is loaded, it's not
		/// decrypted, but the session keys intended for the specified recipient identity are decrypted and added to the
		/// <see cref="SessionKeys"/> list.
		/// </summary>
		/// <param name="recipientIdentity">The private identity which will be used during the decryption of the
		/// document.</param>
		/// <remarks>
		/// <para>This method must be called when decrypting an encrypted envelope which was meant to be decrypted by several
		/// people together.</para>
		/// </remarks>
		public static AsymmDecryptionPassword PrepareToCollect( PrivateIdentity recipientIdentity )
		{
			AsymmDecryptionPassword adp = new AsymmDecryptionPassword();

			adp.IsDecryptDocument = false;
			adp.IgnoreRecipientIdentity = false;
			adp.RecipientIdentity = recipientIdentity;

			return adp;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Merges the session keys from the specified asymmetric decryption password with the session keys from the current
		/// object.
		/// </summary>
		/// <param name="passwords">Asymmetric decryption passwords whose session keys to merge.</param>
		/// <remarks>
		/// <para>The session keys which have the same recipient UPN fingerprint (<see cref="AsymmSessionKey.UpnFingerprint"/>)
		/// are not overwritten.</para>
		/// </remarks>
		public void MergeSessionKeys( params AsymmDecryptionPassword[] passwords )
		{
			ValidationUtils.ArrayHasContent( passwords, "passwords" );

			foreach( AsymmDecryptionPassword password in passwords ) ValidationUtils.ObjectHasContent( password, "password" );

			foreach( AsymmDecryptionPassword password in passwords ) SessionKeys.Merge( password.SessionKeys );
		}

		#endregion Public methods
	}
}
