////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Reflection;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Crypto.Symmetric;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// An asymmetrically encrypted document.
	/// </summary>
	public class AsymmEncryptedEnvelope : EncryptedEnvelope
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AsymmEncryptedEnvelope()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="compressionAlg">Compression algorithm. This may be null.</param>
		/// <param name="symmAlg">Symmetric cryptography algorithm.</param>
		/// <param name="asymmEncryptionPaddingAlg">The asymmetric encryption padding algorithm which is used to encrypt the
		/// recipient's solitary session key (<see cref="AsymmRecipientSessionKey.SolitarySessionKey"/>).</param>
		public AsymmEncryptedEnvelope( Locator instanceLocator, MacAlg macAlg, CompressionAlg compressionAlg, SymmAlg symmAlg
				, IAsymmEncryptionPaddingAlg asymmEncryptionPaddingAlg )
			: base( instanceLocator, macAlg, compressionAlg, symmAlg )
		{
			AsymmEncryptionPaddingAlg = asymmEncryptionPaddingAlg;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// For details, see <see cref="IIdentityPublisher.EncryptToGroup"/>. This is automatically set in the
		/// <see cref="GenerateSymmEncryptionSessionKey"/> method.
		/// </summary>
		private bool EncryptToGroup = false;

		#endregion Private fields

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( base.ExpirationTime, AsymmEncryptionPaddingAlg.ExpirationTime ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private IAsymmEncryptionPaddingAlg _AsymmEncryptionPaddingAlg;
		/// <summary>
		/// The asymmetric encryption padding algorithm which is used to encrypt the recipient's solitary session key
		/// (<see cref="AsymmRecipientSessionKey.SolitarySessionKey"/>). READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public IAsymmEncryptionPaddingAlg AsymmEncryptionPaddingAlg
		{
			get { return _AsymmEncryptionPaddingAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "AsymmEncryptionPaddingAlg" );

				_AsymmEncryptionPaddingAlg = value;
			}
		}

		private AsymmRecipientSessionKeys _RecipientSessionKeys;
		/// <summary>
		/// The list of session keys for all the solitary recipients of the encrypted document.
		/// </summary>
		/// <remarks>
		/// <para>The list of session keys is stored within the encrypted document. It contains a session key for each recipient
		/// of the document.</para>
		/// 
		/// <para>Each session key is generated for its intended recipient. The session key with which the private data of the
		/// document is actually encrypted, is generated from all the session keys.</para>
		/// 
		/// <para>The session keys of each recipient is encrypted with his public identity. After encryption, each session key
		/// from the list will significantly increase its size.</para>
		/// </remarks>
		[SerializeAttribute( ListItemName = "Key" )]
		public AsymmRecipientSessionKeys RecipientSessionKeys
		{
			get { return _RecipientSessionKeys; }
			set
			{
				ValidationUtils.ListHasContent( value, "RecipientSessionKeys" );

				_RecipientSessionKeys = value;
			}
		}

		#endregion Public serializable properties

		#region Group encryption (session keys)

		/// <summary>
		/// Generates the session key used for the actual symmetric encryption of the data to be encrypted. The value of the
		/// session key is random. After calling this, the developer must extract the actual session key used for the encryption
		/// of the data to be encrypted, by calling the <see cref="GetSymmEncryptionSessionKey"/> method.
		/// </summary>
		/// <param name="recipientIdentities">Identity publisher (<see cref="IIdentityPublisher"/>) which belongs to the
		/// recipient of the encrypted document.</param>
		protected override void GenerateSymmEncryptionSessionKey( object recipientIdentities )
		{
			ValidateEncryptionPassword( recipientIdentities );

			// Get the list of recipient identities.
			IIdentityPublisher recipientIdentity = (IIdentityPublisher)recipientIdentities;
			int identityCount = recipientIdentity.IdentityCount;

			#region Enforce guarantees

			EncryptToGroup = recipientIdentity.EncryptToGroup;

			if( recipientIdentity.EncryptToGroup )
			{
				if( identityCount < 2 )
				{
					throw new ValidationException( "When encrypting to a group, the number of solitary recipients must be at"
						+ " least 2." );
				}

				if( recipientIdentity.MissingGroupDecryptorCount > identityCount - 2 )
				{
					throw new ValidationException( string.Format( "The number of solitary recipients which may miss when the"
						+ " document is decrypted must be maximum {0}.", identityCount - 2 ) );
				}
			}

			#endregion Enforce guarantees

			// Create a new recipient session key list.
			AsymmRecipientSessionKeys recipientSessionKeys = new AsymmRecipientSessionKeys();

			byte[] rawGroupSessionKey = null;

			// Generate the raw group session key for all solitary recipients.
			if( !recipientIdentity.EncryptToGroup )
				rawGroupSessionKey = EntropyPool.GetBytes( SymmAlg.AlgSizeBytes );

			// Set up the session keys for the solitary recipients (but not the extra session keys).
			foreach( SolitaryIdentityPublisher identity in recipientIdentity.Identities )
			{
				// Generate the raw solitary session key for each recipient.
				byte[] rawSolitarySessionKey = EntropyPool.GetBytes( SymmAlg.AlgSizeBytes );

				// Set up a new session key for the recipient.
				AsymmSessionKey solitarySessionKey = new AsymmSessionKey( identity.Upn.Fingerprint
					, identity.Identity.GroupUpnFingerprint, rawSolitarySessionKey );

				// Generate the raw group session key for each recipient.
				if( recipientIdentity.EncryptToGroup ) rawGroupSessionKey = EntropyPool.GetBytes( SymmAlg.AlgSizeBytes );

				// Set up a new recipient session key.
				AsymmRecipientSessionKey recipientSessionKey = new AsymmRecipientSessionKey( solitarySessionKey
					, rawGroupSessionKey );

				// Add the new session key to the recipient session key list.
				recipientSessionKeys.Add( recipientSessionKey );
			}

			int sessionKeysCount = recipientSessionKeys.Count;
			if( sessionKeysCount != identityCount )
			{
				throw new ValidationException( string.Format( "The number of recipient session key list items ({0}) is not the"
					+ " same as the number of recipient identities ({1}).", sessionKeysCount, identityCount ) );
			}

			// Set up the extra session keys for each recipient (for when some solitary recipients may miss from the decryption
			// of the document).
			if( recipientIdentity.EncryptToGroup && (recipientIdentity.MissingGroupDecryptorCount > 0) )
			{
				// We need access by index to the recipient session keys.
				AsymmRecipientSessionKey[] recipientSessionKeysByIndex = recipientSessionKeys.GetItems();

				for( int i = 0; i < sessionKeysCount; i++ )
				{
					// Get the current recipient session key.
					AsymmRecipientSessionKey recipient = recipientSessionKeysByIndex[ i ];

					// Set up a new session key list for extra solitary recipients, for the current recipient.
					recipient.ExtraSessionKeys = new AsymmSessionKeys();

					// For each recipient which may miss, add his own session key to the extra session keys of the current
					// recipient.
					for( int j = 1; j <= recipientIdentity.MissingGroupDecryptorCount; j++ )
					{
						// Calculate the index of the recipient which becomes an extra recipient.
						int k = (i + j) % sessionKeysCount;

						// Create a new extra session key.
						AsymmRecipientSessionKey recipientK = recipientSessionKeysByIndex[ k ];
						AsymmSessionKey extraKey = new AsymmSessionKey( recipientK.SolitarySessionKey.UpnFingerprint
							, recipientK.SolitarySessionKey.GroupUpnFingerprint, recipientK.GroupSessionKey );

						// Add the extra session key.
						recipient.ExtraSessionKeys.Add( extraKey );
					}
				}
			}

			RecipientSessionKeys = recipientSessionKeys;
		}

		/// <summary>
		/// Returns the session key used for the actual symmetric encryption of the data to be encrypted.
		/// </summary>
		/// <returns>Returns the session key used for the actual symmetric encryption of the data to be encrypted.</returns>
		protected override byte[] GetSymmEncryptionSessionKey()
		{
			ValidationUtils.ListHasContent( RecipientSessionKeys, "RecipientSessionKeys" );

			byte[] finalSessionKey = new byte[ SymmAlg.AlgSizeBytes ];

			if( EncryptToGroup )
			{
				// The document is to be encrypted to the group of solitary recipients.

				foreach( AsymmRecipientSessionKey item in RecipientSessionKeys )
					finalSessionKey = BufferUtils.Xor( finalSessionKey, item.GroupSessionKey );
			}
			else
			{
				// The document is to be encrypted to the each recipient.

				// Since all recipient session keys are the same, use the first one.
				finalSessionKey = RecipientSessionKeys.GetFirstItem().GroupSessionKey;

				// Verify that all recipient session keys are the same. If they aren't then some solitary recipients wouldn't be
				// able to decrypt the document. If this exception is thrown, something has gone very wrong.
				foreach( AsymmRecipientSessionKey item in RecipientSessionKeys )
				{
					if( !BufferUtils.Equals( finalSessionKey, item.GroupSessionKey ) )
					{
						throw new ValidationException( "The recipient session keys are different even though the document is"
							+ " not encrypted to the group (of solitary recipients)." );
					}
				}
			}

			return finalSessionKey;
		}

		/// <summary>
		/// Returns the session key to be used for the actual symmetric decryption of the encrypted document.
		/// </summary>
		/// <param name="password">This is an <see cref="AsymmDecryptionPassword"/> object.</param>
		/// <returns>Returns the session key to be used for the actual symmetric decryption of the encrypted document.</returns>
		protected override byte[] GetSymmDecryptionSessionKey( object password )
		{
			ValidationUtils.ObjectType<AsymmDecryptionPassword>( password, "password" );

			AsymmDecryptionPassword typedPassword = (AsymmDecryptionPassword)password;

			byte[] finalSessionKey = new byte[ SymmAlg.AlgSizeBytes ];

			foreach( AsymmSessionKey sessionKey in typedPassword.SessionKeys )
				finalSessionKey = BufferUtils.Xor( finalSessionKey, sessionKey.RawSessionKey );

			return finalSessionKey;
		}

		#endregion Group encryption (session keys)

		#region Private methods

		/// <summary>
		/// Validates the identity used for encryption.
		/// </summary>
		/// <param name="identityPublisher">Identity publisher (<see cref="IIdentityPublisher"/>) which belongs to the recipient
		/// entity of the encrypted document.</param>
		protected override void ValidateEncryptionPassword( object identityPublisher )
		{
			ValidationUtils.ObjectType<IIdentityPublisher>( identityPublisher, "identityPublisher" );
		}

		/// <summary>
		/// Validates the password used for decryption.
		/// </summary>
		/// <param name="password">This must be an <see cref="AsymmDecryptionPassword"/> object.</param>
		protected override void ValidateDecryptionPassword( object password )
		{
			ValidationUtils.ObjectType<AsymmDecryptionPassword>( password, "password" );

			AsymmDecryptionPassword typedPassword = (AsymmDecryptionPassword)password;
			if( !typedPassword.IgnoreRecipientIdentity )
			{
				// Check the private decryptor identity.
				ValidationUtils.ObjectHasContent( typedPassword.RecipientIdentity, "password.RecipientIdentity" );

				// Find the recipient session key which was created for the current identity.
				AsymmRecipientSessionKey recipient = RecipientSessionKeys.Find(
					typedPassword.RecipientIdentity.Upn.Fingerprint );
				if( recipient == null )
				{
					throw new ValidationException( string.Format( "The document was not encrypted for the current recipient"
						+ " UPN: \"{0}\".", typedPassword.RecipientIdentity.Upn ) );
				}
			}
		}

		/// <summary>
		/// Asymmetrically encrypts the session key using the specified public identity.
		/// </summary>
		/// <param name="identityPublisher">Identity publisher (<see cref="IIdentityPublisher"/>) which belongs to the recipient
		/// entity of the encrypted document.</param>
		/// <remarks>All the solitary recipients of the encrypted document must be able to use the same cryptographic algorithms
		/// (which are stored in the <see cref="EncryptedEnvelope"/> class), algorithms which are specified during encryption
		/// (and stored in the encrypted document).
		/// </remarks>
		protected override void EncryptSessionKey( object identityPublisher )
		{
			ValidateEncryptionPassword( identityPublisher );

			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				// Get the list of recipient identities.
				IIdentityPublisher recipientIdentity = (IIdentityPublisher)identityPublisher;
				int identityCount = recipientIdentity.IdentityCount;
				int sessionKeyCount = RecipientSessionKeys.Count;

				if( sessionKeyCount != identityCount )
				{
					throw new ValidationException( string.Format( "The number of recipient session key list items is not the"
						+ " same as the number of recipient identities: {0} and {1}.", sessionKeyCount, identityCount ) );
				}

				foreach( AsymmRecipientSessionKey recipient in RecipientSessionKeys )
				{
					// Find the identity of the current recipient.
					SolitaryIdentityPublisher recipientIdentityPublisher = recipientIdentity.Find(
						recipient.SolitarySessionKey.UpnFingerprint );
					if( recipientIdentityPublisher == null )
					{
						throw new ValidationException( string.Format( "A recipient identity publisher for the entity with the"
							+ " UPN fingerprint \"{0}\" could not be found.", recipient.SolitarySessionKey.UpnFingerprint ) );
					}

					PublicIdentity identity = recipientIdentityPublisher.Identity;

					// The group session key and the extra session keys are symmetrically encrypted with the solitary session key
					// of the current recipient.

					// Encrypt the group session key of the current recipient. This doesn't need to use padding or chaining.
					recipient.GroupSessionKey = SymmAlg.Encrypt( recipient.GroupSessionKey, SymmPaddings.None, SymmAlg.Chaining
						, Salt, recipient.SolitarySessionKey.RawSessionKey );

					// Encrypt the extra session keys of the current recipient.
					if( recipient.ExtraSessionKeys != null )
					{
						foreach( AsymmSessionKey extraKey in recipient.ExtraSessionKeys )
						{
							// This doesn't need to use padding or chaining.
							extraKey.RawSessionKey = SymmAlg.Encrypt( extraKey.RawSessionKey, SymmPaddings.None, SymmAlg.Chaining
								, Salt, recipient.SolitarySessionKey.RawSessionKey );
						}
					}

					if( recipient.SolitarySessionKey.RawSessionKey.Length != SymmAlg.AlgSizeBytes )
					{
						throw new ValidationException( string.Format( "The size of the raw session key to encrypt is {0} instead"
							+ " of {1}.", recipient.SolitarySessionKey.RawSessionKey.Length, SymmAlg.AlgSizeBytes ) );
					}

					// The solitary session key is asymmetrically encrypted with the public identity of the current recipient.
					recipient.SolitarySessionKey.RawSessionKey = identity.EncryptData( recipient.SolitarySessionKey.RawSessionKey
						, AsymmEncryptionPaddingAlg );
				}
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		/// <summary>
		/// Asymmetrically decrypts the session key using the specified private identity.
		/// </summary>
		/// <param name="password">This is an <see cref="AsymmDecryptionPassword"/> object.</param>
		/// <returns>Returns true if the decryption must be performed, and false if the session key has been decrypted but other
		/// session keys are necessary in order to decrypt the document.</returns>
		protected override bool DecryptSessionKey( object password )
		{
			ValidateDecryptionPassword( password );

			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				AsymmDecryptionPassword typedPassword = (AsymmDecryptionPassword)password;
				if( typedPassword.IgnoreRecipientIdentity ) return true;

				// Find the recipient session key which was created for the current identity.
				AsymmRecipientSessionKey recipient = RecipientSessionKeys.Find(
					typedPassword.RecipientIdentity.Upn.Fingerprint );

				// The solitary session key is asymmetrically decrypted with the current recipient identity.
				recipient.SolitarySessionKey.RawSessionKey = typedPassword.RecipientIdentity.DecryptData(
					recipient.SolitarySessionKey.RawSessionKey, AsymmEncryptionPaddingAlg );

				if( recipient.SolitarySessionKey.RawSessionKey.Length != SymmAlg.AlgSizeBytes )
				{
					throw new ValidationException( string.Format( "The size of the decrypted raw session key is {0} instead of"
						+ " {1}.", recipient.SolitarySessionKey.RawSessionKey.Length, SymmAlg.AlgSizeBytes ) );
				}

				// The group session key and the extra session keys are symmetrically decrypted with the solitary session key of
				// the current recipient.

				// Decrypt the recipient session key of the current recipient. This doesn't need to use padding or chaining.
				recipient.GroupSessionKey = SymmAlg.Decrypt( recipient.GroupSessionKey, SymmPaddings.None, SymmAlg.Chaining, Salt
					, recipient.SolitarySessionKey.RawSessionKey );

				// Add the session key to the list.
				AsymmSessionKey sessionKey = new AsymmSessionKey( recipient.SolitarySessionKey.UpnFingerprint
					, recipient.SolitarySessionKey.GroupUpnFingerprint, recipient.GroupSessionKey );
				typedPassword.SessionKeys.Add( sessionKey );

				// Decrypt the extra session keys of the current recipient.
				if( recipient.ExtraSessionKeys != null )
				{
					foreach( AsymmSessionKey extraKey in recipient.ExtraSessionKeys )
					{
						// This doesn't need to use padding or chaining.
						extraKey.RawSessionKey = SymmAlg.Decrypt( extraKey.RawSessionKey, SymmPaddings.None, SymmAlg.Chaining
							, Salt, recipient.SolitarySessionKey.RawSessionKey );

						// Add the session key to the list.
						typedPassword.SessionKeys.Add( extraKey );
					}
				}

				return typedPassword.IsDecryptDocument;
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		#endregion Private methods

		#region Public static methods

		/// <summary>
		/// Loads the encrypted document from the specified buffer and decrypts it.
		/// </summary>
		/// <param name="buffer">Buffer from which to load the document. This may be null or empty.</param>
		/// <param name="privateIdentity">The private identity to use for the decryption of the document.</param>
		/// <param name="encryptableDocumentType">The root type of the encryptable document
		/// (<see cref="EncryptedEnvelope.EncryptableDocument"/>) to load; the document to load may either have this type or a type
		/// derived from this (if it's found in the global portable types from <see cref="PortableTypes.Instance"/>). If
		/// this is null, the type specified in the document's root node is searched in the global portable types. If the
		/// name of the document's root node is not found, an exception is thrown.</param>
		/// <param name="deserializer">Is set with the deserializer of the loaded document.</param>
		/// <returns>Returns the loaded document, or null if the input data is null or empty.</returns>
		/// <remarks>
		/// <para>The encrypted document (<see cref="EncryptedEnvelope.EncryptedDocumentValue"/>) is cleared after loading
		/// because it's not necessary anymore.</para>
		/// </remarks>
		public static AsymmEncryptedEnvelope LoadFromBuffer( byte[] buffer, object privateIdentity
			, Type encryptableDocumentType, out IDeserializer deserializer )
		{
			return EncryptedEnvelope.LoadFromBuffer( buffer, privateIdentity, typeof( AsymmEncryptedEnvelope )
				, encryptableDocumentType, out deserializer ) as AsymmEncryptedEnvelope;
		}

		/// <summary>
		/// Loads the encrypted document from the specified file and decrypts it.
		/// </summary>
		/// <param name="fileName">The name of the file from which to load the document. The file may be missing and its content
		/// may be null or empty.</param>
		/// <param name="privateIdentity">The private identity to use for the decryption of the document.</param>
		/// <param name="encryptableDocumentType">The root type of the encryptable document
		/// (<see cref="EncryptedEnvelope.EncryptableDocument"/>) to load; the document to load may either have this type or a
		/// type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <param name="deserializer">Is set with the deserializer of the loaded document.</param>
		/// <returns>Returns the loaded document, or null if the file is missing or its content is null or empty.</returns>
		/// <remarks>
		/// <para>The encrypted document (<see cref="EncryptedEnvelope.EncryptedDocumentValue"/>) is cleared after loading
		/// because it's not necessary anymore.</para>
		/// 
		/// <para>This method implements a recovery mechanism following a crash during a call of the
		/// <see cref="EncryptedEnvelope.SaveToFile"/> method.</para>
		/// </remarks>
		public static AsymmEncryptedEnvelope LoadFromFile( string fileName, object privateIdentity
			, Type encryptableDocumentType, out IDeserializer deserializer )
		{
			return EncryptedEnvelope.LoadFromFile( fileName, privateIdentity, typeof( AsymmEncryptedEnvelope )
				, encryptableDocumentType, out deserializer ) as AsymmEncryptedEnvelope;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Collects the UPN fingerprints of the encrypting entities (from the recipient session keys of the current document).
		/// </summary>
		/// <returns>Returns the UPN fingerprints of the encrypting entities. This may be empty, but not null.</returns>
		public Locators CollectEncryptingUpnFingerprints()
		{
			Locators locators = new Locators();

			foreach( AsymmRecipientSessionKey sessionKey in RecipientSessionKeys )
			{
				locators.Add( sessionKey.SolitarySessionKey.UpnFingerprint );

				// Since several keys may have the same group UPN fingerprint, allow updating.
				if( sessionKey.SolitarySessionKey.GroupUpnFingerprint != null )
					locators.Add( sessionKey.SolitarySessionKey.GroupUpnFingerprint, true );
			}

			return locators;
		}

		#endregion Public methods
	}
}
