////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// A recipient session key item for the <see cref="AsymmRecipientSessionKeys"/> class.
	/// </summary>
	public class AsymmRecipientSessionKey : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AsymmRecipientSessionKey()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="solitarySessionKey">The session key used to symmetrically encrypt the group session key and the extra
		/// session keys of the current recipient.</param>
		/// <param name="groupSessionKey">The group session key of the current recipient.</param>
		public AsymmRecipientSessionKey( AsymmSessionKey solitarySessionKey, byte[] groupSessionKey )
		{
			SolitarySessionKey = solitarySessionKey;
			GroupSessionKey = groupSessionKey;
		}

		#endregion Constructors

		#region Public serializable properties

		private AsymmSessionKey _SolitarySessionKey;
		/// <summary>
		/// The solitary session key which is used to encrypt / decrypt the group session key (<see cref="GroupSessionKey"/>) and
		/// the extra session keys of the current recipient. This key (actually, its <see cref="AsymmSessionKey.RawSessionKey"/>
		/// property) is asymmetrically encrypted with the public identity of the recipient, identity which is identified by the
		/// <see cref="AsymmSessionKey.UpnFingerprint"/> property. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This key exists because the group session key can't be used to symmetrically encrypt the extra session keys
		/// since it may be known by other recipients (which could use it to decrypt the extra session keys of the current
		/// recipient, which in turn would mean that a single recipient from the group could decrpyt a document). Therefore, it
		/// is necessary to have a session key which is know only by the current recipient.</para>
		/// 
		/// <para>This is also used when the document is encrypted to each recipient. In such a case, although the group session
		/// key would be the same for everybody, it is better to encrypt it with a solitary session key in order to not
		/// asymmetrically encrypt the same data for all recipients.</para>
		/// </remarks>
		[SerializeAttribute]
		public AsymmSessionKey SolitarySessionKey
		{
			get { return _SolitarySessionKey; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "SolitarySessionKey" );

				_SolitarySessionKey = value;
			}
		}

		private byte[] _GroupSessionKey;
		/// <summary>
		/// The group session key of the recipient (with the identity identified by the
		/// <see cref="AsymmSessionKey.UpnFingerprint"/> property) which is used to generate the encryption key for the
		/// encryption of the private data, in the <see cref="AsymmEncryptedEnvelope.GetSymmEncryptionSessionKey"/> method. This
		/// value is symmetrically encrypted with the <see cref="SolitarySessionKey"/> key.
		/// </summary>
		[SerializeAttribute]
		public byte[] GroupSessionKey
		{
			get { return _GroupSessionKey; }
			set
			{
				ValidationUtils.ArrayHasContent( value, "GroupSessionKey" );

				_GroupSessionKey = value;
			}
		}

		private AsymmSessionKeys _ExtraSessionKeys;
		/// <summary>
		/// The list of session keys of the extra recipients (for when some recipients may miss from the decryption of the
		/// document). Each session key from this list is symmetrically encrypted with the <see cref="SolitarySessionKey"/>
		/// key. This may be null because there might be no missing group decryptors (in which case there would also be no extra
		/// session key list).
		/// </summary>
		[SerializeAttribute( ListItemName = "Key", MustBeDeserialized = false )]
		public AsymmSessionKeys ExtraSessionKeys
		{
			get { return _ExtraSessionKeys; }
			set { _ExtraSessionKeys = value; }
		}

		#endregion Public serializable properties
	}
}
