////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// A session key item. See <see cref="AsymmRecipientSessionKeys"/> and <see cref="AsymmSessionKeys"/>.
	/// </summary>
	public class AsymmSessionKey : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AsymmSessionKey()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="recipientUpnFingerprint">The UPN fingerprint of the recipient of the encrypted document for which this
		/// session key is meant.</param>
		/// <param name="recipientGroupUpnFingerprint">The UPN fingerprint of the group identity publisher to which the current
		/// object belongs. This may be null.</param>
		/// <param name="rawSessionKey">The raw session key. This is stored in encrypted form. The encryption is done with the
		/// public identity of the intended recipient of the encrypted document, identity which is identified by the
		/// <see cref="UpnFingerprint"/> property.</param>
		public AsymmSessionKey( Locator recipientUpnFingerprint, Locator recipientGroupUpnFingerprint, byte[] rawSessionKey )
		{
			UpnFingerprint = recipientUpnFingerprint;
			GroupUpnFingerprint = recipientGroupUpnFingerprint;
			RawSessionKey = rawSessionKey;
		}

		#endregion Constructors

		#region Public serializable properties

		private Locator _UpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the recipient of the encrypted document for which this session key is meant. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Locator UpnFingerprint
		{
			get { return _UpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "UpnFingerprint" );

				_UpnFingerprint = value;
			}
		}

		private Locator _GroupUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the group identity publisher to which the current object belongs. This may be null. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public Locator GroupUpnFingerprint
		{
			get { return _GroupUpnFingerprint; }
			protected set { _GroupUpnFingerprint = value; }
		}

		private byte[] _RawSessionKey;
		/// <summary>
		/// The raw session key. This is stored either symmetrically or asymmetrically encrypted.
		/// </summary>
		[SerializeAttribute]
		public byte[] RawSessionKey
		{
			get { return _RawSessionKey; }
			set
			{
				ValidationUtils.ArrayHasContent( value, "RawSessionKey" );

				_RawSessionKey = value;
			}
		}

		#endregion Public serializable properties
	}
}
