////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Reflection;
using Vela.Crypto;
using Vela.Crypto.Mac;
using Vela.Crypto.Symmetric;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Contract for an encrypted envelope / document. An encrypted envelope / document is designed to be used as an envelope
	/// for documents, not as a document in itself.
	/// </summary>
	/// <remarks>
	/// <para><b>Buffer size</b></para>
	/// 
	/// <para>A buffer stored in an encrypted document serialized with a textual serializer, increases its size with about
	/// 80%. This is because a buffer is stored in a 6 bit character set instead of the normal 8 bit (= 1 byte) character
	/// set.</para>
	/// 
	/// <para>You might notice that 8 / 6 is actually 1.33 (not 1.80 as implied above). The difference comes from the fact that
	/// the object which is stored in the encrypted document (<see cref="EncryptedEnvelope.EncryptableDocument"/>) is first
	/// serialized in a textual form (which increases the size of the buffer with a factor of 8 / 6) which is encrypted to a
	/// buffer (<see cref="EncryptedEnvelope.EncryptedDocumentValue"/>) which is then again serialized with the envelope
	/// (which increases again the size of the buffer with a factor of 8 / 6). So, the size of the buffer actually
	/// increases (8 / 6) * (8 / 6).</para>
	/// </remarks>
	public abstract class EncryptedEnvelope : ILocatable, IExpirableAlgorithm, IEncryptableDocument
	{
		#region Public constants

		/// <summary>
		/// The value above which the block size (in bytes) of the data has meaning. If the specified block size is bigger than
		/// this value, the size of the encrypted document is aligned to that value. For usage, see the
		/// <see cref="AlignToBlockSize"/> method.
		/// </summary>
		/// <remarks>
		/// <para>This must be at least the size of the type of the <see cref="EncryptedDocumentBlockSize"/> property
		/// (<see cref="int"/>), in bytes. Basically, this represents the space required to store the block size value.</para>
		/// </remarks>
		public const int BlockSizeSize = 4;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected EncryptedEnvelope()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="compressionAlg">Compression algorithm. This may be null.</param>
		/// <param name="symmAlg">Symmetric cryptography algorithm.</param>
		protected EncryptedEnvelope( Locator instanceLocator, MacAlg macAlg, CompressionAlg compressionAlg, SymmAlg symmAlg )
		{
			InstanceLocator = instanceLocator;
			MacAlg = macAlg;
			CompressionAlg = compressionAlg;
			SymmAlg = symmAlg;
		}

		#endregion Constructors

		#region Session keys

		/// <summary>
		/// Generates the session key used for the actual symmetric encryption of the data to be encrypted. The value of the
		/// session key is random. After calling this, the developer must extract the actual session key used for the encryption
		/// of the data to be encrypted, by calling the <see cref="GetSymmEncryptionSessionKey"/> method.
		/// </summary>
		/// <param name="password"><para>For symmetric encryption, this is the password to use for the encryption of the
		/// data. This may be a buffer or a text.</para>
		/// 
		/// <para>In the case of asymmetric encryption, this is the identity publisher (<see cref="IIdentityPublisher"/>), which
		/// is not a password, to use for the encryption of the data.</para></param>
		/// <remarks>
		/// <para>The session key may be encrypted in two ways: symmetrically and asymmetrically.</para>
		/// 
		/// <para>In the case of symmetric encryption, a blended password derived from the user's password is used to
		/// encrypt this key.</para>
		/// 
		/// <para>In the case of asymmetric encryption, instead of this key, a list of session keys, generated for each recipient
		/// of the encrypted document, is stored. The session key with which the encryptable document of the envelope is actually
		/// encrypted, is generated from all the session keys.</para>
		/// 
		/// <para>The session keys of each recipient is encrypted with his public identity. After encryption, each session key
		/// will significantly increase its size.</para>
		/// </remarks>
		protected abstract void GenerateSymmEncryptionSessionKey( object password );

		/// <summary>
		/// Returns the session key used for the actual symmetric encryption of the data to be encrypted.
		/// </summary>
		/// <returns>Returns the session key used for the actual symmetric encryption of the data to be encrypted.</returns>
		/// <remarks>
		/// <para>If session keys were generated, they must be put in an list property because they must be stored; thus, they
		/// don't have to be sent in a parameter.</para>
		/// </remarks>
		protected abstract byte[] GetSymmEncryptionSessionKey();

		/// <summary>
		/// Returns the session key to be used for the actual symmetric decryption of the encrypted document.
		/// </summary>
		/// <param name="password">For symmetric decryption, this is not used. In the case of asymmetric decryption, this is an
		/// <see cref="AsymmDecryptionPassword"/> object.</param>
		/// <returns>Returns the session key to be used for the actual symmetric decryption of the encrypted document.</returns>
		protected abstract byte[] GetSymmDecryptionSessionKey( object password );

		#endregion Session keys

		#region Public non-serializable properties

		private IEncryptableDocument _EncryptableDocument;
		/// <summary>
		/// The document which is to be encrypted. This object is not stored (except when compiled in debug mode), but is
		/// encrypted and stored in a separate object: <see cref="EncryptedDocumentValue"/>.
		/// </summary>
		/// <remarks>
		/// <para>After the document is encrypted (with the <see cref="EncryptEncryptableDocument"/> method) or saved (with a
		/// method like <see cref="SaveToBuffer"/>), this isn't null. Also, if this is null when a document is loaded, an
		/// exception is thrown.</para>
		/// </remarks>
#if DEBUG && SERIALIZEEXPOSEDATTRIBUTE
		[SerializeExposedAttribute]
#endif
		public IEncryptableDocument EncryptableDocument
		{
			get { return _EncryptableDocument; }
			set
			{
				ValidationUtils.ObjectHasContent( value, "EncryptableDocument" );

				_EncryptableDocument = value;
			}
		}

		private byte[] _PriorEncryptedDocumentMacValue;
		/// <summary>
		/// The unencrypted MAC of the encrypted document of the current object. This is relevant only immediately after a
		/// document's encryptable document (<see cref="EncryptableDocument"/>) is encrypted
		/// (<see cref="EncryptEncryptableDocument"/>), like when the current envelope is saved with the
		/// <see cref="SaveToBuffer"/> method. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>You might need this value to compare it with the unencrypted MAC (<see cref="EncryptedDocumentMacValue"/>) of
		/// a loaded document. If the MACs are the same, the objects are the same (at least from a cryptographic point of
		/// view).</para>
		/// </remarks>
		public byte[] PriorEncryptedDocumentMacValue
		{
			get { return _PriorEncryptedDocumentMacValue; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "PriorEncryptedDocumentMacValue" );

				_PriorEncryptedDocumentMacValue = value;
			}
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public virtual DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( MacAlg.ExpirationTime, SymmAlg.ExpirationTime ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Locator _InstanceLocator;
		/// <summary>
		/// The locator of the current document. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Locator InstanceLocator
		{
			get { return _InstanceLocator; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "InstanceLocator" );

				_InstanceLocator = value;
			}
		}

		private byte[] _Salt;
		/// <summary>
		/// The (random) salt used to compute the MAC and encrypt the data. This is stored in unencrypted form. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is used for the:</para>
		/// <list type="bullet">
		///		<item>Derivation of the blended password from the user's password.</item>
		///		<item>Encryption of the session keys (see the <see cref="GetSymmEncryptionSessionKey"/> method) used for the
		///			symmetric encryption of the encryptable document (<see cref="EncryptableDocument"/>).</item>
		///		<item>Encryption of the encryptable document.</item>
		/// </list>
		/// </remarks>
		[SerializeAttribute]
		public byte[] Salt
		{
			get { return _Salt; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "Salt" );

				_Salt = value;
			}
		}

		#region Algorithms

		private MacAlg _MacAlg;
		/// <summary>
		/// MAC algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public MacAlg MacAlg
		{
			get { return _MacAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "MacAlg" );

				_MacAlg = value;
			}
		}

		private CompressionAlg _CompressionAlg;
		/// <summary>
		/// Compression algorithm. This may be null. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is not null, the encryptable document is compressed before it is encrypted, fact which improves the
		/// security and may reduce the required storage space and network bandwidth.</para>
		/// 
		/// <para>Compressing the encryptable document (<see cref="EncryptableDocument"/>) when it has a block size
		/// (<see cref="EncryptedDocumentBlockSize"/>) may decrease the size of the encrypted document
		/// (<see cref="EncryptedDocumentValue"/>) only insignificantly.</para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false )]
		public CompressionAlg CompressionAlg
		{
			get { return _CompressionAlg; }
			protected set { _CompressionAlg = value; }
		}

		private SymmAlg _SymmAlg;
		/// <summary>
		/// Symmetric cryptography algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public SymmAlg SymmAlg
		{
			get { return _SymmAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "SymmAlg" );

				_SymmAlg = value;
			}
		}

		#endregion Algorithms

		private int _EncryptedDocumentBlockSize = 0;
		/// <summary>
		/// The size (in bytes) to which the encrypted document (<see cref="EncryptedDocumentValue"/>) is aligned. Set this to a
		/// value bigger than <see cref="BlockSizeSize"/> when you want to hide the real size of the encrypted document. For
		/// usage, see the <see cref="AlignToBlockSize"/> method.
		/// </summary>
		/// <remarks>
		/// <para>Compressing the encryptable data (<see cref="EncryptableDocument"/>) when it has a block size
		/// (<see cref="EncryptedDocumentBlockSize"/>) may decrease the size of the encrypted document
		/// (<see cref="EncryptedDocumentValue"/>) only insignificantly.</para>
		/// </remarks>
		[SerializeAttribute]
		public int EncryptedDocumentBlockSize
		{
			get { return _EncryptedDocumentBlockSize; }
			set { _EncryptedDocumentBlockSize = value; }
		}

		private byte[] _EncryptedDocumentMacValue;
		/// <summary>
		/// The encrypted or unencrypted MAC of the encrypted document of the current object. This is stored symmetrically
		/// encrypted with the session key. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The encryption of this property isn't technically necessary.</para>
		/// 
		/// <para>After the current document is saved (for example with the <see cref="SaveToBuffer"/> method), this
		/// contains the <b>encrypted</b> MAC of the encrypted document (<see cref="EncryptableDocument"/>). The actual
		/// encryption is made in the <see cref="EncryptEncryptableDocument"/> method. The unencrypted MAC is in
		/// <see cref="PriorEncryptedDocumentMacValue"/>.</para>
		/// 
		/// <para>After the current document is loaded (for example with the <see cref="LoadFromBuffer"/> method), this
		/// contains the <b>unencrypted</b> MAC of the encrypted document (<see cref="EncryptableDocument"/>). The actual
		/// decryption is made in the <see cref="DecryptEncryptableDocument"/> method.</para>
		/// 
		/// <para>This can be used during decryption to verify that the specified password is the same as the one used for
		/// encryption, and that the data to be decrypted is not corrupted.</para>
		/// 
		/// <para>This value is authoritative only if the encrypted envelope is loaded and no exception is thrown.</para>
		/// </remarks>
		[SerializeAttribute]
		public byte[] EncryptedDocumentMacValue
		{
			get { return _EncryptedDocumentMacValue; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "EncryptedDocumentMacValue" );

				_EncryptedDocumentMacValue = value;
			}
		}

		private byte[] _EncryptedDocumentValue;
		/// <summary>
		/// The encrypted document. When the current object is saved, the encryptable document
		/// (<see cref="EncryptableDocument"/>) is serialized and then encrypted with the session key for symmetric encryption
		/// (see the <see cref="GetSymmEncryptionSessionKey"/> method). READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] EncryptedDocumentValue
		{
			get { return _EncryptedDocumentValue; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "EncryptedDocumentValue" );

				_EncryptedDocumentValue = value;
			}
		}

		private IUnencryptedDocument _UnencryptedDocument;
		/// <summary>
		/// Document which may be attached to the encrypted envelope, but is not encrypted. This may be null.
		/// </summary>
		/// <remarks>
		/// <para>This object should contain information which refers to the manipulation and storage of the encrypted envelope
		/// rather than to the content of the document.</para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false )]
		public IUnencryptedDocument UnencryptedDocument
		{
			get { return _UnencryptedDocument; }
			set { _UnencryptedDocument = value; }
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Aligns the specified data to the specified block size.
		/// </summary>
		/// <param name="buffer">The data to align to the specified block size.</param>
		/// <param name="blockSize">The block size (in bytes) to which to align the specified data. If this is smaller than (or
		/// equal with) <see cref="BlockSizeSize"/>, no aligning is done. If you want to hide the real size of the encrypted
		/// document, specify a value bigger than <see cref="BlockSizeSize"/>.</param>
		/// <returns>Returns the aligned buffer.</returns>
		/// <remarks>
		/// <para>When data is aligned to the specified block size, the remaining space up to the specified block size is set
		/// with random data.</para>
		/// </remarks>
		private static byte[] AlignToBlockSize( byte[] buffer, int blockSize )
		{
			ValidationUtils.ObjectHasContent( buffer, "buffer" );

			if( blockSize <= BlockSizeSize )
				return buffer;

			int alignedSize = MathUtils.AlignUp( buffer.Length + BlockSizeSize, blockSize );

			// Allocate the buffer with the aligned size and copy the specified data in it.
			byte[] alignedData = new byte[ alignedSize ];
			Buffer.BlockCopy( buffer, 0, alignedData, 0, buffer.Length );

			// Write the random data, to fill the capacity (except the space for the size).
			int randSize = alignedSize - BlockSizeSize - buffer.Length;
			if( randSize > 0 )
			{
				byte[] randData = EntropyPool.GetBytes( randSize );
				Buffer.BlockCopy( randData, 0, alignedData, buffer.Length, randData.Length );
			}

			// Write the real size of the content (at the end of the aligned buffer).
			byte[] realSizeBuffer = BitConverter.GetBytes( buffer.Length );
			Buffer.BlockCopy( realSizeBuffer, 0, alignedData, alignedData.Length - BlockSizeSize, realSizeBuffer.Length );

			return alignedData;
		}

		/// <summary>
		/// Unpad the aligned buffer which was set up (during saving) with the <see cref="AlignToBlockSize"/> method.
		/// </summary>
		/// <param name="alignedBuffer">The aligned buffer whose size to unpad.</param>
		/// <param name="blockSize">The size (in bytes) to which the data was aligned. If this is smaller than (or equal with)
		/// <see cref="BlockSizeSize"/>, no unpadding is done.</param>
		/// <returns>Returns the unpadded buffer.</returns>
		private static byte[] UnpadBlockSizeData( byte[] alignedBuffer, int blockSize )
		{
			ValidationUtils.ObjectHasContent( alignedBuffer, "alignedBuffer" );

			if( blockSize <= BlockSizeSize )
				return alignedBuffer;

			if( alignedBuffer.Length % blockSize != 0 )
			{
				throw new AlignedToBlockSizeException( string.Format( "The size of the data to unpad using a block size must be"
					+ " aligned to {0}; {1} is not aligned.", blockSize, alignedBuffer.Length ) );
			}

			// Get the real / unpadded size of the content.
			int realSize = BitConverter.ToInt32( alignedBuffer, alignedBuffer.Length - BlockSizeSize );

			byte[] unpaddedData = new byte[ realSize ];

			// Write the unpadded content to the result buffer.
			Buffer.BlockCopy( alignedBuffer, 0, unpaddedData, 0, realSize );

			return unpaddedData;
		}

		/// <summary>
		/// Validates the password used for encryption.
		/// </summary>
		/// <param name="password"><para>For symmetric encryption, this is the password to use for the encryption of the
		/// data. This may be a buffer or a text.</para>
		/// 
		/// <para>For asymmetric encryption, this is an identity publisher (<see cref="IIdentityPublisher"/>), which is not a
		/// password, which belongs to the recipient entity of the encrypted document.</para></param>
		protected abstract void ValidateEncryptionPassword( object password );

		/// <summary>
		/// Validates the password used for decryption.
		/// </summary>
		/// <param name="password"><para>For symmetric decryption, this is the password to use for the decryption of the
		/// data. This may be a buffer or a text.</para>
		/// 
		/// <para>For asymmetric decryption, this is an <see cref="AsymmDecryptionPassword"/> object.</para></param>
		protected abstract void ValidateDecryptionPassword( object password );

		/// <summary>
		/// Encrypts the session key.
		/// </summary>
		/// <param name="password"><para>For symmetric encryption, this is the password to use for the encryption of the
		/// data. This may be a buffer or a text.</para>
		/// 
		/// <para>In the case of asymmetric encryption, this is an identity publisher (<see cref="IIdentityPublisher"/>), which
		/// is not a password, which belongs to the recipient entity of the encrypted document.</para></param>
		protected abstract void EncryptSessionKey( object password );

		/// <summary>
		/// Decrypts the session key.
		/// </summary>
		/// <param name="password"><para>For symmetric decryption, this is the password to use for the decryption of the
		/// data. This may be a buffer or a text.</para>
		/// 
		/// <para>In the case of asymmetric decryption, this is the private identity (<see cref="PrivateIdentity"/>) of the
		/// recipient of the encrypted document (which is the complement of one of the solitary identity publishers which were
		/// used to encrypt the document.</para></param>
		/// <returns>Returns true if the decryption must be performed, and false if the session key has been decrypted but other
		/// session keys are necessary in order to decrypt the document.</returns>
		protected abstract bool DecryptSessionKey( object password );

		/// <summary>
		/// Validates the current object.
		/// </summary>
		protected virtual void Validate()
		{
			if( Salt.Length != SymmAlg.AlgSizeBytes )
			{
				throw new ValidationException( string.Format( "The size of the salt is {0} bytes instead of {1} bytes."
					, Salt.Length, SymmAlg.AlgSizeBytes ) );
			}
		}

		/// <summary>
		/// This method is automatically called before the current object is serialized.
		/// </summary>
		protected virtual void Serializing()
		{
			Validate();
		}

		/// <summary>
		/// This method is automatically called after the current object has been deserialized.
		/// </summary>
		protected virtual void Deserialized()
		{
			Validate();
		}

		#endregion Private methods

		#region Public static methods

		/// <summary>
		/// Loads the encrypted document from the specified encrypted envelope and decrypts it into
		/// <see cref="EncryptableDocument"/>.
		/// </summary>
		/// <param name="envelope">Encrypted envelope to load object from. This may be null.</param>
		/// <param name="password"><para>For symmetric decryption, this is the password to use for the decryption of the
		/// data. This may be a buffer or a text.</para>
		/// <para>In the case of asymmetric decryption, this is the private identity (<see cref="PrivateIdentity"/>) to use for
		/// the decryption of the data.</para></param>
		/// <param name="encryptableDocumentType">The root type of the encryptable document
		/// (<see cref="EncryptedEnvelope.EncryptableDocument"/>) to load; the document to load may either have this type or a
		/// type derived from this (if it's found in the global portable types from <see cref="PortableTypes.Instance"/>). If
		/// this is null, the type specified in the document's root node is searched in the global portable types. If the
		/// name of the document's root node is not found, an exception is thrown.</param>
		/// <param name="deserializer">Is set with the deserializer of the loaded document.</param>
		/// <remarks>
		/// <para>The encrypted document (<see cref="EncryptedDocumentValue"/>) is cleared after loading because it's not
		/// necessary anymore.</para>
		/// </remarks>
		public static void LoadFromEnvelope( EncryptedEnvelope envelope, object password, Type encryptableDocumentType
			, out IDeserializer deserializer )
		{
			if( envelope == null )
			{
				deserializer = null;

				return;
			}

			envelope.DecryptEncryptableDocument( encryptableDocumentType, password, out deserializer );

			// Clear the encrypted document because it's not necessary anymore.
			envelope.ClearEncryptedDocumentValue();
		}

		/// <summary>
		/// Loads the encrypted document from the specified buffer and decrypts it.
		/// </summary>
		/// <param name="buffer">Buffer from which to load the document. This may be null or empty.</param>
		/// <param name="password"><para>For symmetric decryption, this is the password to use for the decryption of the
		/// data. This may be a buffer or a text.</para>
		/// <para>In the case of asymmetric decryption, this is the private identity (<see cref="PrivateIdentity"/>) to use for
		/// the decryption of the data.</para></param>
		/// <param name="documentType">The root type of the document to load; the document to load may either have this type or
		/// a type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <param name="encryptableDocumentType">The root type of the encryptable document
		/// (<see cref="EncryptedEnvelope.EncryptableDocument"/>) to load; the document to load may either have this type or a
		/// type derived from this (if it's found in the global portable types from <see cref="PortableTypes.Instance"/>). If
		/// this is null, the type specified in the document's root node is searched in the global portable types. If the
		/// name of the document's root node is not found, an exception is thrown.</param>
		/// <param name="deserializer">Is set with the deserializer of the loaded document, or null if the data is null or
		/// empty.</param>
		/// <returns>Returns the loaded document, or null if the input data is null or empty.</returns>
		/// <remarks>
		/// <para>The encrypted document (<see cref="EncryptedDocumentValue"/>) is cleared after loading because it's not
		/// necessary anymore.</para>
		/// </remarks>
		public static EncryptedEnvelope LoadFromBuffer( byte[] buffer, object password, Type documentType
			, Type encryptableDocumentType, out IDeserializer deserializer )
		{
			EncryptedEnvelope envelope = SerializationUtils.LoadFromBuffer( buffer, documentType, out deserializer )
				as EncryptedEnvelope;

			IDeserializer encryptedContentDeserializer;
			LoadFromEnvelope( envelope, password, encryptableDocumentType, out encryptedContentDeserializer );

			return envelope;
		}

		/// <summary>
		/// Loads the encrypted document from the specified file and decrypts it.
		/// </summary>
		/// <param name="fileName">The name of the file from which to load the document. The file may be missing and its content
		/// may be null or empty.</param>
		/// <param name="password"><para>For symmetric decryption, this is the password to use for the decryption of the
		/// data. This may be a buffer or a text.</para>
		/// <para>In the case of asymmetric decryption, this is the private identity (<see cref="PrivateIdentity"/>) to use for
		/// the decryption of the data.</para></param>
		/// <param name="documentType">The root type of the document to load; the document to load may either have this type or
		/// a type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <param name="encryptableDocumentType">The root type of the encryptable document
		/// (<see cref="EncryptedEnvelope.EncryptableDocument"/>) to load; the document to load may either have this type or a
		/// type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <param name="deserializer">Is set with the deserializer of the loaded document, or null if the data is null or
		/// empty.</param>
		/// <returns>Returns the loaded document, or null if the file is missing or its content is null or empty.</returns>
		/// <remarks>
		/// <para>The encrypted document (<see cref="EncryptedDocumentValue"/>) is cleared after loading because it's not
		/// necessary anymore.</para>
		/// 
		/// <para>This method implements a recovery mechanism following a crash during a call of the
		/// <see cref="SaveToFile"/> method.</para>
		/// </remarks>
		public static EncryptedEnvelope LoadFromFile( string fileName, object password, Type documentType
			, Type encryptableDocumentType, out IDeserializer deserializer )
		{
			EncryptedEnvelope envelope = SerializationUtils.LoadFromFile( fileName, documentType, out deserializer )
				as EncryptedEnvelope;

			IDeserializer encryptedContentDeserializer;
			LoadFromEnvelope( envelope, password, encryptableDocumentType, out encryptedContentDeserializer );

			return envelope;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Clears the encrypted document. This is normally called after the encrypted document has been decrypted.
		/// </summary>
		public void ClearEncryptedDocumentValue()
		{
			_EncryptedDocumentValue = null;
		}

		/// <summary>
		/// Encrypts the encryptable document (<see cref="EncryptableDocument"/>).
		/// </summary>
		/// <param name="salt">The (random) salt used to compute the MAC and encrypt the data. This is stored in unencrypted
		/// form. If this is null or empty, a random salt is used instead.</param>
		/// <param name="password">
		/// <para>For symmetric encryption, this is the password to use for the encryption of the data. This may be a buffer
		/// or a text.</para>
		/// <para>In the case of asymmetric encryption, this is an identity publisher (<see cref="IIdentityPublisher"/>), which
		/// is not a password, which belongs to the recipient entity of the encrypted document.</para></param>
		/// <param name="serializer">The serializer of the encryptable document (<see cref="EncryptableDocument"/>) of the
		/// current document.</param>
		public virtual void EncryptEncryptableDocument( byte[] salt, object password, ISerializer serializer )
		{
			// Validate the password.
			ValidateEncryptionPassword( password );

			ValidationUtils.ObjectHasContent( serializer, "serializer" );

			ValidationUtils.ObjectHasContent( EncryptableDocument, "EncryptableDocument" );

			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				// Generate the salt.
				Salt = BufferUtils.HasContent( salt ) ? salt : EntropyPool.GetBytes( SymmAlg.AlgSizeBytes );

				// Generate the session key which is to be used to encrypt the encryptable data.
				GenerateSymmEncryptionSessionKey( password );

				// Get the current session and MAC keys.
				byte[] sessionKey = GetSymmEncryptionSessionKey();
				byte[] macKey = MacAlg.HashAlg.Compute( null, true, Salt, sessionKey );

				#region Save the encryptable document in a buffer

				// Serialize the encryptable document.
				byte[] encryptableDocument = serializer.SaveToBuffer( EncryptableDocument, null );

				// Compress the serialized data.
				if( CompressionAlg != null )
					encryptableDocument = CompressionAlg.Compress( encryptableDocument );

				// Align the data to a block size.
				encryptableDocument = AlignToBlockSize( encryptableDocument, EncryptedDocumentBlockSize );

				#endregion Save the encryptable document in a buffer

				// Encrypt the encryptable document.
				EncryptedDocumentValue = SymmAlg.Encrypt( encryptableDocument, Salt, sessionKey );

				// Set the MAC of the encrypted document.
				EncryptedDocumentMacValue = MacAlg.Compute( macKey, null, true, Salt, EncryptedDocumentValue );

				// Backup the unencrypted MAC.
				PriorEncryptedDocumentMacValue = EncryptedDocumentMacValue;

				// Encrypt the MAC of the encrypted document.
				EncryptedDocumentMacValue = SymmAlg.Encrypt( EncryptedDocumentMacValue, Salt, SymmAlg.AlignToAlgSize( macKey ) );

				// Encrypt the session key.
				EncryptSessionKey( password );
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		/// <summary>
		/// Decrypts the encrypted document (<see cref="EncryptedDocumentValue"/>). If the information can't be decrypted, an
		/// exception is thrown.
		/// </summary>
		/// <param name="encryptableDocumentType">The root type of the encryptable document
		/// (<see cref="EncryptedEnvelope.EncryptableDocument"/>) to load; the document to load may either have this type or a
		/// type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <param name="password"><para>For symmetric decryption, this is the password to use for the decryption of the
		/// data. This may be a buffer or a text.</para>
		/// 
		/// <para>In the case of asymmetric decryption, this is the private identity (<see cref="PrivateIdentity"/>) to use for
		/// the decryption of the data.</para></param>
		/// <param name="deserializer">Is set with the deserializer of the (decrypted) encryptable document
		/// (<see cref="EncryptableDocument"/>) of the current document. If the returned value is false, this is null.</param>
		/// <returns>Returns true if the data has been decrypted. There are cases when the decryption is not performed (when
		/// there is no data to decrypt) but no exception is thrown. In such cases, the returned value is false.</returns>
		public virtual bool DecryptEncryptableDocument( Type encryptableDocumentType, object password
			, out IDeserializer deserializer )
		{
			// Validate the password.
			ValidateDecryptionPassword( password );

			ValidationUtils.ArrayHasContent( EncryptedDocumentValue, "EncryptedDocumentValue" );

			// Check if the decryption of the session keys must continue.
			if( !DecryptSessionKey( password ) )
			{
				deserializer = null;

				return false;
			}

			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				// Can't set the property with null, so set the field.
				_PriorEncryptedDocumentMacValue = null;

				// Get the current session and MAC keys.
				byte[] sessionKey = GetSymmDecryptionSessionKey( password );
				byte[] macKey = MacAlg.HashAlg.Compute( null, true, Salt, sessionKey );

				// Decrypt the MAC of the encrypted document.
				EncryptedDocumentMacValue = SymmAlg.Decrypt( EncryptedDocumentMacValue, Salt, SymmAlg.AlignToAlgSize( macKey ) );
				ValidationUtils.ArrayHasContent( EncryptedDocumentMacValue, "EncryptedDocumentMacValue" );

				// Verify the MAC of the encrypted document.
				byte[] mac = MacAlg.Compute( macKey, null, true, Salt, EncryptedDocumentValue );
				if( !BufferUtils.Equals( mac, EncryptedDocumentMacValue ) )
					throw new ValidationException( "The MAC of the document to decrypt is different than the stored MAC." );

				// Decrypt the encryptable document.
				byte[] decryptedData = SymmAlg.Decrypt( EncryptedDocumentValue, Salt, sessionKey );
				ValidationUtils.ArrayHasContent( decryptedData, "decryptedData" );

				// Unpad the aligned buffer.
				decryptedData = UnpadBlockSizeData( decryptedData, EncryptedDocumentBlockSize );

				// Decompress the (unpadded) data.
				if( CompressionAlg != null )
					decryptedData = CompressionAlg.Decompress( decryptedData );

				// Deserialize.
				IEncryptableDocument encryptableDocument = SerializationUtils.LoadFromBuffer( decryptedData
					, encryptableDocumentType, out deserializer ) as IEncryptableDocument;
				ValidationUtils.ObjectHasContent( encryptableDocument, "encryptableDocument" );

				EncryptableDocument = encryptableDocument;

				return true;
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		/// <summary>
		/// Saves the current object to a buffer.
		/// </summary>
		/// <param name="salt">The (random) salt used to compute the MAC and encrypt the data. This is stored in unencrypted
		/// form. If this is null or empty, a random salt is used instead.</param>
		/// <param name="password"><para>For symmetric encryption, this is the password to use for the encryption of the
		/// data. This may be a buffer or a text.</para>
		/// <para>In the case of asymmetric encryption, this is an identity publisher (<see cref="IIdentityPublisher"/>), which
		/// is not a password, which belongs to the recipient entity of the encrypted document.</para></param>
		/// <param name="serializer">The serializer of the document.</param>
		/// <returns>Returns the serialized document.</returns>
		/// <remarks>
		/// <para>The encrypted document (<see cref="EncryptedDocumentValue"/>) is cleared after saving because it's not
		/// necessary anymore.</para>
		/// </remarks>
		public virtual byte[] SaveToBuffer( byte[] salt, object password, ISerializer serializer )
		{
			EncryptEncryptableDocument( salt, password, serializer );

			byte[] buffer = serializer.SaveToBuffer( this, null );

			// Clear the encrypted document because it's not necessary anymore.
			ClearEncryptedDocumentValue();

			return buffer;
		}

		/// <summary>
		/// Saves the current object to the specified file.
		/// </summary>
		/// <param name="fileName">File to save the object to.</param>
		/// <param name="salt">The (random) salt used to compute the MAC and encrypt the data. This is stored in unencrypted
		/// form. If this is null or empty, a random salt is used instead.</param>
		/// <param name="password"><para>For symmetric encryption, this is the password to use for the encryption of the
		/// data. This may be a buffer or a text.</para>
		/// <para>In the case of asymmetric encryption, this is an identity publisher (<see cref="IIdentityPublisher"/>), which
		/// is not a password, which belongs to the recipient entity of the encrypted document.</para></param>
		/// <param name="serializer">The serializer of the document.</param>
		/// <remarks>
		/// <para>The encrypted document (<see cref="EncryptedDocumentValue"/>) is cleared after saving because it's not
		/// necessary anymore.</para>
		/// 
		/// <para>The saving is crash-safe. If the application crashes when the file is being saved, at least 1 of the
		/// following files is still available: a file with the specified name, a file with a backup name, a file with a
		/// temporary name. If the backup file exists, it's the valid one and you should implement your own recovery mechanism
		/// following a crash.</para>
		/// </remarks>
		public virtual void SaveToFile( string fileName, byte[] salt, object password, ISerializer serializer )
		{
			EncryptEncryptableDocument( salt, password, serializer );

			serializer.SaveToFile( this, null, fileName );

			// Clear the encrypted document because it's not necessary anymore.
			ClearEncryptedDocumentValue();
		}

		#endregion Public methods
	}
}
