////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto;
using Vela.Crypto.Mac;
using Vela.Crypto.Symmetric;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// A symmetrically encrypted document.
	/// </summary>
	public class SymmEncryptedEnvelope : EncryptedEnvelope
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected SymmEncryptedEnvelope()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="compressionAlg">Compression algorithm. This may be null.</param>
		/// <param name="symmAlg">Symmetric cryptography algorithm.</param>
		/// <param name="passwordBlenderAlg">Password blender algorithm to use to generate the blended password. This may
		/// be null if the password is not textual.</param>
		public SymmEncryptedEnvelope( Locator instanceLocator, MacAlg macAlg, CompressionAlg compressionAlg, SymmAlg symmAlg
				, PasswordBlenderAlg passwordBlenderAlg )
			: base( instanceLocator, macAlg, compressionAlg, symmAlg )
		{
			PasswordBlenderAlg = passwordBlenderAlg;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private byte[] _BlendedPassword;
		/// <summary>
		/// The blended password which is used to symmetrically encrypt the session key (<see cref="SymmEncryptionSessionKey"/>).
		/// READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is a cache used during saving. When the file is saved, the blended password is computed. Then, this
		/// is used to load the file in order to verify that it was properly saved.</para>
		/// </remarks>
		private byte[] BlendedPassword
		{
			get { return _BlendedPassword; }
			set
			{
				ValidationUtils.ArrayHasContent( value, "BlendedPassword" );

				_BlendedPassword = value;
			}
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( base.ExpirationTime, PasswordBlenderAlg.ExpirationTime ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private PasswordBlenderAlg _PasswordBlenderAlg;
		/// <summary>
		/// Password blender algorithm to use to generate the blended password. This may be null if the password is not
		/// textual. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public PasswordBlenderAlg PasswordBlenderAlg
		{
			get { return _PasswordBlenderAlg; }
			protected set { _PasswordBlenderAlg = value; }
		}

		private byte[] _SymmEncryptionSessionKey;
		/// <summary>
		/// The session key used for the actual symmetric encryption of the data to be encrypted. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The value of this key is random.</para>
		/// 
		/// <para>This key is stored in encrypted form.</para>
		/// 
		/// <para>A blended password derived from the user's password is used to encrypt this key.</para>
		/// </remarks>
		[SerializeAttribute]
		public byte[] SymmEncryptionSessionKey
		{
			get { return _SymmEncryptionSessionKey; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "SymmEncryptionSessionKey" );

				_SymmEncryptionSessionKey = value;
			}
		}

		#endregion Public serializable properties

		#region Session keys

		/// <summary>
		/// Generates the session key used for the actual symmetric encryption of the data to be encrypted. The value of the
		/// session key is random. After calling this, the developer must extract the actual session key used for the encryption
		/// of the data to be encrypted, by calling the <see cref="GetSymmEncryptionSessionKey"/> method.
		/// </summary>
		/// <param name="password">This is not used.</param>
		protected override void GenerateSymmEncryptionSessionKey( object password )
		{
			SymmEncryptionSessionKey = EntropyPool.GetBytes( SymmAlg.AlgSizeBytes );
		}

		/// <summary>
		/// Returns the session key used for the actual symmetric encryption of the data to be encrypted.
		/// </summary>
		/// <returns>Returns the session key used for the actual symmetric encryption of the data to be encrypted.</returns>
		protected override byte[] GetSymmEncryptionSessionKey()
		{
			return SymmEncryptionSessionKey;
		}

		/// <summary>
		/// Returns the session key to be used for the actual symmetric decryption of the encrypted document.
		/// </summary>
		/// <param name="password">This is not used.</param>
		/// <returns>Returns the session key to be used for the actual symmetric decryption of the encrypted document.</returns>
		protected override byte[] GetSymmDecryptionSessionKey( object password )
		{
			return SymmEncryptionSessionKey;
		}

		#endregion Session keys

		#region Private methods

		/// <summary>
		/// Computes the blended password.
		/// </summary>
		/// <param name="password">The password from which to compute the blended password. This may be a buffer or a
		/// text (although it is still aligned to the size of the symmetric cryptographic algorithm).</param>
		/// <returns>Returns the computed blended password.</returns>
		/// <remarks>
		/// <para>The resulted key is aligned with the <see cref="Vela.Crypto.Symmetric.SymmAlg.AlignToAlgSize"/> method.
		/// </para>
		/// </remarks>
		protected byte[] GenerateBlendedPassword( object password )
		{
			ValidateEncryptionPassword( password );

			byte[] blendedPassword;

			if( password is byte[] )
			{
				// The blended password was generated outside the current envelope.
				blendedPassword = (byte[])password;
			}
			else if( password is string )
			{
				// Generate the blended password from the user password.
				blendedPassword = PasswordBlenderAlg.GenerateBlendedPassword( Salt, password, true );
			}
			else
			{
				throw new ValidationException( string.Format( "The password must be either a buffer or a text, not a"
					+ " \"{0}\".", password.GetType().FullName ) );
			}

			// Align the blended password to the key size of the symmetric cryptographic algorithm.
			return SymmAlg.AlignToAlgSize( blendedPassword );
		}

		/// <summary>
		/// Validates the password used for encryption.
		/// </summary>
		/// <param name="password">The password to use for the decryption of the document. This may be a buffer or a
		/// text.</param>
		protected override void ValidateEncryptionPassword( object password )
		{
			ValidationUtils.Password( password, "password" );
		}

		/// <summary>
		/// Validates the password used for decryption.
		/// </summary>
		/// <param name="password">The password to use for the decryption of the document. This may be a buffer or a
		/// text.</param>
		protected override void ValidateDecryptionPassword( object password )
		{
			ValidateEncryptionPassword( password );
		}

		/// <summary>
		/// Symmetrically encrypts the session key.
		/// </summary>
		/// <param name="password">The password to use for the decryption of the document. This may be a buffer or a
		/// text.</param>
		protected override void EncryptSessionKey( object password )
		{
			// Compute the blended password and cache it.
			BlendedPassword = GenerateBlendedPassword( password );

			// Encrypt the session key. This doesn't need to use padding or chaining.
			SymmEncryptionSessionKey = SymmAlg.Encrypt( SymmEncryptionSessionKey, SymmPaddings.None, SymmAlg.Chaining, Salt
				, BlendedPassword );
		}

		/// <summary>
		/// Symmetrically decrypts the session key.
		/// </summary>
		/// <param name="password">The password to use for the decryption of the document. This may be a buffer or a
		/// text.</param>
		/// <returns>Returns true.</returns>
		protected override bool DecryptSessionKey( object password )
		{
			ValidateDecryptionPassword( password );

			// Compute the blended password and cache it.
			BlendedPassword = GenerateBlendedPassword( password );

			// Decrypt the session key. This doesn't need to use padding or chaining.
			SymmEncryptionSessionKey = SymmAlg.Decrypt( SymmEncryptionSessionKey, SymmPaddings.None, SymmAlg.Chaining, Salt
				, BlendedPassword );

			return true;
		}

		#endregion Private methods

		#region Public static methods

		/// <summary>
		/// Saves the specified information in a symmetrically encrypted document and reloads it in order to verify if it was
		/// correctly saved (if "verifySave" is true).
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="encryptableDocument">The object to save in encrypted form.</param>
		/// <param name="salt">The (random) salt used to compute the MAC and encrypt the data. This is stored in unencrypted
		/// form. If this is null or empty, a random salt is used instead.</param>
		/// <param name="password">The password to use for the decryption of the document. This may be a buffer or a text.
		/// </param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="compressionAlg">Compression algorithm. This may be null.</param>
		/// <param name="symmAlg">The symmetric cryptography algorithm.</param>
		/// <param name="passwordBlenderAlg">The parameters of the password blender algorithm used to generate the blended
		/// encryption key from the user's password. This may be null if the password is not textual.</param>
		/// <param name="fileName">The name of the file where the document is to be saved.</param>
		/// <param name="serializer">The serializer used for saving the document.</param>
		/// <param name="verifySave">If this is true then the saved document is reloaded in order to verify if it was correctly
		/// saved.</param>
		/// <returns>Returns the unencrypted MAC of the saved document.</returns>
		/// <remarks>
		/// <para>The saving is crash-safe. If the application crashes when the file is being saved, at least 1 of the
		/// following files is still available: a file with the specified name, a file with a backup name, a file with a
		/// temporary name. If the backup file exists, it's the valid one and you should implement your own recovery mechanism
		/// following a crash.</para>
		/// </remarks>
		public static byte[] SaveToFileVerify( Locator instanceLocator, IEncryptableDocument encryptableDocument, byte[] salt
			, object password, MacAlg macAlg, CompressionAlg compressionAlg, SymmAlg symmAlg
			, PasswordBlenderAlg passwordBlenderAlg, string fileName, ISerializer serializer, bool verifySave )
		{
			// Prepare the encrypted envelope of the document.
			SymmEncryptedEnvelope savedDoc = new SymmEncryptedEnvelope( instanceLocator, macAlg, compressionAlg, symmAlg
				, passwordBlenderAlg );
			savedDoc.EncryptableDocument = encryptableDocument;

			// Save the document. This will compute and cache the blended password.
			savedDoc.SaveToFile( fileName, salt, password, serializer );

			#region Verify the saved document

			// The verification process decreases significantly (= to half) the performance, but the safety is maximum.

			if( verifySave )
			{
				// Reload the saved document.
				IDeserializer envelopeDeserializer;
				SymmEncryptedEnvelope reloadedDoc = LoadFromFile( fileName, savedDoc.BlendedPassword
					, encryptableDocument.GetType(), out envelopeDeserializer );
				ValidationUtils.ObjectHasContent( reloadedDoc, "reloadedDoc" );

				#region Compare the original document with the reloaded document

				// The comparison can be done on MACs because they are the MACs of the data which was stored encrypted.
				// There is no guarantee that the unencrypted content is the uncorrupted original data, that is, while the
				// encrypted document was serialized, the result may have been corrupted (before being encrypted).

				// Verify the content.
				if( !BufferUtils.Equals( savedDoc.PriorEncryptedDocumentMacValue, reloadedDoc.EncryptedDocumentMacValue ) )
				{
					throw new ValidationException( "The original data of the content of this symmetrically encrypted document"
						+ " is different than the reloaded data." );
				}

				#endregion Compare the original document with the reloaded document
			}

			#endregion Verify the saved document

			return savedDoc.PriorEncryptedDocumentMacValue;
		}

		/// <summary>
		/// Loads the encrypted document from the specified buffer and decrypts it.
		/// </summary>
		/// <param name="buffer">Buffer from which to load the document. This may be null or empty.</param>
		/// <param name="password">The password to use for the decryption of the document. This may be a buffer or a
		/// text.</param>
		/// <param name="encryptableDocumentType">The root type of the encryptable document
		/// (<see cref="EncryptedEnvelope.EncryptableDocument"/>) to load; the document to load may either have this type or a
		/// type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <param name="deserializer">Is set with the deserializer of the loaded document.</param>
		/// <returns>Returns the loaded document, or null if the input data is null or empty.</returns>
		/// <remarks>
		/// <para>The encrypted document (<see cref="EncryptedEnvelope.EncryptedDocumentValue"/>) is cleared after loading
		/// because it's not necessary anymore.</para>
		/// </remarks>
		public static SymmEncryptedEnvelope LoadFromBuffer( byte[] buffer, object password, Type encryptableDocumentType
			, out IDeserializer deserializer )
		{
			return EncryptedEnvelope.LoadFromBuffer( buffer, password, typeof( SymmEncryptedEnvelope )
				, encryptableDocumentType, out deserializer ) as SymmEncryptedEnvelope;
		}

		/// <summary>
		/// Loads the encrypted document from the specified file and decrypts it.
		/// </summary>
		/// <param name="fileName">The name of the file from which to load the document. The file may be missing and its content
		/// may be null or empty.</param>
		/// <param name="password">The password to use for the decryption of the document. This may be a buffer or a
		/// text.</param>
		/// <param name="encryptableDocumentType">The root type of the encryptable document
		/// (<see cref="EncryptedEnvelope.EncryptableDocument"/>) to load; the document to load may either have this type or a
		/// type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <param name="deserializer">Is set with the deserializer of the loaded document.</param>
		/// <returns>Returns the loaded document, or null if the file is missing or its content is null or empty.</returns>
		/// <remarks>
		/// <para>The encrypted document (<see cref="EncryptedEnvelope.EncryptedDocumentValue"/>) is cleared after loading
		/// because it's not necessary anymore.</para>
		/// 
		/// <para>This method implements a recovery mechanism following a crash during a call of the
		/// <see cref="EncryptedEnvelope.SaveToFile"/> method.</para>
		/// </remarks>
		public static SymmEncryptedEnvelope LoadFromFile( string fileName, object password, Type encryptableDocumentType
			, out IDeserializer deserializer )
		{
			return EncryptedEnvelope.LoadFromFile( fileName, password, typeof( SymmEncryptedEnvelope )
				, encryptableDocumentType, out deserializer ) as SymmEncryptedEnvelope;
		}

		#endregion Public static methods
	}
}
