////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Crypto.Mac;
using Vela.Crypto.Symmetric;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// A symmetrically encrypted envelope used when a client has no information about a service and wants to interact with a
	/// service.
	/// </summary>
	/// <remarks>
	/// <para>The recipient service of this envelope is located using the of the service's UPN fingerprint, which is stored in
	/// <see cref="UpnEncryptedEnvelope.RecipientUpnFingerprint"/>.</para>
	/// 
	/// <para>This envelope is encrypted with the key generated from the service's UPN, with the
	/// <see cref="Upn.GenerateSymmEncryptionKey"/> method.</para>
	/// 
	/// <para>The content (encrypted by the envelope) is secure only if the UPN is private, specifically only if its known only
	/// to the sender and to the recipient.</para>
	/// </remarks>
	public class UpnEncryptedEnvelope : SymmEncryptedEnvelope, IRecipientUpnFingerprintContainer
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected UpnEncryptedEnvelope()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="compressionAlg">Compression algorithm. This may be null.</param>
		/// <param name="symmAlg">Symmetric cryptography algorithm.</param>
		/// <param name="passwordBlenderAlg">Password blender algorithm to use to generate the blended password. This may
		/// be null if the password is not textual.</param>
		/// <param name="recipientUpnFingerprint">The UPN fingerprint of the recipient of the encrypted envelope.</param>
		public UpnEncryptedEnvelope( Locator instanceLocator, MacAlg macAlg, CompressionAlg compressionAlg, SymmAlg symmAlg
				, PasswordBlenderAlg passwordBlenderAlg, Locator recipientUpnFingerprint )
			: base( instanceLocator, macAlg, compressionAlg, symmAlg, passwordBlenderAlg )
		{
			RecipientUpnFingerprint = recipientUpnFingerprint;
		}

		#endregion Constructors

		#region Public serializable properties

		private Locator _RecipientUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the recipient of the encrypted envelope. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Locator RecipientUpnFingerprint
		{
			get { return _RecipientUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RecipientUpnFingerprint" );

				_RecipientUpnFingerprint = value;
			}
		}

		#endregion Public serializable properties
	}
}
