////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// A group identity publisher which is a representation of a group entity, like an organization.
	/// </summary>
	/// <remarks>
	/// <para>A group identity publisher may be owned by several physical entities (people or organizations).</para>
	/// 
	/// <para>A group identity publisher describes how other entities can interact with it.</para>
	/// 
	/// <para>A group identity publisher may contain only public information.</para>
	/// 
	/// <para>Adding and removing identities must be done through the methods from this class, not through the
	/// <see cref="Identities"/> property. This is because the group's UPN depends on all the identities.</para>
	/// </remarks>
	public class GroupIdentityPublisher : IUpnBodySource, IIdentityPublisher
	{
		#region Constructors

		/// <summary>
		/// Createas a new instance of the current class.
		/// </summary>
		protected GroupIdentityPublisher()
		{
		}

		/// <summary>
		/// Createas a new instance of the current class.
		/// </summary>
		/// <param name="upnAlg">The UPN algorithm.</param>
		/// <param name="identities">List of solitary identity publishers which form the current group.</param>
		public GroupIdentityPublisher( UpnAlg upnAlg, SolitaryIdentityPublishers identities )
		{
			UpnAlg = upnAlg;
			Identities = identities;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private Upn _Upn;
		/// <summary>
		/// The UPN. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public Upn Upn
		{
			get
			{
				if( _Upn != null ) return _Upn;

				// The port is always 0 because a group identity can't be used for asymmetric cryptography, since only the
				// public identities are available.

				byte[] upnBody = UpnAlg.GenerateBody( this );
				_Upn = new Upn( UpnAlg, 0, upnBody );

				return _Upn;
			}
		}

		/// <summary>
		/// The UPN fingerprint of the group identity publisher to which the current object belongs. READ-ONLY.
		/// </summary>
		public Locator GroupUpnFingerprint
		{
			get { return Upn.Fingerprint; }
		}

		/// <summary>
		/// Specifies whether the current signature belongs to a group. READ-ONLY.
		/// </summary>
		public bool IsGroup
		{
			get { return true; }
		}

		/// <summary>
		/// The number of solitary identity publishers which are part of the current identity. READ-ONLY.
		/// </summary>
		public int IdentityCount
		{
			get { return Identities.Count; }
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public virtual DateTime ExpirationTime
		{
			get
			{
				DateTime min = DateTime.MaxValue;

				foreach( SolitaryIdentityPublisher identityPublisher in Identities )
				{
					if( identityPublisher.ExpirationTime < min ) min = identityPublisher.ExpirationTime;
				}

				return KernelUtils.MinTime( min, UpnAlg.ExpirationTime );
			}
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private UpnAlg _UpnAlg;
		/// <summary>
		/// The UPN algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public UpnAlg UpnAlg
		{
			get { return _UpnAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "UpnAlg" );

				_UpnAlg = value;

				// Regenerate the UPN of the group.
				_Upn = null;
			}
		}

		private SolitaryIdentityPublishers _Identities;
		/// <summary>
		/// List of solitary identity publishers which form the current group. READ-ONLY.
		/// </summary>
		[SerializeAttribute( ListItemName = "Identity" )]
		public SolitaryIdentityPublishers Identities
		{
			get { return _Identities; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Identities" );

				_Identities = value;
			}
		}

		private bool _EncryptToGroup = false;
		/// <summary>
		/// If this is true then the all documents sent to the group which owns the current identity are encrypted for the
		/// entire group.
		/// </summary>
		[SerializeAttribute]
		public bool EncryptToGroup
		{
			get { return _EncryptToGroup; }
			set
			{
				_EncryptToGroup = value;

				if( !_EncryptToGroup ) _MissingGroupDecryptorCount = 0;
			}
		}

		private int _MissingGroupDecryptorCount = 0;
		/// <summary>
		/// The number of extra session keys which are encrypted for each recipient of an encrypted document.
		/// </summary>
		[SerializeAttribute]
		public int MissingGroupDecryptorCount
		{
			get { return _MissingGroupDecryptorCount; }
			set
			{
				if( (_MissingGroupDecryptorCount != value) && !EncryptToGroup )
				{
					throw new ValidationException( "You can change the number of missing decryptor solitary recipients only if"
						+ " a document is to be encrypted to a recipient group." );
				}

				_MissingGroupDecryptorCount = value;
			}
		}

		private int _MissingGroupSignerCount = 0;
		/// <summary>
		/// The number of signers from the current group which may miss when the signatures of a signed document are verified
		/// with the current group identity publisher.
		/// </summary>
		[SerializeAttribute]
		public int MissingGroupSignerCount
		{
			get { return _MissingGroupSignerCount; }
			set { _MissingGroupSignerCount = value; }
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Validates the current object.
		/// </summary>
		protected virtual void Validate()
		{
			Locator groupUpnFingerprint = null;

			foreach( SolitaryIdentityPublisher idPublisher in Identities )
			{
				Locator solitaryGroupUpnFingerprint = idPublisher.Identity.GroupUpnFingerprint;
				if( solitaryGroupUpnFingerprint == null )
				{
					throw new ValidationException( string.Format( "The group UPN fingerprint of the identity with the UPN"
							+ " \"{0}\" of the group identity publisher with the group UPN fingerprint \"{1}\" is missing."
						, idPublisher.Upn, GroupUpnFingerprint ) );
				}

				if( groupUpnFingerprint == null ) groupUpnFingerprint = solitaryGroupUpnFingerprint;
				else if( !groupUpnFingerprint.Equals( solitaryGroupUpnFingerprint ) )
				{
					throw new ValidationException( string.Format( "The group UPN fingerprints of the identities of the group"
						+ " identity publisher with the group UPN fingerprint \"{0}\" are different.", GroupUpnFingerprint ) );
				}
			}
		}

		/// <summary>
		/// This method is automatically called before the current object is serialized.
		/// </summary>
		protected virtual void Serializing()
		{
			Validate();
		}

		/// <summary>
		/// This method is automatically called after the current object has been deserialized.
		/// </summary>
		protected virtual void Deserialized()
		{
			Validate();
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Clears the UPN (in order to regenerate it when it's read next time).
		/// </summary>
		public void ClearUpn()
		{
			_Upn = null;
		}

		/// <summary>
		/// Gets the data from which the UPN body will be generated.
		/// </summary>
		/// <returns>Returns the data from which the UPN body will be generated.</returns>
		public byte[] GetUpnBodySource()
		{
			BinaryNodeStream bns = new BinaryNodeStream( 0, (TextEncoding)null );

			foreach( SolitaryIdentityPublisher identity in Identities )
			{
				byte[] publicParams = identity.Identity.GetUpnBodySource();
				bns.WriteValue( publicParams );
			}

			return bns.ToArray();
		}

		/// <summary>
		/// Finds the solitary identity publisher, in the list of identities (<see cref="IIdentityPublisher.Identities"/>), which
		/// is identified by the specified UPN fingerprint.
		/// </summary>
		/// <param name="upnFingerprint">UPN fingerprint to look for.</param>
		/// <returns>Returns the found identity publisher, or null if none is found.</returns>
		public SolitaryIdentityPublisher Find( Locator upnFingerprint )
		{
			return Identities.Find( upnFingerprint );
		}

		/// <summary>
		/// Adds a new identity to the list. If the list already contains an identity with the same fingerprint and
		/// "updateIfSameKey" is false, an exception is thrown. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="identity">Identity to add to the list.</param>
		/// <param name="updateIfSameKey">When an identity is added to the list and the list already contains an identity with
		/// the same fingerprint, if this is true then the existing identity is replaced, else an exception is thrown.</param>
		public void AddIdentity( SolitaryIdentityPublisher identity, bool updateIfSameKey )
		{
			Identities.Add( identity, updateIfSameKey );

			// Regenerate the UPN of the group.
			ClearUpn();
		}

		/// <summary>
		/// Adds a list of new identities to the list. If the list already contains an identity with the same fingerprint and
		/// "updateIfSameKey" is false, an exception is thrown.
		/// </summary>
		/// <param name="identities">Identities to add to the list.</param>
		/// <param name="updateIfSameKey">When an identity is added to the list and the list already contains an identity with
		/// the same fingerprint, if this is true then the existing identity is replaced, else an exception is thrown.</param>
		public void AddIdentity( SolitaryIdentityPublishers identities, bool updateIfSameKey )
		{
			ValidationUtils.ObjectHasContent( identities, "identities" );

			foreach( SolitaryIdentityPublisher identity in identities ) Identities.Add( identity, updateIfSameKey );

			// Regenerate the UPN of the group.
			ClearUpn();
		}

		/// <summary>
		/// Removes the identity with the specified fingerprint from the list. Notifies property observers. Notifies list
		/// observers.
		/// </summary>
		/// <param name="upnFingerprint">The UPN fingerprint of the identity to remove. If an identity with this UPN fingerprint
		/// is not found, nothing happens.</param>
		/// <returns>Returns the removed identity, or null if none was removed.</returns>
		public SolitaryIdentityPublisher RemoveIdentity( Locator upnFingerprint )
		{
			SolitaryIdentityPublisher removedIdentity = Identities.Remove( upnFingerprint );

			// Regenerate the UPN of the group.
			ClearUpn();

			return removedIdentity;
		}

		/// <summary>
		/// Removes all the identities from the list. Notifies property observers. Notifies list observers.
		/// </summary>
		public void ClearIdentities()
		{
			Identities.Clear();

			// Regenerate the UPN of the group.
			ClearUpn();
		}

		#endregion Public methods
	}
}
