////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Contract for an identity publisher.
	/// </summary>
	public interface IIdentityPublisher : IExpirableAlgorithm
	{
		#region Public non-serializable properties

		/// <summary>
		/// The UPN algorithm. READ-ONLY.
		/// </summary>
		UpnAlg UpnAlg { get; }

		/// <summary>
		/// The UPN. READ-ONLY.
		/// </summary>
		Upn Upn { get; }

		/// <summary>
		/// The UPN fingerprint of the group identity publisher to which the current object belongs. This may be null. READ-ONLY.
		/// </summary>
		Locator GroupUpnFingerprint { get; }

		/// <summary>
		/// Specifies whether the current signature belongs to a group. READ-ONLY.
		/// </summary>
		bool IsGroup { get; }

		/// <summary>
		/// The number of solitary identity publishers which are part of the current identity. READ-ONLY.
		/// </summary>
		int IdentityCount { get; }

		/// <summary>
		/// The list of solitary identity publishers which are part of the current identity. All identities must be members of
		/// the same group. READ-ONLY.
		/// </summary>
		SolitaryIdentityPublishers Identities { get; }

		/// <summary>
		/// If this is true then the all documents sent to the group which owns the current identity are encrypted for the
		/// entire group. READ-ONLY.
		/// </summary>
		/// <remarks>A document may be encrypted with several passwords, one for each solitary recipient of the encrypted
		/// document. All solitary recipients form the group recipient. The document is encrypted with a session key which is a
		/// combination of the session keys for all solitary recipients. The session key of each solitary recipient is encrypted
		/// with that recipient's password.
		/// </remarks>
		bool EncryptToGroup { get; }

		/// <summary>
		/// The number of extra session keys which are encrypted for each recipient of an encrypted document. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is relevant only if <see cref="EncryptToGroup"/> is true.</para>
		/// 
		/// <para>If this property is 0, it is guaranteed that all solitary recipients must participate in order to decrypt the
		/// encrypted document.</para>
		/// 
		/// <para>Basically, this specifies how many solitary recipients (from the recipient group) may miss when the document is
		/// decrypted. It doesn't matter which recipient is missing, it may be any of them.</para>
		/// 
		/// <para>Technically, this is done by adding, to each recipient, several session keys of other solitary recipients.
		/// These other solitary recipients are the next "MissingGroupDecryptorCount" solitary recipients.</para>
		/// 
		/// <br/>
		/// 
		/// <para><b>Guarantees</b></para>
		/// 
		/// <para>If the document will be encrypted to a group recipient, assume that <c>NGR</c> is the number of solitary
		/// recipients from the group, and is at least 2.</para>
		/// 
		/// <para>It is guaranteed that:</para>
		/// <list type="bullet">
		///		<item>At least two solitary recipients will be required to decrypt the document.</item>
		///		<item>For "MissingGroupDecryptorCount" will be used a value equal with
		///			<c>Min( MissingGroupDecryptorCount, NGR - 2)</c>, which means that at least 2 solitary recipients are
		///			required in order to encrypt to a group.</item>
		///		<item><c>NGR / (MissingGroupDecryptorCount + 1)</c> (always rounded upwards) is the minimum number of solitary
		///			recipients necessary to decrypt the document.</item>
		/// </list>
		/// 
		/// <br/>
		/// 
		/// <para><b>Issue</b></para>
		/// 
		/// <para>If this property is at least 1, the document could be decrypted even if more than
		/// "MissingGroupDecryptorCount" solitary recipients don't participate to the decryption (something which was the goal of
		/// this mechanism). In fact, the minimum number of solitary recipients which is guaranteed to be necessary in order to
		/// decrypt the document, is equal with <c>NGR / (MissingGroupDecryptorCount + 1)</c>.</para>
		/// 
		/// <para>However, this requires some special conditions. For example, if "MissingGroupDecryptorCount" is 2, if every
		/// third recipient (using the order of the solitary recipients set during the encryption of the document)agrees to
		/// participate to the decryption of the document, the decryption would succeed because each recipient has the
		/// encryption keys of the missing solitary recipients (plus their own encryption key).</para>
		/// 
		/// <para>There is a partial solution which can be used to "fortify" this system by adding in a group identity publisher
		/// (<see cref="GroupIdentityPublisher"/>) information which specifies which solitary recipients
		/// <b>must</b> be present during the decryption of the document. This way, key people (of an entity) would always have
		/// to be present for the decryption of a document.</para>
		/// 
		/// <para>In technical terms, for such solitary recipients there is no "backup" session key stored for other solitary
		/// recipients. Thus, no other recipient would be able to use the session keys of the mandatory solitary recipients.
		/// </para>
		/// 
		/// <para>Presently, there doesn't seem to be enough reason to implement this solution, and actually is undesirable
		/// because it would mean that a document would be irretrievable if a mandatory solitary recipients would be
		/// unavailable.</para>
		/// </remarks>
		int MissingGroupDecryptorCount { get; }

		/// <summary>
		/// The number of signers from the current group which may miss when the signatures of a signed document are verified
		/// with the current group identity publisher. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>Normally, a service, for example, considers a document to have a valid signature only if all group member
		/// solitary identity publishers (<see cref="GroupIdentityPublisher.Identities"/>) have signed the document.</para>
		/// 
		/// <para>However, this property instructs the service to consider the document's signature as valid even if this number
		/// of member solitary identity publishers have not signed the document.</para>
		/// </remarks>
		int MissingGroupSignerCount { get; }

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Finds the solitary identity publisher, in the list of identities (<see cref="Identities"/>), which is identified by
		/// the specified UPN fingerprint.
		/// </summary>
		/// <param name="upnFingerprint">UPN fingerprint to look for.</param>
		/// <returns>Returns the found identity publisher, or null if none is found.</returns>
		SolitaryIdentityPublisher Find( Locator upnFingerprint );

		#endregion Public methods
	}
}
