////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto.Asymmetric;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Private identity which can be used to encrypt and sign data.
	/// </summary>
	/// <remarks>
	/// <para>A private identity contains the public identity.</para>
	/// 
	/// <para>A private identity contains two asymmetric keys. One pair is used for (asymmetric) encryption and decryption, the
	/// other is used sign data and to verify (asymmetric) signatures.</para>
	/// 
	/// <para>To "generate a private identity" means to generate its asymmetric keys.</para>
	/// </remarks>
	public class PrivateIdentity : PublicIdentity, IEncryptableDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected PrivateIdentity()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="upnAlg">The UPN algorithm.</param>
		/// <param name="processingPort">The network port of a service where action requests are received from clients.</param>
		/// <param name="encryptionAca">The asymmetric cryptographic algorithm (ACA), which performs the actual asymmetric
		/// encryption.</param>
		/// <param name="signingAca">The asymmetric cryptographic algorithm (ACA), which performs the actual asymmetric
		/// signing.</param>
		public PrivateIdentity( UpnAlg upnAlg, UInt16 processingPort, AsymmAlg encryptionAca, AsymmAlg signingAca )
			: base( upnAlg, processingPort, encryptionAca, signingAca )
		{
		}

		#endregion Constructors

		#region Public methods

		/// <summary>
		/// Decrypts the specified data with the current private key.
		/// </summary>
		/// <param name="buffer">Buffer to decrypt.</param>
		/// <param name="asymmEncryptionPaddingAlg">The asymmetric encryption padding algorithm. Must be the same as the one
		/// used during encryption.</param>
		/// <returns>Returns the decrypted data.</returns>
		public byte[] DecryptData( byte[] buffer, IAsymmEncryptionPaddingAlg asymmEncryptionPaddingAlg )
		{
			return EncryptionAca.DecryptData( buffer, asymmEncryptionPaddingAlg );
		}

		/// <summary>
		/// Signs the specified data with the current private key.
		/// </summary>
		/// <param name="buffer">Buffer to sign. This should be something small, like a hash.</param>
		/// <param name="asymmSigningPaddingAlg">The asymmetric signing padding algorithm.</param>
		/// <returns>Returns the signature of the specified data.</returns>
		public byte[] SignData( byte[] buffer, IAsymmSigningPaddingAlg asymmSigningPaddingAlg )
		{
			return SigningAca.SignData( buffer, asymmSigningPaddingAlg );
		}

		#endregion Public methods
	}
}
