////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Public identity which can be used to encrypt data and verify data signatures.
	/// </summary>
	/// <remarks>
	/// <para>A public identity contains two public keys. One pair is used for (asymmetric) encryption, the other is used to
	/// verify (asymmetric) signatures.</para>
	/// 
	/// <para>The two key pairs exists because, for example, if the owner of the identity is forced through a court order to
	/// hand over the key pair which is used to decrypt his received documents, he can hand over only that key pair and not the
	/// one which is used to sign documents. This way, the user preserves his exclusive ability to sign documents, even though
	/// someone else may also be able to see the content of the encrypted documents.</para>
	/// 
	/// <para>When an identity is generated (or imported), its parameters are verified for validity.</para>
	/// 
	/// <para>If you want to publish a public identity, create an identity publisher (<see cref="IIdentityPublisher"/>) and
	/// publish that.</para>
	/// </remarks>
	public class PublicIdentity : IUpnBodySource, IExpirableAlgorithm
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected PublicIdentity()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="upnAlg">The UPN algorithm.</param>
		/// <param name="processingPort">The network port of a service where action requests are received from clients.</param>
		/// <param name="encryptionAca">The asymmetric cryptographic algorithm (ACA), which performs the actual asymmetric
		/// encryption.</param>
		/// <param name="signingAca">The asymmetric cryptographic algorithm (ACA), which performs the actual asymmetric
		/// signing.</param>
		public PublicIdentity( UpnAlg upnAlg, UInt16 processingPort, AsymmAlg encryptionAca, AsymmAlg signingAca )
		{
			UpnAlg = upnAlg;
			ProcessingPort = processingPort;
			EncryptionAca = encryptionAca;
			SigningAca = signingAca;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// If this is true then the current identity contains the public keys. READ-ONLY.
		/// </summary>
		public bool HasPublicKeys
		{
			get { return EncryptionAca.HasPublicKey && SigningAca.HasPublicKey; }
		}

		/// <summary>
		/// If this is true then the current identity contains at least a private key (although it should always be both).
		/// READ-ONLY.
		/// </summary>
		public bool HasPrivateKey
		{
			get { return EncryptionAca.HasPrivateKey || SigningAca.HasPrivateKey; }
		}

		private Upn _Upn;
		/// <summary>
		/// The UPN. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public Upn Upn
		{
			get
			{
				if( _Upn != null ) return _Upn;

				_Upn = GenerateUpn();

				return _Upn;
			}
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public virtual DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( UpnAlg.ExpirationTime, EncryptionAca.ExpirationTime, SigningAca.ExpirationTime ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Locator _GroupUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the group identity publisher to which the current object belongs. This may be null.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public Locator GroupUpnFingerprint
		{
			get { return _GroupUpnFingerprint; }
			set { _GroupUpnFingerprint = value; }
		}

		private UpnAlg _UpnAlg;
		/// <summary>
		/// The UPN algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public UpnAlg UpnAlg
		{
			get { return _UpnAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "UpnAlg" );

				_UpnAlg = value;

				// Regenerate the UPN.
				_Upn = null;
			}
		}

		private UInt16 _ProcessingPort;
		/// <summary>
		/// The network port of a service where action requests are received from clients. This must be at least 1. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public UInt16 ProcessingPort
		{
			get { return _ProcessingPort; }
			protected set
			{
				ValidationUtils.Integer( value, 1, UInt16.MaxValue, "ProcessingPort" );

				_ProcessingPort = value;
			}
		}

		/// <summary></summary>
		protected AsymmAlg _EncryptionAca;
		/// <summary>
		/// The asymmetric cryptographic algorithm (ACA), which performs the actual asymmetric encryption. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public AsymmAlg EncryptionAca
		{
			get
			{
				if( _EncryptionAca == null )
					ValidationUtils.ThrowObjectNotInitialized( "EncryptionAca" );

				return _EncryptionAca;
			}
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "EncryptionAca" );

				_EncryptionAca = value;

				// Regenerate the UPN.
				_Upn = null;
			}
		}

		/// <summary></summary>
		protected AsymmAlg _SigningAca;
		/// <summary>
		/// The asymmetric cryptographic algorithm (ACA), which performs the actual asymmetric signing. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public AsymmAlg SigningAca
		{
			get
			{
				if( _SigningAca == null )
					ValidationUtils.ThrowObjectNotInitialized( "SigningAca" );

				return _SigningAca;
			}
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "SigningAca" );

				_SigningAca = value;

				// Regenerate the UPN.
				_Upn = null;
			}
		}

		private DateTime _CreationTime = DateTime.UtcNow;
		/// <summary>
		/// The time (in UTC format) when the current object was created. The current object is valid from and after this time;
		/// before this time, it should not be used in processing. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is automatically set when the current object is created (and you should not modify it).</para>
		/// </remarks>
		[SerializeAttribute]
		public DateTime CreationTime
		{
			get { return _CreationTime; }
			protected set { _CreationTime = value; }
		}

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Gets the public part of the current (private) identity.
		/// </summary>
		/// <returns>Returns the public part of the current (private) identity. This is not a clone.</returns>
		public PublicIdentity GetPublicIdentity()
		{
			// Create a new identity.
			PublicIdentity identity = new PublicIdentity();

			// Copy the public properties.
			identity.GroupUpnFingerprint = GroupUpnFingerprint;
			identity.UpnAlg = UpnAlg;
			identity.ProcessingPort = ProcessingPort;
			identity.EncryptionAca = EncryptionAca.GetPublicAlg();
			identity.SigningAca = SigningAca.GetPublicAlg();
			identity.CreationTime = CreationTime;

			if( identity.HasPrivateKey )
			{
				throw new ValidationException( string.Format( "The identity with the UPN \"{0}\" has the private part, although"
					+ " only the public part was requested.", Upn ) );
			}

			return identity;
		}

		/// <summary>
		/// Gets the data from which the UPN body will be generated.
		/// </summary>
		/// <returns>Returns the data from which the UPN body will be generated.</returns>
		public byte[] GetUpnBodySource()
		{
			BinaryNodeStream bns = new BinaryNodeStream( 0, KernelUtils.GetDeepTextEncoding() );

			// Do not write the type into the stream because then the private identity and the public one extracted with the
			// "GetPublicIdentity" method would have different UPNs.

			EncryptionAca.WriteUpnBodySource( bns );
			SigningAca.WriteUpnBodySource( bns );

			return bns.ToArray();
		}

		/// <summary>
		/// Generates the UPN associated to the current identity.
		/// </summary>
		/// <returns>Returns the generated UPN.</returns>
		public Upn GenerateUpn()
		{
			byte[] upnBody = UpnAlg.GenerateBody( this );
			return new Upn( UpnAlg, ProcessingPort, upnBody );
		}

		/// <summary>
		/// Encrypts the specified data with the current public key.
		/// </summary>
		/// <param name="buffer">Buffer to encrypt. This should be something small, like a hash.</param>
		/// <param name="asymmEncryptionPaddingAlg">The asymmetric encryption padding algorithm.</param>
		/// <returns>Returns the encrypted data.</returns>
		public byte[] EncryptData( byte[] buffer, IAsymmEncryptionPaddingAlg asymmEncryptionPaddingAlg )
		{
			return EncryptionAca.EncryptData( buffer, asymmEncryptionPaddingAlg );
		}

		/// <summary>
		/// Verifies the signature of the specified data with the current public key. An
		/// <see cref="InvalidSignatureException"/> (among others) is thrown if the signature is invalid.
		/// </summary>
		/// <param name="buffer">Buffer whose signature has to be verified.</param>
		/// <param name="signature">The signature against which the data's signature is verified.</param>
		/// <param name="asymmSigningPaddingAlg">The asymmetric signing padding algorithm. Must be the same as the one used
		/// during signing.</param>
		public void VerifyDataSignature( byte[] buffer, byte[] signature, IAsymmSigningPaddingAlg asymmSigningPaddingAlg )
		{
			SigningAca.VerifyDataSignature( buffer, signature, asymmSigningPaddingAlg );
		}

		#endregion Public methods
	}
}
