////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// A solitary identity publisher which is a representation of a single entity, not necessarily a single person.
	/// </summary>
	/// <remarks>
	/// <para>A solitary identity publisher may be owned by only one entity (person or organization). Normally, only people
	/// should use identity publishers, while organizations should use group identity publishers
	/// (<see cref="GroupIdentityPublisher"/>) because they can group several solitary identity publishers. However, small
	/// organizations are usually controlled by a single person.</para>
	/// 
	/// <para>The content of a solitary identity publisher may not be modified, but only created from its associated vault.
	/// </para>
	/// 
	/// <para>A solitary identity publisher may contain only public information.</para>
	/// </remarks>
	public class SolitaryIdentityPublisher : IIdentityPublisher
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected SolitaryIdentityPublisher()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="publicIdentity">Public identity from which to create a solitary identity publisher.</param>
		public SolitaryIdentityPublisher( PublicIdentity publicIdentity )
		{
			Identity = publicIdentity;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The UPN algorithm. READ-ONLY.
		/// </summary>
		public UpnAlg UpnAlg
		{
			get { return Identity.UpnAlg; }
		}

		/// <summary>
		/// The UPN. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public Upn Upn
		{
			get { return Identity.Upn; }
		}

		/// <summary>
		/// The UPN fingerprint of the group identity publisher to which the current object belongs. This may be null. READ-ONLY.
		/// </summary>
		public Locator GroupUpnFingerprint
		{
			get { return Identity.GroupUpnFingerprint; }
		}

		/// <summary>
		/// Specifies whether the current signature belongs to a group. READ-ONLY.
		/// </summary>
		public bool IsGroup
		{
			get { return false; }
		}

		/// <summary>
		/// The number of solitary identity publishers which are part of the current identity. READ-ONLY.
		/// </summary>
		public int IdentityCount
		{
			get { return 1; }
		}

		/// <summary>
		/// The list of solitary identity publishers which are part of the current identity. All identities must be members of
		/// the same group. READ-ONLY.
		/// </summary>
		public SolitaryIdentityPublishers Identities
		{
			get { return new SolitaryIdentityPublishers( this ); }
		}

		/// <summary>
		/// If this is true then the all documents sent to the group which owns the current identity are encrypted for the
		/// entire group. READ-ONLY.
		/// </summary>
		public bool EncryptToGroup
		{
			get { return false; }
		}

		/// <summary>
		/// The number of extra session keys which are encrypted for each recipient of an encrypted document. READ-ONLY.
		/// </summary>
		public int MissingGroupDecryptorCount
		{
			get { return 0; }
		}

		/// <summary>
		/// The number of signers from the current group which may miss when the signatures of a signed document are verified
		/// with the current group identity publisher. READ-ONLY.
		/// </summary>
		public int MissingGroupSignerCount
		{
			get { return 0; }
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public virtual DateTime ExpirationTime
		{
			get { return Identity.ExpirationTime; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private PublicIdentity _Identity;
		/// <summary>
		/// The public identity which is contained by the current identity publisher. This must not contain the private
		/// information of the private identity. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public PublicIdentity Identity
		{
			get { return _Identity; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Identity" );

				if( value.HasPrivateKey )
				{
					throw new ValidationException( string.Format( "The public identity with the UPN \"{0}\" contains the"
						+ " private key.", value.Upn ) );
				}

				_Identity = value;
			}
		}

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Finds the solitary identity publisher, in the list of identities (<see cref="IIdentityPublisher.Identities"/>), which
		/// is identified by the specified UPN fingerprint.
		/// </summary>
		/// <param name="upnFingerprint">UPN fingerprint to look for.</param>
		/// <returns>Returns the found identity publisher, or null if none is found.</returns>
		public SolitaryIdentityPublisher Find( Locator upnFingerprint )
		{
			if( Upn.Fingerprint.Equals( upnFingerprint ) ) return this; // Identity found.
			else return null; // No identity was found.
		}

		#endregion Public methods
	}
}
