////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// List of solitary identity publishers (<see cref="SolitaryIdentityPublisher"/>).
	/// </summary>
	/// <remarks>
	/// <para>A sorted list is used in order to ensure that the UPN (body source) is always the same. For details, see the
	/// <see cref="GroupIdentityPublisher.GetUpnBodySource"/> method.</para>
	/// </remarks>
	public class SolitaryIdentityPublishers : KernelSortedList<Locator, SolitaryIdentityPublisher>
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public SolitaryIdentityPublishers()
			: base( LocatorComparer.Instance )
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="identities">List of identities to add to the new list.</param>
		public SolitaryIdentityPublishers( params SolitaryIdentityPublisher[] identities )
			: base( LocatorComparer.Instance )
		{
			ValidationUtils.ArrayHasContent( identities, "identities" );

			foreach( SolitaryIdentityPublisher identity in identities )
				Add( identity );
		}

		#endregion Constructors

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para>The fingerprint of the <see cref="SolitaryIdentityPublisher.Upn"/> is used as key.</para>
		/// </remarks>
		public override Locator GetKey( SolitaryIdentityPublisher item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.Upn.Fingerprint;
		}

		#endregion Public methods
	}
}
