////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Pool of core algorithms.
	/// </summary>
	public interface ICoreAlgorithmPool
	{
		#region Public methods

		/// <summary>
		/// Gets the compression algorithm.
		/// </summary>
		/// <returns>Returns the compression algorithm. This may be null.</returns>
		CompressionAlg GetCompressionAlg();

		/// <summary>
		/// Gets the password blender.
		/// </summary>
		/// <returns>Returns the password blender.</returns>
		PasswordBlenderAlg GetPasswordBlender();

		/// <summary>
		/// Gets the symmetrically encrypted envelope.
		/// </summary>
		/// <returns>Returns the symmetrically encrypted envelope.</returns>
		SymmEncryptedEnvelope GetSymmEncryptedEnvelope();

		/// <summary>
		/// Gets the envelope symmetrically encrypted with a UPN.
		/// </summary>
		/// <param name="recipientUpnFingerprint">The UPN fingerprint of the recipient of the encrypted envelope.</param>
		/// <returns>Returns the envelope symmetrically encrypted with a UPN.</returns>
		UpnEncryptedEnvelope GetUpnEncryptedEnvelope( Locator recipientUpnFingerprint );

		/// <summary>
		/// Gets the asymmetrically encrypted envelope.
		/// </summary>
		/// <returns>Returns the asymmetrically encrypted envelope.</returns>
		AsymmEncryptedEnvelope GetAsymmEncryptedEnvelope();

		/// <summary>
		/// Gets the algorithm used to generate UPNs.
		/// </summary>
		/// <param name="algorithmSecurityLevel">The security level for the returned algorithm.</param>
		/// <returns>Returns the algorithm used to generate UPNs.</returns>
		UpnAlg GetUpnAlg( AlgorithmSecurityLevel algorithmSecurityLevel );

		/// <summary>
		/// Gets the size (in bytes) to which the encrypted data of an exported document's (encrypted) envelope is aligned. If
		/// this is bigger than <see cref="EncryptedEnvelope.BlockSizeSize"/>, the real size of the encrypted data will be
		/// hidden.
		/// </summary>
		/// <returns>Returns the size (in bytes) to which the encrypted data of an exported document's (encrypted) envelope is
		/// aligned.</returns>
		int GetExportedDocumentEnvelopeEncryptedDocumentBlockSize();

		/// <summary>
		/// Gets the size (in bytes) to which the encrypted data of an action request (encrypted) envelope is aligned. If this is
		/// bigger than <see cref="EncryptedEnvelope.BlockSizeSize"/>, the real size of the encrypted data will be hidden.
		/// </summary>
		/// <returns>Returns the size (in bytes) to which the encrypted data of an action request (encrypted) envelope is
		/// aligned.</returns>
		int GetActionRequestEnvelopeEncryptedDocumentBlockSize();

		/// <summary>
		/// Gets the size (in bytes) to which the (encrypted) data of a secured document is aligned. If this is bigger than 1,
		/// the real size of the data will be hidden.
		/// </summary>
		/// <returns>Returns the size (in bytes) to which the (encrypted) data of a secured document is aligned.</returns>
		int GetSecuredDocumentBlockSize();

		#endregion Public methods
	}
}
