////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

namespace Vela.CoreDocuments
{
	/// <summary>
	/// <para><b>Hashes</b></para>
	/// 
	/// <para>A data hash is used to store a "summary" of some data, in order to check (with ease) the integrity of the
	/// data.</para>
	/// 
	/// <para>When some data is encrypted, the hash of the unencrypted data may be stored as plain-text because it's not
	/// intended to authenticate the data, but is there to merely prove that:</para>
	/// <list type="bullet">
	///		<item>The encrypted data was not accidentally altered.</item>
	///		<item>The user's password is the one used during encryption.</item>
	///		<item>All the other parameters used in cryptographic operations (like the identity with which the encryption was
	///			made) are the ones used during encryption.</item>
	/// </list>
	/// 
	/// <para>When some data is signed, the hash of the data is actually signed (and thus nobody else could sign the hash for
	/// some other data, using the same asymmetric key).</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Serialization</b></para>
	/// 
	/// <para>All core documents are processed entirely into memory. They are not bound to XML. The content of the documents is
	/// independent from the serializer.</para>
	/// 
	/// <para>If you want to serialize the documents to another format, you have to implement your own methods to perform the
	/// serialization (see the <see cref="Vela.Serialization.ISerializer"/> and
	/// <see cref="Vela.Serialization.IDeserializer"/> classes). However, you must keep the hash serializer
	/// (<see cref="Vela.Serialization.IHashSerializer"/>), which is used for signing data.</para>
	/// 
	/// <para>What's more important is that the data which is signed doesn't depend on the document serializer because it uses a
	/// hash serializer which guarantees that the serialization output is the same, regardless of the platform it runs on.</para>
	/// 
	/// <para>Note that you must take into consideration the design-time serialization attributes (like
	/// <see cref="Vela.Kernel.Serialization.SerializeAttribute"/>).</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Self-signing</b></para>
	/// 
	/// <para>A solitary identity publisher (<see cref="IIdentityPublisher"/>) may be self-signed, that is, it may be signed with
	/// the private identity (<see cref="PrivateIdentity"/>) whose public identity (<see cref="PublicIdentity"/>) it
	/// contains.Through self-signing, the information contained in the identity is authenticated by its issuer.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Custom signatures</b></para>
	/// 
	/// <para>When signing a document (<see cref="SignedDocument"/>), you can customize the authenticated information from
	/// signatures (<see cref="Signature"/>) with any class inherited from <see cref="SignatureAuthenticatedInfo"/>.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Signed document detachability</b></para>
	/// 
	/// <para>Detachability is a property of a signed document (<see cref="SignedDocument"/>) which allows for the signed
	/// document to be copied (either programatically or manually from the stored XML), from within its enveloping
	/// document, elsewhere while preserving the validity of its signatures.</para>
	/// 
	/// <para>Moreover, a detached signed document can be signed further.</para>
	/// 
	/// <para>However, the same hash serializer must be used to serialize the document when the signed document is copied.</para>
	/// 
	/// <para>Note that a signed document is detachable from within its envelope, but the document's signatures are not
	/// detachable (from the signed document).</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Attention</b></para>
	/// 
	/// <para><i><u>Padding for symmetric encryption</u></i></para>
	/// 
	/// <para>If padding is used during the symmetric encryption of some data, the result an extra block required for
	/// padding).</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Expose encryptable document</b></para>
	/// 
	/// <para>In debug mode, the encryptable document from all encrypted documents
	/// (<see cref="Vela.CoreDocuments.EncryptedEnvelope.EncryptableDocument"/>) is serialized to see exactly what is
	/// saved.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Various</b></para>
	/// 
	/// <para>Unless otherwise stated in the documentation, all properties exposed outside classes must not hold null or empty
	/// datas.</para>
	/// </summary>
	class NamespaceDoc
	{
		// Only for documentation.
	}
}
