////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Reflection;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// "MAC-ed Password Blender" password blender algorithm.
	/// </summary>
	/// <remarks>
	/// <para>The algorithm first expends a memory effort and then a processing effort.</para>
	/// 
	/// <para>The memory effort consists of allocating a large amount of memory and accessing it as much as possible. This is
	/// presumed to dramatically raise the costs of a parallelized brute-force attack by requiring each used processor to use a
	/// large and slow memory.</para>
	/// 
	/// <para>The processing effort consists of computing a number of chained MACs (= PBKDF2).</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Memory effort</b></para>
	/// 
	/// <para>It's essential to create the content of the large buffer in a pseudo-random manner, using itself, in order to make
	/// it impossible for the attacker to parallelize this step.</para>
	/// 
	/// <para>Due to the unconvincing hardware-based arguments made in support of using a memory effort, most effort has to be
	/// expended with a processing effort (not with a memory effort).</para>
	/// </remarks>
	public class Hpb : PasswordBlenderAlg
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Hpb()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="textEncoding">Text encoding.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="strength">Strength. Use 0 (or a negative number) if you want to execute a single iteration in the
		/// blender.</param>
		/// <param name="memoryEffortBufferSize">The size (in bytes) of the buffer used to expend the memory effort. Use 0 (or a
		/// negative number) if you want to ignore the memory effort.</param>
		/// <param name="memoryEffortCycles">The number of cycles to go over the buffer used to expend the memory effort. This
		/// must be at least 1.</param>
		public Hpb( TextEncoding textEncoding, MacAlg macAlg, int strength, int memoryEffortBufferSize, int memoryEffortCycles )
			: base( textEncoding, macAlg, strength )
		{
			MemoryEffortBufferSize = memoryEffortBufferSize;
			MemoryEffortCycles = memoryEffortCycles;
		}

		#endregion Constructors

		#region Public serializable properties

		private int _MemoryEffortBufferSize;
		/// <summary>
		/// The size (in bytes) of the buffer used to expend the memory effort. Use 0 (or a negative number) if you want to
		/// ignore the memory effort. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is used only when a user password is also blended.</para>
		/// </remarks>
		[SerializeAttribute]
		public int MemoryEffortBufferSize
		{
			get { return _MemoryEffortBufferSize; }
			protected set { _MemoryEffortBufferSize = value; }
		}

		private int _MemoryEffortCycles;
		/// <summary>
		/// The number of cycles to go over the buffer used to expend the memory effort. This must be at least 1. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is used only when a user password is also blended.</para>
		/// </remarks>
		[SerializeAttribute]
		public int MemoryEffortCycles
		{
			get { return _MemoryEffortCycles; }
			protected set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "MemoryEffortCycles" );

				_MemoryEffortCycles = value;
			}
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Gets the real strength of the specified relative strength of the blender.
		/// </summary>
		/// <param name="strength">The relative strength of the blender.</param>
		/// <returns>Returns the real strength of the specified relative strength.</returns>
		protected override Int64 GetRealStrength( int strength )
		{
			ValidateStrength( strength );

			// The real strength is power of 2.
			return (strength > 0) ? (1 << strength) : 0;
		}

		/// <summary>
		/// Expends an effort associated to a large amount of memory.
		/// </summary>
		/// <param name="salt">Salt with which to generate the blended password.</param>
		/// <param name="password">Password with which to generate the blended password.</param>
		/// <returns>Returns a MAC which represents the expended memory effort.</returns>
		private byte[] ExpendMemoryEffort( byte[] salt, byte[] password )
		{
			ValidationUtils.ObjectHasContent( salt, "salt" );
			ValidationUtils.ObjectHasContent( password, "password" );
			ValidationUtils.Integer( MemoryEffortBufferSize, 1, int.MaxValue, "MemoryEffortBufferSize" );

			int passwordSize = password.Length;
			byte[] mac = MacAlg.Compute( password, null, true, salt );

			// Create a state which consumes a large amount of memory.
			byte[] state = new byte[ MemoryEffortBufferSize ];
			BufferUtils.ContiguousCopy( mac, state );

			const UInt32 p1 = 16777619;
			const UInt32 p2 = 2166136261;

			UInt32 accumulator = BitConverter.ToUInt32( mac, 0 );

			for( int c = 0; c < MemoryEffortCycles; c++ )
			{
				for( int i = 0; i < MemoryEffortBufferSize; i++ )
				{
					unchecked
					{
						UInt32 state0 = state[ i ];
						UInt32 state1 = state[ accumulator % MemoryEffortBufferSize ] * p1;
						UInt32 state2 = state[ state1 % MemoryEffortBufferSize ] * p1;

						byte p0 = password[ i % passwordSize ];

						accumulator = (UInt32)((accumulator + state1) * p1 + (state0 + state2 + p0) * p2);

						state[ i ] = (byte)accumulator;
					}
				}
			}

			return MacAlg.Compute( password, null, true, salt, state );
		}

		/// <summary>
		/// Expends an effort associated to a large amount of memory.
		/// </summary>
		/// <param name="salt">Salt with which to generate the blended password.</param>
		/// <param name="password">Password with which to generate the blended password.</param>
		/// <param name="strength">Strength. This must be at least 1.</param>
		/// <returns>Returns a MAC which represents the expended processing effort.</returns>
		private byte[] ExpendProcessingEffort( byte[] salt, byte[] password, Int64 strength )
		{
			ValidationUtils.ObjectHasContent( salt, "salt" );
			ValidationUtils.ObjectHasContent( password, "password" );
			ValidationUtils.Integer( strength, 1, int.MaxValue, "strength" );

			byte[] mac = salt;
			byte[] processingEffort = salt;

			for( Int64 i = 0; i < strength; i++ )
			{
				// The counter should never be (used) 0.
				mac = MacAlg.Compute( password, null, true, mac, strength - i );

				processingEffort = BufferUtils.Xor( processingEffort, mac );
			}

			return processingEffort;
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Generates the blended password from the specified salt and key.
		/// </summary>
		/// <param name="salt">Salt with which to generate the blended password.</param>
		/// <param name="password">Password with which to generate the blended password. This must be a scalar.</param>
		/// <param name="useStrength">If this is true then the <see cref="PasswordBlenderAlg.Strength"/> is also used.</param>
		/// <returns>Returns the blended password.</returns>
		/// <remarks>
		/// <para>The size of the resulted key depends on the used hash algorithm. Therefore, when you have to use this key to
		/// encrypt data, use the <see cref="Vela.Crypto.Symmetric.SymmAlg.AlignToAlgSize"/> method to align the key to
		/// the required size for the used encryption algorithm.</para>
		/// </remarks>
		public override byte[] GenerateBlendedPassword( byte[] salt, object password, bool useStrength )
		{
			ValidationUtils.ObjectHasContent( salt, "salt" );
			ValidationUtils.Password( password, "password" );

			Int64 strength = useStrength ? GetRealStrength( Strength ) : 0;

			LogManager logManager = (strength > 0) ? LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() ) : null;

			byte[] passwordBuffer = (password is string) ? TextEncoding.TextToArray( (string)password, false )
				: (byte[])password;

			// The main MAC.
			byte[] blendedPassword = MacAlg.Compute( passwordBuffer, null, true, salt );

			// Iterate the MAC.
			if( strength > 0 )
			{
				byte[] mac = (MemoryEffortBufferSize > 0) ? ExpendMemoryEffort( blendedPassword, passwordBuffer )
					: blendedPassword;

				blendedPassword = ExpendProcessingEffort( mac, passwordBuffer, strength );
			}

			LogManager.DebugEndMethod( logManager );

			return blendedPassword;
		}

		#endregion Public methods
	}
}
