////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Contract for a password blender which is used to generate blended passwords (BEKs) from user passwords.
	/// </summary>
	/// <remarks>
	/// <para>A password blender is used to slow down dictionary attacks and any other possible attack on user passwords.</para>
	/// 
	/// <para>Documents are encrypted symmetrically, using a random key, so there is no possibility of a dictionary attack on
	/// it. However, since the user who made the symmetric encryption must also decrypt the document, it's necessary to encrypt
	/// the random key with the user's password.</para>
	/// 
	/// <para>The problem is that this password is usually weak, made of few words, and is thus exposed to dictionary attacks.To
	/// counter this problem, the password is passed through a time consuming process, the password blender, in order to obtain
	/// a blended password which will then be used to encrypt the random key.</para>
	/// 
	/// <para>This time consuming process slows down just a little the encryption and decryption for the owner of the
	/// document, but has a much stronger effect on an attacker.</para>
	/// 
	/// <para>The password blender uses multiple MAC iterations. Because the iterations can't be parallelized, a dictionary
	/// attack on an encrypted document is slowed down significantly.</para>
	/// 
	/// <para>The strength of the password blender is proportional with the number of consecutive MAC iterations used by the
	/// blender.</para>
	/// </remarks>
	public abstract class PasswordBlenderAlg : IImmutableAlgorithm, IExpirableAlgorithm
	{
		#region Public constants

		/// <summary>
		/// The minimum strength of the password blender.
		/// </summary>
		public const int StrengthMin = 10;

		/// <summary>
		/// The maximum strength of the password blender.
		/// </summary>
		public const int StrengthMax = 20;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected PasswordBlenderAlg()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="textEncoding">Text encoding.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="strength">Strength. Use 0 (or a negative number) if you want to execute a single MAC iteration.</param>
		public PasswordBlenderAlg( TextEncoding textEncoding, MacAlg macAlg, int strength )
		{
			TextEncoding = textEncoding;
			MacAlg = macAlg;
			Strength = strength;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public virtual DateTime ExpirationTime
		{
			get { return MacAlg.ExpirationTime; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private TextEncoding _TextEncoding;
		/// <summary>
		/// Text encoding. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public TextEncoding TextEncoding
		{
			get { return _TextEncoding; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "TextEncoding" );

				_TextEncoding = value;
			}
		}

		private MacAlg _MacAlg;
		/// <summary>
		/// MAC algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public MacAlg MacAlg
		{
			get { return _MacAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "MacAlg" );

				_MacAlg = value;
			}
		}

		private int _Strength;
		/// <summary>
		/// Strength. Use 0 (or a negative number) if you want to execute a single MAC iteration. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is used only when a user password is also blended.</para>
		/// </remarks>
		[SerializeAttribute]
		public int Strength
		{
			get { return _Strength; }
			protected set
			{
				ValidateStrength( value );

				_Strength = value;
			}
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Gets the real strength of the specified relative strength of the blender.
		/// </summary>
		/// <param name="strength">The relative strength of the blender.</param>
		/// <returns>Returns the real strength of the specified relative strength.</returns>
		protected abstract Int64 GetRealStrength( int strength );

		#endregion Private methods

		#region Public static methods

		/// <summary>
		/// Validates the specified strength (of a password blender).
		/// </summary>
		/// <param name="strength">Strength to validate. If this is at least 1, it must be at least <see cref="StrengthMin"/>.
		/// </param>
		public static void ValidateStrength( int strength )
		{
			// Check if only one iteration of the blender is to be executed.
			if( strength <= 0 ) return;

			ValidationUtils.Integer( strength, StrengthMin, StrengthMax, "strength" );
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Generates the blended password from the specified salt and key.
		/// </summary>
		/// <param name="salt">Salt with which to generate the blended password.</param>
		/// <param name="password">Password with which to generate the blended password. This must be a scalar.</param>
		/// <param name="useStrength">If this is true then the <see cref="PasswordBlenderAlg.Strength"/> is also used.</param>
		/// <returns>Returns the blended password.</returns>
		/// <remarks>
		/// <para>The size of the resulted key depends on the used hash algorithm. Therefore, when you have to use this key to
		/// encrypt data, use the <see cref="Vela.Crypto.Symmetric.SymmAlg.AlignToAlgSize"/> method to align the key to the
		/// required size for the used encryption algorithm.</para>
		/// </remarks>
		public abstract byte[] GenerateBlendedPassword( byte[] salt, object password, bool useStrength );

		#endregion Public methods
	}
}
