////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Crypto.Asymmetric;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// A blinding session. This is used to track all documents exchanged during an interaction based on blinding.
	/// </summary>
	public class BlindingSession : InteractionSession
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected BlindingSession()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpnFingerprint">The UPN fingerprint of the recipient of the current object.</param>
		/// <param name="documentExpirationTime">The expiration time (in UTC format) of the document. The current document is
		/// valid before this time; from and after this time, it should not be used in processing.</param>
		/// <param name="symmEncryptionKeySize">The size (in bytes) that the symmetric encryption key must have. If this is 0
		/// or a negative number, the default <see cref="CoreDocumentsConsts.SymmEncryptionKeySizeDef"/> is used.</param>
		/// <param name="blindingConfiguration">Blinding configuration.</param>
		/// <remarks>
		/// <para>The symmetric encryption key of the session is automatically generated in here.</para>
		/// </remarks>
		public BlindingSession( Locator instanceLocator, Locator initiatorUpnFingerprint
				, Locator recipientUpnFingerprint, DateTime documentExpirationTime, int symmEncryptionKeySize
				, BlindingConfiguration blindingConfiguration )
			: base( instanceLocator, initiatorUpnFingerprint, recipientUpnFingerprint, documentExpirationTime
				, symmEncryptionKeySize )
		{
			BlindingConfiguration = blindingConfiguration;
		}

		#endregion Constructors

		#region Public serializable properties

		private BlindingConfiguration _BlindingConfiguration;
		/// <summary>
		/// The blinding configuration. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public BlindingConfiguration BlindingConfiguration
		{
			get { return _BlindingConfiguration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "BlindingConfiguration" );

				_BlindingConfiguration = value;
			}
		}

		#endregion Public serializable properties
	}
}
