////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// An interaction session. This is used to track all documents exchanged during an interaction.
	/// </summary>
	/// <remarks>
	/// <para>An interaction session must not be sent to a service because it may contain private information. Instead, send
	/// the session obtained with the <see cref="GetPublicSession"/> method.</para>
	/// 
	/// <para>An interaction session allows action requests and responses to be decoupled from their encrypted envelopes and
	/// still reference the same session.</para>
	/// 
	/// <para>The symmetric encryption key from a session (<see cref="SymmEncryptionKey"/>) is used to encrypt the response which
	/// is sent back to the initiator of the current request in order to provide faster encryption and less exposure of the
	/// asymmetric encryption key of the initiator of the request.</para>
	/// </remarks>
	public class InteractionSession : Sheaf, IInitiatorUpnFingerprintContainer, IRecipientUpnFingerprintContainer
		, IExpirableDocument
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="InitiatorUpnFingerprint"/> property.
		/// </summary>
		public const string InitiatorUpnFingerprintFilterIdentifier = "InteractionSession.InitiatorUpnFingerprint#4";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="RecipientUpnFingerprint"/> property.
		/// </summary>
		public const string RecipientUpnFingerprintFilterIdentifier = "InteractionSession.RecipientUpnFingerprint#5";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="DocumentExpirationTime"/> property.
		/// </summary>
		public const string DocumentExpirationTimeFilterIdentifier = "InteractionSession.DocumentExpirationTime#6";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected InteractionSession()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpnFingerprint">The UPN fingerprint of the recipient of the current object.</param>
		/// <param name="documentExpirationTime">The expiration time (in UTC format) of the document. The current document is
		/// valid before this time; from and after this time, it should not be used in processing.</param>
		/// <param name="symmEncryptionKeySize">The size (in bytes) that the symmetric encryption key must have. If this is 0
		/// or a negative number, the default <see cref="CoreDocumentsConsts.SymmEncryptionKeySizeDef"/> is used.</param>
		/// <remarks>
		/// <para>The symmetric encryption key of the session is automatically generated in here.</para>
		/// </remarks>
		public InteractionSession( Locator instanceLocator, Locator initiatorUpnFingerprint
				, Locator recipientUpnFingerprint, DateTime documentExpirationTime, int symmEncryptionKeySize )
			: base( instanceLocator )
		{
			InitiatorUpnFingerprint = initiatorUpnFingerprint;
			RecipientUpnFingerprint = recipientUpnFingerprint;
			DocumentExpirationTime = documentExpirationTime;

			if( symmEncryptionKeySize <= 0 ) symmEncryptionKeySize = CoreDocumentsConsts.SymmEncryptionKeySizeDef;
			SymmEncryptionKey = EntropyPool.GetBytes( symmEncryptionKeySize );
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( base.ExpirationTime, DocumentExpirationTime ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Locator _InitiatorUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the initiator of the current object. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( InitiatorUpnFingerprintFilterIdentifier )]
		public Locator InitiatorUpnFingerprint
		{
			get { return _InitiatorUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "InitiatorUpnFingerprint" );

				_InitiatorUpnFingerprint = value;
			}
		}

		private Locator _RecipientUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the recipient of the current object. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( RecipientUpnFingerprintFilterIdentifier )]
		public Locator RecipientUpnFingerprint
		{
			get { return _RecipientUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RecipientUpnFingerprint" );

				_RecipientUpnFingerprint = value;
			}
		}

		private DateTime _DocumentExpirationTime = DateTime.MaxValue;
		/// <summary>
		/// The expiration time (in UTC format) of the document. The current document is valid before this time; from and after
		/// this time, it should not be used in processing. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( DocumentExpirationTimeFilterIdentifier )]
		public DateTime DocumentExpirationTime
		{
			get { return _DocumentExpirationTime; }
			protected set { _DocumentExpirationTime = value; }
		}

		private byte[] _SymmEncryptionKey;
		/// <summary>
		/// The symmetric encryption key of the session, randomly generated. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] SymmEncryptionKey
		{
			get { return _SymmEncryptionKey; }
			protected set
			{
				ValidationUtils.Array( value, CoreDocumentsConsts.SymmEncryptionKeySizeMin
					, CoreDocumentsConsts.SymmEncryptionKeySizeMax, "SymmEncryptionKey" );

				_SymmEncryptionKey = value;
			}
		}

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Gets the public version of the current session, version which may be sent to a service.
		/// </summary>
		/// <returns>Returns the public version of the current session.</returns>
		public virtual InteractionSession GetPublicSession()
		{
			InteractionSession publicSession = new InteractionSession();

			publicSession.InstanceLocator = InstanceLocator;
			publicSession.DocumentExpirationTime = DocumentExpirationTime;
			publicSession.SymmEncryptionKey = SymmEncryptionKey;

			return publicSession;
		}

		#endregion Public methods
	}
}
