////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Collections.Generic;
using Vela.Kernel;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// List of interaction sessions.
	/// </summary>
	/// <remarks>
	/// <para>Expired sessions may be added to the list.</para>
	/// </remarks>
	public class InteractionSessions : KernelSortedList<Locator, InteractionSession>
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public InteractionSessions()
			: base( LocatorComparer.Instance )
		{
		}

		#endregion Constructors

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para>The instance locator of the interaction session is used as key.</para>
		/// </remarks>
		public override Locator GetKey( InteractionSession item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.InstanceLocator;
		}

		/// <summary>
		/// Removes the expired sessions from the list.
		/// </summary>
		/// <param name="lifetimeMax">The maximum allowed lifetime (in minutes). If this is 0 or a negative number, it's
		/// ignored.</param>
		public void RemoveExpired( int lifetimeMax )
		{
			List<Locator> locatorsToRemove = new List<Locator>();

			// Gather the items which have expired.
			foreach( KeyValuePair<Locator, InteractionSession> kvp in Items )
			{
				if( KernelUtils.HasExpired( kvp.Value, lifetimeMax, false ) )
					locatorsToRemove.Add( kvp.Key );
			}

			// Remove the expired items.
			foreach( Locator locator in locatorsToRemove )
				Remove( locator );
		}

		#endregion Public methods
	}
}
