////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// A signature for digitally signed data. A signature is the result of an asymmetric private encryption applied on a MAC
	/// of the data to sign and of the authenticated information (<see cref="AuthenticatedInfo"/>).
	/// </summary>
	/// <remarks>
	/// <para>A signature authenticates certain information (= a document).</para>
	/// 
	/// <para>When a document is signed, it is cryptographically serialized (= in a binary identical manner, see
	/// <see cref="Vela.Serialization.IHashSerializer"/>) in a buffer which is then MACed. The resulted MAC is then signed.
	/// </para>
	/// </remarks>
	public class Signature : IExpirableAlgorithm
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Signature()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="upnFingerprint">The UPN fingerprint of the identity with which the current signature was created.
		/// </param>
		/// <param name="groupUpnFingerprint">The UPN fingerprint of the group identity publisher to which the current object
		/// belongs. This may be null.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="sigAuthInfo">Information authenticated by the identity which creates the current signature. This
		/// information is granted by the signing identity to the signed document.</param>
		/// <param name="value">The value of the signature.</param>
		public Signature( Locator upnFingerprint, Locator groupUpnFingerprint, MacAlg macAlg
			, IAsymmSigningPaddingAlg asymmSigningPaddingAlg, SignatureAuthenticatedInfo sigAuthInfo, byte[] value )
		{
			UpnFingerprint = upnFingerprint;
			GroupUpnFingerprint = groupUpnFingerprint;
			MacAlg = macAlg;
			AsymmSigningPaddingAlg = asymmSigningPaddingAlg;
			AuthenticatedInfo = sigAuthInfo;
			Value = value;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public virtual DateTime ExpirationTime
		{
			get
			{
				return KernelUtils.MinTime( MacAlg.ExpirationTime, AsymmSigningPaddingAlg.ExpirationTime
					, AuthenticatedInfo.ExpirationTime );
			}
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Locator _UpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the identity with which the current signature was created. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Locator UpnFingerprint
		{
			get { return _UpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "UpnFingerprint" );

				_UpnFingerprint = value;
			}
		}

		private Locator _GroupUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the group identity publisher to which the current object belongs. This may be null. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public Locator GroupUpnFingerprint
		{
			get { return _GroupUpnFingerprint; }
			protected set { _GroupUpnFingerprint = value; }
		}

		private MacAlg _MacAlg;
		/// <summary>
		/// MAC algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public MacAlg MacAlg
		{
			get { return _MacAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "MacAlg" );

				_MacAlg = value;
			}
		}

		private IAsymmSigningPaddingAlg _AsymmSigningPaddingAlg;
		/// <summary>
		/// The asymmetric signing padding algorithm which was used when the current signature was created. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public IAsymmSigningPaddingAlg AsymmSigningPaddingAlg
		{
			get { return _AsymmSigningPaddingAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "AsymmSigningPaddingAlg" );

				_AsymmSigningPaddingAlg = value;
			}
		}

		private SignatureAuthenticatedInfo _AuthenticatedInfo;
		/// <summary>
		/// Information, for a signed document (<see cref="SignedDocument"/>), authenticated by the identity which creates the
		/// current signature. This information is granted by the signing identity to the signed document. Since you pass an
		/// object for this property to the constructor of the current class, there is no need to manually set it later.
		/// READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>When a document is signed, the data of the document and the authenticated information are serialized and put
		/// in a single buffer, one after another. Then, this buffer is MACed and the result is stored in the
		/// <see cref="Value"/> property in encrypted form (using asymmetric private encryption).</para>
		/// 
		/// <para>This way, both the document and the authenticated information are authenticated by the current signature, and
		/// are also dependent of each other, that is, the signature can't be detached from the signed document.</para>
		/// </remarks>
		[SerializeAttribute]
		public SignatureAuthenticatedInfo AuthenticatedInfo
		{
			get { return _AuthenticatedInfo; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "AuthenticatedInfo" );

				_AuthenticatedInfo = value;
			}
		}

		private byte[] _Value;
		/// <summary>
		/// The value of the signature. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] Value
		{
			get { return _Value; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "Value" );

				_Value = value;
			}
		}

		#endregion Public serializable properties
	}
}
