////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// A structure of information authenticated by the identity which creates the owner signature (<see cref="Signature"/>).
	/// This information is granted by the signing identity to the signed document which contains the owner signature.
	/// </summary>
	/// <remarks>
	/// <para>This class serves as base class for all data structures which must be stored in signatures, and must be
	/// authenticated by the identities which create the signatures.</para>
	/// 
	/// <para>All serialized properties (including those from the derived classes) of the current class are authenticated by the
	/// identity which creates the current signature.</para>
	/// </remarks>
	public class SignatureAuthenticatedInfo : IExpirableAlgorithm
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public SignatureAuthenticatedInfo()
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public virtual DateTime ExpirationTime
		{
			get { return DateTime.MaxValue; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private DateTime _CreationTime = DateTime.UtcNow;
		/// <summary>
		/// The creation time (in UTC format) of the document. This is automatically set with the time of the creation of the
		/// current object. The current object is valid from and after this time; before this time, it should not be used in
		/// processing. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public DateTime CreationTime
		{
			get { return _CreationTime; }
			protected set { _CreationTime = value; }
		}

		#endregion Public serializable properties
	}
}
