////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Reflection;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Contract for a class which may be used to digitally sign documents.
	/// </summary>
	/// <remarks>
	/// <para>Documents are signed at data level, not at file level. Actually, when a document is signed, it is cryptographically
	/// serialized (= in a binary identical manner, see
	/// <see cref="Vela.Serialization.IHashSerializer"/>) in a buffer whose MAC is computed, MAC which is then signed.</para>
	/// 
	/// <para>This means that the signature is verified after the data is loaded from the document file into an object of the
	/// current type, and therefore whitespaces are not relevant (so long as the serialization methods are the same on both the
	/// computer which signed the document, and on the computer which verifies the signature).</para>
	/// 
	/// <para>A signed document may have only one signature created by a specific identity (<see cref="PrivateIdentity"/>).
	/// </para>
	/// </remarks>
	public abstract class SignedDocument : Sheaf
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected SignedDocument()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">The locator of the current sheaf.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		public SignedDocument( Locator instanceLocator, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg
				, IHashSerializer signatureHashSerializer )
			: base( instanceLocator )
		{
			MacAlg = macAlg;
			AsymmSigningPaddingAlg = asymmSigningPaddingAlg;
			SignatureHashSerializer = signatureHashSerializer;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">The locator of the current sheaf.</param>
		/// <param name="signedDocument">Signed document from which to copy the necessary information.</param>
		public SignedDocument( Locator instanceLocator, SignedDocument signedDocument )
			: this( instanceLocator, signedDocument.MacAlg, signedDocument.AsymmSigningPaddingAlg
				, signedDocument.SignatureHashSerializer )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get
			{
				DateTime min = DateTime.MaxValue;

				foreach( Signature signature in Signatures )
				{
					if( signature.ExpirationTime < min )
						min = signature.ExpirationTime;
				}

				return KernelUtils.MinTime( min, base.ExpirationTime, MacAlg.ExpirationTime
					, AsymmSigningPaddingAlg.ExpirationTime );
			}
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private MacAlg _MacAlg;
		/// <summary>
		/// MAC algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public MacAlg MacAlg
		{
			get { return _MacAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "MacAlg" );

				_MacAlg = value;
			}
		}

		private IAsymmSigningPaddingAlg _AsymmSigningPaddingAlg;
		/// <summary>
		/// Asymmetric signing padding algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public IAsymmSigningPaddingAlg AsymmSigningPaddingAlg
		{
			get { return _AsymmSigningPaddingAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "AsymmSigningPaddingAlg" );

				_AsymmSigningPaddingAlg = value;
			}
		}

		private IHashSerializer _SignatureHashSerializer;
		/// <summary>
		/// Hash serializer for signatures. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public IHashSerializer SignatureHashSerializer
		{
			get { return _SignatureHashSerializer; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "SignatureHashSerializer" );

				_SignatureHashSerializer = value;
			}
		}

		private Signatures _Signatures = new Signatures();
		/// <summary>
		/// The list of signatures of the document. This is not signed / authenticated together with the document because it
		/// changes every time when a signature is created. The backing field is initialized by default. READ-ONLY.
		/// </summary>
		[SerializeAttribute( ListItemName = "Signature", HashSerialization = HashSerializations.NotInRoot )]
		public Signatures Signatures
		{
			get { return _Signatures; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Signatures" );

				_Signatures = value;
			}
		}

		#endregion Public serializable properties

		#region Private static methods

		/// <summary>
		/// Verifies the signature of the specified document. An <see cref="InvalidSignatureException"/> (among others) is thrown
		/// if the signature is invalid.
		/// </summary>
		/// <param name="serializedContent">Document for which to verify the signature.</param>
		/// <param name="instanceLocator">The instance locator of the document for which to verify the signature.</param>
		/// <param name="identity">The public identity with which the document's signature is checked.</param>
		/// <param name="signature">Signature to verify for validity.</param>
		/// <param name="hashSerializer">The hash serializer used to serialize the document whose signature is to be verified.
		/// </param>
		private static void VerifySignatureCore( byte[] serializedContent, Locator instanceLocator, PublicIdentity identity
			, Signature signature, IHashSerializer hashSerializer )
		{
			ValidationUtils.ArrayHasContent( serializedContent, "serializedContent" );
			ValidationUtils.ObjectHasContent( instanceLocator, "instanceLocator" );
			ValidationUtils.ObjectHasContent( identity, "identity" );
			ValidationUtils.ObjectHasContent( signature, "signature" );
			ValidationUtils.ObjectHasContent( hashSerializer, "hashSerializer" );

			// Check if the signing identity is the specified identity.
			if( !identity.Upn.Fingerprint.Equals( signature.UpnFingerprint ) )
			{
				throw new MissingSignatureException( string.Format( "The UPN fingerprint \"{0}\" of the verifying identity"
						+ " is different than the UPN fingerprint \"{1}\" of the verified signature."
					, identity.Upn.Fingerprint, signature.UpnFingerprint ) );
			}

			// Verify the signature.
			#region Authenticated information

			// Serialize the authenticated information in order to prepare it for processing.
			byte[] serializedAuthInfo = hashSerializer.SaveToBuffer( signature.AuthenticatedInfo, null );

			#endregion Authenticated information

			// The MAC of the data to be verified.
			byte[] macKey = signature.MacAlg.HashAlg.Compute( null, true, instanceLocator.Value
				, signature.UpnFingerprint.Value );
			byte[] macValue = signature.MacAlg.Compute( macKey, null, true, serializedContent, serializedAuthInfo );

			identity.VerifyDataSignature( macValue, signature.Value, signature.AsymmSigningPaddingAlg );
		}

		#endregion Private static methods

		#region Public methods

		/// <summary>
		/// Collects the UPN fingerprints of the signing entities (from the signatures of the current document).
		/// </summary>
		/// <returns>Returns the UPN fingerprints of the signing entities. This may be empty, but not null.</returns>
		public Locators CollectSigningUpnFingerprints()
		{
			Locators locators = new Locators();

			foreach( Signature signature in Signatures )
			{
				locators.Add( signature.UpnFingerprint );

				// Since several signatures may have the same group UPN fingerprint, allow updating.
				if( signature.GroupUpnFingerprint != null )
					locators.Add( signature.GroupUpnFingerprint, true );
			}

			return locators;
		}

		/// <summary>
		/// Signs the document with the specified identity. If the document was already signed with the specified identity then
		/// the existing signature is replaced (with a newly generated signature), else the signature is added to the list of
		/// signatures.
		/// </summary>
		/// <param name="privateIdentity">The private identity with which the document is to be signed.</param>
		/// <param name="sigAuthInfo">Information authenticated by the signing identity. This information is granted by the
		/// signing identity to the signed document. If this is null then a default object is created, if the signing identity
		/// hasn't already signed the document.</param>
		/// <remarks>
		/// <para>Note that what is signed is neither the pure data nor the saved document, but a buffer into which the data
		/// to be signed was serialized. Therefore, when the signature is verified, the whitespaces from the loaded XML are
		/// relevant only if they are part of the data (not part of the structure of the XML).</para>
		/// 
		/// <para>During processing, the <see cref="Signatures"/> list is ignored since it can't be signed (because it will be
		/// modified with the signature).</para>
		/// 
		/// <para>The creation time of the signature (<see cref="SignatureAuthenticatedInfo.CreationTime"/>) and the UPN
		/// fingerprint are automatically set during signing.</para>
		/// </remarks>
		public virtual void Sign( PrivateIdentity privateIdentity, SignatureAuthenticatedInfo sigAuthInfo )
		{
			ValidationUtils.ObjectHasContent( privateIdentity, "privateIdentity" );

			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				// Check if the list of signatures contains a signature created with the specified identity.
				Signature signature = Signatures.Find( privateIdentity.Upn.Fingerprint );

				// The data to be signed must be serialized with the hash serializer because the serialization output must be
				// binary identical, regardless of the platform it runs on.

				// Serialize the document in order to prepare it for processing.
				byte[] serializedContent = SignatureHashSerializer.SaveToBuffer( this, null );

				#region Authenticated information

				// Set up the authenticated information.
				if( signature == null )
				{
					if( sigAuthInfo == null )
						sigAuthInfo = new SignatureAuthenticatedInfo();
				}
				else
				{
					ValidationUtils.ObjectHasContent( signature.AuthenticatedInfo, "signature.AuthenticatedInfo" );

					if( sigAuthInfo == null )
						sigAuthInfo = signature.AuthenticatedInfo;
				}

				// Serialize the authenticated information in order to prepare it for processing.
				byte[] serializedAuthInfo = SignatureHashSerializer.SaveToBuffer( sigAuthInfo, null );

				#endregion Authenticated information

				// The MAC of the data to be signed.
				byte[] macKey = MacAlg.HashAlg.Compute( null, true, InstanceLocator.Value
					, privateIdentity.Upn.Fingerprint.Value );
				byte[] macValue = MacAlg.Compute( macKey, null, true, serializedContent, serializedAuthInfo );

				// Sign the serialized data.
				byte[] signatureValue = privateIdentity.SignData( macValue, AsymmSigningPaddingAlg );

				signature = new Signature( privateIdentity.Upn.Fingerprint, privateIdentity.GroupUpnFingerprint, MacAlg
					, AsymmSigningPaddingAlg, sigAuthInfo, signatureValue );
				Signatures.Add( signature, true );
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		/// <summary>
		/// Verifies the signature of the specified document. An <see cref="InvalidSignatureException"/> (among others) is thrown
		/// if the signature is invalid.
		/// </summary>
		/// <param name="identityPublisher">The identity publisher with which the document's signatures are checked.</param>
		/// <returns>Returns the time of signing of the specified document; for a group identity, it's the newest signing time.
		/// </returns>
		public virtual DateTime VerifySignature( IIdentityPublisher identityPublisher )
		{
			ValidationUtils.ObjectHasContent( identityPublisher, "identityPublisher" );

			// The data to be signed must be serialized with the hash serializer because the serialization output must be binary
			// identical, regardless of the platform it runs on.
			byte[] serializedContent = SignatureHashSerializer.SaveToBuffer( this, null );

			// Used because group signatures may be valid even if not all group members have signed the document.
			int validSigCount = 0;

			DateTime signingTime = DateTime.MinValue;

			// Check if the list of signatures contains signatures created with the specified identity.
			foreach( SolitaryIdentityPublisher identity in identityPublisher.Identities )
			{
				// Get the signature.
				Signature signature = Signatures.Find( identity.Upn.Fingerprint );
				if( signature == null )
					continue;

				try
				{
					VerifySignatureCore( serializedContent, InstanceLocator, identity.Identity, signature
						, SignatureHashSerializer );

					if( signingTime < signature.AuthenticatedInfo.CreationTime )
						signingTime = signature.AuthenticatedInfo.CreationTime;

					validSigCount++;
				}
				catch( Exception ex )
				{
					// The loop must continue (if possible).

					// When a group of people are expected to sign a document, some signatures may miss entirely or some may be
					// invalid, so ignore them.

					if( ExceptionUtils.IsAbort( ex ) )
						throw ExceptionUtils.PreserveStackTrace( ex );

					LogManager.Add( ex );
				}
			}

			// Check if there are enough valid signatures.
			if( validSigCount < identityPublisher.IdentityCount - identityPublisher.MissingGroupSignerCount )
			{
				throw new InvalidSignatureException( string.Format( "The signed document whose signatures are being verified"
						+ " with the identity publisher with the UPN \"{0}\" doesn't contain enough valid signatures. Make sure"
						+ " that you and your contacts are using the newest version of the application."
					, identityPublisher.Upn ) );
			}

			return signingTime;
		}

		/// <summary>
		/// Verifies the signature of the specified document. An <see cref="InvalidSignatureException"/> (among others) is thrown
		/// if the signature is invalid.
		/// </summary>
		/// <param name="identity">The public identity with which the document's signature is checked.</param>
		/// <returns>Returns the time of signing of the specified document; for a group identity, it's the newest signing time.
		/// </returns>
		public virtual DateTime VerifySignature( PublicIdentity identity )
		{
			ValidationUtils.ObjectHasContent( identity, "identity" );

			// Check if there are still no signatures.
			if( Signatures == null )
			{
				throw new MissingSignatureException( string.Format( "The signed document whose signatures are being verified"
					+ " with the identity with the UPN \"{0}\" doesn't contain any signature.", identity.Upn ) );
			}

			// The data to be signed must be serialized with the hash serializer because the serialization output must be binary
			// identical, regardless of the platform it runs on.
			byte[] serializedContent = SignatureHashSerializer.SaveToBuffer( this, null );

			// Get the signature.
			Signature signature = Signatures.Find( identity.Upn.Fingerprint );

			// A signature created with the specified identity must exist.
			if( signature == null )
			{
				throw new MissingSignatureException( string.Format( "The document was not signed with the identity with the"
					+ " UPN \"{0}\".", identity.Upn ) );
			}

			VerifySignatureCore( serializedContent, InstanceLocator, identity, signature, SignatureHashSerializer );

			return signature.AuthenticatedInfo.CreationTime;
		}

		/// <summary>
		/// Verifies the signature of the specified document. An <see cref="InvalidSignatureException"/> (among others) is thrown
		/// if the signature is invalid.
		/// </summary>
		/// <param name="identityPublishers">The identity publishers with which the document's signatures are checked.</param>
		/// <param name="signingLocatorTimes">List where to put the locators of the entities which have actually signed the
		/// document, and the creation times of their signatures. If an item with the same key already exists, it's updated.
		/// This may be empty, but not null.</param>
		public void CollectValidSignatures( Dictionary<Locator, IIdentityPublisher> identityPublishers
			, LocatorTimes signingLocatorTimes )
		{
			foreach( KeyValuePair<Locator, IIdentityPublisher> kvp in identityPublishers )
			{
				try
				{
					DateTime signingTime = VerifySignature( kvp.Value );

					LocatorTime signatureLocatorTime = new LocatorTime( kvp.Key, signingTime );

					signingLocatorTimes.Add( signatureLocatorTime, true );
				}
				catch( Exception ex )
				{
					// The loop must continue (if possible).

					if( ExceptionUtils.IsAbort( ex ) )
						throw ExceptionUtils.PreserveStackTrace( ex );

					LogManager.Add( ex );
				}
			}
		}

		/// <summary>
		/// Verifies the signature of the specified document. An <see cref="InvalidSignatureException"/> (among others) is thrown
		/// if the signature is invalid.
		/// </summary>
		/// <param name="identities">The identities with which the document's signatures are checked.</param>
		/// <param name="signingLocatorTimes">List where to put the locators of the entities which have actually signed the
		/// document, and the creation times of their signatures. If an item with the same key already exists, it's updated.
		/// This may be empty, but not null.</param>
		public void CollectValidSignatures( Dictionary<Locator, PublicIdentity> identities, LocatorTimes signingLocatorTimes )
		{
			foreach( KeyValuePair<Locator, PublicIdentity> kvp in identities )
			{
				try
				{
					DateTime signingTime = VerifySignature( kvp.Value );

					LocatorTime signatureLocatorTime = new LocatorTime( kvp.Key, signingTime );

					signingLocatorTimes.Add( signatureLocatorTime, true );
				}
				catch( Exception ex )
				{
					// The loop must continue (if possible).

					if( ExceptionUtils.IsAbort( ex ) )
						throw ExceptionUtils.PreserveStackTrace( ex );

					LogManager.Add( ex );
				}
			}
		}

		#endregion Public methods
	}
}
