////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto;
using Vela.Crypto.Hash;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// A stamp acts as a proof of work.
	/// </summary>
	/// <remarks>
	/// <para>Stamps are hash collisions which take a lot of time to compute, but can be validated almost instantly. They are
	/// used to ensure that the available resources are not wasted by opportunists or criminals.</para>
	/// 
	/// <para>For example, stamps prove that people are so serious about using a resource or a service, that they are willing to
	/// spend in exchange some of their computer's processing power.</para>
	/// 
	/// <para>The cost of a stamp is the number of bits which form the hash collision.</para>
	/// 
	/// <para>To "generate a stamp" means to generate a hash collision for it.</para>
	/// 
	/// <para>A signed stamp has no guarantee that it fulfills its self-cost (<see cref="Cost"/>), so use the
	/// <see cref="FulfillsSelfCost"/> method to verify the cost.</para>
	/// </remarks>
	public class Stamp : Sheaf, IExpirableDocument
	{
		#region Public constants

		/// <summary>
		/// The maximum cost.
		/// </summary>
		public const int CostMax = 40;

		/// <summary>
		/// The minimum size of the salt.
		/// </summary>
		public const int SaltSizeMin = 16;

		/// <summary>
		/// The maximum size of the salt.
		/// </summary>
		public const int SaltSizeMax = 128;

		/// <summary>
		/// The default size of the salt.
		/// </summary>
		public const int SaltSizeDef = 32;

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="DocumentExpirationTime"/> property.
		/// </summary>
		public const string DocumentExpirationTimeFilterIdentifier = "Stamp.DocumentExpirationTime#4";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="GenerationCompleted"/> property.
		/// </summary>
		public const string GenerationCompletedFilterIdentifier = "Stamp.GenerationCompleted#5";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <remarks>
		/// <para>In order to generate the stamp, call the <see cref="Generate"/> method.</para>
		/// </remarks>
		protected Stamp()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="hashAlg">Hash algorithm.</param>
		/// <param name="hashSerializer">Hash serializer.</param>
		/// <param name="documentExpirationTime">The expiration time (in UTC format) of the document. The current document is
		/// valid before this time; from and after this time, it should not be used in processing.</param>
		/// <param name="cost">The cost of the stamp.</param>
		/// <remarks>
		/// <para>In order to generate the stamp, call the <see cref="Generate"/> method.</para>
		/// 
		/// <para>The salt is randomly generated in here.</para>
		/// </remarks>
		public Stamp( Locator instanceLocator, HashAlg hashAlg, IHashSerializer hashSerializer
				, DateTime documentExpirationTime, int cost )
			: base( instanceLocator )
		{
			HashAlg = hashAlg;
			HashSerializer = hashSerializer;
			DocumentExpirationTime = documentExpirationTime;
			Cost = cost;

			Salt = EntropyPool.GetBytes( SaltSizeDef );
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get
			{
				return KernelUtils.MinTime( base.ExpirationTime, HashAlg.ExpirationTime, DocumentExpirationTime );
			}
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private HashAlg _HashAlg;
		/// <summary>
		/// Hash algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public HashAlg HashAlg
		{
			get { return _HashAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "HashAlg" );

				_HashAlg = value;
			}
		}

		private IHashSerializer _HashSerializer;
		/// <summary>
		/// Hash serializer for signatures. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public IHashSerializer HashSerializer
		{
			get { return _HashSerializer; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "HashSerializer" );

				_HashSerializer = value;
			}
		}

		private DateTime _DocumentExpirationTime = DateTime.MaxValue;
		/// <summary>
		/// The expiration time (in UTC format) of the document. The current document is valid before this time; from and after
		/// this time, it should not be used in processing. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( DocumentExpirationTimeFilterIdentifier )]
		public DateTime DocumentExpirationTime
		{
			get { return _DocumentExpirationTime; }
			protected set { _DocumentExpirationTime = value; }
		}

		private int _Cost = 0;
		/// <summary>
		/// The cost of the stamp (= the number of 0 bits from the hash collision). This must be at least 1. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is the cost that the stamp must have after it's generated. For the current cost, see the
		/// <see cref="GetCost"/> method.</para>
		/// </remarks>
		[SerializeAttribute]
		public int Cost
		{
			get { return _Cost; }
			protected set
			{
				ValidationUtils.Integer( value, 1, CostMax, "Cost" );

				_Cost = value;
			}
		}

		private byte[] _Salt;
		/// <summary>
		/// Random data which makes every stamp unique. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] Salt
		{
			get { return _Salt; }
			protected set
			{
				ValidationUtils.Array( value, SaltSizeMin, SaltSizeMax, "Salt" );

				_Salt = value;
			}
		}

		private UInt64 _Counter;
		/// <summary>
		/// The counter which forms a valid stamp. This must be at least 0. This is not signed / authenticated together with the
		/// stamp because it's meant to be set after the hash collision is generated. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The stamp can be valid only if this value is at least 1. The default value is 0.</para>
		/// 
		/// <para>This is incremented with 1 at every attempt to find a hash collision.</para>
		/// </remarks>
		[SerializeAttribute( HashSerialization = HashSerializations.NotInRoot )]
		public UInt64 Counter
		{
			get { return _Counter; }
			protected set
			{
				ValidationUtils.Integer( value, 0, UInt64.MaxValue, "Counter" );

				_Counter = value;
			}
		}

		private byte[] _CollisionHash;
		/// <summary>
		/// The hash result of the underlying hash collision. This is not signed / authenticated together with the stamp because
		/// it's meant to be set after a hash collision is found. READ-ONLY.
		/// </summary>
		[SerializeAttribute( HashSerialization = HashSerializations.NotInRoot )]
		public byte[] CollisionHash
		{
			get { return _CollisionHash; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "CollisionHash" );

				_CollisionHash = value;
			}
		}

		private bool _GenerationCompleted = false;
		/// <summary>
		/// If this is true then the generation of the stamp has been completed and the stamp fulfills its cost. This is not
		/// signed / authenticated together with the stamp because it's meant to be set after the hash collision is generated.
		/// </summary>
		/// <remarks>
		/// <para>This is used only as a speckle filter, in order to be able to quickly retrieve the stamps whose
		/// generation must continue.</para>
		/// </remarks>
		[SerializeAttribute( HashSerialization = HashSerializations.NotInRoot )]
		[SpeckleFilterAttribute( GenerationCompletedFilterIdentifier )]
		public bool GenerationCompleted
		{
			get { return _GenerationCompleted; }
			set { _GenerationCompleted = value; }
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Gets the suffix buffer. This is appended to the stamp counter in order to create the buffer to loop.
		/// </summary>
		/// <returns>Returns the created buffer.</returns>
		private byte[] GetSuffixBuffer()
		{
			ValidationUtils.Integer( Cost, 1, CostMax, "Cost" );
			ValidationUtils.Array( Salt, SaltSizeMin, SaltSizeMax, "Salt" );

			byte[] buffer = HashSerializer.SaveToBuffer( this, null );

			return HashAlg.Compute( buffer );
		}

		/// <summary>
		/// Gets the buffer to hash.
		/// </summary>
		/// <returns>Returns the buffer to hash.</returns>
		/// <remarks>
		/// <para>This contains the stamp counter at the beginning.</para>
		/// 
		/// <para>This buffer is the very short in order to mVela.Imize the generation speed.</para>
		/// </remarks>
		private byte[] GetBufferToHash()
		{
			byte[] counterBuffer = BitConverter.GetBytes( Counter );
			byte[] suffixBuffer = GetSuffixBuffer();

			byte[] buffer = new byte[ counterBuffer.Length + suffixBuffer.Length ];

			Buffer.BlockCopy( counterBuffer, 0, buffer, 0, counterBuffer.Length );
			Buffer.BlockCopy( suffixBuffer, 0, buffer, counterBuffer.Length, suffixBuffer.Length );

			return buffer;
		}

		/// <summary>
		/// Generates or resumes the generation of a stamp which fulfills the <see cref="Cost"/>.
		/// </summary>
		/// <param name="resume">If this is true then the stamp generation has been resumed.</param>
		/// <param name="lop">Long operation progress used during the generation of a stamp. This may be null.</param>
		/// <returns>Returns the number of iterations consumed by the generation.</returns>
		/// <remarks>
		/// <para>The generation of a stamp is interruptible, that is, its generation state can be saved by the calling
		/// process, and its generation can then be interrupted. The generation stamp of the stamp can be restored later, and its
		/// generation can be then be resumed.</para>
		/// 
		/// <para>Calling processes can save the generation state in the "lop" delegate.</para>
		/// 
		/// <para>If the ability to precalculate a stamp can make the stamp useless, it may be useful to link the stamp to the
		/// data ("linkedData") for which the stamp is intended (to act as proof of work).</para>
		/// </remarks>
		private int GenerateOrResume( bool resume, LongOperationProgress lop )
		{
			// Verify.
			if( (Counter > 0) && (GetCost() >= Cost) ) return 0;

			if( !resume )
			{
				if( Counter > 0 )
					throw new ValidationException( "The stamp is being generated, but the counter is not 0." );
				if( CollisionHash != null )
					throw new ValidationException( "The stamp is being generated, but the hash collision is not missing." );
				if( GenerationCompleted )
					throw new ValidationException( "The stamp is being generated, but the generation is completed." );
			}

			// Do progress.
			if( lop != null ) lop.Progress( 0, "Starting stamp generation...", true );

			// The buffer to hash.
			byte[] bufferToLoop = GetBufferToHash();

			int counterSize = sizeof( UInt64 );
			int counterPos = 0;

			Chronometer chrono = new Chronometer();
			chrono.Start();

			// Find a hash collision.
			int i = 0;
			while( true )
			{
				// Increment the stamp counter. A stamp can be valid only if the stamp counter is at least 1. The default stamp
				// counter is 0, so increment it before the hash is computed.
				BufferUtils.Increment( bufferToLoop, counterPos, counterSize, false );

				i++;

				CollisionHash = HashAlg.Compute( bufferToLoop );

				// If the cost is fulfilled, break the cycle.
				if( GetCost() >= Cost )
				{
					// Extract the stamp counter from the buffer.
					Counter = BitConverter.ToUInt64( bufferToLoop, counterPos );

					// Do progress.
					if( lop != null ) lop.Progress( 0, "Verifying the generated stamp...", true );

					// For safety reasons, validate the collision hash.
					FulfillsCost( Cost, true );

					return i;
				}

				// Do progress.
				if( (lop != null) && chrono.PassedSeconds( lop.UpdateTimespan ) ) lop.Progress( 0, "Generating stamp...", true );
			}
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Gets the cost of the stamp.
		/// </summary>
		/// <returns>Returns the cost of the stamp.</returns>
		public int GetCost()
		{
			return BufferUtils.CountLs0Bits( CollisionHash );
		}

		/// <summary>
		/// Checks that the collision hash of the stamp fulfills the specified cost. If the cost is not fulfilled, an
		/// <see cref="UnfulfilledStampCostException"/> exception is thrown if "throwNotFulfilled" is true.
		/// </summary>
		/// <param name="cost">Cost that the collision hash of the stamp must fulfill in order to be considered valid. This must
		/// be at least 1.</param>
		/// <param name="throwNotFulfilled">If this is true and the current stamp doesn't fulfill the specified cost, an
		/// exception is thrown.</param>
		/// <returns>Returns true if the stamp fulfills the required stamp cost.</returns>
		public bool FulfillsCost( int cost, bool throwNotFulfilled )
		{
			ValidationUtils.Integer( cost, 1, CostMax, "cost" );

			if( Counter <= 0 ) return false;

			// The buffer to hash.
			byte[] buffer = GetBufferToHash();

			// The actual collision hash.
			byte[] collisionHash = HashAlg.Compute( buffer );

			// Check if the cost of the actual collision hash fulfills the required cost.
			int actualCost = BufferUtils.CountLs0Bits( collisionHash );
			if( (actualCost < cost) || (actualCost < Cost) )
			{
				if( !throwNotFulfilled ) return false;

				throw new UnfulfilledStampCostException( string.Format( "The stamp \"{0}\" doesn't fulfill the cost {1}."
					, TextUtils.ArrayToMoteText16( collisionHash ), cost ) );
			}

			// Check if the collision hash and the actual collision hash are a match.
			if( !BufferUtils.Equals( collisionHash, CollisionHash ) )
			{
				if( !throwNotFulfilled ) return false;

				throw new UnfulfilledStampCostException( string.Format( "The stamp \"{0}\" doesn't fulfill the cost {1}."
					, TextUtils.ArrayToMoteText16( CollisionHash ), Cost ) );
			}

			// Cost fulfilled.
			return true;
		}

		/// <summary>
		/// Checks that the collision hash of the stamp fulfills its cost.
		/// </summary>
		/// <returns>Returns true if the stamp fulfills its cost.</returns>
		public bool FulfillsSelfCost()
		{
			return FulfillsCost( Cost, false );
		}

		/// <summary>
		/// Generates a stamp which fulfills the <see cref="Cost"/>.
		/// </summary>
		/// <param name="lop">Long operation progress used during the generation of a stamp. This may be null.</param>
		/// <returns>Returns the number of iterations consumed by the generation.</returns>
		/// <remarks>
		/// <para>The generation of a stamp is interruptible, that is, its generation state can be saved by the calling
		/// process, and its generation can then be interrupted. The generation stamp of the stamp can be restored later, and its
		/// generation can be then be resumed.</para>
		/// 
		/// <para>Calling processes can save the generation state in the "lop" delegate.</para>
		/// 
		/// <para>If the ability to precalculate a stamp can make the stamp useless, it may be useful to link the stamp to the
		/// data ("linkedData") for which the stamp is intended (to act as proof of work).</para>
		/// </remarks>
		public int Generate( LongOperationProgress lop )
		{
			return GenerateOrResume( false, lop );
		}

		/// <summary>
		/// Resumes the generation of a stamp which fulfills the <see cref="Cost"/>.
		/// </summary>
		/// <param name="lop">Long operation progress used during the generation of a stamp. This may be null.</param>
		/// <returns>Returns the number of iterations consumed by the generation.</returns>
		/// <remarks>
		/// <para>The generation of a stamp is interruptible, that is, its generation state can be saved by the calling
		/// process, and its generation can then be interrupted. The generation stamp of the stamp can be restored later, and its
		/// generation can be then be resumed.</para>
		/// 
		/// <para>Calling processes can save the generation state in the "lop" delegate.</para>
		/// </remarks>
		public int Resume( LongOperationProgress lop )
		{
			return GenerateOrResume( true, lop );
		}

		#endregion Public methods
	}
}
