////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Crypto.Hash;
using Vela.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// A stamp with a recipient.
	/// </summary>
	public class StampWithRecipient : Stamp, IInitiatorUpnFingerprintContainer, IRecipientUpnFingerprintContainer
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="InitiatorUpnFingerprint"/> property.
		/// </summary>
		public const string InitiatorUpnFingerprintFilterIdentifier = "StampWithRecipient.InitiatorUpnFingerprint#6";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="RecipientUpnFingerprint"/> property.
		/// </summary>
		public const string RecipientUpnFingerprintFilterIdentifier = "StampWithRecipient.RecipientUpnFingerprint#7";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <remarks>
		/// <para>In order to generate the stamp, call the <see cref="Stamp.Generate"/> method.</para>
		/// </remarks>
		protected StampWithRecipient()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="hashAlg">Hash algorithm.</param>
		/// <param name="hashSerializer">Hash serializer.</param>
		/// <param name="documentExpirationTime">The expiration time (in UTC format) of the document. The current document is
		/// valid before this time; from and after this time, it should not be used in processing.</param>
		/// <param name="cost">The cost of the stamp.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpnFingerprint">The UPN fingerprint of the recipient of the current object.</param>
		/// <remarks>
		/// <para>In order to generate the stamp, call the <see cref="Stamp.Generate"/> method.</para>
		/// 
		/// <para>The salt is randomly generated in here.</para>
		/// </remarks>
		public StampWithRecipient( Locator instanceLocator, HashAlg hashAlg, IHashSerializer hashSerializer
				, DateTime documentExpirationTime, int cost, Locator initiatorUpnFingerprint, Locator recipientUpnFingerprint )
			: base( instanceLocator, hashAlg, hashSerializer, documentExpirationTime, cost )
		{
			InitiatorUpnFingerprint = initiatorUpnFingerprint;
			RecipientUpnFingerprint = recipientUpnFingerprint;
		}

		#endregion Constructors

		#region Public serializable properties

		private Locator _InitiatorUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the initiator of the current object. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( InitiatorUpnFingerprintFilterIdentifier )]
		public Locator InitiatorUpnFingerprint
		{
			get { return _InitiatorUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "InitiatorUpnFingerprint" );

				_InitiatorUpnFingerprint = value;
			}
		}

		private Locator _RecipientUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the recipient of the current object. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( RecipientUpnFingerprintFilterIdentifier )]
		public Locator RecipientUpnFingerprint
		{
			get { return _RecipientUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RecipientUpnFingerprint" );

				_RecipientUpnFingerprint = value;
			}
		}

		#endregion Public serializable properties
	}
}
