////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto.Hash;
using Vela.Crypto.Symmetric;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Contract for a steganographic algorithm which doesn't support unencrypted action responses for the frames of a multiframe
	/// response.
	/// </summary>
	/// <remarks>
	/// <para>This class and the <see cref="MacedSteganographicAlg"/> class are separated for legacy reasons.</para>
	/// </remarks>
	public abstract class HashedSteganographicAlg : SteganographicAlg
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected HashedSteganographicAlg()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="hashAlg">Hash algorithm.</param>
		/// <param name="symmAlg">Symmetric cryptography algorithm.</param>
		/// <param name="disalignmentSizeZenith">The upper bound for the disalignment size. The disalignment size may be between
		/// [0..."disalignmentSizeZenith"/>). If this is 0 or a negative number, it's ignored.</param>
		protected HashedSteganographicAlg( HashAlg hashAlg, SymmAlg symmAlg, int disalignmentSizeZenith )
			: base( symmAlg, disalignmentSizeZenith )
		{
			HashAlg = hashAlg;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( base.ExpirationTime, HashAlg.ExpirationTime ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private HashAlg _HashAlg;
		/// <summary>
		/// Hash algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public HashAlg HashAlg
		{
			get { return _HashAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "HashAlg" );

				_HashAlg = value;
			}
		}

		#endregion Public serializable properties
	}
}
