////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto.Hash;
using Vela.Crypto.Mac;
using Vela.Crypto.Symmetric;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Contract for a steganographic algorithm which supports unencrypted action responses for the frames of a multiframe
	/// response.
	/// </summary>
	/// <remarks>
	/// <para>This class and the <see cref="HashedSteganographicAlg"/> class are separated for legacy reasons.</para>
	/// </remarks>
	public abstract class MacedSteganographicAlg : SteganographicAlg
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected MacedSteganographicAlg()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="symmAlg">Symmetric cryptography algorithm.</param>
		/// <param name="disalignmentSizeZenith">The upper bound for the disalignment size. The disalignment size may be between
		/// [0..."disalignmentSizeZenith"/>). If this is 0 or a negative number, it's ignored.</param>
		protected MacedSteganographicAlg( MacAlg macAlg, SymmAlg symmAlg, int disalignmentSizeZenith )
			: base( symmAlg, disalignmentSizeZenith )
		{
			MacAlg = macAlg;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// Hash algorithm. READ-ONLY.
		/// </summary>
		public HashAlg HashAlg
		{
			get { return MacAlg.HashAlg; }
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( base.ExpirationTime, MacAlg.ExpirationTime ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private MacAlg _MacAlg;
		/// <summary>
		/// MAC algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public MacAlg MacAlg
		{
			get { return _MacAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "MacAlg" );

				_MacAlg = value;
			}
		}

		#endregion Public serializable properties
	}
}
