////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto;
using Vela.Crypto.Hash;
using Vela.Crypto.Mac;
using Vela.Crypto.Symmetric;
using Vela.Kernel;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// "MACed Binary Steganography" steganographic algorithm.
	/// </summary>
	/// <remarks>
	/// <para>The obfuscated data contains exactly the minimum information required by <see cref="SteganographicAlg"/>.</para>
	/// </remarks>
	public class Mist : MacedSteganographicAlg
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Mist()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="symmAlg">Symmetric cryptography algorithm.</param>
		/// <param name="disalignmentSizeZenith">The upper bound for the disalignment size. The disalignment size may be between
		/// [0..."disalignmentSizeZenith"/>). If this is 0 or a negative number, it's ignored.</param>
		public Mist( MacAlg macAlg, SymmAlg symmAlg, int disalignmentSizeZenith )
			: base( macAlg, symmAlg, disalignmentSizeZenith )
		{
		}

		#endregion Constructors

		#region Public methods

		/// <summary>
		/// Extracts the salt and the riddle from the specified data.
		/// </summary>
		/// <param name="buffer">Buffer from which to extract the salt and the riddle.</param>
		/// <param name="salt">Is set with the extracted salt. It will have the same size as the symmetric encryption algorithm.
		/// </param>
		/// <param name="riddle">Is set with the extracted riddle.</param>
		public override void ExtractSaltAndRiddle( byte[] buffer, out byte[] salt, out byte[] riddle )
		{
			int pos = 0;

			salt = new byte[ SymmAlg.AlgSizeBytes ];
			Buffer.BlockCopy( buffer, pos, salt, 0, salt.Length );
			pos += salt.Length;

			riddle = new byte[ HashAlg.AlgSizeBytes ];
			Buffer.BlockCopy( buffer, pos, riddle, 0, riddle.Length );
			pos += riddle.Length;
		}

		/// <summary>
		/// Generates the riddle.
		/// </summary>
		/// <param name="salt">Salt.</param>
		/// <param name="headerEncryptionKey">Header encryption key. This should be the UPN symmetric encryption key of the
		/// recipient.</param>
		/// <param name="fingerprint">Identifying information. This should be the UPN fingerprint of the recipient.</param>
		/// <returns>Returns the generated riddle.</returns>
		public override byte[] GenerateRiddle( byte[] salt, byte[] headerEncryptionKey, byte[] fingerprint )
		{
			return HashAlg.Compute( null, true, salt, headerEncryptionKey, fingerprint );
		}

		/// <summary>
		/// Obfuscates the specified data.
		/// </summary>
		/// <param name="buffer">Buffer to obfuscate.</param>
		/// <param name="headerEncryptionKey">Header encryption key. This should be the UPN symmetric encryption key of the
		/// recipient.</param>
		/// <param name="fingerprint">Identifying information. This should be the UPN fingerprint of the recipient.</param>
		/// <param name="overridingDisalignmentSizeZenith">If this is higher than 0, it's used instead of the
		/// <see cref="SteganographicAlg.DisalignmentSizeZenith"/> property, else it's ignored.</param>
		/// <returns>Returns the obfuscated data.</returns>
		/// <remarks>
		/// <para>The body encryption key and the disalignment size are generated in here.</para>
		/// </remarks>
		public override byte[] Obfuscate( byte[] buffer, byte[] headerEncryptionKey, byte[] fingerprint
			, int overridingDisalignmentSizeZenith = 0 )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );
			ValidationUtils.ArrayHasContent( headerEncryptionKey, "headerEncryptionKey" );
			ValidationUtils.ArrayHasContent( fingerprint, "fingerprint" );

			int algSizeBytes = SymmAlg.AlgSizeBytes;

			byte[] alignedHeaderEncryptionKey = SymmAlg.AlignToAlgSize( headerEncryptionKey );

			byte[] salt = EntropyPool.GetBytes( algSizeBytes );
			byte[] riddle = GenerateRiddle( salt, headerEncryptionKey, fingerprint );

			// Encrypt the encryption key of the body.
			byte[] bodyEncryptionKey = EntropyPool.GetBytes( algSizeBytes );
			byte[] encryptedBodyEncryptionKey = SymmAlg.Encrypt( bodyEncryptionKey, SymmPaddings.None, SymmAlg.Chaining, salt
				, alignedHeaderEncryptionKey );

			// Disalignment size.
			byte[] disalignmentSizeBuffer = EntropyPool.GetBytes( algSizeBytes );
			int disalignmentSize = SetDisalignmentSize( disalignmentSizeBuffer, DisalignmentSizeZenith
				, overridingDisalignmentSizeZenith );
			byte[] encryptedDisalignmentSizeBuffer = SymmAlg.Encrypt( disalignmentSizeBuffer, SymmPaddings.None
				, SymmAlg.Chaining, salt, alignedHeaderEncryptionKey );

			byte[] disalignmentData = (disalignmentSize >= 1) ? EntropyPool.GetBytes( disalignmentSize ) : new byte[ 0 ];

			// Encrypt the body.
			byte[] encryptedBody = SymmAlg.Encrypt( buffer, salt, bodyEncryptionKey );

			// Compute the MAC of the body.
			byte[] macKey = MacAlg.HashAlg.Compute( null, true, salt, bodyEncryptionKey );
			byte[] macBody = MacAlg.Compute( macKey, null, true, salt, encryptedBody );

			byte[] obfuscatedData = new byte[ salt.Length + riddle.Length + macBody.Length + encryptedBodyEncryptionKey.Length
				+ encryptedDisalignmentSizeBuffer.Length + disalignmentData.Length + encryptedBody.Length ];

			int pos = 0;

			Buffer.BlockCopy( salt, 0, obfuscatedData, pos, salt.Length );
			pos += salt.Length;

			Buffer.BlockCopy( riddle, 0, obfuscatedData, pos, riddle.Length );
			pos += riddle.Length;

			Buffer.BlockCopy( macBody, 0, obfuscatedData, pos, macBody.Length );
			pos += macBody.Length;

			Buffer.BlockCopy( encryptedBodyEncryptionKey, 0, obfuscatedData, pos, encryptedBodyEncryptionKey.Length );
			pos += encryptedBodyEncryptionKey.Length;

			Buffer.BlockCopy( encryptedDisalignmentSizeBuffer, 0, obfuscatedData, pos, encryptedDisalignmentSizeBuffer.Length );
			pos += encryptedDisalignmentSizeBuffer.Length;

			Buffer.BlockCopy( disalignmentData, 0, obfuscatedData, pos, disalignmentData.Length );
			pos += disalignmentData.Length;

			Buffer.BlockCopy( encryptedBody, 0, obfuscatedData, pos, encryptedBody.Length );
			pos += encryptedBody.Length;

			return obfuscatedData;
		}

		/// <summary>
		/// Reveals the previously obfuscated data from the specified data.
		/// </summary>
		/// <param name="buffer">Buffer from which to reveal the previously obfuscated data.</param>
		/// <param name="headerEncryptionKey">Header encryption key. This should be the UPN symmetric encryption key of the
		/// recipient.</param>
		/// <param name="fingerprint">Identifying information. This should be the UPN fingerprint of the recipient.</param>
		/// <param name="overridingDisalignmentSizeZenith">If this is higher than 0, it's used instead of the
		/// <see cref="SteganographicAlg.DisalignmentSizeZenith"/> property, else it's ignored.</param>
		/// <returns>Returns the revealed data.</returns>
		[Obsolete( "The 'overridingDisalignmentSizeZenith' parameter (and this attribute) should be removed once all users"
			+ " switch to version 2. Before this version, the disalignment size was extracted from a random buffer as a"
			+ " remainder of a modulo, but now it's set in the buffer (by the 'SetDisalignmentSize' method) as an integer"
			+ " number, so that it can be read by the 'ExtractDisalignmentSize' method without knowing the used modulo /"
			+ " zenith)." )]
		public override byte[] Reveal( byte[] buffer, byte[] headerEncryptionKey, byte[] fingerprint
			, int overridingDisalignmentSizeZenith )
		{
			byte[] alignedHeaderEncryptionKey = SymmAlg.AlignToAlgSize( headerEncryptionKey );

			int algSizeBytes = SymmAlg.AlgSizeBytes;
			int pos = 0;

			byte[] salt, riddle;
			ExtractSaltAndRiddle( buffer, out salt, out riddle );
			pos += salt.Length + riddle.Length;

			byte[] macBody = new byte[ MacAlg.AlgSizeBytes ];
			Buffer.BlockCopy( buffer, pos, macBody, 0, macBody.Length );
			pos += macBody.Length;

			byte[] encryptedBodyEncryptionKey = new byte[ algSizeBytes ];
			Buffer.BlockCopy( buffer, pos, encryptedBodyEncryptionKey, 0, encryptedBodyEncryptionKey.Length );
			pos += encryptedBodyEncryptionKey.Length;

			// Decrypt the body encryption key.
			byte[] bodyEncryptionKey = SymmAlg.Decrypt( encryptedBodyEncryptionKey, SymmPaddings.None, SymmAlg.Chaining, salt
				, alignedHeaderEncryptionKey );

			byte[] encryptedDisalignmentSizeBuffer = new byte[ algSizeBytes ];
			Buffer.BlockCopy( buffer, pos, encryptedDisalignmentSizeBuffer, 0, encryptedDisalignmentSizeBuffer.Length );
			pos += encryptedDisalignmentSizeBuffer.Length;

			// Get the disalignment size and skip the disalignment data.
			byte[] decryptedDisalignmentSizeBuffer = SymmAlg.Decrypt( encryptedDisalignmentSizeBuffer, SymmPaddings.None
				, SymmAlg.Chaining, salt, alignedHeaderEncryptionKey );
			int disalignmentSize = ExtractDisalignmentSize( decryptedDisalignmentSizeBuffer, DisalignmentSizeZenith
				, overridingDisalignmentSizeZenith );
			pos += disalignmentSize;

			// Verify the riddle.
			byte[] generatedRiddle = GenerateRiddle( salt, headerEncryptionKey, fingerprint );
			if( !BufferUtils.Equals( riddle, generatedRiddle ) )
			{
				throw new ValidationException( string.Format( "The generated riddle is \"{0}\" instead of \"{1}\"."
					, TextUtils.ArrayToMoteText16( generatedRiddle ), TextUtils.ArrayToMoteText16( riddle ) ) );
			}

			byte[] encryptedBody = new byte[ buffer.Length - pos ];
			Buffer.BlockCopy( buffer, pos, encryptedBody, 0, buffer.Length - pos );

			// Verify the MAC of the body.
			byte[] macKey = MacAlg.HashAlg.Compute( null, true, salt, bodyEncryptionKey );
			byte[] generatedMacBody = MacAlg.Compute( macKey, null, true, salt, encryptedBody );
			if( !BufferUtils.Equals( macBody, generatedMacBody ) )
			{
				throw new ValidationException( string.Format( "The generated body MAC is \"{0}\" instead of \"{1}\"."
					, TextUtils.ArrayToMoteText16( generatedMacBody ), TextUtils.ArrayToMoteText16( macBody ) ) );
			}

			// Decrypt the body.
			byte[] decryptedData = SymmAlg.Decrypt( encryptedBody, salt, bodyEncryptionKey );

			return decryptedData;
		}

		#endregion Public methods
	}
}
