////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto.Symmetric;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Contract for a steganographic algorithm.
	/// </summary>
	/// <remarks>
	/// <para>A steganographic algorithm is meant to be used by a communication frame algorithm
	/// (<see cref="ICommunicationFrameAlg"/>) in order to prevent an attacker to have a way of detecting that an Vela
	/// communication is taking place. It's purpose is not to ofer security (although it uses encryption), and it's security
	/// is below that offered by an encrypted envelope.</para>
	/// 
	/// <para>The data to obfuscate is encrypted and outputted without any identifying marks. However, the obfuscated data may
	/// be integrated within some common container, like an image or an HTML page which contains an image.</para>
	/// 
	/// <para>The fingerprint is a piece of information used to identify the recipient of obfuscated data. The key used for the
	/// symmetric encryption of the data is taken from the recipient information.</para>
	/// 
	/// <para>For action requests, the fingerprint should be the UPN fingerprint of the recipient, and the symmetric encryption
	/// key should be the UPN encryption key of the recipient.</para>
	/// 
	/// <para>For action responses, the fingerprint should be the UPN fingerprint of the recipient, and the symmetric encryption
	/// key should be the response encryption key from the action request.</para>
	/// 
	/// <para>The obfuscated data should contain the following:</para>
	/// <list type="bullet">
	///		<item>Salt.</item>
	///		<item>Riddle. This is a hash of the salt, header encryption key and fingerprint. This is used to be able to quickly
	///			verify if the revealing (<see cref="SteganographicAlg.Reveal(byte[], Upn)"/>) is correct, since it may be
	///			necessary to iterate through a lot of potential recipients (like UPNs) in order to find the correct one.</item>
	///		<item>MAC of the body and encryption key. This is only for algorithms derived from
	///			<see cref="MacedSteganographicAlg"/></item>
	///		<item>Encrypted body encryption key.</item>
	///		<item>Encrypted buffer with disalignment size.</item>
	///		<item>Disalignment data. This is a random buffer with a size equal with the disalignment size. It exists in
	///			order to break the intrinsic alignment of the cryptographic algorithms (for example, at 32 bytes).</item>
	///		<item>Encrypted body.</item>
	/// </list>
	/// 
	///	<para>If the disalignment size is 0 or a negative number, it's ignored.</para>
	/// </remarks>
	public abstract class SteganographicAlg : IImmutableAlgorithm, IExpirableAlgorithm
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected SteganographicAlg()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="symmAlg">Symmetric cryptography algorithm.</param>
		/// <param name="disalignmentSizeZenith">The upper bound for the disalignment size. The disalignment size may be between
		/// [0...<see cref="DisalignmentSizeZenith"/>). If this is 0 or a negative number, it's ignored.</param>
		protected SteganographicAlg( SymmAlg symmAlg, int disalignmentSizeZenith )
		{
			SymmAlg = symmAlg;
			DisalignmentSizeZenith = disalignmentSizeZenith;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public virtual DateTime ExpirationTime
		{
			get { return SymmAlg.ExpirationTime; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private SymmAlg _SymmAlg;
		/// <summary>
		/// Symmetric cryptography algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public SymmAlg SymmAlg
		{
			get { return _SymmAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "SymmAlg" );

				_SymmAlg = value;
			}
		}

		private int _DisalignmentSizeZenith = CoreDocumentsConsts.DisalignmentSizeZenith;
		/// <summary>
		/// The upper bound for the disalignment size. The disalignment size may be between
		/// [0...<see cref="DisalignmentSizeZenith"/>). If this is 0 or a negative number, it's ignored. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public int DisalignmentSizeZenith
		{
			get { return _DisalignmentSizeZenith; }
			protected set { _DisalignmentSizeZenith = value; }
		}

		#endregion Public serializable properties

		#region Private static methods

		/// <summary>
		/// Sets the disalignment size in the specified buffer.
		/// </summary>
		/// <param name="disalignmentSizeBuffer">Buffer from which to generate the disalignment size, and in which to set it.
		/// </param>
		/// <param name="disalignmentSizeZenith">The upper bound for the disalignment size. The disalignment size may be between
		/// [0..."disalignmentSizeZenith"). If this is 0 or a negative number, it's ignored.</param>
		/// <param name="overridingDisalignmentSizeZenith">If this is higher than 0, it's used instead of the
		/// "disalignmentSizeZenith", else it's ignored.</param>
		/// <returns>Returns the disalignment size.</returns>
		protected static int SetDisalignmentSize( byte[] disalignmentSizeBuffer, int disalignmentSizeZenith
			, int overridingDisalignmentSizeZenith )
		{
			int actualDisalignmentSizeZenith = (overridingDisalignmentSizeZenith > 0) ? overridingDisalignmentSizeZenith
				: disalignmentSizeZenith;

			if( actualDisalignmentSizeZenith <= 0 )
				return 0;

			// Extract the disalignment size.
			UInt32 disalignmentSize = (UInt32)(BitConverter.ToUInt32( disalignmentSizeBuffer, 0 )
				% actualDisalignmentSizeZenith);

			// Set the disalignment size in the buffer as an integer number, so that it can be read (by the
			// "ExtractDisalignmentSize" method) without knowing the used modulo / zenith.
			byte[] strictSizeBuffer = BitConverter.GetBytes( disalignmentSize );
			Buffer.BlockCopy( strictSizeBuffer, 0, disalignmentSizeBuffer, 0, strictSizeBuffer.Length );

			return (int)disalignmentSize;
		}

		/// <summary>
		/// Extracts the disalignment size from the specified buffer.
		/// </summary>
		/// <param name="disalignmentSizeBuffer">Buffer from which to extract the disalignment size</param>
		/// <param name="disalignmentSizeZenith">The upper bound for the disalignment size. The disalignment size may be between
		/// [0..."disalignmentSizeZenith"). If this is 0 or a negative number, it's ignored.</param>
		/// <param name="overridingDisalignmentSizeZenith">If this is higher than 0, it's used instead of the
		/// "disalignmentSizeZenith", else it's ignored.</param>
		/// <returns>Returns the extracted disalignment size.</returns>
		[Obsolete( "The 'overridingDisalignmentSizeZenith' parameter (and this attribute) should be removed once all users"
			+ " switch to version 2. Before this version, the disalignment size was extracted from a random buffer as a"
			+ " remainder of a modulo, but now it's set in the buffer (by the 'SetDisalignmentSize' method) as an integer"
			+ " number, so that it can be read by the 'ExtractDisalignmentSize' method without knowing the used modulo /"
			+ " zenith)." )]
		protected static int ExtractDisalignmentSize( byte[] disalignmentSizeBuffer, int disalignmentSizeZenith
			, int overridingDisalignmentSizeZenith )
		{
			int actualDisalignmentSizeZenith = (overridingDisalignmentSizeZenith > 0) ? overridingDisalignmentSizeZenith
				: disalignmentSizeZenith;

			if( actualDisalignmentSizeZenith <= 0 )
				return 0;

			// Extract the disalignment size.
			int disalignmentSize = (int)(BitConverter.ToUInt32( disalignmentSizeBuffer, 0 ) % actualDisalignmentSizeZenith);

			return disalignmentSize;
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Reveals the raw document from the specified obfuscated document, using the specified recipient UPN.
		/// </summary>
		/// <param name="obfuscatedDocument">Obfuscated document.</param>
		/// <param name="recipientUpn">The UPN of the recipient for which the document is intended.</param>
		/// <param name="overridingDisalignmentSizeZenith">If this is higher than 0, it's used instead of the
		/// <see cref="DisalignmentSizeZenith"/> property, else it's ignored.</param>
		/// <returns>Returns the raw document. This may be null.</returns>
		[Obsolete( "The 'overridingDisalignmentSizeZenith' parameter (and this attribute) should be removed once all users"
			+ " switch to version 2. Before this version, the disalignment size was extracted from a random buffer as a"
			+ " remainder of a modulo, but now it's set in the buffer (by the 'SetDisalignmentSize' method) as an integer"
			+ " number, so that it can be read by the 'ExtractDisalignmentSize' method without knowing the used modulo /"
			+ " zenith)." )]
		public static byte[] Reveal( byte[] obfuscatedDocument, Upn recipientUpn, int overridingDisalignmentSizeZenith )
		{
			ValidationUtils.ObjectHasContent( obfuscatedDocument, "obfuscatedDocument" );
			ValidationUtils.ObjectHasContent( recipientUpn, "recipientUpn" );

			SteganographicAlg steganographicAlg = recipientUpn.UpnAlg.SteganographicAlg;

			byte[] salt, riddle;
			steganographicAlg.ExtractSaltAndRiddle( obfuscatedDocument, out salt, out riddle );

			byte[] computedRiddle = steganographicAlg.GenerateRiddle( salt, recipientUpn.SymmEncryptionKey
				, recipientUpn.Fingerprint.Value );

			if( BufferUtils.Equals( riddle, computedRiddle ) )
			{
				// Reveal the document.
				return steganographicAlg.Reveal( obfuscatedDocument, recipientUpn.SymmEncryptionKey
					, recipientUpn.Fingerprint.Value, overridingDisalignmentSizeZenith );
			}

			return null;
		}

		/// <summary>
		/// Reveals the raw document from the specified obfuscated document, using the specified recipient UPN.
		/// </summary>
		/// <param name="obfuscatedDocument">Obfuscated document.</param>
		/// <param name="recipientUpn">The UPN of the recipient for which the document is intended.</param>
		/// <returns>Returns the raw document. This may be null.</returns>
		public static byte[] Reveal( byte[] obfuscatedDocument, Upn recipientUpn )
		{
			return Reveal( obfuscatedDocument, recipientUpn, 0 );
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Extracts the salt and the riddle from the specified data.
		/// </summary>
		/// <param name="buffer">Buffer from which to extract the salt and the riddle.</param>
		/// <param name="salt">Is set with the extracted salt. It will have the same size as the symmetric encryption algorithm.
		/// </param>
		/// <param name="riddle">Is set with the extracted riddle.</param>
		public abstract void ExtractSaltAndRiddle( byte[] buffer, out byte[] salt, out byte[] riddle );

		/// <summary>
		/// Generates the riddle.
		/// </summary>
		/// <param name="salt">Salt.</param>
		/// <param name="headerEncryptionKey">Header encryption key. This should be the UPN symmetric encryption key of the
		/// recipient.</param>
		/// <param name="fingerprint">Identifying information. This should be the UPN fingerprint of the recipient.</param>
		/// <returns>Returns the generated riddle.</returns>
		public abstract byte[] GenerateRiddle( byte[] salt, byte[] headerEncryptionKey, byte[] fingerprint );

		/// <summary>
		/// Obfuscates the specified data.
		/// </summary>
		/// <param name="buffer">Buffer to obfuscate.</param>
		/// <param name="headerEncryptionKey">Header encryption key. This should be the UPN symmetric encryption key of the
		/// recipient.</param>
		/// <param name="fingerprint">Identifying information. This should be the UPN fingerprint of the recipient.</param>
		/// <param name="overridingDisalignmentSizeZenith">If this is higher than 0, it's used instead of the
		/// <see cref="DisalignmentSizeZenith"/> property, else it's ignored.</param>
		/// <returns>Returns the obfuscated data.</returns>
		/// <remarks>
		/// <para>The body encryption key and the disalignment size are generated in here.</para>
		/// </remarks>
		public abstract byte[] Obfuscate( byte[] buffer, byte[] headerEncryptionKey, byte[] fingerprint
			, int overridingDisalignmentSizeZenith = 0 );

		/// <summary>
		/// Reveals the previously obfuscated data from the specified data.
		/// </summary>
		/// <param name="buffer">Buffer from which to reveal the previously obfuscated data.</param>
		/// <param name="headerEncryptionKey">Header encryption key. This should be the UPN symmetric encryption key of the
		/// recipient.</param>
		/// <param name="fingerprint">Identifying information. This should be the UPN fingerprint of the recipient.</param>
		/// <param name="overridingDisalignmentSizeZenith">If this is higher than 0, it's used instead of the
		/// <see cref="DisalignmentSizeZenith"/> property, else it's ignored.</param>
		/// <returns>Returns the revealed data.</returns>
		[Obsolete( "The 'overridingDisalignmentSizeZenith' parameter (and this attribute) should be removed once all users"
			+ " switch to version 2. Before this version, the disalignment size was extracted from a random buffer as a"
			+ " remainder of a modulo, but now it's set in the buffer (by the 'SetDisalignmentSize' method) as an integer"
			+ " number, so that it can be read by the 'ExtractDisalignmentSize' method without knowing the used modulo /"
			+ " zenith)." )]
		public abstract byte[] Reveal( byte[] buffer, byte[] headerEncryptionKey, byte[] fingerprint
			, int overridingDisalignmentSizeZenith );

		/// <summary>
		/// Reveals the previously obfuscated data from the specified data.
		/// </summary>
		/// <param name="buffer">Buffer from which to reveal the previously obfuscated data.</param>
		/// <param name="headerEncryptionKey">Header encryption key. This should be the UPN symmetric encryption key of the
		/// recipient.</param>
		/// <param name="fingerprint">Identifying information. This should be the UPN fingerprint of the recipient.</param>
		/// <returns>Returns the revealed data.</returns>
		public byte[] Reveal( byte[] buffer, byte[] headerEncryptionKey, byte[] fingerprint )
		{
			return Reveal( buffer, headerEncryptionKey, fingerprint, 0 );
		}

		#endregion Public methods
	}
}
