////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Contract for a document which may be stored in a storehouse.
	/// </summary>
	/// <remarks>
	/// <para>A sheaf is stored encrypted in the storehouse, so its security doesn't depend on the encryption of the
	/// storehouse.</para>
	/// 
	/// <para>The maximum size of a sheaf should be in the order of hundreds of kilobytes.</para>
	/// </remarks>
	public interface ISheaf : ILocatable, IExpirableAlgorithm, IEncryptableDocument
	{
		#region Public non-serializable properties

		/// <summary>
		/// The lock which was used when the document was loaded. RESERVED for internal use. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>A sheaf / speckle may be saved only if it was first locked for writing (with either the
		/// <see cref="SheafLocks.Update"/> or <see cref="SheafLocks.Exclusive"/> locks).</para>
		/// </remarks>
		SheafLocks StorehouseLock { get; }

		/// <summary>
		/// The update version of the sheaf. This may be null or empty. RESERVED for internal use. This is automatically set when
		/// the document is loaded, with the "SheafUpdateVersion" value loaded from the storehouse. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>A document must not be saved if this value is different than the "SheafUpdateVersion" value from the
		/// storehouse (because this is an indication that the document was saved by another thread / process).</para>
		/// </remarks>
		byte[] SheafUpdateVersion { get; }

		#endregion Public non-serializable properties

		#region Public serializable properties

		/// <summary>
		/// The creation time (in UTC format) of the document. This is automatically set with the time of the creation of the
		/// current object. The current object is valid from and after this time; before this time, it should not be used in
		/// processing. READ-ONLY.
		/// </summary>
		DateTime CreationTime { get; }

		/// <summary>
		/// The update time (in UTC format) of the document. This is automatically set with the update time. This is not
		/// signed / authenticated together with the document because it changes every time when the document is saved.
		/// </summary>
		DateTime UpdateTime { get; set; }

		/// <summary>
		/// If this is true then the sheaf is visible to the users. This is not signed / authenticated together with the document
		/// because it's meant to be changed.
		/// </summary>
		/// <remarks>
		/// <para>This should be used only by the applications which can / may not delete sheafs, and which would rather set
		/// this with false. Most applications may ignore this.</para>
		/// </remarks>
		bool Visible { get; set; }

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Sets the <see cref="StorehouseLock"/> and <see cref="SheafUpdateVersion"/> properties with the specified values.
		/// RESERVED for internal use.
		/// </summary>
		/// <param name="storehouseLock">The storehouse lock.</param>
		/// <param name="sheafUpdateVersion">The update version.</param>
		void SetLockAndUpdateVersion( SheafLocks storehouseLock, byte[] sheafUpdateVersion );

		/// <summary>
		/// Sets the <see cref="StorehouseLock"/> property with the specified value. RESERVED for internal use.
		/// </summary>
		/// <param name="storehouseLock">The storehouse lock.</param>
		void SetLock( SheafLocks storehouseLock );

		/// <summary>
		/// Sets the <see cref="SheafUpdateVersion"/> property with the specified value. RESERVED for internal use.
		/// </summary>
		/// <param name="sheafUpdateVersion">The update version.</param>
		void SetUpdateVersion( byte[] sheafUpdateVersion );

		/// <summary>
		/// Prepares the current sheaf to be updated, using the storehouse lock and sheaf update version from the specified
		/// sheaf.
		/// </summary>
		/// <param name="sheaf">Sheaf from which to get the storehouse lock and sheaf update version.</param>
		/// <remarks>
		/// <para>This is useful when an external sheaf (which has no storehouse lock and sheaf update version) updates an
		/// existing sheaf.</para>
		/// </remarks>
		void PrepareToUpdate( ISheaf sheaf );

		/// <summary>
		/// Prepares the current sheaf to be updated, using the storehouse lock and sheaf update version from the specified
		/// speckle.
		/// </summary>
		/// <param name="speckle">Speckle from which to get the storehouse lock and sheaf update version.</param>
		/// <remarks>
		/// <para>This is useful when an external sheaf (which has no storehouse lock and sheaf update version) updates an
		/// existing sheaf. Using this method, the existing sheaf doesn't have to be loaded, but only its associated speckle.
		/// </para>
		/// </remarks>
		void PrepareToUpdate( Speckle speckle );

		/// <summary>
		/// Collects the speckle filters from the current sheaf.
		/// </summary>
		/// <returns>Returns the list of collected speckle filters.</returns>
		SpeckleFilters CollectSpeckleFilters();

		#endregion Public methods
	}
}
