////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Reflection;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Contract for a document which is (normally) stored in a storehouse.
	/// </summary>
	public abstract class Sheaf : ISheaf
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="CreationTime"/> property.
		/// </summary>
		public const string CreationTimeFilterIdentifier = "Sheaf.CreationTime#0";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="UpdateTime"/> property.
		/// </summary>
		public const string UpdateTimeFilterIdentifier = "Sheaf.UpdateTime#1";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="Visible"/> property.
		/// </summary>
		public const string VisibleFilterIdentifier = "Sheaf.Visible#2";

		/// <summary>
		/// The identifier of the speckle filter which contains the UPN fingerprint of the processing identity from the current
		/// sheaf.
		/// </summary>
		/// <remarks>
		/// <para>This is centralized here because it's used dynamically (= untyped) to collect the sheafs which belong to a
		/// given identity UPN fingerprint. In a dynamic case it's not possible to have different filter identifiers for each
		/// type of sheaf to collect.</para>
		/// 
		/// <para>For example, this is used together with <see cref="ILocallyNamedPublicProcessingIdentitySheaf"/> and
		/// <see cref="ILocallyNamedPrivateProcessingIdentitySheaf"/>.</para>
		/// </remarks>
		public const string ProcessingIdentityUpnFingerprintFilterIdentifier = "Sheaf.ProcessingIdentityUpnFingerprint#3";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Sheaf()
		{
			UpdateTime = CreationTime;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">The locator of the current sheaf.</param>
		public Sheaf( Locator instanceLocator )
		{
			InstanceLocator = instanceLocator;
			UpdateTime = CreationTime;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public virtual DateTime ExpirationTime
		{
			get { return DateTime.MaxValue; }
		}

		private SheafLocks _StorehouseLock = SheafLocks.Shared;
		/// <summary>
		/// The lock which was used when the document was loaded. RESERVED for internal use. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>A sheaf / speckle may be saved only if it was first locked for writing (with either the
		/// <see cref="SheafLocks.Update"/> or <see cref="SheafLocks.Exclusive"/> locks).</para>
		/// </remarks>
		public SheafLocks StorehouseLock
		{
			get { return _StorehouseLock; }
			protected set { _StorehouseLock = value; }
		}

		private byte[] _SheafUpdateVersion;
		/// <summary>
		/// The update version of the sheaf. This may be null or empty. RESERVED for internal use. This is automatically set when
		/// the document is loaded, with the "SheafUpdateVersion" value loaded from the storehouse. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>A document must not be saved if this value is different than the "SheafUpdateVersion" value from the
		/// storehouse (because this is an indication that the document was saved by another thread / process).</para>
		/// </remarks>
		public byte[] SheafUpdateVersion
		{
			get { return _SheafUpdateVersion; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "SheafUpdateVersion" );

				_SheafUpdateVersion = value;
			}
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Locator _InstanceLocator;
		/// <summary>
		/// The locator of the current document. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Locator InstanceLocator
		{
			get { return _InstanceLocator; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "InstanceLocator" );

				_InstanceLocator = value;
			}
		}

		private DateTime _CreationTime = DateTime.UtcNow;
		/// <summary>
		/// The creation time (in UTC format) of the document. This is automatically set with the time of the creation of the
		/// current object. The current object is valid from and after this time; before this time, it should not be used in
		/// processing. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( CreationTimeFilterIdentifier )]
		public DateTime CreationTime
		{
			get { return _CreationTime; }
			protected set { _CreationTime = value; }
		}

		private DateTime _UpdateTime;
		/// <summary>
		/// The update time (in UTC format) of the document. This is automatically set with the update time. This is not
		/// signed / authenticated together with the document because it changes every time when the document is saved.
		/// </summary>
		[SerializeAttribute( HashSerialization = HashSerializations.NotInRoot )]
		[SpeckleFilterAttribute( UpdateTimeFilterIdentifier )]
		public DateTime UpdateTime
		{
			get { return _UpdateTime; }
			set { _UpdateTime = value; }
		}

		private bool _Visible = true;
		/// <summary>
		/// If this is true then the sheaf is visible to the users. This is not signed / authenticated together with the document
		/// because it's meant to be changed.
		/// </summary>
		/// <remarks>
		/// <para>This should be used only by the applications which can / may not delete sheafs, and which would rather set
		/// this with false. Most applications may ignore this.</para>
		/// </remarks>
		[SerializeAttribute( HashSerialization = HashSerializations.NotInRoot )]
		[SpeckleFilterAttribute( VisibleFilterIdentifier )]
		public bool Visible
		{
			get { return _Visible; }
			set { _Visible = value; }
		}

		#endregion Public serializable properties

		#region Public static methods

		/// <summary>
		/// Collects the speckle filter identifiers from the specified sheaf type.
		/// </summary>
		/// <param name="sheafType">The sheaf type from which to collect the speckle filter identifiers.</param>
		/// <returns>Returns the list of collected speckle filter identifiers.</returns>
		public static SpeckleFilterIdentifiers CollectSpeckleFilterIdentifiers( Type sheafType )
		{
			SpeckleFilterIdentifiers identifiers = new SpeckleFilterIdentifiers();

			// Process the properties of the class.

			PropertyInfo[] propInfos = sheafType.GetProperties( BindingFlags.Public | BindingFlags.NonPublic
				| BindingFlags.Instance );

			foreach( PropertyInfo propInfo in propInfos )
			{
				// Process only members which can be read.
				if( propInfo.CanRead )
				{
					// Get the attributes of the property.
					object[] attributes = propInfo.GetCustomAttributes( typeof( SpeckleFilterAttribute ), false );

					if( attributes != null )
					{
						foreach( object attribute in attributes )
						{
							if( attribute != null )
							{
								SpeckleFilterAttribute ssfAttr = (SpeckleFilterAttribute)attribute;

								identifiers.Add( ssfAttr.Identifier );
							}
						}
					}
				}
			}

			return identifiers;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Sets the <see cref="StorehouseLock"/> and <see cref="SheafUpdateVersion"/> properties with the specified values.
		/// RESERVED for internal use.
		/// </summary>
		/// <param name="storehouseLock">The storehouse lock.</param>
		/// <param name="sheafUpdateVersion">The update version.</param>
		public void SetLockAndUpdateVersion( SheafLocks storehouseLock, byte[] sheafUpdateVersion )
		{
			StorehouseLock = storehouseLock;
			SheafUpdateVersion = sheafUpdateVersion;
		}

		/// <summary>
		/// Sets the <see cref="StorehouseLock"/> property with the specified value. RESERVED for internal use.
		/// </summary>
		/// <param name="storehouseLock">The storehouse lock.</param>
		public void SetLock( SheafLocks storehouseLock )
		{
			StorehouseLock = storehouseLock;
		}

		/// <summary>
		/// Sets the <see cref="SheafUpdateVersion"/> property with the specified value. RESERVED for internal use.
		/// </summary>
		/// <param name="sheafUpdateVersion">The update version.</param>
		public void SetUpdateVersion( byte[] sheafUpdateVersion )
		{
			SheafUpdateVersion = sheafUpdateVersion;
		}

		/// <summary>
		/// Prepares the current sheaf to be updated, using the storehouse lock and sheaf update version from the specified
		/// sheaf.
		/// </summary>
		/// <param name="sheaf">Sheaf from which to get the storehouse lock and sheaf update version.</param>
		/// <remarks>
		/// <para>This is useful when an external sheaf (which has no storehouse lock and sheaf update version) updates an
		/// existing sheaf.</para>
		/// </remarks>
		public void PrepareToUpdate( ISheaf sheaf )
		{
			ValidationUtils.ObjectHasContent( sheaf, "sheaf" );

			StorehouseLock = sheaf.StorehouseLock;
			SheafUpdateVersion = sheaf.SheafUpdateVersion;
		}

		/// <summary>
		/// Prepares the current sheaf to be updated, using the storehouse lock and sheaf update version from the specified
		/// speckle.
		/// </summary>
		/// <param name="speckle">Speckle from which to get the storehouse lock and sheaf update version.</param>
		/// <remarks>
		/// <para>This is useful when an external sheaf (which has no storehouse lock and sheaf update version) updates an
		/// existing sheaf. Using this method, the existing sheaf doesn't have to be loaded, but only its associated speckle.
		/// </para>
		/// </remarks>
		public void PrepareToUpdate( Speckle speckle )
		{
			ValidationUtils.ObjectHasContent( speckle, "speckle" );

			StorehouseLock = speckle.StorehouseLock;
			SheafUpdateVersion = speckle.SheafUpdateVersion;
		}

		/// <summary>
		/// Collects the speckle filters from the current sheaf.
		/// </summary>
		/// <returns>Returns the list of collected speckle filters.</returns>
		public SpeckleFilters CollectSpeckleFilters()
		{
			SpeckleFilters filters = new SpeckleFilters();

			// Process the properties of the class.

			PropertyInfo[] propInfos = GetType().GetProperties( BindingFlags.Public | BindingFlags.NonPublic
				| BindingFlags.Instance );

			foreach( PropertyInfo propInfo in propInfos )
			{
				// Process only members which can be read.
				if( propInfo.CanRead )
				{
					// Get the attributes of the property.
					object[] attributes = propInfo.GetCustomAttributes( typeof( SpeckleFilterAttribute ), false );

					if( attributes != null )
					{
						foreach( object attribute in attributes )
						{
							if( attribute != null )
							{
								SpeckleFilterAttribute ssfAttr = (SpeckleFilterAttribute)attribute;

								// Get the property value.
								object filterValue = propInfo.GetValue( this, null );

								// Create a speckle filter.
								SpeckleFilter filter = new SpeckleFilter( ssfAttr.Identifier, filterValue );

								filters.Add( filter );
							}
						}
					}
				}
			}

			return filters;
		}

		#endregion Public methods
	}
}
