////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Collections.Generic;
using Vela.Kernel;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// List of sheafs (<see cref="ISheaf"/>).
	/// </summary>
	public class Sheafs : KernelSortedList<Locator, ISheaf>
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public Sheafs()
			: base( LocatorComparer.Instance )
		{
		}

		#endregion Constructors

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para>The <see cref="Sheaf.InstanceLocator"/> property is used as key.</para>
		/// </remarks>
		public override Locator GetKey( ISheaf item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.InstanceLocator;
		}

		/// <summary>
		/// Gets the current list as a dictionary where the items are the sheafs whose speckle filters contain a filter with
		/// the specified filter name.
		/// </summary>
		/// <param name="filterName">The filter name of the speckles of the sheafs which are to be added in the dictionary.
		/// The value of the filter must be a locator.</param>
		/// <returns>Returns the created dictionary.</returns>
		public Dictionary<Locator, Sheaf> GetDictionaryByFilterName( string filterName )
		{
			ValidationUtils.TextHasContent( filterName, "filterName" );

			Dictionary<Locator, Sheaf> output = new Dictionary<Locator, Sheaf>();

			foreach( Sheaf sheaf in this )
			{
				SpeckleFilters speckleFilters = sheaf.CollectSpeckleFilters();

				SpeckleFilter foundFilter = speckleFilters[ filterName ];
				if( foundFilter != null )
				{
					Locator foundLocator = foundFilter.Value as Locator;
					if( foundLocator == null )
					{
						throw new ValidationException( string.Format( "The value of the speckle filter with the filter name"
								+ " \"{0}\" isn't a locator.", filterName ), null, null
							, new ExceptionDetail( "Sheaf locator", sheaf.InstanceLocator ) );
					}

					output.Add( foundLocator, sheaf );
				}
			}

			return output;
		}

		/// <summary>
		/// Finds the sheaf whose speckle filter with the specified name has a value equal with the specified locator.
		/// </summary>
		/// <param name="filterName">The name of the speckle filter whose value to compare with the specified locator.</param>
		/// <param name="locatorToFind">Locator to find.</param>
		/// <returns>Returns the found sheaf, or null if none is found.</returns>
		public Sheaf FindByLocatorFilter( string filterName, Locator locatorToFind )
		{
			ValidationUtils.TextHasContent( filterName, "filterName" );
			ValidationUtils.ObjectHasContent( locatorToFind, "locatorToFind" );

			foreach( Sheaf sheaf in this )
			{
				SpeckleFilters speckleFilters = sheaf.CollectSpeckleFilters();

				SpeckleFilter foundFilter = speckleFilters[ filterName ];
				if( foundFilter != null )
				{
					Locator foundLocator = foundFilter.Value as Locator;
					if( foundLocator == null )
					{
						throw new ValidationException( string.Format( "The value of the speckle filter with the filter name"
								+ " \"{0}\" isn't a locator.", filterName ), null, null
							, new ExceptionDetail( "Sheaf locator", sheaf.InstanceLocator ) );
					}

					if( foundLocator.Equals( locatorToFind ) )
						return sheaf;
				}
			}

			return null;
		}

		#endregion Public methods
	}
}
