////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// A speckle is a data structure (like a row in a database) which provides quick access to the important properties of a
	/// sheaf. Its storehouse key is unique in a storehouse. A speckle refers a sheaf of type <see cref="ISheaf"/>.
	/// </summary>
	/// <remarks>
	/// <para>A speckle is stored unencrypted in the storehouse, so its security depends on the encryption of the storehouse.
	/// </para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Generic design</b></para>
	/// 
	/// <para>A speckle is generic. This way, the storehouse structure doesn't have to change with the structure of the speckles,
	/// which means that dramatically less management is required.</para>
	/// 
	/// <para>For quickly finding speckles, use speckle filters (<see cref="SpeckleFilter"/>).</para>
	/// 
	/// <para>Once you store a sheaf (and its speckle) in a storehouse, with a set of given filters, don't change the names
	/// and types of those filters. You may add new ones (even though the already stored sheafs don't have the new filter stored
	/// in the storehouse; you can find these sheafs by using a null filter value, and you can update those sheafs if you
	/// want).</para>
	/// </remarks>
	public class Speckle : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Speckle()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the referred sheaf. This must be unique in the storehouse.
		/// </param>
		/// <param name="sheafTypeName">The type name of the stored sheaf.</param>
		public Speckle( Locator storehouseKey, string sheafTypeName )
		{
			StorehouseKey = storehouseKey;
			SheafTypeName = sheafTypeName;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the referred sheaf. This must be unique in the storehouse.</param>
		/// <param name="sheafTypeName">The type name of the stored sheaf.</param>
		/// <param name="loadedSpeckleFilters">The filters of the speckle, loaded from the storehouse. This may be null or
		/// empty.</param>
		public Speckle( Locator storehouseKey, string sheafTypeName, SpeckleFilters loadedSpeckleFilters )
			: this( storehouseKey, sheafTypeName )
		{
			LoadedSpeckleFilters = loadedSpeckleFilters;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="sheaf">The sheaf to refer.</param>
		/// <remarks>
		/// <para>All the properties declared in this class are automatically set in here.</para>
		/// </remarks>
		public Speckle( ISheaf sheaf )
		{
			ValidationUtils.ObjectHasContent( sheaf, "sheaf" );

			StorehouseLock = sheaf.StorehouseLock;
			SheafUpdateVersion = sheaf.SheafUpdateVersion;

			StorehouseKey = sheaf.InstanceLocator;
			SheafTypeName = sheaf.GetType().FullName;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private SheafLocks _StorehouseLock = SheafLocks.Shared;
		/// <summary>
		/// The lock which was used when the speckle was loaded. RESERVED for internal use. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>A sheaf / speckle may be saved only if it was first locked for writing (with either the
		/// <see cref="SheafLocks.Update"/> or <see cref="SheafLocks.Exclusive"/> locks).</para>
		/// </remarks>
		public SheafLocks StorehouseLock
		{
			get { return _StorehouseLock; }
			protected set { _StorehouseLock = value; }
		}

		private byte[] _SheafUpdateVersion;
		/// <summary>
		/// The update version of the speckle. This may be null or empty. RESERVED for internal use. This is automatically set
		/// when the speckle is loaded, with the "SheafUpdateVersion" value loaded from the storehouse. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>A speckle must not be saved if this value is different than the "SheafUpdateVersion" value from the storehouse
		/// (because this is an indication that the speckle was saved by another thread / process).</para>
		/// </remarks>
		public byte[] SheafUpdateVersion
		{
			get { return _SheafUpdateVersion; }
			protected set { _SheafUpdateVersion = value; }
		}

		private SpeckleFilters _LoadedSpeckleFilters;
		/// <summary>
		/// The filters of the speckle, loaded from the storehouse. This may be null or empty.
		/// </summary>
		public SpeckleFilters LoadedSpeckleFilters
		{
			get { return _LoadedSpeckleFilters; }
			set { _LoadedSpeckleFilters = value; }
		}

		/// <summary>
		/// The creation time (in UTC format). This is <see cref="DateTime.MinValue"/> if <see cref="LoadedSpeckleFilters"/> is
		/// null or doesn't contain the <see cref="Sheaf.CreationTimeFilterIdentifier"/> speckle filter. READ-ONLY.
		/// </summary>
		public DateTime CreationTime
		{
			get { return TryGetLoadedSpeckleFilter<DateTime>( Sheaf.CreationTimeFilterIdentifier, DateTime.MinValue ); }
		}

		/// <summary>
		/// The update time (in UTC format). This is <see cref="DateTime.MinValue"/> if <see cref="LoadedSpeckleFilters"/> is
		/// null or doesn't contain the <see cref="Sheaf.CreationTimeFilterIdentifier"/> speckle filter. READ-ONLY.
		/// </summary>
		public DateTime UpdateTime
		{
			get { return TryGetLoadedSpeckleFilter<DateTime>( Sheaf.UpdateTimeFilterIdentifier, DateTime.MinValue ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Locator _StorehouseKey;
		/// <summary>
		/// The storehouse key of the referred sheaf. This must be unique in the storehouse. This is the instance locator of the
		/// referred sheaf. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Locator StorehouseKey
		{
			get { return _StorehouseKey; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "StorehouseKey" );

				_StorehouseKey = value;
			}
		}

		private string _SheafTypeName;
		/// <summary>
		/// The type name of the stored sheaf. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string SheafTypeName
		{
			get { return _SheafTypeName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "SheafTypeName" );

				_SheafTypeName = value;
			}
		}

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Sets the <see cref="StorehouseLock"/> and <see cref="SheafUpdateVersion"/> properties with the specified values.
		/// RESERVED for internal use.
		/// </summary>
		/// <param name="storehouseLock">The storehouse lock.</param>
		/// <param name="sheafUpdateVersion">The update version.</param>
		public void SetLockAndUpdateVersion( SheafLocks storehouseLock, byte[] sheafUpdateVersion )
		{
			ValidationUtils.ArrayHasContent( sheafUpdateVersion, "sheafUpdateVersion" );

			StorehouseLock = storehouseLock;
			SheafUpdateVersion = sheafUpdateVersion;
		}

		/// <summary>
		/// Sets the <see cref="StorehouseLock"/> property with the specified value. RESERVED for internal use.
		/// </summary>
		/// <param name="storehouseLock">The storehouse lock.</param>
		public void SetLock( SheafLocks storehouseLock )
		{
			StorehouseLock = storehouseLock;
		}

		/// <summary>
		/// Gets the value of the loaded speckle filter with the specified composite identifier.
		/// </summary>
		/// <typeparam name="TValue">The expected type of the value of the speckle filter to get.</typeparam>
		/// <param name="compositeIdentifier">The composite identifier of the speckle filter to get.</param>
		/// <param name="defaultValue">The value to return if the speckle filter (or its value) isn't found.</param>
		/// <returns>Returns the value of the speckle filter. This may be null or empty.</returns>
		public TValue TryGetLoadedSpeckleFilter<TValue>( string compositeIdentifier, TValue defaultValue )
		{
			ValidationUtils.TextHasContent( compositeIdentifier, "filterIdentifier" );

			if( LoadedSpeckleFilters == null )
				return defaultValue;

			SpeckleFilter filter = LoadedSpeckleFilters.GetByCompositeIdentifier( compositeIdentifier );
			if( (filter == null) || (filter.Value == null) )
				return defaultValue;

			return (TValue)filter.Value;
		}

		/// <summary>
		/// Gets the value of the loaded speckle filter with the specified composite identifier.
		/// </summary>
		/// <typeparam name="TValue">The expected type of the value of the speckle filter to get.</typeparam>
		/// <param name="compositeIdentifier">The composite identifier of the speckle filter to get.</param>
		/// <returns>Returns the value of the speckle filter. This may be null or empty.</returns>
		public TValue TryGetLoadedSpeckleFilter<TValue>( string compositeIdentifier )
			where TValue : class
		{
			return TryGetLoadedSpeckleFilter<TValue>( compositeIdentifier, null );
		}

		/// <summary>
		/// Gets the value of the loaded speckle filter with the specified composite identifier. An exception is thrown if the
		/// speckle filter isn't found, but its value may be null.
		/// </summary>
		/// <typeparam name="TValue">The expected type of the value of the speckle filter to get.</typeparam>
		/// <param name="compositeIdentifier">The composite identifier of the speckle filter to get.</param>
		/// <returns>Returns the value of the speckle filter. This may be null or empty if this is the value of the found speckle
		/// filter.</returns>
		public TValue GetLoadedSpeckleFilterOrNull<TValue>( string compositeIdentifier )
			where TValue : class
		{
			ValidationUtils.TextHasContent( compositeIdentifier, "compositeIdentifier" );

			if( LoadedSpeckleFilters == null )
			{
				throw new ValidationException( string.Format( "There are no loaded speckle filters from which to get the one"
					+ " with the \"{0}\" composite identifier.", compositeIdentifier ) );
			}

			SpeckleFilter filter = LoadedSpeckleFilters.GetByCompositeIdentifier( compositeIdentifier );
			if( filter == null )
			{
				throw new ValidationException( string.Format( "The loaded speckle filter with the \"{0}\" composite identifier"
					+ " is missing.", compositeIdentifier ) );
			}

			return filter.Value as TValue;
		}

		/// <summary>
		/// Gets the value of the loaded speckle filter with the specified composite identifier. An exception is thrown if the
		/// speckle filter (or its value) isn't found or its value is null.
		/// </summary>
		/// <typeparam name="TValue">The expected type of the value of the speckle filter to get.</typeparam>
		/// <param name="compositeIdentifier">The composite identifier of the speckle filter to get.</param>
		/// <returns>Returns the value of the speckle filter.</returns>
		public TValue GetLoadedSpeckleFilter<TValue>( string compositeIdentifier )
		{
			ValidationUtils.TextHasContent( compositeIdentifier, "compositeIdentifier" );

			if( LoadedSpeckleFilters == null )
			{
				throw new ValidationException( string.Format( "There are no loaded speckle filters from which to get the one"
					+ " with the \"{0}\" composite identifier.", compositeIdentifier ) );
			}

			SpeckleFilter filter = LoadedSpeckleFilters.GetByCompositeIdentifier( compositeIdentifier );
			if( (filter == null) || (filter.Value == null) )
			{
				throw new ValidationException( string.Format( "The loaded speckle filter with the \"{0}\" composite identifier"
					+ " is missing.", compositeIdentifier ) );
			}

			return (TValue)filter.Value;
		}

		#endregion Public methods
	}
}
