////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Speckle filter used to search speckles in the storehouse.
	/// </summary>
	/// <remarks>
	/// <para>A speckle filter may be used to store information about the associated speckle, and may be used to quickly
	/// find the speckles which meet certain criteria.</para>
	/// 
	/// <para>The value of a speckle filter is stored (in a storehouse) as a buffer, so (for example) comparisons of texts are
	/// case sensitive.</para>
	/// </remarks>
	public class SpeckleFilter
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected SpeckleFilter()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="identifier">The identifier.</param>
		public SpeckleFilter( SpeckleFilterIdentifier identifier )
		{
			Identifier = identifier;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="identifier">The identifier.</param>
		/// <param name="value">The value. This may be a scalar (see <see cref="ScalarEncoder"/> for details). This may be
		/// null.</param>
		public SpeckleFilter( SpeckleFilterIdentifier identifier, object value )
		{
			Identifier = identifier;
			Value = value;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="compositeIdentifier">The composite identifier.</param>
		/// <param name="value">The value. This may be a scalar (see <see cref="ScalarEncoder"/> for details). This may be
		/// null.</param>
		public SpeckleFilter( string compositeIdentifier, object value )
		{
			Identifier = new SpeckleFilterIdentifier( compositeIdentifier );
			Value = value;
		}

		#endregion Constructors

		#region Public serializable properties

		private SpeckleFilterIdentifier _Identifier;
		/// <summary>
		/// The identifier. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public SpeckleFilterIdentifier Identifier
		{
			get { return _Identifier; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Identifier" );

				_Identifier = value;
			}
		}

		private object _Value;
		/// <summary>
		/// The value. This may be a scalar (see <see cref="ScalarEncoder"/> for details). This may be null. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public object Value
		{
			get { return _Value; }
			protected set
			{
				// Check if the type of the value is valid.
				if( (value != null) && !ScalarEncoder.IsScalar( value ) )
				{
					throw new ValidationException( string.Format( "The type \"{0}\" is not allowed for speckle filters values."
						, value.GetType().FullName ) );
				}

				_Value = value;
			}
		}

		#endregion Public serializable properties
	}
}
