////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// May be applied on properties to indicate that they are used as speckle filters (<see cref="SpeckleFilter"/>).
	/// </summary>
	/// <remarks>
	/// <para>The properties on which this attribute is applied must have a getter.</para>
	/// 
	/// <para>You must ensure that each property from a class (including from its inheritors) has a unique filter index. If
	/// you don't, you'll get exceptions.</para>
	/// </remarks>
	[AttributeUsage( AttributeTargets.Property, AllowMultiple = false )]
	public class SpeckleFilterAttribute : Attribute
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected SpeckleFilterAttribute()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="identifier">The identifier.</param>
		public SpeckleFilterAttribute( SpeckleFilterIdentifier identifier )
		{
			Identifier = identifier;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="compositeIdentifier">The composite identifier.</param>
		public SpeckleFilterAttribute( string compositeIdentifier )
		{
			Identifier = new SpeckleFilterIdentifier( compositeIdentifier );
		}

		#endregion Constructors

		#region Public non-serializable properties

		private SpeckleFilterIdentifier _Identifier;
		/// <summary>
		/// The identifier. READ-ONLY.
		/// </summary>
		public SpeckleFilterIdentifier Identifier
		{
			get { return _Identifier; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Identifier" );

				_Identifier = value;
			}
		}

		#endregion Public non-serializable properties
	}
}
