////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Identifier for a speckle filter.
	/// </summary>
	/// <remarks>
	/// <para>A composite identifier has the form "{Name}#{Index}>{OverriddenName}".</para>
	/// 
	/// <para>Altough the filter name must be unique throughout the application / storehouse, this is not enforced.</para>
	/// 
	/// <para>You must ensure that each property (on which you apply the <see cref="SpeckleFilterAttribute"/> attribute) from a
	/// class (including from its inheritors) has a unique filter index (in its composite identifier). If you don't, you'll get
	/// exceptions.</para>
	/// </remarks>
	public class SpeckleFilterIdentifier
	{
		#region Public constants

		/// <summary>
		/// The separator between the name and the index in a composite identifier.
		/// </summary>
		public const char IndexSeparator = '#';

		/// <summary>
		/// The separator between the index and the overridden name in a composite identifier.
		/// </summary>
		public const char OverriddenNameSeparator = '>';

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected SpeckleFilterIdentifier()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="name">The name. This must be unique throughout the application / storehouse, but is not enforced.
		/// </param>
		/// <param name="index">The index. This must be unique throughout the hierarchy of derived classes. This must be at least
		/// 0.</param>
		public SpeckleFilterIdentifier( string name, int index )
		{
			Name = name;
			Index = index;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="compositeIdentifier">The composite identifier.</param>
		public SpeckleFilterIdentifier( string compositeIdentifier )
		{
			ValidationUtils.TextHasContent( compositeIdentifier, "compositeIdentifier" );

			string overriddenName = null;
			int overriddenNamePos = compositeIdentifier.IndexOf( OverriddenNameSeparator );
			if( (overriddenNamePos > 0) && (overriddenNamePos < compositeIdentifier.Length - 1) )
			{
				overriddenName = compositeIdentifier.Substring( overriddenNamePos + 1 );
			}
			else
			{
				overriddenNamePos = compositeIdentifier.Length;
			}

			int indexPos = compositeIdentifier.IndexOf( IndexSeparator );
			if( (indexPos <= 0) || (indexPos >= compositeIdentifier.Length - 1) )
			{
				throw new ValidationException( string.Format( "The \"{0}\" separator is missing from the composite identifier"
					+ " \"{1}\", or is at its beginning or end.", IndexSeparator, compositeIdentifier ) );
			}

			string textualIndex = compositeIdentifier.Substring( indexPos + 1, overriddenNamePos - (indexPos + 1) );
			ValidationUtils.Integer( textualIndex, 0, CoreDocumentsConsts.SpeckleFilterSizeMax - 1, "textualIndex" );

			Index = int.Parse( textualIndex );
			Name = compositeIdentifier.Substring( 0, indexPos );
			OverriddenName = overriddenName;
		}

		#endregion Constructors

		#region Public serializable properties

		private string _Name;
		/// <summary>
		/// The name. This must be unique throughout the application / storehouse, but is not enforced. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string Name
		{
			get { return _Name; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "Name" );

				_Name = value;
			}
		}

		private string _OverriddenName;
		/// <summary>
		/// The overridden name. This must be unique throughout the application / storehouse, but is not enforced. This may be
		/// null or empty. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is the name of a speckle filter which is overridden by the speckle filter with the <see cref="Name"/>
		/// name (and with the same index), filter which may contain a value with a different type.</para>
		/// </remarks>
		[SerializeAttribute]
		public string OverriddenName
		{
			get { return _OverriddenName; }
			protected set { _OverriddenName = value; }
		}

		private int _Index;
		/// <summary>
		/// The index. This must be unique throughout the hierarchy of derived classes. This must be at least 0. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public int Index
		{
			get { return _Index; }
			protected set
			{
				ValidationUtils.Integer( value, 0, CoreDocumentsConsts.SpeckleFilterSizeMax - 1, "Index" );

				_Index = value;
			}
		}

		#endregion Public serializable properties
	}
}
