////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Text;
using Vela.Kernel;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// List of speckle filters (<see cref="SpeckleFilter"/>).
	/// </summary>
	public class SpeckleFilters : KernelSortedList<string, SpeckleFilter>
	{
		#region Public constants

		/// <summary>
		/// Template for displaying the a filter to the user. Placeholders: 0 - filter name, 1 - filter index, 2 - value.
		/// </summary>
		public const string DisplayItemTemplate = "{0} ({1}) = \"{2}\"";

		/// <summary>
		/// Item separator.
		/// </summary>
		public const string ItemSeparator = ", ";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public SpeckleFilters()
			: base( StringInsensitiveComparer.Instance )
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="speckleFilters">Speckle filters to add to the list.</param>
		public SpeckleFilters( params SpeckleFilter[] speckleFilters )
			: base( StringInsensitiveComparer.Instance )
		{
			ValidationUtils.ArrayHasContent( speckleFilters, "speckleFilters" );

			foreach( SpeckleFilter speckleFilter in speckleFilters )
				Add( speckleFilter );
		}

		#endregion Constructors

		#region Public static methods

		/// <summary>
		/// Packs the specified items into a list.
		/// </summary>
		/// <param name="items">Items to pack into a list. This may be null or empty.</param>
		/// <returns>Returns the created list. This is null if "items" is null.</returns>
		public static SpeckleFilters ToList( params SpeckleFilter[] items )
		{
			if( items == null )
				return null;

			SpeckleFilters itemList = new SpeckleFilters();
			foreach( SpeckleFilter item in items )
				itemList.Add( item );

			return itemList;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Ensures that each filter from the list has a unique index.
		/// </summary>
		public void EnsureFilterIndexUniqueness()
		{
			Dictionary<int, SpeckleFilter> filtersByIndex = new Dictionary<int, SpeckleFilter>();

			foreach( KeyValuePair<string, SpeckleFilter> kvp in Items )
			{
				int index = kvp.Value.Identifier.Index;

				if( filtersByIndex.ContainsKey( index ) )
				{
					throw new ValidationException( string.Format( "The speckle filters with the names \"{0}\" and \"{1}\""
						+ " have the same index {2}."
						, kvp.Value.Identifier.Name, filtersByIndex[ index ].Identifier.Name, index ) );
				}

				filtersByIndex.Add( index, kvp.Value );
			}
		}

		/// <summary>
		/// Validates the specified item before adding or updating it to the list. Inheritors must throw an exception if the
		/// item must not be added to the list.
		/// </summary>
		/// <param name="item">Item to add to the list. There is no need to check if this is null because this is performed
		/// before the method is called.</param>
		/// <remarks>
		/// <para>Don't check for duplicate items here because this is automatically done.</para>
		/// </remarks>
		protected override void ValidateAddOrUpdate( SpeckleFilter item )
		{
			foreach( KeyValuePair<string, SpeckleFilter> kvp in Items )
			{
				if( (kvp.Value.Identifier.Index == item.Identifier.Index)
					&& !TextUtils.Equals( kvp.Value.Identifier.Name, item.Identifier.Name ) )
				{
					string message = string.Format( "Can't add a speckle filter with the name \"{0}\" and the index {1}"
							+ " because a duplicate index already exists for the filter name \"{2}\"."
						, item.Identifier.Name, item.Identifier.Index, kvp.Value.Identifier.Name );

					throw new ValidationException( message );
				}
			}
		}

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para>The name of the identifier of the item is used as key.</para>
		/// </remarks>
		public override string GetKey( SpeckleFilter item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.Identifier.Name;
		}

		/// <summary>
		/// Gets the item with the specified composite identifier in the list.
		/// </summary>
		/// <param name="compositeIdentifier">The composite identifier of the item to get.</param>
		/// <returns>Returns the item with the specified composite identifier. This is null if the composite identifier doesn't
		/// exist in the list.</returns>
		public SpeckleFilter GetByCompositeIdentifier( string compositeIdentifier )
		{
			SpeckleFilterIdentifier identifier = new SpeckleFilterIdentifier( compositeIdentifier );

			return this[ identifier.Name ];
		}

		/// <summary>
		/// Gets the item with the specified filter index in the list.
		/// </summary>
		/// <param name="filterIndex">The filter index of the item to get.</param>
		/// <returns>Returns the item with the specified composite identifier. This is null if the composite identifier doesn't
		/// exist in the list.</returns>
		public SpeckleFilter GetByIndex( int filterIndex )
		{
			ValidationUtils.Integer( filterIndex, 0, CoreDocumentsConsts.SpeckleFilterSizeMax - 1, "filterIndex" );

			foreach( KeyValuePair<string, SpeckleFilter> kvp in Items )
			{
				if( kvp.Value.Identifier.Index == filterIndex )
					return kvp.Value;
			}

			return null;
		}

		/// <summary>
		/// Gets the textual representation of all the items from the list.
		/// </summary>
		/// <returns>Returns the textual representation of all the items from the list.</returns>
		public override string ToString()
		{
			StringBuilder sb = new StringBuilder();

			int i = 0;
			foreach( SpeckleFilter speckleFilter in this )
			{
				string displayText;
				object filterValue = speckleFilter.Value;

				if( filterValue is byte[] )
					displayText = TextUtils.ArrayToMoteText16( (byte[])filterValue );
				else if( filterValue is DateTime )
					displayText = ((DateTime)filterValue).ToString( KernelConsts.TimeFormat );
				else if( filterValue != null )
					displayText = filterValue.ToString();
				else
					displayText = null;

				sb.AppendFormat( DisplayItemTemplate, speckleFilter.Identifier.Name, speckleFilter.Identifier.Index
					, displayText );
				if( i < Count - 1 )
					sb.Append( ItemSeparator );

				i++;
			}

			return sb.ToString();
		}

		#endregion Public methods
	}
}
