////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Collections.Generic;
using Vela.Kernel;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// List of speckles (<see cref="Speckle"/>).
	/// </summary>
	/// <remarks>
	/// <para>The items from this list should have unique speckle keys (<see cref="Speckle.StorehouseKey"/>).</para>
	/// 
	/// <para>This list must be unsorted (but ordered) because the collection of speckles from the storehouse doesn't support
	/// sorting by custom criteria (but only by the underlying database ID), and the speckles might be paged. Therefore, if
	/// this list were sorted, there would appear inconsistencies in the order of items in a page of speckles.</para>
	/// </remarks>
	public class Speckles : KernelSortedList<Locator, Speckle>
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public Speckles()
			: base( LocatorComparer.Instance )
		{
		}

		#endregion Constructors

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para>The <see cref="Speckle.StorehouseKey"/> property is used as key.</para>
		/// </remarks>
		public override Locator GetKey( Speckle item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.StorehouseKey;
		}

		/// <summary>
		/// Gets the current list as a dictionary where the items are the loaded speckle filters
		/// (<see cref="Speckle.LoadedSpeckleFilters"/>) with the specified filter name.
		/// </summary>
		/// <param name="filterName">The filter name of the speckles which are to be added in the dictionary. The value of the
		/// filter must be a locator.</param>
		/// <returns>Returns the created dictionary.</returns>
		public Dictionary<Locator, Speckle> GetDictionaryByFilterName( string filterName )
		{
			ValidationUtils.TextHasContent( filterName, "filterName" );

			Dictionary<Locator, Speckle> output = new Dictionary<Locator, Speckle>();

			foreach( Speckle speckle in this )
			{
				if( speckle.LoadedSpeckleFilters == null )
					continue;

				SpeckleFilter foundFilter = speckle.LoadedSpeckleFilters[ filterName ];
				if( foundFilter != null )
				{
					Locator foundLocator = foundFilter.Value as Locator;
					if( foundLocator == null )
					{
						throw new ValidationException( string.Format( "The value of the speckle filter with the filter name"
								+ " \"{0}\" isn't a locator.", filterName ), null, null
							, new ExceptionDetail( "Speckle storehouse key", speckle.StorehouseKey ) );
					}

					output.Add( foundLocator, speckle );
				}
			}

			return output;
		}

		/// <summary>
		/// Gets the current list as a dictionary where the items are the loaded speckle filters with the specified composite
		/// identifier.
		/// </summary>
		/// <param name="compositeIdentifier">The composite identifier of the loaded speckle filters which are to be added in the
		/// dictionary. The value of the loaded speckle filter must be a locator.</param>
		/// <returns>Returns the created dictionary.</returns>
		public Dictionary<Locator, Speckle> GetDictionaryByFilterIdentifier( string compositeIdentifier )
		{
			SpeckleFilterIdentifier identifier = new SpeckleFilterIdentifier( compositeIdentifier );

			return GetDictionaryByFilterName( identifier.Name );
		}

		/// <summary>
		/// Finds the speckle whose loaded speckle filter with the filter name equal with the filter name has a value equal with
		/// the specified locator.
		/// </summary>
		/// <param name="filterName">The name of the speckle filter whose value to compare with the specified locator.</param>
		/// <param name="locatorToFind">Locator to find.</param>
		/// <returns>Returns the found speckle, or null if none is found.</returns>
		public Speckle FindByLocatorFilter( string filterName, Locator locatorToFind )
		{
			ValidationUtils.TextHasContent( filterName, "filterName" );
			ValidationUtils.ObjectHasContent( locatorToFind, "locatorToFind" );

			foreach( Speckle speckle in this )
			{
				if( speckle.LoadedSpeckleFilters == null )
					continue;

				SpeckleFilter foundFilter = speckle.LoadedSpeckleFilters[ filterName ];
				if( foundFilter != null )
				{
					Locator foundLocator = foundFilter.Value as Locator;
					if( foundLocator == null )
					{
						throw new ValidationException( string.Format( "The value of the speckle filter with the filter name"
								+ " \"{0}\" isn't a locator.", filterName ), null, null
							, new ExceptionDetail( "Speckle storehouse key", speckle.StorehouseKey ) );
					}

					if( foundLocator.Equals( locatorToFind ) )
						return speckle;
				}
			}

			return null;
		}

		#endregion Public methods
	}
}
