////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Contract for the configuration of a storehouse.
	/// </summary>
	/// <remarks>
	/// <para>A storehouse configuration is a normal sheaf, stored in encrypted form.</para>
	/// 
	/// <para>The main reason for the existence of this sheaf is to have a way to store storehouse-wide information, but
	/// without having to load it often.</para>
	/// </remarks>
	public abstract class StorehouseConfiguration : Sheaf
	{
		#region Public constants

		/// <summary>
		/// The text from which the instance locator of a storehouse configuration is created.
		/// </summary>
		public const string PredefinedInstanceLocatorText = "Storehouse configuration";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected StorehouseConfiguration()
			: base( PredefinedInstanceLocator )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		private static Locator _PredefinedInstanceLocator;
		/// <summary>
		/// The instance locator of a storehouse configuration. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The value is the encoded <see cref="PredefinedInstanceLocatorText"/> text.</para>
		/// 
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public static Locator PredefinedInstanceLocator
		{
			get
			{
				if( _PredefinedInstanceLocator == null )
				{
					byte[] buffer = KernelUtils.GetDeepTextEncoding().TextToArray( PredefinedInstanceLocatorText, false );

					_PredefinedInstanceLocator = new Locator( buffer );
				}

				return _PredefinedInstanceLocator;
			}
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private DateTime _LastBackupTime;
		/// <summary>
		/// The time when the last backup was made.
		/// </summary>
		[SerializeAttribute]
		public DateTime LastBackupTime
		{
			get { return _LastBackupTime; }
			set { _LastBackupTime = value; }
		}

		#endregion Public serializable properties
	}
}
