////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto.Hash;
using Vela.Kernel;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Contract for a "Peer name Resolution, Encryption and Authentication" UPN algorithm.
	/// </summary>
	public abstract class Prea : UpnAlg
	{
		#region Public constants

		/// <summary>
		/// The minimum size (in bytes) that the generated body must have. This must be at least 1.
		/// </summary>
		public const int OutputBodySizeMin = 16;

		/// <summary>
		/// The buffer size for the size of a value.
		/// </summary>
		public const int ValueSizeBufferSize = 1;

		/// <summary>
		/// The seed used for the generation of the fingerprint.
		/// </summary>
		public const string FingerprintSeedDef = "Just as a cell doesn't define us, so we aren't just a thought or an instinct.";

		/// <summary>
		/// The seed used for the generation of the long and short host labels.
		/// </summary>
		public const string HostLabelSeedDef = "Upon a frozen see of stars, a ship is sailing in eternity.";

		/// <summary>
		/// The seed used for the generation of the classifier.
		/// </summary>
		public const string ClassifierSeedDef = "I am not a part of the primordial soup where all life forms are one entity.";

		/// <summary>
		/// The seed used for the generation of the symmetric encryption key.
		/// </summary>
		public const string SymmEncryptionKeySeedDef = "Aim for the stars. Even if you fail, you'll be in good company.";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="textEncoding">Text encoding.</param>
		/// <param name="hashAlg">Hash algorithm.</param>
		/// <param name="steganographicAlg">The steganographic algorithm.</param>
		/// <param name="communicationFrameAlg">The communication frame algorithm which a client must use to send action
		/// requests.</param>
		/// <param name="outputBodySize">The size (in bytes) that the generated body must have. If this is smaller than
		/// <see cref="OutputBodySizeMin"/>, all the available bytes are used.</param>
		/// <param name="fingerprintSeed">The seed used for the generation of the fingerprint.</param>
		/// <param name="hostLabelSeed">The seed used for the generation of the long and short host labels.</param>
		/// <param name="classifierSeed">The seed used for the generation of the classifier.</param>
		/// <param name="symmEncryptionKeySeed">The seed used for the generation of the symmetric encryption key.</param>
		protected Prea( TextEncoding textEncoding, HashAlg hashAlg, SteganographicAlg steganographicAlg
				, ICommunicationFrameAlg communicationFrameAlg, int outputBodySize, string fingerprintSeed, string hostLabelSeed
				, string classifierSeed, string symmEncryptionKeySeed )
			: base( textEncoding, hashAlg, steganographicAlg, communicationFrameAlg, outputBodySize, fingerprintSeed
				, hostLabelSeed, classifierSeed, symmEncryptionKeySeed )
		{
		}

		#endregion Constructors

		#region Public static methods

		/// <summary>
		/// Checks if the current algorithm may be used to parse the specified data into a UPN.
		/// </summary>
		/// <param name="buffer">Buffer to parse.</param>
		/// <param name="parsingAlgType">Type whose name must be found in the data to parse. This type will be instantiated as
		/// parsing algorithm.</param>
		/// <returns>Returns an instance of itself if it may parse the specified data, or null if it may not parse it.</returns>
		public static UpnAlg GetParsingAlgorithm( byte[] buffer, Type parsingAlgType )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );
			ValidationUtils.ObjectHasContent( parsingAlgType, "parsingAlgType" );

			BinaryNodeStream bns = new BinaryNodeStream( ValueSizeBufferSize, buffer );

			if( !bns.CanReadFirstNode( KernelConsts.UpnSizeMax ) )
				return null;

			// UPN algorithm name.
			string upnAlgName = (string)bns.ReadValue( typeof( string ) );

			if( TextUtils.Equals( RegistrableAlgorithmUtils.GetAlgorithmName( parsingAlgType ), upnAlgName ) )
				return (UpnAlg)RegistrableAlgorithms.InstantiateGlobal( upnAlgName, parsingAlgType );

			return null;
		}

		/// <summary>
		/// Checks if the current algorithm may be used to parse the specified data into a UPN.
		/// </summary>
		/// <param name="data">Data to parse.</param>
		/// <param name="parsingAlgType">Type whose name must be found in the data to parse. This type will be instantiated as
		/// parsing algorithm.</param>
		/// <returns>Returns an instance of itself if it may parse the specified data, or null if it may not parse it.</returns>
		public static UpnAlg GetParsingAlgorithm( string data, Type parsingAlgType )
		{
			data = TextUtils.DeleteWhitespaces( data );

			int dataSize = data.Length;
			if( (dataSize < 1) || (dataSize > KernelConsts.UpnSizeMax) )
				return null;

			byte[] buffer = BufferUtils.Flip( TextUtils.Text32ToArray( data ) );

			return GetParsingAlgorithm( buffer, parsingAlgType );
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Parses the specified binary UPN.
		/// </summary>
		/// <param name="upnData">Binary form of the UPN to parse.</param>
		/// <returns>Returns the resulted UPN.</returns>
		public override Upn Parse( byte[] upnData )
		{
			ValidationUtils.ArrayHasContent( upnData, "upnData" );

			BinaryNodeStream bns = new BinaryNodeStream( ValueSizeBufferSize, upnData );

			// UPN algorithm name.
			string upnAlgName = (string)bns.ReadValue( typeof( string ) );
			UpnAlg upnAlg = RegistrableAlgorithms.InstantiateGlobal( upnAlgName, GetType() ) as UpnAlg;

			// Port.
			UInt16 port = (UInt16)bns.ReadValue( typeof( UInt16 ) );

			int outputBodySize = (OutputBodySize >= OutputBodySizeMin) ? OutputBodySize : HashAlg.AlgSizeBytes;

			// Body.
			byte[] body = (byte[])bns.ReadValue( typeof( byte[] ) );
			int bodySize = (body != null) ? body.Length : 0;
			if( bodySize != outputBodySize )
			{
				throw new ValidationException( string.Format( "The UPN algorithm \"{0}\" requires a body size {1} not {2}."
					, RegistrableAlgorithmUtils.GetAlgorithmName( upnAlg ), outputBodySize, bodySize ) );
			}

			// Checksum.
			int checksumPos = bns.Position;
			UInt16 checksum = (UInt16)bns.ReadValue( typeof( UInt16 ) );

			#region Verify the checksum

			byte[] bufferForChecksum = bns.ToArray();
			UInt16 actualChecksum = BufferUtils.ComputeChecksum16( bufferForChecksum, checksumPos );

			if( checksum != actualChecksum )
				throw new ValidationException( "The UPN checksum is invalid." );

			#endregion Verify the checksum

			return new Upn( upnAlg, port, body );
		}

		/// <summary>
		/// Parses the specified textual UPN.
		/// </summary>
		/// <param name="upnData">Textual form of the UPN to parse.</param>
		/// <returns>Returns the resulted UPN.</returns>
		public override Upn Parse( string upnData )
		{
			upnData = TextUtils.DeleteWhitespaces( upnData );

			ValidationUtils.Text( upnData, -1, -1, TextValidations.Upn, "upnData" );

			byte[] buffer = BufferUtils.Flip( TextUtils.Text32ToArray( upnData ) );

			return Parse( buffer );
		}

		/// <summary>
		/// Generates the UPN body from the specified data.
		/// </summary>
		/// <param name="upnBodySource">Source for the UPN body.</param>
		/// <returns>Returns the generated body.</returns>
		public override byte[] GenerateBody( IUpnBodySource upnBodySource )
		{
			ValidationUtils.ObjectHasContent( upnBodySource, "upnBodySource" );

			byte[] bodySource = HashAlg.Compute( TextEncoding, true, FingerprintSeed, HostLabelSeed, ClassifierSeed
				, SymmEncryptionKeySeed, upnBodySource.GetUpnBodySource() );

			if( (OutputBodySize >= OutputBodySizeMin) && (OutputBodySize < bodySource.Length) )
			{
				int size = Math.Min( bodySource.Length, OutputBodySize );
				byte[] body = new byte[ size ];
				Buffer.BlockCopy( bodySource, 0, body, 0, size );

				return body;
			}
			else
				return bodySource;
		}

		/// <summary>
		/// Generates the binary form of the specified UPN.
		/// </summary>
		/// <param name="upn">UPN from which to generate the binary form.</param>
		/// <returns>Returns the generated binary UPN.</returns>
		public override byte[] GenerateBinaryUpn( Upn upn )
		{
			ValidationUtils.ObjectHasContent( upn, "upn" );

			BinaryNodeStream bns = new BinaryNodeStream( ValueSizeBufferSize, TextEncoding );

			bns.WriteValue( RegistrableAlgorithmUtils.GetAlgorithmName( upn.UpnAlg ) );
			bns.WriteValue( upn.Port );
			bns.WriteValue( upn.Body );

			// Write the checksum.
			byte[] buffer = bns.ToArray();
			UInt16 checksum = BufferUtils.ComputeChecksum16( buffer );
			bns.WriteValue( checksum );

			return bns.ToArray();
		}

		/// <summary>
		/// Generates the textual form of the specified UPN.
		/// </summary>
		/// <param name="upn">UPN from which to generate the textual form.</param>
		/// <returns>Returns the generated textual UPN.</returns>
		public override string GenerateTextualUpn( Upn upn )
		{
			// The binary UPN is flipped because otherwise the textual UPN would start with the same few characters, fact which
			// would make the user think that any two UPNs are the same.
			byte[] buffer = BufferUtils.Flip( GenerateBinaryUpn( upn ) );

			return TextUtils.ArrayToMoteText32( buffer );
		}

		/// <summary>
		/// Generates the UPN fingerprint from the specified UPN.
		/// </summary>
		/// <param name="upn">UPN from which to generate the fingerprint.</param>
		/// <returns>Returns the generated fingerprint.</returns>
		public override Locator GenerateFingerprint( Upn upn )
		{
			byte[] upnBuffer = GenerateBinaryUpn( upn );
			byte[] hash = HashAlg.Compute( TextEncoding, true, FingerprintSeed, upnBuffer );

			return new Locator( hash );
		}

		/// <summary>
		/// Generates the UPN host label from the specified UPN.
		/// </summary>
		/// <param name="upn">UPN from which to generate the host label.</param>
		/// <param name="generateLongHostLabel">If this is true then the long host label will be generated, else the short one.
		/// </param>
		/// <returns>Returns the generated host label.</returns>
		public override string GenerateHostLabel( Upn upn, bool generateLongHostLabel )
		{
			byte[] upnBuffer = GenerateBinaryUpn( upn );
			byte[] hash = HashAlg.Compute( TextEncoding, true, HostLabelSeed, upnBuffer );

			// Since the host label is used in a host name, it may not use motes. Base 16 provides the maximum stealth.
			string upnHostLabel = TextUtils.ArrayToText16( hash );

			int hostLabelSizeMax = generateLongHostLabel ? KernelConsts.LongHostLabelSizeMax
				: KernelConsts.ShortHostLabelSizeMax;

			return TextUtils.GetBeginning( upnHostLabel, hostLabelSizeMax, false, 1, null );
		}

		/// <summary>
		/// Generates the UPN classifier from the specified UPN.
		/// </summary>
		/// <param name="upn">UPN from which to generate the classifier.</param>
		/// <returns>Returns the generated classifier.</returns>
		public override string GenerateClassifier( Upn upn )
		{
			byte[] upnBuffer = GenerateBinaryUpn( upn );
			byte[] hash = HashAlg.Compute( TextEncoding, true, ClassifierSeed, upnBuffer );

			// Since the classifier is used by the underlying PNRP, it should not use motes. Base 16 provides the maximum
			// stealth; see the comment of the "CoreDocumentsConsts.UpnClassifierSizeMax" constant to understand why.
			string classifierText = TextUtils.ArrayToText16( hash );

			return TextUtils.GetBeginning( classifierText, CoreDocumentsConsts.UpnClassifierSizeMax );
		}

		/// <summary>
		/// Generates the UPN symmetric encryption key from the specified UPN.
		/// </summary>
		/// <param name="upn">UPN from which to generate the symmetric encryption key.</param>
		/// <returns>Returns the generated symmetric encryption key.</returns>
		/// <remarks>
		/// <para>The size of the resulted key depends on the used hash algorithm. Therefore, when you have to use this key to
		/// encrypt data, use the <see cref="Vela.Crypto.Symmetric.SymmAlg.AlignToAlgSize"/> method to align the key to the
		/// required size for the used encryption algorithm.</para>
		/// </remarks>
		public override byte[] GenerateSymmEncryptionKey( Upn upn )
		{
			byte[] upnBuffer = GenerateBinaryUpn( upn );
			byte[] hash = HashAlg.Compute( TextEncoding, true, SymmEncryptionKeySeed, upnBuffer );

			return hash;
		}

		#endregion Public methods
	}
}
