////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto.Hash;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// Contract for a UPN algorithm.
	/// </summary>
	/// <remarks>
	/// <para>It must be impossible to generate the UPN from the UPN fingerprint. This is because the UPN fingerprint is used to
	/// identify the recipient of an encrypted envelope, while the UPN may be shared only among friends.</para>
	/// </remarks>
	public abstract class UpnAlg : IParsingAlgorithm, IExpirableAlgorithm
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="textEncoding">Text encoding.</param>
		/// <param name="hashAlg">Hash algorithm.</param>
		/// <param name="steganographicAlg">The steganographic algorithm.</param>
		/// <param name="communicationFrameAlg">The communication frame algorithm which a client must use to send action
		/// requests.</param>
		/// <param name="outputBodySize">The size (in bytes) that the generated body must have. If this is smaller than the
		/// minimum output body size, all the available bytes are used.</param>
		/// <param name="fingerprintSeed">The seed used for the generation of the fingerprint.</param>
		/// <param name="hostLabelSeed">The seed used for the generation of the long and short host labels.</param>
		/// <param name="classifierSeed">The seed used for the generation of the classifier.</param>
		/// <param name="symmEncryptionKeySeed">The seed used for the generation of the symmetric encryption key.</param>
		protected UpnAlg( TextEncoding textEncoding, HashAlg hashAlg, SteganographicAlg steganographicAlg
			, ICommunicationFrameAlg communicationFrameAlg, int outputBodySize, string fingerprintSeed, string hostLabelSeed
			, string classifierSeed, string symmEncryptionKeySeed )
		{
			ValidationUtils.TextHasContent( fingerprintSeed, "fingerprintSeed" );
			ValidationUtils.TextHasContent( hostLabelSeed, "hostLabelSeed" );
			ValidationUtils.TextHasContent( classifierSeed, "classifierSeed" );
			ValidationUtils.TextHasContent( symmEncryptionKeySeed, "symmEncryptionKeySeed" );

			OutputBodySize = outputBodySize;
			FingerprintSeed = fingerprintSeed;
			HostLabelSeed = hostLabelSeed;
			ClassifierSeed = classifierSeed;
			SymmEncryptionKeySeed = symmEncryptionKeySeed;

			TextEncoding = textEncoding;
			HashAlg = hashAlg;
			SteganographicAlg = steganographicAlg;
			CommunicationFrameAlg = communicationFrameAlg;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The size (in bytes) that the generated body must have. If this is smaller than the minimum output body size, all the
		/// available bytes are used.
		/// </summary>
		protected int OutputBodySize;

		/// <summary>
		/// The seed used for the generation of the fingerprint.
		/// </summary>
		protected string FingerprintSeed;

		/// <summary>
		/// The seed used for the generation of the long and short host labels.
		/// </summary>
		protected string HostLabelSeed;

		/// <summary>
		/// The seed used for the generation of the classifier.
		/// </summary>
		protected string ClassifierSeed;

		/// <summary>
		/// The seed used for the generation of the symmetric encryption key.
		/// </summary>
		protected string SymmEncryptionKeySeed;

		#endregion Private fields

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public virtual DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( HashAlg.ExpirationTime, SteganographicAlg.ExpirationTime ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private TextEncoding _TextEncoding;
		/// <summary>
		/// Text encoding. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public TextEncoding TextEncoding
		{
			get { return _TextEncoding; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "TextEncoding" );

				_TextEncoding = value;
			}
		}

		private HashAlg _HashAlg;
		/// <summary>
		/// Hash algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public HashAlg HashAlg
		{
			get { return _HashAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "HashAlg" );

				_HashAlg = value;
			}
		}

		private SteganographicAlg _SteganographicAlg;
		/// <summary>
		/// The steganographic algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public SteganographicAlg SteganographicAlg
		{
			get { return _SteganographicAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "SteganographicAlg" );

				_SteganographicAlg = value;
			}
		}

		private ICommunicationFrameAlg _CommunicationFrameAlg;
		/// <summary>
		/// The communication frame algorithm which a client must use to send action requests. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public ICommunicationFrameAlg CommunicationFrameAlg
		{
			get { return _CommunicationFrameAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "CommunicationFrameAlg" );

				_CommunicationFrameAlg = value;
			}
		}

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Parses the specified binary UPN.
		/// </summary>
		/// <param name="upnData">Binary form of the UPN to parse.</param>
		/// <returns>Returns the resulted UPN.</returns>
		public abstract Upn Parse( byte[] upnData );

		/// <summary>
		/// Parses the specified textual UPN.
		/// </summary>
		/// <param name="upnData">Textual form of the UPN to parse.</param>
		/// <returns>Returns the resulted UPN.</returns>
		public abstract Upn Parse( string upnData );

		/// <summary>
		/// Generates the UPN body from the specified data.
		/// </summary>
		/// <param name="upnBodySource">Source for the UPN body.</param>
		/// <returns>Returns the generated body.</returns>
		public abstract byte[] GenerateBody( IUpnBodySource upnBodySource );

		/// <summary>
		/// Generates the binary form of the specified UPN.
		/// </summary>
		/// <param name="upn">UPN from which to generate the binary form.</param>
		/// <returns>Returns the generated binary UPN.</returns>
		public abstract byte[] GenerateBinaryUpn( Upn upn );

		/// <summary>
		/// Generates the textual form of the specified UPN.
		/// </summary>
		/// <param name="upn">UPN from which to generate the textual form.</param>
		/// <returns>Returns the generated textual UPN.</returns>
		public abstract string GenerateTextualUpn( Upn upn );

		/// <summary>
		/// Generates the UPN fingerprint from the specified UPN.
		/// </summary>
		/// <param name="upn">UPN from which to generate the fingerprint.</param>
		/// <returns>Returns the generated fingerprint.</returns>
		public abstract Locator GenerateFingerprint( Upn upn );

		/// <summary>
		/// Generates the UPN host label from the specified UPN.
		/// </summary>
		/// <param name="upn">UPN from which to generate the host label.</param>
		/// <param name="generateLongHostLabel">If this is true then the long host label will be generated, else the short one.
		/// </param>
		/// <returns>Returns the generated host label.</returns>
		public abstract string GenerateHostLabel( Upn upn, bool generateLongHostLabel );

		/// <summary>
		/// Generates the UPN classifier from the specified UPN.
		/// </summary>
		/// <param name="upn">UPN from which to generate the classifier.</param>
		/// <returns>Returns the generated classifier.</returns>
		public abstract string GenerateClassifier( Upn upn );

		/// <summary>
		/// Generates the UPN symmetric encryption key from the specified UPN.
		/// </summary>
		/// <param name="upn">UPN from which to generate the symmetric encryption key.</param>
		/// <returns>Returns the generated symmetric encryption key.</returns>
		public abstract byte[] GenerateSymmEncryptionKey( Upn upn );

		#endregion Public methods
	}
}
