////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.CoreDocuments
{
	/// <summary>
	/// A Universal Peer Name (UPN).
	/// </summary>
	/// <remarks>
	/// <para>Two people can communicate securely if they exchange their UPNs.</para>
	/// 
	/// <para>A UPN provides name resolution, encryption and authentication between peers. The encryption is done with the key
	/// generated with the <see cref="GenerateSymmEncryptionKey"/> method, on a <see cref="UpnEncryptedEnvelope"/> encrypted
	/// envelope.</para>
	/// 
	/// <para>The following UPN properties are generated from the UPN: fingerprint, long and short host labels, classifier, and
	/// symmetric encryption key. It's practically impossible to find out any of a UPN's properties from any other of them.
	/// </para>
	/// 
	/// <para>The UPN fingerprint is used as peer identifier, and is exposed by (exported) documents.</para>
	/// 
	/// <para>The UPN long and short host labels may be used in the alternative peer address, to form a host name.</para>
	/// 
	/// <para>The UPN classifier is used in the underlying PNRP.</para>
	/// </remarks>
	public class Upn : IExpirableAlgorithm
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Upn()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="upnAlg">The UPN algorithm.</param>
		/// <param name="port">The network port where action requests must be sent by a client.</param>
		/// <param name="body">The body.</param>
		public Upn( UpnAlg upnAlg, UInt16 port, byte[] body )
		{
			UpnAlg = upnAlg;
			Port = port;
			Body = body;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private Locator _Fingerprint;
		/// <summary>
		/// The fingerprint used to identify the peer. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public Locator Fingerprint
		{
			get
			{
				if( _Fingerprint == null )
					_Fingerprint = GenerateFingerprint();

				return _Fingerprint;
			}
		}

		private string _LongHostLabel;
		/// <summary>
		/// The long host label used in a host name (which can be used in an alternative peer address). READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public string LongHostLabel
		{
			get
			{
				if( _LongHostLabel == null )
					_LongHostLabel = GenerateHostLabel( true );

				return _LongHostLabel;
			}
		}

		private string _ShortHostLabel;
		/// <summary>
		/// The short host label used in a host name (which can be used in an alternative peer address). READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public string ShortHostLabel
		{
			get
			{
				if( _ShortHostLabel == null )
					_ShortHostLabel = GenerateHostLabel( false );

				return _ShortHostLabel;
			}
		}

		private string _Classifier;
		/// <summary>
		/// The classifier used in the underlying PRNP. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public string Classifier
		{
			get
			{
				if( _Classifier == null )
					_Classifier = GenerateClassifier();

				return _Classifier;
			}
		}

		private byte[] _SymmEncryptionKey;
		/// <summary>
		/// The key used for symmetric encryption. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public byte[] SymmEncryptionKey
		{
			get
			{
				if( _SymmEncryptionKey == null )
					_SymmEncryptionKey = GenerateSymmEncryptionKey();

				return _SymmEncryptionKey;
			}
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public virtual DateTime ExpirationTime
		{
			get { return UpnAlg.ExpirationTime; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private UpnAlg _UpnAlg;
		/// <summary>
		/// The UPN algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public UpnAlg UpnAlg
		{
			get { return _UpnAlg; }
			protected set
			{
				_UpnAlg = value;

				_Fingerprint = null;
				_LongHostLabel = null;
				_ShortHostLabel = null;
				_Classifier = null;
				_SymmEncryptionKey = null;
			}
		}

		private UInt16 _Port;
		/// <summary>
		/// The network port where action requests must be sent by a client. This must be at least 1. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public UInt16 Port
		{
			get { return _Port; }
			protected set
			{
				ValidationUtils.Integer( value, 1, UInt16.MaxValue, "Port" );

				_Port = value;

				_Fingerprint = null;
				_LongHostLabel = null;
				_ShortHostLabel = null;
				_Classifier = null;
				_SymmEncryptionKey = null;
			}
		}

		private byte[] _Body;
		/// <summary>
		/// The body. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] Body
		{
			get { return _Body; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "Body" );

				_Body = value;

				_Fingerprint = null;
				_LongHostLabel = null;
				_ShortHostLabel = null;
				_Classifier = null;
				_SymmEncryptionKey = null;
			}
		}

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Returns the textual form of the current UPN.
		/// </summary>
		/// <returns>Returns the textual form of the UPN.</returns>
		public override string ToString()
		{
			return UpnAlg.GenerateTextualUpn( this );
		}

		/// <summary>
		/// Verifies if the current UPN is the same as the specified one.
		/// </summary>
		/// <param name="upn">UPN to verify against the current UPN.</param>
		/// <returns>Returns true if the UPNs have the same fingerprints.</returns>
		public bool Equals( Upn upn )
		{
			ValidationUtils.ObjectHasContent( upn, "upn" );

			return Fingerprint.Equals( upn.Fingerprint );
		}

		/// <summary>
		/// Verifies if the current UPN is the same as one with the specified fingerprint.
		/// </summary>
		/// <param name="upnFingerprint">UPN fingerprint to verify against the fingerprint of the current UPN.</param>
		/// <returns>Returns true if the UPNs have the same fingerprints.</returns>
		public bool Equals( Locator upnFingerprint )
		{
			ValidationUtils.ObjectHasContent( upnFingerprint, "upnFingerprint" );

			return Fingerprint.Equals( upnFingerprint );
		}

		/// <summary>
		/// Generates the fingerprint used to identify the peer.
		/// </summary>
		/// <returns>Returns the generated fingerprint.</returns>
		public Locator GenerateFingerprint()
		{
			return UpnAlg.GenerateFingerprint( this );
		}

		/// <summary>
		/// Generates the host label used in a host name (which can be used as alternative peer address).
		/// </summary>
		/// <param name="generateLongHostLabel">If this is true then the long host label will be generated, else the short one.
		/// </param>
		/// <returns>Returns the generated host label.</returns>
		public string GenerateHostLabel( bool generateLongHostLabel )
		{
			return UpnAlg.GenerateHostLabel( this, generateLongHostLabel );
		}

		/// <summary>
		/// Generates the classifier used in the underlying PRNP.
		/// </summary>
		/// <returns>Returns the generated classifier.</returns>
		public string GenerateClassifier()
		{
			return UpnAlg.GenerateClassifier( this );
		}

		/// <summary>
		/// Generates the key used for symmetric encryption.
		/// </summary>
		/// <returns>Returns the key used for symmetric encryption.</returns>
		public byte[] GenerateSymmEncryptionKey()
		{
			return UpnAlg.GenerateSymmEncryptionKey( this );
		}

		/// <summary>
		/// Generates all the lazy loaded properties.
		/// </summary>
		public void GenerateAllLazyLoaded()
		{
			GenerateFingerprint();
			GenerateHostLabel( true );
			GenerateHostLabel( false );
			GenerateClassifier();
			GenerateSymmEncryptionKey();
		}

		#endregion Public methods
	}
}
