////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;

namespace Vela.Crypto.Asymmetric
{
	/// <summary>
	/// Contract for an asymmetric cryptographic algorithm.
	/// </summary>
	/// <remarks>
	/// <para>The size of the algorithm (<see cref="AlgSizeBits"/>) must be multiple of 16 so that it can be an even number of
	/// bytes; this is not a technical requirement, but a practical one.</para>
	/// 
	/// <para>When a key pair is generated, it is also verified for validity.</para>
	/// 
	/// <para>When you want to sign a value (which is normally a hash of the data to sign), process the value with the private
	/// key (<see cref="ProcessWithPrivateKey"/>). You will thus obtain the digital signature.</para>
	/// 
	/// <para>When you want to get the original hash value in order to verify it against the hash of the data whose signature to
	/// verify, process the digital signature with the public key (<see cref="ProcessWithPublicKey"/>).</para>
	/// </remarks>
	public abstract class AsymmAlg : IMutableAlgorithm, IExpirableAlgorithm
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AsymmAlg()
		{
		}

		#endregion Constructors

		#region Events

		/// <summary>
		/// Event raised when the verification of an asymmetric key's parameters fails.
		/// </summary>
		public static event FailedAsymmKeyVerificationDelegate FailedKeyVerificationEvent;

		#endregion Events

		#region Public non-serializable properties

		/// <summary>
		/// The size (in bits) of the cryptographic algorithm. This must be multiple of 16 so that it can be an even number of
		/// bytes; this is not a technical requirement, but a practical one. READ-ONLY.
		/// </summary>
		public abstract int AlgSizeBits { get; }

		/// <summary>
		/// The size (in bytes) of the asymmetric algorithm. Must be an even number; this is not a technical requirement, but a
		/// practical one. READ-ONLY.
		/// </summary>
		public int AlgSizeBytes
		{
			get { return AlgSizeBits / 8; }
		}

		/// <summary>
		/// The maximum size (in bytes) of the data which can be processed. READ-ONLY.
		/// </summary>
		/// <remarks>If you try to pass, to the cryptographic direct methods, input data whose size is bigger than this value,
		/// an exception is thrown.
		/// </remarks>
		public int MaxProcessableSizeBytes
		{
			get { return AlgSizeBytes - 1; }
		}

		/// <summary>
		/// If this is true then the current algorithm contains the public key. READ-ONLY.
		/// </summary>
		public abstract bool HasPublicKey { get; }

		/// <summary>
		/// If this is true then the current algorithm contains the private key. READ-ONLY.
		/// </summary>
		public abstract bool HasPrivateKey { get; }

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public abstract DateTime ExpirationTime { get; }

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Copies the public part of the current algorithm to the specified algorithm (the properties are deeply cloned).
		/// </summary>
		protected abstract void CopyPublicKeyTo( AsymmAlg asymmAlg );

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Checks if the current key pair is valid, that is, if some random data can be processed with the public key and then
		/// processed with the private key to obtain the original data. If the private key does not exist, an exception is
		/// thrown.
		/// </summary>
		/// <returns>Returns true if the current key pair is valid.</returns>
		public bool VerifyKeyPair()
		{
			if( !HasPrivateKey ) throw new ValidationException( "In order to verify a key pair, the private key must exist." );

			byte[] buffer = null, encrypted = null, decrypted = null;
			bool isEqual = false;

			try
			{
				buffer = EntropyPool.GetBytes( MaxProcessableSizeBytes );

				// Process the data with the public key.
				encrypted = ProcessWithPublicKey( buffer, true );

				// Process the data with the private key.
				decrypted = ProcessWithPrivateKey( encrypted, false );

				// Because of the "Zero-byte issue" of the "BigInteger" library, pad the result.
				decrypted = BufferUtils.PadBeginningWithZeros( decrypted, buffer.Length );

				// Compare the original data with the decrypted data.
				isEqual = BufferUtils.Equals( buffer, decrypted );
			}
			finally
			{
				if( !isEqual && (FailedKeyVerificationEvent != null) )
					FailedKeyVerificationEvent( this, buffer, encrypted, decrypted );
			}

			return isEqual;
		}

		/// <summary>
		/// Gets the public part of the current algorithm (the properties are deeply cloned).
		/// </summary>
		/// <returns>Returns the public part of the current algorithm.</returns>
		public abstract AsymmAlg GetPublicAlg();

		/// <summary>
		/// Writes the public parameters in the specified stream.
		/// </summary>
		/// <param name="bns">Stream in which to write the public parameters.</param>
		public abstract void WriteUpnBodySource( BinaryNodeStream bns );

		#endregion Public methods

		#region Cryptographic direct methods

		/// <summary>
		/// Generates a new (random) key pair of the specified size. The key pair is verified for validity.
		/// </summary>
		/// <param name="lop">Long operation progress used during the generation of a key pair. This may be null.</param>
		public abstract void GenerateKeyPair( LongOperationProgress lop );

		/// <summary>
		/// Processes the specified data with the current private key (so that it could only be extracted with the public
		/// key). RESERVED for internal use.
		/// </summary>
		/// <param name="buffer">Buffer to process. The maximum size of this object is <see cref="MaxProcessableSizeBytes"/>.
		/// </param>
		/// <param name="hide">Specifies whether the data to be processed is to be hidden, that is, processed from its raw form
		/// into a form which can be revealed only by converting it back with the public key. If this is false then the data was
		/// previously hidden with the public key.</param>
		/// <returns>Returns the processed data.</returns>
		/// <remarks>
		/// <para>The resulted data will be accessible only to the owner of the public key (associated to the private key which
		/// is used to process the data).</para>
		/// </remarks>
		public abstract byte[] ProcessWithPrivateKey( byte[] buffer, bool hide );

		/// <summary>
		/// Processes the specified data with the current public key (so that it could only be extracted with the private
		/// key). RESERVED for internal use.
		/// </summary>
		/// <param name="buffer">Buffer to process. The maximum size of this object is <see cref="MaxProcessableSizeBytes"/>.
		/// </param>
		/// <param name="hide">Specifies whether the data to be processed is to be hidden, that is, processed from its raw form
		/// into a form which can be revealed only by converting it back with the private key. If this is false then the data
		/// was previously hidden with the private key.</param>
		/// <returns>Returns the processed data.</returns>
		/// <remarks>
		/// <para>The resulted data will be accessible only to the owner of the private key (associated to the public key which
		/// is used to process the data).</para>
		/// </remarks>
		public abstract byte[] ProcessWithPublicKey( byte[] buffer, bool hide );

		#endregion Cryptographic direct methods

		#region Cryptographic public methods

		/// <summary>
		/// Decrypts the specified data with the current private key.
		/// </summary>
		/// <param name="buffer">Buffer to decrypt.</param>
		/// <param name="asymmEncryptionPaddingAlg">The asymmetric encryption padding algorithm. Must be the same as the one
		/// used during encryption.</param>
		/// <returns>Returns the decrypted data.</returns>
		public abstract byte[] DecryptData( byte[] buffer, IAsymmEncryptionPaddingAlg asymmEncryptionPaddingAlg );

		/// <summary>
		/// Encrypts the specified data with the current public key.
		/// </summary>
		/// <param name="buffer">Buffer to encrypt. This should be something small, like a hash.</param>
		/// <param name="asymmEncryptionPaddingAlg">The asymmetric encryption padding algorithm.</param>
		/// <returns>Returns the encrypted data.</returns>
		public abstract byte[] EncryptData( byte[] buffer, IAsymmEncryptionPaddingAlg asymmEncryptionPaddingAlg );

		/// <summary>
		/// Signs the specified data with the current private key.
		/// </summary>
		/// <param name="buffer">Buffer to sign. This should be something small, like a hash.</param>
		/// <param name="asymmSigningPaddingAlg">The asymmetric signing padding algorithm.</param>
		/// <returns>Returns the signature of the specified data.</returns>
		public abstract byte[] SignData( byte[] buffer, IAsymmSigningPaddingAlg asymmSigningPaddingAlg );

		/// <summary>
		/// Verifies the signature of the specified data with the current public key. An
		/// <see cref="InvalidSignatureException"/> (among others) is thrown if the signature is invalid.
		/// </summary>
		/// <param name="buffer">Buffer whose signature has to be verified.</param>
		/// <param name="signature">The signature against which the data's signature is verified.</param>
		/// <param name="asymmSigningPaddingAlg">The asymmetric signing padding algorithm. Must be the same as the one used
		/// during signing.</param>
		public abstract void VerifyDataSignature( byte[] buffer, byte[] signature
			, IAsymmSigningPaddingAlg asymmSigningPaddingAlg );

		#endregion Cryptographic public methods

		#region Blinding public methods

		/// <summary>
		/// Generates random blinding configuration.
		/// </summary>
		/// <param name="buffer">Buffer to blind.</param>
		/// <returns>Returns the generated blinding configuration.</returns>
		public abstract BlindingConfiguration GenerateBlindingConfiguration( byte[] buffer );

		/// <summary>
		/// Blinds the specified data with the specified parameters.
		/// </summary>
		/// <param name="buffer">Buffer to blind.</param>
		/// <param name="blindingConfiguration">Parameters with which to perform the blinding.</param>
		/// <returns>Returns the blinded data.</returns>
		public abstract byte[] Blind( byte[] buffer, BlindingConfiguration blindingConfiguration );

		/// <summary>
		/// Unblinds the specified data with the specified parameters.
		/// </summary>
		/// <param name="buffer">Buffer to unblind.</param>
		/// <param name="blindingConfiguration">Parameters with which to perform the unblinding.</param>
		/// <returns>Returns the unblinded data.</returns>
		public abstract byte[] Unblind( byte[] buffer, BlindingConfiguration blindingConfiguration );

		#endregion Blinding public methods
	}
}
