////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////
//
// This file is based on the work of Chew Keong Tan, and contains improvements made by the Mono Project, released under the MIT
// X11 license.
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;

namespace Vela.Crypto.Asymmetric.Math
{
	/// <summary>
	/// Big integer number mathematics to use in asymmetric cryptography.
	/// </summary>
	/// <remarks>
	/// <para>The internal data of a big integer is stored as an array of atomic blocks (<see cref="DataBlocks"/>), where a
	/// block is a <see cref="UInt32"/> object.</para>
	/// 
	/// <para>Optimizations were implemented from "Handbook of Applied Cryptography" (HAC), found at
	/// http://www.cacr.math.uwaterloo.ca/hac/ .</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Zero-byte issue</b></para>
	/// 
	/// <para>The first bytes of the buffer from which a big integer is created become the most significant digits of the big
	/// integer. Since multiplying 0 with anything results in 0, and because this library ignores those 0s when a big integer is
	/// converted to a buffer, this issue occurs.</para>
	/// 
	/// <para>A buffer which is converted into a big integer, operated upon with this library (so as the output big integer is
	/// equal with the input big integer) and then converted back to a buffer, is never guaranteed to be an identical buffer
	/// (those ignored 0s make the difference).</para>
	/// 
	/// <para>For example, if we encrypt a buffer whose first byte is 0, after decryption and after <see cref="GetBytes"/> is
	/// called, the obtained buffer will not contain the original first 0. Therefore, the decryption result is not the same as
	/// the encrypted data.</para>
	/// 
	/// <para>The issue can be solved by always padding with 0s the buffers obtained with the <see cref="GetBytes"/> method,
	/// before they are compared to the original buffers.</para>
	/// 
	/// <para>The issue can also be solved by adding a non-zero byte at the beginning of the buffer, before the buffer is
	/// encrypted through this library. After decryption, that byte must be removed.</para>
	/// </remarks>
	public sealed class BigInteger
	{
		#region Public constants

		/// <summary>
		/// The number of bits in an atomic block (<see cref="UInt32"/>).
		/// </summary>
		public const Int32 BlockBitCount = 32;

		/// <summary>
		/// The number of bytes in an atomic block.
		/// </summary>
		public const Int32 BlockByteCount = BlockBitCount / 8;

		/// <summary>
		/// The number of bits in an atomic block expressed as power of 2:
		/// <c><see cref="BlockBitCount"/> = 2 ^ BlockPow2BitCount</c>.
		/// </summary>
		public const Int32 BlockPow2BitCount = 5;

		/// <summary>
		/// The maximum value of an atomic block.
		/// </summary>
		public const UInt32 BlockValMax = UInt32.MaxValue;

		/// <summary>
		/// The bit mask for the number of bits in an atomic block:
		/// <c>2 ^ <see cref="BlockPow2BitCount"/> - 1</c>.
		/// </summary>
		public const Int32 BlockBitCountMask = 31;

		/// <summary>
		/// The maximum size (in bits) of an asymmetric algorithm. This size is not a technical limitation, but a practical one.
		/// This value must be multiple of <see cref="BlockBitCount"/>.
		/// </summary>
		public const Int32 AsymAlgSizeBitsMax = 8192;

		/// <summary>
		/// The maximum size (in bits) of a big integer. This size is not a technical limitation, but a practical one. This
		/// value must be multiple of <see cref="BlockBitCount"/>.
		/// </summary>
		/// <remarks>Currently, this value allows for asymmetric algorithms of sizes of up to <see cref="AsymAlgSizeBitsMax"/>
		/// bits.
		/// </remarks>
		public const Int32 SizeBitsMax = 2 * (AsymAlgSizeBitsMax + BlockBitCount);

		/// <summary>
		/// The maximum size (in bytes) of a big integer. This size is not really a technical limitation, but a practical one.
		/// </summary>
		public const Int32 SizeBytesMax = SizeBitsMax / 8;

		/// <summary>
		/// The maximum size (in blocks) of a big integer. This size is not really a technical limitation, but a practical one.
		/// </summary>
		public const Int32 SizeBlocksMax = SizeBitsMax / 8 / BlockByteCount;

		/// <summary>
		/// The exception message for when an operation would return a negative big integer.
		/// </summary>
		private const string WouldReturnNegVal = "Operation would return a negative big integer.";

		#endregion Public constants

		#region Tables

		/// <summary>
		/// Table of small primes.
		/// </summary>
		/// <remarks>
		/// <para>This table was generated using Mathematica using the following:
		/// <c>PrimeTable[ 6000 ]</c> where
		/// <c>PrimeTable[ x_ ] := Prime[ Range[ 1, PrimePi[ x ] ] ].</c></para>
		/// 
		/// <para>The number of small primes is determined experimentally, so that the generation of primes is fastest.</para>
		/// </remarks>
		private static readonly UInt32[] SmallPrimes =
		{
			2, 3, 5, 7, 11, 13, 17, 19, 23, 29
			, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71
			, 73, 79, 83, 89, 97, 101, 103, 107, 109, 113
			, 127, 131, 137, 139, 149, 151, 157, 163, 167, 173
			, 179, 181, 191, 193, 197, 199, 211, 223, 227, 229
			, 233, 239, 241, 251, 257, 263, 269, 271, 277, 281
			, 283, 293, 307, 311, 313, 317, 331, 337, 347, 349
			, 353, 359, 367, 373, 379, 383, 389, 397, 401, 409
			, 419, 421, 431, 433, 439, 443, 449, 457, 461, 463
			, 467, 479, 487, 491, 499, 503, 509, 521, 523, 541
			, 547, 557, 563, 569, 571, 577, 587, 593, 599, 601
			, 607, 613, 617, 619, 631, 641, 643, 647, 653, 659
			, 661, 673, 677, 683, 691, 701, 709, 719, 727, 733
			, 739, 743, 751, 757, 761, 769, 773, 787, 797, 809
			, 811, 821, 823, 827, 829, 839, 853, 857, 859, 863
			, 877, 881, 883, 887, 907, 911, 919, 929, 937, 941
			, 947, 953, 967, 971, 977, 983, 991, 997, 1009, 1013
			, 1019, 1021, 1031, 1033, 1039, 1049, 1051, 1061, 1063, 1069
			, 1087, 1091, 1093, 1097, 1103, 1109, 1117, 1123, 1129, 1151
			, 1153, 1163, 1171, 1181, 1187, 1193, 1201, 1213, 1217, 1223
			, 1229, 1231, 1237, 1249, 1259, 1277, 1279, 1283, 1289, 1291
			, 1297, 1301, 1303, 1307, 1319, 1321, 1327, 1361, 1367, 1373
			, 1381, 1399, 1409, 1423, 1427, 1429, 1433, 1439, 1447, 1451
			, 1453, 1459, 1471, 1481, 1483, 1487, 1489, 1493, 1499, 1511
			, 1523, 1531, 1543, 1549, 1553, 1559, 1567, 1571, 1579, 1583
			, 1597, 1601, 1607, 1609, 1613, 1619, 1621, 1627, 1637, 1657
			, 1663, 1667, 1669, 1693, 1697, 1699, 1709, 1721, 1723, 1733
			, 1741, 1747, 1753, 1759, 1777, 1783, 1787, 1789, 1801, 1811
			, 1823, 1831, 1847, 1861, 1867, 1871, 1873, 1877, 1879, 1889
			, 1901, 1907, 1913, 1931, 1933, 1949, 1951, 1973, 1979, 1987
			, 1993, 1997, 1999, 2003, 2011, 2017, 2027, 2029, 2039, 2053
			, 2063, 2069, 2081, 2083, 2087, 2089, 2099, 2111, 2113, 2129
			, 2131, 2137, 2141, 2143, 2153, 2161, 2179, 2203, 2207, 2213
			, 2221, 2237, 2239, 2243, 2251, 2267, 2269, 2273, 2281, 2287
			, 2293, 2297, 2309, 2311, 2333, 2339, 2341, 2347, 2351, 2357
			, 2371, 2377, 2381, 2383, 2389, 2393, 2399, 2411, 2417, 2423
			, 2437, 2441, 2447, 2459, 2467, 2473, 2477, 2503, 2521, 2531
			, 2539, 2543, 2549, 2551, 2557, 2579, 2591, 2593, 2609, 2617
			, 2621, 2633, 2647, 2657, 2659, 2663, 2671, 2677, 2683, 2687
			, 2689, 2693, 2699, 2707, 2711, 2713, 2719, 2729, 2731, 2741
			, 2749, 2753, 2767, 2777, 2789, 2791, 2797, 2801, 2803, 2819
			, 2833, 2837, 2843, 2851, 2857, 2861, 2879, 2887, 2897, 2903
			, 2909, 2917, 2927, 2939, 2953, 2957, 2963, 2969, 2971, 2999
			, 3001, 3011, 3019, 3023, 3037, 3041, 3049, 3061, 3067, 3079
			, 3083, 3089, 3109, 3119, 3121, 3137, 3163, 3167, 3169, 3181
			, 3187, 3191, 3203, 3209, 3217, 3221, 3229, 3251, 3253, 3257
			, 3259, 3271, 3299, 3301, 3307, 3313, 3319, 3323, 3329, 3331
			, 3343, 3347, 3359, 3361, 3371, 3373, 3389, 3391, 3407, 3413
			, 3433, 3449, 3457, 3461, 3463, 3467, 3469, 3491, 3499, 3511
			, 3517, 3527, 3529, 3533, 3539, 3541, 3547, 3557, 3559, 3571
			, 3581, 3583, 3593, 3607, 3613, 3617, 3623, 3631, 3637, 3643
			, 3659, 3671, 3673, 3677, 3691, 3697, 3701, 3709, 3719, 3727
			, 3733, 3739, 3761, 3767, 3769, 3779, 3793, 3797, 3803, 3821
			, 3823, 3833, 3847, 3851, 3853, 3863, 3877, 3881, 3889, 3907
			, 3911, 3917, 3919, 3923, 3929, 3931, 3943, 3947, 3967, 3989
		};

		#endregion Tables

		#region Enumerations

		/// <summary>
		/// The available values for the sign of a big integer.
		/// </summary>
		public enum Signs : int
		{
			/// <summary>
			/// Indicates a negative number.
			/// </summary>
			Negative = -1,

			/// <summary>
			/// Indicates a 0 number.
			/// </summary>
			Zero = 0,

			/// <summary>
			/// Indicates a positive number.
			/// </summary>
			Positive = 1
		}

		#endregion Enumerations

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="size">The size (in blocks) of the new big integer. This must be at least 1.</param>
		public BigInteger( Int32 size )
		{
			ValidationUtils.Integer( size, 1, SizeBlocksMax, "size" );

			DataBlocks = new UInt32[ size ];
			SizeBlocks = size;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="bi">The big integer to clone.</param>
		public BigInteger( BigInteger bi )
		{
			ValidationUtils.ObjectHasContent( bi, "bi" );

			DataBlocks = (UInt32[])bi.DataBlocks.Clone();
			SizeBlocks = bi.SizeBlocks;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="bi">The big integer from which to copy data.</param>
		/// <param name="size">The size (in blocks) of the new big integer. This must be at least
		/// <see cref="BigInteger.SizeBlocks"/> of "bi".</param>
		public BigInteger( BigInteger bi, Int32 size )
		{
			ValidationUtils.ObjectHasContent( bi, "bi" );
			ValidationUtils.Integer( size, bi.SizeBlocks, SizeBlocksMax, "size" );

			DataBlocks = new UInt32[ size ];

			for( Int32 i = 0; i < bi.SizeBlocks; i++ ) DataBlocks[ i ] = bi.DataBlocks[ i ];

			SizeBlocks = bi.SizeBlocks;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="bi">The data of the new big integer. The bit count of this buffer must be at least 1.</param>
		public BigInteger( byte[] bi )
		{
			ValidationUtils.ArrayHasContent( bi, "bi" );
			ValidationUtils.Integer( bi.Length * 8, 1, SizeBitsMax, "bi.Length" );

			SizeBlocks = bi.Length >> 2;
			Int32 leftOver = bi.Length & 0x03;

			// Size not multiple of 4.
			if( leftOver != 0 ) SizeBlocks++;

			DataBlocks = new UInt32[ SizeBlocks ];

			for( Int32 i = bi.Length - 1, j = 0; i >= 3; i -= 4, j++ )
			{
				DataBlocks[ j ] =
					((UInt32)bi[ i - 3 ] << 24)
					| ((UInt32)bi[ i - 2 ] << 16)
					| ((UInt32)bi[ i - 1 ] << 8)
					| bi[ i ];
			}

			switch( leftOver )
			{
				case 1:
					DataBlocks[ SizeBlocks - 1 ] = bi[ 0 ];
					break;
				case 2:
					DataBlocks[ SizeBlocks - 1 ] = (UInt32)((bi[ 0 ] << 8) | bi[ 1 ]);
					break;
				case 3:
					DataBlocks[ SizeBlocks - 1 ] = (UInt32)((bi[ 0 ] << 16)
						| (bi[ 1 ] << 8) | bi[ 2 ]);
					break;
			}

			Normalize();
		}

		#endregion Constructors

		#region Conversions

		/// <summary>
		/// Converts the specified integer value to a big integer.
		/// </summary>
		/// <param name="value">Value to convert to a big integer. This may be 0.</param>
		/// <returns>Returns the resulted big integer.</returns>
		public static implicit operator BigInteger( UInt32 value )
		{
			BigInteger bi = new BigInteger( 1 );
			bi.DataBlocks[ 0 ] = value;

			return bi;
		}

		#endregion Conversions

		#region Data storage

		/// <summary>
		/// The size (in blocks, "UInt32") of the current big integer.
		/// </summary>
		private Int32 SizeBlocks = 1;

		/// <summary>
		/// The block data of the current big integer.
		/// </summary>
		private UInt32[] DataBlocks;

		#endregion Data storage

		#region Operators

		/// <summary>
		/// Adds the specified big integers.
		/// </summary>
		/// <param name="bi1">Big integer to add.</param>
		/// <param name="bi2">Big integer to add.</param>
		/// <returns>Returns the added big integers.</returns>
		public static BigInteger operator +( BigInteger bi1, BigInteger bi2 )
		{
			if( bi1 == 0 ) return new BigInteger( bi2 );
			else if( bi2 == 0 ) return new BigInteger( bi1 );
			else return Kernel.AddSameSign( bi1, bi2 );
		}

		/// <summary>
		/// Subtracts the specified big integers.
		/// </summary>
		/// <param name="bi1">Big integer to subtract from.</param>
		/// <param name="bi2">Big integer to subtract.</param>
		/// <returns>Returns the difference between the big integers.</returns>
		public static BigInteger operator -( BigInteger bi1, BigInteger bi2 )
		{
			if( bi2 == 0 ) return new BigInteger( bi1 );

			if( bi1 == 0 ) throw new ValidationException( WouldReturnNegVal );

			Signs sign = Kernel.Compare( bi1, bi2 );
			switch( sign )
			{
				case Signs.Zero:
					return 0;
				case Signs.Positive:
					return Kernel.Subtract( bi1, bi2 );
				case Signs.Negative:
					throw new ValidationException( WouldReturnNegVal );
				default:
					throw ValidationUtils.Unhandled( sign, "sign" );
			}
		}

		/// <summary>
		/// Computes the modulus of the specified big integer.
		/// </summary>
		/// <param name="bi">Big integer whose modulus to compute.</param>
		/// <param name="ui">Modulus.</param>
		/// <returns>Returns the remaining value.</returns>
		public static UInt32 operator %( BigInteger bi, UInt32 ui )
		{
			return Kernel.BlockMod( bi, ui );
		}

		/// <summary>
		/// Computes the modulus of the specified big integer.
		/// </summary>
		/// <param name="bi1">Big integer whose modulus to compute.</param>
		/// <param name="bi2">Modulus.</param>
		/// <returns>Returns the remaining value.</returns>
		public static BigInteger operator %( BigInteger bi1, BigInteger bi2 )
		{
			return Kernel.MultiByteDivide( bi1, bi2 )[ 1 ];
		}

		/// <summary>
		/// Divides the specified big integer.
		/// </summary>
		/// <param name="bi">The dividend big integer to divide.</param>
		/// <param name="i">Divisor.</param>
		/// <returns>Returns the result of division.</returns>
		public static BigInteger operator /( BigInteger bi, UInt32 i )
		{
			if( i > 0 ) return Kernel.BlockDiv( bi, i );

			throw new ValidationException( WouldReturnNegVal );
		}

		/// <summary>
		/// Divides the specified big integer.
		/// </summary>
		/// <param name="bi1">The dividend big integer to divide.</param>
		/// <param name="bi2">Divisor.</param>
		/// <returns>Returns the result of division.</returns>
		public static BigInteger operator /( BigInteger bi1, BigInteger bi2 )
		{
			return Kernel.MultiByteDivide( bi1, bi2 )[ 0 ];
		}

		/// <summary>
		/// Multiplies the specified big integers.
		/// </summary>
		/// <param name="bi1">Big integer to multiply.</param>
		/// <param name="bi2">Big integer to multiply.</param>
		/// <returns>Returns the result of multiplication.</returns>
		public static BigInteger operator *( BigInteger bi1, BigInteger bi2 )
		{
			if( bi1 == 0 || bi2 == 0 ) return 0;

			ValidationUtils.Integer( bi1.DataBlocks.Length, bi1.SizeBlocks, int.MaxValue, "bi1.DataBlocks.Length" );
			ValidationUtils.Integer( bi2.DataBlocks.Length, bi2.SizeBlocks, int.MaxValue, "bi2.DataBlocks.Length" );

			BigInteger ret = new BigInteger( bi1.SizeBlocks + bi2.SizeBlocks );

			Kernel.Multiply( bi1.DataBlocks, 0, bi1.SizeBlocks, bi2.DataBlocks, 0, bi2.SizeBlocks, ret.DataBlocks, 0 );

			ret.Normalize();
			return ret;
		}

		/// <summary>
		/// Multiplies the specified big integer.
		/// </summary>
		/// <param name="bi">Big integer to multiply.</param>
		/// <param name="i">Integer to multiply.</param>
		/// <returns>Returns the result of multiplication.</returns>
		public static BigInteger operator *( BigInteger bi, UInt32 i )
		{
			if( i < 0 ) throw new ValidationException( WouldReturnNegVal );
			if( i == 0 ) return 0;
			if( i == 1 ) return new BigInteger( bi );

			return Kernel.MultiplyByBlock( bi, i );
		}

		/// <summary>
		/// Shifts to the left the bits of the specified big integer, with the specified value.
		/// </summary>
		/// <param name="bi1">Big integer whose bits to shift.</param>
		/// <param name="shiftVal">Number of bits to shift with.</param>
		/// <returns>Returns the shifted big integer.</returns>
		public static BigInteger operator <<( BigInteger bi1, Int32 shiftVal )
		{
			return Kernel.LeftShift( bi1, shiftVal );
		}

		/// <summary>
		/// Shifts to the right the bits of the specified big integer, with the specified value.
		/// </summary>
		/// <param name="bi1">Big integer whose bits to shift.</param>
		/// <param name="shiftVal">Number of bits to shift with.</param>
		/// <returns>Returns the shifted big integer.</returns>
		public static BigInteger operator >>( BigInteger bi1, Int32 shiftVal )
		{
			return Kernel.RightShift( bi1, shiftVal );
		}

		#endregion Operators

		#region Random

		/// <summary>
		/// Generates a random big integer of the specified size.
		/// </summary>
		/// <param name="bitCount">The number of bits for the new number. This must be at least 1.</param>
		/// <returns>Returns the generated random number.</returns>
		public static BigInteger GenerateRandom( Int32 bitCount )
		{
			ValidationUtils.Integer( bitCount, 1, SizeBitsMax, "bitCount" );

			Int32 blocks = bitCount >> BlockPow2BitCount;
			Int32 remBits = bitCount & BlockBitCountMask;

			if( remBits != 0 ) blocks++;

			BigInteger random = new BigInteger( blocks + 1 );
			byte[] randomBuffer = EntropyPool.GetBytes( blocks << 2 );
			Buffer.BlockCopy( randomBuffer, 0, random.DataBlocks, 0, randomBuffer.Length );

			if( remBits != 0 )
			{
				UInt32 mask = (UInt32)1 << (remBits - 1);
				random.DataBlocks[ blocks - 1 ] |= mask;

				mask = BlockValMax >> (BlockBitCount - remBits);
				random.DataBlocks[ blocks - 1 ] &= mask;
			}
			else random.DataBlocks[ blocks - 1 ] |= 0x80000000;

			random.Normalize();

			return random;
		}

		#endregion Random

		#region Bitwise

		/// <summary>
		/// Returns the number of bits from the current big integer where the most significant bit is 1.
		/// </summary>
		/// <returns>Returns the number of bits from the current big integer where the most significant bit is 1.</returns>
		public Int32 GetBitCount()
		{
			Normalize();

			UInt32 val = DataBlocks[ SizeBlocks - 1 ];
			UInt32 mask = 0x80000000;
			Int32 bits = BlockBitCount;

			while( (bits > 0) && ((val & mask) == 0) )
			{
				bits--;
				mask >>= 1;
			}

			bits += (SizeBlocks - 1) << BlockPow2BitCount;

			return bits;
		}

		/// <summary>
		/// Tests if the bit with the specified index is 1.
		/// </summary>
		/// <param name="bitIndex">Index of bit to test. This must be at least 0.</param>
		/// <returns>Returns true if the bit with the specified index is 1.
		/// </returns>
		public bool TestBit( Int32 bitIndex )
		{
			ValidationUtils.Integer( bitIndex, 0, SizeBlocks * BlockBitCount - 1, "bitIndex" );

			Int32 bytePos = bitIndex >> BlockPow2BitCount;
			byte bitPos = (byte)(bitIndex & BlockBitCountMask);

			UInt32 mask = (UInt32)1 << bitPos;
			return (DataBlocks[ bytePos ] | mask) == DataBlocks[ bytePos ];
		}

		/// <summary>
		/// Sets the bit with the specified index with the specified value.
		/// </summary>
		/// <param name="bitIndex">Index of bit to set. This must be at least 0.</param>
		/// <param name="value">If this is true then the specified bit is set with 1.</param>
		public void SetBit( Int32 bitIndex, bool value = true )
		{
			ValidationUtils.Integer( bitIndex, 0, SizeBlocks * BlockBitCount - 1, "bitIndex" );

			Int32 bytePos = bitIndex >> BlockPow2BitCount;

			if( bytePos < SizeBlocks )
			{
				UInt32 mask = (UInt32)1 << (bitIndex & BlockBitCountMask);

				if( value ) DataBlocks[ bytePos ] |= mask;
				else DataBlocks[ bytePos ] &= ~mask;
			}
		}

		/// <summary>
		/// Sets the bit with the specified index to 0.
		/// </summary>
		/// <param name="bitIndex">Index of bit to set.</param>
		public void ClearBit( Int32 bitIndex )
		{
			SetBit( bitIndex, false );
		}

		/// <summary>
		/// Returns the index of the least significant bit with the value 1.
		/// </summary>
		/// <returns>Returns the index of the least significant bit with the value 1.</returns>
		public Int32 LowestSetBit()
		{
			if( this == 0 ) throw new ValidationException( "The big integer is 0, so it has no bit set to 1." );

			Int32 i = 0;
			while( !TestBit( i ) ) i++;

			return i;
		}

		/// <summary>
		/// Gets the data block with the specified index in the current big integer.
		/// </summary>
		/// <remarks>
		/// <para>The index of the data block to get. This must be at least 0.</para>
		/// </remarks>
		public UInt32 GetDataBlock( Int32 index )
		{
			ValidationUtils.Integer( index, 0, SizeBlocks - 1, "index" );

			return DataBlocks[ index ];
		}

		/// <summary>
		/// Returns the value of the number as a buffer. The lowest index contains the MSB.
		/// </summary>
		/// <returns>Returns the binary representation of the current big integer, in big-endian format.</returns>
		public byte[] GetBytes()
		{
			Normalize();

			if( this == 0 ) return new byte[ 1 ];

			Int32 numBits = GetBitCount();
			Int32 numBytes = numBits / 8;
			if( (numBits & 0x07) != 0 ) numBytes++;

			byte[] result = new byte[ numBytes ];

			// The number of bytes from a block, initialized with the number of bytes from the first block.
			Int32 numBytesInBlock = numBytes & 0x03;
			if( numBytesInBlock == 0 ) numBytesInBlock = 4;

			Int32 pos = 0;
			for( Int32 i = SizeBlocks - 1; i >= 0; i-- )
			{
				UInt32 val = DataBlocks[ i ];

				for( Int32 j = numBytesInBlock - 1; j >= 0; j-- )
				{
					result[ pos + j ] = (byte)(val & 0xFF);
					val >>= 8;
				}

				pos += numBytesInBlock;
				numBytesInBlock = 4; // The number of bytes from all the other blocks.
			}

			return result;
		}

		#endregion Bitwise

		#region Compare

		/// <summary>
		/// Checks if the specified big integers are equal.
		/// </summary>
		/// <param name="bi1">Big integer to compare. This may be null.</param>
		/// <param name="bi2">Big integer to compare. This may be null.</param>
		/// <returns>Returns true if the numbers are equal.</returns>
		public static bool operator ==( BigInteger bi1, BigInteger bi2 )
		{
			if( ReferenceEquals( bi1, bi2 ) ) return true;
			if( ((object)bi1 == null) || ((object)bi2 == null) ) return false;

			return Kernel.Compare( bi1, bi2 ) == 0;
		}

		/// <summary>
		/// Checks if the specified big integer and integer are equal.
		/// </summary>
		/// <param name="bi">Big integer to compare. This may be null.</param>
		/// <param name="ui">Integer to compare.</param>
		/// <returns>Returns true if the numbers are equal.</returns>
		public static bool operator ==( BigInteger bi, UInt32 ui )
		{
			if( bi.SizeBlocks != 1 ) bi.Normalize();

			return (bi.SizeBlocks == 1) && (bi.DataBlocks[ 0 ] == ui);
		}

		/// <summary>
		/// Checks if the specified big integers are different.
		/// </summary>
		/// <param name="bi1">Big integer to compare. This may be null.</param>
		/// <param name="bi2">Big integer to compare. This may be null.</param>
		/// <returns>Returns true if the numbers are different.</returns>
		public static bool operator !=( BigInteger bi1, BigInteger bi2 )
		{
			if( ReferenceEquals( bi1, bi2 ) ) return false;
			if( ((object)bi1 == null) || ((object)bi2 == null) ) return true;

			return Kernel.Compare( bi1, bi2 ) != 0;
		}

		/// <summary>
		/// Checks if the specified big integer and integer are different.
		/// </summary>
		/// <param name="bi">Big integer to compare. This may be null.</param>
		/// <param name="ui">Integer to compare.</param>
		/// <returns>Returns true if the numbers are different.</returns>
		public static bool operator !=( BigInteger bi, UInt32 ui )
		{
			if( bi.SizeBlocks != 1 ) bi.Normalize();

			return !((bi.SizeBlocks == 1) && (bi.DataBlocks[ 0 ] == ui));
		}

		/// <summary>
		/// Checks if the first specified big integer is bigger than the second specified big integer.
		/// </summary>
		/// <param name="bi1">First big integer to compare. This may be null.</param>
		/// <param name="bi2">Second big integer to compare. This may be null.</param>
		/// <returns>Returns true if the first number is bigger than the second.</returns>
		public static bool operator >( BigInteger bi1, BigInteger bi2 )
		{
			return Kernel.Compare( bi1, bi2 ) > 0;
		}

		/// <summary>
		/// Checks if the first specified big integer is bigger than or equal with the second specified big integer.
		/// </summary>
		/// <param name="bi1">First big integer to compare. This may be null.</param>
		/// <param name="bi2">Second big integer to compare. This may be null.</param>
		/// <returns>Returns true if the first number is bigger than or equal with the second.</returns>
		public static bool operator >=( BigInteger bi1, BigInteger bi2 )
		{
			return Kernel.Compare( bi1, bi2 ) >= 0;
		}

		/// <summary>
		/// Checks if the first specified big integer is smaller than the second specified big integer.
		/// </summary>
		/// <param name="bi1">First big integer to compare. This may be null.</param>
		/// <param name="bi2">Second big integer to compare. This may be null.</param>
		/// <returns>Returns true if the first number is smaller than the second.</returns>
		public static bool operator <( BigInteger bi1, BigInteger bi2 )
		{
			return Kernel.Compare( bi1, bi2 ) < 0;
		}

		/// <summary>
		/// Checks if the first specified big integer is smaller or equal with the second specified big integer.
		/// </summary>
		/// <param name="bi1">First big integer to compare. This may be null.</param>
		/// <param name="bi2">Second big integer to compare. This may be null.</param>
		/// <returns>Returns true if the first number is smaller or equal with the second.</returns>
		public static bool operator <=( BigInteger bi1, BigInteger bi2 )
		{
			return Kernel.Compare( bi1, bi2 ) <= 0;
		}

		/// <summary>
		/// Compares the current big integer with the specified big integer.
		/// </summary>
		/// <param name="bi">Big integer to compare with. This may be null.</param>
		/// <returns>Returns -1 if the current big integer is smaller than the specified big integer, 0 if they are equal, and 1
		/// if the current big integer is bigger than the specified big integer</returns>
		public Signs Compare( BigInteger bi )
		{
			return Kernel.Compare( this, bi );
		}

		#endregion Compare

		#region Various

		/// <summary>
		/// Normalizes this by setting the size to the actual number of atomic blocks used in data.
		/// </summary>
		private void Normalize()
		{
			// Normalize size.
			while( (SizeBlocks > 0) && (DataBlocks[ SizeBlocks - 1 ] == 0) ) SizeBlocks--;

			// Check for 0.
			if( SizeBlocks <= 0 ) SizeBlocks = 1;
		}

		/// <summary>
		/// Clears the content of the current number.
		/// </summary>
		public void Clear()
		{
			for( Int32 i = 0; i < SizeBlocks; i++ ) DataBlocks[ i ] = 0;
		}

		#endregion Various

		#region "Object" class implementation

		/// <summary>
		/// Gets the hash code of the big integer.
		/// </summary>
		/// <returns>Returns the hash code of the big integer.</returns>
		public override Int32 GetHashCode()
		{
			return BufferUtils.ComputeBlockChecksum32( DataBlocks );
		}

		/// <summary>
		/// Checks if the current big integer is equal with the specified big integer.
		/// </summary>
		/// <param name="bi">Big integer to compare with. This may be null.</param>
		/// <returns>Returns true if the two big integers are equal.</returns>
		public override bool Equals( object bi )
		{
			if( bi as BigInteger == null ) return false;
			if( bi is UInt32 ) return ((UInt32)bi >= 0) && (this == (UInt32)bi);

			return Kernel.Compare( this, (BigInteger)bi ) == 0;
		}

		#endregion "Object" class implementation

		#region Number Theory

		/// <summary>
		/// Computes the Greatest Common Divisor between the current big integer and the specified big integer.
		/// </summary>
		/// <param name="bi">Big integer to use.</param>
		/// <returns>Returns the Greatest Common Divisor.</returns>
		public BigInteger Gcd( BigInteger bi )
		{
			return Kernel.Gcd( this, bi );
		}

		/// <summary>
		/// Computes the inverse modulus.
		/// </summary>
		/// <param name="modulus">Modulus.</param>
		/// <returns>Returns the inverse modulus.</returns>
		public BigInteger ModInverse( BigInteger modulus )
		{
			return Kernel.ModInverse( this, modulus );
		}

		/// <summary>
		/// Raises the current big integer to the specified power, in the specified modulus ring.
		/// </summary>
		/// <param name="exp">Exponent to raise the current big integer to.</param>
		/// <param name="n">Modulus.</param>
		/// <returns>Returns the current big integer raised the to the specified power, in the specified modulus ring.</returns>
		public BigInteger ModPow( BigInteger exp, BigInteger n )
		{
			ModulusRing mr = new ModulusRing( n );
			return mr.Pow( this, exp );
		}

		#endregion Number Theory

		#region Prime number generation

		/// <summary>
		/// Checks if the specified big integer (which must be an odd number) is divisible by a small prime, and thusly is not
		/// prime.
		/// </summary>
		/// <param name="bi">Big integer to check.</param>
		/// <param name="smallestPrimeRem">The remainder of the division between the specified big integer and the product of
		/// the smallest primes.</param>
		/// <returns>Returns true if the specified number is divisible by a small prime, and thusly is not prime.</returns>
		private static bool CheckSmallPrimeDivisors( BigInteger bi, UInt32 smallestPrimeRem )
		{
			// Sieve out the numbers divisible by the smallest primes, for speed.
			//
			// Don't include 2 because it doesn't divide odd numbers.
			//
			// Can't use more than the first 9 primes because the product overflows an "UInt32", which means that
			// "smallestPrimeRem" can't be calculated.
			if( smallestPrimeRem % 3 == 0 ) return true;
			if( smallestPrimeRem % 5 == 0 ) return true;
			if( smallestPrimeRem % 7 == 0 ) return true;
			if( smallestPrimeRem % 11 == 0 ) return true;
			if( smallestPrimeRem % 13 == 0 ) return true;
			if( smallestPrimeRem % 17 == 0 ) return true;
			if( smallestPrimeRem % 19 == 0 ) return true;
			if( smallestPrimeRem % 23 == 0 ) return true;
			if( smallestPrimeRem % 29 == 0 ) return true;

			// Skip the prime divisors tested above.
			for( Int32 i = 10; i < SmallPrimes.Length; i++ )
			{
				if( bi % SmallPrimes[ i ] == 0 ) return true;
			}

			return false;
		}

		/// <summary>
		/// Generates a random prime number.
		/// </summary>
		/// <param name="bitCount">The size in bits of the prime number. This must be at least 1.</param>
		/// <param name="lop">Long operation progress used during the generation of the identity. This may be null.</param>
		/// <returns>Returns the newly generated prime number.</returns>
		/// <remarks>It is guaranteed that the returned number has his <see cref="BigInteger.GetBitCount">bit count</see> equal
		/// with the specified value ("bitCount").
		/// </remarks>
		public static BigInteger GeneratePrime( Int32 bitCount, LongOperationProgress lop )
		{
			ValidationUtils.Integer( bitCount, 1, SizeBitsMax, "bitCount" );

			#region Prepare random number

			// Find a place to do a sequential search.
			BigInteger currVal = GenerateRandom( bitCount );

			// Make the number odd because only odd numbers can be prime.
			currVal.SetBit( 0 );

			// Optimizations required by the generation of an RSA key pair:
			//
			// "M" must have a size equal with the key size, and "P" and "Q" must have about the same size, so their size
			// must equal with half of the key size. To guarantee this, the most significant bit of "P" and "Q" must be 1.
			//
			// In order for "M = P * Q" to have a size equal with the key size, it is necessary to set the second most
			// significant bit of "P" and "Q" to 1. This method guarantees that "P" and "Q" have to be generated only one
			// time.
			//
			// The third most significant bit of "P" and "Q" must be 0 so that their generation would have enough space to
			// search for a prime number without overflowing.

			currVal.SetBit( bitCount - 1 );
			currVal.SetBit( bitCount - 2 );
			currVal.ClearBit( bitCount - 3 );

			#endregion Prepare random number

			// The product of the smallest primes. Can't use more than the first 9 primes because the product overflows an
			// "UInt32".
			const UInt32 smallestPrimeProd = (UInt32)3 * 5 * 7 * 11 * 13 * 17 * 19 * 23 * 29;

			UInt32 smallestPrimeRem = currVal % smallestPrimeProd;

			// Search for primes.
			while( true )
			{
				if( !CheckSmallPrimeDivisors( currVal, smallestPrimeRem ) )
				{
					// Use a primality test.
					if( PrimalityTests.MillerRabin( currVal ) ) return currVal;

					// Keep this after all tests so that it would affect speed the least.
					if( lop != null ) lop.Progress( 0, "Generating prime number..." );
				}

				// Prepare the next cycle for non-prime numbers.
				smallestPrimeRem += 2;
				if( smallestPrimeRem >= smallestPrimeProd ) smallestPrimeRem -= smallestPrimeProd;

				currVal.Incr2();
			}
		}

		/// <summary>
		/// Increments the current big integer with 2.
		/// </summary>
		public void Incr2()
		{
			unchecked
			{
				Int32 i = 0;

				unchecked { DataBlocks[ 0 ] += 2; }

				// If there was no carry, nothing to do.
				if( DataBlocks[ 0 ] < 2 )
				{
					// Account for the first carry.
					i++;
					unchecked { DataBlocks[ i ]++; }

					// Keep adding until no carry.
					while( DataBlocks[ i++ ] == 0 ) unchecked { DataBlocks[ i ]++; }

					// Check if the data size has to be increased.
					if( SizeBlocks == i ) SizeBlocks++;
				}
			}
		}

		#endregion Prime number generation

		#region Private classes

		/// <summary>
		/// Low level functions for big integer numbers.
		/// </summary>
		private static class Kernel
		{
			#region Addition / subtraction

			/// <summary>
			/// Adds two numbers with the same sign.
			/// </summary>
			/// <param name="bi1">A BigInteger.</param>
			/// <param name="bi2">A BigInteger.</param>
			/// <returns>bi1 + bi2</returns>
			public static BigInteger AddSameSign( BigInteger bi1, BigInteger bi2 )
			{
				UInt32[] x, y;
				Int32 yMax, xMax, i = 0;

				// "x" should be bigger.
				if( bi1.SizeBlocks < bi2.SizeBlocks )
				{
					x = bi2.DataBlocks;
					xMax = bi2.SizeBlocks;
					y = bi1.DataBlocks;
					yMax = bi1.SizeBlocks;
				}
				else
				{
					x = bi1.DataBlocks;
					xMax = bi1.SizeBlocks;
					y = bi2.DataBlocks;
					yMax = bi2.SizeBlocks;
				}

				BigInteger result = new BigInteger( xMax + 1 );

				UInt32[] r = result.DataBlocks;

				UInt64 sum = 0;

				// Add common parts of both numbers.
				do
				{
					sum = ((UInt64)x[ i ]) + ((UInt64)y[ i ]) + sum;
					r[ i ] = unchecked( (UInt32)sum );
					sum >>= BlockBitCount;
				}
				while( ++i < yMax );

				// Copy remainder of longer number while carry propagation is required.
				bool carry = sum != 0;

				if( carry )
				{
					if( i < xMax )
					{
						do { carry = (r[ i ] = x[ i ] + 1) == 0; }
						while( (++i < xMax) && carry );
					}

					if( carry )
					{
						r[ i ] = 1;
						result.SizeBlocks = ++i;
						return result;
					}
				}

				// Copy the rest
				if( i < xMax )
				{
					do { r[ i ] = x[ i ]; }
					while( ++i < xMax );
				}

				result.Normalize();
				return result;
			}

			public static BigInteger Subtract( BigInteger big, BigInteger small )
			{
				BigInteger result = new BigInteger( big.SizeBlocks );

				UInt32[] r = result.DataBlocks, b = big.DataBlocks, s = small.DataBlocks;
				UInt32 c = 0;
				Int32 i = 0;

				do
				{
					UInt32 x = s[ i ];

					unchecked
					{
						if( ((x += c) < c) || ((r[ i ] = b[ i ] - x) > ~x) ) c = 1;
						else c = 0;
					}
				}
				while( ++i < small.SizeBlocks );

				if( i != big.SizeBlocks )
				{
					bool stillFix = true;

					if( c == 1 )
					{
						do { r[ i ] = b[ i ] - 1; }
						while( (b[ i++ ] == 0) && (i < big.SizeBlocks) );

						if( i == big.SizeBlocks ) stillFix = false;
					}

					if( stillFix )
					{
						do { r[ i ] = b[ i ]; }
						while( ++i < big.SizeBlocks );
					}
				}

				result.Normalize();

				return result;
			}

			public static void MinusEq( BigInteger big, BigInteger small )
			{
				UInt32[] b = big.DataBlocks, s = small.DataBlocks;
				UInt32 c = 0;
				Int32 i = 0;

				do
				{
					UInt32 x = s[ i ];

					unchecked
					{
						if( ((x += c) < c) || ((b[ i ] -= x) > ~x) ) c = 1;
						else c = 0;
					}
				}
				while( ++i < small.SizeBlocks );

				if( i != big.SizeBlocks )
				{
					if( c == 1 )
					{
						do { b[ i ]--; } while( (b[ i++ ] == 0) && (i < big.SizeBlocks) );
					}
				}

				// Normalize size.
				while( (big.SizeBlocks > 0) && (big.DataBlocks[ big.SizeBlocks - 1 ] == 0) ) big.SizeBlocks--;

				// Check for 0.
				if( big.SizeBlocks == 0 ) big.SizeBlocks++;

			}

			public static void PlusEq( BigInteger bi1, BigInteger bi2 )
			{
				UInt32[] x, y;
				Int32 yMax, xMax, i = 0;
				bool flag = false;

				// x should be bigger
				if( bi1.SizeBlocks < bi2.SizeBlocks )
				{
					flag = true;
					x = bi2.DataBlocks;
					xMax = bi2.SizeBlocks;
					y = bi1.DataBlocks;
					yMax = bi1.SizeBlocks;
				}
				else
				{
					x = bi1.DataBlocks;
					xMax = bi1.SizeBlocks;
					y = bi2.DataBlocks;
					yMax = bi2.SizeBlocks;
				}

				UInt32[] r = bi1.DataBlocks;

				UInt64 sum = 0;

				// Add common parts of both numbers
				do
				{
					sum += ((UInt64)x[ i ]) + ((UInt64)y[ i ]);
					r[ i ] = unchecked( (UInt32)sum );
					sum >>= BlockBitCount;
				}
				while( ++i < yMax );

				// Copy remainder of longer number while carry propagation is required.
				bool carry = sum != 0;

				if( carry )
				{
					if( i < xMax )
					{
						do { carry = (r[ i ] = x[ i ] + 1) == 0; }
						while( (++i < xMax) && carry );
					}

					if( carry )
					{
						r[ i ] = 1;
						bi1.SizeBlocks = ++i;

						return;
					}
				}

				// Copy the rest
				if( flag && i < xMax - 1 )
				{
					do { r[ i ] = x[ i ]; }
					while( ++i < xMax );
				}

				bi1.SizeBlocks = xMax + 1;
				bi1.Normalize();
			}

			#endregion Addition / subtraction

			#region Compare

			/// <summary>
			/// Compares two BigInteger
			/// </summary>
			/// <param name="bi1">A BigInteger</param>
			/// <param name="bi2">A BigInteger</param>
			/// <returns>The sign of bi1 - bi2</returns>
			public static Signs Compare( BigInteger bi1, BigInteger bi2 )
			{
				//
				// Step 1. Compare the lengths.
				//

				Int32 l1 = bi1.SizeBlocks, l2 = bi2.SizeBlocks;

				while( (l1 > 0) && (bi1.DataBlocks[ l1 - 1 ] == 0) ) l1--;
				while( (l2 > 0) && (bi2.DataBlocks[ l2 - 1 ] == 0) ) l2--;

				if( (l1 == 0) && (l2 == 0) ) return Signs.Zero;

				if( l1 < l2 ) return Signs.Negative;
				else if( l1 > l2 ) return Signs.Positive;

				//
				// Step 2. Compare the bits.
				//

				Int32 pos = l1 - 1;

				while( (pos > 0) && (bi1.DataBlocks[ pos ] == bi2.DataBlocks[ pos ]) ) pos--;

				if( bi1.DataBlocks[ pos ] < bi2.DataBlocks[ pos ] ) return Signs.Negative;
				else if( bi1.DataBlocks[ pos ] > bi2.DataBlocks[ pos ] ) return Signs.Positive;
				else return Signs.Zero;
			}

			#endregion Compare

			#region Division

			#region Atomic block

			public static UInt32 BlockMod( BigInteger n, UInt32 d )
			{
				UInt64 r = 0;
				Int32 i = n.SizeBlocks;

				while( i-- > 0 )
				{
					r <<= BlockBitCount;
					r |= n.DataBlocks[ i ];
					r %= d;
				}

				return (UInt32)r;
			}

			public static BigInteger BlockDiv( BigInteger n, UInt32 d )
			{
				BigInteger ret = new BigInteger( n.SizeBlocks );

				UInt64 r = 0;
				Int32 i = n.SizeBlocks;

				while( i-- > 0 )
				{
					r <<= BlockBitCount;
					r |= n.DataBlocks[ i ];
					ret.DataBlocks[ i ] = (UInt32)(r / d);
					r %= d;
				}

				ret.Normalize();

				return ret;
			}

			public static BigInteger[] BlockDivMod( BigInteger n, UInt32 d )
			{
				BigInteger ret = new BigInteger( n.SizeBlocks );

				UInt64 r = 0;
				Int32 i = n.SizeBlocks;

				while( i-- > 0 )
				{
					r <<= BlockBitCount;
					r |= n.DataBlocks[ i ];
					ret.DataBlocks[ i ] = (UInt32)(r / d);
					r %= d;
				}

				ret.Normalize();

				BigInteger rem = (UInt32)r;

				return new BigInteger[] { ret, rem };
			}

			#endregion

			#region BigNum

			/// <summary>
			/// 
			/// </summary>
			/// <param name="bi1"></param>
			/// <param name="bi2"></param>
			/// <returns>Returns the quotient and the remainder.</returns>
			public static BigInteger[] MultiByteDivide( BigInteger bi1, BigInteger bi2 )
			{
				if( Compare( bi1, bi2 ) == Signs.Negative )
					return new BigInteger[ 2 ] { 0, new BigInteger( bi1 ) };

				bi1.Normalize();
				bi2.Normalize();

				if( bi2.SizeBlocks == 1 ) return BlockDivMod( bi1, bi2.DataBlocks[ 0 ] );

				Int32 remainderLen = bi1.SizeBlocks + 1;
				Int32 divisorLen = bi2.SizeBlocks + 1;

				UInt32 mask = 0x80000000;
				UInt32 val = bi2.DataBlocks[ bi2.SizeBlocks - 1 ];
				Int32 shift = 0;
				Int32 resultPos = bi1.SizeBlocks - bi2.SizeBlocks;

				while( (mask != 0) && ((val & mask) == 0) )
				{
					shift++;
					mask >>= 1;
				}

				BigInteger quot = new BigInteger( bi1.SizeBlocks - bi2.SizeBlocks + 1 );
				BigInteger rem = (bi1 << shift);

				UInt32[] remainder = rem.DataBlocks;

				bi2 = bi2 << shift;

				Int32 j = remainderLen - bi2.SizeBlocks;
				Int32 pos = remainderLen - 1;

				UInt32 firstDivisorByte = bi2.DataBlocks[ bi2.SizeBlocks - 1 ];
				UInt64 secondDivisorByte = bi2.DataBlocks[ bi2.SizeBlocks - 2 ];

				while( j > 0 )
				{
					UInt64 dividend = ((UInt64)remainder[ pos ] << BlockBitCount) + remainder[ pos - 1 ];

					UInt64 q_hat = dividend / firstDivisorByte;
					UInt64 r_hat = dividend % firstDivisorByte;

					while( true )
					{

						if( (q_hat == 0x100000000)
							|| ((q_hat * secondDivisorByte) > ((r_hat << BlockBitCount) + remainder[ pos - 2 ])) )
						{
							q_hat--;
							r_hat += firstDivisorByte;

							if( r_hat < 0x100000000 ) continue;
						}

						break;
					}

					// At this point, q_hat is either exact, or one too large (more likely to be exact), so we attempt to
					// multiply the divisor by q_hat, if we get a borrow, we just subtract one from q_hat and add the divisor
					// back.

					Int32 dPos = 0;
					Int32 nPos = 1 + pos - divisorLen;
					UInt64 mc = 0;
					UInt32 q_hat_uint32 = unchecked( (UInt32)q_hat );

					do
					{
						unchecked
						{
							mc += (UInt64)bi2.DataBlocks[ dPos ] * (UInt64)q_hat_uint32;
						}

						UInt32 t = remainder[ nPos ];
						remainder[ nPos ] = unchecked( remainder[ nPos ] - (UInt32)mc );

						mc >>= BlockBitCount;

						if( remainder[ nPos ] > t ) unchecked { mc++; }

						dPos++;
						nPos++;
					}
					while( dPos < divisorLen );

					nPos = 1 + pos - divisorLen;
					dPos = 0;

					// Overestimate
					if( mc != 0 )
					{
						q_hat_uint32--;
						UInt64 sum = 0;

						do
						{
							sum = unchecked( ((UInt64)remainder[ nPos ]) + ((UInt64)bi2.DataBlocks[ dPos ]) + sum );

							remainder[ nPos ] = unchecked( (UInt32)sum );

							sum >>= BlockBitCount;

							dPos++;
							nPos++;
						}
						while( dPos < divisorLen );
					}

					quot.DataBlocks[ resultPos-- ] = q_hat_uint32;

					pos--;
					j--;
				}

				quot.Normalize();
				rem.Normalize();
				BigInteger[] ret = new BigInteger[ 2 ] { quot, rem };

				if( shift != 0 ) ret[ 1 ] >>= shift;

				return ret;
			}

			#endregion

			#endregion Division

			#region Shift

			public static BigInteger LeftShift( BigInteger bi, Int32 n )
			{
				if( n == 0 ) return new BigInteger( bi, bi.SizeBlocks + 1 );

				Int32 w = n >> BlockPow2BitCount;
				n &= (1 << BlockPow2BitCount) - 1;

				BigInteger ret = new BigInteger( bi.SizeBlocks + 1 + w );

				Int32 i = 0, l = bi.SizeBlocks;
				if( n != 0 )
				{
					UInt32 carry = 0;

					while( i < l )
					{
						UInt32 x = bi.DataBlocks[ i ];
						ret.DataBlocks[ i + w ] = (x << n) | carry;

						carry = x >> (BlockBitCount - n);

						i++;
					}

					ret.DataBlocks[ i + w ] = carry;
				}
				else
				{
					while( i < l )
					{
						ret.DataBlocks[ i + w ] = bi.DataBlocks[ i ];
						i++;
					}
				}

				ret.Normalize();
				return ret;
			}

			public static BigInteger RightShift( BigInteger bi, Int32 n )
			{
				if( n == 0 ) return new BigInteger( bi );

				Int32 w = n >> BlockPow2BitCount;
				Int32 s = n & ((1 << BlockPow2BitCount) - 1);

				BigInteger ret = new BigInteger( bi.SizeBlocks - w + 1 );
				Int32 l = ret.DataBlocks.Length - 1;

				if( s != 0 )
				{
					UInt32 carry = 0;

					while( l-- > 0 )
					{
						UInt32 x = bi.DataBlocks[ l + w ];
						ret.DataBlocks[ l ] = (x >> n) | carry;
						carry = x << (BlockBitCount - n);
					}
				}
				else
				{
					while( l-- > 0 ) ret.DataBlocks[ l ] = bi.DataBlocks[ l + w ];
				}

				ret.Normalize();
				return ret;
			}

			#endregion Shift

			#region Multiply

			public static BigInteger MultiplyByBlock( BigInteger n, UInt32 f )
			{
				BigInteger ret = new BigInteger( n.SizeBlocks + 1 );

				Int32 i = 0;
				UInt64 c = 0;

				do
				{
					c += (UInt64)n.DataBlocks[ i ] * (UInt64)f;
					ret.DataBlocks[ i ] = unchecked( (UInt32)c );
					c >>= BlockBitCount;
				}
				while( ++i < n.SizeBlocks );

				ret.DataBlocks[ i ] = unchecked( (UInt32)c );

				ret.Normalize();
				return ret;

			}

			/// <summary>
			/// Multiplies the data in "x[ xOffset ... xOffset + xLen ]" by "y[ yOffset ... yOffset + yLen ]" and puts it into
			/// "d[ dOffset ... dOffset + xLen + yLen ]".
			/// </summary>
			/// <remarks>This code is unsafe! It is the caller's responsibility to make sure that it is safe to access
			/// "x[ xOffset ... xOffset + xLen ]", "y[ yOffset ... yOffset + yLen ]", and
			/// "d[ dOffset ... dOffset + xLen + yLen]".
			/// </remarks>
			public static void Multiply( UInt32[] x, Int32 xOffset, Int32 xLen, UInt32[] y, Int32 yOffset, Int32 yLen
				, UInt32[] d, Int32 dOffset )
			{
				unsafe
				{
					fixed( UInt32* xx = x, yy = y, dd = d )
					{
						UInt32* xP = xx + xOffset;
						UInt32* xE = xP + xLen;
						UInt32* yB = yy + yOffset;
						UInt32* yE = yB + yLen;
						UInt32* dB = dd + dOffset;

						for( ; xP < xE; xP++, dB++ )
						{
							if( *xP == 0 ) continue;

							UInt64 mcarry = 0;

							UInt32* dP = dB;
							for( UInt32* yP = yB; yP < yE; yP++, dP++ )
							{
								unchecked
								{
									mcarry += ((UInt64)(*xP) * (UInt64)(*yP)) + (*dP);
								}

								*dP = unchecked( (UInt32)mcarry );
								mcarry >>= BlockBitCount;
							}

							if( mcarry != 0 ) *dP = unchecked( (UInt32)mcarry );
						}
					}
				}
			}

			/// <summary>
			/// Multiplies the data in "x[ xOffset ... xOffset + xLen ]" by "y[ yOffset ... yOffset + yLen ]" and puts the low
			/// mod blocks into "d[ dOffset ... dOffset + mod ]".
			/// </summary>
			/// <remarks>This code is unsafe! It is the caller's responsibility to make sure that it is safe to access
			/// "x[ xOffset ... xOffset + xLen ]",
			/// "y[ yOffset ... yOffset + yLen ]", and "d[ dOffset ... dOffset + mod ]".
			/// </remarks>
			public static void MultiplyMod2p32pMod( UInt32[] x, Int32 xOffset, Int32 xLen, UInt32[] y, Int32 yOffest, Int32 yLen
				, UInt32[] d, Int32 dOffset, Int32 mod )
			{
				unsafe
				{
					fixed( UInt32* xx = x, yy = y, dd = d )
					{
						UInt32* xP = xx + xOffset;
						UInt32* xE = xP + xLen;
						UInt32* yB = yy + yOffest;
						UInt32* yE = yB + yLen;
						UInt32* dB = dd + dOffset;
						UInt32* dE = dB + mod;

						for( ; xP < xE; xP++, dB++ )
						{
							if( *xP == 0 ) continue;

							UInt64 mcarry = 0;

							UInt32* dP = dB;
							for( UInt32* yP = yB; yP < yE && dP < dE; yP++, dP++ )
							{
								unchecked
								{
									mcarry += ((UInt64)(*xP) * (UInt64)(*yP)) + (*dP);
								}

								*dP = unchecked( (UInt32)mcarry );
								mcarry >>= BlockBitCount;
							}

							if( mcarry != 0 && dP < dE ) *dP = unchecked( (UInt32)mcarry );
						}
					}
				}
			}

			public static void SquarePositive( BigInteger bi, ref UInt32[] workSpace )
			{
				UInt32[] t = workSpace;
				workSpace = bi.DataBlocks;
				UInt32[] d = bi.DataBlocks;
				Int32 dl = bi.SizeBlocks;
				bi.DataBlocks = t;

				unsafe
				{
					fixed( UInt32* dd = d, tt = t )
					{
						UInt32* ttE = tt + t.Length;

						// Clear the destination.
						for( UInt32* ttt = tt; ttt < ttE; ttt++ ) *ttt = 0;

						UInt32* dP = dd, tP = tt;

						for( Int32 i = 0; i < dl; i++, dP++ )
						{
							if( *dP == 0 ) continue;

							UInt64 mcarry = 0;
							UInt32 bi1val = *dP;

							UInt32* dP2 = dP + 1, tP2 = tP + 2 * i + 1;

							for( Int32 j = i + 1; j < dl; j++, tP2++, dP2++ )
							{
								// "k = i + j".
								unchecked
								{
									mcarry += ((UInt64)bi1val * (UInt64)(*dP2)) + *tP2;
								}

								*tP2 = unchecked( (UInt32)mcarry );
								mcarry >>= BlockBitCount;
							}

							if( mcarry != 0 ) *tP2 = unchecked( (UInt32)mcarry );
						}

						// Double "t". Inlined for speed.

						tP = tt;

						UInt32 carry = 0;
						while( tP < ttE )
						{
							UInt32 x = *tP;
							*tP = (x << 1) | carry;

							carry = x >> (BlockBitCount - 1);

							tP++;
						}

						if( carry != 0 ) *tP = carry;

						// Add in the diagonals.

						dP = dd;
						tP = tt;
						for( UInt32* dE = dP + dl; (dP < dE); dP++, tP++ )
						{
							UInt64 val = unchecked( (UInt64)(*dP) * (UInt64)(*dP) + *tP );

							*tP = unchecked( (UInt32)val );
							val >>= BlockBitCount;

							tP++;
							unchecked { *tP += (UInt32)val; }

							if( *tP < unchecked( (UInt32)val ) )
							{
								UInt32* tP3 = tP;

								// Account for the first carry.
								tP3++;
								unchecked { (*tP3)++; }

								// Keep adding until no carry.
								while( *(tP3++) == 0 ) unchecked { (*tP3)++; }
							}
						}

						bi.SizeBlocks <<= 1;

						// Normalize size.
						while( (tt[ bi.SizeBlocks - 1 ] == 0) && (bi.SizeBlocks > 1) )
							bi.SizeBlocks--;
					}
				}
			}

			#endregion Multiply

			#region Number Theory

			public static BigInteger Gcd( BigInteger a, BigInteger b )
			{
				BigInteger x = a;
				BigInteger y = b;

				BigInteger g = y;

				while( x.SizeBlocks > 1 )
				{
					g = x;
					x = y % x;
					y = g;

				}
				if( x == 0 ) return g;

				UInt32 yy = x.DataBlocks[ 0 ];
				UInt32 xx = y % yy;

				Int32 t = 0;

				while( ((xx | yy) & 1) == 0 )
				{
					xx >>= 1;
					yy >>= 1;
					t++;
				}

				while( xx != 0 )
				{
					while( (xx & 1) == 0 ) xx >>= 1;
					while( (yy & 1) == 0 ) yy >>= 1;

					if( xx >= yy ) xx = (xx - yy) >> 1;
					else yy = (yy - xx) >> 1;
				}

				return yy << t;
			}

			public static UInt32 ModInverse( BigInteger bi, UInt32 modulus )
			{
				UInt32 a = modulus, b = bi % modulus;
				UInt32 p0 = 0, p1 = 1;

				while( b != 0 )
				{
					if( b == 1 ) return p1;

					p0 += (a / b) * p1;
					a %= b;

					if( a == 0 ) break;

					if( a == 1 ) return modulus - p0;

					p1 += (b / a) * p0;
					b %= a;

				}

				return 0;
			}

			public static BigInteger ModInverse( BigInteger bi, BigInteger modulus )
			{
				if( modulus.SizeBlocks == 1 ) return ModInverse( bi, modulus.DataBlocks[ 0 ] );

				BigInteger[] p = { 0, 1 };
				BigInteger[] q = new BigInteger[ 2 ]; // Quotients.
				BigInteger[] r = { 0, 0 }; // Remainders.

				Int32 step = 0;

				BigInteger a = modulus;
				BigInteger b = bi;

				ModulusRing mr = new ModulusRing( modulus );

				while( b != 0 )
				{

					if( step > 1 )
					{

						BigInteger pval = mr.Difference( p[ 0 ], p[ 1 ] * q[ 0 ] );
						p[ 0 ] = p[ 1 ];
						p[ 1 ] = pval;
					}

					BigInteger[] divret = MultiByteDivide( a, b );

					q[ 0 ] = q[ 1 ];
					q[ 1 ] = divret[ 0 ];
					r[ 0 ] = r[ 1 ];
					r[ 1 ] = divret[ 1 ];
					a = b;
					b = divret[ 1 ];

					step++;
				}

				if( r[ 0 ] != 1 ) throw new ValidationException( "There is no inverse." );

				return mr.Difference( p[ 0 ], p[ 1 ] * q[ 0 ] );

			}

			#endregion Number Theory
		}

		private sealed class ModulusRing
		{
			BigInteger _Mod, _Constant;

			#region Public methods

			public ModulusRing( BigInteger modulus )
			{
				_Mod = modulus;

				// Calculate "constant = b ^ (2 * k) / m".
				Int32 i = _Mod.SizeBlocks << 1;

				_Constant = new BigInteger( i + 1 );
				_Constant.DataBlocks[ i ] = 1;

				_Constant = _Constant / _Mod;
			}

			public void BarrettReduction( BigInteger x )
			{
				BigInteger n = _Mod;
				Int32 k = n.SizeBlocks;
				Int32 kPlusOne = k + 1;
				Int32 kMinusOne = k - 1;

				// "x < mod", so nothing to do.
				if( x.SizeBlocks < k ) return;

				BigInteger q3;

				// Validate.
				ValidationUtils.Integer( x.DataBlocks.Length, x.SizeBlocks, int.MaxValue, "x.DataBlocks.Length" );

				// "q1 = x / b^ (k-1)".
				// "q2 = q1 * constant".
				// "q3 = q2 / b^ (k+1)", needs to be accessed with an offset of "kPlusOne".

				// TODO: Use the method from HAC p 604 to do this (14.45).
				q3 = new BigInteger( x.SizeBlocks - kMinusOne + _Constant.SizeBlocks );
				Kernel.Multiply( x.DataBlocks, kMinusOne, x.SizeBlocks - kMinusOne, _Constant.DataBlocks, 0
					, _Constant.SizeBlocks, q3.DataBlocks, 0 );

				// "r1 = x mod b^ (k+1)".Keep the lowest "(k+1)" blocks.

				Int32 lengthToCopy = (x.SizeBlocks > kPlusOne) ? kPlusOne : x.SizeBlocks;

				x.SizeBlocks = lengthToCopy;
				x.Normalize();

				// "r2 = (q3 * n) mod b ^ (k + 1)".Partial multiplication of "q3" and "n".

				BigInteger r2 = new BigInteger( kPlusOne );
				Kernel.MultiplyMod2p32pMod( q3.DataBlocks, kPlusOne, q3.SizeBlocks - kPlusOne, n.DataBlocks, 0, n.SizeBlocks
					, r2.DataBlocks, 0, kPlusOne );

				r2.Normalize();

				if( r2 <= x )
				{
					Kernel.MinusEq( x, r2 );
				}
				else
				{
					BigInteger val = new BigInteger( kPlusOne + 1 );
					val.DataBlocks[ kPlusOne ] = 1;

					Kernel.MinusEq( val, r2 );
					Kernel.PlusEq( x, val );
				}

				while( x >= n ) Kernel.MinusEq( x, n );
			}

			public BigInteger Difference( BigInteger a, BigInteger b )
			{
				Signs sign = Kernel.Compare( a, b );
				BigInteger diff;

				switch( sign )
				{
					case Signs.Zero:
						return 0;
					case Signs.Positive:
						diff = a - b;
						break;
					case Signs.Negative:
						diff = b - a;
						break;
					default:
						throw ValidationUtils.Unhandled( sign, "sign" );
				}

				if( diff >= _Mod )
				{
					if( diff.SizeBlocks >= _Mod.SizeBlocks << 1 ) diff %= _Mod;
					else BarrettReduction( diff );
				}

				if( sign == Signs.Negative ) diff = _Mod - diff;

				return diff;
			}

			public BigInteger Pow( BigInteger b, BigInteger exp )
			{
				if( (_Mod.DataBlocks[ 0 ] & 1) == 1 ) return OddPow( b, exp );
				else return EvenPow( b, exp );
			}

			public BigInteger EvenPow( BigInteger b, BigInteger exp )
			{
				BigInteger resultNum = new BigInteger( 1, _Mod.SizeBlocks << 1 );

				// Ensures "(tempNum * tempNum) < b ^ (2 * k)".
				BigInteger tempNum = new BigInteger( b % _Mod, _Mod.SizeBlocks << 1 );

				Int32 totalBits = exp.GetBitCount();

				UInt32[] workSpace = new UInt32[ _Mod.SizeBlocks << 1 ];

				// Perform squaring and multiply exponentiation.
				for( Int32 pos = 0; pos < totalBits; pos++ )
				{
					if( exp.TestBit( pos ) )
					{
						Array.Clear( workSpace, 0, workSpace.Length );
						Kernel.Multiply( resultNum.DataBlocks, 0, resultNum.SizeBlocks, tempNum.DataBlocks, 0
							, tempNum.SizeBlocks, workSpace, 0 );

						resultNum.SizeBlocks += tempNum.SizeBlocks;

						UInt32[] t = workSpace;
						workSpace = resultNum.DataBlocks;
						resultNum.DataBlocks = t;

						BarrettReduction( resultNum );
					}

					Kernel.SquarePositive( tempNum, ref workSpace );
					BarrettReduction( tempNum );

					if( tempNum == 1 ) return resultNum;
				}

				return resultNum;
			}

			private BigInteger OddPow( BigInteger b, BigInteger exp )
			{
				BigInteger resultNum = new BigInteger( Montgomery.Toon( 1, _Mod ), _Mod.SizeBlocks << 1 );

				// Ensures "(tempNum * tempNum) < b ^ (2 * k)".
				BigInteger tempNum = new BigInteger( Montgomery.Toon( b, _Mod ), _Mod.SizeBlocks << 1 );

				UInt32 _Prime = Montgomery.Inverse( _Mod.DataBlocks[ 0 ] );
				Int32 totalBits = exp.GetBitCount();

				UInt32[] workSpace = new UInt32[ _Mod.SizeBlocks << 1 ];

				// Perform squaring and multiply exponentiation.
				for( Int32 pos = 0; pos < totalBits; pos++ )
				{
					if( exp.TestBit( pos ) )
					{

						Array.Clear( workSpace, 0, workSpace.Length );
						Kernel.Multiply( resultNum.DataBlocks, 0, resultNum.SizeBlocks, tempNum.DataBlocks, 0
							, tempNum.SizeBlocks, workSpace, 0 );

						resultNum.SizeBlocks += tempNum.SizeBlocks;

						UInt32[] t = workSpace;
						workSpace = resultNum.DataBlocks;
						resultNum.DataBlocks = t;

						Montgomery.Reduce( resultNum, _Mod, _Prime );
					}

					// "tempNum" is required in the last loop.
					if( pos < totalBits - 1 )
					{
						Kernel.SquarePositive( tempNum, ref workSpace );
						Montgomery.Reduce( tempNum, _Mod, _Prime );
					}
				}

				Montgomery.Reduce( resultNum, _Mod, _Prime );
				return resultNum;
			}

			#region Pow small base

			public BigInteger PowSmallBase( UInt32 b, BigInteger exp )
			{
				if( (_Mod.DataBlocks[ 0 ] & 1) == 1 ) return OddPowSmallBase( b, exp );
				else return EvenPowSmallBase( b, exp );
			}

			private BigInteger OddPowSmallBase( UInt32 b, BigInteger exp )
			{
				exp.Normalize();
				UInt32[] workspace = new UInt32[ (_Mod.SizeBlocks << 1) + 1 ];

				BigInteger resultNum = Montgomery.Toon( b, _Mod );
				resultNum = new BigInteger( resultNum, (_Mod.SizeBlocks << 1) + 1 );

				UInt32 _Prime = Montgomery.Inverse( _Mod.DataBlocks[ 0 ] );

				Int32 pos = exp.GetBitCount() - 2;
				if( pos < 0 ) return Montgomery.Reduce( resultNum, _Mod, _Prime );

				// We know that the first iteration will make the val "b".

				do
				{
					// r = r ^ 2 % m
					Kernel.SquarePositive( resultNum, ref workspace );
					resultNum = Montgomery.Reduce( resultNum, _Mod, _Prime );

					if( exp.TestBit( pos ) )
					{
						// r = r * b % m
						unsafe
						{
							fixed( UInt32* u = resultNum.DataBlocks )
							{
								Int32 i = 0;
								UInt64 mc = 0;

								do
								{
									mc += (UInt64)u[ i ] * (UInt64)b;
									u[ i ] = unchecked( (UInt32)mc );
									mc >>= BlockBitCount;
								}
								while( ++i < resultNum.SizeBlocks );

								if( resultNum.SizeBlocks < _Mod.SizeBlocks )
								{
									if( mc != 0 )
									{
										u[ i ] = unchecked( (UInt32)mc );
										resultNum.SizeBlocks++;

										while( resultNum >= _Mod ) Kernel.MinusEq( resultNum, _Mod );
									}
								}
								else if( mc != 0 )
								{
									// First, we estimate the quotient by dividing the first part of each of the numbers. Then
									// we correct this, if necessary, with a subtraction.

									UInt32 cc = unchecked( (UInt32)mc );

									// We would rather have this estimate overshoot, so we add one to the divisor.
									UInt32 divEstimate;
									if( _Mod.DataBlocks[ _Mod.SizeBlocks - 1 ] < UInt32.MaxValue )
									{
										divEstimate = (UInt32)
											(
												(((UInt64)cc << BlockBitCount) | u[ i - 1 ])
												/ (_Mod.DataBlocks[ _Mod.SizeBlocks - 1 ] + 1)
											);
									}
									else
									{
										// Guess, but don't divide by 0.
										divEstimate = (UInt32)
											(
												(((UInt64)cc << BlockBitCount) | u[ i - 1 ])
												/ (_Mod.DataBlocks[ _Mod.SizeBlocks - 1 ])
											);
									}

									i = 0;
									mc = 0;

									do
									{
										mc += (UInt64)_Mod.DataBlocks[ i ] * (UInt64)divEstimate;
										UInt32 t = u[ i ];

										u[ i ] = unchecked( u[ i ] - (UInt32)mc );
										mc >>= BlockBitCount;

										if( u[ i ] > t ) unchecked { mc++; }

										i++;
									}
									while( i < resultNum.SizeBlocks );

									cc = unchecked( cc - (UInt32)mc );

									if( cc != 0 )
									{
										Int32 j = 0;
										UInt32 sc = 0;
										UInt32[] s = _Mod.DataBlocks;

										do
										{
											UInt32 a = s[ j ];

											if( ((a += sc) < sc) | ((u[ j ] -= a) > ~a) ) sc = 1;
											else sc = 0;

											j++;
										}
										while( j < resultNum.SizeBlocks );
									}

									while( resultNum >= _Mod ) Kernel.MinusEq( resultNum, _Mod );
								}
								else
								{
									while( resultNum >= _Mod ) Kernel.MinusEq( resultNum, _Mod );
								}
							}
						}
					}
				}
				while( pos-- > 0 );

				return Montgomery.Reduce( resultNum, _Mod, _Prime );
			}

			private BigInteger EvenPowSmallBase( UInt32 b, BigInteger exp )
			{
				exp.Normalize();
				UInt32[] workspace = new UInt32[ (_Mod.SizeBlocks << 1) + 1 ];
				BigInteger resultNum = new BigInteger( b, (_Mod.SizeBlocks << 1) + 1 );

				Int32 pos = exp.GetBitCount() - 2;
				if( pos < 0 )
				{
					if( resultNum.SizeBlocks >= _Mod.SizeBlocks ) BarrettReduction( resultNum );

					return resultNum;
				}

				// The first iteration will make the val "b".

				do
				{
					// r = r ^ 2 % m
					Kernel.SquarePositive( resultNum, ref workspace );
					if( resultNum.SizeBlocks >= _Mod.SizeBlocks ) BarrettReduction( resultNum );

					if( exp.TestBit( pos ) )
					{
						// r = r * b % m
						unsafe
						{
							fixed( UInt32* u = resultNum.DataBlocks )
							{
								Int32 i = 0;
								UInt64 mc = 0;

								do
								{
									mc += (UInt64)u[ i ] * (UInt64)b;
									u[ i ] = unchecked( (UInt32)mc );
									mc >>= BlockBitCount;
								}
								while( ++i < resultNum.SizeBlocks );

								if( resultNum.SizeBlocks < _Mod.SizeBlocks )
								{
									if( mc != 0 )
									{
										u[ i ] = unchecked( (UInt32)mc );
										resultNum.SizeBlocks++;

										while( resultNum >= _Mod ) Kernel.MinusEq( resultNum, _Mod );
									}
								}
								else if( mc != 0 )
								{
									// First, we estimate the quotient by dividing the first part of each of the numbers. Then
									// we correct this, if necessary, with a subtraction.

									UInt32 cc = unchecked( (UInt32)mc );

									// We would rather have this estimate overshoot, so we add one to the divisor.
									UInt32 divEstimate = (UInt32)
										(
											(((UInt64)cc << BlockBitCount) | u[ i - 1 ])
											/ (_Mod.DataBlocks[ _Mod.SizeBlocks - 1 ] + 1)
										);

									i = 0;
									mc = 0;

									do
									{
										mc += (UInt64)_Mod.DataBlocks[ i ] * (UInt64)divEstimate;
										UInt32 t = u[ i ];

										u[ i ] = unchecked( u[ i ] - (UInt32)mc );
										mc >>= BlockBitCount;

										if( u[ i ] > t ) unchecked { mc++; }

										i++;
									}
									while( i < resultNum.SizeBlocks );

									cc = unchecked( cc - (UInt32)mc );

									if( cc != 0 )
									{
										Int32 j = 0;
										UInt32 sc = 0;
										UInt32[] s = _Mod.DataBlocks;

										do
										{
											UInt32 a = s[ j ];

											if( ((a += sc) < sc) | ((u[ j ] -= a) > ~a) ) sc = 1;
											else sc = 0;

											j++;
										}
										while( j < resultNum.SizeBlocks );
									}

									while( resultNum >= _Mod ) Kernel.MinusEq( resultNum, _Mod );
								}
								else
								{
									while( resultNum >= _Mod ) Kernel.MinusEq( resultNum, _Mod );
								}
							}
						}
					}
				}
				while( pos-- > 0 );

				return resultNum;
			}

			#endregion Pow small base

			#endregion Public methods
		}

		private static class Montgomery
		{
			#region Public methods

			public static UInt32 Inverse( UInt32 n )
			{
				UInt32 y = n, z;

				unchecked
				{
					while( (z = n * y) != 1 ) y *= 2 - z;
				}

				return unchecked( (UInt32)(-y) );
			}

			public static BigInteger Toon( BigInteger n, BigInteger m )
			{
				n.Normalize();
				m.Normalize();

				n <<= m.SizeBlocks * BlockBitCount;
				n %= m;

				return n;
			}

			public static BigInteger Reduce( BigInteger n, BigInteger m, UInt32 prime )
			{
				unsafe
				{
					fixed( UInt32* a = n.DataBlocks, mm = m.DataBlocks )
					{
						for( Int32 i = 0; i < m.SizeBlocks; i++ )
						{
							// The mod here is taken care of by the CPU, since the multiply will overflow.
							UInt32 u_i = unchecked( a[ 0 ] * prime );

							// n += u_i * m n >>= BlockBitCount

							// pos = position in mod sPos = the source of bits from a dPos = destination for bits
							UInt32* mPos = mm, sPos = a, dPos = a;

							UInt64 c = (UInt64)u_i * ((UInt64)(*(mPos++))) + *(sPos++);
							c >>= BlockBitCount;

							Int32 j = 1;

							// Multiply and add.
							for( ; j < m.SizeBlocks; j++ )
							{
								c += (UInt64)u_i * (UInt64)(*(mPos++)) + *(sPos++);
								*(dPos++) = unchecked( (UInt32)c );
								c >>= BlockBitCount;
							}

							// Account for carry.
							for( ; j < n.SizeBlocks; j++ )
							{
								c += *(sPos++);
								*(dPos++) = unchecked( (UInt32)c );
								c >>= BlockBitCount;

								if( c == 0 ) { j++; break; }
							}

							// Copy the rest.
							for( ; j < n.SizeBlocks; j++ )
							{
								*(dPos++) = *(sPos++);
							}

							*(dPos++) = unchecked( (UInt32)c );
						}

						while( (n.SizeBlocks > 1) && (a[ n.SizeBlocks - 1 ] == 0) ) n.SizeBlocks--;
					}
				}

				if( n >= m ) Kernel.MinusEq( n, m );

				return n;
			}

			#endregion Public methods
		}

		#endregion Private classes

		#region Public classes

		/// <summary>
		/// Tests for verifying is big integers are strong pseudo-primes.
		/// </summary>
		public static class PrimalityTests
		{
			#region Public methods

			/// <summary>
			/// Get the number of rounds to use when testing if a big integer is a strong pseudo-prime.
			/// </summary>
			/// <param name="bc">The number of bits of the big integer to test if it is a strong pseudo-prime. This must be at
			/// least 1000.</param>
			/// <returns>Returns the number of rounds which should be used for testing.</returns>
			public static int GetRoundCount( int bc )
			{
				// This is a practical limitation, not a technical one.
				ValidationUtils.Integer( bc, 1000, int.MaxValue, "bc" );

				// The following values are higher (which is more secure) than but similar with those from HAC 4.49. The more
				// bits the number to test has, the less rounds there need to be.
				if( bc <= 1024 ) return 6;
				else if( bc <= 1536 ) return 5;
				else return 4; // For any other number of bits.
			}

			/// <summary>
			/// Get the number of rounds to use when testing if a big integer is a strong pseudo-prime.
			/// </summary>
			/// <param name="bi">Big integer to test if it is a strong pseudo-prime. The bit count of this object must be at
			/// least 1000.</param>
			/// <returns>Returns the number of rounds which should be used for testing.</returns>
			public static Int32 GetRoundCount( BigInteger bi )
			{
				ValidationUtils.ObjectHasContent( bi, "bi" );

				return GetRoundCount( bi.GetBitCount() );
			}

			/// <summary>
			/// Returns the tipical probability for the Miller-Rabin test to return prime for a non-prime number.
			/// </summary>
			/// <param name="bc">The number of bits of the big integer which is supposed to be prime.</param>
			/// <param name="r">The number of rounds of the Miller-Rabin test.</param>
			/// <returns>Returns the tipical probability for the Miller-Rabin test to return prime for a non-prime number.
			/// </returns>
			public static double MillerRabinProbability( int bc, int r )
			{
				// HAC 4.48.
				//
				// * "k" is the number of bits of the big integer which is supposed to be prime, "k >= 21".
				// * "t" is number of rounds of the Miller-Rabin test, "t = [3 ... k / 9]".

				// The size of the public modulus of the asymmetric key is "2 * bc".

				ValidationUtils.Integer( bc, 21, int.MaxValue, "bc" );
				ValidationUtils.Integer( r, 3, bc / 9, "r" );

				// The probability of a non-prime passing the test as being prime.
				double p =
					System.Math.Pow( bc, 3 / 2 )
					* System.Math.Pow( 2, r )
					* System.Math.Pow( r, -1 / 2 )
					* System.Math.Pow( 4, 2 - System.Math.Sqrt( bc * r ) );

				return p;
			}

			/// <summary>
			/// Probabilistic prime test based on Miller-Rabin's test.
			/// </summary>
			/// <param name="bi">The number to test. The bit count of this object must be at least 1000.</param>
			/// <returns>Returns true if the specified number is a strong pseudo-prime to randomly chosen bases. It returns
			/// false if the specified number is definitely NOT prime.</returns>
			public static bool MillerRabin( BigInteger bi )
			{
				ValidationUtils.ObjectHasContent( bi, "bi" );

				// This is a practical limitation, not a technical one.
				ValidationUtils.Integer( bi.GetBitCount(), 1000, int.MaxValue, "bi.GetBitCount()" );

				Int32 roundCount = GetRoundCount( bi );

				// The number of primality test rounds. The probability for this method to return true for a number which is not
				// prime is maximum "MillerRabinProbability".
				//
				// If a non-prime is declared prime, the verification of the generated asymmetric key may also pass.
				//
				// The problem would be that the key would become more exposed to factoring, and thus the private key could be
				// found much easier.
				//
				// Still, this doesn't mean that a private key could immediately be computed, but that it is easier.

				// Calculate values of "s" and "t".
				BigInteger p_sub1 = bi - 1;
				Int32 s = p_sub1.LowestSetBit();
				BigInteger t = p_sub1 >> s;

				Int32 bits = bi.GetBitCount();
				ModulusRing mr = new ModulusRing( bi );

				// Applying optimization from HAC section 4.50, radix 2 is not a random one, but is fast.
				BigInteger b = mr.PowSmallBase( 2, t );
				if( b != 1 )
				{
					bool result = false;
					for( Int32 j = 0; j < s; j++ )
					{
						if( b == p_sub1 )
						{
							// "a ^ ((2 ^ j) * t) mod p = p - 1" where "0 <= j <= s - 1".
							result = true;
							break;
						}

						b = (b * b) % bi;
					}

					if( !result ) return false;
				}

				// Start from the index 1 because the index 0 was the optimization where "a == 2".
				for( Int32 round = 1; round < roundCount; round++ )
				{
					BigInteger a;

					while( true )
					{
						// Generate "a < n".
						a = GenerateRandom( bits );

						if( (a > 2) && (a < bi) ) break;
					}

					if( a.Gcd( bi ) != 1 ) return false;

					b = mr.Pow( a, t );

					if( b == 1 ) continue; // "a ^ t mod p = 1".

					bool result = false;
					for( Int32 j = 0; j < s; j++ )
					{
						if( b == p_sub1 )
						{
							// "a ^ ((2 ^ j) * t) mod p = p - 1" where "0 <= j <= s - 1".
							result = true;
							break;
						}

						b = (b * b) % bi;
					}

					if( !result ) return false;
				}

				return true;
			}

			#endregion Public methods
		}

		#endregion Public classes
	}
}
