////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto.Hash;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.Crypto.Asymmetric
{
	/// <summary>
	/// Contract for an asymmetric cryptography padding.
	/// </summary>
	public abstract class AsymmPaddingAlg : IImmutableAlgorithm, IExpirableAlgorithm
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AsymmPaddingAlg()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="textEncoding">Text encoding.</param>
		/// <param name="hashAlg">Hash algorithm.</param>
		public AsymmPaddingAlg( TextEncoding textEncoding, HashAlg hashAlg )
		{
			TextEncoding = textEncoding;
			HashAlg = hashAlg;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public abstract DateTime ExpirationTime { get; }

		#endregion Public non-serializable properties

		#region Public serializable properties

		private TextEncoding _TextEncoding;
		/// <summary>
		/// Text encoding. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public TextEncoding TextEncoding
		{
			get { return _TextEncoding; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "TextEncoding" );

				_TextEncoding = value;
			}
		}

		private HashAlg _HashAlg;
		/// <summary>
		/// Hash algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public HashAlg HashAlg
		{
			get { return _HashAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "HashAlg" );

				_HashAlg = value;
			}
		}

		#endregion Public serializable properties
	}
}
