////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto.Hash;
using Vela.Kernel;

namespace Vela.Crypto.Asymmetric
{
	/// <summary>
	/// "VSH" ("Valuable data, Salt, Hash") asymmetric cryptography padding.
	/// </summary>
	/// <remarks>
	/// <para>This is OAEP with binary encoding (not ASN1), with included algorithm names (instead of numeric identifiers), and
	/// an extra hash. The implementation is custom because more information had to be included.</para>
	/// 
	/// <para>The data block which is signed or encrypted is formatted as follows:</para>
	/// <list type="bullet">
	///		<item>The value <see cref="NonZeroByte"/>.</item>
	///		<item>Name of the asymmetric padding algorithm.</item>
	///		<item>Name of the asymmetric cryptographic algorithm.</item>
	///		<item>Name of the hash algorithm used for hashing.</item>
	///		<item>Filling size (in bytes); this occupies <see cref="SizeOfDataSize"/> bytes.</item>
	///		<item>Filling (= random data) whose minimum size is 1. This fills the empty space of the block.</item>
	///		<item>Valuable data size (in bytes); this occupies <see cref="SizeOfDataSize"/> bytes.</item>
	///		<item>Valuable data.</item>
	///		<item>Salt (= random data); this occupies the same space as a hash.</item>
	///		<item>Hash of the entire formatted data block (where the space for the hash itself is 0).</item>
	///		<item>The value <see cref="NonZeroByte"/>.</item>
	/// </list>
	/// 
	/// <para>Before the hash (of the formatted data block) is computed and put in the block, the block from its beginning up to
	/// (but excluding) the salt is XORed byte by byte with a mask generated (through hashing) from the salt, then the salt is
	/// XORed byte by byte with a mask generated (through hashing) from the previous mask result. The <see cref="NonZeroByte"/>
	/// byte from the beginning is not XORed.</para>
	/// 
	/// <para>The hash of the entire formatted data block (where the space for the hash itself is 0) is computed and put in the
	/// block.</para>
	/// 
	/// <para>For signing, the valuable data should be the hash of the data to sign.</para>
	/// 
	/// <para>For encryption, the valuable data should be the session key which was used to encrypt the actual hidden data.
	/// </para>
	/// </remarks>
	public abstract class Vsh : AsymmPaddingAlg
	{
		#region Public constants

		/// <summary>
		/// The first byte of the data to process must be non-zero (because to the "Zero-byte issue" from the
		/// <see cref="Vela.Crypto.Asymmetric.Math.BigInteger"/> class).
		/// </summary>
		public const byte NonZeroByte = 0x7F;

		/// <summary>
		/// The minimum size of the salt.
		/// </summary>
		public const int SaltSizeMin = 32;

		/// <summary>
		/// The size (in bytes) of a data size.
		/// </summary>
		public const int SizeOfDataSize = sizeof( int );

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Vsh()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="textEncoding">Text encoding.</param>
		/// <param name="hashAlg">Hash algorithm.</param>
		public Vsh( TextEncoding textEncoding, HashAlg hashAlg )
			: base( textEncoding, hashAlg )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return HashAlg.ExpirationTime; }
		}

		#endregion Public non-serializable properties

		#region Private static methods

		/// <summary>
		/// Generates a mask with the specified size.
		/// </summary>
		/// <param name="hashAlg">Hash algorithm with which to generate the mask.</param>
		/// <param name="salt">Salt for the generation process.</param>
		/// <param name="maskSize">Mask size.</param>
		/// <returns>Returns the generated mask.</returns>
		private static byte[] GenerateMask( HashAlg hashAlg, byte[] salt, int maskSize )
		{
			ValidationUtils.ObjectHasContent( hashAlg, "hashAlg" );
			ValidationUtils.ArrayHasContent( salt, "salt" );
			ValidationUtils.Integer( maskSize, 1, int.MaxValue, "maskSize" );

			int saltSize = salt.Length;
			int hashSize = hashAlg.AlgSizeBytes;
			int hashBlocks = MathUtils.RoundUp( maskSize, hashSize );

			byte[] t = new byte[ hashSize * hashBlocks ];
			byte[] buffer = new byte[ saltSize + SizeOfDataSize ];

			int pos = 0;
			for( int i = 0; i < hashBlocks; i++ )
			{
				// t = t || Hash( salt || i ).

				// The counter should never be (used) 0.
				byte[] iBuffer = BitConverter.GetBytes( hashBlocks - i );

				Buffer.BlockCopy( salt, 0, buffer, 0, saltSize );
				Buffer.BlockCopy( iBuffer, 0, buffer, saltSize, iBuffer.Length );

				byte[] output = hashAlg.Compute( buffer );
				Buffer.BlockCopy( output, 0, t, pos, hashSize );
				pos += output.Length;
			}

			byte[] mask = new byte[ maskSize ];
			Buffer.BlockCopy( t, 0, mask, 0, maskSize );

			return mask;
		}

		/// <summary>
		/// Masks the specified formatted data block with the specified salt. The margins bytes are not masked.
		/// </summary>
		/// <param name="block">Formatted data block to mask.</param>
		/// <param name="blockOffset">The offset from which to start masking the formatted data block.</param>
		/// <param name="processSize">The number of bytes to mask from the formatted data block.</param>
		/// <param name="salt">Salt with which to mask the formatted data block.</param>
		private static void MaskBlock( byte[] block, int blockOffset, int processSize, byte[] salt )
		{
			ValidationUtils.ArrayHasContent( block, "block" );
			ValidationUtils.Integer( blockOffset, 1, block.Length - 2, "blockOffset" );
			ValidationUtils.Integer( processSize, 1, block.Length - 2, "processSize" );
			ValidationUtils.Array( salt, SaltSizeMin, block.Length - 2, "salt" );

			int blockSize = block.Length;
			int blockEndProcess = blockOffset + processSize - 1;

			for( int i = blockOffset; i <= blockEndProcess; i++ )
				block[ i ] ^= salt[ i - blockOffset ];
		}

		/// <summary>
		/// Returns the size of the filling.
		/// </summary>
		/// <param name="asymmPaddingNameSize">The size of the name of the asymmetric padding algorithm.</param>
		/// <param name="hashAlgNameSize">The size of the name of the hash algorithm which is used to hash the entire formatted
		/// data block.</param>
		/// <param name="asymmAlgNameSize">The size of the name of the asymmetric algorithm used for encryption or
		/// signing.</param>
		/// <param name="valuableDataSize">The size of the valuable data.</param>
		/// <param name="hashAlgSizeBytes">The size of the hash algorithm which is used to hash the entire formatted data block.
		/// </param>
		/// <param name="maxProcessableSize">The maximum processable size of the asymmetric algorithm.</param>
		/// <returns>Returns the size of the filling.</returns>
		/// <remarks>
		/// <para>This must be at least 1.</para>
		/// </remarks>
		private static int GetFillingSize( int asymmPaddingNameSize, int hashAlgNameSize, int asymmAlgNameSize
			, int valuableDataSize, int hashAlgSizeBytes, int maxProcessableSize )
		{
			int storedSize =
				1 // The first byte.
				+ asymmPaddingNameSize
				+ hashAlgNameSize
				+ asymmAlgNameSize
				+ SizeOfDataSize // Size of the filling.
				// Filling
				+ SizeOfDataSize // Size of the valuable data.
				+ valuableDataSize
				+ hashAlgSizeBytes // Salt.
				+ hashAlgSizeBytes // Hash of the entire formatted data block.
				+ 1; // The last byte.

			int fillingSize = maxProcessableSize - storedSize;

			if( fillingSize < 1 )
				throw new ValidationException( "During VSH padding, the stored size has overflown." );

			return fillingSize;
		}

		/// <summary>
		/// Prepares the formatted data block which may be asymmetrically signed or encrypted.
		/// </summary>
		/// <param name="textEncoding"></param>
		/// <param name="asymmPaddingAlg">The asymmetric padding algorithm.</param>
		/// <param name="hashAlg">The hash algorithm.</param>
		/// <param name="asymmAlg">The asymmetric algorithm.</param>
		/// <param name="valuableData">The valuable data to be signed or encrypted. This should be something small, like a hash.
		/// </param>
		/// <returns>Returns the generated block.</returns>
		private static byte[] SetupBlock( TextEncoding textEncoding, AsymmPaddingAlg asymmPaddingAlg, HashAlg hashAlg
			, AsymmAlg asymmAlg, byte[] valuableData )
		{
			ValidationUtils.ObjectHasContent( textEncoding, "textEncoding" );
			ValidationUtils.ObjectHasContent( asymmPaddingAlg, "asymmPaddingAlg" );
			ValidationUtils.ObjectHasContent( hashAlg, "hashAlg" );
			ValidationUtils.ObjectHasContent( asymmAlg, "asymmAlg" );
			ValidationUtils.ArrayHasContent( valuableData, "valuableData" );

			byte[] asymmPaddingName = textEncoding.TextToArray( RegistrableAlgorithmUtils.GetAlgorithmName( asymmPaddingAlg )
				, true );
			byte[] hashAlgName = textEncoding.TextToArray( RegistrableAlgorithmUtils.GetAlgorithmName( hashAlg ), true );
			byte[] asymmAlgName = textEncoding.TextToArray( RegistrableAlgorithmUtils.GetAlgorithmName( asymmAlg ), true );

			// Generate the filling.
			int fillingSize = GetFillingSize( asymmPaddingName.Length, hashAlgName.Length, asymmAlgName.Length
				, valuableData.Length, hashAlg.AlgSizeBytes, asymmAlg.MaxProcessableSizeBytes );
			byte[] filling = EntropyPool.GetBytes( fillingSize );

			// Generate the salt.
			byte[] salt = EntropyPool.GetBytes( hashAlg.AlgSizeBytes );

			byte[] block = new byte[ asymmAlg.MaxProcessableSizeBytes ];

			int blockPos = 0;

			// The first byte of the block must be non-zero.
			block[ blockPos ] = NonZeroByte;
			blockPos++;

			// The name of the asymmetric padding algorithm.
			Buffer.BlockCopy( asymmPaddingName, 0, block, blockPos, asymmPaddingName.Length );
			blockPos += asymmPaddingName.Length;

			// The name of the hash algorithm.
			Buffer.BlockCopy( hashAlgName, 0, block, blockPos, hashAlgName.Length );
			blockPos += hashAlgName.Length;

			// The name of the asymmetric algorithm.
			Buffer.BlockCopy( asymmAlgName, 0, block, blockPos, asymmAlgName.Length );
			blockPos += asymmAlgName.Length;

			// The size of the filling.
			byte[] fillingSizeBuffer = BitConverter.GetBytes( fillingSize );
			Buffer.BlockCopy( fillingSizeBuffer, 0, block, blockPos, fillingSizeBuffer.Length );
			blockPos += SizeOfDataSize;

			// The filling.
			Buffer.BlockCopy( filling, 0, block, blockPos, filling.Length );
			blockPos += filling.Length;

			// The size of the valuable data.
			byte[] valuableDataSizeBuffer = BitConverter.GetBytes( valuableData.Length );
			Buffer.BlockCopy( valuableDataSizeBuffer, 0, block, blockPos, valuableDataSizeBuffer.Length );
			blockPos += SizeOfDataSize;

			// The valuable data.
			Buffer.BlockCopy( valuableData, 0, block, blockPos, valuableData.Length );
			blockPos += valuableData.Length;

			// The salt.
			Buffer.BlockCopy( salt, 0, block, blockPos, salt.Length );
			int saltPos = blockPos;
			blockPos += salt.Length;

			// Clear the space for the hash of the data (in order to compute the hash of the entire block).
			int hashPos = blockPos;
			Array.Clear( block, blockPos, hashAlg.AlgSizeBytes );
			blockPos += hashAlg.AlgSizeBytes;

			// The last byte of the block must be non-zero.
			block[ blockPos ] = NonZeroByte;

			// Mask the block from its beginning up to the salt (but excluding it).
			int maskSize = saltPos - 1 /* Non-zero byte */;
			byte[] mask = GenerateMask( hashAlg, salt, maskSize );
			MaskBlock( block, 1, maskSize, mask );

			// Mask the salt.
			byte[] maskedBuffer = new byte[ maskSize ];
			Buffer.BlockCopy( block, 1, maskedBuffer, 0, maskedBuffer.Length );
			mask = GenerateMask( hashAlg, maskedBuffer, salt.Length );
			MaskBlock( block, saltPos, salt.Length, mask );

			// The hash of the block.
			byte[] hashVal = hashAlg.Compute( block );
			Buffer.BlockCopy( hashVal, 0, block, hashPos, hashVal.Length );

			return block;
		}

		/// <summary>
		/// Compares the hash extracted from the specified formatted data block with the block's computed hash, and throws an
		/// exception if they don't match.
		/// </summary>
		/// <param name="block">Block from which to extract the hash, and whose hash to compute.</param>
		/// <param name="hashPos">Position in the block from where to extract the existing hash.</param>
		/// <param name="hashAlg">The hash algorithm which is to be used to hash the entire block.</param>
		/// <param name="errorDetail">Message to include in the description of the thrown exceptions. This may be null or empty.
		/// </param>
		private static void VerifyBlockHash( byte[] block, int hashPos, HashAlg hashAlg, string errorDetail )
		{
			ValidationUtils.ArrayHasContent( block, "block" );
			ValidationUtils.Integer( hashPos, 1, block.Length - 2, "hashPos" );
			ValidationUtils.ObjectHasContent( hashAlg, "hashAlg" );

			// Check the first byte of the block.
			if( block[ 0 ] != NonZeroByte )
			{
				throw new ValidationException( string.Format( "The first byte of the stored data is {0} instead of {1},"
					+ " when {2}.", block[ 0 ], NonZeroByte, errorDetail ) );
			}

			// Check the last byte of the block.
			if( block[ block.Length - 1 ] != NonZeroByte )
			{
				throw new ValidationException( string.Format( "The last byte of the stored data is {0} instead of {1}, when"
					+ " {2}.", block[ 0 ], NonZeroByte, errorDetail ) );
			}

			// The hash of the block.
			byte[] blockHash = new byte[ hashAlg.AlgSizeBytes ];
			Buffer.BlockCopy( block, hashPos, blockHash, 0, blockHash.Length );

			// Clear the space for the hash of the data (in order to compute the hash of the entire block).
			Array.Clear( block, hashPos, hashAlg.AlgSizeBytes );

			// Compute the hash of the block.
			byte[] hashVal = hashAlg.Compute( block );

			// Compare the hash extracted from the block with the block's computed hash.
			if( !BufferUtils.Equals( blockHash, hashVal ) )
				throw new ValidationException( string.Format( "Invalid data hash when {0}.", errorDetail ) );
		}

		/// <summary>
		/// Verifies the data from the specified formatted data block.
		/// </summary>
		/// <param name="textEncoding"></param>
		/// <param name="asymmPaddingAlg">The asymmetric padding algorithm.</param>
		/// <param name="hashAlg">The hash algorithm.</param>
		/// <param name="asymmAlg">The asymmetric algorithm.</param>
		/// <param name="block">The formatted data block whose data to verify.</param>
		/// <param name="errorDetail">Message to include in the description of the thrown exceptions. This may be null or empty.
		/// </param>
		/// <param name="valuableDataPos">The position in the formatted data block at which the valuable data starts.</param>
		/// <returns>Returns the size of the valuable data.</returns>
		private static int VerifyBlock( TextEncoding textEncoding, AsymmPaddingAlg asymmPaddingAlg, HashAlg hashAlg
			, AsymmAlg asymmAlg, byte[] block, string errorDetail, out int valuableDataPos )
		{
			ValidationUtils.ObjectHasContent( textEncoding, "textEncoding" );
			ValidationUtils.ObjectHasContent( asymmPaddingAlg, "asymmPaddingAlg" );
			ValidationUtils.ObjectHasContent( hashAlg, "hashAlg" );
			ValidationUtils.ObjectHasContent( asymmAlg, "asymmAlg" );
			ValidationUtils.ObjectHasContent( block, "block" );

			byte[] asymmPaddingName = textEncoding.TextToArray( RegistrableAlgorithmUtils.GetAlgorithmName( asymmPaddingAlg )
				, true );
			byte[] hashAlgName = textEncoding.TextToArray( RegistrableAlgorithmUtils.GetAlgorithmName( hashAlg ), true );
			byte[] asymmAlgName = textEncoding.TextToArray( RegistrableAlgorithmUtils.GetAlgorithmName( asymmAlg ), true );

			// Verify the integrity of the block by computing its hash.
			int hashPos = asymmAlg.MaxProcessableSizeBytes - 1 /* Non-zero byte */ - hashAlg.AlgSizeBytes;
			VerifyBlockHash( block, hashPos, hashAlg, errorDetail );

			// Unmask the salt.
			int saltPos = hashPos - hashAlg.AlgSizeBytes;
			int maskSize = saltPos - 1 /* Non-zero byte */;
			byte[] maskedBuffer = new byte[ maskSize ];
			Buffer.BlockCopy( block, 1, maskedBuffer, 0, maskedBuffer.Length );
			byte[] mask = GenerateMask( hashAlg, maskedBuffer, hashAlg.AlgSizeBytes );
			MaskBlock( block, saltPos, hashAlg.AlgSizeBytes, mask );

			// The salt.
			byte[] salt = new byte[ hashAlg.AlgSizeBytes ];
			Buffer.BlockCopy( block, saltPos, salt, 0, hashAlg.AlgSizeBytes );

			// Unmask the block from its beginning up to the salt (but excluding it).
			mask = GenerateMask( hashAlg, salt, maskSize );
			MaskBlock( block, 1, maskSize, mask );

			// The first byte.
			int blockPos = 1;

			// The name of the asymmetric padding algorithm.
			if( !BufferUtils.Equals( asymmPaddingName, 0, block, blockPos, asymmPaddingName.Length ) )
				throw new ValidationException( "Different asymmetric padding algorithm name." );
			blockPos += asymmPaddingName.Length;

			// The name of the hash algorithm.
			if( !BufferUtils.Equals( hashAlgName, 0, block, blockPos, hashAlgName.Length ) )
				throw new ValidationException( "Different hash algorithm name." );
			blockPos += hashAlgName.Length;

			// The name of the asymmetric algorithm.
			if( !BufferUtils.Equals( asymmAlgName, 0, block, blockPos, asymmAlgName.Length ) )
				throw new ValidationException( "Different asymmetric algorithm name." );
			blockPos += asymmAlgName.Length;

			// The filling size.
			int fillingSize = BitConverter.ToInt32( block, blockPos );
			blockPos += SizeOfDataSize;

			// The filling.
			blockPos += fillingSize;

			// The valuable data size.
			int valuableDataSize = BitConverter.ToInt32( block, blockPos );
			blockPos += SizeOfDataSize;

			// The valuable data.
			valuableDataPos = blockPos;
			blockPos += valuableDataSize;

			return valuableDataSize;
		}

		#endregion Private static methods

		#region Private methods

		/// <summary>
		/// Decrypts the specified data using the asymmetric padding algorithm and the private key of the specified asymmetric
		/// algorithm.
		/// </summary>
		/// <param name="block">Block to decrypt.</param>
		/// <param name="asymmAlg">Asymmetric algorithm.</param>
		/// <returns>Returns the decrypted data.</returns>
		protected byte[] InternalDecryptData( byte[] block, AsymmAlg asymmAlg )
		{
			if( !asymmAlg.HasPrivateKey ) throw new ValidationException( "You can decrypt data only with a private key." );

			ValidationUtils.ArrayHasContent( block, "block" );
			ValidationUtils.ObjectHasContent( asymmAlg, "asymmAlg" );

			// Here we must test ">" instead of "!=" because the big integer library ignores the most significant 0 bits, and
			// thus may generate buffers which are shorted than the specified size.
			if( block.Length > asymmAlg.AlgSizeBytes )
			{
				throw new ValidationException( string.Format( "The size of the data to be decrypted is {0} bytes instead of"
					+ " maximum {1} bytes.", block.Length, asymmAlg.AlgSizeBytes ) );
			}

			// Get the content of the encrypted data.
			byte[] decryptedBlock = asymmAlg.ProcessWithPrivateKey( block, false );

			if( BufferUtils.NoContent( decryptedBlock ) ) throw new ValidationException( "The decrypted data is missing." );

			if( decryptedBlock.Length != asymmAlg.MaxProcessableSizeBytes )
			{
				throw new ValidationException( string.Format( "The size of the decrypted data is {0} bytes instead of {1}"
					+ " bytes.", decryptedBlock.Length, asymmAlg.MaxProcessableSizeBytes ) );
			}

			int valuableDataPos;
			int valuableDataSize = VerifyBlock( TextEncoding, this, HashAlg, asymmAlg, decryptedBlock
				, "asymmetrically decrypting data", out valuableDataPos );

			// Get the valuable data from the block.
			byte[] output = new byte[ valuableDataSize ];
			Buffer.BlockCopy( decryptedBlock, valuableDataPos, output, 0, output.Length );

			return output;
		}

		/// <summary>
		/// Encrypts the specified data using the asymmetric padding algorithm and the public key of the specified asymmetric
		/// algorithm.
		/// </summary>
		/// <param name="block">Block to encrypt. This should be something small, like a hash.</param>
		/// <param name="asymmAlg">Asymmetric algorithm.</param>
		/// <returns>Returns the encrypted data.</returns>
		protected byte[] InternalEncryptData( byte[] block, AsymmAlg asymmAlg )
		{
			ValidationUtils.ArrayHasContent( block, "block" );
			ValidationUtils.ObjectHasContent( asymmAlg, "asymmAlg" );

			// Create the block which is to be asymmetrically encrypted.
			byte[] blockToEncrypt = SetupBlock( TextEncoding, this, HashAlg, asymmAlg, block );

			return asymmAlg.ProcessWithPublicKey( blockToEncrypt, true );
		}

		/// <summary>
		/// Signs the specified data using the private key of the specified asymmetric algorithm.
		/// </summary>
		/// <param name="block">Block to sign. This should be something small, like a hash.</param>
		/// <param name="asymmAlg">Asymmetric algorithm.</param>
		/// <returns>Returns the signed data.</returns>
		protected byte[] InternalSignData( byte[] block, AsymmAlg asymmAlg )
		{
			if( !asymmAlg.HasPrivateKey ) throw new ValidationException( "You can sign data only with a private key." );

			ValidationUtils.ArrayHasContent( block, "block" );
			ValidationUtils.ObjectHasContent( asymmAlg, "asymmAlg" );

			// Create the block which is to be asymmetrically signed.
			byte[] blockToSign = SetupBlock( TextEncoding, this, HashAlg, asymmAlg, block );

			return asymmAlg.ProcessWithPrivateKey( blockToSign, true );
		}

		/// <summary>
		/// Verifies if the specified signature is valid, using the public key of the specified asymmetric algorithm. An
		/// <see cref="InvalidSignatureException"/> (among others) is thrown if the signature is invalid.
		/// </summary>
		/// <param name="block">Block whose signature has to be verified.</param>
		/// <param name="signature">The signature against which the data's signature is verified.</param>
		/// <param name="asymmAlg">Asymmetric algorithm.</param>
		protected void InternalVerifyDataSignature( byte[] block, byte[] signature, AsymmAlg asymmAlg )
		{
			ValidationUtils.ArrayHasContent( block, "block" );

			// The big integer library ignores the most significant 0 bits, and thus may generate buffers which are
			// shorted than the specified maximum size.
			ValidationUtils.Array( signature, 1, asymmAlg.AlgSizeBytes, "signature" );

			ValidationUtils.ObjectHasContent( asymmAlg, "asymmAlg" );

			// Get the content of the signature.
			byte[] extractedBlock = asymmAlg.ProcessWithPublicKey( signature, false );

			if( BufferUtils.NoContent( extractedBlock ) )
				throw new ValidationException( "The extracted signature data is missing." );

			if( extractedBlock.Length != asymmAlg.MaxProcessableSizeBytes )
			{
				throw new ValidationException( string.Format( "The size of the content of the signature which is being verified"
					+ " is {0} bytes instead of {1} bytes.", extractedBlock.Length, asymmAlg.MaxProcessableSizeBytes ) );
			}

			int valuableDataPos;
			VerifyBlock( TextEncoding, this, HashAlg, asymmAlg, extractedBlock, "verifying signature", out valuableDataPos );

			// Compare the valuable data.
			if( !BufferUtils.Equals( block, 0, extractedBlock, valuableDataPos, block.Length ) )
				throw new InvalidSignatureException( "Invalid signature." );
		}

		#endregion Private methods
	}
}
