////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto.Asymmetric.Math;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.Crypto.Asymmetric
{
	/// <summary>
	/// Contract for an RSA algorithm.
	/// </summary>
	/// <remarks>
	/// <para>The size of the algorithm (<see cref="AsymmAlg.AlgSizeBits"/>) must be multiple of
	/// <see cref="BigInteger.BlockBitCount"/>. It should be at least 2048 bits, and at most
	/// <see cref="BigInteger.AsymAlgSizeBitsMax"/>; this is not a technical requirement, but a practical one.</para>
	/// </remarks>
	public abstract class RsaAlg : AsymmAlg
	{
		#region Public constants

		/// <summary>
		/// The default value for the exponent. When a new key pair is generated, this value if used. This must be a prime
		/// number, where the minimum value is 3. The recommended value is 65537.
		/// </summary>
		/// <remarks>The smaller this number is, the faster the processing with the public key becomes, in a significant manner.
		/// </remarks>
		public const int ExponentDef = 65537;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected RsaAlg()
		{
		}

		#endregion Constructors

		#region Private fields

		#region Key parameters

		#region Mandatory private key parameters

		/// <summary>
		/// The "D" private key parameter.
		/// </summary>
		private BigInteger DBigInteger;

		#endregion Mandatory private key parameters

		#region Optional private key parameters (used for CRT)

		/// <summary>
		/// The "DP" private key parameter.
		/// </summary>
		private BigInteger DPBigInteger;

		/// <summary>
		/// The "DQ" private key parameter.
		/// </summary>
		private BigInteger DQBigInteger;

		/// <summary>
		/// The "P" private key parameter.
		/// </summary>
		private BigInteger PBigInteger;

		/// <summary>
		/// The "Q" private key parameter.
		/// </summary>
		private BigInteger QBigInteger;

		/// <summary>
		/// The "QI" private key parameter.
		/// </summary>
		private BigInteger QIBigInteger;

		#endregion Optional private key parameters (used for CRT)

		#region Public key parameters

		/// <summary>
		/// The "M" public key parameter.
		/// </summary>
		private BigInteger MBigInteger;

		/// <summary>
		/// The "E" public key parameter.
		/// </summary>
		private BigInteger EBigInteger;

		#endregion Public key parameters

		#endregion Key parameters

		#endregion Private fields

		#region Public non-serializable properties

		/// <summary>
		/// If this is true then the current algorithm contains the public key. READ-ONLY.
		/// </summary>
		public override bool HasPublicKey
		{
			get { return BufferUtils.HasContent( E ) && BufferUtils.HasContent( M ); }
		}

		/// <summary>
		/// If this is true then the current algorithm contains the private key. READ-ONLY.
		/// </summary>
		public override bool HasPrivateKey
		{
			get { return BufferUtils.HasContent( D ); }
		}

		/// <summary>
		/// If this is true then the current object can use the "Chinese Remainder Theorem" to optimize calculations. READ-ONLY.
		/// </summary>
		public bool IsCrtPossible
		{
			get
			{
				return BufferUtils.HasContent( DP ) && BufferUtils.HasContent( DQ ) && BufferUtils.HasContent( P )
					&& BufferUtils.HasContent( Q ) && BufferUtils.HasContent( QI );
			}
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		#region Mandatory key parameters

		private byte[] _D;
		/// <summary>
		/// The private exponent. This may be null or empty. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public byte[] D
		{
			get { return _D; }
			protected set
			{
				if( BufferUtils.HasContent( value ) && (value.Length > BigInteger.SizeBytesMax) )
				{
					throw new ValidationException( string.Format( "The size of the RSA key parameter is {0}, but the maximum"
						+ " allowed is {1}.", value.Length, BigInteger.SizeBytesMax ) );
				}

				_D = value;
			}
		}

		#endregion Mandatory key parameters

		#region Optional key parameters (used for CRT)

		private byte[] _DP;
		/// <summary>
		/// The "DP" parameter. This may be null or empty. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public byte[] DP
		{
			get { return _DP; }
			protected set
			{
				if( BufferUtils.HasContent( value ) && (value.Length > BigInteger.SizeBytesMax / 2) )
				{
					throw new ValidationException( string.Format( "The size of the RSA key parameter is {0}, but the maximum"
						+ " allowed is {1}.", value.Length, BigInteger.SizeBytesMax / 2 ) );
				}

				_DP = value;
			}
		}

		private byte[] _DQ;
		/// <summary>
		/// The "DQ" parameter. This may be null or empty. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public byte[] DQ
		{
			get { return _DQ; }
			protected set
			{
				if( BufferUtils.HasContent( value ) && (value.Length > BigInteger.SizeBytesMax / 2) )
				{
					throw new ValidationException( string.Format( "The size of the RSA key parameter is {0}, but the maximum"
						+ " allowed is {1}.", value.Length, BigInteger.SizeBytesMax / 2 ) );
				}

				_DQ = value;
			}
		}

		private byte[] _P;
		/// <summary>
		/// The "P" parameter. This may be null or empty. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public byte[] P
		{
			get { return _P; }
			protected set
			{
				if( BufferUtils.HasContent( value ) && (value.Length > BigInteger.SizeBytesMax / 2) )
				{
					throw new ValidationException( string.Format( "The size of the RSA key parameter is {0}, but the maximum"
						+ " allowed is {1}.", value.Length, BigInteger.SizeBytesMax / 2 ) );
				}

				_P = value;
			}
		}

		private byte[] _Q;
		/// <summary>
		/// The "Q" parameter. This may be null or empty. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public byte[] Q
		{
			get { return _Q; }
			protected set
			{
				if( BufferUtils.HasContent( value ) && (value.Length > BigInteger.SizeBytesMax / 2) )
				{
					throw new ValidationException( string.Format( "The size of the RSA key parameter is {0}, but the maximum"
						+ " allowed is {1}.", value.Length, BigInteger.SizeBytesMax / 2 ) );
				}

				_Q = value;
			}
		}

		private byte[] _QI;
		/// <summary>
		/// The inverse "Q" parameter. This may be null or empty. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public byte[] QI
		{
			get { return _QI; }
			protected set
			{
				if( BufferUtils.HasContent( value ) && (value.Length > BigInteger.SizeBytesMax / 2) )
				{
					throw new ValidationException( string.Format( "The size of the RSA key parameter is {0}, but the maximum"
						+ " allowed is {1}.", value.Length, BigInteger.SizeBytesMax / 2 ) );
				}

				_QI = value;
			}
		}

		#endregion Optional key parameters (used for CRT)

		private byte[] _E;
		/// <summary>
		/// Exponent. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] E
		{
			get { return _E; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "E" );

				if( value.Length > BigInteger.SizeBytesMax )
				{
					throw new ValidationException( string.Format( "The size of the RSA key parameter is {0}, but the maximum"
						+ " allowed is {1}.", value.Length, BigInteger.SizeBytesMax ) );
				}

				_E = value;
			}
		}

		#region Public key parameters

		private byte[] _M;
		/// <summary>
		/// Modulus. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] M
		{
			get { return _M; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "M" );

				if( value.Length != AlgSizeBytes )
				{
					throw new ValidationException( string.Format( "The size of the RSA key parameter is {0} instead of {1}."
						, value.Length, BigInteger.SizeBytesMax ) );
				}

				_M = value;
			}
		}

		#endregion Public key parameters

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// This method is automatically called after the current object has been deserialized.
		/// </summary>
		protected virtual void Deserialized()
		{
			BuffersToBigIntegers();
		}

		/// <summary>
		/// Clears the current cryptographic object.
		/// </summary>
		private void Clear()
		{
			// Clear the private key.
			_D = null;
			_DP = null;
			_DQ = null;
			_P = null;
			_Q = null;
			_QI = null;

			// Clear the public key.
			_M = null;
			_E = null;

			BuffersToBigIntegers();
		}

		/// <summary>
		/// Copies the buffer properties to the big integer fields.
		/// </summary>
		private void BuffersToBigIntegers()
		{
			// Mandatory private key parameters.
			DBigInteger = BufferUtils.HasContent( D ) ? new BigInteger( D ) : null;

			// Optional private key parameters.
			DPBigInteger = BufferUtils.HasContent( DP ) ? new BigInteger( DP ) : null;
			DQBigInteger = BufferUtils.HasContent( DQ ) ? new BigInteger( DQ ) : null;
			PBigInteger = BufferUtils.HasContent( P ) ? new BigInteger( P ) : null;
			QBigInteger = BufferUtils.HasContent( Q ) ? new BigInteger( Q ) : null;
			QIBigInteger = BufferUtils.HasContent( QI ) ? new BigInteger( QI ) : null;

			// Public key parameters.
			EBigInteger = BufferUtils.HasContent( E ) ? new BigInteger( E ) : null;
			MBigInteger = BufferUtils.HasContent( M ) ? new BigInteger( M ) : null;
		}

		/// <summary>
		/// Copies the big integer fields to the buffer properties.
		/// </summary>
		private void BigIntegersToBuffers()
		{
			// Mandatory private key parameters.
			D = (DBigInteger != null) ? DBigInteger.GetBytes() : null;

			// Optional private key parameters.
			DP = (DPBigInteger != null) ? DPBigInteger.GetBytes() : null;
			DQ = (DQBigInteger != null) ? DQBigInteger.GetBytes() : null;
			P = (PBigInteger != null) ? PBigInteger.GetBytes() : null;
			Q = (QBigInteger != null) ? QBigInteger.GetBytes() : null;
			QI = (QIBigInteger != null) ? QIBigInteger.GetBytes() : null;

			// Public key parameters.
			E = (EBigInteger != null) ? EBigInteger.GetBytes() : null;
			M = (MBigInteger != null) ? MBigInteger.GetBytes() : null;
		}

		/// <summary>
		/// Copies the public part of the current algorithm to the specified algorithm (the properties are deeply cloned).
		/// </summary>
		protected override void CopyPublicKeyTo( AsymmAlg asymmAlg )
		{
			ValidationUtils.ObjectType<RsaAlg>( asymmAlg, "asymmAlg" );

			RsaAlg rsaAlg = (RsaAlg)asymmAlg;

			// Copy the public properties.
			rsaAlg.E = E;
			rsaAlg.M = M;

			rsaAlg.BuffersToBigIntegers();
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Writes the public parameters in the specified stream.
		/// </summary>
		/// <param name="bns">Stream in which to write the public parameters.</param>
		public override void WriteUpnBodySource( BinaryNodeStream bns )
		{
			ValidationUtils.ObjectHasContent( bns, "bns" );

			bns.WriteValue( GetType().FullName );

			bns.WriteValue( E );
			bns.WriteValue( M );
		}

		#endregion Public methods

		#region Cryptographic direct methods

		/// <summary>
		/// Generates a new (random) key pair of the specified size. The key pair is verified for validity.
		/// </summary>
		/// <param name="publicExponent">The public exponent of the key to generate. This must be a prime number (the minimum
		/// value is 3). If this is 2 or less, the default <see cref="ExponentDef"/> value is used.</param>
		/// <param name="lop">Long operation progress used during the generation of a key pair. This may be null.</param>
		/// <remarks>
		/// <para>It is guaranteed that the public modulus of the generated key pair has its
		/// <see cref="BigInteger.GetBitCount">bit count</see> equal with the size of the algorithm
		/// (<see cref="AsymmAlg.AlgSizeBits"/>). Therefore, you can use the bit count of the modulus to get the size of the
		/// algorithm, at any point.</para>
		/// </remarks>
		public void GenerateKeyPair( int publicExponent, LongOperationProgress lop )
		{
			// Clear the previous key pair.
			Clear();

			try
			{
				if( lop != null ) lop.Progress( 0, "Starting key pair generation...", true );

				// "M" must have a size equal with the key size, and "P" and "Q" must have about the same size, so their size
				// must equal with half of the key size. To guarantee this, the most significant bit of "P" and "Q" must be 1.
				// (The "BigInteger.GeneratePrime"method does this.)
				//
				// In order for "M = P * Q" to have a size equal with the key size, it is necessary to set the second most
				// significant bit of "P" and "Q" to 1. This method guarantees that "P" and "Q" have to be generated only one
				// time.
				//
				// The third most significant bit of "P" and "Q" must be 0 so that their generation would have enough space to
				// search for a prime number without overflowing.

				int pPrimeSize = AlgSizeBits / 2;
				int qPrimeSize = pPrimeSize;

				EBigInteger = (publicExponent >= 3) ? (UInt32)publicExponent : ExponentDef;

				// Generate "P" prime, and "P - 1" relatively prime to "E".
				while( true )
				{
					PBigInteger = BigInteger.GeneratePrime( pPrimeSize, lop );
					if( PBigInteger % EBigInteger != 1 ) break;
				}

				// Generate "Q" prime, "Q - 1" relatively prime to "E", and "Q" not equal with "P".
				while( true )
				{
					QBigInteger = BigInteger.GeneratePrime( qPrimeSize, lop );
					if( (QBigInteger % EBigInteger != 1) && (PBigInteger != QBigInteger) ) break;
				}

				// Calculate the modulus.
				MBigInteger = PBigInteger * QBigInteger;

				// Guarantee the bit count of the public modulus.
				if( MBigInteger.GetBitCount() != AlgSizeBits )
				{
					throw new ValidationException( string.Format( "The size of the public modulus of the key pair has"
						+ " {0} bits instead of {1} bits.", MBigInteger.GetBitCount(), AlgSizeBits ) );
				}

				if( lop != null ) lop.Progress( 0, "Generating other parameters...", true );

				BigInteger pSub1 = PBigInteger - 1;
				BigInteger qSub1 = QBigInteger - 1;
				BigInteger phi = pSub1 * qSub1;

				// Calculate the private exponent.
				DBigInteger = EBigInteger.ModInverse( phi );

				// Calculate the CRT factors.
				DPBigInteger = DBigInteger % pSub1;
				DQBigInteger = DBigInteger % qSub1;
				QIBigInteger = QBigInteger.ModInverse( PBigInteger );
			}
			catch( Exception ex )
			{
				Clear(); // Clear the invalid key pair.

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			if( lop != null ) lop.Progress( 1, "Verifying the generated key pair...", true );

			BigIntegersToBuffers();

			// Verify if the key pair does work.
			if( !VerifyKeyPair() )
			{
				// Invalid key pair.

				Clear(); // Clear the invalid key pair.

				throw new ValidationException( "The generated key pair is invalid." );
			}
		}

		/// <summary>
		/// Generates a new (random) key pair of the specified size. The key pair is verified for validity.
		/// </summary>
		/// <param name="lop">Long operation progress used during the generation of a key pair. This may be null.</param>
		/// <remarks>
		/// <para>It is guaranteed that the public modulus of the generated key pair has its
		/// <see cref="BigInteger.GetBitCount">bit count</see> equal with the size of the algorithm
		/// (<see cref="AsymmAlg.AlgSizeBits"/>). Therefore, you can use the bit count of the modulus to get the size of the
		/// algorithm, at any point.</para>
		/// </remarks>
		public override void GenerateKeyPair( LongOperationProgress lop )
		{
			GenerateKeyPair( ExponentDef, lop );
		}

		/// <summary>
		/// Processes the specified data with the current private key (so that it could only be extracted with the public
		/// key). RESERVED for internal use. Formula: "m = c ^ d mod n".
		/// </summary>
		/// <param name="buffer">Buffer to process. The maximum size of this object is
		/// <see cref="AsymmAlg.MaxProcessableSizeBytes"/>.</param>
		/// <param name="hide">Specifies whether the data to be processed is to be hidden, that is, processed from its raw form
		/// into a form which can be revealed only by converting it back with the public key. If this is false then the data was
		/// previously hidden with the public key.</param>
		/// <returns>Returns the processed data.</returns>
		/// <remarks>
		/// <para>The resulted data will be accessible only to the owner of the public key (associated to the private key which
		/// is used to process the data).</para>
		/// 
		/// <para>This method is much slower than the <see cref="ProcessWithPublicKey"/> method.</para>
		/// </remarks>
		public override byte[] ProcessWithPrivateKey( byte[] buffer, bool hide )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );

			if( !HasPrivateKey ) throw new ValidationException( "You can process data only with a private key." );

			if( hide )
			{
				ValidationUtils.Array( buffer, 1, MaxProcessableSizeBytes, "buffer" );
			}
			else
			{
				// The big integer library ignores the most significant 0 bits, and thus may generate buffers which are
				// shorted than the specified maximum size.
				ValidationUtils.Array( buffer, 1, AlgSizeBytes, "buffer" );
			}

			BigInteger input = new BigInteger( buffer );
			BigInteger output;

			if( hide && ((input >= MBigInteger) || (input == PBigInteger) || (input == QBigInteger)) )
				throw new ValidationException( "The data to hide with the private key is invalid." );

			#region Key blinding

			// Key blinding is used to protect against timing attacks. Key blinding can be done multiple times as long as the
			// unblinding is also done (the order of each blinding / unblinding is irrelevant).

			// The timing of a random number is random. GCD( r, M ) must be 1. We can't use all the possible bits to generate
			// the random big integer necause else we might get failed unblinding (because of the way big integers overflow).

			// This is especially used to protect against acoustic cryptanalysis (http://cs.tau.ac.il/~tromer/acoustic).
			// A cryptographic service has to also protect each ENTIRE operation against timing attacks.

			BigInteger r;

			do
			{
				r = BigInteger.GenerateRandom( MaxProcessableSizeBytes * 8 );
			}
			while( !RsaBlindingConfiguration.IsAllowed( input, MBigInteger, r ) );

			// "x = (r ^ e * g) mod m".
			input = (r.ModPow( EBigInteger, MBigInteger ) * input) % MBigInteger;

			#endregion Key blinding

			if( IsCrtPossible )
			{
				// Decryption can be optimized using the "Chinese Remainder Theorem".

				// "m1 = c ^ dp mod p".
				BigInteger m1 = input.ModPow( DPBigInteger, PBigInteger );

				// "m2 = c ^ dq mod q".
				BigInteger m2 = input.ModPow( DQBigInteger, QBigInteger );

				BigInteger h;
				if( m2 > m1 )
				{
					h = PBigInteger - ((m2 - m1) * QIBigInteger % PBigInteger);
					output = m2 + QBigInteger * h;
				}
				else
				{
					// "h = (m1 - m2) * qi mod p".
					h = (m1 - m2) * QIBigInteger % PBigInteger;

					// "m = m2 + q * h".
					output = m2 + QBigInteger * h;
				}
			}
			else
			{
				// "m = c ^ d mod m".
				output = input.ModPow( DBigInteger, MBigInteger );
			}

			#region Key unblinding

			// "g = x / r mod m".
			output = output * r.ModInverse( MBigInteger ) % MBigInteger;

			#endregion Key unblinding

			return output.GetBytes();
		}

		/// <summary>
		/// Processes the specified data with the current public key (so that it could only be extracted with the private
		/// key). RESERVED for internal use. Formula: "c = m ^ e mod n".
		/// </summary>
		/// <param name="buffer">Buffer to process. The maximum size of this object is
		/// <see cref="AsymmAlg.MaxProcessableSizeBytes"/>.</param>
		/// <param name="hide">Specifies whether the data to be processed is to be hidden, that is, processed from its raw form
		/// into a form which can be revealed only by converting it back with the private key. If this is false then the data
		/// was previously hidden with the private key.</param>
		/// <returns>Returns the processed data.</returns>
		/// <remarks>
		/// <para>The resulted data will be accessible only to the owner of the private key (associated to the public key which
		/// is used to process the data).</para>
		/// 
		/// <para>This method is much faster than the <see cref="ProcessWithPrivateKey"/> method.</para>
		/// </remarks>
		public override byte[] ProcessWithPublicKey( byte[] buffer, bool hide )
		{
			#region Validate

			ValidationUtils.ArrayHasContent( buffer, "buffer" );

			if( hide )
			{
				if( buffer.Length > MaxProcessableSizeBytes )
				{
					throw new ValidationException( string.Format( "The size of the data ({0} bytes) to be hidden with the"
							+ " public key is bigger than the maximum possible ({1} bytes)."
						, buffer.Length, MaxProcessableSizeBytes ) );
				}
			}
			else
			{
				// Here we must test ">" instead of "!=" because the big integer library ignores the most significant 0 bits,
				// and thus may generate buffers which are shorter than the specified size.
				if( buffer.Length > AlgSizeBytes )
				{
					throw new ValidationException( string.Format( "The size of the data to be revealed with the public key"
						+ " is {0} instead of maximum {1} bytes.", buffer.Length, AlgSizeBytes ) );
				}
			}

			#endregion Validate

			BigInteger input = new BigInteger( buffer );

			if( hide && ((input >= MBigInteger) || (input == PBigInteger) || (input == QBigInteger)) )
				throw new ValidationException( "The data to hide with the public key is invalid." );

			// "o = i ^ e mod m".
			BigInteger output = input.ModPow( EBigInteger, MBigInteger );

			byte[] result = output.GetBytes();

			return result;
		}

		#endregion Cryptographic direct methods

		#region Cryptographic public methods

		/// <summary>
		/// Decrypts the specified data with the current private key.
		/// </summary>
		/// <param name="buffer">Buffer to decrypt.</param>
		/// <param name="asymmEncryptionPaddingAlg">The asymmetric encryption padding algorithm. Must be the same as the one
		/// used during encryption.</param>
		/// <returns>Returns the decrypted data.</returns>
		public override byte[] DecryptData( byte[] buffer, IAsymmEncryptionPaddingAlg asymmEncryptionPaddingAlg )
		{
			ValidationUtils.ObjectHasContent( asymmEncryptionPaddingAlg, "asymmEncryptionPaddingAlg" );

			if( !HasPrivateKey ) throw new ValidationException( "You can decrypt data only with a private key." );

			return asymmEncryptionPaddingAlg.DecryptData( buffer, this );
		}

		/// <summary>
		/// Encrypts the specified data with the current public key.
		/// </summary>
		/// <param name="buffer">Buffer to encrypt. This should be something small, like a hash.</param>
		/// <param name="asymmEncryptionPaddingAlg">The asymmetric encryption padding algorithm.</param>
		/// <returns>Returns the encrypted data.</returns>
		public override byte[] EncryptData( byte[] buffer, IAsymmEncryptionPaddingAlg asymmEncryptionPaddingAlg )
		{
			ValidationUtils.ObjectHasContent( asymmEncryptionPaddingAlg, "asymmEncryptionPaddingAlg" );

			return asymmEncryptionPaddingAlg.EncryptData( buffer, this );
		}

		/// <summary>
		/// Signs the specified data with the current private key.
		/// </summary>
		/// <param name="buffer">Buffer to sign. This should be something small, like a hash.</param>
		/// <param name="asymmSigningPaddingAlg">The asymmetric signing padding algorithm.</param>
		/// <returns>Returns the signature of the specified data.</returns>
		public override byte[] SignData( byte[] buffer, IAsymmSigningPaddingAlg asymmSigningPaddingAlg )
		{
			ValidationUtils.ObjectHasContent( asymmSigningPaddingAlg, "asymmSigningPaddingAlg" );

			if( !HasPrivateKey ) throw new ValidationException( "You can sign data only with a private key." );

			return asymmSigningPaddingAlg.SignData( buffer, this );
		}

		/// <summary>
		/// Verifies the signature of the specified data with the current public key. An
		/// <see cref="InvalidSignatureException"/> (among others) is thrown if the signature is invalid.
		/// </summary>
		/// <param name="buffer">Buffer whose signature has to be verified.</param>
		/// <param name="signature">The signature against which the data's signature is verified.</param>
		/// <param name="asymmSigningPaddingAlg">The asymmetric signing padding algorithm. Must be the same as the one used
		/// during signing.</param>
		public override void VerifyDataSignature( byte[] buffer, byte[] signature
			, IAsymmSigningPaddingAlg asymmSigningPaddingAlg )
		{
			// The big integer library ignores the most significant 0 bits, and thus may generate buffers which are
			// shorted than the specified maximum size.
			ValidationUtils.Array( signature, 1, AlgSizeBytes, "signature" );

			ValidationUtils.ObjectHasContent( asymmSigningPaddingAlg, "asymmSigningPaddingAlg" );

			asymmSigningPaddingAlg.VerifyDataSignature( buffer, signature, this );
		}

		#endregion Cryptographic public methods

		#region Blinding public methods

		/// <summary>
		/// Generates random blinding configuration.
		/// </summary>
		/// <param name="buffer">Buffer to blind.</param>
		/// <returns>Returns the generated blinding configuration.</returns>
		public override BlindingConfiguration GenerateBlindingConfiguration( byte[] buffer )
		{
			BigInteger dataBi = new BigInteger( buffer );

			int bitCount = MaxProcessableSizeBytes * 8;
			while( true )
			{
				BigInteger rand = BigInteger.GenerateRandom( bitCount );

				// We make sure the random number is not small.
				rand.SetBit( bitCount - 1 );

				if( RsaBlindingConfiguration.IsAllowed( dataBi, MBigInteger, rand ) )
					return new RsaBlindingConfiguration( rand );
			}
		}

		/// <summary>
		/// Blinds the specified data with the specified parameters.
		/// </summary>
		/// <param name="buffer">Buffer to blind.</param>
		/// <param name="blindingConfiguration">Parameters with which to perform the blinding.</param>
		/// <returns>Returns the blinded data.</returns>
		public override byte[] Blind( byte[] buffer, BlindingConfiguration blindingConfiguration )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );
			ValidationUtils.ObjectType<RsaBlindingConfiguration>( blindingConfiguration, "blindingConfiguration" );

			BigInteger bi = new BigInteger( buffer );
			RsaBlindingConfiguration rsaConfig = (RsaBlindingConfiguration)blindingConfiguration;
			BigInteger r = new BigInteger( rsaConfig.R );

			if( !RsaBlindingConfiguration.IsAllowed( bi, MBigInteger, r ) )
				throw new ValidationException( "The data may not be blinded.", null, "bi" );

			BigInteger blinded = (r.ModPow( EBigInteger, MBigInteger ) * bi) % MBigInteger;

			return blinded.GetBytes();
		}

		/// <summary>
		/// Unblinds the specified data with the specified parameters.
		/// </summary>
		/// <param name="buffer">Buffer to unblind.</param>
		/// <param name="blindingConfiguration">Parameters with which to perform the unblinding.</param>
		/// <returns>Returns the unblinded data.</returns>
		public override byte[] Unblind( byte[] buffer, BlindingConfiguration blindingConfiguration )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );
			ValidationUtils.ObjectHasContent( blindingConfiguration, "blindingConfiguration" );

			BigInteger bi = new BigInteger( buffer );

			RsaBlindingConfiguration rsaConfig = (RsaBlindingConfiguration)blindingConfiguration;
			BigInteger r = new BigInteger( rsaConfig.R );

			BigInteger unblinded = (bi * r.ModInverse( MBigInteger )) % MBigInteger;

			return unblinded.GetBytes();
		}

		#endregion Blinding public methods
	}
}
