////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Crypto.Asymmetric.Math;

namespace Vela.Crypto.Asymmetric
{
	/// <summary>
	/// RSA blinding configuration.
	/// </summary>
	public class RsaBlindingConfiguration : BlindingConfiguration
	{
		#region Public constants

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected RsaBlindingConfiguration()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="r">The random multiplier.</param>
		public RsaBlindingConfiguration( BigInteger r )
		{
			R = r.GetBytes();
		}

		#endregion Public constants

		#region Public serializable properties

		private byte[] _R;
		/// <summary>
		/// The random multiplier. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] R
		{
			get { return _R; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "R" );

				_R = value;
			}
		}

		#endregion Public serializable properties

		#region Public static methods

		/// <summary>
		/// Verifies if the specified data may be blinded with the specified random number. An exception is thrown if the
		/// validations not related to the random number fail.
		/// </summary>
		/// <param name="bi">Big integer to check if it may be blinded.</param>
		/// <param name="m">The modulo of the public key parameters which the blinding will be done.</param>
		/// <param name="r">The random number with which the blinding will be done.</param>
		/// <returns>Returns true if the specified data may be blinded.</returns>
		public static bool IsAllowed( BigInteger bi, BigInteger m, BigInteger r )
		{
			ValidationUtils.ObjectHasContent( bi, "bi" );
			ValidationUtils.ObjectHasContent( m, "m" );
			ValidationUtils.ObjectHasContent( r, "r" );

			if( bi <= 1 )
			{
				throw new ValidationException( "Blinding is not allowed because the data to blind is smaller than or"
					+ " equal with 1.", null, "bi" );
			}

			if( m <= 1 )
			{
				throw new ValidationException( "Blinding is not allowed because the module of the RSA algorithm is"
					+ " smaller than or equal with 1.", null, "m" );
			}

			if( bi >= m )
			{
				throw new ValidationException( "Blinding is not allowed because the data to blind is greater than or"
					+ " equal with the module of the RSA algorithm.", null, "bi" );
			}

			return (m.Gcd( r ) == 1) && (m > r) && (bi.Gcd( r ) == 1) && (bi != r) && (r > 1);
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Verifies if the current key has the same content as the specified key.
		/// </summary>
		/// <param name="blindingConfiguration">Object to verify against the current object.</param>
		/// <returns>Returns true if the objects have the same content.</returns>
		public override bool Equals( BlindingConfiguration blindingConfiguration )
		{
			ValidationUtils.ObjectType<RsaBlindingConfiguration>( blindingConfiguration, "blindingConfiguration" );

			RsaBlindingConfiguration parameters = (RsaBlindingConfiguration)blindingConfiguration;

			return BufferUtils.Equals( R, parameters.R );
		}

		/// <summary>
		/// Verifies if the specified data may be blinded.
		/// </summary>
		/// <param name="bi">Big integer to check if it may be blinded.</param>
		/// <param name="asymmAlg">The asymmetric algorithm which the blinding will be done.</param>
		/// <returns>Returns true if the specified data may be blinded.</returns>
		public override bool IsAllowed( BigInteger bi, AsymmAlg asymmAlg )
		{
			ValidationUtils.ObjectType<RsaAlg>( asymmAlg, "asymmAlg" );

			RsaAlg rsaAlg = (RsaAlg)asymmAlg;

			return IsAllowed( bi, new BigInteger( rsaAlg.M ), new BigInteger( R ) );
		}

		#endregion Public methods
	}
}
