////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Microsoft.VisualBasic.Devices;
using System;
using System.Diagnostics;
using System.Net.NetworkInformation;
using Vela.Kernel;

namespace Vela.Crypto
{
	/// <summary>
	/// Utilities for working with core documents.
	/// </summary>
	public static class CryptoUtils
	{
		#region Public static methods

		/// <summary>
		/// Gets entropy from the system.
		/// </summary>
		/// <param name="addFixedAndSlowInfo">If this is true then some fixed and slowly retrievable information related to the
		/// environment are also added.</param>
		/// <returns>Returns entropy from the system.</returns>
		public static byte[] GetSystemEntropy( bool addFixedAndSlowInfo = false )
		{
			BinaryNodeStream bns = new BinaryNodeStream( 0, KernelUtils.GetDeepTextEncoding() );

			// Time information.
			bns.WriteValue( DateTime.UtcNow.Ticks );
			bns.WriteValue( HighResolutionTimer.GetTicks() );

			if( addFixedAndSlowInfo )
			{
				// Computer information.
				try
				{
					ComputerInfo computerInfo = new ComputerInfo();

					bns.WriteValue( computerInfo.AvailablePhysicalMemory );
				}
				catch( Exception ex )
				{
					// The retrieval of information must continue (if possible).

					if( ExceptionUtils.IsAbort( ex ) )
						throw ExceptionUtils.PreserveStackTrace( ex );

					LogManager.Add( ex );
				}

				// Process information.
				try
				{
					Process[] processes = Process.GetProcesses();

					foreach( Process process in processes )
					{
						try
						{
							bns.WriteValue( process.Id );
							bns.WriteValue( process.ProcessName );
							bns.WriteValue( process.HandleCount );
							bns.WriteValue( process.WorkingSet64 );
						}
						catch( Exception ex )
						{
							// The loop must continue (if possible).

							if( ExceptionUtils.IsAbort( ex ) )
								throw ExceptionUtils.PreserveStackTrace( ex );

							LogManager.Add( ex );
						}
					}
				}
				catch( Exception ex )
				{
					// The retrieval of information must continue (if possible).

					if( ExceptionUtils.IsAbort( ex ) )
						throw ExceptionUtils.PreserveStackTrace( ex );

					LogManager.Add( ex );
				}

				// Network information.
				try
				{
					NetworkInterface[] nis = NetworkInterface.GetAllNetworkInterfaces();

					foreach( NetworkInterface ni in nis )
					{
						// If the MAC address is missing, this is not a real network interface.
						string macAddress = ni.GetPhysicalAddress().ToString();
						if( TextUtils.NoContent( macAddress ) )
							continue;

						// Interface identifier.
						bns.WriteValue( ni.Id );

						// Interface MAC.
						bns.WriteValue( macAddress );

						// Interface IP addresses.
						IPInterfaceProperties ipProp = ni.GetIPProperties();
						foreach( UnicastIPAddressInformation ipInfo in ipProp.UnicastAddresses )
							bns.WriteValue( ipInfo.Address.GetAddressBytes() );

						// Interface statistics.
						IPv4InterfaceStatistics stats = ni.GetIPv4Statistics();
						bns.WriteValue( stats.BytesReceived );
						bns.WriteValue( stats.BytesSent );
					}
				}
				catch( Exception ex )
				{
					// The retrieval of information must continue (if possible).

					if( ExceptionUtils.IsAbort( ex ) )
						throw ExceptionUtils.PreserveStackTrace( ex );

					LogManager.Add( ex );
				}
			}

			return bns.ToArray();
		}

		/// <summary>
		/// Gets a new (random) locator.
		/// </summary>
		/// <param name="valueSize">The size (in bytes) that the locator value must have. If this is 0 or a negative
		/// number, the default <see cref="Locator.ValueSizeDef"/> is used.</param>
		public static Locator GetLocator( int valueSize = 0 )
		{
			byte[] buffer = EntropyPool.GetBytes( (valueSize > 0) ? valueSize : Locator.ValueSizeDef );

			return new Locator( buffer );
		}

		#endregion Public static methods
	}
}
