////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.Crypto
{
	/// <summary>
	/// Contract for an entropy algorithm which can be used in cryptographic processes.
	/// </summary>
	/// <remarks>
	/// <para>The generated data is pseudo-random.</para>
	/// 
	/// <para>Using a hash algorithm in the design of the entropy algorithm is safe because this never calls back the entropy
	/// algorithm.</para>
	/// 
	/// <para>Using a symmetric encryption algorithm in the design of the entropy algorithm is safe so long as this never calls
	/// back the entropy algorithm. Normally, the entropy algorithm is not used if the <see cref="SymmPaddings.None"/>
	/// padding is used, but check the algorithm's documentation.</para>
	/// 
	/// <para>The entropy algorithm is required to generate data which is as unpredictable as possible. It's less important to
	/// generate a large amount of data for a single request.</para>
	/// 
	/// <para>The entropy algorithm must be secure even if the attacker is able to inject data in it, for a long time, after the
	/// entropy state is initialized.</para>
	/// 
	/// <para>The entropy state must not be obtainable from the generated data. For this to happen, the data must be generated
	/// from the entropy state through a process which is different than the one which blends the entropy state in order to
	/// prepare it for the next generation.</para>
	/// 
	/// <para>Never replace entropy algorithms, always merge them. This will avoid subtle bugs where the entropy gained while
	/// the application is running, is not saved or not restored later.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Entropy sources</b></para>
	/// 
	/// <para>The most important source of entropy is the one which is manually introduced by a user when the entropy algorithm
	/// is initialized.</para>
	/// 
	/// <para>The entropy algorithm should use the following sources of entropy: user typed entropy, the current time, the
	/// system's high resolution clock, coordinates of user clicks, and the hash of the encrypted document from the decrypted
	/// documents.</para>
	/// 
	/// <para>Services may also use the hash of the encrypted document from the processed action requests.</para>
	/// </remarks>
	public abstract class EntropyAlg : IDocument
	{
		#region Public constants

		/// <summary>
		/// The size (in bytes) of the <see cref="ProgressionCounter"/> property.
		/// </summary>
		public const int ProgressionCounterSize = 32;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected EntropyAlg()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="textEncoding">Text encoding.</param>
		protected EntropyAlg( TextEncoding textEncoding )
		{
			TextEncoding = textEncoding;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The size (in bytes) required to merge the current algorithm with data from another. READ-ONLY.
		/// </summary>
		public abstract int MergingStateSize { get; }

		#endregion Public non-serializable properties

		#region Public serializable properties

		private TextEncoding _TextEncoding;
		/// <summary>
		/// Text encoding. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public TextEncoding TextEncoding
		{
			get { return _TextEncoding; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "TextEncoding" );

				_TextEncoding = value;
			}
		}

		private byte[] _ProgressionCounter = new byte[ ProgressionCounterSize ];
		/// <summary>
		/// The counter of the generated blocks with entropy. The backing field is initialized by default.
		/// </summary>
		/// <remarks>
		/// <para>This should never be reset.</para>
		/// 
		/// <para>Since this is a new property, it must have "MustBeDeserialized = false".</para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false )]
		public byte[] ProgressionCounter
		{
			get { return _ProgressionCounter; }
			set { _ProgressionCounter = value; }
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Validates the current object.
		/// </summary>
		protected abstract void Validate();

		/// <summary>
		/// This method is automatically called before the current object is serialized.
		/// </summary>
		protected virtual void Serializing()
		{
			Validate();
		}

		/// <summary>
		/// This method is automatically called after the current object has been deserialized.
		/// </summary>
		protected virtual void Deserialized()
		{
			Validate();

			SetupState();
		}

		/// <summary>
		/// Sets up the state of the algorithm.
		/// </summary>
		protected abstract void SetupState();

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Adds the specified data into the entropy algorithm.
		/// </summary>
		/// <param name="buffer">Buffer to add.</param>
		public abstract void Add( byte[] buffer );

		/// <summary>
		/// Adds the specified data into the entropy algorithm.
		/// </summary>
		/// <param name="text">Text to add.</param>
		public void Add( string text )
		{
			ValidationUtils.TextHasContent( text, "text" );

			byte[] buffer = TextEncoding.TextToArray( text, false );

			Add( buffer );
		}

		/// <summary>
		/// Merges the current entropy algorithm with the specified entropy algorithm, in both directions.
		/// </summary>
		/// <param name="entropyAlg">The entropy algorithm to merge with the current entropy algorithm.</param>
		public virtual void Merge( EntropyAlg entropyAlg )
		{
			ValidationUtils.ObjectHasContent( entropyAlg, "entropyAlg" );

			byte[] progressionCounter = (BufferUtils.Compare( ProgressionCounter, entropyAlg.ProgressionCounter ) >= 0)
				? ProgressionCounter : entropyAlg.ProgressionCounter;

			ProgressionCounter = (byte[])progressionCounter.Clone();
			Add( entropyAlg.GetBytes( 2 * MergingStateSize ) );

			entropyAlg.ProgressionCounter = (byte[])ProgressionCounter.Clone();
			entropyAlg.Add( GetBytes( 2 * entropyAlg.MergingStateSize ) );
		}

		/// <summary>
		/// Returns an array of random bytes.
		/// </summary>
		/// <param name="size">The number of bytes to return in the array.</param>
		/// <returns>Returns an array of random bytes.</returns>
		public abstract byte[] GetBytes( int size );

		#endregion Public methods
	}
}
