////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Diagnostics;
using System.Text;
using Vela.Kernel;

namespace Vela.Crypto
{
	/// <summary>
	/// The global entropy pool.
	/// </summary>
	public static class EntropyPool
	{
		#region Public constants

		/// <summary>
		/// The threshold for the size of generated entropy from which an exception is thrown if no real entropy was yet added to
		/// the pool (that is, <see cref="ContainsRealEntropy"/> is false).
		/// </summary>
		public const int NoRealEntropyExceptionThreshold = 5;

		#endregion Public constants

		#region Private static fields

		/// <summary>
		/// The synchronization object for the <see cref="Instance"/> property. READ-ONLY.
		/// </summary>
		private static readonly object InstanceLock = new object();

		/// <summary>
		/// If this is true then real entropy was added to the pool, so the pool can be used to generate proper pseudo-entropy.
		/// </summary>
		private static bool ContainsRealEntropy = false;

		#endregion Private static fields

		#region Private static properties

		private static EntropyAlg _Instance;
		/// <summary>
		/// The global entropy algorithm.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// </remarks>
		private static EntropyAlg Instance
		{
			get
			{
				lock( InstanceLock )
				{
					if( _Instance == null )
					{
						throw new ValidationException( "The global entropy algorithm should be set when the application is"
							+ " started, before this property is read." );
					}

					return _Instance;
				}
			}
			set
			{
				ValidationUtils.ObjectHasContent( value, "Instance" );

				lock( InstanceLock )
				{
					_Instance = value;
				}
			}
		}

		#endregion Private static properties

		#region Private static methods

		/// <summary>
		/// Checks if real entropy was added to the pool. An exception is thrown if no real entropy was yet added to the pool.
		/// </summary>
		/// <param name="throwIfNoRealEntropy">If this is true then an exception is thrown if no real entropy was yet added to
		/// the pool, else only a debug message is logged.</param>
		private static void CheckItContainsRealEntropy( bool throwIfNoRealEntropy )
		{
			if( !ContainsRealEntropy )
			{
				if( throwIfNoRealEntropy )
				{
					throw new ValidationException( "Real entropy was not yet added to the pool." );
				}
				else
				{
					string message = string.Format( "Real entropy was not yet added to the pool. Check the stack trace to see if"
						+ " this message indicates a design flaw:\n{0}", new StackTrace() );

					LogManager.DebugAdd( message, null, LogRecordModes.Warning, true );
				}
			}
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Sets the global entropy algorithm with the specified one.
		/// </summary>
		/// <param name="instance">Entropy algorithm.</param>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static void SetInstance( EntropyAlg instance )
		{
			Instance = instance;
		}

		/// <summary>
		/// Sets the <see cref="ContainsRealEntropy"/> property with true. Call this method after real entropy was added to the
		/// pool, or after an entropy algorithm which contains real entropy replaces the global entropy algorithm.
		/// </summary>
		public static void SetContainsRealEntropy()
		{
			ContainsRealEntropy = true;
		}

		/// <summary>
		/// Merges the global entropy algorithm with the specified entropy algorithm, in both directions.
		/// </summary>
		/// <param name="entropyAlg">The entropy algorithm to merge.</param>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static void Merge( EntropyAlg entropyAlg )
		{
			ValidationUtils.ObjectHasContent( entropyAlg, "entropyAlg" );

			lock( InstanceLock )
			{
				entropyAlg.Merge( Instance );
			}
		}

		/// <summary>
		/// Merges the global entropy algorithm with the specified entropy algorithm, in both directions, and replaces the
		/// global entropy algorithm (with the specified one). It's presumed that the specified entropy algorithm already
		/// contains real entropy.
		/// </summary>
		/// <param name="entropyAlg">The entropy algorithm to merge.</param>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// 
		/// <para><see cref="ContainsRealEntropy"/> is set with true in here.</para>
		/// </remarks>
		public static void MergeAndReplace( EntropyAlg entropyAlg )
		{
			ValidationUtils.ObjectHasContent( entropyAlg, "entropyAlg" );

			lock( InstanceLock )
			{
				EntropyAlg oldInstance = Instance;

				Instance = entropyAlg;
				SetContainsRealEntropy();

				entropyAlg.Merge( oldInstance );
			}
		}

		/// <summary>
		/// Adds the specified data into the global entropy algorithm.
		/// </summary>
		/// <param name="buffer">Buffer to add.</param>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static void Add( byte[] buffer )
		{
			lock( InstanceLock )
			{
				Instance.Add( buffer );
			}
		}

		/// <summary>
		/// Adds the specified data into the global entropy algorithm.
		/// </summary>
		/// <param name="text">Text to add.</param>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static void Add( string text )
		{
			lock( InstanceLock )
			{
				Instance.Add( text );
			}
		}

		/// <summary>
		/// Returns an array of random bytes.
		/// </summary>
		/// <param name="size">The number of bytes to return in the array.</param>
		/// <param name="throwIfNoRealEntropy">If this is true then an exception is thrown if no real entropy was yet added to
		/// the pool, else only a debug message is logged.</param>
		/// <returns>Returns an array of random bytes.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static byte[] GetBytes( int size, bool throwIfNoRealEntropy = true )
		{
			lock( InstanceLock )
			{
				CheckItContainsRealEntropy( throwIfNoRealEntropy && (size >= NoRealEntropyExceptionThreshold) );

				return Instance.GetBytes( size );
			}
		}

		/// <summary>
		/// Returns an array of (non-random) bytes used to fill some space which doesn't need to contain random bytes.
		/// </summary>
		/// <param name="throwIfNoRealEntropy">If this is true then an exception is thrown if no real entropy was yet added to
		/// the pool, else only a debug message is logged.</param>
		/// <param name="size">The number of bytes to return in the array.</param>
		/// <returns>Returns an array of non-random bytes. This may be empty, but not null.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static byte[] GetFillBytes( int size, bool throwIfNoRealEntropy = true )
		{
			if( size <= 0 )
				return new byte[ 0 ];

			Random rng = new Random( GetInt( throwIfNoRealEntropy ) );

			byte[] buffer = new byte[ size ];
			rng.NextBytes( buffer );

			return buffer;
		}

		/// <summary>
		/// Returns an array of non-zero random bytes.
		/// </summary>
		/// <param name="size">The number of bytes to return in the array.</param>
		/// <param name="throwIfNoRealEntropy">If this is true then an exception is thrown if no real entropy was yet added to
		/// the pool, else only a debug message is logged.</param>
		/// <returns>Returns an array of non-zero random bytes.</returns>
		/// <remarks>
		/// <para>By analyzing the value returned by this method, it may be possible to distinguish it from true random data.
		/// </para>
		/// 
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static byte[] GetNonZeroBytes( int size, bool throwIfNoRealEntropy = true )
		{
			byte[] random = GetBytes( size, throwIfNoRealEntropy );

			// Make sure that no byte is 0.
			for( int i = 0; i < size; i++ )
			{
				if( random[ i ] == 0 )
					random[ i ] = (i > 0) ? random[ i - 1 ] : (byte)0x7F;
			}

			return random;
		}

		/// <summary>
		/// Returns a text of random letters and digits, created from the base-16 representation of a random buffer.
		/// </summary>
		/// <param name="size">The number of characters to return in the text. If this is an odd number, the returned text will
		/// contain one character more. This must be at least 1.</param>
		/// <param name="throwIfNoRealEntropy">If this is true then an exception is thrown if no real entropy was yet added to
		/// the pool, else only a debug message is logged.</param>
		/// <returns>Returns a text of random letters and digits.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static string GetText16( int size, bool throwIfNoRealEntropy = true )
		{
			ValidationUtils.Integer( size, 1, int.MaxValue, "size" );

			byte[] random = GetBytes( (size + 1) / 2, throwIfNoRealEntropy );

			return TextUtils.ArrayToText16( random );
		}

		/// <summary>
		/// Returns a text of random letters and digits, created from the base-32 representation of a random buffer.
		/// </summary>
		/// <param name="size">The number of characters to return in the text. This must be at least 1.</param>
		/// <param name="throwIfNoRealEntropy">If this is true then an exception is thrown if no real entropy was yet added to
		/// the pool, else only a debug message is logged.</param>
		/// <returns>Returns a text of random letters and digits.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static string GetText32( int size, bool throwIfNoRealEntropy = true )
		{
			ValidationUtils.Integer( size, 1, int.MaxValue, "size" );

			byte[] random = GetBytes( MathUtils.AlignUp( 5 * size, 8 ) / 8, throwIfNoRealEntropy );

			return TextUtils.ArrayToText32( random );
		}

		/// <summary>
		/// Returns a text of random lowercase letters.
		/// </summary>
		/// <param name="size">The number of characters to return in the text. This must be at least 1.</param>
		/// <param name="throwIfNoRealEntropy">If this is true then an exception is thrown if no real entropy was yet added to
		/// the pool, else only a debug message is logged.</param>
		/// <returns>Returns a text of random lowercase letters.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static string GetLowercaseText( int size, bool throwIfNoRealEntropy = true )
		{
			ValidationUtils.Integer( size, 1, int.MaxValue, "size" );

			// Don't optimize the size because it's too complicated to convert to base 26.
			byte[] random = GetBytes( size, throwIfNoRealEntropy );

			StringBuilder sb = new StringBuilder();

			for( int i = 0; i < size; i++ )
			{
				char ch = (char)('a' + (random[ i ] % 26));
				sb.Append( ch );
			}

			return sb.ToString();
		}

		/// <summary>
		/// Returns a random boolean.
		/// </summary>
		/// <param name="throwIfNoRealEntropy">If this is true then an exception is thrown if no real entropy was yet added to
		/// the pool, else only a debug message is logged.</param>
		/// <returns>Returns a random boolean.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static bool GetBool( bool throwIfNoRealEntropy = true )
		{
			byte[] buffer = GetBytes( 1, throwIfNoRealEntropy );

			return (buffer[ 0 ] & 1) == 1;
		}

		/// <summary>
		/// Returns a random byte.
		/// </summary>
		/// <param name="throwIfNoRealEntropy">If this is true then an exception is thrown if no real entropy was yet added to
		/// the pool, else only a debug message is logged.</param>
		/// <returns>Returns a random byte.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static byte GetByte( bool throwIfNoRealEntropy = true )
		{
			byte[] buffer = GetBytes( 1, throwIfNoRealEntropy );

			return buffer[ 0 ];
		}

		/// <summary>
		/// Returns a random integer.
		/// </summary>
		/// <param name="throwIfNoRealEntropy">If this is true then an exception is thrown if no real entropy was yet added to
		/// the pool, else only a debug message is logged.</param>
		/// <returns>Returns a random integer.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static int GetInt( bool throwIfNoRealEntropy = true )
		{
			byte[] buffer = GetBytes( sizeof( int ), throwIfNoRealEntropy );

			return BitConverter.ToInt32( buffer, 0 );
		}

		/// <summary>
		/// Returns a random integer between ["min"..."max").
		/// </summary>
		/// <param name="min">The (inclusive) minimum value to return.</param>
		/// <param name="max">The (exclusive) maximum value to return. This is not checked to be greater than "min".</param>
		/// <param name="throwIfNoRealEntropy">If this is true then an exception is thrown if no real entropy was yet added to
		/// the pool, else only a debug message is logged.</param>
		/// <returns>Returns a random integer.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static int GetInt( int min, int max, bool throwIfNoRealEntropy = true )
		{
			return (int)(min + GetReal( throwIfNoRealEntropy ) * (max - min));
		}

		/// <summary>
		/// Returns a random real number within [0...1).
		/// </summary>
		/// <param name="throwIfNoRealEntropy">If this is true then an exception is thrown if no real entropy was yet added to
		/// the pool, else only a debug message is logged.</param>
		/// <returns>Returns a random integer.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static double GetReal( bool throwIfNoRealEntropy = true )
		{
			return (1 - GetInt( throwIfNoRealEntropy ) / (double)int.MinValue) / 2;
		}

		#endregion Public static methods
	}
}
