////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto.Hash;
using Vela.Crypto.Symmetric;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.Crypto
{
	/// <summary>
	/// "Mash" entropy algorithm.
	/// </summary>
	/// <remarks>
	/// <para>The algorithm design is inspired by Fortuna. The generated data is the symmetrically encrypted result of a block
	/// counter. The encryption salt and key are changed at every data request, and also periodically during the request. For
	/// each data request, the counter starts at 1 and is incremented with 1 for each generated block.</para>
	/// 
	/// <para>The entropy algorithm is faster in generating long rather than short buffers of data.</para>
	/// 
	/// <para>This algorithm produces many collisions, so if you need well distributed random numbers, especially of just a few
	/// bytes, you have to create your own algorithm. However, such a generator would be more predictable because as more entropy
	/// is generated, fewer possible values remain, although this should not be a concern for large requests (of tens of bytes or
	/// more).</para>
	/// </remarks>
	public class Mash : EntropyAlg
	{
		#region Public constants

		/// <summary>
		/// The number of dynamic states. This must be at least 2.
		/// </summary>
		public const int DynamicStatesCount = 32;

		/// <summary>
		/// The number of generated bytes after which the algorithm state has to be advanced.
		/// </summary>
		/// <remarks>
		/// <para>Fortuna recommends to ouput maximum 1 MB of entropy per request.</para>
		/// </remarks>
		public const int AdvanceStateAfterBytes = 1 << 20;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Mash()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="textEncoding">Text encoding.</param>
		/// <param name="hashAlg">Hash algorithm.</param>
		/// <param name="symmAlg">Symmetric encryption algorithm</param>
		public Mash( TextEncoding textEncoding, HashAlg hashAlg, SymmAlg symmAlg )
			: base( textEncoding )
		{
			HashAlg = hashAlg;
			SymmAlg = symmAlg;

			SetupState();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The number of generated blocks after which the algorithm state has to be advanced.
		/// </summary>
		private int AdvanceStateAfterBlocks;

		/// <summary>
		/// List of dynamic states.
		/// </summary>
		private byte[][] DynamicStates = new byte[ DynamicStatesCount ][];

		/// <summary>
		/// The salt used for encryption.
		/// </summary>
		private byte[] EncryptionSalt;

		/// <summary>
		/// The key used for encryption.
		/// </summary>
		private byte[] EncryptionKey;

		/// <summary>
		/// The counter used to periodically update the encryption salt and key with data from <see cref="DynamicStates"/>.
		/// </summary>
		private int EncryptionSaltKeyUpdateCounter;

		#endregion Private fields

		#region Public non-serializable properties

		/// <summary>
		/// The size (in bytes) required to merge the current algorithm with data from another. READ-ONLY.
		/// </summary>
		public override int MergingStateSize
		{
			get { return PersistentState.Length; }
		}

		#endregion Public non-serializable properties

		#region Private serializable properties

		private byte[] _PersistentState;
		/// <summary>
		/// The persistent state.
		/// </summary>
		[SerializeAttribute]
		protected byte[] PersistentState
		{
			get { return _PersistentState; }
			set
			{
				ValidationUtils.ArrayHasContent( value, "PersistentState" );

				_PersistentState = value;
			}
		}

		#endregion Private serializable properties

		#region Public serializable properties

		private HashAlg _HashAlg;
		/// <summary>
		/// Hash algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public HashAlg HashAlg
		{
			get { return _HashAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "HashAlg" );

				_HashAlg = value;
			}
		}

		private SymmAlg _SymmAlg;
		/// <summary>
		/// Symmetric encryption algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public SymmAlg SymmAlg
		{
			get { return _SymmAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "SymmAlg" );

				_SymmAlg = value;
			}
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Validates the current object.
		/// </summary>
		protected override void Validate()
		{
			int hashAlgSize = HashAlg.AlgSizeBytes;
			if( PersistentState.Length != hashAlgSize )
			{
				throw new ValidationException( string.Format( "The persistent state size is {0} not {1}.", PersistentState.Length
					, hashAlgSize ) );
			}

			if( SymmAlg.Padding != SymmPaddings.None )
			{
				throw new ValidationException( string.Format( "The padding of the symmetric encryption algorithm must be"
					+ " \"{1}\", else the algorithm might call back the entropy algorithm.", SymmPaddings.None ) );
			}

			if( ProgressionCounter.Length > SymmAlg.AlgSizeBytes )
			{
				throw new ValidationException( string.Format( "The progression counter size is {0} bytes but the maximum allowed"
					+ " is {1} bytes.", ProgressionCounter.Length, SymmAlg.AlgSizeBytes ) );
			}
		}

		/// <summary>
		/// Sets up the state of the algorithm.
		/// </summary>
		/// <remarks>
		/// <para>The valid serializable properties must not be modified in here because that would cause an entropy algorithm to
		/// fail to be reloaded from a storehouse.</para>
		/// </remarks>
		protected override void SetupState()
		{
			AdvanceStateAfterBlocks = AdvanceStateAfterBytes / SymmAlg.AlgSizeBytes;

			// The counter should never be (used) 0.
			int counter = 0;

			if( BufferUtils.NoContent( PersistentState ) )
			{
				counter++;
				PersistentState = HashAlg.Compute( null, true, CryptoUtils.GetSystemEntropy( true )
					, BitConverter.GetBytes( counter ) );
			}

			for( int i = 0; i < DynamicStatesCount; i++ )
			{
				counter++;
				DynamicStates[ i ] = HashAlg.Compute( null, true, PersistentState, BitConverter.GetBytes( counter ) );
			}

			// The salt and the key must be generated in a different way.
			counter++;
			EncryptionSalt = HashAlg.Compute( null, true, PersistentState, DynamicStates[ 0 ]
				, BitConverter.GetBytes( counter ) );

			// The salt and the key must be generated in a different way.
			counter++;
			EncryptionKey = HashAlg.Compute( null, true, PersistentState, DynamicStates[ DynamicStatesCount - 1 ]
				, EncryptionSalt, BitConverter.GetBytes( counter ) );

			EncryptionSalt = SymmAlg.AlignToAlgSize( EncryptionSalt );
			EncryptionKey = SymmAlg.AlignToAlgSize( EncryptionKey );
		}

		/// <summary>
		/// Advances the state of the algorithm.
		/// </summary>
		/// <param name="buffer">Buffer to add to the algorithm's state. This may be null or empty.</param>
		/// <remarks>
		/// <para><see cref="SymmAlg"/> can't be used in here since this method is called while an encryption stream worker is
		/// active (see the <see cref="GetBytes"/> method).</para>
		/// </remarks>
		private void AdvanceState( byte[] buffer )
		{
			int hashAlgSize = HashAlg.AlgSizeBytes;

			// The counter should never be (used) 0.
			int counter = 0;

			// Compute the hash of the new data.
			counter++;
			byte[] newDataHash = BufferUtils.HasContent( buffer )
				? HashAlg.Compute( null, true, CryptoUtils.GetSystemEntropy(), buffer, BitConverter.GetBytes( counter ) )
				: HashAlg.Compute( null, true, CryptoUtils.GetSystemEntropy(), BitConverter.GetBytes( counter ) );

			counter++;
			PersistentState = HashAlg.Compute( null, true, PersistentState, newDataHash, BitConverter.GetBytes( counter ) );

			for( int i = 0; i < DynamicStatesCount; i++ )
			{
				counter++;
				DynamicStates[ i ] = HashAlg.Compute( null, true, PersistentState, DynamicStates[ i ], newDataHash
					, BitConverter.GetBytes( counter ) );
			}

			// The salt and the key must be generated in a different way.
			counter++;
			EncryptionSalt = HashAlg.Compute( null, true, PersistentState, EncryptionSalt, EncryptionKey
				, BitConverter.GetBytes( counter ) );

			// The salt and the key must be generated in a different way.
			counter++;
			EncryptionKey = HashAlg.Compute( null, true, PersistentState, EncryptionKey, EncryptionSalt
				, BitConverter.GetBytes( counter ) );

			// Periodically change the encryption salt and key with data from the dynamic states.
			for( int i = 0; i < DynamicStatesCount; i++ )
			{
				if( EncryptionSaltKeyUpdateCounter % (1 << (i + 1)) == 0 )
				{
					counter++;
					EncryptionSalt = HashAlg.Compute( null, true, DynamicStates[ i ], EncryptionSalt
						, BitConverter.GetBytes( counter ) );

					counter++;
					EncryptionKey = HashAlg.Compute( null, true, DynamicStates[ i ], EncryptionKey
						, BitConverter.GetBytes( counter ) );
				}
			}

			unchecked { EncryptionSaltKeyUpdateCounter++; }

			EncryptionSalt = SymmAlg.AlignToAlgSize( EncryptionSalt );
			EncryptionKey = SymmAlg.AlignToAlgSize( EncryptionKey );
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Adds the specified data into the state.
		/// </summary>
		/// <param name="buffer">Buffer to add.</param>
		public override void Add( byte[] buffer )
		{
			AdvanceState( buffer );
		}

		/// <summary>
		/// Returns an array of random bytes.
		/// </summary>
		/// <param name="size">The number of bytes to return in the array. This must be at least 1.</param>
		/// <returns>Returns an array of random bytes.</returns>
		public override byte[] GetBytes( int size )
		{
			ValidationUtils.Integer( size, 1, int.MaxValue, "size" );

			byte[] result = new byte[ size ];

			// The encryption worker used to generate entropy.
			SymmAlgStreamWorker encryptionWorker = null;

			int processedSize = 0;
			while( processedSize < size )
			{
				// Advance the state periodically. Fortuna recommends to perform a reseed for every request.
				if( (BufferUtils.FragmentModulo( ProgressionCounter, (UInt64)AdvanceStateAfterBlocks ) == 0)
					|| (encryptionWorker == null) )
				{
					AdvanceState( null );

					encryptionWorker = SymmAlg.CreateEncryptionStreamWorker( EncryptionSalt, EncryptionKey );
				}

				// Increment before encryption because the counter should never be (used) 0.
				BufferUtils.Increment( ProgressionCounter, true );

				// Generate the entropy. The persistent / dynamic / block state may not be extractable from the returned data.
				byte[] blockCounterBuffer = SymmAlg.AlignToAlgSize( ProgressionCounter );
				byte[] exposedData = encryptionWorker.Process( blockCounterBuffer );

				// Copy the hash in the result buffer.
				int resultBlockSize = Math.Min( exposedData.Length, size - processedSize );
				Buffer.BlockCopy( exposedData, 0, result, processedSize, resultBlockSize );

				processedSize += resultBlockSize;
			}

			return result;
		}

		#endregion Public methods
	}
}
