////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;

namespace Vela.Crypto.Hash
{
	/// <summary>
	/// Contract for a cryptographic hash algorithm.
	/// </summary>
	public abstract class HashAlg : IImmutableAlgorithm, IExpirableAlgorithm, IProgressionAlgorithm
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected HashAlg()
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The size (in bits) of the hash algorithm. The resulted hash has to have this size. READ-ONLY.
		/// </summary>
		public abstract int AlgSizeBits { get; }

		/// <summary>
		/// The size (in bytes) of the hash algorithm. The resulted hash has to have this size. READ-ONLY.
		/// </summary>
		public int AlgSizeBytes
		{
			get { return AlgSizeBits / 8; }
		}

		/// <summary>
		/// The size (in bytes) of a super-block. READ-ONLY.
		/// </summary>
		public abstract int SuperBlockSizeBytes { get; }

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public abstract DateTime ExpirationTime { get; }

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Creates a stream worker for the current algorithm.
		/// </summary>
		public abstract HashAlgStreamWorker CreateStreamWorker();

		/// <summary>
		/// Computes the hash of the specified data.
		/// </summary>
		/// <param name="buffer">Buffer to hash. This may be empty, but not null.</param>
		/// <param name="offset">The offset in the data to hash from where to start.</param>
		/// <param name="size">The size (in bytes) of the data to hash, starting from the specified offset.</param>
		/// <returns>Returns the computed hash.</returns>
		public byte[] Compute( byte[] buffer, int offset, int size )
		{
			HashAlgStreamWorker worker = CreateStreamWorker();

			worker.Process( buffer, offset, size );

			return worker.ProcessFinal();
		}

		/// <summary>
		/// Computes the hash of the specified data.
		/// </summary>
		/// <param name="buffer">Buffer to hash. This may be empty, but not null.</param>
		/// <returns>Returns the computed hash.</returns>
		public byte[] Compute( byte[] buffer )
		{
			return Compute( buffer, 0, buffer.Length );
		}

		/// <summary>
		/// Computes the hash of the specified datas.
		/// </summary>
		/// <param name="textEncoding">Text encoding. If this is null, <see cref="KernelUtils.GetDeepTextEncoding"/> is used
		/// instead. This is used only if "datas" contains texts.</param>
		/// <param name="includeDataSize">If this is true then the size of the data is hashed-in.</param>
		/// <param name="datas">List of values whose hash to compute. An item must be a scalar.</param>
		/// <returns>Returns the hash of the specified datas.</returns>
		public byte[] Compute( TextEncoding textEncoding, bool includeDataSize, params object[] datas )
		{
			ValidationUtils.ArrayHasContent( datas, "datas" );

			if( textEncoding == null ) textEncoding = KernelUtils.GetDeepTextEncoding();

			HashAlgStreamWorker worker = CreateStreamWorker();

			int count = datas.Length;
			for( int i = 0; i < count; i++ )
			{
				object item = datas[ i ];
				ValidationUtils.ObjectHasContent( item, string.Format( "datas[ {0} ]", i ) );

				// Convert the item value to a byte array.
				byte[] dataBuffer = ScalarEncoder.ValueToBuffer( item, textEncoding, true );

				// Hash-in the data size.
				if( includeDataSize )
				{
					byte[] dataSizeBuffer = BitConverter.GetBytes( (Int64)dataBuffer.Length );
					worker.Process( dataSizeBuffer );
				}

				// Hash-in the data.
				worker.Process( dataBuffer );
			}

			return worker.ProcessFinal();
		}

		/// <summary>
		/// Returns an array of bytes, as a progression of the current algorithm.
		/// </summary>
		/// <param name="seed">The seed from which to start the progression. This is incremented in here.</param>
		/// <param name="size">The number of bytes to return in the array. This must be at least 1.</param>
		/// <returns>Returns an array of bytes, as a progression of the current algorithm.</returns>
		/// <remarks>
		/// <para>The specified seed is incremented and then progressed, iteratively, until the specified size is bytes is
		/// generated.</para>
		/// </remarks>
		public byte[] GetProgressionBytes( byte[] seed, int size )
		{
			ValidationUtils.ArrayHasContent( seed, "seed" );
			ValidationUtils.Integer( size, 1, int.MaxValue, "size" );

			byte[] result = new byte[ size ];

			int processedSize = 0;
			while( processedSize < size )
			{
				// Increment before hashing because the seed should not be 0.
				BufferUtils.Increment( seed, 0, seed.Length, true );

				HashAlgStreamWorker worker = CreateStreamWorker();
				worker.Process( seed );
				byte[] hash = worker.ProcessFinal();

				// Copy the hash in the result buffer.
				int resultBlockSize = Math.Min( hash.Length, size - processedSize );
				Buffer.BlockCopy( hash, 0, result, processedSize, resultBlockSize );

				processedSize += resultBlockSize;
			}

			return result;
		}

		#endregion Public methods
	}
}
