////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Crypto.Asymmetric;
using Vela.Crypto.Hash;
using Vela.Crypto.Mac;
using Vela.Crypto.Symmetric;
using Vela.Kernel;

namespace Vela.Crypto
{
	/// <summary>
	/// Pool of cryptographic algorithms.
	/// </summary>
	public interface ICryptoAlgorithmPool
	{
		#region Public methods

		/// <summary>
		/// Gets the entropy algorithm.
		/// </summary>
		/// <returns>Returns the entropy algorithm.</returns>
		EntropyAlg GetEntropyAlg();

		/// <summary>
		/// Gets the hash algorithm.
		/// </summary>
		/// <returns>Returns the hash algorithm.</returns>
		HashAlg GetHashAlg();

		/// <summary>
		/// Gets the MAC algorithm.
		/// </summary>
		/// <returns>Returns the MAC algorithm.</returns>
		MacAlg GetMacAlg();

		/// <summary>
		/// Gets the symmetric encryption padding.
		/// </summary>
		/// <returns>Returns the symmetric encryption padding.</returns>
		SymmPaddings GetSymmPadding();

		/// <summary>
		/// Gets the symmetric encryption chaining.
		/// </summary>
		/// <returns>Returns the symmetric encryption chaining.</returns>
		SymmChainings GetSymmChaining();

		/// <summary>
		/// Gets the symmetric encryption algorithm.
		/// </summary>
		/// <returns>Returns the symmetric encryption algorithm.</returns>
		SymmAlg GetSymmAlg();

		/// <summary>
		/// Gets the asymmetric encryption padding.
		/// </summary>
		/// <returns>Returns the asymmetric encryption padding.</returns>
		IAsymmEncryptionPaddingAlg GetAsymmEncryptionPadding();

		/// <summary>
		/// Gets the asymmetric signing padding.
		/// </summary>
		/// <returns>Returns the asymmetric signing padding.</returns>
		IAsymmSigningPaddingAlg GetAsymmSigningPadding();

		/// <summary>
		/// Gets the algorithm used for asymmetric encryption.
		/// </summary>
		/// <param name="algorithmSecurityLevel">The security level for the cryptographic algorithms to use.</param>
		/// <returns>Returns the algorithm used for asymmetric encryption.</returns>
		/// <remarks>
		/// <para>You must later call the <see cref="AsymmAlg.GenerateKeyPair"/> method for the returned algorithm.</para>
		/// </remarks>
		AsymmAlg GetAsymmAlg( AlgorithmSecurityLevel algorithmSecurityLevel );

		/// <summary>
		/// Generates the algorithm used for asymmetric encryption.
		/// </summary>
		/// <param name="algorithmSecurityLevel">The security level for the cryptographic algorithms to use.</param>
		/// <returns>Returns the generated algorithm used for asymmetric encryption.</returns>
		/// <param name="lop">Long operation progress used during the generation of a key pair. This may be null.</param>
		AsymmAlg GenerateAsymmAlg( AlgorithmSecurityLevel algorithmSecurityLevel, LongOperationProgress lop );

		#endregion Public methods
	}
}
