////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Crypto.Hash;
using Vela.Kernel;

namespace Vela.Crypto.Mac
{
	/// <summary>
	/// The stream worker for the <see cref="Bck"/> MAC algorithm.
	/// </summary>
	public class BckAlgStreamWorker : MacAlgStreamWorker
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="macAlg">Parent MAC algorithm.</param>
		/// <param name="key">The key of the MAC.</param>
		public BckAlgStreamWorker( MacAlg macAlg, byte[] key )
			: base( macAlg, key )
		{
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// Hash stream worker.
		/// </summary>
		private HashAlgStreamWorker HashStreamWorker;

		#endregion Private fields

		#region Private methods

		/// <summary>
		/// Sets the key with the specified value.
		/// </summary>
		/// <param name="key">The key of the MAC.</param>
		protected override void SetKey( byte[] key )
		{
			ValidationUtils.ArrayHasContent( key, "key" );

			HashAlg hashAlg = MacAlg.HashAlg;

			if( key.Length == hashAlg.SuperBlockSizeBytes ) Key = key;

			// The key must not exceed the super-block size.
			if( key.Length > hashAlg.SuperBlockSizeBytes ) Key = hashAlg.Compute( key );

			// Align the key to the super-block size.
			if( key.Length < hashAlg.SuperBlockSizeBytes ) Key = BufferUtils.PadEndWithZeros( key, hashAlg.SuperBlockSizeBytes );

			HashStreamWorker = hashAlg.CreateStreamWorker();

			// Hash-in the inner key.
			byte[] innerKeyPad = BufferUtils.GetInitialized( hashAlg.SuperBlockSizeBytes, 0x36 );
			byte[] innerKey = BufferUtils.ContiguousXor( innerKeyPad, Key );
			HashStreamWorker.Process( innerKey );
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Processes the specified data. The <see cref="ProcessFinal"/> method must be called to finish processing the data.
		/// </summary>
		/// <param name="buffer">Buffer to process. This may be empty, but not null.</param>
		/// <param name="offset">Offset in the input data, from where to start processing. This must be at least 0.</param>
		/// <param name="size">The number of bytes to process. This must be at least 0.</param>
		public override void Process( byte[] buffer, int offset, int size )
		{
			HashStreamWorker.Process( buffer, offset, size );
		}

		/// <summary>
		/// Finishes processing the data.
		/// </summary>
		/// <returns>Returns the result of the processing.</returns>
		public override byte[] ProcessFinal()
		{
			// Compute the inner hash.
			byte[] hash = HashStreamWorker.ProcessFinal();

			HashAlg hashAlg = MacAlg.HashAlg;

			#region Outer hash

			// Re-initialize the worker for a new hash.
			HashStreamWorker = hashAlg.CreateStreamWorker();

			// Hash-in the outer key.
			byte[] outerKeyPad = BufferUtils.GetInitialized( hashAlg.SuperBlockSizeBytes, 0x5C );
			byte[] outerKey = BufferUtils.ContiguousXor( outerKeyPad, Key );
			HashStreamWorker.Process( outerKey );

			// Hash-in the inner hash.
			HashStreamWorker.Process( hash );

			// Compute the outer hash.
			hash = HashStreamWorker.ProcessFinal();

			#endregion Outer hash

			// Make the worker unusable.
			MacAlg = null;

			return hash;
		}

		#endregion Public methods
	}
}
