////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto.Hash;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.Crypto.Mac
{
	/// <summary>
	/// Contract for a cryptographic MAC algorithm based on a hash algorithm.
	/// </summary>
	public abstract class MacAlg : IImmutableAlgorithm, IExpirableAlgorithm
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected MacAlg()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="hashAlg">Hash algorithm.</param>
		protected MacAlg( HashAlg hashAlg )
		{
			HashAlg = hashAlg;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The size (in bits) of the MAC algorithm. The resulted MAC has to have this size. READ-ONLY.
		/// </summary>
		public int AlgSizeBits
		{
			get { return HashAlg.AlgSizeBits; }
		}

		/// <summary>
		/// The size (in bytes) of the MAC algorithm. The resulted MAC has to have this size. READ-ONLY.
		/// </summary>
		public int AlgSizeBytes
		{
			get { return HashAlg.AlgSizeBytes; }
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public abstract DateTime ExpirationTime { get; }

		#endregion Public non-serializable properties

		#region Serializable properties

		private HashAlg _HashAlg;
		/// <summary>
		/// Hash algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public HashAlg HashAlg
		{
			get { return _HashAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "HashAlg" );

				_HashAlg = value;
			}
		}

		#endregion Serializable properties

		#region Public methods

		/// <summary>
		/// Creates a stream worker for the current algorithm.
		/// </summary>
		/// <param name="key">The key of the MAC.</param>
		public abstract MacAlgStreamWorker CreateStreamWorker( byte[] key );

		/// <summary>
		/// Computes the MAC of the specified datas.
		/// </summary>
		/// <param name="key">The key of the MAC.</param>
		/// <param name="textEncoding">Text encoding. If this is null, <see cref="KernelUtils.GetDeepTextEncoding"/> is used
		/// instead. This is used only if "datas" contains texts.</param>
		/// <param name="includeDataSize">If this is true then the size of the data is hashed-in.</param>
		/// <param name="datas">List of values whose MAC to compute. An item must be a scalar.</param>
		/// <returns>Returns the MAC of the specified datas.</returns>
		public byte[] Compute( byte[] key, TextEncoding textEncoding, bool includeDataSize, params object[] datas )
		{
			ValidationUtils.ArrayHasContent( datas, "datas" );

			if( textEncoding == null ) textEncoding = KernelUtils.GetDeepTextEncoding();

			MacAlgStreamWorker worker = CreateStreamWorker( key );

			int count = datas.Length;
			for( int i = 0; i < count; i++ )
			{
				object item = datas[ i ];
				ValidationUtils.ObjectHasContent( item, string.Format( "datas[ {0} ]", i ) );

				// Convert the item value to a byte array.
				byte[] dataBuffer = ScalarEncoder.ValueToBuffer( item, textEncoding, true );

				// Hash-in the data size.
				if( includeDataSize )
				{
					byte[] dataSizeBuffer = BitConverter.GetBytes( (Int64)dataBuffer.Length );
					worker.Process( dataSizeBuffer );
				}

				// Hash-in the data.
				worker.Process( dataBuffer );
			}

			return worker.ProcessFinal();
		}

		#endregion Public methods
	}
}
