////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;

namespace Vela.Crypto.Mac
{
	/// <summary>
	/// Contract for the stream worker of a MAC algorithm.
	/// </summary>
	public abstract class MacAlgStreamWorker
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="macAlg">Parent MAC algorithm.</param>
		/// <param name="key">The key of the MAC.</param>
		protected MacAlgStreamWorker( MacAlg macAlg, byte[] key )
		{
			ValidationUtils.ObjectHasContent( macAlg, "macAlg" );

			MacAlg = macAlg;

			SetKey( key );
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// MAC algorithm.
		/// </summary>
		protected MacAlg MacAlg;

		/// <summary>
		/// The key.
		/// </summary>
		protected byte[] Key;

		#endregion Private fields

		#region Private methods

		/// <summary>
		/// Sets the key with the specified value.
		/// </summary>
		/// <param name="key">The key of the MAC.</param>
		protected abstract void SetKey( byte[] key );

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Processes the specified data. The <see cref="ProcessFinal"/> method must be called to finish processing the data.
		/// </summary>
		/// <param name="buffer">Buffer to process. This may be empty, but not null.</param>
		/// <param name="offset">Offset in the input data, from where to start processing. This must be at least 0.</param>
		/// <param name="size">The number of bytes to process. This must be at least 0.</param>
		public abstract void Process( byte[] buffer, int offset, int size );

		/// <summary>
		/// Processes the specified data. The <see cref="ProcessFinal"/> method must be called to finish processing the data.
		/// </summary>
		/// <param name="buffer">Buffer to process. This may be empty, but not null.</param>
		public void Process( byte[] buffer )
		{
			Process( buffer, 0, buffer.Length );
		}

		/// <summary>
		/// Finishes processing the data.
		/// </summary>
		/// <returns>Returns the result of the processing.</returns>
		public abstract byte[] ProcessFinal();

		#endregion Public methods
	}
}
