////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;

namespace Vela.Crypto.Symmetric
{
	/// <summary>
	/// A Rijndael256 symmetric cryptographic algorithm
	/// (<see href="http://csrc.nist.gov/publications/fips/fips197/fips-197.pdf">FIPS197</see>and <see
	/// href="http://csrc.nist.gov/CryptoToolkit/aes/rijndael/Rijndael-ammended.pdf">Rijndael</see>).
	/// </summary>
	/// <remarks>
	/// <para>The entropy algorithm is not used if the <see cref="SymmPaddings.None"/> padding is used.</para>
	/// </remarks>
	public class Rijndael256 : SymmAlg
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Rijndael256()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="padding">The padding of the encrypted data.</param>
		/// <param name="chaining">The chaining of the encrypted data.</param>
		public Rijndael256( SymmPaddings padding, SymmChainings chaining )
			: base( padding, chaining )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The size (in bits) of the symmetric cryptographic algorithm. The input and output data blocks must have this
		/// size. READ-ONLY.
		/// </summary>
		public override int AlgSizeBits
		{
			get { return 256; }
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return new DateTime( 2040, 1, 1 ); }
		}

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Creates an encryption stream worker for the current algorithm.
		/// </summary>
		/// <param name="padding">The padding.</param>
		/// <param name="chaining">The chaining.</param>
		/// <param name="salt">The salt. It must have the same size as the algorithm.</param>
		/// <param name="key">The encryption key. It must have the same size as the algorithm.</param>
		/// <returns>Returns an encryption stream worker for the current algorithm.</returns>
		public override SymmAlgStreamWorker CreateEncryptionStreamWorker( SymmPaddings padding, SymmChainings chaining
			, byte[] salt, byte[] key )
		{
			return new Rijndael256StreamWorker( this, padding, chaining, salt, key, true );
		}

		/// <summary>
		/// Creates a decryption stream worker for the current algorithm.
		/// </summary>
		/// <param name="padding">The padding.</param>
		/// <param name="chaining">The chaining.</param>
		/// <param name="salt">The salt. It must have the same size as the algorithm.</param>
		/// <param name="key">The encryption key. It must have the same size as the algorithm.</param>
		/// <returns>Returns a decryption stream worker for the current algorithm.</returns>
		public override SymmAlgStreamWorker CreateDecryptionStreamWorker( SymmPaddings padding, SymmChainings chaining
			, byte[] salt, byte[] key )
		{
			return new Rijndael256StreamWorker( this, padding, chaining, salt, key, false );
		}

		#endregion Public methods
	}
}
