////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.Crypto.Symmetric
{
	/// <summary>
	/// Contract for a symmetric cryptographic algorithm.
	/// </summary>
	/// <remarks>
	/// <para>The entropy algorithm is not used if the <see cref="SymmPaddings.None"/> padding is used.</para>
	/// </remarks>
	public abstract class SymmAlg : IImmutableAlgorithm, IExpirableAlgorithm
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected SymmAlg()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="padding">The padding.</param>
		/// <param name="chaining">The chaining.</param>
		public SymmAlg( SymmPaddings padding, SymmChainings chaining )
		{
			Padding = padding;
			Chaining = chaining;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The size (in bits) of the symmetric cryptographic algorithm. The input and output data blocks must have this
		/// size. READ-ONLY.
		/// </summary>
		public abstract int AlgSizeBits { get; }

		/// <summary>
		/// The size (in bytes) of the symmetric cryptographic algorithm. The input and output data blocks must have this
		/// size. READ-ONLY.
		/// </summary>
		public int AlgSizeBytes
		{
			get { return AlgSizeBits >> 3; }
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public abstract DateTime ExpirationTime { get; }

		#endregion Public non-serializable properties

		#region Public serializable properties

		private SymmPaddings _Padding;
		/// <summary>
		/// The padding. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public SymmPaddings Padding
		{
			get { return _Padding; }
			protected set { _Padding = value; }
		}

		private SymmChainings _Chaining;
		/// <summary>
		/// The chaining. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public SymmChainings Chaining
		{
			get { return _Chaining; }
			protected set { _Chaining = value; }
		}

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Returns the size in blocks (of algorithm size) of the specified buffer. If the buffer is not aligned to
		/// algorithm size, an exception is thrown.
		/// </summary>
		/// <param name="buffer">Buffer buffer whose size in blocks to get.</param>
		/// <returns>Returns the size in blocks (of algorithm size) of the specified buffer.</returns>
		public int GetBlockCount( byte[] buffer )
		{
			if( buffer.Length % AlgSizeBytes != 0 )
				throw new ValidationException( "The specified buffer is not aligned at algorithm size." );

			return buffer.Length / AlgSizeBytes;
		}

		/// <summary>
		/// Aligns the specified data to the size of the current algorithm.
		/// </summary>
		/// <param name="buffer">Buffer to align.</param>
		/// <returns>Returns the aligned data.</returns>
		public byte[] AlignToAlgSize( byte[] buffer )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );

			int dataSize = buffer.Length;

			// Check if the source data is already aligned.
			if( dataSize == AlgSizeBytes )
			{
				return buffer;
			}
			else
			{
				byte[] alignedBuffer = new byte[ AlgSizeBytes ];
				BufferUtils.ContiguousCopy( buffer, alignedBuffer );

				return alignedBuffer;
			}
		}

		/// <summary>
		/// Creates an encryption stream worker for the current algorithm.
		/// </summary>
		/// <param name="padding">The padding.</param>
		/// <param name="chaining">The chaining.</param>
		/// <param name="salt">The salt. It must have the same size as the algorithm.</param>
		/// <param name="key">The encryption key. It must have the same size as the algorithm.</param>
		/// <returns>Returns an encryption stream worker for the current algorithm.</returns>
		public abstract SymmAlgStreamWorker CreateEncryptionStreamWorker( SymmPaddings padding, SymmChainings chaining
			, byte[] salt, byte[] key );

		/// <summary>
		/// Creates an encryption stream worker for the current algorithm.
		/// </summary>
		/// <param name="salt">The salt. It must have the same size as the algorithm.</param>
		/// <param name="key">The encryption key. It must have the same size as the algorithm.</param>
		/// <returns>Returns an encryption stream worker for the current algorithm.</returns>
		public SymmAlgStreamWorker CreateEncryptionStreamWorker( byte[] salt, byte[] key )
		{
			return CreateEncryptionStreamWorker( Padding, Chaining, salt, key );
		}

		/// <summary>
		/// Creates a decryption stream worker for the current algorithm.
		/// </summary>
		/// <param name="padding">The padding.</param>
		/// <param name="chaining">The chaining.</param>
		/// <param name="salt">The salt. It must have the same size as the algorithm.</param>
		/// <param name="key">The encryption key. It must have the same size as the algorithm.</param>
		/// <returns>Returns a decryption stream worker for the current algorithm.</returns>
		public abstract SymmAlgStreamWorker CreateDecryptionStreamWorker( SymmPaddings padding, SymmChainings chaining
			, byte[] salt, byte[] key );

		/// <summary>
		/// Creates a decryption stream worker for the current algorithm.
		/// </summary>
		/// <param name="salt">The salt. It must have the same size as the algorithm.</param>
		/// <param name="key">The encryption key. It must have the same size as the algorithm.</param>
		/// <returns>Returns a decryption stream worker for the current algorithm.</returns>
		public SymmAlgStreamWorker CreateDecryptionStreamWorker( byte[] salt, byte[] key )
		{
			return CreateDecryptionStreamWorker( Padding, Chaining, salt, key );
		}

		/// <summary>
		/// Encrypts the specified data.
		/// </summary>
		/// <param name="buffer">Buffer to process.</param>
		/// <param name="offset">Offset in the input data, from where to start processing.</param>
		/// <param name="count">The number of bytes to process.</param>
		/// <param name="padding">The padding.</param>
		/// <param name="chaining">The chaining.</param>
		/// <param name="salt">The salt. It must have the same size as the algorithm.</param>
		/// <param name="key">The encryption key. It must have the same size as the algorithm.</param>
		/// <returns>Returns the result of the processing.</returns>
		public byte[] Encrypt( byte[] buffer, int offset, int count, SymmPaddings padding, SymmChainings chaining, byte[] salt
			, byte[] key )
		{
			SymmAlgStreamWorker worker = CreateEncryptionStreamWorker( padding, chaining, salt, key );

			byte[] result = worker.ProcessFinal( buffer, offset, count );

			if( (result.Length % AlgSizeBytes != 0)
				|| ((result.Length < buffer.Length - offset) && (result.Length > buffer.Length - offset + AlgSizeBytes)) )
			{
				throw new ValidationException( string.Format( "The size of the encrypted data must be multiple of {0} bytes and"
						+ " between [{1}...{2}], not {3}."
					, AlgSizeBytes, buffer.Length - offset, buffer.Length - offset + AlgSizeBytes, result.Length ) );
			}

			return result;
		}

		/// <summary>
		/// Encrypts the specified data.
		/// </summary>
		/// <param name="buffer">Buffer to process.</param>
		/// <param name="offset">Offset in the input data, from where to start processing.</param>
		/// <param name="count">The number of bytes to process.</param>
		/// <param name="salt">The salt. It must have the same size as the algorithm.</param>
		/// <param name="key">The encryption key. It must have the same size as the algorithm.</param>
		/// <returns>Returns the result of the processing.</returns>
		public byte[] Encrypt( byte[] buffer, int offset, int count, byte[] salt, byte[] key )
		{
			return Encrypt( buffer, offset, count, Padding, Chaining, salt, key );
		}

		/// <summary>
		/// Encrypts the specified data.
		/// </summary>
		/// <param name="buffer">Buffer to process.</param>
		/// <param name="padding">The padding.</param>
		/// <param name="chaining">The chaining.</param>
		/// <param name="salt">The salt. It must have the same size as the algorithm.</param>
		/// <param name="key">The encryption key. It must have the same size as the algorithm.</param>
		/// <returns>Returns the result of the processing.</returns>
		public byte[] Encrypt( byte[] buffer, SymmPaddings padding, SymmChainings chaining, byte[] salt, byte[] key )
		{
			return Encrypt( buffer, 0, buffer.Length, padding, chaining, salt, key );
		}

		/// <summary>
		/// Encrypts the specified data.
		/// </summary>
		/// <param name="buffer">Buffer to process.</param>
		/// <param name="salt">The salt. It must have the same size as the algorithm.</param>
		/// <param name="key">The encryption key. It must have the same size as the algorithm.</param>
		/// <returns>Returns the result of the processing.</returns>
		public byte[] Encrypt( byte[] buffer, byte[] salt, byte[] key )
		{
			return Encrypt( buffer, 0, buffer.Length, Padding, Chaining, salt, key );
		}

		/// <summary>
		/// Decrypts the specified data.
		/// </summary>
		/// <param name="buffer">Buffer to process.</param>
		/// <param name="offset">Offset in the input data, from where to start processing.</param>
		/// <param name="count">The number of bytes to process.</param>
		/// <param name="padding">The padding.</param>
		/// <param name="chaining">The chaining.</param>
		/// <param name="salt">The salt. It must have the same size as the algorithm.</param>
		/// <param name="key">The encryption key. It must have the same size as the algorithm.</param>
		/// <returns>Returns the result of the processing.</returns>
		public byte[] Decrypt( byte[] buffer, int offset, int count, SymmPaddings padding, SymmChainings chaining, byte[] salt
			, byte[] key )
		{
			SymmAlgStreamWorker worker = CreateDecryptionStreamWorker( padding, chaining, salt, key );

			byte[] result = worker.ProcessFinal( buffer, offset, count );

			if( (result.Length < buffer.Length - offset - AlgSizeBytes) || (result.Length > buffer.Length - offset) )
			{
				throw new ValidationException( string.Format( "The size of the decrypted data must be between [{0}...{1}] bytes,"
					+ " not {2}.", buffer.Length - offset - AlgSizeBytes, buffer.Length - offset, result.Length ) );
			}

			return result;
		}

		/// <summary>
		/// Decrypts the specified data.
		/// </summary>
		/// <param name="buffer">Buffer to process.</param>
		/// <param name="offset">Offset in the input data, from where to start processing.</param>
		/// <param name="count">The number of bytes to process.</param>
		/// <param name="salt">The salt. It must have the same size as the algorithm.</param>
		/// <param name="key">The encryption key. It must have the same size as the algorithm.</param>
		/// <returns>Returns the result of the processing.</returns>
		public byte[] Decrypt( byte[] buffer, int offset, int count, byte[] salt, byte[] key )
		{
			return Decrypt( buffer, offset, count, Padding, Chaining, salt, key );
		}

		/// <summary>
		/// Decrypts the specified data.
		/// </summary>
		/// <param name="buffer">Buffer to process.</param>
		/// <param name="padding">The padding.</param>
		/// <param name="chaining">The chaining.</param>
		/// <param name="salt">The salt. It must have the same size as the algorithm.</param>
		/// <param name="key">The encryption key. It must have the same size as the algorithm.</param>
		/// <returns>Returns the result of the processing.</returns>
		public byte[] Decrypt( byte[] buffer, SymmPaddings padding, SymmChainings chaining, byte[] salt, byte[] key )
		{
			return Decrypt( buffer, 0, buffer.Length, padding, chaining, salt, key );
		}

		/// <summary>
		/// Decrypts the specified data.
		/// </summary>
		/// <param name="buffer">Buffer to process.</param>
		/// <param name="salt">The salt. It must have the same size as the algorithm.</param>
		/// <param name="key">The encryption key. It must have the same size as the algorithm.</param>
		/// <returns>Returns the result of the processing.</returns>
		public byte[] Decrypt( byte[] buffer, byte[] salt, byte[] key )
		{
			return Decrypt( buffer, 0, buffer.Length, Padding, Chaining, salt, key );
		}

		#endregion Public methods
	}
}
