////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;

namespace Vela.Crypto.Symmetric
{
	/// <summary>
	/// Contract for the stream worker of a symmetric cryptographic algorithm.
	/// </summary>
	/// <remarks>
	/// <para>The entropy algorithm is not used if the <see cref="SymmPaddings.None"/> padding is used.</para>
	/// </remarks>
	public abstract class SymmAlgStreamWorker
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="symmAlg">Parent symmetric cryptographic algorithm.</param>
		/// <param name="padding">The padding.</param>
		/// <param name="chaining">The chaining.</param>
		/// <param name="salt">The salt. It must have the same size as the algorithm.</param>
		/// <param name="key">The encryption key. It must have the same size as the algorithm.</param>
		/// <param name="isEncrypt">If this is true then the data will be encrypted, else it will be decrypted.</param>
		protected SymmAlgStreamWorker( SymmAlg symmAlg, SymmPaddings padding, SymmChainings chaining, byte[] salt
			, byte[] key, bool isEncrypt )
		{
			ValidationUtils.ObjectHasContent( symmAlg, "symmAlg" );

			SymmAlg = symmAlg;
			Padding = padding;
			Chaining = chaining;

			int algSizeBytes = SymmAlg.AlgSizeBytes;

			// Chaining buffers.
			ChainingBuffer = new byte[ algSizeBytes ];

			// Transform buffers.
			WorkBuffer = new byte[ algSizeBytes ];
			WorkOut = new byte[ algSizeBytes ];

			// Key expansion may use this, so set this before the key.
			IsEncrypt = isEncrypt;

			Salt = salt;
			Key = key;

			Buffer.BlockCopy( Salt, 0, ChainingBuffer, 0, algSizeBytes );
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// Symmetric cryptographic algorithm.
		/// </summary>
		protected SymmAlg SymmAlg;

		/// <summary>
		/// The padding.
		/// </summary>
		protected SymmPaddings Padding;

		/// <summary>
		/// The chaining.
		/// </summary>
		protected SymmChainings Chaining;

		/// <summary>
		/// If this is true then the data will be encrypted, else it will be decrypted.
		/// </summary>
		protected bool IsEncrypt = true;

		private byte[] ChainingBuffer;
		private byte[] WorkBuffer;
		private byte[] WorkOut;
		private bool IsLastBlock;

		#endregion Private fields

		#region Public non-serializable properties

		private byte[] _Salt;
		/// <summary>
		/// The salt. It must have the same size as the algorithm. READ-ONLY.
		/// </summary>
		public byte[] Salt
		{
			get { return _Salt; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "Salt" );

				if( value.Length != SymmAlg.AlgSizeBytes )
				{
					throw new ValidationException( string.Format( "The size of the salt is {0} instead of {1}."
						, value.Length, SymmAlg.AlgSizeBytes ) );
				}

				_Salt = value;
			}
		}

		private byte[] _Key;
		/// <summary>
		/// The encryption key. It must have the same size as the algorithm. READ-ONLY.
		/// </summary>
		public byte[] Key
		{
			get { return _Key; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "Key" );

				if( value.Length != SymmAlg.AlgSizeBytes )
				{
					throw new ValidationException( string.Format( "The size of the encryption key is {0} instead of {1}."
						, value.Length, SymmAlg.AlgSizeBytes ) );
				}

				_Key = value;

				ExpandKey();
			}
		}

		/// <summary>
		/// Returns whether the last block of data has to be kept. READ-ONLY.
		/// </summary>
		private bool KeepLastBlock
		{
			get { return !IsEncrypt && (Padding != SymmPaddings.None); }
		}

		#endregion Public non-serializable properties

		#region Private static methods

		/// <summary>
		/// Fills the specified buffer with random data.
		/// </summary>
		/// <param name="buffer">Buffer to fill with random data. This may be empty, but not null.</param>
		/// <param name="offset">Offset in the input data, from where to fill with random data. This must be at least 0.</param>
		/// <param name="count">The number of bytes to process. This must be at least 0.</param>
		private static void RandomFill( byte[] buffer, int offset, int count )
		{
			ValidationUtils.ObjectHasContent( buffer, "buffer" );
			ValidationUtils.Integer( offset, 0, (buffer.Length > 0) ? (buffer.Length - 1) : 0, "offset" );
			ValidationUtils.Integer( count, 0, buffer.Length - offset, "count" );

			if( count <= 0 )
				return;

			byte[] randomData = EntropyPool.GetBytes( count );
			Buffer.BlockCopy( randomData, 0, buffer, offset, count );
		}

		#endregion Private static methods

		#region Private methods

		/// <summary>
		/// Expands the encryption key.
		/// </summary>
		protected abstract void ExpandKey();

		/// <summary>
		/// Processes one block of data, with no chaining (also known as ECB). This method must be called with a valid block
		/// size for both input and output.
		/// </summary>
		/// <param name="input">Input data block.</param>
		/// <param name="output">Output data block.</param>
		protected abstract void ProcessBlock( byte[] input, byte[] output );

		/// <summary>
		/// Gets the size of the output for the specified input data size.
		/// </summary>
		/// <param name="intputSize">Size of the input data.</param>
		/// <returns>Returns the size of the output.</returns>
		private int GetTransformOutputSize( int intputSize )
		{
			if( KeepLastBlock )
				return intputSize + (IsLastBlock ? 0 : -SymmAlg.AlgSizeBytes);
			else
				return intputSize;
		}

		/// <summary>
		/// Processes the specified data, but only an integer number of blocks.
		/// </summary>
		/// <param name="inputBuffer">Buffer to process.</param>
		/// <param name="inputOffset"> This must be at least 0.</param>
		/// <param name="inputCount"> This must be at least 1.</param>
		/// <param name="outputBuffer"></param>
		/// <param name="outputOffset"> This must be at least 0.</param>
		/// <returns>Returns the number of processed bytes.</returns>
		private int TransformBlocks( byte[] inputBuffer, int inputOffset, int inputCount, byte[] outputBuffer
			, int outputOffset )
		{
			ValidationUtils.ArrayHasContent( inputBuffer, "inputBuffer" );
			ValidationUtils.Integer( inputOffset, 0, inputBuffer.Length - 1, "inputOffset" );
			ValidationUtils.Integer( inputCount, 1, inputBuffer.Length - inputOffset, "inputCount" );
			ValidationUtils.ArrayHasContent( outputBuffer, "outputBuffer" );

			int algSizeBytes = SymmAlg.AlgSizeBytes;

			int outputSize = GetTransformOutputSize( inputCount );
			ValidationUtils.Integer( outputOffset, 0, outputBuffer.Length - outputSize, "outputOffset" );

			if( BufferUtils.NoContent( Salt ) || BufferUtils.NoContent( Key ) )
				throw new ValidationException( "The state of the symmetric cryptographic algorithm was not set." );

			if( (inputCount % algSizeBytes) != 0 )
				throw new ValidationException( "Input buffer is not aligned to the algorithm size." );

			int offset = inputOffset;
			int fullBlockCount = (inputCount != algSizeBytes) ? (inputCount / algSizeBytes) : 1;

			if( KeepLastBlock )
				fullBlockCount--;

			int total = 0;

			if( IsLastBlock )
			{
				TransformBlockWithChaining( WorkBuffer, WorkOut );
				Buffer.BlockCopy( WorkOut, 0, outputBuffer, outputOffset, algSizeBytes );
				outputOffset += algSizeBytes;
				total += algSizeBytes;
				IsLastBlock = false;
			}

			for( int i = 0; i < fullBlockCount; i++ )
			{
				Buffer.BlockCopy( inputBuffer, offset, WorkBuffer, 0, algSizeBytes );
				TransformBlockWithChaining( WorkBuffer, WorkOut );
				Buffer.BlockCopy( WorkOut, 0, outputBuffer, outputOffset, algSizeBytes );
				offset += algSizeBytes;
				outputOffset += algSizeBytes;
				total += algSizeBytes;
			}

			if( KeepLastBlock )
			{
				Buffer.BlockCopy( inputBuffer, offset, WorkBuffer, 0, algSizeBytes );
				IsLastBlock = true;
			}

			return total;
		}

		/// <summary>
		/// Finishes processing the data for encryption.
		/// </summary>
		/// <param name="inputBuffer">Buffer to process. This may be empty, but not null.</param>
		/// <param name="inputOffset"> This must be at least 0.</param>
		/// <param name="inputCount"> This must be at least 0.</param>
		/// <returns>Returns the result of the processing. This may be empty, but not null.</returns>
		private byte[] TransformFinalEncrypt( byte[] inputBuffer, int inputOffset, int inputCount )
		{
			ValidationUtils.ObjectHasContent( inputBuffer, "inputBuffer" );
			ValidationUtils.Integer( inputOffset, 0, (inputBuffer.Length > 0) ? (inputBuffer.Length - 1) : 0, "inputOffset" );
			ValidationUtils.Integer( inputCount, 0, inputBuffer.Length - inputOffset, "inputCount" );

			int algSizeBytes = SymmAlg.AlgSizeBytes;

			// Check if there are there still full blocks to process.
			int fullBlockCount = inputCount / algSizeBytes;
			int alignedSize = fullBlockCount * algSizeBytes;
			int remaining = inputCount - alignedSize;
			int total = alignedSize;

			switch( Padding )
			{
				case SymmPaddings.Iso10126:
					// Add an extra block for padding.
					total += algSizeBytes;
					break;
				case SymmPaddings.None:
					if( inputCount == 0 )
						return new byte[ 0 ];
					if( remaining != 0 )
					{
						throw new ValidationException( "If no symmetric padding is done, the data must be aligned at algorithm"
							+ " size." );
					}
					break;
				default:
					throw ValidationUtils.Unhandled( Padding, "Padding" );
			}

			byte[] result = new byte[ total ];
			int outputOffset = 0;

			// Process all blocks except the last partial block.
			int processed = 0;
			if( inputBuffer.Length >= algSizeBytes )
				processed = TransformBlocks( inputBuffer, inputOffset, alignedSize, result, outputOffset );

			if( processed != alignedSize )
			{
				throw new ValidationException( string.Format( "The symmetric encryption algorithm has processed {0} bytes"
					+ " instead of {1} bytes.", processed, alignedSize ) );
			}

			inputOffset += alignedSize;
			outputOffset += alignedSize;
			total -= alignedSize;

			// A single (last) block is left to process.
			byte padding = (byte)(algSizeBytes - remaining);
			switch( Padding )
			{
				case SymmPaddings.Iso10126:
					// Random data + padding size.
					RandomFill( result, result.Length - padding, padding - 1 );
					result[ result.Length - 1 ] = padding;
					Buffer.BlockCopy( inputBuffer, inputOffset, result, outputOffset, remaining );

					// The last padded block will be transformed in-place.
					TransformBlocks( result, inputOffset, algSizeBytes, result, outputOffset );
					break;
				case SymmPaddings.None:
					// No padding.
					break;
				default:
					throw ValidationUtils.Unhandled( Padding, "Padding" );
			}

			return result;
		}

		/// <summary>
		/// Finishes processing the data for encryption.
		/// </summary>
		/// <param name="inputBuffer">Buffer to process. This may be empty, but not null.</param>
		/// <param name="inputOffset"> This must be at least 0.</param>
		/// <param name="inputCount"> This must be at least 0.</param>
		/// <returns>Returns the result of the processing. This may be empty, but not null.</returns>
		private byte[] TransformFinalDecrypt( byte[] inputBuffer, int inputOffset, int inputCount )
		{
			ValidationUtils.ObjectHasContent( inputBuffer, "inputBuffer" );
			ValidationUtils.Integer( inputOffset, 0, (inputBuffer.Length > 0) ? (inputBuffer.Length - 1) : 0, "inputOffset" );
			ValidationUtils.Integer( inputCount, 0, inputBuffer.Length - inputOffset, "inputCount" );

			int algSizeBytes = SymmAlg.AlgSizeBytes;

			if( (inputCount % algSizeBytes) != 0 )
				throw new ValidationException( "Input buffer is not aligned to the algorithm size." );

			int total = inputCount;
			if( IsLastBlock )
				total += algSizeBytes;

			byte[] result = new byte[ total ];
			int outputOffset = 0;

			int fullBlockCount = inputCount / algSizeBytes;
			int alignedSize = fullBlockCount * algSizeBytes;

			// Process all blocks except the last (padded) block.
			int processed = TransformBlocks( inputBuffer, inputOffset, alignedSize, result, outputOffset );
			if( (processed != alignedSize) && (processed != alignedSize - algSizeBytes) )
			{
				throw new ValidationException( string.Format( "The symmetric encryption algorithm has processed {0} bytes"
					+ " instead of {1} bytes.", processed, alignedSize ) );
			}

			inputOffset += processed;
			outputOffset += processed;
			inputCount -= processed;

			if( IsLastBlock )
			{
				TransformBlockWithChaining( WorkBuffer, WorkOut );
				Buffer.BlockCopy( WorkOut, 0, result, outputOffset, algSizeBytes );
				IsLastBlock = false;
			}

			// Total may be 0.
			byte padding = (total > 0) ? result[ total - 1 ] : (byte)0;

			switch( Padding )
			{
				case SymmPaddings.Iso10126:
					ValidationUtils.Integer( padding, 1, algSizeBytes, "padding" );
					total -= padding;
					break;
				case SymmPaddings.None:
					// Nothing to do (it's multiple of block size).
					break;
			}

			// Return the output without padding ("total - padding").
			if( total > 0 )
			{
				byte[] buffer = new byte[ total ];
				Buffer.BlockCopy( result, 0, buffer, 0, total );

				return buffer;
			}
			else
			{
				return new byte[ 0 ];
			}
		}

		/// <summary>
		/// Processes one block of data, using the CBC chaining.
		/// </summary>
		/// <param name="input">Input data block.</param>
		/// <param name="output">Output data block.</param>
		private void CbcChaining( byte[] input, byte[] output )
		{
			int algSizeBytes = SymmAlg.AlgSizeBytes;

			if( IsEncrypt )
			{
				// Mask input.
				for( int i = 0; i < algSizeBytes; i++ )
					ChainingBuffer[ i ] ^= input[ i ];

				// Process input.
				ProcessBlock( ChainingBuffer, output );

				// Preserve the data for the next chain.
				Buffer.BlockCopy( output, 0, ChainingBuffer, 0, algSizeBytes );
			}
			else
			{
				// Process input.
				ProcessBlock( input, output );

				// Demask input.
				for( int i = 0; i < algSizeBytes; i++ )
					output[ i ] ^= ChainingBuffer[ i ];

				// Preserve the data for the next chain.
				Buffer.BlockCopy( input, 0, ChainingBuffer, 0, algSizeBytes );
			}
		}

		/// <summary>
		/// Processes one block of data.
		/// </summary>
		/// <param name="input">Input data block.</param>
		/// <param name="output">Output data block.</param>
		/// <remarks>
		/// <para>The size of a block must be "AlgSize". Padding must be done before calling this method.</para>
		/// </remarks>
		protected void TransformBlockWithChaining( byte[] input, byte[] output )
		{
			switch( Chaining )
			{
				case SymmChainings.Cbc:
					CbcChaining( input, output );
					break;
				default:
					throw ValidationUtils.Unhandled( Chaining, "Chaining" );
			}
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Processes the specified data, but only an integer number of blocks. The <see cref="ProcessFinal( byte[] )"/> or
		/// <see cref="ProcessFinal( byte[], int, int )"/> method must be called to finish processing the data.
		/// </summary>
		/// <param name="buffer">Buffer to process. This may be empty, but not null.</param>
		/// <param name="offset">Offset in the input data, from where to start processing.</param>
		/// <param name="size">The number of bytes to process.</param>
		/// <returns>Returns the result of the processing.</returns>
		public byte[] Process( byte[] buffer, int offset, int size )
		{
			byte[] result = new byte[ GetTransformOutputSize( size ) ];

			int processedSize = TransformBlocks( buffer, offset, size, result, 0 );

			if( processedSize != result.Length )
			{
				throw new ValidationException( string.Format( "The symmetric cryptographic algorithm has outputted {0} bytes"
					+ " instead of the expected {1} bytes.", processedSize, result.Length ) );
			}

			return result;
		}

		/// <summary>
		/// Processes the specified data, but only an integer number of blocks. The <see cref="ProcessFinal( byte[] )"/> or
		/// <see cref="ProcessFinal( byte[], int, int )"/> method must be called to finish processing the data.
		/// </summary>
		/// <param name="inputBuffer">Buffer to process. This may be empty, but not null.</param>
		/// <returns>Returns the result of the processing.</returns>
		public byte[] Process( byte[] inputBuffer )
		{
			return Process( inputBuffer, 0, inputBuffer.Length );
		}

		/// <summary>
		/// Finishes processing the specified data.
		/// </summary>
		/// <param name="inputBuffer">Buffer to process. This may be empty, but not null.</param>
		/// <param name="inputOffset">Offset in the input data, from where to start processing.</param>
		/// <param name="inputCount">The number of bytes to process.</param>
		/// <returns>Returns the result of the processing. This may be empty, but not null.</returns>
		public byte[] ProcessFinal( byte[] inputBuffer, int inputOffset, int inputCount )
		{
			byte[] outputBuffer;

			if( IsEncrypt )
				outputBuffer = TransformFinalEncrypt( inputBuffer, inputOffset, inputCount );
			else
				outputBuffer = TransformFinalDecrypt( inputBuffer, inputOffset, inputCount );

			// Make the worker unusable.
			SymmAlg = null;

			return outputBuffer;
		}

		/// <summary>
		/// Finishes processing the specified data.
		/// </summary>
		/// <param name="inputBuffer">Buffer to process. This may be empty, but not null.</param>
		/// <returns>Returns the result of the processing. This may be empty, but not null.</returns>
		public byte[] ProcessFinal( byte[] inputBuffer )
		{
			byte[] outputBuffer = ProcessFinal( inputBuffer, 0, inputBuffer.Length );

			// Make the worker unusable.
			SymmAlg = null;

			return outputBuffer;
		}

		#endregion Public methods
	}
}
