////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Storehouse;

namespace Vela.DatabaseStorehouse
{
	/// <summary>
	/// Configuration for database access.
	/// </summary>
	public class DatabaseAccess : IStorehouseAccess
	{
		#region Public constants

		/// <summary>
		/// The maximum timeout (in seconds) for the execution of a command.
		/// </summary>
		public const int CommandTimeoutMax = 30;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected DatabaseAccess()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseProviderTypeName">The type name of the storehouse provider to instantiate.</param>
		/// <param name="connectionStringTemplate">The template for the database connection string.</param>
		public DatabaseAccess( string storehouseProviderTypeName, string connectionStringTemplate )
		{
			StorehouseProviderTypeName = storehouseProviderTypeName;
			ConnectionStringTemplate = connectionStringTemplate;
		}

		#endregion Constructors

		#region Public serializable properties

		private string _StorehouseProviderTypeName;
		/// <summary>
		/// The type name of the storehouse provider to instantiate. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string StorehouseProviderTypeName
		{
			get { return _StorehouseProviderTypeName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "StorehouseProviderTypeName" );

				_StorehouseProviderTypeName = value;
			}
		}

		private string _ConnectionStringTemplate;
		/// <summary>
		/// Template for the database connection string. Placeholders:
		/// <see cref="StorehouseConsts.StorehouseFileNamePlaceholder"/> - name of file attached to the storehouse,
		/// <see cref="StorehouseConsts.ServerStorehouseNamePlaceholder"/> - server storehouse name,
		/// <see cref="StorehouseConsts.StorehouseAttachmentsPasswordPlaceholder"/> - storehouse attachments password. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string ConnectionStringTemplate
		{
			get { return _ConnectionStringTemplate; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "ConnectionStringTemplate" );

				_ConnectionStringTemplate = value;
			}
		}

		private int _CommandTimeout = 0;
		/// <summary>
		/// The timeout (in seconds) for the execution of a command. This must be at least 0.
		/// </summary>
		[SerializeAttribute]
		public int CommandTimeout
		{
			get { return _CommandTimeout; }
			set
			{
				ValidationUtils.Integer( value, 0, CommandTimeoutMax, "CommandTimeout" );

				_CommandTimeout = value;
			}
		}

		#endregion Public serializable properties
	}
}
