////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.CoreDocuments;

namespace Vela.DatabaseStorehouse
{
	/// <summary>
	/// Column names for databases.
	/// </summary>
	public static class DatabaseColumns
	{
		#region Private constants

		/// <summary>
		/// Template used to format the column name for a speckle filter checksum. Placeholders: 0 - index.
		/// </summary>
		private const string SpeckleFilterChecksumTemplate = "SpeckleFilterChecksum{0:00}";

		/// <summary>
		/// Template used to format the column name for a speckle filter type name. Placeholders: 0 - index.
		/// </summary>
		private const string SpeckleFilterTypeNameTemplate = "SpeckleFilterTypeName{0:00}";

		/// <summary>
		/// Template used to format the column name for a speckle filter name. Placeholders: 0 - index.
		/// </summary>
		private const string SpeckleFilterNameTemplate = "SpeckleFilterName{0:00}";

		/// <summary>
		/// Template used to format the column name for a speckle filter value. Placeholders: 0 - index.
		/// </summary>
		private const string SpeckleFilterValueTemplate = "SpeckleFilterValue{0:00}";

		#endregion Private constants

		#region Public constants

		#region Checksum columns

		/// <summary>
		/// The "StorehouseKeyChecksum" column name.
		/// </summary>
		public const string StorehouseKeyChecksum = "StorehouseKeyChecksum";

		/// <summary>
		/// The "SheafTypeNameChecksum" column name.
		/// </summary>
		public const string SheafTypeNameChecksum = "SheafTypeNameChecksum";

		#endregion Checksum columns

		/// <summary>
		/// The "SheafUpdateVersion" column name.
		/// </summary>
		public const string SheafUpdateVersion = "SheafUpdateVersion";

		/// <summary>
		/// The "StorehouseKeyTypeName" column name.
		/// </summary>
		public const string StorehouseKeyTypeName = "StorehouseKeyTypeName";

		/// <summary>
		/// The "StorehouseKey" column name.
		/// </summary>
		public const string StorehouseKey = "StorehouseKey";

		/// <summary>
		/// The "SheafTypeName" column name.
		/// </summary>
		public const string SheafTypeName = "SheafTypeName";

		/// <summary>
		/// The "ReferredSheaf" column name.
		/// </summary>
		public const string ReferredSheaf = "ReferredSheaf";

		#endregion Public constants

		#region Public static methods

		/// <summary>
		/// Gets the column name for the speckle filter checksum with the specified index.
		/// </summary>
		/// <param name="filterIndex">Filter index. This must be at least 0.</param>
		/// <returns>Returns the column name for the speckle filter checksum.</returns>
		public static string GetSpeckleFilterChecksum( int filterIndex )
		{
			ValidationUtils.Integer( filterIndex, 0, CoreDocumentsConsts.SpeckleFilterSizeMax - 1, "filterIndex" );

			return string.Format( SpeckleFilterChecksumTemplate, filterIndex );
		}

		/// <summary>
		/// Gets the column name for the speckle filter type name with the specified index.
		/// </summary>
		/// <param name="filterIndex">Filter index. This must be at least 0.</param>
		/// <returns>Returns the column name for the speckle filter type name.</returns>
		public static string GetSpeckleFilterTypeName( int filterIndex )
		{
			ValidationUtils.Integer( filterIndex, 0, CoreDocumentsConsts.SpeckleFilterSizeMax - 1, "filterIndex" );

			return string.Format( SpeckleFilterTypeNameTemplate, filterIndex );
		}

		/// <summary>
		/// Gets the column name for the speckle filter name with the specified index.
		/// </summary>
		/// <param name="filterIndex">Filter index. This must be at least 0.</param>
		/// <returns>Returns the column name for the speckle filter.</returns>
		public static string GetSpeckleFilterName( int filterIndex )
		{
			ValidationUtils.Integer( filterIndex, 0, CoreDocumentsConsts.SpeckleFilterSizeMax - 1, "filterIndex" );

			return string.Format( SpeckleFilterNameTemplate, filterIndex );
		}

		/// <summary>
		/// Gets the column name for the speckle filter value with the specified index.
		/// </summary>
		/// <param name="filterIndex">Filter index. This must be at least 0.</param>
		/// <returns>Returns the column name for the speckle filter.</returns>
		public static string GetSpeckleFilterValue( int filterIndex )
		{
			ValidationUtils.Integer( filterIndex, 0, CoreDocumentsConsts.SpeckleFilterSizeMax - 1, "filterIndex" );

			return string.Format( SpeckleFilterValueTemplate, filterIndex );
		}

		#endregion Public static methods
	}
}
