////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.CoreDocuments;

namespace Vela.DatabaseStorehouse
{
	/// <summary>
	/// Parameter names for databases.
	/// </summary>
	/// <remarks>
	/// <para>Indexed parameter names are parameter names which end with a number formatted with two digits
	/// (= with a leading 0).</para>
	/// </remarks>
	public static class DatabaseParameters
	{
		#region Private constants

		/// <summary>
		/// Template used to format the parameter name for a sheaf type name checksum. Placeholders: 0 - index.
		/// </summary>
		private const string SheafTypeNameChecksumTemplate = "@SheafTypeNameChecksum{0:00}";

		/// <summary>
		/// Template used to format the parameter name for a sheaf type name. Placeholders: 0 - index.
		/// </summary>
		private const string SheafTypeNameTemplate = "@SheafTypeName{0:00}";

		/// <summary>
		/// Template used to format the parameter name for a speckle filter checksum. Placeholders: 0 - index.
		/// </summary>
		private const string SpeckleFilterChecksumTemplate = "@SpeckleFilterChecksum{0:00}";

		/// <summary>
		/// Template used to format the parameter name for a speckle filter type name. Placeholders: 0 - index.
		/// </summary>
		private const string SpeckleFilterTypeNameTemplate = "@SpeckleFilterTypeName{0:00}";

		/// <summary>
		/// Template used to format the parameter name for a speckle filter name. Placeholders: 0 - index.
		/// </summary>
		private const string SpeckleFilterNameTemplate = "@SpeckleFilterName{0:00}";

		/// <summary>
		/// Template used to format the parameter name for a speckle filter value. Placeholders: 0 - index.
		/// </summary>
		private const string SpeckleFilterValueTemplate = "@SpeckleFilterValue{0:00}";

		#endregion Private constants

		#region Public constants

		#region Checksum parameters

		/// <summary>
		/// The "StorehouseKeyChecksum" parameter name.
		/// </summary>
		public const string StorehouseKeyChecksum = "@StorehouseKeyChecksum";

		/// <summary>
		/// The "SheafTypeNameChecksum" parameter name.
		/// </summary>
		public const string SheafTypeNameChecksum = "@SheafTypeNameChecksum";

		#endregion Checksum parameters

		/// <summary>
		/// The "SheafUpdateVersion" parameter name.
		/// </summary>
		public const string SheafUpdateVersion = "@SheafUpdateVersion";

		/// <summary>
		/// The "StorehouseKeyTypeName" parameter name.
		/// </summary>
		public const string StorehouseKeyTypeName = "@StorehouseKeyTypeName";

		/// <summary>
		/// The "StorehouseKey" parameter name.
		/// </summary>
		public const string StorehouseKey = "@StorehouseKey";

		/// <summary>
		/// The "SheafTypeName" parameter name. This parameter name may be indexed.
		/// </summary>
		public const string SheafTypeName = "@SheafTypeName";

		/// <summary>
		/// The "ReferredSheaf" parameter name.
		/// </summary>
		public const string ReferredSheaf = "@ReferredSheaf";

		/// <summary>
		/// The "PagedReadOffset" parameter name.
		/// </summary>
		public const string PagedReadOffset = "@PagedReadOffset";

		/// <summary>
		/// The "PagedReadCount" parameter name.
		/// </summary>
		public const string PagedReadCount = "@PagedReadCount";

		#endregion Public constants

		#region Public static methods

		/// <summary>
		/// Gets the parameter name for the sheaf type name checksum with the specified index.
		/// </summary>
		/// <param name="index">Index. This must be at least 0.</param>
		/// <returns>Returns the parameter name for the sheaf type name checksum.</returns>
		public static string GetSheafTypeNameChecksum( int index )
		{
			ValidationUtils.Integer( index, 0, DatabaseConsts.SheafTypeNameSizeMax - 1, "index" );

			return string.Format( SheafTypeNameChecksumTemplate, index );
		}

		/// <summary>
		/// Gets the parameter name for the sheaf type name with the specified index.
		/// </summary>
		/// <param name="index">Index. This must be at least 0.</param>
		/// <returns>Returns the parameter name for the sheaf type name.</returns>
		public static string GetSheafTypeName( int index )
		{
			ValidationUtils.Integer( index, 0, DatabaseConsts.SheafTypeNameSizeMax - 1, "index" );

			return string.Format( SheafTypeNameTemplate, index );
		}

		/// <summary>
		/// Gets the parameter name for the speckle filter checksum with the specified index.
		/// </summary>
		/// <param name="filterIndex">Filter index. This must be at least 0.</param>
		/// <returns>Returns the parameter name for the speckle filter checksum.</returns>
		public static string GetSpeckleFilterChecksum( int filterIndex )
		{
			ValidationUtils.Integer( filterIndex, 0, CoreDocumentsConsts.SpeckleFilterSizeMax - 1, "filterIndex" );

			return string.Format( SpeckleFilterChecksumTemplate, filterIndex );
		}

		/// <summary>
		/// Gets the parameter name for the speckle filter type name with the specified index.
		/// </summary>
		/// <param name="filterIndex">Filter index. This must be at least 0.</param>
		/// <returns>Returns the parameter name for the speckle filter type name.</returns>
		public static string GetSpeckleFilterTypeName( int filterIndex )
		{
			ValidationUtils.Integer( filterIndex, 0, CoreDocumentsConsts.SpeckleFilterSizeMax - 1, "filterIndex" );

			return string.Format( SpeckleFilterTypeNameTemplate, filterIndex );
		}

		/// <summary>
		/// Gets the parameter name for the speckle filter name with the specified index.
		/// </summary>
		/// <param name="filterIndex">Filter index. This must be at least 0.</param>
		/// <returns>Returns the parameter name for the speckle filter name.</returns>
		public static string GetSpeckleFilterName( int filterIndex )
		{
			ValidationUtils.Integer( filterIndex, 0, CoreDocumentsConsts.SpeckleFilterSizeMax - 1, "filterIndex" );

			return string.Format( SpeckleFilterNameTemplate, filterIndex );
		}

		/// <summary>
		/// Gets the parameter name for the speckle filter value with the specified index.
		/// </summary>
		/// <param name="filterIndex">Filter index. This must be at least 0.</param>
		/// <returns>Returns the parameter name for the speckle filter value.</returns>
		public static string GetSpeckleFilterValue( int filterIndex )
		{
			ValidationUtils.Integer( filterIndex, 0, CoreDocumentsConsts.SpeckleFilterSizeMax - 1, "filterIndex" );

			return string.Format( SpeckleFilterValueTemplate, filterIndex );
		}

		#endregion Public static methods
	}
}
