////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Data;
using System.Data.Common;
using System.IO;
using Vela.Kernel;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.DatabaseStorehouse
{
	/// <summary>
	/// Contract for a database provider.
	/// </summary>
	/// <remarks>
	/// <para>Don't share a storehouse provider among multiple threads, even if you synchronize it. Clone it instead.</para>
	/// </remarks>
	public abstract class DatabaseProvider : StorehouseProvider
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected DatabaseProvider()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="access">The database access.</param>
		/// <param name="fileName">The name of the file where the storehouse header is stored.</param>
		/// <param name="headerUserPassword">The user password used to symmetrically encrypt the header.</param>
		/// <param name="header">The storehouse header (which is stored in a symmetrically encrypted form). This may be null if
		/// the storehouse doesn't exist yet.</param>
		/// <param name="serializer">The serializer of the sheafs.</param>
		public DatabaseProvider( DatabaseAccess access, string fileName, string headerUserPassword, StorehouseHeader header
				, ISerializer serializer )
			: base( access, fileName, headerUserPassword, header, serializer )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The typed storehouse access. READ-ONLY.
		/// </summary>
		public DatabaseAccess AccessTyped
		{
			get { return Access as DatabaseAccess; }
		}

		/// <summary>
		/// The typed set of scripts used to work with the storehouse. READ-ONLY.
		/// </summary>
		public IDatabaseScripts ScriptsTyped
		{
			get { return Scripts as IDatabaseScripts; }
		}

		/// <summary>
		/// The typed placeholders used in storehouse scripts. READ-ONLY.
		/// </summary>
		public DatabaseScriptPlaceholders ScriptPlaceholdersTyped
		{
			get { return ScriptPlaceholders as DatabaseScriptPlaceholders; }
		}

		#endregion Public non-serializable properties

		#region Private static methods

		/// <summary>
		/// Executes the parts of the specified command text, split by <see cref="DatabaseConsts.ScriptSeparator"/>.
		/// </summary>
		/// <param name="commandText">Command text to execute.</param>
		/// <param name="command">Command to execute.</param>
		protected static void SplitExecuteNonQuery( string commandText, DbCommand command )
		{
			ValidationUtils.TextHasContent( commandText, "commandText" );
			ValidationUtils.ObjectHasContent( command, "command" );

			string[] scripts = commandText.Split( new string[] { DatabaseConsts.ScriptSeparator }
				, StringSplitOptions.RemoveEmptyEntries );

			foreach( string script in scripts )
			{
				string trimmedScript = script.Trim();
				if( TextUtils.HasContent( trimmedScript ) )
				{
					command.CommandText = trimmedScript;

					command.ExecuteNonQuery();
				}
			}
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Checks if the storehouse with the specified file name exists.
		/// </summary>
		/// <param name="fileName">The name of the file where the storehouse header is stored.</param>
		public static bool HasStorehouseFile( string fileName )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			return File.Exists( fileName );
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Creates a new database connection string.
		/// </summary>
		/// <returns>Returns the new database connection string.</returns>
		public virtual string CreateConnectionString()
		{
			// Loads the storehouse header to get the attachments key, and this could thrown an exception.
			string attachmentsPassword = GetStorehouseAttachmentsPassword();

			return ReplacePlaceholders( AccessTyped.ConnectionStringTemplate, GetStorehouseMainAttachmentFileName()
				, ServerStorehouseName, attachmentsPassword );
		}

		/// <summary>
		/// Creates a new database connection.
		/// </summary>
		/// <param name="connectionString">The connection string to use.</param>
		/// <returns>Returns the new database connection.</returns>
		public abstract DbConnection CreateConnection( string connectionString );

		/// <summary>
		/// Creates a new database command.
		/// </summary>
		/// <param name="connection">Database connection to use.</param>
		/// <param name="transaction">Database transaction to use.</param>
		/// <returns>Returns the new database command.</returns>
		public abstract DbCommand CreateCommand( DbConnection connection, DbTransaction transaction );

		/// <summary>
		/// Adds a new database command parameter.
		/// </summary>
		/// <param name="dbCommand">Database command to which to add the parameter.</param>
		/// <param name="paramName">The name of the parameter.</param>
		/// <param name="dbType">Database type of the parameter.</param>
		/// <param name="isNullable">If this is true then the parameter is nullable (and its value may be null).</param>
		/// <param name="paramValue">The value of the parameter. This may be null if "isNullable"is true.</param>
		public abstract void AddCommandParam( DbCommand dbCommand, string paramName, DbType dbType, bool isNullable
			, object paramValue );

		/// <summary>
		/// Adds a new database command parameter to the specified command. The parameter's data type is a huge buffer.
		/// </summary>
		/// <param name="dbCommand">Database command to which to add the parameter.</param>
		/// <param name="paramName">The name of the parameter.</param>
		/// <param name="paramValue">The value of the parameter. This may be null.</param>
		public abstract void AddBlobCommandParam( DbCommand dbCommand, string paramName, object paramValue );

		#endregion Public methods
	}
}
