////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Data;
using System.Data.Common;
using System.Reflection;
using System.Threading;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.DatabaseStorehouse
{
	/// <summary>
	/// Contract for a database transaction.
	/// </summary>
	public abstract class DatabaseTransaction : IStorehouseTransaction
	{
		#region Private constants

		/// <summary>
		/// The log scope of the database transaction.
		/// </summary>
		private const string DatabaseTransactionLogScope = "Entire database transaction";

		#endregion Private constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected DatabaseTransaction()
		{
			InstanceId = CreateInstanceId();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="storehouseProvider">The storehouse provider which has created the current transaction.</param>
		/// <remarks>
		/// <para>The storehouse is opened in here, but the storehouse header is not loaded (because it's lazy loaded).</para>
		/// </remarks>
		public DatabaseTransaction( IStorehouseProvider storehouseProvider )
		{
			InstanceId = CreateInstanceId();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				InstanceLogManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), DatabaseTransactionLogScope );

				Provider = storehouseProvider;

				string connectionString = ProviderTyped.CreateConnectionString();
				Connection = ProviderTyped.CreateConnection( connectionString );
				Connection.Open();

				Transaction = Connection.BeginTransaction( IsolationLevel.ReadCommitted );

				IsOpen = true;
			}
			catch( Exception ex )
			{
				Close();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		#endregion Constructors

		#region Private static fields

		/// <summary>
		/// The last unique identifier of an instance of this class.
		/// </summary>
		private static Int64 LastInstanceId;

		#endregion Private static fields

		#region Private fields

		/// <summary>
		/// The deep text encoding.
		/// </summary>
		private readonly TextEncoding DeepTextEncoding = KernelUtils.GetDeepTextEncoding();

		private DbConnection _Connection;
		/// <summary>
		/// The database connection.
		/// </summary>
		private DbConnection Connection
		{
			get { return _Connection; }
			set
			{
				ValidationUtils.ObjectHasContent( value, "Connection" );

				_Connection = value;
			}
		}

		private DbTransaction _Transaction;
		/// <summary>
		/// The current database transaction.
		/// </summary>
		protected DbTransaction Transaction
		{
			get { return _Transaction; }
			set
			{
				ValidationUtils.ObjectHasContent( value, "Transaction" );

				_Transaction = value;
			}
		}

		/// <summary>
		/// The log manager for the entire transaction.
		/// </summary>
		private LogManager InstanceLogManager;

		/// <summary>
		/// The unique identifier of the current object.
		/// </summary>
		private Int64 InstanceId;

		#endregion Private fields

		#region Public non-serializable properties

		private IStorehouseProvider _Provider;
		/// <summary>
		/// The storehouse provider which has opened the current transaction. READ-ONLY.
		/// </summary>
		public IStorehouseProvider Provider
		{
			get { return _Provider; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Provider" );

				_Provider = value;
			}
		}

		/// <summary>
		/// The typed storehouse provider which has opened the current transaction. READ-ONLY.
		/// </summary>
		public DatabaseProvider ProviderTyped
		{
			get { return Provider as DatabaseProvider; }
		}

		/// <summary>
		/// The storehouse access. READ-ONLY.
		/// </summary>
		public IStorehouseAccess Access
		{
			get { return Provider.Access; }
		}

		/// <summary>
		/// The typed storehouse access. READ-ONLY.
		/// </summary>
		public DatabaseAccess AccessTyped
		{
			get { return ProviderTyped.AccessTyped; }
		}

		private bool _IsOpen = false;
		/// <summary>
		/// If this is true then the current transaction is open, else it's closed. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>When a transaction is closed, no changes may be made through it to the storehouse.</para>
		/// </remarks>
		public bool IsOpen
		{
			get { return _IsOpen; }
			protected set { _IsOpen = value; }
		}

		#endregion Public non-serializable properties

		#region Private static methods

		/// <summary>
		/// Creates a new unique identifier for an instance of the current class.
		/// </summary>
		/// <returns>Returns a new unique identifier.</returns>
		/// <remarks>
		/// <para><see cref="LastInstanceId"/> is thread safe.</para>
		/// 
		/// <para><see cref="LastInstanceId"/> is incremented here.</para>
		/// </remarks>
		private static Int64 CreateInstanceId()
		{
			return Interlocked.Increment( ref LastInstanceId );
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Checks if the specified type is a storehouse key type.
		/// </summary>
		/// <param name="type">The type to check if it is a storehouse key type.</param>
		/// <returns>Returns the information of the constructor which can be used to instantiate the type.</returns>
		private static ConstructorInfo CheckStorehouseKeyType( Type type )
		{
			ConstructorInfo ci = type.GetConstructor( BindingFlags.Public | BindingFlags.NonPublic
				| BindingFlags.Instance, null, new Type[] { typeof( byte[] ) }, null );

			// The class must be non-abstract and must have an instance constructor.
			if( !type.IsVisible || !type.IsClass || type.IsAbstract || !typeof( Locator ).IsAssignableFrom( type )
				|| (ci == null) )
			{
				throw new ValidationException( string.Format( "Only non-abstract classes derived from \"{0}\", with an"
					+ " instance constructor (with a buffer parameter) is a storehouse key type. Type \"{1}\" doesn't"
					+ " meet these requirements.", typeof( Locator ).FullName, type.FullName ) );
			}

			return ci;
		}

		/// <summary>
		/// Checks if the specified type is a speckle type.
		/// </summary>
		/// <param name="type">The type to check if it is a speckle type.</param>
		/// <returns>Returns the information of the constructor which can be used to instantiate the type.</returns>
		private static ConstructorInfo CheckSpeckleType( Type type )
		{
			ConstructorInfo ci = type.GetConstructor( BindingFlags.Public | BindingFlags.NonPublic
				| BindingFlags.Instance, null
				, new Type[] { typeof( Locator ), typeof( string ), typeof( SpeckleFilters ) }, null );

			// The class must be non-abstract and must have an instance constructor.
			if( !type.IsVisible || !type.IsClass || type.IsAbstract || !typeof( Speckle ).IsAssignableFrom( type )
				|| (ci == null) )
			{
				throw new ValidationException( string.Format( "Only non-abstract classes derived from \"{0}\", with an"
						+ " instance constructor (with a locator, string and speckle filters parameters) is a"
						+ " speckle type. Type \"{1}\" doesn't meet these requirements."
					, typeof( Speckle ).FullName, type.FullName ) );
			}

			return ci;
		}

		#endregion Public static methods

		#region Private methods

		/// <summary>
		/// Commits the current transaction in the storehouse.
		/// </summary>
		protected abstract void InternalCommit();

		/// <summary>
		/// Creates a new log manager instance and starts it.
		/// </summary>
		/// <param name="methodInfo">Information about the method where the tracer is created.</param>
		/// <param name="scope">The scope of the started method. This may be null or empty.</param>
		/// <returns>Returns the created tracer instance.</returns>
		protected LogManager LogManagerStartMethod( MethodBase methodInfo, string scope = null )
		{
			if( !LogManager.LogDebugMessages )
				return null;

			LogRecordFields fields = new LogRecordFields();
			fields.Add( new LogRecordField( "TRID", InstanceId, false, LogRecordFieldPriorities.Beginning ) );
			fields.Add( new LogRecordField( "Scope", scope, false, LogRecordFieldPriorities.Middle ) );

			return LogManager.DebugStartMethod( methodInfo, fields );
		}

		/// <summary>
		/// Creates a new log manager instance and starts it.
		/// </summary>
		/// <param name="methodInfo">Information about the method where the tracer is created.</param>
		/// <param name="storehouseKey">The storehouse key of the sheaf which the calling method has to process. This may be
		/// null.</param>
		/// <param name="sheafType">The type of the document which the calling method has to process. This may be null.</param>
		/// <param name="sheafLock">The lock to be used for the sheaf which the calling method has to process.</param>
		/// <returns>Returns the created tracer instance.</returns>
		protected LogManager LogManagerStartMethod( MethodBase methodInfo, Locator storehouseKey, Type sheafType
			, SheafLocks sheafLock )
		{
			if( !LogManager.LogDebugMessages )
				return null;

			string sheafTypeName = (sheafType != null) ? sheafType.FullName : null;

			LogRecordFields fields = new LogRecordFields();
			fields.Add( new LogRecordField( "TRID", InstanceId, false, LogRecordFieldPriorities.Beginning ) );
			fields.Add( new LogRecordField( "Sheaf lock", sheafLock, false, LogRecordFieldPriorities.Middle ) );
			fields.Add( new LogRecordField( "Sheaf type name", sheafTypeName, false, LogRecordFieldPriorities.Middle ) );
			fields.Add( new LogRecordField( "Instance locator", storehouseKey, false, LogRecordFieldPriorities.Middle ) );

			return LogManager.DebugStartMethod( methodInfo, fields );
		}

		/// <summary>
		/// Creates a new log manager instance and starts it.
		/// </summary>
		/// <param name="methodInfo">Information about the method where the tracer is created.</param>
		/// <param name="storehouseKey">The storehouse key of the sheaf which the calling method has to process. This may be
		/// null.</param>
		/// <param name="sheafTypes">The possible types of the document which the calling method has to process. This may be
		/// null or empty.</param>
		/// <param name="sheafLock">The lock to be used for the sheaf which the calling method has to process.</param>
		/// <returns>Returns the created tracer instance.</returns>
		protected LogManager LogManagerStartMethod( MethodBase methodInfo, Locator storehouseKey, PortableTypes sheafTypes
			, SheafLocks sheafLock )
		{
			if( !LogManager.LogDebugMessages )
				return null;

			string sheafTypeNames = (sheafTypes != null) ? sheafTypes.ToString() : null;

			LogRecordFields fields = new LogRecordFields();
			fields.Add( new LogRecordField( "TRID", InstanceId, false, LogRecordFieldPriorities.Beginning ) );
			fields.Add( new LogRecordField( "Sheaf lock", sheafLock, false, LogRecordFieldPriorities.Middle ) );
			fields.Add( new LogRecordField( "Sheaf type names", sheafTypeNames, false, LogRecordFieldPriorities.Middle ) );
			fields.Add( new LogRecordField( "Instance locator", storehouseKey, false, LogRecordFieldPriorities.Middle ) );

			return LogManager.DebugStartMethod( methodInfo, fields );
		}

		/// <summary>
		/// Creates a new log manager instance and starts it.
		/// </summary>
		/// <param name="methodInfo">Information about the method where the tracer is created.</param>
		/// <param name="sheafTypes">The possible types of the document which the calling method has to process. This may be
		/// null or empty.</param>
		/// <param name="speckleFilters">The speckle filters which the calling method has to use. This may be null or empty.
		/// </param>
		/// <param name="sheafLock">The lock to be used for the sheaf which the calling method has to process.</param>
		/// <returns>Returns the created tracer instance.</returns>
		protected LogManager LogManagerStartMethod( MethodBase methodInfo, PortableTypes sheafTypes
			, SpeckleFilters speckleFilters, SheafLocks sheafLock )
		{
			if( !LogManager.LogDebugMessages )
				return null;

			string sheafTypeNames = (sheafTypes != null) ? sheafTypes.ToString() : null;
			string speckleFiltersText = (speckleFilters != null) ? speckleFilters.ToString() : null;

			LogRecordFields fields = new LogRecordFields();
			fields.Add( new LogRecordField( "TRID", InstanceId, false, LogRecordFieldPriorities.Beginning ) );
			fields.Add( new LogRecordField( "Sheaf lock", sheafLock, false, LogRecordFieldPriorities.Middle ) );
			fields.Add( new LogRecordField( "Sheaf type names", sheafTypeNames, false, LogRecordFieldPriorities.Middle ) );
			fields.Add( new LogRecordField( "Speckle filters", speckleFiltersText, false, LogRecordFieldPriorities.Middle ) );

			return LogManager.DebugStartMethod( methodInfo, fields );
		}

		/// <summary>
		/// Shows the message for the specified log manager.
		/// </summary>
		/// <param name="tracer">Log manager for which to show the message. If this is null, nothing happens.</param>
		/// <param name="scope">The scope of the ended method. This may be null or empty.</param>
		protected void LogManagerEndMethod( LogManager tracer, string scope = null )
		{
			if( !LogManager.LogDebugMessages )
				return;

			LogRecordFields fields = new LogRecordFields();
			fields.Add( new LogRecordField( "TRID", InstanceId, false, LogRecordFieldPriorities.Beginning ) );
			fields.Add( new LogRecordField( "Scope", scope, false, LogRecordFieldPriorities.Middle ) );

			LogManager.DebugEndMethod( tracer, fields );
		}

		/// <summary>
		/// Logs the successful loading of the specified speckle.
		/// </summary>
		/// <param name="speckle">The loaded speckle. This may be null.</param>
		protected void LogManagerSpeckleLoaded( Speckle speckle )
		{
			if( !LogManager.LogDebugMessages || (speckle == null) )
				return;

			Locator storehouseKey = speckle.StorehouseKey;
			string speckleTypeName = speckle.GetType().FullName;

			LogRecordFields fields = new LogRecordFields();
			fields.Add( new LogRecordField( "TRID", InstanceId, false, LogRecordFieldPriorities.Beginning ) );
			fields.Add( new LogRecordField( "Speckle lock", speckle.StorehouseLock, false, LogRecordFieldPriorities.Middle ) );
			fields.Add( new LogRecordField( "Speckle type name", speckleTypeName, false, LogRecordFieldPriorities.Middle ) );
			fields.Add( new LogRecordField( "Instance locator", storehouseKey, false, LogRecordFieldPriorities.Middle ) );

			LogManager.DebugAdd( "Speckle loaded", fields );
		}

		/// <summary>
		/// Logs the successful loading of the specified sheaf.
		/// </summary>
		/// <param name="sheaf">The loaded sheaf. This may be null.</param>
		protected void LogManagerSheafLoaded( ISheaf sheaf )
		{
			if( !LogManager.LogDebugMessages || (sheaf == null) )
				return;

			Locator instanceLocator = sheaf.InstanceLocator;
			string sheafTypeName = sheaf.GetType().FullName;

			LogRecordFields fields = new LogRecordFields();
			fields.Add( new LogRecordField( "TRID", InstanceId, false, LogRecordFieldPriorities.Beginning ) );
			fields.Add( new LogRecordField( "Sheaf lock", sheaf.StorehouseLock, false, LogRecordFieldPriorities.Middle ) );
			fields.Add( new LogRecordField( "Sheaf type name", sheafTypeName, false, LogRecordFieldPriorities.Middle ) );
			fields.Add( new LogRecordField( "Instance locator", instanceLocator, false, LogRecordFieldPriorities.Middle ) );

			LogManager.DebugAdd( "Sheaf loaded", fields );
		}

		/// <summary>
		/// Gets the encryption key of the sheaf with the specified storehouse key.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the document for which to get the encryption key.</param>
		/// <returns>Returns the encryption key of the sheaf.</returns>
		private byte[] GetSheafEncryptionKey( Locator storehouseKey )
		{
			ValidationUtils.ObjectHasContent( storehouseKey, "storehouseKey" );

			return Provider.Header.GenerateSheafEncryptionKey( storehouseKey );
		}

		/// <summary>
		/// Gets the encryption key of the specified sheaf.
		/// </summary>
		/// <param name="sheaf">The document for which to get the encryption key.</param>
		/// <returns>Returns the encryption key of the sheaf.</returns>
		private byte[] GetSheafEncryptionKey( ISheaf sheaf )
		{
			ValidationUtils.ObjectHasContent( sheaf, "sheaf" );

			return Provider.Header.GenerateSheafEncryptionKey( sheaf.InstanceLocator );
		}

		/// <summary>
		/// If the current storehouse transaction is not open, an exception is thrown.
		/// </summary>
		private void CheckIfOpen()
		{
			if( !IsOpen )
			{
				throw new ValidationException( "The transaction is closed.", null, null
					, new ExceptionDetail( "Storehouse file", Provider.FileName ) );
			}
		}

		/// <summary>
		/// Closes the current connection and transaction in the storehouse. You must either commit or rollback the transaction
		/// before you call this method.
		/// </summary>
		private void Close()
		{
			if( !IsOpen )
				return;

			IsOpen = false;

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				if( Transaction != null )
					Transaction.Dispose();
			}
			finally
			{
				try
				{
					if( Connection != null )
						Connection.Close();
				}
				finally
				{
					LogManagerEndMethod( logManager );
					LogManagerEndMethod( InstanceLogManager, DatabaseTransactionLogScope );
				}
			}
		}

		/// <summary>
		/// Includes the specified storehouse realm in the specified script.
		/// </summary>
		/// <param name="script">Script in which to include the storehouse realm.</param>
		/// <param name="storehouseRealm">Storehouse realm to include in the script.</param>
		/// <returns>Returns the script which contains the storehouse realm.</returns>
		private string ReplaceStorehouseRealm( string script, string storehouseRealm )
		{
			ValidationUtils.TextHasContent( script, "script" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			return TextUtils.Replace( script, DatabaseScriptPlaceholders.StorehouseRealm, storehouseRealm );
		}

		/// <summary>
		/// Includes the list of seek sheaf type name parameters in the specified script.
		/// </summary>
		/// <param name="script">Script in which to include the sheaf type name parameters.</param>
		/// <param name="sheafTypes">Sheaf types whose names to include in the script.</param>
		/// <returns>Returns the script which contains the sheaf type name parameters.</returns>
		private string ReplaceSheafTypeNameSeekList( string script, PortableTypes sheafTypes )
		{
			ValidationUtils.TextHasContent( script, "script" );
			ValidationUtils.ListHasContent( sheafTypes, "sheafTypes" );

			if( ListUtils.HasContent( sheafTypes ) )
			{
				string checksumParamList = ProviderTyped.ScriptPlaceholdersTyped.GetSheafTypeNameChecksumSeek( sheafTypes );
				string valueParamList = ProviderTyped.ScriptPlaceholdersTyped.GetSheafTypeNameSeek( sheafTypes );

				script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SheafTypeNameChecksumSeekList
					, checksumParamList );
				script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SheafTypeNameSeekList, valueParamList );
			}
			else
			{
				script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SheafTypeNameChecksumSeekList, "" );
				script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SheafTypeNameSeekList, "" );
			}

			return script;
		}

		/// <summary>
		/// Includes the list of speckle filter columns and parameters in the specified script.
		/// </summary>
		/// <param name="script">Script in which to include the sheaf type name parameters.</param>
		/// <returns>Returns the script which contains the speckle filter columns and parameters.</returns>
		private string ReplaceSpeckleFilterPlaceholders( string script )
		{
			ValidationUtils.TextHasContent( script, "script" );

			// Columns.
			script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterChecksumColumnList
				, ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterChecksumColumnList() );
			script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterTypeNameColumnList
				, ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterTypeNameColumnList() );
			script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterNameColumnList
				, ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterNameColumnList() );
			script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterValueColumnList
				, ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterValueColumnList() );

			// Parameters.
			script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterChecksumParameterList
				, ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterChecksumParameterList() );
			script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterTypeNameParameterList
				, ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterTypeNameParameterList() );
			script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterNameParameterList
				, ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterNameParameterList() );
			script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterValueParameterList
				, ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterValueParameterList() );

			// Column - parameter assignments.
			script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterChecksumColumnParameterList
				, ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterChecksumColumnParameterList() );
			script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterTypeNameColumnParameterList
				, ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterTypeNameColumnParameterList() );
			script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterNameColumnParameterList
				, ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterNameColumnParameterList() );
			script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterValueColumnParameterList
				, ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterValueColumnParameterList() );

			return script;
		}

		/// <summary>
		/// Includes the list of speckle filter parameters in the specified script.
		/// </summary>
		/// <param name="script">Script in which to include the speckle filter parameters.</param>
		/// <param name="seekFilters">The seek filters to include in the script. This may be null or empty.</param>
		/// <returns>Returns the script which contains the speckle filter parameters.</returns>
		private string ReplaceSpeckleFilterSeekList( string script, SpeckleFilters seekFilters )
		{
			ValidationUtils.TextHasContent( script, "script" );

			if( ListUtils.HasContent( seekFilters ) )
			{
				seekFilters.EnsureFilterIndexUniqueness();

				string checksumParamList = ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterChecksumSeek( seekFilters );
				string valueParamList = ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterValueSeek( seekFilters );

				script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterChecksumSeekList
					, checksumParamList );
				script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterValueSeekList, valueParamList );
			}
			else
			{
				script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterChecksumSeekList, "" );
				script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterValueSeekList, "" );
			}

			return script;
		}

		/// <summary>
		/// Includes the list of load speckle filter parameters in the specified script.
		/// </summary>
		/// <param name="script">Script in which to include the speckle filter parameters.</param>
		/// <param name="loadFilterIdentifiers">The speckle filter identifiers to include in the script. This may be
		/// null or empty.</param>
		/// <returns>Returns the script which contains the speckle filter parameters.</returns>
		private string ReplaceSpeckleFilterLoadList( string script, SpeckleFilterIdentifiers loadFilterIdentifiers )
		{
			ValidationUtils.TextHasContent( script, "script" );

			if( ListUtils.HasContent( loadFilterIdentifiers ) )
			{
				string typeNameParamList = ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterTypeNameLoad(
					loadFilterIdentifiers );
				string nameParamList = ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterNameLoad( loadFilterIdentifiers );
				string valueParamList = ProviderTyped.ScriptPlaceholdersTyped.GetSpeckleFilterValueLoad( loadFilterIdentifiers );

				script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterTypeNameLoadList
					, typeNameParamList );
				script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterNameLoadList, nameParamList );
				script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterValueLoadList, valueParamList );
			}
			else
			{
				script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterTypeNameLoadList, "" );
				script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterNameLoadList, "" );
				script = TextUtils.Replace( script, DatabaseScriptPlaceholders.SpeckleFilterValueLoadList, "" );
			}

			return script;
		}

		/// <summary>
		/// Adds the specified storehouse key as a parameter of to the specified database command.
		/// </summary>
		/// <param name="dbCommand">Database command to which to add the storehouse key parameter.</param>
		/// <param name="storehouseKey">Storehouse key to add as parameter to the database command.</param>
		private void AddStorehouseKeyParam( DbCommand dbCommand, Locator storehouseKey )
		{
			ValidationUtils.ObjectHasContent( dbCommand, "dbCommand" );
			ValidationUtils.ObjectHasContent( storehouseKey, "storehouseKey" );

			ProviderTyped.AddCommandParam( dbCommand, DatabaseParameters.StorehouseKeyChecksum, DbType.Int32, false
				, BufferUtils.ComputeChecksum32( storehouseKey.Value ) );
			ProviderTyped.AddCommandParam( dbCommand, DatabaseParameters.StorehouseKey, DbType.Binary, false
				, storehouseKey.Value );
		}

		/// <summary>
		/// Adds the storehouse key of the specified sheaf as a parameter of the specified database command.
		/// </summary>
		/// <param name="dbCommand">Database command to which to add the storehouse key parameter.</param>
		/// <param name="sheaf">Sheaf whose storehouse key to add as parameter to the database command.</param>
		private void AddStorehouseKeyParam( DbCommand dbCommand, ISheaf sheaf )
		{
			ValidationUtils.ObjectHasContent( sheaf, "sheaf" );

			AddStorehouseKeyParam( dbCommand, sheaf.InstanceLocator );
		}

		/// <summary>
		/// Adds the specified sheaf type name as a parameter of the specified database command.
		/// </summary>
		/// <param name="dbCommand">Database command to which to add the sheaf type name parameter.</param>
		/// <param name="sheafTypeName">Sheaf type name to add to the database command.</param>
		private void AddSheafTypeNameParam( DbCommand dbCommand, string sheafTypeName )
		{
			ValidationUtils.ObjectHasContent( dbCommand, "dbCommand" );
			ValidationUtils.ObjectHasContent( sheafTypeName, "sheafTypeName" );

			// Convert to buffer to compute the checksum.
			byte[] docTypeNameBuffer = DeepTextEncoding.TextToArray( sheafTypeName, true );

			ProviderTyped.AddCommandParam( dbCommand, DatabaseParameters.SheafTypeNameChecksum, DbType.Int32, false
				, BufferUtils.ComputeChecksum32( docTypeNameBuffer ) );
			ProviderTyped.AddCommandParam( dbCommand, DatabaseParameters.SheafTypeName, DbType.String, false
				, sheafTypeName );
		}

		/// <summary>
		/// Adds the sheaf type name of the specified sheaf as a parameter of the specified database command.
		/// </summary>
		/// <param name="dbCommand">Database command to which to add the sheaf type name parameter.</param>
		/// <param name="sheaf">Sheaf whose type name to add as parameter to the database command.</param>
		private void AddSheafTypeNameParam( DbCommand dbCommand, ISheaf sheaf )
		{
			ValidationUtils.ObjectHasContent( sheaf, "sheaf" );

			AddSheafTypeNameParam( dbCommand, sheaf.GetType().FullName );
		}

		/// <summary>
		/// Adds the names of the specified sheaf types as parameters of the specified database command.
		/// </summary>
		/// <param name="dbCommand">Database command to which to add the sheaf type name parameters.</param>
		/// <param name="sheafTypes">Sheaf types whose names to add as parameters of the database command.</param>
		private void AddSheafTypeNameParams( DbCommand dbCommand, PortableTypes sheafTypes )
		{
			ValidationUtils.ObjectHasContent( dbCommand, "dbCommand" );
			ValidationUtils.ListHasContent( sheafTypes, "sheafTypes" );

			int sheafTypesCount = sheafTypes.Count;
			ValidationUtils.Integer( sheafTypes.Count, 1, DatabaseConsts.SheafTypeNameSizeMax, "sheafTypes.Count" );

			int i = 0;
			foreach( Type sheafType in sheafTypes )
			{
				string paramNameChecksum = DatabaseParameters.GetSheafTypeNameChecksum( i );
				string paramNameValue = DatabaseParameters.GetSheafTypeName( i );

				string sheafTypeName = sheafType.FullName;
				byte[] valueBuffer = DeepTextEncoding.TextToArray( sheafTypeName, true );
				int checksum = BufferUtils.ComputeChecksum32( valueBuffer );

				ProviderTyped.AddCommandParam( dbCommand, paramNameChecksum, DbType.Int32, false, checksum );
				ProviderTyped.AddCommandParam( dbCommand, paramNameValue, DbType.String, false, sheafTypeName );

				i++;
			}
		}

		/// <summary>
		/// Adds the specified speckle filters as parameters of the specified database command.
		/// </summary>
		/// <param name="dbCommand">Database command to which to add the speckle filter parameters.</param>
		/// <param name="filters">Speckle filters to add as parameters of the database command. This may be null or empty.
		/// </param>
		/// <param name="includeMissing">If this is true then the parameters for the missing speckle filters are also
		/// added.</param>
		private void AddSpeckleFilterParams( DbCommand dbCommand, SpeckleFilters filters, bool includeMissing )
		{
			ValidationUtils.ObjectHasContent( dbCommand, "dbCommand" );

			if( ListUtils.HasContent( filters ) )
				filters.EnsureFilterIndexUniqueness();

			for( int i = 0; i < CoreDocumentsConsts.SpeckleFilterSizeMax; i++ )
			{
				string paramNameChecksum = DatabaseParameters.GetSpeckleFilterChecksum( i );
				string paramNameTypeName = DatabaseParameters.GetSpeckleFilterTypeName( i );
				string paramNameFilterName = DatabaseParameters.GetSpeckleFilterName( i );
				string paramNameFilterValue = DatabaseParameters.GetSpeckleFilterValue( i );

				// Check if the filter was specified.
				SpeckleFilter filter = ListUtils.HasContent( filters ) ? filters.GetByIndex( i ) : null;
				if( filter != null )
				{
					// The filter is present.

					if( filter.Value != null )
					{
						byte[] valueBuffer = ScalarEncoder.ValueToBuffer( filter.Value, DeepTextEncoding, true );

						// Check if the size of the value is valid.
						if( valueBuffer.Length > CoreDocumentsConsts.SpeckleFilterValueSizeMax )
						{
							throw new ValidationException( string.Format( "The size of the value of a speckle filter must be at"
									+ " most {0}, but it's {1}."
								, CoreDocumentsConsts.SpeckleFilterValueSizeMax, valueBuffer.Length ) );
						}

						int checksum = BufferUtils.ComputeChecksum32( valueBuffer );
						string typeName = filter.Value.GetType().FullName;

						ProviderTyped.AddCommandParam( dbCommand, paramNameChecksum, DbType.Int32, true, checksum );
						ProviderTyped.AddCommandParam( dbCommand, paramNameTypeName, DbType.String, true, typeName );
						ProviderTyped.AddCommandParam( dbCommand, paramNameFilterName, DbType.String, true
							, filter.Identifier.Name );
						ProviderTyped.AddCommandParam( dbCommand, paramNameFilterValue, DbType.Binary, true, valueBuffer );
					}
					else
					{
						// The filter value is missing.

						ProviderTyped.AddCommandParam( dbCommand, paramNameChecksum, DbType.Int32, true, null );
						ProviderTyped.AddCommandParam( dbCommand, paramNameTypeName, DbType.String, true, null );
						ProviderTyped.AddCommandParam( dbCommand, paramNameFilterName, DbType.String, true
							, filter.Identifier.Name );
						ProviderTyped.AddCommandParam( dbCommand, paramNameFilterValue, DbType.Binary, true, null );
					}
				}
				else
				{
					// The filter is missing.

					if( includeMissing )
					{
						ProviderTyped.AddCommandParam( dbCommand, paramNameChecksum, DbType.Int32, true, null );
						ProviderTyped.AddCommandParam( dbCommand, paramNameTypeName, DbType.String, true, null );
						ProviderTyped.AddCommandParam( dbCommand, paramNameFilterName, DbType.String, true, null );
						ProviderTyped.AddCommandParam( dbCommand, paramNameFilterValue, DbType.Binary, true, null );
					}
				}
			}
		}

		/// <summary>
		/// Adds the specified paging parameters to the specified database command.
		/// </summary>
		/// <param name="dbCommand">Database command to which to add the paging parameters.</param>
		/// <param name="paging">Information about how to process the documents in a paged manner. If this is null, no paging is
		/// used.</param>
		private void AddPagingParams( DbCommand dbCommand, CollectPaging paging )
		{
			ValidationUtils.ObjectHasContent( dbCommand, "dbCommand" );

			if( paging == null )
				return;

			ProviderTyped.AddCommandParam( dbCommand, DatabaseParameters.PagedReadOffset, DbType.Int32, false
				, paging.Offset );
			ProviderTyped.AddCommandParam( dbCommand, DatabaseParameters.PagedReadCount, DbType.Int32, false
				, paging.Count );
		}

		/// <summary>
		/// Loads the speckle filters from the specified database reader.
		/// </summary>
		/// <param name="dbReader">Database reader from which to read the speckle filters.</param>
		/// <param name="loadFilterIdentifiers">The speckle filter identifiers whose values to load.</param>
		/// <returns>Returns the loaded speckle filters whose values are set with those loaded from the storehouse.</returns>
		private SpeckleFilters LoadSpeckleFilters( DbDataReader dbReader, SpeckleFilterIdentifiers loadFilterIdentifiers )
		{
			ValidationUtils.ObjectHasContent( dbReader, "dbReader" );
			ValidationUtils.ListHasContent( loadFilterIdentifiers, "loadFilterIdentifiers" );

			SpeckleFilters loadedFilters = new SpeckleFilters();

			foreach( SpeckleFilterIdentifier loadFilterIdentifier in loadFilterIdentifiers )
			{
				int i = loadFilterIdentifier.Index;

				// Type name.
				string typeNameColumnName = DatabaseColumns.GetSpeckleFilterTypeName( i );
				string typeNameResult = dbReader[ typeNameColumnName ] as string;

				// Filter name.
				string filterNameColumnName = DatabaseColumns.GetSpeckleFilterName( i );
				string nameResult = dbReader[ filterNameColumnName ] as string;

				// Filter value.
				string filterValueColumnName = DatabaseColumns.GetSpeckleFilterValue( i );
				byte[] valueResult = dbReader[ filterValueColumnName ] as byte[];

				#region Null validation

				if( ((typeNameResult != null) && (valueResult == null)) || ((typeNameResult == null) && (valueResult != null)) )
				{
					throw new ValidationException( string.Format( "For the speckle filter \"{0}\" ({1}), the type name and the"
						+ " value must both be either non-null or null.", loadFilterIdentifier.Name, i ) );
				}

				if( (nameResult == null) && ((typeNameResult != null) || (valueResult != null)) )
				{
					throw new ValidationException( string.Format( "For the speckle filter \"{0}\" ({1}), if the filter name"
						+ " is null, the type name and the value must both be null.", loadFilterIdentifier.Name, i ) );
				}

				#endregion Null validation

				// Filter name validation.
				if( (nameResult != null) && !TextUtils.Equals( loadFilterIdentifier.Name, nameResult ) )
				{
					if( TextUtils.Equals( loadFilterIdentifier.OverriddenName, nameResult ) )
					{
						// The value will be overridden, but null must be returned because the type of the new value may be
						// different (and later casts would fail otherwise). Besides, since it's overridden, the value probably
						// has a different meaning.
						valueResult = null;
					}
					else
					{
						throw new ValidationException( string.Format( "The filter name of the speckle filter {0} to load is"
								+ " \"{1}\", but the loaded filter name is \"{2}\"."
							, i, loadFilterIdentifier.Name, nameResult ) );
					}
				}

				// Create filter.
				if( valueResult != null )
				{
					Type valueType = PortableTypes.FindGlobal( typeNameResult );
					if( valueType == null )
					{
						throw new ValidationException( "Can't find the globally portable type for the value of a speckle"
								+ " filter.", null, null
							, new ExceptionDetail( "Type name", typeNameResult )
							, new ExceptionDetail( "Load filter identifier", loadFilterIdentifier.Name ) );
					}

					// Reconstruct the filter value.
					object filterValue = ScalarEncoder.ValueFromBuffer( valueResult, valueType );

					// Create a new speckle filter.
					SpeckleFilter loadedFilter = new SpeckleFilter( loadFilterIdentifier, filterValue );
					loadedFilters.Add( loadedFilter );
				}
			}

			return loadedFilters;
		}

		/// <summary>
		/// Gets the sheaf in a form which may be saved in the storehouse.
		/// </summary>
		/// <param name="sheaf">Sheaf.</param>
		/// <param name="salt">The (random) salt used to compute the hash and encrypt the sheaf. This is stored in unencrypted
		/// form. If this is null or empty, a random salt is used instead.</param>
		/// <param name="encryptionKey">The key used for the encryption of the sheaf.</param>
		/// <returns>Returns the sheaf in a form which may be saved in the storehouse.</returns>
		private byte[] GetReferredSheaf( ISheaf sheaf, byte[] salt, byte[] encryptionKey )
		{
			ValidationUtils.ObjectHasContent( sheaf, "sheaf" );
			ValidationUtils.ArrayHasContent( encryptionKey, "encryptionKey" );

			// Using the document's instance locator as the envelope's instance locator makes it easy to associate the encrypted
			// envelopes in case they get separated from the storehouse (key / speckle).
			StorehouseHeader header = Provider.Header;
			SymmEncryptedEnvelope documentEnvelope = new SymmEncryptedEnvelope( sheaf.InstanceLocator, header.MacAlg
				, header.CompressionAlg, header.SymmAlg, header.PasswordBlenderAlg );
			documentEnvelope.EncryptableDocument = sheaf;

			byte[] referredSheaf = documentEnvelope.SaveToBuffer( salt, encryptionKey, Provider.Serializer );
			ValidationUtils.ObjectHasContent( referredSheaf, "referredSheaf" );

			return referredSheaf;
		}

		/// <summary>
		/// Verifies if the specified sheaf (which was just saved) can be reloaded. If the sheaf can't be reloaded, a
		/// <see cref="SheafReloadException"/> exception is thrown.
		/// </summary>
		/// <param name="sheaf">The sheaf to reload from the storehouse. The update version of this sheaf is changed in here
		/// with the value from the loaded sheaf.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="encryptionKey">The key used for the encryption of the sheaf.</param>
		/// <remarks>
		/// <para>The reloaded sheaf is locked with <see cref="SheafLocks.Shared"/>.</para>
		/// 
		/// <para>Use this to double-check that a sheaf (which was just saved) can be reloaded. It's been observed that in a
		/// multi-threaded environment, sheafs were sometimes incorrectly saved. The issue has disappeared once the action
		/// service listener has been reorganized (which included an extra cloning for the serializer itself).</para>
		/// 
		/// <para>This is also useful for other potential errors with the storage hardware / software.</para>
		/// </remarks>
		private void VerifyReloadSheaf( ISheaf sheaf, string storehouseRealm, byte[] encryptionKey )
		{
			ValidationUtils.ObjectHasContent( sheaf, "sheaf" );

			byte[] originalSheafUpdateVersion = sheaf.SheafUpdateVersion;

			try
			{
				IDeserializer deserializer;
				ISheaf reloadedSheaf = LoadSheaf( sheaf.InstanceLocator, storehouseRealm
					, new PortableTypes( sheaf.GetType() ), encryptionKey, SheafLocks.Shared, out deserializer );

				if( reloadedSheaf == null )
				{
					throw new SheafReloadException( "Failed to reload the sheaf.", null
						, new ExceptionDetail( "Sheaf type", sheaf.GetType().FullName )
						, new ExceptionDetail( "Sheaf locator", sheaf.InstanceLocator ) );
				}

				IHashSerializer hashSerializer = deserializer.GetHashSerializer( true );
				if( !hashSerializer.EqualDocuments( sheaf, reloadedSheaf ) )
				{
					throw new SheafReloadException( "The reloaded sheaf is different than the original one.", null
						, new ExceptionDetail( "Sheaf type", sheaf.GetType().FullName )
						, new ExceptionDetail( "Sheaf locator", sheaf.InstanceLocator ) );
				}

				// The update version is not serialized, so it's not compared.
				sheaf.SetUpdateVersion( reloadedSheaf.SheafUpdateVersion );
			}
			catch( Exception ex )
			{
				if( BufferUtils.HasContent( originalSheafUpdateVersion ) )
					sheaf.SetUpdateVersion( originalSheafUpdateVersion );

				throw new SheafReloadException( "Failed to reload the sheaf", ex
					, new ExceptionDetail( "Sheaf type", sheaf.GetType().FullName )
					, new ExceptionDetail( "Sheaf locator", sheaf.InstanceLocator ) );
			}
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Commits and closes the current transaction in the storehouse.
		/// </summary>
		public void Commit()
		{
			CheckIfOpen();

			try
			{
				InternalCommit();
			}
			finally
			{
				Close();
			}
		}

		/// <summary>
		/// Drops (= rolls back) and closes the current transaction from the storehouse.
		/// </summary>
		/// <remarks>
		/// <para>Dropping a transaction means closing it without making the changes requested in it.</para>
		/// </remarks>
		public void Rollback()
		{
			try
			{
				if( IsOpen )
					Transaction.Rollback();
			}
			finally
			{
				Close();
			}
		}

		/// <summary>
		/// Adds the specified storehouse entropy to the storehouse. RESERVED for internal use.
		/// </summary>
		/// <param name="entropy">The storehouse entropy to add.</param>
		/// <remarks>
		/// <para>The entropy is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		public void AddEntropy( StorehouseEntropy entropy )
		{
			AddSheaf( entropy, StorehouseRealms.StorehouseConfiguration );
		}

		/// <summary>
		/// Updates the storehouse entropy with the specified one.
		/// </summary>
		/// <param name="entropy">The storehouse entropy to update.</param>
		/// <remarks>
		/// <para>The entropy is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		public void UpdateEntropy( StorehouseEntropy entropy )
		{
			UpdateSheaf( entropy, StorehouseRealms.StorehouseConfiguration );
		}

		/// <summary>
		/// Loads the entropy of the storehouse. An exception is thrown if the entropy is missing.
		/// </summary>
		/// <param name="sheafLock">Lock to put on the storehouse entropy document.</param>
		public StorehouseEntropy LoadEntropy( SheafLocks sheafLock )
		{
			try
			{
				IDeserializer deserializer;

				// Load the entropy.
				StorehouseEntropy entropy = LoadSheaf( StorehouseEntropy.PredefinedInstanceLocator
					, StorehouseRealms.StorehouseConfiguration, typeof( StorehouseEntropy ), sheafLock
					, out deserializer ) as StorehouseEntropy;
				ValidationUtils.ObjectType<StorehouseEntropy>( entropy, "entropy" );

				return entropy;
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsAbort( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				throw new FailedOpenStorehouseException( "Failed to open the storehouse. The entropy was not found.", ex );
			}
		}

		/// <summary>
		/// Adds the specified storehouse configuration to the storehouse. RESERVED for internal use.
		/// </summary>
		/// <param name="configuration">The storehouse configuration to add.</param>
		/// <remarks>
		/// <para>The configuration is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		public void AddConfiguration( StorehouseConfiguration configuration )
		{
			AddSheaf( configuration, StorehouseRealms.StorehouseConfiguration );
		}

		/// <summary>
		/// Updates the storehouse configuration with the specified one.
		/// </summary>
		/// <param name="configuration">The storehouse configuration to update.</param>
		/// <remarks>
		/// <para>The configuration is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		public void UpdateConfiguration( StorehouseConfiguration configuration )
		{
			UpdateSheaf( configuration, StorehouseRealms.StorehouseConfiguration );
		}

		/// <summary>
		/// Loads the configuration of the storehouse. An exception is thrown if the configuration is missing.
		/// </summary>
		/// <param name="sheafLock">Lock to put on the storehouse configuration document.</param>
		public StorehouseConfiguration LoadConfiguration( SheafLocks sheafLock )
		{
			try
			{
				IDeserializer deserializer;

				// Load the configuration.
				StorehouseConfiguration configuration = LoadSheaf( StorehouseConfiguration.PredefinedInstanceLocator
					, StorehouseRealms.StorehouseConfiguration, typeof( StorehouseConfiguration ), sheafLock
					, out deserializer ) as StorehouseConfiguration;
				ValidationUtils.ObjectType<StorehouseConfiguration>( configuration, "configuration" );

				return configuration;
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsAbort( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				throw new FailedOpenStorehouseException( "Failed to open the storehouse. The configuration was not found.", ex );
			}
		}

		/// <summary>
		/// Adds a speckle with the specified storehouse key.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the speckle to add.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="sheafType">The type of the sheaf referred by the added speckle.</param>
		/// <remarks>
		/// <para>The storehouse key of the added speckle might be the same as another storehouse key from the
		/// current storehouse; in such a case, an exception is thrown.</para>
		/// 
		/// <para>The speckle is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// </remarks>
		public void AddSpeckle( Locator storehouseKey, string storehouseRealm, Type sheafType )
		{
			ValidationUtils.ObjectHasContent( storehouseKey, "storehouseKey" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );
			ValidationUtils.ObjectHasContent( sheafType, "sheafType" );

			CheckIfOpen();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), storehouseKey, sheafType
				, SheafLocks.Exclusive );

			try
			{
				// Create the speckle.
				Speckle speckle = new Speckle( storehouseKey, sheafType.FullName );

				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script = ProviderTyped.ScriptsTyped.AddSpeckle_ExclusiveLock;

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					// Include the list of speckle filter columns and parameters in the script.
					script = ReplaceSpeckleFilterPlaceholders( script );

					dbCommand.CommandText = script;

					#region Add the parameters

					// Storehouse key.
					AddStorehouseKeyParam( dbCommand, speckle.StorehouseKey );

					// Sheaf type name.
					AddSheafTypeNameParam( dbCommand, speckle.SheafTypeName );

					// Speckle filters.
					AddSpeckleFilterParams( dbCommand, null, true );

					#endregion Add the parameters

					// Execute the command and check if one (and only one) database record was affected.
					using( DbDataReader reader = dbCommand.ExecuteReader() )
					{
						if( (reader == null) || (reader.RecordsAffected != 1) || reader.Read() )
						{
							throw new ValidationException( "The speckle was not added to the storehouse.", null, null
								, new ExceptionDetail( "Speckle locator", speckle.StorehouseKey )
								, new ExceptionDetail( "Sheaf type", sheafType.FullName ) );
						}
					}
				}
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		/// <summary>
		/// Counts the speckles with the specified speckle filters, which refer a document with the specified types.
		/// </summary>
		/// <param name="sheafTypes">The allowed sheaf types of the speckles to count.</param>
		/// <param name="seekFilters">The seek filters which identify the matching speckles. This may be null or empty.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <returns>Returns the number of matching speckles.</returns>
		/// <remarks>
		/// <para>The storehouse matching speckles are not locked.</para>
		/// 
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// </remarks>
		public int CountSpeckles( PortableTypes sheafTypes, SpeckleFilters seekFilters, string storehouseRealm )
		{
			ValidationUtils.ListHasContent( sheafTypes, "sheafTypes" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			CheckIfOpen();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), sheafTypes, seekFilters
				, SheafLocks.Shared );

			try
			{
				if( !sheafTypes.SameCommonAncestor() )
					throw new ValidationException( "The allowed sheaf types of the speckles to count are different." );

				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script = ProviderTyped.ScriptsTyped.CountSpeckles;

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					// Include the list of seek sheaf type name parameters in the script.
					script = ReplaceSheafTypeNameSeekList( script, sheafTypes );

					// Include the list of seek filter parameters in the script.
					script = ReplaceSpeckleFilterSeekList( script, seekFilters );

					dbCommand.CommandText = script;

					#region Add the parameters

					// Sheaf type names.
					AddSheafTypeNameParams( dbCommand, sheafTypes );

					// Seek filters.
					AddSpeckleFilterParams( dbCommand, seekFilters, false );

					#endregion Add the parameters

					// Read the result.
					object result = dbCommand.ExecuteScalar();
					ValidationUtils.ObjectType<int>( result, "result" );

					return (int)result;
				}
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		/// <summary>
		/// Loads the speckles with the specified speckle filters, which refer a document with the specified types.
		/// </summary>
		/// <param name="sheafTypes">The allowed sheaf types of the speckles to load.</param>
		/// <param name="seekFilters">The seek filters which identify the matching speckles. This may be null or empty.
		/// </param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="paging">Information about how to process the documents in a paged manner. If this is null, no paging is
		/// used.</param>
		/// <param name="collectedSpeckles">List where to add the collected speckles. If this is null, a new instance of
		/// <see cref="Speckles"/> is used instead. This allows you to have lists sorted by different criteria.</param>
		/// <param name="collectedSpeckleType">The type of the items to add to the list of collected speckles. This must inherit
		/// <see cref="Speckle"/> and must have a constructor which receives (as parameters) the storehouse key of the
		/// referred sheaf, the type name of the stored sheaf, and the loaded speckle filters. If this is null,
		/// <see cref="Speckle"/> is used instead. If this is different than null and <see cref="Speckle"/>,
		/// the performance is slightly reduced.</param>
		/// <param name="loadFilterIdentifiers">The speckle filter identifiers whose values to load from the matching
		/// speckles. This may be null or empty.</param>
		/// <returns>Returns the loaded speckles.</returns>
		/// <remarks>
		/// <para>The storehouse matching speckles are not locked.</para>
		/// 
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// 
		/// <para>If you never want (or it's never possible) to collect more than one speckle, for performance reasons you should
		/// collect the first speckle or document. This is because speckle collection doesn't stop at the first matching speckle,
		/// so the seek performance degrades geometrically with the number of documents from the storehouse.</para>
		/// </remarks>
		public IKernelObservableList CollectSpeckles( PortableTypes sheafTypes, SpeckleFilters seekFilters
			, string storehouseRealm, CollectPaging paging, IKernelObservableList collectedSpeckles
			, Type collectedSpeckleType, SpeckleFilterIdentifiers loadFilterIdentifiers )
		{
			ValidationUtils.ListHasContent( sheafTypes, "sheafTypes" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			CheckIfOpen();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), sheafTypes, seekFilters
				, SheafLocks.Shared );

			try
			{
				if( !sheafTypes.SameCommonAncestor() )
					throw new ValidationException( "The allowed sheaf types of the speckles to load are different." );

				#region Prepare dynamic types

				if( collectedSpeckles == null )
					collectedSpeckles = new Speckles();
				if( collectedSpeckleType == null )
					collectedSpeckleType = typeof( Speckle );

				ConstructorInfo collectedSpeckleTypeConstructor = CheckSpeckleType( collectedSpeckleType );

				#endregion Prepare dynamic types

				bool hasLoadFilterIdentifier = ListUtils.HasContent( loadFilterIdentifiers );

				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script = (paging == null) ? ProviderTyped.ScriptsTyped.CollectSpeckles
						: (paging.NewestFirst ? ProviderTyped.ScriptsTyped.CollectSpeckles_PagedNewest
							: ProviderTyped.ScriptsTyped.CollectSpeckles_PagedOldest);

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					// Include the list of seek sheaf type name parameters in the script.
					script = ReplaceSheafTypeNameSeekList( script, sheafTypes );

					// Include the list of seek filter parameters in the script.
					script = ReplaceSpeckleFilterSeekList( script, seekFilters );

					// Include the list of load speckle filter in the script.
					script = ReplaceSpeckleFilterLoadList( script, loadFilterIdentifiers );

					dbCommand.CommandText = script;

					#region Add the parameters

					// Sheaf type names.
					AddSheafTypeNameParams( dbCommand, sheafTypes );

					// Seek filters.
					AddSpeckleFilterParams( dbCommand, seekFilters, false );

					// Paging offset and count.
					AddPagingParams( dbCommand, paging );

					#endregion Add the parameters

					// Read the result.
					using( DbDataReader reader = dbCommand.ExecuteReader() )
					{
						while( (reader != null) && reader.Read() )
						{
							// Storehouse key.
							object storehouseKeyResult = reader[ DatabaseColumns.StorehouseKey ];
							ValidationUtils.ObjectType<byte[]>( storehouseKeyResult, "storehouseKeyResult" );

							// Sheaf type name.
							object sheafTypeNameResult = reader[ DatabaseColumns.SheafTypeName ];
							ValidationUtils.ObjectType<string>( sheafTypeNameResult, "sheafTypeNameResult" );

							// Speckle filters.
							SpeckleFilters loadedSpeckleFilters = hasLoadFilterIdentifier
								? LoadSpeckleFilters( reader, loadFilterIdentifiers ) : null;

							// Get the type (of the document) read from the storehouse, and double-check it.
							string readSheafTypeName = (string)sheafTypeNameResult;
							Type readSheafType = sheafTypes.Find( readSheafTypeName );
							if( readSheafType == null )
							{
								throw new ValidationException( "The type of the loaded sheaf is not allowed.", null, null
									, new ExceptionDetail( "Sheaf type", readSheafTypeName ) );
							}

							// Create the speckle.
							Locator storehouseKey = new Locator( (byte[])storehouseKeyResult );
							object speckle = (collectedSpeckleType == typeof( Speckle ))
								? new Speckle( storehouseKey, readSheafTypeName, loadedSpeckleFilters )
								: collectedSpeckleTypeConstructor.Invoke( new object[] { storehouseKey, readSheafTypeName
								, loadedSpeckleFilters } );

							collectedSpeckles.Add( speckle );
						}
					}
				}

				return collectedSpeckles;
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		/// <summary>
		/// Loads the first speckle with the specified speckle filters, which refer a document with the specified
		/// types.
		/// </summary>
		/// <param name="sheafTypes">The allowed sheaf types of the speckle to load.</param>
		/// <param name="seekFilters">The seek filters which identify the matching speckle. This may be null or empty.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="speckleLock">Lock to put on the speckle.</param>
		/// <param name="collectedSpeckleType">The type of the collected speckle. This must inherit <see cref="Speckle"/> and
		/// must have a constructor which receives (as parameters) the storehouse key of the referred sheaf, the type name of
		/// the stored sheaf, and the loaded speckle filters. If this is null, <see cref="Speckle"/> is used instead.</param>
		/// <param name="loadFilterIdentifiers">The speckle filter identifiers whose values to load from the matching
		/// speckle. This may be null or empty.</param>
		/// <returns>Returns the loaded speckle.</returns>
		/// <remarks>
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// 
		/// <para>The method returns the first encountered speckle, not necessarily the speckle which was added first. This
		/// behavior mVela.Imizes performance.</para>
		/// </remarks>
		public Speckle CollectFirstSpeckle( PortableTypes sheafTypes, SpeckleFilters seekFilters, string storehouseRealm
			, SheafLocks speckleLock, Type collectedSpeckleType, SpeckleFilterIdentifiers loadFilterIdentifiers )
		{
			ValidationUtils.ListHasContent( sheafTypes, "sheafTypes" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			CheckIfOpen();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), sheafTypes, seekFilters, speckleLock );

			try
			{
				if( !sheafTypes.SameCommonAncestor() )
					throw new ValidationException( "The allowed sheaf types of the speckle to load are different." );

				#region Prepare dynamic types

				if( collectedSpeckleType == null )
					collectedSpeckleType = typeof( Speckle );

				ConstructorInfo collectedSpeckleTypeConstructor = CheckSpeckleType( collectedSpeckleType );

				#endregion Prepare dynamic types

				bool hasLoadFilterIdentifier = ListUtils.HasContent( loadFilterIdentifiers );

				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script;

					// The script depends on the lock.
					switch( speckleLock )
					{
						case SheafLocks.Shared:
							script = ProviderTyped.ScriptsTyped.CollectFirstSpeckle_SharedLock;
							break;
						case SheafLocks.Update:
							script = ProviderTyped.ScriptsTyped.CollectFirstSpeckle_UpdateLock;
							break;
						case SheafLocks.Exclusive:
							script = ProviderTyped.ScriptsTyped.CollectFirstSpeckle_ExclusiveLock;
							break;
						default:
							throw ValidationUtils.Unhandled( speckleLock, "speckleLock" );
					}

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					// Include the list of seek sheaf type name parameters in the script.
					script = ReplaceSheafTypeNameSeekList( script, sheafTypes );

					// Include the list of seek filter parameters in the script.
					script = ReplaceSpeckleFilterSeekList( script, seekFilters );

					// Include the list of load speckle filter in the script.
					script = ReplaceSpeckleFilterLoadList( script, loadFilterIdentifiers );

					dbCommand.CommandText = script;

					#region Add the parameters

					// Sheaf type names.
					AddSheafTypeNameParams( dbCommand, sheafTypes );

					// Seek filters.
					AddSpeckleFilterParams( dbCommand, seekFilters, false );

					#endregion Add the parameters

					// Read the result.
					using( DbDataReader reader = dbCommand.ExecuteReader() )
					{
						if( (reader == null) || !reader.Read() )
						{
							// No speckle found.
							return null;
						}

						// Sheaf update version.
						object sheafUpdateVersionResult = reader[ DatabaseColumns.SheafUpdateVersion ];
						ValidationUtils.ObjectType<byte[]>( sheafUpdateVersionResult, "sheafUpdateVersionResult" );

						// Storehouse key.
						object storehouseKeyResult = reader[ DatabaseColumns.StorehouseKey ];
						ValidationUtils.ObjectType<byte[]>( storehouseKeyResult, "storehouseKeyResult" );

						// Sheaf type name.
						object sheafTypeNameResult = reader[ DatabaseColumns.SheafTypeName ];
						ValidationUtils.ObjectType<string>( sheafTypeNameResult, "sheafTypeNameResult" );

						// Speckle filters.
						SpeckleFilters loadedSpeckleFilters = hasLoadFilterIdentifier
							? LoadSpeckleFilters( reader, loadFilterIdentifiers ) : null;

						if( reader.Read() )
						{
							throw new ValidationException( "More than one speckle was found.", null, null
								, new ExceptionDetail( "Sheaf types", sheafTypes.ToString() )
								, new ExceptionDetail( "Seek filters", (seekFilters != null) ? seekFilters.ToString()
									: "" ) );
						}

						// Get the type (of the document) read from the storehouse, and double-check it.
						string readSheafTypeName = (string)sheafTypeNameResult;
						Type readSheafType = sheafTypes.Find( readSheafTypeName );
						if( readSheafType == null )
						{
							throw new ValidationException( "The type of the loaded sheaf is not allowed.", null, null
								, new ExceptionDetail( "Sheaf type", readSheafTypeName ) );
						}

						// Create the speckle.
						Locator storehouseKey = new Locator( (byte[])storehouseKeyResult );
						Speckle speckle = (collectedSpeckleType == typeof( Speckle ))
							? new Speckle( storehouseKey, readSheafTypeName, loadedSpeckleFilters )
							: collectedSpeckleTypeConstructor.Invoke( new object[] { storehouseKey, readSheafTypeName
							, loadedSpeckleFilters } ) as Speckle;

						speckle.SetLockAndUpdateVersion( speckleLock, (byte[])sheafUpdateVersionResult );

						LogManagerSpeckleLoaded( speckle );

						return speckle;
					}
				}
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		/// <summary>
		/// Loads the speckle with the specified storehouse key.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the speckle to load.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="speckleLock">Lock to put on the speckle.</param>
		/// <param name="loadedSpeckleType">The type of the speckle to load. This must inherit <see cref="Speckle"/> and must
		/// have a constructor which receives (as parameters) the storehouse key of the referred sheaf, the type name of the
		/// stored sheaf, and the loaded speckle filters. If this is null, <see cref="Speckle"/> is used instead.</param>
		/// <param name="loadFilterIdentifiers">The speckle filter identifiers whose values to load from the speckle.
		/// This may be null or empty.</param>
		/// <returns>Returns the loaded speckle, or null if none is found.</returns>
		public Speckle LoadSpeckle( Locator storehouseKey, string storehouseRealm, SheafLocks speckleLock, Type loadedSpeckleType
			, SpeckleFilterIdentifiers loadFilterIdentifiers )
		{
			ValidationUtils.ObjectHasContent( storehouseKey, "storehouseKey" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			CheckIfOpen();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), storehouseKey, loadedSpeckleType
				, speckleLock );

			try
			{
				#region Prepare dynamic types

				// Prepare the dynamic type.
				if( loadedSpeckleType == null )
					loadedSpeckleType = typeof( Speckle );

				ConstructorInfo loadedSpeckleTypeConstructor = CheckSpeckleType( loadedSpeckleType );

				#endregion Prepare dynamic types

				bool hasLoadFilterIdentifier = ListUtils.HasContent( loadFilterIdentifiers );

				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script;

					// The script depends on the lock.
					switch( speckleLock )
					{
						case SheafLocks.Shared:
							script = ProviderTyped.ScriptsTyped.LoadSpeckle_SharedLock;
							break;
						case SheafLocks.Update:
							script = ProviderTyped.ScriptsTyped.LoadSpeckle_UpdateLock;
							break;
						case SheafLocks.Exclusive:
							script = ProviderTyped.ScriptsTyped.LoadSpeckle_ExclusiveLock;
							break;
						default:
							throw ValidationUtils.Unhandled( speckleLock, "speckleLock" );
					}

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					// Include the list of load speckle filter in the script.
					script = ReplaceSpeckleFilterLoadList( script, loadFilterIdentifiers );

					dbCommand.CommandText = script;

					#region Add the parameters

					// Storehouse key.
					AddStorehouseKeyParam( dbCommand, storehouseKey );

					#endregion Add the parameters

					// Read the result.
					using( DbDataReader reader = dbCommand.ExecuteReader() )
					{
						if( (reader != null) && reader.Read() )
						{
							// Sheaf update version.
							object sheafUpdateVersionResult = reader[ DatabaseColumns.SheafUpdateVersion ];
							ValidationUtils.ObjectType<byte[]>( sheafUpdateVersionResult, "sheafUpdateVersionResult" );

							// Storehouse key.
							object storehouseKeyResult = reader[ DatabaseColumns.StorehouseKey ];
							ValidationUtils.ObjectType<byte[]>( storehouseKeyResult, "storehouseKeyResult" );

							// Sheaf type name.
							object sheafTypeNameResult = reader[ DatabaseColumns.SheafTypeName ];
							ValidationUtils.ObjectType<string>( sheafTypeNameResult, "sheafTypeNameResult" );

							// Speckle filters.
							SpeckleFilters loadedSpeckleFilters = hasLoadFilterIdentifier
								? LoadSpeckleFilters( reader, loadFilterIdentifiers ) : null;

							if( reader.Read() )
							{
								throw new ValidationException( "More than one speckle matched the storehouse key.", null, null
									, new ExceptionDetail( "Storehouse key", storehouseKey ) );
							}

							// Create the speckle.
							Locator speckleStorehouseKey = new Locator( (byte[])storehouseKeyResult );
							string speckleSheafTypeName = (string)sheafTypeNameResult;
							Speckle speckle = (loadedSpeckleType == typeof( Speckle ))
								? new Speckle( storehouseKey, speckleSheafTypeName, loadedSpeckleFilters )
								: loadedSpeckleTypeConstructor.Invoke( new object[] { speckleStorehouseKey, speckleSheafTypeName
								, loadedSpeckleFilters } ) as Speckle;

							speckle.SetLockAndUpdateVersion( speckleLock, (byte[])sheafUpdateVersionResult );

							return speckle;
						}
					}
				}

				// No speckle found.
				return null;
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		/// <summary>
		/// Checks if the speckle with the specified storehouse key has a referred sheaf.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the speckle to check.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <returns>Returns true if the speckle has a referred sheaf.</returns>
		/// <remarks>
		/// <para>The speckle filters are not loaded from the storehouse.</para>
		/// </remarks>
		public bool SpeckleHasSheaf( Locator storehouseKey, string storehouseRealm )
		{
			ValidationUtils.ObjectHasContent( storehouseKey, "storehouseKey" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			CheckIfOpen();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), storehouseKey, (Type)null
				, SheafLocks.Shared );

			try
			{
				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script = ProviderTyped.ScriptsTyped.SpeckleHasDocument;

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					dbCommand.CommandText = script;

					#region Add the parameters

					// Storehouse key.
					AddStorehouseKeyParam( dbCommand, storehouseKey );

					#endregion Add the parameters

					// Read the result.
					object result = dbCommand.ExecuteScalar();
					ValidationUtils.ObjectType<int>( result, "result" );

					return (int)result == 1;
				}
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		/// <summary>
		/// Loads the storehouse keys of the speckles with the specified speckle filters, which refer a document with the
		/// specified types.
		/// </summary>
		/// <param name="sheafTypes">The allowed sheaf types of the speckles whose keys to load.</param>
		/// <param name="seekFilters">The seek filters which identify the matching speckles. This may be null or empty.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="paging">Information about how to process the documents in a paged manner. If this is null, no paging is
		/// used.</param>
		/// <param name="collectedKeys">List where to add the collected keys. If this is null, a new instance of
		/// <see cref="Locators"/> is used instead. This allows you to have lists sorted by different criteria.</param>
		/// <param name="collectedKeyType">The type of the items to add to the list of collected keys. This must inherit
		/// <see cref="Locator"/> and must have a constructor which receives (as parameter) the locator's value. If this is
		/// null, <see cref="Locator"/> is used instead. If this is different than null and <see cref="Locator"/>, the
		/// performance is slightly reduced.</param>
		/// <returns>Returns the loaded storehouse keys.</returns>
		/// <remarks>
		/// <para>The storehouse matching speckles are not locked.</para>
		/// 
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// </remarks>
		public IKernelObservableList CollectKeys( PortableTypes sheafTypes, SpeckleFilters seekFilters
			, string storehouseRealm, CollectPaging paging, IKernelObservableList collectedKeys, Type collectedKeyType )
		{
			ValidationUtils.ListHasContent( sheafTypes, "sheafTypes" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			CheckIfOpen();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), sheafTypes, seekFilters
				, SheafLocks.Shared );

			try
			{
				if( !sheafTypes.SameCommonAncestor() )
					throw new ValidationException( "The allowed sheaf types of the speckles whose keys to load are different." );

				#region Prepare dynamic types

				if( collectedKeys == null )
					collectedKeys = new Locators();
				if( collectedKeyType == null )
					collectedKeyType = typeof( Locator );

				ConstructorInfo collectedKeyTypeConstructor = CheckStorehouseKeyType( collectedKeyType );

				#endregion Prepare dynamic types

				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script = (paging == null) ? ProviderTyped.ScriptsTyped.CollectKeys
						: (paging.NewestFirst ? ProviderTyped.ScriptsTyped.CollectKeys_PagedNewest
							: ProviderTyped.ScriptsTyped.CollectKeys_PagedOldest);

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					// Include the list of seek sheaf type name parameters in the script.
					script = ReplaceSheafTypeNameSeekList( script, sheafTypes );

					// Include the list of seek filter parameters in the script.
					script = ReplaceSpeckleFilterSeekList( script, seekFilters );

					dbCommand.CommandText = script;

					#region Add the parameters

					// Sheaf type names.
					AddSheafTypeNameParams( dbCommand, sheafTypes );

					// Seek filters.
					AddSpeckleFilterParams( dbCommand, seekFilters, false );

					// Paging offset and count.
					AddPagingParams( dbCommand, paging );

					#endregion Add the parameters

					// Read the result.
					using( DbDataReader reader = dbCommand.ExecuteReader() )
					{
						while( (reader != null) && reader.Read() )
						{
							// Storehouse key.
							object storehouseKeyResult = reader[ DatabaseColumns.StorehouseKey ];
							ValidationUtils.ObjectType<byte[]>( storehouseKeyResult, "storehouseKeyResult" );

							// Create the storehouse key.
							object storehouseKey = (collectedKeyType == typeof( Locator ))
								? new Locator( (byte[])storehouseKeyResult )
								: collectedKeyTypeConstructor.Invoke( new object[] { (byte[])storehouseKeyResult } );

							collectedKeys.Add( storehouseKey );
						}
					}
				}

				return collectedKeys;
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		/// <summary>
		/// Adds the specified sheaf to the storehouse.
		/// </summary>
		/// <param name="sheaf">The sheaf to add to the storehouse.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="salt">The (random) salt used to compute the hash and encrypt the sheaf. This is stored in unencrypted
		/// form. If this is null or empty, a random salt is used instead.</param>
		/// <param name="encryptionKey">The key used for the encryption of the sheaf.</param>
		/// <remarks>
		/// <para>If a sheaf with the same storehouse key already exists, an exception is thrown.</para>
		/// 
		/// <para>The sheaf is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>The <see cref="Sheaf.UpdateTime"/> property is automatically set in here with the current
		/// time.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		public void AddSheaf( ISheaf sheaf, string storehouseRealm, byte[] salt, byte[] encryptionKey )
		{
			ValidationUtils.ObjectHasContent( sheaf, "sheaf" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			CheckIfOpen();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), sheaf.InstanceLocator, sheaf.GetType()
				, SheafLocks.Exclusive );

			try
			{
				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script = ProviderTyped.ScriptsTyped.AddSheaf_ExclusiveLock;

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					// Include the list of speckle filter columns and parameters in the script.
					script = ReplaceSpeckleFilterPlaceholders( script );

					dbCommand.CommandText = script;

					// Set the update time before the parameters are added.
					sheaf.UpdateTime = DateTime.UtcNow;

					#region Add the parameters

					// Storehouse key.
					AddStorehouseKeyParam( dbCommand, sheaf );

					// Sheaf type name.
					AddSheafTypeNameParam( dbCommand, sheaf );

					// Speckle filters.
					SpeckleFilters filters = sheaf.CollectSpeckleFilters();
					AddSpeckleFilterParams( dbCommand, filters, true );

					// Referred sheaf.
					byte[] referredSheaf = GetReferredSheaf( sheaf, salt, encryptionKey );
					ProviderTyped.AddBlobCommandParam( dbCommand, DatabaseParameters.ReferredSheaf, referredSheaf );

					#endregion Add the parameters

					// Execute the command and check if one (and only one) database record was affected.
					using( DbDataReader reader = dbCommand.ExecuteReader() )
					{
						if( (reader == null) || (reader.RecordsAffected != 1) || reader.Read() )
						{
							throw new ValidationException( "The sheaf with the was not added (or more records were affected)."
								, null, null, new ExceptionDetail( "Sheaf locator", sheaf.InstanceLocator ) );
						}
					}
				}

				sheaf.SetLock( SheafLocks.Exclusive );

				VerifyReloadSheaf( sheaf, storehouseRealm, encryptionKey );
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		/// <summary>
		/// Adds the specified sheaf to the storehouse.
		/// </summary>
		/// <param name="sheaf">The sheaf to add to the storehouse.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <remarks>
		/// <para>If a sheaf with the same storehouse key already exists, an exception is thrown.</para>
		/// 
		/// <para>The sheaf is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>The <see cref="Sheaf.UpdateTime"/> property is automatically set in here with the current
		/// time.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		public void AddSheaf( ISheaf sheaf, string storehouseRealm )
		{
			AddSheaf( sheaf, storehouseRealm, null, GetSheafEncryptionKey( sheaf ) );
		}

		/// <summary>
		/// Updates the specified document in the storehouse.
		/// </summary>
		/// <param name="sheaf">The sheaf to update in the storehouse.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="salt">The (random) salt used to compute the hash and encrypt the sheaf. This is stored in unencrypted
		/// form. If this is null or empty, a random salt is used instead.</param>
		/// <param name="encryptionKey">The key used for the encryption of the sheaf.</param>
		/// <remarks>
		/// <para>If a sheaf with the same storehouse key already exists, it is replaced.</para>
		/// 
		/// <para>The sheaf must be locked for writing.</para>
		/// 
		/// <para>The <see cref="Sheaf.UpdateTime"/> property is automatically set in here with the current
		/// time.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		public void UpdateSheaf( ISheaf sheaf, string storehouseRealm, byte[] salt, byte[] encryptionKey )
		{
			ValidationUtils.ObjectHasContent( sheaf, "sheaf" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			CheckIfOpen();

			if( (sheaf.StorehouseLock != SheafLocks.Update) && (sheaf.StorehouseLock != SheafLocks.Exclusive) )
			{
				throw new ValidationException( "The sheaf was not locked for writing, but an attempt to update it is being"
					+ " made.", null, null, new ExceptionDetail( "Sheaf locator", sheaf.InstanceLocator ) );
			}

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), sheaf.InstanceLocator, sheaf.GetType()
				, sheaf.StorehouseLock );

			try
			{
				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script;

					// The script depends on the lock.
					switch( sheaf.StorehouseLock )
					{
						case SheafLocks.Update:
							script = ProviderTyped.ScriptsTyped.UpdateSheaf_UpdateLock;
							break;
						case SheafLocks.Exclusive:
							script = ProviderTyped.ScriptsTyped.UpdateSheaf_ExclusiveLock;
							break;
						default:
							throw ValidationUtils.Unhandled( sheaf.StorehouseLock, "sheaf.StorehouseLock" );
					}

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					// Include the list of speckle filter columns and parameters in the script.
					script = ReplaceSpeckleFilterPlaceholders( script );

					dbCommand.CommandText = script;

					// Set the update time before the parameters are added.
					sheaf.UpdateTime = DateTime.UtcNow;

					#region Add the parameters

					// Storehouse key.
					AddStorehouseKeyParam( dbCommand, sheaf );

					// Sheaf update version.
					ProviderTyped.AddCommandParam( dbCommand, DatabaseParameters.SheafUpdateVersion, DbType.Binary, false
						, sheaf.SheafUpdateVersion );

					// Sheaf type name.
					AddSheafTypeNameParam( dbCommand, sheaf );

					// Speckle filters.
					SpeckleFilters filters = sheaf.CollectSpeckleFilters();
					AddSpeckleFilterParams( dbCommand, filters, true );

					byte[] referredSheaf = GetReferredSheaf( sheaf, salt, encryptionKey );
					ProviderTyped.AddBlobCommandParam( dbCommand, DatabaseParameters.ReferredSheaf, referredSheaf );

					#endregion Add the parameters

					// Execute the command and check if one (and only one) database record was affected.
					using( DbDataReader reader = dbCommand.ExecuteReader() )
					{
						if( (reader == null) || (reader.RecordsAffected != 1) || reader.Read() )
						{
							throw new SheafUpdateVersionException( "The sheaf was not updated (or more records were affected)."
									+ " It's possible that this has happened because another thread has updated the document"
									+ " after the current thread has read it.", null
								, new ExceptionDetail( "Sheaf locator", sheaf.InstanceLocator )
								, new ExceptionDetail( "Sheaf update version"
									, TextUtils.ArrayToText16( sheaf.SheafUpdateVersion ) ) );
						}
					}
				}

				VerifyReloadSheaf( sheaf, storehouseRealm, encryptionKey );
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		/// <summary>
		/// Updates the specified document in the storehouse.
		/// </summary>
		/// <param name="sheaf">The sheaf to update in the storehouse.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <remarks>
		/// <para>If a sheaf with the same storehouse key already exists, it is replaced.</para>
		/// 
		/// <para>The sheaf must be locked for writing.</para>
		/// 
		/// <para>The <see cref="Sheaf.UpdateTime"/> property is automatically set in here with the current
		/// time.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		public void UpdateSheaf( ISheaf sheaf, string storehouseRealm )
		{
			UpdateSheaf( sheaf, storehouseRealm, null, GetSheafEncryptionKey( sheaf ) );
		}

		/// <summary>
		/// Deletes the sheafs with the specified speckle filters, which refer documents with the specified types.
		/// </summary>
		/// <param name="sheafTypes">The allowed sheaf types of the sheafs to delete.</param>
		/// <param name="seekFilters">The seek filters which identify the matching sheafs.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <returns>Returns the number of deleted sheafs.</returns>
		/// <remarks>
		/// <para>The sheafs are locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// </remarks>
		public int DeleteSheafs( PortableTypes sheafTypes, SpeckleFilters seekFilters, string storehouseRealm )
		{
			ValidationUtils.ListHasContent( sheafTypes, "sheafTypes" );
			ValidationUtils.ListHasContent( seekFilters, "seekFilters" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			CheckIfOpen();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), sheafTypes, seekFilters
				, SheafLocks.Exclusive );

			try
			{
				if( !sheafTypes.SameCommonAncestor() )
					throw new ValidationException( "The allowed sheaf types of the sheafs to delete are different." );

				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script = ProviderTyped.ScriptsTyped.DeleteSheafs_ExclusiveLock;

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					// Include the list of seek sheaf type name parameters in the script.
					script = ReplaceSheafTypeNameSeekList( script, sheafTypes );

					// Include the list of seek filter parameters in the script.
					script = ReplaceSpeckleFilterSeekList( script, seekFilters );

					dbCommand.CommandText = script;

					#region Add the parameters

					// Sheaf type names.
					AddSheafTypeNameParams( dbCommand, sheafTypes );

					// Seek filters.
					AddSpeckleFilterParams( dbCommand, seekFilters, false );

					#endregion Add the parameters

					// Execute the command.
					return dbCommand.ExecuteNonQuery();
				}
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		/// <summary>
		/// Deletes the first sheaf with the specified speckle filters, which refer documents with the specified types.
		/// </summary>
		/// <param name="sheafTypes">The allowed sheaf types of the sheaf to delete.</param>
		/// <param name="seekFilters">The seek filters which identify the matching sheaf.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <remarks>
		/// <para>The sheaf is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// 
		/// <para>The method deletes the first encountered sheaf, not necessarily the sheaf which was added first. This behavior
		/// mVela.Imizes performance.</para>
		/// </remarks>
		public void DeleteFirstSheaf( PortableTypes sheafTypes, SpeckleFilters seekFilters, string storehouseRealm )
		{
			ValidationUtils.ListHasContent( sheafTypes, "sheafTypes" );
			ValidationUtils.ListHasContent( seekFilters, "seekFilters" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			CheckIfOpen();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), sheafTypes, seekFilters
				, SheafLocks.Exclusive );

			try
			{
				if( !sheafTypes.SameCommonAncestor() )
					throw new ValidationException( "The allowed sheaf types of the sheaf to delete are different." );

				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script = ProviderTyped.ScriptsTyped.DeleteFirstSheaf_ExclusiveLock;

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					// Include the list of seek sheaf type name parameters in the script.
					script = ReplaceSheafTypeNameSeekList( script, sheafTypes );

					// Include the list of seek filter parameters in the script.
					script = ReplaceSpeckleFilterSeekList( script, seekFilters );

					dbCommand.CommandText = script;

					#region Add the parameters

					// Sheaf type names.
					AddSheafTypeNameParams( dbCommand, sheafTypes );

					// Seek filters.
					AddSpeckleFilterParams( dbCommand, seekFilters, false );

					#endregion Add the parameters

					// Execute the command and check if at most one database record was affected.
					int recordsAffected = dbCommand.ExecuteNonQuery();
					if( (recordsAffected < 0) || (recordsAffected > 1) )
					{
						throw new ValidationException( "More than one sheaf was deleted.", null, null
							, new ExceptionDetail( "Sheaf types", sheafTypes.ToString() )
							, new ExceptionDetail( "Seek filters", (seekFilters != null) ? seekFilters.ToString() : "" ) );
					}
				}
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		/// <summary>
		/// Deletes the sheaf with the specified storehouse key from the storehouse.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the sheaf to delete from the storehouse.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <remarks>
		/// <para>The sheaf is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// </remarks>
		public void DeleteSheaf( Locator storehouseKey, string storehouseRealm )
		{
			ValidationUtils.ObjectHasContent( storehouseKey, "storehouseKey" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			CheckIfOpen();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), storehouseKey, (Type)null
				, SheafLocks.Exclusive );

			try
			{
				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script = ProviderTyped.ScriptsTyped.DeleteSheaf_ExclusiveLock;

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					dbCommand.CommandText = script;

					#region Add the parameters

					// Storehouse key.
					AddStorehouseKeyParam( dbCommand, storehouseKey );

					#endregion Add the parameters

					// Execute the command and check if at most one database record was affected.
					int recordsAffected = dbCommand.ExecuteNonQuery();
					if( (recordsAffected < 0) || (recordsAffected > 1) )
					{
						throw new ValidationException( "More than one sheaf was deleted.", null, null
							, new ExceptionDetail( "Sheaf locator", storehouseKey ) );
					}
				}
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		/// <summary>
		/// Loads the sheafs with the specified speckle filters, which have the specified types.
		/// </summary>
		/// <param name="sheafTypes">The allowed types of the sheafs to load.</param>
		/// <param name="seekFilters">The seek filters which identify the speckles of the matching storehouse documents. This
		/// may be null or empty.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="paging">Information about how to process the documents in a paged manner. If this is null, no paging is
		/// used.</param>
		/// <param name="sheafLock">Lock to put on the sheafs.</param>
		/// <param name="collectedSheafs">List where to add the collected sheafs. If this is null, a new instance of
		/// <see cref="Sheafs"/> is used instead. This allows you to have lists sorted by different criteria.</param>
		/// <param name="deserializer">Is set with the deserializer of the last loaded sheaf.</param>
		/// <returns>Returns the loaded sheafs.</returns>
		/// <remarks>
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// </remarks>
		public IKernelObservableList CollectSheafs( PortableTypes sheafTypes, SpeckleFilters seekFilters
			, string storehouseRealm, CollectPaging paging, SheafLocks sheafLock, IKernelObservableList collectedSheafs
			, out IDeserializer deserializer )
		{
			ValidationUtils.ListHasContent( sheafTypes, "sheafTypes" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			CheckIfOpen();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), sheafTypes, seekFilters, sheafLock );

			try
			{
				if( !sheafTypes.SameCommonAncestor() )
					throw new ValidationException( "The allowed sheaf types of the sheafs to load are different." );

				if( collectedSheafs == null )
					collectedSheafs = new Sheafs();

				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script;

					// The script depends on the lock.
					switch( sheafLock )
					{
						case SheafLocks.Shared:
							script = (paging == null) ? ProviderTyped.ScriptsTyped.CollectSheafs_SharedLock
								: (paging.NewestFirst ? ProviderTyped.ScriptsTyped.CollectSheafs_PagedNewest_SharedLock
									: ProviderTyped.ScriptsTyped.CollectSheafs_PagedOldest_SharedLock);
							break;
						case SheafLocks.Update:
							script = (paging == null) ? ProviderTyped.ScriptsTyped.CollectSheafs_UpdateLock
								: (paging.NewestFirst ? ProviderTyped.ScriptsTyped.CollectSheafs_PagedNewest_UpdateLock
									: ProviderTyped.ScriptsTyped.CollectSheafs_PagedOldest_UpdateLock);
							break;
						case SheafLocks.Exclusive:
							script = (paging == null) ? ProviderTyped.ScriptsTyped.CollectSheafs_ExclusiveLock
								: (paging.NewestFirst ? ProviderTyped.ScriptsTyped.CollectSheafs_PagedNewest_ExclusiveLock
									: ProviderTyped.ScriptsTyped.CollectSheafs_PagedOldest_ExclusiveLock);
							break;
						default:
							throw ValidationUtils.Unhandled( sheafLock, "sheafLock" );
					}

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					// Include the list of seek sheaf type name parameters in the script.
					script = ReplaceSheafTypeNameSeekList( script, sheafTypes );

					// Include the list of seek filter parameters in the script.
					script = ReplaceSpeckleFilterSeekList( script, seekFilters );

					dbCommand.CommandText = script;

					#region Add the parameters

					// Sheaf type names.
					AddSheafTypeNameParams( dbCommand, sheafTypes );

					// Seek filters.
					AddSpeckleFilterParams( dbCommand, seekFilters, false );

					// Paging offset and count.
					AddPagingParams( dbCommand, paging );

					#endregion Add the parameters

					deserializer = null;

					// Read the result.
					using( DbDataReader reader = dbCommand.ExecuteReader() )
					{
						while( (reader != null) && reader.Read() )
						{
							// Sheaf update version.
							object sheafUpdateVersionResult = reader[ DatabaseColumns.SheafUpdateVersion ];
							ValidationUtils.ObjectType<byte[]>( sheafUpdateVersionResult, "sheafUpdateVersionResult" );

							// Storehouse key.
							object storehouseKeyResult = reader[ DatabaseColumns.StorehouseKey ];
							ValidationUtils.ObjectType<byte[]>( storehouseKeyResult, "storehouseKeyResult" );

							// Sheaf type name.
							object sheafTypeNameResult = reader[ DatabaseColumns.SheafTypeName ];
							ValidationUtils.ObjectType<string>( sheafTypeNameResult, "sheafTypeNameResult" );

							// Referred sheaf.
							object referredSheafResult = reader[ DatabaseColumns.ReferredSheaf ];
							ValidationUtils.ObjectType<byte[]>( referredSheafResult, "referredSheafResult" );

							Locator storehouseKey = new Locator( (byte[])storehouseKeyResult );
							byte[] encryptionKey = GetSheafEncryptionKey( storehouseKey );

							// Get the type (of the document) read from the storehouse, and double-check it.
							string readSheafTypeName = (string)sheafTypeNameResult;
							Type readSheafType = sheafTypes.Find( readSheafTypeName );
							if( readSheafType == null )
							{
								throw new ValidationException( "The type of the loaded sheaf is not allowed.", null, null
									, new ExceptionDetail( "Sheaf type", readSheafTypeName ) );
							}

							// Decrypt the sheaf.
							SymmEncryptedEnvelope see = SymmEncryptedEnvelope.LoadFromBuffer( (byte[])referredSheafResult
								, encryptionKey, readSheafType, out deserializer );
							ValidationUtils.ObjectHasContent( see, "see" );

							// Check the sheaf type.
							Type loadedSheafType = see.EncryptableDocument.GetType();
							if( readSheafType != loadedSheafType )
							{
								throw new ValidationException( "The type of the loaded sheaf is not allowed.", null, null
									, new ExceptionDetail( "Sheaf type", loadedSheafType.FullName ) );
							}

							// Create the sheaf.
							ISheaf sheaf = (ISheaf)see.EncryptableDocument;
							sheaf.SetLockAndUpdateVersion( sheafLock, (byte[])sheafUpdateVersionResult );

							collectedSheafs.Add( sheaf );
						}
					}
				}

				return collectedSheafs;
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		/// <summary>
		/// Loads the first sheaf with the specified speckle filters, which has the specified types.
		/// </summary>
		/// <param name="sheafTypes">The allowed types of the sheaf to load.</param>
		/// <param name="seekFilters">The seek filters which identify the speckles of the matching storehouse documents. This
		/// may be null or empty.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="sheafLock">Lock to put on the sheaf.</param>
		/// <param name="deserializer">Is set with the deserializer of the loaded sheaf.</param>
		/// <returns>Returns the loaded sheaf, or null if none is found.</returns>
		/// <remarks>
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// 
		/// <para>The method returns the first encountered sheaf, not necessarily the sheaf which was added first. This behavior
		/// mVela.Imizes performance.</para>
		/// </remarks>
		public ISheaf CollectFirstSheaf( PortableTypes sheafTypes, SpeckleFilters seekFilters, string storehouseRealm
			, SheafLocks sheafLock, out IDeserializer deserializer )
		{
			ValidationUtils.ListHasContent( sheafTypes, "sheafTypes" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			CheckIfOpen();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), sheafTypes, seekFilters, sheafLock );

			try
			{
				if( !sheafTypes.SameCommonAncestor() )
					throw new ValidationException( "The allowed sheaf types of the sheaf to load are different." );

				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script;

					// The script depends on the lock.
					switch( sheafLock )
					{
						case SheafLocks.Shared:
							script = ProviderTyped.ScriptsTyped.CollectFirstSheaf_SharedLock;
							break;
						case SheafLocks.Update:
							script = ProviderTyped.ScriptsTyped.CollectFirstSheaf_UpdateLock;
							break;
						case SheafLocks.Exclusive:
							script = ProviderTyped.ScriptsTyped.CollectFirstSheaf_ExclusiveLock;
							break;
						default:
							throw ValidationUtils.Unhandled( sheafLock, "sheafLock" );
					}

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					// Include the list of seek sheaf type name parameters in the script.
					script = ReplaceSheafTypeNameSeekList( script, sheafTypes );

					// Include the list of seek filter parameters in the script.
					script = ReplaceSpeckleFilterSeekList( script, seekFilters );

					dbCommand.CommandText = script;

					#region Add the parameters

					// Sheaf type names.
					AddSheafTypeNameParams( dbCommand, sheafTypes );

					// Seek filters.
					AddSpeckleFilterParams( dbCommand, seekFilters, false );

					#endregion Add the parameters

					// Read the result.
					using( DbDataReader reader = dbCommand.ExecuteReader() )
					{
						if( (reader == null) || !reader.Read() )
						{
							deserializer = null;

							// No document found.
							return null;
						}

						// Sheaf update version.
						object sheafUpdateVersionResult = reader[ DatabaseColumns.SheafUpdateVersion ];
						ValidationUtils.ObjectType<byte[]>( sheafUpdateVersionResult, "sheafUpdateVersionResult" );

						// Storehouse key.
						object storehouseKeyResult = reader[ DatabaseColumns.StorehouseKey ];
						ValidationUtils.ObjectType<byte[]>( storehouseKeyResult, "storehouseKeyResult" );

						// Sheaf type name.
						object sheafTypeNameResult = reader[ DatabaseColumns.SheafTypeName ];
						ValidationUtils.ObjectType<string>( sheafTypeNameResult, "sheafTypeNameResult" );

						// Referred sheaf.
						object referredSheafResult = reader[ DatabaseColumns.ReferredSheaf ];
						ValidationUtils.ObjectType<byte[]>( referredSheafResult, "referredSheafResult" );

						Locator storehouseKey = new Locator( (byte[])storehouseKeyResult );
						byte[] encryptionKey = GetSheafEncryptionKey( storehouseKey );

						// Get the type (of the document) read from the storehouse, and double-check it.
						string readSheafTypeName = (string)sheafTypeNameResult;
						Type readSheafType = sheafTypes.Find( readSheafTypeName );
						if( readSheafType == null )
						{
							throw new ValidationException( "The type of the loaded sheaf is not allowed.", null, null
								, new ExceptionDetail( "Sheaf type", readSheafTypeName ) );
						}

						// Decrypt the sheaf.
						SymmEncryptedEnvelope see = SymmEncryptedEnvelope.LoadFromBuffer( (byte[])referredSheafResult
							, encryptionKey, readSheafType, out deserializer );
						ValidationUtils.ObjectHasContent( see, "see" );

						// Check the sheaf type.
						Type loadedSheafType = see.EncryptableDocument.GetType();
						if( readSheafType != loadedSheafType )
						{
							throw new ValidationException( "The type of the loaded sheaf is not allowed.", null, null
								, new ExceptionDetail( "Sheaf type", loadedSheafType.FullName ) );
						}

						// Create the sheaf.
						ISheaf sheaf = (ISheaf)see.EncryptableDocument;
						sheaf.SetLockAndUpdateVersion( sheafLock, (byte[])sheafUpdateVersionResult );

						if( reader.Read() )
						{
							throw new ValidationException( "More than one sheaf was found.", null, null
								, new ExceptionDetail( "Sheaf types", sheafTypes.ToString() )
								, new ExceptionDetail( "Seek filters", (seekFilters != null) ? seekFilters.ToString()
									: "" ) );
						}

						LogManagerSheafLoaded( sheaf );

						return sheaf;
					}
				}
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		/// <summary>
		/// Loads the sheaf with the specified storehouse key, which has the specified types.
		/// </summary>
		/// <param name="storehouseKey">Storehouse key (= document instance locator) of the sheaf to load.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="sheafTypes">The allowed types of the sheaf to load.</param>
		/// <param name="encryptionKey">The key used for the decryption of the sheaf.</param>
		/// <param name="sheafLock">Lock to put on the sheaf.</param>
		/// <param name="deserializer">Is set with the deserializer of the loaded sheaf.</param>
		/// <returns>Returns the loaded sheaf, or null if none is found.</returns>
		/// <remarks>
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// </remarks>
		public ISheaf LoadSheaf( Locator storehouseKey, string storehouseRealm, PortableTypes sheafTypes
			, byte[] encryptionKey, SheafLocks sheafLock, out IDeserializer deserializer )
		{
			ValidationUtils.ObjectHasContent( storehouseKey, "storehouseKey" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );
			ValidationUtils.ListHasContent( sheafTypes, "sheafTypes" );
			ValidationUtils.ArrayHasContent( encryptionKey, "encryptionKey" );

			CheckIfOpen();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), storehouseKey, sheafTypes, sheafLock );

			try
			{
				if( !sheafTypes.SameCommonAncestor() )
					throw new ValidationException( "The allowed sheaf types of the sheaf to load are different." );

				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script;

					// The script depends on the lock.
					switch( sheafLock )
					{
						case SheafLocks.Shared:
							script = ProviderTyped.ScriptsTyped.LoadSheaf_SharedLock;
							break;
						case SheafLocks.Update:
							script = ProviderTyped.ScriptsTyped.LoadSheaf_UpdateLock;
							break;
						case SheafLocks.Exclusive:
							script = ProviderTyped.ScriptsTyped.LoadSheaf_ExclusiveLock;
							break;
						default:
							throw ValidationUtils.Unhandled( sheafLock, "sheafLock" );
					}

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					dbCommand.CommandText = script;

					#region Add the parameters

					// Storehouse key.
					AddStorehouseKeyParam( dbCommand, storehouseKey );

					#endregion Add the parameters

					// Read the result.
					using( DbDataReader reader = dbCommand.ExecuteReader() )
					{
						if( (reader == null) || !reader.Read() )
						{
							deserializer = null;

							// No document found.
							return null;
						}

						// Sheaf update version.
						object sheafUpdateVersionResult = reader[ DatabaseColumns.SheafUpdateVersion ];
						ValidationUtils.ObjectType<byte[]>( sheafUpdateVersionResult, "sheafUpdateVersionResult" );

						// Sheaf type name.
						object sheafTypeNameResult = reader[ DatabaseColumns.SheafTypeName ];
						ValidationUtils.ObjectType<string>( sheafTypeNameResult, "sheafTypeNameResult" );

						// Referred sheaf.
						object referredSheafResult = reader[ DatabaseColumns.ReferredSheaf ];
						ValidationUtils.ObjectType<byte[]>( referredSheafResult, "referredSheafResult" );

						if( reader.Read() )
						{
							throw new ValidationException( "More than one sheaf matched the storehouse key.", null, null
								, new ExceptionDetail( "Storehouse key", storehouseKey ) );
						}

						// Get the type (of the document) read from the storehouse, and double-check it.
						string readSheafTypeName = (string)sheafTypeNameResult;
						Type readSheafType = sheafTypes.Find( readSheafTypeName );
						if( readSheafType == null )
						{
							throw new ValidationException( "The type of the loaded sheaf is not allowed.", null, null
								, new ExceptionDetail( "Sheaf type", readSheafTypeName ) );
						}

						// Decrypt the sheaf.
						SymmEncryptedEnvelope see = SymmEncryptedEnvelope.LoadFromBuffer( (byte[])referredSheafResult
							, encryptionKey, readSheafType, out deserializer );
						ValidationUtils.ObjectHasContent( see, "see" );

						// Check the sheaf type.
						Type loadedSheafType = see.EncryptableDocument.GetType();
						if( sheafTypes.FindByItem( loadedSheafType ) == null )
						{
							throw new ValidationException( "The type of the loaded sheaf is not allowed.", null, null
								, new ExceptionDetail( "Sheaf type", loadedSheafType.FullName ) );
						}

						// Create the sheaf.
						ISheaf sheaf = (ISheaf)see.EncryptableDocument;
						sheaf.SetLockAndUpdateVersion( sheafLock, (byte[])sheafUpdateVersionResult );

						return sheaf;
					}
				}
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		/// <summary>
		/// Loads the sheaf with the specified storehouse key, which has the specified type (or a derived one).
		/// </summary>
		/// <param name="storehouseKey">Storehouse key (= document instance locator) of the sheaf to load.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="sheafType">The type of the sheaf to load; the sheaf to load may either have this type or a type derived
		/// from this (if it's found in the global portable types from <see cref="PortableTypes.Instance"/>).</param>
		/// <param name="sheafLock">Lock to put on the sheaf.</param>
		/// <param name="deserializer">Is set with the deserializer of the loaded sheaf.</param>
		/// <returns>Returns the loaded sheaf, or null if none is found.</returns>
		public ISheaf LoadSheaf( Locator storehouseKey, string storehouseRealm, Type sheafType, SheafLocks sheafLock
			, out IDeserializer deserializer )
		{
			PortableTypes sheafTypes = PortableTypes.FindAllDerived( sheafType );

			return LoadSheaf( storehouseKey, storehouseRealm, sheafTypes, GetSheafEncryptionKey( storehouseKey ), sheafLock
				, out deserializer );
		}

		/// <summary>
		/// Checks if the specified sheaf (already) exists in the specified storehouse realm and has the same content.
		/// </summary>
		/// <param name="sheaf">Sheaf to check if it (already) exists.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <returns>Returns the existence mode of the sheaf.</returns>
		/// <remarks>
		/// <para>Use the <see cref="SpeckleHasSheaf"/> method if you want to know if a given instance locator is already
		/// allocated in the specified realm.</para>
		/// </remarks>
		public SheafExistenceModes SheafExists( ISheaf sheaf, string storehouseRealm )
		{
			ValidationUtils.ObjectHasContent( sheaf, "sheaf" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			IDeserializer deserializer;
			ISheaf existingSheaf = LoadSheaf( sheaf.InstanceLocator, storehouseRealm, sheaf.GetType(), SheafLocks.Shared
				, out deserializer );
			if( existingSheaf == null )
				return SheafExistenceModes.DoesntExist;

			return deserializer.EqualDocuments( sheaf, existingSheaf ) ? SheafExistenceModes.IsIdentical
				: SheafExistenceModes.IsDifferent;
		}

		/// <summary>
		/// Gets the type of the sheaf with the specified storehouse key.
		/// </summary>
		/// <param name="storehouseKey">Storehouse key (= document instance locator) of the sheaf whose type to get.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <returns>Returns the type of the sheaf, or null if none is found.</returns>
		public Type GetSheafType( Locator storehouseKey, string storehouseRealm )
		{
			ValidationUtils.ObjectHasContent( storehouseKey, "storehouseKey" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			CheckIfOpen();

			LogManager logManager = LogManagerStartMethod( MethodBase.GetCurrentMethod(), storehouseKey, (Type)null
				, SheafLocks.Shared );

			try
			{
				using( DbCommand dbCommand = ProviderTyped.CreateCommand( Connection, Transaction ) )
				{
					string script = ProviderTyped.ScriptsTyped.GetSheafType_SharedLock;

					// Include the storehouse realm in the script.
					script = ReplaceStorehouseRealm( script, storehouseRealm );

					dbCommand.CommandText = script;

					#region Add the parameters

					// Storehouse key.
					AddStorehouseKeyParam( dbCommand, storehouseKey );

					#endregion Add the parameters

					// Read the result.
					using( DbDataReader reader = dbCommand.ExecuteReader() )
					{
						if( (reader == null) || !reader.Read() )
						{
							// No document found.
							return null;
						}

						// Sheaf type name.
						object sheafTypeNameResult = reader[ DatabaseColumns.SheafTypeName ];
						ValidationUtils.ObjectType<string>( sheafTypeNameResult, "sheafTypeNameResult" );

						if( reader.Read() )
						{
							throw new ValidationException( "More than one sheaf matched the storehouse key.", null, null
								, new ExceptionDetail( "Storehouse key", storehouseKey ) );
						}

						// Get the type (of the document) read from the storehouse.
						string readSheafTypeName = (string)sheafTypeNameResult;

						return PortableTypes.FindGlobal( readSheafTypeName );
					}
				}
			}
			finally
			{
				LogManagerEndMethod( logManager );
			}
		}

		#endregion Public methods
	}
}
