////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Storehouse;

namespace Vela.DatabaseStorehouse
{
	/// <summary>
	/// Contract for a set of scripts used to work with a database.
	/// </summary>
	public interface IDatabaseScripts : IStorehouseScripts
	{
		#region Public non-serializable properties

		/// <summary>
		/// The script used to create the storehouse on a storehouse server engine. READ-ONLY.
		/// </summary>
		string CreateServerStorehouse { get; }

		/// <summary>
		/// The script used to create a storehouse realm. READ-ONLY.
		/// </summary>
		string CreateStorehouseRealm { get; }

		/// <summary>
		/// The script used to check if a storehouse realm exists. READ-ONLY.
		/// </summary>
		string StorehouseHasRealm { get; }

		/// <summary>
		/// The script used to count the speckles which can be collected (with no lock). READ-ONLY.
		/// </summary>
		string CountSpeckles { get; }

		/// <summary>
		/// The script used to add a speckle (with an exclusive lock). READ-ONLY.
		/// </summary>
		string AddSpeckle_ExclusiveLock { get; }

		/// <summary>
		/// The script used to collect speckles (with no lock); the requested speckle filters are loaded. READ-ONLY.
		/// </summary>
		string CollectSpeckles { get; }

		/// <summary>
		/// The script used to collect speckles (with no lock) in a paged manner, starting with the newest ones; the requested
		/// speckle filters are loaded. READ-ONLY.
		/// </summary>
		string CollectSpeckles_PagedNewest { get; }

		/// <summary>
		/// The script used to collect speckles (with no lock) in a paged manner, starting with the oldest ones; the requested
		/// speckle filters are loaded. READ-ONLY.
		/// </summary>
		string CollectSpeckles_PagedOldest { get; }

		/// <summary>
		/// The script used to collect a speckle (with a shared lock); the requested speckle filters are loaded.
		/// READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The script returns the first encountered speckle, not necessarily the speckle which was added first. This
		/// behavior mVela.Imizes performance.</para>
		/// </remarks>
		string CollectFirstSpeckle_SharedLock { get; }

		/// <summary>
		/// The script used to collect a speckle (with an update lock); the requested speckle filters are loaded.
		/// READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The script returns the first encountered speckle, not necessarily the speckle which was added first. This
		/// behavior mVela.Imizes performance.</para>
		/// </remarks>
		string CollectFirstSpeckle_UpdateLock { get; }

		/// <summary>
		/// The script used to collect a speckle (with an exclusive lock); the requested speckle filters are loaded.
		/// READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The script returns the first encountered speckle, not necessarily the speckle which was added first. This
		/// behavior mVela.Imizes performance.</para>
		/// </remarks>
		string CollectFirstSpeckle_ExclusiveLock { get; }

		/// <summary>
		/// The script used to load a speckle (with a shared lock); the requested speckle filters are loaded.
		/// READ-ONLY.
		/// </summary>
		string LoadSpeckle_SharedLock { get; }

		/// <summary>
		/// The script used to load a speckle (with an update lock); the requested speckle filters are loaded.
		/// READ-ONLY.
		/// </summary>
		string LoadSpeckle_UpdateLock { get; }

		/// <summary>
		/// The script used to load a speckle (with an exclusive lock); the requested speckle filters are loaded.
		/// READ-ONLY.
		/// </summary>
		string LoadSpeckle_ExclusiveLock { get; }

		/// <summary>
		/// The script used to check if a speckle has a referred sheaf; no lock is acquired. READ-ONLY.
		/// </summary>
		string SpeckleHasDocument { get; }

		/// <summary>
		/// The script used to collect storehouse keys (with no lock). READ-ONLY.
		/// </summary>
		string CollectKeys { get; }

		/// <summary>
		/// The script used to collect storehouse keys (with no lock) in a paged manner, starting with the newest ones.
		/// READ-ONLY.
		/// </summary>
		string CollectKeys_PagedNewest { get; }

		/// <summary>
		/// The script used to collect storehouse keys (with no lock) in a paged manner, starting with the oldest ones.
		/// READ-ONLY.
		/// </summary>
		string CollectKeys_PagedOldest { get; }

		/// <summary>
		/// The script used to add a sheaf (with an exclusive lock). READ-ONLY.
		/// </summary>
		string AddSheaf_ExclusiveLock { get; }

		/// <summary>
		/// The script used to update a sheaf (with an update lock). READ-ONLY.
		/// </summary>
		string UpdateSheaf_UpdateLock { get; }

		/// <summary>
		/// The script used to update a sheaf (with an exclusive lock). READ-ONLY.
		/// </summary>
		string UpdateSheaf_ExclusiveLock { get; }

		/// <summary>
		/// The script used to delete sheafs (with an exclusive lock). READ-ONLY.
		/// </summary>
		string DeleteSheafs_ExclusiveLock { get; }

		/// <summary>
		/// The script used to delete the first sheaf (with an exclusive lock). READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The script deletes the first encountered sheaf, not necessarily the sheaf which was added first. This behavior
		/// mVela.Imizes performance.</para>
		/// </remarks>
		string DeleteFirstSheaf_ExclusiveLock { get; }

		/// <summary>
		/// The script used to delete a sheaf (with an exclusive lock). READ-ONLY.
		/// </summary>
		string DeleteSheaf_ExclusiveLock { get; }

		/// <summary>
		/// The script used to collect sheafs (with a shared lock). READ-ONLY.
		/// </summary>
		string CollectSheafs_SharedLock { get; }

		/// <summary>
		/// The script used to collect sheafs (with an update lock). READ-ONLY.
		/// </summary>
		string CollectSheafs_UpdateLock { get; }

		/// <summary>
		/// The script used to collect sheafs (with an exclusive lock). READ-ONLY.
		/// </summary>
		string CollectSheafs_ExclusiveLock { get; }

		/// <summary>
		/// The script used to collect sheafs (with a shared lock) in a paged manner, starting with the newest ones. READ-ONLY.
		/// </summary>
		string CollectSheafs_PagedNewest_SharedLock { get; }

		/// <summary>
		/// The script used to collect sheafs (with an update lock) in a paged manner, starting with the newest ones. READ-ONLY.
		/// </summary>
		string CollectSheafs_PagedNewest_UpdateLock { get; }

		/// <summary>
		/// The script used to collect sheafs (with an exclusive lock) in a paged manner, starting with the newest ones.
		/// READ-ONLY.
		/// </summary>
		string CollectSheafs_PagedNewest_ExclusiveLock { get; }

		/// <summary>
		/// The script used to collect sheafs (with a shared lock) in a paged manner, starting with the oldest ones. READ-ONLY.
		/// </summary>
		string CollectSheafs_PagedOldest_SharedLock { get; }

		/// <summary>
		/// The script used to collect sheafs (with an update lock) in a paged manner, starting with the oldest ones. READ-ONLY.
		/// </summary>
		string CollectSheafs_PagedOldest_UpdateLock { get; }

		/// <summary>
		/// The script used to collect sheafs (with an exclusive lock) in a paged manner, starting with the oldest ones.
		/// READ-ONLY.
		/// </summary>
		string CollectSheafs_PagedOldest_ExclusiveLock { get; }

		/// <summary>
		/// The script used to collect a sheaf (with a shared lock). READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The script returns the first encountered sheaf, not necessarily the sheaf which was added first. This behavior
		/// mVela.Imizes performance.</para>
		/// </remarks>
		string CollectFirstSheaf_SharedLock { get; }

		/// <summary>
		/// The script used to collect a sheaf (with an update lock). READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The script returns the first encountered sheaf, not necessarily the sheaf which was added first. This behavior
		/// mVela.Imizes performance.</para>
		/// </remarks>
		string CollectFirstSheaf_UpdateLock { get; }

		/// <summary>
		/// The script used to collect a sheaf (with an exclusive lock). READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The script returns the first encountered sheaf, not necessarily the sheaf which was added first. This behavior
		/// mVela.Imizes performance.</para>
		/// </remarks>
		string CollectFirstSheaf_ExclusiveLock { get; }

		/// <summary>
		/// The script used to load a sheaf (with a shared lock). READ-ONLY.
		/// </summary>
		string LoadSheaf_SharedLock { get; }

		/// <summary>
		/// The script used to load a sheaf (with an update lock). READ-ONLY.
		/// </summary>
		string LoadSheaf_UpdateLock { get; }

		/// <summary>
		/// The script used to load a sheaf (with an exclusive lock). READ-ONLY.
		/// </summary>
		string LoadSheaf_ExclusiveLock { get; }

		/// <summary>
		/// The script used to get the type of a sheaf (with a shared lock). READ-ONLY.
		/// </summary>
		string GetSheafType_SharedLock { get; }

		#endregion Public non-serializable properties
	}
}
