////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;

namespace Vela.Kernel
{
	/// <summary>
	/// Base exception for all Vela-wide exceptions.
	/// </summary>
	/// <remarks>
	/// <para>Don't include private information in exceptions because they might be logged.</para>
	/// </remarks>
	public abstract class BaseException : Exception, IPortableType
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected BaseException()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="message">The description of the exception. This may be null or empty.</param>
		/// <param name="exception">The existing exception when the current exception was created. This may be null.</param>
		/// <param name="details">Exception details. This may be null or empty.</param>
		public BaseException( string message, Exception exception = null, params ExceptionDetail[] details )
			: base( message, exception )
		{
			if( ListUtils.HasContent( details ) )
			{
				foreach( ExceptionDetail detail in details ) Details.Add( detail );
			}
		}

		#endregion Constructors

		#region Public non-serializable properties

		private ExceptionDetails _Details = new ExceptionDetails();
		/// <summary>
		/// Exception details. The backing field is initialized by default. This may be empty, but not null. READ-ONLY.
		/// </summary>
		public ExceptionDetails Details
		{
			get { return _Details; }
			protected set { _Details = value; }
		}

		private bool _DoLog = true;
		/// <summary>
		/// If this is true then the current exception should be automatically logged and saved (when it is caught). The default
		/// value is true.
		/// </summary>
		public bool DoLog
		{
			get { return _DoLog; }
			set { _DoLog = value; }
		}

		private bool _ShowToUser = true;
		/// <summary>
		/// If this is true then the current exception should be shown to the user (when it is caught). The default value is
		/// true.
		/// </summary>
		public bool ShowToUser
		{
			get { return _ShowToUser; }
			set { _ShowToUser = value; }
		}

		#endregion Public non-serializable properties
	}
}
