////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;

namespace Vela.Kernel
{
	/// <summary>
	/// A chronometer counts the time accumulated during a process.
	/// </summary>
	public class Chronometer
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public Chronometer()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="accumulatedTime">The initial accumulated time.</param>
		public Chronometer( TimeSpan accumulatedTime )
		{
			AccumulatedTime = accumulatedTime;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private bool _WasStarted = false;
		/// <summary>
		/// If this is false then the chronometer was never started, so it must never be accessed. READ-ONLY.
		/// </summary>
		public bool WasStarted
		{
			get { return _WasStarted; }
			protected set { _WasStarted = value; }
		}

		private bool _IsRunning = false;
		/// <summary>
		/// If this is true then the chronometer is running. READ-ONLY.
		/// </summary>
		public bool IsRunning
		{
			get { return _IsRunning; }
			protected set { _IsRunning = value; }
		}

		/// <summary>
		/// If this is true then the chronometer is stopped. READ-ONLY.
		/// </summary>
		public bool IsStopped
		{
			get { return !IsRunning; }
		}

		private TimeSpan _AccumulatedTime = TimeSpan.Zero;
		/// <summary>
		/// The accumulated time. READ-ONLY.
		/// </summary>
		public TimeSpan AccumulatedTime
		{
			get
			{
				CheckWasStarted();

				return _AccumulatedTime;
			}
			protected set { _AccumulatedTime = value; }
		}

		private DateTime _StartTime = DateTime.MinValue;
		/// <summary>
		/// The time when the chronometer was started. READ-ONLY.
		/// </summary>
		public DateTime StartTime
		{
			get
			{
				CheckWasStarted();

				return _StartTime;
			}
			protected set { _StartTime = value; }
		}

		/// <summary>
		/// The accumulated time (in milliseconds) of the chronometer. READ-ONLY.
		/// </summary>
		public double Milliseconds
		{
			get
			{
				Update();

				return AccumulatedTime.TotalMilliseconds;
			}
		}

		/// <summary>
		/// The accumulated time (in seconds) of the chronometer. READ-ONLY.
		/// </summary>
		public double Seconds
		{
			get
			{
				Update();

				return AccumulatedTime.TotalSeconds;
			}
		}

		/// <summary>
		/// The accumulated time (as a time span) of the chronometer. READ-ONLY.
		/// </summary>
		public TimeSpan TimeSpan
		{
			get
			{
				Update();

				return AccumulatedTime;
			}
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// If the chronometer was never started, an exception is thrown.
		/// </summary>
		private void CheckWasStarted()
		{
			if( !WasStarted ) throw new ValidationException( "The chronometer was never started." );
		}

		/// <summary>
		/// Updates the accumulated time.
		/// </summary>
		private void Update()
		{
			CheckWasStarted();

			if( IsRunning )
			{
				DateTime currentTime = DateTime.UtcNow;
				AccumulatedTime += currentTime - StartTime;
				StartTime = currentTime;
			}
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Clears the chronometer.
		/// </summary>
		public void Clear()
		{
			AccumulatedTime = TimeSpan.Zero;
			IsRunning = false;
			StartTime = DateTime.MinValue;
		}

		/// <summary>
		/// Starts the chronometer.
		/// </summary>
		/// <param name="accumulatedTime">The initial accumulated time.</param>
		public void Start( TimeSpan accumulatedTime )
		{
			AccumulatedTime = accumulatedTime;
			IsRunning = true;
			StartTime = DateTime.UtcNow;

			WasStarted = true;
		}

		/// <summary>
		/// Starts the chronometer.
		/// </summary>
		public void Start()
		{
			Start( TimeSpan.Zero );
		}

		/// <summary>
		/// Stops the chronometer from accumulating time.
		/// </summary>
		public void Stop()
		{
			CheckWasStarted();

			Update();
			IsRunning = false;
		}

		/// <summary>
		/// Resumes a stopped chronometer.
		/// </summary>
		public void Resume()
		{
			CheckWasStarted();

			IsRunning = true;
			StartTime = DateTime.UtcNow;
		}

		/// <summary>
		/// Checks if a specified timespan (in milliseconds) has passed since the chronometer was started.
		/// </summary>
		public bool PassedMilli( int milliseconds )
		{
			CheckWasStarted();

			bool passed = Milliseconds >= milliseconds;

			// Reset the chronometer so that it can check if a new timespan has passed.
			if( passed ) Start();

			return passed;
		}

		/// <summary>
		/// Checks if a specified timespan (in seconds) has passed since the chronometer was started. If the timespan has passed,
		/// the chronometer is reset.
		/// </summary>
		public bool PassedSeconds( double seconds )
		{
			CheckWasStarted();

			bool passed = Seconds >= seconds;

			// Reset the chronometer so that it can check if a new timespan has passed.
			if( passed ) Start();

			return passed;
		}

		/// <summary>
		/// Checks if a specified timespan (in minutes) has passed since the chronometer was started. If the timespan has passed,
		/// the chronometer is reset.
		/// </summary>
		public bool PassedMinutes( double minutes )
		{
			return PassedSeconds( TimeUtils.MinutesToSeconds( minutes ) );
		}

		/// <summary>
		/// Checks if a specified timespan (in hours) has passed since the chronometer was started. If the timespan has passed,
		/// the chronometer is reset.
		/// </summary>
		public bool PassedHours( double hours )
		{
			return PassedSeconds( TimeUtils.HoursToSeconds( hours ) );
		}

		#endregion Public methods
	}
}
