////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;

namespace Vela.Kernel
{
	/// <summary>
	/// Stream which contains a list of binary nodes.
	/// </summary>
	/// <remarks>
	/// <para>A node is formed by several attributes:</para>
	/// <list type="bullet">
	///		<item>Value size (the size of this attribute is specified in the constructor).</item>
	///		<item>Value (variable size). This is written in the stream only if it's not null and its size is greater than 0.
	///		</item>
	/// </list>
	/// 
	/// <para>It is guaranteed that the generated output is the same, regardless of the platform on which this class runs.</para>
	/// </remarks>
	public class BinaryNodeStream
	{
		#region Public constants

		/// <summary>
		/// The maximum value for <see cref="ValueSizeBufferSize"/>.
		/// </summary>
		public const int ValueSizeBufferSizeMax = sizeof( Int32 );

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="valueSizeBufferSize">The buffer size for the size of a value. If this is 0 or a negative number,
		/// <see cref="ValueSizeBufferSizeMax"/> is used instead.</param>
		/// <param name="textEncoding">Text encoding. This may be null if the data to convert to a buffer is not a
		/// text.</param>
		public BinaryNodeStream( int valueSizeBufferSize, TextEncoding textEncoding )
		{
			ValueSizeBufferSize = (valueSizeBufferSize > 0) ? valueSizeBufferSize : ValueSizeBufferSizeMax;
			TextEncoding = textEncoding;
			UnderlyingStream = new MemoryStream();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="valueSizeBufferSize">The buffer size for the size of a value. If this is 0 or a negative number,
		/// <see cref="ValueSizeBufferSizeMax"/> is used instead.</param>
		/// <param name="buffer">Buffer from which to read nodes.</param>
		public BinaryNodeStream( int valueSizeBufferSize, byte[] buffer )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );

			ValueSizeBufferSize = (valueSizeBufferSize > 0) ? valueSizeBufferSize : ValueSizeBufferSizeMax;
			UnderlyingStream = new MemoryStream( buffer );
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The underlying memory stream.
		/// </summary>
		/// <remarks>
		/// <para>This doesn't have to be disposed because a memory stream does nothing while disposing. This helps avoid
		/// having to require the caller to dispose the current object.</para>
		/// </remarks>
		private MemoryStream UnderlyingStream;

		#endregion Private fields

		#region Public non-serializable properties

		private int _ValueSizeBufferSize;
		/// <summary>
		/// The buffer size for the size of a value. This must be at least 1. READ-ONLY.
		/// </summary>
		public int ValueSizeBufferSize
		{
			get { return _ValueSizeBufferSize; }
			protected set
			{
				ValidationUtils.Integer( value, 1, ValueSizeBufferSizeMax, "ValueSizeBufferSize" );

				_ValueSizeBufferSize = value;
			}
		}

		/// <summary>
		/// The maximum size of a value. READ-ONLY.
		/// </summary>
		public int ValueSizeMax
		{
			get { return (ValueSizeBufferSize >= 4) ? int.MaxValue : (1 << (ValueSizeBufferSize * 8) - 1); }
		}

		private TextEncoding _TextEncoding;
		/// <summary>
		/// Text encoding. This may be null if the data to convert to a buffer is not a text. READ-ONLY.
		/// </summary>
		public TextEncoding TextEncoding
		{
			get { return _TextEncoding; }
			protected set { _TextEncoding = value; }
		}

		/// <summary>
		/// Gets the position in the stream. READ-ONLY.
		/// </summary>
		public int Position
		{
			get { return (int)UnderlyingStream.Position; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Reads the next attribute from the stream.
		/// </summary>
		/// <param name="attributeSize">Attribute size. This must be at least 1.</param>
		/// <returns>Returns the read attribute.</returns>
		private byte[] ReadAttribute( int attributeSize )
		{
			int attributeSizeMax = (int)(UnderlyingStream.Length - UnderlyingStream.Position);
			ValidationUtils.Integer( attributeSize, 1, attributeSizeMax, "attributeSize" );

			byte[] buffer = new byte[ attributeSize ];

			int readCount = UnderlyingStream.Read( buffer, 0, attributeSize );

			if( readCount != attributeSize )
			{
				throw new ValidationException( string.Format( "Only {0} bytes were read instead of {1}.", readCount
					, attributeSize ) );
			}

			return buffer;
		}

		#endregion Private methods

		#region Public static methods

		/// <summary>
		/// Writes the specified datas in a stream.
		/// </summary>
		/// <param name="textEncoding">Text encoding. This may be null if the data to convert to a buffer is not a
		/// text.</param>
		/// <param name="datas">List of datas to write.</param>
		/// <returns>Returns the resulted stream as a buffer.</returns>
		public static byte[] WriteValues( TextEncoding textEncoding, params object[] datas )
		{
			ValidationUtils.ArrayHasContent( datas, "datas" );

			BinaryNodeStream bns = new BinaryNodeStream( 0, textEncoding );

			int count = datas.Length;
			for( int i = 0; i < count; i++ )
			{
				object item = datas[ i ];

				ValidationUtils.ObjectHasContent( item, string.Format( "datas[ {0} ]", i ) );

				bns.WriteValue( item );
			}

			return bns.ToArray();
		}

		/// <summary>
		/// Writes the specified datas in a stream.
		/// </summary>
		/// <param name="buffers">Buffers to write.</param>
		/// <returns>Returns the resulted stream as a buffer.</returns>
		public static byte[] WriteBinaryValues( params byte[][] buffers )
		{
			return WriteValues( null, buffers );
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Checks if more data can be read from the stream.
		/// </summary>
		/// <returns>Returns true if more data can be read from the stream.</returns>
		public bool HasNextNode()
		{
			return UnderlyingStream.Position < UnderlyingStream.Length - 1;
		}

		/// <summary>
		/// Writes the specified data in the stream.
		/// </summary>
		/// <param name="buffer">Buffer to write in the stream. This may be empty, but not null.</param>
		public void WriteValue( object buffer )
		{
			ValidationUtils.ObjectHasContent( buffer, "buffer" );

			byte[] valueBuffer = ScalarEncoder.ValueToBuffer( buffer, TextEncoding, true );
			int valueBufferSize = valueBuffer.Length;

			// Check for empty data value.
			if( valueBufferSize <= 0 )
			{
				// Write the data value size.
				byte[] emptySizeBuffer = BitConverter.GetBytes( 0 );
				UnderlyingStream.Write( emptySizeBuffer, 0, ValueSizeBufferSize );

				return;
			}

			if( valueBufferSize > ValueSizeMax )
			{
				throw new ValidationException( string.Format( "The size {0} of the data exceeds the maximum size allowed {1}."
					, valueBufferSize, ValueSizeMax ) );
			}

			// Write the data value size.
			byte[] bufferSizeBuffer = BitConverter.GetBytes( valueBufferSize );
			UnderlyingStream.Write( bufferSizeBuffer, 0, ValueSizeBufferSize );

			// Write the data value.
			UnderlyingStream.Write( valueBuffer, 0, valueBufferSize );
		}

		/// <summary>
		/// Checks if the first node can be read.
		/// </summary>
		/// <param name="nodeValueSizeMax">The maximum size that the first node's value may have. If this is 0 or a negative
		/// value, it's ignored.</param>
		/// <returns>Returns true if the first node can be read.</returns>
		public bool CanReadFirstNode( int nodeValueSizeMax )
		{
			// Don't use "UnderlyingStream.Length <= ValueSizeBufferSize" because there may be a (single) node and it may be
			// empty.
			if( (UnderlyingStream.Position != 0) || (UnderlyingStream.Length < ValueSizeBufferSize) )
			{
				throw new ValidationException( "Failed to check if the first node can be read because the current position in"
					+ " the underlying stream is not 0, or the stream is too short." );
			}

			byte[] dataSizeBuffer = new byte[ ValueSizeBufferSize ];

			// Read the data value size without moving the current position in the underlying stream.
			Int64 position = UnderlyingStream.Position;
			int readCount = UnderlyingStream.Read( dataSizeBuffer, 0, ValueSizeBufferSize );
			UnderlyingStream.Position = position;

			if( readCount != ValueSizeBufferSize ) return false;

			int valueBufferSize = ConvertUtils.ByteToInt( dataSizeBuffer, 0, dataSizeBuffer.Length );
			int valueBufferSizeMax = (int)(UnderlyingStream.Length - UnderlyingStream.Position) - ValueSizeBufferSize;

			if( nodeValueSizeMax > 0 ) valueBufferSizeMax = Math.Min( valueBufferSizeMax, nodeValueSizeMax );

			return (valueBufferSize >= 0) && (valueBufferSize <= valueBufferSizeMax);
		}

		/// <summary>
		/// Reads a data from the stream.
		/// </summary>
		/// <param name="contentType">Type of the value to convert the read buffer to.</param>
		/// <param name="declaredType">Declared type of the property to which the converted value is to be assigned.</param>
		/// <param name="enumFallbackValue">If the value to convert is an enumeration, if the conversion to the enumeration
		/// fails, this value is assigned instead of an exception being thrown. This may be null.</param>
		/// <returns>Returns the read data. This may be empty, but not null.</returns>
		public object ReadValue( Type contentType, Type declaredType, object enumFallbackValue )
		{
			// Read the data value size.
			byte[] dataSizeBuffer = ReadAttribute( ValueSizeBufferSize );
			int valueBufferSize = ConvertUtils.ByteToInt( dataSizeBuffer, 0, dataSizeBuffer.Length );

			int valueBufferSizeMax = (int)(UnderlyingStream.Length - UnderlyingStream.Position);
			ValidationUtils.Integer( valueBufferSize, 0, valueBufferSizeMax, "valueBufferSize" );

			if( valueBufferSize <= 0 )
			{
				return ScalarEncoder.ValueFromBuffer( new byte[ 0 ], contentType, declaredType, enumFallbackValue );
			}
			else
			{
				// Read the data value.
				byte[] valueBuffer = ReadAttribute( valueBufferSize );

				return ScalarEncoder.ValueFromBuffer( valueBuffer, contentType, declaredType, enumFallbackValue );
			}
		}

		/// <summary>
		/// Reads a data from the stream.
		/// </summary>
		/// <param name="contentType">Type of the value to convert the read buffer to.</param>
		/// <returns>Returns the read data. This may be empty, but not null.</returns>
		public object ReadValue( Type contentType )
		{
			return ReadValue( contentType, contentType, null );
		}

		/// <summary>
		/// Reads a buffer from the stream.
		/// </summary>
		/// <returns>Returns the read buffer. This may be empty, but not null.</returns>
		public byte[] ReadBuffer()
		{
			return (byte[])ReadValue( typeof( byte[] ), typeof( byte[] ), null );
		}

		/// <summary>
		/// Gets the content of the current stream as a buffer.
		/// </summary>
		/// <returns>Returns the content of the current stream as a buffer.</returns>
		public byte[] ToArray()
		{
			return UnderlyingStream.ToArray();
		}

		#endregion Public methods
	}
}
