////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Globalization;

namespace Vela.Kernel
{
	/// <summary>
	/// Encoder for scalar datas.
	/// </summary>
	/// <remarks>
	/// <para>It is guaranteed that the generated output is the same, regardless of the platform on which this class runs.</para>
	/// 
	/// <para>Objects of the following types may be encoded: arrary of <see cref="byte"/>, <see cref="Locator"/>,
	/// <see cref="string"/>, <see cref="char"/>, <see cref="sbyte"/>, <see cref="byte"/>, <see cref="Int16"/>,
	/// <see cref="UInt16"/>, <see cref="Int32"/>, <see cref="UInt32"/>, <see cref="Int64"/>, <see cref="UInt64"/>,
	/// <see cref="double"/>, <see cref="DateTime"/>, <see cref="TimeSpan"/>, <see cref="bool"/>, enumerations.</para>
	/// 
	/// <para>Enumeration values are encoded as texts.</para>
	/// 
	/// <para>In order to ensure a level of versioning, the following scalar type conversions are performed when the scalar is
	/// extracted from a buffer or from a text, if the content type (of the value which is being extracted) is different
	/// than the declared type (of the property to which the extracted value is to be assigned):</para>
	/// <list type="bullet">
	///		<item><see cref="string"/> to enumeration (the enumeration must contain the textual values).</item>
	///		<item><see cref="bool"/> to enumeration (the enumeration must contain the textual values for the boolean true and
	///			false).</item>
	///		<item><see cref="sbyte"/> to <see cref="double"/>.</item>
	///		<item><see cref="byte"/> to <see cref="double"/>.</item>
	///		<item><see cref="System.Int16"/> to <see cref="double"/>.</item>
	///		<item><see cref="System.UInt16"/> to <see cref="double"/>.</item>
	///		<item><see cref="System.Int32"/> to <see cref="double"/>.</item>
	///		<item><see cref="System.UInt32"/> to <see cref="double"/>.</item>
	///		<item><see cref="System.Int64"/> to <see cref="double"/>.</item>
	///		<item><see cref="System.UInt64"/> to <see cref="double"/>.</item>
	///		<item>Enumeration to <see cref="string"/>.</item>
	///	</list>
	///	
	/// <para>The binary and textual converters must use the same data precision. One example is the time where one might be
	/// tempted to use ticks for the binary converter, and seconds for the textual one. However, these converters are used by
	/// serializers, and serializers may be mixed during the processing of a single document, like signing with a binary hash
	/// serializer, and saving with a textual one. But this would lead to signing the document with all time ticks (because of
	/// the binary serializer), while saving and loading the document with only the seconds (because of the textual serializer),
	/// which means that the signature of the document would not verify (after loading).</para>
	/// </remarks>
	public static class ScalarEncoder
	{
		#region Public constants

		/// <summary>
		/// The format for converting a time to text.
		/// </summary>
		/// <remarks>
		/// <para>Make sure that you synchronize the precision of this textual converter with that of the binary converter.
		/// </para>
		/// </remarks>
		public const string ScalarEncoderTimeFormat = "yyyy'-'MM'-'dd'T'HH':'mm':'ss'.'fffffffK";

		/// <summary>
		/// The format for converting a time span to text.
		/// </summary>
		/// <remarks>
		/// <para>Make sure that you synchronize the precision of this textual converter with that of the binary converter.
		/// </para>
		/// </remarks>
		public const string ScalarEncoderTimeSpanFormat = "d'T'hh':'mm':'ss'.'fffffff";

		#endregion Public constants

		#region Public static methods

		/// <summary>
		/// Gets a list with the basic scalar types.
		/// </summary>
		/// <returns>Returns a list with the basic scalar types.</returns>
		public static List<Type> GetBasicScalarTypes()
		{
			List<Type> types = new List<Type>();

			types.Add( typeof( byte[] ) );
			types.Add( typeof( string ) );
			types.Add( typeof( char ) );
			types.Add( typeof( sbyte ) );
			types.Add( typeof( byte ) );
			types.Add( typeof( Int16 ) );
			types.Add( typeof( UInt16 ) );
			types.Add( typeof( Int32 ) );
			types.Add( typeof( UInt32 ) );
			types.Add( typeof( Int64 ) );
			types.Add( typeof( UInt64 ) );
			types.Add( typeof( double ) );
			types.Add( typeof( DateTime ) );
			types.Add( typeof( TimeSpan ) );
			types.Add( typeof( bool ) );

			return types;
		}

		/// <summary>
		/// Checks if the specified type is a scalar.
		/// </summary>
		/// <param name="type">The type to verify if it is a scalar.</param>
		/// <returns>Returns true if the specified type is a scalar.</returns>
		/// <remarks>
		/// <para>A scalar's content may have only one type, its declared one.</para>
		/// 
		/// <para>Check if an object is a scalar, before you check if the object is a list or a class (because the object might
		/// actually be a list or a class (like a <see cref="string"/>), even though it must be serialized as a scalar).</para>
		/// </remarks>
		public static bool IsScalar( Type type )
		{
			ValidationUtils.ObjectHasContent( type, "type" );

			if( type == typeof( byte[] ) ) return true;
			if( type == typeof( Locator ) ) return true;
			if( type == typeof( string ) ) return true;
			if( type == typeof( char ) ) return true;
			if( type == typeof( sbyte ) ) return true;
			if( type == typeof( byte ) ) return true;
			if( type == typeof( Int16 ) ) return true;
			if( type == typeof( UInt16 ) ) return true;
			if( type == typeof( Int32 ) ) return true;
			if( type == typeof( UInt32 ) ) return true;
			if( type == typeof( Int64 ) ) return true;
			if( type == typeof( UInt64 ) ) return true;
			if( type == typeof( double ) ) return true;
			if( type == typeof( DateTime ) ) return true;
			if( type == typeof( TimeSpan ) ) return true;
			if( type == typeof( bool ) ) return true;
			if( type.IsEnum ) return true;

			// Not a scalar.
			return false;
		}

		/// <summary>
		/// Checks if the type of the specified value is a scalar.
		/// </summary>
		/// <param name="data">Value whose type to verify if it is a scalar. This may be empty, but not null.</param>
		/// <returns>Returns true if the type of the specified value is a scalar.</returns>
		/// <remarks>
		/// <para>A scalar's content may have only one type, its declared one.</para>
		/// 
		/// <para>Check if an object is a scalar, before you check if the object is a list or a class (because the object might
		/// actually be a list or a class (like a <see cref="string"/>), even though it must be serialized as a scalar).</para>
		/// </remarks>
		public static bool IsScalar( object data )
		{
			ValidationUtils.ObjectHasContent( data, "data" );

			return IsScalar( data.GetType() );
		}

		/// <summary>
		/// Converts the specified value to a buffer.
		/// </summary>
		/// <param name="data">Value to convert to a buffer. This may be empty, but not null.</param>
		/// <param name="textEncoding">Text encoding. This may be null if the data to convert to a buffer is not a
		/// text.</param>
		/// <param name="addPreamble">If this is true then the preamble of the text encoding is added at the beginning of the
		/// resulted buffer.</param>
		/// <returns>Returns the resulted buffer. This may be empty, but not null.</returns>
		/// <remarks>
		/// <para>There is no need to call the <see cref="IsScalar( object )"/> method before calling this method.</para>
		/// </remarks>
		public static byte[] ValueToBuffer( object data, TextEncoding textEncoding, bool addPreamble )
		{
			ValidationUtils.ObjectHasContent( data, "data" );

			Type dataType = data.GetType();

			if( dataType == typeof( byte[] ) ) return (byte[])data;
			if( dataType == typeof( Locator ) ) return ((Locator)data).Value;
			if( dataType == typeof( string ) )
			{
				ValidationUtils.ObjectHasContent( textEncoding, "textEncoding" );

				return textEncoding.TextToArray( (string)data, addPreamble );
			}
			if( dataType == typeof( char ) ) return BitConverter.GetBytes( (char)data );
			if( dataType == typeof( sbyte ) ) return BitConverter.GetBytes( (sbyte)data );
			if( dataType == typeof( byte ) ) return BitConverter.GetBytes( (byte)data );
			if( dataType == typeof( Int16 ) ) return BitConverter.GetBytes( (Int16)data );
			if( dataType == typeof( UInt16 ) ) return BitConverter.GetBytes( (UInt16)data );
			if( dataType == typeof( Int32 ) ) return BitConverter.GetBytes( (Int32)data );
			if( dataType == typeof( UInt32 ) ) return BitConverter.GetBytes( (UInt32)data );
			if( dataType == typeof( Int64 ) ) return BitConverter.GetBytes( (Int64)data );
			if( dataType == typeof( UInt64 ) ) return BitConverter.GetBytes( (UInt64)data );
			if( dataType == typeof( double ) ) return BitConverter.GetBytes( (double)data );
			if( dataType == typeof( DateTime ) ) return BitConverter.GetBytes( ((DateTime)data).ToBinary() );
			if( dataType == typeof( TimeSpan ) ) return BitConverter.GetBytes( ((TimeSpan)data).Ticks );
			if( dataType == typeof( bool ) ) return BitConverter.GetBytes( (bool)data );
			if( dataType.IsEnum )
			{
				ValidationUtils.ObjectHasContent( textEncoding, "textEncoding" );

				return textEncoding.TextToArray( data.ToString(), addPreamble );
			}

			throw new ValidationException( string.Format( "The type \"{0}\" of the scalar to get as buffer is invalid."
				, dataType.FullName ) );
		}

		/// <summary>
		/// Converts the specified buffer to a value of the specified type.
		/// </summary>
		/// <param name="buffer">Buffer to convert to an object of the specified type. This may be empty, but not null.</param>
		/// <param name="contentType">Type of the value to convert the specified buffer to.</param>
		/// <param name="declaredType">Declared type of the property to which the converted value is to be assigned.</param>
		/// <param name="enumFallbackValue">If the value to convert is an enumeration, if the conversion to the enumeration
		/// fails, this value is assigned instead of an exception being thrown. This may be null.</param>
		/// <returns>Returns the extracted value.</returns>
		/// <remarks>
		/// <para>There is no need to call the <see cref="IsScalar( object )"/> method before or after calling this
		/// method.</para>
		/// 
		/// <para>If the required scalar type conversion is not implemented, an exception is thrown.</para>
		/// </remarks>
		public static object ValueFromBuffer( byte[] buffer, Type contentType, Type declaredType, object enumFallbackValue )
		{
			ValidationUtils.ObjectHasContent( buffer, "buffer" );
			ValidationUtils.ObjectHasContent( contentType, "contentType" );
			ValidationUtils.ObjectHasContent( declaredType, "declaredType" );

			if( !declaredType.IsAssignableFrom( contentType ) )
			{
				// The content type and the declared type are different. Check if the content type can be converted to the
				// declared type.

				#region Scalar type conversions

				if( (contentType == typeof( string )) && declaredType.IsEnum )
				{
					// String to enumeration.

					string text = TextEncoding.ArrayToText( buffer );

					if( enumFallbackValue == null ) return Enum.Parse( declaredType, text, true );
					else if( Enum.IsDefined( declaredType, text ) ) return Enum.Parse( declaredType, text, true );
					else return enumFallbackValue;
				}
				if( (contentType == typeof( bool )) && declaredType.IsEnum )
				{
					// Boolean to enumeration.

					bool b = BitConverter.ToBoolean( buffer, 0 );

					return Enum.Parse( declaredType, b.ToString(), true );
				}
				if( (contentType == typeof( sbyte )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)(unchecked( (sbyte)buffer[ 0 ] ));
				}
				if( (contentType == typeof( byte )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)buffer[ 0 ];
				}
				if( (contentType == typeof( Int16 )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)BitConverter.ToInt16( buffer, 0 );
				}
				if( (contentType == typeof( UInt16 )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)BitConverter.ToUInt16( buffer, 0 );
				}
				if( (contentType == typeof( Int32 )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)BitConverter.ToInt32( buffer, 0 );
				}
				if( (contentType == typeof( UInt32 )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)BitConverter.ToUInt32( buffer, 0 );
				}
				if( (contentType == typeof( Int64 )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)BitConverter.ToInt64( buffer, 0 );
				}
				if( (contentType == typeof( UInt64 )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)BitConverter.ToUInt64( buffer, 0 );
				}
				if( contentType.IsEnum && (declaredType == typeof( string )) )
				{
					// Enumeration to string.

					return TextEncoding.ArrayToText( buffer );
				}

				#endregion Scalar type conversions

				// Since no scalar type conversion is implemented, an exception is thrown.
				throw new ValidationException( string.Format( "The content type \"{0}\" of the binary value is different than"
						+ " the declared type \"{1}\" of the property to which the converted value is to be assigned."
						+ " This scalar type conversion is not implemented."
					, contentType.FullName, declaredType.FullName ) );
			}

			// The content type and the declared type are identical.

			if( contentType == typeof( byte[] ) ) return buffer;
			if( contentType == typeof( Locator ) ) return new Locator( buffer );
			if( contentType == typeof( string ) ) return TextEncoding.ArrayToText( buffer );
			if( contentType == typeof( char ) ) return BitConverter.ToChar( buffer, 0 );
			if( contentType == typeof( sbyte ) ) return unchecked( (sbyte)buffer[ 0 ] );
			if( contentType == typeof( byte ) ) return buffer[ 0 ];
			if( contentType == typeof( Int16 ) ) return BitConverter.ToInt16( buffer, 0 );
			if( contentType == typeof( UInt16 ) ) return BitConverter.ToUInt16( buffer, 0 );
			if( contentType == typeof( Int32 ) ) return BitConverter.ToInt32( buffer, 0 );
			if( contentType == typeof( UInt32 ) ) return BitConverter.ToUInt32( buffer, 0 );
			if( contentType == typeof( Int64 ) ) return BitConverter.ToInt64( buffer, 0 );
			if( contentType == typeof( UInt64 ) ) return BitConverter.ToUInt64( buffer, 0 );
			if( contentType == typeof( double ) ) return BitConverter.ToDouble( buffer, 0 );
			if( contentType == typeof( DateTime ) ) return DateTime.FromBinary( BitConverter.ToInt64( buffer, 0 ) );
			if( contentType == typeof( TimeSpan ) ) return TimeSpan.FromTicks( BitConverter.ToInt64( buffer, 0 ) );
			if( contentType == typeof( bool ) ) return BitConverter.ToBoolean( buffer, 0 );
			if( contentType.IsEnum )
			{
				string text = TextEncoding.ArrayToText( buffer );

				if( enumFallbackValue == null ) return Enum.Parse( contentType, text, true );
				else if( Enum.IsDefined( contentType, text ) ) return Enum.Parse( contentType, text, true );
				else return enumFallbackValue;
			}

			throw new ValidationException( string.Format( "The type \"{0}\" of the scalar to extract from the buffer"
				+ " is invalid.", contentType.FullName ) );
		}

		/// <summary>
		/// Converts the specified buffer to a value of the specified type.
		/// </summary>
		/// <param name="buffer">Buffer to convert to an object of the specified type. This may be empty, but not null.</param>
		/// <param name="contentType">Type of the value to convert the specified buffer to.</param>
		/// <returns>Returns the extracted value.</returns>
		/// <remarks>
		/// <para>There is no need to call the <see cref="IsScalar( object )"/> method before or after calling this
		/// method.</para>
		/// 
		/// <para>If the required scalar type conversion is not implemented, an exception is thrown.</para>
		/// </remarks>
		public static object ValueFromBuffer( byte[] buffer, Type contentType )
		{
			return ValueFromBuffer( buffer, contentType, contentType, null );
		}

		/// <summary>
		/// Converts the specified value to a text.
		/// </summary>
		/// <param name="data">Value to convert to a text. This may be empty, but not null.</param>
		/// <returns>Returns the resulted text. This may be empty, but not null.</returns>
		/// <remarks>
		/// <para>There is no need to call the <see cref="IsScalar( object )"/> method before calling this method.</para>
		/// </remarks>
		public static string ValueToText( object data )
		{
			ValidationUtils.ObjectHasContent( data, "data" );

			Type dataType = data.GetType();

			if( dataType == typeof( byte[] ) ) return Convert.ToBase64String( (byte[])data );
			if( dataType == typeof( Locator ) ) return Convert.ToBase64String( ((Locator)data).Value );
			if( dataType == typeof( string ) ) return (string)data;
			if( dataType == typeof( char ) ) return ((UInt16)(char)data).ToString(); // XML rejects characters with code 0.
			if( dataType == typeof( sbyte ) ) return data.ToString();
			if( dataType == typeof( byte ) ) return data.ToString();
			if( dataType == typeof( Int16 ) ) return data.ToString();
			if( dataType == typeof( UInt16 ) ) return data.ToString();
			if( dataType == typeof( Int32 ) ) return data.ToString();
			if( dataType == typeof( UInt32 ) ) return data.ToString();
			if( dataType == typeof( Int64 ) ) return data.ToString();
			if( dataType == typeof( UInt64 ) ) return data.ToString();
			if( dataType == typeof( double ) ) return ((double)data).ToString( CultureInfo.InvariantCulture );
			if( dataType == typeof( DateTime ) ) return ((DateTime)data).ToString( ScalarEncoderTimeFormat );
			if( dataType == typeof( TimeSpan ) ) return ((TimeSpan)data).ToString( ScalarEncoderTimeSpanFormat );
			if( dataType == typeof( bool ) ) return data.ToString();
			if( dataType.IsEnum ) return data.ToString();

			throw new ValidationException( string.Format( "The type \"{0}\" of the scalar to get as text is invalid."
				, dataType.FullName ) );
		}

		/// <summary>
		/// Converts the specified buffer to a value of the specified type.
		/// </summary>
		/// <param name="text">Text to convert to an object of the specified type. This may be empty, but not null.</param>
		/// <param name="contentType">Type of the value to convert the specified text to.</param>
		/// <param name="declaredType">Declared type of the property to which the converted value is to be assigned.</param>
		/// <param name="enumFallbackValue">If the value to convert is an enumeration, if the conversion to the enumeration
		/// fails, this value is assigned instead of an exception being thrown. This may be null.</param>
		/// <returns>Returns the resulted value.</returns>
		/// <remarks>
		/// <para>There is no need to call the <see cref="IsScalar( object )"/> method before or after calling this
		/// method.</para>
		/// 
		/// <para>If the required scalar type conversion is not implemented, an exception is thrown.</para>
		/// </remarks>
		public static object ValueFromText( string text, Type contentType, Type declaredType, object enumFallbackValue )
		{
			ValidationUtils.ObjectHasContent( text, "text" );
			ValidationUtils.ObjectHasContent( contentType, "contentType" );
			ValidationUtils.ObjectHasContent( declaredType, "declaredType" );

			if( !declaredType.IsAssignableFrom( contentType ) )
			{
				// The content type and the declared type are different. Check if the content type can be converted to the
				// declared type.

				#region Scalar type conversions

				if( (contentType == typeof( string )) && declaredType.IsEnum )
				{
					// String to enumeration.

					if( enumFallbackValue == null ) return Enum.Parse( declaredType, text, true );
					else if( Enum.IsDefined( declaredType, text ) ) return Enum.Parse( declaredType, text, true );
					else return enumFallbackValue;
				}
				if( (contentType == typeof( bool )) && declaredType.IsEnum )
				{
					// Boolean to enumeration.

					return Enum.Parse( declaredType, text, true );
				}
				if( (contentType == typeof( sbyte )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)Convert.ToSByte( text );
				}
				if( (contentType == typeof( byte )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)Convert.ToByte( text );
				}
				if( (contentType == typeof( Int16 )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)Convert.ToInt16( text );
				}
				if( (contentType == typeof( UInt16 )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)Convert.ToUInt16( text );
				}
				if( (contentType == typeof( Int32 )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)Convert.ToInt32( text );
				}
				if( (contentType == typeof( UInt32 )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)Convert.ToUInt32( text );
				}
				if( (contentType == typeof( Int64 )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)Convert.ToInt64( text );
				}
				if( (contentType == typeof( UInt64 )) && (declaredType == typeof( double )) )
				{
					// Integer to double number.

					return (double)Convert.ToUInt64( text );
				}
				if( contentType.IsEnum && (declaredType == typeof( string )) )
				{
					// Enumeration to string.

					return text;
				}

				#endregion Scalar type conversions

				// Since no scalar type conversion is implemented, an exception is thrown.
				throw new ValidationException( string.Format( "The content type \"{0}\" of the textual value is different"
						+ " than the declared type \"{1}\" of the property to which the converted value is to be assigned."
						+ " This scalar type conversion is not implemented."
					, contentType.FullName, declaredType.FullName ) );
			}

			// The content type and the declared type are identical.

			if( contentType == typeof( byte[] ) ) return Convert.FromBase64String( text );
			if( contentType == typeof( Locator ) ) return new Locator( Convert.FromBase64String( text ) );
			if( contentType == typeof( string ) ) return text;
			if( contentType == typeof( char ) ) return (char)Convert.ToInt16( text );
			if( contentType == typeof( sbyte ) ) return Convert.ToSByte( text );
			if( contentType == typeof( byte ) ) return Convert.ToByte( text );
			if( contentType == typeof( Int16 ) ) return Convert.ToInt16( text );
			if( contentType == typeof( UInt16 ) ) return Convert.ToUInt16( text );
			if( contentType == typeof( Int32 ) ) return Convert.ToInt32( text );
			if( contentType == typeof( UInt32 ) ) return Convert.ToUInt32( text );
			if( contentType == typeof( Int64 ) ) return Convert.ToInt64( text );
			if( contentType == typeof( UInt64 ) ) return Convert.ToUInt64( text );
			if( contentType == typeof( double ) ) return Convert.ToDouble( text );
			if( contentType == typeof( DateTime ) )
			{
				return DateTime.ParseExact( text, ScalarEncoderTimeFormat, null, DateTimeStyles.RoundtripKind );
			}
			if( contentType == typeof( TimeSpan ) ) return TimeSpan.ParseExact( text, ScalarEncoderTimeSpanFormat, null );
			if( contentType == typeof( bool ) ) return TextUtils.Equals( text, true.ToString() );
			if( contentType.IsEnum )
			{
				if( enumFallbackValue == null ) return Enum.Parse( contentType, text, true );
				else if( Enum.IsDefined( contentType, text ) ) return Enum.Parse( contentType, text, true );
				else return enumFallbackValue;
			}

			throw new ValidationException( string.Format( "The type \"{0}\" of the scalar to extract from the text"
				+ " is invalid.", contentType.FullName ) );
		}

		#endregion Public static methods
	}
}
