////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Collections.Generic;

namespace Vela.Kernel
{
	/// <summary>
	/// Vela-wide file extensions.
	/// </summary>
	public static class FileExtensions
	{
		#region Public constants

		/// <summary>
		/// The extension for a file which contains an application.
		/// </summary>
		public const string Application = ".exe";

		#region Vela specialized

		/// <summary>
		/// The extension for a file which contains a storehouse (header).
		/// </summary>
		public const string Storehouse = ".vstore";

		/// <summary>
		/// The extension for a file which contains the main storehouse file which is attached to a storehouse header.
		/// </summary>
		public const string StorehouseMainAttachment = ".vstoreattach";

		/// <summary>
		/// The extension for a file which contains a generic document.
		/// </summary>
		public const string GenericDocument = ".vgeneric";

		/// <summary>
		/// The extension for a file which contains a storable configuration.
		/// </summary>
		public const string StorableConfiguration = ".vconfig";

		/// <summary>
		/// The extension for a file which contains a presenter.
		/// </summary>
		public const string Presenter = ".vpresenter";

		/// <summary>
		/// The extension for a file which contains an attesting document.
		/// </summary>
		public const string AttestingDocument = ".vattesting";

		/// <summary>
		/// The extension for a file which contains an securing document.
		/// </summary>
		public const string SecuringDocument = ".vsecuring";

		/// <summary>
		/// The extension for a file which contains a secure file.
		/// </summary>
		public const string SecureFile = ".vsecure";

		/// <summary>
		/// The extension for a file which contains a symmetrically encrypted document.
		/// </summary>
		public const string SymmEncryptedDocument = ".vsymm";

		/// <summary>
		/// The extension for a file which contains an asymmetrically encrypted document.
		/// </summary>
		public const string AsymmEncryptedDocument = ".vasymm";

		#endregion Vela specialized

		#region File extension sets

		/// <summary>
		/// File extensions specialized for Vela.
		/// </summary>
		public static readonly HashSet<string> VelaFile = new HashSet<string> { Storehouse, StorehouseMainAttachment
			, GenericDocument, StorableConfiguration, Presenter, AttestingDocument, SecuringDocument, SecureFile
			, SymmEncryptedDocument, AsymmEncryptedDocument };

		/// <summary>
		/// File extensions for a video file.
		/// </summary>
		public static readonly HashSet<string> VideoFile = new HashSet<string> { ".avi", ".divx", ".mkv", ".mov", ".mp4"
			, ".mpeg", ".mpg", ".m2ts", ".mts", ".ogv", ".webm", ".wmv" };

		/// <summary>
		/// File extensions for an audio file.
		/// </summary>
		public static readonly HashSet<string> AudioFile = new HashSet<string> { ".aac", ".ac3", ".flac", ".m4a", ".mp3"
			, ".oga", ".ogg", ".wav", ".wma" };

		/// <summary>
		/// File extensions for an image file.
		/// </summary>
		public static readonly HashSet<string> ImageFile = new HashSet<string> { ".bmp", ".gif", ".jpg", ".jpeg", ".png" };

		/// <summary>
		/// File extensions for a media file. This is filled later.
		/// </summary>
		public static readonly HashSet<string> MediaFile = new HashSet<string> { };

		#endregion File extension sets

		#region Filters

		/// <summary>
		/// Filter for a dialog to open any file.
		/// </summary>
		public const string AllFilesFilter = "All files (*)|*";

		/// <summary>
		/// Filter for a dialog to open a storehouse (header).
		/// </summary>
		public static readonly string StorehouseFilter = string.Format( "Storehouses (*{0})|*{0}", Storehouse );

		/// <summary>
		/// Filter for a dialog to open a generic document.
		/// </summary>
		public static readonly string GenericDocumentFilter = string.Format( "Generic documents (*{0})|*{0}", GenericDocument );

		/// <summary>
		/// Filter for a dialog to open a storable configuration.
		/// </summary>
		public static readonly string StorableConfigurationFilter = string.Format( "Storable configurations (*{0})|*{0}"
			, StorableConfiguration );

		/// <summary>
		/// Filter for a dialog to open a presenter.
		/// </summary>
		public static readonly string PresenterFilter = string.Format( "Presenters (*{0})|*{0}", Presenter );

		/// <summary>
		/// Filter for a dialog to open an attesting document.
		/// </summary>
		public static readonly string AttestingDocumentFilter = string.Format( "Attesting documents (*{0})|*{0}"
			, AttestingDocument );

		/// <summary>
		/// Filter for a dialog to open an securing document.
		/// </summary>
		public static readonly string SecuringDocumentFilter = string.Format( "Securing documents (*{0})|*{0}"
			, SecuringDocument );

		/// <summary>
		/// Filter for a dialog to open a secure file.
		/// </summary>
		public static readonly string SecureFileFilter = string.Format( "Secure files (*{0})|*{0}", SecureFile );

		/// <summary>
		/// Filter for a dialog to open a symmetrically encrypted document.
		/// </summary>
		public static readonly string SymmEncryptedDocumentFilter = string.Format( "Symmetrically encrypted (*{0})|*{0}"
			, SymmEncryptedDocument );

		/// <summary>
		/// Filter for a dialog to open an asymmetrically encrypted document.
		/// </summary>
		public static readonly string AsymmEncryptedDocumentFilter = string.Format( "Asymmetrically encrypted (*{0})|*{0}"
			, AsymmEncryptedDocument );

		/// <summary>
		/// Filter for a dialog to open an asymmetrically encrypted and (possibly) obfuscated document.
		/// </summary>
		public static readonly string AsymmEncryptedAndPossiblyObfuscatedDocumentFilter = string.Format(
			"Asymmetrically encrypted (*{0})|*{0}|All files (*)|*", AsymmEncryptedDocument );

		/// <summary>
		/// Filter for a dialog to open a video file.
		/// </summary>
		public const string VideoFilter = "Movies (*.avi, *.divx, *.mkv, *.mov, *.mp4, *.mpeg, *.mpg, *.m2ts, *.mts,"
			+ " *.ogv, *.webm, *.wmv)|*.avi; *.divx; *.mkv; *.mov; *.mp4; *.mpeg; *.mpg; *.m2ts; *.mts; *.ogv; *.webm; *.wmv";

		/// <summary>
		/// Filter for a dialog to open an audio file.
		/// </summary>
		public const string AudioFilter = "Sounds (*.aac, *.ac3, *.flac, *.m4a, *.mp3, *.oga, *.ogg, *.wav, *.wma)"
			+ "|*.aac; *.ac3; *.flac; *.m4a; *.mp3; *.oga; *.ogg; *.wav; *.wma";

		/// <summary>
		/// Filter for a dialog to open an image file.
		/// </summary>
		public const string ImageFilter = "Images (*.bmp, *.gif, *.jpg, *.jpeg, *.png)|*.bmp; *.gif; *.jpg; *.jpeg; *.png";

		#endregion Filters

		#endregion Public constants

		#region Static constructors

		/// <summary>
		/// Initializes the current class.
		/// </summary>
		static FileExtensions()
		{
			MediaFile.UnionWith( FileExtensions.VideoFile );
			MediaFile.UnionWith( FileExtensions.AudioFile );
			MediaFile.UnionWith( FileExtensions.ImageFile );
		}

		#endregion Static constructors

		#region Public static methods

		/// <summary>
		/// Checks if the specified file name represents a video file.
		/// </summary>
		/// <param name="fileName">File name to check if it represents a video file.</param>
		/// <returns>Returns true if the file name represents a video file.</returns>
		public static bool IsVideoFile( string fileName )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			return FileUtils.HasExtension( fileName, FileExtensions.VideoFile );
		}

		/// <summary>
		/// Checks if the specified file name represents an audio file.
		/// </summary>
		/// <param name="fileName">File name to check if it represents an audio file.</param>
		/// <returns>Returns true if the file name represents an audio file.</returns>
		public static bool IsAudioFile( string fileName )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			return FileUtils.HasExtension( fileName, FileExtensions.AudioFile );
		}

		/// <summary>
		/// Checks if the specified file name represents an image file.
		/// </summary>
		/// <param name="fileName">File name to check if it represents an image file.</param>
		/// <returns>Returns true if the file name represents an image file.</returns>
		public static bool IsImageFile( string fileName )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			return FileUtils.HasExtension( fileName, FileExtensions.ImageFile );
		}

		/// <summary>
		/// Checks if the specified file name represents a media file.
		/// </summary>
		/// <param name="fileName">File name to check if it represents a media file.</param>
		/// <returns>Returns true if the file name represents a media file.</returns>
		public static bool IsMediaFile( string fileName )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			return FileUtils.HasExtension( fileName, FileExtensions.MediaFile );
		}

		#endregion Public static methods
	}
}
