////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

namespace Vela.Kernel
{
	/// <summary>
	/// List of strings.
	/// </summary>
	/// <remarks>
	/// <para>Null or empty values can't be added to the list.</para>
	/// </remarks>
	public class Strings : KernelSortedList<string, string>
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public Strings()
			: base( StringInsensitiveComparer.Instance )
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="texts">List of texts to add to the list. This may be empty, but not null.</param>
		public Strings( string[] texts )
			: this()
		{
			ValidationUtils.ObjectHasContent( texts, "texts" );

			foreach( string text in texts )
				Add( text );
		}

		#endregion Constructors

		#region Private methods

		/// <summary>
		/// Validates the specified item before adding or updating it to the list. Inheritors must throw an exception if the
		/// item must not be added to the list.
		/// </summary>
		/// <param name="item">Item to add to the list. There is no need to check if this is null because this is performed
		/// before the method is called.</param>
		/// <remarks>
		/// <para>Don't check for duplicate items here because this is automatically done.</para>
		/// </remarks>
		protected override void ValidateAddOrUpdate( string item )
		{
			ValidationUtils.TextHasContent( item, "item" );
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para>The item itself is used as key.</para>
		/// </remarks>
		public override string GetKey( string item )
		{
			ValidationUtils.TextHasContent( item, "item" );

			return item;
		}

		/// <summary>
		/// Gets the list's values as an array.
		/// </summary>
		/// <returns>Returns the list's values as an array.</returns>
		public string[] ToArray()
		{
			string[] values = new string[ Count ];

			int i = 0;
			foreach( string value in this )
			{
				values[ i ] = value;
				i++;
			}

			return values;
		}

		#endregion Public methods
	}
}
