////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;

namespace Vela.Kernel
{
	/// <summary>
	/// A progressing mechanism for a long operation.
	/// </summary>
	public class LongOperationProgress
	{
		#region Public constants

		/// <summary>
		/// The minimum value for <see cref="UpdateTimespan"/>.
		/// </summary>
		public const double UpdateTimespanMin = 0.1;

		/// <summary>
		/// The maximum value for <see cref="UpdateTimespan"/>.
		/// </summary>
		public const double UpdateTimespanMax = 2;

		/// <summary>
		/// The default value for <see cref="UpdateTimespan"/>.
		/// </summary>
		public const double UpdateTimespanDef = 0.3;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="progressing">Event raised when progress occurs. This may be null.</param>
		public LongOperationProgress( LongOperationProgressDelegate progressing )
		{
			Progressing = progressing;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The synchronization object for the <see cref="IsAbort"/> property.
		/// </summary>
		private readonly object IsAbortLock = new object();

		/// <summary>
		/// Event raised when progress occurs. This may be null.
		/// </summary>
		private LongOperationProgressDelegate Progressing;

		/// <summary>
		/// The time elapsed from the beginning of the long operation. This is initialized by default.
		/// </summary>
		private Chronometer ElapsedTimeChrono = new Chronometer();

		/// <summary>
		/// The chronometer which is used to see when enough time has passed so that progress can occur. This is initialized by
		/// default.
		/// </summary>
		private Chronometer ShouldProgressChrono = new Chronometer();

		#endregion Private fields

		#region Public non-serializable properties

		/// <summary>
		/// The time (in seconds) elapsed from the beginning of the long operation. READ-ONLY.
		/// </summary>
		public double ElapsedTime
		{
			get { return ElapsedTimeChrono.Seconds; }
		}

		/// <summary>
		/// The time span elapsed from the beginning of the long operation. READ-ONLY.
		/// </summary>
		public TimeSpan ElapsedTimeSpan
		{
			get { return ElapsedTimeChrono.TimeSpan; }
		}

		private bool _IsAbort = false;
		/// <summary>
		/// The abort state of a long operation. If this is true then an exception
		/// (<see cref="AbortLongOperationException"/>) is thrown in the <see cref="Progress"/> method.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// </remarks>
		public bool IsAbort
		{
			get
			{
				lock( IsAbortLock )
				{
					return _IsAbort;
				}
			}
			set
			{
				lock( IsAbortLock )
				{
					_IsAbort = value;
				}
			}
		}

		private double _UpdateTimespan = UpdateTimespanDef;
		/// <summary>
		/// The timespan (in seconds) which must pass in order for progress to occur. This must be at least
		/// <see cref="UpdateTimespanMin"/>.
		/// </summary>
		public double UpdateTimespan
		{
			get { return _UpdateTimespan; }
			set
			{
				ValidationUtils.Real( value, UpdateTimespanMin, UpdateTimespanMax, "UpdateTimespan" );

				_UpdateTimespan = value;
			}
		}

		private object _Context;
		/// <summary>
		/// Context. This may be null.
		/// </summary>
		public object Context
		{
			get { return _Context; }
			set { _Context = value; }
		}

		private string _PersistentMessage;
		/// <summary>
		/// Persistent message. This may be null or empty.
		/// </summary>
		public string PersistentMessage
		{
			get { return _PersistentMessage; }
			set { _PersistentMessage = value; }
		}

		private double _LastProgress;
		/// <summary>
		/// Last progress value (when progress occurred). This should be within [0...1].
		/// </summary>
		public double LastProgress
		{
			get { return _LastProgress; }
			set { _LastProgress = value; }
		}

		private string _LastMessage;
		/// <summary>
		/// Last message (when progress occurred). This may be null or empty.
		/// </summary>
		public string LastMessage
		{
			get { return _LastMessage; }
			set { _LastMessage = value; }
		}

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Clears progressing.
		/// </summary>
		public void Clear()
		{
			ElapsedTimeChrono.Clear();
			ShouldProgressChrono.Clear();

			IsAbort = false;

			Context = null;
			PersistentMessage = null;
			LastProgress = 0;
			LastMessage = null;

			if( Progressing != null ) Progressing( this );
		}

		/// <summary>
		/// Starts progressing. Call this when you start a long operation.
		/// </summary>
		/// <param name="context">Context.</param>
		/// <param name="persistentMessage">Persistent message.</param>
		/// <param name="message">Message. This may be null or empty.</param>
		public void Start( object context, string persistentMessage, string message = null )
		{
			ElapsedTimeChrono.Start();
			ShouldProgressChrono.Start();

			IsAbort = false;

			Context = context;
			PersistentMessage = persistentMessage;
			LastProgress = 0;
			LastMessage = message;

			if( Progressing != null ) Progressing( this );
		}

		/// <summary>
		/// Starts progressing. Call this when you start a long operation.
		/// </summary>
		/// <param name="message">Message. This may be null or empty.</param>
		public void Start( string message = null )
		{
			ElapsedTimeChrono.Start();
			ShouldProgressChrono.Start();

			IsAbort = false;

			Context = null;
			PersistentMessage = null;
			LastProgress = 0;
			LastMessage = message;

			if( Progressing != null ) Progressing( this );
		}

		/// <summary>
		/// Progresses.
		/// </summary>
		/// <param name="progress">Progress value. This should be within [0...1].</param>
		/// <param name="message">Message. This may be null or empty.</param>
		/// <param name="force">If this is true then the progress occurs even if the update timespan hasn't passed.</param>
		public void Progress( double progress = 0, string message = null, bool force = false )
		{
			if( !force && !ShouldProgressChrono.PassedSeconds( UpdateTimespan ) ) return;

			LastProgress = progress;
			LastMessage = message;

			if( Progressing != null ) Progressing( this );

			// Throw exception if the process must be aborted.
			if( IsAbort ) throw new AbortLongOperationException();
		}

		/// <summary>
		/// Stops progressing. Call this when you stop a long operation.
		/// </summary>
		/// <param name="message">Message. This may be null or empty.</param>
		public void Stop( string message = null )
		{
			ElapsedTimeChrono.Stop();
			ShouldProgressChrono.Stop();

			IsAbort = false;

			LastMessage = message;

			if( Progressing != null ) Progressing( this );
		}

		#endregion Public methods
	}
}
