////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections;
using Vela.Kernel.Serialization;

namespace Vela.Kernel
{
	/// <summary>
	/// A tuple (made from multiple values).
	/// </summary>
	public class Tuple : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Tuple()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="values">The values which form the key. Each value must have content.</param>
		public Tuple( params object[] values )
		{
			ValidationUtils.ArrayHasContent( values, "values" );

			foreach( object value in values )
			{
				ValidateValue( value );

				Values.Add( value );
			}
		}

		#endregion Constructors

		#region Public serializable properties

		private ArrayList _Values = new ArrayList();
		/// <summary>
		/// The values which form the key. Each value must have content, but this is not enforced here. The backing field is
		/// initialized by default. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public ArrayList Values
		{
			get { return _Values; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Values" );

				_Values = value;
			}
		}

		#endregion Public serializable properties

		#region Operators

		/// <summary>
		/// Verifies if the specified tuples have the same value.
		/// </summary>
		/// <param name="tuple1">Tuple to verify. This may be null.</param>
		/// <param name="tuple2">Tuple to verify. This may be null.</param>
		/// <returns>Returns true if the tuples have the same value.</returns>
		/// <remarks>
		/// <para>The tuples are equal if their values are all equal.</para>
		/// </remarks>
		public static bool operator ==( Tuple tuple1, Tuple tuple2 )
		{
			if( ReferenceEquals( tuple1, tuple2 ) )
				return true;

			// Null tuples are not equal.
			if( ((object)tuple1 == null) || ((object)tuple2 == null) )
				return false;

			return tuple1.Equals( tuple2 );
		}

		/// <summary>
		/// Verifies if the specified tuples have a different value.
		/// </summary>
		/// <param name="tuple1">Tuple to verify. This may be null.</param>
		/// <param name="tuple2">Tuple to verify. This may be null.</param>
		/// <returns>Returns true if the tuples have a different value.</returns>
		/// <remarks>
		/// <para>The tuples are different if at least 1 of their values is different.</para>
		/// </remarks>
		public static bool operator !=( Tuple tuple1, Tuple tuple2 )
		{
			if( ReferenceEquals( tuple1, tuple2 ) )
				return false;

			// Null tuples are not equal.
			if( ((object)tuple1 == null) || ((object)tuple2 == null) )
				return true;

			return !tuple1.Equals( tuple2 );
		}

		#endregion Operators

		#region Public methods

		/// <summary>
		/// Gets the hash code of the current tuple.
		/// </summary>
		/// <returns>Returns the hash code of the current tuple.</returns>
		public override int GetHashCode()
		{
			int hashCode = 0;

			foreach( object value in Values )
				hashCode |= value.GetHashCode();

			return hashCode;
		}

		/// <summary>
		/// Verifies if the current and the specified tuple have the same value.
		/// </summary>
		/// <param name="tuple">Tuple to verify. This may be null.</param>
		/// <returns>Returns true if the tuples have the same value.</returns>
		/// <remarks>
		/// <para>The tuples are equal if their values are all equal.</para>
		/// </remarks>
		public bool Equals( Tuple tuple )
		{
			if( tuple == null )
				return false;

			if( Values.Count != tuple.Values.Count )
			{
				throw new ValidationException( string.Format( "The tuples to check for equality contain a different number"
					+ " of values: {0} and {1}.", Values.Count, tuple.Values.Count ) );
			}

			int count = Values.Count;
			for( int i = 0; i < count; i++ )
			{
				object value1 = Values[ i ];
				object value2 = tuple.Values[ i ];

				Type type1 = value1.GetType();
				Type type2 = value2.GetType();

				if( type1 != type2 )
				{
					throw new ValidationException( string.Format( "The types of the compared values of the tuples are"
						+ " different: \"{0}\" and \"{1}\".", type1.FullName, type2.FullName ) );
				}

				if( !value1.Equals( value2 ) )
					return false; // The tuples are NOT equal.
			}

			// The tuples are equal.
			return true;
		}

		/// <summary>
		/// Verifies if the current and the specified tuples have the same value.
		/// </summary>
		/// <param name="tuple">Tuple to verify.</param>
		/// <returns>Returns true if the tuples have the same value.</returns>
		/// <remarks>
		/// <para>The tuples are equal if their values are all equal.</para>
		/// </remarks>
		public override bool Equals( object tuple )
		{
			return Equals( tuple as Tuple );
		}

		/// <summary>
		/// Compares the current and the specified tuples.
		/// </summary>
		/// <param name="tuple">Tuple to compare.</param>
		/// <returns>Returns 0 if the tuples have the same value, -1 if the current tuple is smaller than "tuple", and 1 if the
		/// current tuple is greater than "tuple".</returns>
		public int Compare( Tuple tuple )
		{
			ValidationUtils.ObjectHasContent( tuple, "tuple" );

			if( Values.Count != tuple.Values.Count )
			{
				throw new ValidationException( string.Format( "The tuples to check for equality contain a different number"
					+ " of values: {0} and {1}.", Values.Count, tuple.Values.Count ) );
			}

			int count = Values.Count;
			for( int i = 0; i < count; i++ )
			{
				object value1 = Values[ i ];
				object value2 = tuple.Values[ i ];

				Type type1 = value1.GetType();
				Type type2 = value2.GetType();

				if( type1 != type2 )
				{
					throw new ValidationException( string.Format( "The types of the compared values of the tuples are"
						+ " different: \"{0}\" and \"{1}\".", type1.FullName, type2.FullName ) );
				}

				int comparison = 0;
				if( type1 == typeof( int ) )
				{
					comparison = ((int)value1).CompareTo( (int)value2 );
				}
				else if( type1 == typeof( DateTime ) )
				{
					comparison = DateTime.Compare( (DateTime)value1, (DateTime)value2 );
				}
				else if( type1 == typeof( byte[] ) )
				{
					comparison = BufferUtils.Compare( (byte[])value1, (byte[])value2 );
				}
				else if( type1 == typeof( string ) )
				{
					comparison = string.Compare( (string)value1, (string)value2, StringComparison.OrdinalIgnoreCase );
				}
				else if( type1 == typeof( Locator ) )
				{
					comparison = ((Locator)value1).Compare( (Locator)value2 );
				}
				else if( typeof( ITextualFormContainer ).IsAssignableFrom( type1 ) )
				{
					string text1 = ((ITextualFormContainer)value1).TextualForm;
					string text2 = ((ITextualFormContainer)value2).TextualForm;

					comparison = string.Compare( text1, text2, StringComparison.OrdinalIgnoreCase );
				}
				else
				{
					throw new ValidationException( string.Format( "The type \"{0}\" of a value of the tuples to compare is"
						+ " invalid.", type1.FullName ) );
				}

				if( comparison != 0 )
					return comparison;
			}

			// The tuples are equal.
			return 0;
		}

		/// <summary>
		/// Validates the specified value, including its type. An exception is thrown if the value is not valid.
		/// </summary>
		/// <param name="value">Value to validate.</param>
		public void ValidateValue( object value )
		{
			ValidationUtils.ObjectHasContent( value, "value" );

			Type type = value.GetType();

			if( (type != typeof( int )) && (type != typeof( DateTime )) && (type != typeof( byte[] ))
				&& (type != typeof( string )) && (type != typeof( Locator ))
				&& !typeof( ITextualFormContainer ).IsAssignableFrom( type ) )
			{
				throw new ValidationException( string.Format( "The type \"{0}\" is not valid for the value of a tuple."
					, type.FullName ) );
			}
		}

		#endregion Public methods
	}
}
