////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Net.Sockets;

namespace Vela.Kernel
{
	/// <summary>
	/// Vela-wide constants.
	/// </summary>
	public static class KernelConsts
	{
		#region Public constants

		/// <summary>
		/// The maximum value for a precision progress (representing 100%), that is, a progress which requires a higher precision
		/// than an integer percentage.
		/// </summary>
		public const int PrecisionProgressMax = 10000;

		/// <summary>
		/// The value by which a precision progress has to be divided in order to be converted to a percentage.
		/// </summary>
		public const double PrecisionProgressToPercentageDivisor = PrecisionProgressMax / 100.0;

		/// <summary>
		/// The format for converting a time to text.
		/// </summary>
		public const string TimeFormat = "yyyy'-'MM'-'dd HH':'mm':'ss";

		/// <summary>
		/// The format for converting a time to a compact text.
		/// </summary>
		public const string TimeCompactFormat = "yyyyMMddHHmmss";

		/// <summary>
		/// The format for converting a time to text when it has to be frequently shown as text in the user interface.
		/// </summary>
		public const string TimeUiFormat = "yyyy'-'MM'-'dd HH':'mm':'ss";

		/// <summary>
		/// The format for converting a date to text.
		/// </summary>
		public const string DateFormat = "yyyy'-'MM'-'dd";

		/// <summary>
		/// The format for converting a date to text, meant to be displayed.
		/// </summary>
		public const string DateFormatForDisplay = "yyyy-MM-dd";

		/// <summary>
		/// The short format for converting a date to text when it has to be frequently shown as text in the user interface.
		/// </summary>
		public const string DateUiShortFormat = "d";

		/// <summary>
		/// The long format for converting a date to text when it has to be frequently shown as text in the user interface.
		/// </summary>
		public const string DateUiLongFormat = "D";

		/// <summary>
		/// The format for converting a clock to text.
		/// </summary>
		public const string ClockFormat = "HH':'mm':'ss";

		/// <summary>
		/// The short format for converting a clock to text when it has to be frequently shown as text in the user interface.
		/// </summary>
		public const string ClockUiShortFormat = "t";

		/// <summary>
		/// The long format for converting a clock to text when it has to be frequently shown as text in the user interface.
		/// </summary>
		public const string ClockUiLongFormat = "T";

		/// <summary>
		/// The format for converting a time span to text.
		/// </summary>
		public const string TimeSpanFormat = "d'T'hh':'mm':'ss'.'f";

		/// <summary>
		/// The long format for converting a time span to text when it has to be frequently shown as text in the user interface.
		/// </summary>
		public const string TimeSpanUiLongFormat = "d':'hh':'mm':'ss'.'f";

		/// <summary>
		/// The short format for converting a time span to text when it has to be frequently shown as text in the user interface.
		/// </summary>
		public const string TimeSpanUiShortFormat = "d':'hh':'mm':'ss";

		/// <summary>
		/// The buffer size (in bytes) for reading from or writing to a file.
		/// </summary>
		public const int FileBufferSize = 16 * 1024;

		/// <summary>
		/// The maximum number of characters from a textual representation of a <see cref="UInt16"/> value.
		/// </summary>
		public const int UInt16TextSizeMax = 5;

		/// <summary>
		/// The maximum number of characters from a textual representation of a <see cref="UInt32"/> value.
		/// </summary>
		public const int UInt32TextSizeMax = 10;

		#region Images

		/// <summary>
		/// The maximum width (in pixels) of an image.
		/// </summary>
		public const int ImageWidthMax = 20000;

		/// <summary>
		/// The maximum height (in pixels) of an image.
		/// </summary>
		public const int ImageHeightMax = 20000;

		/// <summary>
		/// The maximum size (in bytes) of a peer image.
		/// </summary>
		public const int PeerImageSizeMax = 10000000;

		/// <summary>
		/// The reference size (in pixels) of an image.
		/// </summary>
		public const int ImageSizeReference = 64;

		/// <summary>
		/// The height (in pixels) of a thumb-sized image preview.
		/// </summary>
		/// <remarks>
		/// <para>XAML requires a <see cref="double"/> value.</para>
		/// </remarks>
		public const double ThumbSizedImagePreviewHeight = ImageSizeReference;

		/// <summary>
		/// The height (in pixels) of a thumb-sized image preview.
		/// </summary>
		public const int ThumbSizedImagePreviewHeightInt = (int)ThumbSizedImagePreviewHeight;

		/// <summary>
		/// The width (in pixels) of a thumb-sized image preview.
		/// </summary>
		/// <remarks>
		/// <para>XAML requires a <see cref="double"/> value.</para>
		/// </remarks>
		public const double ThumbSizedImagePreviewWidth = 3 * ThumbSizedImagePreviewHeight;

		/// <summary>
		/// The width (in pixels) of a thumb-sized image preview.
		/// </summary>
		public const int ThumbSizedImagePreviewWidthInt = (int)ThumbSizedImagePreviewWidth;

		/// <summary>
		/// The height (in pixels) of a full-sized image preview.
		/// </summary>
		/// <remarks>
		/// <para>XAML requires a <see cref="double"/> value.</para>
		/// </remarks>
		public const double FullSizedImagePreviewHeight = 3 * ImageSizeReference;

		/// <summary>
		/// The height (in pixels) of a full-sized image preview.
		/// </summary>
		public const int FullSizedImagePreviewHeightInt = (int)FullSizedImagePreviewHeight;

		/// <summary>
		/// The width (in pixels) of a full-sized image preview.
		/// </summary>
		/// <remarks>
		/// <para>XAML requires a <see cref="double"/> value.</para>
		/// </remarks>
		public const double FullSizedImagePreviewWidth = 3 * FullSizedImagePreviewHeight;

		/// <summary>
		/// The width (in pixels) of a full-sized image preview.
		/// </summary>
		public const int FullSizedImagePreviewWidthInt = (int)FullSizedImagePreviewWidth;

		#endregion Images

		#region Bounds

		/// <summary>
		/// The minimum size of an alias.
		/// </summary>
		public const int AliasSizeMin = 4;

		/// <summary>
		/// The maximum size of an alias.
		/// </summary>
		public const int AliasSizeMax = 50;

		/// <summary>
		/// The maximum size of an alternative peer address.
		/// </summary>
		public const int AlternativePeerAddressSizeMax = 253;

		/// <summary>
		/// The maximum size of a code.
		/// </summary>
		public const int CodeSizeMax = 50;

		/// <summary>
		/// The maximum size of a command line.
		/// </summary>
		public const int CommandLineSizeMax = 8191;

		/// <summary>
		/// The maximum size of an email address.
		/// </summary>
		public const int EmailAddressSizeMax = 50;

		/// <summary>
		/// The minimum size of an emoticon symbol.
		/// </summary>
		public const int EmoticonSymbolSizeMin = 2;

		/// <summary>
		/// The maximum size of an emoticon symbol.
		/// </summary>
		public const int EmoticonSymbolSizeMax = 20;

		/// <summary>
		/// The minimum size of an entity name.
		/// </summary>
		public const int EntityNameSizeMin = 2;

		/// <summary>
		/// The maximum size of an entity name.
		/// </summary>
		public const int EntityNameSizeMax = 50;

		/// <summary>
		/// The maximum size of a file / folder name.
		/// </summary>
		public const int FileNameSizeMax = 259;

		/// <summary>
		/// The maximum size of a file / folder title.
		/// </summary>
		public const int FileTitleSizeMax = 247;

		/// <summary>
		/// The maximum size of an IP address.
		/// </summary>
		public const int IpAddressSizeMax = 40;

		/// <summary>
		/// The maximum size (in characters) of a legal contract.
		/// </summary>
		public const int LegalContractSizeMax = 100000;

		/// <summary>
		/// The maximum size of a long host label (which is meant to be used in a host name).
		/// </summary>
		public const int LongHostLabelSizeMax = 63;

		/// <summary>
		/// The maximum size of a short host label (which is meant to be used in a host name).
		/// </summary>
		public const int ShortHostLabelSizeMax = 32;

		/// <summary>
		/// The maximum size (in characters) of a medium note.
		/// </summary>
		public const int MediumNoteSizeMax = 500;

		/// <summary>
		/// The maximum size (in characters) of a long note.
		/// </summary>
		public const int LongNoteSizeMax = 10000;

		/// <summary>
		/// The maximum size (in characters) of a short note.
		/// </summary>
		public const int ShortNoteSizeMax = 100;

		/// <summary>
		/// The maximum size (in characters) of a tiny note.
		/// </summary>
		public const int TinyNoteSizeMax = 30;

		/// <summary>
		/// The minimum size of a binary password.
		/// </summary>
		public const int BinaryPasswordSizeMin = 32;

		/// <summary>
		/// The minimum size of a password.
		/// </summary>
		public const int PasswordSizeMin = 10;

		/// <summary>
		/// The maximum size of a password.
		/// </summary>
		public const int PasswordSizeMax = 100;

		/// <summary>
		/// The maximum size of a phone number.
		/// </summary>
		public const int PhoneNumberSizeMax = 30;

		/// <summary>
		/// The maximum size (in characters) of a short command.
		/// </summary>
		public const int ShortCommandSizeMax = 1000;

		/// <summary>
		/// The maximum size of an UPN.
		/// </summary>
		public const int UpnSizeMax = 150;

		/// <summary>
		/// The maximum size of an URL.
		/// </summary>
		public const int UrlSizeMax = 32766;

		/// <summary>
		/// The minimum valid date (this allows for the oldest possible birth year).
		/// </summary>
		public static readonly DateTime DateMin = new DateTime( 1900, 1, 1 );

		#endregion Bounds

		/// <summary>
		/// The non-alphanumeric characters which are valid in a <see cref="TextValidations.Alias"/>.
		/// </summary>
		public static readonly HashSet<char> AliasNanValidCharacters = new HashSet<char> { '@', '.', '_' };

		/// <summary>
		/// The non-alphanumeric characters which are valid in a <see cref="TextValidations.EmailAddress"/>.
		/// </summary>
		public static readonly HashSet<char> EmailAddressNanValidCharacters = new HashSet<char> { '@', '.', '_' };

		/// <summary>
		/// The non-alphanumeric characters which are valid in a <see cref="TextValidations.EmoticonSymbol"/>.
		/// </summary>
		public static readonly HashSet<char> EmoticonSymbolNanValidCharacters = new HashSet<char> { '!', '"', '#', '$', '&'
			, '\'', '(', ')', '*', '+', ',', '/', ':', ';', '<', '=', '>', '?', '@', '|', '[', ']', '%', '\\', '{', '}', '.'
			, '_', '-' };

		/// <summary>
		/// The non-alphanumeric characters which are valid in a <see cref="TextValidations.Password"/>.
		/// </summary>
		public static readonly HashSet<char> PasswordNanValidCharacters = new HashSet<char> { '!', '"', '#', '$', '&'
			, '\'', '(', ')', '*', '+', ',', '/', ':', ';', '<', '=', '>', '?', '@', '|', '[', ']', '%', '\\', '{', '}', '.'
			, '_', '-' };

		/// <summary>
		/// The non-alphanumeric characters which are valid in a <see cref="TextValidations.PhoneNumber"/>.
		/// </summary>
		public static readonly HashSet<char> PhoneNumberNanValidCharacters = new HashSet<char> { '+', '.' };

		#region Network

		/// <summary>
		/// The socket packet type used in network connections.
		/// </summary>
		public const SocketType SocketPacketType = SocketType.Stream;

		/// <summary>
		/// The socket protocol type used in network connections.
		/// </summary>
		public const ProtocolType SocketProtocolType = ProtocolType.Tcp;

		#endregion Network

		#endregion Public constants
	}
}
