////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.ComponentModel;

namespace Vela.Kernel
{
	/// <summary>
	/// The available modes for a log record (<see cref="LogRecord"/>).
	/// </summary>
	public enum LogRecordModes
	{
		/// <summary>
		/// The log record is an error.
		/// </summary>
		Error,

		/// <summary>
		/// The log record is a warning.
		/// </summary>
		Warning,

		/// <summary>
		/// The log record is an information.
		/// </summary>
		Info
	}

	/// <summary>
	/// The available display priorities for a log record field (<see cref="LogRecordField"/>).
	/// </summary>
	public enum LogRecordFieldPriorities
	{
		/// <summary>
		/// The log record field should be displayed at the beginning of the log message.
		/// </summary>
		Beginning,

		/// <summary>
		/// The log record field should be displayed in the middle of the log message.
		/// </summary>
		Middle,

		/// <summary>
		/// The log record field should be displayed at the end of the log message.
		/// </summary>
		End
	}

	/// <summary>
	/// The available validation text types.
	/// </summary>
	/// <remarks>
	/// <para>General rules of validation (except for <see cref="TextValidations.None"/>):</para>
	/// <list type="bullet">
	///		<item>Control characters are not allowed.</item>
	///		<item>Whitespaces are not allowed at the beginning or end of the text.</item>
	///		<item>Double whitespaces are not allowed.</item>
	/// </list>
	/// </remarks>
	public enum TextValidations
	{
		/// <summary>
		/// All characters are valid.
		/// </summary>
		None,

		/// <summary>
		/// The valid characters are: letters, digits, the special characters from
		/// <see cref="KernelConsts.AliasNanValidCharacters"/>. The default maximum size is
		/// <see cref="KernelConsts.AliasSizeMax"/>.
		/// </summary>
		Alias,

		/// <summary>
		/// The valid characters are defined by the infrastructure for either an IP address or a host name. It may also contain
		/// the <see cref="Vela.Kernel.AlternativePeerAddress.UpnLongHostLabelPlaceholder"/> placeholder (which will be
		/// replaced with a UPN long host label, limited to <see cref="KernelConsts.LongHostLabelSizeMax"/>) or
		/// the <see cref="Vela.Kernel.AlternativePeerAddress.UpnShortHostLabelPlaceholder"/> placeholder (which will be
		/// replaced with a UPN short host label, limited to <see cref="KernelConsts.ShortHostLabelSizeMax"/>).
		/// </summary>
		AlternativePeerAddress,

		/// <summary>
		/// The valid characters are: letters, digits. The default maximum size is <see cref="KernelConsts.CodeSizeMax"/>.
		/// </summary>
		Code,

		/// <summary>
		/// The valid characters are: letters, digits, the special characters from
		/// <see cref="KernelConsts.EmailAddressNanValidCharacters"/>. The first character must be a letter; the '@' character
		/// must exist but not at the beginning or end of the text. The default maximum size is
		/// <see cref="KernelConsts.EmailAddressSizeMax"/>.
		/// </summary>
		EmailAddress,

		/// <summary>
		/// The valid characters are: letters, digits, the special characters from
		/// <see cref="KernelConsts.EmoticonSymbolNanValidCharacters"/>. The default maximum size is
		/// <see cref="KernelConsts.EmoticonSymbolSizeMax"/>.
		/// </summary>
		EmoticonSymbol,

		/// <summary>
		/// The valid characters are: letters, spaces. The first character must be a letter. The default maximum size is
		/// <see cref="KernelConsts.EntityNameSizeMax"/>.
		/// </summary>
		EntityName,

		/// <summary>
		/// The valid characters are defined by the infrastructure. The default maximum size is
		/// <see cref="KernelConsts.FileNameSizeMax"/>.
		/// </summary>
		FileName,

		/// <summary>
		/// The valid characters are defined by the infrastructure. The default maximum size is
		/// <see cref="KernelConsts.FileTitleSizeMax"/>.
		/// </summary>
		FileTitle,

		/// <summary>
		/// The valid characters are defined by the infrastructure. The default maximum size is
		/// <see cref="KernelConsts.IpAddressSizeMax"/>.
		/// </summary>
		IpAddress,

		/// <summary>
		/// The valid characters are: letters.
		/// </summary>
		Letters,

		/// <summary>
		/// All characters are valid. The default maximum size is <see cref="KernelConsts.LongNoteSizeMax"/>.
		/// </summary>
		LongNote,

		/// <summary>
		/// All characters are valid. The default maximum size is <see cref="KernelConsts.MediumNoteSizeMax"/>.
		/// </summary>
		MediumNote,

		/// <summary>
		/// The valid characters are: letters, digits, the special characters from
		/// <see cref="KernelConsts.PasswordNanValidCharacters"/>. The default minimum size is
		/// <see cref="KernelConsts.PasswordSizeMin"/>. The default maximum size is <see cref="KernelConsts.PasswordSizeMax"/>.
		/// </summary>
		Password,

		/// <summary>
		/// The valid characters are: digits, the special characters from
		/// <see cref="KernelConsts.PhoneNumberNanValidCharacters"/>. The default maximum size is
		/// <see cref="KernelConsts.PhoneNumberSizeMax"/>.
		/// </summary>
		PhoneNumber,

		/// <summary>
		/// All characters are valid. The default maximum size is <see cref="KernelConsts.ShortCommandSizeMax"/>.
		/// </summary>
		ShortCommand,

		/// <summary>
		/// All characters are valid. The default maximum size is <see cref="KernelConsts.ShortNoteSizeMax"/>.
		/// </summary>
		ShortNote,

		/// <summary>
		/// All characters are valid. The default maximum size is <see cref="KernelConsts.TinyNoteSizeMax"/>.
		/// </summary>
		TinyNote,

		/// <summary>
		/// The valid characters are: letters, digits, spaces. The default maximum size is <see cref="KernelConsts.UpnSizeMax"/>.
		/// </summary>
		Upn,

		/// <summary>
		/// The valid characters are defined by the infrastructure as those from a relative or absolute URL. The '.' character
		/// must exist but not at the beginning or end of the text. The default maximum size is
		/// <see cref="KernelConsts.UrlSizeMax"/>.
		/// </summary>
		Url,
	}

	/// <summary>
	/// The available security levels for the cryptographic algorithms to use.
	/// </summary>
	/// <remarks>
	/// <para>The higher the security level is, the slower the algorithms are.</para>
	/// </remarks>
	public enum AlgorithmSecurityLevel
	{
		/// <summary>
		/// The security level is normal. This should be used most of the time.
		/// </summary>
		Normal,

		/// <summary>
		/// The security level is strong. This should be used only if you actually need very high security.
		/// </summary>
		Strong,

		/// <summary>
		/// The security level is strongest. This should be used only if you actually need the highest security.
		/// </summary>
		Strongest
	}

	/// <summary>
	/// The available genders.
	/// </summary>
	public enum Genders
	{
		/// <summary>
		/// No gender has been specified.
		/// </summary>
		[Description( "Not specified" )]
		NotSpecified,

		/// <summary>
		/// The gender is female.
		/// </summary>
		Female,

		/// <summary>
		/// The gender is male.
		/// </summary>
		Male,

		/// <summary>
		/// The gender is another.
		/// </summary>
		Other
	}

	/// <summary>
	/// The available countries.
	/// </summary>
	public enum Countries
	{
		/// <summary>
		/// No country has been specified.
		/// </summary>
		[Description( "Not specified" )]
		NotSpecified,

		/// <summary></summary>
		Afghanistan,

		/// <summary></summary>
		[Description( "Aland Islands" )]
		AlandIslands,

		/// <summary></summary>
		Albania,

		/// <summary></summary>
		Algeria,

		/// <summary></summary>
		[Description( "American Samoa" )]
		AmericanSamoa,

		/// <summary></summary>
		Andorra,

		/// <summary></summary>
		Angola,

		/// <summary></summary>
		Anguilla,

		/// <summary></summary>
		Antarctica,

		/// <summary></summary>
		[Description( "Antigua Barbuda" )]
		AntiguaBarbuda,

		/// <summary></summary>
		Argentina,

		/// <summary></summary>
		Armenia,

		/// <summary></summary>
		Aruba,

		/// <summary></summary>
		Australia,

		/// <summary></summary>
		Austria,

		/// <summary></summary>
		Azerbaijan,

		/// <summary></summary>
		Bahamas,

		/// <summary></summary>
		Bahrain,

		/// <summary></summary>
		Bangladesh,

		/// <summary></summary>
		Barbados,

		/// <summary></summary>
		Belarus,

		/// <summary></summary>
		Belgium,

		/// <summary></summary>
		Belize,

		/// <summary></summary>
		Benin,

		/// <summary></summary>
		Bermuda,

		/// <summary></summary>
		Bhutan,

		/// <summary></summary>
		Bolivia,

		/// <summary></summary>
		[Description( "Bonaire (Sint Eustatius and Saba)" )]
		Bonaire,

		/// <summary></summary>
		[Description( "Bosnia Herzegovina" )]
		BosniaHerzegovina,

		/// <summary></summary>
		Botswana,

		/// <summary></summary>
		[Description( "Bouvet Island" )]
		BouvetIsland,

		/// <summary></summary>
		Brazil,

		/// <summary></summary>
		Brunei,

		/// <summary></summary>
		Bulgaria,

		/// <summary></summary>
		[Description( "Burkina Faso" )]
		BurkinaFaso,

		/// <summary></summary>
		Burundi,

		/// <summary></summary>
		Cambodia,

		/// <summary></summary>
		Cameroon,

		/// <summary></summary>
		Canada,

		/// <summary></summary>
		[Description( "Cape Verde" )]
		CapeVerde,

		/// <summary></summary>
		[Description( "Cayman Islands" )]
		CaymanIslands,

		/// <summary></summary>
		[Description( "Central African Republic" )]
		CentralAfricanRepublic,

		/// <summary></summary>
		Chad,

		/// <summary></summary>
		Chile,

		/// <summary></summary>
		China,

		/// <summary></summary>
		[Description( "Christmas Island" )]
		ChristmasIsland,

		/// <summary></summary>
		[Description( "Cocos Islands" )]
		CocosIslands,

		/// <summary></summary>
		Colombia,

		/// <summary></summary>
		Comoros,

		/// <summary></summary>
		Congo,

		/// <summary></summary>
		[Description( "Cook Islands" )]
		CookIslands,

		/// <summary></summary>
		[Description( "Costa Rica" )]
		CostaRica,

		/// <summary></summary>
		[Description( "Cote d'Ivoire" )]
		CoteDIvoire,

		/// <summary></summary>
		Croatia,

		/// <summary></summary>
		Cuba,

		/// <summary></summary>
		Curacao,

		/// <summary></summary>
		Cyprus,

		/// <summary></summary>
		[Description( "Czech Republic" )]
		CzechRepublic,

		/// <summary></summary>
		Denmark,

		/// <summary></summary>
		Djibouti,

		/// <summary></summary>
		Dominica,

		/// <summary></summary>
		[Description( "Dominican Republic" )]
		DominicanRepublic,

		/// <summary></summary>
		Ecuador,

		/// <summary></summary>
		Egypt,

		/// <summary></summary>
		[Description( "El Salvador" )]
		ElSalvador,

		/// <summary></summary>
		[Description( "Equatorial Guinea" )]
		EquatorialGuinea,

		/// <summary></summary>
		Eritrea,

		/// <summary></summary>
		Estonia,

		/// <summary></summary>
		Ethiopia,

		/// <summary></summary>
		[Description( "Falkland Islands" )]
		FalklandIslands,

		/// <summary></summary>
		[Description( "Faroe Islands" )]
		FaroeIslands,

		/// <summary></summary>
		Fiji,

		/// <summary></summary>
		Finland,

		/// <summary></summary>
		France,

		/// <summary></summary>
		[Description( "French Guiana" )]
		FrenchGuiana,

		/// <summary></summary>
		[Description( "French Polynesia" )]
		FrenchPolynesia,

		/// <summary></summary>
		[Description( "French Southern Territories" )]
		FrenchSouthernTerritories,

		/// <summary></summary>
		Gabon,

		/// <summary></summary>
		Gambia,

		/// <summary></summary>
		Georgia,

		/// <summary></summary>
		Germany,

		/// <summary></summary>
		Ghana,

		/// <summary></summary>
		Gibraltar,

		/// <summary></summary>
		Greece,

		/// <summary></summary>
		Greenland,

		/// <summary></summary>
		Grenada,

		/// <summary></summary>
		Guadeloupe,

		/// <summary></summary>
		Guam,

		/// <summary></summary>
		Guatemala,

		/// <summary></summary>
		Guernsey,

		/// <summary></summary>
		Guinea,

		/// <summary></summary>
		[Description( "Guinea Bissau" )]
		GuineaBissau,

		/// <summary></summary>
		Guyana,

		/// <summary></summary>
		Haiti,

		/// <summary></summary>
		[Description( "Heard Island and McDonald Islands" )]
		HeardIslandAndMcDonaldIslands,

		/// <summary></summary>
		Honduras,

		/// <summary></summary>
		Hungary,

		/// <summary></summary>
		Iceland,

		/// <summary></summary>
		India,

		/// <summary></summary>
		Indonesia,

		/// <summary></summary>
		Iran,

		/// <summary></summary>
		Iraq,

		/// <summary></summary>
		Ireland,

		/// <summary></summary>
		[Description( "Isle of Man" )]
		IsleOfMan,

		/// <summary></summary>
		Israel,

		/// <summary></summary>
		Italy,

		/// <summary></summary>
		Jamaica,

		/// <summary></summary>
		Japan,

		/// <summary></summary>
		Jersey,

		/// <summary></summary>
		Jordan,

		/// <summary></summary>
		Kazakhstan,

		/// <summary></summary>
		Kenya,

		/// <summary></summary>
		Kiribati,

		/// <summary></summary>
		Kosovo,

		/// <summary></summary>
		Kuwait,

		/// <summary></summary>
		Kyrgyzstan,

		/// <summary></summary>
		Laos,

		/// <summary></summary>
		Latvia,

		/// <summary></summary>
		Lebanon,

		/// <summary></summary>
		Lesotho,

		/// <summary></summary>
		Liberia,

		/// <summary></summary>
		Libya,

		/// <summary></summary>
		Liechtenstein,

		/// <summary></summary>
		Lithuania,

		/// <summary></summary>
		Luxembourg,

		/// <summary></summary>
		Macedonia,

		/// <summary></summary>
		Madagascar,

		/// <summary></summary>
		Malawi,

		/// <summary></summary>
		Malaysia,

		/// <summary></summary>
		Maldives,

		/// <summary></summary>
		Mali,

		/// <summary></summary>
		Malta,

		/// <summary></summary>
		[Description( "Marshall Islands" )]
		MarshallIslands,

		/// <summary></summary>
		Martinique,

		/// <summary></summary>
		Mauritania,

		/// <summary></summary>
		Mauritius,

		/// <summary></summary>
		Mayotte,

		/// <summary></summary>
		Mexico,

		/// <summary></summary>
		Micronesia,

		/// <summary></summary>
		Moldova,

		/// <summary></summary>
		Monaco,

		/// <summary></summary>
		Mongolia,

		/// <summary></summary>
		Montenegro,

		/// <summary></summary>
		Montserrat,

		/// <summary></summary>
		Morocco,

		/// <summary></summary>
		Mozambique,

		/// <summary></summary>
		Myanmar,

		/// <summary></summary>
		Namibia,

		/// <summary></summary>
		Nauru,

		/// <summary></summary>
		Nepal,

		/// <summary></summary>
		Netherlands,

		/// <summary></summary>
		[Description( "New Caledonia" )]
		NewCaledonia,

		/// <summary></summary>
		[Description( "New Zealand" )]
		NewZealand,

		/// <summary></summary>
		Nicaragua,

		/// <summary></summary>
		Niger,

		/// <summary></summary>
		Nigeria,

		/// <summary></summary>
		Niue,

		/// <summary></summary>
		[Description( "Norfolk Island" )]
		NorfolkIsland,

		/// <summary></summary>
		[Description( "Northern Mariana Islands" )]
		NorthernMarianaIslands,

		/// <summary></summary>
		Norway,

		/// <summary></summary>
		Oman,

		/// <summary></summary>
		Pakistan,

		/// <summary></summary>
		Palau,

		/// <summary></summary>
		Palestine,

		/// <summary></summary>
		Panama,

		/// <summary></summary>
		[Description( "Papua New Guinea" )]
		PapuaNewGuinea,

		/// <summary></summary>
		Paraguay,

		/// <summary></summary>
		Peru,

		/// <summary></summary>
		Philippines,

		/// <summary></summary>
		[Description( "Pitcairn Islands" )]
		PitcairnIslands,

		/// <summary></summary>
		Poland,

		/// <summary></summary>
		Portugal,

		/// <summary></summary>
		[Description( "Puerto Rico" )]
		PuertoRico,

		/// <summary></summary>
		Qatar,

		/// <summary></summary>
		Reunion,

		/// <summary></summary>
		Romania,

		/// <summary></summary>
		Russia,

		/// <summary></summary>
		Rwanda,

		/// <summary></summary>
		[Description( "Saint Barthelemy" )]
		SaintBarthelemy,

		/// <summary></summary>
		[Description( "Saint Helena" )]
		SaintHelena,

		/// <summary></summary>
		[Description( "Saint Kitts and Nevis" )]
		SaintKittsAndNevis,

		/// <summary></summary>
		[Description( "Saint Lucia" )]
		SaintLucia,

		/// <summary></summary>
		[Description( "Saint Martin" )]
		SaintMartin,

		/// <summary></summary>
		[Description( "Saint Pierre and Miquelon" )]
		SaintPierreAndMiquelon,

		/// <summary></summary>
		[Description( "Saint Vincent and the Grenadines" )]
		SaintVincentAndTheGrenadines,

		/// <summary></summary>
		Samoa,

		/// <summary></summary>
		[Description( "San Marino" )]
		SanMarino,

		/// <summary></summary>
		[Description( "Sao Tome and Principe" )]
		SaoTomeAndPrincipe,

		/// <summary></summary>
		[Description( "Saudi Arabia" )]
		SaudiArabia,

		/// <summary></summary>
		Senegal,

		/// <summary></summary>
		Serbia,

		/// <summary></summary>
		Seychelles,

		/// <summary></summary>
		[Description( "Sierra Leone" )]
		SierraLeone,

		/// <summary></summary>
		Singapore,

		/// <summary></summary>
		Slovakia,

		/// <summary></summary>
		Slovenia,

		/// <summary></summary>
		[Description( "Solomon Islands" )]
		SolomonIslands,

		/// <summary></summary>
		Somalia,

		/// <summary></summary>
		[Description( "South Africa" )]
		SouthAfrica,

		/// <summary></summary>
		[Description( "South Georgia and the South Sandwich Islands" )]
		SouthGeorgiaAndTheSouthSandwichIslands,

		/// <summary></summary>
		[Description( "South Korea" )]
		SouthKorea,

		/// <summary></summary>
		[Description( "South Sudan" )]
		SouthSudan,

		/// <summary></summary>
		Spain,

		/// <summary></summary>
		[Description( "Sri Lanka" )]
		SriLanka,

		/// <summary></summary>
		Sudan,

		/// <summary></summary>
		Suriname,

		/// <summary></summary>
		[Description( "Svalbard and Jan Mayen" )]
		SvalbardAndJanMayen,

		/// <summary></summary>
		Swaziland,

		/// <summary></summary>
		Sweden,

		/// <summary></summary>
		Switzerland,

		/// <summary></summary>
		Syria,

		/// <summary></summary>
		Taiwan,

		/// <summary></summary>
		Tajikistan,

		/// <summary></summary>
		Tanzania,

		/// <summary></summary>
		Thailand,

		/// <summary></summary>
		[Description( "Timor-Leste" )]
		TimorLeste,

		/// <summary></summary>
		Togo,

		/// <summary></summary>
		Tokelau,

		/// <summary></summary>
		Tonga,

		/// <summary></summary>
		[Description( "Trinidad and Tobago" )]
		TrinidadAndTobago,

		/// <summary></summary>
		Tunisia,

		/// <summary></summary>
		Turkey,

		/// <summary></summary>
		Turkmenistan,

		/// <summary></summary>
		[Description( "Turks and Caicos Islands" )]
		TurksAndCaicosIslands,

		/// <summary></summary>
		Tuvalu,

		/// <summary></summary>
		Uganda,

		/// <summary></summary>
		Ukraine,

		/// <summary></summary>
		[Description( "United Arab Emirates" )]
		UnitedArabEmirates,

		/// <summary></summary>
		[Description( "United Kingdom" )]
		UnitedKingdom,

		/// <summary></summary>
		[Description( "United States of America" )]
		USA,

		/// <summary></summary>
		[Description( "United States Minor Outlying Islands" )]
		USA_MinorOutlyingIslands,

		/// <summary></summary>
		Uruguay,

		/// <summary></summary>
		Uzbekistan,

		/// <summary></summary>
		Vanuatu,

		/// <summary></summary>
		Vatican,

		/// <summary></summary>
		Venezuela,

		/// <summary></summary>
		Vietnam,

		/// <summary></summary>
		[Description( "Virgin Islands (British)" )]
		VirginIslandsBritish,

		/// <summary></summary>
		[Description( "Virgin Islands (USA)" )]
		VirginIslandsUSA,

		/// <summary></summary>
		[Description( "Wallis and Futuna" )]
		WallisAndFutuna,

		/// <summary></summary>
		[Description( "Western Sahara" )]
		WesternSahara,

		/// <summary></summary>
		Yemen,

		/// <summary></summary>
		Zambia,

		/// <summary></summary>
		Zimbabwe
	}
}
