////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;

namespace Vela.Kernel
{
	/// <summary>
	/// Exception thrown when a critical error has occurred and the application must be aborted.
	/// </summary>
	public class CriticalException : BaseException
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected CriticalException()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="message">The description of the exception. This may be null or empty.</param>
		/// <param name="exception">The existing exception when the current exception was created. This may be null.</param>
		/// <param name="details">Exception details. This may be null or empty.</param>
		public CriticalException( string message, Exception exception = null, params ExceptionDetail[] details )
			: base( message, exception, details )
		{
		}

		#endregion Constructors
	}

	/// <summary>
	/// Exception thrown when a long operation must be aborted.
	/// </summary>
	public class AbortLongOperationException : BaseException
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <remarks>
		/// <para>Leave this public.</para>
		/// </remarks>
		public AbortLongOperationException()
		{
			DoLog = false;
			ShowToUser = false;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="message">The description of the exception. This may be null or empty.</param>
		/// <param name="exception">The existing exception when the current exception was created. This may be null.</param>
		/// <param name="context">The context of the exception. This may be null.</param>
		/// <param name="details">Exception details. This may be null or empty.</param>
		public AbortLongOperationException( string message, Exception exception = null, object context = null
				, params ExceptionDetail[] details )
			: base( message, exception, details )
		{
			DoLog = false;
			ShowToUser = false;

			Context = context;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private object _Context;
		/// <summary>
		/// The context of the exception. This may be null. READ-ONLY.
		/// </summary>
		public object Context
		{
			get { return _Context; }
			protected set { _Context = value; }
		}

		#endregion Public non-serializable properties
	}

	/// <summary>
	/// Exception thrown when an error occurs while the user interacts with the application.
	/// </summary>
	/// <remarks>
	/// <para>This exception is not saved by the log manager (<see cref="LogManager"/>).</para>
	/// </remarks>
	public class UserInteractionException : BaseException
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected UserInteractionException()
		{
			DoLog = false;
			ShowToUser = true;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="message">The description of the exception. This may be null or empty.</param>
		/// <param name="exception">The existing exception when the current exception was created. This may be null.</param>
		/// <param name="details">Exception details. This may be null or empty.</param>
		public UserInteractionException( string message, Exception exception = null, params ExceptionDetail[] details )
			: base( message, exception, details )
		{
			DoLog = false;
			ShowToUser = true;
		}

		#endregion Constructors
	}

	/// <summary>
	/// Generic validation exception thrown when data fails to pass validation, and the current code flow is in an invalid state.
	/// </summary>
	public class ValidationException : BaseException
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ValidationException()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="message">The description of the exception. This may be null or empty.</param>
		/// <param name="exception">The existing exception when the current exception was created. This may be null.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="details">Exception details. This may be null or empty.</param>
		public ValidationException( string message, Exception exception = null, string errorContext = null
				, params ExceptionDetail[] details )
			: base( message, exception, details )
		{
			ErrorContext = errorContext;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private string _ErrorContext;
		/// <summary>
		/// The context which may help the user to better understand the error. This may be null or empty. READ-ONLY.
		/// </summary>
		public string ErrorContext
		{
			get { return _ErrorContext; }
			protected set { _ErrorContext = value; }
		}

		#endregion Public non-serializable properties
	}

	/// <summary>
	/// Exception thrown when an object is checked if it's initialized and it's not.
	/// </summary>
	public class ObjectNotInitializedException : ValidationException
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ObjectNotInitializedException()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="message">The description of the exception. This may be null or empty.</param>
		/// <param name="exception">The existing exception when the current exception was created. This may be null.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="details">Exception details. This may be null or empty.</param>
		public ObjectNotInitializedException( string message, Exception exception, string errorContext = null
				, params ExceptionDetail[] details )
			: base( message, exception, errorContext, details )
		{
		}

		#endregion Constructors
	}

	/// <summary>
	/// Exception which can cross a service's boundaries, and indicates that an error message must be passed from the service to
	/// the client.
	/// </summary>
	/// <remarks>
	/// <para>All exceptions other than those derived from this exception must not be forwarded to the client because they might
	/// leak exploitable information.</para>
	/// </remarks>
	public abstract class CrossServiceException : BaseException
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected CrossServiceException()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="message">The description of the exception. This may be null or empty.</param>
		/// <param name="exception">The existing exception when the current exception was created. This may be null.</param>
		/// <param name="details">Exception details. This may be null or empty.</param>
		public CrossServiceException( string message, Exception exception, params ExceptionDetail[] details )
			: base( message, exception, details )
		{
		}

		#endregion Constructors
	}

	/// <summary>
	/// Exception which indicates that an error message must be passed from the service to the client. The action client is
	/// strongly recommended not to try sending the action request again to the service (because the error would occur again).
	/// </summary>
	public class FatalCrossServiceException : CrossServiceException
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected FatalCrossServiceException()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="message">The description of the exception. This may be null or empty.</param>
		/// <param name="exception">The existing exception when the current exception was created. This may be null.</param>
		/// <param name="details">Exception details. This may be null or empty.</param>
		public FatalCrossServiceException( string message, Exception exception = null, params ExceptionDetail[] details )
			: base( message, exception, details )
		{
		}

		#endregion Constructors
	}

	/// <summary>
	/// Exception which indicates that an error message must be passed from the service to the client. The action client may
	/// try sending the action request again to the service.
	/// </summary>
	public class RetryCrossServiceException : CrossServiceException
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected RetryCrossServiceException()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="message">The description of the exception. This may be null or empty.</param>
		/// <param name="exception">The existing exception when the current exception was created. This may be null.</param>
		/// <param name="details">Exception details. This may be null or empty.</param>
		public RetryCrossServiceException( string message, Exception exception = null, params ExceptionDetail[] details )
			: base( message, exception, details )
		{
		}

		#endregion Constructors
	}

	/// <summary>
	/// Exception which indicates that a timeout has occurred.
	/// </summary>
	public class TimeoutException : RetryCrossServiceException
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected TimeoutException()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="message">The description of the exception. This may be null or empty.</param>
		/// <param name="exception">The existing exception when the current exception was created. This may be null.</param>
		/// <param name="details">Exception details. This may be null or empty.</param>
		public TimeoutException( string message, Exception exception = null, params ExceptionDetail[] details )
			: base( message, exception, details )
		{
		}

		#endregion Constructors
	}

	/// <summary>
	/// Exception which indicates that an algorithm / document has expired.
	/// </summary>
	public class HasExpiredException : FatalCrossServiceException
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected HasExpiredException()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="message">The description of the exception. This may be null or empty.</param>
		/// <param name="exception">The existing exception when the current exception was created. This may be null.</param>
		/// <param name="details">Exception details. This may be null or empty.</param>
		public HasExpiredException( string message, Exception exception = null, params ExceptionDetail[] details )
			: base( message, exception, details )
		{
		}

		#endregion Constructors
	}

	/// <summary>
	/// Exception thrown when an error has occurred in a background operation.
	/// </summary>
	/// <remarks>
	/// <para>The exceptions which occur in background operations should not be shown to the user, since the user may have not
	/// initiated the operation, so he may be absent. For example, exceptions which occur in the (synchronous) delegate of a
	/// dispatcher are forwarded to the application's global exception handler, which means that an error dialog would be
	/// shown; this may happen when the UI is updated as a result of an action request being received by an action service.
	/// </para>
	/// </remarks>
	public class BackgroundOperationException : BaseException
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected BackgroundOperationException()
		{
			DoLog = true;
			ShowToUser = false;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="exception">The existing exception when the current exception was created. This may be null.</param>
		public BackgroundOperationException( Exception exception )
			: base( null, exception )
		{
			DoLog = true;
			ShowToUser = false;
		}

		#endregion Constructors
	}

	/// <summary>
	/// Exception thrown when (user inputted) data fails to pass validation.
	/// </summary>
	/// <remarks>
	/// <para>This exception is not saved by the log manager (<see cref="LogManager"/>).</para>
	/// 
	/// <para>This exception should be thrown when the user interface is being validated, for example, when a user inputted form
	/// is accepted (by the user).</para>
	/// </remarks>
	public class PresentedValidationException : ValidationException
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected PresentedValidationException()
		{
			DoLog = false;
			ShowToUser = true;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="message">The description of the exception. This may be null or empty.</param>
		/// <param name="exception">The existing exception when the current exception was created. This may be null.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error.</param>
		/// <param name="details">Exception details. This may be null or empty.</param>
		public PresentedValidationException( string message, Exception exception, string errorContext, object errorControl
				, object errorPresenter, params ExceptionDetail[] details )
			: base( message, exception, errorContext, details )
		{
			DoLog = false;
			ShowToUser = true;

			ErrorControl = errorControl;
			ErrorPresenter = errorPresenter;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private object _ErrorControl;
		/// <summary>
		/// The control which caused the exception. READ-ONLY.
		/// </summary>
		public object ErrorControl
		{
			get { return _ErrorControl; }
			protected set
			{
				if( value == null )
					throw new ValidationException( "\"ErrorControl\" is null." );

				_ErrorControl = value;
			}
		}

		private object _ErrorPresenter;
		/// <summary>
		/// The control which must show that there is a validation error. READ-ONLY.
		/// </summary>
		public object ErrorPresenter
		{
			get { return _ErrorPresenter; }
			protected set
			{
				if( value == null )
					throw new ValidationException( "\"ErrorPresenter\" is null." );

				_ErrorPresenter = value;
			}
		}

		#endregion Public non-serializable properties
	}
}
