////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;

namespace Vela.Kernel
{
	/// <summary>
	/// Kernel utilities.
	/// </summary>
	public static class KernelUtils
	{
		#region Public static methods

		/// <summary>
		/// Gets the deep text encoding.
		/// </summary>
		/// <remarks>
		/// <para>This is used when it's (virtually) impossible to store the name of an configuration pool (together with the
		/// data).</para>
		/// 
		/// <para>This should never change so long as <see cref="Vela"/> exists. If it changes, it may become impossible
		/// to access / compare some (stored) information.</para>
		/// </remarks>
		public static TextEncoding GetDeepTextEncoding()
		{
			return new Utf8();
		}

		/// <summary>
		/// Gets the minimum time from the specified times.
		/// </summary>
		/// <param name="times">List of times from which to get the minimum.</param>
		/// <returns>Returns the minimum time.</returns>
		public static DateTime MinTime( params DateTime[] times )
		{
			ValidationUtils.ArrayHasContent( times, "times" );

			DateTime min = DateTime.MaxValue;

			foreach( DateTime time in times )
			{
				if( time < min )
					min = time;
			}

			return min;
		}

		/// <summary>
		/// Verifies if the specified algorithm has expired. An <see cref="HasExpiredException"/> exception is thrown if the
		/// algorithm has expired and "throwIfExpired" is true.
		/// </summary>
		/// <param name="expirableAlgorithm">The algorithm to verify. If this is not an expirable algorithm, the method returns
		/// false. If this is null, the method returns false.</param>
		/// <param name="lifetimeMax">The maximum allowed lifetime (in minutes). If this is 0 or a negative number, it's
		/// ignored.</param>
		/// <param name="throwIfExpired">If this is true and the algorithm has expired then an exception is thrown.</param>
		/// <returns>Returns true if the algorithm has expired and "throwIfExpired" is false.</returns>
		public static bool HasExpired( IExpirableAlgorithm expirableAlgorithm, int lifetimeMax = 0, bool throwIfExpired = true )
		{
			ValidationUtils.ObjectHasContent( expirableAlgorithm, "expirableAlgorithm" );

			// The current time.
			DateTime now = DateTime.UtcNow;

			// Verify the algorithm expiration time.
			if( now >= expirableAlgorithm.ExpirationTime )
			{
				if( !throwIfExpired )
					return true;

				string message = "The document with the type \"{0}\" and with the locator \"{1}\" has expired at time {2}. The"
					+ " current time is {3}.";

				Locator instanceLocator = (expirableAlgorithm as ILocatable != null)
					? ((ILocatable)expirableAlgorithm).InstanceLocator : null;

				throw new HasExpiredException( string.Format( message, expirableAlgorithm.GetType().FullName
					, instanceLocator, expirableAlgorithm.ExpirationTime.ToLocalTime().ToString( KernelConsts.TimeFormat )
					, now.ToLocalTime().ToString( KernelConsts.TimeFormat ) ) );
			}

			if( expirableAlgorithm as IExpirableDocument != null )
			{
				IExpirableDocument expirableDocument = (IExpirableDocument)expirableAlgorithm;

				// Verify the maximum lifetime.
				double actualLifetime = (expirableDocument.DocumentExpirationTime - expirableDocument.CreationTime).TotalMinutes;
				if( (lifetimeMax > 0) && (actualLifetime > lifetimeMax) )
				{
					if( !throwIfExpired )
						return true;

					string message = "The document with the type \"{0}\" and identifier \"{1}\" has a maximum allowed"
						+ " lifetime of {2} minutes. The current lifetime is {3} minutes.";

					throw new HasExpiredException( string.Format( message, expirableDocument.GetType().FullName
						, expirableDocument.InstanceLocator, lifetimeMax, actualLifetime ) );
				}

				// Verify the document expiration time. This is done in case "DocumentExpirationTime" in not included in
				// "ExpirationTime" (which should be).
				if( now >= expirableDocument.DocumentExpirationTime )
				{
					if( !throwIfExpired )
						return true;

					string message = "The document with the type \"{0}\" and identifier \"{1}\" has expired at time {2}."
						+ " The current time is {3}.";

					throw new HasExpiredException( string.Format( message, expirableDocument.GetType().FullName
						, expirableDocument.InstanceLocator
						, expirableDocument.DocumentExpirationTime.ToLocalTime().ToString( KernelConsts.TimeFormat )
						, now.ToLocalTime().ToString( KernelConsts.TimeFormat ) ) );
				}
			}

			// Not expired.
			return false;
		}

		#endregion Public static methods
	}
}
