////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;

namespace Vela.Kernel
{
	/// <summary>
	/// Contract for a sorted list with a limited number of items.
	/// </summary>
	public abstract class KernelDictionaryWithLimit<TKey, TItem> : KernelDictionary<TKey, TItem>
		where TItem : class
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="comparer">The equality comparer used to compare keys.</param>
		protected KernelDictionaryWithLimit( IEqualityComparer<TKey> comparer )
			: base( comparer )
		{
		}

		#endregion Constructors

		#region Private methods

		/// <summary>
		/// Gets the age (in the current list) of the specified item.
		/// </summary>
		/// <param name="item">The item whose age to get. Validation is not performed.</param>
		/// <returns>Returns the age of the specified item.</returns>
		/// <remarks>
		/// <para>This is used by the <see cref="MakeRoomWhenLimitExceeded"/> method.</para>
		/// </remarks>
		protected abstract DateTime GetItemAge( TItem item );

		/// <summary>
		/// When the <see cref="AddOrUpdateWithLimit"/> method is called and the specified limit is exceeded, this method is
		/// called in order to move the newest existing items into the specified list of new items.
		/// </summary>
		/// <param name="newItems">List of items where to add the existing items which have to be preserved.</param>
		/// <remarks>
		/// <para>This method will make room in the list of items, by moving (approximately) half of the existing items to keep
		/// into a new list.</para>
		/// </remarks>
		protected virtual void MakeRoomWhenLimitExceeded( Dictionary<TKey, TItem> newItems )
		{
			// Find the minimum and maximum age.
			DateTime minAge = DateTime.MaxValue, maxAge = DateTime.MinValue;
			foreach( TItem item in this )
			{
				DateTime itemAge = GetItemAge( item );

				if( itemAge > maxAge )
					maxAge = itemAge;
				if( itemAge < minAge )
					minAge = itemAge;
			}

			// Approximate the time which splits the list in half.
			DateTime keepFromTime = minAge.AddTicks( (maxAge.Ticks - minAge.Ticks) / 2 );

			foreach( TItem item in this )
			{
				if( GetItemAge( item ) > keepFromTime )
					newItems.Add( GetKey( item ), item );
			}
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Adds or updates an existing item in the list, regardless of the value of the
		/// <see cref="KernelKeyedList{ TKey, TItem }.UpdateIfSameKey"/> property. If the list doesn't contain an item with the
		/// same key, it's added; if the list already contains "limit" items, (approximately) half of the existing items are
		/// removed from the list before the specified item is added. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="item">Item to update in the list.</param>
		/// <param name="limit">Limit from which items are removed from the list before an item is added. This must be at least
		/// 1.</param>
		/// <remarks>
		/// <para>This method will make room in the list of items, by moving (approximately) half of the existing items to keep
		/// into a new list.</para>
		/// </remarks>
		public virtual void AddOrUpdateWithLimit( TItem item, int limit )
		{
			ValidationUtils.Integer( limit, 1, int.MaxValue, "limit" );

			// If the list is full, make room.
			if( Count >= limit )
			{
				Dictionary<TKey, TItem> newItems = new Dictionary<TKey, TItem>( Comparer );

				MakeRoomWhenLimitExceeded( newItems );

				Items = newItems;

				// If there is still no room, remove the first item.
				if( Count >= limit )
					RemoveItem( GetFirstItem() );

				// Notify property and list observers.
				NotifyRefreshList();
			}

			AddOrUpdate( item );
		}

		#endregion Public methods
	}
}
