////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections;
using System.Collections.Generic;

namespace Vela.Kernel
{
	/// <summary>
	/// Contract for a keyed list, that is, for a list in which items can be quickly found based on their key
	/// (<see cref="GetKey"/>), with the items stored in a <see cref="IDictionary"/>.
	/// </summary>
	/// <remarks>
	/// <para>Avoid using unsorted dictionaries in (signed) documents because the order of the items is not guaranteed.</para>
	/// 
	/// <para>The key must be obtainable from the item, with the <see cref="GetKey"/> method.</para>
	/// 
	/// <para>Inheritors must provide an equality comparer for keys.</para>
	/// 
	/// <para>If the list already contains an item with the same key and <see cref="UpdateIfSameKey"/> is false (which is by
	/// default), an exception is thrown.</para>
	/// 
	/// <para>Null items can't be added to the list.</para>
	/// 
	/// <para>If you add an item to the list, then modify the properties with which the item's key is created, then update the
	/// item in the list, the updated item would actually be a new item with a new key.</para>
	/// 
	/// <para>You should <b>not</b> modify the properties of an item from which its key is created, after the item was added to
	/// the list. This is because in some cases, the item's key may be used from the key list, but sometimes it may be
	/// dynamically generated).</para>
	/// 
	/// <para>If an event is attached to <see cref="KernelObservableList{TItem}.CollectionChanged"/>, the performance (for add,
	/// update, remove) is dramatically affected; this happens for example when the list is bound to an UI control.</para>
	/// </remarks>
	public abstract class KernelKeyedList<TKey, TItem> : KernelObservableList<TItem>
		where TItem : class
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected KernelKeyedList()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="comparer">The equality comparer used to compare keys.</param>
		protected KernelKeyedList( IEqualityComparer<TKey> comparer )
		{
			ValidationUtils.ObjectHasContent( comparer, "comparer" );

			Comparer = comparer;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The equality comparer used to compare keys.
		/// </summary>
		protected IEqualityComparer<TKey> Comparer;

		/// <summary>
		/// List of items.
		/// </summary>
		protected IDictionary<TKey, TItem> Items;

		#endregion Private fields

		#region Indexers

		/// <summary>
		/// Gets the item with the specified key in the list. READ-ONLY.
		/// </summary>
		/// <param name="key">The key of the item to get.</param>
		/// <returns>Returns the item with the specified key, or null if the item is not found.</returns>
		public virtual TItem this[ TKey key ]
		{
			get
			{
				TItem item;

				return Items.TryGetValue( key, out item ) ? item : null;
			}
		}

		#endregion Indexers

		#region Public non-serializable properties

		/// <summary>
		/// The number of items from the list. READ-ONLY.
		/// </summary>
		public override int Count
		{
			get { return Items.Count; }
		}

		private bool _UpdateIfSameKey = false;
		/// <summary>
		/// When an item is added to the list and the list already contains an item with the same key, if this is true then the
		/// existing item is replaced, else an exception is thrown.
		/// </summary>
		public bool UpdateIfSameKey
		{
			get { return _UpdateIfSameKey; }
			set { _UpdateIfSameKey = value; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Inheritors must return the exception which is to be thrown if an attempt is made to add an item which has a key
		/// which already exists in the list.
		/// </summary>
		/// <param name="key">Key which already exists in the list.</param>
		/// <param name="item">Item whose key to already exists in the list.</param>
		protected virtual BaseException GetAlreadyExistingKeyException( TKey key, TItem item )
		{
			return new ValidationException( string.Format( "An item with the key \"{0}\" already exists in the list.", key ) );
		}

		/// <summary>
		/// Validates the specified item before adding or updating it to the list. Inheritors must throw an exception if the
		/// item must not be added to the list.
		/// </summary>
		/// <param name="item">Item to add to the list. There is no need to check if this is null because this is performed
		/// before the method is called.</param>
		/// <remarks>
		/// <para>Don't check for duplicate items here because this is automatically done.</para>
		/// </remarks>
		protected virtual void ValidateAddOrUpdate( TItem item )
		{
		}

		/// <summary>
		/// Validates the specified item before updating in the list. Inheritors must throw an exception if the item must not be
		/// updated.
		/// </summary>
		/// <param name="existingItem">The item from the list, which has the same key as the new item. There is no need to check
		/// if this is null because this is performed before the method is called.</param>
		/// <param name="newItem">The item which is to replace the existing item. There is no need to check if this is null
		/// because this is performed before the method is called.</param>
		protected virtual void ValidateUpdate( TItem existingItem, TItem newItem )
		{
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Returns the enumerator of the items of the list.
		/// </summary>
		/// <returns>Returns the enumerator of the items of the list.</returns>
		public override IEnumerator GetEnumerator()
		{
			return Items.Values.GetEnumerator();
		}

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para>The returned key must not depend on the current object, but only on the specified item. This must not be null
		/// or empty.</para>
		/// </remarks>
		public abstract TKey GetKey( TItem item );

		/// <summary>
		/// Checks if the list contains an item with the specified key.
		/// </summary>
		/// <param name="key">Key to search for.</param>
		/// <returns>Returns true if the list contains an item with the specified key.</returns>
		public virtual bool ContainsKey( TKey key )
		{
			return Items.ContainsKey( key );
		}

		/// <summary>
		/// Checks if the list contains an item with the key of the specified item.
		/// </summary>
		/// <param name="item">Item whose key to search for.</param>
		/// <returns>Returns true if the list contains an item with the key of the specified item.</returns>
		public virtual bool ContainsItem( TItem item )
		{
			return Items.ContainsKey( GetKey( item ) );
		}

		/// <summary>
		/// Finds an item with the specified key.
		/// </summary>
		/// <param name="key">Key to search for.</param>
		/// <returns>Returns the found item, or null if none is found.</returns>
		public virtual TItem Find( TKey key )
		{
			return this[ key ];
		}

		/// <summary>
		/// Finds an item with the same key as the specified item.
		/// </summary>
		/// <param name="item">Item whose key to search for.</param>
		/// <returns>Returns the found item, or null if none is found.</returns>
		public virtual TItem FindByItem( TItem item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return this[ GetKey( item ) ];
		}

		/// <summary>
		/// Finds the item immediately after the specified item.
		/// </summary>
		/// <param name="item">The item after which to find the next item.</param>
		/// <param name="getFirstIfNoNext">If this is true and there is no next item, the first one is returned.</param>
		/// <returns>Returns the found item, or the first / previous / last item (depending on "getFirstIfNoNext") if there is no
		/// next item (which allows circular movement), or null if the list is empty or the next item was not yet set.</returns>
		public virtual TItem FindNext( TItem item, bool getFirstIfNoNext )
		{
			TKey itemKey = GetKey( item );
			TItem firstItem = null;
			TItem nextItem = null;

			bool foundItem = false;
			foreach( TItem currentItem in this )
			{
				if( firstItem == null )
					firstItem = currentItem;

				// Check if the specified item was found during the previous cycle.
				if( foundItem )
					return currentItem;

				if( Comparer.Equals( GetKey( currentItem ), itemKey ) )
					foundItem = true;
				else
					nextItem = currentItem;
			}

			return getFirstIfNoNext ? firstItem : nextItem;
		}

		/// <summary>
		/// Finds the item immediately before the specified item.
		/// </summary>
		/// <param name="item">The item before which to find the previous item.</param>
		/// <param name="getLastIfNoPrevious">If this is true and there is no previous item, the last one is returned.</param>
		/// <returns>Returns the found item, or the last / previous item (depending on "getLastIfNoPrevious") if there is no
		/// previous item (which allows circular movement), or null if the list is empty or the previous item was not yet set.
		/// </returns>
		public virtual TItem FindPrevious( TItem item, bool getLastIfNoPrevious )
		{
			TKey itemKey = GetKey( item );
			TItem lastItem = null;
			TItem previousItem = null;

			foreach( TItem currentItem in this )
			{
				if( Comparer.Equals( GetKey( currentItem ), itemKey ) )
				{
					return ((previousItem == null) && getLastIfNoPrevious) ? GetLastItem() : previousItem;
				}
				else
				{
					previousItem = currentItem;
				}
			}

			return getLastIfNoPrevious ? lastItem : previousItem;
		}

		/// <summary>
		/// Finds an item with the same key as one of the specified items.
		/// </summary>
		/// <param name="items">Items whose keys to search for. This may be empty, but not null.</param>
		/// <returns>Returns the found item, or null if none is found.</returns>
		public virtual TItem FindAny( KernelObservableList<TItem> items )
		{
			ValidationUtils.ObjectHasContent( items, "items" );

			foreach( TItem item in items )
			{
				TItem foundItem = this[ GetKey( item ) ];

				if( foundItem != null )
					return foundItem;
			}

			return null;
		}

		/// <summary>
		/// Gets an array with all the items.
		/// </summary>
		/// <returns>Returns an array with all the items.</returns>
		public TItem[] GetItems()
		{
			int size = Items.Count;
			TItem[] items = new TItem[ Items.Count ];

			int i = 0;
			foreach( TItem currentItem in this )
			{
				items[ i ] = currentItem;
				i++;
			}

			return items;
		}

		/// <summary>
		/// Gets the first item from the list.
		/// </summary>
		/// <returns>Returns the first item from the list, or null if the list is empty.</returns>
		public TItem GetFirstItem()
		{
			if( Count <= 0 )
				return null;

			IEnumerator enumerator = GetEnumerator();
			enumerator.Reset();
			enumerator.MoveNext();

			return enumerator.Current as TItem;
		}

		/// <summary>
		/// Gets the last item from the list.
		/// </summary>
		/// <returns>Returns the last item from the list, or null if the list is empty.</returns>
		public TItem GetLastItem()
		{
			if( Count <= 0 )
				return null;

			IEnumerator enumerator = GetEnumerator();
			enumerator.Reset();

			TItem lastItem = null;
			while( enumerator.MoveNext() )
				lastItem = enumerator.Current as TItem;

			return lastItem;
		}

		/// <summary>
		/// Adds a new item to the list. If the list already contains an item with the same key and "updateIfSameKey" is false,
		/// an exception is thrown. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="item">Item to add to the list.</param>
		/// <param name="updateIfSameKey">When an item is added to the list and the list already contains an item with the same
		/// key, if this is true then the existing item is replaced, else an exception is thrown.</param>
		public virtual void Add( TItem item, bool updateIfSameKey )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			// Validate the adding.
			ValidateAddOrUpdate( item );

			TKey key = GetKey( item );
			bool alreadyContains = Items.ContainsKey( key );

			if( updateIfSameKey )
			{
				if( alreadyContains )
				{
					// Get the existing item.
					TItem existingItem = Items[ key ];

					// Validate the updating.
					ValidateUpdate( existingItem, item );

					if( HasCollectionChanged && !FrozenCollectionChangedNotifications )
					{
						// Dictionaries have no item index, so find the index.
						int existingIndex = FindIndex( existingItem );
						if( existingIndex < 0 )
						{
							throw new ValidationException( string.Format( "The item with the key \"{0}\" is missing even tough"
								+ " it was just found.", key ) );
						}

						// Replace the existing item.
						Items[ key ] = item;

						// Dictionaries have no item index, so find the index.
						int newIndex = FindIndex( item );
						if( newIndex < 0 )
						{
							throw new ValidationException( string.Format( "The item with the key \"{0}\" is missing even tough"
								+ " it was just added.", key ) );
						}

						// Notify list observers.
						OnItemUpdated( existingItem, item, existingIndex, newIndex );
					}
					else
					{
						// Replace the existing item.
						Items[ key ] = item;
					}
				}
				else
				{
					// Add the new item.
					Items.Add( key, item );

					if( HasCollectionChanged && !FrozenCollectionChangedNotifications )
					{
						// Dictionaries have no item index, so find the index.
						int newIndex = FindIndex( item );
						if( newIndex < 0 )
						{
							throw new ValidationException( string.Format( "The item with the key \"{0}\" is missing even tough"
								+ " it was just added.", key ) );
						}

						// Notify list observers.
						OnItemAdded( item, newIndex );
					}

					// Notify property observers.
					OnPropertyChanged( CountPropertyName );
				}
			}
			else
			{
				// Add the new item or throw an exception if an item with the same key already exists.

				if( alreadyContains )
					throw GetAlreadyExistingKeyException( key, item );
				else
				{
					// Add the new item.
					Items.Add( key, item );

					if( HasCollectionChanged && !FrozenCollectionChangedNotifications )
					{
						// Dictionaries have no item index, so find the index.
						int newIndex = FindIndex( item );
						if( newIndex < 0 )
						{
							throw new ValidationException( string.Format( "The item with the key \"{0}\" is missing even tough"
								+ " it was just added.", key ) );
						}

						// Notify list observers.
						OnItemAdded( item, newIndex );
					}

					// Notify property observers.
					OnPropertyChanged( CountPropertyName );
				}
			}
		}

		/// <summary>
		/// Adds a new item to the list. If the list already contains an item with the same key and
		/// <see cref="UpdateIfSameKey"/> is false, an exception is thrown. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="item">Item to add to the list.</param>
		public void Add( TItem item )
		{
			Add( item, UpdateIfSameKey );
		}

		/// <summary>
		/// Adds a list of new items to the list. If the list already contains an item with the same key and "updateIfSameKey"
		/// is false, an exception is thrown. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="items">Items to add to the list. This may be empty, but not null.</param>
		/// <param name="updateIfSameKey">When an item is added to the list and the list already contains an item with the same
		/// key, if this is true then the existing item is replaced, else an exception is thrown.</param>
		public void Add( KernelObservableList<TItem> items, bool updateIfSameKey )
		{
			ValidationUtils.ObjectHasContent( items, "items" );

			foreach( TItem item in items )
				Add( item, updateIfSameKey );
		}

		/// <summary>
		/// Adds a list of new items to the list. If the list already contains an item with the same key and
		/// <see cref="UpdateIfSameKey"/> is false, an exception is thrown. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="items">Items to add to the list. This may be empty, but not null.</param>
		public void Add( KernelObservableList<TItem> items )
		{
			Add( items, UpdateIfSameKey );
		}

		/// <summary>
		/// Adds a new item to the list. If the list already contains an item with the same key and
		/// <see cref="UpdateIfSameKey"/> is false, an exception is thrown. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="item">Item to add to the list.</param>
		public override void Add( object item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			Type itemType = item.GetType();
			Type typedItemType = typeof( TItem );

			if( !typedItemType.IsAssignableFrom( itemType ) )
			{
				throw new ValidationException( string.Format( "The type of the item to add to the list is \"{0}\""
					+ " instead of (one derived from) \"{1}\".", itemType.FullName, typedItemType.FullName ) );
			}

			Add( (TItem)item );
		}

		/// <summary>
		/// Adds a new item to the list. If the list already contains an item with the same key, nothing happens. Notifies
		/// property observers. Notifies list observers.
		/// </summary>
		/// <param name="item">Item to add to the list.</param>
		/// <returns>Returns the existing item (the one with the same key), or null if there is no such item.</returns>
		public TItem AddOnlyNew( TItem item )
		{
			TItem existingItem = this[ GetKey( item ) ];
			if( existingItem != null )
				return existingItem;

			Add( item, false );

			return null;
		}

		/// <summary>
		/// Adds a list of new items to the list. If the list already contains an item with the same key, nothing happens.
		/// Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="items">Items to add to the list. This may be empty, but not null.</param>
		public void AddOnlyNew( KernelObservableList<TItem> items )
		{
			ValidationUtils.ObjectHasContent( items, "items" );

			foreach( TItem item in items )
				AddOnlyNew( item );
		}

		/// <summary>
		/// Adds or updates an existing item in the list, regardless of the value of the <see cref="UpdateIfSameKey"/> property.
		/// If the list doesn't contain an item with the same key, it's added. Notifies property observers. Notifies list
		/// observers.
		/// </summary>
		/// <param name="item">Item to update in the list.</param>
		public virtual void AddOrUpdate( TItem item )
		{
			Add( item, true );
		}

		/// <summary>
		/// Replaces an existing item in the list, regardless of the value of the <see cref="UpdateIfSameKey"/> property. If the
		/// list doesn't contain an item with the same key as the new item, it's added. Notifies property observers. Notifies
		/// list observers.
		/// </summary>
		/// <param name="oldItem">Item to be replaced in the list. This may be null.</param>
		/// <param name="newItem">Item with which to do the replacing in the list.</param>
		/// <remarks>
		/// <para>The old item is first removed from the list, then the new item is added.</para>
		/// 
		/// <para>The old and the new items may have different keys.</para>
		/// </remarks>
		public void AddOrReplace( TItem oldItem, TItem newItem )
		{
			// Remove the existing old item.
			RemoveItem( oldItem );

			// Add the new item.
			Add( newItem );
		}

		/// <summary>
		/// Merges the specified list with the current list. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="list">List whose items to merge. This may be empty, but not null.</param>
		/// <remarks>
		/// <para>The items which have the same key are not overwritten.</para>
		/// </remarks>
		public void Merge( KernelKeyedList<TKey, TItem> list )
		{
			ValidationUtils.ObjectHasContent( list, "list" );

			foreach( TItem item in list )
			{
				if( Find( GetKey( item ) ) == null )
					Add( item );
			}
		}

		/// <summary>
		/// Removes the item with the specified key from the list. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="key">The key of the item to remove. If an item with this key is not found, nothing happens.</param>
		/// <returns>Returns the removed item, or null if none was removed.</returns>
		public virtual TItem Remove( TKey key )
		{
			ValidationUtils.ObjectHasContent( key, "key" );

			if( !Items.ContainsKey( key ) )
				return null;

			// Get the existing item.
			TItem existingItem = Items[ key ];

			if( HasCollectionChanged && !FrozenCollectionChangedNotifications )
			{
				// Dictionaries have no item index, so find the index.
				int existingIndex = FindIndex( existingItem );
				if( existingIndex < 0 )
				{
					throw new ValidationException( string.Format( "The item with the key \"{0}\" is missing even tough it was"
						+ " just found.", key ) );
				}

				// Remove the item.
				Items.Remove( key );

				// Notify list observers.
				OnItemRemoved( existingItem, existingIndex );
			}
			else
			{
				// Remove the item.
				Items.Remove( key );
			}

			// Notify property observers.
			OnPropertyChanged( CountPropertyName );

			return existingItem;
		}

		/// <summary>
		/// Removes the item with the key of the specified item from the list. Notifies property observers. Notifies list
		/// observers.
		/// </summary>
		/// <param name="item">Item to remove from the list. If an item with the key of the specified item is not found, nothing
		/// happens. This may be null.</param>
		/// <returns>Returns the removed item, or null if none was removed.</returns>
		public TItem RemoveItem( TItem item )
		{
			if( item == null )
				return null;

			return Remove( GetKey( item ) );
		}

		/// <summary>
		/// Removes all the items from the list. Notifies property observers. Notifies list observers.
		/// </summary>
		public virtual void Clear()
		{
			Items.Clear();

			// Notify property and list observers.
			NotifyRefreshList();
		}

		#endregion Public methods
	}
}
