////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections;
using System.Collections.Generic;

namespace Vela.Kernel
{
	/// <summary>
	/// Contract for a list.
	/// </summary>
	/// <remarks>
	/// <para>This list doesn't have the concept of item updating.</para>
	/// 
	/// <para>For limiting the size of the list, see <see cref="CountLimit"/>.</para>
	/// </remarks>
	public abstract class KernelList<TItem> : KernelObservableList<TItem>
		where TItem : class
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected KernelList()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class and adds a list of new items to the list. No duplicate verification
		/// is made; inheritors should use the <see cref="ValidateAdd"/> method to verify duplicates. Notifies property
		/// observers. Notifies list observers.
		/// </summary>
		/// <param name="items">Items to add to the list. This may be empty, but not null.</param>
		public KernelList( KernelObservableList<TItem> items )
			: base( items )
		{
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// List of items.
		/// </summary>
		protected List<TItem> Items = new List<TItem>();

		/// <summary>
		/// The maximum items that the list may have. If this is equal with or smaller than 1, it's ignored.
		/// </summary>
		protected virtual int CountLimit
		{
			get { return 0; }
		}

		#endregion Private fields

		#region Indexers

		/// <summary>
		/// Gets the item with the specified index in the list. READ-ONLY.
		/// </summary>
		/// <param name="index">The index of the item. This must be at least 0.</param>
		/// <returns>Returns the item with the specified index.</returns>
		public TItem this[ int index ]
		{
			get { return Items[ index ]; }
		}

		#endregion Indexers

		#region Public non-serializable properties

		/// <summary>
		/// The number of items from the list. READ-ONLY.
		/// </summary>
		public override int Count
		{
			get { return Items.Count; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Makes room in the list of items by removing half of the items (from the beginning of the list).
		/// </summary>
		protected void MakeRoomWhenLimitExceeded()
		{
			if( (CountLimit <= 1) || (Count < CountLimit) )
				return;

			// If the list is full, make room.
			Items.RemoveRange( 0, Count / 2 );

			// Notify property and list observers.
			NotifyRefreshList();
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Returns the enumerator of the list.
		/// </summary>
		/// <returns>Returns the enumerator of the list.</returns>
		public override IEnumerator GetEnumerator()
		{
			return Items.GetEnumerator();
		}

		/// <summary>
		/// Finds the item immediately after the specified item.
		/// </summary>
		/// <param name="item">The item after which to find the next item.</param>
		/// <param name="getFirstIfNoNext">If this is true and there is no next item, the first one is returned.</param>
		/// <returns>Returns the found item, or the first / previous item (depending on "getFirstIfNoNext") if "item" is the last
		/// item in the list, or the first / last item (depending on "getFirstIfNoNext") if there is no next item, or null if the
		/// list is empty.</returns>
		/// <remarks>
		/// <para>Object equality is determined by reference, not content.</para>
		/// </remarks>
		public virtual TItem FindNext( TItem item, bool getFirstIfNoNext )
		{
			TItem firstItem = null;
			TItem nextItem = null;

			bool foundItem = false;
			foreach( TItem currentItem in this )
			{
				if( firstItem == null )
					firstItem = currentItem;

				// Check if the specified item was found during the previous cycle.
				if( foundItem )
					return currentItem;

				if( currentItem == item )
					foundItem = true;
				else
					nextItem = currentItem;
			}

			return getFirstIfNoNext ? firstItem : nextItem;
		}

		/// <summary>
		/// Validates the specified item before adding it to the list. Inheritors must throw an exception if the item must not
		/// be added to the list.
		/// </summary>
		/// <param name="item">Item to add to the list. There is no need to check if this is null because this is performed
		/// before the method is called.</param>
		public virtual void ValidateAdd( TItem item )
		{
		}

		/// <summary>
		/// Adds a new item to the list. No duplicate verification is made; inheritors should use the <see cref="ValidateAdd"/>
		/// method to verify duplicates. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="item">Item to add to the list.</param>
		/// <remarks>
		/// <para>If the number of items is at least <see cref="CountLimit"/>, this method first removes half of the items (from
		/// the beginning of the list).</para>
		/// </remarks>
		public virtual void Add( TItem item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			ValidateAdd( item );

			MakeRoomWhenLimitExceeded();

			Items.Add( item );

			// Notify property observers.
			OnPropertyChanged( CountPropertyName );

			// Notify list observers.
			OnItemAdded( item, Count - 1 );
		}

		/// <summary>
		/// Adds a list of new items to the list. No duplicate verification is made; inheritors should use the
		/// <see cref="ValidateAdd"/> method to verify duplicates.
		/// </summary>
		/// <param name="items">Items to add to the list. This may be empty, but not null.</param>
		public void Add( KernelObservableList<TItem> items )
		{
			ValidationUtils.ObjectHasContent( items, "items" );

			foreach( TItem item in items )
				Add( item );
		}

		/// <summary>
		/// Adds a new item to the list. No duplicate verification is made; inheritors should use the <see cref="ValidateAdd"/>
		/// method to verify duplicates. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="item">Item to add to the list.</param>
		public override void Add( object item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			Type itemType = item.GetType();
			Type typedItemType = typeof( TItem );

			if( !typedItemType.IsAssignableFrom( itemType ) )
			{
				throw new ValidationException( string.Format( "The type of the item to add to the list is \"{0}\""
					+ " instead of (one derived from) \"{1}\".", itemType.FullName, typedItemType.FullName ) );
			}

			Add( (TItem)item );
		}

		/// <summary>
		/// Replaces an existing item in the list. If the list doesn't contain the old item, the new one is added. Notifies
		/// property observers. Notifies list observers.
		/// </summary>
		/// <param name="oldItem">Item to be replaced in the list. This may be null.</param>
		/// <param name="newItem">Item with which to do the replacing in the list.</param>
		/// <remarks>
		/// <para>The old item is replaced in place, that is, it's not removed before the new item is added.</para>
		/// 
		/// <para>The old and the new items may have different keys.</para>
		/// </remarks>
		public void AddOrReplace( TItem oldItem, TItem newItem )
		{
			int existingIndex = -1;

			if( oldItem != null )
				existingIndex = FindIndex( oldItem );

			if( existingIndex >= 0 )
			{
				// Replace the existing item.

				ValidateAdd( newItem );

				Items[ existingIndex ] = newItem;

				// Notify property observers.
				OnPropertyChanged( CountPropertyName );

				// Notify list observers.
				OnItemUpdated( oldItem, newItem, existingIndex, existingIndex );
			}
			else
			{
				// Add the new item.
				Add( newItem );
			}
		}

		/// <summary>
		/// Removes the item with the specified index from the list. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="index">The index of the item to remove from the list. This must be at least 0.</param>
		public virtual void RemoveAt( int index )
		{
			ValidationUtils.Integer( index, 0, Count - 1, "index" );

			TItem itemToRemove = Items[ index ];

			Items.RemoveAt( index );

			// Notify property observers.
			OnPropertyChanged( CountPropertyName );

			// Notify list observers.
			OnItemRemoved( itemToRemove, index );
		}

		/// <summary>
		/// Removes the specified item from the list. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="item">Item to remove from the list. If the item is not found, nothing happens. This may be null.</param>
		public void RemoveItem( TItem item )
		{
			if( item == null )
				return;

			int existingIndex = FindIndex( item );

			// Remove the item.
			if( existingIndex >= 0 )
			{
				Items.RemoveAt( existingIndex );

				// Notify list observers.
				OnItemRemoved( item, existingIndex );
			}

			// Notify property observers.
			OnPropertyChanged( CountPropertyName );
		}

		/// <summary>
		/// Removes all the items from the list. Notifies property observers. Notifies list observers.
		/// </summary>
		public virtual void Clear()
		{
			Items.Clear();

			// Notify property and list observers.
			NotifyRefreshList();
		}

		#endregion Public methods
	}
}
