////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections;
using System.Collections.Specialized;

namespace Vela.Kernel
{
	/// <summary>
	/// Contract for an observable list.
	/// </summary>
	public abstract class KernelObservableList<TItem> : NotifyPropertyChanged, INotifyCollectionChanged, IKernelObservableList
		where TItem : class
	{
		#region Public constants

		/// <summary>
		/// The name of the "Count" property.
		/// </summary>
		public const string CountPropertyName = "Count";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected KernelObservableList()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class and adds a list of new items to the list. Notifies property
		/// observers. Notifies list observers.
		/// </summary>
		/// <param name="items">Items to add to the list. This may be empty, but not null.</param>
		public KernelObservableList( KernelObservableList<TItem> items )
		{
			ValidationUtils.ObjectHasContent( items, "items" );

			foreach( TItem item in items )
				Add( item );
		}

		#endregion Constructors

		#region Events

		/// <summary>
		/// Event raised when the list changes (that is, when an item is added or removed). This may be null.
		/// </summary>
		public event NotifyCollectionChangedEventHandler CollectionChanged;

		#endregion Events

		#region Private fields

		/// <summary>
		/// Is true if the list has a <see cref="CollectionChanged"/> event. READ-ONLY.
		/// </summary>
		protected bool HasCollectionChanged
		{
			get { return CollectionChanged != null; }
		}

		#endregion Private fields

		#region Public non-serializable properties

		/// <summary>
		/// The declared type of the items from the list. READ-ONLY.
		/// </summary>
		public Type ItemDeclaredType
		{
			get { return typeof( TItem ); }
		}

		/// <summary>
		/// The number of items from the list. READ-ONLY.
		/// </summary>
		public abstract int Count { get; }

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Notifies list observers when the list was reset.
		/// </summary>
		protected virtual void OnListReset()
		{
			if( !HasCollectionChanged || FrozenCollectionChangedNotifications )
				return;

			NotifyCollectionChangedEventArgs eventArgs = new NotifyCollectionChangedEventArgs(
				NotifyCollectionChangedAction.Reset );

			CollectionChanged( this, eventArgs );
		}

		/// <summary>
		/// Notifies list observers when an item was added to the list.
		/// </summary>
		/// <param name="addedItem">The added item. Validation is not performed.</param>
		/// <param name="addedItemIndex">The index in the list of the added item. Validation is not performed.</param>
		protected virtual void OnItemAdded( TItem addedItem, int addedItemIndex )
		{
			if( !HasCollectionChanged || FrozenCollectionChangedNotifications )
				return;

			NotifyCollectionChangedEventArgs eventArgs = new NotifyCollectionChangedEventArgs(
				NotifyCollectionChangedAction.Add, addedItem, addedItemIndex );

			CollectionChanged( this, eventArgs );
		}

		/// <summary>
		/// Notifies list observers when an item was updated in the list.
		/// </summary>
		/// <param name="oldItem">The item which was in the list before the update. Validation is not performed.</param>
		/// <param name="newItem">The updated item. Validation is not performed.</param>
		/// <param name="oldItemIndex">The index in the list of the item before it was updated. Validation is not performed.
		/// </param>
		/// <param name="newItemIndex">The index in the list of the item after it was updated. Validation is not performed.
		/// </param>
		protected virtual void OnItemUpdated( TItem oldItem, TItem newItem, int oldItemIndex, int newItemIndex )
		{
			if( !HasCollectionChanged || FrozenCollectionChangedNotifications )
				return;

			NotifyCollectionChangedEventArgs eventArgs;

			if( oldItemIndex != newItemIndex )
			{
				eventArgs = new NotifyCollectionChangedEventArgs( NotifyCollectionChangedAction.Move, newItem, newItemIndex
					, oldItemIndex );
			}
			else
			{
				eventArgs = new NotifyCollectionChangedEventArgs( NotifyCollectionChangedAction.Replace, newItem, oldItem
					, oldItemIndex );
			}

			CollectionChanged( this, eventArgs );
		}

		/// <summary>
		/// Notifies list observers when an item was removed from the list.
		/// </summary>
		/// <param name="removedItem">The removed item. Validation is not performed.</param>
		/// <param name="removedItemIndex">The index in the list of the removed item. Validation is not performed.</param>
		protected virtual void OnItemRemoved( TItem removedItem, int removedItemIndex )
		{
			if( !HasCollectionChanged || FrozenCollectionChangedNotifications )
				return;

			NotifyCollectionChangedEventArgs eventArgs = new NotifyCollectionChangedEventArgs(
				NotifyCollectionChangedAction.Remove, removedItem, removedItemIndex );

			CollectionChanged( this, eventArgs );
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Notifies the observers that the list UI has to be refreshed. Notifies property observers. Notifies list observers.
		/// </summary>
		public void NotifyRefreshList()
		{
			// Notify property observers.
			OnPropertyChanged( CountPropertyName );

			// Notify list observers.
			OnListReset();
		}

		/// <summary>
		/// Freezes or unfreezes the notification events.
		/// </summary>
		/// <param name="freeze">If this is true then the notification events are not raised while the content of the current
		/// object is changed.</param>
		public override void FreezeCollectionChangedNotifications( bool freeze )
		{
			base.FreezeCollectionChangedNotifications( freeze );

			if( !FrozenCollectionChangedNotifications )
				NotifyRefreshList();
		}

		/// <summary>
		/// Returns the enumerator of the items of the list.
		/// </summary>
		/// <returns>Returns the enumerator of the items of the list.</returns>
		public abstract IEnumerator GetEnumerator();

		/// <summary>
		/// Finds the index of the specified item in the list.
		/// </summary>
		/// <param name="item">Item whose index to find.</param>
		/// <returns>Returns the index of the found item, or null if none is found.</returns>
		/// <remarks>
		/// <para>The item index is found by iterating through the items until the specified item is found.</para>
		/// 
		/// <para>Object equality is determined by reference, not content.</para>
		/// </remarks>
		public int FindIndex( TItem item )
		{
			int i = 0;

			foreach( TItem currentItem in this )
			{
				if( currentItem == item )
					return i;

				i++;
			}

			return -1;
		}

		/// <summary>
		/// Adds a new item to the list. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="item">Item to add to the list.</param>
		public abstract void Add( object item );

		#endregion Public methods
	}
}
