////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Collections.Generic;

namespace Vela.Kernel
{
	/// <summary>
	/// Contract for a sorted list, that is, for a list in which items can be quickly found based on their key
	/// (<see cref="KernelKeyedList&lt;TKey, TItem&gt;.GetKey"/>), using a binary search algorithm.
	/// </summary>
	public abstract class KernelSortedList<TKey, TItem> : KernelKeyedList<TKey, TItem>
		where TItem : class
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="comparer">The equality comparer used to compare keys.</param>
		protected KernelSortedList( IItemEqualityComparer<TKey> comparer )
			: base( comparer )
		{
			// A performance comparison which shows that "SortedList" is the better choice for a sorted list with maximum
			// retrieval speed: http://blog.bodurov.com/Performance-SortedList-SortedDictionary-Dictionary-Hashtable/

			Items = new SortedList<TKey, TItem>( comparer );
		}

		#endregion Constructors

		#region Public methods

		/// <summary>
		/// Removes the item with the specified index from the list. Notifies property observers. Notifies list observers.
		/// </summary>
		/// <param name="index">The index of the item to remove from the list. This must be at least 0.</param>
		public virtual void RemoveAt( int index )
		{
			ValidationUtils.Integer( index, 0, Count - 1, "index" );

			SortedList<TKey, TItem> ItemsTyped = (SortedList<TKey, TItem>)Items;

			TItem itemToRemove = Items[ ItemsTyped.Keys[ index ] ];

			ItemsTyped.RemoveAt( index );

			// Notify property observers.
			OnPropertyChanged( CountPropertyName );

			// Notify list observers.
			OnItemRemoved( itemToRemove, index );
		}

		#endregion Public methods
	}
}
