////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel.Serialization;

namespace Vela.Kernel
{
	/// <summary>
	/// A locator. This can be used to uniquely identify and locate a document, a session, and so on.
	/// </summary>
	public class Locator
	{
		#region Public constants

		/// <summary>
		/// The minimum size (in bytes) of the value (<see cref="Value"/>).
		/// </summary>
		public const int ValueSizeMin = 16;

		/// <summary>
		/// The maximum size (in bytes) of the value (<see cref="Value"/>).
		/// </summary>
		public const int ValueSizeMax = 128;

		/// <summary>
		/// The default size (in bytes) of the value (<see cref="Value"/>).
		/// </summary>
		public const int ValueSizeDef = 32;

		/// <summary>
		/// A dummy locator to use where it's not possible to use a null locator.
		/// </summary>
		public static readonly Locator Dummy = new Locator( new byte[ ValueSizeMin ] );

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Locator()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="value">The locator's value.</param>
		public Locator( byte[] value )
		{
			Value = value;
		}

		#endregion Constructors

		#region Public serializable properties

		private byte[] _Value;
		/// <summary>
		/// The value of the locator. Normally, this has to be unique (= randomly generated). READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] Value
		{
			get { return _Value; }
			protected set
			{
				ValidationUtils.Array( value, ValueSizeMin, ValueSizeMax, "Value" );

				_Value = value;
			}
		}

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Gets the hash code of the specified locator.
		/// </summary>
		/// <returns>Returns the hash code of the specified locator.</returns>
		/// <remarks>
		/// <para>The hash code is generated from the <see cref="Value"/> property.</para>
		/// </remarks>
		public override int GetHashCode()
		{
			return BufferUtils.ComputeChecksum32( Value );
		}

		/// <summary>
		/// Returns the textual form of the current locator.
		/// </summary>
		/// <returns>Returns the textual form of the locator.</returns>
		public override string ToString()
		{
			return TextUtils.ArrayToMoteText32( Value );
		}

		/// <summary>
		/// Verifies if the current locator has the same value as the specified locator.
		/// </summary>
		/// <param name="locator">Locator to be verified against the current locator.</param>
		/// <returns>Returns true if the locators have the same value.</returns>
		/// <remarks>
		/// <para>The locators are equal if their <see cref="Value"/> properties are equal.</para>
		/// </remarks>
		public bool Equals( Locator locator )
		{
			ValidationUtils.ObjectHasContent( locator, "locator" );

			return BufferUtils.Equals( Value, locator.Value );
		}

		/// <summary>
		/// Verifies if the current locator has the same value as the specified locator.
		/// </summary>
		/// <param name="locator">Locator to be verified against the current locator.</param>
		/// <returns>Returns true if the locators have the same value.</returns>
		/// <remarks>
		/// <para>The locators are equal if their <see cref="Value"/> properties are equal.</para>
		/// </remarks>
		public override bool Equals( object locator )
		{
			return Equals( locator as Locator );
		}

		/// <summary>
		/// Compares the current locator with the specified locator.
		/// </summary>
		/// <param name="locator">Locator to be compared against the current locator.</param>
		/// <returns>Returns 0 if the locators have the same value, -1 if the current locator is smaller than "locator", and 1
		/// if the current locator is greater than "locator".
		/// </returns>
		/// <remarks>
		/// <para>The locators are equal if their <see cref="Value"/> properties are equal.</para>
		/// </remarks>
		public int Compare( Locator locator )
		{
			ValidationUtils.ObjectHasContent( locator, "locator" );

			return BufferUtils.Compare( Value, locator.Value );
		}

		#endregion Public methods
	}
}
