////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Threading;

namespace Vela.Kernel
{
	/// <summary>
	/// A log manager which writes to the system trace and to a log file.
	/// </summary>
	/// <remarks>
	/// <para>The log messages are written to the system trace if <see cref="LogDebugMessages"/> is true.</para>
	/// </remarks>
	public class LogManager
	{
		#region Public constants

		/// <summary>
		/// The name of the folder which contains the log files.
		/// </summary>
		public const string LogsFolderName = "Logs";

		/// <summary>
		/// The name of the current log file.
		/// </summary>
		public const string CurrentLogFileNameDef = "CurrentLog.txt";

		/// <summary>
		/// The name of the archive log file.
		/// </summary>
		public const string ArchiveLogFileNameDef = "ArchiveLog.txt";

		/// <summary>
		/// The maximum size of the log file.
		/// </summary>
		/// <remarks>
		/// <para>If the log file exceeds this size then it is archived and the logging starts over with a clean file.</para>
		/// 
		/// <para>The log file might exceed this size with the size of one log item.</para>
		/// </remarks>
		public const int LogFileSizeMax = 10000000;

		/// <summary>
		/// Template for the system-wide name of the system mutex. Placeholders: 0 - application name.
		/// </summary>
		public const string GlobalSystemMutexNameTemplate = "Global\\{0}.LogManager";

		/// <summary>
		/// The separator between consecutive keywords in <see cref="LogMessagesWhitelist"/> and
		/// <see cref="LogMessagesBlacklist"/>.
		/// </summary>
		public const char LogMessagesKeywordsSeparator = ';';

		#endregion Public constants

		#region Static constructors

		/// <summary>
		/// Initializes the current class.
		/// </summary>
		static LogManager()
		{
			if( AssemblyUtils.ApplicationBuildInfo == null )
			{
				string initializeMessage = string.Format( "You must initialize \"AssemblyUtils.ApplicationBuildInfo\" before"
					+ " calling \"{0}\".", typeof( LogManager ).FullName );
				Trace.WriteLine( initializeMessage, BuildInfo.LibrariesBuildInfo.ProductFullNameDef );

				throw new CriticalException( initializeMessage );
			}

			GlobalSystemMutexName = string.Format( GlobalSystemMutexNameTemplate
				, AssemblyUtils.ApplicationBuildInfo.ProductFullName );

			LogFolderName = AssemblyUtils.GetApplicationDataFolderName( LogsFolderName );
			CurrentLogFileName = FileUtils.MakeFullName( LogFolderName, CurrentLogFileNameDef );
			ArchiveLogFileName = FileUtils.MakeFullName( LogFolderName, ArchiveLogFileNameDef );

			string message = string.Format( "Logging to folder \"{0}\".", LogFolderName );
			Trace.WriteLine( message, AssemblyUtils.ApplicationBuildInfo.ProductFullName );
		}

		#endregion Static constructors

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="methodName">The name of the method where the new instance is created. This may be null or empty.</param>
		public LogManager( string methodName )
		{
			Stopwatch = Stopwatch.StartNew();
			InstanceId = CreateInstanceId();
			MethodName = methodName;
		}

		#endregion Constructors

		#region Events

		/// <summary>
		/// Event raised when a log record is added to the log manager.
		/// </summary>
		/// <remarks>
		/// <para>Setting this event may have dramatic performance implications, especially if it's used to update the UI
		/// through methods like "Dispatcher.Invoke".</para>
		/// 
		/// <para>This event is not raised for informational debug log records (= it is raised for debug errors and warnings).
		/// </para>
		/// </remarks>
		public static event LogRecordAddedDelegate Added;

		#endregion Events

		#region Private static fields

		/// <summary>
		/// The last unique identifier of an instance of this class.
		/// </summary>
		private static Int64 LastInstanceId;

		/// <summary>
		/// The name of the folder where the log files are stored.
		/// </summary>
		private static string LogFolderName;

		/// <summary>
		/// The name of the archive log file.
		/// </summary>
		private static string ArchiveLogFileName;

		/// <summary>
		/// The (default) system-wide name of the system mutex.
		/// </summary>
		private static string GlobalSystemMutexName = "Global\\Vela.LogManager";

		/// <summary>
		/// System mutex for writing to the log file.
		/// </summary>
		private static Mutex FileWriteSystemMutex = new Mutex( false, GlobalSystemMutexName );

		/// <summary>
		/// The list of keywords which must be present in the messages to log. This may be null or empty.
		/// </summary>
		private static string[] LogMessagesWhitelistSplit;

		/// <summary>
		/// The list of keywords which must NOT be present in the messages to log. This may be null or empty.
		/// </summary>
		private static string[] LogMessagesBlacklistSplit;

		#endregion Private static fields

		#region Private fields

		/// <summary>
		/// Stopwatch which measures the time elapsed in the current object.
		/// </summary>
		private Stopwatch Stopwatch;

		/// <summary>
		/// The unique identifier of the current object.
		/// </summary>
		private Int64 InstanceId;

		/// <summary>
		/// The name of the method where the current object was created. This may be null or empty.
		/// </summary>
		private string MethodName;

		#endregion Private fields

		#region Public static properties

		/// <summary>
		/// If this is true then the debug messages are logged.
		/// </summary>
		/// <remarks>
		/// <para>Setting this with true may have dramatic performance implications because it activates writing the log
		/// messages to the system trace.</para>
		/// </remarks>
		public static bool LogDebugMessages = false;

		/// <summary>
		/// If this is true then the debug messages are saved.
		/// </summary>
		public static bool SaveDebugMessages = false;

		private static string _LogMessagesWhitelist;
		/// <summary>
		/// The list of keywords which must be present in the messages to log, separated by
		/// <see cref="LogMessagesKeywordsSeparator"/>. This may be null or empty.
		/// </summary>
		public static string LogMessagesWhitelist
		{
			get { return _LogMessagesWhitelist; }
			set
			{
				_LogMessagesWhitelist = value;

				LogMessagesWhitelistSplit = TextUtils.HasContent( value ) ? value.Split( LogMessagesKeywordsSeparator ) : null;
			}
		}

		private static string _LogMessagesBlacklist;
		/// <summary>
		/// The list of keywords which must NOT be present in the messages to log, separated by
		/// <see cref="LogMessagesKeywordsSeparator"/>. This may be null or empty.
		/// </summary>
		public static string LogMessagesBlacklist
		{
			get { return _LogMessagesBlacklist; }
			set
			{
				_LogMessagesBlacklist = value;

				LogMessagesBlacklistSplit = TextUtils.HasContent( value ) ? value.Split( LogMessagesKeywordsSeparator ) : null;
			}
		}

		private static string _CurrentLogFileName;
		/// <summary>
		/// The name of the current log file. READ-ONLY.
		/// </summary>
		public static string CurrentLogFileName
		{
			get { return _CurrentLogFileName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "CurrentLogFileName" );

				_CurrentLogFileName = value;
			}
		}

		#endregion Public static properties

		#region Private static methods

		/// <summary>
		/// Creates a new unique identifier for an instance of the current class.
		/// </summary>
		/// <returns>Returns a new unique identifier.</returns>
		/// <remarks>
		/// <para><see cref="LastInstanceId"/> is thread safe.</para>
		/// 
		/// <para><see cref="LastInstanceId"/> is incremented here.</para>
		/// </remarks>
		private static Int64 CreateInstanceId()
		{
			return Interlocked.Increment( ref LastInstanceId );
		}

		/// <summary>
		/// Gets the full name of the specified method.
		/// </summary>
		/// <param name="methodInfo">Information about the method whose name to get. This is not validated.</param>
		/// <returns>Returns the full name of the specified method.</returns>
		private static string GetMethodFullName( MethodBase methodInfo )
		{
			return string.Format( "{0}.{1}", methodInfo.DeclaringType.FullName, methodInfo.Name );
		}

		/// <summary>
		/// If the current log file has exceeded the maximum size, it's archived.
		/// </summary>
		private static void ArchiveLog()
		{
			// Check if the log file exists.
			if( File.Exists( CurrentLogFileName ) )
			{
				// Check if the size of the log file is above the limit.
				if( FileUtils.GetFileSize( CurrentLogFileName ) > LogFileSizeMax )
				{
					// Archive the log file by renaming it.

					// Delete the existing archive log file.
					if( File.Exists( ArchiveLogFileName ) )
						File.Delete( ArchiveLogFileName );

					// Rename the current log file.
					File.Move( CurrentLogFileName, ArchiveLogFileName );
				}
			}
		}

		/// <summary>
		/// Writes the specified message in the log file.
		/// </summary>
		/// <param name="message">Message to write in the log file.</param>
		private static void WriteToFile( string message )
		{
			FileUtils.CreateFolder( LogFolderName );

			// Open the system mutex.
			if( FileWriteSystemMutex.WaitOne( -1 ) )
			{
				try
				{
					// Archive the log file is necessary.
					ArchiveLog();

					File.AppendAllText( CurrentLogFileName, message + "\n" );
				}
				finally
				{
					FileWriteSystemMutex.ReleaseMutex();
				}
			}
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Adds the specified record to the log.
		/// </summary>
		/// <param name="logRecord">Record to log. If this is null, nothing happens.</param>
		/// <param name="save">If this is true then the specified message is saved in the log file.</param>
		/// <returns>Returns true if the log message should be shown to the user.</returns>
		/// <remarks>
		/// <para>If the log record is a debug one and <see cref="LogDebugMessages"/> is false, nothing happens and false is
		/// returned.</para>
		/// 
		/// <para>If the log record should not be logged, nothing happens and the <see cref="LogRecord.ShouldShow"/> method is
		/// returned.</para>
		/// 
		/// <para>If the log record is a debug one, and <see cref="LogDebugMessages"/> and <see cref="SaveDebugMessages"/> are
		/// true, the message is saved regardless of "save".</para>
		/// 
		/// <para>The <see cref="ThreadAbortException"/> exception is never logged.</para>
		/// 
		/// <para>The <see cref="Added"/> event is raised only if the log record is not a debug one or it's log mode is not
		/// <see cref="LogRecordModes.Info"/>.</para>
		/// 
		/// <para>After an exception is logged, its log mode is changed so that the exception is not logged again.</para>
		/// </remarks>
		public static bool Add( LogRecord logRecord, bool save )
		{
			if( (logRecord == null) || (logRecord.IsDebug && !LogDebugMessages) )
				return false;

			if( !logRecord.ShouldLog() )
				return logRecord.ShouldShow();

			try
			{
				string logMessage = logRecord.GetLogMessage();
				bool isDebug = logRecord.IsDebug && LogDebugMessages;

				// Filter through the whitelist and blacklist.
				bool filteringAllowsLog =
					(
						ArrayUtils.NoContent( LogMessagesWhitelistSplit )
						|| (TextUtils.FindPatternInText( LogMessagesWhitelistSplit, logMessage ) >= 0)
					)
					&& (
						ArrayUtils.NoContent( LogMessagesBlacklistSplit )
						|| (TextUtils.FindPatternInText( LogMessagesBlacklistSplit, logMessage ) < 0)
					);

				if( filteringAllowsLog )
				{
					// Write the log message to the system trace.
					if( save || (logRecord.Mode != LogRecordModes.Info) || !logRecord.IsDebug || isDebug )
						Trace.WriteLine( logMessage );

					if( !logRecord.IsDebug )
					{
						// Write the log message in the log file.
						if( save || (logRecord.Mode != LogRecordModes.Info) || (isDebug && SaveDebugMessages) )
							WriteToFile( logMessage );
					}

					if( (!logRecord.IsDebug || (logRecord.Mode != LogRecordModes.Info))
						&& (Added != null) && !logRecord.HasCriticalException() )
					{
						Added( logRecord );
					}
				}

				return true;
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsAbort( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				string message = string.Format( "Failed to add record to log:\n{0}", ex );
				Trace.WriteLine( message, AssemblyUtils.ApplicationBuildInfo.ProductFullName );

				return false;
			}
		}

		/// <summary>
		/// Adds a new record to the log.
		/// </summary>
		/// <param name="message">Message to log. This may be null or empty.</param>
		/// <param name="exception">Exception to log. This may be null.</param>
		/// <param name="logMode">Log mode.</param>
		/// <param name="save">If this is true then the specified message is saved in the log file.</param>
		/// <returns>Returns true if the exception should be shown to the user.</returns>
		/// <remarks>
		/// <para>If the log record is a debug one and <see cref="LogDebugMessages"/> is false, nothing happens and false is
		/// returned.</para>
		/// 
		/// <para>If the log record should not be logged, nothing happens and the <see cref="LogRecord.ShouldShow"/> method is
		/// returned.</para>
		/// 
		/// <para>If the log record is a debug one, and <see cref="LogDebugMessages"/> and <see cref="SaveDebugMessages"/> are
		/// true, the message is saved regardless of "save".</para>
		/// 
		/// <para>The <see cref="ThreadAbortException"/> exception is never logged.</para>
		/// 
		/// <para>The <see cref="Added"/> event is raised only if the log record is not a debug one or it's log mode is not
		/// <see cref="LogRecordModes.Info"/>.</para>
		/// 
		/// <para>After an exception is logged, its log mode is changed so that the exception is not logged again.</para>
		/// </remarks>
		public static bool Add( string message, Exception exception, LogRecordModes logMode, bool save )
		{
			LogRecord logRecord = new LogRecord( -1, false, logMode );
			logRecord.Fields.Add( new LogRecordField( null, message, true, LogRecordFieldPriorities.End ) );
			logRecord.Fields.Add( new LogRecordField( null, exception, true, LogRecordFieldPriorities.End ) );

			return Add( logRecord, save );
		}

		/// <summary>
		/// Adds a message to the log.
		/// </summary>
		/// <param name="message">Message to log. This may be null or empty.</param>
		/// <param name="save">If this is true then the specified message is saved in the log file.</param>
		public static void Add( string message, bool save = false )
		{
			LogRecord logRecord = new LogRecord( -1, false, LogRecordModes.Info );
			logRecord.Fields.Add( new LogRecordField( null, message, true, LogRecordFieldPriorities.End ) );

			Add( logRecord, save );
		}

		/// <summary>
		/// Adds an exception to the log. The exception is saved in the log file.
		/// </summary>
		/// <param name="exception">Exception to log. This may be null.</param>
		/// <returns>Returns true if the exception should be shown.</returns>
		/// <remarks>
		/// <para>If the log record is a debug one and <see cref="LogDebugMessages"/> is false, nothing happens and false is
		/// returned.</para>
		/// 
		/// <para>If the log record should not be logged, nothing happens and the <see cref="LogRecord.ShouldShow"/> method is
		/// returned.</para>
		/// 
		/// <para>If the log record is a debug one, and <see cref="LogDebugMessages"/> and <see cref="SaveDebugMessages"/> are
		/// true, the message is saved regardless of "save".</para>
		/// 
		/// <para>The <see cref="ThreadAbortException"/> exception is never logged.</para>
		/// 
		/// <para>The <see cref="Added"/> event is raised only if the log record is not a debug one or it's log mode is not
		/// <see cref="LogRecordModes.Info"/>.</para>
		/// 
		/// <para>After an exception is logged, its log mode is changed so that the exception is not logged again.</para>
		/// </remarks>
		public static bool Add( Exception exception )
		{
			LogRecord logRecord = new LogRecord( -1, false, LogRecordModes.Error );
			logRecord.Fields.Add( new LogRecordField( null, exception, true, LogRecordFieldPriorities.End ) );

			return Add( logRecord, true );
		}

		/// <summary>
		/// Adds a new record to the log.
		/// </summary>
		/// <param name="message">Message to log. This may be null or empty.</param>
		/// <param name="exception">Exception to log. This may be null.</param>
		/// <param name="logMode">Log mode.</param>
		/// <param name="save">If this is true then the specified message is saved in the log file.</param>
		/// <remarks>
		/// <para>If <see cref="LogDebugMessages"/> is false, nothing happens.</para>
		/// 
		/// <para>If the log record should not be logged, nothing happens and the <see cref="LogRecord.ShouldShow"/> method is
		/// returned.</para>
		/// 
		/// <para>If the log record is a debug one, and <see cref="LogDebugMessages"/> and <see cref="SaveDebugMessages"/> are
		/// true, the message is saved regardless of "save".</para>
		/// 
		/// <para>The <see cref="ThreadAbortException"/> exception is never logged.</para>
		/// 
		/// <para>The <see cref="Added"/> event is raised only if the log record is not a debug one or it's log mode is not
		/// <see cref="LogRecordModes.Info"/>.</para>
		/// 
		/// <para>After an exception is logged, its log mode is changed so that the exception is not logged again.</para>
		/// </remarks>
		public static void DebugAdd( string message, Exception exception, LogRecordModes logMode, bool save )
		{
			if( !LogDebugMessages )
				return;

			LogRecord logRecord = new LogRecord( -1, true, logMode );
			logRecord.Fields.Add( new LogRecordField( null, message, true, LogRecordFieldPriorities.End ) );
			logRecord.Fields.Add( new LogRecordField( null, exception, true, LogRecordFieldPriorities.End ) );

			Add( logRecord, save );
		}

		/// <summary>
		/// Adds a message to the log.
		/// </summary>
		/// <param name="message">Message to log. This may be null or empty.</param>
		/// <param name="recordFields">Extra fields to add to the created log record. This may be null or empty.</param>
		/// <remarks>
		/// <para>If <see cref="LogDebugMessages"/> is false, nothing happens.</para>
		/// 
		/// <para>If <see cref="SaveDebugMessages"/> is true, the message is saved.</para>
		/// </remarks>
		public static void DebugAdd( string message, LogRecordFields recordFields )
		{
			if( !LogDebugMessages )
				return;

			LogRecord logRecord = new LogRecord( -1, true, LogRecordModes.Info );
			logRecord.Fields.Add( new LogRecordField( null, message, true, LogRecordFieldPriorities.End ) );

			if( recordFields != null )
				logRecord.Fields.Add( recordFields );

			Add( logRecord, SaveDebugMessages );
		}

		/// <summary>
		/// Adds a message to the log.
		/// </summary>
		/// <param name="message">Message to log. This may be null or empty.</param>
		/// <param name="datas">Data to format in "message". This may be null or empty.</param>
		/// <remarks>
		/// <para>If <see cref="LogDebugMessages"/> is false, nothing happens.</para>
		/// 
		/// <para>If <see cref="SaveDebugMessages"/> is true, the message is saved.</para>
		/// </remarks>
		public static void DebugAddFormat( string message, params object[] datas )
		{
			if( !LogDebugMessages )
				return;

			string fullMessage = (TextUtils.HasContent( message ) && (datas != null) && (datas.Length > 0))
				? string.Format( message, datas ) : message;

			LogRecord logRecord = new LogRecord( -1, true, LogRecordModes.Info );
			logRecord.Fields.Add( new LogRecordField( null, fullMessage, true, LogRecordFieldPriorities.End ) );

			Add( logRecord, SaveDebugMessages );
		}

		/// <summary>
		/// Adds an exception to the log. The exception is saved in the log file.
		/// </summary>
		/// <param name="exception">Exception to log. This may be null.</param>
		/// <remarks>
		/// <para>If the log record is a debug one and <see cref="LogDebugMessages"/> is false, nothing happens and false is
		/// returned.</para>
		/// 
		/// <para>If the log record should not be logged, nothing happens and the <see cref="LogRecord.ShouldShow"/> method is
		/// returned.</para>
		/// 
		/// <para>If the log record is a debug one, and <see cref="LogDebugMessages"/> and <see cref="SaveDebugMessages"/> are
		/// true, the message is saved regardless of "save".</para>
		/// 
		/// <para>The <see cref="ThreadAbortException"/> exception is never logged.</para>
		/// 
		/// <para>The <see cref="Added"/> event is raised only if the log record is not a debug one or it's log mode is not
		/// <see cref="LogRecordModes.Info"/>.</para>
		/// 
		/// <para>After an exception is logged, its log mode is changed so that the exception is not logged again.</para>
		/// </remarks>
		public static void DebugAdd( Exception exception )
		{
			if( !LogDebugMessages )
				return;

			LogRecord logRecord = new LogRecord( -1, true, LogRecordModes.Error );
			logRecord.Fields.Add( new LogRecordField( null, exception, true, LogRecordFieldPriorities.End ) );

			Add( logRecord, true );
		}

		/// <summary>
		/// Creates a new instance of the current class and logs a debug message.
		/// </summary>
		/// <param name="methodName">The name of the method where the instance is created.</param>
		/// <param name="recordFields">Extra fields to add to the created log record. This may be null or empty.</param>
		/// <returns>Returns the created instance. This is null if <see cref="LogDebugMessages"/> is false.</returns>
		/// <remarks>
		/// <para>If <see cref="LogDebugMessages"/> is false, nothing happens.</para>
		/// 
		/// <para>If <see cref="SaveDebugMessages"/> is true, the message is saved.</para>
		/// </remarks>
		public static LogManager DebugStartMethod( string methodName, LogRecordFields recordFields = null )
		{
			if( !LogDebugMessages )
				return null;

			ValidationUtils.TextHasContent( methodName, "methodName" );

			LogManager log = new LogManager( methodName );

			LogRecord logRecord = new LogRecord( -1, true, LogRecordModes.Info );
			logRecord.Fields.Add( new LogRecordField( "LMID", log.InstanceId, false, LogRecordFieldPriorities.Beginning ) );
			logRecord.Fields.Add( new LogRecordField( "Started", methodName, false, LogRecordFieldPriorities.Middle ) );

			if( recordFields != null )
				logRecord.Fields.Add( recordFields );

			Add( logRecord, SaveDebugMessages );

			return log;
		}

		/// <summary>
		/// Creates a new instance of the current class and logs a debug message.
		/// </summary>
		/// <param name="methodInfo">Information about the method where the instance is created.</param>
		/// <param name="recordFields">Extra fields to add to the created log record. This may be null or empty.</param>
		/// <returns>Returns the created instance. This is null if <see cref="LogDebugMessages"/> is false.</returns>
		/// <remarks>
		/// <para>If <see cref="LogDebugMessages"/> is false, nothing happens.</para>
		/// 
		/// <para>If <see cref="SaveDebugMessages"/> is true, the message is saved.</para>
		/// </remarks>
		public static LogManager DebugStartMethod( MethodBase methodInfo, LogRecordFields recordFields = null )
		{
			if( !LogDebugMessages )
				return null;

			return DebugStartMethod( GetMethodFullName( methodInfo ), recordFields );
		}

		/// <summary>
		/// Shows the message for the specified log manager. A debug message is logged.
		/// </summary>
		/// <param name="log">Log manager for which to show the message. If this is null, nothing happens.</param>
		/// <param name="recordFields">Extra fields to add to the created log record. This may be null or empty.</param>
		/// <remarks>
		/// <para>If <see cref="LogDebugMessages"/> is false, nothing happens.</para>
		/// 
		/// <para>If <see cref="SaveDebugMessages"/> is true, the message is saved.</para>
		/// </remarks>
		public static void DebugEndMethod( LogManager log, LogRecordFields recordFields = null )
		{
			if( !LogDebugMessages || (log == null) )
				return;

			string time = log.GetElapsedTime().TotalSeconds.ToString( "0.##" );

			LogRecord logRecord = new LogRecord( -1, true, LogRecordModes.Info );
			logRecord.Fields.Add( new LogRecordField( "LMID", log.InstanceId, false, LogRecordFieldPriorities.Beginning ) );
			logRecord.Fields.Add( new LogRecordField( "Ended", log.MethodName, false, LogRecordFieldPriorities.Middle ) );
			logRecord.Fields.Add( new LogRecordField( "Time (seconds)", time, false, LogRecordFieldPriorities.Middle ) );

			if( recordFields != null )
				logRecord.Fields.Add( recordFields );

			Add( logRecord, SaveDebugMessages );
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Gets the elapsed time of the stopwatch.
		/// </summary>
		/// <returns>Returns the elapsed time.</returns>
		public TimeSpan GetElapsedTime()
		{
			return Stopwatch.Elapsed;
		}

		/// <summary>
		/// Gets the elapsed number of seconds of the stopwatch.
		/// </summary>
		/// <returns>Returns the elapsed number of seconds.</returns>
		public double GetElapsedSeconds()
		{
			return Stopwatch.Elapsed.TotalSeconds;
		}

		#endregion Public methods
	}
}
