////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Text;
using System.Threading;

namespace Vela.Kernel
{
	/// <summary>
	/// A log record.
	/// </summary>
	public class LogRecord
	{
		#region Public constants

		/// <summary>
		/// The separator between consecutive fields with scalar (textual) values.
		/// </summary>
		public const string ScalarFieldsSeparator = " | ";

		/// <summary>
		/// The separator between the name and the value of a field with a scalar (textual) value.
		/// </summary>
		public const string ScalarFieldNameValueSeparator = ": ";

		/// <summary>
		/// The separator between the name and the value of a field with an exception value.
		/// </summary>
		public const string ExceptionFieldNameValueSeparator = ":\n";

		/// <summary>
		/// Template for the message prefix. Placeholders: 0 - application name, 1 - creation time, 2 - thread ID, 3 - mode.
		/// </summary>
		public const string MessagePrefixTemplate = "{0} | Time {1} | THID {2} | Mode {3}";

		/// <summary>
		/// The format for converting a time to text.
		/// </summary>
		public const string TimeFormat = "yyyy'-'MM'-'dd HH':'mm':'ss.fff";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected LogRecord()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="threadId">The identifier of the thread which created the log record. If this is a negative number, the
		/// ID of the current thread is used instead.</param>
		/// <param name="isDebug">If this is true then the log record contains a debug message.</param>
		/// <param name="mode">Log mode.</param>
		public LogRecord( int threadId, bool isDebug, LogRecordModes mode )
		{
			ThreadId = (threadId >= 0) ? threadId : Thread.CurrentThread.ManagedThreadId;
			IsDebug = isDebug;
			Mode = mode;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private DateTime _CreationTime = DateTime.UtcNow;
		/// <summary>
		/// The time (in UTC format) when the log record was created. This is automatically set with the time of the creation of
		/// the current object (and you should not modify it). READ-ONLY.
		/// </summary>
		public DateTime CreationTime
		{
			get { return _CreationTime; }
			protected set { _CreationTime = value; }
		}

		private int _ThreadId;
		/// <summary>
		/// The identifier of the thread which created the log record. READ-ONLY.
		/// </summary>
		public int ThreadId
		{
			get { return _ThreadId; }
			protected set { _ThreadId = value; }
		}

		private bool _IsDebug;
		/// <summary>
		/// If this is true then the log record contains a debug message. READ-ONLY.
		/// </summary>
		public bool IsDebug
		{
			get { return _IsDebug; }
			protected set { _IsDebug = value; }
		}

		private LogRecordModes _Mode = LogRecordModes.Info;
		/// <summary>
		/// Mode. READ-ONLY.
		/// </summary>
		public LogRecordModes Mode
		{
			get { return _Mode; }
			protected set { _Mode = value; }
		}

		private LogRecordFields _Fields = new LogRecordFields();
		/// <summary>
		/// List of fields. The backing field is initialized by default. READ-ONLY.
		/// </summary>
		public LogRecordFields Fields
		{
			get { return _Fields; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Fields" );

				_Fields = value;
			}
		}

		/// <summary>
		/// Message for a log list. This may be null or empty. READ-ONLY.
		/// </summary>
		public string MessageForLogList
		{
			get { return GetMessageForLogList(); }
		}

		#endregion Public non-serializable properties

		#region Private static methods

		/// <summary>
		/// Adds the textual representation of the specified field, in the specified string builder.
		/// </summary>
		/// <param name="sb">String builder in which to add the textual field. This is modified in here. If this is null, nothing
		/// happens.</param>
		/// <param name="field">Field whose textual representation to add to the string builder. If this is null, nothing
		/// happens.</param>
		/// <param name="showToUser">If this is true then the field will be shown to the user.</param>
		private static void AddFieldText( StringBuilder sb, LogRecordField field, bool showToUser )
		{
			if( (sb == null) || (field == null) ) return;

			object value = field.Value;
			if( value == null ) return;

			Exception exception = value as Exception;
			bool isException = exception != null;

			string valutText;
			if( isException )
			{
				valutText = showToUser ? ExceptionUtils.GetPresenterMessage( null, exception, false )
					: GetInnerExceptions( exception, null );
			}
			else
			{
				valutText = value.ToString();
			}

			isException = isException || ((valutText != null) && (valutText.IndexOf( '\n' ) >= 0));

			if( isException ) TextUtils.AddWithLineBreaks( sb, 2, null );
			else if( sb.Length > 0 ) sb.Append( ScalarFieldsSeparator );

			if( TextUtils.HasContent( field.Name ) )
			{
				sb.Append( field.Name );
				sb.Append( isException ? ExceptionFieldNameValueSeparator : ScalarFieldNameValueSeparator );
			}

			sb.Append( valutText );
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Gets all the inner exceptions from the specified exception.
		/// </summary>
		/// <param name="exception">Exception for which to get the messages. This may be null.</param>
		/// <param name="indent">Indentation text to add in front of every line of text.</param>
		/// <returns>Returns all inner exceptions from the specified exception.</returns>
		/// <remarks>
		/// <para>Some information about the current exception (like "StackTrace") is set (by .Net) only when the exception is
		/// actually raised.</para>
		/// </remarks>
		public static string GetInnerExceptions( Exception exception, string indent )
		{
			if( exception == null ) return null;

			string exceptions = "";
			Exception ex = exception;
			int imbrication = 0; // The level of imbrication of the current exception.
			string prefix = "\n" + indent;

			while( ex != null )
			{
				#region Get the current exception

				Type type = ex.GetType();

				exceptions +=
					prefix + "<Exception Imbrication=" + imbrication + ">"
						+ prefix + "\t<Type>" + type.FullName + "</Type>"
						+ prefix + "\t<Message>" + ex.Message + "</Message>"
						+ (
							(ex as BaseException != null)
								? (
									prefix + "\t<Details>" + ExceptionUtils.GetDetails( ex )
									+ "</Details>"
								)
								: ""
						)
						+ (
							(ex as ValidationException != null)
								? (
									prefix + "\t<ErrorContext>" + ((ValidationException)ex).ErrorContext
									+ "</ErrorContext>"
								)
								: ""
						)
						+ prefix + "\t<StackTrace>"
							+ prefix + "\t\t"
							+ (
								(ex.StackTrace != null)
									? ex.StackTrace.Replace( "\n", prefix + "\t\t" )
									: ""
							)
						+ prefix + "\t</StackTrace>"
						+ prefix + "</Exception>";

				#endregion Get the current exception

				// Move to the next exception.
				ex = ex.InnerException;
				imbrication++;
			}

			return exceptions;
		}

		#endregion Public static methods

		#region Private methods

		/// <summary>
		/// Adds the message prefix to the specified string builder.
		/// </summary>
		/// <param name="sb">String builder in which to add the textual field. If this is null, nothing happens.</param>
		private void AddMessagePrefix( StringBuilder sb )
		{
			if( sb == null ) return;

			sb.AppendFormat( MessagePrefixTemplate, AssemblyUtils.ApplicationBuildInfo.ProductFullName
				, CreationTime.ToLocalTime().ToString( TimeFormat ), ThreadId, Mode );
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Returns true if a contained exception is critical.
		/// </summary>
		/// <returns>Returns true if a contained exception is critical.</returns>
		public bool HasCriticalException()
		{
			foreach( LogRecordField field in Fields )
			{
				if( ExceptionUtils.IsCritical( field.Value as Exception ) ) return true;
			}

			return false;
		}

		/// <summary>
		/// Returns true if a contained exception should be logged.
		/// </summary>
		/// <returns>Returns true if a contained exception should be logged.</returns>
		public bool ShouldLog()
		{
			foreach( LogRecordField field in Fields )
			{
				if( field.Value == null ) continue;

				Exception exception = field.Value as Exception;

				// If a field's (non-null) value is not an exception or is an exception which should be logged, return true.
				if( (exception == null) || ExceptionUtils.ShouldLog( exception ) ) return true;
			}

			return false;
		}

		/// <summary>
		/// Returns true if a contained exception should be shown to the user.
		/// </summary>
		/// <returns>Returns true if a contained exception should be shown to the user.</returns>
		public bool ShouldShow()
		{
			foreach( LogRecordField field in Fields )
			{
				if( ExceptionUtils.ShouldShow( field.Value as Exception ) ) return true;
			}

			return false;
		}

		/// <summary>
		/// Gets the message for logging.
		/// </summary>
		/// <returns>Returns the message for logging. This may be null or empty.</returns>
		public string GetLogMessage()
		{
			StringBuilder sb = new StringBuilder();

			AddMessagePrefix( sb );

			// The items with the "LogRecordFieldPriorities.Beginning" priority.
			foreach( LogRecordField field in Fields )
			{
				if( field.Priority == LogRecordFieldPriorities.Beginning ) AddFieldText( sb, field, false );
			}

			// The items with the "LogRecordFieldPriorities.Middle" priority.
			foreach( LogRecordField field in Fields )
			{
				if( field.Priority == LogRecordFieldPriorities.Middle ) AddFieldText( sb, field, false );
			}

			// The items with the "LogRecordFieldPriorities.End" priority.
			foreach( LogRecordField field in Fields )
			{
				if( field.Priority == LogRecordFieldPriorities.End ) AddFieldText( sb, field, false );
			}

			return sb.ToString();
		}

		/// <summary>
		/// Gets the message for a log list.
		/// </summary>
		/// <returns>Returns the message for a log list. This may be null or empty.</returns>
		public string GetMessageForLogList()
		{
			StringBuilder sb = new StringBuilder();

			// The items with the "LogRecordFieldPriorities.Beginning" priority.
			foreach( LogRecordField field in Fields )
			{
				if( (field.Priority == LogRecordFieldPriorities.Beginning) && field.ShowToUser ) AddFieldText( sb, field, true );
			}

			// The items with the "LogRecordFieldPriorities.Middle" priority.
			foreach( LogRecordField field in Fields )
			{
				if( (field.Priority == LogRecordFieldPriorities.Middle) && field.ShowToUser ) AddFieldText( sb, field, true );
			}

			// The items with the "LogRecordFieldPriorities.End" priority.
			foreach( LogRecordField field in Fields )
			{
				if( (field.Priority == LogRecordFieldPriorities.End) && field.ShowToUser ) AddFieldText( sb, field, true );
			}

			if( sb.Length <= 0 ) sb.Append( "Error" );

			return sb.ToString();
		}

		#endregion Public methods
	}
}
