﻿////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////


using System.Collections.Generic;

namespace Vela.Kernel
{
	/// <summary>
	/// Parsing engine for extracting name-value fields from a text.
	/// </summary>
	/// <remarks>
	/// <para>Blank spaces from a field's name and value are trimmed.</para>
	/// 
	/// <para>If duplicate names are encountered, the new value is ignored (and the existing value is kept).</para>
	/// 
	/// <para>Fields without values are ignored (to minimize the impact of the parsing on the resources).</para>
	/// </remarks>
	public class NameValueParser
	{
		#region Public constants

		/// <summary>
		/// The maximum size of the name of a field.
		/// </summary>
		public const int NameSizeMax = 50;

		/// <summary>
		/// The separator between the name and the value of a field.
		/// </summary>
		public const string NameValueSeparator = "=";

		/// <summary>
		/// The possible separators between two consecutive fields.
		/// </summary>
		/// <remarks>
		/// <para>Each separator must be a single character.</para>
		/// </remarks>
		public static readonly char[] FieldsSeparators = new char[] { ';', '\n' };

		/// <summary>
		/// The texts which are recognized as a boolean true value.
		/// </summary>
		/// <remarks>
		/// <para>These must be in lowercase.</para>
		/// </remarks>
		public static readonly List<string> BooleanTrueValues = new List<string>() { "true", "yes", "on" };

		/// <summary>
		/// The texts which are recognized as a boolean false value.
		/// </summary>
		/// <remarks>
		/// <para>These must be in lowercase.</para>
		/// </remarks>
		public static readonly List<string> BooleanFalseValues = new List<string>() { "false", "no", "off" };

		#endregion Public constants

		#region Private declarations

		/// <summary>
		/// List with the fields resulted from parsing.
		/// </summary>
		private NameValueFields Fields = new NameValueFields();

		#endregion Private declarations

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="textualFields">Text to parse. This may be null or empty.</param>
		/// <param name="fieldNameWhichEndsParsing">The field name which ends the parsing.</param>
		public NameValueParser( string textualFields, string fieldNameWhichEndsParsing )
		{
			ParseFields( textualFields, fieldNameWhichEndsParsing );
		}

		#endregion Constructors

		#region Private methods

		/// <summary>
		/// Parses the specified text in multiple fields.
		/// </summary>
		/// <param name="textualFields">Text to parse. This may be null or empty.</param>
		/// <param name="fieldNameWhichEndsParsing">The field name which ends the parsing.</param>
		private void ParseFields( string textualFields, string fieldNameWhichEndsParsing )
		{
			if( TextUtils.NoContent( textualFields ) )
				return;

			// Don't use "string.Split" because the text may be too large and most of it may be unrelated to a command line.

			bool continueParsing = true;
			int textualFieldsSize = textualFields.Length;
			int lastTextualFieldPos = 0;

			while( continueParsing && (lastTextualFieldPos >= 0) && (lastTextualFieldPos < textualFieldsSize) )
			{
				// Optimize the memory use because "textualFields" may be quite large and not include (any more) fields, so there
				// is no point in creating a substring of the remaining text in order to parse it as a field.
				int nameValueSeparatorPos = textualFields.IndexOf( NameValueSeparator, lastTextualFieldPos );
				if( (nameValueSeparatorPos < 0) || (nameValueSeparatorPos - lastTextualFieldPos > NameSizeMax) )
				{
					continueParsing = false;

					continue;
				}

				int fieldsSeparatorPos = textualFields.IndexOfAny( FieldsSeparators, lastTextualFieldPos );
				if( fieldsSeparatorPos < 0 )
					fieldsSeparatorPos = textualFieldsSize;

				string textualField = textualFields.Substring( lastTextualFieldPos, fieldsSeparatorPos - lastTextualFieldPos );

				continueParsing = ParseField( textualField, fieldNameWhichEndsParsing );

				lastTextualFieldPos = fieldsSeparatorPos + 1;
			}
		}

		/// <summary>
		/// Parses the specified text to a single field.
		/// </summary>
		/// <param name="textualField">Text to parse. This may be null or empty.</param>
		/// <param name="fieldNameWhichEndsParsing">The field name which ends the parsing.</param>
		/// <remarks>Returns true if the parsing must continue, and false if it must stop.</remarks>
		private bool ParseField( string textualField, string fieldNameWhichEndsParsing )
		{
			if( TextUtils.NoContent( textualField ) )
				return true;

			int nameValueSeparatorPos = textualField.IndexOf( NameValueSeparator );

			// If there is no name-value separator, there is no field.
			if( (nameValueSeparatorPos < 0) || (nameValueSeparatorPos > NameSizeMax) )
				return false;

			string name = textualField.Substring( 0, nameValueSeparatorPos ).Trim().ToLowerInvariant();
			if( TextUtils.NoContent( name ) || Fields.ContainsKey( name ) )
				return true;

			if( TextUtils.Equals( name, fieldNameWhichEndsParsing ) )
				return false;

			string value = textualField.Substring( nameValueSeparatorPos + NameValueSeparator.Length ).Trim();

			Fields.Add( new NameValueField( name, value ), true );

			return true;
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Checks if the field with the specified name exists (and has a value).
		/// </summary>
		/// <param name="fieldName">The name of the field to check if it exists (and has a value).</param>
		/// <returns>Returns true if the field with the specified name exists (and has a value).</returns>
		public bool HasField( string fieldName )
		{
			ValidationUtils.TextHasContent( fieldName, "fieldName" );

			NameValueField field = Fields[ fieldName.ToLowerInvariant() ];

			return field != null; // Only fields with a value are in the list of fields.
		}

		/// <summary>
		/// Gets the value of the field with the specified name.
		/// </summary>
		/// <param name="fieldName">The name of the field whose value to get.</param>
		/// <param name="valueDef">The default value which is returned if the field's name is not in the parsed list. This may be
		/// null or empty.</param>
		/// <returns>Returns the value of the field with the specified name. This may be null or empty.</returns>
		public string GetTextValue( string fieldName, string valueDef )
		{
			ValidationUtils.TextHasContent( fieldName, "fieldName" );

			NameValueField field = Fields[ fieldName.ToLowerInvariant() ];

			return (field != null) ? field.Value : valueDef;
		}

		/// <summary>
		/// Gets the integer value of the field with the specified name.
		/// </summary>
		/// <param name="fieldName">The name of the field whose value to get.</param>
		/// <param name="valueMin">Minimum value allowed.</param>
		/// <param name="valueMax">Maximum value allowed.</param>
		/// <param name="valueDef">The default value which is returned if the field's name is not in the parsed list.</param>
		/// <returns>Returns the integer value of the field with the specified name.</returns>
		public int GetIntegerValue( string fieldName, int valueMin, int valueMax, int valueDef )
		{
			if( !HasField( fieldName ) )
				return valueDef;

			string fieldValue = GetTextValue( fieldName, null );

			int integer;
			bool parsed = int.TryParse( fieldValue, out integer );

			if( !parsed )
			{
				string message = string.Format( "The field \"{0}\" has a value \"{1}\" which doesn't contain a valid integer."
					, fieldName, fieldValue );

				throw new ValidationException( message );
			}

			if( (integer < valueMin) || (integer > valueMax) )
			{
				string message = string.Format( "The field \"{0}\" has an integer value \"{1}\" which is outside the required"
					+ " bounds [{2}...{3}].", fieldName, fieldValue, valueMin, valueMax );

				throw new ValidationException( message );
			}

			return integer;
		}

		/// <summary>
		/// Gets the boolean value of the field with the specified name.
		/// </summary>
		/// <param name="fieldName">The name of the field whose value to get.</param>
		/// <param name="valueDef">The default value which is returned if the field's name is not in the parsed list.</param>
		/// <returns>Returns the boolean value of the field with the specified name.</returns>
		public bool GetBooleanValue( string fieldName, bool valueDef )
		{
			if( !HasField( fieldName ) )
				return valueDef;

			string fieldValue = GetTextValue( fieldName, null );

			fieldValue = fieldValue.ToLowerInvariant();

			if( BooleanTrueValues.Contains( fieldValue ) )
				return true;

			if( BooleanFalseValues.Contains( fieldValue ) )
				return false;

			string message = string.Format( "The field \"{0}\" has a value \"{1}\" which doesn't contain a valid boolean."
				, fieldName, fieldValue );

			throw new ValidationException( message );
		}

		#endregion Public methods
	}
}
