////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Net;
using Vela.Kernel.Serialization;

namespace Vela.Kernel
{
	/// <summary>
	/// An alternative peer address used for the alternative peer resolution.
	/// </summary>
	public class AlternativePeerAddress : IDocument
	{
		#region Public constants

		/// <summary>
		/// Placeholder which is automatically replaced with the UPN long host label (of the remote peer); this may be used in
		/// <see cref="Address"/>.
		/// </summary>
		public const string UpnLongHostLabelPlaceholder = "{HostLabel}";

		/// <summary>
		/// Placeholder which is automatically replaced with the UPN short host label (of the remote peer); this may be used in
		/// <see cref="Address"/>.
		/// </summary>
		public const string UpnShortHostLabelPlaceholder = "{ShortHostLabel}";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public AlternativePeerAddress()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="address">The alternative peer address (of the remote peer) to use for the alternative peer resolution.
		/// This may be null or empty.</param>
		/// <param name="processingPort">The network port where the local peer must send action requests to the remote peer.
		/// </param>
		public AlternativePeerAddress( string address, UInt16 processingPort )
		{
			Address = address;
			ProcessingPort = processingPort;
		}

		#endregion Constructors

		#region Public serializable properties

		private string _Address;
		/// <summary>
		/// The alternative peer address (of the remote peer) to use for the alternative peer resolution. This may be null or
		/// empty.
		/// </summary>
		/// <remarks>
		/// <para>This may be either an IP address or a host name. If this is a host name, the UPN long host label of the remote
		/// peer will automatically replace the <see cref="UpnLongHostLabelPlaceholder"/> placeholder, while the UPN short host
		/// label of the remote peer will automatically replace the <see cref="UpnShortHostLabelPlaceholder"/> placeholder.
		/// </para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string Address
		{
			get { return _Address; }
			set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.AlternativePeerAddress, "Address" );

				_Address = value;
			}
		}

		private UInt16 _ProcessingPort;
		/// <summary>
		/// The network port where the local peer must send action requests to the remote peer.
		/// </summary>
		/// <remarks>
		/// <para>Normally, action requests are sent to the UPN's processing port, but if this is higher than 0, action requests
		/// are sent to this port instead.</para>
		/// 
		/// <para>This exists in order to allow peers to bypass network restrictions (like firewalls and filters). This is
		/// possible by making all the vaults available for all the ports of a service application, and by registering an action
		/// service listener for each application port.</para>
		/// 
		/// <para>The downside is that the local peer may become unable to communicate with the remote peer (the reverse is still
		/// possible) until the remote peer adds this port in the configuration of his / her storehouse.</para>
		/// </remarks>
		[SerializeAttribute]
		public UInt16 ProcessingPort
		{
			get { return _ProcessingPort; }
			set
			{
				ValidationUtils.Integer( value, 0, UInt16.MaxValue, "ProcessingPort" );

				_ProcessingPort = value;
			}
		}

		#endregion Public serializable properties

		#region Public static methods

		/// <summary>
		/// Checks if the specified alternative peer address contains an IP address.
		/// </summary>
		/// <param name="address">The alternative peer address. This may be null or empty.</param>
		/// <returns>Returns true if the alternative peer address contains an IP address.</returns>
		public static bool HasIpAddress( string address )
		{
			if( TextUtils.NoContent( address ) )
				return false;

			IPAddress ipAddress;
			return IPAddress.TryParse( address, out ipAddress );
		}

		/// <summary>
		/// Checks if the specified alternative peer address contains a host name.
		/// </summary>
		/// <param name="address">The alternative peer address. This may be null or empty.</param>
		/// <returns>Returns true if the alternative peer address contains a host name.</returns>
		public static bool HasHostName( string address )
		{
			if( TextUtils.NoContent( address ) )
				return false;

			// Check for an IP address.
			IPAddress ipAddress;
			IPAddress.TryParse( address, out ipAddress );
			if( ipAddress != null )
				return false;

			// Allow the "UpnLongHostLabelPlaceholder" and "UpnShortHostLabelPlaceholder" placeholders; use a dummy text instead
			// the placeholders.
			string noPlaceholderAddress = address;
			noPlaceholderAddress = TextUtils.Replace( noPlaceholderAddress, UpnLongHostLabelPlaceholder, "x" );
			noPlaceholderAddress = TextUtils.Replace( noPlaceholderAddress, UpnShortHostLabelPlaceholder, "x" );

			// Check for a host name.
			return Uri.CheckHostName( noPlaceholderAddress ) == UriHostNameType.Dns;
		}

		/// <summary>
		/// Checks if the specified alternative peer address contains an IP address or a host name.
		/// </summary>
		/// <param name="address">The alternative peer address to check. This may be null or empty.</param>
		/// <returns>Returns true if the alternative peer address contains an IP address or a host name.</returns>
		public static bool HasAddress( string address )
		{
			if( TextUtils.NoContent( address ) )
				return false;

			// Check for an IP address.
			IPAddress ipAddress;
			IPAddress.TryParse( address, out ipAddress );
			if( ipAddress != null )
				return true;

			// Allow the "UpnLongHostLabelPlaceholder" and "UpnShortHostLabelPlaceholder" placeholders; use a dummy text instead
			// the placeholders.
			string noPlaceholderAddress = address;
			noPlaceholderAddress = TextUtils.Replace( noPlaceholderAddress, UpnLongHostLabelPlaceholder, "x" );
			noPlaceholderAddress = TextUtils.Replace( noPlaceholderAddress, UpnShortHostLabelPlaceholder, "x" );

			// Check for a host name.
			return Uri.CheckHostName( noPlaceholderAddress ) == UriHostNameType.Dns;
		}

		/// <summary>
		/// Gets the formatted host name, where the <see cref="UpnLongHostLabelPlaceholder"/> placeholder is replaced with the
		/// specified UPN host label, and the <see cref="UpnShortHostLabelPlaceholder"/> placeholder is replaced with the
		/// specified UPN short host label.
		/// </summary>
		/// <param name="address">The alternative peer address. This may be null or empty.</param>
		/// <param name="upnLongHostLabel">UPN long host label with which to replace the
		/// <see cref="UpnLongHostLabelPlaceholder"/> placeholder. This may be null or empty.</param>
		/// <param name="upnShortHostLabel">UPN short host label with which to replace the
		/// <see cref="UpnShortHostLabelPlaceholder"/> placeholder. This may be null or empty.</param>
		/// <returns>Returns the formatted host name.</returns>
		public static string GetFormattedHostName( string address, string upnLongHostLabel, string upnShortHostLabel )
		{
			if( !HasHostName( address ) )
			{
				throw new ValidationException( string.Format( "The alternative peer address \"{0}\" doesn't contain a host name."
					, address ) );
			}

			string formattedHostName = address;

			formattedHostName = TextUtils.Replace( formattedHostName, UpnLongHostLabelPlaceholder, upnLongHostLabel );
			formattedHostName = TextUtils.Replace( formattedHostName, UpnShortHostLabelPlaceholder, upnShortHostLabel );

			return formattedHostName;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Updates the current alternative peer address with the data which may override the communication between peers, from
		/// the specified alternative peer address.
		/// </summary>
		/// <param name="newAddress">Alternative peer address from which to get the override data. This may be null.</param>
		public void UpdateOverridingData( AlternativePeerAddress newAddress )
		{
			if( newAddress == null )
				return;

			if( newAddress.HasAddress() )
				Address = newAddress.Address;

			if( newAddress.ProcessingPort > 0 )
				ProcessingPort = newAddress.ProcessingPort;
		}

		/// <summary>
		/// Checks if the alternative peer address contains an IP address.
		/// </summary>
		/// <returns>Returns true if the alternative peer address contains an IP address.</returns>
		public bool HasIpAddress()
		{
			return HasIpAddress( Address );
		}

		/// <summary>
		/// Checks if the alternative peer address contains a host name.
		/// </summary>
		/// <returns>Returns true if the alternative peer address contains a host name.</returns>
		public bool HasHostName()
		{
			return HasHostName( Address );
		}

		/// <summary>
		/// Checks if the alternative peer address contains an IP address or a host name.
		/// </summary>
		/// <returns>Returns true if the alternative peer address contains an IP address or a host name.</returns>
		public bool HasAddress()
		{
			return HasAddress( Address );
		}

		/// <summary>
		/// Gets the alternative IP address.
		/// </summary>
		/// <remarks>Returns the alternative IP address. This may be null.</remarks>
		public IPAddress GetIpAddress()
		{
			if( !HasIpAddress() )
				return null;

			return IPAddress.Parse( Address );
		}

		/// <summary>
		/// Gets the alternative IP address as a buffer.
		/// </summary>
		/// <remarks>Returns the alternative IP address buffer. This may be null or empty.</remarks>
		public byte[] GetIpAddressBuffer()
		{
			if( !HasIpAddress() )
				return null;

			return IPAddress.Parse( Address ).GetAddressBytes();
		}

		/// <summary>
		/// Gets the alternative host name.
		/// </summary>
		/// <remarks>Returns the alternative host name. This may be null or empty.</remarks>
		public string GetHostName()
		{
			if( !HasHostName() )
				return null;

			return Address;
		}

		/// <summary>
		/// Gets the formatted host name, where the <see cref="UpnLongHostLabelPlaceholder"/> placeholder is replaced with the
		/// specified UPN long host label, and the <see cref="UpnShortHostLabelPlaceholder"/> placeholder is replaced with the
		/// specified UPN short host label.
		/// </summary>
		/// <param name="upnLongHostLabel">UPN long host label with which to replace the
		/// <see cref="UpnLongHostLabelPlaceholder"/> placeholder. This may be null or empty.</param>
		/// <param name="upnShortHostLabel">UPN short host label with which to replace the
		/// <see cref="UpnShortHostLabelPlaceholder"/> placeholder. This may be null or empty.</param>
		/// <returns>Returns the formatted host name.</returns>
		public string GetFormattedHostName( string upnLongHostLabel, string upnShortHostLabel )
		{
			return GetFormattedHostName( Address, upnLongHostLabel, upnShortHostLabel );
		}

		#endregion Public methods
	}
}
