////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Reflection;
using System.Text;

namespace Vela.Kernel
{
	/// <summary>
	/// List of portable types.
	/// </summary>
	/// <remarks>
	/// <para>All the enumeration types and the types derived from <see cref="IPortableType"/> are collected in
	/// <see cref="PortableTypes.Instance"/> when this type is first accessed.</para>
	/// 
	/// <para>The portable types are used to deserialize documents without knowing their type at that moment. This is possible
	/// because the serialized form of an object contains metadata with the name of each serialized type (and that name is mapped
	/// to a type from <see cref="PortableTypes.Instance"/>).</para>
	/// 
	/// <para>All the assemblies (directly or indirectly referenced by the application's entry assembly) are loaded (into the
	/// application domain) when this class is first used. Some assemblies used by the application may not be referenced by the
	/// application's entry assembly, so you have to either reference them in the entry assembly or you have register them
	/// manually (<see cref="RegisterGlobal( Assembly )"/>).</para>
	/// 
	/// <para>All the assemblies referenced by the entry assembly are searched, recursively, for portable types. Be aware
	/// that only the <b>used</b> assemblies are searched, not those that are simply referenced. If you want a specific assembly
	/// to be searched, you must at least reference (in code) a type declared in it.</para>
	/// </remarks>
	public class PortableTypes : KernelSortedList<string, Type>
	{
		#region Public constants

		/// <summary>
		/// Template for displaying the a type to the user. Placeholders: 0 - type name.
		/// </summary>
		public const string DisplayItemTemplate = "\"{0}\"";

		/// <summary>
		/// Item separator.
		/// </summary>
		public const string ItemSeparator = ", ";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public PortableTypes()
			: base( StringInsensitiveComparer.Instance )
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="types">Portable types to add to the list.</param>
		public PortableTypes( params Type[] types )
			: base( StringInsensitiveComparer.Instance )
		{
			ValidationUtils.ArrayHasContent( types, "types" );

			foreach( Type type in types ) Add( type );
		}

		#endregion Constructors

		#region Private static fields

		/// <summary>
		/// The synchronization object for the <see cref="Instance"/> property. READ-ONLY.
		/// </summary>
		private static readonly object InstanceLock = new object();

		#endregion Private static fields

		#region Private static properties

		private static PortableTypes _Instance;
		/// <summary>
		/// The global list of portable types.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// 
		/// <para>This property uses lazy loading. The most know types (<see cref="RegisterGlobalKnownTypes"/>), the enumeration
		/// types and the types which derive from <see cref="IPortableType"/> (see <see cref="RegisterGlobalAssemblies"/>) are
		/// registered during the lazy loading.</para>
		/// </remarks>
		protected static PortableTypes Instance
		{
			get
			{
				lock( InstanceLock )
				{
					if( _Instance != null ) return _Instance;

					_Instance = new PortableTypes();

					RegisterGlobalKnownTypes();
					RegisterGlobalAssemblies();

					return _Instance;
				}
			}
		}

		#endregion Private static properties

		#region Private static methods

		/// <summary>
		/// Checks if the specified type may be registered as a portable type.
		/// </summary>
		/// <param name="type">Type to check if it may be registered as a portable type.</param>
		/// <returns>Returns true if the specified type may be registered as a portable type.</returns>
		private static bool IsRegistrable( Type type )
		{
			ValidationUtils.ObjectHasContent( type, "type" );

			if( type.IsEnum ) return true;

			// The class must have an instance constructor with no parameters.
			if( type.IsVisible && type.IsClass && !type.IsAbstract && typeof( IPortableType ).IsAssignableFrom( type )
				&& (
					type.GetConstructor( BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance, null
						, Type.EmptyTypes, null ) != null
				) )
			{
				return true;
			}

			if( ScalarEncoder.IsScalar( type ) ) return true;

			return false;
		}

		/// <summary>
		/// Finds the ancestor class type with the specified name, from which the specified descendent type is derived.
		/// </summary>
		/// <param name="ancestorTypeName">The name of the ancestor type to search for.</param>
		/// <param name="descendentType">Descendent type to look into.</param>
		/// <returns>Returns the found type, or null if none is found.</returns>
		/// <remarks>
		/// <para>The parameters are not validated.</para>
		/// </remarks>
		private static Type FindAncestorClass( string ancestorTypeName, Type descendentType )
		{
			// Get the base type.
			Type baseType = descendentType.BaseType;
			if( baseType == null ) return null; // Ancestor type not found.

			if( TextUtils.Equals( baseType.FullName, ancestorTypeName ) )
				return baseType; // Ancestor type found.

			// Ancestor type not found.
			return FindAncestorClass( ancestorTypeName, baseType );
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Registers globally the types from the specified assembly. If a type with the same name already exists, an exception
		/// is thrown.
		/// </summary>
		/// <param name="assembly">The assembly in which to look for types.</param>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static void RegisterGlobal( Assembly assembly )
		{
			ValidationUtils.ObjectHasContent( assembly, "assembly" );

			// Get all the types of the assembly.
			Type[] assemblyTypes = assembly.GetTypes();

			foreach( Type type in assemblyTypes )
			{
				if( !IsRegistrable( type ) ) continue;

				lock( InstanceLock )
				{
					// Register the type.
					Instance.Add( type );
				}
			}
		}

		/// <summary>
		/// Registers globally the types from the specified assembly. If a type with the same name already exists, an exception
		/// is thrown.
		/// </summary>
		/// <param name="assemblyName">The name of the assembly in which to look for types.</param>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static void RegisterGlobal( string assemblyName )
		{
			ValidationUtils.TextHasContent( assemblyName, "assemblyName" );

			RegisterGlobal( Assembly.Load( assemblyName ) );
		}

		/// <summary>
		/// Registers globally the specified type with the specified key. If a type with the same name already exists, it's
		/// updated.
		/// </summary>
		/// <param name="type">The type to register globally.</param>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static void RegisterGlobal( Type type )
		{
			ValidationUtils.ObjectHasContent( type, "type" );

			if( !IsRegistrable( type ) )
			{
				throw new ValidationException( string.Format( "The type \"{0}\" to register as a global portable type is"
					+ " invalid.", type.FullName ) );
			}

			lock( InstanceLock )
			{
				Instance.AddOrUpdate( type );
			}
		}

		/// <summary>
		/// Registers globally the most known types.
		/// </summary>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static void RegisterGlobalKnownTypes()
		{
			List<Type> types = ScalarEncoder.GetBasicScalarTypes();

			foreach( Type type in types )
				RegisterGlobal( type );
		}

		/// <summary>
		/// Registers globally all the types from the application which are enumeration types or the types derived from
		/// <see cref="IPortableType"/>. If a type with the same name already exists, an exception is thrown.
		/// </summary>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static void RegisterGlobalAssemblies()
		{
			Dictionary<string, Assembly> assemblies = AssemblyUtils.GetApplicationAssemblies();

			// Register the types from all the assemblies.
			foreach( KeyValuePair<string, Assembly> kvp in assemblies ) RegisterGlobal( kvp.Value );
		}

		/// <summary>
		/// Finds a globally portable type with the specified name.
		/// </summary>
		/// <param name="typeName">The name of the type to search for.</param>
		/// <returns>Returns the found type, or null if none is found.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static Type FindGlobal( string typeName )
		{
			lock( InstanceLock )
			{
				return Instance.Find( typeName );
			}
		}

		/// <summary>
		/// Finds a globally portable type with the same key as the specified type.
		/// </summary>
		/// <param name="type">The type to search for.</param>
		/// <returns>Returns the found type, or null if none is found.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static Type FindGlobal( Type type )
		{
			lock( InstanceLock )
			{
				return FindGlobal( Instance.GetKey( type ) );
			}
		}

		/// <summary>
		/// Gets a globally portable type with the specified name. An exception is thrown if a type with the specified name
		/// isn't found and "throwIfNotFound" is true.
		/// </summary>
		/// <param name="typeName">The name of the type to search for.</param>
		/// <param name="ancestorType">The type from which the found type must derive. This may be null.</param>
		/// <param name="throwIfNotFound">If this is true then an exception is thrown if a type is not found.</param>
		/// <returns>Returns the found type, or throws an exception if none is found (or is not derived from "ancestorType") and
		/// "throwIfNotFound" is true, else returns null.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static Type GetGlobal( string typeName, Type ancestorType, bool throwIfNotFound )
		{
			ValidationUtils.ObjectHasContent( ancestorType, "ancestorType" );

			Type foundType = FindGlobal( typeName );

			if( foundType == null )
			{
				if( !throwIfNotFound ) return null;

				throw new ValidationException( string.Format( "The globally portable type \"{0}\" was not found."
					, typeName ) );
			}

			if( (ancestorType != null) && !ancestorType.IsAssignableFrom( foundType ) )
			{
				if( !throwIfNotFound ) return null;

				throw new ValidationException( string.Format( "The globally portable type \"{0}\" is not derived from the"
					+ " \"{1}\" type.", typeName, ancestorType.FullName ) );
			}

			return foundType;
		}

		/// <summary>
		/// Finds the ancestor type with the specified name, from which the specified descendent type is derived.
		/// </summary>
		/// <param name="ancestorTypeName">The name of the ancestor type to search for.</param>
		/// <param name="descendentType">Descendent type whose ancestor type to find.</param>
		/// <returns>Returns the found type, or null if none is found.</returns>
		/// <remarks>
		/// <para>The ancestor type is first searched in the globally portable types, and then directly into the descendent
		/// type.</para>
		/// 
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static Type FindAncestor( string ancestorTypeName, Type descendentType )
		{
			ValidationUtils.TextHasContent( ancestorTypeName, "ancestorTypeName" );
			ValidationUtils.ObjectHasContent( descendentType, "descendentType" );

			lock( InstanceLock )
			{
				Type ancestorType = Instance.Find( ancestorTypeName );
				if( (ancestorType != null) && ancestorType.IsAssignableFrom( descendentType ) )
					return ancestorType; // Ancestor type found.

				// Check if the descendent type is derived from an interface.
				ancestorType = descendentType.GetInterface( ancestorTypeName );
				if( ancestorType != null ) return ancestorType; // Ancestor type found.

				// Look directly into the descendent type.
				return FindAncestorClass( ancestorTypeName, descendentType );
			}
		}

		/// <summary>
		/// Finds all the globally portable types which are derived from the specified ancestor type.
		/// </summary>
		/// <param name="ancestorType">Type from which the found types must derive.</param>
		/// <param name="includeAbstract">If this is true then the abstract types are also included in the result.</param>
		/// <returns>Returns the found types. The specified ancestor type is included (if it fulfills the criteria).</returns>
		public static PortableTypes FindAllDerived( Type ancestorType, bool includeAbstract = false )
		{
			ValidationUtils.ObjectHasContent( ancestorType, "ancestorType" );

			lock( InstanceLock )
			{
				PortableTypes foundTypes = new PortableTypes();

				if( includeAbstract || !ancestorType.IsAbstract ) foundTypes.Add( ancestorType );

				// Iterate through the global types.
				foreach( Type currentType in Instance )
				{
					if( (ancestorType != currentType) && ancestorType.IsAssignableFrom( currentType )
						&& (includeAbstract || !currentType.IsAbstract) )
					{
						foundTypes.Add( currentType );
					}
				}

				return foundTypes;
			}
		}

		/// <summary>
		/// Checks if the specified type is an ancestor (or the same) type of all the descendent types from the specified list.
		/// </summary>
		/// <param name="ancestorType">Type from which the derived types must derive.</param>
		/// <param name="descendentTypes">Descendent types to check for ancestor.</param>
		/// <returns>Returns true if the specified type is an ancestor (or the same) type of all the descendent types from the
		/// specified list.</returns>
		public static bool IsAncestor( Type ancestorType, PortableTypes descendentTypes )
		{
			ValidationUtils.ObjectHasContent( ancestorType, "ancestorType" );
			ValidationUtils.ListHasContent( descendentTypes, "types" );

			foreach( Type type in descendentTypes )
			{
				if( !ancestorType.IsAssignableFrom( type ) ) return false; // Different ancestor type.
			}

			// The ancestor type is the same.
			return true;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Gets the textual representation of all the items from the list.
		/// </summary>
		/// <returns>Returns the textual representation of all the items from the list.</returns>
		public override string ToString()
		{
			StringBuilder sb = new StringBuilder();

			int i = 0;
			foreach( Type currentItem in this )
			{
				sb.AppendFormat( DisplayItemTemplate, currentItem.FullName );
				if( i < Count - 1 ) sb.Append( ItemSeparator );

				i++;
			}

			return sb.ToString();
		}

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para>The full type name is used as key.</para>
		/// </remarks>
		public override string GetKey( Type item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.FullName;
		}

		/// <summary>
		/// Finds either a type from the current object which has the specified name, or a a globally portable type with the
		/// specified name which is derived from one of the types from the current object.
		/// </summary>
		/// <param name="typeName">The name of the type to search for.</param>
		/// <returns>Returns the found type.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public Type FindDerived( string typeName )
		{
			ValidationUtils.TextHasContent( typeName, "typeName" );

			// Try to find a type with the specified name.
			Type type = Find( typeName );
			if( type != null ) return type;

			// Try to find a global type which derives from one of the types from the current object, and which has the
			// specified name.
			lock( InstanceLock )
			{
				// Try to find a global type with the specified name.
				Type globalType = Instance.Find( typeName );
				if( globalType == null ) return null; // No type found.

				// Iterate through the types from the current object in order to check if the global type is derived from a
				// type from the current object.
				foreach( Type currentItem in this )
				{
					if( currentItem.IsAssignableFrom( globalType ) ) return globalType;
				}
			}

			// No type found.
			return null;
		}

		/// <summary>
		/// Finds either a type from the current object which has the same key as the specified type, or a globally registered
		/// type with the same key as the specified type which is derived from one of the types from the current object.
		/// </summary>
		/// <param name="type">The type to search for.</param>
		/// <returns>Returns the found type.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public Type FindDerived( Type type )
		{
			return FindDerived( GetKey( type ) );
		}

		/// <summary>
		/// Checks if one of the types from the current list is the same ancestor type of all the others.
		/// </summary>
		/// <returns>Returns true if one of the types from the current list is the same ancestor type of all the others.
		/// </returns>
		public bool SameCommonAncestor()
		{
			foreach( Type currentItem in this )
			{
				if( IsAncestor( currentItem, this ) ) return true; // The ancestor type is the same.
			}

			// Different ancestor type.
			return true;
		}

		#endregion Public methods
	}
}
