////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel.Serialization;

namespace Vela.Kernel
{
	/// <summary>
	/// Contract for a registrable algorithm.
	/// </summary>
	/// <remarks>
	/// <para>A registrable algorithm is an algorithm which must be found (globally) by the short name of its type.</para>
	/// 
	/// <para>All registrable algorithms must derive from this interface.</para>
	/// 
	/// <para>The short name of the type of a registrable algorithm must be unique.</para>
	/// 
	/// <para>Registrable algorithms are of two kinds:</para>
	/// <list type="bullet">
	///		<item>Immutable: A portable type which may NOT contain fields which could be modified directly (by assignment)
	///			or indirectly through (the parameters of) a public method, after the algorithm was object created. The
	///			non-serializable fields may be modified indirectly (as a consequence of internal processing). The serializable
	///			fields may be modified in the serialized form of the algorithm, but this isn't relevant because each
	///			deserialization creates a new instance of the algorithm. The immutability must be deep. Since the
	///			non-serializable fields may be modified indirectly, the immutability doesn't provide thread safety.</item>
	///		<item>Mutable: A portable type which may contain fields which could be modified directly (by assignment)
	///			or indirectly through (the parameters of) a public method, after the algorithm object was created.</item>
	/// </list>
	/// 
	/// <para>Note that both kinds of algorithms may have an internal state which could change as the algorithm is being used.
	/// The difference between them is in whether or not they expose something to the outside for modification.</para>
	/// 
	/// <para>The kind (= immutable, mutable) of an algorithm is not enforced by the underlying infrastructure.</para>
	/// 
	/// <para>Immutable algorithms with identical names are identical. Even though the non-serializable fields may be modified
	/// indirectly, a hash serializer would not be able to see the differences between several instances of an immutable
	/// algorithm with the same name (since it can't serialize the non-serializable fields).</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Reusability</b></para>
	/// 
	/// <para>Algorithms are intrinsically meant to be reusable. This means that once a consistent state of the algorithm was
	/// achieved, using the algorithm multiple times with the same input data must produce the same output data. This doesn't
	/// mean that the algorithm must also be synchronized for a multi-threaded environment.</para>
	/// 
	/// <para>The same instance or a clone of an immutable algorithm can be used in a different place without (significant)
	/// difficulties.</para>
	/// 
	/// <para>The same instance or a clone of a mutable algorithm can be reused only to a limited extent because its modifiable
	/// fields may have been changed in one place where the algorithm is used, but the changes should not propagate to another
	/// place where the algorithm is used.</para>
	/// 
	/// <para>Using an immutable algorithm (which has non-serializable properties) from different code flows (of an application)
	/// may still result in state collisions if the algorithm is used in an interlaced manner.</para>
	/// 
	/// <para>An interlaced manner is one where the code flow of an algorithm is interrupted to execute the code flow of
	/// another algorithm which in turn uses the first algorithm, or where the code flow of an algorithm spans several public
	/// methods (like "Begin" and "End") between which a call is made to another algorithm which in turn uses the first
	/// algorithm. Basically, if the internal state of the algorithm has changed from one code flow to the other, there are
	/// inconsistencies in how the algorithm behaves in interlaced uses compared to non-interlaced uses.</para>
	/// 
	/// <para>Using a mutable algorithm from different code flows (of an application) may easily result in state collisions
	/// if the object instance is the same, so mutable algorithms must be used with care.</para>
	/// </remarks>
	public interface IRegistrableAlgorithm : IDocument
	{
	}
}
