////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Reflection;

namespace Vela.Kernel
{
	/// <summary>
	/// Utilities for registered algorithms.
	/// </summary>
	public static class RegistrableAlgorithmUtils
	{
		#region Public constants

		/// <summary>
		/// The name of a static method which is called to get a parsing algorithm instance.
		/// </summary>
		public const string GetParsingAlgorithmMethodName = "GetParsingAlgorithm";

		#endregion Public constants

		#region Public static methods

		/// <summary>
		/// Gets the name of the specified algorithm.
		/// </summary>
		/// <param name="algorithm">The algorithm whose name to get.</param>
		/// <returns>Returns the name of the specified algorithm.</returns>
		public static string GetAlgorithmName( IRegistrableAlgorithm algorithm )
		{
			ValidationUtils.ObjectHasContent( algorithm, "algorithm" );

			return algorithm.GetType().Name;
		}

		/// <summary>
		/// Gets the name of the specified algorithm type.
		/// </summary>
		/// <param name="algorithmType">The type of the algorithm whose name to get.</param>
		/// <returns>Returns the name of the specified algorithm type.</returns>
		public static string GetAlgorithmName( Type algorithmType )
		{
			ValidationUtils.ObjectHasContent( algorithmType, "algorithmType" );

			if( !typeof( IRegistrableAlgorithm ).IsAssignableFrom( algorithmType ) )
			{
				throw new ValidationException( string.Format( "The type \"{0}\" whose algorithm name to get must derive from"
					+ " \"{1}\".", algorithmType.FullName, typeof( IRegistrableAlgorithm ).FullName ) );
			}

			return algorithmType.Name;
		}

		/// <summary>
		/// Gets a parsing algorithm which derives from the specified ancestor type, which can parse the specified data.An
		/// exception is thrown if the data can't be parsed.
		/// </summary>
		/// <param name="data">Data to parse. This may be either a buffer or a text; the buffer may be a binary encoded
		/// text.</param>
		/// <param name="ancestorType">The type from which the parsing algorithm must derive.</param>
		/// <returns>Returns an instance of the parsing algorithm which can parse the specified data.</returns>
		public static IParsingAlgorithm GetParser( object data, Type ancestorType )
		{
			ValidationUtils.BufferOrText( data, "data" );

			RegistrableAlgorithms algs = RegistrableAlgorithms.FindAllGlobal( ancestorType );

			foreach( Type algType in algs )
			{
				MethodInfo eventMethodInfo = null;

				// Check if there is an appropriate method to call.
				if( data is byte[] )
				{
					eventMethodInfo = algType.GetMethod( GetParsingAlgorithmMethodName, BindingFlags.Public
						| BindingFlags.NonPublic | BindingFlags.Static, null, new Type[] { typeof( byte[] ) }, null );
				}
				else if( data is string )
				{
					eventMethodInfo = algType.GetMethod( GetParsingAlgorithmMethodName, BindingFlags.Public
						| BindingFlags.NonPublic | BindingFlags.Static, null, new Type[] { typeof( string ) }, null );
				}
				else
				{
					throw ValidationUtils.Unhandled( data, "data" );
				}

				try
				{
					// Call the "GetParsingAlgorithmMethodName" method in order to determine if the specified data may be parsed.
					IParsingAlgorithm parsingAlg = eventMethodInfo.Invoke( null, new object[] { data } ) as IParsingAlgorithm;
					if( parsingAlg != null ) return parsingAlg;
				}
				catch( Exception ex )
				{
					// The loop must continue (if possible).

					if( ExceptionUtils.IsAbort( ex ) ) throw ExceptionUtils.PreserveStackTrace( ex );
				}
			}

			throw new ValidationException( string.Format( "Can't parse the data with a registered algorithm derived from"
					+ " \"{0}\". Make sure that you and your contacts are using the newest version of the application."
				, ancestorType.FullName ) );
		}

		#endregion Public static methods
	}
}
