////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Reflection;

namespace Vela.Kernel
{
	/// <summary>
	/// List of registrable algorithm (<see cref="IRegistrableAlgorithm"/>) types.
	/// </summary>
	public class RegistrableAlgorithms : KernelSortedList<string, Type>
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public RegistrableAlgorithms()
			: base( StringInsensitiveComparer.Instance )
		{
		}

		#endregion Constructors

		#region Private static fields

		/// <summary>
		/// The synchronization object for the <see cref="Instance"/> property. READ-ONLY.
		/// </summary>
		private static readonly object InstanceLock = new object();

		#endregion Private static fields

		#region Private static properties

		private static RegistrableAlgorithms _Instance;
		/// <summary>
		/// The global registered algorithm types. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// 
		/// <para>This property uses lazy loading. All algorithm types are registered during the lazy loading.</para>
		/// </remarks>
		private static RegistrableAlgorithms Instance
		{
			get
			{
				lock( InstanceLock )
				{
					if( _Instance != null ) return _Instance;

					_Instance = new RegistrableAlgorithms();

					RegisterGlobalAssemblies();

					return _Instance;
				}
			}
		}

		#endregion Private static properties

		#region Private static methods

		/// <summary>
		/// Checks if the specified type may be registered as an algorithm type.
		/// </summary>
		/// <param name="type">Type to check if it may be registered as an algorithm type.</param>
		/// <returns>Returns true if the specified type may be registered as an algorithm type.</returns>
		private static bool IsRegistrable( Type type )
		{
			ValidationUtils.ObjectHasContent( type, "type" );

			// The class must be non-abstract and must have an instance constructor with no parameters.
			return type.IsVisible && type.IsClass && !type.IsAbstract
				&& typeof( IRegistrableAlgorithm ).IsAssignableFrom( type )
				&& (
					type.GetConstructor( BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance, null
						, Type.EmptyTypes, null ) != null
				);
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Registers globally all the algorithm types from the specified assembly. If an algorithm type with the same name
		/// already exists, an exception is thrown.
		/// </summary>
		/// <param name="assembly">The assembly in which to look for algorithm types.</param>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static void RegisterGlobal( Assembly assembly )
		{
			ValidationUtils.ObjectHasContent( assembly, "assembly" );

			// Get all the types of the assembly.
			Type[] assemblyTypes = assembly.GetTypes();

			foreach( Type type in assemblyTypes )
			{
				if( !IsRegistrable( type ) ) continue;

				lock( InstanceLock )
				{
					// Register the algorithm type.
					Instance.Add( type );
				}
			}
		}

		/// <summary>
		/// Registers globally all the algorithm types from the assembly with the specified name. If an algorithm type with the
		/// same name already exists, an exception is thrown.
		/// </summary>
		/// <param name="assemblyName">The name of the assembly in which to look for algorithm types.</param>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static void RegisterGlobal( string assemblyName )
		{
			ValidationUtils.TextHasContent( assemblyName, "assemblyName" );

			RegisterGlobal( Assembly.Load( assemblyName ) );
		}

		/// <summary>
		/// Registers globally the specified algorithm type with the specified key. If an algorithm type with the same name
		/// already exists, it's updated.
		/// </summary>
		/// <param name="algorithmType">The algorithm type to register globally.</param>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static void RegisterGlobal( Type algorithmType )
		{
			ValidationUtils.ObjectHasContent( algorithmType, "algorithmType" );

			if( !IsRegistrable( algorithmType ) )
			{
				throw new ValidationException( string.Format( "The type \"{0}\" to register as a global algorithm type"
					+ " is invalid.", algorithmType.FullName ) );
			}

			lock( InstanceLock )
			{
				Instance.AddOrUpdate( algorithmType );
			}
		}

		/// <summary>
		/// Registers globally all the algorithm type from the application (which derive from
		/// <see cref="IRegistrableAlgorithm"/>). If an algorithm type with the same name already exists, an exception is
		/// thrown.
		/// </summary>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static void RegisterGlobalAssemblies()
		{
			Dictionary<string, Assembly> assemblies = AssemblyUtils.GetApplicationAssemblies();

			// Register the algorithm types from all the assemblies.
			foreach( KeyValuePair<string, Assembly> kvp in assemblies ) RegisterGlobal( kvp.Value );
		}

		/// <summary>
		/// Finds all globally registered algorithm types derived from the specified type.
		/// </summary>
		/// <param name="ancestorType">The type from which the found algorithm type must derive.</param>
		/// <returns>Returns the found algorithm types.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static RegistrableAlgorithms FindAllGlobal( Type ancestorType )
		{
			ValidationUtils.ObjectHasContent( ancestorType, "ancestorType" );

			lock( InstanceLock )
			{
				RegistrableAlgorithms algorithms = new RegistrableAlgorithms();

				foreach( Type type in Instance )
				{
					if( ancestorType.IsAssignableFrom( type ) ) algorithms.Add( type );
				}

				return algorithms;
			}
		}

		/// <summary>
		/// Finds a globally registered algorithm type with the specified name.
		/// </summary>
		/// <param name="algorithmName">The name of the algorithm type to search for.</param>
		/// <returns>Returns the found algorithm type, or null if none is found.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static Type FindGlobal( string algorithmName )
		{
			lock( InstanceLock )
			{
				return Instance.Find( algorithmName );
			}
		}

		/// <summary>
		/// Gets a globally registered algorithm type with the specified name. An exception is thrown if an algorithm type with
		/// the specified name isn't found.
		/// </summary>
		/// <param name="algorithmName">The name of the algorithm type to search for.</param>
		/// <param name="ancestorType">The type from which the found algorithm type must derive. This may be null.</param>
		/// <returns>Returns the found algorithm type, or throws an exception if none is found.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static Type GetGlobal( string algorithmName, Type ancestorType )
		{
			ValidationUtils.ObjectHasContent( ancestorType, "ancestorType" );

			Type foundType = FindGlobal( algorithmName );

			if( foundType == null )
			{
				throw new ValidationException( string.Format( "The globally registered algorithm type \"{0}\" was not found."
					, algorithmName ) );
			}

			if( (ancestorType != null) && !ancestorType.IsAssignableFrom( foundType ) )
			{
				throw new ValidationException( string.Format( "The globally registered algorithm type \"{0}\" is not derived"
					+ " from the \"{1}\" type.", algorithmName, ancestorType.FullName ) );
			}

			return foundType;
		}

		/// <summary>
		/// Instantiates a globally registered algorithm type with the specified name. An exception is thrown if an algorithm
		/// type with the specified name isn't found.
		/// </summary>
		/// <param name="algorithmName">The name of the algorithm type to instantiate.</param>
		/// <param name="ancestorType">The type from which the algorithm type to instantiate must derive.</param>
		/// <param name="args">Parameters to pass to the constructor of the algorithm type to instantiate. This may be null or
		/// empty.</param>
		/// <returns>Returns the instantiated algorithm, or throws an exception if none is found.</returns>
		/// <remarks>
		/// <para>The <see cref="Instance"/> property is thread safe within this method.</para>
		/// </remarks>
		public static IRegistrableAlgorithm InstantiateGlobal( string algorithmName, Type ancestorType, params object[] args )
		{
			Type algType = GetGlobal( algorithmName, ancestorType );

			return (IRegistrableAlgorithm)Activator.CreateInstance( algType, args );
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para><see cref="RegistrableAlgorithmUtils.GetAlgorithmName( Type )"/> is used to generate the key.</para>
		/// </remarks>
		public override string GetKey( Type item )
		{
			return RegistrableAlgorithmUtils.GetAlgorithmName( item );
		}

		#endregion Public methods
	}
}
