////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;

namespace Vela.Kernel.Serialization
{
	/// <summary>
	/// May be applied on properties to indicate that they can be (de)serialized.
	/// </summary>
	/// <remarks>
	/// <para>The serialization name of the property on which this attribute is applied is not its declared name, but the
	/// <see cref="SerializationName"/> value.</para>
	/// 
	/// <para>You must ensure that each property from a class has a unique serialization name. If you don't, you'll get
	/// serialization exceptions.</para>
	/// </remarks>
	[AttributeUsage( AttributeTargets.Property, AllowMultiple = false )]
	public class SerializeAttribute : Attribute
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public SerializeAttribute()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="serializationName">The name which will be used to identify the serialized document in the output of the
		/// serialization. If this is null or empty, the name of the property on which the current attribute is applied, is
		/// used.</param>
		public SerializeAttribute( string serializationName )
		{
			SerializationName = serializationName;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private string _SerializationName;
		/// <summary>
		/// The name which will be used to identify the serialized document in the output of the serialization. If this is null or
		/// empty, the name of the property on which the current attribute is applied, is used.
		/// </summary>
		public string SerializationName
		{
			get { return _SerializationName; }
			set { _SerializationName = value; }
		}

		private string _ListItemName;
		/// <summary>
		/// The list item name which will be used to identify the items of a serialized list object, in the output of the
		/// serialization. This may be null or empty. This is used only if the property on which this attribute is applied, is a
		/// list.
		/// </summary>
		public string ListItemName
		{
			get { return _ListItemName; }
			set { _ListItemName = value; }
		}

		private HashSerializations _HashSerialization = HashSerializations.Always;
		/// <summary>
		/// The mode in which the property on which the current attribute is applied will be serialized with a hash serializer.
		/// </summary>
		public HashSerializations HashSerialization
		{
			get { return _HashSerialization; }
			set { _HashSerialization = value; }
		}

		private object _DontHashSerializeValue;
		/// <summary>
		/// If the property on which the current attribute is applied has this value when it's being hash serialized, it's not
		/// included in the hash serialization. This should be a scalar. This may be null.
		/// </summary>
		public object DontHashSerializeValue
		{
			get { return _DontHashSerializeValue; }
			set { _DontHashSerializeValue = value; }
		}

		private bool _MustBeDeserialized = true;
		/// <summary>
		/// If this is true then the document which is being deserialized must have a node which contains the non-null value of
		/// the property on which the current attribute is applied.
		/// </summary>
		/// <remarks>
		/// <para>The default value of this property is true, so all the properties on which the current attribute is applied
		/// must exist in the deserialized documents. If you want to have optional serializable properties, set this property
		/// (in the attribute applied on the serializable property) with false.</para>
		/// </remarks>
		public bool MustBeDeserialized
		{
			get { return _MustBeDeserialized; }
			set { _MustBeDeserialized = value; }
		}

		private object _EnumFallbackValue;
		/// <summary>
		/// The fallback value for enumerations. If the property on which the current attribute is applied is an enumeration, if
		/// the deserialization of the enumeration fails, this value is assigned instead of an exception being thrown. This may
		/// be null.
		/// </summary>
		public object EnumFallbackValue
		{
			get { return _EnumFallbackValue; }
			set { _EnumFallbackValue = value; }
		}

		#endregion Public non-serializable properties
	}
}
