////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Reflection;

namespace Vela.Kernel
{
	/// <summary>
	/// List of tests (<see cref="Test"/>).
	/// </summary>
	public class Tests : KernelSortedList<string, Test>
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public Tests()
			: base( StringInsensitiveComparer.Instance )
		{
		}

		#endregion Constructors

		#region Private static methods

		/// <summary>
		/// Checks if the specified type is a test.
		/// </summary>
		/// <param name="type">Type to check if it is a test.</param>
		/// <returns>Returns true if the specified type is a test.</returns>
		private static bool IsTest( Type type )
		{
			ValidationUtils.ObjectHasContent( type, "type" );

			// The class must be non-abstract and must have an instance constructor with no parameters, and must be derived
			// from "Test".
			return type.IsVisible && type.IsClass && !type.IsAbstract
				&& typeof( Test ).IsAssignableFrom( type )
				&& (
					type.GetConstructor( BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance, null
						, Type.EmptyTypes, null ) != null
				);
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Collects the tests from the specified assembly.
		/// </summary>
		/// <param name="assembly">Assembly from which to collect the tests.</param>
		/// <returns>Returns the collected tests.</returns>
		public static Tests Collect( Assembly assembly )
		{
			ValidationUtils.ObjectHasContent( assembly, "assembly" );

			Tests tests = new Tests();

			// Get all the types of the assembly.
			Type[] assemblyTypes = assembly.GetTypes();

			foreach( Type type in assemblyTypes )
			{
				if( !IsTest( type ) ) continue;

				// Create the test and add it to the list.
				Test test = (Test)Activator.CreateInstance( type );
				tests.Add( test );
			}

			return tests;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para>The <see cref="Test.SortKey"/> of the item is used as key.</para>
		/// </remarks>
		public override string GetKey( Test item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.SortKey;
		}

		#endregion Public methods
	}
}
