////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Text;

namespace Vela.Kernel
{
	/// <summary>
	/// Contract for a text encoding.
	/// </summary>
	public abstract class TextEncoding : IImmutableAlgorithm
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected TextEncoding()
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		private Encoding _Ute;
		/// <summary>
		/// The underlying text encoding. READ-ONLY.
		/// </summary>
		public Encoding Ute
		{
			get { return _Ute; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Ute" );

				_Ute = value;
			}
		}

		private Encoding _UteNoInvalidException;
		/// <summary>
		/// The underlying text encoding. This must not throw an exception if a buffer to convert to text contains invalid
		/// characters. READ-ONLY.
		/// </summary>
		public Encoding UteNoInvalidException
		{
			get { return _UteNoInvalidException; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "UteNoInvalidException" );

				_UteNoInvalidException = value;
			}
		}

		#endregion Public non-serializable properties

		#region Private static methods

		/// <summary>
		/// Verifies if the preamble of the specified text encoding starts the specified buffer.
		/// </summary>
		/// <param name="buffer">Buffer in which to look for the preamble of the specified text encoding.</param>
		/// <param name="lookTextEncoding">Text encoding whose preamble to look for.</param>
		/// <param name="lastMatchPreambleSize">The size of the last preamble which was found to start the specified byte
		/// array.</param>
		/// <param name="preambleSize">Is set with the size of the preamble of the specified text encoding.</param>
		/// <returns>Returns true if the preamble of the specified text encoding is found to start the specified buffer.
		/// </returns>
		/// <remarks>
		/// <para>The parameters are not validated.</para>
		/// </remarks>
		private static bool VerifyTextEncoding( byte[] buffer, TextEncoding lookTextEncoding, ref int lastMatchPreambleSize
			, out int preambleSize )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );
			ValidationUtils.ObjectHasContent( lookTextEncoding, "lookTextEncoding" );

			preambleSize = 0;

			// Get the encoding's preamble.
			byte[] preamble = lookTextEncoding.Ute.GetPreamble();
			int extractedPreambleSize = preamble.Length;

			// Skip preambles with a size equal with (or smaller than) the last matching preamble.
			if( extractedPreambleSize <= lastMatchPreambleSize ) return false;

			// The buffer must be longer than the encoding's preamble.
			if( buffer.Length < extractedPreambleSize ) return false;

			for( int i = 0; i < extractedPreambleSize; i++ )
			{
				// If a byte is different, skip the current encoding.
				if( preamble[ i ] != buffer[ i ] ) return false;
			}

			// Preamble found.

			preambleSize = extractedPreambleSize;
			lastMatchPreambleSize = preambleSize;

			return true;
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Gets the text encoding from the name extracted from the specified buffer.
		/// </summary>
		/// <param name="buffer">Buffer from where to extract the text encoding name.</param>
		/// <param name="preambleSize">Is set with the size (in bytes) of the text encoding preamble (which starts the
		/// specified buffer).</param>
		/// <returns>Returns the text encoding from the name extracted from the specified buffer.</returns>
		public static TextEncoding ExtractTextEncoding( byte[] buffer, out int preambleSize )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );

			TextEncoding deepTextEncoding = KernelUtils.GetDeepTextEncoding();

			preambleSize = 0;

			// The default text encoding name.
			if( buffer.Length <= 0 ) return deepTextEncoding;

			// Make sure that the longest preambles are checked (because shorted ones may start with the same bytes as the
			// longer one, and if they are first in the encodings list, they may be picked).
			int lastMatchPreambleSize = 0;

			// First verify the default text encoding.
			if( VerifyTextEncoding( buffer, deepTextEncoding, ref lastMatchPreambleSize, out preambleSize ) )
				return deepTextEncoding;

			// Check if each available encoding starts the buffer.
			Type ancestorType = typeof( TextEncoding );
			RegistrableAlgorithms algorithms = RegistrableAlgorithms.FindAllGlobal( ancestorType );
			foreach( Type algType in algorithms )
			{
				string algName = RegistrableAlgorithmUtils.GetAlgorithmName( algType );
				TextEncoding textEncoding = (TextEncoding)RegistrableAlgorithms.InstantiateGlobal( algName, ancestorType );

				if( VerifyTextEncoding( buffer, textEncoding, ref lastMatchPreambleSize, out preambleSize ) )
					return textEncoding;
			}

			// The default text encoding name.
			return deepTextEncoding;
		}

		/// <summary>
		/// Converts the specified buffer to a text. An exception is thrown if the specified buffer contains invalid characters.
		/// </summary>
		/// <param name="buffer">Buffer to convert to text. If this has a text encoding preamble, the used text encoding is
		/// extracted from it. This may be null or empty.</param>
		/// <param name="textEncoding">Is set with the text encoding used to convert the buffer to text.</param>
		/// <returns>Returns the resulted text. This may be null or empty.</returns>
		public static string ArrayToText( byte[] buffer, out TextEncoding textEncoding )
		{
			// Without the separation between null and empty arrays / texts, objects with empty texts would differ between
			// serialization and deserialization, which would result in exceptions being thrown when saved sheafs are reloaded
			// to be verified (that they have been correctly saved).
			if( buffer == null )
			{
				textEncoding = null;

				return null;
			}
			else if( buffer.Length <= 0 )
			{
				textEncoding = null;

				return "";
			}

			// Extract the text encoding.
			int preambleSizeArg;
			textEncoding = ExtractTextEncoding( buffer, out preambleSizeArg );

			// Get the text from the buffer. Skip the text encoding preamble.
			return textEncoding.Ute.GetString( buffer, preambleSizeArg, buffer.Length - preambleSizeArg );
		}

		/// <summary>
		/// Converts the specified buffer to a text. An exception is thrown if the specified buffer contains invalid characters.
		/// </summary>
		/// <param name="buffer">Buffer to convert to text. If this has a text encoding preamble, the used text encoding is
		/// extracted from it. This may be null or empty.</param>
		/// <returns>Returns the resulted text. This may be null or empty.</returns>
		public static string ArrayToText( byte[] buffer )
		{
			TextEncoding textEncoding;

			return ArrayToText( buffer, out textEncoding );
		}

		/// <summary>
		/// Converts the specified buffer to a text. No exception is thrown if the specified buffer contains invalid characters.
		/// </summary>
		/// <param name="buffer">Buffer to convert to text. If this has a text encoding preamble, the used text encoding is
		/// extracted from it. This may be null or empty.</param>
		/// <param name="textEncoding">Is set with the text encoding used to convert the buffer to text.</param>
		/// <returns>Returns the resulted text. This may be null or empty.</returns>
		public static string TryArrayToText( byte[] buffer, out TextEncoding textEncoding )
		{
			// Without the separation between null and empty arrays / texts, objects with empty texts would differ between
			// serialization and deserialization, which would result in exceptions being thrown when saved sheafs are reloaded
			// to be verified (that they have been correctly saved).
			if( buffer == null )
			{
				textEncoding = null;

				return null;
			}
			else if( buffer.Length <= 0 )
			{
				textEncoding = null;

				return "";
			}

			// Extract the text encoding.
			int preambleSizeArg;
			textEncoding = ExtractTextEncoding( buffer, out preambleSizeArg );

			// Get the text from the buffer. Skip the text encoding preamble.
			return textEncoding.UteNoInvalidException.GetString( buffer, preambleSizeArg, buffer.Length - preambleSizeArg );
		}

		/// <summary>
		/// Converts the specified buffer to a text. No exception is thrown if the specified buffer contains invalid characters.
		/// </summary>
		/// <param name="buffer">Buffer to convert to text. If this has a text encoding preamble, the used text encoding is
		/// extracted from it. This may be null or empty.</param>
		/// <returns>Returns the resulted text. This may be null or empty.</returns>
		public static string TryArrayToText( byte[] buffer )
		{
			TextEncoding textEncoding;

			return TryArrayToText( buffer, out textEncoding );
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Converts the specified text to a buffer.
		/// </summary>
		/// <param name="text">Text to convert to buffer. This may be null or empty.</param>
		/// <param name="addPreamble">If this is true then the preamble of the text encoding is added at the beginning of the
		/// resulted buffer.</param>
		/// <returns>Returns the resulted buffer. This is null if "text" is null, and empty if "text" is empty.</returns>
		public byte[] TextToArray( string text, bool addPreamble )
		{
			if( text == null ) return null;
			else if( text == "" ) return new byte[ 0 ];

			if( addPreamble )
			{
				byte[] preamble = Ute.GetPreamble();

				int byteCount = Ute.GetByteCount( text );
				byte[] buffer = new byte[ preamble.Length + byteCount ];

				Buffer.BlockCopy( preamble, 0, buffer, 0, preamble.Length );

				int writtenCount = Ute.GetBytes( text, 0, text.Length, buffer, preamble.Length );
				if( writtenCount != byteCount )
				{
					throw new ValidationException( string.Format( "The text encoding has converted {0} bytes instead of the"
						+ " calculated {1} bytes.", writtenCount, byteCount ) );
				}

				return buffer;
			}
			else return Ute.GetBytes( text );
		}

		#endregion Public methods
	}
}
