////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using Vela.BuildInfo;

namespace Vela.Kernel
{
	/// <summary>
	/// Assembly utilities.
	/// </summary>
	public static class AssemblyUtils
	{
		#region Public constants

#if DEBUG
		/// <summary>
		/// Template for a log message for build information. Placeholders: 0 - release date, 1 - version.
		/// </summary>
		public const string BuildInfoLogMessageTemplate = "Build information: date {0}, version {1}, DEBUG build";
#else
		/// <summary>
		/// Template for a log message for build information. Placeholders: 0 - release date, 1 - version.
		/// </summary>
		public const string BuildInfoLogMessageTemplate = "Build information: date {0}, version {1}, RELEASE build";
#endif

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		static AssemblyUtils()
		{
			ApplicationAssembly = Assembly.GetEntryAssembly();
		}

		#endregion Constructors

		#region Private static methods

		/// <summary>
		/// Checks if the specified assembly name belongs to the system.
		/// </summary>
		/// <param name="assemblyName">Assembly name to check.</param>
		/// <returns>Returns true if the specified assembly name belongs to the system.
		/// </returns>
		private static bool IsSystemAssembly( AssemblyName assemblyName )
		{
			// Check if the assembly's public key token belongs to the system.
			byte[] keyTokenBuffer = assemblyName.GetPublicKeyToken();
			if( BufferUtils.HasContent( keyTokenBuffer ) )
			{
				string keyToken = TextUtils.ArrayToText16( keyTokenBuffer );
				HashSet<string> systemKeyTokens = new HashSet<string> { "b77a5c561934e089", "31bf3856ad364e35" };
				if( systemKeyTokens.Contains( keyToken.ToLowerInvariant() ) )
					return true;
			}

			// Check if the assembly's name belongs to the system.
			HashSet<string> systemNames = new HashSet<string> { "mscorlib", "windowsbase", "presentationcore"
				, "presentationframework" };
			if( systemNames.Contains( assemblyName.Name.ToLowerInvariant() ) )
				return true;

			// Check if the assembly's name belongs to the system.
			string[] systemPrefixes = new string[] { "microsoft.", "system.", "windows." };
			if( TextUtils.TextStartsWithList( assemblyName.Name, systemPrefixes ) )
				return true;

			return false;
		}

		/// <summary>
		/// Collects all the assemblies referenced by the specified assembly.
		/// </summary>
		/// <param name="assembly">Assembly from which to collect the referenced assemblies.</param>
		/// <param name="assemblies">List where to collect the referenced assemblies.</param>
		private static void CollectReferencedAssemblies( Assembly assembly, Dictionary<string, Assembly> assemblies )
		{
			// Get the referenced assemblies.
			AssemblyName[] assemblyNames = assembly.GetReferencedAssemblies();
			if( assemblyNames == null )
				return;

			// Collect the assemblies.
			foreach( AssemblyName assemblyName in assemblyNames )
			{
				// Check if the assembly was already collected.
				if( assemblies.ContainsKey( assemblyName.FullName ) )
					continue;

				// Check if the assembly belongs to the system.
				if( IsSystemAssembly( assemblyName ) )
					continue;

				// Collect the assembly.
				Assembly loadedAssembly = Assembly.Load( assemblyName );
				assemblies.Add( assemblyName.FullName, loadedAssembly );

				// Recurse.
				CollectReferencedAssemblies( loadedAssembly, assemblies );
			}
		}

		#endregion Private static methods

		#region Public static properties

		private static Assembly _ApplicationAssembly;
		/// <summary>
		/// The assembly which represents the current application.
		/// </summary>
		public static Assembly ApplicationAssembly
		{
			get
			{
				if( _ApplicationAssembly == null )
				{
					throw new ValidationException( "The assembly which represents the current application should be set when"
						+ " the application is started, before this property is read." );
				}

				return _ApplicationAssembly;
			}
			set
			{
				ValidationUtils.ObjectHasContent( value, "ApplicationAssembly" );

				_ApplicationAssembly = value;
			}
		}

		/// <summary>
		/// The folder name of the current application.
		/// </summary>
		public static string ApplicationFolderName
		{
			get
			{
				string fullName = ApplicationAssembly.ManifestModule.FullyQualifiedName;

				return Path.GetDirectoryName( fullName );
			}
		}

		/// <summary>
		/// The file name of the current application.
		/// </summary>
		public static string ApplicationFileName
		{
			get { return ApplicationAssembly.ManifestModule.FullyQualifiedName; }
		}

		/// <summary>
		/// The file title of the current application.
		/// </summary>
		public static string ApplicationFileTitle
		{
			get { return Path.GetFileName( ApplicationAssembly.ManifestModule.FullyQualifiedName ); }
		}

		private static IBuildInfo _ApplicationBuildInfo;
		/// <summary>
		/// The build information of the current application.
		/// </summary>
		public static IBuildInfo ApplicationBuildInfo
		{
			get { return _ApplicationBuildInfo; }
			set
			{
				ValidationUtils.ObjectHasContent( value, "ApplicationBuildInfo" );

				_ApplicationBuildInfo = value;

				string message = string.Format( BuildInfoLogMessageTemplate, value.ReleaseDate, value.Version );
				LogManager.Add( message );
			}
		}

		#endregion Public static properties

		#region Public static methods

		/// <summary>
		/// Gets the name of the data folder of the specified application. If the folder doesn't exist, it's created.
		/// </summary>
		/// <param name="applicationName">The title of the application for which to get the data folder.</param>
		/// <param name="folderPath">Folder path within the application data folder; if this doesn't exist, it's created. This
		/// may be null or empty.</param>
		/// <returns>Returns the name of the data folder of the specified application.</returns>
		public static string GetApplicationDataFolderName( string applicationName, string folderPath )
		{
			ValidationUtils.TextHasContent( applicationName, "applicationName" );

			string dataFolderName = Environment.GetFolderPath( Environment.SpecialFolder.ApplicationData );
			string applicationFolderName = FileUtils.MakeFullName( dataFolderName, applicationName );

			if( TextUtils.HasContent( folderPath ) )
				applicationFolderName = FileUtils.MakeFullName( applicationFolderName, folderPath );

			FileUtils.CreateFolder( applicationFolderName );

			return applicationFolderName;
		}

		/// <summary>
		/// Gets the name of the data folder of the current application. If the folder doesn't exist, it's created.
		/// </summary>
		/// <param name="folderPath">Folder path within the application data folder; if this doesn't exist, it's created. This
		/// may be null or empty.</param>
		/// <returns>Returns the name of the data folder of the current application.</returns>
		public static string GetApplicationDataFolderName( string folderPath )
		{
			return GetApplicationDataFolderName( ApplicationBuildInfo.ProductFullName, folderPath );
		}

		/// <summary>
		/// Gets the name of the data folder of the current application. If the folder doesn't exist, it's created.
		/// </summary>
		/// <returns>Returns the name of the data folder of the current application.</returns>
		public static string GetApplicationDataFolderName()
		{
			return GetApplicationDataFolderName( ApplicationBuildInfo.ProductFullName, null );
		}

		/// <summary>
		/// Gets the name of a file stored in the data folder of the specified application. If the data folder doesn't exist,
		/// it's created.
		/// </summary>
		/// <param name="applicationName">The name of the application in whose data folder in file name has to be.</param>
		/// <param name="folderPath">Folder path within the application data folder; if this doesn't exist, it's created. This
		/// may be null or empty.</param>
		/// <param name="fileTitle">File title for which to get the file name.</param>
		/// <returns>Returns the name of the file stored in the data folder of the specified application.</returns>
		public static string GetApplicationDataFileName( string applicationName, string folderPath, string fileTitle )
		{
			ValidationUtils.TextHasContent( fileTitle, "fileTitle" );

			string applicationFolderName = GetApplicationDataFolderName( applicationName, folderPath );

			return FileUtils.MakeFullName( applicationFolderName, fileTitle );
		}

		/// <summary>
		/// Gets the name of a file stored in the data folder of the current application. If the data folder doesn't exist, it's
		/// created.
		/// </summary>
		/// <param name="folderPath">Folder path within the application data folder; if this doesn't exist, it's created. This
		/// may be null or empty.</param>
		/// <param name="fileTitle">File title for which to get the file name.</param>
		/// <returns>Returns the name of the file stored in the data folder of the current application.</returns>
		public static string GetApplicationDataFileName( string folderPath, string fileTitle )
		{
			return GetApplicationDataFileName( ApplicationBuildInfo.ProductFullName, folderPath, fileTitle );
		}

		/// <summary>
		/// Gets the working folder of the current application. If the working folder doesn't exist, it's created.
		/// </summary>
		/// <param name="folderPath">Folder path within the application working folder; if this doesn't exist, it's created. This
		/// may be null or empty.</param>
		/// <returns>Returns the name of the working folder of the current application.</returns>
		public static string GetApplicationWorkingFolderName( string folderPath )
		{
			string applicationFolderName = ApplicationFolderName;

			if( TextUtils.HasContent( folderPath ) )
				applicationFolderName = FileUtils.MakeFullName( applicationFolderName, folderPath );

			FileUtils.CreateFolder( applicationFolderName );

			return applicationFolderName;
		}

		/// <summary>
		/// Gets the name of a file stored in the working folder of the current application.
		/// </summary>
		/// <param name="folderPath">Folder path within the application working folder; if this doesn't exist, it's created. This
		/// may be null or empty.</param>
		/// <param name="fileTitle">File title for which to get the file name.</param>
		/// <returns>Returns the full name of a file stored in the working folder of the current application.</returns>
		public static string GetApplicationWorkingFileName( string folderPath, string fileTitle )
		{
			ValidationUtils.TextHasContent( fileTitle, "fileTitle" );

			string applicationFolderName = GetApplicationWorkingFolderName( folderPath );

			return FileUtils.MakeFullName( applicationFolderName, fileTitle );
		}

		/// <summary>
		/// Gets the name of a file stored in the temporary folder of the application. If the temporary folder doesn't
		/// exist, it's created.
		/// </summary>
		/// <param name="folderPath">Folder path within the application temporary folder; if this doesn't exist, it's created.
		/// This may be null or empty.</param>
		/// <returns>Returns the name of the file stored in the temporary folder of the application.</returns>
		public static string GetApplicationTempFolderName( string folderPath )
		{
			string tempFolderName = FileUtils.MakeFullName( Path.GetTempPath(), ApplicationBuildInfo.ProductFullName );

			if( TextUtils.HasContent( folderPath ) )
				tempFolderName = FileUtils.MakeFullName( tempFolderName, folderPath );

			FileUtils.CreateFolder( tempFolderName );

			return tempFolderName;
		}

		/// <summary>
		/// Gets the name of a file stored in the temporary folder of the application. If the temporary folder doesn't
		/// exist, it's created.
		/// </summary>
		/// <param name="folderPath">Folder path within the application temporary folder; if this doesn't exist, it's created.
		/// This may be null or empty.</param>
		/// <param name="fileTitle">File title for which to get the file name.</param>
		/// <returns>Returns the name of the file stored in the temporary folder of the application.</returns>
		public static string GetApplicationTempFileName( string folderPath, string fileTitle )
		{
			ValidationUtils.TextHasContent( fileTitle, "fileTitle" );

			string tempFolderName = GetApplicationTempFolderName( folderPath );

			return FileUtils.MakeFullName( tempFolderName, fileTitle );
		}

		/// <summary>
		/// Gets a list with all the assemblies referenced by the application, be they directly or indirectly referenced by the
		/// application's entry assembly.
		/// </summary>
		/// <returns>Returns a list with all the assemblies referenced by the application.
		/// </returns>
		public static Dictionary<string, Assembly> GetApplicationAssemblies()
		{
			Dictionary<string, Assembly> assemblies = new Dictionary<string, Assembly>();

			// Get the entry assembly.
			Assembly currentAssembly = Assembly.GetEntryAssembly();

			// Collect the entry assembly.
			AssemblyName currentAssemblyName = currentAssembly.GetName();
			assemblies.Add( currentAssemblyName.FullName, currentAssembly );

			// Collect the assembly's referenced assemblies.
			CollectReferencedAssemblies( currentAssembly, assemblies );

			return assemblies;
		}

		#endregion Public static methods
	}
}
