////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Microsoft.Win32;

namespace Vela.Kernel
{
	/// <summary>
	/// Application auto-start utilities.
	/// </summary>
	/// <remarks>
	/// <para>An application can be marked to auto-start when the system starts.</para>
	/// </remarks>
	public static class AutoStartUtils
	{
		#region Private constants

		/// <summary>
		/// Template for the command line to be executed when the system starts. Placeholders: 0 - application file name,
		/// 1 - command line parameter which indicates that an auto-start is being executed.
		/// </summary>
		private const string CommandLineTemplate = "\"{0}\" {1}";

		/// <summary>
		/// The path of the registry key which contains the applications marked to auto-start.
		/// </summary>
		private const string RunRegistryKey = @"Software\Microsoft\Windows\CurrentVersion\Run";

		#endregion Private constants

		#region Public constants

		/// <summary>
		/// Command line parameter which indicates that an auto-start is being executed.
		/// </summary>
		public const string AutoStartCommandLineParameter = "-AutoStart";

		#endregion Public constants

		#region Public static methods

		/// <summary>
		/// Gets the auto-start mode of the current application.
		/// </summary>
		/// <returns>Returns the auto-start mode of the current application. True is returned only if the application's path
		/// stored in the registry is the same as the path of the current application.
		/// </returns>
		public static bool Get()
		{
			string commandLine = string.Format( CommandLineTemplate, AssemblyUtils.ApplicationFileName, AutoStartCommandLineParameter );
			string appName = AssemblyUtils.ApplicationBuildInfo.ProductFullName;

			using( RegistryKey key = Registry.CurrentUser.OpenSubKey( RunRegistryKey, false ) )
			{
				if( key == null )
					return false;

				string storedLocation = key.GetValue( appName ) as string;
				if( storedLocation == null )
					return false;

				return TextUtils.Equals( storedLocation, commandLine );
			}
		}

		/// <summary>
		/// Sets the auto-start mode of the current application.
		/// </summary>
		/// <param name="on">The auto-start mode of the current application is set with this value.</param>
		public static void Set( bool on )
		{
			string commandLine = string.Format( CommandLineTemplate, AssemblyUtils.ApplicationFileName, AutoStartCommandLineParameter );
			string appName = AssemblyUtils.ApplicationBuildInfo.ProductFullName;

			if( on )
			{
				using( RegistryKey key = Registry.CurrentUser.CreateSubKey( RunRegistryKey ) )
				{
					key.SetValue( appName, commandLine );
				}
			}
			else
			{
				using( RegistryKey key = Registry.CurrentUser.CreateSubKey( RunRegistryKey ) )
				{
					key.DeleteValue( appName, false );
				}
			}
		}

		/// <summary>
		/// Checks if the specified command line parameters indicate that an auto-start is being executed.
		/// </summary>
		/// <param name="commandLineParameters">Command line parameters in which to look. This may be null or empty.</param>
		/// <returns>Returns true if the command line parameters indicate that an auto-start is being executed.</returns>
		public static bool IsAutoStarting( string[] commandLineParameters )
		{
			if( ListUtils.NoContent( commandLineParameters ) )
				return false;

			return TextUtils.Equals( commandLineParameters[ 0 ], AutoStartCommandLineParameter );
		}

		#endregion Public static methods
	}
}
