////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;

namespace Vela.Kernel
{
	/// <summary>
	/// Bit utilities.
	/// </summary>
	public static class BitUtils
	{
		#region Public static methods

		/// <summary>
		/// Returns the index (from 0) of the most significant bit 1 of the specified value.
		/// </summary>
		/// <param name="value">Value for whose most significant bit 1 to look. If this is 0, the result is -1.</param>
		/// <returns>Returns the index (from 0) of the most significant bit 1.</returns>
		public static int HighestBit1( int value )
		{
			if( value == 0 ) return -1;

			int bits = sizeof( int ) * 8;
			for( int i = bits - 1; i >= 0; i-- )
			{
				if( (value & (1 << i)) != 0 ) return i;
			}

			throw new ValidationException( "The value has no bit set to 1." );
		}

		/// <summary>
		/// Returns the number of bits which are 1, from the specified value.
		/// </summary>
		/// <param name="value">Value from which to count the 1 bits.</param>
		/// <returns>Returns the number of bits which are 1.</returns>
		public static int Bit1Count( int value )
		{
			const int bitCount = sizeof( int ) * 8;

			// Shifting a negative number to the right, fills the empty space with "1" bits, so we must use an unsigned integer.
			UInt32 uncheckedValue = unchecked( (UInt32)value );

			int count = 0;

			for( int i = 0; i < bitCount; i++ )
			{
				if( (uncheckedValue & 1) == 1 ) count++;
				uncheckedValue >>= 1;
			}

			return count;
		}

		/// <summary>
		/// Rotates the bits toward the left with the specified value.
		/// </summary>
		/// <param name="value">The value to rotate.</param>
		/// <param name="shift">The number of bits to rotate.</param>
		/// <returns>Returns the rotated value.</returns>
		/// <remarks>
		/// <para>Shifting a negative number to the right, fills the empty space with "1" bits, so we must use an unsigned
		/// integer.</para>
		/// </remarks>
		public static UInt32 RotateLeft( UInt32 value, int shift )
		{
			return (value << shift) | (value >> (32 - shift));
		}

		/// <summary>
		/// Rotates the bits toward the left with the specified value.
		/// </summary>
		/// <param name="value">The value to rotate.</param>
		/// <param name="shift">The number of bits to rotate.</param>
		/// <returns>Returns the rotated value.</returns>
		/// <remarks>
		/// <para>Shifting a negative number to the right, fills the empty space with "1" bits, so we must use an unsigned
		/// integer.</para>
		/// </remarks>
		public static UInt64 RotateLeft( UInt64 value, int shift )
		{
			return (value << shift) | (value >> (64 - shift));
		}

		/// <summary>
		/// Rotates the bits toward the right with the specified value.
		/// </summary>
		/// <param name="value">The value to rotate.</param>
		/// <param name="shift">The number of bits to rotate.</param>
		/// <returns>Returns the rotated value.</returns>
		/// <remarks>
		/// <para>Shifting a negative number to the right, fills the empty space with "1" bits, so we must use an unsigned
		/// integer.</para>
		/// </remarks>
		public static UInt32 RotateRight( UInt32 value, int shift )
		{
			return (value >> shift) | (value << (32 - shift));
		}

		/// <summary>
		/// Rotates the bits toward the right with the specified value.
		/// </summary>
		/// <param name="value">The value to rotate.</param>
		/// <param name="shift">The number of bits to rotate.</param>
		/// <returns>Returns the rotated value.</returns>
		/// <remarks>
		/// <para>Shifting a negative number to the right, fills the empty space with "1" bits, so we must use an unsigned
		/// integer.</para>
		/// </remarks>
		public static UInt64 RotateRight( UInt64 value, int shift )
		{
			return (value >> shift) | (value << (64 - shift));
		}

		#endregion Public static methods
	}
}
