////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections;

namespace Vela.Kernel
{
	/// <summary>
	/// Buffer utilities.
	/// </summary>
	public static class BufferUtils
	{
		#region Public static methods

		/// <summary>
		/// Checks if the specified buffer has content.
		/// </summary>
		/// <param name="buffer">Buffer to be checked. This may be null or empty.</param>
		/// <returns>Returns true if the specified buffer has content.</returns>
		public static bool HasContent( byte[] buffer )
		{
			return (buffer != null) && (buffer.Length > 0);
		}

		/// <summary>
		/// Checks if the specified buffer is null or empty.
		/// </summary>
		/// <param name="buffer">Buffer to be checked. This may be null or empty.</param>
		/// <returns>Returns true if the specified buffer is null or empty.</returns>
		public static bool NoContent( byte[] buffer )
		{
			return (buffer == null) || (buffer.Length <= 0);
		}

		/// <summary>
		/// Gets a buffer with the specified size, initialized with the specified value.
		/// </summary>
		/// <param name="size">The size of the buffer to get.</param>
		/// <param name="value">The value with which each byte of the buffer to get is initialized.</param>
		/// <returns>Returns the resulted buffer.</returns>
		public static byte[] GetInitialized( int size, byte value )
		{
			byte[] buffer = new byte[ size ];

			for( int i = 0; i < size; i++ ) buffer[ i ] = value;

			return buffer;
		}

		/// <summary>
		/// Verifies if the specified buffers have the same content.
		/// </summary>
		/// <param name="buffer1">Buffer whose content to compare. This may be null or empty.</param>
		/// <param name="buffer2">Buffer whose content to compare. This may be null or empty.</param>
		/// <returns>Returns true if the two buffers have the same content.</returns>
		public static bool Equals( byte[] buffer1, byte[] buffer2 )
		{
			// If both are null, they're equal.
			if( (buffer1 == null) && (buffer2 == null) ) return true;

			// If only one is null, they're not equal.
			if( (buffer1 == null) || (buffer2 == null) ) return false;

			// If they have different sizes, they're not equal.
			if( buffer1.Length != buffer2.Length ) return false;

			int size = buffer1.Length;

			for( int i = 0; i < size; i++ )
			{
				if( buffer1[ i ] != buffer2[ i ] ) return false;
			}

			// Finally, the arrays ar equal.
			return true;
		}

		/// <summary>
		/// Verifies if the specified buffers have the same content.
		/// </summary>
		/// <param name="buffer1">Buffer whose content to compare. This may be null or empty.</param>
		/// <param name="pos1">Index in the first buffer from where to start comparing.</param>
		/// <param name="buffer2">Buffer whose content to compare. This may be null or empty.</param>
		/// <param name="pos2">Index in the second buffer from where to start comparing.</param>
		/// <param name="size">The number of bytes to compare.</param>
		/// <returns>Returns true if the two buffers have the same content.</returns>
		public static bool Equals( byte[] buffer1, int pos1, byte[] buffer2, int pos2, int size )
		{
			// If both are null, they're equal.
			if( (buffer1 == null) && (buffer2 == null) ) return true;

			// If only one is null, they're not equal.
			if( (buffer1 == null) || (buffer2 == null) ) return false;

			// If the first buffer overflows, they're not equal.
			if( pos1 + size > buffer1.Length ) return false;

			// If the second buffer overflows, they're not equal.
			if( pos2 + size > buffer2.Length ) return false;

			for( int i = 0; i < size; i++ )
			{
				if( buffer1[ i + pos1 ] != buffer2[ i + pos2 ] ) return false;
			}

			// Finally, the arrays ar equal.
			return true;
		}

		/// <summary>
		/// Compares the specified buffers.
		/// </summary>
		/// <param name="buffer1">Buffer to compare. This may be null or empty.</param>
		/// <param name="buffer2">Buffer to compare. This may be null or empty.</param>
		/// <returns>Returns 0 if the buffers have the same value, -1 if "buffer1" is smaller than "buffer2", and 1 if
		/// "buffer1"is greater than "buffer2".
		/// </returns>
		public static int Compare( byte[] buffer1, byte[] buffer2 )
		{
			// If both are null, they're equal.
			if( (buffer1 == null) && (buffer2 == null) ) return 0;

			// If the first one is null, the second is greater.
			if( (buffer1 == null) && (buffer2 != null) ) return -1;

			// If the second one is null, the first is greater.
			if( (buffer1 != null) && (buffer2 == null) ) return 1;

			int size = Math.Min( buffer1.Length, buffer2.Length );

			for( int i = 0; i < size; i++ )
			{
				int comparison = buffer1[ i ] - buffer2[ i ];
				if( comparison != 0 ) return comparison;
			}

			// If they are equal so far but have different sizes, the longer one is greater.
			return buffer1.Length - buffer2.Length;
		}

		/// <summary>
		/// Increments the least significant bytes of the specified buffer with 1.
		/// </summary>
		/// <param name="buffer">Buffer to increment.</param>
		/// <param name="offset">The position at which the bytes to increment start. This must be at least 0.</param>
		/// <param name="size">The number of bytes which may be incremented. This must be at least 1.</param>
		/// <param name="allowOverflow">If this is true then the incrementation may overflow (= the incrementation is
		/// performed in a ring).</param>
		public static void Increment( byte[] buffer, int offset, int size, bool allowOverflow )
		{
			ValidationUtils.ObjectHasContent( buffer, "buffer" );
			ValidationUtils.Integer( offset, 0, buffer.Length - 1, "offset" );
			ValidationUtils.Integer( size, 1, buffer.Length - offset, "size" );

			int i = offset;
			while( i < offset + size )
			{
				if( buffer[ i ] < 255 )
				{
					buffer[ i ]++;

					return;
				}

				buffer[ i ] = 0;

				if( i < offset + size - 1 ) i++;
				else
				{
					if( allowOverflow ) i = offset;
					else throw new ValidationException( "Buffer incrementation has overflown." );
				}
			}
		}

		/// <summary>
		/// Increments the least significant bytes of the specified buffer with 1.
		/// </summary>
		/// <param name="buffer">Buffer to increment. This is modified in here.</param>
		/// <param name="allowOverflow">If this is true then the incrementation may overflow (= the incrementation is
		/// performed in a ring).</param>
		public static void Increment( byte[] buffer, bool allowOverflow )
		{
			ValidationUtils.ObjectHasContent( buffer, "buffer" );

			Increment( buffer, 0, buffer.Length, allowOverflow );
		}

		/// <summary>
		/// Divides the least significant bytes of the specified buffer with the specified modulo.
		/// </summary>
		/// <param name="buffer">Buffer to increment.</param>
		/// <param name="modulo">Modulo.</param>
		/// <remarks>
		/// <para>The fragment's size is the size of the <see cref="UInt64"/> type.</para>
		/// </remarks>
		public static UInt64 FragmentModulo( byte[] buffer, UInt64 modulo )
		{
			ValidationUtils.ObjectHasContent( buffer, "buffer" );

			UInt64 dividend = BitConverter.ToUInt64( buffer, 0 );

			return dividend % modulo;
		}

		/// <summary>
		/// Copies the specified source buffer to another buffer, in a contiguous manner.
		/// </summary>
		/// <param name="source">Source buffer to copy.</param>
		/// <param name="dest">Destination buffer where to copy the source buffer.</param>
		public static void ContiguousCopy( byte[] source, byte[] dest )
		{
			ValidationUtils.ObjectHasContent( source, "source" );
			ValidationUtils.ObjectHasContent( dest, "dest" );

			int sourceSize = source.Length;
			int destSize = dest.Length;

			if( sourceSize >= destSize )
			{
				// The size of the source buffer is bigger than (or equal with) the result size, so copy only the beginning.

				Buffer.BlockCopy( source, 0, dest, 0, destSize );
			}
			else
			{
				// The size of the source buffer is smaller than the result size. The source buffer is copied to the result, in a
				// contiguous manner.

				int blocks = destSize / sourceSize;
				int remaining = destSize - sourceSize * blocks;

				// Copy the full blocks.
				int resultOffset = 0;
				for( int i = 0; i < blocks; i++ )
				{
					Buffer.BlockCopy( source, 0, dest, resultOffset, sourceSize );
					resultOffset += sourceSize;
				}

				// Copy the leftover.
				Buffer.BlockCopy( source, 0, dest, resultOffset, remaining );
			}
		}

		/// <summary>
		/// XORs a buffer with another. The arrays must have the same size.
		/// </summary>
		/// <param name="buffer1">Buffer to XOR.</param>
		/// <param name="buffer2">Buffer to XOR.</param>
		/// <returns>Returns the result of the XOR.</returns>
		public static byte[] Xor( byte[] buffer1, byte[] buffer2 )
		{
			ValidationUtils.ObjectHasContent( buffer1, "buffer1" );
			ValidationUtils.ObjectHasContent( buffer2, "buffer2" );

			if( buffer1.Length != buffer2.Length )
			{
				throw new ValidationException( string.Format( "The size of the two arrays to XOR is different: {0} and {1}."
					, buffer1.Length, buffer2.Length ) );
			}

			int size = buffer1.Length;
			byte[] result = new byte[ size ];

			for( int i = 0; i < size; i++ ) result[ i ] = (byte)(buffer1[ i ] ^ buffer2[ i ]);

			return result;
		}

		/// <summary>
		/// XORs a buffer with another, in a contiguous manner. The arrays may have different sizes.
		/// </summary>
		/// <param name="buffer1">Buffer to XOR.</param>
		/// <param name="buffer2">Buffer to XOR.</param>
		/// <param name="size">The number of bytes to XOR. If this is 0 or a negative number, all the bytes (= the higher
		/// number) are XORed.</param>
		/// <returns>Returns the result of the XOR.</returns>
		public static byte[] ContiguousXor( byte[] buffer1, byte[] buffer2, int size = 0 )
		{
			ValidationUtils.ObjectHasContent( buffer1, "buffer1" );
			ValidationUtils.ObjectHasContent( buffer2, "buffer2" );

			int size1 = buffer1.Length;
			int size2 = buffer2.Length;

			int actualSize = (size > 0) ? size : Math.Max( size1, size2 );
			byte[] result = new byte[ actualSize ];

			int i1 = 0;
			int i2 = 0;
			for( int i = 0; i < actualSize; i++ )
			{
				result[ i ] = (byte)(buffer1[ i1 ] ^ buffer2[ i2 ]);

				i1++;
				if( i1 >= size1 ) i1 = 0;

				i2++;
				if( i2 >= size2 ) i2 = 0;
			}

			return result;
		}

		/// <summary>
		/// Adds each item from the specified buffer with the previous, twice.
		/// </summary>
		/// <param name="buffer">Buffer to chain-add.</param>
		/// <returns>Returns the result of the chain-add.</returns>
		public static byte[] ChainAdd( byte[] buffer )
		{
			ValidationUtils.ObjectHasContent( buffer, "buffer" );

			int size = buffer.Length;
			byte[] result = new byte[ size ];

			Buffer.BlockCopy( buffer, 0, result, 0, size );

			unchecked
			{
				// Add the buffer twice.
				for( int c = 0; c < 2; c++ )
				{
					result[ 0 ] += result[ size - 1 ];
					for( int i = 1; i < size; i++ ) result[ i ] += result[ i - 1 ];
				}
			}

			return result;
		}

		/// <summary>
		/// Flips the bytes of the specified buffer.
		/// </summary>
		/// <param name="buffer">Buffer whose bytes to flip.</param>
		/// <returns>Returns the flipped buffer.</returns>
		public static byte[] Flip( byte[] buffer )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );

			int size = buffer.Length;
			byte[] flippedBuffer = new byte[ size ];

			for( int i = 0; i < size; i++ )
				flippedBuffer[ size - i - 1 ] = buffer[ i ];

			return flippedBuffer;
		}

		/// <summary>
		/// Pads with 0s the beginning of the specified buffer.
		/// </summary>
		/// <param name="buffer">Buffer to pad with 0s.</param>
		/// <param name="padToSize">Size of the padded buffer. This must be at least equal with the size of "buffer".</param>
		/// <returns>Returns the padded buffer.</returns>
		public static byte[] PadBeginningWithZeros( byte[] buffer, int padToSize )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );

			int size = buffer.Length;
			ValidationUtils.Integer( padToSize, size, int.MaxValue, "padToSize" );

			if( padToSize <= size ) return buffer;

			byte[] paddedBuffer = new byte[ padToSize ];
			Buffer.BlockCopy( buffer, 0, paddedBuffer, padToSize - size, size );

			return paddedBuffer;
		}

		/// <summary>
		/// Pads with 0s the end of the specified buffer.
		/// </summary>
		/// <param name="buffer">Buffer to pad with 0s.</param>
		/// <param name="padToSize">Size of the padded buffer. This must be at least equal with the size of "buffer".</param>
		/// <returns>Returns the padded buffer.</returns>
		public static byte[] PadEndWithZeros( byte[] buffer, int padToSize )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );

			int size = buffer.Length;
			ValidationUtils.Integer( padToSize, size, int.MaxValue, "padToSize" );

			if( padToSize <= size ) return buffer;

			byte[] paddedBuffer = new byte[ padToSize ];
			Buffer.BlockCopy( buffer, 0, paddedBuffer, 0, size );

			return paddedBuffer;
		}

		/// <summary>
		/// Searches for the specified buffer in the specified list of buffers.
		/// </summary>
		/// <param name="arrayList">List of buffers to search in. This may be empty, but not null.</param>
		/// <param name="buffer">Buffer to search for.</param>
		/// <returns>Returns true if the specified buffer is found.</returns>
		public static bool Find( ArrayList arrayList, byte[] buffer )
		{
			ValidationUtils.ObjectHasContent( arrayList, "arrayList" );
			ValidationUtils.ObjectHasContent( buffer, "buffer" );

			foreach( byte[] array in arrayList )
			{
				if( Equals( array, buffer ) ) return true;
			}

			return false;
		}

		#region Bit level

		/// <summary>
		/// Counts the number of least-significant 0 bits, of the specified buffer.
		/// </summary>
		/// <param name="buffer">Buffer whose 0 bits to count.</param>
		/// <returns>Returns the number of least-significant 0 bits.</returns>
		public static int CountLs0Bits( byte[] buffer )
		{
			ValidationUtils.ObjectHasContent( buffer, "buffer" );

			int zeros = 0;

			int count = buffer.Length;
			for( int i = 0; i < count; i++ )
			{
				byte b = buffer[ i ];

				// Optimize comparison.
				if( b == 0 ) zeros += 8;
				else
				{
					for( int j = 0; j < 8; j++ )
					{
						if( (b & (1 << j)) == 0 ) zeros++;
						else return zeros;
					}
				}
			}

			return zeros;
		}

		#endregion Bit level

		#region Checksums

		/// <summary>
		/// Computes the checksum of the specified buffer. The checksum can be used as hash code in a hash table.
		/// </summary>
		/// <param name="buffer">Buffer whose checksum to compute. This may be empty (but not null).</param>
		/// <param name="size">The number of bytes to include in the checksum. If this is 0 or a negative number, all the
		/// bytes are included.</param>
		/// <returns>Returns the calculated checksum.</returns>
		/// <remarks>
		/// <para>It is guaranteed that the used algorithm generates the same output, regardless of the platform it runs
		/// on.</para>
		/// 
		/// <para>This uses a FNV (Fowler / Noll / Vo) algorithm modified by Bret Mulvey
		/// ( http://papa.bretmulvey.com/post/124027987928/hash-functions ).</para>
		/// </remarks>
		public static Int32 ComputeChecksum32( byte[] buffer, int size = 0 )
		{
			ValidationUtils.ObjectHasContent( buffer, "buffer" );

			int actualSize = (size > 0) ? size : buffer.Length;

			unchecked
			{
				const UInt32 p = 16777619;
				UInt32 checksum = 2166136261;

				for( int i = 0; i < actualSize; i++ ) checksum = (checksum ^ buffer[ i ]) * p;

				checksum += checksum << 13;
				checksum ^= checksum >> 7;
				checksum += checksum << 3;
				checksum ^= checksum >> 17;
				checksum += checksum << 5;

				return (Int32)checksum;
			}
		}

		/// <summary>
		/// Computes the checksum of the specified buffer. The checksum can be used as hash code in a hash table.
		/// </summary>
		/// <param name="buffer">Buffer whose checksum to compute. This may be empty (but not null).</param>
		/// <param name="size">The number of bytes to include in the checksum. If this is 0 or a negative number, all the
		/// bytes are included.</param>
		/// <returns>Returns the calculated checksum.</returns>
		/// <remarks>
		/// <para>It is guaranteed that the used algorithm generates the same output, regardless of the platform it runs
		/// on.</para>
		/// 
		/// <para>This uses a FNV (Fowler / Noll / Vo) algorithm modified by Bret Mulvey
		/// ( http://papa.bretmulvey.com/post/124027987928/hash-functions ).</para>
		/// 
		/// <para>The checksum returned by this method is not equal with the one returned by the <see cref="ComputeChecksum32"/>
		/// method.</para>
		/// </remarks>
		public static Int32 ComputeBlockChecksum32( UInt32[] buffer, int size = 0 )
		{
			ValidationUtils.ObjectHasContent( buffer, "buffer" );

			int actualSize = (size > 0) ? size : buffer.Length;

			unchecked
			{
				const UInt32 p = 16777619;
				UInt32 checksum = 2166136261;

				for( int i = 0; i < actualSize; i++ ) checksum = (checksum ^ buffer[ i ]) * p;

				checksum += checksum << 13;
				checksum ^= checksum >> 7;
				checksum += checksum << 3;
				checksum ^= checksum >> 17;
				checksum += checksum << 5;

				return (Int32)checksum;
			}
		}

		/// <summary>
		/// Computes the checksum of the specified buffer. The checksum can be used as hash code in a hash table.
		/// </summary>
		/// <param name="buffer">Buffer whose checksum to compute. This may be empty (but not null).</param>
		/// <param name="size">The number of bytes to include in the checksum. If this is 0 or a negative number, all the
		/// bytes are included.</param>
		/// <returns>Returns the calculated checksum.</returns>
		/// <remarks>
		/// <para>It is guaranteed that the used algorithm generates the same output, regardless of the platform it runs
		/// on.</para>
		/// 
		/// <para>This uses a FNV (Fowler / Noll / Vo) algorithm modified by Bret Mulvey
		/// ( http://papa.bretmulvey.com/post/124027987928/hash-functions ).</para>
		/// </remarks>
		public static UInt16 ComputeChecksum16( byte[] buffer, int size = 0 )
		{
			Int32 hash = ComputeChecksum32( buffer, size );

			return unchecked( (UInt16)((hash & 0xFFFF) ^ (hash >> 16)) );
		}

		#endregion Checksums

		#endregion Public static methods
	}
}
