////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;

namespace Vela.Kernel
{
	/// <summary>
	/// Conversion utilities.
	/// </summary>
	public static class ConvertUtils
	{
		#region Public static methods

		/// <summary>
		/// Converts the specified buffer to an integer.
		/// </summary>
		/// <param name="buffer">The data to convert.</param>
		/// <param name="offset">Position in the buffer to begin the conversion. This must be at least 0.</param>
		/// <param name="size">The number of bytes (from the buffer) to convert. This must be at least 0.</param>
		/// <returns>Returns the resulted integer.</returns>
		public static int ByteToInt( byte[] buffer, int offset, int size )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );
			ValidationUtils.Integer( offset, 0, buffer.Length - 1, "offset" );
			ValidationUtils.Integer( size, 0, buffer.Length - offset, "size" );

			int result = 0;
			int bitShift = 0;

			for( int i = offset; i < offset + size; i++ )
			{
				result |= (buffer[ i ] << bitShift);

				bitShift += 8;
			}

			return result;
		}

		/// <summary>
		/// Converts an array of <see cref="UInt32"/> values to an array of bytes.
		/// </summary>
		/// <param name="array">The array to convert.</param>
		/// <param name="offset">Position in the array to begin the conversion. This must be at least 0.</param>
		/// <param name="size">How many bytes in the array to convert. This must be at least 0.</param>
		/// <param name="littleEndian">Specifies whether to use little endian. If this is false, big endian is used.</param>
		/// <returns>Returns the <see cref="UInt32"/> values represented as a buffer.
		/// </returns>
		public static UInt32[] ByteToUInt( byte[] array, int offset, int size, bool littleEndian )
		{
			ValidationUtils.ArrayHasContent( array, "array" );
			ValidationUtils.Integer( offset, 0, array.Length - 1, "offset" );
			ValidationUtils.Integer( size, 0, array.Length - offset, "size" );

			if( size % 4 != 0 )
			{
				throw new ValidationException( string.Format( "The number of bytes ({0}) to convert must be a multiple of 4."
					, size ) );
			}

			UInt32[] temp = new UInt32[ size / 4 ];

			if( littleEndian )
			{
				for( int i = 0, j = offset; i < temp.Length; i++ )
				{
					temp[ i ] =
						((UInt32)array[ j++ ] & 0xFF)
						| (((UInt32)array[ j++ ] & 0xFF) << 8)
						| (((UInt32)array[ j++ ] & 0xFF) << 16)
						| (((UInt32)array[ j++ ] & 0xFF) << 24);
				}
			}
			else
			{
				for( int i = 0, j = offset; i < temp.Length; i++ )
				{
					temp[ i ] =
						(((UInt32)array[ j++ ] & 0xFF) << 24)
						| (((UInt32)array[ j++ ] & 0xFF) << 16)
						| (((UInt32)array[ j++ ] & 0xFF) << 8)
						| (((UInt32)array[ j++ ] & 0xFF));
				}
			}

			return temp;
		}

		/// <summary>
		/// Converts an array of <see cref="UInt32"/> values to an array of bytes.
		/// </summary>
		/// <param name="array">The array to convert.</param>
		/// <param name="offset">Position in the array to begin the conversion. This must be at least 0.</param>
		/// <param name="size">How many <see cref="UInt32"/> values are in the array to convert. This must be at least 0.</param>
		/// <param name="littleEndian">Specifies whether to use little endian. If this is false, big endian is used.</param>
		/// <returns>Returns the <see cref="UInt32"/> values represented as a buffer.
		/// </returns>
		public static byte[] UIntToByte( UInt32[] array, int offset, int size, bool littleEndian )
		{
			ValidationUtils.ArrayHasContent( array, "array" );
			ValidationUtils.Integer( offset, 0, array.Length - 1, "offset" );
			ValidationUtils.Integer( size, 0, array.Length - offset, "size" );

			byte[] temp = new byte[ size * 4 ];

			if( littleEndian )
			{
				for( int i = offset; i < offset + size; i++ )
				{
					for( int j = 0; j < 4; j++ )
					{
						temp[ (i - offset) * 4 + j ] = unchecked( (byte)(array[ i ] >> (j * 8)) );
					}
				}
			}
			else
			{
				for( int i = offset; i < offset + size; i++ )
				{
					for( int j = 0; j < 4; j++ )
					{
						temp[ (i - offset) * 4 + (3 - j) ] = unchecked( (byte)(array[ i ] >> (j * 8)) );
					}
				}
			}

			return temp;
		}

		/// <summary>
		/// Converts an array of <see cref="UInt64"/> values to an array of bytes.
		/// </summary>
		/// <param name="array">The array to convert.</param>
		/// <param name="offset">Position in the array to begin the conversion. This must be at least 0.</param>
		/// <param name="size">How many bytes in the array to convert. This must be at least 0.</param>
		/// <param name="littleEndian">Specifies whether to use little endian. If this is false, big endian is used.</param>
		/// <returns>Returns the <see cref="UInt64"/> values represented as a buffer.</returns>
		public static UInt64[] ByteToUInt64( byte[] array, int offset, int size, bool littleEndian )
		{
			ValidationUtils.ArrayHasContent( array, "array" );
			ValidationUtils.Integer( offset, 0, array.Length - 1, "offset" );
			ValidationUtils.Integer( size, 0, array.Length - offset, "size" );

			if( size % 8 != 0 )
			{
				throw new ValidationException( string.Format( "The number of bytes ({0}) to convert must be a multiple of 4."
					, size ) );
			}

			UInt64[] temp = new UInt64[ size / 8 ];

			if( littleEndian )
			{
				for( int i = 0, j = offset; i < temp.Length; i++ )
				{
					temp[ i ] =
						(((UInt64)array[ j++ ] & 0xFF))
						| (((UInt64)array[ j++ ] & 0xFF) << 8)
						| (((UInt64)array[ j++ ] & 0xFF) << 16)
						| (((UInt64)array[ j++ ] & 0xFF) << 24)
						| (((UInt64)array[ j++ ] & 0xFF) << 32)
						| (((UInt64)array[ j++ ] & 0xFF) << 40)
						| (((UInt64)array[ j++ ] & 0xFF) << 48)
						| (((UInt64)array[ j++ ] & 0xFF) << 56);
				}
			}
			else
			{
				for( int i = 0, j = offset; i < temp.Length; i++ )
				{
					temp[ i ] =
						(((UInt64)array[ j++ ] & 0xFF) << 56)
						| (((UInt64)array[ j++ ] & 0xFF) << 48)
						| (((UInt64)array[ j++ ] & 0xFF) << 40)
						| (((UInt64)array[ j++ ] & 0xFF) << 32)
						| (((UInt64)array[ j++ ] & 0xFF) << 24)
						| (((UInt64)array[ j++ ] & 0xFF) << 16)
						| (((UInt64)array[ j++ ] & 0xFF) << 8)
						| (((UInt64)array[ j++ ] & 0xFF));
				}
			}

			return temp;
		}

		/// <summary>
		/// Converts an array of <see cref="UInt64"/> values to an array of bytes.
		/// </summary>
		/// <param name="array">The array to convert.</param>
		/// <param name="offset">Position in the array to begin the conversion. This must be at least 0.</param>
		/// <param name="size">How many <see cref="UInt64"/> values are in the array to convert. This must be at least 0.</param>
		/// <param name="littleEndian">Specifies whether to use little endian. If this is false, big endian is used.</param>
		/// <returns>Returns the <see cref="UInt64"/> values represented as a buffer.</returns>
		public static byte[] UInt64ToByte( UInt64[] array, int offset, int size, bool littleEndian )
		{
			ValidationUtils.ArrayHasContent( array, "array" );
			ValidationUtils.Integer( offset, 0, array.Length - 1, "offset" );
			ValidationUtils.Integer( size, 0, array.Length - offset, "size" );

			byte[] temp = new byte[ size * 8 ];

			if( littleEndian )
			{
				for( int i = offset; i < offset + size; i++ )
				{
					for( int j = 0; j < 8; j++ )
					{
						temp[ (i - offset) * 8 + j ] = unchecked( (byte)(array[ i ] >> (j * 8)) );
					}
				}
			}
			else
			{
				for( int i = offset; i < offset + size; i++ )
				{
					for( int j = 0; j < 8; j++ )
					{
						temp[ (i - offset) * 8 + (7 - j) ] = unchecked( (byte)(array[ i ] >> (j * 8)) );
					}
				}
			}

			return temp;
		}

		/// <summary>
		/// Converts an <see cref="UInt64"/> values to an array of bytes.
		/// </summary>
		/// <param name="data">The <see cref="UInt64"/> values to convert.</param>
		/// <param name="littleEndian">Specifies whether to use little endian. If this is false, big endian is used.</param>
		/// <returns>Returns the <see cref="UInt64"/> values represented as a buffer. </returns>
		public static byte[] UInt64ToByte( UInt64 data, bool littleEndian )
		{
			return UInt64ToByte( new UInt64[ 1 ] { data }, 0, 1, littleEndian );
		}

		#endregion Public static methods
	}
}
