////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Reflection;
using System.Runtime.ExceptionServices;
using System.Text;
using System.Threading;

namespace Vela.Kernel
{
	/// <summary>
	/// Exception utilities.
	/// </summary>
	public static class ExceptionUtils
	{
		#region Public constants

		/// <summary>
		/// Template for an exception detail. Placeholders: 0 - name, 1 - value.
		/// </summary>
		public const string DetailTemplate = "\t* {0}: {1}";

		#endregion Public constants

		#region Private static methods

		/// <summary>
		/// Returns true if the specified exception (not its inner exceptions) should be logged.
		/// </summary>
		/// <param name="exception">Exception to check if it should be logged. This may be null.</param>
		/// <returns>Returns true if the specified exception should be logged.</returns>
		private static bool ShouldLogNonRecursive( Exception exception )
		{
			if( exception == null )
				return false;

			BaseException baseException = exception as BaseException;
			if( baseException != null )
			{
				if( baseException.DoLog )
					return true;
			}
			else
			{
				if( (exception as TargetInvocationException == null)
					&& (exception as ThreadAbortException == null)
					&& (exception as ThreadInterruptedException == null)
					&& (exception as OperationCanceledException == null) )
				{
					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Returns true if the specified exception (not its inner exceptions) should be shown to the user.
		/// </summary>
		/// <param name="exception">Exception to check if it should be shown to the user. This may be null.</param>
		/// <returns>Returns true if the specified exception should be shown to the user.</returns>
		private static bool ShouldShowNonRecursive( Exception exception )
		{
			if( exception == null )
				return false;

			BaseException baseException = exception as BaseException;
			if( baseException != null )
			{
				if( baseException.ShowToUser )
					return true;
			}
			else
			{
				if( (exception as TargetInvocationException == null)
					&& (exception as ThreadAbortException == null)
					&& (exception as ThreadInterruptedException == null)
					&& (exception as OperationCanceledException == null) )
				{
					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Returns true if the specified exception (not its inner exceptions) is critical and the application must be
		/// terminated.
		/// </summary>
		/// <param name="exception">Exception to check if it's critical. This may be null.</param>
		/// <returns>Returns true if the specified exception is critical.</returns>
		private static bool IsCriticalNonRecursive( Exception exception )
		{
			if( exception == null )
				return false;

			return
				// System exceptions.
				(exception as AccessViolationException != null)
				|| (exception as DataMisalignedException != null)
				|| (exception as InsufficientExecutionStackException != null)
				|| (exception as InvalidProgramException != null)
				|| (exception as OutOfMemoryException != null)
				|| (exception as StackOverflowException != null)

				// Vela exceptions.
				|| (exception as CriticalException != null);
		}

		/// <summary>
		/// Checks if it should be avoided to use the specified exception (not its inner exceptions) in a presenter message.
		/// </summary>
		/// <param name="exception">Exception to check. This may be null.</param>
		/// <returns>Returns true if should be avoided to use the specified exception in a presenter message.</returns>
		private static bool ShouldAvoidExceptionInPresenterMessage( Exception exception )
		{
			if( exception == null )
				return true;

			BaseException baseException = exception as BaseException;
			if( (baseException != null) && !baseException.ShowToUser )
				return false;

			return (exception as TargetInvocationException != null)
				|| (exception as ThreadAbortException != null)
				|| (exception as ThreadInterruptedException != null)
				|| (exception as OperationCanceledException != null);
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Returns true if the specified exception (or any of its inner exceptions) should be logged.
		/// </summary>
		/// <param name="exception">Exception to check if it should be logged. This may be null.</param>
		/// <returns>Returns true if the specified exception should be logged.</returns>
		public static bool ShouldLog( Exception exception )
		{
			if( exception == null )
				return false;

			while( exception != null )
			{
				if( ShouldLogNonRecursive( exception ) )
					return true;

				exception = exception.InnerException;
			}

			return false;
		}

		/// <summary>
		/// Returns true if the specified exception (or any of its inner exceptions) should be shown to the user.
		/// </summary>
		/// <param name="exception">Exception to check if it should be shown to the user. This may be null.</param>
		/// <returns>Returns true if the specified exception should be shown to the user.</returns>
		public static bool ShouldShow( Exception exception )
		{
			if( exception == null )
				return false;

			// If there is a "BackgroundOperationException" exception, don't show the exception.
			if( HasException<BackgroundOperationException>( exception ) )
				return false;

			while( exception != null )
			{
				if( ShouldShowNonRecursive( exception ) )
					return true;

				exception = exception.InnerException;
			}

			return false;
		}

		/// <summary>
		/// Returns true if the specified exception (or any of its inner exceptions) is critical and the application must be
		/// terminated.
		/// </summary>
		/// <param name="exception">Exception to check if it's critical. This may be null.</param>
		/// <returns>Returns true if the specified exception (or any of its inner exceptions) is critical.</returns>
		public static bool IsCritical( Exception exception )
		{
			while( exception != null )
			{
				if( IsCriticalNonRecursive( exception ) )
					return true;

				exception = exception.InnerException;
			}

			return false;
		}

		/// <summary>
		/// Checks if the specified error type name is derived from any of the specified exception types.
		/// </summary>
		/// <param name="errorTypeName">The error type name to check. This may be null or empty.</param>
		/// <param name="exceptionTypes">Exception types which may be the ancestors of "errorTypeName". This may be null or
		/// empty.</param>
		/// <returns>Returns true if specified error type name is derived from any of the specified exception types.</returns>
		public static bool IsDerivedFromException( string errorTypeName, params Type[] exceptionTypes )
		{
			if( ArrayUtils.NoContent( exceptionTypes ) || TextUtils.NoContent( errorTypeName ) )
				return false;

			Type responseExceptionType = PortableTypes.GetGlobal( errorTypeName, typeof( Exception ), false );
			if( responseExceptionType == null )
				return false;

			foreach( Type exceptionType in exceptionTypes )
			{
				if( exceptionType.IsAssignableFrom( responseExceptionType ) )
					return true;
			}

			return false;
		}

		/// <summary>
		/// Finds an exception with the specified type, in the inner exceptions of the specified exception.
		/// </summary>
		/// <param name="exception">Exception in which to find the exception with the specified type. This may be null.</param>
		/// <returns>Returns the found exception, or null if none is found.</returns>
		public static TException FindException<TException>( Exception exception )
			where TException : Exception
		{
			while( exception != null )
			{
				if( exception as TException != null )
					return (TException)exception;

				exception = exception.InnerException;
			}

			return null;
		}

		/// <summary>
		/// Checks if an exception with the specified type is in the inner exceptions of the specified exception.
		/// </summary>
		/// <param name="exception">Exception in which to find the exception with the specified type. This may be null.</param>
		/// <returns>Returns true if an exception with the specified type is found.</returns>
		public static bool HasException<TException>( Exception exception )
			where TException : Exception
		{
			return FindException<TException>( exception ) != null;
		}

		/// <summary>
		/// Checks if the specified exception (but not its inner exceptions) is cross service.
		/// </summary>
		/// <param name="exception">Exception to check. This may be null.</param>
		/// <returns>Returns true the specified exception (but not its inner exceptions) is cross service.</returns>
		public static bool IsCrossServiceException( Exception exception )
		{
			// It's important to keep this verification limited to the inheritance from "CrossServiceException" because this is
			// what the code presumes elsewhere.
			return exception as CrossServiceException != null;
		}

		/// <summary>
		/// Checks if the specified exception (but not its inner exceptions) has a message which is cross service (even if the
		/// exception itself is not cross service).
		/// </summary>
		/// <param name="exception">Exception to check. This may be null.</param>
		/// <returns>Returns true the specified exception (but not its inner exceptions) has a message which is cross service.
		/// </returns>
		public static bool IsCrossServiceMessageException( Exception exception )
		{
			if( IsCrossServiceException( exception ) )
				return true;

			if( exception as System.ComponentModel.Win32Exception != null )
			{
				int nativeErrorCode = ((System.ComponentModel.Win32Exception)exception).NativeErrorCode;

				if( SystemUtils.IsCrossServiceWindowsErrorCode( nativeErrorCode ) )
					return true;
			}

			return false;
		}

		/// <summary>
		/// Checks if the specified error type name identifies a fatal cross-service exception.
		/// </summary>
		/// <param name="errorTypeName">The error type name to check. This may be null or empty.</param>
		/// <returns>Returns true if the specified error type name identifies a fatal cross-service exception.</returns>
		public static bool IsFatalCrossServiceException( string errorTypeName )
		{
			return IsDerivedFromException( errorTypeName
				// Vela exceptions.
				, typeof( FatalCrossServiceException )
				, typeof( CriticalException )

				// System exceptions.
				, typeof( AccessViolationException )
				, typeof( DataMisalignedException )
				, typeof( InsufficientExecutionStackException )
				, typeof( InvalidProgramException )
				, typeof( OutOfMemoryException )
				, typeof( StackOverflowException ) );
		}

		/// <summary>
		/// Finds a fatal cross-service exception, in the inner exceptions of the specified exception.
		/// </summary>
		/// <param name="exception">Exception in which to find the exception with the specified type. This may be null.</param>
		/// <returns>Returns the found exception, or null if none is found.</returns>
		public static Exception FindFatalCrossServiceException( Exception exception )
		{
			while( exception != null )
			{
				if(
					// Vela exceptions.
					(exception as FatalCrossServiceException != null)
					|| (exception as CriticalException != null)

					// System exceptions.
					|| (exception as AccessViolationException != null)
					|| (exception as DataMisalignedException != null)
					|| (exception as InsufficientExecutionStackException != null)
					|| (exception as InvalidProgramException != null)
					|| (exception as OutOfMemoryException != null)
					|| (exception as StackOverflowException != null) )
				{
					return exception;
				}

				exception = exception.InnerException;
			}

			return null;
		}

		/// <summary>
		/// Returns true if the specified exception (or any of its inner exceptions) requires the current processing to be
		/// terminated.
		/// </summary>
		/// <param name="exception">Exception to check. This may be null.</param>
		/// <returns>Returns true if the current processing must be terminated.</returns>
		public static bool IsAbort( Exception exception )
		{
			while( exception != null )
			{
				if(
					// System exceptions.
					(exception as ThreadAbortException != null)
					|| (exception as ThreadInterruptedException != null)
					|| (exception as OperationCanceledException != null)

					// Vela exceptions.
					|| (exception as AbortLongOperationException != null)

					// Critical exceptions.
					|| IsCriticalNonRecursive( exception ) )
				{
					return true;
				}

				exception = exception.InnerException;
			}

			return false;
		}

		/// <summary>
		/// Gets the message, from the specified exception, which has to be shown to the user.
		/// </summary>
		/// <param name="message">Message which is to be shown to the user. This may be null or empty.</param>
		/// <param name="exception">Exception to get message for. This may be null.</param>
		/// <param name="onlyCrossService">If this is true then only the cross service exception are included.</param>
		/// <returns>Returns the message which has to be shown to the user. This may be null or empty.</returns>
		public static string GetPresenterMessage( string message, Exception exception, bool onlyCrossService )
		{
			if( exception == null )
				return message;

			StringBuilder sb = new StringBuilder();

			// Add the error context.
			ValidationException ve = exception as ValidationException;
			if( (ve != null) && TextUtils.HasContent( ve.ErrorContext ) )
				TextUtils.AddWithLineBreaks( sb, 0, ve.ErrorContext );

			// Add the message.
			if( TextUtils.HasContent( message ) )
				TextUtils.AddWithLineBreaks( sb, 2, message );

			while( exception != null )
			{
				bool avoid = ShouldAvoidExceptionInPresenterMessage( exception )
					|| (onlyCrossService && !IsCrossServiceMessageException( exception ));

				if( !avoid )
				{
					// Add the error message.
					if( TextUtils.HasContent( exception.Message ) )
						TextUtils.AddWithLineBreaks( sb, 2, exception.Message );

					// Add the error details.
					string details = GetDetails( exception );
					if( TextUtils.HasContent( details ) )
						TextUtils.AddWithLineBreaks( sb, 1, details );
				}

				exception = exception.InnerException;
			}

			return sb.ToString();
		}

		/// <summary>
		/// Gets the details of the specified exception.
		/// </summary>
		/// <param name="exception">Exception for which to get the details. This may be null.</param>
		/// <returns>Returns the exception details. This may be null or empty.</returns>
		public static string GetDetails( Exception exception )
		{
			BaseException baseException = exception as BaseException;
			if( baseException == null )
				return null;

			StringBuilder sb = new StringBuilder();

			int count = baseException.Details.Count;
			int i = 0;

			foreach( ExceptionDetail detail in baseException.Details )
			{
				if( i < count - 1 )
					sb.AppendLine( string.Format( DetailTemplate, detail.Name, detail.Value ) );
				else
					sb.Append( string.Format( DetailTemplate, detail.Name, detail.Value ) );

				i++;
			}

			return sb.ToString();
		}

		/// <summary>
		/// Preserves the stack trace of the specified exception and rethrows it.
		/// </summary>
		/// <param name="exception">Exception which will be rethrown. This is modified in here. This may be null.</param>
		/// <returns>Returns the thrown exception. The return value would normally be superfluous since the exception is also
		/// thrown in here, but the return is still necessary (sometimes) because the compiler generates a compile-time error
		/// when the calling method should either return a value or throw an exception, yet the compiler sees neither; in such a
		/// case, the return value can be returned or thrown from the caller.</returns>
		/// <remarks>
		/// <para>This method is meant to be called just before the specified exception is about to be rethrown, after being
		/// caught to do some cleanup.</para>
		/// 
		/// <para>Rethrowing an exception with "throw exception" (in a "catch" statement) resets the stack trace at the rethrow
		/// point, so this should be avoided at all costs. However, rethrowing with "throw" is still an incomplete way to rethrow
		/// exceptions.</para>
		/// 
		/// <para>If this method is not called and the current exception is rethrown by calling "throw", the stack trace will
		/// contain, for the calling method, the source code line number of the rethrow (or of the end of the calling method)
		/// instead of the line number where the exception has originally occurred.</para>
		/// </remarks>
		public static Exception PreserveStackTrace( Exception exception )
		{
			if( exception != null )
				ExceptionDispatchInfo.Capture( exception ).Throw();

			return exception;
		}

		#endregion Public static methods
	}
}
