////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;

namespace Vela.Kernel
{
	/// <summary>
	/// File utilities.
	/// </summary>
	public static class FileUtils
	{
		#region Public constants

		/// <summary>
		/// The extension for a backup file used for saving data, including the prefix dot.
		/// </summary>
		public const string BackupFileExtension = ".bku";

		/// <summary>
		/// The extension for a temporary file used for saving data, including the prefix dot.
		/// </summary>
		public const string TempFileExtension = ".tmp";

		/// <summary>
		/// The character which is used to mark the beginning and end of a single textual unit (which contains whitespaces).
		/// </summary>
		public const char SystemCommandUnitMarker = '"';

		/// <summary>
		/// The character which is separates the command and command parameters.
		/// </summary>
		public const char SystemCommandParametersSeparator = ' ';

		#endregion Public constants

		#region Private static methods

		/// <summary>
		/// Extracts the plain command and command parameters from the specified system command.
		/// </summary>
		/// <param name="command">System command to parse. This may contain command parameters.</param>
		/// <param name="plainCommand">Is set with the extracted plain command.</param>
		/// <param name="parameters">Is set with the extracted command parameters.</param>
		/// <returns>Returns the extracted command parameters. This may be null or empty.</returns>
		private static void ParseSystemCommand( string command, out string plainCommand, out string parameters )
		{
			ValidationUtils.TextHasContent( command, "command" );

			string trimmedCommand = command.TrimStart();
			int commandStart, commandEnd, parametersStart;

			// Check if the command starts with a unit marker.
			if( trimmedCommand[ 0 ] == SystemCommandUnitMarker )
			{
				commandStart = 1;

				// Find the end marker.
				int endMarker = trimmedCommand.IndexOf( SystemCommandUnitMarker, 1 );

				commandEnd = (endMarker >= 0) ? (endMarker - 1) : (trimmedCommand.Length - 1);
				parametersStart = (endMarker >= 0) ? (endMarker + 1) : -1;
			}
			else
			{
				commandStart = 0;

				// Find the first blankspace.
				int parametersSeparator = trimmedCommand.IndexOf( SystemCommandParametersSeparator );

				commandEnd = (parametersSeparator >= 0) ? (parametersSeparator - 1) : (trimmedCommand.Length - 1);
				parametersStart = (parametersSeparator >= 0) ? (parametersSeparator + 1) : -1;
			}

			plainCommand = trimmedCommand.Substring( commandStart, commandEnd - commandStart + 1 );
			parameters = (parametersStart >= 0) ? trimmedCommand.Substring( parametersStart ) : null;
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Checks if the specified file names represent the same file.
		/// </summary>
		/// <param name="fileName1">File name to compare. This may be null or empty.</param>
		/// <param name="fileName2">File name to compare. This may be null or empty.</param>
		/// <returns>Returns true if the specified file names represent the same file.</returns>
		public static bool EqualNames( string fileName1, string fileName2 )
		{
			if( TextUtils.NoContent( fileName1 ) || TextUtils.NoContent( fileName2 ) )
				return false;

			return TextUtils.Equals( Path.GetFullPath( fileName1 ), Path.GetFullPath( fileName2 ) );
		}

		/// <summary>
		/// Creates a full file / folder name from the specified absolute path and relative file / folder name.
		/// </summary>
		/// <param name="absolutePath">Absolute path.</param>
		/// <param name="relativeFileName">Relative file / folder name</param>
		/// <returns>Returns the created full file / folder name.</returns>
		public static string MakeFullName( string absolutePath, string relativeFileName )
		{
			ValidationUtils.TextHasContent( absolutePath, "absolutePath" );
			ValidationUtils.TextHasContent( relativeFileName, "relativeFileName" );

			// Cut the prefix folder separator character because "Path.Combine" doesn't handle it.
			int pos = 0;
			while( (relativeFileName[ pos ] == Path.DirectorySeparatorChar)
				|| (relativeFileName[ pos ] == Path.AltDirectorySeparatorChar) )
			{
				pos++;
			}

			relativeFileName = relativeFileName.Substring( pos );

			return Path.GetFullPath( Path.Combine( absolutePath, relativeFileName ) );
		}

		/// <summary>
		/// Checks if the specified file name has an extension which is in the specified list.
		/// </summary>
		/// <param name="fileName">File name whose extension to check. If this is null or empty, or if it has no extension, false
		/// is returned.</param>
		/// <param name="fileExtensions">List of acceptable file extensions. If this is null or empty, true is returned.</param>
		/// <returns>Returns true if the specified file name has an extension which is in the specified list.</returns>
		public static bool HasExtension( string fileName, HashSet<string> fileExtensions )
		{
			if( TextUtils.NoContent( fileName ) )
				return false;

			if( (fileExtensions == null) || (fileExtensions.Count <= 0) )
				return true;

			string fileExtension = Path.GetExtension( fileName );
			if( TextUtils.NoContent( fileExtension ) )
				return false;

			return fileExtensions.Contains( fileExtension.ToLowerInvariant() );
		}

		/// <summary>
		/// Creates the specified folder if it doesn't exist.
		/// </summary>
		/// <param name="folderName">Name of folder to create.</param>
		public static void CreateFolder( string folderName )
		{
			ValidationUtils.TextHasContent( folderName, "folderName" );

			if( Directory.Exists( folderName ) )
				return;

			Directory.CreateDirectory( folderName );
		}

		/// <summary>
		/// Gets the available free space of the drive on which the file with the specified name is stored.
		/// </summary>
		/// <param name="fileName">File name.</param>
		/// <returns>Returns the drive's available free space.</returns>
		public static Int64 GetDriveAvailableFreeSpace( string fileName )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			string driveName = Path.GetPathRoot( fileName );

			foreach( DriveInfo drive in DriveInfo.GetDrives() )
			{
				if( drive.IsReady && TextUtils.Equals( drive.Name, driveName ) )
					return drive.AvailableFreeSpace;
			}

			throw new ValidationException( string.Format( "The drive \"{0}\" was not found.", driveName ) );
		}

		/// <summary>
		/// Gets the size of the specified file.
		/// </summary>
		/// <param name="fileName">Name of file whose size to get.</param>
		/// <param name="throwIfMissing">If this is true and the file is missing, an exception is thrown.</param>
		/// <returns>Returns the size of the specified file.</returns>
		public static Int64 GetFileSize( string fileName, bool throwIfMissing = false )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			FileInfo fi = new FileInfo( fileName );

			if( throwIfMissing && !fi.Exists )
				throw new ValidationException( string.Format( "The \"{0}\" file is missing or can't be accessed.", fileName ) );

			return fi.Exists ? fi.Length : 0;
		}

		/// <summary>
		/// Gets the time (in UTC format) when the content of the specified file was last changed.
		/// </summary>
		/// <param name="fileName">Name of file whose time to get.</param>
		/// <param name="throwIfMissing">If this is true and the file is missing, an exception is thrown.</param>
		/// <returns>Returns the time when the content of the specified file was last changed.</returns>
		public static DateTime GetFileLastChangeTime( string fileName, bool throwIfMissing = false )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			FileInfo fi = new FileInfo( fileName );

			if( throwIfMissing && !fi.Exists )
				throw new ValidationException( string.Format( "The \"{0}\" file is missing or can't be accessed.", fileName ) );

			return fi.Exists ? fi.LastWriteTimeUtc : DateTime.MinValue;
		}

		/// <summary>
		/// Gets the size of the specified file, and the time (in UTC format) when the content of the specified file was last
		/// changed.
		/// </summary>
		/// <param name="fileName">Name of file whose size and time to get.</param>
		/// <param name="throwIfMissing">If this is true and the file is missing, an exception is thrown.</param>
		/// <param name="lastChangeTime">Is set with the time when the content of the file was last changed.</param>
		/// <returns>Returns the size of the specified file.</returns>
		public static Int64 GetFileSizeAndLastChangeTime( string fileName, bool throwIfMissing, out DateTime lastChangeTime )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			FileInfo fi = new FileInfo( fileName );

			if( throwIfMissing && !fi.Exists )
				throw new ValidationException( string.Format( "The \"{0}\" file is missing or can't be accessed.", fileName ) );

			lastChangeTime = fi.Exists ? fi.LastWriteTimeUtc : DateTime.MinValue;

			return fi.Exists ? fi.Length : 0;
		}

		/// <summary>
		/// Gets the name of the root (= left most) parent folder from the specified folder.
		/// </summary>
		/// <param name="folderName">The name of the folder whose root parent folder to get. This may be null or empty.</param>
		/// <returns>Returns the root parent folder, or null is none exists.</returns>
		public static string GetRootParentFolderName( string folderName )
		{
			folderName = folderName.TrimStart( Path.DirectorySeparatorChar, Path.AltDirectorySeparatorChar );

			while( TextUtils.HasContent( folderName ) )
			{
				string parentFolderName = Path.GetDirectoryName( folderName );
				if( TextUtils.NoContent( parentFolderName ) )
					return folderName;

				folderName = parentFolderName;
			}

			return null;
		}

		/// <summary>
		/// Gets the name of the folder which exists, starting with the specified folder and walking through its parents.
		/// </summary>
		/// <param name="folderName">The name of the folder from which to start. This may be null or empty.</param>
		/// <returns>Returns the name of the folder which exists.</returns>
		public static string GetExistingFolderName( string folderName )
		{
			while( TextUtils.HasContent( folderName ) && !Directory.Exists( folderName ) )
				folderName = Path.GetDirectoryName( folderName );

			if( TextUtils.NoContent( folderName ) || !Directory.Exists( folderName ) )
				return null;

			return folderName;
		}

		/// <summary>
		/// Explores the parent folder of the specified file, and selects the specified file.
		/// </summary>
		/// <param name="fileName">The name of the file whose parent folder to explore.</param>
		public static void ExploreFile( string fileName )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			if( !File.Exists( fileName ) )
				fileName = GetExistingFolderName( fileName );

			if( TextUtils.NoContent( fileName ) || (!File.Exists( fileName ) && !Directory.Exists( fileName )) )
			{
				throw new ValidationException( "The file to explore is missing.", null, null
					, new ExceptionDetail( "File name", fileName ) );
			}

			try
			{
				string parameters = Directory.Exists( fileName ) ? string.Format( " \"{0}\"", fileName )
					: string.Format( " /select,\"{0}\"", fileName );

				Process.Start( "explorer.exe", parameters );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsAbort( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				throw new ValidationException( "Failed to explore the file.", ex, null
					, new ExceptionDetail( "File name", fileName ) );
			}
		}

		/// <summary>
		/// Runs the specified system command.
		/// </summary>
		/// <param name="command">The system command to run. This may NOT contain command parameters. The file name of the
		/// application to run does NOT have to be within quotes.</param>
		/// <param name="parameters">Command parameters. This may be null or empty.</param>
		/// <remarks>
		/// <para>This can be used (for example) to start applications, open files or navigate to URLs.</para>
		/// </remarks>
		public static void RunSystemCommand( string command, string parameters )
		{
			ValidationUtils.TextHasContent( command, "command" );

			try
			{
				if( TextUtils.HasContent( parameters ) )
					Process.Start( command, parameters );
				else
					Process.Start( command );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsAbort( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				throw new ValidationException( string.Format( "Failed to run the system command \"{0}\".", command ), ex );
			}
		}

		/// <summary>
		/// Runs the specified system command.
		/// </summary>
		/// <param name="command">The system command to run. This may contain command parameters. The file name of the
		/// application to run must be within quotes.</param>
		/// <remarks>
		/// <para>This can be used (for example) to start applications, open files or navigate to URLs.</para>
		/// </remarks>
		public static void RunSystemCommandWithParameters( string command )
		{
			string plainCommand, parameters;
			ParseSystemCommand( command, out plainCommand, out parameters );

			RunSystemCommand( plainCommand, parameters );
		}

		/// <summary>
		/// Runs the file with the specified name.
		/// </summary>
		/// <param name="fileName">The name of the file to run. The file name to run does NOT have to be within quotes.</param>
		/// <remarks>
		/// <para>This can be used (for example) to start applications, open files or navigate to URLs, without command line
		/// parameters.</para>
		/// </remarks>
		public static void RunFile( string fileName )
		{
			RunSystemCommand( fileName, null );
		}

		/// <summary>
		/// Runs the command interpreter with the specified command line.
		/// </summary>
		/// <param name="command">The command line to run. This may contain command parameters.</param>
		/// <param name="keepOpen">If this is true then the command line window is kept open after the application ends.</param>
		public static void RunCommandLine( string command, bool keepOpen )
		{
			ValidationUtils.TextHasContent( command, "command" );

			try
			{
				ProcessStartInfo processInfo = new ProcessStartInfo( "cmd.exe" );

				processInfo.Arguments = (keepOpen ? "/K " : "/C ") + command;

				Process.Start( processInfo );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsAbort( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				throw new ValidationException( string.Format( "Failed to run the system command \"{0}\".", command ), ex );
			}
		}

		/// <summary>
		/// Deletes the specified folder recursively, and all its content.
		/// </summary>
		/// <param name="folderName">Folder to delete.</param>
		/// <remarks>
		/// <para>No exception is thrown if the folder doesn't exist.</para>
		/// </remarks>
		public static void DeleteFolder( string folderName )
		{
			ValidationUtils.TextHasContent( folderName, "folderName" );

			if( Directory.Exists( folderName ) )
				Directory.Delete( folderName, true );
		}

		/// <summary>
		/// Copies the content of the specified folder recursively, and all its files, into the specified destination folder.
		/// Existing (destination) files are overwritten.
		/// </summary>
		/// <param name="sourceFolderName">Source folder.</param>
		/// <param name="destFolderName">Destination folder.</param>
		/// <param name="includeSourceFolderTitle">If this is true then the title of the source folder is included in the name
		/// of the destination files.</param>
		/// <param name="filterFileExtensions">List of extensions for the files which may be copied. If this is null or empty,
		/// all the files are copied.</param>
		public static void CopyFolderContent( string sourceFolderName, string destFolderName, bool includeSourceFolderTitle
			, HashSet<string> filterFileExtensions )
		{
			ValidationUtils.TextHasContent( sourceFolderName, "sourceFolderName" );
			ValidationUtils.TextHasContent( destFolderName, "destFolderName" );

			if( includeSourceFolderTitle )
				destFolderName = MakeFullName( destFolderName, Path.GetFileName( sourceFolderName ) );

			CreateFolder( destFolderName );
			DirectoryInfo sourceFolderInfo = new DirectoryInfo( sourceFolderName );

			// Copy sub-folders.
			foreach( DirectoryInfo folderInfo in sourceFolderInfo.EnumerateDirectories() )
			{
				CopyFolderContent( MakeFullName( sourceFolderName, folderInfo.Name )
					, MakeFullName( destFolderName, folderInfo.Name ), false, filterFileExtensions );
			}

			// Copy (and overwrite) files.
			foreach( FileInfo fileInfo in sourceFolderInfo.EnumerateFiles() )
			{
				if( !HasExtension( fileInfo.Name, filterFileExtensions ) )
					continue;

				fileInfo.CopyTo( MakeFullName( destFolderName, fileInfo.Name ), true );
			}
		}

		/// <summary>
		/// Copies the content of the specified folders recursively, and all their files, and the specified files, into the
		/// specified destination folder. Existing (destination) files are overwritten.
		/// </summary>
		/// <param name="sourceFolderFileNames">Source folder and file names. This may be null or empty.</param>
		/// <param name="destinationFolderName">Destination folder.</param>
		/// <param name="filterFileExtensions">List of extensions for the files which may be copied. If this is null or empty,
		/// all the files are copied.</param>
		public static void CopyFolderContent( string[] sourceFolderFileNames, string destinationFolderName
			, HashSet<string> filterFileExtensions )
		{
			ValidationUtils.TextHasContent( destinationFolderName, "destinationFolderName" );

			if( ArrayUtils.NoContent( sourceFolderFileNames ) )
				return;

			foreach( string fileName in sourceFolderFileNames )
			{
				if( !HasExtension( fileName, filterFileExtensions ) )
					continue;

				if( Directory.Exists( fileName ) )
				{
					CopyFolderContent( fileName, destinationFolderName, true, FileExtensions.ImageFile );
				}
				else if( FileExtensions.IsImageFile( fileName ) )
				{
					string sourceFileTitle = Path.GetFileName( fileName );
					string destinationFileName = MakeFullName( destinationFolderName, sourceFileTitle );

					File.Copy( fileName, destinationFileName, true );
				}
			}
		}

		/// <summary>
		/// Collects all the file names from the specified folder, recursively from all the child folders.
		/// </summary>
		/// <param name="folderName">The name of the folder from which to collect all the file names.</param>
		/// <param name="fileNames">The collected file names are added in here.</param>
		public static void CollectFileNamesFromFolder( string folderName, Strings fileNames )
		{
			ValidationUtils.TextHasContent( folderName, "folderName" );
			ValidationUtils.ObjectHasContent( fileNames, "fileNames" );

			if( !Directory.Exists( folderName ) )
				return;

			DirectoryInfo folderInfo = new DirectoryInfo( folderName );

			foreach( DirectoryInfo childFolderInfo in folderInfo.EnumerateDirectories() )
			{
				CollectFileNamesFromFolder( MakeFullName( folderName, childFolderInfo.Name ), fileNames );
			}

			foreach( FileInfo childFileInfo in folderInfo.EnumerateFiles() )
			{
				fileNames.Add( MakeFullName( folderName, childFileInfo.Name ) );
			}
		}

		/// <summary>
		/// Collects the file and folder names from the specified folder, without recursing the child folders.
		/// </summary>
		/// <param name="folderName">The name of the folder from which to collect the file and folder names.</param>
		/// <returns>Returns the collected file and folder names.</returns>
		public static Strings CollectFileNamesFromFolderNoRecursing( string folderName )
		{
			ValidationUtils.TextHasContent( folderName, "folderName" );

			Strings files = new Strings();
			DirectoryInfo folderInfo = new DirectoryInfo( folderName );

			foreach( DirectoryInfo childFolderInfo in folderInfo.EnumerateDirectories() )
			{
				files.Add( MakeFullName( folderName, childFolderInfo.Name ) );
			}

			foreach( FileInfo childFileInfo in folderInfo.EnumerateFiles() )
			{
				files.Add( MakeFullName( folderName, childFileInfo.Name ) );
			}

			return files;
		}

		/// <summary>
		/// Checks if the specified folder contains more than the specified number of files, including the current count.
		/// </summary>
		/// <param name="folderName">Folder whose files to count.</param>
		/// <param name="countMax">The maximum count. If this is 0 or a negative number, it's ignored.</param>
		/// <param name="currentCount">Is incremented for each counted file.</param>
		/// <returns>Returns true if the specified folder contains more than the specified number of files, including the current
		/// count.</returns>
		public static bool FileCountExceeds( string folderName, int countMax, ref int currentCount )
		{
			ValidationUtils.TextHasContent( folderName, "folderName" );

			if( countMax <= 0 )
				return false;

			DirectoryInfo folderInfo = new DirectoryInfo( folderName );

			foreach( DirectoryInfo childFolderInfo in folderInfo.EnumerateDirectories() )
			{
				if( FileCountExceeds( MakeFullName( folderName, childFolderInfo.Name ), countMax, ref currentCount ) )
					return true;
			}

			foreach( FileInfo childFileInfo in folderInfo.EnumerateFiles() )
			{
				currentCount++;

				if( currentCount > countMax )
					return true;
			}

			return false;
		}

		/// <summary>
		/// Checks if the specified list of files and folders contain more than the specified number of files.
		/// </summary>
		/// <param name="fileNames">List of files and folders whose files to count. This may be null or empty.</param>
		/// <param name="countMax">The maximum count. If this is 0 or a negative number, it's ignored.</param>
		/// <returns>Returns true if the specified list of files and folders contain more than the specified number of files.
		/// </returns>
		public static bool FileCountExceeds( string[] fileNames, int countMax )
		{
			if( ArrayUtils.NoContent( fileNames ) || (countMax <= 0) )
				return false;

			int currentCount = 0;
			foreach( string fileName in fileNames )
			{
				if( Directory.Exists( fileName ) )
				{
					if( FileCountExceeds( fileName, countMax, ref currentCount ) )
						return true;
				}
				else
				{
					currentCount++;

					if( currentCount > countMax )
						return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Saves the specified buffer to the specified file. Existing files are overwritten.
		/// </summary>
		/// <param name="fileName">Name of the file where to save the buffer.</param>
		/// <param name="buffer">Buffer to save.</param>
		/// <remarks>
		/// <para>The saving is crash-safe. If the application crashes when the file is being saved, at least 1 of the
		/// following files is still available: a file with the specified name, a file with a backup name, a file with a
		/// temporary name. If the backup file exists, it's the valid one and you should implement your own recovery mechanism
		/// following a crash.</para>
		/// </remarks>
		public static void Save( string fileName, byte[] buffer )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );
			ValidationUtils.ArrayHasContent( buffer, "buffer" );

			// Don't use "File.Replace" because there is no indication that it's transactional.

			// Create the names for the backup and temporary files.
			string backupFileName = fileName + BackupFileExtension;
			string tempFileName = fileName + TempFileExtension;

			// Save the content to a temporary file.
			File.Delete( tempFileName ); // Delete the previous temporary file.
			CreateFolder( Path.GetDirectoryName( tempFileName ) );
			File.WriteAllBytes( tempFileName, buffer );

			// Backup the original file.
			if( File.Exists( fileName ) )
			{
				File.Delete( backupFileName ); // Delete the previous backup file.
				File.Move( fileName, backupFileName );
			}

			// Rename the temporary file as the specified one.
			File.Move( tempFileName, fileName );

			// Delete the backup file.
			File.Delete( backupFileName );
		}


		/// <summary>
		/// Gets the name of the actual file whose content has to be read, associated to the specified file whose content should
		/// be read.
		/// </summary>
		/// <param name="fileName">Name of the file whose content to read. The file may be missing and its content may be null
		/// or empty.</param>
		/// <returns>Returns the file content, or null if the file is missing or its content is null or empty.</returns>
		/// <remarks>
		/// <para>Call this method to get the name of the file whose content has to be read, from a method which implements a
		/// recovery mechanism following a crash during a call of the <see cref="Save"/> method.</para>
		/// </remarks>
		public static string GetFileNameToRead( string fileName )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			// Check if the temporary file exists.
			string tempFileName = fileName + TempFileExtension;
			FileInfo tempInfo = File.Exists( tempFileName ) ? new FileInfo( tempFileName ) : null;
			Int64 tempSize = (tempInfo != null) ? tempInfo.Length : 0;
			if( tempSize > 0 )
				return tempFileName;

			// Check if the backup file exists.
			string backupFileName = fileName + BackupFileExtension;
			FileInfo backupInfo = File.Exists( backupFileName ) ? new FileInfo( backupFileName ) : null;
			Int64 backupSize = (backupInfo != null) ? backupInfo.Length : 0;
			if( backupSize > 0 )
				return backupFileName;

			return fileName;
		}

		/// <summary>
		/// Reads the content of the specified file.
		/// </summary>
		/// <param name="fileName">Name of the file whose content to read. The file may be missing and its content may be null
		/// or empty.</param>
		/// <param name="throwIfMissing">If this is true and the file to read is missing, an exception is thrown.</param>
		/// <returns>Returns the file content, or null if the file is missing or its content is null or empty.</returns>
		/// <remarks>
		/// <para>This method implements a recovery mechanism following a crash during a call of the <see cref="Save"/> method.
		/// </para>
		/// </remarks>
		public static byte[] Read( string fileName, bool throwIfMissing )
		{
			string actualFileName = GetFileNameToRead( fileName );

			if( !File.Exists( actualFileName ) )
			{
				if( throwIfMissing )
					throw new ValidationException( string.Format( "The file \"{0}\" is missing.", actualFileName ) );
				else
					return null;
			}

			return File.ReadAllBytes( actualFileName );
		}

		/// <summary>
		/// Reads a part of the specified file.
		/// </summary>
		/// <param name="fileName">Name of the file whose content to read.</param>
		/// <param name="offset">The position in the file from which to read the file part.</param>
		/// <param name="partSize">The size of the file part to read. This may refer beyond the end of the file.</param>
		/// <returns>Returns the read file part.</returns>
		public static byte[] Read( string fileName, Int64 offset, int partSize )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );
			ValidationUtils.Integer( offset, 0, Int64.MaxValue - 1, "partSize" );
			ValidationUtils.Integer( partSize, 1, int.MaxValue, "partSize" );

			Int64 fileSize = GetFileSize( fileName );
			if( (offset >= fileSize) || (fileSize <= 0) )
			{
				throw new ValidationException( "Can't read beyond the end of the file.", null, null
					, new ExceptionDetail( "File name", fileName ) );
			}

			int actualSizeToRead = (int)Math.Min( partSize, fileSize - offset );

			byte[] filePart = new byte[ actualSizeToRead ];
			using( FileStream fs = File.OpenRead( fileName ) )
			{
				fs.Position = offset;
				int readSize = fs.Read( filePart, 0, actualSizeToRead );

				if( readSize != actualSizeToRead )
				{
					throw new ValidationException( string.Format( "{0} bytes were read instead of {1}, from the file."
						, readSize, actualSizeToRead ), null, null, new ExceptionDetail( "File name", fileName ) );
				}
			}

			return filePart;
		}

		/// <summary>
		/// Verifies that the specified file is readable. An exception is thrown if the file is not readable.
		/// </summary>
		/// <param name="fileName">Name of the file to verify if it's readable.</param>
		public static void VerifyIsReadable( string fileName )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			using( FileStream fs = File.OpenRead( fileName ) )
			{
				// Try to read 1 byte.
				fs.ReadByte();
			}
		}

		#endregion Public static methods
	}
}
