////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Diagnostics;

namespace Vela.Kernel
{
	/// <summary>
	/// Utilities for the high resolution timer.
	/// </summary>
	public static class HighResolutionTimer
	{
		#region Public constants

		/// <summary>
		/// The minimum allowed frequency.
		/// </summary>
		public const int FrequencyMin = 500000;

		#endregion Public constants

		#region Static constructors

		/// <summary>
		/// Initializes the current class.
		/// </summary>
		/// <remarks>
		/// <para>An exception is thrown if <see cref="Frequency"/> is smaller than <see cref="FrequencyMin"/>.</para>
		/// </remarks>
		static HighResolutionTimer()
		{
			if( !Stopwatch.IsHighResolution || (Frequency < FrequencyMin) )
			{
				throw new ValidationException( string.Format( "The system's high resolution timer is unusable. It's frequency is"
					+ " {0} Hz.", Frequency ) );
			}
		}

		#endregion Static constructors

		#region Public static properties

		/// <summary>
		/// The frequency of the timer. READ-ONLY.
		/// </summary>
		public static Int64 Frequency
		{
			get { return Stopwatch.Frequency; }
		}

		#endregion Public static properties

		#region Public static methods

		/// <summary>
		/// Converts the specified number of seconds in ticks.
		/// </summary>
		/// <returns>Returns the resulted number of ticks.</returns>
		public static Int64 SecondsToTicks( double seconds )
		{
			return (Int64)(seconds * Frequency);
		}

		/// <summary>
		/// Converts the specified number of ticks in seconds.
		/// </summary>
		/// <returns>Returns the resulted number of seconds.</returns>
		public static double TicksToSeconds( Int64 ticks )
		{
			return (double)ticks / Frequency;
		}

		/// <summary>
		/// Gets the value (in ticks) of the timer.
		/// </summary>
		/// <returns>Returns the value (in ticks) of the timer.</returns>
		public static Int64 GetTicks()
		{
			return Stopwatch.GetTimestamp();
		}

		#endregion Public static methods
	}
}
