////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections;

namespace Vela.Kernel
{
	/// <summary>
	/// List utilities.
	/// </summary>
	public static class ListUtils
	{
		#region Public static methods

		/// <summary>
		/// Checks if the specified list has content.
		/// </summary>
		/// <param name="list">List to be checked. This may be null or empty.</param>
		/// <returns>Returns true if the specified list has content.</returns>
		public static bool HasContent( IKernelObservableList list )
		{
			return (list != null) && (list.Count > 0);
		}

		/// <summary>
		/// Checks if the specified list has content.
		/// </summary>
		/// <param name="list">List to be checked. This may be null or empty.</param>
		/// <returns>Returns true if the specified list has content.</returns>
		public static bool HasContent( ICollection list )
		{
			return (list != null) && (list.Count > 0);
		}

		/// <summary>
		/// Checks if the specified list is null or empty.
		/// </summary>
		/// <param name="list">List to be checked. This may be null or empty.</param>
		/// <returns>Returns true if the specified list is null or empty.</returns>
		public static bool NoContent( IKernelObservableList list )
		{
			return (list == null) || (list.Count <= 0);
		}

		/// <summary>
		/// Checks if the specified list is null or empty.
		/// </summary>
		/// <param name="list">List to be checked. This may be null or empty.</param>
		/// <returns>Returns true if the specified list is null or empty.</returns>
		public static bool NoContent( ICollection list )
		{
			return (list == null) || (list.Count <= 0);
		}

		/// <summary>
		/// Checks if the specified lists have the same items (in any order).
		/// </summary>
		/// <typeparam name="TItem">The type of the items from the lists.</typeparam>
		/// <param name="list1">List to compare.</param>
		/// <param name="list2">List to compare.</param>
		/// <param name="areItemsEqual">Delegate which checks if items are equal.</param>
		/// <returns>Returns true if the specified lists have the same items (in any order).</returns>
		public static bool Equals<TItem>( KernelObservableList<TItem> list1, KernelObservableList<TItem> list2
				, Func<TItem, TItem, bool> areItemsEqual )
			where TItem : class
		{
			ValidationUtils.ObjectHasContent( areItemsEqual, "areItemsEqual" );

			if( list1 == null )
			{
				if( list2 == null )
					return true;
				else
					return false;
			}
			else
			{
				if( list2 == null )
					return false;
			}

			if( list1.Count != list2.Count )
				return false;

			foreach( TItem item1 in list1 )
			{
				foreach( TItem item2 in list2 )
				{
					if( !areItemsEqual( item1, item2 ) )
						return false;
				}
			}

			return true;
		}

		#endregion Public static methods
	}
}
