////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;

namespace Vela.Kernel
{
	/// <summary>
	/// Mathematical utilities.
	/// </summary>
	public static class MathUtils
	{
		#region Public static methods

		/// <summary>
		/// Limits the specified value within the specified bounds.
		/// </summary>
		/// <param name="value">Value to limit.</param>
		/// <param name="valueMin">Minimum value.</param>
		/// <param name="valueMax">Maximum value.</param>
		/// <returns>Returns the limited value.</returns>
		public static int Limit( int value, int valueMin, int valueMax )
		{
			if( value < valueMin )
				return valueMin;

			if( value > valueMax )
				return valueMax;

			return value;
		}

		/// <summary>
		/// Rounds up quotient of the specified dividend and divisor.
		/// </summary>
		/// <param name="dividend">Dividend to divide.</param>
		/// <param name="divisor">Divisor with which to divide.</param>
		/// <returns>Returns the aligned value.</returns>
		public static int RoundUp( int dividend, int divisor )
		{
			int quotient = dividend / divisor;
			int remainder = dividend % divisor;

			if( remainder > 0 )
				return quotient + 1;
			else if( remainder < 0 )
				return quotient - 1;
			else
				return quotient;
		}

		/// <summary>
		/// Rounds up quotient of the specified dividend and divisor.
		/// </summary>
		/// <param name="dividend">Dividend to divide.</param>
		/// <param name="divisor">Divisor with which to divide.</param>
		/// <returns>Returns the aligned value.</returns>
		public static Int64 RoundUp( Int64 dividend, Int64 divisor )
		{
			Int64 quotient = dividend / divisor;
			Int64 remainder = dividend % divisor;

			if( remainder > 0 )
				return quotient + 1;
			else if( remainder < 0 )
				return quotient - 1;
			else
				return quotient;
		}

		/// <summary>
		/// Aligns the specified value to an integer number of the specified alignment.
		/// </summary>
		/// <param name="value">Value to align.</param>
		/// <param name="alignment">Alignment to which to align "value".</param>
		/// <returns>Returns the aligned value.</returns>
		public static int AlignUp( int value, int alignment )
		{
			return alignment * RoundUp( value, alignment );
		}

		/// <summary>
		/// Aligns the specified value to an integer number of the specified alignment.
		/// </summary>
		/// <param name="value">Value to align.</param>
		/// <param name="alignment">Alignment to which to align "value".</param>
		/// <returns>Returns the aligned value.</returns>
		public static Int64 AlignUp( Int64 value, Int64 alignment )
		{
			return alignment * RoundUp( value, alignment );
		}

		/// <summary>
		/// Aligns the specified value to an integer number of the specified alignment.
		/// </summary>
		/// <param name="value">Value to align.</param>
		/// <param name="alignment">Alignment to which to align "value".</param>
		/// <returns>Returns the aligned value.</returns>
		public static double AlignUp( double value, double alignment )
		{
			return alignment * Math.Ceiling( value / alignment );
		}

		/// <summary>
		/// Checks if the specified value is prime.
		/// </summary>
		/// <param name="value">Value to check if it's prime. This must be at least 1.</param>
		/// <returns>Returns true if the specified value is prime.</returns>
		public static bool CheckSmallPrime( int value )
		{
			ValidationUtils.Integer( value, 1, int.MaxValue, "value" );

			if( value <= 3 )
				return true;

			if( (value & 1) == 0 )
				throw new ValidationException( string.Format( "An even number can never be prime: {0}.", value ) );

			int maxTest = (int)Math.Sqrt( value );

			for( int i = 3; i <= maxTest; i += 2 )
				if( value % i == 0 )
					return false; // The number is NOT prime.

			// The number is prime.
			return true;
		}

		/// <summary>
		/// Generates a small prime bigger than the specified value.
		/// </summary>
		/// <param name="startValue">The generated prime must be bigger than or equal with this value. This must be at least 1.
		/// </param>
		/// <returns>Returns a small prime bigger than the specified value.</returns>
		public static int GenerateSmallPrime( int startValue )
		{
			ValidationUtils.Integer( startValue, 1, int.MaxValue, "startValue" );

			// The start value must be an odd number.
			if( (startValue & 1) == 0 )
				startValue++;

			for( int prime = startValue; prime <= int.MaxValue - 3; prime += 2 )
			{
				if( CheckSmallPrime( prime ) )
					return prime; // Prime found.
			}

			throw new ValidationException( string.Format( "Failed to generate a small prime starting from {0}.", startValue ) );
		}

		#endregion Public static methods
	}
}
