////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using System.Reflection;

namespace Vela.Kernel
{
	/// <summary>
	/// Resource utilities.
	/// </summary>
	public static class ResourceUtils
	{
		#region Public static methods

		/// <summary>
		/// Gets all the names of the embedded resources from the specified assembly.
		/// </summary>
		/// <param name="assembly">The assembly from where to get the resource names. If this is null, the application's
		/// assembly is used.</param>
		/// <param name="prefix">Only resource names which start with this prefix are returned. This may be null or empty.
		/// </param>
		/// <returns>Returns all the names of resources from the specified assembly.</returns>
		public static string[] GetEmbeddedNames( Assembly assembly, string prefix )
		{
			if( assembly == null ) assembly = AssemblyUtils.ApplicationAssembly;

			// Get all names.
			string[] names = assembly.GetManifestResourceNames();

			// Count the prefixed names.
			int count = 0;
			if( TextUtils.HasContent( prefix ) )
			{
				foreach( string name in names )
				{
					if( name.StartsWith( prefix, StringComparison.OrdinalIgnoreCase ) ) count++;
				}
			}
			else count = names.Length;

			// Get the prefixed names.
			string[] result = new string[ count ];
			int i = 0;
			if( TextUtils.HasContent( prefix ) )
			{
				foreach( string name in names )
				{
					if( name.StartsWith( prefix, StringComparison.OrdinalIgnoreCase ) )
					{
						result[ i ] = name;
						i++;
					}
				}
			}
			else
			{
				foreach( string name in names )
				{
					result[ i ] = name;
					i++;
				}
			}

			return result;
		}

		/// <summary>
		/// Returns the content of an embedded binary resource with the specified name, from the specified assembly.
		/// </summary>
		/// <param name="assembly">The assembly from where to get the resource names. If this is null, the application's
		/// assembly is used.</param>
		/// <param name="resName">Name of the resource to get the content for.</param>
		/// <returns>Returns the content of a binary resource with the specified name.</returns>
		/// <remarks>
		/// <para>This method can extract the content of binary files.</para>
		/// </remarks>
		public static byte[] GetEmbeddedBinary( Assembly assembly, string resName )
		{
			if( assembly == null ) assembly = AssemblyUtils.ApplicationAssembly;

			// Get the resource stream.
			using( Stream resourceStream = assembly.GetManifestResourceStream( resName ) )
			{
				if( resourceStream == null )
				{
					throw new ValidationException( string.Format( "Missing binary resource \"{0}\".", resName ) );
				}

				int size = (int)resourceStream.Length;
				byte[] buffer = new byte[ size ];

				resourceStream.Read( buffer, 0, buffer.Length );

				return buffer;
			}
		}

		/// <summary>
		/// Returns the content of an embedded resource string with the specified name, from the specified assembly.
		/// </summary>
		/// <param name="assembly">The assembly from where to get the resource names. If this is null, the application's
		/// assembly is used.</param>
		/// <param name="resName">Name of the resource to get the content for.</param>
		/// <returns>Returns the content of a resource string with the specified name.</returns>
		/// <remarks>
		/// <para>This method can extract the content of text files.</para>
		/// </remarks>
		public static string GetEmbeddedString( Assembly assembly, string resName )
		{
			if( assembly == null ) assembly = AssemblyUtils.ApplicationAssembly;

			// Get the resource stream.
			using( Stream resourceStream = assembly.GetManifestResourceStream( resName ) )
			{
				if( resourceStream == null )
				{
					throw new ValidationException( string.Format( "Missing string resource \"{0}\".", resName ) );
				}

				using( StreamReader sr = new StreamReader( resourceStream ) )
				{
					return sr.ReadToEnd();
				}
			}
		}

		#endregion Public static methods
	}
}
