////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Globalization;

namespace Vela.Kernel
{
	/// <summary>
	/// System utilities.
	/// </summary>
	public static class SystemUtils
	{
		#region Public static methods

		/// <summary>
		/// Checks if the operating system is at least Windows 8.
		/// </summary>
		/// <returns>Returns true if the operating system is at least Windows 8.</returns>
		public static bool IsOsAtLeastWindows8()
		{
			return (Environment.OSVersion.Version.Major >= 6) && (Environment.OSVersion.Version.Minor >= 2);
		}

		/// <summary>
		/// Gets the number of physical and virtual processor cores of the system.
		/// </summary>
		/// <returns>Returns the number of physical and virtual processor cores of the system.</returns>
		public static int GetProcessorCoresCount()
		{
			return Environment.ProcessorCount;
		}

		/// <summary>
		/// Gets the language code (= two-letter ISO 639-1) of the user.
		/// </summary>
		/// <returns>Returns the language code of the user.</returns>
		public static string GetUserCountryCode()
		{
			return CultureInfo.CurrentUICulture.TwoLetterISOLanguageName;
		}

		/// <summary>
		/// Checks if the specified Windows error code can be sent across the bounds of an action service, to an action
		/// client.
		/// </summary>
		/// <param name="errorCode">Error code to check.</param>
		/// <returns>Returns true if the specified Windows / native error code can be sent across the bounds of an action
		/// service, to an action client.</returns>
		public static bool IsCrossServiceWindowsErrorCode( int errorCode )
		{
			HashSet<int> crossServiceCodes = new HashSet<int>()
				{
					0x00000002 /* ERROR_FILE_NOT_FOUND */
					, 0x00000003 /* ERROR_PATH_NOT_FOUND */
					, 0x00000005 /* ERROR_ACCESS_DENIED */
					, 0x0000000B /* ERROR_BAD_FORMAT */
					, 0x0000000F /* ERROR_INVALID_DRIVE */
				};

			return crossServiceCodes.Contains( errorCode );
		}

		/// <summary>
		/// Gets a user friendly text for the specified application exit code.
		/// </summary>
		/// <param name="exitCode">Application exit code.</param>
		/// <returns>Returns a user friendly text for the specified application exit code. This may be null or empty.</returns>
		public static string GetUserFriendlyApplicationExitCode( int exitCode )
		{
			switch( exitCode )
			{
				case 0:
					return null;
				case 1:
					return "The command name may have been misspelled (command not recognized).";
				case 2:
					return "The file name can't be found.";
				case 3:
					return "The file path can't be found.";
				case 5:
					return "The user doesn't have enough access rights (access is denied).";
				case 9009:
					return "The application file name or path may have been misspelled (program not recognized).";
				case -1073741510:
					return "The application was terminated by the user (Ctrl + C was pressed).";
				default:
					return null;
			}
		}

		#endregion Public static methods
	}
}
