////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Globalization;
using System.Security;
using System.Text;

namespace Vela.Kernel
{
	/// <summary>
	/// Text utilities.
	/// </summary>
	public static class TextUtils
	{
		#region Public constants

		/// <summary>
		/// The mote size for numbers.
		/// </summary>
		public const int NumberMoteSize = 3;

		/// <summary>
		/// The mote separator for numbers.
		/// </summary>
		public const string NumberMoteSeparator = "'";

		/// <summary>
		/// The default mote size.
		/// </summary>
		public const int MoteSize = 3;

		/// <summary>
		/// The default mote separator.
		/// </summary>
		public const string MoteSeparator = " ";

		/// <summary>
		/// The default mote size for a buffer.
		/// </summary>
		public const int BufferMoteSize = 4;

		/// <summary>
		/// The default mote separator for a buffer.
		/// </summary>
		public const string BufferMoteSeparator = " ";

		/// <summary>
		/// The characters which are used to convert a buffer to a radix-16 textual representation.
		/// </summary>
		public const string TextRadix16Chars = "0123456789abcdef";

		/// <summary>
		/// The characters which are used to convert a buffer to a radix-32 textual representation.
		/// </summary>
		/// <remarks>
		/// <para>Some of the letters from the latin alphabet have been excluded because they can be easily confused with other
		/// characters.</para>
		/// </remarks>
		public const string TextRadix32Chars = "0123456789abcdefghjklmnprstuvxyz";

		/// <summary>
		/// The URL prefixes for which an HTTPS prefix will be added (instead of HTTP).
		/// </summary>
		public static readonly string[] RecognizedSecureUrlPrefixes = new string[]
		{
			"accounts.google", "drive.google", "facebook", "google", "login.yahoo", "mail.google", "mail.yahoo"
			, "maps.google", "news.google", "play.google", "plus.google", "twitter", "youtube"
		};

		#endregion Public constants

		#region Private static methods

		/// <summary>
		/// Prefixes the specified URL with either the secure prefix (if the URL starts with a text from
		/// <see cref="RecognizedSecureUrlPrefixes"/>), or with the unsecure one. If the specified URL already has a (recognized)
		/// prefix, nothing happens.
		/// </summary>
		/// <param name="url">URL to prefix.</param>
		/// <returns>Returns the prefixed URL.</returns>
		private static string PrefixUrl( string url )
		{
			ValidationUtils.TextHasContent( url, "url" );

			string[] urlPrefixes = new string[] { "https://", "http://", "www." };
			if( TextStartsWithList( url, urlPrefixes ) )
				return url;

			return (TextStartsWithList( url, RecognizedSecureUrlPrefixes ) ? "https://" : "http://") + url;
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Checks if the specified text has content.
		/// </summary>
		/// <param name="text">Text to be checked. This may be null or empty.</param>
		/// <returns>Returns true if the specified text has content.</returns>
		public static bool HasContent( string text )
		{
			return (text != null) && (text.Length > 0);
		}

		/// <summary>
		/// Checks if the specified text is null or empty.
		/// </summary>
		/// <param name="text">Text to be checked. This may be null or empty.</param>
		/// <returns>Returns true if the specified text is null or empty.</returns>
		public static bool NoContent( string text )
		{
			return (text == null) || (text.Length <= 0);
		}

		/// <summary>
		/// Verifies if the specified texts have the same content.
		/// </summary>
		/// <param name="text1">Text whose content to compare. This may be null or empty.</param>
		/// <param name="text2">Text whose content to compare. This may be null or empty.</param>
		/// <returns>Returns true if the two texts have the same content.</returns>
		/// <remarks>
		/// <para>The comparison is ordinal and case-insensitive.</para>
		/// </remarks>
		public static bool Equals( string text1, string text2 )
		{
			// If both are null, they're equal.
			if( (text1 == null) && (text2 == null) )
				return true;

			// If only one is null, they're not equal.
			if( (text1 == null) || (text2 == null) )
				return false;

			// If they have different sizes, they're not equal.
			if( text1.Length != text2.Length )
				return false;

			return string.Compare( text1, text2, StringComparison.OrdinalIgnoreCase ) == 0;
		}

		/// <summary>
		/// Checks if the specified text contains the specified text to find.
		/// </summary>
		/// <param name="text">Text in which to find a text. This may be null or empty.</param>
		/// <param name="textToFind">Text to find. This may be null or empty.</param>
		/// <returns>Returns true if the text contains the text to find.</returns>
		public static bool Contains( string text, string textToFind )
		{
			if( NoContent( text ) )
			{
				return NoContent( textToFind );
			}
			else if( NoContent( textToFind ) )
			{
				return true;
			}

			return text.IndexOf( textToFind, StringComparison.OrdinalIgnoreCase ) >= 0;
		}

		/// <summary>
		/// Replaces the specified replaceable text with the specified replacing text, in the specified text.
		/// </summary>
		/// <param name="text">Text in which to perform the replace. This may be null or empty.</param>
		/// <param name="replaceableText">Text to replace. This may be null or empty.</param>
		/// <param name="replacingText">Text with which to replace the replaceable text. This may be null or empty.</param>
		/// <returns>Returns the resulted text.</returns>
		/// <remarks>
		/// <para>The replace is case insensitive.</para>
		/// </remarks>
		public static string Replace( string text, string replaceableText, string replacingText )
		{
			if( NoContent( text ) || NoContent( replaceableText ) )
				return text;

			int currentIndex = 0;
			int nextIndex = text.IndexOf( replaceableText, currentIndex, StringComparison.OrdinalIgnoreCase );
			if( nextIndex < 0 )
				return text;

			StringBuilder sb = new StringBuilder();
			bool hasReplacingText = HasContent( replacingText );

			while( nextIndex >= 0 )
			{
				sb.Append( text.Substring( currentIndex, nextIndex - currentIndex ) );

				if( hasReplacingText )
					sb.Append( replacingText );

				nextIndex += replaceableText.Length;
				currentIndex = nextIndex;

				nextIndex = text.IndexOf( replaceableText, nextIndex, StringComparison.OrdinalIgnoreCase );
			}

			sb.Append( text.Substring( currentIndex ) );

			return sb.ToString();
		}

		/// <summary>
		/// Finds the position of the first whitespace which comes after the specified position, in the specified text.
		/// </summary>
		/// <param name="text">Text in which to find the position of the whitespace. This may be null or empty.</param>
		/// <param name="startPos">Position in "text" before which to find the whitespace; the search starts from this position.
		/// </param>
		/// <returns>Returns the position of the found whitespace, or -1 if none is found.</returns>
		public static int IndexOfWhitespace( string text, int startPos )
		{
			if( NoContent( text ) )
				return -1;

			if( startPos < 0 )
				startPos = 0;

			int count = text.Length;
			for( int i = startPos; i < count; i++ )
			{
				if( char.IsWhiteSpace( text[ i ] ) )
					return i;
			}

			return -1;
		}

		/// <summary>
		/// Finds the position of the first non-whitespace which comes after the specified position, in the specified text.
		/// </summary>
		/// <param name="text">Text in which to find the position of the non-whitespace. This may be null or empty.</param>
		/// <param name="startPos">Position in "text" before which to find the non-whitespace; the search starts from this
		/// position.</param>
		/// <returns>Returns the position of the found non-whitespace, or -1 if none is found.</returns>
		public static int IndexOfNonWhitespace( string text, int startPos )
		{
			if( NoContent( text ) )
				return -1;

			if( startPos < 0 )
				startPos = 0;

			int count = text.Length;
			for( int i = startPos; i < count; i++ )
			{
				if( !char.IsWhiteSpace( text[ i ] ) )
					return i;
			}

			return -1;
		}

		/// <summary>
		/// Finds the position of the first whitespace which comes before the specified position, in the specified text.
		/// </summary>
		/// <param name="text">Text in which to find the position of the whitespace. This may be null or empty.</param>
		/// <param name="startPos">Position in "text" before which to find the whitespace; the search starts from this position.
		/// </param>
		/// <returns>Returns the position of the found whitespace, or -1 if none is found.</returns>
		public static int LastIndexOfWhitespace( string text, int startPos )
		{
			if( NoContent( text ) )
				return -1;

			int count = text.Length;
			if( startPos >= count )
				startPos = count - 1;

			for( int i = startPos; i >= 0; i-- )
			{
				if( char.IsWhiteSpace( text[ i ] ) )
					return i;
			}

			return -1;
		}

		/// <summary>
		/// Deletes all the whitespaces from the specified text.
		/// </summary>
		/// <param name="text">Text to process. This may be null or empty.</param>
		/// <returns>Returns the resulted text. This may be null or empty.</returns>
		public static string DeleteWhitespaces( string text )
		{
			if( NoContent( text ) )
				return text;

			StringBuilder sb = new StringBuilder();

			for( int i = 0; i < text.Length; i++ )
			{
				char ch = text[ i ];
				if( !char.IsWhiteSpace( ch ) )
					sb.Append( ch );
			}

			return sb.ToString();
		}

		/// <summary>
		/// Reduces multiple consecutive line breaks to a single one in the specified text.
		/// </summary>
		/// <param name="text">Text to process. This may be null or empty.</param>
		/// <returns>Returns the resulted text. This may be null or empty.</returns>
		public static string ReduceMultipleConsecutiveLineBreaks( string text )
		{
			if( NoContent( text ) )
				return text;

			string lineBreak = Environment.NewLine;

			StringBuilder sb = new StringBuilder();
			int size = text.Length;
			bool foundLineBreak = false;

			int i = 0;
			while( i < size )
			{
				char ch = text[ i ];

				if( (ch == '\r') || (ch == '\n') )
				{
					if( !foundLineBreak )
					{
						foundLineBreak = true;
						sb.Append( lineBreak );
					}
				}
				else
				{
					foundLineBreak = false;
					sb.Append( ch );
				}

				i++;
			}

			return sb.ToString();
		}

		/// <summary>
		/// Gets the number of times that the specified character is found in the specified text.
		/// </summary>
		/// <param name="text">Text in which to count the specified character. This may be null or empty.</param>
		/// <param name="characterToCount">The character to count.</param>
		/// <returns>Returns the number of times that the specified character is found in the specified text.</returns>
		public static int GetCharacterCount( string text, char characterToCount )
		{
			if( NoContent( text ) )
				return 0;

			int count = 0;
			int size = text.Length;

			for( int i = 0; i < size; i++ )
			{
				if( text[ i ] == characterToCount )
					count++;
			}

			return count;
		}

		/// <summary>
		/// Gets the specified size (in characters) from the specified text.
		/// </summary>
		/// <param name="text">Text whose beginning to get. This may be null or empty.</param>
		/// <param name="sizeMax">Text maximum size (in characters) to get. If this is 0 or a negative number, it's ignored.
		/// </param>
		/// <returns>Returns the beginning of the specified text. This may be null or empty.</returns>
		public static string GetBeginning( string text, int sizeMax )
		{
			if( NoContent( text ) || (sizeMax <= 0) )
				return text;

			return (sizeMax >= text.Length) ? text : text.Substring( 0, sizeMax );
		}

		/// <summary>
		/// Gets the specified size (in characters) from the specified text.
		/// </summary>
		/// <param name="text">Text whose beginning to get. This may be null or empty.</param>
		/// <param name="sizeMax">Text maximum size (in characters) to get, including the size of the "textHasMoreContentSuffix"
		/// text. If this is 0 or a negative number, it's ignored.</param>
		/// <param name="reduceMultipleConsecutiveLineBreaks">If this is true then multiple consecutive line breaks in the
		/// specified text are reduced to a single one.</param>
		/// <param name="lineCountMax">The maximum number of lines of the text to get. If this is 0 or a negative number,
		/// it's ignored.</param>
		/// <param name="textHasMoreContentSuffix">The suffix added to the returned text in order indicate that the specified
		/// text has more content (which is not included in the returned text). This may be null or empty.</param>
		/// <returns>Returns the beginning of the specified text. This may be null or empty.</returns>
		public static string GetBeginning( string text, int sizeMax, bool reduceMultipleConsecutiveLineBreaks, int lineCountMax
			, string textHasMoreContentSuffix )
		{
			if( NoContent( text ) || (sizeMax <= 0) )
				return text;

			// Trim the text.
			text = text.Trim();
			int textSize = text.Length;

			// "textHasMoreContentSuffix" has to fit in the result. At least 1 character from "text" must remain in the result.
			if( HasContent( textHasMoreContentSuffix ) )
				ValidationUtils.Integer( sizeMax, textHasMoreContentSuffix.Length + 1, int.MaxValue, "sizeMax" );

			textHasMoreContentSuffix = (textHasMoreContentSuffix != null) ? textHasMoreContentSuffix : "";

			// Limit the size (in characters).
			string sizeLimitedText = (sizeMax >= textSize) ? text
				: text.Substring( 0, sizeMax - textHasMoreContentSuffix.Length );
			int sizeLimitedTextSize = sizeLimitedText.Length;
			bool hasMoreContent = sizeLimitedTextSize != textSize;

			// Reduce multiple consecutive line breaks.
			string reducedLineBreaks = reduceMultipleConsecutiveLineBreaks
				? ReduceMultipleConsecutiveLineBreaks( sizeLimitedText ) : sizeLimitedText;
			int reducedLineBreaksSize = reducedLineBreaks.Length;

			// Limit the number of lines.
			if( lineCountMax > 0 )
			{
				int lineCount = 0;

				for( int i = 0; i < reducedLineBreaksSize; i++ )
				{
					if( reducedLineBreaks[ i ] == '\n' )
						lineCount++;

					// Append the "has more content" suffix.
					if( lineCount >= lineCountMax )
						return reducedLineBreaks.Substring( 0, i ).Trim() + textHasMoreContentSuffix;
				}
			}

			// Append the "has more content" suffix.
			return reducedLineBreaks.Trim() + (hasMoreContent ? textHasMoreContentSuffix : "");
		}

		/// <summary>
		/// Gets maximum "sizeMax" characters (letters and digits) from the beginning of the specified text.
		/// </summary>
		/// <param name="text">Text from which to get the characters. This may be null or empty.</param>
		/// <param name="sizeMax">The maximum number of characters to get.</param>
		/// <param name="stopAtWhiteSpaceButGetAtLeast">If this is at least 1, the characters will be retrieved up to the first
		/// whitespace character, but not less than this number of non-whitespace characters.</param>
		/// <returns>Returns the extracted characters, as lowercase.</returns>
		public static string GetBeginningLettersAndDigits( string text, int sizeMax, int stopAtWhiteSpaceButGetAtLeast )
		{
			if( NoContent( text ) )
				return text;

			StringBuilder sb = new StringBuilder();

			foreach( char ch in text )
			{
				if( (stopAtWhiteSpaceButGetAtLeast > 0) && (sb.Length >= stopAtWhiteSpaceButGetAtLeast)
					&& char.IsWhiteSpace( ch ) )
				{
					break;
				}

				if( char.IsLetterOrDigit( ch ) )
					sb.Append( ch );

				if( sb.Length >= sizeMax )
					break;
			}

			return sb.ToString().ToLowerInvariant();
		}

		/// <summary>
		/// Deletes the existing line breaks from the end of the specified string builder, adds the specified number of line
		/// breaks, and adds the specified text.
		/// </summary>
		/// <param name="sb">String builder in which to add the line breaks and text. This is modified in here.</param>
		/// <param name="lineBreakCount">The number of line breaks to add. This must be at least 0.</param>
		/// <param name="text">Text to add. If this is null or empty, only the line breaks are added.</param>
		public static void AddWithLineBreaks( StringBuilder sb, int lineBreakCount, string text )
		{
			ValidationUtils.ObjectHasContent( sb, "sb" );
			ValidationUtils.Integer( lineBreakCount, 0, int.MaxValue, "lineBreakCount" );

			// Delete the existing line breaks from the end of the string builder.
			int sbSize = sb.Length;
			for( int i = sbSize - 1; i >= 0; i-- )
			{
				char ch = sb[ i ];
				if( (ch != '\r') && (ch != '\n') )
					break;

				sb.Length--;
			}

			// Add the new line breaks at the end of the string builder.
			if( sb.Length > 0 )
			{
				for( int i = 0; i < lineBreakCount; i++ )
					sb.AppendLine();
			}

			if( HasContent( text ) )
				sb.Append( text );
		}

		/// <summary>
		/// Returns true if any of the texts from the specified list starts the specified text.
		/// </summary>
		/// <param name="text">Text in which to look.</param>
		/// <param name="textList">List of texts to look for in the specified text.</param>
		/// <returns>Returns true if any of the texts from the specified list starts the specified text.</returns>
		public static bool TextStartsWithList( string text, string[] textList )
		{
			ValidationUtils.TextHasContent( text, "text" );
			ValidationUtils.ArrayHasContent( textList, "textList" );

			foreach( string textItem in textList )
			{
				if( text.StartsWith( textItem, StringComparison.OrdinalIgnoreCase ) )
					return true;
			}

			return false;
		}

		/// <summary>
		/// Finds the specified pattern in the specified text.
		/// </summary>
		/// <param name="patternList">The list of patterns to find in the specified text.</param>
		/// <param name="text">The text in which to find the specified pattern.</param>
		/// <returns>If a pattern is found, returns its index in the list of patterns, else returns -1.</returns>
		public static int FindPatternInText( string[] patternList, string text )
		{
			ValidationUtils.ArrayHasContent( patternList, "patternList" );
			ValidationUtils.TextHasContent( text, "text" );

			int count = patternList.Length;
			for( int i = 0; i < count; i++ )
			{
				if( text.IndexOf( patternList[ i ], StringComparison.OrdinalIgnoreCase ) >= 0 )
					return i; // Matching pattern found.
			}

			// No matching pattern found.
			return -1;
		}

		/// <summary>
		/// Concatenates the texts from the specified list.
		/// </summary>
		/// <param name="textList">List of texts to concatenate.</param>
		/// <param name="separator">The separator between the concatenated texts. If this is null or empty, "; " is used
		/// instead.</param>
		/// <returns>Returns the concatenated texts from the specified list.</returns>
		public static string Concatenate( string[] textList, string separator )
		{
			ValidationUtils.ArrayHasContent( textList, "textList" );

			if( NoContent( separator ) )
				separator = "; ";

			StringBuilder sb = new StringBuilder();

			int count = textList.Length;
			for( int i = 0; i < count; i++ )
			{
				sb.Append( textList[ i ] );
				if( i < count - 1 )
					sb.Append( separator );
			}

			return sb.ToString();
		}

		/// <summary>
		/// Converts the specified number to a textual representation (with motes of 3 characters).
		/// </summary>
		/// <param name="number">Number to convert to a textual representation.</param>
		/// <returns>Returns the textual representation of the specified number.</returns>
		public static string NumberToMoteText( int number )
		{
			NumberFormatInfo nfi = (NumberFormatInfo)CultureInfo.InvariantCulture.NumberFormat.Clone();
			nfi.NumberGroupSeparator = NumberMoteSeparator;
			nfi.NumberGroupSizes = new int[] { NumberMoteSize };

			return number.ToString( "N0", nfi );
		}

		/// <summary>
		/// Converts the specified number to a textual representation (with motes of 3 characters).
		/// </summary>
		/// <param name="number">Number to convert to a textual representation.</param>
		/// <returns>Returns the textual representation of the specified number.</returns>
		public static string NumberToMoteText( Int64 number )
		{
			NumberFormatInfo nfi = (NumberFormatInfo)CultureInfo.InvariantCulture.NumberFormat.Clone();
			nfi.NumberGroupSeparator = NumberMoteSeparator;
			nfi.NumberGroupSizes = new int[] { NumberMoteSize };

			return number.ToString( "N0", nfi );
		}

		/// <summary>
		/// Breaks the specified text into motes.
		/// </summary>
		/// <param name="text">Text to break into motes. This may be null or empty.</param>
		/// <param name="moteSize">Mote size.</param>
		/// <param name="moteSeparator">Mote separator.</param>
		/// <returns>Returns the resulted text. This may be null or empty.</returns>
		public static string Mote( string text, int moteSize = MoteSize, string moteSeparator = MoteSeparator )
		{
			if( NoContent( text ) || (moteSize <= 0) || NoContent( moteSeparator ) )
				return text;

			StringBuilder sb = new StringBuilder();

			int count = text.Length;
			for( int i = 0; i < count; i++ )
			{
				if( (i > 0) && (i % moteSize == 0) )
					sb.Append( moteSeparator );
				sb.Append( text[ i ] );
			}

			return sb.ToString();
		}

		#region Radix-16

		/// <summary>
		/// Converts the specified buffer to a radix-16 textual representation (where each byte has two character radix-16
		/// digits).
		/// </summary>
		/// <param name="buffer">Buffer to convert. This may be empty, but not null.</param>
		/// <param name="prefixHexa">If this is true then "0x" will prefix the result.</param>
		/// <param name="moteSize">Mote size. If this is greater than 0, the textual representation is split in motes of this
		/// many characters, separated by a "moteSeparator".</param>
		/// <param name="moteSeparator">Mote separator. This may be null or empty if "moteSize" is 0 or a negative
		/// number.</param>
		/// <returns>Returns a textual radix-16 representation of the specified buffer. This may be empty, but not null.
		/// </returns>
		/// <remarks>The radix-16 text represents the least significant byte of the radix-16 buffer at the left / beginning
		/// of the text. The first radix-16 digit from each pair of two represents the most significant 4 bits of the associated
		/// byte.
		/// </remarks>
		public static string ArrayToText16( byte[] buffer, bool prefixHexa = false, int moteSize = 0
			, string moteSeparator = null )
		{
			ValidationUtils.ObjectHasContent( buffer, "buffer" );
			if( moteSize > 0 )
				ValidationUtils.TextHasContent( moteSeparator, "moteSeparator" );

			int arraySize = buffer.Length;

			if( arraySize <= 0 )
				return "";

			StringBuilder sb = new StringBuilder();
			if( prefixHexa )
				sb.Append( "0x" );

			int ci = 0;
			for( int bi = 0; bi < arraySize; bi++ )
			{
				// First radix-16 digit.
				char ch = TextRadix16Chars[ (buffer[ bi ] & 0xF0) >> 4 ];
				sb.Append( ch );
				ci++;

				// Add the mote separator.
				if( (moteSize > 0) && (ci % moteSize == 0) && (ci < 2 * arraySize) )
					sb.Append( moteSeparator );

				// Second radix-16 digit.
				ch = TextRadix16Chars[ (buffer[ bi ] & 0x0F) ];
				sb.Append( ch );
				ci++;

				// Add the mote separator.
				if( (moteSize > 0) && (ci % moteSize == 0) && (ci < 2 * arraySize) )
					sb.Append( moteSeparator );
			}

			return sb.ToString().ToUpperInvariant();
		}

		/// <summary>
		/// Converts the specified buffer to a radix-16 textual representation (where each byte has two character radix-16
		/// digits).
		/// </summary>
		/// <param name="buffer">Buffer to convert. This may be empty, but not null.</param>
		/// <param name="moteSize">Mote size. If this is greater than 0, the textual representation is split in motes of this
		/// many characters, separated by a "moteSeparator".</param>
		/// <param name="moteSeparator">Mote separator. This may be null or empty if "moteSize" is 0 or a negative
		/// number.</param>
		/// <returns>Returns a textual radix-16 representation of the specified buffer. This may be empty, but not null.
		/// </returns>
		/// <remarks>The radix-16 text represents the least significant byte of the radix-16 buffer at the left / beginning
		/// of the text. The first radix-16 digit from each pair of two represents the most significant 4 bits of the associated
		/// byte.
		/// </remarks>
		public static string ArrayToMoteText16( byte[] buffer, int moteSize = BufferMoteSize
			, string moteSeparator = BufferMoteSeparator )
		{
			return ArrayToText16( buffer, false, moteSize, moteSeparator );
		}

		/// <summary>
		/// Parses the specified character radix-16 digit to an integer radix-16 digit.
		/// </summary>
		/// <param name="digit">Character radix-16 digit to parse. Must be within <see cref="TextRadix16Chars"/>.</param>
		/// <returns>Returns the integer representation of the specified character radix-16 digit.</returns>
		public static int Digit16ToInt( char digit )
		{
			int radix16Value = TextRadix16Chars.IndexOf( digit );

			if( radix16Value < 0 )
			{
				throw new ValidationException( string.Format( "The character \"{0}\" is not a valid radix-16 character."
					, digit ) );
			}

			return radix16Value;
		}

		/// <summary>
		/// Converts the specified radix-16 text to a buffer (where each byte is represented by two character radix-16
		/// digits in the text).
		/// </summary>
		/// <param name="text">Text to convert to a buffer. This may be empty, but not null.</param>
		/// <returns>Returns the resulted buffer. This is empty if "text" is empty.</returns>
		/// <remarks>The radix-16 text represents the least significant byte of the radix-16 buffer at the left / beginning
		/// of the text. The first radix-16 digit from each pair of two represents the most significant 4 bits of the associated
		/// byte.
		/// </remarks>
		public static byte[] Text16ToArray( string text )
		{
			ValidationUtils.ObjectHasContent( text, "text" );

			if( text == "" )
				return new byte[ 0 ];

			// The specified text must have an even number of radix-16 digits in order to fill an integer number of bytes.
			if( (text.Length & 1) != 0 )
			{
				throw new ValidationException( string.Format( "When converting the radix-16 text \"{0}\" to a byte"
					+ " array, the number of radix-16 digits is not even.", text ) );
			}

			text = text.ToLowerInvariant();

			int textSize = text.Length;
			int resultSize = textSize / 2;
			byte[] result = new byte[ resultSize ];

			for( int bi = 0, ci = 0; bi < resultSize; bi++ )
			{
				int d1 = Digit16ToInt( text[ ci ] );
				ci++;
				int d2 = Digit16ToInt( text[ ci ] );
				ci++;

				result[ bi ] = (byte)(d2 + (d1 << 4));
			}

			return result;
		}

		#endregion Radix-16

		#region Radix-32

		/// <summary>
		/// Converts the specified buffer to a radix-32 textual representation.
		/// </summary>
		/// <param name="buffer">Buffer to convert.</param>
		/// <param name="moteSize">Mote size. If this is greater than 0, the textual representation is split in motes of this
		/// many characters, separated by a "moteSeparator".</param>
		/// <param name="moteSeparator">Mote separator. This may be null or empty if "moteSize" is 0 or a negative
		/// number.</param>
		/// <returns>Returns a textual radix-32 representation of the specified buffer.</returns>
		public static string ArrayToText32( byte[] buffer, int moteSize = 0, string moteSeparator = null )
		{
			const int DigitBitCount = 5;

			ValidationUtils.ArrayHasContent( buffer, "buffer" );
			if( moteSize > 0 )
				ValidationUtils.TextHasContent( moteSeparator, "moteSeparator" );

			StringBuilder sb = new StringBuilder();

			int itemsCount = buffer.Length;
			int actualBitCount = itemsCount * 8;
			int digitCount = MathUtils.AlignUp( actualBitCount, DigitBitCount ) / DigitBitCount;

			int firstBitIndex = 0;
			for( int dc = 0; dc < digitCount; dc++ )
			{
				// The bits which are necessary to generate a single character may be distributed in two buffer items. If
				// there are two items, the bits from the second item are at its beginning.

				int firstItemIndex = firstBitIndex >> 3;
				int firstItem = buffer[ firstItemIndex ];

				// The second buffer item may fall outside the specified array.
				int secondItemIndex = firstItemIndex + 1;
				int secondItem = (secondItemIndex < itemsCount) ? buffer[ secondItemIndex ] : 0;

				int firstBitIndexInItem = firstBitIndex & 7;

				int bitsInFirstItem = Math.Min( 8 - firstBitIndexInItem, DigitBitCount );
				int bitsInSecondItem = DigitBitCount - bitsInFirstItem;

				int firstItemMask = (1 << bitsInFirstItem) - 1;
				int secondItemMask = (1 << bitsInSecondItem) - 1;

				int firstItemPart = (firstItem >> firstBitIndexInItem) & firstItemMask;
				int secondItemPart = secondItem & secondItemMask;

				int charIndex = firstItemPart | (secondItemPart << bitsInFirstItem);

				sb.Append( TextRadix32Chars[ charIndex ] );

				// Add the mote separator.
				if( (moteSize > 0) && ((dc + 1) % moteSize == 0) && (dc < digitCount - 1) )
					sb.Append( moteSeparator );

				firstBitIndex += DigitBitCount;
			}

			return sb.ToString().ToUpperInvariant();
		}

		/// <summary>
		/// Converts the specified buffer to a radix-32 textual representation.
		/// </summary>
		/// <param name="buffer">Buffer to convert.</param>
		/// <param name="moteSize">Mote size. If this is greater than 0, the textual representation is split in motes of this
		/// many characters, separated by a "moteSeparator".</param>
		/// <param name="moteSeparator">Mote separator. This may be null or empty if "moteSize" is 0 or a negative
		/// number.</param>
		/// <returns>Returns a textual radix-32 representation of the specified buffer.</returns>
		public static string ArrayToMoteText32( byte[] buffer, int moteSize = BufferMoteSize
			, string moteSeparator = BufferMoteSeparator )
		{
			return ArrayToText32( buffer, moteSize, moteSeparator );
		}

		/// <summary>
		/// Converts the specified radix-32 text to a buffer.
		/// </summary>
		/// <param name="text">Text to convert to a buffer.</param>
		/// <returns>Returns the resulted buffer.</returns>
		public static byte[] Text32ToArray( string text )
		{
			const int DigitBitCount = 5;

			ValidationUtils.TextHasContent( text, "text" );

			text = text.ToLowerInvariant();

			int digitCount = text.Length;
			int paddedBitCount = digitCount * DigitBitCount;
			int itemsCount = paddedBitCount / 8;
			int actualBitCount = itemsCount * 8;
			int paddingBitCount = paddedBitCount - actualBitCount;

			if( paddingBitCount >= DigitBitCount )
			{
				throw new ValidationException( string.Format( "The padding {0} of the radix-32 text is greater than the"
					+ " maximum padding allowed {1}.", paddingBitCount, DigitBitCount - 1 ) );
			}

			#region Verify the last character

			char lastChar = text[ digitCount - 1 ];
			int lastCharRadix32Value = TextRadix32Chars.IndexOf( lastChar );
			int paddingMask = (1 << (DigitBitCount - paddingBitCount)) - 1;

			if( (lastCharRadix32Value < 0) || (lastCharRadix32Value > paddingMask) )
			{
				throw new ValidationException( string.Format( "The last radix-32 character {0} is outside the allowed padding"
					+ " {1}.", lastChar, paddingMask ) );
			}

			#endregion Verify the last character

			byte[] buffer = new byte[ itemsCount ];

			int firstBitIndex = 0;
			for( int dc = 0; dc < digitCount; dc++ )
			{
				char ch = text[ dc ];
				int radix32Value = TextRadix32Chars.IndexOf( ch );

				if( radix32Value < 0 )
				{
					throw new ValidationException( string.Format( "The character \"{0}\" is not a valid radix-32 character."
						, ch ) );
				}

				// A character may have to be distributed in two buffer items. If there are two items, the bits from the
				// second item are at its beginning.

				int firstItemIndex = firstBitIndex >> 3;
				int secondItemIndex = firstItemIndex + 1;

				int firstBitIndexInItem = firstBitIndex & 7;

				int bitsInFirstItem = Math.Min( 8 - firstBitIndexInItem, DigitBitCount );
				int bitsInSecondItem = DigitBitCount - bitsInFirstItem;

				int firstItemMask = (1 << bitsInFirstItem) - 1;
				int secondItemMask = (1 << bitsInSecondItem) - 1;

				buffer[ firstItemIndex ] |= unchecked((byte)((radix32Value & firstItemMask) << firstBitIndexInItem));

				if( (bitsInSecondItem > 0) && (secondItemIndex < itemsCount) )
					buffer[ secondItemIndex ] |= (byte)((radix32Value >> bitsInFirstItem) & secondItemMask);

				firstBitIndex += DigitBitCount;
			}

			return buffer;
		}

		#endregion Radix-32

		/// <summary>
		/// Converts the XML reserved characters into escape sequences.
		/// </summary>
		/// <param name="text">Text to escape. This may be null or empty.</param>
		/// <returns>Returns the escaped text. This may be null or empty.</returns>
		public static string EscapeXml( string text )
		{
			if( NoContent( text ) )
				return text;

			return SecurityElement.Escape( text );
		}

		/// <summary>
		/// Converts the URL escapable characters into escape sequences.
		/// </summary>
		/// <param name="url">URL to escape. This may be null or empty.</param>
		/// <returns>Returns the escaped URL. This may be null or empty.</returns>
		public static string EscapeUrl( string url )
		{
			if( NoContent( url ) )
				return url;

			return Uri.EscapeUriString( url );
		}

		/// <summary>
		/// Converts the URL parameter escapable characters into escape sequences.
		/// </summary>
		/// <param name="urlParameter">URL parameter to escape. This may be null or empty.</param>
		/// <returns>Returns the escaped URL parameter. This may be null or empty.</returns>
		public static string EscapeUrlParameter( string urlParameter )
		{
			if( NoContent( urlParameter ) )
				return urlParameter;

			return Uri.EscapeDataString( urlParameter );
		}

		/// <summary>
		/// Extracts the URL from the specified position in the specified text.
		/// </summary>
		/// <param name="text">Text from which to extract the URL. This may be null or empty.</param>
		/// <param name="position">Position in "text" from which to extract the URL.</param>
		/// <returns>Returns the extracted URL. This may be null or empty.</returns>
		/// <remarks>
		/// <para>The URL is extracted from the specified position, toward left and right until the first whitespace.</para>
		/// </remarks>
		public static string ExtractUrl( string text, int position )
		{
			if( NoContent( text ) )
				return null;
			if( position < 0 )
				return null;

			int size = text.Length;
			int urlBeginning = 0;
			int urlEnd = size - 1;

			position = Math.Min( position, size - 1 );

			// Search the URL beginning toward the left of the text.
			for( int i = position - 1; i >= 0; i-- )
			{
				if( char.IsWhiteSpace( text[ i ] ) )
				{
					urlBeginning = i + 1;
					break;
				}
			}

			// Search the URL end toward the right of the text.
			for( int i = position; i <= size - 1; i++ )
			{
				if( char.IsWhiteSpace( text[ i ] ) )
				{
					urlEnd = i - 1;
					break;
				}
			}

			if( urlEnd <= urlBeginning )
				return null;

			string url = text.Substring( urlBeginning, urlEnd - urlBeginning + 1 );
			if( NoContent( url ) )
				return null;

			if( ValidationUtils.Text( url, -1, -1, TextValidations.Url, null, null, null, false ) != null )
				return null;

			return PrefixUrl( url );
		}

		/// <summary>
		/// Extracts the word from the specified position in the specified text.
		/// </summary>
		/// <param name="text">Text from which to extract the word. This may be null or empty.</param>
		/// <param name="position">Position in "text" from which to extract the word.</param>
		/// <returns>Returns the extracted word. This may be null or empty.</returns>
		public static string ExtractWord( string text, int position )
		{
			if( NoContent( text ) )
				return null;
			if( position < 0 )
				return null;

			int size = text.Length;
			int wordBeginning = 0;
			int wordEnd = size - 1;

			// Search the word beginning toward the left of the text.
			for( int i = position; i >= 1; i-- )
			{
				if( !char.IsLetter( text[ i - 1 ] ) )
				{
					wordBeginning = i;
					break;
				}
			}

			// Search the word end toward the right of the text.
			for( int i = position; i <= size - 2; i++ )
			{
				if( !char.IsLetter( text[ i + 1 ] ) )
				{
					wordEnd = i;
					break;
				}
			}

			return (wordEnd > wordBeginning) ? text.Substring( wordBeginning, wordEnd - wordBeginning + 1 ) : null;
		}

		/// <summary>
		/// Computes the checksum of the specified texts.
		/// </summary>
		/// <param name="texts">Texts to include in the calculation of the checksum.</param>
		/// <returns>Returns the checksum of the specified texts.</returns>
		/// <remarks>
		/// <para>This uses a FNV (Fowler / Noll / Vo) algorithm modified by Bret Mulvey
		/// ( http://papa.bretmulvey.com/post/124027987928/hash-functions ).</para>
		/// 
		/// <para>The method is an alteration which is compatible with JavaScript. This is compatible with the algorithm from
		/// a web application.</para>
		/// 
		/// <para>The returned value is a 16-bit positive integer.</para>
		/// </remarks>
		public static int ComputeChecksum16( params string[] texts )
		{
			ValidationUtils.ArrayHasContent( texts, "texts" );

			foreach( string text in texts )
				ValidationUtils.TextHasContent( text, "text" );

			TextEncoding textEncoding = KernelUtils.GetDeepTextEncoding();

			int p = 401;
			int checksum = 40387;

			foreach( string text in texts )
			{
				byte[] buffer = textEncoding.TextToArray( text, false );

				int size = buffer.Length;
				for( int i = 0; i < size; i++ )
					checksum = unchecked(((checksum ^ buffer[ i ]) * p) & 0xFFFF);
			}

			unchecked
			{
				checksum += checksum << 7;
				checksum ^= checksum >> 4;
				checksum += checksum << 2;
				checksum ^= checksum >> 9;
				checksum += checksum << 3;
			}

			return checksum & 0xFFFF;
		}

		#endregion Public static methods
	}
}
