////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Net;

namespace Vela.Kernel
{
	/// <summary>
	/// Validation utilities.
	/// </summary>
	public static class ValidationUtils
	{
		#region Private constants

		/// <summary>
		/// Exception reason template for a text with an invalid size.
		/// </summary>
		private const string InvalidSizeTextReasonTemplate = "The size of the text must be between {0} and {1} characters, but"
			+ " it's {2}.";

		/// <summary>
		/// Exception reason template for a text with a whitespace in a margin.
		/// </summary>
		private const string MarginWhitespaceTextReasonTemplate = "The text is invalid because it contains a whitespace either"
			+ " at the beginning or at the end.";

		/// <summary>
		/// Exception reason template for a text with a double whitespace.
		/// </summary>
		private const string DoubleWhitespaceTextReasonTemplate = "The text is invalid because it contains a double whitespace.";

		/// <summary>
		/// Exception reason template for a text with an invalid character.
		/// </summary>
		private const string InvalidCharTextReasonTemplate = "The text is invalid because it contains the invalid character"
			+ " \"{0}\".";

		/// <summary>
		/// Exception reason template for a text which must start with a letter, but doesn't.
		/// </summary>
		private const string InvalidFirstLetterTextReasonTemplate = "The text must start with a letter, but doesn't.";

		/// <summary>
		/// Exception reason template for a text which must start with a specific character, but doesn't.
		/// </summary>
		private const string InvalidFirstCharacterTextReasonTemplate = "The text must start with \"{0}\", not \"{1}\".";

		/// <summary>
		/// Exception reason template for an invalid email address.
		/// </summary>
		private const string InvalidEmailAddressTextReasonTemplate = "An email address must contain the \"{0}\" character"
			+ " but not at the beginning or end.";

		/// <summary>
		/// Exception reason template for an invalid alternative peer address.
		/// </summary>
		private const string InvalidAlternativePeerAddressTextReasonTemplate = "Invalid alternative peer address.";

		/// <summary>
		/// Exception reason template for an invalid IP address.
		/// </summary>
		private const string InvalidIpAddressTextReasonTemplate = "Invalid IP address.";

		/// <summary>
		/// Exception reason template for an invalid URL.
		/// </summary>
		private const string InvalidUrlTextReasonTemplate = "Invalid URL.";

		/// <summary>
		/// Exception reason template for an array with an invalid size.
		/// </summary>
		private const string InvalidSizeArrayReasonTemplate = "The size of the array must be between {0} and {1} elements, but"
			+ " it's {2}.";

		/// <summary>
		/// Exception reason template for a list with an invalid size.
		/// </summary>
		private const string InvalidSizeListReasonTemplate = "The size of the list must be between {0} and {1} elements, but"
			+ " it's {2}.";

		#endregion Private constants

		#region Static constructors

		/// <summary>
		/// Initializes the current class.
		/// </summary>
		static ValidationUtils()
		{
			InvalidFileTitleChars = new HashSet<char>( Path.GetInvalidFileNameChars() );
			InvalidFileNameChars = new HashSet<char>( Path.GetInvalidPathChars() );
		}

		#endregion Static constructors

		#region Private fields

		/// <summary>
		/// The characters which are invalid for file titles, as indicated by the system.
		/// </summary>
		private static HashSet<char> InvalidFileTitleChars;

		/// <summary>
		/// The characters which are invalid for file names, as indicated by the system.
		/// </summary>
		private static HashSet<char> InvalidFileNameChars;

		#endregion Private fields

		#region Private static methods

		/// <summary>
		/// Throws an exception with the specified type.
		/// </summary>
		/// <param name="message">The description of the exception. This may be null or empty.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		/// <returns>Returns the thrown exception. Since the exception is thrown in here, this is superfluous but is necessary
		/// sometimes because the compiler generates an error when a method should either return a value or throw an exception,
		/// but sees neither.</returns>
		private static Exception ThrowException( string message, string errorContext, object errorControl
			, object errorPresenter )
		{
			Exception exception;

			if( (errorControl == null) && (errorPresenter == null) )
			{
				exception = new ValidationException( message, null, errorContext );
			}
			else
			{
				exception = new PresentedValidationException( message, null, errorContext, errorControl, errorPresenter );
			}

			throw exception;
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Throws an <see cref="ObjectNotInitializedException"/> exception.
		/// </summary>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <remarks>
		/// <para>This should be used in property getters when the property is not initialized.</para>
		/// </remarks>
		public static void ThrowObjectNotInitialized( string errorContext )
		{
			throw new ObjectNotInitializedException( "The object is missing. The instance data is null, but the object doesn't"
				+ " allow null a data.", null, errorContext );
		}

		/// <summary>
		/// Validates the specified data. A <see cref="ValidationException"/> exception is thrown if the data is invalid and
		/// "throwIfInvalid" is true.
		/// </summary>
		/// <param name="data">Data to validate. This may be null or empty, if the minimum size allows it.</param>
		/// <param name="sizeMin">Minimum size allowed. If this is 0 then an empty (but not null) data is allowed, if it's -1
		/// then the data must have content (if there is a default minimum size, it's applied), if it's -2 (or less) then the
		/// data may be null or empty but if it has content then the default minimum size (if there is any) is applied.</param>
		/// <param name="sizeMax">Maximum size allowed. If this is 0 then only an empty or null data is allowed, if it's -1
		/// then any maximum size is allowed except if there is a default maximum size, if it's -2 (or less) then any maximum
		/// size is allowed.</param>
		/// <param name="contentType">The type of content of the data.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		/// <param name="throwIfInvalid">If this is true and the text is invalid then an exception is thrown.</param>
		/// <returns>Returns null if the text is valid, else (if "throwIfInvalid" is false) returns the reason why the text is
		/// invalid.</returns>
		public static string Text( string data, int sizeMin, int sizeMax, TextValidations contentType, string errorContext
			, object errorControl = null, object errorPresenter = null, bool throwIfInvalid = true )
		{
			string reason;

			if( data == null )
			{
				if( sizeMin <= -2 )
					return null; // Valid null text.
				else
				{
					reason = "The text is missing.";
					if( !throwIfInvalid )
						return reason;

					ThrowException( reason, errorContext, errorControl, errorPresenter );
				}
			}

			int size = data.Length;

			#region Validate size

			if( (size <= 0) && ((sizeMin == 0) || (sizeMin <= -2)) )
				return null;

			// Get the default minimum size.
			if( sizeMin < 0 )
			{
				switch( contentType )
				{
					case TextValidations.Alias:
						sizeMin = KernelConsts.AliasSizeMin;
						break;
					case TextValidations.EmoticonSymbol:
						sizeMin = KernelConsts.EmoticonSymbolSizeMin;
						break;
					case TextValidations.EntityName:
						sizeMin = KernelConsts.EntityNameSizeMin;
						break;
					case TextValidations.Password:
						sizeMin = KernelConsts.PasswordSizeMin;
						break;
					default:
						// From this point on, the text has at least 1 character.
						sizeMin = 1;
						break;
				}
			}

			// Get the default maximum size.
			if( sizeMax < 0 )
			{
				switch( contentType )
				{
					case TextValidations.Alias:
						sizeMax = KernelConsts.AliasSizeMax;
						break;
					case TextValidations.AlternativePeerAddress:
						sizeMax = KernelConsts.AlternativePeerAddressSizeMax;
						break;
					case TextValidations.Code:
						sizeMax = KernelConsts.CodeSizeMax;
						break;
					case TextValidations.EmailAddress:
						sizeMax = KernelConsts.EmailAddressSizeMax;
						break;
					case TextValidations.EmoticonSymbol:
						sizeMax = KernelConsts.EmoticonSymbolSizeMax;
						break;
					case TextValidations.EntityName:
						sizeMax = KernelConsts.EntityNameSizeMax;
						break;
					case TextValidations.FileName:
						sizeMax = KernelConsts.FileNameSizeMax;
						break;
					case TextValidations.FileTitle:
						sizeMax = KernelConsts.FileTitleSizeMax;
						break;
					case TextValidations.IpAddress:
						sizeMax = KernelConsts.IpAddressSizeMax;
						break;
					case TextValidations.LongNote:
						sizeMax = KernelConsts.LongNoteSizeMax;
						break;
					case TextValidations.MediumNote:
						sizeMax = KernelConsts.MediumNoteSizeMax;
						break;
					case TextValidations.Password:
						sizeMax = KernelConsts.PasswordSizeMax;
						break;
					case TextValidations.PhoneNumber:
						sizeMax = KernelConsts.PhoneNumberSizeMax;
						break;
					case TextValidations.ShortCommand:
						sizeMax = KernelConsts.ShortCommandSizeMax;
						break;
					case TextValidations.ShortNote:
						sizeMax = KernelConsts.ShortNoteSizeMax;
						break;
					case TextValidations.TinyNote:
						sizeMax = KernelConsts.TinyNoteSizeMax;
						break;
					case TextValidations.Upn:
						sizeMax = KernelConsts.UpnSizeMax;
						break;
					case TextValidations.Url:
						sizeMax = KernelConsts.UrlSizeMax;
						break;
				}
			}

			if( ((sizeMin >= 0) && (size < sizeMin)) || ((sizeMax >= 0) && (size > sizeMax)) )
			{
				reason = string.Format( InvalidSizeTextReasonTemplate, sizeMin, sizeMax, size );
				if( !throwIfInvalid )
					return reason;

				ThrowException( reason, errorContext, errorControl, errorPresenter );
			}

			#endregion Validate size

			if( (contentType == TextValidations.None) || (contentType == TextValidations.LongNote)
				|| (contentType == TextValidations.MediumNote) || (contentType == TextValidations.ShortCommand)
				|| (contentType == TextValidations.ShortNote) || (contentType == TextValidations.TinyNote) )
			{
				// Don't check the content because anything is valid.
				return null;
			}

			#region Validate spaces

			// Don't allow whitespaces at the beginning or end of the text.

			// Beginning of text.
			if( char.IsWhiteSpace( data[ 0 ] ) )
			{
				reason = MarginWhitespaceTextReasonTemplate;
				if( !throwIfInvalid )
					return reason;

				ThrowException( reason, errorContext, errorControl, errorPresenter );
			}

			// End of text.
			if( char.IsWhiteSpace( data[ size - 1 ] ) )
			{
				reason = MarginWhitespaceTextReasonTemplate;
				if( !throwIfInvalid )
					return reason;

				ThrowException( reason, errorContext, errorControl, errorPresenter );
			}

			// Don't allow double whitespaces.
			bool hasPrevSpace = false;
			for( int i = 0; i < size; i++ )
			{
				bool isWhiteSpace = char.IsWhiteSpace( data[ i ] );

				if( hasPrevSpace && isWhiteSpace )
				{
					reason = DoubleWhitespaceTextReasonTemplate;
					if( !throwIfInvalid )
						return reason;

					ThrowException( reason, errorContext, errorControl, errorPresenter );
				}

				hasPrevSpace = isWhiteSpace;
			}

			#endregion Validate spaces

			#region Validate first character

			switch( contentType )
			{
				case TextValidations.EmailAddress:
				case TextValidations.EntityName:
					// The first character must be a letter.

					char ch = data[ 0 ];
					UnicodeCategory category = char.GetUnicodeCategory( ch );

					if( (category != UnicodeCategory.LowercaseLetter) && (category != UnicodeCategory.UppercaseLetter) )
					{
						reason = InvalidFirstLetterTextReasonTemplate;
						if( !throwIfInvalid )
							return reason;

						ThrowException( reason, errorContext, errorControl, errorPresenter );
					}

					break;
				default:
					// Nothing to do.
					break;
			}

			#endregion Validate first character

			#region Validate content

			switch( contentType )
			{
				case TextValidations.Alias:
				case TextValidations.Code:
				case TextValidations.EmoticonSymbol:
				case TextValidations.FileName:
				case TextValidations.FileTitle:
				case TextValidations.Letters:
				case TextValidations.EntityName:
				case TextValidations.Password:
				case TextValidations.PhoneNumber:
				case TextValidations.Upn:
					// Nothing to do.
					break;
				case TextValidations.AlternativePeerAddress:
					if( !AlternativePeerAddress.HasAddress( data ) )
					{
						reason = InvalidAlternativePeerAddressTextReasonTemplate;
						if( !throwIfInvalid )
							return reason;

						ThrowException( reason, errorContext, errorControl, errorPresenter );
					}

					break;
				case TextValidations.EmailAddress:
					// "@" must exist (but not at the beginning or end).

					int aroundPos = data.IndexOf( '@' );
					if( (aroundPos <= 0) || (aroundPos >= data.Length - 1) )
					{
						reason = string.Format( InvalidEmailAddressTextReasonTemplate, '@' );
						if( !throwIfInvalid )
							return reason;

						ThrowException( reason, errorContext, errorControl, errorPresenter );
					}

					break;
				case TextValidations.IpAddress:
					IPAddress ipAddress;
					if( !IPAddress.TryParse( data, out ipAddress ) )
					{
						reason = InvalidIpAddressTextReasonTemplate;
						if( !throwIfInvalid )
							return reason;

						ThrowException( reason, errorContext, errorControl, errorPresenter );
					}

					break;
				case TextValidations.Url:
					// "." must exist (but not at the beginning or end).

					int dotPos = data.IndexOf( '.' );
					if( (dotPos <= 0) || (dotPos >= data.Length - 1)
						|| !Uri.IsWellFormedUriString( data, UriKind.RelativeOrAbsolute ) )
					{
						reason = InvalidUrlTextReasonTemplate;
						if( !throwIfInvalid )
							return reason;

						ThrowException( reason, errorContext, errorControl, errorPresenter );
					}

					break;
				default:
					string message = string.Format( "Invalid text content type \"{0}\" while validating the text \"{1}\"."
						, contentType, data );
					ThrowException( message, errorContext, errorControl, errorPresenter );
					break;
			}

			#endregion Validate content

			if( (contentType == TextValidations.AlternativePeerAddress) || (contentType == TextValidations.Url) )
				return null;

			#region Validate all characters

			for( int i = 0; i < size; i++ )
			{
				char ch = data[ i ];
				UnicodeCategory category = char.GetUnicodeCategory( ch );

				// Control characters are not allowed.
				if( category == UnicodeCategory.Control )
				{
					reason = string.Format( InvalidCharTextReasonTemplate, ch );
					if( !throwIfInvalid )
						return reason;

					ThrowException( reason, errorContext, errorControl, errorPresenter );
				}

				switch( contentType )
				{
					case TextValidations.Alias:
						if( (category != UnicodeCategory.LowercaseLetter)
							&& (category != UnicodeCategory.UppercaseLetter)
							&& (category != UnicodeCategory.DecimalDigitNumber)
							&& !KernelConsts.AliasNanValidCharacters.Contains( ch ) )
						{
							reason = string.Format( InvalidCharTextReasonTemplate, ch );
							if( !throwIfInvalid )
								return reason;

							ThrowException( reason, errorContext, errorControl, errorPresenter );
						}
						break;
					case TextValidations.Code:
						if( (category != UnicodeCategory.LowercaseLetter)
							&& (category != UnicodeCategory.UppercaseLetter)
							&& (category != UnicodeCategory.DecimalDigitNumber) )
						{
							reason = string.Format( InvalidCharTextReasonTemplate, ch );
							if( !throwIfInvalid )
								return reason;

							ThrowException( reason, errorContext, errorControl, errorPresenter );
						}
						break;
					case TextValidations.EmailAddress:
						if( (category != UnicodeCategory.LowercaseLetter)
							&& (category != UnicodeCategory.UppercaseLetter)
							&& (category != UnicodeCategory.DecimalDigitNumber)
							&& !KernelConsts.EmailAddressNanValidCharacters.Contains( ch ) )
						{
							reason = string.Format( InvalidCharTextReasonTemplate, ch );
							if( !throwIfInvalid )
								return reason;

							ThrowException( reason, errorContext, errorControl, errorPresenter );
						}
						break;
					case TextValidations.EmoticonSymbol:
						if( (category != UnicodeCategory.LowercaseLetter)
							&& (category != UnicodeCategory.UppercaseLetter)
							&& (category != UnicodeCategory.DecimalDigitNumber)
							&& !KernelConsts.EmoticonSymbolNanValidCharacters.Contains( ch ) )
						{
							reason = string.Format( InvalidCharTextReasonTemplate, ch );
							if( !throwIfInvalid )
								return reason;

							ThrowException( reason, errorContext, errorControl, errorPresenter );
						}
						break;
					case TextValidations.EntityName:
						if( (category != UnicodeCategory.LowercaseLetter)
							&& (category != UnicodeCategory.UppercaseLetter)
							&& (category != UnicodeCategory.SpaceSeparator) )
						{
							reason = string.Format( InvalidCharTextReasonTemplate, ch );
							if( !throwIfInvalid )
								return reason;

							ThrowException( reason, errorContext, errorControl, errorPresenter );
						}
						break;
					case TextValidations.FileName:
						if( InvalidFileNameChars.Contains( ch ) )
						{
							reason = string.Format( InvalidCharTextReasonTemplate, ch );
							if( !throwIfInvalid )
								return reason;

							ThrowException( reason, errorContext, errorControl, errorPresenter );
						}
						break;
					case TextValidations.FileTitle:
						if( InvalidFileTitleChars.Contains( ch ) )
						{
							reason = string.Format( InvalidCharTextReasonTemplate, ch );
							if( !throwIfInvalid )
								return reason;

							ThrowException( reason, errorContext, errorControl, errorPresenter );
						}
						break;
					case TextValidations.IpAddress:
						// The validation is separate.
						break;
					case TextValidations.Letters:
						if( (category != UnicodeCategory.LowercaseLetter)
							&& (category != UnicodeCategory.UppercaseLetter) )
						{
							reason = string.Format( InvalidCharTextReasonTemplate, ch );
							if( !throwIfInvalid )
								return reason;

							ThrowException( reason, errorContext, errorControl, errorPresenter );
						}
						break;
					case TextValidations.Password:
						if( (category != UnicodeCategory.LowercaseLetter)
							&& (category != UnicodeCategory.UppercaseLetter)
							&& (category != UnicodeCategory.DecimalDigitNumber)
							&& !KernelConsts.PasswordNanValidCharacters.Contains( ch ) )
						{
							reason = string.Format( InvalidCharTextReasonTemplate, ch );
							if( !throwIfInvalid )
								return reason;

							ThrowException( reason, errorContext, errorControl, errorPresenter );
						}
						break;
					case TextValidations.PhoneNumber:
						if( (category != UnicodeCategory.DecimalDigitNumber)
							&& !KernelConsts.PhoneNumberNanValidCharacters.Contains( ch ) )
						{
							reason = string.Format( InvalidCharTextReasonTemplate, ch );
							if( !throwIfInvalid )
								return reason;

							ThrowException( reason, errorContext, errorControl, errorPresenter );
						}
						break;
					case TextValidations.Upn:
						if( (category != UnicodeCategory.LowercaseLetter)
							&& (category != UnicodeCategory.UppercaseLetter)
							&& (category != UnicodeCategory.DecimalDigitNumber)
							&& (category != UnicodeCategory.SpaceSeparator) )
						{
							reason = string.Format( InvalidCharTextReasonTemplate, ch );
							if( !throwIfInvalid )
								return reason;

							ThrowException( reason, errorContext, errorControl, errorPresenter );
						}
						break;
					default:
						string message = string.Format( "Invalid text content type \"{0}\" while validating the text \"{1}\"."
							, contentType, data );
						ThrowException( message, errorContext, errorControl, errorPresenter );
						break;
				}
			}

			#endregion Validate all characters

			// Valid text.
			return null;
		}

		/// <summary>
		/// Validates the specified data. A <see cref="ValidationException"/> exception is thrown if the data is missing.
		/// </summary>
		/// <param name="data">Data to validate.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void TextHasContent( string data, string errorContext, object errorControl = null
			, object errorPresenter = null )
		{
			if( TextUtils.NoContent( data ) )
				ThrowException( "The text has no content.", errorContext, errorControl, errorPresenter );
		}

		/// <summary>
		/// Validates the specified data. A <see cref="ValidationException"/> exception is thrown if the data is null.
		/// </summary>
		/// <param name="data">Data to validate.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void ObjectHasContent( object data, string errorContext, object errorControl = null
			, object errorPresenter = null )
		{
			if( data == null )
				ThrowException( "The object is missing.", errorContext, errorControl, errorPresenter );
		}

		/// <summary>
		/// Checks if the specified data has the specified type. A <see cref="ValidationException"/> exception is thrown if the
		/// data is null or has a different type.
		/// </summary>
		/// <typeparam name="TObject">The type that the data must have.</typeparam>
		/// <param name="data">Data to validate.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void ObjectType<TObject>( object data, string errorContext, object errorControl = null
			, object errorPresenter = null )
		{
			if( data == null )
			{
				string message = string.Format( "The object whose type must be \"{0}\" is missing."
					, typeof( TObject ).FullName );
				ThrowException( message, errorContext, errorControl, errorPresenter );
			}
			else
			{
				if( !typeof( TObject ).IsInstanceOfType( data ) )
				{
					string message = string.Format( "The object of type \"{0}\" must be of type \"{1}\"."
						, data.GetType().FullName, typeof( TObject ).FullName );
					ThrowException( message, errorContext, errorControl, errorPresenter );
				}
			}
		}

		/// <summary>
		/// Checks if the specified data has one of the specified types. A <see cref="ValidationException"/> exception is thrown
		/// if the data is null or has a different type.
		/// </summary>
		/// <typeparam name="TObject1">Type that the data may have.</typeparam>
		/// <typeparam name="TObject2">Type that the data may have.</typeparam>
		/// <param name="data">Data to validate.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void ObjectType<TObject1, TObject2>( object data, string errorContext, object errorControl = null
			, object errorPresenter = null )
		{
			if( data == null )
			{
				string message = string.Format( "The object whose type must be \"{0}\" or \"{1}\" is missing."
					, typeof( TObject1 ).FullName, typeof( TObject2 ).FullName );
				ThrowException( message, errorContext, errorControl, errorPresenter );
			}
			else
			{
				if( !typeof( TObject1 ).IsInstanceOfType( data ) && !typeof( TObject2 ).IsInstanceOfType( data ) )
				{
					string message = string.Format( "The object of type \"{0}\" must be of type \"{1}\" or \"{2}\"."
						, data.GetType().FullName, typeof( TObject1 ).FullName, typeof( TObject2 ).FullName );
					ThrowException( message, errorContext, errorControl, errorPresenter );
				}
			}
		}

		/// <summary>
		/// Checks if the specified data has the specified type. A <see cref="ValidationException"/> exception is thrown if the
		/// data is null or has a different type.
		/// </summary>
		/// <typeparam name="TObject">The type that the data must have.</typeparam>
		/// <param name="data">Data to validate.</param>
		/// <param name="locator">The locator of the data in the storehouse.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void StorehouseContains<TObject>( object data, Locator locator, object errorControl = null
			, object errorPresenter = null )
		{
			if( data == null )
			{
				string message = string.Format( "The storehouse object with the locator \"{0}\" and type \"{1}\" is missing."
					, locator, typeof( TObject ).FullName );
				ThrowException( message, null, errorControl, errorPresenter );
			}
			else
			{
				if( !typeof( TObject ).IsInstanceOfType( data ) )
				{
					string message = string.Format( "The storehouse object with the locator \"{0}\" must be of type \"{1}\","
						+ " not \"{2}\".", locator, typeof( TObject ).FullName, data.GetType().FullName );
					ThrowException( message, null, errorControl, errorPresenter );
				}
			}
		}

		/// <summary>
		/// Checks if the specified data is either a buffer or a text. A <see cref="ValidationException"/> exception is
		/// thrown if the data is invalid.
		/// </summary>
		/// <param name="data">Data to validate. This may be either a buffer or a text. This may be empty, but not null.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void BufferOrText( object data, string errorContext, object errorControl = null
			, object errorPresenter = null )
		{
			ObjectHasContent( data, "data" );

			if( !(data is byte[]) && !(data is string) )
			{
				string message = string.Format( "The data must be either a buffer or a text, not a \"{0}\"."
					, data.GetType().FullName );
				ThrowException( message, errorContext, errorControl, errorPresenter );
			}
		}

		/// <summary>
		/// Validates the specified data. A <see cref="ValidationException"/> exception is thrown if the data is invalid.
		/// </summary>
		/// <param name="data">Data to validate. This may be null or empty, if the minimum size allows it.</param>
		/// <param name="sizeMin">Minimum size allowed. If this is 0 then an empty (but not null) data is allowed, if it's a
		/// negative number then any minimum size or empty or null is allowed. </param>
		/// <param name="sizeMax">Maximum size allowed. If this is 0 then only an empty or null data is allowed, if it's a
		/// negative number then any maximum size is allowed.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void Array( Array data, int sizeMin, int sizeMax, string errorContext, object errorControl = null
			, object errorPresenter = null )
		{
			if( data == null )
			{
				if( sizeMin < 0 )
					return; // Valid null array.
				else
					ThrowException( "The array is missing.", errorContext, errorControl, errorPresenter );
			}

			int size = data.Length;

			if( ((sizeMin >= 0) && (size < sizeMin)) || ((sizeMax >= 0) && (size > sizeMax)) )
			{
				string reason = string.Format( InvalidSizeArrayReasonTemplate, sizeMin, sizeMax, size );
				ThrowException( reason, errorContext, errorControl, errorPresenter );
			}
		}

		/// <summary>
		/// Validates the specified data. A <see cref="ValidationException"/> exception is thrown if the data is missing.
		/// </summary>
		/// <param name="data">Data to validate.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void ArrayHasContent( Array data, string errorContext, object errorControl = null
			, object errorPresenter = null )
		{
			if( ArrayUtils.NoContent( data ) )
				ThrowException( "The array has no content.", errorContext, errorControl, errorPresenter );
		}

		/// <summary>
		/// Validates the specified data. A <see cref="ValidationException"/> exception is thrown if the data is invalid.
		/// </summary>
		/// <param name="data">Data to validate.</param>
		/// <param name="sizeMin">Minimum size allowed. If this is 0 then an empty (but not null) data is allowed, if it's a
		/// negative number then any minimum size or empty or null is allowed. </param>
		/// <param name="sizeMax">Maximum size allowed. If this is 0 then only an empty or null data is allowed, if it's a
		/// negative number then any maximum size is allowed.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void List( IKernelObservableList data, int sizeMin, int sizeMax, string errorContext
			, object errorControl = null, object errorPresenter = null )
		{
			if( data == null )
			{
				if( sizeMin < 0 )
					return; // Valid null list.
				else
					ThrowException( "The list is missing.", errorContext, errorControl, errorPresenter );
			}

			int size = data.Count;

			if( ((sizeMin >= 0) && (size < sizeMin)) || ((sizeMax >= 0) && (size > sizeMax)) )
			{
				string reason = string.Format( InvalidSizeListReasonTemplate, sizeMin, sizeMax, size );
				ThrowException( reason, errorContext, errorControl, errorPresenter );
			}
		}

		/// <summary>
		/// Validates the specified data. A <see cref="ValidationException"/> exception is thrown if the data is invalid.
		/// </summary>
		/// <param name="data">Data to validate.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void ListHasContent( IKernelObservableList data, string errorContext, object errorControl = null
			, object errorPresenter = null )
		{
			if( ListUtils.NoContent( data ) )
				ThrowException( "The list has no content.", errorContext, errorControl, errorPresenter );
		}

		/// <summary>
		/// Validates the specified data. A <see cref="ValidationException"/> exception is thrown if the data is invalid.
		/// </summary>
		/// <param name="data">Data to validate.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void ListHasContent( ICollection data, string errorContext, object errorControl = null
			, object errorPresenter = null )
		{
			if( ListUtils.NoContent( data ) )
				ThrowException( "The list has no content.", errorContext, errorControl, errorPresenter );
		}

		/// <summary>
		/// Validates the specified data. A <see cref="ValidationException"/> exception is thrown if the data is invalid.
		/// </summary>
		/// <param name="data">Data to validate.</param>
		/// <param name="dataMin">Minimum data allowed.</param>
		/// <param name="dataMax">Maximum data allowed.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void Integer( int data, int dataMin, int dataMax, string errorContext, object errorControl = null
			, object errorPresenter = null )
		{
			if( (data < dataMin) || (data > dataMax) )
			{
				string message = string.Format( "The integer data {0} is outside the required bounds [{1}...{2}]."
					, data, dataMin, dataMax );
				ThrowException( message, errorContext, errorControl, errorPresenter );
			}
		}

		/// <summary>
		/// Validates the specified data. A <see cref="ValidationException"/> exception is thrown if the data is invalid.
		/// </summary>
		/// <param name="data">Data to validate.</param>
		/// <param name="dataMin">Minimum data allowed.</param>
		/// <param name="dataMax">Maximum data allowed.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void Integer( Int64 data, Int64 dataMin, Int64 dataMax, string errorContext
			, object errorControl = null, object errorPresenter = null )
		{
			if( (data < dataMin) || (data > dataMax) )
			{
				string message = string.Format( "The integer data {0} is outside the required bounds [{1}...{2}]."
					, data, dataMin, dataMax );
				ThrowException( message, errorContext, errorControl, errorPresenter );
			}
		}

		/// <summary>
		/// Validates the specified data. A <see cref="ValidationException"/> exception is thrown if the data is invalid.
		/// </summary>
		/// <param name="data">Data to validate.</param>
		/// <param name="dataMin">Minimum data allowed.</param>
		/// <param name="dataMax">Maximum data allowed.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void Integer( UInt64 data, UInt64 dataMin, UInt64 dataMax, string errorContext
			, object errorControl = null, object errorPresenter = null )
		{
			if( (data < dataMin) || (data > dataMax) )
			{
				string message = string.Format( "The integer data {0} is outside the required bounds [{1}...{2}]."
					, data, dataMin, dataMax );
				ThrowException( message, errorContext, errorControl, errorPresenter );
			}
		}

		/// <summary>
		/// Validates the specified data. A <see cref="ValidationException"/> exception is thrown if the data is invalid.
		/// </summary>
		/// <param name="data">Data (in textual representation) to validate.</param>
		/// <param name="dataMin">Minimum data allowed.</param>
		/// <param name="dataMax">Maximum data allowed.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		/// <returns>Returns the validated data.</returns>
		public static int Integer( string data, int dataMin, int dataMax, string errorContext, object errorControl = null
			, object errorPresenter = null )
		{
			int integer;
			bool parsed = int.TryParse( data, out integer );

			if( !parsed || (integer < dataMin) || (integer > dataMax) )
			{
				string message = string.Format( "The integer data {0} is outside the required bounds [{1}...{2}]."
					, data, dataMin, dataMax );
				ThrowException( message, errorContext, errorControl, errorPresenter );
			}

			return integer;
		}

		/// <summary>
		/// Validates the specified data. A <see cref="ValidationException"/> exception is thrown if the data is invalid.
		/// </summary>
		/// <param name="data">Data to validate.</param>
		/// <param name="dataMin">Minimum data allowed.</param>
		/// <param name="dataMax">Maximum data allowed.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void Real( double data, double dataMin, double dataMax, string errorContext, object errorControl = null
			, object errorPresenter = null )
		{
			if( (data < dataMin) || (data > dataMax) )
			{
				string message = string.Format( "The real data {0} is outside the required bounds [{1}...{2}]."
					, data, dataMin, dataMax );
				ThrowException( message, errorContext, errorControl, errorPresenter );
			}
		}

		/// <summary>
		/// Validates the specified data. A <see cref="ValidationException"/> exception is thrown if the data is invalid.
		/// </summary>
		/// <param name="data">Data to validate.</param>
		/// <param name="dataMin">Minimum data allowed.</param>
		/// <param name="dataMax">Maximum data allowed.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void Time( DateTime data, DateTime dataMin, DateTime dataMax, string errorContext
			, object errorControl = null, object errorPresenter = null )
		{
			if( (data < dataMin) || (data > dataMax) )
			{
				string message = string.Format( "The time {0} is outside the required bounds [{1}...{2}]."
					, data.ToString( KernelConsts.TimeFormat ), dataMin.ToString( KernelConsts.TimeFormat )
					, dataMax.ToString( KernelConsts.TimeFormat ) );
				ThrowException( message, errorContext, errorControl, errorPresenter );
			}
		}

		/// <summary>
		/// Validates the specified data. A <see cref="ValidationException"/> exception is thrown if the data is invalid.
		/// </summary>
		/// <param name="data">Data to validate.</param>
		/// <param name="dataMin">Minimum data allowed.</param>
		/// <param name="dataMax">Maximum data allowed.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void Date( string data, DateTime dataMin, DateTime dataMax, string errorContext
			, object errorControl = null, object errorPresenter = null )
		{
			string[] dateFormats = { KernelConsts.DateFormat, KernelConsts.DateUiLongFormat, KernelConsts.DateUiShortFormat };

			DateTime date = DateTime.MinValue;

			try
			{
				date = DateTime.ParseExact( data, dateFormats, null, DateTimeStyles.RoundtripKind );
			}
			catch( Exception ex )
			{
				ThrowException( ex.Message, errorContext, errorControl, errorPresenter );
			}

			Time( date, dataMin, dataMax, errorContext, errorControl, errorPresenter );
		}

		/// <summary>
		/// Considers that the specified data is unhandled (like an enumeration), and throws a
		/// <see cref="ValidationException"/> exception.
		/// </summary>
		/// <param name="data">Data to validate.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		/// <returns>Returns the thrown exception. Since the exception is thrown in here, this is superfluous but is necessary
		/// sometimes because the compiler generates an error when a method should either return a value or throw an exception,
		/// but sees neither.</returns>
		public static Exception Unhandled( object data, string errorContext, object errorControl = null
			, object errorPresenter = null )
		{
			string message = string.Format( "The data \"{0}\" of the \"{1}\" type is invalid.", data, data.GetType().FullName );

			return ThrowException( message, errorContext, errorControl, errorPresenter );
		}

		/// <summary>
		/// Validates the specified password. A <see cref="ValidationException"/> exception is thrown if the data is invalid.
		/// </summary>
		/// <param name="password">Password to validate. This is to be used for symmetric encryption. This may be a buffer
		/// or a text.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		public static void Password( object password, string errorContext, object errorControl = null
			, object errorPresenter = null )
		{
			if( TextUtils.NoContent( errorContext ) )
				errorContext = "Password";

			if( password == null )
				ThrowException( "The password is missing.", errorContext, errorControl, errorPresenter );

			if( password is byte[] )
			{
				byte[] typedPassword = (byte[])password;
				if( typedPassword.Length < KernelConsts.BinaryPasswordSizeMin )
				{
					string message = string.Format( "The buffer password size must be at least {0} bytes."
						, KernelConsts.BinaryPasswordSizeMin );
					ThrowException( message, errorContext, errorControl, errorPresenter );
				}
			}
			else if( password is string )
			{
				Text( password as string, -1, -1, TextValidations.Password, errorContext, errorControl, errorPresenter );
			}
			else
			{
				string message = string.Format( string.Format( "The password must be either a buffer or a text, not a"
					+ " \"{0}\".", password.GetType().FullName ) );
				ThrowException( message, errorContext, errorControl, errorPresenter );
			}

			// The password is valid.
		}

		/// <summary>
		/// Checks if the specified passwords are identical. A <see cref="ValidationException"/> exception is thrown if the
		/// passwords are different and "throwIfDifferent" is true.
		/// </summary>
		/// <param name="password">Password to compare. This may be null or empty.</param>
		/// <param name="confirmationPassword">Password to compare. This may be null or empty.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <param name="errorControl">The control which caused the exception. This may be null.</param>
		/// <param name="errorPresenter">The control which must show that there is a validation error. This may be null.</param>
		/// <param name="throwIfDifferent">If this is true and the passwords are different then an exception is thrown.</param>
		/// <returns>Returns true if the passwords are identical; returns false if the passwords are different and
		/// "throwIfDifferent" is false.</returns>
		public static bool PasswordConfirm( string password, string confirmationPassword, string errorContext
			, object errorControl = null, object errorPresenter = null, bool throwIfDifferent = true )
		{
			if( TextUtils.NoContent( errorContext ) )
				errorContext = "Confirmation password";

			if( password == null )
			{
				if( confirmationPassword == null )
					return true;
				else
				{
					if( !throwIfDifferent )
						return false;

					throw ThrowException( "The password is not confirmed.", errorContext, errorControl, errorPresenter );
				}
			}
			else
			{
				if( confirmationPassword == null )
				{
					if( !throwIfDifferent )
						return false;

					throw ThrowException( "The password is not confirmed.", errorContext, errorControl, errorPresenter );
				}
				else
				{
					// Compare the passwords.
					if( password == confirmationPassword )
					{
						// The passwords are identical.
						return true;
					}
					else
					{
						if( !throwIfDifferent )
							return false;

						throw ThrowException( "The password is not confirmed.", errorContext, errorControl, errorPresenter );
					}
				}
			}
		}

		/// <summary>
		/// Checks if there is a password.
		/// </summary>
		/// <param name="password">Password to check (its content isn't fully validated as a password). This is to be used for
		/// symmetric encryption. This may be a buffer or a text.</param>
		/// <param name="errorContext">The context which may help the user to better understand the error. This may be null or
		/// empty.</param>
		/// <returns>Returns true if there is a password.</returns>
		public static bool HasPassword( object password, string errorContext )
		{
			if( password == null )
				return false;

			if( password is byte[] )
				return ((byte[])password).Length > 0;
			else if( password is string )
				return (string)password != "";
			else
			{
				if( TextUtils.NoContent( errorContext ) )
					errorContext = "Password";

				string message = string.Format( "The password must be either a buffer or a text, not a \"{0}\"."
					, password.GetType().FullName );

				throw new ValidationException( message, null, errorContext );
			}
		}

		#endregion Public static methods
	}
}
