////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Data;
using System.Data.Common;
using System.Data.SqlClient;
using System.IO;
using Vela.DatabaseStorehouse;
using Vela.Kernel;
using Vela.Serialization;
using Vela.SqlDatabase;
using Vela.Storehouse;

namespace Vela.MsSqlFeDatabase
{
	/// <summary>
	/// Provider for an SQL Server FE (= Full Edition) database.
	/// </summary>
	/// <remarks>
	/// <para>Don't share a storehouse provider among multiple threads, even if you synchronize it. Clone it instead.</para>
	/// </remarks>
	public class MsSqlFeDatabaseProvider : SqlDatabaseProvider
	{
		#region Private constants

		/// <summary>
		/// The error codes which represent a timeout / deadlock error.
		/// </summary>
		private static readonly HashSet<int> TimeoutErrorCodes = new HashSet<int> { -2, 1204, 1205, 1222, 8645 };

		#endregion Private constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected MsSqlFeDatabaseProvider()
		{
			Scripts = new MsSqlFeDatabaseScripts();
			ScriptPlaceholders = new SqlScriptPlaceholders();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="access">The database access.</param>
		/// <param name="fileName">The name of the file where the storehouse header is stored.</param>
		/// <param name="headerUserPassword">The user password used to symmetrically encrypt the header.</param>
		/// <param name="header">The storehouse header (which is stored in a symmetrically encrypted form). This may be null if
		/// the storehouse doesn't exist yet.</param>
		/// <param name="serializer">The serializer of the sheafs.</param>
		public MsSqlFeDatabaseProvider( DatabaseAccess access, string fileName, string headerUserPassword
				, StorehouseHeader header, ISerializer serializer )
			: base( access, fileName, headerUserPassword, header, serializer )
		{
			Scripts = new MsSqlFeDatabaseScripts();
			ScriptPlaceholders = new SqlScriptPlaceholders();
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The size of the password used to encrypt the storehouse files which are attached to the storehouse header. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is used to derive the encryption password (see
		/// <see cref="IStorehouseProvider.GetStorehouseAttachmentsPassword"/>).</para>
		/// </remarks>
		public override int StorehouseAttachmentsPasswordSize
		{
			get { return 128; }
		}

		/// <summary>
		/// If this is true then the storehouse is managed by a storehouse server engine, else it's standalone / embedded.
		/// This is true. READ-ONLY.
		/// </summary>
		public override bool UsesServerEngine
		{
			get { return true; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Creates a new database command.
		/// </summary>
		/// <param name="connection">Database connection to use.</param>
		/// <returns>Returns the new database command.</returns>
		private DbCommand CreateCommandWithoutTransaction( DbConnection connection )
		{
			ValidationUtils.ObjectHasContent( connection, "connection" );

			DbCommand dbCommand = new SqlCommand();
			dbCommand.Connection = connection;
			dbCommand.CommandTimeout = AccessTyped.CommandTimeout;
			dbCommand.CommandType = CommandType.Text;

			return dbCommand;
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Checks if the specified exception (or any of its inner exceptions) represents a timeout (which has occurred while
		/// accessing the storehouse).
		/// </summary>
		/// <param name="exception">Exception to check if it represents a timeout. This may be null.</param>
		/// <returns>Returns true if the specified exception (or any of its inner exceptions) represents a timeout.</returns>
		public override bool IsTimeoutException( Exception exception )
		{
			while( exception != null )
			{
				SqlException sqlException = exception as SqlException;
				if( (sqlException != null) && (sqlException.Number != 0) && TimeoutErrorCodes.Contains( sqlException.Number ) )
					return true;

				exception = exception.InnerException;
			}

			return false;
		}

		/// <summary>
		/// Creates new a transaction in the storehouse.
		/// </summary>
		public override IStorehouseTransaction CreateTransaction()
		{
			return new MsSqlFeDatabaseTransaction( this );
		}

		/// <summary>
		/// If the storehouse doesn't already exist, an exception is thrown.
		/// </summary>
		/// <param name="adminConnectionString">The administration connection string with which a storehouse server can create
		/// a storehouse. This may be null or empty for a storehouse provider which doesn't use a storehouse server.</param>
		public override void CreateStorehouse( string adminConnectionString )
		{
			ValidationUtils.TextHasContent( adminConnectionString, "adminConnectionString" );

			// Create the database.
			using( DbConnection connection = new SqlConnection( adminConnectionString ) )
			{
				connection.Open();

				using( DbCommand command = CreateCommandWithoutTransaction( connection ) )
				{
					string commandText = ReplacePlaceholders( ScriptsTyped.CreateServerStorehouse
						, GetStorehouseMainAttachmentFileName(), ServerStorehouseName, GetStorehouseAttachmentsPassword() );

					SplitExecuteNonQuery( commandText, command );
				}
			}

			// Create the storehouse realms.
			CreateStorehouseRealms( GetStandardStorehouseRealms() );

			// Delete the storehouse file only if the database was created.
			File.Delete( FileName );
		}

		/// <summary>
		/// Compacts the storehouse.
		/// </summary>
		/// <param name="adminConnectionString">The administration connection string with which a storehouse server can compact
		/// the storehouse. This may be null or empty for a storehouse provider which doesn't use a storehouse server.</param>
		/// <param name="alsoVerifyAndRepair">If this is true then, before being compacted, the storehouse is verified for
		/// consistency, and is repaired if it's necessary and possible.</param>
		/// <remarks>
		/// <para>The empty space is removed.</para>
		/// 
		/// <para>Compacting this kind of storehouse is not supported. No exception is thrown.</para>
		/// </remarks>
		public override void CompactStorehouse( string adminConnectionString, bool alsoVerifyAndRepair )
		{
			LogManager.Add( "Compacting this kind of storehouse is not supported." );
		}

		/// <summary>
		/// Creates a new database connection.
		/// </summary>
		/// <param name="connectionString">The connection string to use.</param>
		/// <returns>Returns the new database connection.</returns>
		public override DbConnection CreateConnection( string connectionString )
		{
			return new SqlConnection( connectionString );
		}

		/// <summary>
		/// Creates a new database command.
		/// </summary>
		/// <param name="connection">Database connection to use.</param>
		/// <param name="transaction">Database transaction to use.</param>
		/// <returns>Returns the new database command.</returns>
		public override DbCommand CreateCommand( DbConnection connection, DbTransaction transaction )
		{
			ValidationUtils.ObjectHasContent( connection, "connection" );
			ValidationUtils.ObjectHasContent( transaction, "transaction" );

			DbCommand dbCommand = new SqlCommand();
			dbCommand.Connection = connection;
			dbCommand.Transaction = transaction;
			dbCommand.CommandTimeout = AccessTyped.CommandTimeout;
			dbCommand.CommandType = CommandType.Text;

			return dbCommand;
		}

		/// <summary>
		/// Adds a new database command parameter.
		/// </summary>
		/// <param name="dbCommand">Database command to which to add the parameter.</param>
		/// <param name="paramName">The name of the parameter.</param>
		/// <param name="dbType">Database type of the parameter.</param>
		/// <param name="isNullable">If this is true then the parameter is nullable (and its value may be null).</param>
		/// <param name="paramValue">The value of the parameter. This may be null if "isNullable"is true.</param>
		public override void AddCommandParam( DbCommand dbCommand, string paramName, DbType dbType, bool isNullable
			, object paramValue )
		{
			ValidationUtils.ObjectHasContent( dbCommand, "dbCommand" );
			ValidationUtils.ObjectHasContent( paramName, "paramName" );

			if( !isNullable )
				ValidationUtils.ObjectHasContent( paramValue, "paramValue" );

			// Do not pass "dbType" in the constructor because the constructor may expect it to be of a specialized type like
			// "SqlDbType" (not "DbType" as it is incoming); that would cause the wrong overload to be called and the type to
			// be wrongly set.
			DbParameter param = new SqlParameter();
			param.ParameterName = paramName;
			param.DbType = dbType;
			param.IsNullable = isNullable;
			param.SourceColumnNullMapping = isNullable;
			param.Value = (paramValue != null) ? paramValue : DBNull.Value;

			dbCommand.Parameters.Add( param );
		}

		/// <summary>
		/// Adds a new database command parameter to the specified command. The parameter's data type is a huge buffer.
		/// </summary>
		/// <param name="dbCommand">Database command to which to add the parameter.</param>
		/// <param name="paramName">The name of the parameter.</param>
		/// <param name="paramValue">The value of the parameter. This may be null.</param>
		public override void AddBlobCommandParam( DbCommand dbCommand, string paramName, object paramValue )
		{
			ValidationUtils.ObjectHasContent( dbCommand, "dbCommand" );
			ValidationUtils.ObjectHasContent( paramName, "paramName" );

			SqlParameter param = new SqlParameter();
			param.ParameterName = paramName;
			param.SqlDbType = SqlDbType.Image;
			param.IsNullable = true;
			param.SourceColumnNullMapping = true;
			param.Value = (paramValue != null) ? paramValue : DBNull.Value;

			dbCommand.Parameters.Add( param );
		}

		#endregion Public methods
	}
}
