////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Net;
using System.Net.Sockets;
using System.Reflection;
using Vela.Kernel;

namespace Vela.Networking
{
	/// <summary>
	/// A communication frame which a service may process in a single send or receive.
	/// </summary>
	public class CommunicationFrame
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="communicationFrameAlg">Communication frame algorithm.</param>
		/// <param name="communicationTimeoutPerBlob">The timeout (in seconds) used for sending and receiving data through the
		/// network, per blob, per frame. This must be at least 1.</param>
		/// <param name="communicationFrameSizeMax">The maximum size (in bytes) of a communication frame which a service may
		/// process (in a single send or receive). This must be at least 1.</param>
		public CommunicationFrame( ICommunicationFrameAlg communicationFrameAlg, int communicationTimeoutPerBlob
			, int communicationFrameSizeMax )
		{
			CommunicationFrameAlg = communicationFrameAlg;
			CommunicationTimeoutPerBlob = communicationTimeoutPerBlob;
			CommunicationFrameSizeMax = communicationFrameSizeMax;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private ICommunicationFrameAlg _CommunicationFrameAlg;
		/// <summary>
		/// Communication frame algorithm. READ-ONLY.
		/// </summary>
		public ICommunicationFrameAlg CommunicationFrameAlg
		{
			get { return _CommunicationFrameAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "CommunicationFrameAlg" );

				_CommunicationFrameAlg = value;
			}
		}

		private int _CommunicationTimeoutPerBlob;
		/// <summary>
		/// The timeout (in seconds) used for sending and receiving data through the network, per blob, per frame. This must be
		/// at least 1. READ-ONLY.
		/// </summary>
		public int CommunicationTimeoutPerBlob
		{
			get { return _CommunicationTimeoutPerBlob; }
			protected set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "CommunicationTimeoutPerBlob" );

				_CommunicationTimeoutPerBlob = value;
			}
		}

		private int _CommunicationFrameSizeMax;
		/// <summary>
		/// The maximum size (in bytes) of a communication frame which a service may process (in a single send or receive). This
		/// must be at least 1. READ-ONLY.
		/// </summary>
		public int CommunicationFrameSizeMax
		{
			get { return _CommunicationFrameSizeMax; }
			protected set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "CommunicationFrameSizeMax" );

				_CommunicationFrameSizeMax = value;
			}
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Waits to receive the header of the communication frame.
		/// </summary>
		/// <param name="socket">The socket where the header is to be received.</param>
		/// <param name="chrono">Communication timeout chronometer. This is restarted here.</param>
		/// <returns>Returns the size of the body of the communication frame. This is 0 if
		/// <see cref="ICommunicationFrameAlg.HeaderSize"/> is 0 or a negative number.</returns>
		protected int WaitHeader( Socket socket, Chronometer chrono )
		{
			int headerSize = CommunicationFrameAlg.HeaderSize;
			if( headerSize <= 0 )
				return 0;

			ValidationUtils.ObjectHasContent( socket, "socket" );
			ValidationUtils.ObjectHasContent( chrono, "chrono" );

			int timeoutPerData = NetworkingUtils.GetCommunicationTimeoutPerData( CommunicationTimeoutPerBlob );

			socket.ReceiveTimeout = TimeUtils.SecondsToMilli( timeoutPerData );
			socket.SendTimeout = socket.ReceiveTimeout;

			chrono.Start();

			// Wait for the entire header to accumulate.
			int availableSize = 0;
			while( !chrono.PassedSeconds( timeoutPerData ) && (availableSize < headerSize) )
			{
				// If the socket has been disconnected, stop waiting.
				if( socket.Poll( NetworkingConsts.SleepWaitingConnection, SelectMode.SelectRead ) && (socket.Available <= 0) )
					break;

				availableSize = socket.Available;
			}

			if( availableSize < headerSize )
			{
				string message = "Failed to read a response from the network. Ask the remote peer if he / she sees an error in"
					+ " the application's log. Make sure that you and your contacts are using the newest version of the"
					+ " application."
					+ "\n\n"
					+ "The size of the received header is {0} instead of the expected size {1}. Increasing the communication"
					+ " timeout (per blob, per frame) to more than {2} seconds might solve this problem if it occurs regularly"
					+ " (but not always).";
				throw new ValidationException( string.Format( message, availableSize, headerSize
					, CommunicationTimeoutPerBlob ) );
			}

			// Read the header.
			byte[] header = new byte[ headerSize ];
			socket.Receive( header, 0, headerSize, SocketFlags.None );

			int headerOffset = 0;

			#region The size of the body

			int bodySize = BitConverter.ToInt32( header, headerOffset );

			if( bodySize <= 0 )
				throw new ValidationException( "The size of the body of the communication frame must be greater than 0." );

			if( bodySize > CommunicationFrameSizeMax - headerSize )
			{
				throw new ValidationException( string.Format( "The size ({0} bytes) of the body of the communication frame is"
					+ " bigger than the maximum allowed size ({1} bytes).", bodySize, CommunicationFrameSizeMax - headerSize ) );
			}

			headerOffset += sizeof( Int32 );

			#endregion The size of the body

			return bodySize;
		}

		/// <summary>
		/// Waits to receive the body of the communication frame.
		/// </summary>
		/// <param name="socket">The socket where the body is to be received.</param>
		/// <param name="chrono">Communication timeout chronometer. This is NOT restarted here.</param>
		/// <param name="bodySize">The expected size of the communication frame. This must be at least 1.</param>
		/// <returns>Returns the received body.</returns>
		/// <remarks>
		/// <para>If the received data exceeds the size indicated in the header ("bodySize"), it's likely that a new frame is
		/// following the current one, so the rest is ignored (by the current call) and may be read with a new call to this
		/// method.</para>
		/// </remarks>
		protected byte[] WaitBody( Socket socket, Chronometer chrono, int bodySize )
		{
			ValidationUtils.ObjectHasContent( socket, "socket" );
			ValidationUtils.ObjectHasContent( chrono, "chrono" );
			ValidationUtils.Integer( bodySize, 1, int.MaxValue, "bodySize" );

			int timeoutPerData = NetworkingUtils.GetCommunicationTimeoutPerData( CommunicationTimeoutPerBlob, bodySize );

			socket.ReceiveTimeout = TimeUtils.SecondsToMilli( timeoutPerData );
			socket.SendTimeout = socket.ReceiveTimeout;

			// Buffer for the received body.
			byte[] body = new byte[ bodySize ];

			// Read the body as it comes, don't wait for it to accumulate in its entirety (because the system cache may
			// overflow).
			int totalReadSize = 0;
			while( !chrono.PassedSeconds( timeoutPerData ) && (totalReadSize < bodySize) )
			{
				// If the socket has been disconnected, stop waiting.
				if( socket.Poll( NetworkingConsts.SleepWaitingConnection, SelectMode.SelectRead ) && (socket.Available <= 0) )
					break;

				int availableSize = socket.Available;

				if( availableSize > 0 )
				{
					// Read data in order to free the system's receive buffer.

					int sizeToRead = Math.Min( availableSize, bodySize - totalReadSize );

					// Read the available body.
					socket.Receive( body, totalReadSize, sizeToRead, SocketFlags.None );

					totalReadSize += sizeToRead;
				}
			}

			if( totalReadSize != bodySize )
			{
				string message = "Failed to read a response from the network. Ask the remote peer if he / she sees an error in"
					+ " the application's log. Make sure that you and your contacts are using the newest version of the"
					+ " application."
					+ "\n\n"
					+ "The size of the received body is {0} instead of the expected size {1}. Increasing the communication"
					+ " timeout (per blob, per frame) to more than {2} seconds might solve this problem if it occurs regularly"
					+ " (but not always).";
				throw new ValidationException( string.Format( message, totalReadSize, bodySize, CommunicationTimeoutPerBlob ) );
			}

			return body;
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Creates a socket through which to send a message to the specified remote IP address.
		/// </summary>
		/// <param name="remoteIpAddress">The remote IP address of the service where to send the message.</param>
		/// <param name="remotePort">The remote network port of the service where to send the message. This must be at least 1.
		/// </param>
		/// <returns>Returns the socket through which to send a message to the specified remote IP address.</returns>
		public Socket CreateSocketForSendMessage( byte[] remoteIpAddress, UInt16 remotePort )
		{
			ValidationUtils.ArrayHasContent( remoteIpAddress, "remoteIpAddress" );
			ValidationUtils.Integer( remotePort, 1, UInt16.MaxValue, "remotePort" );

			// Create the remote end-point.
			IPAddress remoteEndpointIpAddress = new IPAddress( remoteIpAddress );
			IPEndPoint remoteEndpoint = new IPEndPoint( remoteEndpointIpAddress, remotePort );

			// Create the local socket.
			Socket socket = new Socket( remoteEndpointIpAddress.AddressFamily, KernelConsts.SocketPacketType
				, KernelConsts.SocketProtocolType );
			socket.SetIPProtectionLevel( IPProtectionLevel.Unrestricted ); // Allow NAT traversal.
			socket.SetSocketOption( SocketOptionLevel.IPv6, SocketOptionName.IPv6Only, 0 ); // Allow IPv4 traffic.

			// Don't locally bind the socket because it would, at some point, say that that is already in use, if you try to
			// quickly connect to it mutiple times; the sockets don't get released on demand, they have a delay.

			// Connect to the service.
			socket.Connect( remoteEndpoint );

			return socket;
		}

		/// <summary>
		/// Sends the specified message through the specified socket.
		/// </summary>
		/// <param name="socket">The socket through which to send the specified message.</param>
		/// <param name="message">The message to send.</param>
		public void SendMessage( Socket socket, byte[] message )
		{
			ValidationUtils.ObjectHasContent( socket, "socket" );
			ValidationUtils.ArrayHasContent( message, "message" );

			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				UInt16 localPort, remotePort;
				byte[] localIpAddress = NetworkingUtils.GetLocalIpAddress( socket, out localPort );
				byte[] remoteIpAddress = NetworkingUtils.GetRemoteIpAddress( socket, out remotePort );

				LogManager.DebugAddFormat( "The IP address \"{0}\" on port {1} is sending a message to the IP address"
						+ " \"{2}\" on port {3}, with a message size of {4} bytes."
					, NetworkingUtils.GetTextualIpAddress( remoteIpAddress ), localPort
					, NetworkingUtils.GetTextualIpAddress( remoteIpAddress ), remotePort
					, message.Length );

				byte[] body = CommunicationFrameAlg.EncodeBody( message );
				byte[] header = CommunicationFrameAlg.SetupHeader( body );

				// Send the header.
				int timeoutPerHeader = NetworkingUtils.GetCommunicationTimeoutPerData( CommunicationTimeoutPerBlob );
				socket.ReceiveTimeout = TimeUtils.SecondsToMilli( timeoutPerHeader );
				socket.SendTimeout = socket.ReceiveTimeout;
				socket.Send( header );

				// Send the body.
				int timeoutPerBody = NetworkingUtils.GetCommunicationTimeoutPerData( CommunicationTimeoutPerBlob, body.Length );
				socket.ReceiveTimeout = TimeUtils.SecondsToMilli( timeoutPerBody );
				socket.SendTimeout = socket.ReceiveTimeout;
				socket.Send( body );

				LogManager.DebugAddFormat( "The IP address \"{0}\" on port {1} has sent a message to the IP address"
						+ " \"{2}\" on port {3}, with a message size of {4} bytes."
					, NetworkingUtils.GetTextualIpAddress( localIpAddress ), localPort
					, NetworkingUtils.GetTextualIpAddress( remoteIpAddress ), remotePort
					, message.Length );
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		/// <summary>
		/// Receives a message.
		/// </summary>
		/// <param name="socket">The socket where the message is to be received.</param>
		/// <returns>Returns the received message.</returns>
		/// <remarks>
		/// <para>If the received data exceeds the size indicated in the header, it's likely that a new frame is following the
		/// current one, so the rest is ignored (by the current call) and may be read with a new call to this method.</para>
		/// </remarks>
		public byte[] ReceiveMessage( Socket socket )
		{
			ValidationUtils.ObjectHasContent( socket, "socket" );

			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				UInt16 localPort, remotePort;
				byte[] localIpAddress = NetworkingUtils.GetLocalIpAddress( socket, out localPort );
				byte[] remoteIpAddress = NetworkingUtils.GetRemoteIpAddress( socket, out remotePort );

				LogManager.DebugAddFormat( "The IP address \"{0}\" on port {1} is waiting for a message from the IP address"
						+ " \"{2}\" on port {3}."
					, NetworkingUtils.GetTextualIpAddress( localIpAddress ), localPort
					, NetworkingUtils.GetTextualIpAddress( remoteIpAddress ), remotePort );

				// Timeout chronometer.
				Chronometer chrono = new Chronometer();

				// Wait for the header.
				int bodySize = WaitHeader( socket, chrono );

				// Wait for the body.
				byte[] body = WaitBody( socket, chrono, bodySize );

				byte[] message = CommunicationFrameAlg.DecodeBody( body );

				LogManager.DebugAddFormat( "The IP address \"{0}\" on port {1} has received a message from the IP address"
						+ " \"{2}\" on port {3}, with a message size of {4} bytes."
					, NetworkingUtils.GetTextualIpAddress( localIpAddress ), localPort
					, NetworkingUtils.GetTextualIpAddress( remoteIpAddress ), remotePort
					, message.Length );

				return message;
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		/// <summary>
		/// Sends a request to the specified network address of a service, and receives a response.
		/// </summary>
		/// <param name="remoteIpAddress">The remote IP address of the service where to send the request.</param>
		/// <param name="remotePort">The remote network port of the service where to send the request. This must be at least 1.
		/// </param>
		/// <param name="request">The request to send.</param>
		/// <returns>Returns the received response.</returns>
		public byte[] SendRequestAndReceiveResponse( byte[] remoteIpAddress, UInt16 remotePort, byte[] request )
		{
			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			Socket socket = null;

			try
			{
				socket = CreateSocketForSendMessage( remoteIpAddress, remotePort );

				// Send the request to the service.
				SendMessage( socket, request );

				// Receive the response from the service.
				return ReceiveMessage( socket );
			}
			finally
			{
				try
				{
					NetworkingUtils.CloseSocket( socket );
				}
				finally
				{
					LogManager.DebugEndMethod( logManager );
				}
			}
		}

		#endregion Public methods
	}
}
