////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Net;
using System.Net.NetworkInformation;
using System.Net.PeerToPeer;
using System.Net.Sockets;
using Vela.Kernel;

namespace Vela.Networking
{
	/// <summary>
	/// Networking utilities.
	/// </summary>
	public static class NetworkingUtils
	{
		#region Public static methods

		/// <summary>
		/// Gets the communication timeout (in seconds) for sending and receiving data through the network, for the entire
		/// specified data size, from the specified timeout (in seconds) per blob, per frame.
		/// </summary>
		/// <param name="communicationTimeoutPerBlob">The timeout (in seconds) used for sending and receiving data through the
		/// network, per blob, per frame. This must be at least 1.</param>
		/// <param name="dataSize">The size (in bytes) of the data to send or receive. This must be at least 1.</param>
		/// <returns>Returns the communication timeout (in seconds).</returns>
		public static int GetCommunicationTimeoutPerData( int communicationTimeoutPerBlob, Int64 dataSize = 1 )
		{
			ValidationUtils.Integer( communicationTimeoutPerBlob, 1, int.MaxValue, "communicationTimeoutPerBlob" );
			ValidationUtils.Integer( dataSize, 1, int.MaxValue, "dataSize" );

			int blobCount = (int)MathUtils.RoundUp( dataSize, (Int64)NetworkingConsts.CommunicationTimeoutBlobSize );

			return blobCount * communicationTimeoutPerBlob;
		}

		/// <summary>
		/// Gets the binary representation of the specified textual IP address.
		/// </summary>
		/// <param name="ipAddress">The textual IP address whose binary representation to get. This may be null or empty.</param>
		/// <returns>Returns the binary representation of the specified IP address. This may be null or empty.</returns>
		public static byte[] GetBinaryIpAddressBuffer( string ipAddress )
		{
			if( TextUtils.NoContent( ipAddress ) )
				return null;

			return IPAddress.Parse( ipAddress ).GetAddressBytes();
		}

		/// <summary>
		/// Gets the textual representation of the specified binary IP address.
		/// </summary>
		/// <param name="ipAddress">The binary IP address whose textual representation to get. This may be null or empty.</param>
		/// <returns>Returns the textual representation of the specified IP address. This may be null or empty.</returns>
		public static string GetTextualIpAddress( byte[] ipAddress )
		{
			if( ArrayUtils.NoContent( ipAddress ) )
				return null;

			return new IPAddress( ipAddress ).ToString();
		}

		/// <summary>
		/// Gets the local port of the specified socket.
		/// </summary>
		/// <param name="socket">Socket from which to get the local port.</param>
		/// <returns>Returns the local port of the socket.</returns>
		public static UInt16 GetLocalPort( Socket socket )
		{
			ValidationUtils.ObjectHasContent( socket, "socket" );

			IPEndPoint localEndpoint = socket.LocalEndPoint as IPEndPoint;
			ValidationUtils.ObjectHasContent( localEndpoint, "localEndpoint" );

			return (UInt16)localEndpoint.Port;
		}

		/// <summary>
		/// Gets the IP address of the local endpoint of the specified socket.
		/// </summary>
		/// <param name="socket">Socket from which to get the IP address of the local endpoint.</param>
		/// <param name="localPort">Is set with the port of the local endpoint.</param>
		/// <returns>Returns the IP address of the local endpoint.</returns>
		public static byte[] GetLocalIpAddress( Socket socket, out UInt16 localPort )
		{
			ValidationUtils.ObjectHasContent( socket, "socket" );

			IPEndPoint localEndpoint = socket.LocalEndPoint as IPEndPoint;
			ValidationUtils.ObjectHasContent( localEndpoint, "localEndpoint" );

			byte[] localIpAddress = localEndpoint.Address.GetAddressBytes();
			ValidationUtils.ArrayHasContent( localIpAddress, "localIpAddress" );

			localPort = (UInt16)localEndpoint.Port;

			return localIpAddress;
		}

		/// <summary>
		/// Gets the remote port of the specified socket.
		/// </summary>
		/// <param name="socket">Socket from which to get the remote port.</param>
		/// <returns>Returns the remote port of the socket.</returns>
		public static UInt16 GetRemotePort( Socket socket )
		{
			ValidationUtils.ObjectHasContent( socket, "socket" );

			IPEndPoint remoteEndpoint = socket.RemoteEndPoint as IPEndPoint;
			ValidationUtils.ObjectHasContent( remoteEndpoint, "remoteEndpoint" );

			return (UInt16)remoteEndpoint.Port;
		}

		/// <summary>
		/// Gets the IP address of the remote endpoint of the specified socket.
		/// </summary>
		/// <param name="socket">Socket from which to get the IP address of the remote endpoint.</param>
		/// <param name="remotePort">Is set with the port of the remote endpoint.</param>
		/// <returns>Returns the IP address of the remote endpoint.</returns>
		public static byte[] GetRemoteIpAddress( Socket socket, out UInt16 remotePort )
		{
			ValidationUtils.ObjectHasContent( socket, "socket" );

			IPEndPoint remoteEndpoint = socket.RemoteEndPoint as IPEndPoint;
			ValidationUtils.ObjectHasContent( remoteEndpoint, "remoteEndpoint" );

			byte[] remoteIpAddress = remoteEndpoint.Address.GetAddressBytes();
			ValidationUtils.ArrayHasContent( remoteIpAddress, "remoteIpAddress" );

			remotePort = (UInt16)remoteEndpoint.Port;

			return remoteIpAddress;
		}

		/// <summary>
		/// Gets the local fully qualified host name.
		/// </summary>
		/// <returns>Returns the local fully qualified host name. This may be null or empty.</returns>
		public static string GetLocalFullHostName()
		{
			IPGlobalProperties ipGlobalProperties = IPGlobalProperties.GetIPGlobalProperties();

			return TextUtils.NoContent( ipGlobalProperties.DomainName ) ? ipGlobalProperties.HostName
				: string.Format( "{0}.{1}", ipGlobalProperties.HostName, ipGlobalProperties.DomainName );
		}

		/// <summary>
		/// Gets the IPv4 address which is assigned to the current computer and may be used for port forwarding.
		/// </summary>
		/// <returns>Returns the IPv4 address which is assigned to the current computer and may be used for port forwarding.
		/// </returns>
		public static string GetLocalIpAddressForPortForwarding()
		{
			IPAddress[] localIps = Dns.GetHostAddresses( GetLocalFullHostName() );
			ValidationUtils.ArrayHasContent( localIps, "localIps" );

			foreach( IPAddress localIp in localIps )
			{
				if( localIp.AddressFamily == AddressFamily.InterNetwork )
				{
					byte[] ipBuffer = localIp.GetAddressBytes();

					if( (ipBuffer[ 0 ] == 192) && (ipBuffer[ 1 ] == 168) )
						return localIp.ToString();
				}
			}

			foreach( IPAddress localIp in localIps )
			{
				if( localIp.AddressFamily == AddressFamily.InterNetwork )
				{
					byte[] ipBuffer = localIp.GetAddressBytes();

					if( (ipBuffer[ 0 ] == 172) && (ipBuffer[ 1 ] >= 16) && (ipBuffer[ 1 ] <= 31) )
						return localIp.ToString();
				}
			}

			foreach( IPAddress localIp in localIps )
			{
				if( localIp.AddressFamily == AddressFamily.InterNetwork )
				{
					byte[] ipBuffer = localIp.GetAddressBytes();

					if( ipBuffer[ 0 ] == 10 )
						return localIp.ToString();
				}
			}

			throw new ValidationException( "No local IPv4 address was found." );
		}

		/// <summary>
		/// Resolves the specified URL to a list of IP addresses.
		/// </summary>
		/// <param name="url">URL to resolve.</param>
		/// <returns>Returns the found IP addresses, or null if none is found.</returns>
		public static IPAddress[] ResolveUrl( string url )
		{
			ValidationUtils.TextHasContent( url, "url" );

			IPHostEntry hostEntry = Dns.GetHostEntry( url );
			if( (hostEntry == null) || ArrayUtils.NoContent( hostEntry.AddressList ) )
				return null;

			return hostEntry.AddressList;
		}

		/// <summary>
		/// Sorts the IP addresses from the specified peer records by their priority, first the global ones, then the local ones.
		/// </summary>
		/// <param name="ipAddresses">Peer records whose IP addresses to sort. This may be empty, but not null.</param>
		/// <returns>Returns the sorted IP addresses.</returns>
		/// <remarks>
		/// <para>IPv6 addresses have priority over the IPv4 ones.</para>
		/// </remarks>
		public static List<IPAddress> GetPeerIpAddressesByPriority( IPAddress[] ipAddresses )
		{
			ValidationUtils.ObjectHasContent( ipAddresses, "ipAddresses" );

			List<IPAddress> ipAddressesResult = new List<IPAddress>();

			// Prioritize global IPv6 addresses.
			foreach( IPAddress ipAddress in ipAddresses )
			{
				if( ipAddress.AddressFamily != AddressFamily.InterNetworkV6 )
					continue;

				if( ipAddress.IsIPv6LinkLocal || ipAddress.IsIPv6SiteLocal || ipAddress.IsIPv6Multicast )
					continue;

				ipAddressesResult.Add( ipAddress );
			}

			// Get the local IPv6 addresses.
			foreach( IPAddress ipAddress in ipAddresses )
			{
				if( ipAddress.AddressFamily != AddressFamily.InterNetworkV6 )
					continue;

				if( ipAddress.IsIPv6LinkLocal || ipAddress.IsIPv6SiteLocal )
					ipAddressesResult.Add( ipAddress );
			}

			// Get the IPv4 addresses.
			foreach( IPAddress ipAddress in ipAddresses )
			{
				if( ipAddress.AddressFamily != AddressFamily.InterNetwork )
					continue;

				if( (ipAddress == IPAddress.Any) || (ipAddress == IPAddress.None) || (ipAddress == IPAddress.Broadcast) )
					continue;

				ipAddressesResult.Add( ipAddress );
			}

			return ipAddressesResult;
		}

		/// <summary>
		/// Sorts the IP addresses from the specified peer records by their priority, first the global ones, then the local ones.
		/// </summary>
		/// <param name="peerRecords">Peer records whose IP addresses to sort. This may be empty, but not null.</param>
		/// <returns>Returns the sorted IP addresses.</returns>
		/// <remarks>
		/// <para>IPv6 addresses have priority over the IPv4 ones. This is because IPv6 network support is available since the
		/// PNRP has worked (= returned the specified records).</para>
		/// </remarks>
		public static List<IPAddress> GetPeerIpAddressesByPriority( PeerNameRecordCollection peerRecords )
		{
			ValidationUtils.ObjectHasContent( peerRecords, "peerRecords" );

			List<IPAddress> ipAddresses = new List<IPAddress>();

			// Prioritize global IPv6 addresses.
			foreach( PeerNameRecord peerRecord in peerRecords )
			{
				foreach( IPEndPoint ipe in peerRecord.EndPointCollection )
				{
					if( ipe.AddressFamily != AddressFamily.InterNetworkV6 )
						continue;

					IPAddress ipAddress = ipe.Address;
					if( ipAddress.IsIPv6LinkLocal || ipAddress.IsIPv6SiteLocal || ipAddress.IsIPv6Multicast )
						continue;

					ipAddresses.Add( ipAddress );
				}
			}

			// Get the local IPv6 addresses.
			foreach( PeerNameRecord peerRecord in peerRecords )
			{
				foreach( IPEndPoint ipe in peerRecord.EndPointCollection )
				{
					if( ipe.AddressFamily != AddressFamily.InterNetworkV6 )
						continue;

					IPAddress ipAddress = ipe.Address;
					if( ipAddress.IsIPv6LinkLocal || ipAddress.IsIPv6SiteLocal )
						ipAddresses.Add( ipAddress );
				}
			}

			// Get the IPv4 addresses.
			foreach( PeerNameRecord peerRecord in peerRecords )
			{
				foreach( IPEndPoint ipe in peerRecord.EndPointCollection )
				{
					if( ipe.AddressFamily != AddressFamily.InterNetwork )
						continue;

					ipAddresses.Add( ipe.Address );
				}
			}

			return ipAddresses;
		}

		/// <summary>
		/// Closes the specified socket.
		/// </summary>
		/// <param name="socket">The socket to close. This may be null.</param>
		public static void CloseSocket( Socket socket )
		{
			if( socket == null )
				return;

			try
			{
				// Since the communication protocol is message based (= frame size + frame), and since the entire message is read
				// by the protocol, there is no need to call "socket.Shutdown". Besides, this actually causes problems / delays
				// when trying to quickly reuse the same port.

				socket.Close();
			}
			catch( SocketException ex )
			{
				LogManager.DebugAdd( ex );
			}
		}

		/// <summary>
		/// Waits to receive a response on the specified socket.
		/// </summary>
		/// <param name="socket">The socket where the response is to be received.</param>
		/// <param name="receiveTimeout">The timeout (in seconds) for receiving the response.</param>
		/// <param name="responseSizeMax">The maximum size (in bytes) of the response.</param>
		/// <param name="stopIfZero">If this is true then if a 0 byte is received, the method returns.</param>
		/// <returns>Returns the received response.</returns>
		public static byte[] WaitResponse( Socket socket, double receiveTimeout, int responseSizeMax, bool stopIfZero )
		{
			ValidationUtils.ObjectHasContent( socket, "socket" );

			socket.ReceiveTimeout = TimeUtils.SecondsToMilli( receiveTimeout );

			Chronometer chrono = new Chronometer();
			chrono.Start();

			// Buffer for the received response.
			byte[] response = new byte[ responseSizeMax ];

			// Read the response as it comes, don't wait for it to accumulate in its entirety (because it may overflow).
			int readSize = 0;
			while( !chrono.PassedSeconds( receiveTimeout ) && (readSize < responseSizeMax) )
			{
				// If the socket has been disconnected, stop waiting.
				if( socket.Poll( NetworkingConsts.SleepWaitingConnection, SelectMode.SelectRead ) && (socket.Available <= 0) )
					break;

				int availableSize = socket.Available;

				if( availableSize > 0 )
				{
					// Read data in order to free the system's receive buffer.

					if( readSize + availableSize > responseSizeMax )
					{
						throw new ValidationException( string.Format( "The size of the received response is bigger than"
							+ " the expected size {0}.", responseSizeMax ) );
					}

					// Read the available response.
					socket.Receive( response, readSize, availableSize, SocketFlags.None );

					readSize += availableSize;

					// Scan for a 0 byte (which ends the response).
					if( stopIfZero && (Array.IndexOf<byte>( response, 0 ) >= 0) )
						break;
				}
			}

			if( readSize <= 0 )
				throw new ValidationException( "A response was not received in time." );

			return response;
		}

		#endregion Public static methods
	}
}
