////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Net;
using System.Net.Sockets;
using System.Reflection;
using System.Xml;
using Vela.Kernel;
using Vela.Networking.Properties;

namespace Vela.Networking
{
	/// <summary>
	/// Port forwarder.
	/// </summary>
	public class PortForwarder
	{
		#region Private constants

		/// <summary>
		/// The IP address where to make a request to discover the port forwarding service.
		/// </summary>
		private const string ServiceIpAddress = "239.255.255.250";

		/// <summary>
		/// The port where to make a request to discover the port forwarding service.
		/// </summary>
		private const UInt16 ServicePort = 1900;

		/// <summary>
		/// The timeout (in seconds) for receiving a response from the port forwarding service.
		/// </summary>
		/// <remarks>
		/// <para>The specifications say that this must be within [1...5].</para>
		/// </remarks>
		private const int ReceiveTimeout = 5;

		/// <summary>
		/// The maximum size (in bytes) of a response from the port forwarding service.
		/// </summary>
		private const int ResponseSizeMax = 2000;

		/// <summary>
		/// The error message for when the port forwarding service's response contains invalid data. Placeholders:
		/// 0 - description of missing information, 1 - URL to which a request was sent and from which a response was received,
		/// 2 - service response.
		/// </summary>
		private const string ServiceResponseDoesntContain = "The response of the port forwarding service doesn't contain"
			+ " \"{0}\". Try to disable and reenable the UPnP in your router. Connection URL \"{1}\". Response: \"{2}\"";

		/// <summary>
		/// The UPnP root device field from a port forwarding service response.
		/// </summary>
		private const string UpnpRootDevice = "::upnp:rootdevice";

		/// <summary>
		/// The Internet gateway device field from a port forwarding service response.
		/// </summary>
		private const string InternetGatewayDevice = "InternetGatewayDevice";

		/// <summary>
		/// The X-path used to select the device type node from a port forwarding service response.
		/// </summary>
		private const string DeviceTypeNodeXPath = "//tns:device/tns:deviceType/text()";

		/// <summary>
		/// The X-path used to select the control URL node from a port forwarding service response.
		/// </summary>
		private const string ControlUrlNodeNodeXPath = "//tns:service[tns:serviceType=\"urn:schemas-upnp-org:service:WANIPConnection:1\"]/tns:controlURL/text()";

		/// <summary>
		/// The X-path used to select the external IP address node from a port forwarding service response.
		/// </summary>
		private const string ExternalIpAddressNodeXPath = "//NewExternalIPAddress/text()";

		/// <summary>
		/// The template for the message to show to the user when no port forwarding service was discovered. Placeholders:
		/// 0 - IP address, 1 - remote port.
		/// </summary>
		private const string NoServiceMessageTemplate = "A port forwarding service was not found. Make sure that the firewall"
			+ " allows outbound traffic for the application, to the IP address {0}, to the remote port {1}, using the UDP"
			+ " protocol."
			+ "\n\n"
			+ " Also ensure that the UPnP protocol is enabled in your router, and that the application's folder is"
			+ " inside your computer's \"Program Files\" folder."
			+ "\n\n"
			+ "If you are sure that your computer is not connected to a router, or that your contacts do not use your"
			+ " alternative peer address to contact you (through your IPv4 address), you can uncheck the"
			+ " \"Forward ports\" check-box from the configuration of your storehouse(s).";

		#endregion Private constants

		#region Public constants

		/// <summary>
		/// The message to show to the user when no port forwarding service was discovered.
		/// </summary>
		public static readonly string NoServiceMessage = string.Format( NoServiceMessageTemplate, ServiceIpAddress
			, ServicePort );

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public PortForwarder()
		{
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// Text encoding for the messages sent to the port forwarding service.
		/// </summary>
		private TextEncoding TextEncoding = KernelUtils.GetDeepTextEncoding();

		/// <summary>
		/// The URL of the port forwarding service.
		/// </summary>
		private string ServiceUrl;

		#endregion Private fields

		#region Private static methods

		/// <summary>
		/// Gets the root descriptor URL from the specified message.
		/// </summary>
		/// <param name="message">Message from which to get the root descriptor URL. This may be null or empty.</param>
		/// <returns>Returns the root descriptor URL. This may be null or empty.</returns>
		private static string GetRootDescriptorUrl( string message )
		{
			if( TextUtils.NoContent( message ) )
				return null;

			const string prefix = "location:";

			int startPos = message.IndexOf( prefix, StringComparison.OrdinalIgnoreCase );
			if( startPos < 0 )
				return null;

			startPos += prefix.Length;

			int endPos = message.IndexOf( "\r", startPos, StringComparison.OrdinalIgnoreCase );
			if( endPos < 0 )
				endPos = message.Length;

			return message.Substring( startPos, endPos - startPos ).Trim();
		}

		/// <summary>
		/// Creates the URL of the port forwarding service.
		/// </summary>
		/// <param name="rootDescriptorUrl">Root descriptor URL. This may be null or empty.</param>
		/// <param name="controlUrl">Service control URL. This may be null or empty.</param>
		/// <returns>Returns the created URL. This may be null or empty.</returns>
		private static string CreateServiceUrl( string rootDescriptorUrl, string controlUrl )
		{
			if( TextUtils.NoContent( rootDescriptorUrl ) )
				return null;
			if( TextUtils.NoContent( controlUrl ) )
				return null;

			Uri uri = new Uri( rootDescriptorUrl );

			return uri.AbsoluteUri + controlUrl;
		}

		#endregion Private static methods

		#region Public static properties

		/// <summary>
		/// If this is true then the port forwarding service was discovered.
		/// </summary>
		public bool HasDiscoveredService
		{
			get { return TextUtils.HasContent( ServiceUrl ); }
		}

		#endregion Public static properties

		#region Private methods

		/// <summary>
		/// Creates an exception which indicates that a response received from the port forwarding service is missing some
		/// information.
		/// </summary>
		/// <param name="missingInformation">Description of missing information. This may be null or empty.</param>
		/// <param name="connectionUrl">URL to which a request was sent and from which a response was received. This may be
		/// null or empty.</param>
		/// <param name="serviceResponse">Service response. This may be null or empty.</param>
		/// <returns></returns>
		private Exception GetServiceResponseDoesntContainException( string missingInformation, string connectionUrl
			, string serviceResponse )
		{
			string message = string.Format( ServiceResponseDoesntContain, missingInformation, connectionUrl, serviceResponse );

			return new ValidationException( message );
		}

		/// <summary>
		/// Checks if the current class instance was initialized.
		/// </summary>
		private void CheckInitialized()
		{
			if( !HasDiscoveredService )
			{
				throw new Exception( "Either no port forwarding service is available, or \"DiscoverService\" has not been"
					+ " called." );
			}
		}

		/// <summary>
		/// Discovers the descriptor URL of the router to which the current computer is connected.
		/// </summary>
		/// <returns>Returns the descriptor URL of the router.</returns>
		private string DiscoverRouterDescriptorUrl()
		{
			Socket socket = new Socket( AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp );
			socket.SendTimeout = TimeUtils.SecondsToMilli( ReceiveTimeout );

			string localIpAddress = NetworkingUtils.GetLocalIpAddressForPortForwarding();
			LogManager.DebugAddFormat( "Attempting to discover a router from the IP address {0}.", localIpAddress );

			IPAddress ipAddress = IPAddress.Parse( localIpAddress );
			socket.Bind( new IPEndPoint( ipAddress, 0 ) );

			string requestMessage = string.Format( Resources.DiscoverService, ServiceIpAddress, ServicePort, ReceiveTimeout );
			LogManager.DebugAddFormat( "Request \"{0}\":\n\n{1}", MethodBase.GetCurrentMethod().Name, requestMessage );

			byte[] requestBuffer = TextEncoding.TextToArray( requestMessage, false );
			IPEndPoint remoteEndpoint = new IPEndPoint( IPAddress.Parse( ServiceIpAddress ), ServicePort );

			socket.SendTo( requestBuffer, remoteEndpoint );

			byte[] responseBuffer = NetworkingUtils.WaitResponse( socket, ReceiveTimeout, ResponseSizeMax, true );

			string responseMessage = TextEncoding.ArrayToText( responseBuffer );
			LogManager.DebugAddFormat( "Response \"{0}\":\n\n{1}", MethodBase.GetCurrentMethod().Name, responseMessage );

			if( !TextUtils.Contains( responseMessage, UpnpRootDevice ) )
				throw GetServiceResponseDoesntContainException( UpnpRootDevice, localIpAddress, responseMessage );

			string rootDescriptorUrl = GetRootDescriptorUrl( responseMessage );
			if( TextUtils.NoContent( rootDescriptorUrl ) )
				throw GetServiceResponseDoesntContainException( "root descriptor URL", localIpAddress, responseMessage );

			return rootDescriptorUrl;
		}

		/// <summary>
		/// Makes a request to the port forwarding service.
		/// </summary>
		/// <param name="url">Service URL where to make the request.</param>
		/// <param name="message">Message to send to the service.</param>
		/// <param name="action">Action requested from the service to perform.</param>
		/// <returns>Returns the service's response.</returns>
		private XmlDocument MakeRequest( string url, string message, string action )
		{
			ValidationUtils.TextHasContent( url, "url" );
			ValidationUtils.TextHasContent( message, "message" );
			ValidationUtils.TextHasContent( action, "action" );

			string requestMessage = string.Format( Resources.MakeRequest, message );
			byte[] requestBuffer = TextEncoding.TextToArray( requestMessage, false );

			LogManager.DebugAddFormat( "Request \"{0}\":\n\n{1}", MethodBase.GetCurrentMethod().Name, requestMessage );

			WebRequest requestWeb = WebRequest.Create( url );
			requestWeb.Method = "POST";
			requestWeb.ContentType = "text/xml; charset=\"utf-8\"";
			requestWeb.ContentLength = requestBuffer.Length;
			requestWeb.Headers.Add( "SOAPACTION"
				, string.Format( "\"urn:schemas-upnp-org:service:WANIPConnection:1#{0}\"", action ) );

			requestWeb.GetRequestStream().Write( requestBuffer, 0, requestBuffer.Length );

			HttpWebResponse responseWeb = requestWeb.GetResponse() as HttpWebResponse;
			if( (responseWeb.ContentLength <= 0) || (responseWeb.StatusCode != HttpStatusCode.OK) )
			{
				throw new ValidationException( string.Format( "The request to the URL \"{0}\" has returned the state code {1}."
					, url, responseWeb.StatusCode ) );
			}

			XmlDocument responseXml = new XmlDocument();
			responseXml.Load( responseWeb.GetResponseStream() );

			LogManager.DebugAddFormat( "Response \"{0}\":\n\n{1}", MethodBase.GetCurrentMethod().Name, responseXml.OuterXml );

			return responseXml;
		}

		#endregion Private methods

		#region Public static methods

		/// <summary>
		/// Checks if the specified port may be forwarded.
		/// </summary>
		/// <param name="port">Port to check if it may be forwarded.</param>
		/// <returns>Returns true if the specified port may be forwarded.</returns>
		public static bool MayForward( UInt16 port )
		{
			return port >= 1024;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Makes a request to discover the port forwarding service.
		/// </summary>
		/// <remarks>
		/// <para><see cref="ServiceUrl"/> is set in here.</para>
		/// </remarks>
		public void DiscoverService()
		{
			ServiceUrl = null;

			// Discover the router.
			string rootDescriptorUrl = DiscoverRouterDescriptorUrl();

			#region Request the service URL

			LogManager.DebugAddFormat( "Request \"{0}\": {1}", MethodBase.GetCurrentMethod().Name, rootDescriptorUrl );

			HttpWebResponse responseWeb = WebRequest.Create( rootDescriptorUrl ).GetResponse() as HttpWebResponse;
			if( (responseWeb.ContentLength <= 0) || (responseWeb.StatusCode != HttpStatusCode.OK) )
			{
				throw new ValidationException( string.Format( "The request to the URL \"{0}\" has returned the state code {1}."
					, rootDescriptorUrl, responseWeb.StatusCode ) );
			}

			XmlDocument responseXml = new XmlDocument();
			responseXml.Load( responseWeb.GetResponseStream() );

			LogManager.DebugAddFormat( "Response \"{0}\":\n\n{1}", MethodBase.GetCurrentMethod().Name, responseXml.OuterXml );

			XmlNamespaceManager nsManager = new XmlNamespaceManager( responseXml.NameTable );
			nsManager.AddNamespace( "tns", "urn:schemas-upnp-org:device-1-0" );

			XmlNode deviceTypeNode = responseXml.SelectSingleNode( DeviceTypeNodeXPath, nsManager );
			string deviceTypeNodeText = (deviceTypeNode != null) ? deviceTypeNode.Value : null;
			if( !TextUtils.Contains( deviceTypeNodeText, InternetGatewayDevice ) )
				throw GetServiceResponseDoesntContainException( InternetGatewayDevice, rootDescriptorUrl, responseXml.OuterXml );

			XmlNode controlUrlNode = responseXml.SelectSingleNode( ControlUrlNodeNodeXPath, nsManager );
			string controlUrlNodeText = (controlUrlNode != null) ? controlUrlNode.Value : null;
			if( TextUtils.NoContent( controlUrlNodeText ) )
				throw GetServiceResponseDoesntContainException( "control URL", rootDescriptorUrl, responseXml.OuterXml );

			#endregion Request the service URL

			ServiceUrl = CreateServiceUrl( rootDescriptorUrl, controlUrlNode.Value );
		}

		/// <summary>
		/// Gets the external IP address of the current computer, as reported by the port forwarding service.
		/// </summary>
		/// <returns>Returns the external IP address of the current computer, as reported by the service.</returns>
		public IPAddress GetExternalIpAddress()
		{
			CheckInitialized();

			string message = Resources.GetExternalIpAddress;

			XmlDocument responseXml = MakeRequest( ServiceUrl, message, "GetExternalIPAddress" );

			XmlNamespaceManager nsManager = new XmlNamespaceManager( responseXml.NameTable );
			nsManager.AddNamespace( "tns", "urn:schemas-upnp-org:device-1-0" );
			string ipAddressText = responseXml.SelectSingleNode( ExternalIpAddressNodeXPath, nsManager ).Value;

			if( TextUtils.NoContent( ipAddressText ) )
				throw GetServiceResponseDoesntContainException( "external IP address", ServiceUrl, responseXml.OuterXml );

			return IPAddress.Parse( ipAddressText );
		}

		/// <summary>
		/// Adds a forwarding rule for the specified port, in the port forwarding service.
		/// </summary>
		/// <param name="description">The description of the port forwarding rule. This may be null or empty.</param>
		/// <param name="port">Port to forward.</param>
		/// <param name="protocol">The network protocol for which to forward the port.</param>
		public void AddRule( string description, UInt16 port, ProtocolType protocol = ProtocolType.Tcp )
		{
			CheckInitialized();

			ValidationUtils.Integer( port, 1, UInt16.MaxValue, "port" );

			string message = string.Format( Resources.AddRule, port, protocol.ToString()
				, NetworkingUtils.GetLocalIpAddressForPortForwarding(), description );

			MakeRequest( ServiceUrl, message, "AddPortMapping" );
		}

		/// <summary>
		/// Removes the forwarding rule of the specified port from the port forwarding service.
		/// </summary>
		/// <param name="port">Port whose forwarding rule to remove.</param>
		/// <param name="protocol">The network protocol for which to remove the forwarding rule of the port.</param>
		public void RemoveRule( UInt16 port, ProtocolType protocol = ProtocolType.Tcp )
		{
			if( TextUtils.NoContent( ServiceUrl ) )
				return;

			ValidationUtils.Integer( port, 1, UInt16.MaxValue, "port" );

			string message = string.Format( Resources.RemoveRule, port, protocol.ToString() );

			MakeRequest( ServiceUrl, message, "DeletePortMapping" );
		}

		#endregion Public methods
	}
}
