////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.ActionProcessor;
using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.PaymentDocuments
{
	/// <summary>
	/// An action request make a payment.
	/// </summary>
	/// <remarks>
	/// <para>The payer must sign the action request.</para>
	/// </remarks>
	public class MakePaymentActionRequest : ActionRequestWithBeneficiary
	{
		#region Public constants

		/// <summary>
		/// The minimum amount of money which can be payed.
		/// </summary>
		public const double AmountToPayMin = 1e-4;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected MakePaymentActionRequest()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpn">The UPN of the recipient of the current action request.</param>
		/// <param name="actionRequestValidity">The validity timespan (in minutes) of the current action request.
		/// <see cref="ActionRequest.DocumentExpirationTime"/> is set with <see cref="Sheaf.CreationTime"/> +
		/// "actionRequestValidity". If this is 0 or a negative number, the action request is valid indefinitely.</param>
		/// <param name="responseEncryptionKeySize">The size (in bytes) that the response's symmetric encryption key must
		/// have. If this is 0 or a negative number, the default <see cref="CoreDocumentsConsts.SymmEncryptionKeySizeDef"/> is
		/// used.</param>
		/// <param name="beneficiaryUpn">The UPN of the beneficiary of the action request.</param>
		/// <remarks>
		/// <para>The symmetric encryption key of the response is automatically generated in here.</para>
		/// </remarks>
		public MakePaymentActionRequest( Locator instanceLocator, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg
				, IHashSerializer signatureHashSerializer, Locator initiatorUpnFingerprint, Upn recipientUpn
				, int actionRequestValidity, int responseEncryptionKeySize, Upn beneficiaryUpn )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer, initiatorUpnFingerprint
				, recipientUpn, actionRequestValidity, responseEncryptionKeySize, beneficiaryUpn )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The type of the action response associated to the current action request. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>An action service must return an action response derived either from this type or from
		/// <see cref="ErrorActionResponse"/>.</para>
		/// 
		/// <para>This type must have a constructor with the following parameters: the instance locator of the action response,
		/// the action request to which the current action response is associated. This isn't enforced.</para>
		/// </remarks>
		public override Type ActionResponseType
		{
			get { return typeof( MakePaymentActionResponse ); }
		}

		/// <summary>
		/// If this is true then action request is automatically saved if it was processed successfully (by the service which
		/// received it). The value is true. READ-ONLY.
		/// </summary>
		public override bool ServiceSavesReceivedActionRequest
		{
			get { return true; }
		}

		/// <summary>
		/// If this is true then action response is automatically saved if the action request was processed successfully (by the
		/// service which received it). The value is true. READ-ONLY.
		/// </summary>
		public override bool ServiceSavesSentActionResponse
		{
			get { return true; }
		}

		/// <summary>
		/// If this is true then the account of the initiator is automatically saved if the action request is processed
		/// successfully (by the service which received it). The value is true. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is true then an action service will lock the account for update, so it will no longer be able to
		/// process multiple action requests in parallel for the same account.</para>
		/// 
		/// <para>Don't change this value while an action request is processed because the storehouse locking of the account is
		/// optimized.</para>
		/// </remarks>
		public override bool ServiceSavesInitiatorAccount
		{
			get { return true; }
		}

		/// <summary>
		/// If this is true then the account of the beneficiary is automatically saved if the action request if processed
		/// successfully (by the service which received it). The value is true. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is true then an action service will lock the account for update, so it will no longer be able to
		/// process multiple action requests in parallel for the same account.</para>
		/// 
		/// <para>Don't change this value while an action request is processed because the storehouse locking of the account is
		/// optimized.</para>
		/// </remarks>
		public override bool ServiceSavesBeneficiaryAccount
		{
			get { return true; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private string _CurrencyName;
		/// <summary>
		/// The name of the currency to pay with.
		/// </summary>
		[SerializeAttribute]
		public string CurrencyName
		{
			get { return _CurrencyName; }
			set
			{
				ValidationUtils.TextHasContent( value, "CurrencyName" );

				_CurrencyName = value;
			}
		}

		private double _CurrencyAmount;
		/// <summary>
		/// The amount of currency, expressed in <see cref="CurrencyName"/> currency units, which the initiator of the action
		/// request pays to the beneficiary of the payment. This must be at least <see cref="double.Epsilon"/>.
		/// </summary>
		[SerializeAttribute]
		public double CurrencyAmount
		{
			get { return _CurrencyAmount; }
			set
			{
				ValidationUtils.Real( value, double.Epsilon, double.MaxValue, "CurrencyAmount" );

				_CurrencyAmount = value;
			}
		}

		#endregion Public serializable properties
	}
}
