////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.PaymentDocuments
{
	/// <summary>
	/// Fee charged by a payment service.
	/// </summary>
	public class ChargedPaymentFee : SignedDocument
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="ServiceUpnFingerprint"/> property.
		/// </summary>
		public const string ServiceUpnFingerprintFilterIdentifier = "ChargedPaymentFee.ServiceUpnFingerprint#4";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="ActionRequestLocator"/> property.
		/// </summary>
		public const string ActionRequestLocatorFilterIdentifier = "ChargedPaymentFee.ActionRequestLocator#5";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="ActionResponseLocator"/> property.
		/// </summary>
		public const string ActionResponseLocatorFilterIdentifier = "ChargedPaymentFee.ActionResponseLocator#6";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="InitiatorUpnFingerprint"/> property.
		/// </summary>
		public const string InitiatorUpnFingerprintFilterIdentifier = "ChargedPaymentFee.InitiatorUpnFingerprint#7";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="BeneficiaryUpnFingerprint"/> property.
		/// </summary>
		public const string BeneficiaryUpnFingerprintFilterIdentifier = "ChargedPaymentFee.BeneficiaryUpnFingerprint#8";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="Fee"/> property.
		/// </summary>
		public const string FeeFilterIdentifier = "ChargedPaymentFee.Fee#9";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ChargedPaymentFee()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">The locator of the sheaf.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="serviceUpnFingerprint">The UPN fingerprint of the service which has executed the associated payment.
		/// </param>
		/// <param name="actionRequestLocator">The locator of the associated action request.</param>
		/// <param name="actionResponseLocator">The locator of the associated action response.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the associated payment.</param>
		/// <param name="beneficiaryUpnFingerprint">The UPN fingerprint of the beneficiary of the associated payment.</param>
		/// <param name="fee">The fee charged for a payment. This must be at least 0.</param>
		public ChargedPaymentFee( Locator instanceLocator, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg
				, IHashSerializer signatureHashSerializer, Locator serviceUpnFingerprint, Locator actionRequestLocator
				, Locator actionResponseLocator, Locator initiatorUpnFingerprint, Locator beneficiaryUpnFingerprint, double fee )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer )
		{
			ServiceUpnFingerprint = serviceUpnFingerprint;
			ActionRequestLocator = actionRequestLocator;
			ActionResponseLocator = actionResponseLocator;
			InitiatorUpnFingerprint = initiatorUpnFingerprint;
			BeneficiaryUpnFingerprint = beneficiaryUpnFingerprint;
			Fee = fee;
		}

		#endregion Constructors

		#region Public serializable properties

		private Locator _ServiceUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the service which has executed the associated payment. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( ServiceUpnFingerprintFilterIdentifier )]
		public Locator ServiceUpnFingerprint
		{
			get { return _ServiceUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ServiceUpnFingerprint" );

				_ServiceUpnFingerprint = value;
			}
		}

		private Locator _ActionRequestLocator;
		/// <summary>
		/// The locator of the associated action request. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( ActionRequestLocatorFilterIdentifier )]
		public Locator ActionRequestLocator
		{
			get { return _ActionRequestLocator; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionRequestLocator" );

				_ActionRequestLocator = value;
			}
		}

		private Locator _ActionResponseLocator;
		/// <summary>
		/// The locator of the associated action response. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( ActionResponseLocatorFilterIdentifier )]
		public Locator ActionResponseLocator
		{
			get { return _ActionResponseLocator; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionResponseLocator" );

				_ActionResponseLocator = value;
			}
		}

		private Locator _InitiatorUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the initiator of the associated payment. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( InitiatorUpnFingerprintFilterIdentifier )]
		public Locator InitiatorUpnFingerprint
		{
			get { return _InitiatorUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "InitiatorUpnFingerprint" );

				_InitiatorUpnFingerprint = value;
			}
		}

		private Locator _BeneficiaryUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the beneficiary of the associated payment. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( BeneficiaryUpnFingerprintFilterIdentifier )]
		public Locator BeneficiaryUpnFingerprint
		{
			get { return _BeneficiaryUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "BeneficiaryUpnFingerprint" );

				_BeneficiaryUpnFingerprint = value;
			}
		}

		private double _Fee;
		/// <summary>
		/// The fee charged for a payment. This must be at least 0. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( FeeFilterIdentifier )]
		public double Fee
		{
			get { return _Fee; }
			protected set
			{
				ValidationUtils.Real( value, 0, double.MaxValue, "Fee" );

				_Fee = value;
			}
		}

		#endregion Public serializable properties
	}
}
