////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.PaymentDocuments
{
	/// <summary>
	/// Fee plan charged by a payment service for every payment.
	/// </summary>
	public class PaymentFeePlan : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected PaymentFeePlan()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="currencyName">The name of the currency of the payment.</param>
		/// <param name="fraction">The fraction from the paid amount of currency, which is charged as fee. This must be between
		/// 0 and 1.</param>
		public PaymentFeePlan( string currencyName, double fraction )
		{
			CurrencyName = currencyName;
			Fraction = fraction;
		}

		#endregion Constructors

		#region Public serializable properties

		private string _CurrencyName;
		/// <summary>
		/// The name of the currency of the payment. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string CurrencyName
		{
			get { return _CurrencyName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "CurrencyName" );

				_CurrencyName = value;
			}
		}

		private double _Fraction;
		/// <summary>
		/// The fraction from the paid amount of currency, which is charged as fee. This must be between 0 and 1. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public double Fraction
		{
			get { return _Fraction; }
			protected set
			{
				ValidationUtils.Real( value, 0, 1 - double.Epsilon, "Fraction" );

				_Fraction = value;
			}
		}

		private double _MinPaidCurrencyAmountToCharge;
		/// <summary>
		/// The minimum amount of paid currency from which to charge a fee. This must be at least 0.
		/// </summary>
		[SerializeAttribute]
		public double MinPaidCurrencyAmountToCharge
		{
			get { return _MinPaidCurrencyAmountToCharge; }
			set
			{
				ValidationUtils.Real( value, 0, double.MaxValue, "MinPaidCurrencyAmountToCharge" );

				_MinPaidCurrencyAmountToCharge = value;
			}
		}

		private double _FeeMin;
		/// <summary>
		/// The minimum amount of currency to charge as fee. This must be at least 0.
		/// </summary>
		[SerializeAttribute]
		public double FeeMin
		{
			get { return _FeeMin; }
			set
			{
				ValidationUtils.Real( value, 0, double.MaxValue, "FeeMin" );

				_FeeMin = value;
			}
		}

		private double _FeeMax;
		/// <summary>
		/// The maximum amount of currency to charge as fee. This must be at least 0.
		/// </summary>
		[SerializeAttribute]
		public double FeeMax
		{
			get { return _FeeMax; }
			set
			{
				ValidationUtils.Real( value, 0, double.MaxValue, "FeeMax" );

				_FeeMax = value;
			}
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Validates the current object.
		/// </summary>
		protected virtual void Validate()
		{
			if( FeeMin > FeeMax )
			{
				throw new ValidationException( string.Format( "The minimum fee {0} is higher than the maximum fee {1}."
					, FeeMin, FeeMax ) );
			}
		}

		/// <summary>
		/// This method is automatically called before the current object is serialized.
		/// </summary>
		protected virtual void Serializing()
		{
			Validate();
		}

		/// <summary>
		/// This method is automatically called after the current object has been deserialized.
		/// </summary>
		protected virtual void Deserialized()
		{
			Validate();
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Calculates the fee to charge from the specified amount of paid currency.
		/// </summary>
		/// <param name="currencyAmount">Amount of paid currency for which to calculate a fee. This must be at least
		/// <see cref="double.Epsilon"/>.</param>
		/// <param name="percentage">Is set with the percentage represented by the calculated fee, from the specified amount of
		/// paid currency.</param>
		/// <returns>Returns the fee to charge from the specified amount of paid currency. This can never be bigger than the
		/// specified amount of paid currency.</returns>
		public double Calculate( double currencyAmount, out double percentage )
		{
			ValidationUtils.Real( currencyAmount, double.Epsilon, double.MaxValue, "currencyAmount" );

			percentage = 0;

			if( currencyAmount < MinPaidCurrencyAmountToCharge ) return 0;

			double fee = currencyAmount * Fraction;

			if( fee < FeeMin ) fee = FeeMin;
			if( fee > FeeMax ) fee = FeeMax;

			if( fee >= currencyAmount )
			{
				throw new FatalCrossServiceException( string.Format( "The calculated fee {0} is equal with or higher than the"
					+ " amount {1} of paid currency.", fee, currencyAmount ) );
			}

			percentage = 100 * (fee / currencyAmount);

			return fee;
		}

		#endregion Public methods
	}
}
