////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// Account.
	/// </summary>
	/// <remarks>
	/// <para>An account is uniquely associated to a pair of a local and a remote peer, and is normally stored together with a
	/// tally and a presenter.</para>
	/// </remarks>
	public class Account : Sheaf, ILocalUpnFingerprintContainer, IRemoteUpnFingerprintContainer
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="LocalUpnFingerprint"/> property.
		/// </summary>
		public const string LocalUpnFingerprintFilterIdentifier = "Account.LocalUpnFingerprint#4";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="RemoteUpnFingerprint"/> property.
		/// </summary>
		public const string RemoteUpnFingerprintFilterIdentifier = "Account.RemoteUpnFingerprint#5";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="AliasTextualForm"/> property.
		/// </summary>
		public const string AliasFilterIdentifier = "Account.AliasTextualForm#6";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="RemotePeerMayRunLocalCommandLines"/> property.
		/// </summary>
		public const string RemotePeerMayRunLocalCommandLinesFilterIdentifier = "Account.RemotePeerMayRunLocalCommandLines#7";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Account()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="localUpn">The UPN of the local peer.</param>
		/// <param name="remoteUpn">The UPN of the remote peer.</param>
		/// <param name="alias">The account alias. This may be null.</param>
		public Account( Locator instanceLocator, Upn localUpn, Upn remoteUpn, AccountAlias alias )
			: base( instanceLocator )
		{
			LocalUpn = localUpn;
			RemoteUpn = remoteUpn;
			Alias = alias;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The UPN fingerprint of the local peer. READ-ONLY.
		/// </summary>
		[SpeckleFilterAttribute( LocalUpnFingerprintFilterIdentifier )]
		public Locator LocalUpnFingerprint
		{
			get { return LocalUpn.Fingerprint; }
		}

		/// <summary>
		/// The UPN fingerprint of the remote peer. READ-ONLY.
		/// </summary>
		[SpeckleFilterAttribute( RemoteUpnFingerprintFilterIdentifier )]
		public Locator RemoteUpnFingerprint
		{
			get { return RemoteUpn.Fingerprint; }
		}

		/// <summary>
		/// The textual form of the account alias. This may be null or empty. READ-ONLY.
		/// </summary>
		[SpeckleFilterAttribute( AliasFilterIdentifier )]
		public string AliasTextualForm
		{
			get { return (Alias != null) ? Alias.TextualForm : null; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Upn _LocalUpn;
		/// <summary>
		/// The UPN of the local peer. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Upn LocalUpn
		{
			get { return _LocalUpn; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "LocalUpn" );

				_LocalUpn = value;
			}
		}

		private Upn _RemoteUpn;
		/// <summary>
		/// The UPN of the remote peer. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Upn RemoteUpn
		{
			get { return _RemoteUpn; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RemoteUpn" );

				_RemoteUpn = value;
			}
		}

		private AccountAlias _Alias = new AccountAlias();
		/// <summary>
		/// The account alias. The backing field is initialized by default. For legacy reasons, this may be null, but may not be
		/// set with null.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public AccountAlias Alias
		{
			get { return _Alias; }
			set
			{
				ValidationUtils.ObjectHasContent( value, "Alias" );

				_Alias = value;
			}
		}

		/// <summary>
		/// Must be false by default!
		/// </summary>
		private bool _RemotePeerMayRunLocalCommandLines = false;
		/// <summary>
		/// If this is true then the remote peer may run command lines on the local computer (by sending letters whose messages
		/// are used as command lines).
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		[SpeckleFilterAttribute( RemotePeerMayRunLocalCommandLinesFilterIdentifier )]
		public bool RemotePeerMayRunLocalCommandLines
		{
			get { return _RemotePeerMayRunLocalCommandLines; }
			set { _RemotePeerMayRunLocalCommandLines = value; }
		}

		#endregion Public serializable properties
	}
}
