////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// Account (<see cref="Account"/>) alias.
	/// </summary>
	/// <remarks>
	/// <para>An account alias is meant to uniquely identify an account, but the uniqueness is not mandatory.</para>
	/// 
	/// <para>An account alias is an alias (<see cref="TextValidations.Alias"/>).</para>
	/// </remarks>
	public class AccountAlias : IDocument, ITextualFormContainer
	{
		#region Public constants

		/// <summary>
		/// The maximum number of characters to use from the remote entity name when composing the textual form of an account
		/// alias.
		/// </summary>
		public const int ComposedAliasRemoteEntityNameSizeMax = 6;

		/// <summary>
		/// The prefix for the textual form of an alias composed from a local and a remote entity name.
		/// </summary>
		public const string ComposedAliasTextualFormPrefix = "@";

		/// <summary>
		/// The template for the textual form of an alias composed from a local and a remote entity name. Placeholders:
		/// 0 - local entity name, 1 - remote entity name.
		/// </summary>
		public const string ComposedAliasTextualFormTemplate = ComposedAliasTextualFormPrefix + "{0}{1}";

		/// <summary>
		/// The minimum number of characters that the remote entity name must have before a whitespace, when composing the
		/// textual form of an account alias.
		/// </summary>
		public static readonly int ComposedAliasRemoteEntityNameSizeMinBeforeWhitespace = KernelConsts.AliasSizeMin
			- ComposedAliasTextualFormPrefix.Length;

		#endregion Public constants

		#region Constructor

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public AccountAlias()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="textualForm">The textual form of the account alias. This may be null or empty.</param>
		public AccountAlias( string textualForm )
		{
			TextualForm = textualForm;
		}

		#endregion Constructor

		#region Public serializable properties

		private string _TextualForm;
		/// <summary>
		/// The textual form of the account alias. This may be null or empty.
		/// </summary>
		[SerializeAttribute]
		public string TextualForm
		{
			get { return _TextualForm; }
			set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.Alias, "TextualForm" );

				_TextualForm = (value != null) ? value.ToLowerInvariant() : null;
			}
		}

		#endregion Public serializable properties

		#region Public static methods

		/// <summary>
		/// Returns true if the account alias is not null and its textual form has content.
		/// </summary>
		/// <param name="accountAlias">Account alias. This may be null.</param>
		/// <returns>Returns true if the account alias is not null and its textual form has content.</returns>
		public static bool HasContent( AccountAlias accountAlias )
		{
			return accountAlias != null && TextUtils.HasContent( accountAlias.TextualForm );
		}

		/// <summary>
		/// Returns true if the account alias is null or its textual form has no content.
		/// </summary>
		/// <param name="accountAlias">Account alias. This may be null.</param>
		/// <returns>Returns true if the account alias is null or its textual form has no content.</returns>
		public static bool NoContent( AccountAlias accountAlias )
		{
			return accountAlias == null || TextUtils.NoContent( accountAlias.TextualForm );
		}

		/// <summary>
		/// Checks if the specified account aliases have the same textual form.
		/// </summary>
		/// <param name="accountAlias1">Account alias to check for equality. This may be null.</param>
		/// <param name="accountAlias2">Account alias to check for equality. This may be null.</param>
		/// <returns>Returns true if the specified account aliases have the same textual form.</returns>
		public static bool Equals( AccountAlias accountAlias1, AccountAlias accountAlias2 )
		{
			if( NoContent( accountAlias1 ) || NoContent( accountAlias2 ) )
				return false;

			return TextUtils.Equals( accountAlias1.TextualForm, accountAlias2.TextualForm );
		}

		/// <summary>
		/// Checks if the specified account aliases have the same textual form.
		/// </summary>
		/// <param name="accountAlias1">Account alias to check for equality. This may be null or empty.</param>
		/// <param name="accountAlias2">Account alias to check for equality. This may be null or empty.</param>
		/// <returns>Returns true if the specified account aliases have the same textual form.</returns>
		public static bool Equals( string accountAlias1, string accountAlias2 )
		{
			if( TextUtils.NoContent( accountAlias1 ) || TextUtils.NoContent( accountAlias2 ) )
				return false;

			return TextUtils.Equals( accountAlias1, accountAlias2 );
		}

		/// <summary>
		/// Gets the textual form of the account alias from the specified message of a post.
		/// </summary>
		/// <param name="message">Message from which to get the account alias. This may be null or empty.</param>
		/// <param name="indexOfMessageWithoutAlias">Is set with the index from where the message without the account alias
		/// starts.</param>
		/// <returns>Returns the textual form of the found account alias. This may be null or empty.</returns>
		/// <remarks>
		/// <para>An account alias start from the beginning of the message and ends at the first white space.</para>
		/// </remarks>
		public static string GetTextualFormFromPostMessage( string message, out int indexOfMessageWithoutAlias )
		{
			indexOfMessageWithoutAlias = 0;

			if( TextUtils.NoContent( message ) )
				return null;

			int endPos = TextUtils.IndexOfWhitespace( message, 0 );

			if( (endPos <= 0) || (endPos < KernelConsts.AliasSizeMin) )
				return null;

			string alias = message.Substring( 0, endPos );
			string validationError = ValidationUtils.Text( alias, -2, -1, TextValidations.Alias, null, null, null, false );

			if( TextUtils.HasContent( validationError ) )
				return null;

			indexOfMessageWithoutAlias = TextUtils.IndexOfNonWhitespace( message, endPos );

			return alias;
		}

		/// <summary>
		/// Gets the account alias from the specified message of a post.
		/// </summary>
		/// <param name="message">Message from which to get the account alias. This may be null or empty.</param>
		/// <returns>Returns the found account alias. This may be null.</returns>
		/// <remarks>
		/// <para>An account alias start from the beginning of the message and ends at the first white space.</para>
		/// </remarks>
		public static AccountAlias FromPostMessage( string message )
		{
			int indexOfMessageWithoutAlias;
			string alias = GetTextualFormFromPostMessage( message, out indexOfMessageWithoutAlias );

			if( TextUtils.NoContent( alias ) )
				return null;

			return new AccountAlias( alias );
		}

		/// <summary>
		/// Gets the account alias from the specified message of a post.
		/// </summary>
		/// <param name="message">Message from which to get the account alias. This may be null or empty.</param>
		/// <param name="messageWithoutAlias">Is set with the message without the account alias. This may be null or empty.
		/// This is "message" if "message" contains no account alias.</param>
		/// <returns>Returns the found account alias. This may be null.</returns>
		/// <remarks>
		/// <para>An account alias start from the beginning of the message and ends at the first white space.</para>
		/// </remarks>
		public static AccountAlias FromPostMessage( string message, out string messageWithoutAlias )
		{
			int indexOfMessageWithoutAlias;
			string alias = GetTextualFormFromPostMessage( message, out indexOfMessageWithoutAlias );

			if( TextUtils.NoContent( alias ) )
			{
				messageWithoutAlias = message;

				return null;
			}

			messageWithoutAlias = (indexOfMessageWithoutAlias >= 0) ? message.Substring( indexOfMessageWithoutAlias ).Trim()
				: null;

			return new AccountAlias( alias );
		}

		/// <summary>
		/// Composes an account alias from the specified local and remote entity names.
		/// </summary>
		/// <param name="localEntityName">Local entity name.</param>
		/// <param name="remoteEntityName">Remote entity name.</param>
		/// <returns>Returns the composed account alias.</returns>
		public static AccountAlias FromLocalAndRemoteEntityNames( string localEntityName, string remoteEntityName )
		{
			ValidationUtils.TextHasContent( localEntityName, "localEntityName" );
			ValidationUtils.TextHasContent( remoteEntityName, "remoteEntityName" );

			string remotePart = TextUtils.GetBeginningLettersAndDigits( remoteEntityName, ComposedAliasRemoteEntityNameSizeMax
				, ComposedAliasRemoteEntityNameSizeMinBeforeWhitespace );

			string alias = string.Format( ComposedAliasTextualFormTemplate, "", remotePart );

			return (alias.Length >= KernelConsts.AliasSizeMin) ? new AccountAlias( alias ) : new AccountAlias();
		}

		#endregion Public static methods
	}
}
