////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// Appendix (which is shared by <see cref="Vault"/> and <see cref="Presenter"/>).
	/// </summary>
	public class Appendix : IExpirableAlgorithm
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public Appendix()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="entityName">The name of the identified entity.</param>
		public Appendix( string entityName )
		{
			EntityName = entityName;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public virtual DateTime ExpirationTime
		{
			get { return DateTime.MaxValue; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private bool _IsRevoked = false;
		/// <summary>
		/// If this is true then the associated vault is revoked and should not be used in processing.
		/// </summary>
		[SerializeAttribute]
		public bool IsRevoked
		{
			get { return _IsRevoked; }
			set { _IsRevoked = value; }
		}

		private string _EntityName;
		/// <summary>
		/// The name of the identified entity.
		/// </summary>
		[SerializeAttribute]
		public string EntityName
		{
			get { return _EntityName; }
			set
			{
				ValidationUtils.Text( value, -1, -1, TextValidations.EntityName, "EntityName" );

				_EntityName = value;
			}
		}

		private bool _UseOnlyEmbeddedActionServices = false;
		/// <summary>
		/// If this is true then the service application must allow the current vault to process action requests meant only for
		/// the embedded action services.
		/// </summary>
		[SerializeAttribute]
		public bool UseOnlyEmbeddedActionServices
		{
			get { return _UseOnlyEmbeddedActionServices; }
			set { _UseOnlyEmbeddedActionServices = value; }
		}

		private Strings _ActionServiceTypeNames = new Strings();
		/// <summary>
		/// The type names of the action services which can process action request whose recipient is the current vault; these
		/// are in addition to the default type names of the service application. This may be empty, but not null. The backing
		/// field is initialized by default. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If items are added to or removed from this list, make sure that the action service is restarted either
		/// automatically or by requesting the user to reopen the storehouse (in which the current vault is stored).</para>
		/// </remarks>
		[SerializeAttribute]
		public Strings ActionServiceTypeNames
		{
			get { return _ActionServiceTypeNames; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ActionServiceTypeNames" );

				_ActionServiceTypeNames = value;
			}
		}

		private string _ActionProcessingUrl;
		/// <summary>
		/// The URL of the action service which processes the received action requests. This may be null or empty.
		/// </summary>
		/// <remarks>
		/// <para>This may be used as an alternative to the UPN.</para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string ActionProcessingUrl
		{
			get { return _ActionProcessingUrl; }
			set { _ActionProcessingUrl = value; }
		}

		private int _ActionRequestValidity = PeerDocumentsConsts.ActionRequestValidity;
		/// <summary>		
		/// The timespan (in minutes) which must be added to the action request's "CreationTime" property in order to
		/// obtain the value of the "ExpirationTime" property. An action request expires at the time specified by its
		/// "ExpirationTime" property. This must be at least 1.
		/// </summary>
		/// <remarks>
		/// <para>Through this property, an action service instructs client applications to enforce the following condition:
		/// "ExpirationTime" has to be smaller than (or equal with) "CreationTime + ActionRequestValidity".</para>
		/// 
		/// <para>This property exists so that the action service may delete old action requests.</para>
		/// </remarks>
		[SerializeAttribute]
		public int ActionRequestValidity
		{
			get { return _ActionRequestValidity; }
			set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "ActionRequestValidity" );

				_ActionRequestValidity = value;
			}
		}

		private VirtualIdentity _VirtualIdentity;
		/// <summary>
		/// The virtual identity of the peer. This may be null.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public VirtualIdentity VirtualIdentity
		{
			get { return _VirtualIdentity; }
			set { _VirtualIdentity = value; }
		}

		private PhysicalAddress _PhysicalAddress;
		/// <summary>
		/// The physical address where the owner of the identity lives. This may be null.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public PhysicalAddress PhysicalAddress
		{
			get { return _PhysicalAddress; }
			set { _PhysicalAddress = value; }
		}

		private string _PublicNote;
		/// <summary>
		/// Public note. This may be null or empty.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string PublicNote
		{
			get { return _PublicNote; }
			set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.LongNote, "PublicNote" );

				_PublicNote = value;
			}
		}

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Generates the automated UPNs associated to the current appendix.
		/// </summary>
		/// <param name="upnAlg">The algorithm with which to generate the UPNs.</param>
		/// <returns>Returns the generated UPNs. This may be null or empty.</returns>
		/// <remarks>
		/// <para>The automated UPNs are not accessible to the users.</para>
		/// </remarks>
		public virtual Upns GenerateAutomatedUpns( UpnAlg upnAlg )
		{
			ValidationUtils.ObjectHasContent( upnAlg, "upnAlg" );

			return null;
		}

		#endregion Public methods
	}
}
