///////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
///////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// Attached file.
	/// </summary>
	public class AttachedFile : AttachedFileName
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AttachedFile()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="name">File name.</param>
		/// <param name="size">File size. This must be at least 1.</param>
		/// <param name="lastChangeTime">The time when the content of the file was last changed.</param>
		public AttachedFile( string name, Int64 size, DateTime lastChangeTime )
			: base( name )
		{
			Size = size;
			LastChangeTime = lastChangeTime;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="name">File name.</param>
		/// <remarks>
		/// <para><see cref="Size"/> and <see cref="LastChangeTime"/> are automatically set.</para>
		/// </remarks>
		public AttachedFile( string name )
			: base( name )
		{
			DateTime lastChangeTime;
			Size = FileUtils.GetFileSizeAndLastChangeTime( name, true, out lastChangeTime );
			LastChangeTime = lastChangeTime;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="attachedFileName">Attached file name.</param>
		/// <remarks>
		/// <para><see cref="Size"/> and <see cref="LastChangeTime"/> are automatically set.</para>
		/// </remarks>
		public AttachedFile( AttachedFileName attachedFileName )
			: base( attachedFileName )
		{
			DateTime lastChangeTime;
			Size = FileUtils.GetFileSizeAndLastChangeTime( AbsoluteFileName, true, out lastChangeTime );
			LastChangeTime = lastChangeTime;
		}

		#endregion Constructors

		#region Public serializable properties

		private Int64 _Size;
		/// <summary>
		/// File size. This must be at least 1. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Int64 Size
		{
			get { return _Size; }
			protected set
			{
				ValidationUtils.Integer( value, 1, Int64.MaxValue, "Size" );

				_Size = value;
			}
		}

		private DateTime _LastChangeTime;
		/// <summary>
		/// The time when the content of the file was last changed. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public DateTime LastChangeTime
		{
			get { return _LastChangeTime; }
			protected set { _LastChangeTime = value; }
		}

		private Int64 _CurrentSize;
		/// <summary>
		/// Current file size. This must be at least 0.
		/// </summary>
		/// <remarks>
		/// <para>This is used to keep track of how much of the current attached file has been transferred.</para>
		/// </remarks>
		[SerializeAttribute]
		public Int64 CurrentSize
		{
			get { return _CurrentSize; }
			set
			{
				ValidationUtils.Integer( value, 0, Int64.MaxValue, "CurrentSize" );

				_CurrentSize = value;
			}
		}

		#endregion Public serializable properties
	}
}
