///////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
///////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// Attached file name.
	/// </summary>
	public class AttachedFileName : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AttachedFileName()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="attachedFileName">The name of the file to attach.</param>
		public AttachedFileName( AttachedFileName attachedFileName )
		{
			ValidationUtils.ObjectHasContent( attachedFileName, "attachedFileName" );

			RootFolderName = attachedFileName.RootFolderName;
			RelativeFileName = attachedFileName.RelativeFileName;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="fileNameToAttach">The name of the file to attach. This is split (in here) in the root folder and the
		/// relative file name.</param>
		public AttachedFileName( string fileNameToAttach )
		{
			ValidationUtils.TextHasContent( fileNameToAttach, "fileNameToAttach" );

			RootFolderName = Path.GetDirectoryName( fileNameToAttach );
			RelativeFileName = Path.GetFileName( fileNameToAttach );
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="rootFolderName">The root folder name of the specified file name to attach. This will not be saved on the
		/// computer of the recipient of the attached file.</param>
		/// <param name="fileNameToAttach">The name of the file to attach.</param>
		public AttachedFileName( string rootFolderName, string fileNameToAttach )
		{
			ValidationUtils.TextHasContent( fileNameToAttach, "fileNameToAttach" );
			ValidationUtils.TextHasContent( fileNameToAttach, "fileNameToAttach" );

			if( !fileNameToAttach.StartsWith( rootFolderName, StringComparison.OrdinalIgnoreCase ) )
			{
				throw new ValidationException( "The name of the file to attach must start with the specified root folder"
						+ " name.", null, null
					, new ExceptionDetail( "File name to attach", fileNameToAttach )
					, new ExceptionDetail( "Root folder name", rootFolderName ) );
			}

			RootFolderName = rootFolderName;
			RelativeFileName = fileNameToAttach.Substring( rootFolderName.Length );
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// Absolute file name on the computer of the sender of the attached file. READ-ONLY.
		/// </summary>
		public string AbsoluteFileName
		{
			get { return FileUtils.MakeFullName( RootFolderName, RelativeFileName ); }
		}

		/// <summary>
		/// The sort key. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This must be <see cref="RelativeFileName"/> because that property is unique (see the property's comment about
		/// why it must be unique).</para>
		/// </remarks>
		public string SortKey
		{
			get { return RelativeFileName; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private string _RootFolderName;
		/// <summary>
		/// The root folder name which together with the relative file name (<see cref="RelativeFileName"/>) make an absolute
		/// file name on the computer of the sender of the attached file. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string RootFolderName
		{
			get { return _RootFolderName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "RootFolderName" );

				_RootFolderName = value;
			}
		}

		private string _RelativeFileName;
		/// <summary>
		/// Relative file name, the one which will be saved (by the recipient). This will be saved on the computer of the
		/// recipient of the attached file. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This must be unique in the list of attached files of a letter because the files are meant to be downloaded in
		/// the same folder (on the recipient's computer), so they must have unique file names in that folder, that is, this
		/// name.</para>
		/// 
		/// <para>This is relative to <see cref="RootFolderName"/> on the computer of the sender of the attached file, and
		/// relative to the root save folder on the computer of the recipient of the attached file.</para>
		/// </remarks>
		[SerializeAttribute]
		public string RelativeFileName
		{
			get { return _RelativeFileName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "RelativeFileName" );

				_RelativeFileName = value;
			}
		}

		#endregion Public serializable properties
	}
}
