///////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
///////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// List of attached files (<see cref="AttachedFile"/>).
	/// </summary>
	/// <remarks>
	/// <para>In order to ensure optimum UI performance (regarding the order of the downloaded and shown image previews), this
	/// and the <see cref="LocalAttachedFiles"/> lists have to be sorted in the same way.</para>
	/// </remarks>
	public class AttachedFiles : KernelSortedList<string, AttachedFile>
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public AttachedFiles()
			: base( StringInsensitiveComparer.Instance )
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="fileNames">The names of the attached files. This may be null or empty.</param>
		public AttachedFiles( AttachedFileNames fileNames )
			: base( StringInsensitiveComparer.Instance )
		{
			if( ListUtils.NoContent( fileNames ) )
				return;

			foreach( AttachedFileName fileName in fileNames )
				Add( new AttachedFile( fileName ) );
		}

		#endregion Constructors

		#region Indexers

		/// <summary>
		/// Gets the item with the specified file name in the list. READ-ONLY.
		/// </summary>
		/// <param name="fileName">The key of the item to get.</param>
		/// <returns>Returns the item with the file name, or null if the item is not found.</returns>
		public virtual AttachedFile this[ AttachedFileName fileName ]
		{
			get { return this[ fileName.SortKey ]; }
		}

		#endregion Indexers

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para><see cref="AttachedFileName.SortKey"/> is used as key.</para>
		/// </remarks>
		public override string GetKey( AttachedFile item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.SortKey;
		}

		/// <summary>
		/// Gets the total size of the files from the list.
		/// </summary>
		/// <returns>Returns the total size of the files from the list.</returns>
		public Int64 GetTotalSize()
		{
			Int64 totalSize = 0;

			foreach( AttachedFile attachedFile in this )
				totalSize += attachedFile.Size;

			return totalSize;
		}

		/// <summary>
		/// Checks if the current list contains the same files (= names, sizes and last change times) as the specified list.
		/// </summary>
		/// <param name="attachedFiles">List to compare with the current list. This may be empty, but not null.</param>
		/// <returns>Returns true if the current list contains the same files as the specified list.</returns>
		public bool SameFiles( AttachedFiles attachedFiles )
		{
			ValidationUtils.ObjectHasContent( attachedFiles, "attachedFiles" );

			foreach( AttachedFile attachedFile in this )
			{
				// Getting the existing file guarantees that the relative file names are the same (because that's the key).
				AttachedFile existingFile = attachedFiles.FindByItem( attachedFile );

				if( (existingFile == null) || (existingFile.Size != attachedFile.Size)
					|| (existingFile.LastChangeTime != attachedFile.LastChangeTime)
					|| !TextUtils.Equals( existingFile.RootFolderName, attachedFile.RootFolderName ) )
				{
					return false;
				}
			}

			return true;
		}

		#endregion Public methods
	}
}
