///////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
///////////////////////////////////////////////////////////////////////////////

using System.Collections.Generic;
using Vela.Kernel;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// List of attached files (<see cref="AttachedFile"/>).
	/// </summary>
	/// <remarks>
	/// <para>This is a sorted list to ensure that the image previews are always in the same order, to maintain consistency in
	/// the UI.</para>
	/// 
	/// <para>In order to ensure optimum UI performance (regarding the order of the downloaded and shown image previews), this
	/// and the <see cref="AttachedFiles"/> lists have to be sorted in the same way.</para>
	/// </remarks>
	public class LocalAttachedFiles : KernelSortedList<string, LocalAttachedFile>
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public LocalAttachedFiles()
			: base( StringInsensitiveComparer.Instance )
		{
		}

		#endregion Constructors

		#region Public static methods

		/// <summary>
		/// Gets the local file name of the first attached file from the list.
		/// </summary>
		/// <param name="localAttachedFiles">The local attached files from which to get the first file name. This may be null or
		/// empty.</param>
		/// <returns>Returns the local file name of the first attached file from the list. This may be null or empty.</returns>
		public static string GetFirstFileName( LocalAttachedFiles localAttachedFiles )
		{
			return (localAttachedFiles != null) ? localAttachedFiles.GetFirstItem().LocalFileName : null;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para><see cref="LocalAttachedFile.SortKey"/> is used as key.</para>
		/// </remarks>
		public override string GetKey( LocalAttachedFile item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.SortKey;
		}

		/// <summary>
		/// Adds a new item to the list.
		/// </summary>
		/// <param name="attachedFile">Attached file.</param>
		/// <param name="localFileName">Local file name.</param>
		public void Add( AttachedFile attachedFile, string localFileName )
		{
			LocalAttachedFile localAttachedFile = new LocalAttachedFile( attachedFile, localFileName );

			Add( localAttachedFile );
		}

		/// <summary>
		/// Gets the names of the current local attached files.
		/// </summary>
		/// <returns>Returns the names of the current local attached files.</returns>
		public List<string> GetFileNames()
		{
			List<string> fileNames = new List<string>();

			foreach( KeyValuePair<string, LocalAttachedFile> kvp in Items )
				fileNames.Add( kvp.Value.LocalFileName );

			return fileNames;
		}

		#endregion Public methods
	}
}
