////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// An account context.
	/// </summary>
	public class AccountContext : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class. READ-ONLY.
		/// </summary>
		protected AccountContext()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="account">The account.</param>
		/// <param name="vault">The vault of the local peer (who has sought over the network and found the contained remote
		/// presenter).</param>
		/// <param name="localPhoto">The photo of the local peer. This may be null.</param>
		/// <param name="presenter">The presenter of the remote peer.</param>
		/// <param name="tally">The tally.</param>
		/// <param name="remotePhoto">The photo of the remote peer. This may be null.</param>
		public AccountContext( Account account, Vault vault, Photo localPhoto, Presenter presenter, Tally tally
			, Photo remotePhoto )
		{
			Account = account;
			Vault = vault;
			LocalPhoto = localPhoto;
			Presenter = presenter;
			Tally = tally;
			RemotePhoto = remotePhoto;

			ValidateLocators();
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The account locator. READ-ONLY.
		/// </summary>
		public Locator AccountLocator
		{
			get { return Account.InstanceLocator; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Account _Account;
		/// <summary>
		/// The account. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Account Account
		{
			get { return _Account; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Account" );

				_Account = value;
			}
		}

		private Vault _Vault;
		/// <summary>
		/// The vault of the local peer (who has sought over the network and found the contained remote presenter). READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Vault Vault
		{
			get { return _Vault; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Vault" );

				_Vault = value;
			}
		}

		private Photo _LocalPhoto;
		/// <summary>
		/// The photo of the local peer. This may be null. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Photo LocalPhoto
		{
			get { return _LocalPhoto; }
			protected set { _LocalPhoto = value; }
		}

		private Presenter _Presenter;
		/// <summary>
		/// The presenter of the remote peer. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Presenter Presenter
		{
			get { return _Presenter; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Presenter" );

				_Presenter = value;
			}
		}

		private Tally _Tally;
		/// <summary>
		/// The tally. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Tally Tally
		{
			get { return _Tally; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Tally" );

				_Tally = value;
			}
		}

		private Photo _RemotePhoto;
		/// <summary>
		/// The photo of the remote peer. This may be null. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>Although the photo is normally stored in the presenter, it's possible that a presenter with a photo was
		/// previously used to create an account, but another presenter with no photo was used later used to update the account.
		/// So, if the photo from the presenter were shown to the local peer, the local peer would not see a photo of the remote
		/// peer. To avoid this, the photo sheaf is shown instead of the one from the presenter, so that is stored here.</para>
		/// </remarks>
		[SerializeAttribute]
		public Photo RemotePhoto
		{
			get { return _RemotePhoto; }
			protected set { _RemotePhoto = value; }
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Validates the locators.
		/// </summary>
		protected void ValidateLocators()
		{
			if( !Vault.LocalUpnFingerprint.Equals( Account.LocalUpnFingerprint ) )
			{
				throw new ValidationException( "Different vault-account UPN fingerprints."
					, null, "Vault.LocalUpnFingerprint != Account.LocalUpnFingerprint" );
			}

			if( !Presenter.CreatorUpnFingerprint.Equals( Account.RemoteUpnFingerprint ) )
			{
				throw new ValidationException( "Different presenter-account UPN fingerprints."
					, null, "Presenter.RemoteUpnFingerprint != Account.RemoteUpnFingerprint" );
			}
		}

		#endregion Private methods
	}
}
