////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// Ephemerid.
	/// </summary>
	/// <remarks>
	/// <para>An ephemerid is like a tally, but it's meant to store non-critical information (like an IP address). Therefore,
	/// all the ephemerids from a storehouse may be deleted at any time.</para>
	/// 
	/// <para>This is meant to be loaded and updated on a transaction other than the main transaction of an action client /
	/// service, so as to not affect the scalability.</para>
	/// </remarks>
	public class Ephemerid : Sheaf, IRemoteUpnFingerprintContainer
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="RemoteUpnFingerprint"/> property.
		/// </summary>
		public const string RemoteUpnFingerprintFilterIdentifier = "Ephemerid.RemoteUpnFingerprint#4";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Ephemerid()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="remoteUpnFingerprint">The UPN fingerprint of the remote peer.</param>
		public Ephemerid( Locator instanceLocator, Locator remoteUpnFingerprint )
			: base( instanceLocator )
		{
			RemoteUpnFingerprint = remoteUpnFingerprint;
		}

		#endregion Constructors

		#region Public serializable properties

		private Locator _RemoteUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the remote peer. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( RemoteUpnFingerprintFilterIdentifier )]
		public Locator RemoteUpnFingerprint
		{
			get { return _RemoteUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RemoteUpnFingerprint" );

				_RemoteUpnFingerprint = value;
			}
		}

		private PeerIpAddress _IpAddress;
		/// <summary>
		/// The (last known) IP addresses of the remote peer. This may be null.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public PeerIpAddress IpAddress
		{
			get { return _IpAddress; }
			set { _IpAddress = value; }
		}

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Gets the (last known) IP address of the remote peer, as a buffer.
		/// </summary>
		/// <remarks>Returns the IP address buffer. This may be null or empty.</remarks>
		public byte[] GetIpAddressBuffer()
		{
			return (IpAddress != null) ? IpAddress.IpAddress : null;
		}

		#endregion Public methods
	}
}
