////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
namespace Vela.PeerDocuments
{
	/// <summary>
	/// Constants used for peer documents.
	/// </summary>
	public static class PeerDocumentsConsts
	{
		#region Public constants

		/// <summary>
		/// The default value (in days) for the validity of an action request.
		/// </summary>
		public const int ActionRequestValidityInDays = 16;

		/// <summary>
		/// The default value (in minutes) for the validity of an action request.
		/// </summary>
		/// <remarks>
		/// <para>For anonymous action requests, this is always used because there is no presenter from which to get the
		/// actual action request validity of the service.</para>
		/// </remarks>
		public const int ActionRequestValidity = ActionRequestValidityInDays * 24 * 60;

		/// <summary>
		/// The minimum tally approval which an account must have in order for the remote peer to be considered a "known" peer.
		/// </summary>
		public const TallyApprovals KnownPeerTallyApprovalMin = TallyApprovals.Yellow;

		/// <summary>
		/// The minimum tally approval which an account must have in order for the local peer to process non-anonymous action
		/// requests from the remote peer.
		/// </summary>
		public const TallyApprovals ProcessNonAnonymousActionRequestsTallyApprovalMin = TallyApprovals.Red;

		/// <summary>
		/// The minimum tally approval which an account must have in order to be considered not blocked. If an account is
		/// blocked, all received action requests from the remote peer return an error.
		/// </summary>
		public const TallyApprovals NotBlockedAccountTallyApprovalMin = TallyApprovals.Red;

		/// <summary>
		/// The minimum tally approval which an account must have in order to allow the local peer to receive files from the
		/// remote peer.
		/// </summary>
		public const TallyApprovals ReceiveFilesTallyApprovalMin = TallyApprovals.Blue;

		/// <summary>
		/// The minimum tally approval which an account must have in order to send to the remote peer the photo of the local
		/// peer.
		/// </summary>
		public const TallyApprovals SendPhotoTallyApprovalMin = TallyApprovals.Blue;

		/// <summary>
		/// The minimum tally approval which an account must have in order to send to the remote peer the intimate information of
		/// the local peer.
		/// </summary>
		public const TallyApprovals SendIntimateInfoTallyApprovalMin = TallyApprovals.Green;

		/// <summary>
		/// The minimum tally approval which an account must have in order for a remote peer to be allowed to run command lines
		/// on the local computer (by sending letters whose messages are used as command lines).
		/// </summary>
		public const TallyApprovals RemotePeerMayRunLocalCommandLinesTallyApprovalMin = TallyApprovals.White;

		/// <summary>
		/// The tally approval which is used when (the tally of) an account is created.
		/// </summary>
		public const TallyApprovals CreateAccountApprovalMin = TallyApprovals.Blue;

		/// <summary>
		/// The width (in pixels) of a thumb-sized photo.
		/// </summary>
		/// <remarks>
		/// <para>XAML requires a <see cref="double"/> value.</para>
		/// </remarks>
		public const double ThumbSizedPhotoWidth = 3 * KernelConsts.ImageSizeReference;

		/// <summary>
		/// The width (in pixels) of a thumb-sized photo.
		/// </summary>
		public const int ThumbSizedPhotoWidthInt = (int)ThumbSizedPhotoWidth;

		/// <summary>
		/// The height (in pixels) of a thumb-sized photo.
		/// </summary>
		/// <remarks>
		/// <para>XAML requires a <see cref="double"/> value.</para>
		/// </remarks>
		public const double ThumbSizedPhotoHeight = 3 * KernelConsts.ImageSizeReference;

		/// <summary>
		/// The height (in pixels) of a thumb-sized photo.
		/// </summary>
		public const int ThumbSizedPhotoHeightInt = (int)ThumbSizedPhotoHeight;

		/// <summary>
		/// The width (in pixels) of a full-sized photo.
		/// </summary>
		/// <remarks>
		/// <para>XAML requires a <see cref="double"/> value.</para>
		/// </remarks>
		public const double FullSizedPhotoWidth = 6 * KernelConsts.ImageSizeReference;

		/// <summary>
		/// The width (in pixels) of a full-sized photo.
		/// </summary>
		public const int FullSizedPhotoWidthInt = (int)FullSizedPhotoWidth;

		/// <summary>
		/// The height (in pixels) of a full-sized photo.
		/// </summary>
		/// <remarks>
		/// <para>XAML requires a <see cref="double"/> value.</para>
		/// </remarks>
		public const double FullSizedPhotoHeight = 9 * KernelConsts.ImageSizeReference;

		/// <summary>
		/// The height (in pixels) of a full-sized photo.
		/// </summary>
		public const int FullSizedPhotoHeightInt = (int)FullSizedPhotoHeight;

		/// <summary>
		/// Compression quality (in percentages) for a photo.
		/// </summary>
		public const int PhotoQuality = 80;

		/// <summary>
		/// The age (in years, of the user) after which sensual emoticons may be shown to the user.
		/// </summary>
		public const int MayShowSensualEmoticonsAge = 16;

		/// <summary>
		/// The maximum size (in characters) of the message of a letter.
		/// </summary>
		public const int PostMessageSizeMax = 100000;

		/// <summary>
		/// The maximum size (in characters) of the message summary.
		/// </summary>
		public const int PostMessageSummarySizeMax = 500;

		/// <summary>
		/// The maximum number of lines of the message summary.
		/// </summary>
		public const int PostMessageSummaryLineCountMax = 5;

		/// <summary>
		/// The suffix added to a letter's message summary in order indicate that the full message contains more text.
		/// </summary>
		/// <remarks>
		/// <para>There has to be a blank space at the beginning in order to avoid conflicts with the URL detection mechanism,
		/// when this text is added to the message summary.</para>
		/// </remarks>
		public const string PostMessageHasMoreContentSuffix = " ......";

		/// <summary>
		/// The suffix added to a text in order indicate that it was truncated (and the original had more content).
		/// </summary>
		public const string TextWasTruncatedSuffix = "";

		/// <summary>
		/// The maximum size (in bytes) of an image whose preview may be loaded.
		/// </summary>
		public const int LetterImagePreviewSizeMax = 10000000;

		/// <summary>
		/// The maximum number of attached image files for which image previews can be shown.
		/// </summary>
		public const int ShowPreviewsForAttachedImageFilesCountMax = 3;

		#endregion Public constants
	}
}
