////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.ComponentModel;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// The available approvals of a tally.
	/// </summary>
	/// <remarks>
	/// <para>It's guaranteed that the trust level of the approval increases with the underlying number of the enumeration value.
	/// </para>
	/// 
	/// <para>All the approvals are allowed to process anonymous action requests without mutual knowledge because trying load the
	/// tally for every incoming such request in order to verify the approval could be devastating in terms of performance (for
	/// example, UPN resolution uses an anonymous request).</para>
	/// 
	/// <para>Presenter updates are not sent to blocked peers (see the <see cref="PeerDocumentsUtils.IsBlockedPeer"/> method for
	/// details).</para>
	/// </remarks>
	public enum TallyApprovals
	{
		/// <summary>
		/// Only the anonymous action requests without mutual knowledge coming for such a tally are processed, the others are
		/// rejected.
		/// </summary>
		Black,

		/// <summary>
		/// The associated remote vault is revoked. Only the anonymous action requests without mutual knowledge coming for such a
		/// tally are processed, the others are rejected.
		/// </summary>
		Revoked,

		/// <summary>
		/// The tally is temporarily suspended. Only the anonymous action requests without mutual knowledge coming for such a
		/// tally are processed, the others are rejected.
		/// </summary>
		Suspended,

		/// <summary>
		/// The tally is pending review. This is used only when the tally is automatically created, that is, when the local peer
		/// did not approve the creation of the tally. Only the anonymous action requests coming for such a tally are processed,
		/// the others are rejected. Sending letters and files are not anonymous operations.
		/// </summary>
		Red,

		/// <summary>
		/// The tally belongs to a far acquaintance. Offers to download files are not accepted. This is the minimum tally
		/// approval which an account must have in order for the remote peer to be considered a "known" peer, a peer with which
		/// communication is allowed with only minor limitations.
		/// </summary>
		Yellow,

		/// <summary>
		/// The tally belongs to a close acquaintance, coworker or friend. The remote peer may see the following information:
		/// photo.
		/// </summary>
		Blue,

		/// <summary>
		/// The tally belongs to a close friend or family member. The remote peer may see the following information: photo,
		/// intimate information (= email address, phone number, website URL, birth date, intimate note, gender, sentimental
		/// relationship, physical address).
		/// </summary>
		Green,

		/// <summary>
		/// This is like <see cref="TallyApprovals.Green"/>, but remote peer may change the wallpaper of the local peer; the
		/// wallpaper may be changed with an image file attached to a letter; the emoticons may be updated with custom ones
		/// which are attached to a letter. The remote peer may run command lines on the local computer (by sending letters
		/// whose messages are used as command lines); for this to be possible, there is a dedicated option on the account which
		/// must be checked (<see cref="Account.RemotePeerMayRunLocalCommandLines"/>).
		/// </summary>
		White
	}

	/// <summary>
	/// The available sentimental relationships.
	/// </summary>
	public enum SentimentalRelationships
	{
		/// <summary>
		/// No state has been specified.
		/// </summary>
		[Description( "Not specified" )]
		NotSpecified,

		/// <summary>
		/// The peer is single.
		/// </summary>
		Single,

		/// <summary>
		/// The peer is involved in a sentimental relationship with a partner.
		/// </summary>
		[Description( "With partner" )]
		WithPartner,

		/// <summary>
		/// The peer is married.
		/// </summary>
		Married
	}

	/// <summary>
	/// The available flags of a post.
	/// </summary>
	public enum PostFlags
	{
		/// <summary>
		/// The post is archived (for historic purposes).
		/// </summary>
		Archive,

		/// <summary>
		/// The post is favorite.
		/// </summary>
		Favorite,

		/// <summary>
		/// The post is meant for main use, working as both inbox and outbox.
		/// </summary>
		Main,

		/// <summary>
		/// The post contains an ongoing task (meant to be handled as soon as possible).
		/// </summary>
		Ongoing,

		/// <summary>
		/// The post is trash (meant to be deleted later).
		/// </summary>
		Trash
	}
}
