////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Kernel;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// Utilities for working with peer documents.
	/// </summary>
	public static class PeerDocumentsUtils
	{
		#region Public static methods

		/// <summary>
		/// Returns true if the specified approval identifies a blocked peer.
		/// </summary>
		/// <param name="approval">Approval to check if it identifies a blocked peer.</param>
		/// <returns>Returns true if the specified approval identifies a blocked peer.</returns>
		public static bool IsBlockedPeer( TallyApprovals approval )
		{
			return approval < PeerDocumentsConsts.NotBlockedAccountTallyApprovalMin;
		}

		/// <summary>
		/// Returns true if the specified approval identifies a "known" peer. The approval has to be at least
		/// <see cref="PeerDocumentsConsts.KnownPeerTallyApprovalMin"/>.
		/// </summary>
		/// <param name="approval">Approval to check if it identifies a "known" peer.</param>
		/// <returns>Returns true if the specified approval identifies a "known" peer.</returns>
		public static bool IsKnownPeer( TallyApprovals approval )
		{
			return approval >= PeerDocumentsConsts.KnownPeerTallyApprovalMin;
		}

		/// <summary>
		/// Returns true if the specified approval is enough for a remote peer to run command lines on the local computer, by
		/// sending letters whose messages are used as command lines. The approval has to be at least
		/// <see cref="PeerDocumentsConsts.RemotePeerMayRunLocalCommandLinesTallyApprovalMin"/>.
		/// </summary>
		/// <param name="approval">Approval to check if it's enough.</param>
		/// <returns>Returns true if the specified approval is enough for a remote peer to run command lines on the local
		/// computer.</returns>
		public static bool RemotePeerMayRunLocalCommandLines( TallyApprovals approval )
		{
			return approval >= PeerDocumentsConsts.RemotePeerMayRunLocalCommandLinesTallyApprovalMin;
		}

		/// <summary>
		/// Gets the summary of the specified message of a post.
		/// </summary>
		/// <param name="message">Message whose summary to get. This may be null or empty.</param>
		/// <returns>Returns the summary of the specified message.</returns>
		public static string GetPostMessageSummary( string message )
		{
			return TextUtils.GetBeginning( message, PeerDocumentsConsts.PostMessageSummarySizeMax, true
				, PeerDocumentsConsts.PostMessageSummaryLineCountMax, PeerDocumentsConsts.PostMessageHasMoreContentSuffix );
		}

		/// <summary>
		/// Finds the speckle with the specified local and remote UPN fingerprints.
		/// </summary>
		/// <param name="accountSpeckles">Account speckles in which to look. This may be empty, but not null.</param>
		/// <param name="localUpnFingerprint">The local UPN fingerprint of the account speckle to find.</param>
		/// <param name="remoteUpnFingerprint">The remote UPN fingerprint of the account speckle to find.</param>
		/// <returns>Returns the found speckle, or null if none is found.</returns>
		public static Speckle FindAccountByLocalAndRemoteUpnFingerprints( Speckles accountSpeckles, Locator localUpnFingerprint
			, Locator remoteUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( accountSpeckles, "accountSpeckles" );
			ValidationUtils.ObjectHasContent( localUpnFingerprint, "localUpnFingerprint" );
			ValidationUtils.ObjectHasContent( remoteUpnFingerprint, "remoteUpnFingerprint" );

			foreach( Speckle speckle in accountSpeckles )
			{
				Locator currentLocalUpnFingerprint = speckle.GetLoadedSpeckleFilter<Locator>(
					Account.LocalUpnFingerprintFilterIdentifier );
				Locator currentRemoteUpnFingerprint = speckle.GetLoadedSpeckleFilter<Locator>(
					Account.RemoteUpnFingerprintFilterIdentifier );

				if( localUpnFingerprint.Equals( currentLocalUpnFingerprint )
					&& remoteUpnFingerprint.Equals( currentRemoteUpnFingerprint ) )
				{
					return speckle;
				}
			}

			return null;
		}

		/// <summary>
		/// Finds the speckle with the specified initiator and recipient UPN fingerprints.
		/// </summary>
		/// <param name="accountSpeckles">Account speckles in which to look. This may be empty, but not null.</param>
		/// <param name="initiatorUpnFingerprint">The initiator UPN fingerprint of the account speckle to find.</param>
		/// <param name="recipientUpnFingerprint">The recipient UPN fingerprint of the account speckle to find.</param>
		/// <returns>Returns the found speckle, or null if none is found.</returns>
		/// <remarks>
		/// <para>Use this method when the local and the remote UPN fingerprints are not yet available because the local peer
		/// direction is not (yet) known.</para>
		/// </remarks>
		public static Speckle FindAccountByInitiatorAndRecipientUpnFingerprints( Speckles accountSpeckles
			, Locator initiatorUpnFingerprint, Locator recipientUpnFingerprint )
		{
			ValidationUtils.ObjectHasContent( accountSpeckles, "accountSpeckles" );
			ValidationUtils.ObjectHasContent( initiatorUpnFingerprint, "initiatorUpnFingerprint" );
			ValidationUtils.ObjectHasContent( recipientUpnFingerprint, "recipientUpnFingerprint" );

			foreach( Speckle speckle in accountSpeckles )
			{
				Locator localUpnFingerprint = speckle.GetLoadedSpeckleFilter<Locator>(
					Account.LocalUpnFingerprintFilterIdentifier );
				Locator remoteUpnFingerprint = speckle.GetLoadedSpeckleFilter<Locator>(
					Account.RemoteUpnFingerprintFilterIdentifier );

				// Since the actual local peer direction is not known, verify both directions. Only one combination can be valid
				// for the local peer.
				if(
					(
						localUpnFingerprint.Equals( initiatorUpnFingerprint )
						&& remoteUpnFingerprint.Equals( recipientUpnFingerprint )
					)
					|| (
						localUpnFingerprint.Equals( recipientUpnFingerprint )
						&& remoteUpnFingerprint.Equals( initiatorUpnFingerprint )
					) )
				{
					return speckle;
				}
			}

			return null;
		}

		#endregion Public static methods
	}
}
