////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// A peer IP address.
	/// </summary>
	public class PeerIpAddress : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected PeerIpAddress()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="ipAddress">The IP address.</param>
		public PeerIpAddress( byte[] ipAddress )
		{
			IpAddress = ipAddress;
		}

		#endregion Constructors

		#region Public serializable properties

		private DateTime _CreationTime = DateTime.UtcNow;
		/// <summary>
		/// The time (in UTC format) when the peer IP address was created. This is automatically set with the time of the
		/// creation of the current object (and you should not modify it). READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public DateTime CreationTime
		{
			get { return _CreationTime; }
			protected set { _CreationTime = value; }
		}

		private byte[] _IpAddress;
		/// <summary>
		/// The IP address. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] IpAddress
		{
			get { return _IpAddress; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "IpAddress" );

				_IpAddress = value;
			}
		}

		#endregion Public serializable properties

		#region Public static methods

		/// <summary>
		/// Checks if the specified IP addresses are equal.
		/// </summary>
		/// <param name="ipAddress1">IP address to compare. This may be null.</param>
		/// <param name="ipAddress2">IP address to compare. This may be null.</param>
		/// <returns>Returns true if the two IP addresses are equal.</returns>
		public static bool Equals( PeerIpAddress ipAddress1, PeerIpAddress ipAddress2 )
		{
			if( (ipAddress1 == null) && (ipAddress2 != null) ) return false;
			if( (ipAddress1 != null) && (ipAddress2 == null) ) return false;

			return BufferUtils.Equals( ipAddress1.IpAddress, ipAddress2.IpAddress );
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Checks if the current IP address is equal with the specified IP address.
		/// </summary>
		/// <param name="ipAddress">IP address to compare with. This may be null.</param>
		/// <returns>Returns true if the two IP addresses are equal.</returns>
		public bool Equals( PeerIpAddress ipAddress )
		{
			if( ipAddress == null ) return false;

			return BufferUtils.Equals( IpAddress, ipAddress.IpAddress );
		}

		#endregion Public methods
	}
}
