////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// The main image which may be used to physically identify a peer.
	/// </summary>
	/// <remarks>
	/// <para>A photo is uniquely associated to a creator peer, and is normally stored together with a vault or tally.</para>
	/// </remarks>
	public class Photo : Sheaf, ICreatorUpnFingerprintContainer
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="CreatorUpnFingerprint"/> property.
		/// </summary>
		public const string CreatorUpnFingerprintFilterIdentifier = "Photo.CreatorUpnFingerprint#4";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Photo()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="creatorUpnFingerprint">The UPN fingerprint of the creator peer.</param>
		/// <param name="image">Peer image. This may be null.</param>
		public Photo( Locator instanceLocator, Locator creatorUpnFingerprint, IPeerImage image = null )
			: base( instanceLocator )
		{
			CreatorUpnFingerprint = creatorUpnFingerprint;
			Image = image;
		}

		#endregion Constructors

		#region Public serializable properties

		private Locator _CreatorUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the creator peer. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( CreatorUpnFingerprintFilterIdentifier )]
		public Locator CreatorUpnFingerprint
		{
			get { return _CreatorUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "CreatorUpnFingerprint" );

				_CreatorUpnFingerprint = value;
			}
		}

		private IPeerImage _Image;
		/// <summary>
		/// Peer image. This may be null.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public IPeerImage Image
		{
			get { return _Image; }
			set { _Image = value; }
		}

		#endregion Public serializable properties
	}
}
