////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// Contract for a virtual identity which contains information related to an entity.
	/// </summary>
	public abstract class VirtualIdentity : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected VirtualIdentity()
		{
		}

		#endregion Constructors

		#region Public serializable properties

		private string _EmailAddress;
		/// <summary>
		/// The email address of the peer. This may be null or empty.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string EmailAddress
		{
			get { return _EmailAddress; }
			set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.EmailAddress, "EmailAddress" );

				_EmailAddress = value;
			}
		}

		private string _PhoneNumber;
		/// <summary>
		/// The phone number of the peer. This may be null or empty.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string PhoneNumber
		{
			get { return _PhoneNumber; }
			set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.PhoneNumber, "PhoneNumber" );

				_PhoneNumber = value;
			}
		}

		private string _WebsiteUrl;
		/// <summary>
		/// The URL of the website of the peer. This may be null or empty.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string WebsiteUrl
		{
			get { return _WebsiteUrl; }
			set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.Url, "WebsiteUrl" );

				_WebsiteUrl = value;
			}
		}

		private DateTime _BirthDate = KernelConsts.DateMin;
		/// <summary>
		/// The birth date (in UTC format). If this has the <see cref="KernelConsts.DateMin"/> value, it is considered that no
		/// birth date has been specified.
		/// </summary>
		[SerializeAttribute]
		public DateTime BirthDate
		{
			get { return _BirthDate; }
			set
			{
				ValidationUtils.Time( value, KernelConsts.DateMin, DateTime.MaxValue, "BirthDate" );

				_BirthDate = value;
			}
		}

		private string _IntimateNote;
		/// <summary>
		/// Intimate note. This may be null or empty.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string IntimateNote
		{
			get { return _IntimateNote; }
			set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.LongNote, "IntimateNote" );

				_IntimateNote = value;
			}
		}

		#endregion Public serializable properties
	}
}
