////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// A bulletin which can be read by other entities (in order to get news about the poster).
	/// </summary>
	/// <remarks>
	/// <para>When a bulletin is created, a letter is also created for each remote peer who has a specified minimum approval,
	/// letter which has the content of the bulletin. While the letters are sent to their recipients, the bulletin is not sent
	/// anywhere.</para>
	/// </remarks>
	public class Bulletin : Post
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Bulletin()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="message">The message of the post.</param>
		/// <param name="emoticonSymbol">The emoticon symbol of the post. This may be null or empty.</param>
		/// <param name="isTransient">If this is true then the current letter is transient.</param>
		/// <param name="minimumApproval">The minimum approval of the tallies of the remote peers for whom a letter will be sent
		/// with the content of the current bulletin.</param>
		public Bulletin( Locator instanceLocator, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg
				, IHashSerializer signatureHashSerializer, Locator initiatorUpnFingerprint, string message
				, string emoticonSymbol, bool isTransient, TallyApprovals minimumApproval )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer, initiatorUpnFingerprint, message
				, emoticonSymbol, isTransient )
		{
			MinimumApproval = minimumApproval;
		}

		#endregion Constructors

		#region Public serializable properties

		private TallyApprovals _MinimumApproval = PeerDocumentsConsts.KnownPeerTallyApprovalMin;
		/// <summary>
		/// The minimum approval of the tallies of the remote peers for whom a letter will be sent with the content of the
		/// current bulletin. READ-ONLY.
		/// </summary>
		public TallyApprovals MinimumApproval
		{
			get { return _MinimumApproval; }
			protected set { _MinimumApproval = value; }
		}

		#endregion Public serializable properties
	}
}
