////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// A letter which can be sent to other entities in order to communicate.
	/// </summary>
	public class Letter : Post, IRecipientUpnFingerprintContainer
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="RecipientUpnFingerprint"/> property.
		/// </summary>
		public const string RecipientUpnFingerprintFilterIdentifier = "Letter.RecipientUpnFingerprint#10";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="AttachedFilesCount"/> property.
		/// </summary>
		/// <remarks>
		/// <para>For legacy reasons, this has to overwrite another filter.</para>
		/// </remarks>
		public const string AttachedFilesCountFilterIdentifier = "Letter.AttachedFilesCount#11>Letter.HasAttachedFiles";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="HasAttachedMediaFiles"/> property.
		/// </summary>
		public const string HasAttachedMediaFilesFilterIdentifier = "Letter.HasAttachedMediaFiles#12";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="HasImagePreview"/> property.
		/// </summary>
		public const string HasImagePreviewFilterIdentifier = "Letter.HasImagePreview#13";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Letter()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="message">The message of the letter.</param>
		/// <param name="emoticonSymbol">The emoticon symbol of the letter. This may be null or empty.</param>
		/// <param name="isTransient">If this is true then the letter is transient.</param>
		/// <param name="recipientUpnFingerprint">The UPN fingerprint of the recipient of the current object.</param>
		/// <param name="attachedFiles">Files that the remote peer may download.</param>
		/// <param name="hasAttachedMediaFiles">If this is true then some of the attached files are media files.</param>
		/// <param name="hasImagePreview">If this is true then the letter has an image preview, that is, it has an attached image
		/// file.</param>
		public Letter( Locator instanceLocator, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg
				, IHashSerializer signatureHashSerializer, Locator initiatorUpnFingerprint, string message, string emoticonSymbol
				, bool isTransient, Locator recipientUpnFingerprint, AttachedFiles attachedFiles, bool hasAttachedMediaFiles
				, bool hasImagePreview )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer, initiatorUpnFingerprint, message
				, emoticonSymbol, isTransient )
		{
			RecipientUpnFingerprint = recipientUpnFingerprint;
			AttachedFiles = attachedFiles;
			HasAttachedMediaFiles = hasAttachedMediaFiles;
			HasImagePreview = hasImagePreview;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The number of files attached to the current letter. READ-ONLY.
		/// </summary>
		[SpeckleFilterAttribute( AttachedFilesCountFilterIdentifier )]
		public int AttachedFilesCount
		{
			get { return AttachedFiles.Count; }
		}

		/// <summary>
		/// If this is true then the current letter has files attached (, files which the remote peer may download). READ-ONLY.
		/// </summary>
		public bool HasAttachedFiles
		{
			get { return AttachedFiles.Count > 0; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Locator _RecipientUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the recipient of the current object. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( RecipientUpnFingerprintFilterIdentifier )]
		public Locator RecipientUpnFingerprint
		{
			get { return _RecipientUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RecipientUpnFingerprint" );

				_RecipientUpnFingerprint = value;
			}
		}

		private AttachedFiles _AttachedFiles = new AttachedFiles();
		/// <summary>
		/// Files that the remote peer may download. The backing field is initialized by default. This may be empty, but not
		/// null. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The attached files are meant to be downloaded by the recipient, not uploaded by the sender. This makes it
		/// easier to code the transfer process, especially the resume feature. However, this also means that files can't be sent
		/// from inside a company whose firewall blocks inbound traffic (because the recipient downloads the files from the
		/// sender).</para>
		/// </remarks>
		[SerializeAttribute]
		public AttachedFiles AttachedFiles
		{
			get { return _AttachedFiles; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "AttachedFiles" );

				_AttachedFiles = value;
			}
		}

		private bool _HasAttachedMediaFiles;
		/// <summary>
		/// If this is true then some of the attached files are media files. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>Since this is a new property, it must have "MustBeDeserialized = false".</para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false, DontHashSerializeValue = false )]
		[SpeckleFilterAttribute( HasAttachedMediaFilesFilterIdentifier )]
		public bool HasAttachedMediaFiles
		{
			get { return _HasAttachedMediaFiles; }
			protected set { _HasAttachedMediaFiles = value; }
		}

		private bool _HasImagePreview;
		/// <summary>
		/// If this is true then the current letter has an image preview, that is, it has an attached image file. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>Since this is a new property, it must have "MustBeDeserialized = false".</para>
		/// 
		/// <para>Using "DontHashSerializeValue = false" (like for <see cref="HasAttachedMediaFiles"/>) would mean that a letter
		/// which has an image preview and which is sent from the current version of Vela would not have its signature verified
		/// on older versions of Vela, so it could not be received. Therefore, it's more practical to not sign / authenticate
		/// this property together with the document, so that older versions of Vela can receive such letters.</para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false, HashSerialization = HashSerializations.Never )]
		[SpeckleFilterAttribute( HasImagePreviewFilterIdentifier )]
		public bool HasImagePreview
		{
			get { return _HasImagePreview; }
			protected set { _HasImagePreview = value; }
		}

		#endregion Public serializable properties
	}
}
