////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// Contract for a post which can be sent to (or retrieved by) other entities in order to communicate.
	/// </summary>
	public abstract class Post : SignedDocument, IInitiatorUpnFingerprintContainer
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="InitiatorUpnFingerprint"/> property.
		/// </summary>
		public const string InitiatorUpnFingerprintFilterIdentifier = "Post.InitiatorUpnFingerprint#4";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="MessageSummary"/> property.
		/// </summary>
		public const string MessageSummaryFilterIdentifier = "Post.MessageSummary#5";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="EmoticonSymbol"/> property.
		/// </summary>
		public const string EmoticonSymbolFilterIdentifier = "Post.EmoticonSymbol#6";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="Flag"/> property.
		/// </summary>
		public const string FlagFilterIdentifier = "Post.Flag#7";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="IsImportant"/> property.
		/// </summary>
		[Obsolete( "Use the \"DeskNote\" emoticon instead.", true )]
		public const string IsImportantFilterIdentifier = "Post.IsImportant#8";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="IsTransient"/> property.
		/// </summary>
		public const string IsTransientFilterIdentifier = "Post.IsTransient#9";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Post()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="message">The message of the post.</param>
		/// <param name="emoticonSymbol">The emoticon symbol of the post. This may be null or empty.</param>
		/// <param name="isTransient">If this is true then the current letter is transient.</param>
		public Post( Locator instanceLocator, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg
				, IHashSerializer signatureHashSerializer, Locator initiatorUpnFingerprint, string message
				, string emoticonSymbol, bool isTransient )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer )
		{
			InitiatorUpnFingerprint = initiatorUpnFingerprint;
			Message = message;
			EmoticonSymbol = emoticonSymbol;
			IsTransient = isTransient;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// Summary of the message of the post.
		/// </summary>
		[SpeckleFilterAttribute( MessageSummaryFilterIdentifier )]
		public string MessageSummary
		{
			get { return PeerDocumentsUtils.GetPostMessageSummary( Message ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Locator _InitiatorUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the initiator of the current object. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( InitiatorUpnFingerprintFilterIdentifier )]
		public Locator InitiatorUpnFingerprint
		{
			get { return _InitiatorUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "InitiatorUpnFingerprint" );

				_InitiatorUpnFingerprint = value;
			}
		}

		private string _Message;
		/// <summary>
		/// The message of the post. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string Message
		{
			get { return _Message; }
			protected set
			{
				ValidationUtils.Text( value, 1, PeerDocumentsConsts.PostMessageSizeMax, TextValidations.None, "Message" );

				_Message = value;
			}
		}

		private string _EmoticonSymbol;
		/// <summary>
		/// The emoticon symbol of the post. This may be null or empty. READ-ONLY.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		[SpeckleFilterAttribute( EmoticonSymbolFilterIdentifier )]
		public string EmoticonSymbol
		{
			get { return _EmoticonSymbol; }
			protected set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.EmoticonSymbol, "EmoticonSymbol" );

				_EmoticonSymbol = value;
			}
		}

		private string _Flag = PostFlags.Main.ToString();
		/// <summary>
		/// The flag of the post. This is not signed / authenticated together with the document because it's meant to be set by
		/// both the sender and by the recipient. The backing field is initialized by default.
		/// </summary>
		[SerializeAttribute( HashSerialization = HashSerializations.NotInRoot )]
		[SpeckleFilterAttribute( FlagFilterIdentifier )]
		public string Flag
		{
			get { return _Flag; }
			set
			{
				ValidationUtils.TextHasContent( value, "Flag" );

				_Flag = value;
			}
		}

		private bool _IsImportant;
		/// <summary>
		/// This is kept in order to preserve the validity of the existing signatures. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[Obsolete( "Use the \"DeskNote\" emoticon instead.", true )]
		public bool IsImportant
		{
			get { return _IsImportant; }
			protected set { _IsImportant = value; }
		}

		private bool _IsTransient;
		/// <summary>
		/// If this is true then the current letter is transient. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>A transient letter must not be signed when it's sent, its signature must not be verified when it's received,
		/// and it must be deleted from the storehouse when the application closes.</para>
		/// </remarks>
		[SerializeAttribute]
		[SpeckleFilterAttribute( IsTransientFilterIdentifier )]
		public bool IsTransient
		{
			get { return _IsTransient; }
			protected set { _IsTransient = value; }
		}

		#endregion Public serializable properties

		#region Public static methods

		/// <summary>
		/// Checks if the specified flags are equal.
		/// </summary>
		/// <param name="flag1">Flag to compare. This may be null or empty.</param>
		/// <param name="flag2">Flag to compare. This may be null or empty.</param>
		/// <returns>Returns true if the specified flags are equal. If either specified flag is null or empty, false is returned.
		/// </returns>
		public static bool EqualFlags( string flag1, string flag2 )
		{
			// Because the speckle filters are binary compared, this comparison must also be case-sensitive.
			return TextUtils.HasContent( flag1 ) && TextUtils.HasContent( flag2 ) && (flag1 == flag2);
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Checks if the current letter has the specified flag.
		/// </summary>
		/// <param name="flag">Flag to check if the current letter has.</param>
		/// <returns>Returns true if the current letter has the specified flag.</returns>
		public bool HasFlag( string flag )
		{
			return EqualFlags( Flag, flag );
		}

		#endregion Public methods
	}
}
