////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// Presenter. Presenters are published by some entities so that other entities can use them for interaction.
	/// </summary>
	/// <remarks>
	/// <para>A presenter is uniquely associated to a creator peer, and is normally stored together with a tally.
	/// </para>
	/// 
	/// <para>Published presenters must be self-signed. Applications and action services must verify the self-signature.</para>
	/// 
	/// <para>A presenter may contain only public information.</para>
	/// </remarks>
	public class Presenter : SignedDocument, INamedEntity, IPublicProcessingIdentitySheaf, ICreatorUpnFingerprintContainer
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="CreatorUpnFingerprint"/> property.
		/// </summary>
		public const string CreatorUpnFingerprintFilterIdentifier = ProcessingIdentityUpnFingerprintFilterIdentifier;

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="EntityName"/> property.
		/// </summary>
		public const string EntityNameFilterIdentifier = "Presenter.EntityName#4";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Presenter()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="processingIdentity">The identity publisher used for processing.</param>
		/// <param name="appendix">Appendix.</param>
		/// <param name="photo">Photo. This may be null.</param>
		/// <remarks>
		/// <para>The initiator UPN fingerprint is set with the UPN fingerprint of the specified identity.</para>
		/// </remarks>
		public Presenter( Locator instanceLocator, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg
				, IHashSerializer signatureHashSerializer, IIdentityPublisher processingIdentity, Appendix appendix
				, Photo photo )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer )
		{
			ProcessingIdentity = processingIdentity;
			Appendix = appendix;
			Photo = photo;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The UPN of the creator peer. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public Upn CreatorUpn
		{
			get { return ProcessingIdentity.Upn; }
		}

		/// <summary>
		/// The UPN fingerprint of the creator peer. READ-ONLY.
		/// </summary>
		[SpeckleFilterAttribute( CreatorUpnFingerprintFilterIdentifier )]
		public Locator CreatorUpnFingerprint
		{
			get { return CreatorUpn.Fingerprint; }
		}

		/// <summary>
		/// The entity name. READ-ONLY.
		/// </summary>
		[SpeckleFilterAttribute( EntityNameFilterIdentifier )]
		public string EntityName
		{
			get { return Appendix.EntityName; }
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( base.ExpirationTime, ProcessingIdentity.ExpirationTime, Appendix.ExpirationTime ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private IIdentityPublisher _ProcessingIdentity;
		/// <summary>
		/// The identity publisher used for processing. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public IIdentityPublisher ProcessingIdentity
		{
			get { return _ProcessingIdentity; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ProcessingIdentity" );

				_ProcessingIdentity = value;
			}
		}

		private Appendix _Appendix;
		/// <summary>
		/// Appendix. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Appendix Appendix
		{
			get { return _Appendix; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Appendix" );

				_Appendix = value;
			}
		}

		private Photo _Photo;
		/// <summary>
		/// Photo. This may be null.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public Photo Photo
		{
			get { return _Photo; }
			set { _Photo = value; }
		}

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Verifies the self-signature of the current presenter.
		/// </summary>
		public void VerifySelfSignature()
		{
			VerifySignature( ProcessingIdentity );
		}

		/// <summary>
		/// Collects the identities from the current document.
		/// </summary>
		/// <param name="identityPublishers">List where to collect the identity publishers. The dictionary key is the instance
		/// locator of the current document. This may be empty, but not null.</param>
		/// <param name="identities">List where to collect the identities. The dictionary key is the instance locator of the
		/// current document. This may be empty, but not null.</param>
		public void CollectIdentities( Dictionary<Locator, IIdentityPublisher> identityPublishers
			, Dictionary<Locator, PublicIdentity> identities )
		{
			ValidationUtils.ObjectHasContent( identityPublishers, "identityPublishers" );
			ValidationUtils.ObjectHasContent( identities, "identities" );

			identityPublishers.Add( InstanceLocator, ProcessingIdentity );
		}

		/// <summary>
		/// Generates the automated UPNs associated to the current document.
		/// </summary>
		/// <returns>Returns the generated UPNs. This may be null or empty.</returns>
		/// <remarks>
		/// <para>The automated UPNs are not accessible to the users.</para>
		/// </remarks>
		public Upns GenerateAutomatedUpns()
		{
			return Appendix.GenerateAutomatedUpns( CreatorUpn.UpnAlg );
		}

		/// <summary>
		/// Generates all the UPNs (and their lazy loaded properties) associated to the current document.
		/// </summary>
		/// <returns>Returns the generated UPNs.</returns>
		public Upns GenerateUpns()
		{
			Upns upns = GenerateAutomatedUpns();

			if( upns == null ) upns = new Upns();
			upns.Add( CreatorUpn );

			foreach( Upn upn in upns ) upn.GenerateAllLazyLoaded();

			return upns;
		}

		#endregion Public methods
	}
}
