////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// Tally.
	/// </summary>
	/// <remarks>
	/// <para>A tally is uniquely associated to a remote peer, and is normally stored together with a presenter and an account.
	/// </para>
	/// </remarks>
	public class Tally : Sheaf, IRemoteUpnFingerprintContainer, ILocallyNamedPublicProcessingIdentitySheaf
		, IAlternativePeerAddressContainer
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="RemoteUpnFingerprint"/> property.
		/// </summary>
		public const string RemoteUpnFingerprintFilterIdentifier = ProcessingIdentityUpnFingerprintFilterIdentifier;

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="EntityName"/> property.
		/// </summary>
		public const string EntityNameFilterIdentifier = "Tally.EntityName#4";

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="Approval"/> property.
		/// </summary>
		public const string ApprovalFilterIdentifier = "Tally.Approval#5";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Tally()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="processingIdentity">The processing identity publisher from the associated presenter.</param>
		/// <param name="actionRequestValidity">The timespan (in minutes) which must be added to the action request's
		/// "CreationTime" property in order to obtain the value of the "ExpirationTime" property. An action request expires at
		/// the time specified by its "ExpirationTime" property. This must be at least 1.</param>
		/// <param name="entityName">The name of the identified entity.</param>
		/// <remarks>
		/// <para>The initiator UPN fingerprint is set with the UPN fingerprint of the specified identity.</para>
		/// </remarks>
		public Tally( Locator instanceLocator, IIdentityPublisher processingIdentity, int actionRequestValidity
				, string entityName )
			: base( instanceLocator )
		{
			ProcessingIdentity = processingIdentity;
			ActionRequestValidity = actionRequestValidity;
			EntityName = entityName;

			AlternativePeerAddress = new AlternativePeerAddress();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="processingIdentity">The processing identity publisher from the associated presenter.
		/// </param>
		/// <param name="actionRequestValidity">The timespan (in minutes) which must be added to the action request's
		/// "CreationTime" property in order to obtain the value of the "ExpirationTime" property. An action request expires at
		/// the time specified by its "ExpirationTime" property. This must be at least 1.</param>
		/// <param name="entityName">The name of the identified entity.</param>
		/// <param name="approval">The approval of the tally.</param>
		public Tally( Locator instanceLocator, IIdentityPublisher processingIdentity, int actionRequestValidity
				, string entityName, TallyApprovals approval )
			: this( instanceLocator, processingIdentity, actionRequestValidity, entityName )
		{
			Approval = approval;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The UPN of the remote peer. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public Upn RemoteUpn
		{
			get { return ProcessingIdentity.Upn; }
		}

		/// <summary>
		/// The UPN fingerprint of the remote peer. READ-ONLY.
		/// </summary>
		[SpeckleFilterAttribute( RemoteUpnFingerprintFilterIdentifier )]
		public Locator RemoteUpnFingerprint
		{
			get { return RemoteUpn.Fingerprint; }
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( base.ExpirationTime, ProcessingIdentity.ExpirationTime ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private IIdentityPublisher _ProcessingIdentity;
		/// <summary>
		/// The processing identity publisher copied from the associated presenter. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is a copy of the identity publisher from the presenter. This exists in order eliminate the need to
		/// load the presenter every time an action request is processed.</para>
		/// </remarks>
		[SerializeAttribute]
		public IIdentityPublisher ProcessingIdentity
		{
			get { return _ProcessingIdentity; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ProcessingIdentity" );

				_ProcessingIdentity = value;
			}
		}

		private int _ActionRequestValidity = PeerDocumentsConsts.ActionRequestValidity;
		/// <summary>		
		/// The timespan (in minutes) which must be added to the action request's "CreationTime" property in order to
		/// obtain the value of the "ExpirationTime" property. An action request expires at the time specified by its
		/// "ExpirationTime" property. This must be at least 1.
		/// </summary>
		/// <remarks>
		/// <para>Through this property, an action service instructs client applications to enforce the following condition:
		/// "ExpirationTime" has to be smaller than (or equal with) "CreationTime + ActionRequestValidity".</para>
		/// 
		/// <para>This property exists so that the action service may delete old action requests.</para>
		/// 
		/// <para>This is a copy of the <see cref="Appendix.ActionRequestValidity"/>. This exists in order eliminate the need to
		/// load the presenter every time an action request is processed.</para>
		/// </remarks>
		[SerializeAttribute]
		public int ActionRequestValidity
		{
			get { return _ActionRequestValidity; }
			set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "ActionRequestValidity" );

				_ActionRequestValidity = value;
			}
		}

		private string _EntityName;
		/// <summary>
		/// The name of the identified entity.
		/// </summary>
		/// <remarks>
		/// <para>Initially, this is a copy of the entity name from the appendix, copy made when the associated account is
		/// created. This exists in order eliminate the need to load the presenter every time the name has to be displayed,
		/// and also to allow the local peer to change the (displayed) name.</para>
		/// </remarks>
		[SerializeAttribute]
		[SpeckleFilterAttribute( EntityNameFilterIdentifier )]
		public string EntityName
		{
			get { return _EntityName; }
			set
			{
				ValidationUtils.Text( value, -1, -1, TextValidations.EntityName, "EntityName" );

				_EntityName = value;
			}
		}

		private TallyApprovals _Approval = PeerDocumentsConsts.CreateAccountApprovalMin;
		/// <summary>
		/// The approval of the tally.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( ApprovalFilterIdentifier )]
		public TallyApprovals Approval
		{
			get { return _Approval; }
			set { _Approval = value; }
		}

		private string _Note;
		/// <summary>
		/// Note. This may be null or empty.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string Note
		{
			get { return _Note; }
			set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.LongNote, "Note" );

				_Note = value;
			}
		}

		private AlternativePeerAddress _AlternativePeerAddress;
		/// <summary>
		/// The alternative peer address (of the remote peer).
		/// </summary>
		[SerializeAttribute]
		public AlternativePeerAddress AlternativePeerAddress
		{
			get { return _AlternativePeerAddress; }
			set
			{
				ValidationUtils.ObjectHasContent( value, "AlternativePeerAddress" );

				_AlternativePeerAddress = value;
			}
		}

		private bool _UseTransientInteraction;
		/// <summary>
		/// If this is true then the interaction between the local and the remote peers must be as transient as possible.
		/// </summary>
		/// <remarks>
		/// <para>See <see cref="Post.IsTransient"/> for details.</para>
		/// </remarks>
		[SerializeAttribute]
		public bool UseTransientInteraction
		{
			get { return _UseTransientInteraction; }
			set { _UseTransientInteraction = value; }
		}

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Collects the identities from the current document.
		/// </summary>
		/// <param name="identityPublishers">List where to collect the identity publishers. The dictionary key is the instance
		/// locator of the current document. This may be empty, but not null.</param>
		/// <param name="identities">List where to collect the identities. The dictionary key is the instance locator of the
		/// current document. This may be empty, but not null.</param>
		public void CollectIdentities( Dictionary<Locator, IIdentityPublisher> identityPublishers
			, Dictionary<Locator, PublicIdentity> identities )
		{
			ValidationUtils.ObjectHasContent( identityPublishers, "identityPublishers" );
			ValidationUtils.ObjectHasContent( identities, "identities" );

			identityPublishers.Add( InstanceLocator, ProcessingIdentity );
		}

		/// <summary>
		/// Returns true if the current peer is a "known" peer. The approval has to be at least
		/// <see cref="PeerDocumentsConsts.KnownPeerTallyApprovalMin"/>.
		/// </summary>
		/// <returns>Returns true if the current peer is a "known" peer.</returns>
		public bool IsKnownPeer()
		{
			return PeerDocumentsUtils.IsKnownPeer( Approval );
		}

		#endregion Public methods
	}
}
