////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using Vela.Cloning;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.PeerDocuments
{
	/// <summary>
	/// Vault.
	/// </summary>
	/// <remarks>
	/// <para>A vault is uniquely associated to a local peer.</para>
	/// </remarks>
	public class Vault : Sheaf, ILocallyNamedPrivateProcessingIdentitySheaf, ILocalUpnFingerprintContainer
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="LocalUpnFingerprint"/> property.
		/// </summary>
		public const string LocalUpnFingerprintFilterIdentifier = ProcessingIdentityUpnFingerprintFilterIdentifier;

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="EntityName"/> property.
		/// </summary>
		public const string EntityNameFilterIdentifier = "Vault.EntityName#4";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected Vault()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="rootSecretSize">The size (in bytes) that the root secret must have. If this is 0 or a negative number,
		/// the default <see cref="CoreDocumentsConsts.SymmEncryptionKeySizeDef"/> is used.</param>
		/// <param name="processingIdentity">The private identity used for processing.</param>
		/// <param name="appendix">The appendix.</param>
		/// <remarks>
		/// <para>The initiator UPN fingerprint is set with the UPN fingerprint of the specified identity.</para>
		/// </remarks>
		public Vault( Locator instanceLocator, int rootSecretSize, PrivateIdentity processingIdentity, Appendix appendix )
			: base( instanceLocator )
		{
			if( rootSecretSize <= 0 ) rootSecretSize = CoreDocumentsConsts.SymmEncryptionKeySizeDef;
			RootSecret = EntropyPool.GetBytes( rootSecretSize );

			ProcessingIdentity = processingIdentity;
			Appendix = appendix;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The UPN of the local peer. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public Upn LocalUpn
		{
			get { return ProcessingIdentity.Upn; }
		}

		/// <summary>
		/// The UPN fingerprint of the local peer. READ-ONLY.
		/// </summary>
		[SpeckleFilterAttribute( LocalUpnFingerprintFilterIdentifier )]
		public Locator LocalUpnFingerprint
		{
			get { return LocalUpn.Fingerprint; }
		}

		/// <summary>
		/// The entity name. READ-ONLY.
		/// </summary>
		[SpeckleFilterAttribute( EntityNameFilterIdentifier )]
		public string EntityName
		{
			get { return Appendix.EntityName; }
		}

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get
			{
				return KernelUtils.MinTime( base.ExpirationTime, ProcessingIdentity.ExpirationTime, Appendix.ExpirationTime );
			}
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private byte[] _RootSecret;
		/// <summary>
		/// The root secret, randomly generated, which is used to generate other secrets associated to the current vault.
		/// READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] RootSecret
		{
			get { return _RootSecret; }
			protected set
			{
				ValidationUtils.Array( value, CoreDocumentsConsts.SymmEncryptionKeySizeMin
					, CoreDocumentsConsts.SymmEncryptionKeySizeMax, "RootSecret" );

				_RootSecret = value;
			}
		}

		private PrivateIdentity _ProcessingIdentity;
		/// <summary>
		/// The private identity used for processing actions. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This must be unique for a peer.</para>
		/// 
		/// <para>Action requests are dispatched to the appropriate service based on the equality between the UPN fingerprint of
		/// this identity and the UPN fingerprint of one of the recipient session keys of the action requests.</para>
		/// </remarks>
		[SerializeAttribute]
		public PrivateIdentity ProcessingIdentity
		{
			get { return _ProcessingIdentity; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ProcessingIdentity" );

				_ProcessingIdentity = value;
			}
		}

		private Appendix _Appendix;
		/// <summary>
		/// Appendix. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Appendix Appendix
		{
			get { return _Appendix; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Appendix" );

				_Appendix = value;
			}
		}

		private string _PrivateNote;
		/// <summary>
		/// Private note. This may be null or empty.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string PrivateNote
		{
			get { return _PrivateNote; }
			set
			{
				ValidationUtils.Text( value, -2, -1, TextValidations.LongNote, "PrivateNote" );

				_PrivateNote = value;
			}
		}

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Collects the identities from the current document.
		/// </summary>
		/// <param name="identityPublishers">List where to collect the identity publishers. The dictionary key is the instance
		/// locator of the current document. This may be empty, but not null.</param>
		/// <param name="identities">List where to collect the identities. The dictionary key is the instance locator of the
		/// current document. This may be empty, but not null.</param>
		public void CollectIdentities( Dictionary<Locator, IIdentityPublisher> identityPublishers
			, Dictionary<Locator, PublicIdentity> identities )
		{
			ValidationUtils.ObjectHasContent( identityPublishers, "identityPublishers" );
			ValidationUtils.ObjectHasContent( identities, "identities" );

			identities.Add( InstanceLocator, ProcessingIdentity );
		}

		/// <summary>
		/// Collects the private identities from the current document.
		/// </summary>
		/// <param name="identities">List where to collect the identities. The dictionary key is the instance locator of the
		/// current document. This may be empty, but not null.</param>
		public void CollectPrivateIdentities( Dictionary<Locator, PrivateIdentity> identities )
		{
			ValidationUtils.ObjectHasContent( identities, "identities" );

			identities.Add( InstanceLocator, ProcessingIdentity );
		}

		/// <summary>
		/// Generates the automated UPNs associated to the current document.
		/// </summary>
		/// <returns>Returns the generated UPNs. This may be null or empty.</returns>
		/// <remarks>
		/// <para>The automated UPNs are not accessible to the users.</para>
		/// </remarks>
		public Upns GenerateAutomatedUpns()
		{
			return Appendix.GenerateAutomatedUpns( ProcessingIdentity.UpnAlg );
		}

		/// <summary>
		/// Generates all the UPNs (and their lazy loaded properties) associated to the current document.
		/// </summary>
		/// <returns>Returns the generated UPNs.</returns>
		public Upns GenerateUpns()
		{
			Upns upns = GenerateAutomatedUpns();

			if( upns == null ) upns = new Upns();
			upns.Add( LocalUpn );

			foreach( Upn upn in upns ) upn.GenerateAllLazyLoaded();

			return upns;
		}

		/// <summary>
		/// Gets the solitary identity publisher of the current vault.
		/// </summary>
		/// <returns>Returns a solitary identity publisher.</returns>
		public SolitaryIdentityPublisher GetSolitaryIdentity()
		{
			return new SolitaryIdentityPublisher( ProcessingIdentity.GetPublicIdentity() );
		}

		/// <summary>
		/// Gets the presenter of the current vault.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="photo">The photo to include in the presenter. This may be null.</param>
		/// <param name="includeIntimateInfo">If this is true then then intimate information is included in the presenter.
		/// </param>
		/// <param name="hashSerializer">Hash serializer for signing.</param>
		/// <returns>Returns the presenter of the current vault.</returns>
		public virtual Presenter GetPresenter( Locator instanceLocator, MacAlg macAlg
			, IAsymmSigningPaddingAlg asymmSigningPaddingAlg, Photo photo, bool includeIntimateInfo
			, IHashSerializer hashSerializer )
		{
			ValidationUtils.ObjectHasContent( hashSerializer, "hashSerializer" );

			Appendix appendix;

			// Include the intimate information only if allowed.
			if( includeIntimateInfo )
			{
				appendix = Appendix;
			}
			else
			{
				// Clone the appendix because the intimate information is changed. Because the vault edit dialog creates
				// instances (for the following types) even if no information is set in them, if null were used here instead of
				// new instances, the remote peer would see a difference (due to the binding of scalars, which have default
				// values) and hence would know the approval.
				appendix = Cloner.Deep( Appendix );
				appendix.VirtualIdentity = new PersonVirtualIdentity();
				appendix.PhysicalAddress = new PhysicalAddress();
			}

			SolitaryIdentityPublisher identityPublisher = CoreDocumentsUtils.GetSolitaryIdentity( ProcessingIdentity );

			Presenter presenter = new Presenter( instanceLocator, macAlg, asymmSigningPaddingAlg, hashSerializer
				, identityPublisher, appendix, photo );

			// Sign the presenter.
			presenter.Sign( ProcessingIdentity, null );

			return presenter;
		}

		#endregion Public methods
	}
}
