////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Net.PeerToPeer;
using System.Reflection;
using Vela.Kernel;

namespace Vela.Pnrp
{
	/// <summary>
	/// Network registration for a list of UPN classifiers.
	/// </summary>
	public class ClassifiersPeerRegistration : PeerRegistration
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="parentLocator">The instance locator of the registered parent document.</param>
		/// <param name="useOnlyDebugLogs">If this is true then all the logs from this instance are debug ones.</param>
		/// <param name="classifiersCount">The number of classifiers to register.</param>
		/// <param name="classifiers">The classifiers to register.</param>
		public ClassifiersPeerRegistration( Locator parentLocator, bool useOnlyDebugLogs, int classifiersCount
				, Strings classifiers )
			: base( parentLocator, useOnlyDebugLogs )
		{
			ValidationUtils.ObjectHasContent( classifiers, "classifiers" );

			ClassifiersCount = classifiersCount;
			Classifiers = classifiers;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The number of classifiers to register.
		/// </summary>
		private int ClassifiersCount;

		/// <summary>
		/// The classifiers to register.
		/// </summary>
		private Strings Classifiers;

		#endregion Private fields

		#region Public methods

		/// <summary>
		/// Starts the registration.
		/// </summary>
		/// <remarks>
		/// <para>If not all the UPN classifiers (/ UPNs) are successfully registered then all the registration which did
		/// succeed are unregistered. This ensures that either all the UPN classifiers (/ UPNs) are registered or none is.
		/// </para>
		/// </remarks>
		public override void Start()
		{
			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				PeerClassifierRegistrations = new List<PeerNameRegistration>();

				// Register all classifiers.
				int count = 0;
				foreach( string classifier in Classifiers )
				{
					if( count >= ClassifiersCount ) break;

					LogRecordFields fields = new LogRecordFields();
					fields.Add( new LogRecordField( "UPN classifier", classifier, false, LogRecordFieldPriorities.Middle ) );

					LogManager upnLogManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod(), fields );

					try
					{
						PeerName peerName = new PeerName( classifier, PeerNameType.Unsecured );

						PeerNameRegistration registration = new PeerNameRegistration( peerName, RegistrationPort
							, PnrpConsts.Cloud );
						registration.UseAutoEndPointSelection = true;

						registration.Start();

						PeerClassifierRegistrations.Add( registration );

						string message = string.Format( "The peer registration for the classifier \"{0}\" was started."
							, classifier );
						if( UseOnlyDebugLogs ) LogManager.DebugAdd( message, null );
						else LogManager.Add( message );

						count++;
					}
					finally
					{
						LogManager.DebugEndMethod( upnLogManager );
					}
				}
			}
			catch( Exception ex )
			{
				Stop();

				string message = string.Format( "Failed to start the peer registration for the classifiers." );
				throw new ValidationException( message, ex );
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		#endregion Public methods
	}
}
