////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Net.PeerToPeer;
using Vela.Kernel;

namespace Vela.Pnrp
{
	/// <summary>
	/// Network registration for a peer.
	/// </summary>
	public abstract class PeerRegistration
	{
		#region Public constants

		/// <summary>
		/// The port for the registration. Use 0 for automatic selection.
		/// </summary>
		public const int RegistrationPort = 0;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="parentLocator">The instance locator of the registered parent document.</param>
		/// <param name="useOnlyDebugLogs">If this is true then all the logs from this instance are debug ones.</param>
		public PeerRegistration( Locator parentLocator, bool useOnlyDebugLogs )
		{
			ParentLocator = parentLocator;
			UseOnlyDebugLogs = useOnlyDebugLogs;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The underlying UPN classifier registrations.
		/// </summary>
		protected List<PeerNameRegistration> PeerClassifierRegistrations;

		/// <summary>
		/// If this is true then all the logs from this instance are debug ones.
		/// </summary>
		protected bool UseOnlyDebugLogs;

		#endregion Private fields

		#region Public non-serializable properties

		private Locator _ParentLocator;
		/// <summary>
		/// The instance locator of the registered parent document. READ-ONLY.
		/// </summary>
		public Locator ParentLocator
		{
			get { return _ParentLocator; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ParentLocator" );

				_ParentLocator = value;
			}
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Returns true if the peer is registered on the network.
		/// </summary>
		/// <remarks>
		/// <para>If a single UPN classifier (/ UPN) is not registered on the network, the peer is considered not registered.
		/// </para>
		/// </remarks>
		protected bool IsRegistered()
		{
			foreach( PeerNameRegistration registration in PeerClassifierRegistrations )
			{
				if( !registration.IsRegistered() ) return false;
			}

			return true;
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Returns true if all the UPN classifiers (/ UPNs) are registered.
		/// </summary>
		/// <returns>Returns true if all the UPN classifiers (/ UPNs) are registered.</returns>
		public bool AreAllRegistered()
		{
			foreach( PeerNameRegistration peerClassifierRegistration in PeerClassifierRegistrations )
			{
				if( !peerClassifierRegistration.IsRegistered() ) return false;
			}

			return true;
		}

		/// <summary>
		/// Starts the registration.
		/// </summary>
		/// <remarks>
		/// <para>If not all the UPN classifiers (/ UPNs) are successfully registered then all the registration which did
		/// succeed are unregistered. This ensures that either all the UPN classifiers (/ UPNs) are registered or none is.
		/// </para>
		/// </remarks>
		public abstract void Start();

		/// <summary>
		/// Stops the registration.
		/// </summary>
		public void Stop()
		{
			if( PeerClassifierRegistrations == null ) return;

			Exception rethrowAbortEx = null;

			try
			{
				foreach( PeerNameRegistration registration in PeerClassifierRegistrations )
				{
					try
					{
						if( (registration != null) && registration.IsRegistered() )
						{
							string peerClassifier = registration.PeerName.Classifier;

							registration.Stop();

							string message = string.Format( "The peer registration for the classifier \"{0}\" of the document"
								+ " with the locator \"{1}\" was stopped.", peerClassifier, ParentLocator );
							if( UseOnlyDebugLogs ) LogManager.DebugAdd( message, null );
							else LogManager.Add( message );
						}
					}
					catch( Exception ex )
					{
						// The loop must continue (if possible).

						if( ExceptionUtils.IsCritical( ex ) ) throw ExceptionUtils.PreserveStackTrace( ex );

						if( ExceptionUtils.IsAbort( ex ) ) rethrowAbortEx = ex;

						LogManager.Add( ex );
					}
				}
			}
			finally
			{
				if( rethrowAbortEx != null ) throw ExceptionUtils.PreserveStackTrace( rethrowAbortEx );
			}
		}

		#endregion Public methods
	}
}
