////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Diagnostics;
using System.Net.PeerToPeer;
using Vela.Kernel;

namespace Vela.Pnrp
{
	/// <summary>
	/// Utilities for working with the PNRP.
	/// </summary>
	public static class PnrpUtils
	{
		#region Private constants

		/// <summary>
		/// The category of the PNRP performance counters.
		/// </summary>
		private const string PnrpPerformanceCounterCategory = "Peer Name Resolution Protocol";

		/// <summary>
		/// The cloud instance of the PNRP performance counters.
		/// </summary>
		private const string PnrpPerformanceCounterInstance = "Global_";

		/// <summary>
		/// The name of the PNRP performance counters for the number of cache entries.
		/// </summary>
		private const string CacheEntriesCountPnrpPerformanceCounterName = "Cache Entry";

		/// <summary>
		/// The name of the PNRP performance counters for the estimated cloud size.
		/// </summary>
		private const string EstimatedCloudSizePnrpPerformanceCounterName = "Estimated cloud size";

		#endregion Private constants

		#region Public constants

		/// <summary>
		/// The minimum value that the number of cache entries must have in order to consider that the PNRP is able to resolve
		/// UPNs (over the Internet).
		/// </summary>
		public const int CacheEntriesCountThreshold = 40;

		/// <summary>
		/// The minimum value that the number of cache entries must have in order to consider that the PNRP is able to resolve
		/// UPNs (over the Internet).
		/// </summary>
		public const int EstimatedCloudSizeThreshold = 100;

		#endregion Public constants

		#region Private static methods

		/// <summary>
		/// Gets the value of the <see cref="CacheEntriesCountPnrpPerformanceCounterName"/> PNRP performance counter.
		/// </summary>
		/// <returns>Returns the value of the performance counter.</returns>
		private static Int64 GetCacheEntriesCountPnrpPerformanceCounter()
		{
			if( !PerformanceCounterCategory.InstanceExists( PnrpPerformanceCounterInstance, PnrpPerformanceCounterCategory ) )
				return 0;

			if( !PerformanceCounterCategory.CounterExists( CacheEntriesCountPnrpPerformanceCounterName
				, PnrpPerformanceCounterCategory ) )
			{
				return 0;
			}

			using( PerformanceCounter pc = new PerformanceCounter( PnrpPerformanceCounterCategory
				, CacheEntriesCountPnrpPerformanceCounterName, PnrpPerformanceCounterInstance, true ) )
			{
				return pc.RawValue;
			}
		}

		/// <summary>
		/// Gets the value of the <see cref="EstimatedCloudSizePnrpPerformanceCounterName"/> PNRP performance counter.
		/// </summary>
		/// <returns>Returns the value of the performance counter.</returns>
		private static Int64 GetEstimatedCloudSizePnrpPerformanceCounter()
		{
			if( !PerformanceCounterCategory.InstanceExists( PnrpPerformanceCounterInstance, PnrpPerformanceCounterCategory ) )
				return 0;

			if( !PerformanceCounterCategory.CounterExists( EstimatedCloudSizePnrpPerformanceCounterName
				, PnrpPerformanceCounterCategory ) )
			{
				return 0;
			}

			using( PerformanceCounter pc = new PerformanceCounter( PnrpPerformanceCounterCategory
				, EstimatedCloudSizePnrpPerformanceCounterName, PnrpPerformanceCounterInstance, true ) )
			{
				return pc.RawValue;
			}
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Gets true if the global PNRP cloud is available (and so the local peer is visible on the Internet).
		/// </summary>
		/// <returns>Returns true if the global PNRP cloud is available.</returns>
		public static bool IsGlobalCloudAvailable()
		{
			try
			{
				// Don't access the "Global" cloud directly because an exception is thrown if it doesn't exist.

				CloudCollection clouds = Cloud.GetAvailableClouds();
				if( ListUtils.NoContent( clouds ) )
					return false;

				foreach( Cloud cloud in clouds )
				{
					if( cloud.Scope == PnrpScope.Global )
					{
						try
						{
							Int64 cacheEntriesCount = GetCacheEntriesCountPnrpPerformanceCounter();
							Int64 estimatedCloudSize = GetEstimatedCloudSizePnrpPerformanceCounter();

							return (cacheEntriesCount >= CacheEntriesCountThreshold)
								&& (estimatedCloudSize >= EstimatedCloudSizeThreshold);
						}
						catch( InvalidOperationException )
						{
							// The performance counter is unavailable for the moment.
							return false;
						}
					}
				}
			}
			catch( PeerToPeerException )
			{
				// Nothing to do.
			}

			return false;
		}

		#endregion Public static methods
	}
}
