////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Net.PeerToPeer;
using System.Reflection;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.PeerDocuments;

namespace Vela.Pnrp
{
	/// <summary>
	/// Network registration for a peer vault.
	/// </summary>
	public class VaultPeerRegistration : PeerRegistration
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="vault">The vault for which to create the registration.</param>
		public VaultPeerRegistration( Vault vault )
			: base( vault.InstanceLocator, false )
		{
			ValidationUtils.ObjectHasContent( vault, "vault" );

			Vault = vault;

			Upns = Vault.GenerateUpns();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The vault for which to create the registration.
		/// </summary>
		private Vault Vault;

		/// <summary>
		/// The UPNs associated to <see cref="Vault"/>.
		/// </summary>
		private Upns Upns;

		#endregion Private fields

		#region Public methods

		/// <summary>
		/// Starts the registration.
		/// </summary>
		/// <remarks>
		/// <para>Each UPN of the vault <see cref="Vault"/> is registered as a UPN classifier.</para>
		/// 
		/// <para>If not all the UPNs are successfully registered then all the registration which did succeed are unregistered.
		/// This ensures that either all the UPNs are registered or none is.</para>
		/// </remarks>
		public override void Start()
		{
			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				PeerClassifierRegistrations = new List<PeerNameRegistration>();

				// Register all UPNs.
				foreach( Upn upn in Upns )
				{
					LogRecordFields fields = new LogRecordFields();
					fields.Add( new LogRecordField( "UPN", upn, false, LogRecordFieldPriorities.Middle ) );

					LogManager upnLogManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod(), fields );

					try
					{
						PeerName peerName = new PeerName( upn.Classifier, PeerNameType.Unsecured );

						PeerNameRegistration registration = new PeerNameRegistration( peerName, RegistrationPort
							, PnrpConsts.Cloud );
						registration.UseAutoEndPointSelection = true;

						registration.Start();

						PeerClassifierRegistrations.Add( registration );

						string message = string.Format( "The peer registration for the UPN \"{0}\" (classifier \"{1}\") of the"
							+ " vault with the locator \"{2}\" was started.", upn, upn.Classifier, ParentLocator );
						if( UseOnlyDebugLogs ) LogManager.DebugAdd( message, null );
						else LogManager.Add( message );
					}
					finally
					{
						LogManager.DebugEndMethod( upnLogManager );
					}
				}
			}
			catch( Exception ex )
			{
				Stop();

				string message = string.Format( "Failed to start the peer registration for the vault with the locator \"{0}\"."
					, ParentLocator );
				throw new ValidationException( message, ex );
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		#endregion Public methods
	}
}
