////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

namespace Vela.Serialization
{
	/// <summary>
	/// Contract for a hash serializer. This serializer may be used to serialize documents, and then digitally sign the result.
	/// The serialization output can't be deserialized.
	/// </summary>
	/// <remarks>
	/// <para>You can construct a so called "hash serializer" which guarantees that it generates the same output, regardless of
	/// the platform it runs on. This is crucial for signing data because the same serialization output must be signed in order
	/// to have the same signature.</para>
	/// 
	/// <para>The declaration order of the serialized properties is irrelevant because they are sorted.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Serialization versioning</b></para>
	/// 
	/// <para>Serialization versioning is essential if signatures of documents saved with old versions of their types must be
	/// verifiable after they are loaded in new versions their types.</para>
	/// 
	/// <para>In order to ensure some compatibility level among different versions of a document, a hash serializer must
	/// serialize only non-null scalars, and not store any information about the type of the scalars.</para>
	/// 
	/// <para>This allows you to change the type of a scalar in a way which generates the same serialization output, for
	/// example, from an integer to a real number, or from an enumeration to a text.</para>
	/// 
	/// <para>You can add new serializable properties to classes if they have a null default value, or if you specify the
	/// default value using <see cref="Vela.Kernel.Serialization.SerializeAttribute.DontHashSerializeValue"/>. This is
	/// because null values and default values are not hash serialized and therefore do not change the hash serializer's output.
	/// This pattern provides bidirectional compatibility of the hash serializer's output.</para>
	/// 
	/// <para>If you have a legacy property which must no longer be hash serialized but must still be available after
	/// deserialization, set the <see cref="Vela.Kernel.Serialization.SerializeAttribute.HashSerialization"/> property of
	/// the attribute applied on the legacy property to <see cref="Vela.Kernel.Serialization.HashSerializations.Never"/>.</para>
	/// </remarks>
	public interface IHashSerializer : IOneWaySerializer
	{
		#region Public serializable properties

		/// <summary>
		/// If this is true then the type names of the serialized values are stored in the serialization output. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is true, the compatibility level of different versions of a serialized document is limited (see the
		/// "Serialization versioning" section from the documentation of the current interface).</para>
		/// 
		/// <para>This must be true for a hash serializer which will be used to test the equality of documents. This is because
		/// the parameterless constructor may set some non-serializable properties which can make two (otherwise identical) types
		/// different.</para>
		/// </remarks>
		bool SerializeValueTypeName { get; }

		#endregion Public serializable properties
	}
}
