////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.Serialization
{
	/// <summary>
	/// Contract for a serialization factory for serializers (<see cref="ISerializer"/> and <see cref="IHashSerializer"/>) and
	/// deserializers (<see cref="IDeserializer"/>).
	/// </summary>
	public interface ISerializationFactory : IImmutableAlgorithm
	{
		#region Public non-serializable properties

		/// <summary>
		/// Text encoding. READ-ONLY.
		/// </summary>
		TextEncoding TextEncoding { get; }

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Returns a serializer with the configuration from the current serialization factory.
		/// </summary>
		/// <returns>Returns the serializer with the configuration from the current serialization factory.</returns>
		ISerializer GetSerializer();

		/// <summary>
		/// Returns a deserializer with the configuration from the current serialization factory.
		/// </summary>
		/// <returns>Returns the deserializer with the configuration from the current serialization factory.</returns>
		IDeserializer GetDeserializer();

		/// <summary>
		/// Returns a hash serializer with the configuration from the current serialization factory.
		/// </summary>
		/// <param name="serializeValueTypeName">If this is true then the type names of the serialized values are stored in the
		/// serialization output. If this is true, the compatibility level of different versions of a serialized document is
		/// limited (see the "Serialization versioning" section from the documentation of the <see cref="IHashSerializer"/>
		/// interface). This must be true for a hash serializer which will be used to test the equality of documents.</param>
		/// <returns>Returns the hash serializer with the configuration from the current serialization factory.</returns>
		IHashSerializer GetHashSerializer( bool serializeValueTypeName );

		/// <summary>
		/// Creates a deep clone of the specified document by serializing and then deserializing it.
		/// </summary>
		/// <param name="source">Document to clone. This may be null.</param>
		/// <returns>Returns the cloned document. This is null if "source" is null.</returns>
		IDocument CloneDocument( IDocument source );

		/// <summary>
		/// Checks if the specified documents are equal by comparing the output of their serializations.
		/// </summary>
		/// <param name="document1">Document to compare. This may be null.</param>
		/// <param name="document2">Document to compare. This may be null.</param>
		/// <returns>Returns the true if the documents are equal.</returns>
		/// <remarks>
		/// <para>The documents to compare may have different types.</para>
		/// </remarks>
		bool EqualDocuments( IDocument document1, IDocument document2 );

		#endregion Public methods
	}
}
