////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

namespace Vela.Serialization
{
	/// <summary>
	/// <para><b>Serialization and deserialization</b></para>
	/// 
	/// <para>A serialization followed by a deserialization must restore the consistent state of the original object.
	/// "Consistent" doesn't mean the workload state, but the one from which processing may start.</para>
	/// 
	/// <para>An object must fulfill the following conditions to be serializable:</para>
	/// <list type="bullet">
	///		<item>Must be an instance property with getter and setter. Fields are not serialized.</item>
	///		<item>Must have a <see cref="Vela.Kernel.Serialization.SerializeAttribute"/> attribute.</item>
	///		<item>If the property is a class then it must be non-abstract and must have an instance constructor with no
	///			parameters, must derive from <see cref="Vela.Kernel.Serialization.IDocument"/> and must not derive from
	///			<see cref="Vela.Kernel.Serialization.IUnserializable"/>; the constructor is required during
	///			deserialization to create an instance of that class.</item>
	///		<item>If the property is an array, it may only be a uni-dimensional array; the content of its items may have
	///			different types.</item>
	///		<item>If the property is a list, the content of its items may have different types. List items with a null value are
	///			ignored. The list must have an instance method named <see cref="SerializationConsts.ListAddMethodName"/> with a
	///			single "object" parameter; this method is required during deserialization to add items to the list.</item>
	/// </list>
	/// 
	/// <para>All properties and methods / constructors may be protected (and higher visibility). This allows you to protect a
	/// class from being accidentally altered from code, by declaring mostly read-only properties and methods.</para>
	/// 
	/// <para>The following property attributes are processed:</para>
	/// <list type="bullet">
	///		<item><see cref="Vela.Kernel.Serialization.SerializeAttribute"/> to specify that the object must be
	///			serialized.</item>
	/// </list>
	/// 
	/// <para>The serialization name of an object is its declared name.</para>
	/// 
	/// <para>When an object is serialized, its content must have the declared type of the object. If an object may also contain
	/// types derived from the declared type, those types must be registered in
	/// <see cref="Vela.Kernel.PortableTypes.Instance"/>.</para>
	/// 
	/// <para>Strings and buffers are considered scalars.</para>
	/// 
	/// <para>Objects which derive from <see cref="System.Collections.IEnumerable"/> are serialized as lists, through their
	/// enumerator property (not through their indexer). Arrays of scalars are also serialized as lists.</para>
	/// 
	/// <para>Before a class object is serialized, if the class type contains an instance method named
	/// <see cref="SerializationConsts.SerializingMethodName"/>, it is called. Note that this method is called for a class
	/// <b>after</b> it is called for its parent class.</para>
	/// 
	/// <para>After a class object is serialized, if the class type contains an instance method named
	/// <see cref="SerializationConsts.SerializedMethodName"/>, it is called. Note that this method is called for a class
	/// <b>before</b> it is called for its parent class.</para>
	/// 
	/// <para>Before a class object starts being deserialized, if the class type contains an instance method named
	/// <see cref="SerializationConsts.DeserializingMethodName"/>, it is called. Note that this method is called for a class
	/// <b>after</b> it is called for its parent class.</para>
	/// 
	/// <para>After a class object has been deserialized, if the class type contains an instance method named
	/// <see cref="SerializationConsts.DeserializedMethodName"/>, it is called. Note that this method is called for a class
	/// <b>before</b> it is called for its parent class.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Circular references</b></para>
	/// 
	/// <para>Circular references are allowed, but they must not be created in the constructor (because that would create an
	/// infinite circular reference).</para>
	/// 
	/// <para>A circular reference is a class which contains a property whose type is the (parent) class itself.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Serialization versioning</b></para>
	/// 
	/// <para>Null properties (from a class) are not serialized.</para>
	/// 
	/// <para>Properties for which no serialization node is found (during deserialization) in the data to be deserialized, remain
	/// unset. Therefore, adding a new property with a non-null default value to a signed document would invalidate the existing
	/// signatures.</para>
	/// 
	/// <para>You can add new serializable properties to classes if they have a null default value, or if you specify the
	/// default value using <see cref="Vela.Kernel.Serialization.SerializeAttribute.DontHashSerializeValue"/>. This is
	/// because null values and default values are not hash serialized and therefore do not change the hash serializer's output,
	/// fact which maintains the validity of the existing signatures. This pattern provides bidirectional (= old - new versions)
	/// compatibility of the hash serializer's output.</para>
	/// 
	/// <para>If you change the type of a property after documents (which contain it) have been serialized, you must ensure than
	/// the type of the property from the already serialized documents is derived from the new declared type
	/// (<see cref="object"/> is the base class for all the types).</para>
	/// 
	/// <para>For serialization versioning for the hash serializer, see <see cref="IHashSerializer"/>.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Binary versus XML</b></para>
	/// 
	/// <para>The performances of the binary and XML serialization mechanisms are smilar.</para>
	/// 
	/// <para>The only scenario where the binary serialization may be significantly better, is when large amounts of binary data
	/// are serialized. The XML serializer must convert the binary data to text, and this significantly increases the size of
	/// the output, especially if the output is encrypted later (which causes a second conversion of binary data to text).</para>
	/// </summary>
	class NamespaceDoc
	{
		// Only for documentation.
	}
}
