////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections;
using System.Reflection;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.Serialization
{
	/// <summary>
	/// Serialization utilities.
	/// </summary>
	public static class SerializationUtils
	{
		#region Public static methods

		/// <summary>
		/// Formats the name of a property which is written by a serializer.
		/// </summary>
		/// <param name="parentName">Parent class serialization name.</param>
		/// <param name="propName">Property serialization name.</param>
		/// <returns>Returns the resulted name.</returns>
		public static string FormatSerializationName( string parentName, string propName )
		{
			ValidationUtils.TextHasContent( parentName, "parentName" );
			ValidationUtils.TextHasContent( propName, "propName" );

			return parentName + SerializationConsts.FormattedSerializationNameSeparator + propName;
		}

		/// <summary>
		/// Formats the name of a list item which is written by a serializer.
		/// </summary>
		/// <param name="parentName">Parent list serialization name.</param>
		/// <param name="listItemName">The serialization name of a list item.</param>
		/// <param name="itemsCounter">The index of specified property in its parent list. This must be at least 0.</param>
		/// <returns>Returns the resulted name.</returns>
		public static string FormatSerializationListItemName( string parentName, string listItemName, int itemsCounter )
		{
			ValidationUtils.TextHasContent( parentName, "parentName" );
			ValidationUtils.TextHasContent( listItemName, "listItemName" );
			ValidationUtils.Integer( itemsCounter, 0, int.MaxValue, "itemsCounter" );

			return parentName
				+ SerializationConsts.FormattedSerializationNameSeparator + listItemName
				+ SerializationConsts.FormattedSerializationListItemNameSeparator + (itemsCounter + 1);
		}

		/// <summary>
		/// Verifies if the specified object type is a list / array.
		/// </summary>
		/// <param name="type">The type to verify if it is a list.</param>
		/// <returns>Returns true if the specified type is a list.</returns>
		/// <remarks>
		/// <para>Verify if the object is scalar before this (because some scalars are lists, but must be serialized as
		/// scalars).</para>
		/// </remarks>
		public static bool IsList( Type type )
		{
			ValidationUtils.ObjectHasContent( type, "type" );

			Type itemDeclaredType = type.GetElementType();
			if( itemDeclaredType != null )
				return true;

			if( typeof( IEnumerable ).IsAssignableFrom( type ) )
			{
				// The type is a list.

				// Check if the type has an instance method named "ListAddMethodName" with a single "object" parameter.
				MethodInfo mi = type.GetMethod( SerializationConsts.ListAddMethodName, BindingFlags.Public
					| BindingFlags.NonPublic | BindingFlags.Instance, null, new Type[] { typeof( object ) }, null );

				if( mi != null )
					return true;
				else
				{
					throw new ValidationException( string.Format( "The type \"{0}\" is a list but has no instance \"{1}\""
							+ " method with a single \"object\" parameter."
						, type.FullName, SerializationConsts.ListAddMethodName ) );
				}
			}

			return false;
		}

		/// <summary>
		/// Verifies if the specified object type is a class.
		/// </summary>
		/// <param name="type">The type to verify if it is a class.</param>
		/// <returns>Returns true if the specified type is a class.</returns>
		/// <remarks>
		/// <para>Verify if the object is a scalar or a list before this (because some scalars and all lists are actually
		/// classes, but must be serialized as scalars or lists).</para>
		/// </remarks>
		public static bool IsClass( Type type )
		{
			ValidationUtils.ObjectHasContent( type, "type" );

			return type.IsClass;
		}

		/// <summary>
		/// Checks if the specified type is a serializable class. An exception is thrown if the type is not a serializable class.
		/// </summary>
		/// <param name="type">Type to check if it's a serializable class.</param>
		/// <returns>Returns the information of the constructor which can be used to instantiate the type.</returns>
		public static ConstructorInfo CheckSerializableClass( Type type )
		{
			if( !IsClass( type ) )
			{
				throw new ValidationException( string.Format( "The type \"{0}\" to check if it's serializable is not a class."
					, type.FullName ) );
			}

			ConstructorInfo ci = type.GetConstructor( BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance, null
				, Type.EmptyTypes, null );

			// The class must be non-abstract and must have an instance constructor with no parameters, must derive from
			// "IDocument" and must not derive from "IUnserializable".
			if( !type.IsVisible || !type.IsClass || type.IsAbstract || !typeof( IDocument ).IsAssignableFrom( type )
				|| typeof( IUnserializable ).IsAssignableFrom( type ) || (ci == null) )
			{
				throw new ValidationException( string.Format( "Only non-abstract classes derived from \"{0}\" and not derived"
						+ " from \"{1}\", with an instance constructor with no parameters may be (de)serialized. Type \"{2}\""
						+ " doesn't meet these requirements."
					, typeof( IDocument ).FullName, typeof( IUnserializable ).FullName, type.FullName ) );
			}

			return ci;
		}

		/// <summary>
		/// Checks if the specified property is serializable.
		/// </summary>
		/// <param name="propInfo">Information about the property whose attributes to check.</param>
		/// <param name="serializableData">The parent class object of the specified property.</param>
		/// <param name="rootSerializableData">The object which is currently being serialized.</param>
		/// <param name="isHashSerializer">This is true if the current serializer is a hash serializer.</param>
		/// <param name="serializationName">Is set with the serialization name of the specified property. This may be different
		/// from the name of the specified property.</param>
		/// <param name="listItemName">Is set with the serialization name of list items.</param>
		/// <param name="propValue">Is set with the value of specified property.</param>
		/// <returns>Return true if the object must be serialized, and false if it must NOT be serialized.</returns>
		public static bool CheckSerializeAttribute( PropertyInfo propInfo, object serializableData, object rootSerializableData
			, bool isHashSerializer, out string serializationName, out string listItemName, out object propValue )
		{
			ValidationUtils.ObjectHasContent( propInfo, "propInfo" );
			ValidationUtils.ObjectHasContent( serializableData, "serializableData" );
			ValidationUtils.ObjectHasContent( rootSerializableData, "rootSerializableData" );

			serializationName = propInfo.Name;
			listItemName = null;
			propValue = null;

			// Get the attributes of the property.
			object[] attributes = propInfo.GetCustomAttributes( false );

			if( attributes != null )
			{
				#region Check if the object must be or not be serialized

				bool doSerialize = false;
				object dontHashSerializeValue = null;
				bool mustBeDeserialized = false;

				foreach( object attribute in attributes )
				{
					if( (attribute != null)
						&& (
							(attribute.GetType() == typeof( SerializeAttribute ))
#if DEBUG && SERIALIZEEXPOSEDATTRIBUTE
							|| (attribute.GetType() == typeof( SerializeExposedAttribute ))
#endif
) )
					{
						SerializeAttribute serAttr = (SerializeAttribute)attribute;

						if( isHashSerializer )
						{
							// Check if the property must not be serialized with a hash serializer.
							if( (serAttr.HashSerialization == HashSerializations.Never)
								|| (
									(serAttr.HashSerialization == HashSerializations.NotInRoot)
									&& (serializableData == rootSerializableData)
								) )
							{
								return false;
							}

							dontHashSerializeValue = serAttr.DontHashSerializeValue;
						}
						else
						{
							mustBeDeserialized = serAttr.MustBeDeserialized;
						}

						if( TextUtils.HasContent( serAttr.SerializationName ) )
							serializationName = serAttr.SerializationName;

						listItemName = serAttr.ListItemName;

						// The property is serializable, break the cycle.
						doSerialize = true;
						break;
					}
				}

				if( !doSerialize )
					return false;

				#endregion Check if the object must be or not be serialized

				// Get the property value.
				propValue = propInfo.GetValue( serializableData, null );

				// Null objects.
				if( propValue == null )
				{
					// Check if the property must be deserialized but is null.
					if( mustBeDeserialized )
					{
						throw new ValidationException( string.Format( "The property \"{0}\" from class \"{1}\" which is being"
								+ " serialized must be deserialized but is null."
							, propInfo.Name, serializableData.GetType().FullName ) );
					}
					else
					{
						return false; // Don't serialize null objects.
					}
				}

				Type contentType = propValue.GetType();

				// Check the derived type of the object.
				if( (contentType != propInfo.PropertyType) && (PortableTypes.FindGlobal( contentType ) == null) )
				{
					throw new ValidationException( string.Format( "The derived type \"{0}\" of the property \"{1}\" which"
							+ " is being serialized was not found in the list of global portable types."
						, contentType.FullName, propInfo.Name ) );
				}

				// If the property has the default value, it must not be serialized with a hash serializer.
				if( (dontHashSerializeValue != null) && propValue.Equals( dontHashSerializeValue ) )
				{
					return false;
				}

				// The object must be serialized.
				return true;
			}

			// The object must NOT be serialized.
			return false;
		}

		/// <summary>
		/// Checks if the specified property is deserializable.
		/// </summary>
		/// <param name="propInfo">Information about the property whose attributes to check.</param>
		/// <param name="mustBeDeserialized">Is set with a value which indicates whether the document which is being
		/// deserialized must have a node which contains the value of the specified property.</param>
		/// <param name="serializationName">Is set with the serialization name of the specified property. This may be different
		/// from the name of the specified property.</param>
		/// <param name="listItemName">Is set with the serialization name of list items.</param>
		/// <param name="enumFallbackValue">Is set with the fallback value for enumerations. If the object to deserialize is an
		/// enumeration, if the deserialization of the enumeration fails, this value is assigned instead of an exception being
		/// thrown. This may be null.</param>
		/// <returns>Returns true if the object must be deserialized, and false if it must NOT be deserialized.</returns>
		public static bool CheckDeserializeAttribute( PropertyInfo propInfo, out bool mustBeDeserialized
			, out string serializationName, out string listItemName, out object enumFallbackValue )
		{
			mustBeDeserialized = false;
			serializationName = propInfo.Name;
			listItemName = null;
			enumFallbackValue = null;

			// Get the attributes of the property.
			object[] attributes = propInfo.GetCustomAttributes( typeof( SerializeAttribute ), false );

			if( attributes != null )
			{
				#region Check if the object must be or not be deserialized

				bool doSerialize = false;

				foreach( object attribute in attributes )
				{
					if( (attribute != null) && (attribute.GetType() == typeof( SerializeAttribute )) )
					{
						SerializeAttribute serAttr = (SerializeAttribute)attribute;

						mustBeDeserialized = serAttr.MustBeDeserialized;

						if( TextUtils.HasContent( serAttr.SerializationName ) )
							serializationName = serAttr.SerializationName;

						listItemName = serAttr.ListItemName;

						enumFallbackValue = serAttr.EnumFallbackValue;

						// Object is serializable, break the cycle.
						doSerialize = true;
						break;
					}
				}

				if( !doSerialize )
					return false;

				#endregion Check if the object must be or not be deserialized

				// The object must be deserialized.
				return true;
			}

			// The object must NOT be deserialized.
			return false;
		}

		/// <summary>
		/// Extracts the deserializer associated to the serializer which was used to serialize the specified data. An exception
		/// is thrown if a deserializer can't be extracted.
		/// </summary>
		/// <param name="data">Serialized data. This may be either a buffer or a text.</param>
		/// <returns>Returns the extracted deserializer.</returns>
		/// <remarks>
		/// <para>The text encoding of the deserializer is also extracted.</para>
		/// </remarks>
		public static IDeserializer ExtractDeserializer( object data )
		{
			return RegistrableAlgorithmUtils.GetParser( data, typeof( IDeserializer ) ) as IDeserializer;
		}

		/// <summary>
		/// Loads the class object from the specified buffer.
		/// </summary>
		/// <param name="buffer">Buffer to deserialize. This may be null or empty.</param>
		/// <param name="documentType">The root type of the document to load; the document to load may either have this type or
		/// a type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <param name="deserializer">Is set with the deserializer extracted from the data to deserialize.</param>
		/// <returns>Returns the deserialized document, or null if the input data is null or empty.</returns>
		public static IDocument LoadFromBuffer( byte[] buffer, Type documentType, out IDeserializer deserializer )
		{
			if( BufferUtils.NoContent( buffer ) )
			{
				deserializer = null;

				return null;
			}

			deserializer = ExtractDeserializer( buffer );

			return deserializer.LoadFromBuffer( buffer, documentType );
		}

		/// <summary>
		/// Loads the class object from the specified file.
		/// </summary>
		/// <param name="fileName">The name of the file to deserialize. The file may be missing and its content may be null or
		/// empty.</param>
		/// <param name="documentType">The root type of the document to load; the document to load may either have this type or
		/// a type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <param name="deserializer">Is set with the deserializer extracted from the data to deserialize.</param>
		/// <returns>Returns the deserialized document, or null if the file is missing or its content is null or empty.</returns>
		/// <remarks>
		/// <para>This method implements a recovery mechanism following a crash during a call of the
		/// <see cref="ISerializer.SaveToFile"/> method.</para>
		/// </remarks>
		public static IDocument LoadFromFile( string fileName, Type documentType, out IDeserializer deserializer )
		{
			byte[] buffer = FileUtils.Read( fileName, false );

			return LoadFromBuffer( buffer, documentType, out deserializer );
		}

		/// <summary>
		/// Loads the class object from the specified text.
		/// </summary>
		/// <param name="data">Text to deserialize. This may be null or empty.</param>
		/// <param name="documentType">The root type of the document to load; the document to load may either have this type or
		/// a type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <param name="deserializer">Is set with the deserializer extracted from the data to deserialize.</param>
		/// <returns>Returns the deserialized document, or null if the input data is null or empty.</returns>
		public static IDocument LoadFromText( string data, Type documentType, out IDeserializer deserializer )
		{
			if( TextUtils.NoContent( data ) )
			{
				deserializer = null;

				return null;
			}

			byte[] buffer = KernelUtils.GetDeepTextEncoding().TextToArray( data, false );

			return LoadFromBuffer( buffer, documentType, out deserializer );
		}

		/// <summary>
		/// Checks if the specified type may be signed, and throws an exception if it's not.
		/// </summary>
		/// <param name="type">Type to check. This may be null.</param>
		public static void ValidateSignableType( Type type )
		{
			if( type == null )
				return;

			if( typeof( IUnsignable ).IsAssignableFrom( type ) )
			{
				string messageTemplate = "The type \"{0}\" is derived from \"{1}\" and therefore can't be signed.";

				throw new ValidationException( string.Format( messageTemplate, type.FullName, typeof( IUnsignable ) ) );
			}
		}

		/// <summary>
		/// Checks if the type of the specified value may be signed, and throws an exception if it's not.
		/// </summary>
		/// <param name="value">Value whose type to check. This may be null.</param>
		public static void ValidateSignableObject( object value )
		{
			if( value == null )
				return;

			ValidateSignableType( value.GetType() );
		}

		#endregion Public static methods
	}
}
