////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.Serialization
{
	/// <summary>
	/// Contract for a class from which all storable configuration classes must derive.
	/// </summary>
	public abstract class StorableConfiguration : NotifyPropertyChanged, IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public StorableConfiguration()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="fileName">The name of the file where the configuration is stored.</param>
		public StorableConfiguration( string fileName )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			FileName = fileName;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The name of the file where the configuration is stored.
		/// </summary>
		private string FileName;

		#endregion Private fields

		#region Public static methods

		/// <summary>
		/// Loads the configuration from the specified file.
		/// </summary>
		/// <param name="fileName">The name of the file from which to load the configuration.</param>
		/// <param name="configurationType">The type of the configuration object.</param>
		/// <param name="deserializer">Is set with the deserializer of the loaded sheaf. If there is no configuration to
		/// load, this is null.</param>
		/// <returns>Returns the loaded configuration object. If the specified file doesn't exists then a new configuration
		/// object, of the specified type, is returned.</returns>
		/// <remarks>
		/// <para>This method implements a recovery mechanism following a crash during a call of the <see cref="Save"/>
		/// method.</para>
		/// </remarks>
		public static StorableConfiguration Load( string fileName, Type configurationType, out IDeserializer deserializer )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );
			ValidationUtils.ObjectHasContent( configurationType, "configurationType" );

			StorableConfiguration configuration;

			try
			{
				byte[] buffer = FileUtils.Read( fileName, false );

				if( BufferUtils.HasContent( buffer ) )
				{
					configuration = SerializationUtils.LoadFromBuffer( buffer, configurationType, out deserializer )
						as StorableConfiguration;
				}
				else
				{
					configuration = (StorableConfiguration)Activator.CreateInstance( configurationType, true );

					deserializer = null;
				}

				configuration.FileName = fileName;
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsAbort( ex ) ) throw ExceptionUtils.PreserveStackTrace( ex );

				throw new ValidationException( string.Format( "Failed to load the configuration from file \"{0}\".", fileName )
					, ex );
			}

			return configuration;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Saves the configuration to the associated file.
		/// </summary>
		/// <param name="serializer">The serializer with which to serialize the configuration object to save.</param>
		/// <remarks>
		/// <para>The saving is crash-safe. If the application crashes when the file is being saved, at least 1 of the
		/// following files is still available: a file with the specified name, a file with a backup name, a file with a
		/// temporary name. If the backup file exists, it's the valid one and you should implement your own recovery mechanism
		/// following a crash.</para>
		/// </remarks>
		public void Save( ISerializer serializer )
		{
			ValidationUtils.ObjectHasContent( serializer, "serializer" );
			ValidationUtils.TextHasContent( FileName, "FileName" );

			try
			{
				serializer.SaveToFile( this, null, FileName );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsAbort( ex ) ) throw ExceptionUtils.PreserveStackTrace( ex );

				throw new ValidationException( string.Format( "Failed to save the configuration to file \"{0}\".", FileName )
					, ex );
			}
		}

		#endregion Public methods
	}
}
