////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Data;
using System.Data.Common;
using Vela.DatabaseStorehouse;
using Vela.Kernel;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.SqlDatabase
{
	/// <summary>
	/// Provider for an SQL Server database.
	/// </summary>
	/// <remarks>
	/// <para>Don't share a storehouse provider among multiple threads, even if you synchronize it. Clone it instead.</para>
	/// </remarks>
	public abstract class SqlDatabaseProvider : DatabaseProvider
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected SqlDatabaseProvider()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="access">The database access.</param>
		/// <param name="fileName">The name of the file where the storehouse header is stored.</param>
		/// <param name="headerUserPassword">The user password used to symmetrically encrypt the header.</param>
		/// <param name="header">The storehouse header (which is stored in a symmetrically encrypted form). This may be null if
		/// the storehouse doesn't exist yet.</param>
		/// <param name="serializer">The serializer of the sheafs.</param>
		public SqlDatabaseProvider( DatabaseAccess access, string fileName, string headerUserPassword, StorehouseHeader header
				, ISerializer serializer )
			: base( access, fileName, headerUserPassword, header, serializer )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		private IStorehouseScripts _Scripts;
		/// <summary>
		/// Set of scripts used to work with the storehouse. READ-ONLY.
		/// </summary>
		public override IStorehouseScripts Scripts
		{
			get { return _Scripts; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Scripts" );

				_Scripts = value;
			}
		}

		private IStorehouseScriptPlaceholders _ScriptPlaceholders;
		/// <summary>
		/// Placeholders used in storehouse scripts. READ-ONLY.
		/// </summary>
		public override IStorehouseScriptPlaceholders ScriptPlaceholders
		{
			get { return _ScriptPlaceholders; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ScriptPlaceholders" );

				_ScriptPlaceholders = value;
			}
		}

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Creates the specified storehouse realms.
		/// </summary>
		/// <param name="storehouseRealms">Storehouse realms to create. This may be empty, but not null.</param>
		public override void CreateStorehouseRealms( Strings storehouseRealms )
		{
			ValidationUtils.ObjectHasContent( storehouseRealms, "storehouseRealms" );

			// When "CreateConnectionString" was called inside "CreateConnection", when a storehouse was automatically loaded at
			// the program's start and an exception was thrown because the password was not the correct one,
			// "System.Data.SqlServerCe.dll" was throwing an odd exception simply because it was loaded (by the assembly which
			// implements "CreateConnection"). The exception was in a "Dispose" (called in a finalizer) of an "SqlCeConnection"
			// (which was not created by Vela).
			string connectionString = CreateConnectionString();

			using( DbConnection connection = CreateConnection( connectionString ) )
			{
				connection.Open();

				DbTransaction transaction = null;
				try
				{
					transaction = connection.BeginTransaction( IsolationLevel.ReadCommitted );

					using( DbCommand command = CreateCommand( connection, transaction ) )
					{
						foreach( string realm in storehouseRealms )
						{
							string script;

							// If the storehouse contains the realm, skip creating it.
							command.CommandText = TextUtils.Replace( ScriptsTyped.StorehouseHasRealm
								, DatabaseScriptPlaceholders.StorehouseRealm, realm );
							if( command.ExecuteScalar() != null )
								continue;

							// Create the storehouse realm.
							script = TextUtils.Replace( ScriptsTyped.CreateStorehouseRealm
								, DatabaseScriptPlaceholders.StorehouseRealm, realm );
							SplitExecuteNonQuery( script, command );
						}
					}

					transaction.Commit();
					transaction.Dispose();
				}
				catch( Exception ex )
				{
					if( transaction != null )
					{
						transaction.Rollback();
						transaction.Dispose();
					}

					throw ExceptionUtils.PreserveStackTrace( ex );
				}
			}
		}

		#endregion Public methods
	}
}
