////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.DatabaseStorehouse;

namespace Vela.SqlDatabase
{
	/// <summary>
	/// Set of scripts used to work with an SQL Server database.
	/// </summary>
	public abstract class SqlDatabaseScripts : IDatabaseScripts
	{
		#region Public constants

		/// <summary>
		/// The separator for a list of parameters used in a script.
		/// </summary>
		public const string ParameterListSeparator = ",";

		#endregion Public constants

		#region Public non-serializable properties

		/// <summary>
		/// The script used to create the storehouse on a storehouse server engine. READ-ONLY.
		/// </summary>
		public virtual string CreateServerStorehouse
		{
			get { return Properties.Resources.CreateServerStorehouse; }
		}

		/// <summary>
		/// The script used to create a storehouse realm. READ-ONLY.
		/// </summary>
		public virtual string CreateStorehouseRealm
		{
			get { return Properties.Resources.CreateStorehouseRealm; }
		}

		/// <summary>
		/// The script used to check if a storehouse realm exists. READ-ONLY.
		/// </summary>
		public virtual string StorehouseHasRealm
		{
			get { return Properties.Resources.StorehouseHasRealm; }
		}

		/// <summary>
		/// The script used to count the speckles which can be collected (with no lock). READ-ONLY.
		/// </summary>
		public virtual string CountSpeckles
		{
			get { return Properties.Resources.CountSpeckles; }
		}

		/// <summary>
		/// The script used to add a speckle (with an exclusive lock). READ-ONLY.
		/// </summary>
		public virtual string AddSpeckle_ExclusiveLock
		{
			get { return Properties.Resources.AddSpeckle_ExclusiveLock; }
		}

		/// <summary>
		/// The script used to collect speckles (with no lock); the requested speckle filters are loaded. READ-ONLY.
		/// </summary>
		public virtual string CollectSpeckles
		{
			get { return Properties.Resources.CollectSpeckles; }
		}

		/// <summary>
		/// The script used to collect speckles (with no lock) in a paged manner, starting with the newest ones; the requested
		/// speckle filters are loaded. READ-ONLY.
		/// </summary>
		public virtual string CollectSpeckles_PagedNewest
		{
			get { return Properties.Resources.CollectSpeckles_PagedNewest; }
		}

		/// <summary>
		/// The script used to collect speckles (with no lock) in a paged manner, starting with the oldest ones; the requested
		/// speckle filters are loaded. READ-ONLY.
		/// </summary>
		public virtual string CollectSpeckles_PagedOldest
		{
			get { return Properties.Resources.CollectSpeckles_PagedOldest; }
		}

		/// <summary>
		/// The script used to collect a speckle (with a shared lock); the requested speckle filters are loaded.
		/// READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The script returns the first encountered speckle, not necessarily the speckle which was added first. This
		/// behavior mVela.Imizes performance.</para>
		/// </remarks>
		public virtual string CollectFirstSpeckle_SharedLock
		{
			get { return Properties.Resources.CollectFirstSpeckle_SharedLock; }
		}

		/// <summary>
		/// The script used to collect a speckle (with an update lock); the requested speckle filters are loaded.
		/// READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The script returns the first encountered speckle, not necessarily the speckle which was added first. This
		/// behavior mVela.Imizes performance.</para>
		/// </remarks>
		public virtual string CollectFirstSpeckle_UpdateLock
		{
			get { return Properties.Resources.CollectFirstSpeckle_UpdateLock; }
		}

		/// <summary>
		/// The script used to collect a speckle (with an exclusive lock); the requested speckle filters are loaded.
		/// READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The script returns the first encountered speckle, not necessarily the speckle which was added first. This
		/// behavior mVela.Imizes performance.</para>
		/// </remarks>
		public virtual string CollectFirstSpeckle_ExclusiveLock
		{
			get { return Properties.Resources.CollectFirstSpeckle_ExclusiveLock; }
		}

		/// <summary>
		/// The script used to load a speckle (with a shared lock); the requested speckle filters are loaded.
		/// READ-ONLY.
		/// </summary>
		public virtual string LoadSpeckle_SharedLock
		{
			get { return Properties.Resources.LoadSpeckle_SharedLock; }
		}

		/// <summary>
		/// The script used to load a speckle (with an update lock); the requested speckle filters are loaded.
		/// READ-ONLY.
		/// </summary>
		public virtual string LoadSpeckle_UpdateLock
		{
			get { return Properties.Resources.LoadSpeckle_UpdateLock; }
		}

		/// <summary>
		/// The script used to load a speckle (with an exclusive lock); the requested speckle filters are loaded.
		/// READ-ONLY.
		/// </summary>
		public virtual string LoadSpeckle_ExclusiveLock
		{
			get { return Properties.Resources.LoadSpeckle_ExclusiveLock; }
		}

		/// <summary>
		/// The script used to check if a speckle has a referred sheaf; no lock is acquired. READ-ONLY.
		/// </summary>
		public virtual string SpeckleHasDocument
		{
			get { return Properties.Resources.SpeckleHasSheaf; }
		}

		/// <summary>
		/// The script used to collect storehouse keys (with no lock). READ-ONLY.
		/// </summary>
		public virtual string CollectKeys
		{
			get { return Properties.Resources.CollectKeys; }
		}

		/// <summary>
		/// The script used to collect storehouse keys (with no lock) in a paged manner, starting with the newest ones.
		/// READ-ONLY.
		/// </summary>
		public virtual string CollectKeys_PagedNewest
		{
			get { return Properties.Resources.CollectKeys_PagedNewest; }
		}

		/// <summary>
		/// The script used to collect storehouse keys (with no lock) in a paged manner, starting with the oldest ones.
		/// READ-ONLY.
		/// </summary>
		public virtual string CollectKeys_PagedOldest
		{
			get { return Properties.Resources.CollectKeys_PagedOldest; }
		}

		/// <summary>
		/// The script used to add a sheaf (with an exclusive lock). READ-ONLY.
		/// </summary>
		public virtual string AddSheaf_ExclusiveLock
		{
			get { return Properties.Resources.AddSheaf_ExclusiveLock; }
		}

		/// <summary>
		/// The script used to update a sheaf (with an update lock). READ-ONLY.
		/// </summary>
		public virtual string UpdateSheaf_UpdateLock
		{
			get { return Properties.Resources.UpdateSheaf_UpdateLock; }
		}

		/// <summary>
		/// The script used to update a sheaf (with an exclusive lock). READ-ONLY.
		/// </summary>
		public virtual string UpdateSheaf_ExclusiveLock
		{
			get { return Properties.Resources.UpdateSheaf_ExclusiveLock; }
		}

		/// <summary>
		/// The script used to delete sheafs (with an exclusive lock). READ-ONLY.
		/// </summary>
		public virtual string DeleteSheafs_ExclusiveLock
		{
			get { return Properties.Resources.DeleteSheafs_ExclusiveLock; }
		}

		/// <summary>
		/// The script used to delete the first sheaf (with an exclusive lock). READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The script deletes the first encountered sheaf, not necessarily the sheaf which was added first. This behavior
		/// mVela.Imizes performance.</para>
		/// </remarks>
		public virtual string DeleteFirstSheaf_ExclusiveLock
		{
			get { return Properties.Resources.DeleteFirstSheaf_ExclusiveLock; }
		}

		/// <summary>
		/// The script used to delete a sheaf (with an exclusive lock). READ-ONLY.
		/// </summary>
		public virtual string DeleteSheaf_ExclusiveLock
		{
			get { return Properties.Resources.DeleteSheaf_ExclusiveLock; }
		}

		/// <summary>
		/// The script used to collect sheafs (with a shared lock). READ-ONLY.
		/// </summary>
		public virtual string CollectSheafs_SharedLock
		{
			get { return Properties.Resources.CollectSheafs_SharedLock; }
		}

		/// <summary>
		/// The script used to collect sheafs (with an update lock). READ-ONLY.
		/// </summary>
		public virtual string CollectSheafs_UpdateLock
		{
			get { return Properties.Resources.CollectSheafs_UpdateLock; }
		}

		/// <summary>
		/// The script used to collect sheafs (with an exclusive lock). READ-ONLY.
		/// </summary>
		public virtual string CollectSheafs_ExclusiveLock
		{
			get { return Properties.Resources.CollectSheafs_ExclusiveLock; }
		}

		/// <summary>
		/// The script used to collect sheafs (with a shared lock) in a paged manner, starting with the newest ones. READ-ONLY.
		/// </summary>
		public virtual string CollectSheafs_PagedNewest_SharedLock
		{
			get { return Properties.Resources.CollectSheafs_PagedNewest_SharedLock; }
		}

		/// <summary>
		/// The script used to collect sheafs (with an update lock) in a paged manner, starting with the newest ones. READ-ONLY.
		/// </summary>
		public virtual string CollectSheafs_PagedNewest_UpdateLock
		{
			get { return Properties.Resources.CollectSheafs_PagedNewest_UpdateLock; }
		}

		/// <summary>
		/// The script used to collect sheafs (with an exclusive lock) in a paged manner, starting with the newest ones.
		/// READ-ONLY.
		/// </summary>
		public virtual string CollectSheafs_PagedNewest_ExclusiveLock
		{
			get { return Properties.Resources.CollectSheafs_PagedNewest_ExclusiveLock; }
		}

		/// <summary>
		/// The script used to collect sheafs (with a shared lock) in a paged manner, starting with the oldest ones. READ-ONLY.
		/// </summary>
		public virtual string CollectSheafs_PagedOldest_SharedLock
		{
			get { return Properties.Resources.CollectSheafs_PagedOldest_SharedLock; }
		}

		/// <summary>
		/// The script used to collect sheafs (with an update lock) in a paged manner, starting with the oldest ones. READ-ONLY.
		/// </summary>
		public virtual string CollectSheafs_PagedOldest_UpdateLock
		{
			get { return Properties.Resources.CollectSheafs_PagedOldest_UpdateLock; }
		}

		/// <summary>
		/// The script used to collect sheafs (with an exclusive lock) in a paged manner, starting with the oldest ones.
		/// READ-ONLY.
		/// </summary>
		public virtual string CollectSheafs_PagedOldest_ExclusiveLock
		{
			get { return Properties.Resources.CollectSheafs_PagedOldest_ExclusiveLock; }
		}

		/// <summary>
		/// The script used to collect a sheaf (with a shared lock). READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The script returns the first encountered sheaf, not necessarily the sheaf which was added first. This behavior
		/// mVela.Imizes performance.</para>
		/// </remarks>
		public virtual string CollectFirstSheaf_SharedLock
		{
			get { return Properties.Resources.CollectFirstSheaf_SharedLock; }
		}

		/// <summary>
		/// The script used to collect a sheaf (with an update lock). READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The script returns the first encountered sheaf, not necessarily the sheaf which was added first. This behavior
		/// mVela.Imizes performance.</para>
		/// </remarks>
		public virtual string CollectFirstSheaf_UpdateLock
		{
			get { return Properties.Resources.CollectFirstSheaf_UpdateLock; }
		}

		/// <summary>
		/// The script used to collect a sheaf (with an exclusive lock). READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The script returns the first encountered sheaf, not necessarily the sheaf which was added first. This behavior
		/// mVela.Imizes performance.</para>
		/// </remarks>
		public virtual string CollectFirstSheaf_ExclusiveLock
		{
			get { return Properties.Resources.CollectFirstSheaf_ExclusiveLock; }
		}

		/// <summary>
		/// The script used to load a sheaf (with a shared lock). READ-ONLY.
		/// </summary>
		public virtual string LoadSheaf_SharedLock
		{
			get { return Properties.Resources.LoadSheaf_SharedLock; }
		}

		/// <summary>
		/// The script used to load a sheaf (with an update lock). READ-ONLY.
		/// </summary>
		public virtual string LoadSheaf_UpdateLock
		{
			get { return Properties.Resources.LoadSheaf_UpdateLock; }
		}

		/// <summary>
		/// The script used to load a sheaf (with an exclusive lock). READ-ONLY.
		/// </summary>
		public virtual string LoadSheaf_ExclusiveLock
		{
			get { return Properties.Resources.LoadSheaf_ExclusiveLock; }
		}

		/// <summary>
		/// The script used to get the type of a sheaf (with a shared lock). READ-ONLY.
		/// </summary>
		public virtual string GetSheafType_SharedLock
		{
			get { return Properties.Resources.GetSheafType_SharedLock; }
		}

		#endregion Public non-serializable properties
	}
}
