////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Text;
using Vela.CoreDocuments;
using Vela.DatabaseStorehouse;
using Vela.Kernel;

namespace Vela.SqlDatabase
{
	/// <summary>
	/// Script placeholders for an SQL Server database.
	/// </summary>
	public class SqlScriptPlaceholders : DatabaseScriptPlaceholders
	{
		#region Public methods

		/// <summary>
		/// Gets the script fragment which contains the condition to load information from a database, according to the
		/// specified seek sheaf type name checksums.
		/// </summary>
		/// <param name="sheafTypes">Sheaf types whose names to include in the script.</param>
		/// <returns>Returns the created script fragment.</returns>
		public override string GetSheafTypeNameChecksumSeek( PortableTypes sheafTypes )
		{
			ValidationUtils.ListHasContent( sheafTypes, "sheafTypes" );

			StringBuilder sb = new StringBuilder();

			sb.Append( "(" );

			int count = sheafTypes.Count;
			for( int i = 0; i < count; i++ )
			{
				string template = (i <= 0) ? "({0} = {1})" : " OR ({0} = {1})";

				sb.AppendFormat( template, DatabaseColumns.SheafTypeNameChecksum
					, DatabaseParameters.GetSheafTypeNameChecksum( i ) );
			}

			sb.Append( ")" );

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the condition to load information from a database, according to the
		/// specified seek sheaf type names.
		/// </summary>
		/// <param name="sheafTypes">Sheaf types whose names to include in the script.</param>
		/// <returns>Returns the created script fragment.</returns>
		public override string GetSheafTypeNameSeek( PortableTypes sheafTypes )
		{
			ValidationUtils.ListHasContent( sheafTypes, "sheafTypes" );

			StringBuilder sb = new StringBuilder();

			sb.Append( "AND (" );

			int count = sheafTypes.Count;
			for( int i = 0; i < count; i++ )
			{
				string template = (i <= 0) ? "({0} = {1})" : " OR ({0} = {1})";

				sb.AppendFormat( template, DatabaseColumns.SheafTypeName, DatabaseParameters.GetSheafTypeName( i ) );
			}

			sb.Append( ")" );

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the all the speckle filter checksum columns.
		/// </summary>
		/// <returns>Returns the created script fragment.</returns>
		public override string GetSpeckleFilterChecksumColumnList()
		{
			StringBuilder sb = new StringBuilder();

			for( int i = 0; i < CoreDocumentsConsts.SpeckleFilterSizeMax; i++ )
				sb.AppendFormat( ", {0}", DatabaseColumns.GetSpeckleFilterChecksum( i ) );

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the all the speckle filter type name columns.
		/// </summary>
		/// <returns>Returns the created script fragment.</returns>
		public override string GetSpeckleFilterTypeNameColumnList()
		{
			StringBuilder sb = new StringBuilder();

			for( int i = 0; i < CoreDocumentsConsts.SpeckleFilterSizeMax; i++ )
				sb.AppendFormat( ", {0}", DatabaseColumns.GetSpeckleFilterTypeName( i ) );

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the all the speckle filter name columns.
		/// </summary>
		/// <returns>Returns the created script fragment.</returns>
		public override string GetSpeckleFilterNameColumnList()
		{
			StringBuilder sb = new StringBuilder();

			for( int i = 0; i < CoreDocumentsConsts.SpeckleFilterSizeMax; i++ )
				sb.AppendFormat( ", {0}", DatabaseColumns.GetSpeckleFilterName( i ) );

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the all the speckle filter value columns.
		/// </summary>
		/// <returns>Returns the created script fragment.</returns>
		public override string GetSpeckleFilterValueColumnList()
		{
			StringBuilder sb = new StringBuilder();

			for( int i = 0; i < CoreDocumentsConsts.SpeckleFilterSizeMax; i++ )
				sb.AppendFormat( ", {0}", DatabaseColumns.GetSpeckleFilterValue( i ) );

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the all the speckle filter checksum parameters.
		/// </summary>
		/// <returns>Returns the created script fragment.</returns>
		public override string GetSpeckleFilterChecksumParameterList()
		{
			StringBuilder sb = new StringBuilder();

			for( int i = 0; i < CoreDocumentsConsts.SpeckleFilterSizeMax; i++ )
				sb.AppendFormat( ", {0}", DatabaseParameters.GetSpeckleFilterChecksum( i ) );

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the all the speckle filter type name parameters.
		/// </summary>
		/// <returns>Returns the created script fragment.</returns>
		public override string GetSpeckleFilterTypeNameParameterList()
		{
			StringBuilder sb = new StringBuilder();

			for( int i = 0; i < CoreDocumentsConsts.SpeckleFilterSizeMax; i++ )
				sb.AppendFormat( ", {0}", DatabaseParameters.GetSpeckleFilterTypeName( i ) );

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the all the speckle filter name parameters.
		/// </summary>
		/// <returns>Returns the created script fragment.</returns>
		public override string GetSpeckleFilterNameParameterList()
		{
			StringBuilder sb = new StringBuilder();

			for( int i = 0; i < CoreDocumentsConsts.SpeckleFilterSizeMax; i++ )
				sb.AppendFormat( ", {0}", DatabaseParameters.GetSpeckleFilterName( i ) );

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the all the speckle filter value parameters.
		/// </summary>
		/// <returns>Returns the created script fragment.</returns>
		public override string GetSpeckleFilterValueParameterList()
		{
			StringBuilder sb = new StringBuilder();

			for( int i = 0; i < CoreDocumentsConsts.SpeckleFilterSizeMax; i++ )
				sb.AppendFormat( ", {0}", DatabaseParameters.GetSpeckleFilterValue( i ) );

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the all the speckle filter checksum assignments of parameters to columns.
		/// </summary>
		/// <returns>Returns the created script fragment.</returns>
		public override string GetSpeckleFilterChecksumColumnParameterList()
		{
			StringBuilder sb = new StringBuilder();

			for( int i = 0; i < CoreDocumentsConsts.SpeckleFilterSizeMax; i++ )
			{
				sb.AppendFormat( ", {0} = {1}", DatabaseColumns.GetSpeckleFilterChecksum( i )
					, DatabaseParameters.GetSpeckleFilterChecksum( i ) );
			}

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the all the speckle filter type name assignments of parameters to columns.
		/// </summary>
		/// <returns>Returns the created script fragment.</returns>
		public override string GetSpeckleFilterTypeNameColumnParameterList()
		{
			StringBuilder sb = new StringBuilder();

			for( int i = 0; i < CoreDocumentsConsts.SpeckleFilterSizeMax; i++ )
			{
				sb.AppendFormat( ", {0} = {1}", DatabaseColumns.GetSpeckleFilterTypeName( i )
					, DatabaseParameters.GetSpeckleFilterTypeName( i ) );
			}

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the all the speckle filter name assignments of parameters to columns.
		/// </summary>
		/// <returns>Returns the created script fragment.</returns>
		public override string GetSpeckleFilterNameColumnParameterList()
		{
			StringBuilder sb = new StringBuilder();

			for( int i = 0; i < CoreDocumentsConsts.SpeckleFilterSizeMax; i++ )
			{
				sb.AppendFormat( ", {0} = {1}", DatabaseColumns.GetSpeckleFilterName( i )
					, DatabaseParameters.GetSpeckleFilterName( i ) );
			}

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the all the speckle filter value assignments of parameters to columns.
		/// </summary>
		/// <returns>Returns the created script fragment.</returns>
		public override string GetSpeckleFilterValueColumnParameterList()
		{
			StringBuilder sb = new StringBuilder();

			for( int i = 0; i < CoreDocumentsConsts.SpeckleFilterSizeMax; i++ )
			{
				sb.AppendFormat( ", {0} = {1}", DatabaseColumns.GetSpeckleFilterValue( i )
					, DatabaseParameters.GetSpeckleFilterValue( i ) );
			}

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the condition to load information from a database, according to the
		/// specified speckle filter checksums.
		/// </summary>
		/// <param name="filters">The speckle filters to include in the script fragment. This may be null or empty.</param>
		/// <returns>Returns the created script fragment. This may be empty, but not null.</returns>
		public override string GetSpeckleFilterChecksumSeek( SpeckleFilters filters )
		{
			if( ListUtils.NoContent( filters ) ) return "";

			StringBuilder sb = new StringBuilder();

			sb.Append( "AND (" );

			bool first = true;
			foreach( SpeckleFilter filter in filters )
			{
				string prefix = first ? "" : " AND "; first = false;
				string template = (filter.Value != null) ? "{0}({1} = {2})" : "{0}({1} IS NULL)";

				sb.AppendFormat( template, prefix, DatabaseColumns.GetSpeckleFilterChecksum( filter.Identifier.Index )
					, DatabaseParameters.GetSpeckleFilterChecksum( filter.Identifier.Index ) );
			}

			sb.Append( ")" );

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the condition to load information from a database, about the speckle filter
		/// values, according to the specified speckle filters.
		/// </summary>
		/// <param name="filters">The speckle filters to include in the script fragment. This may be null or empty.</param>
		/// <returns>Returns the created script fragment. This may be empty, but not null.</returns>
		public override string GetSpeckleFilterValueSeek( SpeckleFilters filters )
		{
			if( ListUtils.NoContent( filters ) ) return "";

			StringBuilder sb = new StringBuilder();

			sb.Append( "AND (" );

			bool first = true;
			foreach( SpeckleFilter filter in filters )
			{
				string prefix = first ? "" : " AND "; first = false;
				string template = (filter.Value != null) ? "{0}({1} = {2})" : "{0}({1} IS NULL)";

				sb.AppendFormat( template, prefix, DatabaseColumns.GetSpeckleFilterValue( filter.Identifier.Index )
					, DatabaseParameters.GetSpeckleFilterValue( filter.Identifier.Index ) );
			}

			sb.Append( ")" );

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the speckle filter type names to load from a database, according to the
		/// specified speckle filter identifiers.
		/// </summary>
		/// <param name="filterIdentifiers">The speckle filters to include in the script fragment. This may be null or empty.
		/// </param>
		/// <returns>Returns the created script fragment. This may be empty, but not null.</returns>
		public override string GetSpeckleFilterTypeNameLoad( SpeckleFilterIdentifiers filterIdentifiers )
		{
			if( ListUtils.NoContent( filterIdentifiers ) ) return "";

			StringBuilder sb = new StringBuilder();

			foreach( SpeckleFilterIdentifier identifier in filterIdentifiers )
				sb.AppendFormat( ", {0}", DatabaseColumns.GetSpeckleFilterTypeName( identifier.Index ) );

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the speckle filters to load from a database, about the speckle filter
		/// names, according to the specified speckle filter identifiers.
		/// </summary>
		/// <param name="filterIdentifiers">The speckle filter identifiers to include in the script fragment. This may be null or
		/// empty.</param>
		/// <returns>Returns the created script fragment. This may be empty, but not null.</returns>
		public override string GetSpeckleFilterNameLoad( SpeckleFilterIdentifiers filterIdentifiers )
		{
			if( ListUtils.NoContent( filterIdentifiers ) ) return "";

			StringBuilder sb = new StringBuilder();

			foreach( SpeckleFilterIdentifier identifier in filterIdentifiers )
				sb.AppendFormat( ", {0}", DatabaseColumns.GetSpeckleFilterName( identifier.Index ) );

			return sb.ToString();
		}

		/// <summary>
		/// Gets the script fragment which contains the speckle filters to load from a database, about the speckle filter
		/// values, according to the specified speckle filter identifiers.
		/// </summary>
		/// <param name="filterIdentifiers">The speckle filter identifiers to include in the script fragment. This may be null or
		/// empty.</param>
		/// <returns>Returns the created script fragment. This may be empty, but not null.</returns>
		public override string GetSpeckleFilterValueLoad( SpeckleFilterIdentifiers filterIdentifiers )
		{
			if( ListUtils.NoContent( filterIdentifiers ) ) return "";

			StringBuilder sb = new StringBuilder();

			foreach( SpeckleFilterIdentifier identifier in filterIdentifiers )
				sb.AppendFormat( ", {0}", DatabaseColumns.GetSpeckleFilterValue( identifier.Index ) );

			return sb.ToString();
		}

		#endregion Public methods
	}
}
