////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.IO;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.StarkDocuments
{
	/// <summary>
	/// An attesting document is a signed document which attests a file, and is a satellite of that file.
	/// </summary>
	public class AttestingDocument : SignedDocument
	{
		#region Public constants

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="InitiatorUpnFingerprint"/> property.
		/// </summary>
		public const string InitiatorUpnFingerprintFilterIdentifier = "AttestingDocument.InitiatorUpnFingerprint#4";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AttestingDocument()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		public AttestingDocument( Locator instanceLocator, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg
				, IHashSerializer signatureHashSerializer, Locator initiatorUpnFingerprint )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer )
		{
			InitiatorUpnFingerprint = initiatorUpnFingerprint;
		}

		#endregion Constructors

		#region Public serializable properties

		private Locator _InitiatorUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the initiator of the current object. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( InitiatorUpnFingerprintFilterIdentifier )]
		public Locator InitiatorUpnFingerprint
		{
			get { return _InitiatorUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "InitiatorUpnFingerprint" );

				_InitiatorUpnFingerprint = value;
			}
		}

		private Int64 _OriginalDataSize;
		/// <summary>
		/// The size (in bytes) of the original (/ unencrypted) information. This must be at least 0. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Int64 OriginalDataSize
		{
			get { return _OriginalDataSize; }
			protected set
			{
				ValidationUtils.Integer( value, 0, Int64.MaxValue, "OriginalDataSize" );

				_OriginalDataSize = value;
			}
		}

		private byte[] _AttestedMacValue;
		/// <summary>
		/// The MAC of the attested file.
		/// </summary>
		[SerializeAttribute]
		public byte[] AttestedMacValue
		{
			get { return _AttestedMacValue; }
			set
			{
				ValidationUtils.ArrayHasContent( value, "AttestedMacValue" );

				_AttestedMacValue = value;
			}
		}

		private string _Description;
		/// <summary>
		/// The description of the attested file. This may be null or empty.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string Description
		{
			get { return _Description; }
			set { _Description = value; }
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Verifies if the specified MAC value of an attested document is the same as the one from the current attesting
		/// document. An exception is thrown if the MACs are different.
		/// </summary>
		/// <param name="attestedMacValue">The MAC value of the attested document.</param>
		private void VerifyDocumentMac( byte[] attestedMacValue )
		{
			ValidationUtils.ArrayHasContent( attestedMacValue, "attestedMacValue" );

			if( !BufferUtils.Equals( attestedMacValue, AttestedMacValue ) )
			{
				throw new ValidationException( "The MAC of the attested document is different than the one from the attesting"
						+ " document.", null, null
					, new ExceptionDetail( "Computed MAC", TextUtils.ArrayToMoteText16( attestedMacValue ) )
					, new ExceptionDetail( "Stored MAC", TextUtils.ArrayToMoteText16( AttestedMacValue ) ) );
			}
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Verifies the signature of the specified document; the attested MAC is also verified. An
		/// <see cref="InvalidSignatureException"/> (among others) is thrown if the signature is invalid.
		/// </summary>
		/// <param name="identityPublisher">The identity publisher with which the document's signatures are checked.</param>
		/// <param name="attestedMacValue">The MAC value of the attested document.</param>
		public void VerifySignature( IIdentityPublisher identityPublisher, byte[] attestedMacValue )
		{
			VerifyDocumentMac( attestedMacValue );
			VerifySignature( identityPublisher );
		}

		/// <summary>
		/// Verifies the signature of the specified document; the attested MAC is also verified. An
		/// <see cref="InvalidSignatureException"/> (among others) is thrown if the signature is invalid.
		/// </summary>
		/// <param name="identity">The public identity with which the document's signature is checked.</param>
		/// <param name="attestedMacValue">The MAC value of the attested document.</param>
		public void VerifySignature( PublicIdentity identity, byte[] attestedMacValue )
		{
			VerifyDocumentMac( attestedMacValue );
			VerifySignature( identity );
		}

		/// <summary>
		/// Verifies the signature of the specified document; the attested MAC is also verified. An
		/// <see cref="InvalidSignatureException"/> (among others) is thrown if the signature is invalid.
		/// </summary>
		/// <param name="identityPublishers">The identity publishers with which the document's signatures are checked.</param>
		/// <param name="attestedMacValue">The MAC value of the attested document.</param>
		/// <param name="signingLocatorTimes">List where to put the locators of the entities which have actually signed the
		/// document, and the creation times of their signatures. If an item with the same key already exists, it's updated.
		/// This may be empty, but not null.</param>
		public void CollectValidSignatures( Dictionary<Locator, IIdentityPublisher> identityPublishers, byte[] attestedMacValue
			, LocatorTimes signingLocatorTimes )
		{
			VerifyDocumentMac( attestedMacValue );
			CollectValidSignatures( identityPublishers, signingLocatorTimes );
		}

		/// <summary>
		/// Verifies the signature of the specified document; the attested MAC is also verified. An
		/// <see cref="InvalidSignatureException"/> (among others) is thrown if the signature is invalid.
		/// </summary>
		/// <param name="identities">The identities with which the document's signatures are checked.</param>
		/// <param name="attestedMacValue">The MAC value of the attested document.</param>
		/// <param name="signingLocatorTimes">List where to put the locators of the entities which have actually signed the
		/// document, and the creation times of their signatures. If an item with the same key already exists, it's updated.
		/// This may be empty, but not null.</param>
		public void CollectValidSignatures( Dictionary<Locator, PublicIdentity> identities, byte[] attestedMacValue
			, LocatorTimes signingLocatorTimes )
		{
			VerifyDocumentMac( attestedMacValue );
			CollectValidSignatures( identities, signingLocatorTimes );
		}

		/// <summary>
		/// Computes the MAC of the specified stream.
		/// </summary>
		/// <param name="stream">Stream whose MAC to compute.</param>
		/// <param name="lop">Long operation progress used during the computation of the MAC. This may be null.</param>
		/// <returns>Returns the MAC of the specified stream.</returns>
		public virtual byte[] Process( Stream stream, LongOperationProgress lop )
		{
			ValidationUtils.ObjectHasContent( stream, "stream" );

			if( !stream.CanSeek || (stream.Length <= 0) )
				throw new ValidationException( "The stream whose MAC to compute is empty or its size is unknown." );

			if( lop != null ) lop.Progress( 0 );

			OriginalDataSize = stream.Length;

			byte[] macKey = MacAlg.HashAlg.Compute( null, true, InstanceLocator, InitiatorUpnFingerprint.Value );

			MacAlgStreamWorker macWorker = MacAlg.CreateStreamWorker( macKey );
			Int64 currentDataSize = 0;

			// Hash-in the data size.
			byte[] totalSizeBuffer = BitConverter.GetBytes( OriginalDataSize );
			macWorker.Process( totalSizeBuffer );

			byte[] buffer = new byte[ KernelConsts.FileBufferSize ];

			// Stop processing the first time the stream read returns fewer bytes than the size of the buffer.
			while( true )
			{
				int readCount = stream.Read( buffer, 0, KernelConsts.FileBufferSize );

				// Hash-in the data.
				macWorker.Process( buffer, 0, readCount );
				currentDataSize += readCount;

				if( lop != null ) lop.Progress( (double)currentDataSize / OriginalDataSize );

				if( readCount < KernelConsts.FileBufferSize ) break;
			}

			if( OriginalDataSize != currentDataSize )
			{
				throw new ValidationException( string.Format( "The size of the data must be {0} not {1}."
					, OriginalDataSize, currentDataSize ) );
			}

			// Compute the MAC.
			byte[] mac = macWorker.ProcessFinal();

			if( lop != null ) lop.Progress( 1 );

			return mac;
		}

		#endregion Public methods
	}
}
