////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Collections.Generic;

using Vela.Kernel;

namespace Vela.StarkDocuments
{
	/// <summary>
	/// Emoticon manager.
	/// </summary>
	public class EmoticonManager
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected EmoticonManager()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="configuration">Configuration.</param>
		public EmoticonManager( EmoticonManagerConfiguration configuration )
		{
			Configuration = configuration;
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="configurationData">Configuration data.</param>
		/// <param name="customEmoticonsFolderName">The name of the folder which contains the custom emoticons.</param>
		public EmoticonManager( byte[] configurationData, string customEmoticonsFolderName )
		{
			Configuration = EmoticonManagerConfiguration.Load( configurationData, customEmoticonsFolderName );
		}

		#endregion Constructors

		#region Public non-serializable properties

		private static EmoticonManagerConfiguration _Configuration;
		/// <summary>
		/// Configuration. READ-ONLY.
		/// </summary>
		public EmoticonManagerConfiguration Configuration
		{
			get { return _Configuration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Configuration" );

				_Configuration = value;
			}
		}

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Gets the emoticon descriptor of the specified emoticon symbol.
		/// </summary>
		/// <param name="emoticonSymbol">Emoticon symbol whose emoticon descriptor to get. This may be null or empty.</param>
		/// <returns>Returns the emoticon descriptor. This may be null.</returns>
		public EmoticonSymbolDescriptor GetEmoticonDescriptorFromSymbol( string emoticonSymbol )
		{
			if( TextUtils.NoContent( emoticonSymbol ) ) return null;

			return Configuration.SymbolDescriptors.Find( emoticonSymbol );
		}

		/// <summary>
		/// Gets the emoticon symbols from the specified text.
		/// </summary>
		/// <param name="message">Message from which to get the emoticon symbols. This may be null or empty.</param>
		/// <param name="limit">The maximum number of emoticon symbols to return. If this is 0 or a negative number, it's
		/// ignored.</param>
		/// <returns>Returns the found emoticon symbols. This may be empty, but not null.</returns>
		public List<string> GetEmoticonSymbols( string message, int limit )
		{
			List<string> foundEmoticonSymbols = new List<string>();

			if( TextUtils.NoContent( message ) ) return foundEmoticonSymbols;

			int endPos = message.Length - 1;
			int foundCount = 0;

			bool mustContinue = true;
			while( mustContinue && (endPos >= 0) )
			{
				int startPos = TextUtils.LastIndexOfWhitespace( message, endPos );

				if( startPos < 0 )
				{
					startPos = 0;
					mustContinue = false;
				}
				else
				{
					startPos++;
				}

				int emoticonSymbolSize = endPos - startPos + 1;

				if( (emoticonSymbolSize > 0) && (emoticonSymbolSize <= 20) )
				{
					string emoticonSymbol = message.Substring( startPos, emoticonSymbolSize );

					if( Configuration.SymbolDescriptors.ContainsKey( emoticonSymbol ) )
					{
						foundEmoticonSymbols.Add( emoticonSymbol );
						foundCount++;

						if( (limit > 0) && (foundCount >= limit) ) break;
					}
				}

				endPos = startPos - 2;
			}

			// Since the search was performed from the end toward the beginning, the emoticon symbols were added in reverse
			// order, so reverse it again to get to the intended order.
			if( foundEmoticonSymbols.Count > 1 )
			{
				List<string> orderedSymbols = new List<string>();

				for( int i = foundEmoticonSymbols.Count - 1; i >= 0; i-- ) orderedSymbols.Add( foundEmoticonSymbols[ i ] );

				return orderedSymbols;
			}

			return foundEmoticonSymbols;
		}

		/// <summary>
		/// Gets the emoticon descriptors of the emoticon symbols from the specified text.
		/// </summary>
		/// <param name="message">Message from which to get the emoticon symbols. This may be null or empty.</param>
		/// <param name="limit">The maximum number of emoticon symbols to return. If this is 0 or a negative number, it's
		/// ignored.</param>
		/// <returns>Returns the emoticon descriptors of the found emoticon symbols. This may be empty, but not null.</returns>
		public List<EmoticonSymbolDescriptor> GetEmoticonDescriptors( string message, int limit )
		{
			List<EmoticonSymbolDescriptor> emoticonSymbolDescriptors = new List<EmoticonSymbolDescriptor>();

			List<string> emoticonSymbols = GetEmoticonSymbols( message, limit );

			foreach( string emoticonSymbol in emoticonSymbols )
			{
				EmoticonSymbolDescriptor descriptor = Configuration.SymbolDescriptors[ emoticonSymbol ];
				if( descriptor == null ) continue;

				emoticonSymbolDescriptors.Add( descriptor );
			}

			return emoticonSymbolDescriptors;
		}

		/// <summary>
		/// Gets the emoticon symbol from the specified text.
		/// </summary>
		/// <param name="message">Message from which to get the emoticon symbol. This may be null or empty.</param>
		/// <returns>Returns the found emoticon symbol. This is null if "message" doesn't contain an emoticon symbol.</returns>
		public string GetEmoticonSymbol( string message )
		{
			List<string> emoticonSymbols = GetEmoticonSymbols( message, 1 );
			if( ListUtils.NoContent( emoticonSymbols ) ) return null;

			return emoticonSymbols[ 0 ];
		}

		/// <summary>
		/// Gets the emoticon descriptor of the emoticon symbol from the specified text.
		/// </summary>
		/// <param name="message">Message from which to get the emoticon symbol. This may be null or empty.</param>
		/// <returns>Returns the emoticon descriptor of the found emoticon symbol. This is null if "message" doesn't contain an
		/// emoticon symbol.</returns>
		public EmoticonSymbolDescriptor GetEmoticonDescriptor( string message )
		{
			List<EmoticonSymbolDescriptor> emoticonDescriptors = GetEmoticonDescriptors( message, 1 );
			if( ListUtils.NoContent( emoticonDescriptors ) ) return null;

			return emoticonDescriptors[ 0 ];
		}

		/// <summary>
		/// Checks if the specified text contains an emoticon symbol whose associated command is
		/// <see cref="EmoticonCommands.Wallpaper"/>.
		/// </summary>
		/// <param name="message">Message from which to get the emoticon symbol. This may be null or empty.</param>
		/// <returns>Returns true if the specified text contains an emoticon symbol whose associated command is
		/// <see cref="EmoticonCommands.Wallpaper"/>.</returns>
		public bool ContainsSendWallpaperEmoticonSymbol( string message )
		{
			List<EmoticonSymbolDescriptor> emoticonDescriptors = GetEmoticonDescriptors( message, 0 );

			foreach( EmoticonSymbolDescriptor emoticonDescriptor in emoticonDescriptors )
			{
				if( emoticonDescriptor.EmoticonCommand == EmoticonCommands.Wallpaper ) return true;
			}

			return false;
		}

		/// <summary>
		/// Gets the file name of the emoticon full-sized image from the specified text.
		/// </summary>
		/// <param name="message">Message from which to get the emoticon symbol and full-sized image file name. This may be null
		/// or empty.</param>
		/// <returns>Returns the file name of the emoticon full-sized image. This may be null or empty.</returns>
		public string GetEmoticonFullSizedImageFileName( string message )
		{
			EmoticonSymbolDescriptor descriptor = GetEmoticonDescriptor( message );
			if( descriptor == null ) return null;

			return descriptor.FullSizedImageFileName;
		}

		/// <summary>
		/// Gets the file name of the emoticon thumb-sized image for the emoticon symbol from the specified text.
		/// </summary>
		/// <param name="message">Message from which to get the emoticon symbol and thumb-sized image file name. This may be null
		/// or empty.</param>
		/// <returns>Returns the file name of the emoticon thumb-sized image. This may be null or empty.</returns>
		public string GetEmoticonThumbSizedImageFileName( string message )
		{
			EmoticonSymbolDescriptor descriptor = GetEmoticonDescriptor( message );
			if( descriptor == null ) return null;

			return descriptor.ThumbSizedImageFileName;
		}

		#endregion Public methods
	}
}
