////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.StarkDocuments
{
	/// <summary>
	/// Emoticon manager configuration.
	/// </summary>
	public class EmoticonManagerConfiguration : IDocument
	{
		#region Public constants

		/// <summary>
		/// The display category for custom emoticons.
		/// </summary>
		public const string CustomDisplayCategory = "Custom";

		/// <summary>
		/// The placeholders which may be used in the file title of a custom emoticon to represent various characters.
		/// </summary>
		public static readonly Dictionary<string, string> CustomEmoticonFileTitlePlaceholders = new Dictionary<string, string>()
		{
			{ " ", "" }, { "{Apostrophe}", "'" }, { "{Backslash}", "\\" }, { "{Close}", ")" }, { "{Colon}", ":" }
			, { "{Greater}", ">" }, { "{Open}", "(" }, { "{Pipe}", "|" }, { "{Question}", "?" }, { "{Quote}", "\"" }
			, { "{Quotes}", "\"" }, { "{Semicolon}", ";" }, { "{Slash}", "/" }, { "{Smaller}", "<" }, { "{Star}", "*" }
		};

		/// <summary>
		/// The maximum number of custom emoticons to load.
		/// </summary>
		public const int CustomEmoticonsCountMax = 300;

		/// <summary>
		/// The maximum number of emoticons with the <see cref="CustomDisplayCategory"/> display category.
		/// </summary>
		public const int EmoticonsWithCustomDisplayCategoryCountMax = 100;

		/// <summary>
		/// The maximum size (in bytes) of a custom emoticon image file.
		/// </summary>
		public const int CustomEmoticonFileSizeMax = 5000000;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected EmoticonManagerConfiguration()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="symbolDescriptors">List of emoticon symbol descriptors.</param>
		/// <param name="displayCategoryToShowFirst">Display category to show first to the user, in a window where the user can
		/// select an emoticon.</param>
		/// <param name="sensualDisplayCategory">Display category with sensual emoticons, category which is hidden from the user
		/// if the user is not adult, in a window where the user can select an emoticon.</param>
		public EmoticonManagerConfiguration( EmoticonSymbolDescriptors symbolDescriptors, string displayCategoryToShowFirst
			, string sensualDisplayCategory )
		{
			SymbolDescriptors = symbolDescriptors;
			DisplayCategoryToShowFirst = displayCategoryToShowFirst;
			SensualDisplayCategory = sensualDisplayCategory;
		}

		#endregion Constructors

		#region Public serializable properties

		private EmoticonSymbolDescriptors _SymbolDescriptors;
		/// <summary>
		/// List of emoticon symbol descriptors. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public EmoticonSymbolDescriptors SymbolDescriptors
		{
			get { return _SymbolDescriptors; }
			protected set
			{
				ValidationUtils.ListHasContent( value, "SymbolDescriptors" );

				_SymbolDescriptors = value;
			}
		}

		private string _DisplayCategoryToShowFirst;
		/// <summary>
		/// Display category to show first to the user, in a window where the user can select an emoticon. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string DisplayCategoryToShowFirst
		{
			get { return _DisplayCategoryToShowFirst; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "DisplayCategoryToShowFirst" );

				_DisplayCategoryToShowFirst = value;
			}
		}

		private string _SensualDisplayCategory;
		/// <summary>
		/// Display category with sensual emoticons, category which is hidden from the user if the user is not adult, in a window
		/// where the user can select an emoticon. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string SensualDisplayCategory
		{
			get { return _SensualDisplayCategory; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "SensualDisplayCategory" );

				_SensualDisplayCategory = value;
			}
		}

		#endregion Public serializable properties

		#region Private static methods

		/// <summary>
		/// Gets the emoticon symbol from the specified file name.
		/// </summary>
		/// <param name="fileName">File name from which to get the emoticon symbol. This is not validated.</param>
		/// <returns>Returns the emoticon symbol.</returns>
		private static string GetCustomEmoticonSymbol( string fileName )
		{
			string symbol = Path.GetFileNameWithoutExtension( fileName );

			foreach( KeyValuePair<string, string> kvp in CustomEmoticonFileTitlePlaceholders )
			{
				symbol = TextUtils.Replace( symbol, kvp.Key, kvp.Value );
			}

			return symbol;
		}

		/// <summary>
		/// Gets the number of emoticons with the <see cref="CustomDisplayCategory"/> display category.
		/// </summary>
		/// <param name="symbolDescriptors">List of emoticon symbol descriptors.</param>
		/// <returns>Returns the number of emoticons with the <see cref="CustomDisplayCategory"/> display category.</returns>
		private static int CountEmoticonsWithCustomDisplayCategory( EmoticonSymbolDescriptors symbolDescriptors )
		{
			ValidationUtils.ListHasContent( symbolDescriptors, "symbolDescriptors" );

			int count = 0;

			foreach( EmoticonSymbolDescriptor symbolDescriptor in symbolDescriptors )
			{
				if( TextUtils.Equals( symbolDescriptor.DisplayCategory, CustomDisplayCategory ) )
					count++;
			}

			return count;
		}

		/// <summary>
		/// Removes the emoticons with the <see cref="CustomDisplayCategory"/> display category. The overridden ones with
		/// predefined categories remain overridden.
		/// </summary>
		/// <param name="symbolDescriptors">List of emoticon symbol descriptors.</param>
		private static void RemoveEmoticonsWithCustomDisplayCategory( EmoticonSymbolDescriptors symbolDescriptors )
		{
			ValidationUtils.ListHasContent( symbolDescriptors, "symbolDescriptors" );

			List<string> emoticonSymbolToRemove = new List<string>();
			foreach( EmoticonSymbolDescriptor symbolDescriptor in symbolDescriptors )
			{
				if( TextUtils.Equals( symbolDescriptor.DisplayCategory, CustomDisplayCategory ) )
					emoticonSymbolToRemove.Add( symbolDescriptor.Symbol );
			}

			foreach( string emoticonSymbol in emoticonSymbolToRemove )
				symbolDescriptors.Remove( emoticonSymbol );
		}

		/// <summary>
		/// Restores the file names for the emoticon full-sized and thumb-sized images, using the backup file names.
		/// </summary>
		/// <param name="symbolDescriptors">List of emoticon symbol descriptors.</param>
		private static void RestoreEmoticonFileNames( EmoticonSymbolDescriptors symbolDescriptors )
		{
			ValidationUtils.ListHasContent( symbolDescriptors, "symbolDescriptors" );

			foreach( EmoticonSymbolDescriptor symbolDescriptor in symbolDescriptors )
				symbolDescriptor.RestoreFileNames();
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Loads the configuration.
		/// </summary>
		/// <param name="buffer">Buffer from which to load the configuration</param>
		/// <param name="customEmoticonsFolderName">The name of the folder which contains the custom emoticons.</param>
		/// <returns>Returns the loaded configuration object.</returns>
		public static EmoticonManagerConfiguration Load( byte[] buffer, string customEmoticonsFolderName )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );
			ValidationUtils.TextHasContent( customEmoticonsFolderName, "customEmoticonsFolderName" );

			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				IDeserializer deserializer;
				EmoticonManagerConfiguration configuration = SerializationUtils.LoadFromBuffer( buffer
					, typeof( EmoticonManagerConfiguration ), out deserializer ) as EmoticonManagerConfiguration;

				// Update the emoticons with the custom ones.
				Strings fileNames = new Strings();
				FileUtils.CollectFileNamesFromFolder( customEmoticonsFolderName, fileNames );
				configuration.UpdateWithCustomEmoticons( fileNames );

				return configuration;
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		/// <summary>
		/// Gets the valid emoticon symbol from the specified file name.
		/// </summary>
		/// <param name="fileName">File name from which to get the valid emoticon symbol.</param>
		/// <returns>Returns the valid emoticon symbol, or null if the specified file name doesn't contain a valid emoticon
		/// symbol.</returns>
		public static string GetValidCustomEmoticonSymbol( string fileName )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			if( !FileExtensions.IsImageFile( fileName ) )
				return null;

			Int64 fileSize = FileUtils.GetFileSize( fileName );
			if( (fileSize <= 0) || (fileSize > CustomEmoticonFileSizeMax) )
				return null;

			string customSymbol = GetCustomEmoticonSymbol( fileName );

			if( ValidationUtils.Text( customSymbol, -1, -1, TextValidations.EmoticonSymbol, null, null, null, false ) != null )
				return null;

			return customSymbol;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Updates the current configuration with the custom emoticons.
		/// </summary>
		/// <param name="fileNames">The name of the files of the custom emoticons. If this contains folder names, the content of
		/// those folders is processed recursively. This may be null or empty.</param>
		public void UpdateWithCustomEmoticons( Strings fileNames )
		{
			RemoveEmoticonsWithCustomDisplayCategory( SymbolDescriptors );
			RestoreEmoticonFileNames( SymbolDescriptors );

			if( ListUtils.NoContent( fileNames ) )
				return;

			int customEmoticonsCount = 0;
			int emoticonsWithCustomDisplayCategoryCount = CountEmoticonsWithCustomDisplayCategory( SymbolDescriptors );

			foreach( string fileName in fileNames )
			{
				string customSymbol = GetValidCustomEmoticonSymbol( fileName );
				if( TextUtils.NoContent( customSymbol ) )
					continue;

				EmoticonSymbolDescriptor existingDescriptor = SymbolDescriptors[ customSymbol ];

				if( existingDescriptor != null )
				{
					// Update an emoticon with an existing symbol. The default emoticon (full-sized and thumb-sized) images are
					// being overridden by the custom one.

					existingDescriptor.ReplaceFileNames( fileName, fileName );
				}
				else
				{
					// Add an emoticon with a new symbol.

					emoticonsWithCustomDisplayCategoryCount++;
					if( emoticonsWithCustomDisplayCategoryCount <= EmoticonsWithCustomDisplayCategoryCountMax )
					{
						EmoticonSymbolDescriptor descriptor = new EmoticonSymbolDescriptor( customSymbol, CustomDisplayCategory
							, EmoticonCommands.None, fileName, fileName );

						SymbolDescriptors.Add( descriptor );
					}
				}

				customEmoticonsCount++;
				if( customEmoticonsCount >= CustomEmoticonsCountMax )
					break;
			}
		}

		#endregion Public methods
	}
}
