////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.StarkDocuments
{
	/// <summary>
	/// Emoticon symbol descriptor.
	/// </summary>
	public class EmoticonSymbolDescriptor : IDocument
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected EmoticonSymbolDescriptor()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="symbol">Emoticon symbol.</param>
		/// <param name="displayCategory">The display category. This may be null or empty.</param>
		/// <param name="emoticonCommand">The associated emoticon command.</param>
		/// <param name="fullSizedImageFileName">File name for the emoticon full-sized image.</param>
		/// <param name="thumbSizedImageFileName">File name for the emoticon thumb-sized image.</param>
		public EmoticonSymbolDescriptor( string symbol, string displayCategory, EmoticonCommands emoticonCommand
			, string fullSizedImageFileName, string thumbSizedImageFileName )
		{
			Symbol = symbol;
			DisplayCategory = displayCategory;
			EmoticonCommand = emoticonCommand;
			FullSizedImageFileName = fullSizedImageFileName;
			ThumbSizedImageFileName = thumbSizedImageFileName;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// Backup for the file name for the emoticon full-sized image.
		/// </summary>
		private string BackupFullSizedImageFileName;

		/// <summary>
		/// Backup for the file name for the emoticon thumb-sized image.
		/// </summary>
		private string BackupThumbSizedImageFileName;

		#endregion Private fields

		#region Public non-serializable properties

		/// <summary>
		/// If this is true then the emoticon has an associated emoticon command. READ-ONLY.
		/// </summary>
		public bool HasEmoticonCommand
		{
			get { return EmoticonCommand != EmoticonCommands.None; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private string _Symbol;
		/// <summary>
		/// Emoticon symbol. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string Symbol
		{
			get { return _Symbol; }
			protected set
			{
				ValidationUtils.Text( value, -1, -1, TextValidations.EmoticonSymbol, "Symbol" );

				_Symbol = value;
			}
		}

		private string _DisplayCategory;
		/// <summary>
		/// The display category. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string DisplayCategory
		{
			get { return _DisplayCategory; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "DisplayCategory" );

				_DisplayCategory = value;
			}
		}

		private EmoticonCommands _EmoticonCommand;
		/// <summary>
		/// The associated emoticon command. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public EmoticonCommands EmoticonCommand
		{
			get { return _EmoticonCommand; }
			protected set { _EmoticonCommand = value; }
		}

		private string _FullSizedImageFileName;
		/// <summary>
		/// File name for the emoticon full-sized image. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The backup file name is overwritten.</para>
		/// </remarks>
		[SerializeAttribute]
		public string FullSizedImageFileName
		{
			get { return _FullSizedImageFileName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "FullSizedImageFileName" );

				_FullSizedImageFileName = value;
				BackupFullSizedImageFileName = value;
			}
		}

		private string _ThumbSizedImageFileName;
		/// <summary>
		/// File name for the emoticon thumb-sized image. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The backup file name is overwritten.</para>
		/// </remarks>
		[SerializeAttribute]
		public string ThumbSizedImageFileName
		{
			get { return _ThumbSizedImageFileName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "ThumbSizedImageFileName" );

				_ThumbSizedImageFileName = value;
				BackupThumbSizedImageFileName = value;
			}
		}

		#endregion Public serializable properties

		#region Public static methods

		/// <summary>
		/// Checks if the specified emoticon symbols are the same (when trimmed).
		/// </summary>
		/// <param name="emoticonSymbol1">Emoticon symbol to check. This may be null or empty.</param>
		/// <param name="emoticonSymbol2">Emoticon symbol to check. This may be null or empty.</param>
		/// <returns>Returns true if the specified emoticon symbols are the same (when trimmed). If the emoticon symbols are both
		/// missing, false is returned.</returns>
		public static bool EqualEmoticonSymbols( string emoticonSymbol1, string emoticonSymbol2 )
		{
			if( TextUtils.NoContent( emoticonSymbol1 ) || TextUtils.NoContent( emoticonSymbol2 ) ) return false;

			return TextUtils.Equals( emoticonSymbol1.Trim(), emoticonSymbol2.Trim() );
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Replaces the file names for the emoticon full-sized and thumb-sized images.
		/// </summary>
		/// <param name="fullSizedImageFileName">File name for the emoticon full-sized image.</param>
		/// <param name="thumbSizedImageFileName">File name for the emoticon thumb-sized image.</param>
		/// <remarks>
		/// <para>The backup file names are not overwritten.</para>
		/// </remarks>
		public void ReplaceFileNames( string fullSizedImageFileName, string thumbSizedImageFileName )
		{
			ValidationUtils.TextHasContent( fullSizedImageFileName, "fullSizedImageFileName" );
			ValidationUtils.TextHasContent( thumbSizedImageFileName, "thumbSizedImageFileName" );

			// Make sure that the backup file names are not overwritten.
			_FullSizedImageFileName = fullSizedImageFileName;
			_ThumbSizedImageFileName = thumbSizedImageFileName;
		}

		/// <summary>
		/// Restores the file names for the emoticon full-sized and thumb-sized images, using the backup file names.
		/// </summary>
		public void RestoreFileNames()
		{
			_FullSizedImageFileName = BackupFullSizedImageFileName;
			_ThumbSizedImageFileName = BackupThumbSizedImageFileName;
		}

		#endregion Public methods
	}
}
