////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;

namespace Vela.StarkDocuments
{
	/// <summary>
	/// List of emoticon symbol descriptors (<see cref="EmoticonSymbolDescriptor"/>).
	/// </summary>
	public class EmoticonSymbolDescriptors : KernelSortedList<string, EmoticonSymbolDescriptor>
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public EmoticonSymbolDescriptors()
			: base( StringInsensitiveComparer.Instance )
		{
		}

		#endregion Constructors

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para><see cref="EmoticonSymbolDescriptor.Symbol"/> is used as key.</para>
		/// </remarks>
		public override string GetKey( EmoticonSymbolDescriptor item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.Symbol;
		}

		/// <summary>
		/// Collects the sorted list items with the specified display category.
		/// </summary>
		/// <param name="displayCategory">The display category of the list items to collect. This may be null or empty.</param>
		/// <returns>Returns the collected list items. This may be empty, but not null.</returns>
		public EmoticonSymbolDescriptors CollectByDisplayCategory( string displayCategory )
		{
			EmoticonSymbolDescriptors list = new EmoticonSymbolDescriptors();

			foreach( EmoticonSymbolDescriptor descriptor in this )
			{
				if( TextUtils.NoContent( descriptor.DisplayCategory ) && TextUtils.NoContent( displayCategory ) )
				{
					list.Add( descriptor );
				}
				else
				{
					if( TextUtils.Equals( descriptor.DisplayCategory, displayCategory ) )
						list.Add( descriptor );
				}
			}

			return list;
		}

		/// <summary>
		/// Collects the display categories of the list items.
		/// </summary>
		/// <param name="displayCategoryToHide">The display category which has to be hidden from the user. This may be null or
		/// empty.</param>
		/// <returns>Returns the display categories of the list items. This may be empty, but not null.</returns>
		public Strings CollectDisplayCategories( string displayCategoryToHide )
		{
			Strings categories = new Strings();

			foreach( EmoticonSymbolDescriptor descriptor in this )
			{
				if( !TextUtils.Equals( descriptor.DisplayCategory, displayCategoryToHide ) )
					categories.AddOnlyNew( descriptor.DisplayCategory );
			}

			return categories;
		}

		#endregion Public methods
	}
}
