////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Mac;
using Vela.Crypto.Symmetric;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.StarkDocuments
{
	/// <summary>
	/// A securing document is an encrypted document which attests and encrypts a file, and is a satellite of that file.
	/// </summary>
	public class SecuringDocument : AttestingDocument
	{
		#region Public constants

		/// <summary>
		/// The default value for <see cref="ProcessingBlockSizeDef"/>.
		/// </summary>
		/// <remarks>
		/// <para>This must be multiple of the size of the <see cref="SymmAlg"/> algorithm because buffers of this size are
		/// symmetrically encrypted / decrypted.</para>
		/// </remarks>
		public const int ProcessingBlockSizeDef = 16 * 1024;

		/// <summary>
		/// The identifier of the speckle filter which contains the <see cref="RecipientUpnFingerprint"/> property.
		/// </summary>
		public const string RecipientUpnFingerprintFilterIdentifier = "SecuringDocument.RecipientUpnFingerprint#5";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected SecuringDocument()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="asymmSigningPaddingAlg">Asymmetric signing padding algorithm.</param>
		/// <param name="signatureHashSerializer">Hash serializer for signatures.</param>
		/// <param name="initiatorUpnFingerprint">The UPN fingerprint of the initiator of the current object.</param>
		/// <param name="recipientUpnFingerprint">The UPN fingerprint of the recipient of the current object.</param>
		/// <param name="symmAlg">The algorithm with which to perform the symmetric crypting.</param>
		/// <param name="salt">The (random) salt used for the symmetric encryption of the data. It must have the same size as the
		/// symmetric encryption algorithm.</param>
		/// <param name="symmEncryptionKey">The key used for the symmetric encryption of the data. It must have the same size as
		/// the symmetric encryption algorithm.</param>
		/// <param name="securedDocumentBlockSize">The size (in bytes) to which the (encrypted) data of a secured document is
		/// aligned. If this is bigger than 1, the real size of the data will be hidden.</param>
		public SecuringDocument( Locator instanceLocator, MacAlg macAlg, IAsymmSigningPaddingAlg asymmSigningPaddingAlg
				, IHashSerializer signatureHashSerializer, Locator initiatorUpnFingerprint, Locator recipientUpnFingerprint
				, SymmAlg symmAlg, byte[] salt, byte[] symmEncryptionKey, int securedDocumentBlockSize )
			: base( instanceLocator, macAlg, asymmSigningPaddingAlg, signatureHashSerializer, initiatorUpnFingerprint )
		{
			RecipientUpnFingerprint = recipientUpnFingerprint;
			SymmAlg = symmAlg;
			Salt = salt;
			SymmEncryptionKey = symmEncryptionKey;
			SecuredDocumentBlockSize = securedDocumentBlockSize;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get { return KernelUtils.MinTime( base.ExpirationTime, SymmAlg.ExpirationTime ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private Locator _RecipientUpnFingerprint;
		/// <summary>
		/// The UPN fingerprint of the recipient of the current object. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		[SpeckleFilterAttribute( RecipientUpnFingerprintFilterIdentifier )]
		public Locator RecipientUpnFingerprint
		{
			get { return _RecipientUpnFingerprint; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "RecipientUpnFingerprint" );

				_RecipientUpnFingerprint = value;
			}
		}

		private SymmAlg _SymmAlg;
		/// <summary>
		/// The algorithm with which to perform the symmetric crypting. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public SymmAlg SymmAlg
		{
			get { return _SymmAlg; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "SymmAlg" );

				_SymmAlg = value;
			}
		}

		private byte[] _Salt;
		/// <summary>
		/// The (random) salt used for the symmetric encryption of the data. It must have the same size as the symmetric
		/// encryption algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] Salt
		{
			get { return _Salt; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "Salt" );

				_Salt = value;
			}
		}

		private byte[] _SymmEncryptionKey;
		/// <summary>
		/// The key used for the symmetric encryption of the data. It must have the same size as the symmetric encryption
		/// algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] SymmEncryptionKey
		{
			get { return _SymmEncryptionKey; }
			protected set
			{
				ValidationUtils.ArrayHasContent( value, "SymmEncryptionKey" );

				_SymmEncryptionKey = value;
			}
		}

		private int _SecuredDocumentBlockSize;
		/// <summary>
		/// The size (in bytes) to which the (encrypted) data of a secured document is aligned. If this is bigger than 1, the
		/// real size of the data will be hidden. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public int SecuredDocumentBlockSize
		{
			get { return _SecuredDocumentBlockSize; }
			protected set { _SecuredDocumentBlockSize = value; }
		}

		private int _ProcessingBlockSize = ProcessingBlockSizeDef;
		/// <summary>
		/// The size (in bytes) to which the unencrypted document is processed. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This must be multiple of the size of the <see cref="SymmAlg"/> algorithm because buffers of this size are
		/// symmetrically encrypted / decrypted.</para>
		/// </remarks>
		[SerializeAttribute]
		public int ProcessingBlockSize
		{
			get { return _ProcessingBlockSize; }
			protected set { _ProcessingBlockSize = value; }
		}

		private Int64 _EncryptedDataSize;
		/// <summary>
		/// The size (in bytes) of the encrypted document. This must be at least 0. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public Int64 EncryptedDataSize
		{
			get { return _EncryptedDataSize; }
			protected set
			{
				ValidationUtils.Integer( value, 0, Int64.MaxValue, "EncryptedDataSize" );

				_EncryptedDataSize = value;
			}
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Validates the current object.
		/// </summary>
		protected virtual void Validate()
		{
			if( ProcessingBlockSize < 2 * SymmAlg.AlgSizeBytes )
			{
				throw new ValidationException( string.Format( "The processing block size is {0} bytes, but the smallest allowed"
					+ " size is {1}.", ProcessingBlockSize, 2 * SymmAlg.AlgSizeBytes ) );
			}

			if( ProcessingBlockSize % SymmAlg.AlgSizeBytes != 0 )
			{
				throw new ValidationException( string.Format( "The processing block size ({0} bytes) is not aligned to the size"
					+ " ({1} bytes) of the symmetric encryption algorithm.", ProcessingBlockSize, SymmAlg.AlgSizeBytes ) );
			}
		}

		/// <summary>
		/// This method is automatically called before the current object is serialized.
		/// </summary>
		protected virtual void Serializing()
		{
			Validate();
		}

		/// <summary>
		/// This method is automatically called after the current object has been deserialized.
		/// </summary>
		protected virtual void Deserialized()
		{
			Validate();
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Computes the MAC of the specified stream.
		/// </summary>
		/// <param name="stream">Stream whose MAC to compute.</param>
		/// <param name="lop">Long operation progress used during the computation of the MAC. This may be null.</param>
		/// <returns>Returns the MAC of the specified stream.</returns>
		public override byte[] Process( Stream stream, LongOperationProgress lop )
		{
			throw new ValidationException( "Use the \"Process\" method specific for a securing document." );
		}

		/// <summary>
		/// Computes the MAC and symmetrically crypts the specified source stream.
		/// </summary>
		/// <param name="sourceStream">Stream whose MAC to compute and to symmetrically crypt.</param>
		/// <param name="destinationStream">Stream where to write the crypted data.</param>
		/// <param name="isEncrypt">If this is true then the source stream will be encrypted, else it will be decrypted.</param>
		/// <param name="lop">Long operation progress. This may be null.</param>
		/// <returns>Returns the MAC of the specified stream.</returns>
		public virtual byte[] Process( Stream sourceStream, Stream destinationStream, bool isEncrypt, LongOperationProgress lop )
		{
			ValidationUtils.ObjectHasContent( sourceStream, "sourceStream" );
			ValidationUtils.ObjectHasContent( destinationStream, "destinationStream" );

			if( !sourceStream.CanSeek || (sourceStream.Length <= 0) )
				throw new ValidationException( "The stream whose MAC to compute is empty or its size is unknown." );

			if( lop != null ) lop.Progress( 0 );

			if( isEncrypt ) OriginalDataSize = sourceStream.Length;

			Validate();

			byte[] macKey = MacAlg.HashAlg.Compute( null, true, Salt, SymmEncryptionKey, InstanceLocator
				, InitiatorUpnFingerprint.Value );

			MacAlgStreamWorker macWorker = MacAlg.CreateStreamWorker( macKey );
			SymmAlgStreamWorker cryptionWorker = isEncrypt ? SymmAlg.CreateEncryptionStreamWorker( Salt, SymmEncryptionKey )
				: SymmAlg.CreateDecryptionStreamWorker( Salt, SymmEncryptionKey );
			Int64 currentEncryptedDataSize = 0;
			Int64 currentUnencryptedDataSize = 0;

			// Hash-in the (original) data size.
			byte[] totalSizeBuffer = BitConverter.GetBytes( OriginalDataSize );
			macWorker.Process( totalSizeBuffer );

			byte[] readData = new byte[ ProcessingBlockSize ];

			Int64 sizeToRead = isEncrypt ? OriginalDataSize : EncryptedDataSize;
			Int64 sizeReadSoFar = 0;
			while( sizeReadSoFar < sizeToRead )
			{
				int thisSizeToRead = (int)Math.Min( ProcessingBlockSize, sizeToRead - sizeReadSoFar );

				int thisReadSize = sourceStream.Read( readData, 0, thisSizeToRead );
				if( thisReadSize != thisSizeToRead )
				{
					throw new ValidationException( string.Format( "The stream size must be {0} not {1}.", sizeToRead
						, sizeReadSoFar ) );
				}

				sizeReadSoFar += thisSizeToRead;
				bool isLastRead = sizeReadSoFar >= sizeToRead;

				// Hash-in the data (before the decryption).
				if( !isEncrypt ) macWorker.Process( readData, 0, thisSizeToRead );

				// Crypt the data.
				byte[] processedData = isLastRead ? cryptionWorker.ProcessFinal( readData, 0, thisReadSize )
					: cryptionWorker.Process( readData, 0, thisReadSize );
				int processedDataSize = processedData.Length;

				// Hash-in the data (after the encryption).
				if( isEncrypt ) macWorker.Process( processedData );

				if( isEncrypt )
				{
					// Save the encrypted data.
					destinationStream.Write( processedData, 0, processedDataSize );

					currentEncryptedDataSize += processedDataSize;
					currentUnencryptedDataSize += thisReadSize;
				}
				else
				{
					// Save the decrypted data.
					destinationStream.Write( processedData, 0, processedData.Length );

					currentEncryptedDataSize += thisReadSize;
					currentUnencryptedDataSize += processedDataSize;
				}

				if( lop != null ) lop.Progress( (double)currentUnencryptedDataSize / OriginalDataSize );
			}

			if( OriginalDataSize != currentUnencryptedDataSize )
			{
				throw new ValidationException( string.Format( "The size of the unencrypted data must be {0} not {1}."
					, OriginalDataSize, currentUnencryptedDataSize ) );
			}

			if( isEncrypt )
			{
				EncryptedDataSize = currentEncryptedDataSize;

				// Save the alignment data.
				if( SecuredDocumentBlockSize > 1 )
				{
					Int64 alignedSize = MathUtils.AlignUp( EncryptedDataSize, SecuredDocumentBlockSize );
					int alignmentDataSize = (int)(alignedSize - EncryptedDataSize);

					if( alignmentDataSize > 0 )
					{
						byte[] alignmentData = EntropyPool.GetBytes( alignmentDataSize );

						destinationStream.Write( alignmentData, 0, alignmentData.Length );
					}
				}
			}

			// Compute the MAC.
			byte[] mac = macWorker.ProcessFinal();

			if( lop != null ) lop.Progress( 1 );

			return mac;
		}

		#endregion Public methods
	}
}
