////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.StarkDocuments
{
	/// <summary>
	/// Short command.
	/// </summary>
	public class ShortCommand : NotifyPropertyChanged, IDocument
	{
		#region Public constants

		/// <summary>
		/// Template for the full display name (which includes the category). Placeholders: 0 - display name, 1 - category.
		/// </summary>
		public const string FullDisplayNameTemplate = "{0} - {1}";

		/// <summary>
		/// The placeholder for the raw parameter of the short command.
		/// </summary>
		public const string RawParameterPlaceholder = "{Parameter}";

		/// <summary>
		/// The placeholder for the URL-escaped parameter of the short command.
		/// </summary>
		public const string UrlEscapedParameterPlaceholder = "{UrlEscaped}";

		/// <summary>
		/// The shortcut key which is ignored.
		/// </summary>
		public const char NoShortcutKey = ' ';

		/// <summary>
		/// Template for displaying a shortcut key.
		/// </summary>
		public const string ShortcutKeyTemplate = "Alt + {0}";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public ShortCommand()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="displayName">Display name.</param>
		/// <param name="category">Category.</param>
		/// <param name="shortcutKey">Shortcut key. If this is <see cref="NoShortcutKey"/>, it's ignored.</param>
		/// <param name="contextWithoutShortcutKey">If this is true then the current short command is shown in a menu even if it
		/// has no shortcut key.</param>
		/// <param name="command">Command to run.</param>
		public ShortCommand( string displayName, string category, char shortcutKey, bool contextWithoutShortcutKey
			, string command )
		{
			DisplayName = displayName;
			Category = category;
			ShortcutKey = shortcutKey;
			ContextWithoutShortcutKey = contextWithoutShortcutKey;
			Command = command;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// Full display name (which includes the category).
		/// </summary>
		public string FullDisplayName
		{
			get { return string.Format( FullDisplayNameTemplate, DisplayName, Category ); }
		}

		/// <summary>
		/// Display text for the shortcut key. This may be empty, but not null. READ-ONLY.
		/// </summary>
		public string DisplayShortcutKey
		{
			get { return GetDisplayShortcutKey( ShortcutKey ); }
		}

		/// <summary>
		/// If this is true then the current short command is shown in a menu.
		/// </summary>
		/// <remarks>
		/// <para>A short command is shown in a menu if it has a shortcut key or if its <see cref="ContextWithoutShortcutKey"/>
		/// property is true.</para>
		/// </remarks>
		public bool ShowInMenu
		{
			get { return (ShortcutKey != ShortCommand.NoShortcutKey) || ContextWithoutShortcutKey; }
		}

		/// <summary>
		/// The sort key. READ-ONLY.
		/// </summary>
		public Tuple SortKey
		{
			get { return GetSortKey( DisplayName, Category ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private string _DisplayName;
		/// <summary>
		/// Display name. Notifies property observers.
		/// </summary>
		[SerializeAttribute]
		public string DisplayName
		{
			get { return _DisplayName; }
			set
			{
				ValidationUtils.Text( value, -1, -1, TextValidations.EntityName, "DisplayName" );

				_DisplayName = value;

				// Notify property observers.
				OnPropertyChanged( "DisplayName" );
				OnPropertyChanged( "FullDisplayName" );
			}
		}

		private string _Category;
		/// <summary>
		/// Category. Notifies property observers.
		/// </summary>
		[SerializeAttribute]
		public string Category
		{
			get { return _Category; }
			set
			{
				ValidationUtils.Text( value, -1, -1, TextValidations.EntityName, "Category" );

				_Category = value;

				// Notify property observers.
				OnPropertyChanged( "Category" );
				OnPropertyChanged( "FullDisplayName" );
			}
		}

		private char _ShortcutKey = NoShortcutKey;
		/// <summary>
		/// Shortcut key. If this is <see cref="NoShortcutKey"/>, it's ignored. Notifies property observers.
		/// </summary>
		/// <remarks>
		/// <para>A short command is shown in a menu if it has a shortcut key or if its <see cref="ContextWithoutShortcutKey"/>
		/// property is true.</para>
		/// </remarks>
		[SerializeAttribute]
		public char ShortcutKey
		{
			get { return _ShortcutKey; }
			set
			{
				_ShortcutKey = value;

				// Notify property observers.
				OnPropertyChanged( "ShortcutKey" );
				OnPropertyChanged( "DisplayShortcutKey" );
				OnPropertyChanged( "ShowInMenu" );
			}
		}

		private bool _ContextWithoutShortcutKey;
		/// <summary>
		/// If this is true then the current short command is shown in a menu even if it has no shortcut key. Notifies property
		/// observers.
		/// </summary>
		/// <remarks>
		/// <para>The current short command is shown in a menu if it has a shortcut key or this property is true.</para>
		/// </remarks>
		[SerializeAttribute]
		public bool ContextWithoutShortcutKey
		{
			get { return _ContextWithoutShortcutKey; }
			set
			{
				_ContextWithoutShortcutKey = value;

				// Notify property observers.
				OnPropertyChanged( "ContextWithoutShortcutKey" );
				OnPropertyChanged( "ShowInMenu" );
			}
		}

		private string _Command;
		/// <summary>
		/// Command to run. Notifies property observers.
		/// </summary>
		[SerializeAttribute]
		public string Command
		{
			get { return _Command; }
			set
			{
				ValidationUtils.Text( value, -1, -1, TextValidations.ShortCommand, "Command" );

				_Command = value;

				// Notify property observers.
				OnPropertyChanged( "Command" );
			}
		}

		#endregion Public serializable properties

		#region Public static methods

		/// <summary>
		/// Gets the display text for the specified shortcut key.
		/// </summary>
		/// <param name="shortcutKey">Shortcut key for which to get the display text.</param>
		/// <returns>Returns the display text for the shortcut key. This may be empty, but not null.</returns>
		public static string GetDisplayShortcutKey( char shortcutKey )
		{
			if( (shortcutKey == NoShortcutKey) || (shortcutKey == char.MinValue) )
				return "";

			return string.Format( ShortcutKeyTemplate, shortcutKey );
		}

		/// <summary>
		/// Gets the sort key.
		/// </summary>
		/// <param name="displayName">Display name.</param>
		/// <param name="category">Category.</param>
		public Tuple GetSortKey( string displayName, string category )
		{
			return new Tuple( category, displayName );
		}

		#endregion Public static methods
	}
}
