////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;

namespace Vela.StarkDocuments
{
	/// <summary>
	/// List of short commands (<see cref="ShortCommand"/>).
	/// </summary>
	public class ShortCommands : KernelSortedList<Tuple, ShortCommand>
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public ShortCommands()
			: base( TupleComparer.Instance )
		{
		}

		#endregion Constructors

		#region Public static methods

		/// <summary>
		/// Gets a new initialized instance of the current class.
		/// </summary>
		/// <returns>Returns a new initialized instance of the current class.</returns>
		public static ShortCommands GetInitializedInstance()
		{
			ShortCommands shortCommands = new ShortCommands();
			shortCommands.AddDefaultOrUpdateExistingCommands( false );

			return shortCommands;
		}

		#endregion Public static methods

		#region Private methods

		/// <summary>
		/// Updates the shortcut key (only if it's missing) and the <see cref="ShortCommand.Command"/> property of the existing
		/// short command, with the values form the specified short command.
		/// </summary>
		/// <param name="shortCommand">Short command to update in the list.</param>
		/// <remarks>
		/// <para>Un update has to be done only after all the short commands are added to the list, so that all the used shortcut
		/// keys can be found in the list.</para>
		/// </remarks>
		private void UpdateExistingCommand( ShortCommand shortCommand )
		{
			ShortCommand existingCommand = Find( GetKey( shortCommand ) );

			// If there is no existing command, return.
			if( existingCommand == null )
				return;

			// Update the shortcut key of the existing command.
			if(
				// If the existing command has no shortcut key.
				(existingCommand.ShortcutKey == ShortCommand.NoShortcutKey)

				// If the specified short command has a shortcut key.
				&& (shortCommand.ShortcutKey != ShortCommand.NoShortcutKey)

				// If the shortcut key (of the specified command) is not used by any existing command.
				&& IsShortcutKeyUnused( shortCommand.ShortcutKey ) )
			{
				existingCommand.ShortcutKey = shortCommand.ShortcutKey;
			}

			// Update the "Command" property of the existing command.
			existingCommand.Command = shortCommand.Command;
		}

		/// <summary>
		/// Adds the specified default short command to the list, or updates the existing short command with the values from the
		/// specified short command.
		/// </summary>
		/// <param name="shortCommand">Default short command to add to or update in the list.</param>
		/// <param name="alsoUpdateExisting">If this is true then existing short commands are updated; only the
		/// shortcut key and the <see cref="ShortCommand.Command"/> property are updated.</param>
		private void AddDefaultOrUpdateExistingCommand( ShortCommand shortCommand, bool alsoUpdateExisting )
		{
			AddOnlyNew( shortCommand );

			if( alsoUpdateExisting )
				UpdateExistingCommand( shortCommand );
		}

		/// <summary>
		/// Adds the default short commands, but only if they don't already exist in the list.
		/// </summary>
		/// <param name="alsoUpdateExisting">If this is true then existing short commands are updated; only the
		/// shortcut key and the <see cref="ShortCommand.Command"/> property are updated.</param>
		private void AddDefaultOrUpdateExistingCommands( bool alsoUpdateExisting )
		{
			// The short commands are sorted by category (and then display name).

			#region Applications

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Calculator", "Application", GetShortcutKeyIfUnused( 'C' )
				, false, "calc.exe" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Notepad", "Application", GetShortcutKeyIfUnused( 'N' )
				, false, "notepad.exe" ), alsoUpdateExisting );

			#endregion Applications

			#region URLs

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "DeviantArt", "Art", ShortCommand.NoShortcutKey
				, false, "http://www.deviantart.com/browse/all/?q={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Thesaurus", "Dictionary", ShortCommand.NoShortcutKey
				, false, "http://thesaurus.com/browse/{UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Wiktionary", "Dictionary", ShortCommand.NoShortcutKey
				, true, "https://en.wiktionary.org/w/index.php?search={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Britannica", "Encyclopedia", ShortCommand.NoShortcutKey
				, false, "https://www.britannica.com/search?query={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Wikipedia", "Encyclopedia", ShortCommand.NoShortcutKey
				, true, "https://en.wikipedia.org/w/index.php?search={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "SkipThePie", "Health", ShortCommand.NoShortcutKey
				, true, "http://skipthepie.org/?q={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Bing", "Map", ShortCommand.NoShortcutKey
				, false, "https://www.bing.com/mapspreview?where1={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Google", "Map", ShortCommand.NoShortcutKey
				, true, "https://maps.google.com/maps?q={UrlEscaped}&zoom=14&maptype=roadmap" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "IMDB", "Movies", ShortCommand.NoShortcutKey
				, true, "http://www.imdb.com/find?q={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Google", "News", ShortCommand.NoShortcutKey
				, false, "https://www.google.com/search?q={UrlEscaped}&tbm=nws" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Scholar", "Science", ShortCommand.NoShortcutKey
				, false, "https://www.semanticscholar.org/search?q={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Bing", "Search", ShortCommand.NoShortcutKey
				, false, "https://www.bing.com/search?q={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Google", "Search", ShortCommand.NoShortcutKey
				, true, "https://www.google.com/search?q={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "IxQuick", "Search", ShortCommand.NoShortcutKey
				, true, "https://www.ixquick.com/do/search?query={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Yahoo", "Search", ShortCommand.NoShortcutKey
				, false, "https://search.yahoo.com/search?p={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Amazon USA", "Shop", ShortCommand.NoShortcutKey
				, false, "https://www.amazon.com/s?field-keywords={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Facebook", "Social", ShortCommand.NoShortcutKey
				, false, "https://www.facebook.com/search/top/?q={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Booking", "Travel", ShortCommand.NoShortcutKey
				, false, "https://www.booking.com/searchresults.html?ss={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Expedia", "Travel", ShortCommand.NoShortcutKey
				, false, "https://www.expedia.com/Hotel-Search?destination={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Google", "Travel", ShortCommand.NoShortcutKey
				, false, "https://www.google.com/search?q=travel%20{UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "Hotels", "Travel", ShortCommand.NoShortcutKey
				, false, "https://www.hotels.com/search.do?destination={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "TripAdvisor", "Travel", ShortCommand.NoShortcutKey
				, false, "https://www.tripadvisor.com/search?q={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "WikiVoyage", "Travel", ShortCommand.NoShortcutKey
				, false, "https://en.wikivoyage.org/w/index.php?search={UrlEscaped}" ), alsoUpdateExisting );

			AddDefaultOrUpdateExistingCommand( new ShortCommand( "YouTube", "Video", ShortCommand.NoShortcutKey
				, false, "https://www.youtube.com/results?search_query={UrlEscaped}" ), alsoUpdateExisting );

			#endregion URLs
		}

		/// <summary>
		/// This method is automatically called after the current object has been deserialized.
		/// </summary>
		private void Deserialized()
		{
			AddDefaultOrUpdateExistingCommands( true );
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para><see cref="ShortCommand.SortKey"/> is used as key.</para>
		/// </remarks>
		public override Tuple GetKey( ShortCommand item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.SortKey;
		}

		/// <summary>
		/// Checks if the list contains at least 2 short commands with the specified display name, commands whose
		/// <see cref="ShortCommand.ShowInMenu"/> property is true.
		/// </summary>
		/// <param name="displayName">The display name to check.</param>
		/// <returns>Returns true if the list contains at least 2 short commands with the specified display name.</returns>
		public bool HasMultipleVisibleDisplayNames( string displayName )
		{
			int count = 0;
			foreach( ShortCommand shortCommand in this )
			{
				if( !shortCommand.ShowInMenu )
					continue;

				if( TextUtils.Equals( shortCommand.DisplayName, displayName ) )
				{
					count++;

					if( count >= 2 )
						return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Finds the short command with the specified shortcut key.
		/// </summary>
		/// <param name="shortcutKey">The shortcut key of the short command to find.</param>
		/// <returns>Returns the found short command, or null if none is found.</returns>
		public ShortCommand FindByShortcutKey( char shortcutKey )
		{
			if( (shortcutKey == ShortCommand.NoShortcutKey) || (shortcutKey == char.MinValue) )
				return null;

			foreach( ShortCommand shortCommand in this )
			{
				if( shortCommand.ShortcutKey == shortcutKey )
					return shortCommand;
			}

			return null;
		}

		/// <summary>
		/// Returns the specified shortcut key if it's not used by any of the short commands from the list.
		/// </summary>
		/// <param name="shortcutKey">The shortcut key to find.</param>
		/// <returns>Returns the specified shortcut key if it's not used, or <see cref="ShortCommand.NoShortcutKey"/> if the
		/// shortcut key is used.</returns>
		public char GetShortcutKeyIfUnused( char shortcutKey )
		{
			if( (shortcutKey == ShortCommand.NoShortcutKey) || (shortcutKey == char.MinValue) )
				return ShortCommand.NoShortcutKey;

			foreach( ShortCommand shortCommand in this )
			{
				if( shortCommand.ShortcutKey == shortcutKey )
					return ShortCommand.NoShortcutKey;
			}

			return shortcutKey;
		}

		/// <summary>
		/// Checks if the specified shortcut key is not used by any of the short commands from the list.
		/// </summary>
		/// <param name="shortcutKey">The shortcut key to check.</param>
		/// <returns>Returns true if the specified shortcut key is not used by any of the short commands from the list.</returns>
		public bool IsShortcutKeyUnused( char shortcutKey )
		{
			return GetShortcutKeyIfUnused( shortcutKey ) != ShortCommand.NoShortcutKey;
		}

		/// <summary>
		/// Moves the specified shortcut key from the existing short command to the specified short command.
		/// </summary>
		/// <param name="shortcutKey">Shortcut key to move.</param>
		/// <param name="newShortCommand">Short command where to move the shortcut key.</param>
		public void MoveShortcutKey( char shortcutKey, ShortCommand newShortCommand )
		{
			ValidationUtils.ObjectHasContent( newShortCommand, "shortCommand" );

			// Clear the existing shortcut key.
			if( shortcutKey != ShortCommand.NoShortcutKey )
			{
				foreach( ShortCommand oldShortCommand in this )
				{
					if( oldShortCommand.ShortcutKey == shortcutKey )
						oldShortCommand.ShortcutKey = ShortCommand.NoShortcutKey;
				}
			}

			newShortCommand.ShortcutKey = shortcutKey;
		}

		#endregion Public methods
	}
}
