////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;

namespace Vela.Storehouse
{
	/// <summary>
	/// Information about how to process documents from a storehouse in a paged manner.
	/// </summary>
	public class CollectPaging
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public CollectPaging()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="newestFirst">If this is true then the documents to process from the storehouse are sorted by when they
		/// were last added (to the storehouse), else by when they were first added.</param>
		/// <param name="offset">The offset in the list of documents, sorted according to <see cref="NewestFirst"/>, from which
		/// to process the documents from the storehouse. This must be at least 0.</param>
		/// <param name="count">The number of documents to process from the storehouse. This must be at least 1.</param>
		public CollectPaging( bool newestFirst, int offset, int count )
		{
			NewestFirst = newestFirst;
			Offset = offset;
			Count = count;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private bool _NewestFirst;
		/// <summary>
		/// If this is true then the documents to process from the storehouse are sorted by when they were last added (to the
		/// storehouse), else by when they were first added. READ-ONLY.
		/// </summary>
		public bool NewestFirst
		{
			get { return _NewestFirst; }
			protected set { _NewestFirst = value; }
		}

		private int _Offset;
		/// <summary>
		/// The offset in the list of documents, sorted according to <see cref="NewestFirst"/>, from which to process the
		/// documents from the storehouse. This must be at least 0. READ-ONLY.
		/// </summary>
		public int Offset
		{
			get { return _Offset; }
			protected set
			{
				ValidationUtils.Integer( value, 0, int.MaxValue, "Offset" );

				_Offset = value;
			}
		}

		private int _Count;
		/// <summary>
		/// The number of documents to process from the storehouse. This must be at least 1. READ-ONLY.
		/// </summary>
		public int Count
		{
			get { return _Count; }
			protected set
			{
				ValidationUtils.Integer( value, 1, int.MaxValue, "Count" );

				_Count = value;
			}
		}

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Modifies the <see cref="Offset"/> with the specified value.
		/// </summary>
		/// <param name="shift">Value with which to modify the offset.</param>
		public void ModifyOffset( int shift )
		{
			Offset += shift;
		}

		#endregion Public methods
	}
}
