////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.Storehouse
{
	/// <summary>
	/// Contract for a storehouse provider.
	/// </summary>
	/// <remarks>
	/// <para>Don't share a storehouse provider among multiple threads, even if you synchronize it. Clone it instead.</para>
	/// </remarks>
	public interface IStorehouseProvider : IDocument
	{
		#region Public non-serializable properties

		/// <summary>
		/// The size of the password used to encrypt the storehouse files which are attached to the storehouse header. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is used to derive the encryption password (see <see cref="GetStorehouseAttachmentsPassword"/>).</para>
		/// </remarks>
		int StorehouseAttachmentsPasswordSize { get; }

		/// <summary>
		/// Is true if the referred storehouse exists. READ-ONLY.
		/// </summary>
		bool StorehouseExists { get; }

		/// <summary>
		/// If this is true then the storehouse is managed by a storehouse server engine, else it's standalone / embedded.
		/// READ-ONLY.
		/// </summary>
		bool UsesServerEngine { get; }

		/// <summary>
		/// Set of scripts used to work with the storehouse. READ-ONLY.
		/// </summary>
		IStorehouseScripts Scripts { get; }

		/// <summary>
		/// Placeholders used in storehouse scripts. READ-ONLY.
		/// </summary>
		IStorehouseScriptPlaceholders ScriptPlaceholders { get; }

		/// <summary>
		/// The name of the storehouse as it must be used by the storehouse server (if one exists). READ-ONLY.
		/// </summary>
		string ServerStorehouseName { get; }

		#endregion Public non-serializable properties

		#region Public serializable properties

		/// <summary>
		/// The storehouse access. Notifies property observers. READ-ONLY.
		/// </summary>
		IStorehouseAccess Access { get; }

		/// <summary>
		/// The name of the file where the storehouse header is stored. READ-ONLY.
		/// </summary>
		string FileName { get; }

		/// <summary>
		/// The user password used to symmetrically encrypt the header. READ-ONLY.
		/// </summary>
		string HeaderUserPassword { get; }

		/// <summary>
		/// The storehouse header (which is stored in a symmetrically encrypted form). READ-ONLY.
		/// </summary>
		StorehouseHeader Header { get; }

		/// <summary>
		/// The serializer used to serialize the documents. READ-ONLY.
		/// </summary>
		ISerializer Serializer { get; }

		#endregion Public serializable properties

		#region Public methods

		/// <summary>
		/// Checks if the specified exception (or any of its inner exceptions) represents a timeout (which has occurred while
		/// accessing the storehouse).
		/// </summary>
		/// <param name="exception">Exception to check if it represents a timeout. This may be null.</param>
		/// <returns>Returns true if the specified exception (or any of its inner exceptions) represents a timeout.</returns>
		bool IsTimeoutException( Exception exception );

		/// <summary>
		/// Creates a new storehouse. If the storehouse already exists, it's overwritten.
		/// </summary>
		/// <param name="adminConnectionString">The administration connection string with which a storehouse server can create
		/// a storehouse. This may be null or empty for a storehouse provider which doesn't use a storehouse server.</param>
		void CreateStorehouse( string adminConnectionString );

		/// <summary>
		/// Opens the storehouse with the current provider.
		/// </summary>
		void OpenStorehouse();

		/// <summary>
		/// Compacts the storehouse.
		/// </summary>
		/// <param name="adminConnectionString">The administration connection string with which a storehouse server can compact
		/// the storehouse. This may be null or empty for a storehouse provider which doesn't use a storehouse server.</param>
		/// <param name="alsoVerifyAndRepair">If this is true then, before being compacted, the storehouse is verified for
		/// consistency, and is repaired if it's necessary and possible.</param>
		/// <remarks>
		/// <para>The empty space is removed.</para>
		/// 
		/// <para>If the storehouse doesn't support this operation, no exception is thrown.</para>
		/// </remarks>
		void CompactStorehouse( string adminConnectionString, bool alsoVerifyAndRepair );

		/// <summary>
		/// Creates new a transaction in the storehouse.
		/// </summary>
		IStorehouseTransaction CreateTransaction();

		/// <summary>
		/// Gets the file name of the main storehouse file which is attached to the storehouse header.
		/// </summary>
		/// <returns>Returns the file name of the main storehouse file which is attached to the storehouse header.</returns>
		string GetStorehouseMainAttachmentFileName();

		/// <summary>
		/// Gets the password used to encrypt the storehouse files which are attached to the header.
		/// </summary>
		/// <returns>Returns the password used to encrypt the storehouse files which are attached to the header.</returns>
		string GetStorehouseAttachmentsPassword();

		/// <summary>
		/// Saves the specified storehouse header to the storehouse. RESERVED for internal use.
		/// </summary>
		/// <param name="header">The storehouse header to save.</param>
		/// <param name="headerUserPassword">The user password used to symmetrically encrypt the header. If this is null or
		/// empty, existing password is used instead. If this has content and is different than the existing password, the
		/// header password is essentially changed.</param>
		/// <remarks>
		/// <para>Make sure that you always call this method from the same thread in order to be sure that no other thread
		/// accidentally overrides the header with the old header password.</para>
		/// </remarks>
		void SaveHeader( StorehouseHeader header, string headerUserPassword );

		/// <summary>
		/// Loads the storehouse header. An exception is thrown if the header is missing. RESERVED for internal use.
		/// </summary>
		/// <remarks>
		/// <para>If the header is already loaded, it's not loaded again.</para>
		/// 
		/// <para>This is automatically called when the header is lazy loaded.</para>
		/// 
		/// <para>The <see cref="Serializer"/> property (of the storehouse provider) is set in here.</para>
		/// </remarks>
		void LoadHeader();

		/// <summary>
		/// Changes the password of the user used to symmetrically encrypt the header.
		/// </summary>
		/// <param name="headerUserPassword">The new user password used to symmetrically encrypt the header.</param>
		/// <remarks>
		/// <para>Make sure that you always call this method from the same thread in order to be sure that no other thread
		/// accidentally overrides the header with the old header password.</para>
		/// </remarks>
		void ChangeHeaderUserPassword( string headerUserPassword );

		#endregion Public methods
	}
}
