////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.Serialization;

namespace Vela.Storehouse
{
	/// <summary>
	/// Contract for a storehouse transaction.
	/// </summary>
	/// <remarks>
	/// <para>Don't share a storehouse transaction among multiple threads, even if you synchronize it. Clone it instead.</para>
	/// 
	/// <para>Only the methods which add or load (but not collect) a speckle, or delete a sheaf, (also) work with speckles which
	/// don't contain the referred sheaf. All the other methods work only with speckles which do contain the referred sheaf.
	/// </para>
	/// </remarks>
	public interface IStorehouseTransaction
	{
		#region Public non-serializable properties

		/// <summary>
		/// The storehouse provider which has opened the current transaction. READ-ONLY.
		/// </summary>
		IStorehouseProvider Provider { get; }

		/// <summary>
		/// If this is true then the current transaction is open, else it's closed. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>When a transaction is closed, no changes may be made through it to the storehouse.</para>
		/// </remarks>
		bool IsOpen { get; }

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Commits and closes the current transaction in the storehouse.
		/// </summary>
		void Commit();

		/// <summary>
		/// Rolls back and closes the current transaction from the storehouse.
		/// </summary>
		/// <remarks>
		/// <para>Rolling back a transaction means closing it without making the changes requested in it.</para>
		/// </remarks>
		void Rollback();

		/// <summary>
		/// Adds the specified storehouse entropy to the storehouse. RESERVED for internal use.
		/// </summary>
		/// <param name="entropy">The storehouse entropy to add.</param>
		/// <remarks>
		/// <para>The entropy is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		void AddEntropy( StorehouseEntropy entropy );

		/// <summary>
		/// Updates the storehouse entropy with the specified one.
		/// </summary>
		/// <param name="entropy">The storehouse entropy to update.</param>
		/// <remarks>
		/// <para>The entropy is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		void UpdateEntropy( StorehouseEntropy entropy );

		/// <summary>
		/// Loads the entropy of the storehouse. An exception is thrown if the entropy is missing.
		/// </summary>
		/// <param name="sheafLock">Lock to put on the storehouse entropy document.</param>
		StorehouseEntropy LoadEntropy( SheafLocks sheafLock );

		/// <summary>
		/// Adds the specified storehouse configuration to the storehouse. RESERVED for internal use.
		/// </summary>
		/// <param name="configuration">The storehouse configuration to add.</param>
		/// <remarks>
		/// <para>The configuration is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		void AddConfiguration( StorehouseConfiguration configuration );

		/// <summary>
		/// Updates the storehouse configuration with the specified one.
		/// </summary>
		/// <param name="configuration">The storehouse configuration to update.</param>
		/// <remarks>
		/// <para>The configuration is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		void UpdateConfiguration( StorehouseConfiguration configuration );

		/// <summary>
		/// Loads the configuration of the storehouse. An exception is thrown if the configuration is missing.
		/// </summary>
		/// <param name="sheafLock">Lock to put on the storehouse configuration document.</param>
		StorehouseConfiguration LoadConfiguration( SheafLocks sheafLock );

		/// <summary>
		/// Adds a speckle with the specified storehouse key.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the speckle to add.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="sheafType">The type of the sheaf referred by the added speckle.</param>
		/// <remarks>
		/// <para>The storehouse key of the added speckle might be the same as another storehouse key from the current
		/// storehouse; in such a case, an exception is thrown.</para>
		/// 
		/// <para>The speckle is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// </remarks>
		void AddSpeckle( Locator storehouseKey, string storehouseRealm, Type sheafType );

		/// <summary>
		/// Counts the speckles with the specified speckle filters, which refer a document with the specified types.
		/// </summary>
		/// <param name="sheafTypes">The allowed sheaf types of the speckles to count.</param>
		/// <param name="seekFilters">The seek filters which identify the matching speckles. This may be null or empty.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <returns>Returns the number of matching speckles.</returns>
		/// <remarks>
		/// <para>The storehouse matching speckles are not locked.</para>
		/// 
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// </remarks>
		int CountSpeckles( PortableTypes sheafTypes, SpeckleFilters seekFilters, string storehouseRealm );

		/// <summary>
		/// Loads the speckles with the specified speckle filters, which refer a document with the specified types.
		/// </summary>
		/// <param name="sheafTypes">The allowed sheaf types of the speckles to load.</param>
		/// <param name="seekFilters">The seek filters which identify the matching speckles. This may be null or empty.
		/// </param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="paging">Information about how to process the documents in a paged manner. If this is null, no paging is
		/// used.</param>
		/// <param name="collectedSpeckles">List where to add the collected speckles. If this is null, a new instance of
		/// <see cref="Speckles"/> is used instead. This allows you to have lists sorted by different criteria.</param>
		/// <param name="collectedSpeckleType">The type of the items to add to the list of collected speckles. This must inherit
		/// <see cref="Speckle"/> and must have a constructor which receives (as parameters) the storehouse key of the referred
		/// sheaf, the type name of the stored sheaf, and the loaded speckle filters. If this is null,
		/// <see cref="Speckle"/> is used instead. If this is different than null and <see cref="Speckle"/>, the performance is
		/// slightly reduced.</param>
		/// <param name="loadFilterIdentifiers">The speckle filter identifiers whose values to load from the matching speckles.
		/// This may be null or empty.</param>
		/// <returns>Returns the loaded speckles.</returns>
		/// <remarks>
		/// <para>The storehouse matching speckles are not locked.</para>
		/// 
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// 
		/// <para>If you never want (or it's never possible) to collect more than one speckle, for performance reasons you should
		/// collect the first speckle or document. This is because speckle collection doesn't stop at the first matching speckle,
		/// so the seek performance degrades geometrically with the number of documents from the storehouse.</para>
		/// </remarks>
		IKernelObservableList CollectSpeckles( PortableTypes sheafTypes, SpeckleFilters seekFilters, string storehouseRealm
			, CollectPaging paging, IKernelObservableList collectedSpeckles, Type collectedSpeckleType
			, SpeckleFilterIdentifiers loadFilterIdentifiers );

		/// <summary>
		/// Loads the first speckle with the specified speckle filters, which refer a document with the specified types.
		/// </summary>
		/// <param name="sheafTypes">The allowed sheaf types of the speckle to load.</param>
		/// <param name="seekFilters">The seek filters which identify the matching speckle. This may be null or empty.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="speckleLock">Lock to put on the speckle.</param>
		/// <param name="collectedSpeckleType">The type of the collected speckle. This must inherit <see cref="Speckle"/> and
		/// must have a constructor which receives (as parameters) the storehouse key of the referred sheaf, the type name of
		/// the stored sheaf, and the loaded speckle filters. If this is null, <see cref="Speckle"/> is used instead.</param>
		/// <param name="loadFilterIdentifiers">The speckle filter identifiers whose values to load from the matching speckle.
		/// This may be null or empty.</param>
		/// <returns>Returns the loaded speckle.</returns>
		/// <remarks>
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// 
		/// <para>The method returns the first encountered speckle, not necessarily the speckle which was added first. This
		/// behavior mVela.Imizes performance.</para>
		/// </remarks>
		Speckle CollectFirstSpeckle( PortableTypes sheafTypes, SpeckleFilters seekFilters, string storehouseRealm
			, SheafLocks speckleLock, Type collectedSpeckleType, SpeckleFilterIdentifiers loadFilterIdentifiers );

		/// <summary>
		/// Loads the speckle with the specified storehouse key.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the speckle to load.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="speckleLock">Lock to put on the speckle.</param>
		/// <param name="loadedSpeckleType">The type of the speckle to load. This must inherit <see cref="Speckle"/> and must
		/// have a constructor which receives (as parameters) the storehouse key of the referred sheaf, the type name of the
		/// stored sheaf, and the loaded speckle filters. If this is null, <see cref="Speckle"/> is used instead.</param>
		/// <param name="loadFilterIdentifiers">The speckle filter identifiers whose values to load from the speckle. This may be
		/// null or empty.</param>
		/// <returns>Returns the loaded speckle, or null if none is found.</returns>
		Speckle LoadSpeckle( Locator storehouseKey, string storehouseRealm, SheafLocks speckleLock, Type loadedSpeckleType
			, SpeckleFilterIdentifiers loadFilterIdentifiers );

		/// <summary>
		/// Checks if the speckle with the specified storehouse key has a referred sheaf.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the speckle to check.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <returns>Returns true if the speckle has a referred sheaf.</returns>
		/// <remarks>
		/// <para>The speckle filters are not loaded from the storehouse.</para>
		/// </remarks>
		bool SpeckleHasSheaf( Locator storehouseKey, string storehouseRealm );

		/// <summary>
		/// Loads the storehouse keys of the speckles with the specified speckle filters, which refer a document with the
		/// specified types.
		/// </summary>
		/// <param name="sheafTypes">The allowed sheaf types of the speckles whose keys to load.</param>
		/// <param name="seekFilters">The seek filters which identify the matching speckles. This may be null or empty.
		/// </param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="paging">Information about how to process the documents in a paged manner. If this is null, no paging is
		/// used.</param>
		/// <param name="collectedKeys">List where to add the collected keys. If this is null, a new instance of
		/// <see cref="Locators"/> is used instead. This allows you to have lists sorted by different criteria.</param>
		/// <param name="collectedKeyType">The type of the items to add to the list of collected keys. This must inherit
		/// <see cref="Locator"/> and must have a constructor which receives (as parameter) the locator's value. If this is
		/// null, <see cref="Locator"/> is used instead. If this is different than null and <see cref="Locator"/>, the
		/// performance is slightly reduced.</param>
		/// <returns>Returns the loaded storehouse keys.</returns>
		/// <remarks>
		/// <para>The storehouse matching speckles are not locked.</para>
		/// 
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// </remarks>
		IKernelObservableList CollectKeys( PortableTypes sheafTypes, SpeckleFilters seekFilters, string storehouseRealm
			, CollectPaging paging, IKernelObservableList collectedKeys, Type collectedKeyType );

		/// <summary>
		/// Adds the specified sheaf to the storehouse.
		/// </summary>
		/// <param name="sheaf">The sheaf to add to the storehouse.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="salt">The (random) salt used to compute the hash and encrypt the sheaf. This is stored in unencrypted
		/// form. If this is null or empty, a random salt is used instead.</param>
		/// <param name="encryptionKey">The key used for the encryption of the sheaf.</param>
		/// <remarks>
		/// <para>If a sheaf with the same storehouse key already exists, an exception is thrown.</para>
		/// 
		/// <para>The sheaf is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>The <see cref="Sheaf.UpdateTime"/> property is automatically set in here with the current time.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		void AddSheaf( ISheaf sheaf, string storehouseRealm, byte[] salt, byte[] encryptionKey );

		/// <summary>
		/// Adds the specified sheaf to the storehouse.
		/// </summary>
		/// <param name="sheaf">The sheaf to add to the storehouse.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <remarks>
		/// <para>If a sheaf with the same storehouse key already exists, an exception is thrown.</para>
		/// 
		/// <para>The sheaf is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>The <see cref="Sheaf.UpdateTime"/> property is automatically set in here with the current time.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		void AddSheaf( ISheaf sheaf, string storehouseRealm );

		/// <summary>
		/// Updates the specified document in the storehouse.
		/// </summary>
		/// <param name="sheaf">The sheaf to update in the storehouse.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="salt">The (random) salt used to compute the hash and encrypt the sheaf. This is stored in unencrypted
		/// form. If this is null or empty, a random salt is used instead.</param>
		/// <param name="encryptionKey">The key used for the encryption of the sheaf.</param>
		/// <remarks>
		/// <para>If a sheaf with the same storehouse key already exists, it is replaced.</para>
		/// 
		/// <para>The sheaf must be locked for writing.</para>
		/// 
		/// <para>The <see cref="Sheaf.UpdateTime"/> property is automatically set in here with the current time.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		void UpdateSheaf( ISheaf sheaf, string storehouseRealm, byte[] salt, byte[] encryptionKey );

		/// <summary>
		/// Updates the specified document in the storehouse.
		/// </summary>
		/// <param name="sheaf">The sheaf to update in the storehouse.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <remarks>
		/// <para>If a sheaf with the same storehouse key already exists, it is replaced.</para>
		/// 
		/// <para>The sheaf must be locked for writing.</para>
		/// 
		/// <para>The <see cref="Sheaf.UpdateTime"/> property is automatically set in here with the current time.</para>
		/// 
		/// <para>You can save the same sheaf again (without reloading it) because its update version is refreshed in here.
		/// </para>
		/// </remarks>
		void UpdateSheaf( ISheaf sheaf, string storehouseRealm );

		/// <summary>
		/// Deletes the sheafs with the specified speckle filters, which refer documents with the specified types.
		/// </summary>
		/// <param name="sheafTypes">The allowed sheaf types of the sheafs to delete.</param>
		/// <param name="seekFilters">The seek filters which identify the matching sheafs.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <returns>Returns the number of deleted sheafs.</returns>
		/// <remarks>
		/// <para>The sheafs are locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// </remarks>
		int DeleteSheafs( PortableTypes sheafTypes, SpeckleFilters seekFilters, string storehouseRealm );

		/// <summary>
		/// Deletes the first sheaf with the specified speckle filters, which refer documents with the specified types.
		/// </summary>
		/// <param name="sheafTypes">The allowed sheaf types of the sheaf to delete.</param>
		/// <param name="seekFilters">The seek filters which identify the matching sheaf.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <remarks>
		/// <para>The sheaf is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// 
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// 
		/// <para>The method deletes the first encountered sheaf, not necessarily the sheaf which was added first. This behavior
		/// mVela.Imizes performance.</para>
		/// </remarks>
		void DeleteFirstSheaf( PortableTypes sheafTypes, SpeckleFilters seekFilters, string storehouseRealm );

		/// <summary>
		/// Deletes the sheaf with the specified storehouse key from the storehouse.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the sheaf to delete from the storehouse.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <remarks>
		/// <para>The sheaf is locked with <see cref="SheafLocks.Exclusive"/>.</para>
		/// </remarks>
		void DeleteSheaf( Locator storehouseKey, string storehouseRealm );

		/// <summary>
		/// Loads the sheafs with the specified speckle filters, which have the specified types.
		/// </summary>
		/// <param name="sheafTypes">The allowed types of the sheafs to load.</param>
		/// <param name="seekFilters">The seek filters which identify the speckles of the matching storehouse documents. This
		/// may be null or empty.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="paging">Information about how to process the documents in a paged manner. If this is null, no paging is
		/// used.</param>
		/// <param name="sheafLock">Lock to put on the sheafs.</param>
		/// <param name="collectedSheafs">List where to add the collected sheafs. If this is null, a new instance of
		/// <see cref="Sheafs"/> is used instead. This allows you to have lists sorted by different criteria.</param>
		/// <param name="deserializer">Is set with the deserializer of the last loaded sheaf.</param>
		/// <returns>Returns the loaded sheafs.</returns>
		/// <remarks>
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// </remarks>
		IKernelObservableList CollectSheafs( PortableTypes sheafTypes, SpeckleFilters seekFilters, string storehouseRealm
			, CollectPaging paging, SheafLocks sheafLock, IKernelObservableList collectedSheafs
			, out IDeserializer deserializer );

		/// <summary>
		/// Loads the first sheaf with the specified speckle filters, which has the specified types.
		/// </summary>
		/// <param name="sheafTypes">The allowed types of the sheaf to load.</param>
		/// <param name="seekFilters">The seek filters which identify the speckles of the matching storehouse documents. This
		/// may be null or empty.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="sheafLock">Lock to put on the sheaf.</param>
		/// <param name="deserializer">Is set with the deserializer of the loaded sheaf.</param>
		/// <returns>Returns the loaded sheaf, or null if none is found.</returns>
		/// <remarks>
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// 
		/// <para>The method returns the first encountered sheaf, not necessarily the sheaf which was added first. This behavior
		/// mVela.Imizes performance.</para>
		/// </remarks>
		ISheaf CollectFirstSheaf( PortableTypes sheafTypes, SpeckleFilters seekFilters, string storehouseRealm
			, SheafLocks sheafLock, out IDeserializer deserializer );

		/// <summary>
		/// Loads the sheaf with the specified storehouse key, which has the specified types.
		/// </summary>
		/// <param name="storehouseKey">Storehouse key (= document instance locator) of the sheaf to load.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="sheafTypes">The allowed types of the sheaf to load.</param>
		/// <param name="encryptionKey">The key used for the decryption of the sheaf.</param>
		/// <param name="sheafLock">Lock to put on the sheaf.</param>
		/// <param name="deserializer">Is set with the deserializer of the loaded sheaf.</param>
		/// <returns>Returns the loaded sheaf, or null if none is found.</returns>
		/// <remarks>
		/// <para>Types belonging to multiple inheritance trees can't be used in "sheafTypes" because the speckle filters are
		/// unique only within a single inheritance tree, which means that if multiple types are specified, the indexes of
		/// speckle filter which contain different informations may overlap.</para>
		/// </remarks>
		ISheaf LoadSheaf( Locator storehouseKey, string storehouseRealm, PortableTypes sheafTypes, byte[] encryptionKey
			, SheafLocks sheafLock, out IDeserializer deserializer );

		/// <summary>
		/// Loads the sheaf with the specified storehouse key, which has the specified type (or a derived one).
		/// </summary>
		/// <param name="storehouseKey">Storehouse key (= document instance locator) of the sheaf to load.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <param name="sheafType">The type of the sheaf to load; the sheaf to load may either have this type or a type derived
		/// from this (if it's found in the global portable types from <see cref="PortableTypes.Instance"/>).</param>
		/// <param name="sheafLock">Lock to put on the sheaf.</param>
		/// <param name="deserializer">Is set with the deserializer of the loaded sheaf.</param>
		/// <returns>Returns the loaded sheaf, or null if none is found.</returns>
		ISheaf LoadSheaf( Locator storehouseKey, string storehouseRealm, Type sheafType, SheafLocks sheafLock
			, out IDeserializer deserializer );

		/// <summary>
		/// Checks if the specified sheaf (already) exists in the specified storehouse realm and has the same content.
		/// </summary>
		/// <param name="sheaf">Sheaf to check if it (already) exists.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <returns>Returns the existence mode of the sheaf.</returns>
		/// <remarks>
		/// <para>Use the <see cref="SpeckleHasSheaf"/> method if you want to know if a given instance locator is already
		/// allocated in the specified realm.</para>
		/// </remarks>
		SheafExistenceModes SheafExists( ISheaf sheaf, string storehouseRealm );

		/// <summary>
		/// Gets the type of the sheaf with the specified storehouse key.
		/// </summary>
		/// <param name="storehouseKey">Storehouse key (= document instance locator) of the sheaf whose type to get.</param>
		/// <param name="storehouseRealm">The storehouse realm where this method has effect.</param>
		/// <returns>Returns the type of the sheaf, or null if none is found.</returns>
		Type GetSheafType( Locator storehouseKey, string storehouseRealm );

		#endregion Public methods
	}
}
