////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Crypto;
using Vela.CoreDocuments;

namespace Vela.Storehouse
{
	/// <summary>
	/// The entropy pool of a storehouse.
	/// </summary>
	/// <remarks>
	/// <para>A storehouse entropy is a normal sheaf, stored in encrypted form.</para>
	/// 
	/// <para>The main reason for the existence of this sheaf is to have a way to store storehouse-wide information, but
	/// without having to load it often.</para>
	/// </remarks>
	public class StorehouseEntropy : Sheaf
	{
		#region Public constants

		/// <summary>
		/// The text from which the instance locator of a storehouse entropy is created.
		/// </summary>
		public const string PredefinedInstanceLocatorText = "Storehouse entropy";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected StorehouseEntropy()
			: base( PredefinedInstanceLocator )
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="entropyAlg">Entropy algorithm.</param>
		public StorehouseEntropy( EntropyAlg entropyAlg )
			: base( PredefinedInstanceLocator )
		{
			EntropyAlg = entropyAlg;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private static Locator _PredefinedInstanceLocator;
		/// <summary>
		/// The instance locator of a storehouse entropy. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The value is the encoded <see cref="PredefinedInstanceLocatorText"/> text.</para>
		/// 
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		public static Locator PredefinedInstanceLocator
		{
			get
			{
				if( _PredefinedInstanceLocator == null )
				{
					byte[] buffer = KernelUtils.GetDeepTextEncoding().TextToArray( PredefinedInstanceLocatorText, false );

					_PredefinedInstanceLocator = new Locator( buffer );
				}

				return _PredefinedInstanceLocator;
			}
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private EntropyAlg _EntropyAlg;
		/// <summary>
		/// Entropy algorithm.
		/// </summary>
		[SerializeAttribute]
		public EntropyAlg EntropyAlg
		{
			get { return _EntropyAlg; }
			set
			{
				ValidationUtils.ObjectHasContent( value, "EntropyAlg" );

				_EntropyAlg = value;
			}
		}

		#endregion Public serializable properties
	}
}
