////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Mac;
using Vela.Crypto.Symmetric;
using Vela.Kernel;
using Vela.Kernel.Serialization;

namespace Vela.Storehouse
{
	/// <summary>
	/// Header for a storehouse.
	/// </summary>
	/// <remarks>
	/// <para>The header is stored in encrypted form, as an individual file.</para>
	/// 
	/// <para>The main reason for the existence of this sheaf is to have a way to store storehouse-wide information, like the
	/// password blender (which is used to generate the encryption key for the root key) and the root key.</para>
	/// 
	/// <para>The header must be kept to the minimum possible size (because it may be loaded often), and its content must never
	/// change.</para>
	/// </remarks>
	public sealed class StorehouseHeader : Sheaf
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		private StorehouseHeader()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="instanceLocator">Instance locator.</param>
		/// <param name="textEncoding">Text encoding.</param>
		/// <param name="macAlg">MAC algorithm.</param>
		/// <param name="compressionAlg">Compression algorithm. This may be null.</param>
		/// <param name="symmAlg">Symmetric cryptography algorithm.</param>
		/// <param name="passwordBlenderAlg">Password blender algorithm to use to generate the blended password. This is
		/// used to generate the encryption key of each referred sheaf.</param>
		/// <remarks>
		/// <para><see cref="SheafsRootKey"/> and <see cref="AttachmentsKey"/> are set here with random values.</para>
		/// </remarks>
		public StorehouseHeader( Locator instanceLocator, TextEncoding textEncoding, MacAlg macAlg
				, CompressionAlg compressionAlg, SymmAlg symmAlg, PasswordBlenderAlg passwordBlenderAlg )
			: base( instanceLocator )
		{
			TextEncoding = textEncoding;
			MacAlg = macAlg;
			CompressionAlg = compressionAlg;
			SymmAlg = symmAlg;
			PasswordBlenderAlg = passwordBlenderAlg;

			AttachmentsKey = EntropyPool.GetBytes( SymmAlg.AlgSizeBytes );
			SheafsRootKey = EntropyPool.GetBytes( SymmAlg.AlgSizeBytes );
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The time (in UTC format) from which the current object expires and should no longer be used in processing.
		/// READ-ONLY.
		/// </summary>
		public override DateTime ExpirationTime
		{
			get
			{
				return KernelUtils.MinTime( base.ExpirationTime, MacAlg.ExpirationTime, SymmAlg.ExpirationTime
					, PasswordBlenderAlg.ExpirationTime );
			}
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private TextEncoding _TextEncoding;
		/// <summary>
		/// Text encoding. READ-ONLY.
		/// </summary>
		public TextEncoding TextEncoding
		{
			get { return _TextEncoding; }
			private set
			{
				ValidationUtils.ObjectHasContent( value, "TextEncoding" );

				_TextEncoding = value;
			}
		}

		private MacAlg _MacAlg;
		/// <summary>
		/// MAC algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public MacAlg MacAlg
		{
			get { return _MacAlg; }
			private set
			{
				ValidationUtils.ObjectHasContent( value, "MacAlg" );

				_MacAlg = value;
			}
		}

		private CompressionAlg _CompressionAlg;
		/// <summary>
		/// Compression algorithm. This may be null. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is not null, the encryptable document is compressed before it is encrypted, fact which improves
		/// the security and may reduce the required storage space and network bandwidth.</para>
		/// </remarks>
		[SerializeAttribute( MustBeDeserialized = false )]
		public CompressionAlg CompressionAlg
		{
			get { return _CompressionAlg; }
			private set { _CompressionAlg = value; }
		}

		private SymmAlg _SymmAlg;
		/// <summary>
		/// Symmetric cryptography algorithm. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public SymmAlg SymmAlg
		{
			get { return _SymmAlg; }
			private set
			{
				ValidationUtils.ObjectHasContent( value, "SymmAlg" );

				_SymmAlg = value;
			}
		}

		private PasswordBlenderAlg _PasswordBlenderAlg;
		/// <summary>
		/// Password blender algorithm to use to generate the blended password. This is used to generate the encryption
		/// key of each referred sheaf. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>DO NOT EVER change this. If you do, the storehouse sheafs would become un-decryptable.</para>
		/// 
		/// <para>The encryption key of the referred sheaf must be a MAC between this key and the storehouse key of the referred
		/// sheaf. See the <see cref="Vela.CoreDocuments.PasswordBlenderAlg.GenerateBlendedPassword"/> method for how to
		/// generate the encryption key.</para>
		/// </remarks>
		[SerializeAttribute]
		public PasswordBlenderAlg PasswordBlenderAlg
		{
			get { return _PasswordBlenderAlg; }
			private set
			{
				ValidationUtils.ObjectHasContent( value, "PasswordBlenderAlg" );

				_PasswordBlenderAlg = value;
			}
		}

		private byte[] _AttachmentsKey;
		/// <summary>
		/// The key used to encrypt the storehouse files which are attached to the header. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>A storehouse provider derives an encryption password from this key. See
		/// <see cref="IStorehouseProvider.GetStorehouseAttachmentsPassword"/>.</para>
		/// </remarks>
		[SerializeAttribute]
		public byte[] AttachmentsKey
		{
			get { return _AttachmentsKey; }
			private set
			{
				ValidationUtils.ArrayHasContent( value, "AttachmentsKey" );

				_AttachmentsKey = value;
			}
		}

		private byte[] _SheafsRootKey;
		/// <summary>
		/// The root key used to symmetrically encrypt the sheafs. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public byte[] SheafsRootKey
		{
			get { return _SheafsRootKey; }
			private set
			{
				ValidationUtils.ArrayHasContent( value, "SheafsRootKey" );

				_SheafsRootKey = value;
			}
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Validates the current object.
		/// </summary>
		private void Validate()
		{
			if( AttachmentsKey.Length != SymmAlg.AlgSizeBytes )
			{
				throw new ValidationException( string.Format( "The size of the \"AttachmentsKey\" is {0} bytes instead of"
					+ " {1} bytes.", AttachmentsKey.Length, SymmAlg.AlgSizeBytes ) );
			}

			if( SheafsRootKey.Length != SymmAlg.AlgSizeBytes )
			{
				throw new ValidationException( string.Format( "The size of the \"SheafsRootKey\" is {0} bytes instead of"
					+ " {1} bytes.", SheafsRootKey.Length, SymmAlg.AlgSizeBytes ) );
			}
		}

		/// <summary>
		/// This method is automatically called before the current object is serialized.
		/// </summary>
		private void Serializing()
		{
			Validate();
		}

		/// <summary>
		/// This method is automatically called after the current object has been deserialized.
		/// </summary>
		private void Deserialized()
		{
			Validate();
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Generates the encryption key for the header, using the specified salt and user password.
		/// </summary>
		/// <param name="salt">Salt with which to generate the blended password.</param>
		/// <param name="userPassword">The user password.</param>
		/// <returns>Returns the generated the encryption key.</returns>
		/// <remarks>
		/// <para>There is no need to align the generated key because it will be aligned during the encryption, later.</para>
		/// </remarks>
		public byte[] GenerateBlendedUserPassword( byte[] salt, string userPassword )
		{
			return PasswordBlenderAlg.GenerateBlendedPassword( salt, userPassword, true );
		}

		/// <summary>
		/// Generates an encryption key for the sheaf with the specified storehouse key, using the specified root key.
		/// </summary>
		/// <param name="storehouseKey">The storehouse key of the sheaf for which to generate the encryption key.</param>
		/// <returns>Returns the generated the encryption key.</returns>
		/// <remarks>
		/// <para>There is no need to align the generated key because it will be aligned during the encryption, later.</para>
		/// </remarks>
		public byte[] GenerateSheafEncryptionKey( Locator storehouseKey )
		{
			ValidationUtils.ObjectHasContent( storehouseKey, "storehouseKey" );

			return PasswordBlenderAlg.GenerateBlendedPassword( storehouseKey.Value, SheafsRootKey, false );
		}

		#endregion Public methods
	}
}
