////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.Storehouse
{
	/// <summary>
	/// Contract for a storehouse provider.
	/// </summary>
	public abstract class StorehouseProvider : IStorehouseProvider
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected StorehouseProvider()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="access">The storehouse access.</param>
		/// <param name="fileName">The name of the file where the storehouse header is stored.</param>
		/// <param name="headerUserPassword">The user password used to symmetrically encrypt the header.</param>
		/// <param name="header">The storehouse header (which is stored in a symmetrically encrypted form). This may be null if
		/// the storehouse doesn't exist yet.</param>
		/// <param name="serializer">The serializer of the sheafs.</param>
		protected StorehouseProvider( IStorehouseAccess access, string fileName, string headerUserPassword
			, StorehouseHeader header, ISerializer serializer )
		{
			Access = access;
			FileName = fileName;
			HeaderUserPassword = headerUserPassword;
			if( header != null )
				Header = header;
			Serializer = serializer;
		}

		#endregion Constructors

		#region Public non-serializable properties

		/// <summary>
		/// The size of the password used to encrypt the storehouse files which are attached to the storehouse header. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This is used to derive the encryption password (see
		/// <see cref="IStorehouseProvider.GetStorehouseAttachmentsPassword"/>).</para>
		/// </remarks>
		public abstract int StorehouseAttachmentsPasswordSize { get; }

		private bool _StorehouseExists;
		/// <summary>
		/// Is true if the referred storehouse exists and is accessible. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>The getter checks if the file exists and if the storehouse is accessible. If the storehouse is accessible,
		/// this is not checked on subsequent calls (for the current provider), it's always considered true.</para>
		/// </remarks>
		public virtual bool StorehouseExists
		{
			get
			{
				bool fileExists = File.Exists( FileName );

				// If the file doesn't exist, check again if the storehouse is accessible.
				if( !fileExists )
					_StorehouseExists = false;

				// Check if the storehouse is truly accessible.
				if( fileExists && !_StorehouseExists )
				{
					IStorehouseTransaction transaction = CreateTransaction();

					try
					{
						// The type of the speckles to count is irrelevant because if the result is 0, it's fine; however, an
						// exception is not fine.
						int count = transaction.CountSpeckles( new PortableTypes( typeof( StorehouseConfiguration ) )
							, null, StorehouseRealms.StorehouseConfiguration );

						_StorehouseExists = count >= 0;

						// Commit the transaction.
						transaction.Commit();
					}
					catch( Exception ex )
					{
						// Rolls back the transaction.
						transaction.Rollback();

						throw ExceptionUtils.PreserveStackTrace( ex );
					}
				}

				return fileExists && _StorehouseExists;
			}
		}

		/// <summary>
		/// If this is true then the storehouse is managed by a storehouse server engine, else it's standalone / embedded.
		/// READ-ONLY.
		/// </summary>
		public abstract bool UsesServerEngine { get; }

		/// <summary>
		/// Set of scripts used to work with the storehouse. READ-ONLY.
		/// </summary>
		public abstract IStorehouseScripts Scripts { get; protected set; }

		/// <summary>
		/// Placeholders used in storehouse scripts. READ-ONLY.
		/// </summary>
		public abstract IStorehouseScriptPlaceholders ScriptPlaceholders { get; protected set; }

		/// <summary>
		/// The name of the storehouse as it must be used by the storehouse server (if one exists). READ-ONLY.
		/// </summary>
		public string ServerStorehouseName
		{
			get { return Path.GetFileNameWithoutExtension( FileName ); }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private IStorehouseAccess _Access;
		/// <summary>
		/// The storehouse access. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public IStorehouseAccess Access
		{
			get { return _Access; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Access" );

				_Access = value;
			}
		}

		private string _FileName;
		/// <summary>
		/// The name of the file where the storehouse header is stored. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string FileName
		{
			get { return _FileName; }
			protected set
			{
				ValidationUtils.TextHasContent( value, "FileName" );

				_FileName = value;
			}
		}

		private string _HeaderUserPassword;
		/// <summary>
		/// The user password used to symmetrically encrypt the header. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public string HeaderUserPassword
		{
			get { return _HeaderUserPassword; }
			protected set
			{
				ValidationUtils.Password( value, "HeaderUserPassword" );

				_HeaderUserPassword = value;
			}
		}

		private StorehouseHeader _Header;
		/// <summary>
		/// The storehouse header (which is stored in a symmetrically encrypted form). READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		[SerializeAttribute]
		public StorehouseHeader Header
		{
			get
			{
				if( _Header != null )
					return _Header;

				LoadHeader();

				return _Header;
			}
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Header" );

				_Header = value;
			}
		}

		private ISerializer _Serializer;
		/// <summary>
		/// The serializer used to serialize the documents. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public ISerializer Serializer
		{
			get { return _Serializer; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Serializer" );

				_Serializer = value;
			}
		}

		#endregion Public serializable properties

		#region Private methods

		/// <summary>
		/// Gets the standard storehouse realms.
		/// </summary>
		/// <returns>Returns the standard storehouse realms.</returns>
		protected virtual Strings GetStandardStorehouseRealms()
		{
			return StorehouseUtils.GetStandardStorehouseRealms();
		}

		/// <summary>
		/// Replaces the placeholders (defined in <see cref="StorehouseConsts"/>) with the specified values.
		/// </summary>
		/// <param name="text">Text which contains the placeholders to replace.</param>
		/// <param name="storehouseFileName">Name of file attached to the storehouse.</param>
		/// <param name="serverStorehouseName">Server storehouse name.</param>
		/// <param name="attachmentsPassword">Storehouse attachments password.</param>
		/// <returns>Returns the text with the replaced placeholders.</returns>
		protected string ReplacePlaceholders( string text, string storehouseFileName, string serverStorehouseName
			, string attachmentsPassword )
		{
			ValidationUtils.TextHasContent( text, "text" );
			ValidationUtils.TextHasContent( storehouseFileName, "storehouseFileName" );
			ValidationUtils.TextHasContent( serverStorehouseName, "serverStorehouseName" );
			ValidationUtils.TextHasContent( attachmentsPassword, "attachmentsPassword" );

			text = TextUtils.Replace( text, StorehouseConsts.StorehouseFileNamePlaceholder, storehouseFileName );
			text = TextUtils.Replace( text, StorehouseConsts.ServerStorehouseNamePlaceholder, serverStorehouseName );
			text = TextUtils.Replace( text, StorehouseConsts.StorehouseAttachmentsPasswordPlaceholder, attachmentsPassword );

			return text;
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Checks if the specified exception represents a timeout (which has occurred while accessing the storehouse).
		/// </summary>
		/// <param name="exception">Exception to check if it represents a timeout. This may be null.</param>
		/// <returns>Returns true if the specified exception represents a timeout.</returns>
		public abstract bool IsTimeoutException( Exception exception );

		/// <summary>
		/// Creates a new storehouse. If the storehouse already exists, it's overwritten.
		/// </summary>
		/// <param name="adminConnectionString">The administration connection string with which a storehouse server can create
		/// a storehouse. This may be null or empty for a storehouse provider which doesn't use a storehouse server.</param>
		public abstract void CreateStorehouse( string adminConnectionString );

		/// <summary>
		/// Creates the specified storehouse realms.
		/// </summary>
		/// <param name="storehouseRealms">Storehouse realms to create. This may be empty, but not null.</param>
		public abstract void CreateStorehouseRealms( Strings storehouseRealms );

		/// <summary>
		/// Opens the storehouse with the current provider.
		/// </summary>
		public virtual void OpenStorehouse()
		{
			// Create the storehouse realms.
			CreateStorehouseRealms( GetStandardStorehouseRealms() );
		}

		/// <summary>
		/// Compacts the storehouse.
		/// </summary>
		/// <param name="adminConnectionString">The administration connection string with which a storehouse server can compact
		/// the storehouse. This may be null or empty for a storehouse provider which doesn't use a storehouse server.</param>
		/// <param name="alsoVerifyAndRepair">If this is true then, before being compacted, the storehouse is verified for
		/// consistency, and is repaired if it's necessary and possible.</param>
		/// <remarks>
		/// <para>The empty space is removed.</para>
		/// 
		/// <para>If the storehouse doesn't support this operation, no exception is thrown.</para>
		/// </remarks>
		public abstract void CompactStorehouse( string adminConnectionString, bool alsoVerifyAndRepair );

		/// <summary>
		/// Creates new a transaction in the storehouse.
		/// </summary>
		public abstract IStorehouseTransaction CreateTransaction();

		/// <summary>
		/// Gets the file name of the main storehouse file which is attached to the storehouse header.
		/// </summary>
		/// <returns>Returns the file name of the main storehouse file which is attached to the storehouse header.</returns>
		public virtual string GetStorehouseMainAttachmentFileName()
		{
			return Path.ChangeExtension( FileName, FileExtensions.StorehouseMainAttachment );
		}

		/// <summary>
		/// Gets the password used to encrypt the storehouse files which are attached to the header.
		/// </summary>
		/// <returns>Returns the password used to encrypt the storehouse files which are attached to the header.</returns>
		/// <remarks>
		/// <para>The storehouse header is loaded in here in order to get the attachments key, and this could thrown an
		/// exception.</para>
		/// </remarks>
		public virtual string GetStorehouseAttachmentsPassword()
		{
			string password = Convert.ToBase64String( Header.AttachmentsKey );

			return password.Substring( 0, Math.Min( password.Length, StorehouseAttachmentsPasswordSize ) );
		}

		/// <summary>
		/// Saves the specified storehouse header to the storehouse. RESERVED for internal use.
		/// </summary>
		/// <param name="header">The storehouse header to save.</param>
		/// <param name="headerUserPassword">The user password used to symmetrically encrypt the header.</param>
		/// <remarks>
		/// <para>Make sure that you always call this method from the same thread in order to be sure that no other thread
		/// accidentally overrides the header with the old header password.</para>
		/// </remarks>
		public virtual void SaveHeader( StorehouseHeader header, string headerUserPassword )
		{
			ValidationUtils.ObjectHasContent( header, "header" );
			ValidationUtils.Password( headerUserPassword, "headerUserPassword" );

			// Generate the salt and the blended header user password.
			byte[] salt = EntropyPool.GetBytes( header.SymmAlg.AlgSizeBytes );
			byte[] blendedHeaderUserPassword = header.GenerateBlendedUserPassword( salt, headerUserPassword );

			SymmEncryptedEnvelope.SaveToFileVerify( header.InstanceLocator, header, salt, blendedHeaderUserPassword
				, header.MacAlg, header.CompressionAlg, header.SymmAlg, header.PasswordBlenderAlg, FileName, Serializer, true );

			// Set the underlying field (not the property) in order to avoid to lazy load the header.
			_Header = header;

			// Change the password.
			HeaderUserPassword = headerUserPassword;
		}

		/// <summary>
		/// Loads the storehouse header. An exception is thrown if the header is missing. RESERVED for internal use.
		/// </summary>
		/// <remarks>
		/// <para>If the header is already loaded, it's not loaded again.</para>
		/// 
		/// <para>This is automatically called when the header is lazy loaded.</para>
		/// 
		/// <para>The <see cref="Serializer"/> property (of the storehouse provider) is set in here.</para>
		/// </remarks>
		public virtual void LoadHeader()
		{
			// Don't load the header if it's already loaded.
			if( _Header != null )
				return;

			try
			{
				StorehouseHeader header;

				// Load the envelope.
				IDeserializer deserializer;
				SymmEncryptedEnvelope see = SerializationUtils.LoadFromFile( FileName, typeof( SymmEncryptedEnvelope )
					, out deserializer ) as SymmEncryptedEnvelope;
				ValidationUtils.ObjectHasContent( see, "see" );

				// Generate the blended header user password.
				byte[] blendedHeaderUserPassword = see.PasswordBlenderAlg.GenerateBlendedPassword( see.Salt
					, HeaderUserPassword, true );

				// Decrypt the content.
				IDeserializer encryptedContentDeserializer;
				EncryptedEnvelope.LoadFromEnvelope( see, blendedHeaderUserPassword, typeof( StorehouseHeader )
					, out encryptedContentDeserializer );

				header = see.EncryptableDocument as StorehouseHeader;
				ValidationUtils.ObjectType<StorehouseHeader>( header, "header" );

				// Set the underlying field (not the property) in order to avoid to lazy load the header (since this method is
				// used to lazy load the header).
				_Header = header;
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsAbort( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				throw new FailedOpenStorehouseException( "Failed to open the storehouse. The header was not found."
					+ " Ensure that the password is correct.", ex );
			}
		}

		/// <summary>
		/// Changes the password of the user used to symmetrically encrypt the header.
		/// </summary>
		/// <param name="headerUserPassword">The new user password used to symmetrically encrypt the header.</param>
		/// <remarks>
		/// <para>Make sure that you always call this method from the same thread in order to be sure that no other thread
		/// accidentally overrides the header with the old header password.</para>
		/// </remarks>
		public virtual void ChangeHeaderUserPassword( string headerUserPassword )
		{
			SaveHeader( Header, headerUserPassword );
		}

		#endregion Public methods
	}
}
