////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

namespace Vela.Storehouse
{
	/// <summary>
	/// Storehouse realms.
	/// </summary>
	/// <remarks>
	/// <para>Realms exist in order to increase the performance and the scalability of a storehouse, while maintaining the
	/// ability to work in an object-oriented-programming manner, like being able to load in a single storehouse instruction all
	/// the sheafs derived from a given type.</para>
	/// 
	/// <para>It is highly recommended to put in a realm only sheafs derived from a single type. The more base types are mixed,
	/// the higher is the probability of storehouse resource contention.</para>
	/// 
	/// <para>Converting realms to a speckle filter leads to a catastrophic level of storehouse resource contention, at least for
	/// the service tester application.</para>
	/// 
	/// <br/>
	/// 
	/// <para><b>Client-service realms</b></para>
	/// 
	/// <para>It is highly recommended to create separate client-service realms based only on inheritance / business
	/// requirements, else this separation can create huge difficulties in managing the sheafs from the realms, especially from a
	/// UI perspective because that's where it's likely that the sheafs have to be handled together (from both realms).</para>
	/// 
	/// <para>An example for this is given by letters. If the sent and the received letters were to be put in different realms,
	/// the entire code which works with letters would have to know both the instance locator and the realm of each letter.
	/// However, there would be a need to actually have 3 realms: client, service and offline. Moreover, the pattern would
	/// have to maintained for all the sheafs which can be stored in the same way.</para>
	/// 
	/// <para>The costs appear too great relative to the possible gain: better scalability due to the fact that the client and
	/// the service modules of the same application would encounter less storehouse resource contention. It also has to be
	/// considered the fact that it's likely that this is an asymmetric argument, that is, only large services would benefit
	/// from this, fact which makes the possible gain irrelevant since the client module would be underused in such a case.
	/// </para>
	/// 
	/// <para>As a bottom line, it has to be considered that the current storehouse design (with sheafs and speckles) was
	/// chosen for the ability to use object-oriented-programming patterns (while working with the storehouse). Therefore,
	/// chosing realms based on "if" coding structures rather than inheritance / business requirements is a bad choice; such
	/// "if" coding structures would have to be used for example like this: for a client letter, use a client action progress.
	/// </para>
	/// </remarks>
	public static class StorehouseRealms
	{
		#region Public constants

		/// <summary>
		/// Contains storehouse configuration sheafs (like the configuration or the entropy).
		/// </summary>
		public const string StorehouseConfiguration = "StorehouseConfiguration";

		/// <summary>
		/// Contains UPN fingerprint reservations.
		/// </summary>
		public const string UpnFingerprintReservations = "UpnFingerprintReservations";

		/// <summary>
		/// Contains sheafs which contain a processing identity.
		/// </summary>
		public const string ProcessingIdentityContainers = "ProcessingIdentityContainers";

		/// <summary>
		/// Contains photos.
		/// </summary>
		public const string Photos = "Photos";

		/// <summary>
		/// Contains accounts.
		/// </summary>
		public const string Accounts = "Accounts";

		/// <summary>
		/// Contains ephemerids.
		/// </summary>
		public const string Ephemerids = "Ephemerids";

		/// <summary>
		/// Contains action progresses.
		/// </summary>
		public const string ActionProgresses = "ActionProgresses";

		/// <summary>
		/// Contains action requests which are to be sent by an action client to an action service, for processing.
		/// </summary>
		public const string ActionRequestsToSend = "ActionRequestsToSend";

		/// <summary>
		/// Contains action requests which were received by an action service from an action client, for processing.
		/// </summary>
		public const string ServiceReceivedActionRequests = "ServiceReceivedActionRequests";

		/// <summary>
		/// Contains action responses which were sent by an action service to an action client (after processing the associated
		/// received action requests).
		/// </summary>
		public const string ServiceSentActionResponses = "ServiceSentActionResponses";

		/// <summary>
		/// Contains interaction sessions which are currently active.
		/// </summary>
		public const string InteractionSessions = "InteractionSessions";

		/// <summary>
		/// Contains stamps.
		/// </summary>
		public const string Stamps = "Stamps";

		/// <summary>
		/// Contains posts (like letters sent to or received from another peer).
		/// </summary>
		public const string Posts = "Posts";

		/// <summary>
		/// Contains charged fee sheafs.
		/// </summary>
		public const string ChargedFees = "ChargedFees";

		/// <summary>
		/// Contains generic storage sheafs.
		/// </summary>
		public const string Storage = "Storage";

		/// <summary>
		/// Contains archived sheafs.
		/// </summary>
		/// <remarks>
		/// <para>This should be used only for old sheafs, sheafs which are no longer used in processing.</para>
		/// </remarks>
		public const string Archive = "Archive";

		#endregion Public constants
	}
}
