////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Reflection;
using Vela.Cloning;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Kernel;
using Vela.Serialization;

namespace Vela.Storehouse
{
	/// <summary>
	/// Storehouse utilities.
	/// </summary>
	public static class StorehouseUtils
	{
		#region Public static methods

		/// <summary>
		/// Gets the storehouse realms from the specified type.
		/// </summary>
		/// <param name="storehouseRealmsType">Type which contains the storehouse realms to get. All the public string constants
		/// from this type are considered realm names.</param>
		/// <returns>Returns the storehouse realms from the specified type.</returns>
		public static Strings GetStorehouseRealms( Type storehouseRealmsType )
		{
			ValidationUtils.ObjectHasContent( storehouseRealmsType, "storehouseRealmsType" );

			if( !storehouseRealmsType.IsClass )
			{
				throw new ValidationException( string.Format( "Storehouse realms can be extracted only from a class, not"
					+ " from \"{0}\".", storehouseRealmsType.FullName ) );
			}

			Strings storehouseRealms = new Strings();

			FieldInfo[] fieldInfos = storehouseRealmsType.GetFields( BindingFlags.Public | BindingFlags.Static );
			if( ArrayUtils.NoContent( fieldInfos ) ) return storehouseRealms;

			foreach( FieldInfo fieldInfo in fieldInfos )
			{
				if( fieldInfo.IsLiteral && !fieldInfo.IsInitOnly )
				{
					string fieldValue = fieldInfo.GetRawConstantValue() as string;
					storehouseRealms.Add( fieldValue );
				}
			}

			return storehouseRealms;
		}

		/// <summary>
		/// Gets the standard storehouse realms.
		/// </summary>
		/// <returns>Returns the standard storehouse realms.</returns>
		public static Strings GetStandardStorehouseRealms()
		{
			return GetStorehouseRealms( typeof( StorehouseRealms ) );
		}

		/// <summary>
		/// Gets the storehouse provider type from the specified type name.
		/// </summary>
		/// <param name="providerTypeName">The type name of the storehouse provider to instantiate.</param>
		/// <returns>Returns the storehouse provider type.</returns>
		public static Type GetStorehouseProviderType( string providerTypeName )
		{
			Type storehouseProviderType = PortableTypes.FindGlobal( providerTypeName );
			ValidationUtils.ObjectHasContent( storehouseProviderType, "storehouseProviderType" );

			// Check if the type implements "IStorehouseProvider".
			if( !typeof( IStorehouseProvider ).IsAssignableFrom( storehouseProviderType ) )
			{
				throw new ValidationException( string.Format( "The storehouse provider type \"{0}\" doesn't implement"
					+ " interface \"{1}\".", storehouseProviderType.FullName, typeof( IStorehouseProvider ).FullName ) );
			}

			return storehouseProviderType;
		}

		/// <summary>
		/// Creates a storehouse provider from the specified type name. The storehouse is not created in here, and the
		/// <see cref="CreateStorehouse"/> method should be called afterwards.
		/// </summary>
		/// <param name="providerTypeName">The type name of the storehouse provider to instantiate.</param>
		/// <param name="access">The storehouse access.</param>
		/// <param name="fileName">The name of the file where the storehouse header is stored.</param>
		/// <param name="headerUserPassword">The user password used to symmetrically encrypt the header.</param>
		/// <param name="header">The storehouse header (which is to be symmetrically encrypted).</param>
		/// <param name="serializer">The serializer of the sheafs.</param>
		/// <returns>Returns the newly created storehouse provider.</returns>
		public static IStorehouseProvider CreateStorehouseProvider( string providerTypeName, IStorehouseAccess access
			, string fileName, string headerUserPassword, StorehouseHeader header, ISerializer serializer )
		{
			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				return Activator.CreateInstance( GetStorehouseProviderType( providerTypeName ), access, fileName
					, headerUserPassword, header, serializer ) as IStorehouseProvider;
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		/// <summary>
		/// Creates the storehouse with the specified provider.
		/// </summary>
		/// <param name="storehouseProvider">The storehouse provider.</param>
		/// <param name="adminConnectionString">The administration connection string with which a storehouse server can create
		/// a storehouse. This may be null or empty for a storehouse provider which doesn't use a storehouse server.</param>
		/// <param name="entropy">The storehouse entropy.</param>
		/// <param name="configuration">The storehouse configuration (which is to be symmetrically encrypted).</param>
		public static void CreateStorehouse( IStorehouseProvider storehouseProvider, string adminConnectionString
			, StorehouseEntropy entropy, StorehouseConfiguration configuration )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( entropy, "entropy" );
			ValidationUtils.ObjectHasContent( configuration, "configuration" );

			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				// Create the physical storehouse.
				storehouseProvider.CreateStorehouse( adminConnectionString );

				storehouseProvider.OpenStorehouse();

				// Save the header.
				storehouseProvider.SaveHeader( storehouseProvider.Header, storehouseProvider.HeaderUserPassword );

				if( !storehouseProvider.StorehouseExists )
				{
					throw new ValidationException( "There storehouse is missing or is invalid.", null, null
						, new ExceptionDetail( "Server storehouse name", storehouseProvider.ServerStorehouseName )
						, new ExceptionDetail( "File name", storehouseProvider.FileName ) );
				}

				#region Add the entropy and configuration

				IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

				try
				{
					// Clone the entropy algorithm because it is likely also the global entropy algorithm. This means that if it
					// weren't cloned, when the entropy were saved, its content would change and when it were reloaded to be
					// verified, an exception would be thrown.
					StorehouseEntropy entropyClone = Cloner.Deep<StorehouseEntropy>( entropy );

					// Add the entropy and configuration.
					transaction.AddEntropy( entropyClone );
					transaction.AddConfiguration( configuration );

					// Commit the transaction.
					transaction.Commit();
				}
				catch( Exception ex )
				{
					// Rolls back the transaction.
					transaction.Rollback();

					throw ExceptionUtils.PreserveStackTrace( ex );
				}

				#endregion Add the entropy and configuration
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		/// <summary>
		/// Opens the storehouse with the specified file name.
		/// </summary>
		/// <param name="providerTypeName">The type name of the storehouse provider to instantiate.</param>
		/// <param name="access">The storehouse access.</param>
		/// <param name="fileName">The name of the file where the storehouse header is stored.</param>
		/// <param name="headerUserPassword">The user password used to symmetrically encrypt the header.</param>
		/// <param name="serializer">The serializer of the sheafs.</param>
		/// <returns>Returns the newly created storehouse provider.</returns>
		public static IStorehouseProvider OpenStorehouse( string providerTypeName, IStorehouseAccess access, string fileName
			, string headerUserPassword, ISerializer serializer )
		{
			LogManager logManager = LogManager.DebugStartMethod( MethodBase.GetCurrentMethod() );

			try
			{
				IStorehouseProvider storehouseProvider = Activator.CreateInstance( GetStorehouseProviderType( providerTypeName )
					, access, fileName, headerUserPassword, null, serializer ) as IStorehouseProvider;

				storehouseProvider.OpenStorehouse();

				// Ensure that the header is loaded.
				storehouseProvider.LoadHeader();

				return storehouseProvider;
			}
			finally
			{
				LogManager.DebugEndMethod( logManager );
			}
		}

		/// <summary>
		/// Updates the entropy of the specified storehouse (with the current entropy).
		/// </summary>
		/// <param name="storehouseProvider">The storehouse provider whose entropy to update.</param>
		/// <remarks>
		/// <para>The entropy is first loaded from the storehouse, the application's entropy state is saved in it, and then the
		/// entropy is updated in the storehouse.</para>
		/// 
		/// <para>During the execution of this method, the entropy is locked with <see cref="SheafLocks.Update"/>.</para>
		/// 
		/// <para>Because a transaction is created in here, if you call this method while another transaction is active, it's
		/// possible that a lock timeout will occur.</para>
		/// </remarks>
		public static void UpdateEntropy( IStorehouseProvider storehouseProvider )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the entropy document.
				StorehouseEntropy entropy = transaction.LoadEntropy( SheafLocks.Update );

				// Merge the entropy algorithm from the storehouse with the global one.
				EntropyPool.Merge( entropy.EntropyAlg );

				// Update the entropy document.
				transaction.UpdateEntropy( entropy );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the entropy of the specified storehouse. An exception is thrown if the entropy is missing.
		/// </summary>
		/// <param name="storehouseProvider">The storehouse provider whose entropy to load.</param>
		/// <remarks>
		/// <para>During the execution of this method, the entropy is locked with <see cref="SheafLocks.Shared"/>.</para>
		/// 
		/// <para>Because a transaction is created in here, if you call this method while another transaction is active, it's
		/// possible that a lock timeout will occur.</para>
		/// </remarks>
		public static StorehouseEntropy LoadEntropy( IStorehouseProvider storehouseProvider )
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the entropy.
				StorehouseEntropy entropy = transaction.LoadEntropy( SheafLocks.Shared );

				// Commit the transaction.
				transaction.Commit();

				return entropy;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the configuration from a storehouse. An exception is thrown if the configuration is missing.
		/// </summary>
		/// <param name="transaction">The transaction with which to load the configuration.</param>
		/// <param name="sheafLock">The lock to be used for the sheaf to load.</param>
		/// <returns>Returns the loaded configuration.</returns>
		/// <remarks>
		/// <para>During the execution of this method, the configuration is locked with "sheafLock".</para>
		/// </remarks>
		public static TConfiguration LoadConfiguration<TConfiguration>( IStorehouseTransaction transaction
				, SheafLocks sheafLock = SheafLocks.Shared )
			where TConfiguration : StorehouseConfiguration
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );

			TConfiguration configuration = transaction.LoadConfiguration( sheafLock ) as TConfiguration;
			if( configuration == null )
			{
				throw new ValidationException( string.Format( "A storehouse configuration must be derived from \"{0}\"."
					, typeof( TConfiguration ).FullName ) );
			}

			return configuration;
		}

		/// <summary>
		/// Loads the configuration of the specified storehouse. An exception is thrown if the configuration is missing.
		/// </summary>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="sheafLock">The lock to be used for the sheaf to load.</param>
		/// <returns>Returns the loaded configuration.</returns>
		/// <remarks>
		/// <para>During the execution of this method, the configuration is locked with <see cref="SheafLocks.Shared"/>.</para>
		/// 
		/// <para>Because a transaction is created in here, if you call this method while another transaction is active, it's
		/// possible that a lock timeout will occur.</para>
		/// </remarks>
		public static TConfiguration LoadConfiguration<TConfiguration>( IStorehouseProvider storehouseProvider
				, SheafLocks sheafLock = SheafLocks.Shared )
			where TConfiguration : StorehouseConfiguration
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Load the configuration.
				TConfiguration configuration = LoadConfiguration<TConfiguration>( transaction, sheafLock );

				// Commit the transaction.
				transaction.Commit();

				return configuration;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		/// <summary>
		/// Loads the sheaf with the specified type and instance locator, from the specified storehouse and realm. An exception
		/// is the sheaf is not found.
		/// </summary>
		/// <typeparam name="TSheaf">The type of the sheaf to load.</typeparam>
		/// <param name="storehouseProvider">Storehouse provider from which to create a storehouse transaction.</param>
		/// <param name="instanceLocator">The instance locator of the sheaf to load.</param>
		/// <param name="storehouseRealm">The storehouse realm from which to load the sheaf.</param>
		/// <returns>Returns the loaded sheaf.</returns>
		/// <remarks>
		/// <para>During the execution of this method, the loaded sheaf is locked with <see cref="SheafLocks.Shared"/>.</para>
		/// </remarks>
		public static TSheaf LoadSheaf<TSheaf>( IStorehouseProvider storehouseProvider, Locator instanceLocator
				, string storehouseRealm )
			where TSheaf : class, ISheaf
		{
			ValidationUtils.ObjectHasContent( storehouseProvider, "storehouseProvider" );
			ValidationUtils.ObjectHasContent( instanceLocator, "instanceLocator" );
			ValidationUtils.TextHasContent( storehouseRealm, "storehouseRealm" );

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				IDeserializer deserializer;
				TSheaf loadedSheaf = transaction.LoadSheaf( instanceLocator, storehouseRealm, typeof( TSheaf )
					, SheafLocks.Shared, out deserializer ) as TSheaf;

				if( loadedSheaf == null )
				{
					throw new FatalCrossServiceException( "The sheaf is missing.", null
						, new ExceptionDetail( "Sheaf type", typeof( TSheaf ).FullName )
						, new ExceptionDetail( "Sheaf locator", instanceLocator ) );
				}

				// Commit the transaction.
				transaction.Commit();

				return loadedSheaf;
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		#endregion Public static methods
	}
}
