////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Text;
using Vela.Kernel;
using Vela.Serialization;

namespace Vela.XmlSerialization
{
	/// <summary>
	/// A textual hash serializer writer.
	/// </summary>
	/// <remarks>
	/// <para>The nodes with null or non-scalar values are not written to the output.</para>
	/// 
	/// <para>The items (<see cref="TextSerializationNode"/>) from the writer are sorted, so the order of serialization doesn't
	/// change on various platforms.</para>
	/// </remarks>
	public class TextHashSerializerWriter : KernelSortedList<string, TextSerializationNode>, ITextSerializerWriter
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="serializeValueTypeName">If this is true then the type names of the serialized values are stored in the
		/// accumulator. If this is true, the compatibility level of different versions of a serialized document is
		/// limited (see the "Serialization versioning" section from the documentation of the 
		/// <see cref="Vela.Serialization.IHashSerializer"/> interface). This must be true for a hash serializer which will be
		/// used to test the equality of documents.</param>
		/// <param name="accumulator">The accumulator where to write the output.</param>
		public TextHashSerializerWriter( bool serializeValueTypeName, StringBuilder accumulator )
			: base( StringInsensitiveComparer.Instance )
		{
			ValidationUtils.ObjectHasContent( accumulator, "accumulator" );

			SerializeValueTypeName = serializeValueTypeName;
			OutputAccumulator = accumulator;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// If this is true then the type names of the serialized values are stored in the accumulator. READ-ONLY.
		/// </summary>
		/// <remarks>
		/// <para>If this is true, the compatibility level of different versions of a serialized document is limited (see the
		/// "Serialization versioning" section from the documentation of the 
		/// <see cref="Vela.Serialization.IHashSerializer"/> interface).</para>
		/// 
		/// <para>This must be true for a hash serializer which will be used to test the equality of documents. This is because
		/// the parameterless constructor may set some non-serializable properties which can make two (otherwise identical) types
		/// different.</para>
		/// </remarks>
		private bool SerializeValueTypeName;

		/// <summary>
		/// The accumulator where to write the serializer's output.
		/// </summary>
		private StringBuilder OutputAccumulator;

		#endregion Private fields

		#region Private methods

		/// <summary>
		/// Writes an item to the output of the serializer.
		/// </summary>
		/// <param name="serializationName">The serialization name of the object which is to be written to the output of the
		/// serializer writer.</param>
		/// <param name="value">The value of the object which is to be written to the output of the serializer writer. This may
		/// be null or empty.</param>
		/// <param name="valueTypeName">The type name of the value to be written to the output of the serializer writer.</param>
		/// <remarks>
		/// <para>Before calling this method, ensure that the value is a scalar object.</para>
		/// </remarks>
		private void WriteItemOutput( string serializationName, string value, string valueTypeName )
		{
			ValidationUtils.TextHasContent( serializationName, "serializationName" );
			ValidationUtils.TextHasContent( valueTypeName, "valueTypeName" );

			// For maximum compatibility between different serialization versions, ignore null values.
			if( (value == null) && !SerializeValueTypeName )
				return;

			// The serialization name.
			OutputAccumulator.Append( serializationName );

			// The value type name.
			if( SerializeValueTypeName )
			{
				// The separator between name and value type name.
				OutputAccumulator.Append( TextHashSerializerConsts.NameTypeSeparator );

				OutputAccumulator.Append( valueTypeName );
			}

			if( value != null )
			{
				// The separator between metadata (= name or value type name) and value.
				OutputAccumulator.Append( TextHashSerializerConsts.MetadataValueSeparator );

				// The value.
				OutputAccumulator.Append( value );
			}

			// The separator between items.
			OutputAccumulator.Append( TextHashSerializerConsts.ItemSeparator );
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Gets the key from the specified item.
		/// </summary>
		/// <param name="item">Item to get the key from.</param>
		/// <returns>Returns the key from the specified item.</returns>
		/// <remarks>
		/// <para><see cref="TextSerializationNode.SerializationName"/> is used as key.</para>
		/// </remarks>
		public override string GetKey( TextSerializationNode item )
		{
			ValidationUtils.ObjectHasContent( item, "item" );

			return item.SerializationName;
		}

		/// <summary>
		/// Adds a new class type name node to the writer.
		/// </summary>
		/// <param name="serializationName">Serialization name.</param>
		/// <param name="type">Class type.</param>
		public void AddClassTypeName( string serializationName, Type type )
		{
			ValidationUtils.ObjectHasContent( type, "type" );

			SerializationUtils.ValidateSignableType( type );

			TextSerializationNode node = TextSerializationNode.GetClassNode( serializationName, type.FullName );

			Add( node );
		}

		/// <summary>
		/// Adds a new list type name node to the writer.
		/// </summary>
		/// <param name="serializationName">Serialization name.</param>
		/// <param name="type">List type.</param>
		public void AddListTypeName( string serializationName, Type type )
		{
			ValidationUtils.ObjectHasContent( type, "type" );

			SerializationUtils.ValidateSignableType( type );

			TextSerializationNode node = TextSerializationNode.GetListNode( serializationName, type.FullName );

			Add( node );
		}

		/// <summary>
		/// Adds a new scalar node to the writer.
		/// </summary>
		/// <param name="serializationName">Serialization name.</param>
		/// <param name="value">Value. This must be a scalar.</param>
		public void AddScalar( string serializationName, object value )
		{
			ValidationUtils.ObjectHasContent( value, "value" );

			string valueTypeName = value.GetType().FullName;
			string valueBuffer = ScalarEncoder.ValueToText( value );

			TextSerializationNode node = TextSerializationNode.GetScalarNode( serializationName, valueTypeName, valueBuffer );

			Add( node );
		}

		/// <summary>
		/// Writes the output of the serializer.
		/// </summary>
		public void WriteOutput()
		{
			foreach( TextSerializationNode writerItem in this )
				WriteItemOutput( writerItem.SerializationName, writerItem.Value, writerItem.ValueTypeName );
		}

		#endregion Public methods
	}
}
