////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections;
using System.Reflection;
using System.Xml;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;

namespace Vela.XmlSerialization
{
	/// <summary>
	/// A deserializer for documents. It can deserialize documents which were serialized with <see cref="XmlSerializer"/>.
	/// </summary>
	public class XmlDeserializer : IDeserializer
	{
		#region Public constants

		/// <summary>
		/// The size (in bytes) of the data which is analyzed to see what kind of serializer was used to serialize the data.
		/// </summary>
		private const int AnalyzeSize = 300;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected XmlDeserializer()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="textEncoding">Text encoding.</param>
		public XmlDeserializer( TextEncoding textEncoding )
		{
			TextEncoding = textEncoding;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private int _CandidatePropCount;
		/// <summary>
		/// The number of properties which were candidates for deserialization, and had to be verified whether they are to be
		/// deserialized or not. This does not include the root node. READ-ONLY.
		/// </summary>
		public int CandidatePropCount
		{
			get { return _CandidatePropCount; }
			protected set { _CandidatePropCount = value; }
		}

		private int _ProcessedPropCount;
		/// <summary>
		/// The number of properties which were deserialized. This does not include the root node. READ-ONLY.
		/// </summary>
		public int ProcessedPropCount
		{
			get { return _ProcessedPropCount; }
			protected set { _ProcessedPropCount = value; }
		}

		#endregion Public non-serializable properties

		#region Public serializable properties

		private TextEncoding _TextEncoding;
		/// <summary>
		/// Text encoding. READ-ONLY.
		/// </summary>
		[SerializeAttribute]
		public TextEncoding TextEncoding
		{
			get { return _TextEncoding; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "TextEncoding" );

				_TextEncoding = value;
			}
		}

		#endregion Public serializable properties

		#region Helper methods

		/// <summary>
		/// Gets the serializer type name attribute (<see cref="XmlSerializationConsts.SerializerTypeNameAttributeName"/>) of the
		/// specified XML attribute list.
		/// </summary>
		/// <param name="attributes">XML attribute list to look into.</param>
		/// <returns>Returns the serializer type name attribute of the specified XML attribute list.</returns>
		private static string GetSerializerTypeNameAttr( XmlAttributeCollection attributes )
		{
			// Get the serializer type name attribute.
			XmlAttribute typeAttr = attributes[ XmlSerializationConsts.SerializerTypeNameAttributeName ];

			if( (typeAttr != null) && TextUtils.HasContent( typeAttr.Value ) )
				return typeAttr.Value;

			return null;
		}

		/// <summary>
		/// Gets the content type name attribute (<see cref="XmlSerializationConsts.ContentTypeNameAttributeName"/>) of the
		/// specified XML attribute list.
		/// </summary>
		/// <param name="attributes">XML attribute list to look into.</param>
		/// <returns>Returns the content type name attribute of the specified XML attribute list.</returns>
		private static string GetContentTypeNameAttr( XmlAttributeCollection attributes )
		{
			// Get the content type name attribute.
			XmlAttribute typeAttr = attributes[ XmlSerializationConsts.ContentTypeNameAttributeName ];

			if( (typeAttr != null) && TextUtils.HasContent( typeAttr.Value ) )
				return typeAttr.Value;

			return null;
		}

		/// <summary>
		/// Finds a serialization node, sibling of the specified node, whose name matches the specified serialization name.
		/// </summary>
		/// <param name="node">Node from which to start searching.</param>
		/// <param name="serializationName">The serialization name for which to find the serialization node.</param>
		/// <param name="declaredType">The declared type of the property for which to find a serialization node.</param>
		/// <param name="contentType">Is set with the content type of the found serialization node. By default, this is
		/// "declaredType".</param>
		/// <param name="nextNode">Is set with the serialization node next sibling of the matching node. This is the input node
		/// if no matching node was found.</param>
		/// <returns>Returns the found serialization node, or null if none if found.</returns>
		private static XmlNode FindSerializationNode( XmlNode node, string serializationName, Type declaredType
			, out Type contentType, out XmlNode nextNode )
		{
			ValidationUtils.ObjectHasContent( node, "node" );
			ValidationUtils.TextHasContent( serializationName, "serializationName" );
			ValidationUtils.ObjectHasContent( declaredType, "declaredType" );

			contentType = declaredType; // As default, use the declared type.
			nextNode = node;

			// Start searching from the specified node until the last sibling. In normal circumstances, the specified node
			// itself should match (because the deserialization is performed in the same order as the serialization).
			XmlNode currNode = node;
			while( currNode != null )
			{
				if( TextUtils.Equals( currNode.LocalName, serializationName ) )
				{
					// Node found.

					// Get the content type document attribute.
					string typeAttr = GetContentTypeNameAttr( currNode.Attributes );

					if( TextUtils.HasContent( typeAttr ) )
					{
						// We must find a type.

						if( TextUtils.Equals( typeAttr, declaredType.FullName ) )
						{
							// The type attribute has the same name as the declared type, so we use the declared type.

							contentType = declaredType;
						}
						else
						{
							// The type specified in the document must be derived from the declared type and must be found in
							// the list of global portable types.
							contentType = PortableTypes.FindGlobal( typeAttr );
							if( contentType == null )
							{
								throw new ValidationException( string.Format( "The type attribute \"{0}\" of the serialization"
									+ " node \"{1}\" of the property \"{2}\" was not found in the list of global serializable"
									+ " types.", typeAttr, currNode.LocalName, serializationName ) );
							}

							// Check if the found type is derived from the declared type. Don't check scalar types because
							// scalar type conversion may be implemented by the value deserialization method ("ReadScalar").
							if( !ScalarEncoder.IsScalar( contentType ) && !declaredType.IsAssignableFrom( contentType ) )
							{
								throw new ValidationException( string.Format( "The type attribute \"{0}\" of the property"
										+ " \"{1}\" of the serialization node \"{2}\" is not derived from the declared"
										+ " type \"{3}\" of the property."
									, typeAttr, serializationName, currNode.LocalName, declaredType.FullName ) );
							}
						}
					}
					else
					{
						throw new ValidationException( string.Format( "The type attribute of the property \"{0}\" of the item's"
								+ " serialization node \"{1}\" is missing even though it's content is not missing."
							, serializationName, currNode.LocalName ) );
					}

					// Set the next node.
					nextNode = currNode.NextSibling;

					// Return the found node.
					return currNode;
				}
				else currNode = currNode.NextSibling; // Node not found.
			}

			// Start searching from the first sibling until the specified node.
			currNode = node.ParentNode.FirstChild;
			while( (currNode != null) && (currNode != node) )
			{
				if( TextUtils.Equals( currNode.LocalName, serializationName ) )
				{
					// Node found.

					// Get the content type document attribute.
					string typeAttr = GetContentTypeNameAttr( currNode.Attributes );

					if( TextUtils.HasContent( typeAttr ) ) contentType = PortableTypes.FindGlobal( typeAttr );

					// Set the next node.
					nextNode = currNode.NextSibling;

					// Return the found node.
					return currNode;
				}
				else currNode = currNode.NextSibling; // Node not found.
			}

			// No matching serialization name was found for the current property. Don't move to the next sibling node of the
			// input node because the same node must be checked again for the next property.
			return null;
		}

		#endregion Helper methods

		#region Private static methods

		/// <summary>
		/// Checks if the current deserializer may be used to deserialize the data with the specified prefix.
		/// </summary>
		/// <param name="prefix">Prefix of the serialized data. This may be null or empty.</param>
		/// <returns>Returns true if the current deserializer may be used to deserialize the data with the specified prefix.
		/// </returns>
		private static bool CheckPrefixForParser( string prefix )
		{
			if( TextUtils.NoContent( prefix ) ) return false;

			prefix = prefix.TrimStart();

			// Check if the content was serialized with the associated serializer.
			if( !prefix.StartsWith( XmlSerializerWriter.PrologTemplate, StringComparison.OrdinalIgnoreCase ) ) return false;

			// The "XmlSerializationConsts.SerializerTypeNameAttributeName" doesn't contain blank spaces around the "=".
			string serializerTypeName_XmlSerializer = string.Format( " {0}=\"{1}\""
				, XmlSerializationConsts.SerializerTypeNameAttributeName, typeof( XmlSerializer ).FullName );

			return TextUtils.Contains( prefix, serializerTypeName_XmlSerializer );
		}

		/// <summary>
		/// Reads a scalar object from the specified serialization node.
		/// </summary>
		/// <param name="objectNode">The serialization node to read the scalar from.</param>
		/// <param name="contentType">The type of the scalar object to read.</param>
		/// <param name="declaredType">Declared type of the property to which the converted value is to be assigned.</param>
		/// <param name="enumFallbackValue">The fallback value for enumerations. If the object to deserialize is an enumeration,
		/// if the deserialization of the enumeration fails, this value is assigned instead of an exception being thrown. This
		/// may be null.</param>
		/// <returns>Returns the read scalar object.</returns>
		private static object ReadScalar( XmlNode objectNode, Type contentType, Type declaredType, object enumFallbackValue )
		{
			if( !ScalarEncoder.IsScalar( contentType ) )
			{
				throw new ValidationException( string.Format( "The object to deserialize (\"{0}\") is not a scalar, but a"
					+ " \"{1}\".", objectNode.Name, contentType.FullName ) );
			}

			if( objectNode.HasChildNodes )
			{
				bool isScalarNode = (objectNode.ChildNodes.Count == 1)
					&& (objectNode.ChildNodes[ 0 ].NodeType == XmlNodeType.Text);

				if( isScalarNode )
				{
					string xmlText = objectNode.InnerText; // This unescapes the content.

					return ScalarEncoder.ValueFromText( xmlText, contentType, declaredType, enumFallbackValue );
				}
				else
				{
					throw new ValidationException( string.Format( "Type \"{0}\" being deserialized is not a scalar."
						, contentType.FullName ) );
				}
			}
			else
			{
				return ScalarEncoder.ValueFromText( "", contentType, declaredType, enumFallbackValue );
			}
		}

		#endregion Private static methods

		#region Private methods

		/// <summary>
		/// Processes a list (which may actually be a class derived from <see cref="IEnumerable"/>)with its items.
		/// </summary>
		/// <param name="objectNode">The serialization node to read the list from.</param>
		/// <param name="listType">The type of the list object to read.</param>
		/// <param name="enumFallbackValue">The fallback value for enumerations. If the list item to deserialize is an
		/// enumeration, if the deserialization of the enumeration fails, this value is assigned instead of an exception being
		/// thrown. This may be null.</param>
		/// <returns>Returns the read list object.</returns>
		private object ProcessList( XmlNode objectNode, Type listType, object enumFallbackValue )
		{
			ValidationUtils.ObjectHasContent( objectNode, "objectNode" );
			ValidationUtils.ObjectHasContent( listType, "listType" );

			// The deserialized list object.
			object listObject;

			// The serialization nodes for the list items.
			XmlNodeList nodes = objectNode.HasChildNodes ? objectNode.ChildNodes : null;
			int itemsCount = objectNode.HasChildNodes ? objectNode.ChildNodes.Count : 0;

			// Prepare to call the method named "DeserializingMethodName" before the list is deserialized, so that the method is
			// called before it is called for its items.
			MethodInfo eventMethodInfo = listType.GetMethod( SerializationConsts.DeserializingMethodName
				, BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance, null, Type.EmptyTypes, null );

			Array array = null; // Used if the list is an array.
			MethodInfo addMethodInfo = null; // Used if the list isn't an array.
			Type itemDeclaredType = null;

			if( listType.IsArray )
			{
				// The type is an array.

				itemDeclaredType = listType.GetElementType();

				if( itemDeclaredType == null )
				{
					throw new ValidationException( string.Format( "The list of type \"{0}\" doesn't have a type for items."
						, listType.FullName ) );
				}

				// Create an instance of the specified list type.
				listObject = Array.CreateInstance( itemDeclaredType, itemsCount );

				array = (Array)listObject;

				// Multi-dimensional arrays can't be deserialized.
				if( array.Rank != 1 )
				{
					throw new ValidationException( string.Format( "Can't deserialize list of type \"{0}\" because it is"
						+ " multi-dimensional ({1}).", listType.FullName, array.Rank ) );
				}

				// Call the method "DeserializingMethodName" before the list is deserialized, so that the method is called
				// before it is called for its items.
				if( eventMethodInfo != null ) eventMethodInfo.Invoke( listObject, null );
			}
			else
			{
				// The type is a list.

				// Create an instance of the specified list type.
				listObject = Activator.CreateInstance( listType, true );

				// Call the method "DeserializingMethodName" before the list is deserialized, so that the method is called
				// before it is called for its items.
				if( eventMethodInfo != null ) eventMethodInfo.Invoke( listObject, null );

				// Get the type's instance method named "ListAddMethodName" with a single "object" parameter.
				addMethodInfo = listType.GetMethod( SerializationConsts.ListAddMethodName, BindingFlags.Public
					| BindingFlags.NonPublic | BindingFlags.Instance, null, new Type[] { typeof( object ) }, null );

				if( typeof( IItemTypeAwareList ).IsAssignableFrom( listType ) )
				{
					IItemTypeAwareList itaList = (IItemTypeAwareList)listObject;
					itemDeclaredType = itaList.ItemDeclaredType;

					if( itemDeclaredType == null )
					{
						throw new ValidationException( string.Format( "The list of type \"{0}\" doesn't have a type for items."
							, listType.FullName ) );
					}
				}
			}

			if( (itemsCount > 0) && (nodes != null) )
			{
				int i = 0;
				foreach( XmlNode itemNode in nodes )
				{
					// Get the item content type document attribute.
					string itemTypeAttr = GetContentTypeNameAttr( itemNode.Attributes );
					if( itemTypeAttr == null )
					{
						throw new ValidationException( string.Format( "The type attribute of the list of type \"{0}\" of the"
							+ " item's serialization node \"{1}\" is missing.", listType.FullName, itemNode.LocalName ) );
					}

					#region Check global type

					// Item-type-aware lists know the declared type of their items, so the types of the items from an
					// item-type-aware list don't need to be globally registered (unless those types are derived from the
					// declared type of the items).

					Type itemType = null;

					bool checkGlobalType = true;

					if( (itemDeclaredType != null) && TextUtils.HasContent( itemTypeAttr ) )
					{
						// Check if the item's type is the declared type of the list's items.
						checkGlobalType = !TextUtils.Equals( itemDeclaredType.FullName, itemTypeAttr );

						if( !checkGlobalType ) itemType = itemDeclaredType;
					}

					if( checkGlobalType )
					{
						// The types of the items from a list which is not item-type-aware need to be globally registered.

						itemType = TextUtils.HasContent( itemTypeAttr ) ? PortableTypes.FindGlobal( itemTypeAttr ) : null;
					}

					if( itemType == null )
					{
						throw new ValidationException( string.Format( "The type attribute \"{0}\" of the list of type \"{1}\""
								+ " of the item's serialization node \"{2}\" is unknown."
							, itemTypeAttr, listType.FullName, itemNode.LocalName ) );
					}

					#endregion Check global type

					// The deserialized list item.
					object listItem;

					if( ScalarEncoder.IsScalar( itemType ) )
					{
						listItem = ReadScalar( itemNode, itemType, itemDeclaredType, enumFallbackValue );
					}
					else if( SerializationUtils.IsList( itemType ) )
					{
						throw new ValidationException( string.Format( "Can't deserialize nested list of type \"{0}\"."
							, itemType.FullName ) );
					}
					else if( SerializationUtils.IsClass( itemType ) )
					{
						listItem = ProcessClass( itemNode, itemType );
					}
					else
					{
						throw new ValidationException( string.Format( "List items of type \"{0}\" can't be deserialized."
							, itemType.FullName ) );
					}

					// Deserialize only non-null items.
					if( listType.IsArray )
					{
						if( listItem != null )
						{
							array.SetValue( listItem, i );
							i++;
						}
					}
					else
					{
						if( listItem != null ) addMethodInfo.Invoke( listObject, new object[] { listItem } );
					}
				}
			}

			// Call the method "DeserializedMethodName" after the class is deserialized, so that the method is called after it
			// was called for its items.
			eventMethodInfo = listType.GetMethod( SerializationConsts.DeserializedMethodName, BindingFlags.Public
				| BindingFlags.NonPublic | BindingFlags.Instance, null, Type.EmptyTypes, null );
			if( eventMethodInfo != null ) eventMethodInfo.Invoke( listObject, null );

			return listObject;
		}

		/// <summary>
		/// Processes all serializable members of the specified class type, and sets them with objects read from the specified
		/// serialization node, using serialization name matching.
		/// </summary>
		/// <param name="objectNode">The serialization node to read the class object from.</param>
		/// <param name="classType">The type of the class object to read.</param>
		/// <returns>Returns the read class object.</returns>
		private IDocument ProcessClass( XmlNode objectNode, Type classType )
		{
			if( objectNode == null )
			{
				throw new ValidationException( string.Format( "The serialization node associated to class type \"{0}\" is"
					+ " missing.", classType.FullName ) );
			}

			ConstructorInfo classConstructor = SerializationUtils.CheckSerializableClass( classType );

			// Create an instance of the specified class type. "Activator.CreateInstance" is not used for performance reasons.
			IDocument serializableData = (IDocument)classConstructor.Invoke( null );

			// The current serialization node. This is used to optimize the searching for the serialization node associated to
			// the property which is being deserialized, because in the ideal case they should have the same order.
			XmlNode currNode = objectNode.FirstChild;

			// Call the method "DeserializingMethodName" before the class is deserialized, so that the method is called before
			// it is called for child class properties.
			MethodInfo eventMethodInfo = classType.GetMethod( SerializationConsts.DeserializingMethodName
				, BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance, null, Type.EmptyTypes, null );
			if( eventMethodInfo != null ) eventMethodInfo.Invoke( serializableData, null );

			// Process the properties of the class.

			PropertyInfo[] propInfos = classType.GetProperties( BindingFlags.Public | BindingFlags.NonPublic
				| BindingFlags.Instance );

			foreach( PropertyInfo propInfo in propInfos )
			{
				// Serialize only members which can be read and written.
				if( propInfo.CanRead && propInfo.CanWrite )
				{
					string serializationName, listItemName;
					bool mustBeDeserialized;
					object enumFallbackValue;

					if( SerializationUtils.CheckDeserializeAttribute( propInfo, out mustBeDeserialized, out serializationName
						, out listItemName, out enumFallbackValue ) )
					{
						// If the value of the serialization node is null, it's name would be mapped to the declared
						// serialization name and type of the current property.

						Type contentType; // The content type.
						XmlNode nextNode;

						XmlNode thisNode = null;

						if( currNode != null )
						{
							thisNode = FindSerializationNode( currNode, serializationName, propInfo.PropertyType
								, out contentType, out nextNode );

							if( thisNode != null )
							{
								// A node was found, so process the current property.
								ProcessObject( thisNode, propInfo, contentType, serializableData, enumFallbackValue );
							}

							// Go to the next node (because the matching node was processed). If there is no next node, start
							// from the beginning.
							currNode = (nextNode != null) ? nextNode : objectNode.FirstChild;
						}

						if( thisNode == null )
						{
							if( mustBeDeserialized )
							{
								throw new ValidationException( string.Format( "The property \"{0}\" from class \"{1}\" must"
										+ " have a value in the document which is being deserialized, but it has none."
									, propInfo.Name, classType.FullName ) );
							}
							else
							{
								// Don't deserialize null objects.
							}
						}

						ProcessedPropCount++;
					}

					CandidatePropCount++;
				}
			}

			// Call the method "DeserializedMethodName" after the class is deserialized, so that the method is called after it
			// was called for child class properties.
			eventMethodInfo = classType.GetMethod( SerializationConsts.DeserializedMethodName, BindingFlags.Public
				| BindingFlags.NonPublic | BindingFlags.Instance, null, Type.EmptyTypes, null );
			if( eventMethodInfo != null ) eventMethodInfo.Invoke( serializableData, null );

			return serializableData;
		}

		/// <summary>
		/// Processes the specified object, according to the type of the object's content.
		/// </summary>
		/// <param name="objectNode">The serialization node to read the object from.</param>
		/// <param name="propInfo">Information about the property which is to be set with the processed object.</param>
		/// <param name="contentType">The type of the object to read.</param>
		/// <param name="serializableData">The parent class object for the property which is to be set with the read
		/// object.</param>
		/// <param name="enumFallbackValue">The fallback value for enumerations. If the object to deserialize is an enumeration,
		/// if the deserialization of the enumeration fails, this value is assigned instead of an exception being thrown. This
		/// may be null.</param>
		private void ProcessObject( XmlNode objectNode, PropertyInfo propInfo, Type contentType, object serializableData
			, object enumFallbackValue )
		{
			ValidationUtils.ObjectHasContent( objectNode, "objectNode" );
			ValidationUtils.ObjectHasContent( propInfo, "propInfo" );
			ValidationUtils.ObjectHasContent( contentType, "contentType" );
			ValidationUtils.ObjectHasContent( serializableData, "serializableData" );

			// The deserialized document.
			object val;

			if( ScalarEncoder.IsScalar( contentType ) )
			{
				// Deserialize scalar.
				val = ReadScalar( objectNode, contentType, propInfo.PropertyType, enumFallbackValue );

				// Set the property content with the deserialized document.
				propInfo.SetValue( serializableData, val, null );
			}
			else
			{
				if( SerializationUtils.IsList( contentType ) )
				{
					// Process the items of the list.
					val = ProcessList( objectNode, contentType, enumFallbackValue );
				}
				else if( SerializationUtils.IsClass( contentType ) )
				{
					// Process class.
					val = ProcessClass( objectNode, contentType );
				}
				else
				{
					throw new ValidationException( string.Format( "Objects of type \"{0}\" can't be deserialized."
						, contentType.FullName ) );
				}

				// Set the property content with the deserialized document.
				propInfo.SetValue( serializableData, val, null );
			}
		}

		/// <summary>
		/// Loads the class object from the specified text.
		/// </summary>
		/// <param name="data">Text to deserialize. This may be null or empty.</param>
		/// <param name="documentType">The root type of the document to load; the document to load may either have this type or
		/// a type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <returns>Returns the deserialized document, or null if the input data is null or empty.</returns>
		private IDocument DeserializeCore( string data, Type documentType )
		{
			if( TextUtils.NoContent( data ) ) return null;

			#region Load the XML document

			XmlDocument xmlDoc = new XmlDocument();

			xmlDoc.PreserveWhitespace = false;
			xmlDoc.Schemas = null;
			xmlDoc.XmlResolver = null;

			xmlDoc.LoadXml( data );

			#endregion Load the XML document

			// Get the document root name.
			string documentRootName = xmlDoc.DocumentElement.LocalName;

			// Get the serializer type name attribute (this must be "XmlSerializer").
			string serializerTypeNameAttr = GetSerializerTypeNameAttr( xmlDoc.DocumentElement.Attributes );
			if( TextUtils.NoContent( serializerTypeNameAttr )
				|| !TextUtils.Equals( serializerTypeNameAttr, typeof( XmlSerializer ).FullName ) )
			{
				throw new ValidationException( string.Format( "The extracted serializer type name is \"{0}\" instead of the"
					+ " expected \"{1}\".", serializerTypeNameAttr, typeof( XmlSerializer ).FullName ) );
			}

			// Get the content type document attribute.
			string nodeContentTypeNameAttr = GetContentTypeNameAttr( xmlDoc.DocumentElement.Attributes );
			if( TextUtils.NoContent( nodeContentTypeNameAttr ) )
			{
				throw new ValidationException( string.Format( "The content type of the root node"
					+ " of the document to deserialize (\"{0}\") is missing.", documentRootName ) );
			}

			// Get the document's root type.
			Type rootType = documentType;
			if( (documentType == null) || !TextUtils.Equals( documentType.FullName, nodeContentTypeNameAttr ) )
			{
				rootType = PortableTypes.FindGlobal( nodeContentTypeNameAttr );
				if( rootType == null )
				{
					throw new ValidationException( string.Format( "The type \"{0}\" of the root node of the document to"
							+ " deserialize was not found in the list of global portable types."
						, nodeContentTypeNameAttr ) );
				}

				// If the found global type must be derived from the specified type, check that.
				if( (documentType != null) && !documentType.IsAssignableFrom( rootType ) )
				{
					throw new ValidationException( string.Format( "The type \"{0}\" of the root node of the document to"
							+ " deserialize is not derived from the requested type \"{1}\"."
						, rootType.FullName, documentType.FullName ) );
				}
			}

			// Deserialize the specified class type.
			CandidatePropCount = 0;
			ProcessedPropCount = 0;

			IDocument serializableData;

			if( SerializationUtils.IsList( rootType ) )
			{
				serializableData = ProcessList( xmlDoc.DocumentElement, rootType, null ) as IDocument;
			}
			else serializableData = ProcessClass( xmlDoc.DocumentElement, rootType );

			return serializableData;
		}

		#endregion Private methods

		#region Public static methods

		/// <summary>
		/// Checks if the current deserializer may be used to deserialize the specified data.
		/// </summary>
		/// <param name="buffer">Serialized data. This may be a binary encoded text.</param>
		/// <returns>Returns an instance of itself if it may parse the specified data, or null if it may not parse it.</returns>
		/// <remarks>
		/// <para>The text encoding of the deserializer is extracted from the specified data.</para>
		/// </remarks>
		public static XmlDeserializer GetParsingAlgorithm( byte[] buffer )
		{
			ValidationUtils.ArrayHasContent( buffer, "buffer" );

			int minSize = Math.Min( AnalyzeSize, buffer.Length );
			byte[] analyzeData = new byte[ minSize ];
			Buffer.BlockCopy( buffer, 0, analyzeData, 0, minSize );

			TextEncoding textEncoding;
			string analyzeText = TextEncoding.TryArrayToText( analyzeData, out textEncoding );

			if( CheckPrefixForParser( analyzeText ) ) return new XmlDeserializer( textEncoding );

			return null;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Returns a serializer with the configuration from the current serialization factory.
		/// </summary>
		/// <returns>Returns the serializer with the configuration from the current serialization factory.</returns>
		public ISerializer GetSerializer()
		{
			return new XmlSerializer( TextEncoding );
		}

		/// <summary>
		/// Returns a deserializer with the configuration from the current serialization factory.
		/// </summary>
		/// <returns>Returns the deserializer with the configuration from the current serialization factory.</returns>
		public IDeserializer GetDeserializer()
		{
			return new XmlDeserializer( TextEncoding );
		}

		/// <summary>
		/// Returns a hash serializer with the configuration from the current serialization factory.
		/// </summary>
		/// <param name="serializeValueTypeName">If this is true then the type names of the serialized values are stored in the
		/// serialization output. If this is true, the compatibility level of different versions of a serialized document is
		/// limited (see the "Serialization versioning" section from the documentation of the <see cref="IHashSerializer"/>
		/// interface). This must be true for a hash serializer which will be used to test the equality of documents.</param>
		/// <returns>Returns the hash serializer with the configuration from the current serialization factory.</returns>
		public IHashSerializer GetHashSerializer( bool serializeValueTypeName )
		{
			return new TextHashSerializer( TextEncoding, serializeValueTypeName );
		}

		/// <summary>
		/// Creates a deep clone of the specified document by serializing and then deserializing it.
		/// </summary>
		/// <param name="source">Document to clone. This may be null.</param>
		/// <returns>Returns the cloned document. This is null if "source" is null.</returns>
		public IDocument CloneDocument( IDocument source )
		{
			return GetSerializer().CloneDocument( source );
		}

		/// <summary>
		/// Checks if the specified documents are equal by comparing the output of their serializations.
		/// </summary>
		/// <param name="document1">Document to compare. This may be null.</param>
		/// <param name="document2">Document to compare. This may be null.</param>
		/// <returns>Returns the true if the documents are equal.</returns>
		/// <remarks>
		/// <para>The documents to compare may have different types.</para>
		/// </remarks>
		public bool EqualDocuments( IDocument document1, IDocument document2 )
		{
			return GetHashSerializer( true ).EqualDocuments( document1, document2 );
		}

		/// <summary>
		/// Loads the class object from the specified buffer.
		/// </summary>
		/// <param name="buffer">Buffer to deserialize. This may be null or empty.</param>
		/// <param name="documentType">The root type of the document to load; the document to load may either have this type or
		/// a type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <returns>Returns the deserialized document, or null if the input data is null or empty.</returns>
		public IDocument LoadFromBuffer( byte[] buffer, Type documentType )
		{
			if( BufferUtils.NoContent( buffer ) ) return null;

			// Read the serialized textual data from the buffer.
			string textualData = TextEncoding.ArrayToText( buffer );

			// Deserialize.
			return DeserializeCore( textualData, documentType );
		}

		/// <summary>
		/// Loads the class object from the specified file.
		/// </summary>
		/// <param name="fileName">The name of the file to deserialize. The file may be missing and its content may be null or
		/// empty.</param>
		/// <param name="documentType">The root type of the document to load; the document to load may either have this type or
		/// a type derived from this (if it's found in the global portable types from
		/// <see cref="PortableTypes.Instance"/>). If this is null, the type specified in the document's root node is
		/// searched in the global portable types. If the name of the document's root node is not found, an exception is
		/// thrown.</param>
		/// <returns>Returns the deserialized document, or null if the file is missing or its content is null or empty.</returns>
		/// <remarks>
		/// <para>This method implements a recovery mechanism following a crash during a call of the
		/// <see cref="ISerializer.SaveToFile"/> method.</para>
		/// </remarks>
		public IDocument LoadFromFile( string fileName, Type documentType )
		{
			ValidationUtils.TextHasContent( fileName, "fileName" );

			byte[] buffer;

			try
			{
				// Read the serialized data from the file.
				buffer = FileUtils.Read( fileName, false );

				// Deserialize.
				return LoadFromBuffer( buffer, documentType );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsAbort( ex ) ) throw ExceptionUtils.PreserveStackTrace( ex );

				throw new ValidationException( string.Format( "Failed to load the document from file \"{0}\".", fileName ), ex );
			}
		}

		#endregion Public methods
	}
}
