////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Text;
using Vela.Kernel;

namespace Vela.XmlSerialization
{
	/// <summary>
	/// An XML writer.
	/// </summary>
	public class XmlSerializerWriter : IXmlSerializerWriter
	{
		#region Public constants

		/// <summary>
		/// The text which is written, before the XML node, for each level of indentation.
		/// </summary>
		public const char LevelIndentChar = '\t';

		/// <summary>
		/// The text which is written at the end of each XML line.
		/// </summary>
		public const string LineSeparatorText = "\n";

		/// <summary>
		/// Template of the XML prolog. Placeholders: 0 - standard XML text encoding.
		/// </summary>
		public const string PrologTemplate = "<?xml version=\"1.0\"?>";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="builder">The string builder where to write the XML output.</param>
		public XmlSerializerWriter( StringBuilder builder )
		{
			ValidationUtils.ObjectHasContent( builder, "builder" );

			OutputBuilder = builder;

			StartDocument();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The string builder where to write the serializer's output.
		/// </summary>
		private StringBuilder OutputBuilder;

		private int _IndentLevel;
		/// <summary>
		/// The current level of indentation of the XML document. This must be at least 0.
		/// </summary>
		private int IndentLevel
		{
			get { return _IndentLevel; }
			set
			{
				ValidationUtils.Integer( value, 0, int.MaxValue, "IndentLevel" );

				_IndentLevel = value;
			}
		}

		#endregion Private fields

		#region Private methods

		/// <summary>
		/// Gets the text which begins the current line from the XML document.
		/// </summary>
		/// <returns>Returns the text which begins the current line from the XML document.</returns>
		private string GetLineBeginning()
		{
			return LineSeparatorText + new string( LevelIndentChar, IndentLevel );
		}

		/// <summary>
		/// Writes the start of the current XML document to the output.
		/// </summary>
		private void StartDocument()
		{
			// Don't check if "OutputBuilder" is empty in order to allow multiple serializations into the same string builder.

			IndentLevel = 0;

			OutputBuilder.AppendFormat( PrologTemplate );
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Writes the end of the current XML document to the output.
		/// </summary>
		public void EndDocument()
		{
		}

		/// <summary>
		/// Writes the start of the named XML node to the output. The XML node can't have attributes.
		/// </summary>
		/// <param name="nodeName">The name of the XML node.</param>
		public void StartNodeNoAttr( string nodeName )
		{
			OutputBuilder.Append( GetLineBeginning() );
			OutputBuilder.Append( "<" );
			OutputBuilder.Append( nodeName );
			OutputBuilder.Append( ">" );

			IndentLevel++;
		}

		/// <summary>
		/// Writes the start of the named XML node to the output. The XML node has attributes. The start of the XML node must be
		/// closed using the <see cref="CloseStartNodeWithAttr"/> method.
		/// </summary>
		/// <param name="nodeName">The name of the XML node.</param>
		public void StartNodeWithAttr( string nodeName )
		{
			OutputBuilder.Append( GetLineBeginning() );
			OutputBuilder.Append( "<" );
			OutputBuilder.Append( nodeName );

			IndentLevel++;
		}

		/// <summary>
		/// Writes the end of the start of an XML node which has attributes, to the output.
		/// </summary>
		public void CloseStartNodeWithAttr()
		{
			OutputBuilder.Append( ">" );
		}

		/// <summary>
		/// Writes the end of the named XML node to the output.
		/// </summary>
		/// <param name="nodeName">The name of the XML node.</param>
		public void EndNode( string nodeName )
		{
			IndentLevel--;

			OutputBuilder.Append( GetLineBeginning() );
			OutputBuilder.Append( "</" );
			OutputBuilder.Append( nodeName );
			OutputBuilder.Append( ">" );
		}

		/// <summary>
		/// Writes an XML node with the specified name and value, to the output.
		/// </summary>
		/// <param name="nodeName">The name of the XML node.</param>
		/// <param name="contentTypeName">The name of the type of the node's value of the serialized property.</param>
		/// <param name="value">Buffer to write in the node's value.</param>
		public void WriteScalar( string nodeName, string contentTypeName, object value )
		{
			string valueText = ScalarEncoder.ValueToText( value );

			OutputBuilder.Append( GetLineBeginning() );
			OutputBuilder.Append( "<" );
			OutputBuilder.Append( nodeName );
			OutputBuilder.Append( " " );
			OutputBuilder.Append( XmlSerializationConsts.ContentTypeNameAttributeName );
			OutputBuilder.Append( "=\"" );
			OutputBuilder.Append( contentTypeName );
			OutputBuilder.Append( "\"" );
			OutputBuilder.Append( ">" );
			OutputBuilder.Append( TextUtils.EscapeXml( valueText ) );
			OutputBuilder.Append( "</" );
			OutputBuilder.Append( nodeName );
			OutputBuilder.Append( ">" );
		}

		/// <summary>
		/// Writes the specified text in the currently opened XML node, to the output.
		/// </summary>
		/// <param name="attrName">The name of the XML attribute to write.</param>
		/// <param name="value">Buffer to write in the attribute's value.</param>
		public void WriteAttribute( string attrName, string value )
		{
			OutputBuilder.Append( " " );
			OutputBuilder.Append( attrName );
			OutputBuilder.Append( "=\"" );
			OutputBuilder.Append( TextUtils.EscapeXml( value ) );
			OutputBuilder.Append( "\"" );
		}

		#endregion Public methods
	}
}
