﻿using NetFwTypeLib;
using System;
using Vela.Kernel;

namespace Vela.ConfigureFirewall
{
	/// <summary>
	/// Application entry point.
	/// </summary>
	public class Program
	{
		#region Private constants

		/// <summary>
		/// Template for a firewall rule name. Placeholders: 0 - rule name prefix, 1 - rule-specific name.
		/// </summary>
		private const string RuleNameTemplate = "{0} - {1}";

		#endregion Private constants

		#region Private static methods

		/// <summary>
		/// Changes the state of the firewall, per profile.
		/// </summary>
		/// <param name="firewall">Firewall.</param>
		private static void ChangeState( WindowsFirewall firewall )
		{
			ValidationUtils.ObjectHasContent( firewall, "firewall" );

			Console.WriteLine( "The Windows Firewall:" );

			if( firewall.IsEnabledForPrivateProfile() )
			{
				if( firewall.PrivateInboundConnectionsBlocked() )
				{
					Console.Write( "* Blocks private inbound connections... " );

					firewall.BlockPrivateInboundConnections( false );
					Console.WriteLine( "reconfigured to use rules." );
				}
			}
			else
			{
				Console.WriteLine( "* Is not enabled for the private profile." );
			}

			if( firewall.IsEnabledForDomainProfile() )
			{
				if( firewall.DomainInboundConnectionsBlocked() )
				{
					Console.Write( "* Blocks domain inbound connections... " );

					firewall.BlockDomainInboundConnections( false );
					Console.WriteLine( "reconfigured to use rules." );
				}
			}
			else
			{
				Console.WriteLine( "* Is not enabled for the domain profile." );
			}

			if( firewall.IsEnabledForPublicProfile() )
			{
				if( firewall.PublicInboundConnectionsBlocked() )
				{
					Console.Write( "* Blocks public inbound connections... " );

					firewall.BlockPublicInboundConnections( false );
					Console.WriteLine( "reconfigured to use rules." );
				}
			}
			else
			{
				Console.WriteLine( "* Is not enabled for the public profile." );
			}

			Console.WriteLine();
		}

		/// <summary>
		/// Adds the firewall rules.
		/// </summary>
		/// <param name="ruleNamePrefix">The prefix for a rule name.</param>
		/// <param name="externalApplicationFileName">The file name for the external application for which to configure the
		/// firewall.</param>
		private static void AddFirewallRules( string ruleNamePrefix, string externalApplicationFileName )
		{
			ValidationUtils.TextHasContent( ruleNamePrefix, "ruleNamePrefix" );
			ValidationUtils.TextHasContent( externalApplicationFileName, "externalApplicationFileName" );

			try
			{
				WindowsFirewall firewall = new WindowsFirewall();

				ChangeState( firewall );

				Console.WriteLine( string.Format( "Rule name prefix: {0}", ruleNamePrefix ) );
				Console.WriteLine( string.Format( "Target application: {0}", externalApplicationFileName ) );
				Console.WriteLine();

				// Rules for the external application (specifically, for Vela).
				string ruleName = string.Format( RuleNameTemplate, ruleNamePrefix, "Main - Internal" );
				firewall.AddRule( ruleName, true, false, externalApplicationFileName, true
					, NET_FW_RULE_DIRECTION_.NET_FW_RULE_DIR_IN, NET_FW_IP_PROTOCOL_.NET_FW_IP_PROTOCOL_TCP );
				firewall.AddRule( ruleName, true, false, externalApplicationFileName, true
					, NET_FW_RULE_DIRECTION_.NET_FW_RULE_DIR_OUT, NET_FW_IP_PROTOCOL_.NET_FW_IP_PROTOCOL_TCP );

				// Rule for the external application (specifically, for Vela). This is needed because some UDP traffic originates
				// from the application for UPnP. This is not critical.
				ruleName = string.Format( RuleNameTemplate, ruleNamePrefix, "Main - UPnP" );
				firewall.AddRule( ruleName, true, false, externalApplicationFileName, true
					, NET_FW_RULE_DIRECTION_.NET_FW_RULE_DIR_OUT, NET_FW_IP_PROTOCOL_.NET_FW_IP_PROTOCOL_UDP, null, null
					, "239.255.255.250", "1900" );

				// Rules for the PNRP.
				ruleName = string.Format( RuleNameTemplate, ruleNamePrefix, "PNRP" );
				firewall.AddRule( ruleName, true, true, "PNRPsvc", true, NET_FW_RULE_DIRECTION_.NET_FW_RULE_DIR_IN
					, NET_FW_IP_PROTOCOL_.NET_FW_IP_PROTOCOL_UDP, null, "3540", null, null );
				firewall.AddRule( ruleName, true, true, "PNRPsvc", true, NET_FW_RULE_DIRECTION_.NET_FW_RULE_DIR_OUT
					, NET_FW_IP_PROTOCOL_.NET_FW_IP_PROTOCOL_UDP, null, null, null, "3540" );

				// Rule for Windows Update. Allowing access for the "Windows Update" service isn't not enough. Other services
				// (hosted by "svchost") must still get their own explicit firewall rule in order to get outbound access.
				ruleName = string.Format( RuleNameTemplate, ruleNamePrefix, "Windows Update" );
				firewall.AddRule( ruleName, false, false, @"%SystemRoot%\System32\svchost.exe", true
					, NET_FW_RULE_DIRECTION_.NET_FW_RULE_DIR_OUT, NET_FW_IP_PROTOCOL_.NET_FW_IP_PROTOCOL_TCP, null, null, null
					, "80,443" );

				// Rule for Windows Time.
				ruleName = string.Format( RuleNameTemplate, ruleNamePrefix, "Windows Time" );
				firewall.AddRule( ruleName, false, true, "W32Time", true, NET_FW_RULE_DIRECTION_.NET_FW_RULE_DIR_OUT
					, NET_FW_IP_PROTOCOL_.NET_FW_IP_PROTOCOL_UDP, null, null, null, "123" );
			}
			catch( UnauthorizedAccessException )
			{
				Console.WriteLine( "Restart the application as an administrator." );
			}
		}

		#endregion Private static methods

		#region Public static methods

		/// <summary>
		/// Application entry point.
		/// </summary>
		public static void Main( string[] args )
		{
			try
			{
				AssemblyUtils.ApplicationBuildInfo = new ConfigureFirewallBuildInfo();

				if( !System.Net.Sockets.Socket.OSSupportsIPv6 )
					Console.WriteLine( "This computer doesn't support IPv6, so the PNRP can't be used to resolve UPNs." );

				// Get the rule name prefix and the file name of the external application from the command line parameters.
				if( (args != null) && (args.Length == 2) && TextUtils.HasContent( args[ 0 ] )
					&& TextUtils.HasContent( args[ 1 ] ) )
				{
					AddFirewallRules( args[ 0 ], args[ 1 ] );
				}
				else
				{
					Console.WriteLine( "You must specify a rule name prefix and an application file name for which to configure"
						+ " the firewall." );
				}
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				Console.WriteLine( ex.ToString() );
			}

			Console.Write( "\nPress a key to finish..." );
			Console.ReadKey( true );
		}

		#endregion Public static methods
	}
}
