﻿using NetFwTypeLib;
using System;

namespace Vela.ConfigureFirewall
{
	/// <summary>
	/// Manager for the Windows Firewall.
	/// </summary>
	public class WindowsFirewall
	{
		#region Private constants

		/// <summary>
		/// Indicates that all the ports are affected by a firewall rule.
		/// </summary>
		private const string AllPorts = "*";

		private const string ManagerProgId = "HNetCfg.FwMgr";
		private const string PolicyProgId = "HNetCfg.FwPolicy2";
		private const string RuleProgId = "HNetCfg.FWRule";

		#endregion Private constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public WindowsFirewall()
		{
			Manager = GetManager();
			Policy = GetPolicy();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// Underlying firewall manager.
		/// </summary>
		private INetFwMgr Manager;

		/// <summary>
		/// Underlying firewall policy.
		/// </summary>
		private INetFwPolicy2 Policy;

		#endregion Private fields

		#region Private static methods

		/// <summary>
		/// Creates an underlying firewall manager.
		/// </summary>
		/// <returns>Returns the created firewall manager.</returns>
		private static INetFwMgr GetManager()
		{
			Type type = Type.GetTypeFromProgID( ManagerProgId, false );

			return (INetFwMgr)Activator.CreateInstance( type );
		}

		/// <summary>
		/// Creates an underlying firewall policy.
		/// </summary>
		/// <returns>Returns the created firewall policy.</returns>
		private static INetFwPolicy2 GetPolicy()
		{
			Type policyType = Type.GetTypeFromProgID( PolicyProgId );

			return (INetFwPolicy2)Activator.CreateInstance( policyType );
		}

		#endregion Private static methods

		#region Private methods

		/// <summary>
		/// Checks if the firewall already has a rule with the specified name and direction.
		/// </summary>
		/// <param name="ruleName">Firewall rule name to check if it already exists.</param>
		/// <param name="direction">Direction of the firewall rule to check if it already exists.</param>
		/// <returns>Returns true if the firewall rule already exists.</returns>
		private bool HasRule( string ruleName, NET_FW_RULE_DIRECTION_ direction )
		{
			foreach( INetFwRule rule in Policy.Rules )
			{
				if( (string.Compare( rule.Name, ruleName, StringComparison.OrdinalIgnoreCase ) == 0)
					&& (rule.Direction == direction) )
				{
					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Gets the firewall rule with the specified name and direction.
		/// </summary>
		/// <param name="ruleName">Name of the firewall rule to get.</param>
		/// <param name="direction">Direction of the firewall rule to get.</param>
		/// <returns>Returns true if the firewall rule already exists.</returns>
		private INetFwRule FindRule( string ruleName, NET_FW_RULE_DIRECTION_ direction )
		{
			foreach( INetFwRule rule in Policy.Rules )
			{
				if( (string.Compare( rule.Name, ruleName, StringComparison.OrdinalIgnoreCase ) == 0)
					&& (rule.Direction == direction) )
				{
					return rule;
				}
			}

			return null;
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Checks if the firewall is enabled for the private profile.
		/// </summary>
		/// <returns>Returns true if the firewall is enabled for the private profile.</returns>
		public bool IsEnabledForPrivateProfile()
		{
			return Policy.get_FirewallEnabled( NET_FW_PROFILE_TYPE2_.NET_FW_PROFILE2_PRIVATE );
		}

		/// <summary>
		/// Checks if the firewall blocks the inbound connections for the private profile.
		/// </summary>
		/// <returns>Returns true if the firewall blocks the inbound connections for the private profile.</returns>
		public bool PrivateInboundConnectionsBlocked()
		{
			return Policy.get_BlockAllInboundTraffic( NET_FW_PROFILE_TYPE2_.NET_FW_PROFILE2_PRIVATE );
		}

		/// <summary>
		/// Sets the block state for the inbound connections for the private profile.
		/// </summary>
		/// <param name="block">If this is true then the state is set to block, else to allow.</param>
		public void BlockPrivateInboundConnections( bool block )
		{
			Policy.set_BlockAllInboundTraffic( NET_FW_PROFILE_TYPE2_.NET_FW_PROFILE2_PRIVATE, block );
		}

		/// <summary>
		/// Checks if the firewall is enabled for the domain profile.
		/// </summary>
		/// <returns>Returns true if the firewall is enabled for the domain profile.</returns>
		public bool IsEnabledForDomainProfile()
		{
			return Policy.get_FirewallEnabled( NET_FW_PROFILE_TYPE2_.NET_FW_PROFILE2_DOMAIN );
		}

		/// <summary>
		/// Checks if the firewall blocks the inbound connections for the domain profile.
		/// </summary>
		/// <returns>Returns true if the firewall blocks the inbound connections for the domain profile.</returns>
		public bool DomainInboundConnectionsBlocked()
		{
			return Policy.get_BlockAllInboundTraffic( NET_FW_PROFILE_TYPE2_.NET_FW_PROFILE2_DOMAIN );
		}

		/// <summary>
		/// Sets the block state for the inbound connections for the domain profile.
		/// </summary>
		/// <param name="block">If this is true then the state is set to block, else to allow.</param>
		public void BlockDomainInboundConnections( bool block )
		{
			Policy.set_BlockAllInboundTraffic( NET_FW_PROFILE_TYPE2_.NET_FW_PROFILE2_DOMAIN, block );
		}

		/// <summary>
		/// Checks if the firewall is enabled for the public profile.
		/// </summary>
		/// <returns>Returns true if the firewall is enabled for the public profile.</returns>
		public bool IsEnabledForPublicProfile()
		{
			return Policy.get_FirewallEnabled( NET_FW_PROFILE_TYPE2_.NET_FW_PROFILE2_PUBLIC );
		}

		/// <summary>
		/// Checks if the firewall blocks the inbound connections for the public profile.
		/// </summary>
		/// <returns>Returns true if the firewall blocks the inbound connections for the public profile.</returns>
		public bool PublicInboundConnectionsBlocked()
		{
			return Policy.get_BlockAllInboundTraffic( NET_FW_PROFILE_TYPE2_.NET_FW_PROFILE2_PUBLIC );
		}

		/// <summary>
		/// Sets the block state for the inbound connections for the public profile.
		/// </summary>
		/// <param name="block">If this is true then the state is set to block, else to allow.</param>
		public void BlockPublicInboundConnections( bool block )
		{
			Policy.set_BlockAllInboundTraffic( NET_FW_PROFILE_TYPE2_.NET_FW_PROFILE2_PUBLIC, block );
		}

		/// <summary>
		/// Adds a firewall rule for for the specified application / service.
		/// </summary>
		/// <param name="ruleName">Firewall rule name.</param>
		/// <param name="isEnabled">If this is true then the rule will be enabled, else disabled.</param>
		/// <param name="isService">If this is true then "applicationFileName" is the file name of an application, else it's the
		/// name of a service.</param>
		/// <param name="applicationFileName">The file name of the application / service for which to add firewall rules.</param>
		/// <param name="alsoPublicProfile">If this is true then the rule will also apply for the public profile, else it will
		/// apply only for the private and domain profiles.</param>
		/// <param name="direction">Direction of the firewall rule.</param>
		/// <param name="protocol">Protocol of the firewall rule.</param>
		/// <param name="localIpAddresses">Local IP addresses of the firewall rule.</param>
		/// <param name="localPorts">Local ports of the firewall rule.</param>
		/// <param name="remoteIpAddresses">Remote IP addresses of the firewall rule.</param>
		/// <param name="remotePorts">Remote ports of the firewall rule.</param>
		/// <remarks>
		/// <para>The existing rule with the specified name is overwritten.</para>
		/// </remarks>
		public void AddRule( string ruleName, bool isEnabled, bool isService, string applicationFileName, bool alsoPublicProfile
			, NET_FW_RULE_DIRECTION_ direction, NET_FW_IP_PROTOCOL_ protocol, string localIpAddresses = null
			, string localPorts = null, string remoteIpAddresses = null, string remotePorts = null )
		{
			string messageTemplate;
			string directionText = (direction == NET_FW_RULE_DIRECTION_.NET_FW_RULE_DIR_IN) ? "inbound" : "outbound";

			// Overwrite the existing rule.
			INetFwRule rule = FindRule( ruleName, direction );

			bool newRule = false;
			if( rule != null )
			{
				messageTemplate = "The {0} rule \"{1}\" will be overwritten.";
				Console.WriteLine( string.Format( messageTemplate, directionText, ruleName ) );

				// Preserve the existing "Enabled" state.
				isEnabled = rule.Enabled;
			}
			else
			{
				messageTemplate = "The {0} rule \"{1}\" will be added.";
				Console.WriteLine( string.Format( messageTemplate, directionText, ruleName ) );

				Type ruleType = Type.GetTypeFromProgID( RuleProgId );
				rule = (INetFwRule)Activator.CreateInstance( ruleType );

				newRule = true;
			}

			rule.Name = ruleName;
			rule.ApplicationName = isService ? null : applicationFileName;
			rule.serviceName = isService ? applicationFileName : null;
			rule.Action = NET_FW_ACTION_.NET_FW_ACTION_ALLOW;
			rule.Direction = direction;
			rule.InterfaceTypes = "All";
			rule.Protocol = (int)protocol;
			rule.Profiles = alsoPublicProfile ? (int)NET_FW_PROFILE_TYPE2_.NET_FW_PROFILE2_ALL
				: (int)(NET_FW_PROFILE_TYPE2_.NET_FW_PROFILE2_PRIVATE | NET_FW_PROFILE_TYPE2_.NET_FW_PROFILE2_DOMAIN);
			if( direction == NET_FW_RULE_DIRECTION_.NET_FW_RULE_DIR_IN )
				rule.EdgeTraversal = true;

			rule.LocalAddresses = localIpAddresses;
			rule.RemoteAddresses = remoteIpAddresses;

			rule.LocalPorts = string.IsNullOrEmpty( localPorts ) ? AllPorts : localPorts;
			rule.RemotePorts = string.IsNullOrEmpty( remotePorts ) ? AllPorts : remotePorts;

			rule.Enabled = isEnabled;

			if( newRule )
				Policy.Rules.Add( rule );
		}

		/// <summary>
		/// Adds firewall rules for the specified application, for both directions.
		/// </summary>
		/// <param name="ruleName">Firewall rule name.</param>
		/// <param name="isEnabled">If this is true then the rule will be enabled, else disabled.</param>
		/// <param name="applicationFileName">The file name of the application for which to add firewall rules.</param>
		/// <param name="alsoPublicProfile">If this is true then the rule will also apply for the public profile, else it will
		/// apply only for the private and domain profiles.</param>
		/// <param name="protocol">Protocol of the firewall rule.</param>
		/// <remarks>
		/// <para>The existing rules with the specified name is overwritten.</para>
		/// </remarks>
		public void AddApplicationRules( string ruleName, bool isEnabled, string applicationFileName, bool alsoPublicProfile
			, NET_FW_IP_PROTOCOL_ protocol )
		{
			AddRule( ruleName, isEnabled, false, applicationFileName, alsoPublicProfile
				, NET_FW_RULE_DIRECTION_.NET_FW_RULE_DIR_IN, protocol );
			AddRule( ruleName, isEnabled, false, applicationFileName, alsoPublicProfile
				, NET_FW_RULE_DIRECTION_.NET_FW_RULE_DIR_OUT, protocol );
		}

		/// <summary>
		/// Adds firewall rules for the specified service, for both directions.
		/// </summary>
		/// <param name="ruleName">Firewall rule name.</param>
		/// <param name="isEnabled">If this is true then the rule will be enabled, else disabled.</param>
		/// <param name="serviceName">The name of the service for which to add firewall rules.</param>
		/// <param name="alsoPublicProfile">If this is true then the rule will also apply for the public profile, else it will
		/// apply only for the private and domain profiles.</param>
		/// <param name="protocol">Protocol of the firewall rule.</param>
		/// <remarks>
		/// <para>The existing rules with the specified name is overwritten.</para>
		/// </remarks>
		public void AddServiceRules( string ruleName, bool isEnabled, string serviceName, bool alsoPublicProfile
			, NET_FW_IP_PROTOCOL_ protocol )
		{
			AddRule( ruleName, isEnabled, true, serviceName, alsoPublicProfile, NET_FW_RULE_DIRECTION_.NET_FW_RULE_DIR_IN
				, protocol );
			AddRule( ruleName, isEnabled, true, serviceName, alsoPublicProfile, NET_FW_RULE_DIRECTION_.NET_FW_RULE_DIR_OUT
				, protocol );
		}

		#endregion Public methods
	}
}
