////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows;
using System.Windows.Threading;
using Vela.ApplicationBackend;
using Vela.Controls;
using Vela.Crypto;
using Vela.ImBackend;
using Vela.Kernel;
using Vela.MsSqlCeDatabase;
using Vela.MsSqlFeDatabase;

namespace Vela.Im
{
	/// <summary>
	/// Application.
	/// </summary>
	public partial class App : Application
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public App()
		{
			InitializeComponent();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The application model.
		/// </summary>
		private ApplicationModel ApplicationModel;

		#endregion Private fields

		#region Public static properties

		/// <summary>
		/// The closing mode of the application.
		/// </summary>
		public static ApplicationCloseModes CloseMode = ApplicationCloseModes.Minimize;

		#endregion Public static properties

		#region Event handlers

		private void Application_DispatcherUnhandledException( object sender, DispatcherUnhandledExceptionEventArgs e )
		{
			if( ExceptionUtils.IsCritical( e.Exception ) )
			{
				// Log even critical exceptions.
				LogManager.Add( e.Exception );

				Current.Shutdown( -1 );

				return;
			}

			ControlsUtils.ShowException( e.Exception );

			// Don't let the exception propagate.
			e.Handled = true;
		}

		private void Application_SessionEnding( object sender, SessionEndingCancelEventArgs e )
		{
			if( ApplicationModel == null )
				return;

			try
			{
				ApplicationModel.CloseOrMinimize( ApplicationCloseModes.CleanUp
					, ImConsts.TimePassedFromLastReceivedLetterTimeToClearIt );
			}
			catch( Exception ex )
			{
				// Log even critical exceptions.
				LogManager.Add( ex );
			}
			finally
			{
				CloseMode = ApplicationCloseModes.Force;
			}
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Sets the regional configuration for bindings.
		/// </summary>
		/// <param name="e">Event data.</param>
		protected override void OnStartup( StartupEventArgs e )
		{
			base.OnStartup( e );

			try
			{
				AssemblyUtils.ApplicationBuildInfo = new ImBuildInfo();

				// Adding this event from XAML (instead of code) raises it twice (consecutively).
				DispatcherUnhandledException += Application_DispatcherUnhandledException;

				// Register (potentially) unused types.
				PortableTypes.RegisterGlobal( typeof( MsSqlCeDatabaseProvider ) );
				PortableTypes.RegisterGlobal( typeof( MsSqlFeDatabaseProvider ) );

				EntropyPool.SetInstance( AlgorithmPool.GetDefaultEntropyAlg() );

				AccountModel.MissingPeerPhotoFileName = ImConsts.MissingPeerPhotoFileName;

				ControlsUtils.SetApplicationInstanceGlobalSettings( this );
				ControlsUtils.SetApplicationGlobalUiSettings();

				ApplicationNotificationManager notificationManager = new ApplicationNotificationManager(
					AssemblyUtils.ApplicationAssembly, ImConsts.AppNotificationResourcePrefix );
				ApplicationModel = new ApplicationModel( Dispatcher, notificationManager, e.Args );

				BindingErrorsListener.Initialize();

				MainWindow mainWindow = new MainWindow( ApplicationModel );

				// This may create windows, so create it after the main window was created.
				ApplicationModel.AppViewWindowsManager = new AppViewWindowsManager( ApplicationModel );

				RecentUserInputGlobalManager.Instance = new RecentUserInputGlobalManager( mainWindow );

				mainWindow.Show();
			}
			catch( Exception ex )
			{
				// Log even critical exceptions.
				LogManager.Add( ex );

				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				MessageWindow.ShowDialog( ex.Message, MessageWindowModes.Error );

				Shutdown( -1 );
			}
		}

		#endregion Private methods
	}
}
