////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.ActionProcessor;
using Vela.ApplicationBackend;
using Vela.Cloning;
using Vela.ConfigureFirewall;
using Vela.Controls;
using Vela.CoreDocuments;
using Vela.ImBackend;
using Vela.ImDocuments;
using Vela.Kernel;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.Im
{
	/// <summary>
	/// Application utilities.
	/// </summary>
	public static class ImUtils
	{
		#region Public static methods

		/// <summary>
		/// Creates a new storehouse.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		/// <returns>Returns true if the storehouse was created.</returns>
		public static bool CreateStorehouse( ApplicationModel applicationModel )
		{
			ValidationUtils.ObjectHasContent( applicationModel, "applicationModel" );

			bool created = false;
			Type navigateType = typeof( StorehouseUserControl );

			try
			{
				created = applicationModel.CreateStorehouse();

				if( created && applicationModel.StorehouseExists )
					navigateType = typeof( VaultsUserControl );

				applicationModel.ViewsModel.Navigate( navigateType );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				applicationModel.ViewsModel.Navigate( navigateType );

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			return created;
		}

		/// <summary>
		/// Opens an existing storehouse.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		/// <param name="browseStorehouseFile">If this is true then the dialog to open a storehouse file is shown.</param>
		/// <param name="storehousePassword">If this has content, the last opened storehouse is automatically opened with this
		/// password; no password dialog is displayed.</param>
		/// <returns>Returns true if the storehouse was opened.</returns>
		/// <remarks>
		/// <para>The storehouse's entropy state (<see cref="Vela.Crypto.EntropyPool"/>) is restored in the application's
		/// entropy pool.</para>
		/// </remarks>
		public static bool OpenStorehouse( ApplicationModel applicationModel, bool browseStorehouseFile
			, string storehousePassword )
		{
			ValidationUtils.ObjectHasContent( applicationModel, "applicationModel" );

			bool opened = false;
			Type navigateType = typeof( StorehouseUserControl );

			try
			{
				opened = applicationModel.OpenStorehouse( browseStorehouseFile, storehousePassword );

				IStorehouseProvider storehouseProvider = applicationModel.StorehouseProvider;
				StorehouseModel storehouseModel = applicationModel.StorehouseModel;
				Locator singleAccountLocator = null;

				if( opened && applicationModel.StorehouseExists )
				{
					if( ApplicationBackendStorehouse.HasLetterWithMainFlag( storehouseProvider, out singleAccountLocator ) )
					{
						navigateType = typeof( LettersUserControl );
					}
					else if( ApplicationBackendStorehouse.HasAccount( storehouseProvider, out singleAccountLocator ) )
					{
						navigateType = (singleAccountLocator != null) ? typeof( LettersUserControl )
							: typeof( AccountsUserControl );
					}
					else
					{
						navigateType = typeof( VaultsUserControl );
					}
				}

				applicationModel.ViewsModel.Navigate( navigateType );

				if( singleAccountLocator != null )
					storehouseModel.SetAccountModel( singleAccountLocator, false );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				applicationModel.ViewsModel.Navigate( navigateType );

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			return opened;
		}

		/// <summary>
		/// Opens the previously opened storehouse (if it still exists), or creates a new one.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		/// <returns>Returns true if the storehouse was created or opened.</returns>
		/// <remarks>
		/// <para>The storehouse's entropy state (<see cref="Vela.Crypto.EntropyPool"/>) is restored in the application's
		/// entropy pool.</para>
		/// </remarks>
		public static bool OpenOrCreateStorehouse( ApplicationModel applicationModel )
		{
			ValidationUtils.ObjectHasContent( applicationModel, "applicationModel" );

			if( TextUtils.HasContent( applicationModel.Configuration.LastOpenedStorehouseFileName ) )
			{
				// Check if the last opened storehouse has to be automatically opened.
				if( applicationModel.Configuration.AutoStartStorehouseWithApplication )
				{
					bool opened = OpenStorehouse( applicationModel, false, ImBackendConsts.StorehousePasswordDef );

					// Hide the application if the storehouse was successfully opened after an auto-start.
					if( AutoStartUtils.IsAutoStarting( applicationModel.CommandLineParameters ) )
						ControlsUtils.SetApplicationVisibility( ApplicationVisibilities.Hidden );

					return opened;
				}
				else
				{
					return OpenStorehouse( applicationModel, false, null );
				}
			}
			else
			{
				// Consider that this is the first time when the application is started (on the current computer).

				ConfigureFirewall();

				QuickHelpWindow.ShowDialogInstance();

				string message = "It appears that this is the first time when you use the application, so a storehouse has to be"
					+ " created."
					+ "\n\n"
					+ "If you have previously created a storehouse, you should open it now.";
				ChoiceMessageResults choiceResult = ChoiceWindow.ShowDialog( message, "Create", "Open" );

				if( choiceResult == ChoiceMessageResults.Left )
				{
					return CreateStorehouse( applicationModel );
				}
				else if( choiceResult == ChoiceMessageResults.Right )
				{
					return OpenStorehouse( applicationModel, true, null );
				}
				else
				{
					applicationModel.ViewsModel.Navigate( typeof( StorehouseUserControl ) );

					return false;
				}
			}
		}

		/// <summary>
		/// Edits the configuration of the application.
		/// </summary>
		public static void EditApplicationConfiguration( ApplicationModel applicationModel )
		{
			ValidationUtils.ObjectHasContent( applicationModel, "applicationModel" );

			// Store a copy of the action processor application configuration because it has to be compared with the one after
			// the editing is finished, to see if the action processors have to be restarted.
			ActionProcessorConfiguration oldActionProcessorConfiguration
				= Cloner.Deep<ActionProcessorConfiguration>( applicationModel.Configuration.ActionProcessor );

			ApplicationConfigurationModel configurationModel = new ApplicationConfigurationModel( applicationModel );
			ApplicationConfigurationWindow window = new ApplicationConfigurationWindow( configurationModel );
			if( !ControlsUtils.ShowDialog( window ) )
				return;

			// Save the data.
			applicationModel.Configuration.Save();

			// Restart the action processors.
			if( applicationModel.StorehouseModel != null )
			{
				IHashSerializer hashSerializer = applicationModel.AlgorithmPool.GetHashSerializerForEquality();
				bool actionProcessorConfigurationModified = !hashSerializer.EqualDocuments( oldActionProcessorConfiguration
					, applicationModel.Configuration.ActionProcessor );

				if( actionProcessorConfigurationModified )
					applicationModel.StorehouseModel.Restart( true );
			}
		}

		/// <summary>
		/// Edits the configuration of the currently opened storehouse.
		/// </summary>
		public static void EditStorehouseConfiguration( ApplicationModel applicationModel )
		{
			ValidationUtils.ObjectHasContent( applicationModel, "applicationModel" );

			IStorehouseProvider storehouseProvider = applicationModel.StorehouseProvider;
			if( storehouseProvider == null )
				return;

			#region Load the data

			// Load the configuration.
			ImStorehouseConfiguration storehouseConfiguration
				= StorehouseUtils.LoadConfiguration<ImStorehouseConfiguration>( storehouseProvider, SheafLocks.Update );

			// Store a copy of the storehouse configuration because it has to be compared with the one after the editing is
			// finished, to see if the action processors have to be restarted.
			ImStorehouseConfiguration oldStorehouseConfiguration = Cloner.Deep<ImStorehouseConfiguration>(
				storehouseConfiguration );

			#endregion Load the data

			#region Edit the data

			StorehouseConfigurationWindow window = new StorehouseConfigurationWindow( storehouseConfiguration );
			if( !ControlsUtils.ShowDialog( window ) )
				return;

			#endregion Edit the data

			#region Save the data

			storehouseConfiguration.ActionServiceListeners.AddOrUpdateAsMostRecent( window.ProcessingPort
				, window.RemoveOlderProcessingPorts );

			IHashSerializer hashSerializer = applicationModel.AlgorithmPool.GetHashSerializerForEquality();
			bool storehouseConfigurationModified = !hashSerializer.EqualDocuments( oldStorehouseConfiguration
				, storehouseConfiguration );

			if( !storehouseConfigurationModified )
				return;

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				transaction.UpdateConfiguration( storehouseConfiguration );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			#endregion Save the data

			// Restart the action processors.
			if( window.ProcessingPortWasModified )
				applicationModel.StorehouseModel.Restart( true );
		}

		/// <summary>
		/// Starts the firewall configuration application.
		/// </summary>
		public static void ConfigureFirewall()
		{
			string message = "Do you want to automatically configure the Windows Firewall?"
				+ "\n\nIf you are sure that Vela must not access the network, choose \"No\"."
				+ "\n\nIf you are not sure what to do, choose \"Yes\".";
			if( MessageWindow.ShowDialog( message, MessageWindowModes.Query ) != QueryMessageResults.Yes )
				return;

			string configureFirewallApplicationFileName = FileUtils.MakeFullName( AssemblyUtils.ApplicationFolderName
				, ConfigureFirewallBuildInfo.ProductFullNameDef + FileExtensions.Application );

			string parameters = string.Format( "{0} \"{1}\"", ImBuildInfo.ProductFullNameDef
				, AssemblyUtils.ApplicationFileName );
			FileUtils.RunSystemCommand( configureFirewallApplicationFileName, parameters );
		}

		#endregion Public static methods
	}
}
