////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.ActionProcessor;
using Vela.ApplicationBackend;
using Vela.Cloning;
using Vela.Controls;
using Vela.ImBackend;
using Vela.StarkDocuments;

namespace Vela.Im
{
	/// <summary>
	/// Application configuration model.
	/// </summary>
	public class ApplicationConfigurationModel : BaseModel
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ApplicationConfigurationModel()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		public ApplicationConfigurationModel( ApplicationModel applicationModel )
			: base( applicationModel )
		{
		}

		#endregion Constructors

		#region Public non-serializable properties

		private ShortCommand _CurrentShortCommand;
		/// <summary>
		/// The current short command. Notifies property observers. This may be null.
		/// </summary>
		public ShortCommand CurrentShortCommand
		{
			get { return _CurrentShortCommand; }
			set
			{
				_CurrentShortCommand = value;

				// Notify property observers.
				OnPropertyChanged( "CurrentShortCommand" );
				OnPropertyChanged( "HasCurrentShortCommand" );
			}
		}

		/// <summary>
		/// This is true if a short command is selected. READ-ONLY.
		/// </summary>
		public bool HasCurrentShortCommand
		{
			get { return CurrentShortCommand != null; }
		}

		#endregion Public non-serializable properties

		#region Public methods

		/// <summary>
		/// Adds a short command to the list.
		/// </summary>
		public void AddShortCommand()
		{
			ShortCommand shortCommand = new ShortCommand();

			// Edit the short command.
			EditShortCommandWindow window = new EditShortCommandWindow( shortCommand, null
				, ApplicationModel.Configuration.ShortCommands );
			if( !ControlsUtils.ShowDialog( window ) )
				return;

			// Update the list of short commands.
			ApplicationModel.Configuration.ShortCommands.Add( window.ShortCommand );

			// Update the selected short command.
			CurrentShortCommand = window.ShortCommand;
		}

		/// <summary>
		/// Edits the current short command from the list.
		/// </summary>
		public void EditShortCommand()
		{
			ShortCommand shortCommand = Cloner.Deep( CurrentShortCommand );

			// Edit the short command.
			EditShortCommandWindow window = new EditShortCommandWindow( shortCommand, shortCommand.SortKey
				, ApplicationModel.Configuration.ShortCommands );
			if( !ControlsUtils.ShowDialog( window ) )
				return;

			// Update the list of short commands.
			ApplicationModel.Configuration.ShortCommands.AddOrReplace( CurrentShortCommand, window.ShortCommand );

			// Update the selected short command.
			CurrentShortCommand = window.ShortCommand;
		}

		/// <summary>
		/// Deletes the current short command from the list.
		/// </summary>
		public void DeleteShortCommand()
		{
			string message = string.Format( "Do you want to delete the short command \"{0}\"?"
				, CurrentShortCommand.DisplayName );
			if( MessageWindow.ShowDialog( message, MessageWindowModes.Query ) != QueryMessageResults.Yes )
				return;

			// Find the next short command.
			ShortCommand nextShortCommand = ApplicationModel.Configuration.ShortCommands.FindNext( CurrentShortCommand
				, false );

			// Update the list of short commands.
			ApplicationModel.Configuration.ShortCommands.RemoveItem( CurrentShortCommand );

			// Update the selected short command.
			CurrentShortCommand = nextShortCommand;
		}

		/// <summary>
		/// Sets the shortcut key of the current short command.
		/// </summary>
		/// <param name="shortcutKey">Shortcut key.</param>
		public void SetShortCommandShortcutKey( char shortcutKey )
		{
			if( CurrentShortCommand == null )
				return;

			ApplicationModel.Configuration.ShortCommands.MoveShortcutKey( shortcutKey, CurrentShortCommand );
		}

		/// <summary>
		/// Generates a random processing port.
		/// </summary>
		/// <returns>Returns the random processing port.</returns>
		public UInt16 GenerateRandomProcessingPort()
		{
			return ActionProcessorUtils.GetRandomServiceProcessingPort(
				ApplicationModel.Configuration.ActionServiceListenerTemplate );
		}

		#endregion Public methods
	}
}
