////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Globalization;
using System.Windows;
using System.Windows.Data;
using System.Windows.Media;
using Vela.ApplicationDocuments;
using Vela.ImBackend;

namespace Vela.Im
{
	/// <summary>
	/// Converter for a brush used to render the background of a letter.
	/// </summary>
	public class LetterBackgroundBrushConverter : IMultiValueConverter
	{
		#region Public methods

		/// <summary>
		/// Determines the brush which has to be used.
		/// </summary>
		/// <param name="values">The values to convert, as follows: account model, letter speckle, inactive brush, initiator
		/// brush (for when the local peer of the current account is the initiator), recipient brush (for when the local peer of
		/// the current account is the recipient).</param>
		/// <param name="targetType">Target types to which to convert.</param>
		/// <param name="parameter">Parameter.</param>
		/// <param name="culture">Culture.</param>
		/// <returns>Returns the brush which has to be used.</returns>
		public object Convert( object[] values, Type targetType, object parameter, CultureInfo culture )
		{
			if( (values == null) || (values.Length != 5) )
				return null;

			// Validate the mandatory parameters.
			if( (values[ 0 ] == null) || (values[ 1 ] == null) || (values[ 2 ] == null) || (values[ 3 ] == null)
				|| (values[ 4 ] == null)
				|| (values[ 0 ] == DependencyProperty.UnsetValue) || (values[ 1 ] == DependencyProperty.UnsetValue)
				|| (values[ 2 ] == DependencyProperty.UnsetValue) || (values[ 3 ] == DependencyProperty.UnsetValue)
				|| (values[ 4 ] == DependencyProperty.UnsetValue) )
			{
				return null;
			}

			// Mandatory parameters.
			StorehouseModel storehouseModel = (StorehouseModel)values[ 0 ];
			LetterSpeckle letterSpeckle = (LetterSpeckle)values[ 1 ];
			Brush inactiveBrush = (Brush)values[ 2 ];
			Brush initiatorBrush = (Brush)values[ 3 ];
			Brush recipientBrush = (Brush)values[ 4 ];

			if( (storehouseModel == null) || (letterSpeckle == null) )
				return inactiveBrush;

			AccountModel accountModel = storehouseModel.AccountModel;
			if( accountModel == null )
				return inactiveBrush;

			// Check if the local peer of the current account is the initiator.
			if( accountModel.LocalUpnFingerprint.Equals( letterSpeckle.InitiatorUpnFingerprint )
				&& accountModel.RemoteUpnFingerprint.Equals( letterSpeckle.RecipientUpnFingerprint ) )
			{
				return initiatorBrush;
			}

			// Check if the local peer of the current account is the recipient.
			if( accountModel.LocalUpnFingerprint.Equals( letterSpeckle.RecipientUpnFingerprint )
				&& accountModel.RemoteUpnFingerprint.Equals( letterSpeckle.InitiatorUpnFingerprint ) )
			{
				return recipientBrush;
			}

			return inactiveBrush;
		}

		/// <summary>
		/// Converts back the specified value to the specified target types. Not implemented.
		/// </summary>
		/// <param name="value">Value to convert.</param>
		/// <param name="targetTypes">Target types to which to convert.</param>
		/// <param name="parameter">Parameter.</param>
		/// <param name="culture">Culture.</param>
		/// <returns>Returns the converted values.</returns>
		public object[] ConvertBack( object value, Type[] targetTypes, object parameter, CultureInfo culture )
		{
			throw new NotImplementedException();
		}

		#endregion Public methods
	}
}
