////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Globalization;
using System.Windows;
using System.Windows.Data;
using Vela.ApplicationBackend;
using Vela.ApplicationDocuments;
using Vela.ImBackend;
using Vela.Storehouse;

namespace Vela.Im
{
	/// <summary>
	/// Loader for the group image preview for a letter speckle (<see cref="LetterSpeckle"/>).
	/// </summary>
	public class LetterGroupImagePreviewConverter : IMultiValueConverter
	{
		#region Public methods

		/// <summary>
		/// Loads the image preview.
		/// </summary>
		/// <param name="values">The values to convert, as follows: application model, letter speckle, the control to which the
		/// letter speckle is bound, flag to force a rebind (see <see cref="ViewSpeckle.ForceRebind"/>; this is a dummy, optional
		/// parameter, used only to receive notifications).</param>
		/// <param name="targetType">Target types to which to convert.</param>
		/// <param name="parameter">Parameter.</param>
		/// <param name="culture">Culture.</param>
		/// <returns>Returns the image preview which was loaded synchronously. This may be null.</returns>
		public object Convert( object[] values, Type targetType, object parameter, CultureInfo culture )
		{
			if( (values == null) || ((values.Length != 3) && (values.Length != 4))
				|| (values[ 0 ] == null) || (values[ 1 ] == null) || (values[ 2 ] == null)
				|| (values[ 0 ] == DependencyProperty.UnsetValue) || (values[ 1 ] == DependencyProperty.UnsetValue)
				|| (values[ 2 ] == DependencyProperty.UnsetValue) )
			{
				return null;
			}

			ApplicationModel applicationModel = (ApplicationModel)values[ 0 ];
			LetterSpeckle letterSpeckle = (LetterSpeckle)values[ 1 ];
			FrameworkElement parentControl = (FrameworkElement)values[ 2 ];

			if( !letterSpeckle.HasImagePreview )
				return null;

			IStorehouseProvider storehouseProvider = applicationModel.StorehouseProvider;
			string rootSaveFolderName = applicationModel.Configuration.ActionProcessor.RootSaveFolderName;

			LetterGroupImagePreviewLoader imageLoader = new LetterGroupImagePreviewLoader( storehouseProvider, rootSaveFolderName
				, applicationModel.Configuration.ShowPreviewsForAttachedImageFilesCount
				, ImBackendConsts.ShowFullSizedGroupImagePreview, letterSpeckle, parentControl, applicationModel.Dispatcher );

			return imageLoader.Load();
		}

		/// <summary>
		/// Converts back the specified value to the specified target types. Not implemented.
		/// </summary>
		/// <param name="value">Value to convert.</param>
		/// <param name="targetTypes">Target types to which to convert.</param>
		/// <param name="parameter">Parameter.</param>
		/// <param name="culture">Culture.</param>
		/// <returns>Returns the converted values.</returns>
		public object[] ConvertBack( object value, Type[] targetTypes, object parameter, CultureInfo culture )
		{
			throw new NotImplementedException();
		}

		#endregion Public methods
	}
}
