////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Globalization;
using System.Windows;
using System.Windows.Data;
using Vela.ActionProcessor;
using Vela.ApplicationDocuments;
using Vela.CoreDocuments;
using Vela.ImBackend;
using Vela.Kernel;

namespace Vela.Im
{
	/// <summary>
	/// Converter for the information about when a new attempt to process an action request will be made.
	/// </summary>
	/// <remarks>
	/// <para>If multiple action requests are being processed by an action client for the same recipient, it's possible that the
	/// progress information (including the time for the next processing attempt) of an action request isn't updated in the UI of
	/// the client. For example, if a client sends multiple letters to the same recipient (who is not accessible), the "Retry"
	/// progress of the letter which is processed first will be updated in the UI of the client, while the progress of the other
	/// letters isn't because the action client may not try to send them (during the same processing cycle).</para>
	/// 
	/// <para>Also, if one of the action requests meant to be sent to the same recipient is processed only in the background,
	/// like <see cref="PushUpdatePresenterActionRequest"/> is, it's possible that no progress of a letter will be updated in the
	/// UI of the client.</para>
	/// </remarks>
	public class NextProcessingAttemptConverter : IMultiValueConverter
	{
		#region Public constants

		/// <summary>
		/// The template for the time format. Placeholders: 0 - the time of the next processing attempt, 1 - The maximum delay
		/// (in seconds) after the time of the next processing attempt, after which an action client automatically loads action
		/// requests (from the storehouse) to process.
		/// </summary>
		public const string TimeFormatTemplate = "{0} (+0...{1} seconds)";

		#endregion Public constants

		#region Public methods

		/// <summary>
		/// Determines the time when a new attempt to process an action request will be made.
		/// </summary>
		/// <param name="values">The values to convert, as follows: storehouse model, speckle of document derived from both
		/// <see cref="PossibleActionRequestSpeckle"/> and <see cref="IRecipientUpnFingerprintContainer"/>.</param>
		/// <param name="targetType">Target types to which to convert.</param>
		/// <param name="parameter">Parameter.</param>
		/// <param name="culture">Culture.</param>
		/// <returns>Returns the brush which has to be used.</returns>
		public object Convert( object[] values, Type targetType, object parameter, CultureInfo culture )
		{
			if( (values == null) || (values.Length != 2)
				|| (values[ 0 ] == null) || (values[ 1 ] == null)
				|| (values[ 0 ] == DependencyProperty.UnsetValue) || (values[ 1 ] == DependencyProperty.UnsetValue) )
			{
				return null;
			}

			StorehouseModel sm = (StorehouseModel)values[ 0 ];
			PossibleActionRequestSpeckle documentSpeckle = (PossibleActionRequestSpeckle)values[ 1 ];
			IRecipientUpnFingerprintContainer upnFingerprintContainer = (IRecipientUpnFingerprintContainer)values[ 1 ];

			if( (sm.ActionClientManager == null) || (documentSpeckle.LastProcessingAttempt == null) )
				return null;

			IRecipientAccessStatesManager recipientAccessStatesManager = sm.ActionClientManager.RecipientAccessStatesManager;
			ActionNextProcessingAttempt nextAttempt = recipientAccessStatesManager.GetNextProcessingAttempt(
				upnFingerprintContainer.RecipientUpnFingerprint, documentSpeckle.LastProcessingAttempt.ProcessingState );

			DateTime nextTime = (nextAttempt != null) ? nextAttempt.NextTime.ToLocalTime() : DateTime.Now;

			string nextTimeText = string.Format( TimeFormatTemplate, nextTime.ToString( KernelConsts.TimeUiFormat )
				, ActionProcessorConsts.ClientStorehouseLoadTimespanMax );

			return nextTimeText;
		}

		/// <summary>
		/// Converts back the specified value to the specified target types. Not implemented.
		/// </summary>
		/// <param name="value">Value to convert.</param>
		/// <param name="targetTypes">Target types to which to convert.</param>
		/// <param name="parameter">Parameter.</param>
		/// <param name="culture">Culture.</param>
		/// <returns>Returns the converted values.</returns>
		public object[] ConvertBack( object value, Type[] targetTypes, object parameter, CultureInfo culture )
		{
			throw new NotImplementedException();
		}

		#endregion Public methods
	}
}
