////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Globalization;
using System.Windows;
using System.Windows.Data;
using Vela.ActionProcessor;
using Vela.ApplicationDocuments;
using Vela.ImBackend;

namespace Vela.Im
{
	/// <summary>
	/// Converter for the information about when a new attempt to process the files attached to a letter will be made.
	/// </summary>
	public class NextProcessingAttemptOfAttachedFilesConverter : IMultiValueConverter
	{
		#region Public methods

		/// <summary>
		/// Determines the time when a new attempt to process the files attached to a letter will be made.
		/// </summary>
		/// <param name="values">The values to convert, as follows: storehouse model, letter speckle.</param>
		/// <param name="targetType">Target types to which to convert.</param>
		/// <param name="parameter">Parameter.</param>
		/// <param name="culture">Culture.</param>
		/// <returns>Returns the brush which has to be used.</returns>
		public object Convert( object[] values, Type targetType, object parameter, CultureInfo culture )
		{
			if( (values == null) || (values.Length != 2)
				|| (values[ 0 ] == null) || (values[ 1 ] == null)
				|| (values[ 0 ] == DependencyProperty.UnsetValue) || (values[ 1 ] == DependencyProperty.UnsetValue) )
			{
				return null;
			}

			StorehouseModel sm = (StorehouseModel)values[ 0 ];
			LetterSpeckle letterSpeckle = (LetterSpeckle)values[ 1 ];

			if( (sm.ActionClientManager == null) || (letterSpeckle.LastProcessingAttemptOfAttachedFiles == null) ) return null;

			IRecipientAccessStatesManager recipientAccessStatesManager = sm.ActionClientManager.RecipientAccessStatesManager;
			ActionNextProcessingAttempt nextAttempt = recipientAccessStatesManager.GetNextProcessingAttempt(
				letterSpeckle.InitiatorUpnFingerprint, letterSpeckle.LastProcessingAttemptOfAttachedFiles.ProcessingState );

			return (nextAttempt != null) ? nextAttempt.NextTime.ToLocalTime() : DateTime.UtcNow;
		}

		/// <summary>
		/// Converts back the specified value to the specified target types. Not implemented.
		/// </summary>
		/// <param name="value">Value to convert.</param>
		/// <param name="targetTypes">Target types to which to convert.</param>
		/// <param name="parameter">Parameter.</param>
		/// <param name="culture">Culture.</param>
		/// <returns>Returns the converted values.</returns>
		public object[] ConvertBack( object value, Type[] targetTypes, object parameter, CultureInfo culture )
		{
			throw new NotImplementedException();
		}

		#endregion Public methods
	}
}
