////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Microsoft.VisualBasic.ApplicationServices;
using System;
using Vela.Controls;
using Vela.Kernel;

namespace Vela.Im
{
	/// <summary>
	/// Manager for a unique application instance.
	/// </summary>
	public class UniqueApplicationInstance : WindowsFormsApplicationBase
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public UniqueApplicationInstance()
		{
			IsSingleInstance = true;
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The unique application instance.
		/// </summary>
		private App ApplicationInstance;

		#endregion Private fields

		#region Private methods

		/// <summary>
		/// Handles the first start of the application.
		/// </summary>
		/// <param name="e">Event data.</param>
		/// <returns>Returns true if the application should continue starting up.</returns>
		protected override bool OnStartup( StartupEventArgs e )
		{
			try
			{
				ApplicationInstance = new App();
				ApplicationInstance.Run();
			}
			catch( Exception ex )
			{
				// Log even critical exceptions.
				LogManager.Add( ex );

				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				MessageWindow.ShowDialog( ex.Message, MessageWindowModes.Error );

				if( ApplicationInstance != null )
					ApplicationInstance.Shutdown( -1 );
			}

			return false;
		}

		/// <summary>
		/// Handles the subsequent starting attempts of the application.
		/// </summary>
		/// <param name="e">Event data.</param>
		protected override void OnStartupNextInstance( StartupNextInstanceEventArgs e )
		{
			// Show the main window.
			ControlsUtils.SetApplicationVisibility( ApplicationVisibilities.Normal );
		}

		#endregion Private methods

		#region Public static methods

		/// <summary>
		/// Application entry point.
		/// </summary>
		[STAThreadAttribute()]
		public static void Main( string[] args )
		{
			try
			{
				UniqueApplicationInstance uai = new UniqueApplicationInstance();
				uai.Run( args );
			}
			catch( Exception ex )
			{
				// Log even critical exceptions.
				LogManager.Add( ex );

				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		#endregion Public static methods
	}
}
