////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Specialized;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Input;
using System.Windows.Threading;
using Vela.ActionProcessor;
using Vela.ApplicationBackend;
using Vela.Controls;
using Vela.Crypto;
using Vela.ImBackend;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.StarkDocuments;

namespace Vela.Im
{
	/// <summary>
	/// View for working with letters.
	/// </summary>
	[FunctionViewRegistrationAttribute( "500" )]
	public partial class LettersUserControl : AppViewUserControl
	{
		#region Private constants

		/// <summary>
		/// The shortcut key prefix used for updating a letter flag.
		/// </summary>
		private const string UpdateFlagShortcutKeyPrefix = "Shift + ";

		/// <summary>
		/// The resource key which contains the context menu of a letter.
		/// </summary>
		private const string LetterMenuResourceKey = "LetterMenu";

		/// <summary>
		/// The resource key which contains the context menu of the list of letters.
		/// </summary>
		private const string LettersListMenuResourceKey = "LettersListMenu";

		/// <summary>
		/// The resource key which contains the menu used to update the flag of a letter.
		/// </summary>
		private const string UpdateLetterFlagMenuResourceKey = "UpdateLetterFlagMenu";

		/// <summary>
		/// The number of random lowercase letters to insert in the draft letter.
		/// </summary>
		private const int InsertRandomTextSize = 15;

		/// <summary>
		/// The timespan (in seconds) for the <see cref="LoadTimeConsumingInformationOfVisibleLettersTimer"/> timer.
		/// </summary>
		private const double LoadTimeConsumingInformationOfVisibleLettersTimerTimespan = 0.5;

		#endregion Private constants

		#region Inner classes

		/// <summary>
		/// Context for a menu which starts a short command.
		/// </summary>
		private class ShortCommandMenuContext
		{
			public ShortCommandMenuContext( ShortCommand shortCommand, TextBox textBox )
			{
				ShortCommand = shortCommand;
				TextBox = textBox;
			}

			public ShortCommand ShortCommand;
			public TextBox TextBox;
		}

		#endregion Inner classes

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected LettersUserControl()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		public LettersUserControl( ApplicationModel applicationModel )
			: base( applicationModel )
		{
			InitializeComponent();

			LoadUi();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// If this is true then the list of letters is scrolled to the bottom when the current control is loaded.
		/// </summary>
		private bool ScrollListBoxBottomIntoViewOnLoad = true;

		/// <summary>
		/// If this is true then the list-box with letters is scrolled down near the bottom.
		/// </summary>
		private bool IsLettersScrolledNearBottom;

		/// <summary>
		/// The time (in UTC format) when the list of letters last became out of sight. For example, this is the time when the
		/// context menu of the list of letters was last opened. This is used to ensure that the letters which arrive after the
		/// list became out of sight are not affected by page operations.
		/// </summary>
		private DateTime ListOutOfSightTime;

		/// <summary>
		/// Timer which loads the visible letter speckles with the information which takes a significant amount of time to load.
		/// </summary>
		private DispatcherTimer LoadTimeConsumingInformationOfVisibleLettersTimer;

		#endregion Private fields

		#region Public static properties

		/// <summary>
		/// Title.
		/// </summary>
		public static string Title
		{
			get { return "Letters"; }
		}

		/// <summary>
		/// Tooltip.
		/// </summary>
		public static string Tooltip
		{
			get
			{
				return "A letter contains text that you type and send to a contact of yours, or that a contact sends to you";
			}
		}

		#endregion Public static properties

		#region Public non-serializable properties

		private CompositeCollection _UpdateLetterPredefinedFlagMenuItems = new CompositeCollection();
		/// <summary>
		/// The menu items for the context menu of the list of letters, which show the predefined flags to which the current
		/// letter may be updated. The backing field is initialized by default. This may be empty, but not null. READ-ONLY.
		/// </summary>
		public CompositeCollection UpdateLetterPredefinedFlagMenuItems
		{
			get { return _UpdateLetterPredefinedFlagMenuItems; }
		}

		private CompositeCollection _ShowLettersWithFlagMenuItems = new CompositeCollection();
		/// <summary>
		/// The menu items for the context menu of the list of letters, which can be used to show the letters with a specific
		/// flag. The backing field is initialized by default. This may be empty, but not null. READ-ONLY.
		/// </summary>
		public CompositeCollection ShowLettersWithFlagMenuItems
		{
			get { return _ShowLettersWithFlagMenuItems; }
		}

		private CompositeCollection _UpdateLetterFlagMenuItems = new CompositeCollection();
		/// <summary>
		/// The menu items for the context menu of the list of letters, which show the flags to which the current letter may be
		/// updated. The backing field is initialized by default. This may be empty, but not null. READ-ONLY.
		/// </summary>
		public CompositeCollection UpdateLetterFlagMenuItems
		{
			get { return _UpdateLetterFlagMenuItems; }
		}

		private CompositeCollection _UpdateLettersFlagMenuItems = new CompositeCollection();
		/// <summary>
		/// The menu items for the context menu of the list of letters, which show the flags to which the letters with the
		/// current flag may be updated. The backing field is initialized by default. This may be empty, but not null. READ-ONLY.
		/// </summary>
		public CompositeCollection UpdateLettersFlagMenuItems
		{
			get { return _UpdateLettersFlagMenuItems; }
		}

		private CompositeCollection _LetterShortCommandsMenuItems = new CompositeCollection();
		/// <summary>
		/// The menu items with short commands for the context menu of the current letter. The backing field is initialized by
		/// default. This may be empty, but not null. READ-ONLY.
		/// </summary>
		public CompositeCollection LetterShortCommandsMenuItems
		{
			get { return _LetterShortCommandsMenuItems; }
		}

		private CompositeCollection _MessageShortCommandsMenuItems = new CompositeCollection();
		/// <summary>
		/// The menu items with short commands for the context menu of the message. The backing field is initialized by default.
		/// This may be empty, but not null. READ-ONLY.
		/// </summary>
		public CompositeCollection MessageShortCommandsMenuItems
		{
			get { return _MessageShortCommandsMenuItems; }
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void AppViewUserControl_Loaded( object sender, RoutedEventArgs e )
		{
			if( (ApplicationModel.StorehouseModel == null) || (ApplicationModel.StorehouseModel.LettersModel == null) )
				return;

			// Load the letters (only now is the storehouse model available).
			ApplicationModel.StorehouseModel.LettersModel.LoadLettersFirstTime();

			if( ScrollListBoxBottomIntoViewOnLoad )
				ControlsUtils.ScrollListBoxBottomIntoView( LettersListBox );
			ScrollListBoxBottomIntoViewOnLoad = false;

			// Ensure that the user sees no delay when loading the information, when the list of letters is loaded for the first
			// time.
			LoadTimeConsumingInformationOfLetters();

			((INotifyCollectionChanged)LettersListBox.Items).CollectionChanged += LettersListBox_CollectionChanged;

			ControlsUtils.FocusTextBox( DraftMessageTextBox, true );

			if( !ImBackendConsts.LoadTimeConsumingInformationWhenCollectingLetters )
			{
				LoadTimeConsumingInformationOfVisibleLettersTimer = new DispatcherTimer();
				LoadTimeConsumingInformationOfVisibleLettersTimer.Interval = TimeSpan.FromSeconds(
					LoadTimeConsumingInformationOfVisibleLettersTimerTimespan );
				LoadTimeConsumingInformationOfVisibleLettersTimer.Tick
					+= new EventHandler( LoadTimeConsumingInformationOfVisibleLettersTimer_Tick );
				LoadTimeConsumingInformationOfVisibleLettersTimer.IsEnabled = true;
			}
		}

		private void LoadTimeConsumingInformationOfVisibleLettersTimer_Tick( object sender, EventArgs e )
		{
			if( MayLoadTimeConsumingInformationOfVisibleLetters() )
				LoadTimeConsumingInformationOfLetters();
		}

		private void Tooltip_Opening( object sender, ToolTipEventArgs e )
		{
			ControlsUtils.UpdateBindingTargets( ((FrameworkElement)sender).ToolTip as FrameworkElement );
		}

		private void DeleteLetterButton_Click( object sender, RoutedEventArgs e )
		{
			DeleteLetter();
		}

		private void DeleteLettersButton_Click( object sender, RoutedEventArgs e )
		{
			DeleteLetters();
		}

		private void FlagFiltersListBox_SelectionChanged( object sender, SelectionChangedEventArgs e )
		{
			if( LettersListBox != null )
				ControlsUtils.ScrollListBoxBottomIntoView( LettersListBox );

			ControlsUtils.FocusTextBox( DraftMessageTextBox, true );
		}

		private void ShowOldestPageButton_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.ShowOldestPage();

			ControlsUtils.ScrollListBoxTopIntoView( LettersListBox );
		}

		private void ShowOlderPageButton_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.ShowOlderPage();

			ControlsUtils.ScrollListBoxBottomIntoView( LettersListBox );
		}

		private void ShowNewerPageButton_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.ShowNewerPage();

			ControlsUtils.ScrollListBoxTopIntoView( LettersListBox );
		}

		private void ShowNewestPageButton_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.ShowNewestPage();

			ControlsUtils.ScrollListBoxBottomIntoView( LettersListBox );
		}

		private void LettersListBox_PreviewMouseDown( object sender, MouseButtonEventArgs e )
		{
			DependencyObject originalSource = e.OriginalSource as DependencyObject;
			if( ControlsUtils.FindVisualParent<System.Windows.Controls.Primitives.ScrollBar>( originalSource ) != null )
				return;

			ControlsUtils.FocusParentListBoxItem( LettersListBox, originalSource );

			if( e.ChangedButton == MouseButton.Left )
			{
				if( e.ClickCount == 1 )
				{
					if( Keyboard.Modifiers == ModifierKeys.None )
					{
						ControlsUtils.SetCaretAtClick( sender as TextBox, e );
					}
					else if( Keyboard.Modifiers == ModifierKeys.Control )
					{
						VisitLetterUrl( e.GetPosition( GetCurrentLetterMessageSummaryTextBox() ) );
					}
				}
				else if( e.ClickCount == 2 )
				{
					if( !ViewClickedAttachedImageFile( originalSource as AnimatedImage ) )
						ApplicationModel.StorehouseModel.LettersModel.ViewLetterInWindow();
				}
			}
			else if( e.ChangedButton == MouseButton.Right )
			{
				ShowCurrentLetterMenu();
			}
			else if( e.ChangedButton == MouseButton.Middle )
			{
				Reply();
			}
		}

		private void LettersListBox_PreviewKeyDown( object sender, KeyEventArgs e )
		{
			ModifierKeys modifierKeys = Keyboard.Modifiers;

			if( modifierKeys == ModifierKeys.None )
			{
				// Edit in a separate window.
				if( (e.Key == Key.F2) || (e.Key == Key.Enter) )
				{
					ApplicationModel.StorehouseModel.LettersModel.ViewLetterInWindow();
					e.Handled = true;
					return;
				}

				// Delete the letter.
				if( e.Key == Key.Delete )
				{
					DeleteLetter();
					e.Handled = true;
					return;
				}
			}

			if( modifierKeys == ModifierKeys.Control )
			{
				// Clip text.
				if( e.Key == Key.C )
				{
					ClipLetterMessageSummary();
					e.Handled = true;
					return;
				}

				// Reply.
				if( e.Key == Key.R )
				{ Reply(); e.Handled = true; return; }
			}

			if( modifierKeys == (ModifierKeys.Control | ModifierKeys.Shift) )
			{
				if( e.Key == Key.C )
				{
					ClipLetterSecondaryProcessingStateMessage();
					e.Handled = true;
					return;
				}
			}

			// Update the letter flag.
			if( modifierKeys == ModifierKeys.Shift )
			{
				if( e.Key == Key.A )
				{ UpdateLetterFlag( PostFlags.Archive ); e.Handled = true; return; }
				else if( e.Key == Key.F )
				{ UpdateLetterFlag( PostFlags.Favorite ); e.Handled = true; return; }
				else if( e.Key == Key.M )
				{ UpdateLetterFlag( PostFlags.Main ); e.Handled = true; return; }
				else if( e.Key == Key.O )
				{ UpdateLetterFlag( PostFlags.Ongoing ); e.Handled = true; return; }
				else if( e.Key == Key.T )
				{ UpdateLetterFlag( PostFlags.Trash ); e.Handled = true; return; }
			}
		}

		private void LettersListBox_CollectionChanged( object sender, NotifyCollectionChangedEventArgs e )
		{
			if( e.Action == NotifyCollectionChangedAction.Add )
				ScrollNewestLetterIntoView();
		}

		private void LettersListBox_ScrollChanged( object sender, ScrollChangedEventArgs e )
		{
			IsLettersScrolledNearBottom = ControlsUtils.IsListBoxScrolledNearBottom( LettersListBox );
		}

		private void UpdateLetterFlagLabel_MouseDown( object sender, MouseButtonEventArgs e )
		{
			TextBlock textBlock = sender as TextBlock;
			if( (textBlock == null) || (e.ChangedButton != MouseButton.Left) )
				return;

			ControlsUtils.FocusParentListBoxItem( LettersListBox, textBlock );

			CreateFlagFilterMenuItems( UpdateLetterPredefinedFlagMenuItems, UpdateLetterFlagMenuItem_Click, false, false );

			ContextMenu updateLetterFlagMenu = FindResource( UpdateLetterFlagMenuResourceKey ) as ContextMenu;
			updateLetterFlagMenu.PlacementTarget = textBlock;
			updateLetterFlagMenu.Placement = System.Windows.Controls.Primitives.PlacementMode.MousePoint;
			updateLetterFlagMenu.IsOpen = true;

			e.Handled = true;
		}

		private void LettersListBox_MessageSummary_PreviewKeyDown( object sender, KeyEventArgs e )
		{
			// Pass all the keys (from the message summary text-box) to the list-box, for navigation.
			ControlsUtils.PassKeyDownEventToParent<ListBox>( sender as TextBox, e );
		}

		private void ReplyMenuItem_Click( object sender, RoutedEventArgs e )
		{
			Reply();
		}

		private void ExploreAttachedFilesMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.ExploreAttachedFiles();
		}

		private void PlayAttachedMediaFilesMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.PlayAttachedMediaFiles();
		}

		private void ViewLetterInWindowMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.ViewLetterInWindow();
		}

		private void ShowLettersWithFlagMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ShowLettersWithFlag( (string)((MenuItem)sender).Header );
		}

		private void UpdateLetterFlagMenuItem_Click( object sender, RoutedEventArgs e )
		{
			UpdateLetterFlag( (string)((MenuItem)sender).Header );
		}

		private void UpdateLettersFlagMenuItem_Click( object sender, RoutedEventArgs e )
		{
			UpdateLettersFlag( (string)((MenuItem)sender).Header );
		}

		private void VisitLetterUrlMenuItem_Click( object sender, RoutedEventArgs e )
		{
			VisitLetterUrl();
		}

		private void ClipAllLettersMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.ClipAllLetters();
		}

		private void ExportLetterMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.ExportLetter( false );
		}

		private void ExportObfuscatedLetterMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.ExportLetter( true );
		}

		private void ImportLettersMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.ImportLetters();
		}

		private void ClipLetterMessageSummaryMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ClipLetterMessageSummary();
		}

		private void ClipLetterSecondaryProcessingStateMessageMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ClipLetterSecondaryProcessingStateMessage();
		}

		private void DeleteLetterMenuItem_Click( object sender, RoutedEventArgs e )
		{
			DeleteLetter();
		}

		private void DeleteLettersMenuItem_Click( object sender, RoutedEventArgs e )
		{
			DeleteLetters();
		}

		private void ViewAttachedFilesNamesMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.ViewAttachedFilesNames();
		}

		private void ChangeWallpaperMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.ChangeWallpaper();
		}

		private void UpdateWithCustomEmoticonsMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.UpdateWithCustomEmoticons();
		}

		private void RetrySendLetterMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.RetrySendLetter();
		}

		private void SplittableGrid_SizeChanged( object sender, SizeChangedEventArgs e )
		{
			// Bug fix (in .Net): The bottom part of the letter message area slides out of view when the splitter is dragged
			// upward more than what the minimum size for the list-box area allows.
			LetterMessageAreaRow.MaxHeight = e.NewSize.Height - (LettersListAreaRow.MinHeight + AreaSplitter.Height);
		}

		private void DraftMessageTextBox_PreviewMouseDown( object sender, MouseButtonEventArgs e )
		{
			ClearLastReceivedLetterTime();

			// Visit URL.
			if( Keyboard.Modifiers == ModifierKeys.Control )
				DraftMessageVisitUrl( e.GetPosition( DraftMessageTextBox ) );
		}

		private void DraftMessageTextBox_PreviewKeyDown( object sender, KeyEventArgs e )
		{
			ModifierKeys modifierKeys = Keyboard.Modifiers;

			if( modifierKeys == ModifierKeys.None )
			{
				// Queue the letter.
				if( (e.Key == Key.Enter) && ApplicationModel.Configuration.SendDraftLetterWithPlainEnter )
				{
					DraftQueueSendLetter();
					e.Handled = true;
					return;
				}

				// Edit in a separate window.
				if( e.Key == Key.F2 )
				{ EditDraftMessageInWindow(); e.Handled = true; return; }
			}

			if( modifierKeys == ModifierKeys.Control )
			{
				// Queue the letter or insert a new line.
				if( e.Key == Key.Enter )
				{
					if( ApplicationModel.Configuration.SendDraftLetterWithPlainEnter )
						DraftInsertNewLine();
					else
						DraftQueueSendLetter();

					e.Handled = true;
					return;
				}

				// Select emoticon.
				if( e.Key == Key.E )
				{ SelectEmoticon(); e.Handled = true; return; }

				// Attach files.
				if( e.Key == Key.H )
				{ DraftAttachFileNames(); e.Handled = true; return; }

				// Attach the clipboard image.
				if( e.Key == Key.V )
				{ e.Handled = ListUtils.HasContent( DraftAttachClipboardImage() ); return; }

				// Visit URL.
				if( e.Key == Key.W )
				{ DraftVisitMessageUrl(); e.Handled = true; return; }

				// Insert random text.
				if( e.Key == Key.M )
				{ DraftInsertRandomText(); e.Handled = true; return; }
			}

			// Short command.
			if( (modifierKeys == ModifierKeys.Alt) || (modifierKeys == (ModifierKeys.Alt | ModifierKeys.Control)) )
			{
				char ch = ControlsUtils.ConvertKeyToAlpha( e.Key, e.SystemKey );
				ShortCommand shortCommand = ApplicationModel.Configuration.ShortCommands.FindByShortcutKey( ch );
				if( shortCommand != null )
					RunShortCommand( shortCommand.Command, DraftMessageTextBox, modifierKeys.HasFlag( ModifierKeys.Control ) );
			}
		}

		private void DraftMessageTextBox_KeyDown( object sender, KeyEventArgs e )
		{
			ControlsUtils.AddKeyboardEntropy( e );
		}

		private void DraftMessageTextBox_TextChanged( object sender, TextChangedEventArgs e )
		{
			if( e.Changes.Count <= 0 )
				return;

			ClearLastReceivedLetterTime();
			SetAccountModelBasedOnDraftMessageAccountAlias();
		}

		private void DraftMessageTextBox_ContextMenuOpening( object sender, ContextMenuEventArgs e )
		{
			CreateShortCommandMenuItems( MessageShortCommandsMenuItems, DraftMessageTextBox );
		}

		private void DraftMessageTextBox_PreviewDragEnter( object sender, DragEventArgs e )
		{
			if( DraftMayAttachDroppableFileNames( e.Data ) )
				e.Effects = DragDropEffects.Link;
		}

		private void DraftMessageTextBox_PreviewDragOver( object sender, DragEventArgs e )
		{
			e.Handled = true;
		}

		private void DraftMessageTextBox_Drop( object sender, DragEventArgs e )
		{
			try
			{
				DraftAttachDroppedFileNames( e.Data );

				ControlsUtils.FocusTextBox( DraftMessageTextBox, false );
			}
			catch( Exception ex )
			{
				// The operating system swallows drag-and-drop exceptions, so show them before that happens.
				ControlsUtils.ShowException( ex );
			}
		}

		private void DraftMessageTextBox_Pasting( object sender, DataObjectPastingEventArgs e )
		{
			ControlsUtils.BlockPasteIfTooLong( (TextBox)sender, e );
		}

		private void DraftQueueSendLetterMenuItem_Click( object sender, RoutedEventArgs e )
		{
			DraftQueueSendLetter();
		}

		private void DraftEditMessageInWindowMenuItem_Click( object sender, RoutedEventArgs e )
		{
			EditDraftMessageInWindow();
		}

		private void DraftAttachFileNamesMenuItem_Click( object sender, RoutedEventArgs e )
		{
			DraftAttachFileNames();
		}

		private void DraftRemoveAttachedFilesMenuItem_Click( object sender, RoutedEventArgs e )
		{
			DraftRemoveAttachedFiles();
		}

		private void DraftVisitUrlMenuItem_Click( object sender, RoutedEventArgs e )
		{
			DraftVisitMessageUrl();
		}

		private void ShortCommandMenuItem_Click( object sender, RoutedEventArgs e )
		{
			MenuItem menuItem = (MenuItem)sender;

			ShortCommandMenuContext context = menuItem.Tag as ShortCommandMenuContext;
			RunShortCommand( context.ShortCommand.Command, context.TextBox, Keyboard.Modifiers.HasFlag( ModifierKeys.Control ) );
		}

		private void SelectEmoticonMenuItem_Click( object sender, RoutedEventArgs e )
		{
			SelectEmoticon();
		}

		private void DraftInsertRandomTextMenuItem_Click( object sender, RoutedEventArgs e )
		{
			DraftInsertRandomText();
		}

		private void DraftAttachClipboardImageMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.DraftAttachClipboardImage();
		}

		private void DraftAttachCustomEmoticonsMenuItem_Click( object sender, RoutedEventArgs e )
		{
			DraftAttachCustomEmoticons();
		}

		private void DraftViewAttachedFilesNamesMenuItem_Click( object sender, RoutedEventArgs e )
		{
			ApplicationModel.StorehouseModel.LettersModel.ViewDraftAttachedFilesNames();
		}

		private void SelectEmoticonButton_Click( object sender, RoutedEventArgs e )
		{
			SelectEmoticon();
		}

		private void DraftQueueSendLetterButton_Click( object sender, RoutedEventArgs e )
		{
			DraftQueueSendLetter();
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Checks if it's allowed to load the visible letter speckles with the information which takes a significant amount
		/// of time to load.
		/// </summary>
		/// <returns>Returns true if it's allowed to load the visible letter speckles with the information which takes a
		/// significant amount of time to load.</returns>
		/// <remarks>
		/// <para>The information should not be loaded if the user is using the keyboard or the mouse, because that would
		/// interfere with the user's ability to scroll or perform other operations.</para>
		/// </remarks>
		private bool MayLoadTimeConsumingInformationOfVisibleLetters()
		{
			return !RecentUserInputGlobalManager.Instance.HasRecentUserInput();
		}

		/// <summary>
		/// Gets the indexes of the first and last visible items in the list of letters.
		/// </summary>
		/// <param name="firstVisibleItemIndex">Is set with the index of the first visible item.</param>
		/// <param name="lastVisibleItemIndex">Is set with the index of the last visible item.</param>
		/// <returns>Returns the number of visible items.</returns>
		private int GetVisibleItems( out int firstVisibleItemIndex, out int lastVisibleItemIndex )
		{
			return ControlsUtils.GetVisibleItems( LettersListBox, out firstVisibleItemIndex, out lastVisibleItemIndex );
		}

		/// <summary>
		/// Loads the visible letter speckles with the information which takes a significant amount of time to load.
		/// </summary>
		private void LoadTimeConsumingInformationOfLetters()
		{
			int firstItemIndex;
			int lastItemIndex;
			GetVisibleItems( out firstItemIndex, out lastItemIndex );

			ApplicationModel.StorehouseModel.LettersModel.LoadTimeConsumingInformationOfLetters( firstItemIndex, lastItemIndex );
		}

		/// <summary>
		/// When a new letter is added to the list, the list is scrolled to the bottom (because that's where the new letter is).
		/// </summary>
		/// <remarks>
		/// <para>The scroll is performed only if the list is already scrolled down near the bottom. This allows the user to
		/// scroll the list without the scroll position being reset every time a new letter is received.</para>
		/// </remarks>
		private void ScrollNewestLetterIntoView()
		{
			if( (ApplicationModel.StorehouseModel == null) || (ApplicationModel.StorehouseModel.LettersModel == null) )
				return;

			if( ApplicationModel.StorehouseModel.LettersModel.MayContainNewlyReceivedLetter()
				&& (IsLettersScrolledNearBottom || ControlsUtils.IsListBoxScrolledNearBottom( LettersListBox )) )
			{
				ControlsUtils.ScrollListBoxBottomIntoView( LettersListBox );
			}
		}

		/// <summary>
		/// Gets the text-box with a message summary whose associated letter item is currently selected in the letters list-box.
		/// This may be null.
		/// </summary>
		private TextBox GetCurrentLetterMessageSummaryTextBox()
		{
			if( (ApplicationModel.StorehouseModel == null) || (ApplicationModel.StorehouseModel.LettersModel == null) )
				return null;

			return ControlsUtils.GetListBoxItemChild<TextBox>( LettersListBox
				, ApplicationModel.StorehouseModel.LettersModel.CurrentLetter );
		}

		/// <summary>
		/// Clears the time when the last letter was received.
		/// </summary>
		private void ClearLastReceivedLetterTime()
		{
			if( (ApplicationModel.StorehouseModel == null) || (ApplicationModel.StorehouseModel.LettersModel == null) )
				return;

			ApplicationModel.StorehouseModel.LettersModel.ClearLastReceivedLetterTime(
				ImConsts.TimePassedFromLastReceivedLetterTimeToClearIt );
		}

		/// <summary>
		/// Replys to the remote peer of the current letter.
		/// </summary>
		/// <remarks>
		/// <para>The account model is set with the account associated with remote peer of the current letter.</para>
		/// </remarks>
		private void Reply()
		{
			ControlsUtils.ExecuteAndMaintainListBoxScrollIfActionSuccessful<LettersModel>(
				( lettersModel ) => { return lettersModel.SetAccountModel(); }
				, ApplicationModel.StorehouseModel.LettersModel, LettersListBox );

			// Focus the message text-box.
			ControlsUtils.FocusTextBox( DraftMessageTextBox, true );
		}

		/// <summary>
		/// Sets the account model with the account whose alias is at the beginning of the draft message.
		/// </summary>
		private void SetAccountModelBasedOnDraftMessageAccountAlias()
		{
			if( ApplicationModel.StorehouseModel == null )
				return;

			AccountAlias accountAlias = AccountAlias.FromPostMessage( DraftMessageTextBox.Text );

			ApplicationModel.StorehouseModel.SetAccountModel( accountAlias );
		}

		/// <summary>
		/// Shows the context menu of the current letter.
		/// </summary>
		private void ShowCurrentLetterMenu()
		{
			ListOutOfSightTime = DateTime.UtcNow;

			CreateFlagFilterMenuItems( ShowLettersWithFlagMenuItems, ShowLettersWithFlagMenuItem_Click, true, false );
			CreateFlagFilterMenuItems( UpdateLettersFlagMenuItems, UpdateLettersFlagMenuItem_Click, true, false );

			TextBox currentLetterMessageSummaryTextBox = GetCurrentLetterMessageSummaryTextBox();
			ListBoxItem listBoxItem = ControlsUtils.FindVisualParent<ListBoxItem>( currentLetterMessageSummaryTextBox );

			if( listBoxItem != null )
			{
				CreateFlagFilterMenuItems( UpdateLetterFlagMenuItems, UpdateLetterFlagMenuItem_Click, true, true );
				CreateShortCommandMenuItems( LetterShortCommandsMenuItems, currentLetterMessageSummaryTextBox );

				ContextMenu letterMenu = FindResource( LetterMenuResourceKey ) as ContextMenu;
				letterMenu.PlacementTarget = listBoxItem;
				letterMenu.Placement = System.Windows.Controls.Primitives.PlacementMode.MousePoint;
				letterMenu.IsOpen = true;
			}
			else
			{
				ContextMenu letterMenu = FindResource( LettersListMenuResourceKey ) as ContextMenu;
				letterMenu.PlacementTarget = LettersListBox;
				letterMenu.Placement = System.Windows.Controls.Primitives.PlacementMode.MousePoint;
				letterMenu.IsOpen = true;
			}
		}

		/// <summary>
		/// Creates the menu items with flags, in the context menu of the letters.
		/// </summary>
		/// <param name="menuItems">The menu items to add to the context menu. This is filled with menu items which represent
		/// the available flags.</param>
		/// <param name="menuItemHandler">Method which is called when a menu item is clicked.</param>
		/// <param name="useAllAvailableFlags">If this is true then all the available flags are used, else only the predefined
		/// ones.</param>
		/// <param name="showShortcutKey">If this is true then the shortcut key is shown in the menu items.</param>
		private void CreateFlagFilterMenuItems( CompositeCollection menuItems, RoutedEventHandler menuItemHandler
			, bool useAllAvailableFlags, bool showShortcutKey )
		{
			// Clear the existing menu items.
			foreach( MenuItem menuItem in menuItems )
				menuItem.Click -= menuItemHandler;
			menuItems.Clear();

			Strings flags = useAllAvailableFlags ? ApplicationModel.StorehouseModel.LettersModel.PostFlags
				: ApplicationModel.StorehouseModel.LettersModel.PredefinedPostFlags;
			string flagToSkip = ApplicationModel.StorehouseModel.LettersModel.FlagFilter;

			foreach( string flag in flags )
			{
				MenuItem menuItem = new MenuItem();
				menuItem.Header = flag;
				menuItem.IsEnabled = !Post.EqualFlags( flagToSkip, flag );
				menuItem.Click += menuItemHandler;

				if( showShortcutKey )
				{
					string shortcutKey = null;

					if( Post.EqualFlags( flag, PostFlags.Archive.ToString() ) )
						shortcutKey = "A";
					else if( Post.EqualFlags( flag, PostFlags.Favorite.ToString() ) )
						shortcutKey = "F";
					else if( Post.EqualFlags( flag, PostFlags.Main.ToString() ) )
						shortcutKey = "M";
					else if( Post.EqualFlags( flag, PostFlags.Ongoing.ToString() ) )
						shortcutKey = "O";
					else if( Post.EqualFlags( flag, PostFlags.Trash.ToString() ) )
						shortcutKey = "T";

					if( TextUtils.HasContent( shortcutKey ) )
						menuItem.InputGestureText = UpdateFlagShortcutKeyPrefix + shortcutKey;
				}

				menuItems.Add( menuItem );
			}
		}

		/// <summary>
		/// Shows the letters with the specified flag.
		/// </summary>
		private void ShowLettersWithFlag( string flagFilter )
		{
			ApplicationModel.StorehouseModel.LettersModel.FlagFilter = flagFilter;
		}

		/// <summary>
		/// Updates the flag of the current letter with the specified flag.
		/// </summary>
		private void UpdateLetterFlag( string flag )
		{
			ControlsUtils.ExecuteAndMaintainListBoxScroll<LettersModel>(
				( lettersModel ) => { lettersModel.UpdateLetterFlag( flag ); }
				, ApplicationModel.StorehouseModel.LettersModel, LettersListBox );
		}

		/// <summary>
		/// Updates the flag of the current letter with the specified flag.
		/// </summary>
		private void UpdateLetterFlag( PostFlags flag )
		{
			UpdateLetterFlag( flag.ToString() );
		}

		/// <summary>
		/// Updates the flags of the letters with the current flag filter, with the specified flag.
		/// </summary>
		/// <param name="flag"></param>
		private void UpdateLettersFlag( string flag )
		{
			ControlsUtils.ExecuteAndMaintainListBoxScroll<LettersModel>(
				( lettersModel ) => { lettersModel.UpdateLettersFlag( flag, ListOutOfSightTime ); }
				, ApplicationModel.StorehouseModel.LettersModel, LettersListBox );
		}

		/// <summary>
		/// Updates the flags of the letters with the current flag filter, with the specified flag.
		/// </summary>
		/// <param name="flag"></param>
		private void UpdateLettersFlag( PostFlags flag )
		{
			UpdateLettersFlag( flag.ToString() );
		}

		/// <summary>
		/// Deletes the current letter and maintains the (before-after) scroll.
		/// </summary>
		private void DeleteLetter()
		{
			ControlsUtils.ExecuteAndMaintainListBoxScroll<LettersModel>(
				( lettersModel ) => { lettersModel.DeleteLetter(); }
				, ApplicationModel.StorehouseModel.LettersModel, LettersListBox );
		}

		/// <summary>
		/// Deletes all the letters with the current flag filter and maintains the (before-after) scroll.
		/// </summary>
		private void DeleteLetters()
		{
			ControlsUtils.ExecuteAndMaintainListBoxScroll<LettersModel>(
				( lettersModel ) => { lettersModel.DeleteLetters(); }
				, ApplicationModel.StorehouseModel.LettersModel, LettersListBox );
		}

		/// <summary>
		/// Visits the URL extracted from the selected letter.
		/// </summary>
		private void VisitLetterUrl()
		{
			TextBox currentLetterMessageSummaryTextBox = GetCurrentLetterMessageSummaryTextBox();
			if( currentLetterMessageSummaryTextBox == null )
				return;

			ApplicationBackendUtils.VisitUrl( currentLetterMessageSummaryTextBox.Text
				, currentLetterMessageSummaryTextBox.CaretIndex );
		}

		/// <summary>
		/// Visits the URL extracted from the selected letter.
		/// </summary>
		/// <param name="clickPoint">The point in the message text-box where the user has clicked.</param>
		private void VisitLetterUrl( Point clickPoint )
		{
			TextBox currentLetterMessageSummaryTextBox = GetCurrentLetterMessageSummaryTextBox();
			if( currentLetterMessageSummaryTextBox == null )
				return;

			int clickedIndex = currentLetterMessageSummaryTextBox.GetCharacterIndexFromPoint( clickPoint, false );
			ApplicationBackendUtils.VisitUrl( currentLetterMessageSummaryTextBox.Text, clickedIndex );
		}

		/// <summary>
		/// Views the image file (attached to a letter) whose image preview was clicked.
		/// </summary>
		/// <param name="imagePreview">The clicked image preview. This may be null.</param>
		/// <returns>Returns true if the image file was opened for viewing.</returns>
		private bool ViewClickedAttachedImageFile( AnimatedImage imagePreview )
		{
			if( (imagePreview == null) || (imagePreview.Tag == null) )
				return false;

			string fileName = imagePreview.Tag.ToString();
			FileUtils.RunFile( fileName );

			return true;
		}

		/// <summary>
		/// Creates the menu items with short commands in the specified menu.
		/// </summary>
		/// <param name="list">List where to add the menu items.</param>
		/// <param name="targetTextBox">The text-box from which to extract the command parameter.</param>
		private void CreateShortCommandMenuItems( CompositeCollection list, TextBox targetTextBox )
		{
			ValidationUtils.ObjectHasContent( list, "list" );

			// Clear the existing menu items.
			foreach( MenuItem menuItem in list )
				menuItem.Click -= ShortCommandMenuItem_Click;
			list.Clear();

			ShortCommands shortCommands = ApplicationModel.Configuration.ShortCommands;
			foreach( ShortCommand shortCommand in shortCommands )
			{
				if( !shortCommand.ShowInMenu )
					continue;

				MenuItem menuItem = new MenuItem();
				menuItem.Header = shortCommands.HasMultipleVisibleDisplayNames( shortCommand.DisplayName )
					? shortCommand.FullDisplayName : shortCommand.DisplayName;
				menuItem.InputGestureText = shortCommand.DisplayShortcutKey;
				menuItem.Tag = new ShortCommandMenuContext( shortCommand, targetTextBox );
				menuItem.Click += ShortCommandMenuItem_Click;

				list.Add( menuItem );
			}
		}

		/// <summary>
		/// Runs the specified short command, using as command parameter the text extracted from the specified text.
		/// </summary>
		/// <param name="shortCommand">Short command to run. This may be null or empty.</param>
		/// <param name="textBox">Text-box from which to extract the command parameter. This may be null.</param>
		/// <param name="useAllText">If this is true then all the text (from the specified message) is used as command parameter.
		/// This has priority over "selectedText".</param>
		/// <remarks>
		/// <para>In the short command, the placeholder <see cref="ShortCommand.RawParameterPlaceholder"/> is replaced with the
		/// raw command parameter, and the placeholder <see cref="ShortCommand.UrlEscapedParameterPlaceholder"/> is replaced
		/// with the URL-escaped command parameter.</para>
		/// </remarks>
		private void RunShortCommand( string shortCommand, TextBox textBox, bool useAllText )
		{
			if( textBox == null )
				return;

			ApplicationBackendUtils.RunShortCommand( shortCommand, textBox.Text, useAllText, textBox.SelectedText
				, textBox.CaretIndex );
		}

		/// <summary>
		/// Copies to the clipboard the message summary of the focused letter.
		/// </summary>
		private void ClipLetterMessageSummary()
		{
			TextBox currentLetterMessageSummaryTextBox = GetCurrentLetterMessageSummaryTextBox();
			if( currentLetterMessageSummaryTextBox == null )
				return;

			string selectedText = currentLetterMessageSummaryTextBox.SelectedText;

			if( TextUtils.HasContent( selectedText ) )
			{
				Clipboard.SetText( selectedText );

				TooltipWindow.ShowTooltip( "The selected message summary of the focused letter has been copied to the clipboard." );
			}
			else
			{
				Clipboard.SetText( currentLetterMessageSummaryTextBox.Text );

				TooltipWindow.ShowTooltip( "The message summary of the focused letter has been copied to the clipboard." );
			}
		}

		/// <summary>
		/// Copies to the clipboard the message form the secondary processing state of the focused letter.
		/// </summary>
		private void ClipLetterSecondaryProcessingStateMessage()
		{
			if( (ApplicationModel.StorehouseModel == null) || (ApplicationModel.StorehouseModel.LettersModel == null) )
				return;

			ApplicationModel.StorehouseModel.LettersModel.ClipCurrentLetterSecondaryProcessingStateMessage();
		}

		/// <summary>
		/// Edits the draft message in a separate window.
		/// </summary>
		private void EditDraftMessageInWindow()
		{
			EditMultiLineNoteWindow window = new EditMultiLineNoteWindow( "Letter", null, DraftMessageTextBox.Text, false
				, TextValidations.LongNote, true, PeerDocumentsConsts.PostMessageSizeMax );
			if( !ControlsUtils.ShowDialog( window ) )
				return;

			DraftMessageTextBox.Text = window.Note;
		}

		/// <summary>
		/// Adds an indicator in the draft message to show what files have been attached to the draft letter.
		/// </summary>
		/// <param name="attachedFileNames">The names of the files which were attached. This may be null or empty.</param>
		/// <remarks>
		/// <para>The indicator is attached only if the draft letter doesn't contain an emoticon symbol which sends a wallpaper,
		/// and doesn't contain any text.</para>
		/// </remarks>
		private void AddDraftMessageIndicatorToShowAttachedFiles( AttachedFileNames attachedFileNames )
		{
			if( ListUtils.NoContent( attachedFileNames )
				|| ApplicationModel.EmoticonManager.ContainsSendWallpaperEmoticonSymbol( DraftMessageTextBox.Text )
				|| TextUtils.HasContent( DraftMessageTextBox.Text ) )
			{
				return;
			}

			string indicator = AttachedFilesUtils.GetDraftMessageIndicatorToShowAttachedFiles( attachedFileNames );

			ControlsUtils.AppendText( DraftMessageTextBox, string.Format( " {0} ", indicator ) );
		}

		/// <summary>
		/// Attaches file names to the draft letter.
		/// </summary>
		/// <returns>Returns the names of the files which were attached. This may be null or empty.</returns>
		private AttachedFileNames DraftAttachFileNames()
		{
			if( !ApplicationModel.StorehouseModel.HasAccountModel )
				return null;

			AttachedFileNames attachedFileNames = ApplicationModel.StorehouseModel.LettersModel.DraftAttachFileNames();
			AddDraftMessageIndicatorToShowAttachedFiles( attachedFileNames );

			return attachedFileNames;
		}

		/// <summary>
		/// Attaches the image from the clipboard to the draft letter.
		/// </summary>
		/// <returns>Returns the names of the files which were attached. This may be null or empty.</returns>
		private AttachedFileNames DraftAttachClipboardImage()
		{
			if( !ApplicationModel.StorehouseModel.HasAccountModel )
				return null;

			AttachedFileNames attachedFileNames = ApplicationModel.StorehouseModel.LettersModel.DraftAttachClipboardImage();
			AddDraftMessageIndicatorToShowAttachedFiles( attachedFileNames );

			return attachedFileNames;
		}

		/// <summary>
		/// Checks if the specified droppable file names may be attached to the draft letter.
		/// </summary>
		/// <param name="dropData">The drop data. This may be null or empty.</param>
		private bool DraftMayAttachDroppableFileNames( IDataObject dropData )
		{
			return ApplicationModel.StorehouseModel.HasAccountModel && (dropData != null)
				&& dropData.GetDataPresent( DataFormats.FileDrop );
		}

		/// <summary>
		/// Attaches the dropped file names to the draft letter.
		/// </summary>
		/// <param name="dropData">The drop data. This may be null or empty.</param>
		/// <returns>Returns the names of the files which were attached. This may be null or empty.</returns>
		private AttachedFileNames DraftAttachDroppedFileNames( IDataObject dropData )
		{
			if( !DraftMayAttachDroppableFileNames( dropData ) )
				return null;

			string[] droppedFiles = (string[])dropData.GetData( DataFormats.FileDrop );

			AttachedFileNames attachedFileNames = ApplicationModel.StorehouseModel.LettersModel.DraftAttachDroppedFileNames(
				droppedFiles );
			AddDraftMessageIndicatorToShowAttachedFiles( attachedFileNames );

			return attachedFileNames;
		}

		/// <summary>
		/// Removes all the files attached to the draft letter.
		/// </summary>
		private void DraftRemoveAttachedFiles()
		{
			if( !ApplicationModel.StorehouseModel.HasAccountModel )
				return;

			ApplicationModel.StorehouseModel.LettersModel.DraftRemoveAttachedFiles();
		}

		/// <summary>
		/// Visits the URL extracted from the draft message.
		/// </summary>
		private void DraftVisitMessageUrl()
		{
			ApplicationBackendUtils.VisitUrl( DraftMessageTextBox.Text, DraftMessageTextBox.CaretIndex );
		}

		/// <summary>
		/// Visits the URL extracted from the draft message.
		/// </summary>
		/// <param name="clickPoint">The point in the message text-box where the user has clicked.</param>
		private void DraftMessageVisitUrl( Point clickPoint )
		{
			int clickedIndex = DraftMessageTextBox.GetCharacterIndexFromPoint( clickPoint, false );
			ApplicationBackendUtils.VisitUrl( DraftMessageTextBox.Text, clickedIndex );
		}

		/// <summary>
		/// Selects an emoticon and inserts it in the draft message.
		/// </summary>
		private void SelectEmoticon()
		{
			SelectEmoticonWindow window = new SelectEmoticonWindow( ApplicationModel.EmoticonManager.Configuration
				, MayShowSensualEmoticons(), ApplicationModel.UpdateWithCustomEmoticons );
			if( !ControlsUtils.ShowDialog( window ) )
				return;

			// Allow users to see the emoticons even if an account model is not active, but don't add the symbol to the current
			// message.
			if( !ApplicationModel.StorehouseModel.HasAccountModel )
				return;

			ControlsUtils.AppendText( DraftMessageTextBox, string.Format( " {0} ", window.SelectedSymbol ) );
			ControlsUtils.FocusTextBox( DraftMessageTextBox, false );

			// Show the letter's emoticon in a tooltip.
			string emoticonFileName = ApplicationModel.EmoticonManager.GetEmoticonFullSizedImageFileName(
				DraftMessageTextBox.Text );
			if( TextUtils.HasContent( emoticonFileName ) )
				TooltipWindow.ShowTooltip( emoticonFileName, TooltipWindowModes.CachedImage );
		}

		/// <summary>
		/// Inserts <see cref="InsertRandomTextSize"/> random lowercase letters in the draft letter, at the caret position.
		/// </summary>
		private void DraftInsertRandomText()
		{
			ControlsUtils.InsertText( DraftMessageTextBox, EntropyPool.GetLowercaseText( InsertRandomTextSize ) );
		}

		/// <summary>
		/// Inserts a new line in the draft letter, at the caret position.
		/// </summary>
		private void DraftInsertNewLine()
		{
			ControlsUtils.InsertText( DraftMessageTextBox, Environment.NewLine );
		}

		/// <summary>
		/// Attaches the (local peer's) custom emoticons to the draft letter.
		/// </summary>
		/// <returns>Returns the names of the files which were attached. This may be null or empty.</returns>
		private AttachedFileNames DraftAttachCustomEmoticons()
		{
			if( !ApplicationModel.StorehouseModel.HasAccountModel )
				return null;

			AttachedFileNames fileNamesToAttach = ApplicationModel.StorehouseModel.LettersModel.DraftAttachCustomEmoticons();
			if( ListUtils.NoContent( fileNamesToAttach ) )
				return null;

			ControlsUtils.AppendText( DraftMessageTextBox
				, string.Format( " {0} ", ImBackendConsts.UpdateWithCustomEmoticonsEmoticon ) );
			ControlsUtils.FocusTextBox( DraftMessageTextBox, false );

			return fileNamesToAttach;
		}

		/// <summary>
		/// Queues an action request which contains the draft letter, for sending to the remote peer of the current account.
		/// </summary>
		private void DraftQueueSendLetter()
		{
			if( !ApplicationModel.StorehouseModel.HasAccountModel )
				return;

			string message = DraftMessageTextBox.Text;

			if( TextUtils.NoContent( message.Trim() ) )
			{
				ControlsUtils.FocusTextBox( DraftMessageTextBox, false );

				return;
			}

			string emoticonSymbol = ApplicationModel.EmoticonManager.GetEmoticonSymbol( message );

			bool queued = ApplicationModel.StorehouseModel.LettersModel.DraftQueueSendLetter( message, emoticonSymbol
				, ApplicationModel.StorehouseModel.AccountModel.Tally.UseTransientInteraction );

			// Clear the message.
			if( queued )
				DraftMessageTextBox.Clear();

			ControlsUtils.FocusTextBox( DraftMessageTextBox, false );
		}

		/// <summary>
		/// Checks if the sensual emoticons may be shown to the user.
		/// </summary>
		/// <returns>Returns true if sensual emoticons may be shown to the user.</returns>
		private bool MayShowSensualEmoticons()
		{
			AccountModel accountModel = ApplicationModel.StorehouseModel.AccountModel;

			return (accountModel != null) && accountModel.MayShowSensualEmoticons();
		}

		#endregion Private methods
	}
}
