﻿////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Microsoft.Win32;
using System;
using System.IO;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using Vela.ActionProcessor;
using Vela.ApplicationBackend;
using Vela.Controls;
using Vela.ImBackend;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.StarkDocuments;

namespace Vela.Im
{
	/// <summary>
	/// Window for editing the application's configuration.
	/// </summary>
	public partial class ApplicationConfigurationWindow : AppViewWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ApplicationConfigurationWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationConfiguration">Application configuration</param>
		public ApplicationConfigurationWindow( ApplicationConfigurationModel applicationConfiguration )
			: base( applicationConfiguration.ApplicationModel )
		{
			ApplicationConfiguration = applicationConfiguration;

			InitializeComponent();

			CreateShortCommandMenuItems();

			LoadUi();
		}

		#endregion Constructors

		#region Public non-serializable properties

		ApplicationConfigurationModel _ApplicationConfiguration;
		/// <summary>
		/// The application configuration. Notifies property observers. READ-ONLY.
		/// </summary>
		public ApplicationConfigurationModel ApplicationConfiguration
		{
			get { return _ApplicationConfiguration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "ApplicationConfiguration" );

				_ApplicationConfiguration = value;

				// Notify property observers.
				OnPropertyChanged( "ApplicationConfiguration" );
				OnPropertyChanged( "ApplicationModel" );
			}
		}

		private CompositeCollection _ShortCommandShortcutKeysMenuItems = new CompositeCollection();
		/// <summary>
		/// The menu items with short command shortcut keys. The backing field is initialized by default. This may be empty, but
		/// not null. READ-ONLY.
		/// </summary>
		public CompositeCollection ShortCommandShortcutKeysMenuItems
		{
			get { return _ShortCommandShortcutKeysMenuItems; }
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void PeerIsOnlineCheckBox_Click( object sender, RoutedEventArgs e )
		{
			UseUpnResolversCheckBox.IsEnabled = PeerIsOnlineCheckBox.IsChecked == true;
			if( !UseUpnResolversCheckBox.IsEnabled )
				UseUpnResolversCheckBox.IsChecked = false;

			ForwardPortsCheckBox.IsEnabled = PeerIsOnlineCheckBox.IsChecked == true;
			if( !ForwardPortsCheckBox.IsEnabled )
				ForwardPortsCheckBox.IsChecked = false;
		}

		private void GenerateProcessingPortButton_Click( object sender, RoutedEventArgs e )
		{
			ProcessingPortTextBox.Text = ApplicationConfiguration.GenerateRandomProcessingPort().ToString();
		}

		private void ChangeRootSaveFolderNameButton_Click( object sender, RoutedEventArgs e )
		{
			ChangeRootSaveFolderName( RootSaveFolderNameTextBox );
		}

		private void AddShortCommandButton_Click( object sender, RoutedEventArgs e )
		{
			ApplicationConfiguration.AddShortCommand();
		}

		private void EditShortCommandButton_Click( object sender, RoutedEventArgs e )
		{
			ApplicationConfiguration.EditShortCommand();
		}

		private void DeleteShortCommandButton_Click( object sender, RoutedEventArgs e )
		{
			ApplicationConfiguration.DeleteShortCommand();
		}

		private void ShortCommandsListBox_PreviewMouseDoubleClick( object sender, System.Windows.Input.MouseButtonEventArgs e )
		{
			ApplicationConfiguration.EditShortCommand();
		}

		private void SetShortCommandShortcutMenuItem_Click( object sender, RoutedEventArgs e )
		{
			char ch = (char)((MenuItem)sender).Tag;
			ApplicationConfiguration.SetShortCommandShortcutKey( ch );
		}

		private void ChangeBellAudioNotificationFileNameButton_Click( object sender, RoutedEventArgs e )
		{
			ChangeAudioNotificationFileName( BellAudioNotificationFileNameTextBox );
		}

		private void PlayBellAudioNotificationButton_Click( object sender, RoutedEventArgs e )
		{
			AudioNotificationManager.Play( BellAudioNotificationFileNameTextBox.Text, true );
		}

		private void ChangeHornAudioNotificationFileNameButton_Click( object sender, RoutedEventArgs e )
		{
			ChangeAudioNotificationFileName( HornAudioNotificationFileNameTextBox );
		}

		private void PlayHornAudioNotificationButton_Click( object sender, RoutedEventArgs e )
		{
			AudioNotificationManager.Play( HornAudioNotificationFileNameTextBox.Text, true );
		}

		private void OkayButton_Click( object sender, RoutedEventArgs e )
		{
			try
			{
				ValidateUi( ErrorPresenter );
				SaveUi( ErrorPresenter );

				DialogResult = true;
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				ControlsUtils.ShowException( ex );
			}
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Loads the user interface with the data to edit.
		/// </summary>
		protected override void LoadUi()
		{
			base.LoadUi();

			for( int i = 1; i <= PeerDocumentsConsts.ShowPreviewsForAttachedImageFilesCountMax; i++ )
				ShowPreviewsForAttachedImageFilesCountComboBox.Items.Add( i );
		}

		/// <summary>
		/// Lets the user select a folder and sets with it the content of the specified text-box.
		/// </summary>
		/// <param name="textBox">Text-box in which to put the selected file name.</param>
		private void ChangeRootSaveFolderName( TextBox textBox )
		{
			ValidationUtils.ObjectHasContent( textBox, "textBox" );

			string initialFileName = textBox.Text;

			OpenFileDialog ofd = new OpenFileDialog();
			ofd.Title = "Select Folder";
			ofd.AddExtension = false;
			ofd.Filter = FileExtensions.AllFilesFilter;
			ofd.InitialDirectory = FileUtils.GetExistingFolderName( initialFileName );
			ofd.CheckPathExists = true;
			ofd.CheckFileExists = false;

			if( !ControlsUtils.ShowDialog( ofd ) )
				return;

			if( TextUtils.HasContent( ofd.FileName ) )
				textBox.Text = Path.GetDirectoryName( ofd.FileName );
		}

		/// <summary>
		/// Creates the menu items with short commands.
		/// </summary>
		private void CreateShortCommandMenuItems()
		{
			// Clear the existing menu items.
			foreach( MenuItem menuItem in ShortCommandShortcutKeysMenuItems )
				menuItem.Click -= SetShortCommandShortcutMenuItem_Click;
			ShortCommandShortcutKeysMenuItems.Clear();

			#region "None" shortcut key menu

			MenuItem noneMenuItem = new MenuItem();
			noneMenuItem.Header = "None";
			noneMenuItem.Tag = ShortCommand.NoShortcutKey;
			noneMenuItem.Click += SetShortCommandShortcutMenuItem_Click;

			ShortCommandShortcutKeysMenuItems.Add( noneMenuItem );

			#endregion "None" shortcut key menu

			for( char i = 'A'; i <= 'Z'; i++ )
			{
				MenuItem menuItem = new MenuItem();
				menuItem.Header = ShortCommand.GetDisplayShortcutKey( i );
				menuItem.Tag = i;
				menuItem.Click += SetShortCommandShortcutMenuItem_Click;

				ShortCommandShortcutKeysMenuItems.Add( menuItem );
			}
		}

		/// <summary>
		/// Lets the user select a sound file and sets with it the content of the specified text-box.
		/// </summary>
		/// <param name="textBox">Text-box in which to put the selected file name.</param>
		private void ChangeAudioNotificationFileName( TextBox textBox )
		{
			ValidationUtils.ObjectHasContent( textBox, "textBox" );

			string initialFileName = textBox.Text;

			OpenFileDialog ofd = new OpenFileDialog();
			ofd.Title = "Select Sound File";
			ofd.AddExtension = true;
			ofd.Filter = FileExtensions.AudioFilter;
			ofd.InitialDirectory = FileUtils.GetExistingFolderName( initialFileName );
			if( TextUtils.HasContent( initialFileName ) )
				ofd.FileName = Path.GetFileNameWithoutExtension( initialFileName );
			ofd.CheckFileExists = true;

			if( !ControlsUtils.ShowDialog( ofd ) )
				return;

			if( TextUtils.HasContent( ofd.FileName ) )
				textBox.Text = ofd.FileName;
		}

		/// <summary>
		/// Validates the input data.
		/// </summary>
		/// <param name="errorPresenter">The control which shows that there is a validation error. This may be null.</param>
		/// <remarks>
		/// <para>If you don't validate of all the editable fields in here, it's possible that some bindings are commited even
		/// if the user cancels the edit operation. This can happen if the user edits some fields, clicks the "Okay" button, the
		/// validation fails (during binding commit) inside some properties (which are not validated in here), and then the user
		/// clicks the "Cancel" button. This is likely to matter only if those (successfully) bound properties are used outside
		/// the view.</para>
		/// </remarks>
		protected override void ValidateUi( FrameworkElement errorPresenter )
		{
			base.ValidateUi( errorPresenter );

			ValidationUtils.Integer( ProcessingPortTextBox.Text, 1, UInt16.MaxValue, "Processing port"
				, ProcessingPortTextBox, errorPresenter );
			ValidationUtils.Text( RootSaveFolderNameTextBox.Text, -1, -1, TextValidations.FileName, "Root save folder name"
				, RootSaveFolderNameTextBox, errorPresenter );
			ValidationUtils.Text( BellAudioNotificationFileNameTextBox.Text, 0, -1, TextValidations.FileName
				, "Bell audio notification file", BellAudioNotificationFileNameTextBox, errorPresenter );
			ValidationUtils.Text( HornAudioNotificationFileNameTextBox.Text, 0, -1, TextValidations.FileName
				, "Horn audio notification file", HornAudioNotificationFileNameTextBox, errorPresenter );
			ValidationUtils.Integer( LargeCommunicationCacheSizeTextBox.Text, 0
				, ActionProcessorConsts.LargeCommunicationCacheSizeMax, "Large communication cache size"
				, LargeCommunicationCacheSizeTextBox, errorPresenter );
			ValidationUtils.Integer( FillerUpnClassifiersCountTextBox.Text
				, FillerUpnClassifiersConfiguration.ClassifiersCountMin
				, FillerUpnClassifiersConfiguration.ClassifiersCountMax, "Filler UPN classifiers count"
				, FillerUpnClassifiersCountTextBox, errorPresenter );
		}

		#endregion Private methods
	}
}
