////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.ComponentModel;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Threading;
using Vela.ApplicationBackend;
using Vela.Controls;
using Vela.ImBackend;
using Vela.Kernel;

namespace Vela.Im
{
	/// <summary>
	/// The main window of the application.
	/// </summary>
	public partial class MainWindow : AppViewWindow, IFloatingWindow
	{
		#region Public constants

		/// <summary>
		/// The timespan (in seconds) for the main timer.
		/// </summary>
		public const int MainTimerTimespan = 2;

		/// <summary>
		/// The timespan (in seconds) for the storehouse timer.
		/// </summary>
		public const double StorehouseTimerTimespan = 0.05;

		/// <summary>
		/// The timespan (in seconds) for the notification removal timer.
		/// </summary>
		public const double RemoveNotificationTimerTimespan = 0.1;

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected MainWindow()
		{
			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="applicationModel">Application model.</param>
		public MainWindow( ApplicationModel applicationModel )
			: base( applicationModel )
		{
			InitializeComponent();

			LoadUi();

			ApplicationModel.Configuration.MainWindowLayoutConfiguration = WindowLayoutConfiguration.LoadUi( this
				, ApplicationModel.Configuration.MainWindowLayoutConfiguration );
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The main timer.
		/// </summary>
		private DispatcherTimer MainTimer;

		/// <summary>
		/// The timer from which to create / open a storehouse, when the program starts.
		/// </summary>
		private DispatcherTimer StorehouseTimer;

		/// <summary>
		/// The timer from which to remove the current application notification.
		/// </summary>
		private DispatcherTimer RemoveNotificationTimer;

		#endregion Private fields

		#region Public non-serializable properties

		/// <summary>
		/// Is visible if the program was compiled in debug mode, and should not be used in production environments. READ-ONLY.
		/// </summary>
		public Visibility DebugModeVisibility
		{
			get
			{
#if DEBUG
				return Visibility.Visible;
#else
				return Visibility.Collapsed;
#endif
			}
		}

		/// <summary>
		/// Is visible if the program was compiled so that it exposes private information in serialized objects, and must not be
		/// used in production environments. READ-ONLY.
		/// </summary>
		public Visibility SerializeExposedAttributeVisibility
		{
			get
			{
#if DEBUG && SERIALIZEEXPOSEDATTRIBUTE
				return Visibility.Visible;
#else
				return Visibility.Collapsed;
#endif
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void AppViewWindow_Loaded( object sender, RoutedEventArgs e )
		{
			ApplicationModel.NotifyOnStart();

			MainTimer = new DispatcherTimer();
			MainTimer.Interval = TimeSpan.FromSeconds( MainTimerTimespan );
			MainTimer.Tick += new EventHandler( MainTimer_Tick );
			MainTimer.IsEnabled = true;

			StorehouseTimer = new DispatcherTimer();
			StorehouseTimer.Interval = TimeSpan.FromSeconds( StorehouseTimerTimespan );
			StorehouseTimer.Tick += new EventHandler( StorehouseTimer_Tick );
			StorehouseTimer.IsEnabled = true;

			RemoveNotificationTimer = new DispatcherTimer();
			RemoveNotificationTimer.IsEnabled = false;
			RemoveNotificationTimer.Interval = TimeSpan.FromSeconds( RemoveNotificationTimerTimespan );
			RemoveNotificationTimer.Tick += new EventHandler( RemoveNotificationTimer_Tick );

			MainPanel.IsEnabled = true;

			ControlsUtils.FocusListBoxItem( FunctionViewsListBox );
		}

		private void AppViewWindow_Closing( object sender, CancelEventArgs e )
		{
			// When the OS shuts down, it looks like the application ends in the "App.Run" method, not in the "App.Main", so
			// everything after "App.Run" is not executed.

			if( ApplicationModel == null )
				return;

			try
			{
				App.CloseMode = ApplicationModel.CloseOrMinimize( App.CloseMode
					, ImConsts.TimePassedFromLastReceivedLetterTimeToClearIt );

				if( App.CloseMode == ApplicationCloseModes.Minimize )
					e.Cancel = true;
			}
			catch( Exception ex )
			{
				// The cleanup operations may have thrown an exception, so don't try to do it again.
				if( App.CloseMode == ApplicationCloseModes.CleanUp )
					App.CloseMode = ApplicationCloseModes.Force;

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		private void AppViewWindow_Activated( object sender, EventArgs e )
		{
			if( RemoveNotificationTimer != null )
				RemoveNotificationTimer.IsEnabled = true;
		}

		private void AppViewWindow_Deactivated( object sender, EventArgs e )
		{
			if( RemoveNotificationTimer != null )
				RemoveNotificationTimer.IsEnabled = false;
		}

		private void AppViewWindow_StateChanged( object sender, EventArgs e )
		{
			ApplicationVisibilities applicationVisibility = ControlsUtils.GetApplicationVisibility();

			if( RemoveNotificationTimer != null )
				RemoveNotificationTimer.IsEnabled = applicationVisibility == ApplicationVisibilities.Normal;

			if( applicationVisibility != ApplicationVisibilities.Normal )
				TooltipWindow.CloseTooltip();
		}

		private void AppViewWindow_PreviewMouseMove( object sender, MouseEventArgs e )
		{
			ControlsUtils.AddMouseEntropy( e );

			AudioNotificationManager.Stop( false );
		}

		private void AppViewWindow_PreviewMouseDown( object sender, MouseButtonEventArgs e )
		{
			ControlsUtils.AddMouseEntropy( e );

			AudioNotificationManager.Stop( true );
		}

		private void AppViewWindow_PreviewKeyDown( object sender, KeyEventArgs e )
		{
			if( Keyboard.Modifiers == ModifierKeys.None )
			{
				// Hide the application. If this were handled in a key-up event, it would also be executed when a child
				// window were closed after "Escape" is pressed.
				if( e.Key == Key.Escape )
				{
					if( (ApplicationModel.StorehouseModel != null) && (ApplicationModel.StorehouseModel.LettersModel != null) )
					{
						ApplicationModel.StorehouseModel.LettersModel.ClearLastReceivedLetterTime(
							ImConsts.TimePassedFromLastReceivedLetterTimeToClearIt );
					}

					ControlsUtils.SetApplicationVisibility( ApplicationVisibilities.Hidden );

					e.Handled = true;

					return;
				}
			}

			AudioNotificationManager.Stop( true );
		}

		private void MainTimer_Tick( object sender, EventArgs e )
		{
			ApplicationModel.SetNotification();
			ApplicationModel.EnsureNewLetterNotificationInTaskbar();
		}

		private void StorehouseTimer_Tick( object sender, EventArgs e )
		{
			StorehouseTimer.IsEnabled = false;

			ImUtils.OpenOrCreateStorehouse( ApplicationModel );
		}

		private void RemoveNotificationTimer_Tick( object sender, EventArgs e )
		{
			if( !IsActive || (ControlsUtils.GetApplicationVisibility() != ApplicationVisibilities.Normal)
				|| (ApplicationModel.StorehouseModel == null) || (ApplicationModel.StorehouseModel.LettersModel == null) )
			{
				RemoveNotificationTimer.IsEnabled = false;

				return;
			}

			bool timeCleared = ApplicationModel.StorehouseModel.LettersModel.ClearLastReceivedLetterTime(
				ImConsts.TimePassedFromLastReceivedLetterTimeToClearIt );

			if( timeCleared )
				RemoveNotificationTimer.IsEnabled = false;
		}

		private void FunctionViewsListBox_SelectionChanged( object sender, SelectionChangedEventArgs e )
		{
			ControlsUtils.FocusListBoxItem( FunctionViewsListBox );
		}

		#endregion Event handlers
	}
}
