﻿////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows;
using Vela.ActionProcessor;
using Vela.Controls;
using Vela.ImDocuments;
using Vela.Kernel;

namespace Vela.Im
{
	/// <summary>
	/// Window for editing a storehouse's configuration.
	/// </summary>
	public partial class StorehouseConfigurationWindow : ViewWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected StorehouseConfigurationWindow()
		{
			InitializeComponent();

			LoadUi();
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="configuration">Storehouse configuration.</param>
		public StorehouseConfigurationWindow( ImStorehouseConfiguration configuration )
		{
			Configuration = configuration;

			InitializeComponent();

			LoadUi();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The original value for <see cref="ProcessingPort"/>.
		/// </summary>
		private UInt16 OriginalProcessingPort;

		#endregion Private fields

		#region Public non-serializable properties

		private ImStorehouseConfiguration _Configuration;
		/// <summary>
		/// Storehouse configuration. Notifies property observers. READ-ONLY.
		/// </summary>
		public ImStorehouseConfiguration Configuration
		{
			get { return _Configuration; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Configuration" );

				_Configuration = value;

				ActionServiceListenerConfiguration listenerConfiguration = Configuration.ActionServiceListeners.GetMostRecent();
				OriginalProcessingPort = listenerConfiguration.ProcessingPort;
				ProcessingPort = OriginalProcessingPort;

				// Notify property observers.
				OnPropertyChanged( "Configuration" );
			}
		}

		private UInt16 _ProcessingPort;
		/// <summary>
		/// The network port of a service where action requests are received from clients. If the changes from the current window
		/// are confirmed, the caller of the window has to create a new action service configuration (in the storehouse), or
		/// update the existing one (and also make it the most recent configuration). Notifies property observers.
		/// </summary>
		/// <remarks>
		/// <para>Must be writable for binding.</para>
		/// </remarks>
		public UInt16 ProcessingPort
		{
			get { return _ProcessingPort; }
			set
			{
				ValidationUtils.Integer( value, 1, UInt16.MaxValue, "ProcessingPort", null, null );

				_ProcessingPort = value;

				// Notify property observers.
				OnPropertyChanged( "ProcessingPort" );
			}
		}

		private bool _RemoveOlderProcessingPorts;
		/// <summary>
		/// If the changes from the current window are confirmed and this is true, the caller of the window has to remove all the
		/// processing ports other than <see cref="ProcessingPort"/> from the storehouse configuration. Notifies property
		/// observers. READ-ONLY.
		/// </summary>
		public bool RemoveOlderProcessingPorts
		{
			get { return _RemoveOlderProcessingPorts; }
			protected set
			{
				_RemoveOlderProcessingPorts = value;

				// Notify property observers.
				OnPropertyChanged( "RemoveOlderProcessingPorts" );
			}
		}

		/// <summary>
		/// Is true if <see cref="ProcessingPort"/> was modified (from its original value), or if
		/// <see cref="RemoveOlderProcessingPorts"/> is true.
		/// </summary>
		public bool ProcessingPortWasModified
		{
			get { return (OriginalProcessingPort != ProcessingPort) || RemoveOlderProcessingPorts; }
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void ViewWindow_Loaded( object sender, RoutedEventArgs e )
		{
		}

		private void GenerateProcessingPortButton_Click( object sender, RoutedEventArgs e )
		{
			UInt16 port = ActionProcessorUtils.GetRandomServiceProcessingPort( Configuration.ActionServiceListeners );
			ProcessingPortTextBox.Text = port.ToString();
		}

		private void RemoveOlderProcessingPortsButton_Click( object sender, RoutedEventArgs e )
		{
			string message = "Do you want to remove all the processing ports other than the displayed one, from the"
				+ " configuration of the storehouse?"
				+ "\n\n"
				+ "If you remove them, action requests will no longer be received on the removed ports (unless they are part"
				+ " of your vaults).";

			if( MessageWindow.ShowDialog( message, MessageWindowModes.Query ) != QueryMessageResults.Yes )
				return;

			RemoveOlderProcessingPorts = true;
		}

		private void OkayButton_Click( object sender, RoutedEventArgs e )
		{
			try
			{
				ValidateUi( ErrorPresenter );
				SaveUi( ErrorPresenter );

				DialogResult = true;
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				ControlsUtils.ShowException( ex );
			}
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Validates the input data.
		/// </summary>
		/// <param name="errorPresenter">The control which shows that there is a validation error. This may be null.</param>
		/// <remarks>
		/// <para>If you don't validate of all the editable fields in here, it's possible that some bindings are commited even
		/// if the user cancels the edit operation. This can happen if the user edits some fields, clicks the "Okay" button, the
		/// validation fails (during binding commit) inside some properties (which are not validated in here), and then the user
		/// clicks the "Cancel" button. This is likely to matter only if those (successfully) bound properties are used outside
		/// the view.</para>
		/// </remarks>
		protected override void ValidateUi( FrameworkElement errorPresenter )
		{
			base.ValidateUi( errorPresenter );

			ValidationUtils.Integer( ProcessingPortTextBox.Text, 1, UInt16.MaxValue, "Processing port"
				, ProcessingPortTextBox, errorPresenter );
		}

		#endregion Private methods
	}
}
