////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.ActionProcessor;
using Vela.ApplicationBackend;
using Vela.DatabaseStorehouse;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.PaymentServiceOps
{
	/// <summary>
	/// Application configuration.
	/// </summary>
	public class AppConfiguration : StorableConfiguration
	{
		#region Public constants

		/// <summary>
		/// The title of the file where the configuration is stored.
		/// </summary>
		public const string FileTitle = "App" + FileExtensions.StorableConfiguration;

		/// <summary>
		/// The default value for <see cref="LogDebugMessages"/>.
		/// </summary>
		/// <remarks>
		/// <para>Setting this with true may have dramatic performance implications because it activates writing the log
		/// messages to the system trace, and the usage of the <see cref="LogManager.Added"/> event (event which in turn may
		/// update the UI through the "Dispatcher.Invoke" method).</para>
		/// </remarks>
		public const bool LogDebugMessagesDef = true;

		/// <summary>
		/// The default value for <see cref="SaveDebugMessages"/>.
		/// </summary>
		public const bool SaveDebugMessagesDef = false;

		/// <summary>
		/// The type name of the storehouse provider to instantiate.
		/// </summary>
		/// <remarks>
		/// <para>For SQL Server CE, use "Vela.MsSqlCeDatabase.MsSqlCeDatabaseProvider".</para>
		/// 
		/// <para>For SQL Server FE, use "Vela.MsSqlFeDatabase.MsSqlFeDatabaseProvider".</para>
		/// </remarks>
		public const string StorehouseProviderTypeName = "Vela.MsSqlCeDatabase.MsSqlCeDatabaseProvider";

		/// <summary>
		/// Template for the database connection string. Placeholders:
		/// <see cref="StorehouseConsts.StorehouseFileNamePlaceholder"/> - name of file attached to the storehouse,
		/// <see cref="StorehouseConsts.ServerStorehouseNamePlaceholder"/> - server storehouse name,
		/// <see cref="StorehouseConsts.StorehouseAttachmentsPasswordPlaceholder"/> - storehouse attachments password.
		/// </summary>
		/// <remarks>
		/// <para>For SQL Server CE, use "Data source='{StorehouseFileName}'; Max database size=4091; File mode=Read Write;
		/// Encryption mode=Engine Default; Persist security info=False; Password={StorehouseAttachmentsPassword}; Enlist=False;
		/// Default lock timeout=5000; Autoshrink threshold=100;".</para>
		/// 
		/// <para>For SQL Server FE, use "Server=(local); Database={ServerStorehouseName}; Persist security info=False;
		/// User Id={ServerStorehouseName}; Password={StorehouseAttachmentsPassword}; Enlist=False;".</para>
		/// </remarks>
		public const string DatabaseConnectionTemplate = "Data source='{StorehouseFileName}'; Max database size=4091;"
			+ " File mode=Read Write; Encryption mode=Engine Default; Persist security info=False;"
			+ " Password={StorehouseAttachmentsPassword}; Enlist=False; Default lock timeout=5000; Autoshrink threshold=100;";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public AppConfiguration()
		{
		}

		#endregion Constructors

		#region Public serializable properties

		private bool _LogDebugMessages = LogDebugMessagesDef;
		/// <summary>
		/// <see cref="LogManager.LogDebugMessages"/> is set with this.
		/// </summary>
		[SerializeAttribute]
		public bool LogDebugMessages
		{
			get { return _LogDebugMessages; }
			set { _LogDebugMessages = value; }
		}

		private bool _SaveDebugMessages = SaveDebugMessagesDef;
		/// <summary>
		/// <see cref="LogManager.SaveDebugMessages"/> is set with this.
		/// </summary>
		[SerializeAttribute]
		public bool SaveDebugMessages
		{
			get { return _SaveDebugMessages; }
			set { _SaveDebugMessages = value; }
		}

		private string _LogMessagesWhitelist;
		/// <summary>
		/// <see cref="LogManager.LogMessagesWhitelist"/> is set with this. This may be null or empty.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string LogMessagesWhitelist
		{
			get { return _LogMessagesWhitelist; }
			set { _LogMessagesWhitelist = value; }
		}

		private string _LogMessagesBlacklist;
		/// <summary>
		/// <see cref="LogManager.LogMessagesBlacklist"/> is set with this. This may be null or empty.
		/// </summary>
		[SerializeAttribute( MustBeDeserialized = false )]
		public string LogMessagesBlacklist
		{
			get { return _LogMessagesBlacklist; }
			set { _LogMessagesBlacklist = value; }
		}

		private IStorehouseAccess _StorehouseAccess;
		/// <summary>
		/// The storehouse access.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		[SerializeAttribute]
		public IStorehouseAccess StorehouseAccess
		{
			get
			{
				if( _StorehouseAccess != null )
					return _StorehouseAccess;

				_StorehouseAccess = new DatabaseAccess( StorehouseProviderTypeName, DatabaseConnectionTemplate );

				return _StorehouseAccess;
			}
			set
			{
				ValidationUtils.ObjectHasContent( value, "StorehouseAccess" );

				_StorehouseAccess = value;
			}
		}

		private ActionProcessorConfiguration _ActionProcessor;
		/// <summary>		
		/// The action processor configuration.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		[SerializeAttribute]
		public ActionProcessorConfiguration ActionProcessor
		{
			get
			{
				if( _ActionProcessor != null )
					return _ActionProcessor;

				_ActionProcessor = AlgorithmPool.Instance.GetActionProcessorConfiguration();

				return _ActionProcessor;
			}
			set
			{
				ValidationUtils.ObjectHasContent( value, "ActionProcessor" );

				_ActionProcessor = value;
			}
		}

		private ActionServiceListenerConfiguration _ActionServiceListenerTemplate;
		/// <summary>
		/// Template for an action service listener configuration.
		/// </summary>
		/// <remarks>
		/// <para>This property uses lazy loading.</para>
		/// </remarks>
		[SerializeAttribute]
		public ActionServiceListenerConfiguration ActionServiceListenerTemplate
		{
			get
			{
				if( _ActionServiceListenerTemplate != null )
					return _ActionServiceListenerTemplate;

				_ActionServiceListenerTemplate = AlgorithmPool.Instance.GetActionServiceListenerConfiguration();

				return _ActionServiceListenerTemplate;
			}
			set
			{
				ValidationUtils.ObjectHasContent( value, "ActionServiceListenerTemplate" );

				_ActionServiceListenerTemplate = value;
			}
		}

		#endregion Public serializable properties

		#region Public static methods

		/// <summary>
		/// Loads the configuration.
		/// </summary>
		/// <returns>Returns the loaded configuration object. If the configuration file doesn't exists then a new configuration
		/// object is returned.</returns>
		/// <remarks>
		/// <para>This method implements a recovery mechanism following a crash during a call of the <see cref="Save"/>
		/// method.</para>
		/// </remarks>
		public static AppConfiguration Load()
		{
			string fileName = AssemblyUtils.GetApplicationDataFileName( null, FileTitle );

			IDeserializer deserializer;

			AppConfiguration configuration = (AppConfiguration)Load( fileName, typeof( AppConfiguration ), out deserializer );

			LogManager.LogDebugMessages = configuration.LogDebugMessages;
			LogManager.SaveDebugMessages = configuration.SaveDebugMessages;
			LogManager.LogMessagesWhitelist = configuration.LogMessagesWhitelist;
			LogManager.LogMessagesBlacklist = configuration.LogMessagesBlacklist;

			return configuration;
		}

		#endregion Public static methods

		#region Public methods

		/// <summary>
		/// Saves the configuration to the associated file.
		/// </summary>
		/// <remarks>
		/// <para>The saving is crash-safe. If the application crashes when the file is being saved, at least 1 of the
		/// following files is still available: a file with the specified name, a file with a backup name, a file with a
		/// temporary name. If the backup file exists, it's the valid one and you should implement your own recovery mechanism
		/// following a crash.</para>
		/// </remarks>
		public void Save()
		{
			Save( AlgorithmPool.Instance.GetXmlSerializer() );
		}

		#endregion Public methods
	}
}
