////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Windows.Threading;
using Vela.Kernel;
using Vela.ActionProcessor;

namespace Vela.PaymentServiceOps
{
	/// <summary>
	/// Contract for an action processor host.
	/// </summary>
	public abstract class ActionProcessorHost : IActionProcessorHost
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected ActionProcessorHost()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="mainWindow">The application's main window.</param>
		public ActionProcessorHost( MainWindow mainWindow )
		{
			MainWindow = mainWindow;
		}

		#endregion Constructors

		#region Public non-serializable properties

		private MainWindow _MainWindow;
		/// <summary>
		/// The application's main window. READ-ONLY.
		/// </summary>
		public MainWindow MainWindow
		{
			get { return _MainWindow; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "MainWindow" );

				_MainWindow = value;
			}
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Checks if the action processor is under maintenance.
		/// </summary>
		/// <remarks>
		/// <para>This is already running on the dispatcher of the application model.</para>
		/// </remarks>
		private bool CheckMaintenanceOnUiThread()
		{
			try
			{
				return MainWindow.CheckMaintenance();
			}
			catch( Exception ex )
			{
				throw new BackgroundOperationException( ex );
			}
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Checks if the action processor is under maintenance.
		/// </summary>
		public bool CheckMaintenance()
		{
			// Process on the UI thread.
			object isMaintenance = MainWindow.Dispatcher.Invoke(
				new Func<bool>( delegate { return CheckMaintenanceOnUiThread(); } ), DispatcherPriority.Background );

			return (isMaintenance != null) ? (bool)isMaintenance : false;
		}

		#endregion Public methods
	}
}
