////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.ComponentModel;
using System.IO;
using System.Windows;
using System.Windows.Media;
using Vela.ActionProcessor;
using Vela.ActionService;
using Vela.ApplicationBackend;
using Vela.Controls;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.DatabaseStorehouse;
using Vela.Kernel;
using Vela.PaymentDocuments;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.PaymentServiceOps
{
	/// <summary>
	/// The main window of the application.
	/// </summary>
	public partial class MainWindow : ViewWindow
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public MainWindow()
		{
			InitializeComponent();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The synchronization object for the <see cref="IsMaintenance"/> property.
		/// </summary>
		private readonly object IsMaintenanceLock = new object();

		private bool _IsMaintenance = false;
		/// <summary>
		/// If this is true then the service is in maintenance mode.
		/// </summary>
		/// <remarks>
		/// <para>This property is thread safe.</para>
		/// </remarks>
		private bool IsMaintenance
		{
			get
			{
				lock( IsMaintenanceLock )
				{
					return _IsMaintenance;
				}
			}
			set
			{
				lock( IsMaintenanceLock )
				{
					_IsMaintenance = value;
				}
			}
		}

		/// <summary>
		/// The storehouse provider.
		/// </summary>
		private IStorehouseProvider StorehouseProvider;

		/// <summary>
		/// The action service manager.
		/// </summary>
		private ActionServiceManager ActionServiceManager;

		/// <summary>
		/// Action client host.
		/// </summary>
		private ActionClientHost ActionClientHost;

		/// <summary>
		/// Action service host. READ-ONLY.
		/// </summary>
		private ActionServiceHost ActionServiceHost;

		#endregion Private fields

		#region Event handlers

		private void ViewWindow_Loaded( object sender, RoutedEventArgs e )
		{
			ActionClientHost = new ActionClientHost( this );
			ActionServiceHost = new ActionServiceHost( this );

			SetDisplayInfo();
			SetState( "Not started", false );

			EntropyPool.SetContainsRealEntropy();

			// Save the application configuration.
			App.Configuration.Save();
		}

		private void ViewWindow_Closing( object sender, CancelEventArgs e )
		{
			string message = "Are you sure that you want to shutdown the service?";
			if( MessageWindow.ShowDialog( message, MessageWindowModes.Query ) == QueryMessageResults.Yes )
			{
				try
				{
					StopService();
				}
				finally
				{
					// Save the entropy.
					if( StorehouseProvider != null )
						StorehouseUtils.UpdateEntropy( StorehouseProvider );
				}
			}
			else
			{
				e.Cancel = true;
			}
		}

		private void ImportServiceVaultButton_Click( object sender, RoutedEventArgs e )
		{
			ImportServiceVault();
		}

		/// <summary>
		/// Enters the service in maintenance mode.
		/// </summary>
		/// <param name="sender">The raiser of the event.</param>
		/// <param name="e">Event parameters.</param>
		private void EnterMaintenanceButton_Click( object sender, RoutedEventArgs e )
		{
			if( ActionServiceManager == null )
				return;

			bool isMaintenance = IsMaintenance;

			#region Ask the user

			// This is not put in the "lock" statement because we don't want the query dialog to block the service thread. This
			// is possible because the service itself never sets the maintenance mode, and so the mode can't be modified while
			// the dialog is shown.

			string message = string.Format( "Are you sure that you want the service to {0} maintenance mode?"
				, isMaintenance ? "EXIT" : "ENTER" );
			if( MessageWindow.ShowDialog( message, MessageWindowModes.Query ) != QueryMessageResults.Yes )
				return;

			#endregion Ask the user

			// Toggle the maintenance mode.
			IsMaintenance = !isMaintenance;

			MaintenanceModeButton.Content = IsMaintenance ? "Exit maintenance" : "Enter maintenance";
		}

		#endregion Event handlers

		#region Private methods

		/// <summary>
		/// Disables the user interaction with the application.
		/// </summary>
		private void DisableUserInput()
		{
			OpenServiceVaultButton.IsEnabled = false;
		}

		/// <summary>
		/// Sets the display information of the service.
		/// </summary>
		private void SetDisplayInfo()
		{
			if( ActionServiceManager != null )
			{
				// Show the properties of the first service vault.

				Sheafs vaults = ApplicationBackendStorehouse.CollectVaults( StorehouseProvider );
				PaymentVault vault = vaults.GetFirstItem() as PaymentVault;
				ValidationUtils.ObjectType<PaymentVault>( vault, "vault" );

				ProcessingIdentityUpnLabel.Text = vault.ProcessingIdentity.Upn.ToString();
			}
			else
			{
				ProcessingIdentityUpnLabel.Text = "";
			}
		}

		/// <summary>
		/// Starts the service.
		/// </summary>
		public void StartService()
		{
			Strings actionServiceTypeNames = new Strings();
			actionServiceTypeNames.Add( typeof( PaymentService.PaymentService ).FullName );

			ActionServiceManager = new ActionServiceManager( StorehouseProvider, actionServiceTypeNames
				, AlgorithmPool.Instance.GetActionProcessorConfiguration(), ActionServiceHost, new DummyActionClientManager() );
		}

		/// <summary>
		/// Stops the service.
		/// </summary>
		public void StopService()
		{
			if( ActionServiceManager != null )
				ActionServiceManager.Stop();

			ActionServiceManager = null;
		}

		/// <summary>
		/// Imports the service vault from a file, and starts the service.
		/// </summary>
		private void ImportServiceVault()
		{
			/*
			// Read password.
			PasswordWindow passwordWindow = new PasswordWindow();
			ControlsUtils.SetWindowOwner( passwordWindow );
			if( passwordWindow.ShowDialog() != true) return;
			string serviceVaultPassword = passwordWindow.Password;

			// Add new entropy.
			EntropyWindow entropyWindow = new EntropyWindow();
			ControlsUtils.SetWindowOwner( entropyWindow );
			if( entropyWindow.ShowDialog() != true) return;
			*/
			//? Not during development.

			#region Create the storehouse provider

			//? For testing only.

			string storehousePassword = "1234567890";

			string workPath = AssemblyUtils.GetApplicationDataFolderName( "Vela.Tester", "Documents" );
			string storehouseFileName = Path.Combine( workPath, "ServiceStorehouse.vstore" );

			// Create the storehouse provider.
			ISerializer serializer = AlgorithmPool.Instance.GetSerializer();
			DatabaseAccess databaseAccess = App.Configuration.StorehouseAccess as DatabaseAccess;
			StorehouseProvider = StorehouseUtils.OpenStorehouse( databaseAccess.StorehouseProviderTypeName
				, databaseAccess, storehouseFileName, storehousePassword, serializer ) as IStorehouseProvider;

			// Restore the entropy state.
			StorehouseUtils.LoadEntropy( StorehouseProvider );

			#endregion Create the storehouse provider

			// If there is an error after this point, something went terribly wrong, so block the main functions.
			DisableUserInput();

			try
			{
				// Start listening for client connections.
				StartService();

				// Allow the administrator to enter maintenance mode.
				MaintenanceModeButton.IsEnabled = true;

				SetDisplayInfo();
				SetState( "Running", false );
			}
			catch( Exception ex )
			{
				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		#endregion Private methods

		#region Public methods

		/// <summary>
		/// Checks if the action processor is under maintenance.
		/// </summary>
		public bool CheckMaintenance()
		{
			return IsMaintenance;
		}

		/// <summary>
		/// Sets the state of the service.
		/// </summary>
		/// <param name="message">State message to show.</param>
		/// <param name="isError">If this is true then the message to show is an error.</param>
		public void SetState( string message, bool isError )
		{
			ServiceStateLabel.Foreground = isError ? Brushes.Tomato : Brushes.LightGreen;
			ServiceStateLabel.Text = message;
		}

		#endregion Public methods
	}
}
