////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.ComponentModel;
using System.Windows;
using System.Windows.Input;
using Vela.Controls;
using Vela.Crypto;
using Vela.Kernel;
using Vela.TesterBackend;

namespace Vela.Tester
{
	/// <summary>
	/// The main window of the application.
	/// </summary>
	public partial class MainWindow : ViewWindow
	{
		#region Public constants

		/// <summary>
		/// The time format for the long operation progress.
		/// </summary>
		public const string LopTimeFormat = "0.#";

		/// <summary>
		/// The message template for the long operation progress.
		/// </summary>
		public const string LopMessageTemplate = "{0} | {1} seconds elapsed | {2}";

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		public MainWindow()
		{
			InitializeComponent();
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// The application's configuration.
		/// </summary>
		private AppConfiguration AppConfig;

		private AppTestContext AppContext;

		#endregion Private fields

		#region Public non-serializable properties

		private Tests _Tests;
		/// <summary>
		/// The available tests. Notifies property observers. READ-ONLY.
		/// </summary>
		public Tests Tests
		{
			get { return _Tests; }
			protected set
			{
				ValidationUtils.ObjectHasContent( value, "Tests" );

				_Tests = value;

				// Notify property observers.
				OnPropertyChanged( "Tests" );
			}
		}

		#endregion Public non-serializable properties

		#region Event handlers

		private void ViewWindow_Loaded( object sender, RoutedEventArgs e )
		{
			try
			{
				Tests = Tests.Collect( typeof( AppTestContext ).Assembly );

				EntropyPool.SetContainsRealEntropy();
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				ErrorPresenter( ex );

				Close();
			}

			ControlsUtils.FocusListBoxItem( TestsListBox );

			AppConfig = AppConfiguration.Load();

			AppContext = new AppTestContext( MessagePresenter, ErrorPresenter, Progress, AppConfig );
			AppContext.SaveConfiguration();
		}

		private void ViewWindow_Closing( object sender, CancelEventArgs e )
		{
			if( AppContext != null )
			{
				AppContext.IsAbortGeneration = true;
				AppContext.SaveConfiguration();
			}
		}

		private void ShowDocumentsFolderMenuItem_Click( object sender, RoutedEventArgs e )
		{
			FileUtils.ExploreFile( AppContext.AccountAliceFileName );
		}

		private void OpenErrorLogMenuItem_Click( object sender, RoutedEventArgs e )
		{
			FileUtils.RunFile( LogManager.CurrentLogFileName );
		}

		private void MillerRabinProbabilityMenuItem_Click( object sender, RoutedEventArgs e )
		{
			MessageWindow.ShowDialog( AppContext.GetMillerRabinProbability(), MessageWindowModes.Info );
		}

		private void GenerateSmallPrimeMenu_Click( object sender, RoutedEventArgs e )
		{
			AppContext.GenerateSmallPrime();
		}

		private void GenerateEntropyFileMenuItem_Click( object sender, RoutedEventArgs e )
		{
			EntropyWindow entropyWindow = new EntropyWindow();
			if( ControlsUtils.ShowDialog( entropyWindow ) )
				AppContext.GenerateEntropyFile();
		}

		private void DoAllTestMenuItem_Click( object sender, RoutedEventArgs e )
		{
			Lock();

			ExecuteTests();

			Unlock( false );
		}

		private void AboutMenuItem_Click( object sender, RoutedEventArgs e )
		{
			AboutWindow window = new AboutWindow( null );
			ControlsUtils.ShowDialog( window );
		}

		private void TestsListBox_MouseDoubleClick( object sender, MouseButtonEventArgs e )
		{
			if( TestsListBox.SelectedItem == null )
				return;

			Lock();

			ExecuteTest( TestsListBox.SelectedItem as Test );

			Unlock( false );
		}

		#endregion Event handlers

		#region Private methods

		#region Locking

		int _LockCount = 0;

		private void Lock()
		{
			if( _LockCount > 0 )
				_LockCount = 0;
			_LockCount++;

			MainPanel.IsEnabled = false;
		}

		private void Unlock( bool pForce )
		{
			_LockCount--;

			if( (_LockCount == 0) || pForce )
			{
				MainPanel.IsEnabled = true;
			}
		}

		#endregion Locking

		/// <summary>
		/// Method called when a message has to be saved, as the progress of an action.
		/// </summary>
		/// <param name="message">Message to log.</param>
		/// <param name="isStart">Specifies whether this message is the starting message for an action.</param>
		private void MessagePresenter( string message, bool isStart )
		{
			Window mainWindow = ControlsUtils.GetAppMainWindow();
			if( mainWindow == null )
				return;

			if( TextUtils.NoContent( message ) )
				message = "No text to log";

			if( isStart )
			{
				int lineCount = LogTextBox.LineCount;
				bool isLineStart = ((lineCount >= 1) ? LogTextBox.GetLineLength( LogTextBox.LineCount - 1 ) : 0) <= 0;

				message = (isLineStart ? "" : "\n") + message;
			}

			LogManager.Add( message );
			LogTextBox.AppendText( message );

			// Make the message visible.
			LogTextBox.ScrollToEnd();

			ControlsUtils.ProcessAppUiEvents();
		}

		/// <summary>
		/// Method called when an exception has to be shown.
		/// </summary>
		/// <param name="exception">Exception to show.</param>
		private void ErrorPresenter( Exception exception )
		{
			try
			{
				if( !LogManager.Add( null, exception, LogRecordModes.Error, true ) )
					return;

				string message = ExceptionUtils.GetPresenterMessage( null, exception, false );

				MessagePresenter( AppTestContext.TestSeparator, true );
				MessagePresenter( "ERROR\n", true );
				MessagePresenter( message, true );
				MessagePresenter( AppTestContext.TestSeparator, true );

				Unlock( true );
			}
			catch( Exception ex )
			{
				if( ExceptionUtils.IsCritical( ex ) )
					throw ExceptionUtils.PreserveStackTrace( ex );

				ControlsUtils.ShowException( exception );
			}
		}

		/// <summary>
		/// Method called when the progress has to be changed.
		/// </summary>
		/// <param name="lop">The long operation progress object.</param>
		private void Progress( LongOperationProgress lop )
		{
			Window mainWindow = ControlsUtils.GetAppMainWindow();
			if( mainWindow == null )
				return;

			if( TextUtils.NoContent( lop.PersistentMessage ) && TextUtils.NoContent( lop.LastMessage ) )
			{
				Title = AssemblyUtils.ApplicationBuildInfo.ProductDisplayName;
			}
			else
			{
				string messageToShow = string.Format( LopMessageTemplate, lop.PersistentMessage
					, lop.ElapsedTime.ToString( LopTimeFormat ), lop.LastMessage );

				Title = AssemblyUtils.ApplicationBuildInfo.ProductDisplayName + " | " + messageToShow;
			}

			ControlsUtils.ProcessAppUiEvents();
		}

		/// <summary>
		/// Executes the specified test.
		/// </summary>
		/// <param name="test">The test to execute.</param>
		private void ExecuteTest( Test test )
		{
			ValidationUtils.ObjectHasContent( test, "test" );

			test.Execute( AppContext );
		}

		/// <summary>
		/// Executes all the non-long test.
		/// </summary>
		private void ExecuteTests()
		{
			MessagePresenter( AppTestContext.TestSeparator + "\nThe starred tests will not be executed.", true );
			MessagePresenter( AppTestContext.SubtestSeparator, true );

			foreach( Test test in Tests )
			{
				if( !test.IsLong )
					test.Execute( AppContext );
			}

			MessagePresenter( "All tests have been executed okay.", true );
			MessagePresenter( AppTestContext.TestSeparator, true );
		}

		#endregion Private methods

		#region Public static methods

		/// <summary>
		/// Shows the specified exception.
		/// </summary>
		/// <param name="exception">Exception to show.</param>
		public static void ShowException( Exception exception )
		{
			Window mainWindow = ControlsUtils.GetAppMainWindow();
			if( mainWindow == null )
			{
				LogManager.Add( exception );

				return;
			}

			((MainWindow)mainWindow).ErrorPresenter( exception );
		}

		#endregion Public static methods
	}
}
