////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using System.Text;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Asymmetric.Math;
using Vela.Kernel;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Application test context.
	/// </summary>
	public class AppTestContext : TestContext
	{
		#region Public constants

		/// <summary>
		/// The number of identities which are generated during the identity speed test.
		/// </summary>
		public const int IdentitySpeedCount = 50;

		/// <summary>
		/// The number of locators which are generated during the entropy endurance test.
		/// </summary>
		public const int EntropyEnduranceCount = 10000000;

		/// <summary>
		/// The number of documents which are stored in the storehouse during the storehouse speed test, in a single commit
		/// batch.
		/// </summary>
		public const int StorehouseSpeedAddBatchSize = 1000;

		/// <summary>
		/// The number of commit batches which are made during the storehouse speed test.
		/// </summary>
		public const int StorehouseSpeedAddBatchCount = 50;

		/// <summary>
		/// The number of speckles which are seeked during the storehouse speed test, in a single seek batch, after
		/// all the documents are added to the storehouse.
		/// </summary>
		public const int StorehouseSpeedSeekBatchSize = 20;

		/// <summary>
		/// The number of sheafs to collect during the storehouse speed test, after all the documents are added to the
		/// storehouse.
		/// </summary>
		public const int StorehouseSpeedCollectSheafsCount = 10000;

		/// <summary>
		/// The password used for the symmetric encryption of documents.
		/// </summary>
		public static readonly byte[] DocumentPassword
			= new byte[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2 };

		/// <summary>
		/// The storehouse password.
		/// </summary>
		public static readonly string StorehousePassword = "1234567890";

		/// <summary>
		/// The block size (in bytes) used for the symmetric encryption of documents. Those documents are aligned to this size.
		/// </summary>
		public const int EncryptedDocumentBlockSize = 1000;

		/// <summary>
		/// Template for the information shown in the text-box, at the program's startup.
		/// </summary>
		public const string StartupInfoTemplate = "Before you run the \"Payment service endurance\""
			+ " test, start the service (Vela.PaymentServiceOps)."
			+ "\n{2}"
			+ "\nHigh resolution clock frequency = {0}"
			+ "\nSystem supports IPv6 = {1}"
			+ "\n{2}\n";

		/// <summary>
		/// The text which is shown between tests.
		/// </summary>
		public static readonly string SubtestSeparator = new string( '-', 90 );

		/// <summary>
		/// The text which is shown when a batch of tests ends.
		/// </summary>
		public static readonly string TestSeparator = new string( '*', 90 );

		#endregion Public constants

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected AppTestContext()
		{
		}

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		/// <param name="messagePresenter">Event to raise when a message has to be saved.</param>
		/// <param name="errorPresenter">Event to raise when an exception has to be shown.</param>
		/// <param name="progress">Event to raise when the progress has to be changed.</param>
		/// <param name="configuration">The application's configuration.</param>
		public AppTestContext( MessagePresenterDelegate messagePresenter, ErrorPresenterDelegate errorPresenter
			, LongOperationProgressDelegate progress, AppConfiguration configuration )
		{
			ValidationUtils.ObjectHasContent( messagePresenter, "messagePresenter" );
			ValidationUtils.ObjectHasContent( errorPresenter, "errorPresenter" );
			ValidationUtils.ObjectHasContent( progress, "progress" );
			ValidationUtils.ObjectHasContent( configuration, "configuration" );

			MessagePresenter = messagePresenter;
			ErrorPresenter = errorPresenter;
			LongOperationProgress = new LongOperationProgress( progress );
			Configuration = configuration;

			WorkPath = AssemblyUtils.GetApplicationDataFolderName( "Documents" );

			ServiceStorehouseFileName = Path.Combine( WorkPath, "ServiceStorehouse.vstore" );
			SpeedStorehouseFileName = Path.Combine( WorkPath, "SpeedStorehouse.vstore" );

			#region File names

			SaveLoadSymmEnvelopeFileName = Path.Combine( WorkPath, "SaveLoadSymmEnvelope.vgeneric" );

			VaultAliceFileName = Path.Combine( WorkPath, "PeerVaultAlice.vgeneric" );
			PresenterAliceFileName = Path.Combine( WorkPath, "PeerPresenterAlice.vgeneric" );

			VaultBobFileName = Path.Combine( WorkPath, "PeerVaultBob.vgeneric" );
			PresenterBobFileName = Path.Combine( WorkPath, "PeerPresenterBob.vgeneric" );

			VaultClaudiaFileName = Path.Combine( WorkPath, "PeerVaultClaudia.vgeneric" );
			PresenterClaudiaFileName = Path.Combine( WorkPath, "PeerPresenterClaudia.vgeneric" );

			VaultDanielFileName = Path.Combine( WorkPath, "PeerVaultDaniel.vgeneric" );
			PresenterDanielFileName = Path.Combine( WorkPath, "PeerPresenterDaniel.vgeneric" );

			GroupIdentityPublisherFileName = Path.Combine( WorkPath, "GroupIdPresenter.vgeneric" );
			TallyAliceFileName = Path.Combine( WorkPath, "TallyAlice.vgeneric" );
			AccountAliceFileName = Path.Combine( WorkPath, "AccountAlice.vgeneric" );
			TallyBobFileName = Path.Combine( WorkPath, "TallyBob.vgeneric" );
			AccountBobFileName = Path.Combine( WorkPath, "AccountBob.vgeneric" );
			ServicePresenterFileName = Path.Combine( WorkPath, "ServicePresenter.vgeneric" );
			PaymentVaultFileName = Path.Combine( WorkPath, "ServiceVault.vgeneric" );

			ActionRequestFileName = Path.Combine( WorkPath, "ActionRequest.vgeneric" );
			ActionResponseFileName = Path.Combine( WorkPath, "ActionResponse.vgeneric" );
			AsymmEnvelopeFileName = Path.Combine( WorkPath, "AsymmEncrypted.vgeneric" );
			DocumentToAttestFileName = Path.Combine( WorkPath, "DocumentToAttest.vgeneric" );
			AttestingDocumentFileName = Path.Combine( WorkPath, "AttestingDocument.vgeneric" );
			DocumentToSecureFileName = Path.Combine( WorkPath, "DocumentToSecure.vgeneric" );
			SecuringDocumentFileName = Path.Combine( WorkPath, "SecuringDocument.vgeneric" );
			SecuredDocumentFileName = Path.Combine( WorkPath, "SecuredDocument.vgeneric" );
			UnpackedDocumentFileName = Path.Combine( WorkPath, "UnpackedDocument.vgeneric" );
			SerializerFileName = Path.Combine( WorkPath, "Serializer.vgeneric" );
			HashSerializerFileName = Path.Combine( WorkPath, "HashSerializer.raw" );

			EntropyFileName = AssemblyUtils.GetApplicationTempFileName( null, "EntropyPool.bin" );

			#endregion File names

			AsymmAlg.FailedKeyVerificationEvent += FailedKeyVerification;

			// Show the startup information.
			MessagePresenter( string.Format( StartupInfoTemplate, HighResolutionTimer.Frequency
				, System.Net.Sockets.Socket.OSSupportsIPv6, TestSeparator ), true );
		}

		#endregion Constructors

		#region Private fields

		/// <summary>
		/// Event to raise when a message has to be saved.
		/// </summary>
		private MessagePresenterDelegate MessagePresenter;

		/// <summary>
		/// Event to raise when anexception has to be shown.
		/// </summary>
		private ErrorPresenterDelegate ErrorPresenter;

		/// <summary>
		/// The time elapsed from the beginning of the processing. This is initialized by default.
		/// </summary>
		private Chronometer ElapsedTime = new Chronometer();

		/// <summary>
		/// Folder where the documents are.
		/// </summary>
		private string WorkPath;

		#endregion Private fields

		#region Public fields

		#region File names

		/// <summary>
		/// The name of the file which stores the document to be saved and reloaded.
		/// </summary>
		public string SaveLoadSymmEnvelopeFileName;

		/// <summary>
		/// The name of the file which stores the vault of Alice.
		/// </summary>
		public string VaultAliceFileName;

		/// <summary>
		/// The name of the file which stores the presenter of Alice.
		/// </summary>
		public string PresenterAliceFileName;

		/// <summary>
		/// The name of the file which stores the vault of Bob.
		/// </summary>
		public string VaultBobFileName;

		/// <summary>
		/// The name of the file which stores the presenter of Bob.
		/// </summary>
		public string PresenterBobFileName;

		/// <summary>
		/// The name of the file which stores the vault of Claudia.
		/// </summary>
		public string VaultClaudiaFileName;

		/// <summary>
		/// The name of the file which stores the presenter of Claudia.
		/// </summary>
		public string PresenterClaudiaFileName;

		/// <summary>
		/// The name of the file which stores the vault of Daniel.
		/// </summary>
		public string VaultDanielFileName;

		/// <summary>
		/// The name of the file which stores the presenter of Daniel.
		/// </summary>
		public string PresenterDanielFileName;

		/// <summary>
		/// The name of the file which stores the group identity publisher.
		/// </summary>
		public string GroupIdentityPublisherFileName;

		/// <summary>
		/// The name of the file which stores the tally for Alice.
		/// </summary>
		public string TallyAliceFileName;

		/// <summary>
		/// The name of the file which stores the account for Alice.
		/// </summary>
		public string AccountAliceFileName;

		/// <summary>
		/// The name of the file which stores the tally for Bob.
		/// </summary>
		public string TallyBobFileName;

		/// <summary>
		/// The name of the file which stores the account for Bob.
		/// </summary>
		public string AccountBobFileName;

		/// <summary>
		/// The name of the file which stores the service presenter.
		/// </summary>
		public string ServicePresenterFileName;

		/// <summary>
		/// The name of the file which stores the payment vault.
		/// </summary>
		public string PaymentVaultFileName;

		/// <summary>
		/// The name of the file which stores the action request.
		/// </summary>
		public string ActionRequestFileName;

		/// <summary>
		/// The name of the file which stores the action response.
		/// </summary>
		public string ActionResponseFileName;

		/// <summary>
		/// The name of the file which stores the asymmetrically encrypted document.
		/// </summary>
		public string AsymmEnvelopeFileName;

		/// <summary>
		/// The name of the file which stores the document to attest.
		/// </summary>
		public string DocumentToAttestFileName;

		/// <summary>
		/// The name of the file which stores the attesting document.
		/// </summary>
		public string AttestingDocumentFileName;

		/// <summary>
		/// The name of the file which stores the document to secure.
		/// </summary>
		public string DocumentToSecureFileName;

		/// <summary>
		/// The name of the file which stores the securing document.
		/// </summary>
		public string SecuringDocumentFileName;

		/// <summary>
		/// The name of the file which stores the secured document.
		/// </summary>
		public string SecuredDocumentFileName;

		/// <summary>
		/// The name of the file which stores the unpacked document.
		/// </summary>
		public string UnpackedDocumentFileName;

		/// <summary>
		/// The name of the file which stores the output of the serializer.
		/// </summary>
		public string SerializerFileName;

		/// <summary>
		/// The name of the file which stores the output of the hash serializer.
		/// </summary>
		public string HashSerializerFileName;

		/// <summary>
		/// The name of the file which stores generated entropy which can be used for testing with another application.
		/// </summary>
		public string EntropyFileName;

		#endregion File names

		#endregion Public fields

		#region Public non-serializable properties

		private LongOperationProgress _LongOperationProgress;
		/// <summary>
		/// Handler for the progress of long operations. READ-ONLY.
		/// </summary>
		public LongOperationProgress LongOperationProgress
		{
			get { return _LongOperationProgress; }
			protected set { _LongOperationProgress = value; }
		}

		private AppConfiguration _Configuration;
		/// <summary>
		/// The application's configuration. READ-ONLY.
		/// </summary>
		public AppConfiguration Configuration
		{
			get { return _Configuration; }
			protected set { _Configuration = value; }
		}

		/// <summary>
		/// The abort state of a long operation. If this is true then an exception (<see cref="AbortLongOperationException"/>)
		/// is thrown when the progress changes.
		/// </summary>
		public bool IsAbortGeneration
		{
			get { return LongOperationProgress.IsAbort; }
			set { LongOperationProgress.IsAbort = value; }
		}

		private string _ServiceStorehouseFileName;
		/// <summary>
		/// The full name of the file of the storehouse where the service documents are stored. READ-ONLY.
		/// </summary>
		public string ServiceStorehouseFileName
		{
			get { return _ServiceStorehouseFileName; }
			protected set { _ServiceStorehouseFileName = value; }
		}

		private string _SpeedStorehouseFileName;
		/// <summary>
		/// The full name of the file of the storehouse used for the storehouse speed test. READ-ONLY.
		/// </summary>
		public string SpeedStorehouseFileName
		{
			get { return _SpeedStorehouseFileName; }
			protected set { _SpeedStorehouseFileName = value; }
		}

		#endregion Public non-serializable properties

		#region Private static methods

		/// <summary>
		/// Log data when the verification of a key's parameters has failed.
		/// </summary>
		/// <param name="asymmAlg">The asymmetric algorithm where the key verification failed.</param>
		/// <param name="buffer">Buffer which is used for testing.</param>
		/// <param name="encrypted">The encrypted data (processed with the public key).</param>
		/// <param name="decrypted">The decrypted data (processed with the private key).</param>
		private static void FailedKeyVerification( AsymmAlg asymmAlg, byte[] buffer, byte[] encrypted, byte[] decrypted )
		{
			ValidationUtils.ObjectType<RsaAlg>( asymmAlg, "asymmAlg" );

			RsaAlg rsaAlg = asymmAlg as RsaAlg;

			LogManager.Add( "Plain data: " + buffer, true );
			LogManager.Add( "Encrypted data: " + encrypted, true );
			LogManager.Add( "Decrypted data: " + decrypted, true );

			LogManager.Add( "Key parameter D: " + rsaAlg.D, true );
			LogManager.Add( "Key parameter DP: " + rsaAlg.DP, true );
			LogManager.Add( "Key parameter DQ: " + rsaAlg.DQ, true );
			LogManager.Add( "Key parameter P: " + rsaAlg.P, true );
			LogManager.Add( "Key parameter Q: " + rsaAlg.Q, true );
			LogManager.Add( "Key parameter QI: " + rsaAlg.QI, true );

			LogManager.Add( "Key parameter E: " + rsaAlg.E, true );
			LogManager.Add( "Key parameter M: " + rsaAlg.M, true );
		}

		#endregion Private static methods

		#region Private methods

		/// <summary>
		/// Logs the end of a test and shows speed information.
		/// </summary>
		/// <param name="speedMode">The speed mode of the executed test.</param>
		/// <param name="executed">The number of executed iterations, or the number of processed bytes. This may be null if the
		/// speed mode is <see cref="SpeedModes.Sec"/>.</param>
		/// <param name="separator">Separator at the end of the log message. This may be null or empty.</param>
		private void LogEnd( SpeedModes speedMode, object executed, string separator )
		{
			if( speedMode != SpeedModes.Sec )
				ValidationUtils.ObjectHasContent( executed, "executed" );

			double time = ElapsedTime.Seconds;

			string timeText = string.Format( "{0:0.#} seconds", time );

			string speed;
			switch( speedMode )
			{
				case SpeedModes.Sec:
					speed = timeText;
					break;
				case SpeedModes.Oper_Sec:
					speed = string.Format( "{0:0} operations / second ({1})", (int)executed / time, timeText );
					break;
				case SpeedModes.Sec_Oper:
					speed = string.Format( "{0:0.#} seconds / operation ({1})", time / (int)executed, timeText );
					break;
				case SpeedModes.Msec_Oper:
					speed = string.Format( "{0:0.#} milliseconds / operation ({1})", 1000 * time / (int)executed, timeText );
					break;
				case SpeedModes.MB_Sec:
					speed = string.Format( "{0:0.#} megabytes / seconds ({1})", (int)executed / time / (1 << 20), timeText );
					break;
				default:
					throw ValidationUtils.Unhandled( speedMode, "speedMode" );
			}

			string text = string.Format( "Speed {0}", speed );
			MessagePresenter( text, true );

			if( TextUtils.HasContent( separator ) )
				MessagePresenter( separator, true );

			LongOperationProgress.Clear();
		}

		#endregion Private methods

		#region Public static methods

		/// <summary>
		/// Returns a text which indicates the tipical probability for the Miller-Rabin test to return prime for a non-prime
		/// number.
		/// </summary>
		/// <param name="bc">The number of bits of the big integer which is supposed to be prime.</param>
		/// <param name="r">The number of rounds of the Miller-Rabin test.</param>
		/// <returns>Returns a text which indicates the tipical probability for the Miller-Rabin test to return prime for a
		/// non-prime number.</returns>
		public static string MillerRabinProbability( int bc, int r )
		{
			// The probability of a non-prime passing the test as being prime.
			double p = BigInteger.PrimalityTests.MillerRabinProbability( bc, r );

			// The power of 10, of "p".
			double p_log10 = Math.Log10( p );

			return string.Format( "* \"p\" = 10 ^ {0} for \"bc\" = {1}, \"r\" = {2}.", p_log10.ToString( "#.#" ), bc, r );
		}

		#endregion Public static methods

		#region Public methods

		#region Logging

		/// <summary>
		/// Logs the start of a test. The performance timer is reset.
		/// </summary>
		/// <param name="text">Message to log.</param>
		public void LogStart( string text )
		{
			ValidationUtils.TextHasContent( text, "text" );

			MessagePresenter( string.Format( "TEST {0}... ", text ), true );

			ElapsedTime.Start();

			LongOperationProgress.Start();
		}

		/// <summary>
		/// Logs the processing of a test. The performance timer is reset.
		/// </summary>
		public void LogProcessing()
		{
			MessagePresenter( "Processing... ", true );

			ElapsedTime.Start();

			LongOperationProgress.Start();
		}

		/// <summary>
		/// Logs a message.
		/// </summary>
		/// <param name="text">Message to log.</param>
		public void Log( string text )
		{
			MessagePresenter( text, false );
		}

		/// <summary>
		/// Logs a message.
		/// </summary>
		/// <param name="text">Message to log.</param>
		/// <param name="isStart">Specifies whether this message is the starting message for an action (and thus is shown on a
		/// new line).</param>
		public void Log( string text, bool isStart )
		{
			MessagePresenter( text, isStart );
		}

		/// <summary>
		/// Logs the speed information of the test.
		/// </summary>
		/// <param name="speedMode">The speed mode of the executed test.</param>
		/// <param name="executed">The number of executed iterations, or the number of processed bytes. This may be null if the
		/// speed mode is <see cref="SpeedModes.Sec"/>.</param>
		public void LogSpeed( SpeedModes speedMode, object executed )
		{
			LogEnd( speedMode, executed, null );
		}

		/// <summary>
		/// Logs the end of a test (and its timespan).
		/// </summary>
		public void LogEnd()
		{
			LogEnd( SpeedModes.Sec, null, TestSeparator );

			ElapsedTime.Stop();

			LongOperationProgress.Clear();
		}

		/// <summary>
		/// Logs the start of a test step. The performance timer is reset.
		/// </summary>
		/// <param name="text">Message to log.</param>
		public void LogSubtest( string text )
		{
			ValidationUtils.TextHasContent( text, "text" );

			MessagePresenter( string.Format( "SUBTEST {0}... ", text ), true );

			ElapsedTime.Start();

			LongOperationProgress.Start();
		}

		/// <summary>
		/// Logs the end of a sub-test.
		/// </summary>
		public void LogSubtestEnd()
		{
			LogEnd( SpeedModes.Sec, null, SubtestSeparator );
		}

		/// <summary>
		/// Logs a sub-test separator.
		/// </summary>
		public void LogSubtestSeparator()
		{
			MessagePresenter( SubtestSeparator, true );
		}

		/// <summary>
		/// Logs the start of a test step.
		/// </summary>
		/// <param name="text">Message to log.</param>
		public void LogStep( string text )
		{
			ValidationUtils.TextHasContent( text, "text" );

			MessagePresenter( string.Format( "\t* {0}... ", text ), true );
		}

		#endregion Logging

		/// <summary>
		/// Save the configuration file.
		/// </summary>
		public void SaveConfiguration()
		{
			Configuration.Save();
		}

		/// <summary>
		/// Returns a text which indicates the tipical probability for the Miller-Rabin test to return prime for a non-prime
		/// number.
		/// </summary>
		/// <returns>Returns a text which indicates the tipical probability for the Miller-Rabin test to return prime for a
		/// non-prime number.</returns>
		public string GetMillerRabinProbability()
		{
			StringBuilder sb = new StringBuilder();

			sb.AppendLine( "The typical probability \"p\" for the Miller-Rabin test to return prime for a non-prime number, for"
				+ " \"bc\" bits (the size of the modulus of the asymmetric key is twice this value) and \"r\" rounds is:" );

			sb.AppendLine( "\nIMPLEMENTED:" );
			sb.AppendLine( MillerRabinProbability( 1024, BigInteger.PrimalityTests.GetRoundCount( 1024 ) ) );
			sb.AppendLine( MillerRabinProbability( 1536, BigInteger.PrimalityTests.GetRoundCount( 1536 ) ) );
			sb.AppendLine( MillerRabinProbability( 2048, BigInteger.PrimalityTests.GetRoundCount( 2048 ) ) );
			sb.AppendLine( MillerRabinProbability( 3072, BigInteger.PrimalityTests.GetRoundCount( 3072 ) ) );
			sb.AppendLine( MillerRabinProbability( 4096, BigInteger.PrimalityTests.GetRoundCount( 4096 ) ) );

			sb.AppendLine( "\nCOMPARATIVE:" );
			sb.AppendLine( MillerRabinProbability( 1024, 6 ) );
			sb.AppendLine( MillerRabinProbability( 1024, 5 ) );
			sb.AppendLine( MillerRabinProbability( 1024, 4 ) );
			sb.AppendLine( MillerRabinProbability( 1024, 3 ) );
			sb.AppendLine( MillerRabinProbability( 2048, 6 ) );
			sb.AppendLine( MillerRabinProbability( 2048, 5 ) );
			sb.AppendLine( MillerRabinProbability( 2048, 4 ) );
			sb.AppendLine( MillerRabinProbability( 2048, 3 ) );

			return sb.ToString();
		}

		/// <summary>
		/// Generates a small prime number.
		/// </summary>
		public void GenerateSmallPrime()
		{
			const int primeStart = int.MaxValue - 20;

			LogStart( string.Format( "Generate prime at least {0}", primeStart ) );

			Log( MathUtils.GenerateSmallPrime( primeStart ).ToString() );

			LogEnd();
		}

		/// <summary>
		/// Generates an entropy file which can be used for testing with another application.
		/// </summary>
		public void GenerateEntropyFile()
		{
			const int bufferSize = 10000000;
			const int oneTimeSizeMin = 1;
			const int oneTimeSizeMax = 1000;

			using( FileStream fs = File.OpenWrite( EntropyFileName ) )
			{
				fs.SetLength( 0 );

				LogStart( string.Format( "Generating {0} random bytes in file \"{1}\"", bufferSize, EntropyFileName ) );

				int totalSize = 0;
				while( totalSize < bufferSize )
				{
					double rand = EntropyPool.GetReal();
					int oneTimeSize = (int)(oneTimeSizeMin + rand * (oneTimeSizeMax - oneTimeSizeMin));
					if( totalSize + oneTimeSize > bufferSize )
						oneTimeSize = bufferSize - totalSize;
					totalSize += oneTimeSize;

					// Generate the random data.
					byte[] buffer = EntropyPool.GetBytes( oneTimeSize );

					// Write the random data to the file.
					fs.Write( buffer, 0, buffer.Length );

					LongOperationProgress.PersistentMessage = string.Format( "Size {0} / {1}", totalSize, bufferSize );
					LongOperationProgress.Progress();
				}

				LogSpeed( SpeedModes.MB_Sec, totalSize );
				LogEnd();
			}
		}

		#endregion Public methods
	}
}
