////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Kernel;
using Vela.PaymentDocuments;
using Vela.PeerDocuments;
using Vela.Serialization;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class AsymmetricDecryptionTest : AppTest
	{
		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Asymmetric decryption"; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "0010 0000 01"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			IDeserializer deserializer;

			SymmEncryptedEnvelope seeIdDaniel = SymmEncryptedEnvelope.LoadFromFile( Context.VaultDanielFileName
				, AppTestContext.DocumentPassword, typeof( Vault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeIdDaniel, "seeIdDaniel" );
			Vault vaultDaniel = (Vault)seeIdDaniel.EncryptableDocument;

			SymmEncryptedEnvelope seeIdClaudia = SymmEncryptedEnvelope.LoadFromFile( Context.VaultClaudiaFileName
				, AppTestContext.DocumentPassword, typeof( Vault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeIdClaudia, "seeIdClaudia" );
			Vault vaultClaudia = (Vault)seeIdClaudia.EncryptableDocument;

			SymmEncryptedEnvelope seeIdBob = SymmEncryptedEnvelope.LoadFromFile( Context.VaultBobFileName
				, AppTestContext.DocumentPassword, typeof( Vault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeIdBob, "seeIdBob" );
			Vault vaultBob = (Vault)seeIdBob.EncryptableDocument;

			SymmEncryptedEnvelope seeIdAlice = SymmEncryptedEnvelope.LoadFromFile( Context.VaultAliceFileName
				, AppTestContext.DocumentPassword, typeof( Vault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeIdAlice, "seeIdAlice" );
			Vault vaultAlice = (Vault)seeIdAlice.EncryptableDocument;

			Presenter presenterAlice = SerializationUtils.LoadFromFile( Context.PresenterAliceFileName
				, typeof( Presenter ), out deserializer ) as Presenter;
			ValidationUtils.ObjectHasContent( presenterAlice, "presenterAlice" );

			Context.LogProcessing();

			// Note that due to the fact that sessions keys are replaced in the resulting decryption password (and since every
			// session key has, in its extra session key, the group session key of the next recipient), an alteration of a group
			// session key may not necessarily lead to a decryption failure (since the wrong key could be replaced with the
			// correct key).

			// Use an "AsymmDecryptionPassword" object for each decryption because normally each decryption is done on the
			// computer of each recipient, and then all session keys are gathered on the computer of a single recipient.

			#region Collect the session keys

			// Get the session keys.
			AsymmDecryptionPassword adpDaniel = AsymmDecryptionPassword.PrepareToCollect( vaultDaniel.ProcessingIdentity );
			AsymmEncryptedEnvelope.LoadFromFile( Context.AsymmEnvelopeFileName, adpDaniel, typeof( MakePaymentActionRequest )
				, out deserializer );

			// Get the session keys.
			AsymmDecryptionPassword adpClaudia = AsymmDecryptionPassword.PrepareToCollect( vaultClaudia.ProcessingIdentity );
			AsymmEncryptedEnvelope.LoadFromFile( Context.AsymmEnvelopeFileName, adpClaudia
				, typeof( MakePaymentActionRequest ), out deserializer );

			// Get the session keys.
			AsymmDecryptionPassword adpBob = AsymmDecryptionPassword.PrepareToCollect( vaultBob.ProcessingIdentity );
			AsymmEncryptedEnvelope.LoadFromFile( Context.AsymmEnvelopeFileName, adpBob
				, typeof( MakePaymentActionRequest ), out deserializer );

			// Get the session keys.
			AsymmDecryptionPassword adpAlice = AsymmDecryptionPassword.PrepareToCollect( vaultAlice.ProcessingIdentity );
			AsymmEncryptedEnvelope.LoadFromFile( Context.AsymmEnvelopeFileName, adpAlice
				, typeof( MakePaymentActionRequest ), out deserializer );

			#endregion Collect the session keys

			// Decrypt the document.
			AsymmDecryptionPassword adp = new AsymmDecryptionPassword( adpDaniel, adpClaudia, adpBob, adpAlice );
			AsymmEncryptedEnvelope ase = AsymmEncryptedEnvelope.LoadFromFile( Context.AsymmEnvelopeFileName, adp
				, typeof( MakePaymentActionRequest ), out deserializer );
			ValidationUtils.ObjectHasContent( ase, "ase" );

			// Verify signature.
			MakePaymentActionRequest actionRequest = (MakePaymentActionRequest)ase.EncryptableDocument;
			actionRequest.VerifySignature( presenterAlice.ProcessingIdentity );
		}

		#endregion Private methods
	}
}
