////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Kernel;
using Vela.PaymentDocuments;
using Vela.PeerDocuments;
using Vela.Serialization;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class AsymmetricEncryptionTest : AppTest
	{
		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Asymmetric encryption"; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "0010 0000 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			IDeserializer deserializer;

			// Load the service vault.
			SymmEncryptedEnvelope seeServiceVault = SymmEncryptedEnvelope.LoadFromFile( Context.PaymentVaultFileName
				, AppTestContext.DocumentPassword, typeof( PaymentVault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeServiceVault, "seeServiceVault" );
			PaymentVault serviceVault = (PaymentVault)seeServiceVault.EncryptableDocument;

			// Load the vault which creates the document.
			SymmEncryptedEnvelope seeIdAliceAccount = SymmEncryptedEnvelope.LoadFromFile( Context.VaultAliceFileName
				, AppTestContext.DocumentPassword, typeof( Vault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeIdAliceAccount, "seeIdAliceAccount" );
			Vault vaultAlice = (Vault)seeIdAliceAccount.EncryptableDocument;

			// Load the presenter of the beneficiary of the action request.
			Presenter presenterBob = SerializationUtils.LoadFromFile( Context.PresenterBobFileName
				, typeof( Presenter ), out deserializer ) as Presenter;
			ValidationUtils.ObjectHasContent( presenterBob, "presenterBob" );

			// Load the account of the beneficiary of the action request.
			SymmEncryptedEnvelope seeBobAccount = SymmEncryptedEnvelope.LoadFromFile( Context.AccountBobFileName
				, AppTestContext.DocumentPassword, typeof( Account ), out deserializer );
			ValidationUtils.ObjectHasContent( seeBobAccount, "seeBobAccount" );
			Account accountBob = (Account)seeBobAccount.EncryptableDocument;

			GroupIdentityPublisher groupIdentityPublisher = SerializationUtils.LoadFromFile(
					Context.GroupIdentityPublisherFileName, typeof( GroupIdentityPublisher ), out deserializer )
				as GroupIdentityPublisher;
			ValidationUtils.ObjectHasContent( groupIdentityPublisher, "groupIdentityPublisher" );

			#region Prepare the document to encrypt

			// Create the document to sign, which is also to be encrypted.
			MakePaymentActionRequest actionRequest = new MakePaymentActionRequest( CryptoUtils.GetLocator()
				, AlgorithmPool.GetMacAlg(), AlgorithmPool.GetAsymmSigningPadding(), AlgorithmPool.GetHashSerializer( false )
				, vaultAlice.LocalUpnFingerprint, serviceVault.LocalUpn, PeerDocumentsConsts.ActionRequestValidity, 0
				, presenterBob.CreatorUpn );

			// Set the document's properties.
			actionRequest.CurrencyName = "GAU";
			actionRequest.CurrencyAmount = 1;

			// Sign the data which is to be encrypted.
			actionRequest.Sign( vaultAlice.ProcessingIdentity, null );

			#endregion Prepare the document to encrypt

			AsymmEncryptedEnvelope ase = AlgorithmPool.GetAsymmEncryptedEnvelope();
			ase.EncryptedDocumentBlockSize = AppTestContext.EncryptedDocumentBlockSize;

			ase.EncryptableDocument = actionRequest;

			// Encrypt document.
			ase.SaveToFile( Context.AsymmEnvelopeFileName, null, groupIdentityPublisher, deserializer.GetSerializer() );
		}

		#endregion Private methods
	}
}
