////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Asymmetric.Math;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class BlindingEnduranceTest : AppTest
	{
		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Blinding endurance (infinite cycle) *"; }
		}

		/// <summary>
		/// This is true, meaning that the test is long.
		/// </summary>
		public override bool IsLong
		{
			get { return true; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "9000 0005 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			#region Load the identity

			IDeserializer deserializer;

			SymmEncryptedEnvelope seeIdAlice = SymmEncryptedEnvelope.LoadFromFile( Context.VaultAliceFileName
				, AppTestContext.DocumentPassword, typeof( Vault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeIdAlice, "seeIdAlice" );
			Vault vaultAlice = (Vault)seeIdAlice.EncryptableDocument;

			AsymmAlg asymmAlg = vaultAlice.ProcessingIdentity.SigningAca;

			#endregion Load the identity

			int i = 1;

			while( true )
			{
				Context.LongOperationProgress.Progress( 0, string.Format( "Blinding {0}", i ) );

				// Prepare some random data to blind.
				BigInteger randBi = 0;
				while( randBi < 2 ) randBi = BigInteger.GenerateRandom( asymmAlg.MaxProcessableSizeBytes * 8 );
				byte[] buffer = randBi.GetBytes();

				// Prepare the blinding configuration.
				BlindingConfiguration blindingConfiguration = asymmAlg.GenerateBlindingConfiguration( buffer );

				// Create a blinding session. Normally, this has to be saved.
				BlindingSession blindingSession = new BlindingSession( CryptoUtils.GetLocator()
					, vaultAlice.LocalUpnFingerprint, CryptoUtils.GetLocator(), DateTime.UtcNow.AddDays( 1 ), 0
					, blindingConfiguration );

				#region Blind and sign

				// Blind.
				byte[] blinded = asymmAlg.Blind( buffer, blindingConfiguration );

				// Sign.
				byte[] signed = asymmAlg.ProcessWithPrivateKey( blinded, false );

				#endregion Blind and sign

				#region Un-blind and un-sign

				// Un-blind.
				byte[] unblinded = asymmAlg.Unblind( signed, blindingSession.BlindingConfiguration );

				// Un-sign.
				byte[] unsigned = asymmAlg.ProcessWithPublicKey( unblinded, false );

				#endregion Un-blind and un-sign

				#region Check result

				// Because of the "Zero-byte issue" of the "BigInteger" library, pad the result.
				unsigned = BufferUtils.PadBeginningWithZeros( unsigned, buffer.Length );

				if( !BufferUtils.Equals( buffer, unsigned ) ) throw new ValidationException( "Failed unblinding." );

				#endregion Check result

				i++;
			}
		}

		#endregion Private methods
	}
}
