////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Kernel;
using Vela.PaymentDocuments;
using Vela.PeerDocuments;
using Vela.Serialization;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class CreateActionRequestTest : AppTest
	{
		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Create action request"; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "0011 0000 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			IDeserializer deserializer;

			// Load the service vault.
			SymmEncryptedEnvelope seeServiceVault = SymmEncryptedEnvelope.LoadFromFile( Context.PaymentVaultFileName
				, AppTestContext.DocumentPassword, typeof( PaymentVault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeServiceVault, "seeServiceVault" );
			PaymentVault serviceVault = (PaymentVault)seeServiceVault.EncryptableDocument;

			// Load the presenter of the service.
			Presenter servicePresenter = SerializationUtils.LoadFromFile( Context.ServicePresenterFileName
				, typeof( Presenter ), out deserializer ) as Presenter;
			ValidationUtils.ObjectHasContent( servicePresenter, "presenterClaudia" );

			// Load the vault which creates the document.
			SymmEncryptedEnvelope seeIdAliceAccount = SymmEncryptedEnvelope.LoadFromFile( Context.VaultAliceFileName
				, AppTestContext.DocumentPassword, typeof( Vault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeIdAliceAccount, "seeIdAliceAccount" );
			Vault vaultAlice = (Vault)seeIdAliceAccount.EncryptableDocument;

			// Load the presenter of the beneficiary of the action request.
			Presenter presenterBob = SerializationUtils.LoadFromFile( Context.PresenterBobFileName
				, typeof( Presenter ), out deserializer ) as Presenter;
			ValidationUtils.ObjectHasContent( presenterBob, "presenterBob" );

			// Load the account of the beneficiary of the action request.
			SymmEncryptedEnvelope seeBobAccount = SymmEncryptedEnvelope.LoadFromFile( Context.AccountBobFileName
				, AppTestContext.DocumentPassword, typeof( Account ), out deserializer );
			ValidationUtils.ObjectHasContent( seeBobAccount, "seeBobAccount" );
			Account accountBob = (Account)seeBobAccount.EncryptableDocument;

			#region Prepare the action request

			// Create the action request.
			MakePaymentActionRequest actionRequest = new MakePaymentActionRequest( CryptoUtils.GetLocator()
				, AlgorithmPool.GetMacAlg(), AlgorithmPool.GetAsymmSigningPadding(), AlgorithmPool.GetHashSerializer( false )
				, vaultAlice.LocalUpnFingerprint, serviceVault.LocalUpn, PeerDocumentsConsts.ActionRequestValidity, 0
				, presenterBob.CreatorUpn );

			// Set the document's properties.
			actionRequest.CurrencyName = "GAU";
			actionRequest.CurrencyAmount = 1;

			#endregion Prepare the action request

			// Sign the action request.
			actionRequest.Sign( vaultAlice.ProcessingIdentity, null );

			#region Encrypt

			AsymmEncryptedEnvelope ase = AlgorithmPool.GetAsymmEncryptedEnvelope();
			ase.EncryptableDocument = actionRequest;

			// Save the (encrypted) action request.
			ase.SaveToFile( Context.ActionRequestFileName, null, servicePresenter.ProcessingIdentity
				, deserializer.GetSerializer() );

			#endregion Encrypt
		}

		#endregion Private methods
	}
}
