////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.IO;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Mac;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.StarkDocuments;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class CreateAttestingDocumentTest : AppTest
	{
		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Create attesting document"; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "0008 0000 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			IDeserializer deserializer;

			// Load the vault which signs the document.
			SymmEncryptedEnvelope seeIdAliceAccount = SymmEncryptedEnvelope.LoadFromFile( Context.VaultAliceFileName
				, AppTestContext.DocumentPassword, typeof( Vault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeIdAliceAccount, "seeIdAliceAccount" );
			Vault vaultAlice = (Vault)seeIdAliceAccount.EncryptableDocument;

			#region Prepare the document which is to be attested

			byte[] buffer = EntropyPool.GetBytes( 10000 );
			File.WriteAllBytes( Context.DocumentToAttestFileName, buffer );

			#endregion Prepare the document which is to be attested

			#region Create the attesting document

			MacAlg macAlg = AlgorithmPool.GetMacAlg();

			// Create the attesting document.
			AttestingDocument attestingDocument = new AttestingDocument( CryptoUtils.GetLocator(), macAlg
				, AlgorithmPool.GetAsymmSigningPadding(), AlgorithmPool.GetHashSerializer( false )
				, vaultAlice.LocalUpnFingerprint );
			attestingDocument.Description = "Attesting document description";

			byte[] macKey = macAlg.HashAlg.Compute( null, true, attestingDocument.InstanceLocator
				, attestingDocument.InitiatorUpnFingerprint.Value );
			byte[] manualMacValue = macAlg.Compute( macKey, null, false, buffer.LongLength, buffer );

			byte[] macValue;
			using( FileStream fs = File.OpenRead( Context.DocumentToAttestFileName ) )
			{
				macValue = attestingDocument.Process( fs, null );
			}

			if( !BufferUtils.Equals( manualMacValue, macValue ) )
				throw new ValidationException( "The MAC computed from the buffer and from the file are different." );

			attestingDocument.AttestedMacValue = macValue;

			// Sign the attesting document.
			attestingDocument.Sign( vaultAlice.ProcessingIdentity, null );

			// Save the attesting document.
			ISerializer serializer = AlgorithmPool.GetSerializerForExport();
			serializer.SaveToFile( attestingDocument, null, Context.AttestingDocumentFileName );

			#endregion Create the attesting document
		}

		#endregion Private methods
	}
}
