////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.IO;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Mac;
using Vela.Crypto.Symmetric;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.StarkDocuments;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class CreateSecuringDocumentTest : AppTest
	{
		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Create securing document"; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "0009 0000 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			IDeserializer deserializer;

			// Load the vault which signs the document.
			SymmEncryptedEnvelope seeIdAliceAccount = SymmEncryptedEnvelope.LoadFromFile( Context.VaultAliceFileName
				, AppTestContext.DocumentPassword, typeof( Vault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeIdAliceAccount, "seeIdAliceAccount" );
			Vault vaultAlice = (Vault)seeIdAliceAccount.EncryptableDocument;

			#region Prepare the document which is to be secured

			byte[] buffer = EntropyPool.GetBytes( 10000 );
			File.WriteAllBytes( Context.DocumentToSecureFileName, buffer );

			#endregion Prepare the document which is to be secured

			#region Create the securing document

			MacAlg macAlg = AlgorithmPool.GetMacAlg();
			SymmAlg symmAlg = AlgorithmPool.GetSymmAlg();

			// Create the securing document.
			SecuringDocument securingDocument = new SecuringDocument( CryptoUtils.GetLocator(), macAlg
				, AlgorithmPool.GetAsymmSigningPadding(), AlgorithmPool.GetHashSerializer( false )
				, vaultAlice.LocalUpnFingerprint, CryptoUtils.GetLocator(), symmAlg
				, EntropyPool.GetBytes( symmAlg.AlgSizeBytes ), EntropyPool.GetBytes( symmAlg.AlgSizeBytes )
				, AlgorithmPool.GetSecuredDocumentBlockSize() );
			securingDocument.Description = "Securing document description";

			byte[] macValue;
			byte[] securedDocumentBuffer;
			using( FileStream documentToSecureFs = File.OpenRead( Context.DocumentToSecureFileName ) )
			using( FileStream securedDocumentFs = File.Open( Context.SecuredDocumentFileName, FileMode.Create ) )
			{
				macValue = securingDocument.Process( documentToSecureFs, securedDocumentFs, true, null );

				securedDocumentBuffer = new byte[ (int)securingDocument.EncryptedDataSize ];
				securedDocumentFs.Position = 0;
				securedDocumentFs.Read( securedDocumentBuffer, 0, securedDocumentBuffer.Length );
			}

			byte[] macKey = macAlg.HashAlg.Compute( null, true, securingDocument.Salt, securingDocument.SymmEncryptionKey
				, securingDocument.InstanceLocator, securingDocument.InitiatorUpnFingerprint.Value );
			byte[] manualMacValue = macAlg.Compute( macKey, null, false, buffer.LongLength, securedDocumentBuffer );

			if( !BufferUtils.Equals( manualMacValue, macValue ) )
				throw new ValidationException( "The MAC computed from the buffer and from the file are different." );

			securingDocument.AttestedMacValue = macValue;

			// Sign the securing document.
			securingDocument.Sign( vaultAlice.ProcessingIdentity, null );

			// Save the securing document (WITHOUT ENCRYPTING IT).
			ISerializer serializer = AlgorithmPool.GetSerializerForExport();
			serializer.SaveToFile( securingDocument, null, Context.SecuringDocumentFileName );

			#endregion Create the securing document
		}

		#endregion Private methods
	}
}
