////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.ActionProcessor;
using Vela.ApplicationBackend;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.DatabaseStorehouse;
using Vela.Kernel;
using Vela.PaymentDocuments;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class CreateServiceStorehouseTest : AppTest
	{
		#region Private constants

		private const int AccountCurrencyAmount = 1000000000;

		#endregion Private constants

		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Create service storehouse"; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "0007 0000 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		private void CreateServiceVault( IStorehouseTransaction transaction )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );

			IDeserializer deserializer;

			Context.LogSubtest( "Create service vault" );

			// Load the processing vault to use for the service.
			SymmEncryptedEnvelope procSeeIdDaniel = SymmEncryptedEnvelope.LoadFromFile( Context.VaultDanielFileName
				, AppTestContext.DocumentPassword, typeof( Vault ), out deserializer );
			ValidationUtils.ObjectHasContent( procSeeIdDaniel, "procSeeIdDaniel" );
			Vault procVaultDaniel = (Vault)procSeeIdDaniel.EncryptableDocument;

			// Create the service appendix.
			PaymentAppendix appendix = new PaymentAppendix( "Payment Service" );

			// Set the properties of the service appendix.
			appendix.ActionProcessingUrl = Context.Configuration.ActionProcessingUrl;
			appendix.FeePlans = Context.Configuration.ServiceFeePlans;

			PaymentVault serviceVault = new PaymentVault( CryptoUtils.GetLocator(), 0, procVaultDaniel.ProcessingIdentity
				, appendix );

			// Save the encrypted service vault.
			SymmEncryptedEnvelope seeServiceVault = AlgorithmPool.GetSymmEncryptedEnvelope();
			seeServiceVault.EncryptableDocument = serviceVault;
			seeServiceVault.SaveToFile( Context.PaymentVaultFileName, null, AppTestContext.DocumentPassword
				, deserializer.GetSerializer() );

			transaction.AddSheaf( serviceVault, StorehouseRealms.ProcessingIdentityContainers );

			Context.LogSubtestEnd();
		}

		private void CreateServicePresenter( IStorehouseTransaction transaction )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );

			IDeserializer deserializer;

			Context.LogSubtest( "Create service presenter" );

			// Load the service vault.
			SymmEncryptedEnvelope seeServiceVault = SymmEncryptedEnvelope.LoadFromFile( Context.PaymentVaultFileName
				, AppTestContext.DocumentPassword, typeof( PaymentVault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeServiceVault, "seeServiceVault" );
			PaymentVault vault = (PaymentVault)seeServiceVault.EncryptableDocument;

			// Load the processing presenter to use for the service.
			Presenter procPresenterDaniel = SerializationUtils.LoadFromFile( Context.PresenterDanielFileName
				, typeof( Presenter ), out deserializer ) as Presenter;
			ValidationUtils.ObjectHasContent( procPresenterDaniel, "procPresenterDaniel" );

			// Create the service presenter.
			Presenter servicePresenter = new Presenter( CryptoUtils.GetLocator(), AlgorithmPool.GetMacAlg()
				, AlgorithmPool.GetAsymmSigningPadding(), AlgorithmPool.GetHashSerializer( false )
				, (SolitaryIdentityPublisher)procPresenterDaniel.ProcessingIdentity, vault.Appendix, null );

			// Self-sign the service presenter.
			servicePresenter.Sign( vault.ProcessingIdentity, null );

			// Save the service presenter.
			ISerializer serializer = deserializer.GetSerializer();
			serializer.SaveToFile( servicePresenter, null, Context.ServicePresenterFileName );

			transaction.AddSheaf( servicePresenter, StorehouseRealms.ProcessingIdentityContainers );

			Context.LogSubtestEnd();
		}

		private void CreateAccounts( IStorehouseTransaction transaction )
		{
			ValidationUtils.ObjectHasContent( transaction, "transaction" );

			IDeserializer deserializer;

			Context.LogSubtest( "Create account" );

			// Load the service vault.
			SymmEncryptedEnvelope seeServiceVault = SymmEncryptedEnvelope.LoadFromFile( Context.PaymentVaultFileName
				, AppTestContext.DocumentPassword, typeof( PaymentVault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeServiceVault, "seeServiceVault" );
			PaymentVault serviceVault = (PaymentVault)seeServiceVault.EncryptableDocument;

			#region Alice's account

			// Load the account identity.
			Presenter presenterAlice = SerializationUtils.LoadFromFile( Context.PresenterAliceFileName
				, typeof( Presenter ), out deserializer ) as Presenter;
			ValidationUtils.ObjectHasContent( presenterAlice, "presenterAlice" );

			#region Tally

			// Create the tally.
			Tally aliceTally = new Tally( CryptoUtils.GetLocator(), presenterAlice.ProcessingIdentity
				, presenterAlice.Appendix.ActionRequestValidity, presenterAlice.EntityName, TallyApprovals.Yellow );

			// Encrypt the tally.
			SymmEncryptedEnvelope seeAliceState = AlgorithmPool.GetSymmEncryptedEnvelope();
			seeAliceState.EncryptableDocument = aliceTally;

			// Save the tally.
			seeAliceState.SaveToFile( Context.TallyAliceFileName, null, AppTestContext.DocumentPassword
				, deserializer.GetSerializer() );

			#endregion Tally

			#region Account

			// Create the account.
			AccountAlias aliceAccountAlias = AccountAlias.FromLocalAndRemoteEntityNames( serviceVault.EntityName
				, presenterAlice.EntityName );
			PaymentAccount aliceAccount = new PaymentAccount( CryptoUtils.GetLocator(), serviceVault.LocalUpn
				, presenterAlice.CreatorUpn, aliceAccountAlias );

			// Set the tally properties.
			aliceTally.Approval = TallyApprovals.Green;

			// Set a balance.
			PaymentBalance balance = new PaymentBalance();
			balance.CurrencyName = "GAU";
			balance.CurrencyAmount = AccountCurrencyAmount;

			aliceAccount.Balances.Add( balance );

			// Encrypt the account.
			SymmEncryptedEnvelope seeAliceAccount = AlgorithmPool.GetSymmEncryptedEnvelope();
			seeAliceAccount.EncryptableDocument = aliceAccount;

			// Save the account.
			seeAliceAccount.SaveToFile( Context.AccountAliceFileName, null, AppTestContext.DocumentPassword
				, deserializer.GetSerializer() );

			#endregion Account

			#endregion Alice's account

			#region Bob's account

			Presenter presenterBob = SerializationUtils.LoadFromFile( Context.PresenterBobFileName
				, typeof( Presenter ), out deserializer ) as Presenter;
			ValidationUtils.ObjectHasContent( presenterBob, "presenterBob" );

			#region Tally

			// Create the tally.
			Tally bobTally = new Tally( CryptoUtils.GetLocator(), presenterBob.ProcessingIdentity
				, presenterBob.Appendix.ActionRequestValidity, presenterBob.EntityName, TallyApprovals.Yellow );

			// Encrypt the tally.
			SymmEncryptedEnvelope seeBobState = AlgorithmPool.GetSymmEncryptedEnvelope();
			seeBobState.EncryptableDocument = bobTally;

			// Save the tally.
			seeBobState.SaveToFile( Context.TallyBobFileName, null, AppTestContext.DocumentPassword
				, deserializer.GetSerializer() );

			#endregion Tally

			#region Account

			// Create the account.
			AccountAlias bobAccountAlias = AccountAlias.FromLocalAndRemoteEntityNames( serviceVault.EntityName
				, presenterAlice.EntityName );
			PaymentAccount bobAccount = new PaymentAccount( CryptoUtils.GetLocator(), serviceVault.LocalUpn
				, presenterBob.CreatorUpn, bobAccountAlias );

			// Set the tally properties.
			bobTally.Approval = TallyApprovals.Green;

			// Set a balance.
			balance = new PaymentBalance();
			balance.CurrencyName = "GAU";
			balance.CurrencyAmount = AccountCurrencyAmount;

			bobAccount.Balances.Add( balance );

			// Encrypt the account.
			SymmEncryptedEnvelope seeBobAccount = AlgorithmPool.GetSymmEncryptedEnvelope();
			seeBobAccount.EncryptableDocument = bobAccount;

			// Save the account.
			seeBobAccount.SaveToFile( Context.AccountBobFileName, null, AppTestContext.DocumentPassword
				, deserializer.GetSerializer() );

			#endregion Account

			#endregion Bob's account

			transaction.AddSheaf( presenterAlice, StorehouseRealms.ProcessingIdentityContainers );
			transaction.AddSheaf( aliceTally, StorehouseRealms.ProcessingIdentityContainers );
			transaction.AddSheaf( aliceAccount, StorehouseRealms.Accounts );
			transaction.AddSheaf( presenterBob, StorehouseRealms.ProcessingIdentityContainers );
			transaction.AddSheaf( bobTally, StorehouseRealms.ProcessingIdentityContainers );
			transaction.AddSheaf( bobAccount, StorehouseRealms.Accounts );

			Context.LogSubtestEnd();
		}

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			#region Create the storehouse provider

			ISerializer serializer = AlgorithmPool.GetSerializer();

			StorehouseHeader header = AlgorithmPool.GetStorehouseHeader();

			StorehouseEntropy entropy = new StorehouseEntropy( AlgorithmPool.GetEntropyAlg() );

			ActionServiceListenerConfigurations serviceConfigurations = new ActionServiceListenerConfigurations();
			serviceConfigurations.Add( Configuration.ActionServiceListenerTemplate );
			ServiceListenersStorehouseConfiguration configuration = new ServiceListenersStorehouseConfiguration(
				serviceConfigurations );

			IStorehouseProvider storehouseProvider = StorehouseUtils.CreateStorehouseProvider(
				Context.Configuration.StorehouseAccess.StorehouseProviderTypeName
				, Context.Configuration.StorehouseAccess as DatabaseAccess, Context.ServiceStorehouseFileName
				, AppTestContext.StorehousePassword, header, serializer ) as IStorehouseProvider;

			StorehouseUtils.CreateStorehouse( storehouseProvider, Configuration.StorehouseAdminConnectionString, entropy
				, configuration );

			#endregion Create the storehouse provider

			// Create a transaction.
			IStorehouseTransaction transaction = storehouseProvider.CreateTransaction();

			try
			{
				// Create and save the documents.
				CreateServiceVault( transaction );
				CreateServicePresenter( transaction );
				CreateAccounts( transaction );

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}
		}

		#endregion Private methods
	}
}
