////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto;
using Vela.Kernel;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class EntropyDistributionTest : AppTest
	{
		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Entropy distribution"; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "0001 0000 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Tests the quality of the entropy algorithm.
		/// </summary>
		/// <param name="bufferSize">The size of the buffer which must contain the generated random data. This must be at least
		/// 1.</param>
		/// <param name="meanSum">Is set with the mean sum of the generated random data, after a number of iterations.</param>
		/// <param name="squareDiff">Is set with the sum of square differences of the generated random data, after a number of
		/// iterations.</param>
		/// <returns>Returns the number of iterations, where for each iteration the buffer was filled with random data.</returns>
		private int GetEntropyDistribution( int bufferSize, out Int64 meanSum, out Int64 squareDiff )
		{
			ValidationUtils.Integer( bufferSize, 1, int.MaxValue, "bufferSize" );

			const int iter = 10000;

			meanSum = 0;
			squareDiff = 0;

			Context.Log( string.Format( "{0} bytes * {1} iterations", bufferSize, iter ) );

			for( int i = 0; i < iter; i++ )
			{
				// Get random data.
				byte[] buffer = EntropyPool.GetBytes( bufferSize );

				// Sum all random data.
				Int64 meanValue = 0;
				for( int b = 0; b < bufferSize; b++ ) meanValue += buffer[ b ];

				meanSum += meanValue;

				meanValue = meanValue / bufferSize;

				// Sum the square differences of all random data.
				for( int b = 0; b < bufferSize; b++ )
					squareDiff += (buffer[ b ] - meanValue) * (buffer[ b ] - meanValue);
			}

			return iter;
		}

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			const int bufferSize = 1024;

			Int64 meanSum;
			Int64 squareDiff;

			int iter = GetEntropyDistribution( bufferSize, out meanSum, out squareDiff );

			// Calculate distribution (as percentages).
			int byteCount = iter * bufferSize;
			double meanDistrib = 100.0 / 255 * meanSum / byteCount;
			double squareDistrib = 100.0 / 255 * Math.Sqrt( squareDiff / byteCount );

			Context.LogStep( string.Format( "Mean = {0:0.00}%", meanDistrib ) );
			Context.LogStep( string.Format( "Square = {0:0.00}%", squareDistrib ) );
		}

		#endregion Private methods
	}
}
