////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class GenerateIdentitiesTest : AppTest
	{
		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Generate identities"; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "0005 0000 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			Context.Log( "4 identities" );

			UpnAlg upnAlg = AlgorithmPool.GetUpnAlg( AlgorithmSecurityLevel.Normal );
			UInt16 processingPort = Configuration.ActionServiceListenerTemplate.ProcessingPort;
			ISerializer serializer = AlgorithmPool.GetSerializerForExport();
			IHashSerializer hashSerializer = serializer.GetHashSerializer( false );

			Locator groupUpnFingerprint = CryptoUtils.GetLocator();

			#region Identity "Daniel"

			Context.LogSubtest( "Identity \"Daniel\"" );

			// Create identity.
			AsymmAlg aeaDaniel = AlgorithmPool.GenerateAsymmAlg( AlgorithmSecurityLevel.Normal, Context.LongOperationProgress );
			AsymmAlg asaDaniel = AlgorithmPool.GenerateAsymmAlg( AlgorithmSecurityLevel.Normal, Context.LongOperationProgress );
			PrivateIdentity piDaniel = new PrivateIdentity( upnAlg, processingPort, aeaDaniel, asaDaniel );
			piDaniel.GroupUpnFingerprint = groupUpnFingerprint;

			// Set the appendix and physical identity.
			Appendix paDaniel = new Appendix( "Daniel Dellmar" );
			PersonVirtualIdentity virtualIdDaniel = new PersonVirtualIdentity();
			paDaniel.VirtualIdentity = virtualIdDaniel;

			Vault vaultDaniel = new Vault( CryptoUtils.GetLocator(), 0, piDaniel, paDaniel );

			// Save the vault.
			SymmEncryptedEnvelope seeIdDaniel = AlgorithmPool.GetSymmEncryptedEnvelope();
			seeIdDaniel.EncryptableDocument = vaultDaniel;
			seeIdDaniel.SaveToFile( Context.VaultDanielFileName, null, AppTestContext.DocumentPassword, serializer );

			// Create the solitary identity publisher.
			SolitaryIdentityPublisher identityPublisherDaniel = CoreDocumentsUtils.GetSolitaryIdentity(
				vaultDaniel.ProcessingIdentity );

			// Create and self-sign the presenter.
			Presenter presenterDaniel = new Presenter( CryptoUtils.GetLocator(), AlgorithmPool.GetMacAlg()
				, AlgorithmPool.GetAsymmSigningPadding(), AlgorithmPool.GetHashSerializer( false ), identityPublisherDaniel
				, paDaniel, null );
			presenterDaniel.Sign( vaultDaniel.ProcessingIdentity, null );

			// Save the presenter.
			serializer.SaveToFile( presenterDaniel, null, Context.PresenterDanielFileName );

			Context.LogSubtestEnd();

			#endregion Identity "Daniel"

			#region Identity "Claudia"

			Context.LogSubtest( "Identity \"Claudia\"" );

			// Create identity.
			AsymmAlg aeaClaudia = AlgorithmPool.GenerateAsymmAlg( AlgorithmSecurityLevel.Normal, Context.LongOperationProgress );
			AsymmAlg asaClaudia = AlgorithmPool.GenerateAsymmAlg( AlgorithmSecurityLevel.Normal, Context.LongOperationProgress );
			PrivateIdentity piClaudia = new PrivateIdentity( upnAlg, processingPort, aeaClaudia, asaClaudia );
			piClaudia.GroupUpnFingerprint = groupUpnFingerprint;

			// Set the appendix and physical identity.
			Appendix paClaudia = new Appendix( "Claudia Cousteau" );
			PersonVirtualIdentity virtualIdClaudia = new PersonVirtualIdentity();
			paClaudia.VirtualIdentity = virtualIdClaudia;

			virtualIdClaudia.IntimateNote = "Claudia's note.";

			Vault vaultClaudia = new Vault( CryptoUtils.GetLocator(), 0, piClaudia, paClaudia );

			// Save the vault.
			SymmEncryptedEnvelope seeIdClaudia = AlgorithmPool.GetSymmEncryptedEnvelope();
			seeIdClaudia.EncryptableDocument = vaultClaudia;
			seeIdClaudia.SaveToFile( Context.VaultClaudiaFileName, null, AppTestContext.DocumentPassword, serializer );

			// Create the solitary identity publisher.
			SolitaryIdentityPublisher identityPublisherClaudia = CoreDocumentsUtils.GetSolitaryIdentity(
				vaultClaudia.ProcessingIdentity );

			// Create and self-sign the presenter.
			Presenter presenterClaudia = new Presenter( CryptoUtils.GetLocator(), AlgorithmPool.GetMacAlg()
				, AlgorithmPool.GetAsymmSigningPadding(), AlgorithmPool.GetHashSerializer( false ), identityPublisherClaudia
				, paClaudia, null );
			presenterClaudia.Sign( vaultClaudia.ProcessingIdentity, null );

			// Save the presenter.
			serializer.SaveToFile( presenterClaudia, null, Context.PresenterClaudiaFileName );

			Context.LogSubtestEnd();

			#endregion Identity "Claudia"

			#region Identity "Bob"

			Context.LogSubtest( "Identity \"Bob\"" );

			// Create identity.
			AsymmAlg aeaBob = AlgorithmPool.GenerateAsymmAlg( AlgorithmSecurityLevel.Normal, Context.LongOperationProgress );
			AsymmAlg asaBob = AlgorithmPool.GenerateAsymmAlg( AlgorithmSecurityLevel.Normal, Context.LongOperationProgress );
			PrivateIdentity piBob = new PrivateIdentity( upnAlg, processingPort, aeaBob, asaBob );
			piBob.GroupUpnFingerprint = groupUpnFingerprint;

			// Set the appendix and physical identity.
			Appendix paBob = new Appendix( "Bob Bonkers" );
			PersonVirtualIdentity virtualIdBob = new PersonVirtualIdentity();
			paBob.VirtualIdentity = virtualIdBob;

			Vault vaultBob = new Vault( CryptoUtils.GetLocator(), 0, piBob, paBob );

			// Save the vault.
			SymmEncryptedEnvelope seeIdBob = AlgorithmPool.GetSymmEncryptedEnvelope();
			seeIdBob.EncryptableDocument = vaultBob;
			seeIdBob.SaveToFile( Context.VaultBobFileName, null, AppTestContext.DocumentPassword, serializer );

			// Create the solitary identity publisher.
			SolitaryIdentityPublisher identityPublisherBob = CoreDocumentsUtils.GetSolitaryIdentity(
				vaultBob.ProcessingIdentity );

			// Create and self-sign the presenter.
			Presenter presenterBob = new Presenter( CryptoUtils.GetLocator(), AlgorithmPool.GetMacAlg()
				, AlgorithmPool.GetAsymmSigningPadding(), AlgorithmPool.GetHashSerializer( false ), identityPublisherBob, paBob
				, null );
			presenterBob.Sign( vaultBob.ProcessingIdentity, null );

			// Save the presenter.
			serializer.SaveToFile( presenterBob, null, Context.PresenterBobFileName );

			Context.LogSubtestEnd();

			#endregion Identity "Bob"

			#region Identity "Alice"

			Context.LogSubtest( "Identity \"Alice\"" );

			// Create identity.
			AsymmAlg aeaAlice = AlgorithmPool.GenerateAsymmAlg( AlgorithmSecurityLevel.Normal, Context.LongOperationProgress );
			AsymmAlg asaAlice = AlgorithmPool.GenerateAsymmAlg( AlgorithmSecurityLevel.Normal, Context.LongOperationProgress );
			PrivateIdentity piAlice = new PrivateIdentity( upnAlg, processingPort, aeaAlice, asaAlice );
			piAlice.GroupUpnFingerprint = groupUpnFingerprint;

			// Set the appendix and physical identity.
			Appendix paAlice = new Appendix( "Alice Akira" );
			PersonVirtualIdentity virtualIdAlice = new PersonVirtualIdentity();
			paAlice.VirtualIdentity = virtualIdAlice;

			Vault vaultAlice = new Vault( CryptoUtils.GetLocator(), 0, piAlice, paAlice );

			// Save the vault.
			SymmEncryptedEnvelope seeIdAlice = AlgorithmPool.GetSymmEncryptedEnvelope();
			seeIdAlice.EncryptableDocument = vaultAlice;
			seeIdAlice.SaveToFile( Context.VaultAliceFileName, null, AppTestContext.DocumentPassword, serializer );

			// Create the solitary identity publisher.
			SolitaryIdentityPublisher identityPublisherAlice = CoreDocumentsUtils.GetSolitaryIdentity(
				vaultAlice.ProcessingIdentity );

			// Create and self-sign the presenter.
			Presenter presenterAlice = new Presenter( CryptoUtils.GetLocator(), AlgorithmPool.GetMacAlg()
				, AlgorithmPool.GetAsymmSigningPadding(), AlgorithmPool.GetHashSerializer( false ), identityPublisherAlice
				, paAlice, null );
			presenterAlice.Sign( vaultAlice.ProcessingIdentity, null );

			// Sign with the Daniel's identity.
			presenterAlice.Sign( vaultDaniel.ProcessingIdentity, null );

			// Save the presenter.
			serializer.SaveToFile( presenterAlice, null, Context.PresenterAliceFileName );

			Context.LogSubtestEnd();

			#endregion Identity "Alice"

			#region Group identity publishers

			Context.LogSubtest( "Group identity publishers" );

			#region Group of all

			// Create the list of members of the group identity publisher.
			SolitaryIdentityPublishers solitaryIdentityPublishers = new SolitaryIdentityPublishers( identityPublisherAlice
				, identityPublisherBob, identityPublisherClaudia, identityPublisherDaniel );

			// Create the group identity publisher.
			GroupIdentityPublisher groupIdentityPublisher = new GroupIdentityPublisher(
				AlgorithmPool.GetUpnAlg( AlgorithmSecurityLevel.Normal ), solitaryIdentityPublishers );

			// Set the information of the group identity publisher.
			groupIdentityPublisher.EncryptToGroup = true;
			groupIdentityPublisher.MissingGroupDecryptorCount = 1;
			groupIdentityPublisher.MissingGroupSignerCount = 1;

			// Save the group identity publisher.
			serializer.SaveToFile( groupIdentityPublisher, null, Context.GroupIdentityPublisherFileName );

			#endregion Group of all

			Context.LogSubtestEnd();

			#endregion Group identity publishers
		}

		#endregion Private methods
	}
}
