////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.Net;
using Vela.ActionProcessor;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Kernel;
using Vela.Networking;
using Vela.PaymentDocuments;
using Vela.PeerDocuments;
using Vela.Serialization;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	/// <remarks>
	/// <para>This test simulates the action processing performed by a payment service.</para>
	/// 
	/// <para>This test consists of the main operations which are executed by a service: signing, signature verification,
	/// encryption, decryption, network communication.</para>
	/// </remarks>
	public class PaymentServiceEnduranceTest : AppTest
	{
		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Payment service endurance (infinite cycle) *"; }
		}

		/// <summary>
		/// This is true, meaning that the test is long.
		/// </summary>
		public override bool IsLong
		{
			get { return true; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "9000 0007 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Makes the specified action request.
		/// </summary>
		/// <param name="processorConfiguration">The action processor configuration.</param>
		/// <param name="serviceListenerConfiguration">The action service listener configuration.</param>
		/// <param name="clientProcessingIdentity">The processing identity of the initiator of the action request.</param>
		/// <param name="actionRequest">Action request to make.</param>
		/// <param name="deserializer">Deserializer.</param>
		/// <param name="remoteIp">The IP address of the service where the action request must be sent.</param>
		/// <param name="remoteProcessingIdentity">The processing identity of the service who has to process the action
		/// request.</param>
		/// <param name="remoteUpn">The UPN of the service.</param>
		private void MakeActionRequest( ActionProcessorConfiguration processorConfiguration
			, ActionServiceListenerConfiguration serviceListenerConfiguration, PrivateIdentity clientProcessingIdentity
			, ActionRequest actionRequest, IDeserializer deserializer, IPAddress remoteIp
			, SolitaryIdentityPublisher remoteProcessingIdentity, Upn remoteUpn )
		{
			#region Encrypt the request

			#region Prepare the document to encrypt

			// Sign the data which is to be encrypted.
			actionRequest.Sign( clientProcessingIdentity, null );

			#endregion Prepare the document to encrypt

			AsymmEncryptedEnvelope ase = AlgorithmPool.GetAsymmEncryptedEnvelope();
			ase.EncryptableDocument = actionRequest;

			// Encrypt the action request.
			byte[] encryptedRequest = ase.SaveToBuffer( null, remoteProcessingIdentity, deserializer.GetSerializer() );

			#endregion Encrypt the request

			#region Make the service request

			// Obfuscate the action request.
			byte[] obfuscatedRequest = remoteUpn.UpnAlg.SteganographicAlg.Obfuscate( encryptedRequest
				, remoteUpn.SymmEncryptionKey, remoteUpn.Fingerprint.Value );

			ICommunicationFrameAlg communicationFrameAlg = AlgorithmPool.GetCommunicationFrameAlg();
			CommunicationFrame communicationFrame = new CommunicationFrame( communicationFrameAlg
				, processorConfiguration.CommunicationTimeoutPerBlob, processorConfiguration.CommunicationFrameSizeMax );

			byte[] obfuscatedResponse = communicationFrame.SendRequestAndReceiveResponse( remoteIp.GetAddressBytes()
				, serviceListenerConfiguration.ProcessingPort, obfuscatedRequest );

			#endregion Make the service request

			// Reveal the action response.
			byte[] revealedResponse = remoteUpn.UpnAlg.SteganographicAlg.Reveal( obfuscatedResponse
				, actionRequest.ResponseEncryptionKey, remoteUpn.Fingerprint.Value );

			// Decrypt the action response.
			SymmEncryptedEnvelope see = SymmEncryptedEnvelope.LoadFromBuffer( revealedResponse
				, actionRequest.ResponseEncryptionKey, typeof( ActionResponse ), out deserializer );
			ValidationUtils.ObjectHasContent( see, "see" );

			// Verify signature.
			ActionResponse actionResponse = (ActionResponse)see.EncryptableDocument;
			actionResponse.VerifySignature( remoteProcessingIdentity );

			ErrorActionResponse errorActionResponse = actionResponse as ErrorActionResponse;
			if( errorActionResponse != null )
			{
				string message = string.Format( "Failed to process action request: {0}", errorActionResponse.Message );

				if( !errorActionResponse.HasException( typeof( TimeoutException ) ) )
					throw new FatalCrossServiceException( message );

				Context.Log( message, true );
			}
		}

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			IDeserializer deserializer;

			#region Initialize

			// Load the service vault.
			SymmEncryptedEnvelope seeServiceVault = SymmEncryptedEnvelope.LoadFromFile( Context.PaymentVaultFileName
				, AppTestContext.DocumentPassword, typeof( PaymentVault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeServiceVault, "seeServiceVault" );
			PaymentVault serviceVault = (PaymentVault)seeServiceVault.EncryptableDocument;

			// Load the service presenter.
			Presenter servicePresenter = SerializationUtils.LoadFromFile( Context.ServicePresenterFileName
				, typeof( Presenter ), out deserializer ) as Presenter;
			ValidationUtils.ObjectHasContent( servicePresenter, "servicePresenter" );

			// Load the vault which creates the document.
			SymmEncryptedEnvelope seeIdAliceAccount = SymmEncryptedEnvelope.LoadFromFile( Context.VaultAliceFileName
				, AppTestContext.DocumentPassword, typeof( Vault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeIdAliceAccount, "seeIdAliceAccount" );
			Vault vaultAlice = (Vault)seeIdAliceAccount.EncryptableDocument;

			// Load the presenter of the beneficiary of the action request.
			Presenter presenterAlice = SerializationUtils.LoadFromFile( Context.PresenterAliceFileName
				, typeof( Presenter ), out deserializer ) as Presenter;
			ValidationUtils.ObjectHasContent( presenterAlice, "presenterAlice" );

			// Load the presenter of the beneficiary of the action request.
			Presenter presenterBob = SerializationUtils.LoadFromFile( Context.PresenterBobFileName
				, typeof( Presenter ), out deserializer ) as Presenter;
			ValidationUtils.ObjectHasContent( presenterBob, "presenterBob" );

			// Load the account of the beneficiary of the action request.
			SymmEncryptedEnvelope seeBobAccount = SymmEncryptedEnvelope.LoadFromFile( Context.AccountBobFileName
				, AppTestContext.DocumentPassword, typeof( Account ), out deserializer );
			ValidationUtils.ObjectHasContent( seeBobAccount, "seeBobAccount" );
			Account accountBob = (Account)seeBobAccount.EncryptableDocument;

			string serviceUrl = servicePresenter.Appendix.ActionProcessingUrl;
			IPAddress[] remoteIpAddresses = NetworkingUtils.ResolveUrl( serviceUrl );
			if( ListUtils.NoContent( remoteIpAddresses ) )
				throw new ValidationException( string.Format( "Failed to resolve the service URL \"{0}\".", serviceUrl ) );

			#endregion Initialize

			int i = 0;

			while( true )
			{
				#region Prepare the action request

				ActionRequest actionRequest;

				// Create the document to sign, which is also to be encrypted.
				if( i % 10 != 0 )
				{
					MakePaymentActionRequest makePaymentActionRequest = new MakePaymentActionRequest( CryptoUtils.GetLocator()
						, AlgorithmPool.GetMacAlg(), AlgorithmPool.GetAsymmSigningPadding()
						, AlgorithmPool.GetHashSerializer( false ), vaultAlice.LocalUpnFingerprint, servicePresenter.CreatorUpn
						, PeerDocumentsConsts.ActionRequestValidity, 0, presenterBob.CreatorUpn );

					// Set the document's properties.
					makePaymentActionRequest.CurrencyName = "GAU";
					makePaymentActionRequest.CurrencyAmount = 1;

					actionRequest = makePaymentActionRequest;
				}
				else
				{
					// From time to time, perform a different action request.
					actionRequest = new GetFileAnonymousActionRequest( CryptoUtils.GetLocator(), AlgorithmPool.GetMacAlg()
						, AlgorithmPool.GetAsymmSigningPadding(), AlgorithmPool.GetHashSerializer( false )
						, vaultAlice.LocalUpnFingerprint, servicePresenter.CreatorUpn, PeerDocumentsConsts.ActionRequestValidity
						, 0, "index.html" );
				}

				// Sign the data which is to be encrypted.
				actionRequest.Sign( vaultAlice.ProcessingIdentity, null );

				#endregion Prepare the action request

				// Make the action request.
				MakeActionRequest( Configuration.ActionProcessor, Configuration.ActionServiceListenerTemplate
					, vaultAlice.ProcessingIdentity, actionRequest, deserializer, remoteIpAddresses[ 0 ]
					, (SolitaryIdentityPublisher)servicePresenter.ProcessingIdentity, servicePresenter.CreatorUpn );

				i++;

				double transactionTimespan = Context.LongOperationProgress.ElapsedTime / i;
				string message = string.Format( "Transaction {0} ({1:0.00} seconds / transaction)", i, transactionTimespan );
				Context.LongOperationProgress.Progress( 0, message );
			}
		}

		#endregion Private methods
	}
}
