////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class SerializationSpeedTest : AppTest
	{
		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Serialization speed"; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "0014 0000 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		private void SerializerSpeed()
		{
			IDeserializer deserializer;

			#region Data

			// Load the document which is to be serialized.
			Presenter presenter = SerializationUtils.LoadFromFile( Context.PresenterAliceFileName, typeof( Presenter )
				, out deserializer ) as Presenter;
			ValidationUtils.ObjectHasContent( presenter, "presenter" );

			#endregion Data

			ISerializer serializer = AlgorithmPool.GetSerializer();
			byte[] serializedData = null;

			const int iter = 1000;

			Context.LogSubtest( "Serializer" );

			for( int i = 0; i < iter; i++ )
				serializedData = serializer.SaveToBuffer( presenter, null );

			Context.LogSpeed( SpeedModes.Msec_Oper, iter );
			Context.LogSubtestSeparator();

			// Save the serialized data.
			serializer = AlgorithmPool.GetSerializerForExport();
			serializer.SaveToFile( presenter, null, Context.SerializerFileName );
		}

		private void HashSerializerSpeed()
		{
			IDeserializer deserializer;

			#region Data

			// Load the document which is to be serialized.
			Presenter presenter = SerializationUtils.LoadFromFile( Context.PresenterAliceFileName, typeof( Presenter )
				, out deserializer ) as Presenter;
			ValidationUtils.ObjectHasContent( presenter, "presenter" );

			#endregion Data

			IHashSerializer hashSerializer = AlgorithmPool.GetHashSerializer( false );
			byte[] serializedData = null;

			const int iter = 1000;

			Context.LogSubtest( "Hash serializer" );

			for( int i = 0; i < iter; i++ )
				serializedData = hashSerializer.SaveToBuffer( presenter, null );

			Context.LogSpeed( SpeedModes.Msec_Oper, iter );
			Context.LogSubtestSeparator();

			// Save the serialized data.
			hashSerializer = AlgorithmPool.GetSerializerForExport().GetHashSerializer( false );
			hashSerializer.SaveToFile( presenter, null, Context.HashSerializerFileName );
		}

		private void DeserializerSpeed()
		{
			byte[] buffer = FileUtils.Read( Context.SerializerFileName, true );

			const int iter = 1000;

			IDeserializer deserializer = AlgorithmPool.GetSerializerForExport().GetDeserializer();

			Context.LogSubtest( "Deserializer" );

			for( int i = 0; i < iter; i++ )
				deserializer.LoadFromBuffer( buffer, typeof( Presenter ) );

			Context.LogSpeed( SpeedModes.Msec_Oper, iter );
		}

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			SerializerSpeed();
			HashSerializerSpeed();
			DeserializerSpeed();
		}

		#endregion Private methods
	}
}
