////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Crypto.Hash;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class SignatureEnduranceTest : AppTest
	{
		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Signature endurance (infinite cycle) *"; }
		}

		/// <summary>
		/// This is true, meaning that the test is long.
		/// </summary>
		public override bool IsLong
		{
			get { return true; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "9000 0006 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			#region Load the identity

			IDeserializer deserializer;

			SymmEncryptedEnvelope seeIdClaudia = SymmEncryptedEnvelope.LoadFromFile( Context.VaultClaudiaFileName
				, AppTestContext.DocumentPassword, typeof( Vault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeIdClaudia, "seeIdClaudia" );
			Vault vaultClaudia = (Vault)seeIdClaudia.EncryptableDocument;

			AsymmAlg asymmAlg = vaultClaudia.ProcessingIdentity.SigningAca;

			#endregion Load the identity

			HashAlg hashAlg = AlgorithmPool.GetHashAlg();
			IAsymmSigningPaddingAlg asymmSigningPaddingAlg = AlgorithmPool.GetAsymmSigningPadding();

			int i = 1;

			while( true )
			{
				Context.LongOperationProgress.Progress( 0, string.Format( "Signature {0}", i ) );

				byte[] buffer = EntropyPool.GetBytes( 32 );

				byte[] hashValue = hashAlg.Compute( buffer );

				// Sign data.
				byte[] signature = asymmAlg.SignData( hashValue, asymmSigningPaddingAlg );

				// Verify signature.
				try
				{
					asymmAlg.VerifyDataSignature( hashValue, signature, asymmSigningPaddingAlg );
				}
				catch( Exception ex )
				{
					if( ExceptionUtils.IsCritical( ex ) ) throw ExceptionUtils.PreserveStackTrace( ex );

					LogManager.Add( TextUtils.ArrayToText16( buffer ), true );
					LogManager.Add( TextUtils.ArrayToText16( signature ), true );

					throw ExceptionUtils.PreserveStackTrace( ex );
				}

				i++;
			}
		}

		#endregion Private methods
	}
}
