////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.DatabaseStorehouse;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class StorehouseSeekSpeedTest : AppTest
	{
		#region Private fields

		private IStorehouseProvider StorehouseProvider;

		#endregion Private fields

		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Storehouse seek speed *"; }
		}

		/// <summary>
		/// This is true, meaning that the test is long.
		/// </summary>
		public override bool IsLong
		{
			get { return true; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "9000 0004 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		private void CreateStorehouseProvider()
		{
			ISerializer serializer = AlgorithmPool.GetSerializer();

			// Create the storehouse provider.
			StorehouseProvider = StorehouseUtils.OpenStorehouse(
				Context.Configuration.StorehouseAccess.StorehouseProviderTypeName
				, Context.Configuration.StorehouseAccess as DatabaseAccess, Context.SpeedStorehouseFileName
				, AppTestContext.StorehousePassword, serializer ) as IStorehouseProvider;
		}

		private void StorehouseRandomSpeckleSeekSpeed()
		{
			// Create a transaction.
			IStorehouseTransaction transaction = StorehouseProvider.CreateTransaction();

			Context.LogStart( "Storehouse speed: load inexistent speckle" );

			int iter = 0;

			try
			{
				Chronometer chrono = new Chronometer();
				chrono.Start();
				while( !chrono.PassedSeconds( 1 ) )
				{
					Speckle speckle = transaction.LoadSpeckle( CryptoUtils.GetLocator(), StorehouseRealms.Posts
						, SheafLocks.Shared, null, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Letter ) ) );

					if( speckle != null ) LogManager.DebugAddFormat( "Found speckle \"{0}\".", speckle.StorehouseKey );

					iter++;
				}

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			Context.LogSpeed( SpeedModes.Oper_Sec, iter );
		}

		private void SpeckleFilterSpeed()
		{
			// Create a transaction.
			IStorehouseTransaction transaction = StorehouseProvider.CreateTransaction();

			Context.LogStart( "Storehouse speed: collect inexistent filtered speckles" );

			int iter = 0;

			try
			{
				Chronometer chrono = new Chronometer();
				chrono.Start();
				while( !chrono.PassedSeconds( 1 ) )
				{
					SpeckleFilter seekFilter = new SpeckleFilter( Post.InitiatorUpnFingerprintFilterIdentifier
						, CryptoUtils.GetLocator() );

					IKernelObservableList speckles = transaction.CollectSpeckles(
						PortableTypes.FindAllDerived( typeof( Letter ) ), new SpeckleFilters( seekFilter )
						, StorehouseRealms.Posts, null, null, null, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Letter ) ) );

					if( ListUtils.HasContent( speckles ) ) LogManager.DebugAddFormat( "Found {0} speckles.", speckles.Count );

					iter++;
				}

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			Context.LogSpeed( SpeedModes.Oper_Sec, iter );
		}

		private void StorehouseCollectSpecklesSpeed()
		{
			// Create a transaction.
			IStorehouseTransaction transaction = StorehouseProvider.CreateTransaction();

			Context.LogStart( "Storehouse speed: collect all speckles" );

			int iter = 0;

			try
			{
				Chronometer chrono = new Chronometer();
				chrono.Start();
				while( !chrono.PassedSeconds( 1 ) )
				{
					IKernelObservableList speckles = transaction.CollectSpeckles(
						PortableTypes.FindAllDerived( typeof( Letter ) ), null, StorehouseRealms.Posts, null, null, null
						, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Letter ) ) );

					if( ListUtils.HasContent( speckles ) )
					{
						LogManager.DebugAddFormat( "Found {0} speckles.", speckles.Count );

						iter += speckles.Count;
					}
				}

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			Context.LogSpeed( SpeedModes.Oper_Sec, iter );
		}

		private void StorehouseCollectKeysSpeed()
		{
			// Create a transaction.
			IStorehouseTransaction transaction = StorehouseProvider.CreateTransaction();

			Context.LogStart( "Storehouse speed: collect all keys" );

			int iter = 0;

			try
			{
				Chronometer chrono = new Chronometer();
				chrono.Start();
				while( !chrono.PassedSeconds( 1 ) )
				{
					IKernelObservableList keys = transaction.CollectKeys( PortableTypes.FindAllDerived( typeof( Letter ) )
						, null, StorehouseRealms.Posts, null, null, null );

					if( ListUtils.HasContent( keys ) )
					{
						LogManager.DebugAddFormat( "Found {0} storehouse keys.", keys.Count );

						iter += keys.Count;
					}
				}

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			Context.LogSpeed( SpeedModes.Oper_Sec, iter );
		}

		private void StorehouseCollectSheafsSpeed()
		{
			// Create a transaction.
			IStorehouseTransaction transaction = StorehouseProvider.CreateTransaction();

			Context.LogStart( string.Format( "Storehouse speed: collect {0} sheafs"
				, AppTestContext.StorehouseSpeedCollectSheafsCount ) );

			CollectPaging collectPaging = new CollectPaging( true, 0, AppTestContext.StorehouseSpeedCollectSheafsCount );
			int iter = 0;

			try
			{
				Chronometer chrono = new Chronometer();
				chrono.Start();
				while( !chrono.PassedSeconds( 1 ) )
				{
					IDeserializer deserializer;
					IKernelObservableList sheafs = transaction.CollectSheafs( PortableTypes.FindAllDerived( typeof( Letter ) )
						, null, StorehouseRealms.Posts, collectPaging, SheafLocks.Shared, null, out deserializer );

					if( ListUtils.HasContent( sheafs ) )
					{
						LogManager.DebugAddFormat( "Found {0} sheafs.", sheafs.Count );

						iter += sheafs.Count;
					}
				}

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			Context.LogSpeed( SpeedModes.Oper_Sec, iter );
		}

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			CreateStorehouseProvider();

			StorehouseRandomSpeckleSeekSpeed();
			SpeckleFilterSpeed();
			StorehouseCollectSpecklesSpeed();
			StorehouseCollectKeysSpeed();
			StorehouseCollectSheafsSpeed();
		}

		#endregion Private methods
	}
}
