////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.ActionProcessor;
using Vela.ApplicationBackend;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.DatabaseStorehouse;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.Storehouse;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class StorehouseSpeedTest : AppTest
	{
		#region Private fields

		private IStorehouseProvider StorehouseProvider;

		/// <summary>
		/// Locators to find in the storehouse, after all the documents are added.
		/// </summary>
		private Locators SeekLocators;

		/// <summary>
		/// Locators used for the speckle filters to find in the storehouse, after all the documents are added.
		/// </summary>
		private Locators InitiatorUpnFingerprints;

		#endregion Private fields

		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Storehouse speed *"; }
		}

		/// <summary>
		/// This is true, meaning that the test is long.
		/// </summary>
		public override bool IsLong
		{
			get { return true; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "9000 0003 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		private void CreateStorehouse()
		{
			SeekLocators = new Locators();
			InitiatorUpnFingerprints = new Locators();

			StorehouseHeader header = AlgorithmPool.GetStorehouseHeader();

			StorehouseEntropy entropy = new StorehouseEntropy( AlgorithmPool.GetEntropyAlg() );

			ActionServiceListenerConfigurations serviceListenerConfigurations = new ActionServiceListenerConfigurations();
			serviceListenerConfigurations.Add( Configuration.ActionServiceListenerTemplate );
			ServiceListenersStorehouseConfiguration configuration = new ServiceListenersStorehouseConfiguration(
				serviceListenerConfigurations );

			ISerializer serializer = AlgorithmPool.GetSerializer();

			StorehouseProvider = StorehouseUtils.CreateStorehouseProvider(
				Context.Configuration.StorehouseAccess.StorehouseProviderTypeName
				, Context.Configuration.StorehouseAccess as DatabaseAccess, Context.SpeedStorehouseFileName
				, AppTestContext.StorehousePassword, header, serializer ) as IStorehouseProvider;

			StorehouseUtils.CreateStorehouse( StorehouseProvider, Configuration.StorehouseAdminConnectionString, entropy
				, configuration );
		}

		private void StorehouseAddSpeed()
		{
			int docCount = AppTestContext.StorehouseSpeedAddBatchSize * AppTestContext.StorehouseSpeedAddBatchCount;

			// The instance locator of one document out of every "seekOnceEvery" is stored in order to seek it after all the
			// documents are added to the storehouse.
			int seekOnceEvery = docCount / AppTestContext.StorehouseSpeedSeekBatchSize;
			if( seekOnceEvery < 1 ) seekOnceEvery = 1;

			string letterMessage = EntropyPool.GetText16( 1000 );

			Context.LogStart( "Storehouse add sheaf speed" );

			// Go through all the commit batches.
			int docIndex = 0;
			for( int b = 1; b <= AppTestContext.StorehouseSpeedAddBatchCount; b++ )
			{
				// Create a transaction.
				IStorehouseTransaction batchTransaction = StorehouseProvider.CreateTransaction();

				try
				{
					// Add the documents from the current batch.
					for( int i = 0; i < AppTestContext.StorehouseSpeedAddBatchSize; i++ )
					{
						Context.LongOperationProgress.PersistentMessage = string.Format( "Documents added {0} / {1} in batch"
								+ " {2} / {3}", i, AppTestContext.StorehouseSpeedAddBatchSize, b
							, AppTestContext.StorehouseSpeedAddBatchCount );
						Context.LongOperationProgress.Progress();

						Letter letter = new Letter( CryptoUtils.GetLocator(), AlgorithmPool.GetMacAlg()
							, AlgorithmPool.GetAsymmSigningPadding(), AlgorithmPool.GetHashSerializer( false )
							, CryptoUtils.GetLocator(), letterMessage, null, false, CryptoUtils.GetLocator()
							, new AttachedFiles(), false, false );

						// Add the document.
						batchTransaction.AddSheaf( letter, StorehouseRealms.Posts );

						// Add a locator to be seeked.
						if( docIndex % seekOnceEvery == 0 )
						{
							SeekLocators.Add( letter.InstanceLocator );
							InitiatorUpnFingerprints.Add( letter.InitiatorUpnFingerprint );
						}

						docIndex++;
					}

					double progress = (double)b / AppTestContext.StorehouseSpeedAddBatchCount;

					Context.LongOperationProgress.PersistentMessage = string.Format( "Committing {0} documents in batch"
							+ " {1} / {2}", AppTestContext.StorehouseSpeedAddBatchSize, b
						, AppTestContext.StorehouseSpeedAddBatchCount );
					Context.LongOperationProgress.Progress( progress, null, true );

					// Commit the transaction.
					batchTransaction.Commit();
				}
				catch( Exception ex )
				{
					// Rolls back the transaction.
					batchTransaction.Rollback();

					throw ExceptionUtils.PreserveStackTrace( ex );
				}
			}

			Context.LogSpeed( SpeedModes.Oper_Sec, docCount );
		}

		private void SpeckleLoadSpeed()
		{
			// Create a transaction.
			IStorehouseTransaction transaction = StorehouseProvider.CreateTransaction();

			Context.LogStart( "Storehouse speed: load existing speckle" );

			int iter = 0;

			try
			{
				Chronometer chrono = new Chronometer();
				chrono.Start();
				while( !chrono.PassedSeconds( 1 ) )
				{
					foreach( Locator seekLocator in SeekLocators )
					{
						Speckle speckle = transaction.LoadSpeckle( seekLocator, StorehouseRealms.Posts, SheafLocks.Shared, null
							, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Letter ) ) );

						LogManager.DebugAddFormat( "Found speckle \"{0}\".", speckle.StorehouseKey );

						iter++;
					}
				}

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			Context.LogSpeed( SpeedModes.Oper_Sec, iter );
		}

		private void SheafLoadSpeed()
		{
			// Create a transaction.
			IStorehouseTransaction transaction = StorehouseProvider.CreateTransaction();

			Context.LogStart( "Storehouse speed: load existing sheaf" );

			int iter = 0;

			try
			{
				Chronometer chrono = new Chronometer();
				chrono.Start();
				while( !chrono.PassedSeconds( 1 ) )
				{
					foreach( Locator seekLocator in SeekLocators )
					{
						IDeserializer deserializer;
						ISheaf sheaf = transaction.LoadSheaf( seekLocator, StorehouseRealms.Posts, typeof( Letter )
							, SheafLocks.Shared, out deserializer );

						LogManager.DebugAddFormat( "Found sheaf with the locator \"{0}\".", sheaf.InstanceLocator );

						iter++;
					}
				}

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			Context.LogSpeed( SpeedModes.Oper_Sec, iter );
		}

		private void StorehouseSeekSpeed()
		{
			// Create a transaction.
			IStorehouseTransaction transaction = StorehouseProvider.CreateTransaction();

			Context.LogStart( "Storehouse speed: collect existing filtered speckles" );
			LogManager.DebugAddFormat( "Seeking {0} speckle initiator locators...", InitiatorUpnFingerprints.Count );

			int iter = 0;

			try
			{
				Chronometer chrono = new Chronometer();
				chrono.Start();
				while( !chrono.PassedSeconds( 1 ) )
				{
					foreach( Locator initiatorUpnFingerprint in InitiatorUpnFingerprints )
					{
						SpeckleFilter seekFilter = new SpeckleFilter( Letter.InitiatorUpnFingerprintFilterIdentifier
							, initiatorUpnFingerprint );

						// The speckle scan operation is run for the entire database, for each iteration, so this test gets
						// geometrically slower with the number of stored sheafs.
						IKernelObservableList speckles = transaction.CollectSpeckles(
							PortableTypes.FindAllDerived( typeof( Letter ) ), new SpeckleFilters( seekFilter )
							, StorehouseRealms.Posts, null, null, null
							, Sheaf.CollectSpeckleFilterIdentifiers( typeof( Letter ) ) );

						if( ListUtils.HasContent( speckles ) )
						{
							LogManager.DebugAddFormat( "Found {0} speckles for the initiator UPN fingerprint \"{1}\"."
								, speckles.Count, initiatorUpnFingerprint );
						}

						iter++;
					}
				}

				// Commit the transaction.
				transaction.Commit();
			}
			catch( Exception ex )
			{
				// Rolls back the transaction.
				transaction.Rollback();

				throw ExceptionUtils.PreserveStackTrace( ex );
			}

			Context.LogSpeed( SpeedModes.Oper_Sec, iter );
		}

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			CreateStorehouse();
			StorehouseAddSpeed();
			SpeckleLoadSpeed();
			SheafLoadSpeed();
			StorehouseSeekSpeed();
		}

		#endregion Private methods
	}
}
