////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Crypto.Asymmetric;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class VerifyAsymmetricTest : AppTest
	{
		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Verify asymmetric"; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "0006 0000 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		private void VerifyBigInteger()
		{
			IDeserializer deserializer;

			Context.LogStart( "Verify big integer" );

			// Data to process.
			byte[] buffer = EntropyPool.GetBytes( 32 );

			// Load the vault from the specified file, and verify its self-signature.
			SymmEncryptedEnvelope seeIdAlice = SymmEncryptedEnvelope.LoadFromFile( Context.VaultAliceFileName
				, AppTestContext.DocumentPassword, typeof( Vault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeIdAlice, "seeIdAlice" );
			Vault vaultAlice = (Vault)seeIdAlice.EncryptableDocument;

			// Encrypt the data.
			byte[] encrypted = vaultAlice.ProcessingIdentity.EncryptionAca.ProcessWithPublicKey( buffer, true );

			// Decrypt the data.
			byte[] decrypted = vaultAlice.ProcessingIdentity.EncryptionAca.ProcessWithPrivateKey( encrypted, false );

			// Because of the "Zero-byte issue" of the "BigInteger" library, pad the result.
			decrypted = BufferUtils.PadBeginningWithZeros( decrypted, buffer.Length );

			// Compare the original data with the decrypted data.
			bool isEqual = BufferUtils.Equals( buffer, decrypted );

			if( !isEqual ) throw new ValidationException( "Different big integers." );

			Context.Log( "Equal data" );

			Context.LogSubtestEnd();
		}

		private void VerifyAsymmPadding( string testName, IAsymmEncryptionPaddingAlg asymmEncryptionPaddingAlg
			, IAsymmSigningPaddingAlg asymmSigningPaddingAlg )
		{
			Context.LogStart( "Verify " + testName );

			// Data to process.
			byte[] buffer = EntropyPool.GetBytes( 32 );

			IDeserializer deserializer;

			// Load the vault from the specified file, and verify its self-signature.
			SymmEncryptedEnvelope seeIdAlice = SymmEncryptedEnvelope.LoadFromFile( Context.VaultAliceFileName
				, AppTestContext.DocumentPassword, typeof( Vault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeIdAlice, "seeIdAlice" );
			Vault vaultAlice = (Vault)seeIdAlice.EncryptableDocument;

			#region Encryption

			// Encrypt the data.
			byte[] encrypted = vaultAlice.ProcessingIdentity.EncryptionAca.EncryptData( buffer, asymmEncryptionPaddingAlg );

			// Decrypt the data.
			byte[] decrypted = vaultAlice.ProcessingIdentity.EncryptionAca.DecryptData( encrypted, asymmEncryptionPaddingAlg );

			// Because of the "Zero-byte issue" of the "BigInteger" library, pad the result.
			decrypted = BufferUtils.PadBeginningWithZeros( decrypted, buffer.Length );

			// Compare the original data with the decrypted data.
			if( !BufferUtils.Equals( buffer, decrypted ) ) throw new ValidationException( "Failed to asymmetrically decrypt." );

			#endregion Encryption

			#region Signing

			// Sign the data.
			byte[] signed = vaultAlice.ProcessingIdentity.SigningAca.SignData( buffer, asymmSigningPaddingAlg );

			// Verify the signature.
			vaultAlice.ProcessingIdentity.SigningAca.VerifyDataSignature( buffer, signed, asymmSigningPaddingAlg );

			#endregion Signing

			Context.LogSubtestEnd();
		}

		private void VerifyAsymmPadding()
		{
			VerifyAsymmPadding( "VSH", AlgorithmPool.GetAsymmEncryptionPadding(), AlgorithmPool.GetAsymmSigningPadding() );
		}

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			VerifyBigInteger();
			VerifyAsymmPadding();
		}

		#endregion Private methods
	}
}
