////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.Cloning;
using Vela.CoreDocuments;
using Vela.Crypto.Asymmetric;
using Vela.Kernel;
using Vela.Kernel.Serialization;
using Vela.PeerDocuments;
using Vela.Serialization;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class VerifyCloningTest : AppTest
	{
		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Verify cloning"; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "0013 0001 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		private void Clone( IDocument sourceObject )
		{
			IDocument cloneObject = Cloner.Deep( sourceObject );

			ISerializer serializer = AlgorithmPool.GetSerializer();
			if( !serializer.EqualDocuments( cloneObject, sourceObject ) )
				throw new ValidationException( "The cloned objects are different." );
		}

		private void CloneEmptyArray()
		{
			ArrayItem[ , ] array = new ArrayItem[ , ]
			{
			};

			ArrayItem[ , ] arrayClone = Cloner.Deep( array );

			if( arrayClone == array )
				throw new ValidationException( "The cloned objects are identical." );
		}

		private void CloneArray()
		{
			ArrayItem[ , ] array = new ArrayItem[ , ]
			{
				{ new ArrayItem( 0 ), new ArrayItem( 1 ) },
				{ new ArrayItem( 2 ), new ArrayItem( 3 ) },
				{ new ArrayItem( 4 ), new ArrayItem( 5 ) }
			};

			ArrayItem[ , ] arrayClone = Cloner.Deep( array );

			if( arrayClone == array )
				throw new ValidationException( "The cloned objects are identical." );
		}

		private void CloneVault()
		{
			IDeserializer deserializer;

			SymmEncryptedEnvelope seeIdDaniel = SymmEncryptedEnvelope.LoadFromFile( Context.VaultDanielFileName
				, AppTestContext.DocumentPassword, typeof( Vault ), out deserializer );
			ValidationUtils.ObjectHasContent( seeIdDaniel, "seeIdDaniel" );
			Vault vaultDaniel = (Vault)seeIdDaniel.EncryptableDocument;

			Clone( vaultDaniel );
		}

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			CloneEmptyArray();
			CloneArray();
			Clone( new Rsa2048() );
			Clone( AlgorithmPool.GetActionClientConfiguration() );
			CloneVault();
		}

		#endregion Private methods

		#region Classes

		private class Identifier
		{
			private static int StaticIndex;
			private string PrivateIndex;
			protected int ProtectedIndex;
			public int PublicIndex;

			public Identifier( int i )
			{
				StaticIndex = i;
				PrivateIndex = i.ToString();
				ProtectedIndex = i;
				PublicIndex = i;
			}
		}

		private class ArrayItemBase
		{
			private static Identifier StaticIdentifierBase;
			private Identifier PrivateIdentifierBase;
			protected Identifier ProtectedIdentifierBase;
			public Identifier PublicIdentifierBase;

			public ArrayItemBase( int i )
			{
				StaticIdentifierBase = new Identifier( i );
				PrivateIdentifierBase = new Identifier( i );
				ProtectedIdentifierBase = new Identifier( i );
				PublicIdentifierBase = new Identifier( i );
			}
		}

		private class ArrayItem : ArrayItemBase
		{
			private static Identifier StaticIdentifier;
			private Identifier PrivateIdentifier;
			protected Identifier ProtectedIdentifier;
			public Identifier PublicIdentifier;

			public ArrayItem( int i )
				: base( i )
			{
				StaticIdentifier = new Identifier( i );
				PrivateIdentifier = new Identifier( i );
				ProtectedIdentifier = new Identifier( i );
				PublicIdentifier = new Identifier( i );
			}
		}

		#endregion Classes
	}
}
