////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System.IO;
using Vela.Kernel;
using Vela.PeerDocuments;
using Vela.Serialization;
using Vela.StarkDocuments;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class VerifySecuringDocumentTest : AppTest
	{
		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Verify securing document"; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "0009 0000 01"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			IDeserializer deserializer;

			// Load the public identity of the issuer of the document.
			Presenter presenterAlice = SerializationUtils.LoadFromFile( Context.PresenterAliceFileName
				, typeof( Presenter ), out deserializer ) as Presenter;
			ValidationUtils.ObjectHasContent( presenterAlice, "presenterAlice" );

			// Load the securing document.
			SecuringDocument securingDocument = SerializationUtils.LoadFromFile( Context.SecuringDocumentFileName
				, typeof( SecuringDocument ), out deserializer ) as SecuringDocument;
			ValidationUtils.ObjectHasContent( securingDocument, "securingDocument" );

			// Check if the signature of the document is valid, using the identity publisher of its issuer.
			securingDocument.VerifySignature( presenterAlice.ProcessingIdentity );

			#region Verify the secured document

			using( FileStream securedDocumentFs = File.OpenRead( Context.SecuredDocumentFileName ) )
			using( FileStream documentToSecureFs = File.OpenRead( Context.DocumentToSecureFileName ) )
			using( FileStream unpackedDocumentFs = File.Open( Context.UnpackedDocumentFileName, FileMode.Create ) )
			{
				byte[] macValue = securingDocument.Process( securedDocumentFs, unpackedDocumentFs, false, null );

				if( !BufferUtils.Equals( macValue, securingDocument.AttestedMacValue ) )
					throw new ValidationException( "The MAC from the securing document is different from the computed one." );

				byte[] unpackedDocumentBuffer = new byte[ unpackedDocumentFs.Length ];
				unpackedDocumentFs.Position = 0;
				unpackedDocumentFs.Read( unpackedDocumentBuffer, 0, unpackedDocumentBuffer.Length );

				byte[] documentToSecureBuffer = new byte[ documentToSecureFs.Length ];
				documentToSecureFs.Read( documentToSecureBuffer, 0, documentToSecureBuffer.Length );

				if( !BufferUtils.Equals( unpackedDocumentBuffer, documentToSecureBuffer ) )
					throw new ValidationException( "The unpacked document is different from the document to secure." );
			}

			#endregion Verify the secured document
		}

		#endregion Private methods
	}
}
