////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using System;
using Vela.Crypto;
using Vela.Crypto.Hash;
using Vela.Crypto.Mac;
using Vela.Crypto.Symmetric;
using Vela.Kernel;

namespace Vela.TesterBackend
{
	/// <summary>
	/// Test.
	/// </summary>
	public class VerifySymmetricTest : AppTest
	{
		#region Public non-serializable properties

		/// <summary>
		/// The name of the test.
		/// </summary>
		public override string Name
		{
			get { return "Verify symmetric"; }
		}

		/// <summary>
		/// The prefix for the sort order in which to run the test.
		/// </summary>
		public override string SortPrefix
		{
			get { return "0004 0000 00"; }
		}

		#endregion Public non-serializable properties

		#region Private methods

		private void VerifyHash( HashAlg hashAlg, string testVector, string result )
		{
			byte[] buffer = AlgorithmPool.GetTextEncoding().TextToArray( testVector, false );
			byte[] hash = hashAlg.Compute( buffer );
			string textHash = TextUtils.ArrayToText16( hash );

			if( !TextUtils.Equals( textHash, result ) )
				throw new ValidationException( string.Format( "Hash \"{0}\" failed.", hashAlg.GetType().Name ) );
		}

		private void VerfiySha512WithDotNet( params int[] sizes )
		{
			foreach( int size in sizes )
			{
				HashAlg hashAlg = new Sha512();
				byte[] buffer = EntropyPool.GetBytes( size );

				byte[] hash = hashAlg.Compute( buffer );

				#region DotNet

				System.Security.Cryptography.SHA512Managed hashAlgDN = new System.Security.Cryptography.SHA512Managed();

				byte[] hashDN = hashAlgDN.ComputeHash( buffer );

				#endregion DotNet

				if( !BufferUtils.Equals( hash, hashDN ) ) throw new ValidationException( "Sha512 failed." );
			}
		}

		private void VerfiyBckSha512WithDotNet( params int[] sizes )
		{
			foreach( int size in sizes )
			{
				Bck macAlg = new Bck( new Sha512() );
				byte[] key = EntropyPool.GetBytes( macAlg.HashAlg.AlgSizeBytes - 5 );
				byte[] buffer = EntropyPool.GetBytes( size );

				byte[] mac = macAlg.Compute( key, null, true, buffer );

				#region DotNet

				System.Security.Cryptography.HMACSHA512 macAlgDN = new System.Security.Cryptography.HMACSHA512( key );

				byte[] dataSizeBuffer = BitConverter.GetBytes( (Int64)buffer.Length );

				macAlgDN.TransformBlock( dataSizeBuffer, 0, dataSizeBuffer.Length, null, 0 );
				macAlgDN.TransformFinalBlock( buffer, 0, buffer.Length );

				byte[] macDN = macAlgDN.Hash;

				#endregion DotNet

				if( !BufferUtils.Equals( mac, macDN ) ) throw new ValidationException( "BckSha512 failed." );
			}
		}

		private void VerfiyRijndael256WithDotNet( params int[] sizes )
		{
			foreach( int size in sizes )
			{
				Rijndael256 rijndael = new Rijndael256( SymmPaddings.Iso10126, SymmChainings.Cbc );

				byte[] salt = EntropyPool.GetBytes( rijndael.AlgSizeBytes );
				byte[] key = EntropyPool.GetBytes( rijndael.AlgSizeBytes );
				byte[] buffer = EntropyPool.GetBytes( size );

				byte[] encrypted = rijndael.Encrypt( buffer, salt, key );
				byte[] decrypted = rijndael.Decrypt( encrypted, salt, key );

				if( !BufferUtils.Equals( buffer, decrypted ) ) throw new ValidationException( "Rijndael256 failed." );

				#region DotNet

				System.Security.Cryptography.RijndaelManaged rijndaelDN = new System.Security.Cryptography.RijndaelManaged();
				rijndaelDN.BlockSize = 256;
				rijndaelDN.KeySize = 256;
				rijndaelDN.Key = key;
				rijndaelDN.IV = salt;
				rijndaelDN.Mode = System.Security.Cryptography.CipherMode.CBC;
				rijndaelDN.Padding = System.Security.Cryptography.PaddingMode.ISO10126;

				System.Security.Cryptography.ICryptoTransform decryptor = rijndaelDN.CreateDecryptor();

				decrypted = decryptor.TransformFinalBlock( encrypted, 0, encrypted.Length );

				#endregion DotNet

				if( !BufferUtils.Equals( buffer, decrypted ) ) throw new ValidationException( "Rijndael256 failed." );
			}
		}

		private void VerifySha512()
		{
			Context.LogSubtest( "Verifying Sha512." );

			#region Vectors samples

			HashAlg hashAlg = new Sha512();

			VerifyHash( hashAlg, "qwertyuiop", "6308d8f6a7ccc9f77e41be5331a52c71c0bb28ecbd4669b960d60dd505dfde9ddd7a30cd26bb308010b3819699daba7caeb791bf6a4153605fe56d1fd3d5df41" );

			VerifyHash( hashAlg, "a1b2c3d4e5f6g7h8i9j0", "b4b9f4f2a0ecd89a5369c5d5c1d019869ed1690c49018a3f67a3812bf8b78bc2c88888e8f58a410f996e3ed01d2f98c123cb2b9c61ccbc7fcc96a9d77acae6bc" );

			VerifyHash( hashAlg, "qazwsxedcrfvtgbyhnujmikolp", "0376a1a26f48545cf71a74c5080e8ea6b3beec9eebaded04ce95ceaa1789aa6dd2c57a03788f616e7e248924522af60c170761eadc1a44304b166751380d643a" );

			VerifyHash( hashAlg, "The quick brown fox jumps over the lazy dog", "07e547d9586f6a73f73fbac0435ed76951218fb7d0c8d788a309d785436bbb642e93a252a954f23912547d1e8a3b5ed6e1bfd7097821233fa0538f3db854fee6" );

			#endregion Vectors samples

			VerfiySha512WithDotNet( 15, 16, 17, 31, 32, 33, 63, 64, 65, 127, 128, 129, 255, 256, 257 );

			Context.LogSubtestEnd();
		}

		private void VerifyBckSha512()
		{
			Context.LogSubtest( "Verifying BckSha512." );

			VerfiyBckSha512WithDotNet( 15, 16, 17, 31, 32, 33, 63, 64, 65, 127, 128, 129, 255, 256, 257 );

			Context.LogSubtestEnd();
		}

		private void VerifyRijndael256()
		{
			Context.LogSubtest( "Verifying Rijndael256." );

			VerfiyRijndael256WithDotNet( 15, 16, 17, 31, 32, 33, 63, 64, 65, 127, 128, 129, 255, 256, 257 );

			Context.LogSubtestEnd();
		}

		/// <summary>
		/// Executes the test.
		/// </summary>
		protected override void Execute()
		{
			VerifySha512();

			VerifyBckSha512();

			VerifyRijndael256();
		}

		#endregion Private methods
	}
}
