////////////////////////////////////////////////////////////////////////////////
//
// Developed by the Vela Entity (vela.im).
// Released under a "Common Sense License".
//
// Original author: Gardener of Thoughts (gardenerofthoughts.org).
//
////////////////////////////////////////////////////////////////////////////////

using Vela.ActionProcessor;
using Vela.ActionService;
using Vela.CoreDocuments;
using Vela.Crypto;
using Vela.Kernel;
using Vela.PaymentDocuments;
using Vela.Storehouse;

namespace Vela.PaymentService
{
	/// <summary>
	/// Service for processing payment action requests (<see cref="ActionRequest"/>).
	/// </summary>
	/// <remarks>
	/// <para>A fee is charged from the amount of currency which is being paid, so it's like the beneficiary of the payment pays
	/// the fee.</para>
	/// 
	/// <para>The charged fee documents (<see cref="ChargedPaymentFee"/>) are not saved in a single document for the entire
	/// service in order to allow the service to be scalable up to the storehouse's performance level (which can process several
	/// action requests in parallel).</para>
	/// </remarks>
	public class PaymentService : StarkActionService
	{
		#region Constructors

		/// <summary>
		/// Initializes a new instance of the current class.
		/// </summary>
		protected PaymentService()
		{
		}

		#endregion Constructors

		#region Public methods

		/// <summary>
		/// Processes an action request to make a payment.
		/// </summary>
		/// <param name="actionRequest">Action request to process.</param>
		/// <param name="requestContext">Action request context.</param>
		/// <returns>Returns the action response which is to be sent back to the initiator of the action request.</returns>
		/// <remarks>
		/// <para>The beneficiary pays the fee from the received amount of currency.</para>
		/// </remarks>
		public MakePaymentActionResponse ProcessMakePayment( MakePaymentActionRequest actionRequest
			, ActionRequestWithBeneficiaryContext requestContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( requestContext, "requestContext" );

			#region Process the action request

			// Get the accounts.
			PaymentAccount initiatorAccount = (PaymentAccount)requestContext.InitiatorAccountContext.Account;
			PaymentAccount beneficiaryAccount = (PaymentAccount)requestContext.BeneficiaryAccountContext.Account;

			// Get the balance of the initiator.
			PaymentBalance initiatorBalance = initiatorAccount.Balances.Find( actionRequest.CurrencyName );
			if( initiatorBalance == null )
			{
				throw new FatalCrossServiceException( string.Format( "The initiator account with the locator \"{0}\" has no"
						+ " balance for the \"{1}\" currency to pay from."
					, requestContext.InitiatorAccountContext.Account.InstanceLocator, actionRequest.CurrencyName ) );
			}

			// Get the balance of the beneficiary.
			PaymentBalance beneficiaryBalance = beneficiaryAccount.Balances.Find( actionRequest.CurrencyName );
			if( beneficiaryBalance == null )
			{
				throw new FatalCrossServiceException( string.Format( "The beneficiary account with the locator \"{0}\" has no"
						+ " balance for the \"{1}\" currency in which to receive payment."
					, requestContext.BeneficiaryAccountContext.Account.InstanceLocator, actionRequest.CurrencyName ) );
			}

			PaymentVault serviceVault = requestContext.ServiceContext.Vault as PaymentVault;
			double fee = 0;

			if( initiatorBalance.CurrencyAmount >= actionRequest.CurrencyAmount )
			{
				// Make payment.

				// Amout of currency to pay.
				double currencyAmount = actionRequest.CurrencyAmount;

				#region Calculate the fee

				// Get the fee plans.
				PaymentAppendix serviceAppendix = (PaymentAppendix)serviceVault.Appendix;
				PaymentFeePlans feePlans = serviceAppendix.FeePlans;

				// Find the fee plan for the paid currency.
				PaymentFeePlan feePlan = serviceAppendix.FeePlans.Find( actionRequest.CurrencyName );

				// Calculate the fee.
				double percentage;
				fee = (feePlan != null) ? feePlan.Calculate( currencyAmount, out percentage ) : 0;

				#endregion Calculate the fee

				// Spend the currency from the initiator.
				initiatorBalance.CurrencyAmount -= currencyAmount;

				// Receive the currency for the beneficiary. Charge the fee. The beneficiary pays the fee from the received
				// amount of currency.
				beneficiaryBalance.CurrencyAmount += currencyAmount - fee;
			}
			else
			{
				// The initiator doesn't have enough currency to spend.

				throw new FatalCrossServiceException( string.Format( "The initiator account with the locator \"{0}\" doesn't"
						+ " have {1} units of currency \"{2}\" to spend."
					, requestContext.InitiatorAccountContext.Account.InstanceLocator, actionRequest.CurrencyAmount
					, actionRequest.CurrencyName ) );
			}

			#endregion Process the action request

			#region Handle the fee document

			// Create the fee document.
			ChargedPaymentFee chargedFee = new ChargedPaymentFee( CryptoUtils.GetLocator(), actionRequest.MacAlg
				, actionRequest.AsymmSigningPaddingAlg, requestContext.ServiceContext.Serializer.GetHashSerializer( false )
				, serviceVault.InstanceLocator, actionRequest.InstanceLocator, requestContext.ActionResponseLocator
				, actionRequest.InitiatorUpnFingerprint, actionRequest.BeneficiaryUpnFingerprint, fee );

			// Sign the fee document.
			if( serviceVault.SignChargedFee )
				chargedFee.Sign( serviceVault.ProcessingIdentity, null );

			// Save the fee.
			requestContext.ServiceContext.Transaction.AddSheaf( chargedFee, StorehouseRealms.ChargedFees );

			#endregion Handle the fee document

			#region Prepare the action response

			MakePaymentActionResponse actionResponse = new MakePaymentActionResponse( requestContext.ActionResponseLocator
				, actionRequest, null, initiatorAccount.Balances );

			#endregion Prepare the action response

			return actionResponse;
		}

		//? Test
		/// <summary>
		/// Processes an action request to get a file.
		/// </summary>
		/// <param name="actionRequest">Action request to process.</param>
		/// <param name="requestContext">Action request context.</param>
		/// <returns>Returns the action response which is to be sent back to the initiator of the action request.</returns>
		public GetFileAnonymousActionResponse ProcessGetFile( GetFileAnonymousActionRequest actionRequest
			, ActionRequestContext requestContext )
		{
			ValidationUtils.ObjectHasContent( actionRequest, "actionRequest" );
			ValidationUtils.ObjectHasContent( requestContext, "requestContext" );

			#region Prepare the action response

			TextEncoding textEncoding = new Utf8();
			string message = "File content";
			byte[] buffer = textEncoding.TextToArray( message, true );

			IIdentityPublisher serviceIdentityPublisher = CoreDocumentsUtils.GetSolitaryIdentity(
				requestContext.ServiceContext.Vault.ProcessingIdentity );

			GetFileAnonymousActionResponse actionResponse = new GetFileAnonymousActionResponse(
				requestContext.ActionResponseLocator, actionRequest, null, serviceIdentityPublisher, buffer );

			#endregion Prepare the action response

			return actionResponse;
		}

		#endregion Public methods
	}
}
